diff -pruN 1.5.1-1/bin/sequel 3.0.0-1/bin/sequel
--- 1.5.1-1/bin/sequel	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/bin/sequel	2009-03-23 17:57:23.000000000 +0000
@@ -0,0 +1,105 @@
+#!/usr/bin/env ruby
+
+require 'rubygems'
+require 'optparse'
+require 'sequel'
+
+db_opts = {}
+echo = nil
+env = nil
+logfile = nil
+migrate_dir = nil
+migrate_ver = nil
+load_dir = nil
+
+opts = OptionParser.new do |opts|
+  opts.banner = "Sequel: The Database Toolkit for Ruby"
+  opts.define_head "Usage: sequel <uri|path> [options]"
+  opts.separator ""
+  opts.separator "Examples:"
+  opts.separator "  sequel sqlite://blog.db"
+  opts.separator "  sequel postgres://localhost/my_blog"
+  opts.separator "  sequel config/database.yml"
+  opts.separator ""
+  opts.separator "For more information see http://sequel.rubyforge.org"
+  opts.separator ""
+  opts.separator "Options:"
+
+  opts.on_tail("-?", "--help", "Show this message") do
+    puts opts
+    exit
+  end
+
+  opts.on("-e", "--env ENV", "use environment config for database") do |v|
+    env = v
+  end
+  
+  opts.on("-E", "--echo", "echo SQL statements") do
+    echo = true
+  end
+  
+  opts.on("-l", "--log logfile", "log SQL statements to log file") do |v|
+    logfile = v
+  end
+  
+  opts.on("-L", "--load-dir DIR", "loads all *.rb from specifed directory") do |v|
+    load_dir = v
+  end
+  
+  opts.on("-m", "--migrate-directory DIR", "run the migrations in directory") do |v|
+    migrate_dir = v
+  end
+  
+  opts.on("-M", "--migrate-version VER", "migrate the database to version given") do |v|
+    migrate_ver = Integer(v)
+  end
+
+  opts.on_tail("-v", "--version", "Show version") do
+    puts "sequel #{Sequel.version}"
+    exit
+  end
+end
+opts.parse!
+
+db = ARGV.shift
+
+if db.nil? || db.empty?
+  puts opts
+  exit 1
+end
+
+if logfile || echo
+  require 'logger'
+  db_opts[:loggers] = []
+  db_opts[:loggers] << Logger.new(logfile) if logfile
+  db_opts[:loggers] << Logger.new($stdout) if echo
+end
+
+if File.exist?(db)
+  require 'yaml'
+  env ||= "development"
+  db_config = YAML.load_file(db)
+  db_config = db_config[env] || db_config[env.to_sym] || db_config
+  db_config.each{|(k,v)| db_config[k.to_sym] = db_config.delete(k)}
+  db_config.merge!(db_opts)
+end
+
+begin
+  DB = Sequel.connect(*(db_config ? [db_config] : [db, db_opts]))
+  DB.test_connection
+  if migrate_dir
+    require 'sequel/extensions/migration'
+    Sequel::Migrator.apply(DB, migrate_dir, migrate_ver)
+    exit
+  end
+rescue => e
+  puts "#{e.class}: #{e.message}"
+  puts e.backtrace.first
+  exit 1
+end
+
+Dir["#{load_dir}/**/*.rb"].each{|f| load(f)} if load_dir
+
+require 'irb'
+puts "Your database is stored in DB..."
+IRB.start
diff -pruN 1.5.1-1/CHANGELOG 3.0.0-1/CHANGELOG
--- 1.5.1-1/CHANGELOG	2008-04-30 17:58:41.000000000 +0100
+++ 3.0.0-1/CHANGELOG	2009-05-04 17:36:33.000000000 +0100
@@ -1,78 +1,2327 @@
+=== 3.0.0 (2009-05-04)
+
+* Remove dead threads from connection pool if the pool is full and a connection is requested (jeremyevans)
+
+* Add autoincrementing primary key support in the Oracle adapter, using a sequence and trigger (jeremyevans, Mike Golod)
+
+* Make Model#save use the same server it uses for saving as for retrieving the saved record (jeremyevans)
+
+* Add Database#database_type method, for identifying which type of database the object is connecting to (jeremyevans)
+
+* Add ability to reset primary key sequences in the PostgreSQL adapter (jeremyevans)
+
+* Fix parsing of non-simple sequence names (that contain uppercase, spaces, etc.) in the PostgreSQL adapter (jeremyevans)
+
+* Support dumping indexes in the schema_dumper extension (jeremyevans)
+
+* Add index parsing to PostgreSQL, MySQL, SQLite, and JDBC adapters (jeremyevans)
+
+* Correctly quote SQL Array references, and handle qualified identifiers with them (e.g. :table__column.sql_subscript(1)) (jeremyevans)
+
+* Allow dropping an index with a name different than the default name (jeremyevans)
+
+* Allow Dataset#from to remove existing FROM tables when called without an argument, instead of raising an error later (jeremyevans)
+
+* Fix string quoting on Oracle so it doesn't double backslashes (jeremyevans)
+
+* Alias the count function call in Dataset#count, fixes use on MSSQL (akitaonrails, jeremyevans)
+
+* Allow QualifiedIdentifiers to be qualified, to allow :column.qualify(:table).qualify(:schema) (jeremyevans)
+
+* Allow :db_type=>'mssql' option to be respected when using the DBI adapter (akitaonrails)
+
+* Add schema_dumper extension, for dumping schema of tables (jeremyevans)
+
+* Allow generic database types specified as ruby types to take options (jeremyevans)
+
+* Change Dataset#exclude to invert given hash argument, not negate it (jeremyevans)
+
+* Make Dataset#filter and related methods treat multiple arguments more intuitively (jeremyevans)
+
+* Fix full text searching with multiple search terms on MySQL (jeremyevans)
+
+* Fix altering a column name, type, default, or NULL/NOT NULL status on MySQL (jeremyevans)
+
+* Fix index type syntax on MySQL (jeremyevans)
+
+* Add temporary table support, via :temp option to Database#create_table (EppO, jeremyevans)
+
+* Add Amalgalite adapter (jeremyevans)
+
+* Remove Sequel::Metaprogramming#metaattr_accessor and metaattr_reader (jeremyevans)
+
+* Remove Dataset#irregular_function_sql (jeremyevans)
+
+* Add Dataset#full_text_sql to the MySQL adapter (dusty)
+
+* Fix schema type parsing of decimal types on MySQL (jeremyevans)
+
+* Make Dataset#quote_identifier work with SQL::Identifiers (jeremyevans)
+
+* Remove methods and features deprecated in 2.12.0 (jeremyevans)
+
+=== 2.12.0 (2009-04-03)
+
+* Deprecate Java::JavaSQL::Timestamp#usec (jeremyevans)
+
+* Fix Model.[] optimization introduced in 2.11.0 for databases that don't use LIMIT (jacaetevha)
+
+* Don't use the model association plugin if SEQUEL_NO_ASSOCIATIONS constant or environment variable is defined (jeremyevans)
+
+* Don't require core_sql if SEQUEL_NO_CORE_EXTENSIONS constant or environment variable is defined (jeremyevans)
+
+* Add validation_helpers model plugin, which adds instance level validation support similar to previously standard validations, with a different API (jeremyevans)
+
+* Split multi_insert into 2 methods with separate APIs, multi_insert for hashes, import for arrays of columns and values (jeremyevans)
+
+* Deprecate Dataset#transform and Model.serialize, and model serialization plugin (jeremyevans)
+
+* Add multi_insert_update to the MySQL adapter, used for setting specific update behavior when an error occurs when using multi_insert (dusty)
+
+* Add multi_insert_ignore to the MySQL adapter, used for skipping errors on row inserts when using multi_insert (dusty)
+
+* Add Sequel::MySQL.convert_invalid_date_time accessor for dealing with dates like "0000-00-00" and times like "25:00:00" (jeremyevans, epugh)
+
+* Eliminate internal dependence on core_sql extensions (jeremyevans)
+
+* Deprecate Migration and Migrator, require 'sequel/extensions/migration' if you want them (jeremyevans)
+
+* Denamespace Sequel::Error decendants (e.g. use Sequel::Rollback instead of Sequel::Error::Rollback) (jeremyevans)
+
+* Deprecate Error::InvalidTransform, Error::NoExistingFilter, and Error::InvalidStatement (jeremyevans)
+
+* Deprecate Dataset#[] when called without an argument, and Dataset#map when called with an argument and a block (jeremyevans)
+
+* Fix aliasing columns in the JDBC adapter (per.melin) (#263)
+
+* Make Database#rename_table remove the cached schema entry for the table (jeremyevans)
+
+* Make Database schema sql methods private (jeremyevans)
+
+* Deprecate Database #multi_threaded? and #logger (jeremyevans)
+
+* Make Dataset#where always affect the WHERE clause (jeremyevans)
+
+* Deprecate Object#blank? and related extensions, require 'sequel/extensions/blank' to get them back (jeremyevans)
+
+* Move lib/sequel_core into lib/sequel and lib/sequel_model into lib/sequel/model (jeremyevans)
+
+* Remove Sequel::Schema::SQL module, move methods into Sequel::Database (jeremyevans)
+
+* Support creating and dropping schema qualified views (jeremyevans)
+
+* Fix saving a newly inserted record in an after_create or after_save hook (jeremyevans)
+
+* Deprecate Dataset#print and PrettyTable, require 'sequel/extensions/pretty_table' if you want them (jeremyevans)
+
+* Deprecate Database#query and Dataset#query, require 'sequel/extensions/query' if you want them (jeremyevans)
+
+* Deprecate Dataset#paginate and #each_page, require 'sequel/extensions/pagination' if you want them (jeremyevans)
+
+* Fix ~{:bool_col=>true} and related inversions of boolean values (jeremyevans)
+
+* Add disable_insert_returning method to PostgreSQL datasets, so they fallback to just using INSERT (jeremyevans)
+
+* Don't use savepoints by default on PostgreSQL, use the :savepoint option to Database#transaction to use a savepoint (jeremyevans)
+
+* Deprecate Database#transaction accepting a server symbol argument, use an options hash with the :server option (jeremyevans)
+
+* Add Model.use_transactions for setting whether models should use transactions when destroying/saving records (jeremyevans, mjwillson)
+
+* Deprecate Model::Validation::Errors, use Model::Errors (jeremyevans)
+
+* Deprecate string inflection methods, require 'sequel/extensions/inflector' if you use them (jeremyevans)
+
+* Deprecate Model validation class methods, override Model#validate instead or Model.plugin validation_class_methods (jeremyevans)
+
+* Deprecate Model schema methods, use Model.plugin :schema (jeremyevans)
+
+* Deprecate Model hook class methods, use instance methods instead or Model.plugin :hook_class_methods (jeremyevans)
+
+* Deprecate Model.set_sti_key, use Model.plugin :single_table_inheritance (jeremyevans)
+
+* Deprecate Model.set_cache, use Model.plugin :caching (jeremyevans)
+
+* Move most model instance methods into Model::InstanceMethods, for easier overriding of instance methods for all models (jeremyevans)
+
+* Move most model class methods into Model::ClassMethods, for easier overriding of class methods for all models (jeremyevans)
+
+* Deprecate String#to_date, #to_datetime, #to_time, and #to_sequel_time, use require 'sequel/extensions/string_date_time' if you want them (jeremyevans)
+
+* Deprecate Array#extract_options! and Object#is_one_of? (jeremyevans)
+
+* Deprecate Object#meta_def, #meta_eval, and #metaclass (jeremyevans)
+
+* Deprecate Module#class_def, #class_attr_overridable, #class_attr_reader, #metaalias, #metaattr_reader, and #metaatt_accessor (jeremyevans)
+
+* Speed up the calling of most column accessor methods, and reduce memory overhead of creating them (jeremyevans)
+
+* Deprecate Model#set_restricted using Model#[] if no setter method exists, a symbol is used, and the columns are not set (jeremyevans)
+
+* Deprecate Model#set_with_params and #update_with_params (jeremyevans)
+
+* Deprecate Model#save!, use Model.save(:validate=>false) (jeremyevans)
+
+* Deprecate Model#dataset (jeremyevans)
+
+* Deprecate Model.is and Model.is_a, use Model.plugin for plugins (jeremyevans)
+
+* Deprecate Model.str_columns, Model#str_columns, #set_values, #update_values (jeremyevans)
+
+* Deprecate Model.delete_all, .destroy_all, .size, and .uniq (jeremyevans)
+
+* Copy all current dataset options when calling Model.db= (jeremyevans)
+
+* Deprecate Model.belongs_to, Model.has_many, and Model.has_and_belongs_to_many (jeremyevans)
+
+* Remove SQL::SpecificExpression, have subclasses inherit from SQL::Expression instead (jeremyevans)
+
+* Deprecate SQL::CastMethods#cast_as (jeremyevans)
+
+* Deprecate calling Database#schema without a table argument (jeremyevans)
+
+* Remove cached version of @db_schema in model instances to reduce memory and marshalling overhead (tmm1)
+
+* Deprecate Dataset#quote_column_ref and Dataset#symbol_to_column_ref (jeremyevans)
+
+* Deprecate Dataset#size and Dataset#uniq (jeremyevans)
+
+* Deprecate passing options to Dataset#each, #all, #single_record, #single_value, #sql, #select_sql, #update, #update_sql, #delete, #delete_sql, and #exists (jeremyevans)
+
+* Deprecate Dataset#[Integer] (jeremyevans)
+
+* Deprecate Dataset#create_view and Dataset#create_or_replace_view (jeremyevans)
+
+* Model datasets now have a model accessor that returns the related model (jeremyevans)
+
+* Model datasets no longer have :models and :polymorphic_key options (jeremyevans)
+
+* Deprecate Dataset.dataset_classes, Dataset#model_classes, Dataset#polymorphic_key, and Dataset#set_model (jeremyevans)
+
+* Allow Database#get and Database#select to take a block (jeremyevans)
+
+* Deprecate Database#>> (jeremyevans)
+
+* Deprecate String#to_blob and Sequel::SQL::Blob#to_blob (jeremyevans)
+
+* Deprecate use of Symbol#| for SQL array subscripts, add Symbol#sql_subscript (jeremyevans)
+
+* Deprecate Symbol#to_column_ref (jeremyevans)
+
+* Deprecate String#expr (jeremyevans)
+
+* Deprecate Array#to_sql, String#to_sql, and String#split_sql (jeremyevans)
+
+* Deprecate passing an array to Database#<< (jeremyevans)
+
+* Deprecate Range#interval (jeremyevans)
+
+* Deprecate Enumerable#send_each (jeremyevans)
+
+* Deprecate Hash#key on ruby 1.8, change some SQLite adapter constants (jeremyevans)
+
+* Deprecate Sequel.open, Sequel.use_parse_tree=?, and the upcase_identifier methods (jeremyevans)
+
+* Deprecate virtual row blocks without block arguments, unless Sequel.virtual_row_instance_eval is enabled (jeremyevans)
+
+* Support schema parsing in the Oracle adapter (jacaetevha)
+
+* Allow virtual row blocks to be instance_evaled, add Sequel.virtual_row_instance_eval= (jeremyevans)
+
+=== 2.11.0 (2009-03-02)
+
+* Optimize Model.[] by using static sql when possible, for a 30-40% speed increase (jeremyevans)
+
+* Add Dataset#with_sql, which returns a clone of the datatset with static SQL (jeremyevans)
+
+* Refactor Dataset#literal so it doesn't need to be overridden in subadapters, for a 20-25% performance increase (jeremyevans)
+
+* Remove SQL::IrregularFunction, no longer used internally (jeremyevans)
+
+* Allow String#lit to take arguments and return a SQL::PlaceholderLiteralString (jeremyevans)
+
+* Add Model#set_associated_object, used by the many_to_one setter method, for easier overriding (jeremyevans)
+
+* Allow use of database independent types when casting (jeremyevans)
+
+* Give association datasets knowledge of the model object that created them and the related association reflection (jeremyevans)
+
+* Make Dataset#select, #select_more, #order, #order_more, and #get take a block that yields a SQL::VirtualRow, similar to #filter (jeremyevans)
+
+* Fix stored procedures in MySQL adapter when multiple arguments are used (clivecrous)
+
+* Add :conditions association option, for easier filtering of associated objects (jeremyevans)
+
+* Add :clone association option, for making clones of existing associations (jeremyevans)
+
+* Handle typecasting invalid date strings (and possible other types) correctly (jeremyevans)
+
+* Add :compress=>false option to MySQL adapter to turn off compression of client-server connection (tmm1)
+
+* Set SQL_AUTO_IS_NULL=0 on MySQL connections, disable with :auto_is_null=>false (tmm1)
+
+* Add :timeout option to MySQL adapter, default to 30 days (tmm1)
+
+* Set MySQL encoding using Mysql#options so it works across reconnects (tmm1)
+
+* Fully support blobs on SQLite (jeremyevans)
+
+* Add String#to_sequel_blob, alias String#to_blob to that (jeremyevans)
+
+* Fix default index names when a non-String or Symbol column is used (jeremyevans)
+
+* Fix some ruby -w warnings (jeremyevans) (#259)
+
+* Fix issues with default column values, table names, and quoting in the rename_column and drop_column support in shared SQLite adapter (jeremyevans)
+
+* Add rename_column support to SQLite shared adapter (jmhodges)
+
+* Add validates_inclusion_of validation (jdunphy)
+
+=== 2.10.0 (2009-02-03)
+
+* Don't use a default schema any longer in the shared PostgreSQL adapter (jeremyevans)
+
+* Make Dataset#quote_identifier return LiteralStrings as-is (jeremyevans)
+
+* Support symbol keys and unnested hashes in the sequel command line tool's yaml config support (jeremyevans)
+
+* Add schema parsing support to the JDBC adapter (jeremyevans)
+
+* Add per-database type translation support for schema changes, translating ruby classes to database specific types (jeremyevans)
+
+* Add Sequel::DatabaseConnectionError, for indicating that Sequel wasn't able to connect to the database (jeremyevans)
+
+* Add validates_not_string validation, useful in conjunction with raise_on_typecast_failure = false (jeremyevans)
+
+* Don't modify Model#new? and Model#changed_columns when saving a record until after the after hooks have been run (tamas, jeremyevans)
+
+* Database#quote_identifiers= now affects future schema modification statements, even if it is not used before one of the schema modification statements (jeremyevans)
+
+* Fix literalization of blobs when using the PostreSQL JDBC subadapter (jeremyevans)
+
+* Fix literalization of date and time types when using the MySQL JDBC subadapter (jeremyevans)
+
+* Convert some Java specific types to ruby types on output in the JDBC adapter (jeremyevans)
+
+* Add Database#tables method to JDBC adapter (jeremyevans)
+
+* Add H2 JDBC subadapter (logan_barnett, david_koontz, james_britt, jeremyevans)
+
+* Add identifer_output_method, used for converting identifiers coming out of the database, replacing the lowercase support on some databases (jeremyevans)
+
+* Add identifier_input_method, used for converting identifiers going into the database, replacing upcase_identifiers (jeremyevans)
+
+* Add :allow_missing validation option, useful if the database provides a good default (jeremyevans)
+
+* Fix literalization of SQL::Blobs in DataObjects and JDBC adapter's postgresql subadapters when ruby 1.9 is used (jeremyevans)
+
+* When using standard strings in the postgres adapter with the postgres-pr driver, use custom string escaping to prevent errors (jeremyevans)
+
+* Before hooks now run in reverse order of being added, so later ones are run first (tamas)
+
+* Add Firebird adapter, requires Firebird ruby driver located at http://github.com/wishdev/fb (wishdev)
+
+* Don't clobber the following Symbol instance methods when using ruby 1.9: [], <, <=, >, >= (jeremyevans)
+
+* Quote the table name and the index for PostgreSQL index creation (jeremyevans)
+
+* Add DataObjects adapter, supporting PostgreSQL, MySQL, and SQLite (jeremyevans)
+
+* Add ability for Database#create_table to take options, support specifying MySQL engine, charset, and collate per table (pusewicz, jeremyevans)
+
+* Add Model.add_hook_type class method, for adding your own hook types, mostly for use by plugin authors (pkondzior, jeremyevans)
+
+* Add Sequel.version for getting the internal version of Sequel (pusewicz, jeremyevans)
+
+=== 2.9.0 (2009-01-12)
+
+* Add -L option to sequel command line tool to load all .rb files in the given directory (pkondzior, jeremyevans)
+
+* Fix Dataset#destroy for model datasets that can't handle nested queries (jeremyevans)
+
+* Improve the error messages in parts of Sequel::Model (jeremyevans, pusewicz)
+
+* Much better support for Dataset#{union,except,intersect}, allowing chaining and respecting order (jeremyevans)
+
+* Default to logging only WARNING level messages when connecting to PostgreSQL (jeremyevans)
+
+* Fix add_foreign_key for MySQL (jeremyevans, aphyr)
+
+* Correctly literalize BigDecimal NaN and (+-)Infinity values (jeremyevans) (#256)
+
+* Make Sequel raise an Error if you attempt to subclass Sequel::Model before setting up a database connection (jeremyevans)
+
+* Add Sequel::BeforeHookFailed exception to be raised when a record fails because a before hook fails (bougyman)
+
+* Add Sequel::ValidationFailed exception to be raised when a record fails because a validation fails (bougyman)
+
+* Make Database#schema raise an error if given a table that doesn't exist (jeremyevans) (#255)
+
+* Make Model#inspect call Model#inspect_values private method for easier overloading (bougyman)
+
+* Add methods to create and drop functions, triggers, and procedural languages on PostgreSQL (jeremyevans)
+
+* Fix Dataset#count when using UNION, EXCEPT, or INTERSECT (jeremyevans)
+
+* Make SQLite keep table's primary key information when dropping columns (jmhodges)
+
+* Support dropping indicies on SQLite (jmhodges)
+
+=== 2.8.0 (2008-12-05)
+
+* Support drop column operations inside a transaction on sqlite (jeremyevans)
+
+* Support literal strings with placeholders and subselects in prepared statements (jeremyevans)
+
+* Have the connection pool remove disconnected connections when the adapter supports it (jeremyevans)
+
+* Make Dataset#exists return a LiteralString (jeremyevans)
+
+* Support multiple SQL statements in one query in the MySQL adapter (jeremyevans)
+
+* Add stored procedure support for the MySQL and JDBC adapters (jeremyevans, krsgoss) (#252)
+
+* Support options when altering a column's type (for changing enums, varchar size, etc.) (jeremyevans)
+
+* Support AliasedExpressions in tables when using implicitly qualified arguments in joins (jeremyevans)
+
+* Support Dataset#except on Oracle (jeremyevans)
+
+* Raise errors when EXCEPT/INTERSECT is used when not supported (jeremyevans)
+
+* Fix ordering of UNION, INTERSECT, and EXCEPT statements (jeremyevans) (#253)
+
+* Support aliasing subselects in the Oracle adapter (jeremyevans)
+
+* Add a subadapter for the Progress RDBMS to the ODBC adapter (:db_type=>'progress') (groveriffic) (#251)
+
+* Make MySQL and Oracle adapters raise an Error if asked to do a SELECT DISTINCT ON (jeremyevans)
+
+* Set standard_conforming_strings = ON by default when using PostgreSQL, turn off with Sequel::Postgres.force_standard_strings = false (jeremyevans) (#247)
+
+* Fix Database#rename_table when using PostgreSQL (jeremyevans) (#248)
+
+* Whether to upcase or quote identifiers can now be set separately, via Sequel.upcase_identifiers= or the :upcase_identifiers database option (jeremyevans)
+
+* Support transactions in the ODBC adapter (dlee)
+
+* Support multi_insert_sql and unicode string literals in MSSQL shared adapter (dlee)
+
+* Make PostgreSQL use the default schema if parsing the schema for all tables at once, even if :schema=>nil option is used (jeremyevans)
+
+* Make MySQL adapter not raise an error when giving an SQL::Identifier object to the schema modification methods such as create_table (jeremyevans)
+
+* The keys of the hash returned by Database#schema without a table name are now quoted strings instead of symbols (jeremyevans)
+
+* Make Database#schema to handle implicit schemas on all databases and multiple identifier object types (jeremyevans)
+
+* Remove Sequel.odbc_mssql method (jeremyevans) (#249)
+
+* More optimization of Model#initialize (jeremyevans)
+
+* Treat interval as it's own type, not an integer type (jeremyevans)
+
+* Allow use of implicitly qualified symbol as argument to Symbol#qualify (:a.qualify(:b__c)=>b.c.a), fixes model associations in different schemas (jeremyevans) (#246)
+
+=== 2.7.1 (2008-11-04)
+
+* Fix PostgreSQL Date optimization so that it doesn't reject dates like 11/03/2008 (jeremyevans)
+
+=== 2.7.0 (2008-11-03)
+
+* Transform AssociationReflection from a single class to a class hierarchy (jeremyevans)
+
+* Optimize Date object creation in PostgreSQL adapter (jeremyevans)
+
+* Allow easier creation of custom association types, though support for them may still be suboptimal (jeremyevans)
+
+* Add :eager_grapher option to associations, which the user can use to override the default eager_graph code (jeremyevans)
+
+* Associations are now inherited when a model class is subclassed (jeremyevans)
+
+* Instance methods added by associations are now added to an anonymous module the class includes, allowing you to override them and use super (jeremyevans)
+
+* Add #add_graph_aliases (select_more for graphs), and allow use of arbitrary expressions when graphing (jeremyevans)
+
+* Fix a corner case where the wrong table name is used in eager_graph (jeremyevans)
+
+* Make Dataset#join_table take an option hash instead of a table_alias argument, add support for :implicit_qualifier option (jeremyevans)
+
+* Add :left_primary_key and :right_primary_key options to many_to_many associations (jeremyevans)
+
+* Add :primary_key option to one_to_many and many_to_one associations (jeremyevans)
+
+* Make after_load association callbacks take effect when eager loading via eager (jeremyevans)
+
+* Add a :uniq association option to many_to_many associations (jeremyevans)
+
+* Support using any expression as the argument to Symbol#like (jeremyevans)
+
+* Much better support for multiple schemas in PostgreSQL (jeremyevans) (#243)
+
+* The first argument to Model#initalize can no longer be nil, it must be a hash if it is given (jeremyevans)
+
+* Remove Sequel::Model.lazy_load_schema= setting (jeremyevans)
+
+* Lazily load model instance options such as raise_on_save_failure, for better performance (jeremyevans)
+
+* Make Model::Validiation::Errors more Rails-compatible (jeremyevans)
+
+* Refactor model hooks for performance (jeremyevans)
+
+* Major performance enhancement when fetching rows using PostgreSQL (jeremyevans)
+
+* Don't typecast serialized columns in models (jeremyevans)
+
+* Add Array#sql_array to handle ruby arrays of all two pairs as SQL arrays (jeremyevans) (#245)
+
+* Add ComplexExpression#== and #eql?, for checking equality (rubymage) (#244)
+
+* Allow full text search on PostgreSQL to include rows where a search column is NULL (jeremyevans)
+
+* PostgreSQL full text search queries with multiple columns are joined with space to prevent joining border words to one (michalbugno)
+
+* Don't modify a dataset's cached column information if calling #each with an option that modifies the columns (jeremyevans)
+
+* The PostgreSQL adapter will now generally default to using a unix socket in /tmp if no host is specified, instead of a tcp socket to localhost (jeremyevans)
+
+* Make Dataset#sql call Dataset#select_sql instead of being an alias, to allow for easier subclassing (jeremyevans)
+
+* Split Oracle adapter into shared and unshared parts, so Oracle is better supported when using JDBC (jeremyevans)
+
+* Fix automatic loading of Oracle driver when using JDBC adapter (bburton333) (#242)
+
+=== 2.6.0 (2008-10-11)
+
+* Make the sqlite adapter respect the Sequel.datetime_class setting, for timestamp and datetime types (jeremyevans)
+
+* Enhance the CASE statement support to include an optional expression (jarredholman)
+
+* Default to using the simple language if no language is specified for a full text index on PostgreSQL (michalbugno)
+
+* Add Model.raise_on_typecast_failure=, which makes it possible to not raise errors on invalid typecasts (michalbugno)
+
+* Add schema.rdoc file, which provides an brief description of the various parts of Sequel related to schema modification (jeremyevans)
+
+* Fix constraint generation when not using a proc or interpolated string (jeremyevans)
+
+* Make eager_graph respect associations' :order options (use :order_eager_graph=>false to disable) (jeremyevans)
+
+* Cache negative lookup when eagerly loading many_to_one associations where no objects have an associated object (jeremyevans)
+
+* Allow string keys to be used when using Dataset#multi_insert (jeremyevans)
+
+* Fix join_table when doing the first join for a dataset where the first source is a dataset when using unqualified columns (jeremyevans)
+
+* Fix a few corner cases in eager_graph (jeremyevans)
+
+* Support transactions on MSSQL (jeremyevans)
+
+* Use string literals in AS clauses on SQLite (jeremyevans) (#241)
+
+* AlterTableGenerator#set_column_allow_null was added to SET/DROP NOT NULL for columns (divoxx)
+
+* Database#tables now works for MySQL databases using the JDBC adapter (jeremyevans)
+
+* Database#drop_view can now take multiple arguments to drop multiple views at once (jeremyevans)
+
+* Schema modification methods (e.g. drop_table, create_table!) now remove the cached schema entry (jeremyevans)
+
+* Models can now determine their primary keys by looking at the schema (jeremyevans)
+
+* No longer include :numeric_precision and :max_chars entries in the schema column hashes, use the :db_type entry instead (jeremyevans)
+
+* Make schema parsing on PostgreSQL handle implicit schemas (e.g. schema(:schema__table)), so it works with models for tables outside the public schema (jeremyevans)
+
+* Significantly speed up schema parsing on MySQL (jeremyevans)
+
+* Include primary key information when parsing the schema (jeremyevans)
+
+* Fix schema generation of composite foreign keys on MySQL (clivecrous, jeremyevans)
+
+=== 2.5.0 (2008-09-03)
+
+* Add Dataset #set_defaults and #set_overrides, used for scoping the values used in insert/update statements (jeremyevans)
+
+* Allow Models to use the RETURNING clause when inserting records on PostgreSQL (jeremyevans)
+
+* Raise Sequel::DatabaseError instead of generic Sequel::Error for database errors, don't swallow tracebacks (jeremyevans)
+
+* Use INSERT ... RETURNING ... with PostgreSQL 8.2 and higher (jeremyevans)
+
+* Make insert_sql, delete_sql, and update_sql respect the :sql option (jeremyevans)
+
+* Default to converting 2 digit years, use Sequel.convert_two_digit_years = false to get back the old behavior (jeremyevans)
+
+* Make the PostgreSQL adapter with the pg driver use async_exec, so it doesn't block the entire interpreter (jeremyevans)
+
+* Make the schema generators support composite primary and foreign keys and unique constraints (jarredholman)
+
+* Work with the 2008.08.17 version of the pg gem (erikh)
+
+* Disallow abuse of SQL function syntax for types (use :type=>:varchar, :size=>255 instead of :type=>:varchar[255]) (jeremyevans)
+
+* Quote index names when creating or dropping indexes (jeremyevans, SanityInAnarchy)
+
+* Don't have column accessor methods override plugin instance methods (jeremyevans)
+
+* Allow validation of multiple attributes at once, with built in support for uniqueness checking of multiple columns (jeremyevans)
+
+* In PostgreSQL adapter, fix inserting a row with a primary key value inside a transaction (jeremyevans)
+
+* Allow before_save and before_update to affect the columns saved by save_changes (jeremyevans)
+
+* Make Dataset#single_value work when graphing, which fixes count and paginate on graphed datasets (jeremyevans)
+
+=== 2.4.0 (2008-08-06)
+
+* Handle Java::JavaSql::Date type in the JDBC adapter (jeremyevans)
+
+* Add support for read-only slave/writable master databases and database sharding (jeremyevans)
+
+* Remove InvalidExpression, InvalidFilter, InvalidJoinType, and WorkerStop exceptions (jeremyevans)
+
+* Add prepared statement/bound variable support (jeremyevans)
+
+* Fix anonymous column names in the ADO adapter (nusco)
+
+* Remove odbc_mssql adapter, use :db_type=>'mssql' option instead (jeremyevans)
+
+* Split MSSQL specific syntax into separate file, usable by ADO and ODBC adapters (nusco, jeremyevans)
+
+=== 2.3.0 (2008-07-25)
+
+* Enable almost full support for MySQL using JDBC (jeremyevans)
+
+* Fix ODBC adapter's conversion of ::ODBC::Time values (Michael Xavier)
+
+* Enable full support for SQLite-JDBC using the JDBC adapter (jeremyevans)
+
+* Minor changes to allow for full Ruby 1.9 compatibility (jeremyevans)
+
+* Make Database#disconnect work for the ADO adapter (spicyj)
+
+* Don't raise an exception in the ADO adapter if the dataset contains no records (nusco)
+
+* Enable almost full support of PostgreSQL-JDBC using the JDBC adapter (jeremyevans)
+
+* Remove Sequel::Worker (jeremyevans)
+
+* Make PostgreSQL adapter not raise an error when inserting records into a table without a primary key (jeremyevans)
+
+* Make Database.uri_to_options a private class method (jeremyevans)
+
+* Make JDBC load drivers automatically for PostgreSQL, MySQL, SQLite, Oracle, and MSSQL (jeremyevans)
+
+* Make Oracle adapter work with a nonstandard Oracle database port (pavel.lukin)
+
+* Typecast '' to nil by default for non-string non-blob columns, add typecast_empty_string_to_nil= model class and instance methods (jeremyevans)
+
+* Use a simpler select in Dataset#empty?, fixes use with MySQL (jeremyevans)
+
+* Add integration test suite, testing sequel against a real database, with nothing mocked (jeremyevans)
+
+* Make validates_length_of default tag depend on presence of options passed to it (jeremyevans)
+
+* Combine the directory structure for sequel_model and sequel_core, now there is going to be only one gem named sequel (jeremyevans)
+
+=== 2.2.0 (2008-07-05)
+
+* Add :extend association option, extending the dataset with module(s) (jeremyevans)
+
+* Add :after_load association callback option, called after associated objects have been loaded from the database (jeremyevans)
+
+* Make validation methods support a :tag option, to work correctly with source reloading (jeremyevans)
+
+* Add :before_add, :after_add, :before_remove, :after_remove association callback options (jeremyevans)
+
+* Break many_to_one association setter method in two parts, for easier overriding (jeremyevans)
+
+* Model.validates_presence_of now considers false as present instead of absent (jeremyevans)
+
+* Add Model.raise_on_save_failure, raising errors on save failure instead of return false (now nil), default to true (jeremyevans)
+
+* Add :eager_loader association option, to specify code to be run when eager loading (jeremyevans)
+
+* Make :many_to_one associations support :dataset, :order, :limit association options, as well as block arguments (jeremyevans)
+
+* Add :dataset association option, which overrides the default base dataset to use (jeremyevans)
+
+* Add :eager_graph association option, works just like :eager except it uses #eager_graph (jeremyevans)
+
+* Add :graph_join_table_join_type association option (jeremyevans)
+
+* Add :graph_only_conditions and :graph_join_table_only_conditions association options (jeremyevans)
+
+* Add :graph_block and :graph_join_table_block association options (jeremyevans)
+
+* Set the model's dataset's columns in addition to the model's columns when loading the schema for a model (jeremyevans)
+
+* Make caching work correctly with subclasses (jeremyevans)
+
+* Add the Model.to_hash dataset method (jeremyevans)
+
+* Filter blocks now yield a SQL::VirtualRow argument, which is useful if another library defines operator methods on Symbol (jeremyevans)
+
+* Add Symbol#identifier method, to make x__a be treated as "x__a" instead of "x"."a" (jeremyevans)
+
+* Dataset#update no longer takes a block, please use a hash argument with the expression syntax instead (jeremyevans)
+
+* ParseTree support has been removed from Sequel (jeremyevans)
+
+* Database#drop_column is now supported in the SQLite adapter (abhay)
+
+* Tinyint columns can now be considered integers instead of booleans by setting Sequel.convert_tinyint_to_bool = false (samsouder)
+
+* Allow the use of URL parameters in connection strings (jeremyevans)
+
+* Ignore any previously selected columns when using Dataset#graph for the first time (jeremyevans)
+
+* Dataset#graph now accepts a block which is passed to join_table (jeremyevans)
+
+* Make Dataset#columns ignore any filtering, ordering, and distinct clauses (jeremyevans)
+
+* Use the safer connection-specific string escaping methods for PostgreSQL (jeremyevans)
+
+* Database#transaction now yields a connection when using the Postgres adapter, just like it does for other adapters (jeremyevans)
+
+* Dataset#count now works for a limited dataset (divoxx)
+
+* Database#add_index is now supported in the SQLite adapter (abhay)
+
+* Sequel's MySQL adapter should no longer conflict with ActiveRecord's use of MySQL (careo)
+
+* Treat Hash as expression instead of column alias when used in DISTINCT, ORDER BY, and GROUP BY clauses (jeremyevans)
+
+* PostgreSQL bytea fields are now fully supported (dlee)
+
+* For PostgreSQL, don't raise an error when assigning a value to a SERIAL PRIMARY KEY field when inserting records (jeremyevans)
+
+=== 2.1.0 (2008-06-17)
+
+* Break association add_/remove_/remove_all_ methods into two parts, for easier overriding (jeremyevans)
+
+* Add Model.strict_param_setting, on by default, which raises errors if a missing/restricted method is called via new/set/update/etc. (jeremyevans)
+
+* Raise errors when using association methods on objects without valid primary keys (jeremyevans)
+
+* The model's primary key is a restricted column by default, Add model.unrestrict_primary_key to get the old behavior (jeremyevans)
+
+* Add Model.set_(allowed|restricted)_columns, which affect which columns create/new/set/update/etc. modify (jeremyevans)
+
+* Calls to Model.def_dataset_method with a block are cached and reapplied to the new dataset if set_dataset is called, even in a subclass (jeremyevans)
+
+* The :reciprocal option to associations should now be the symbol name of the reciprocal association, not an instance variable symbol (jeremyevans)
+
+* Add Model#associations, which is a hash holding a cache of associated objects, with each association being a separate key (jeremyevans)
+  
+* Make all associations support a :graph_select option, specifying a column or array of columns to select when using eager_graph (jeremyevans)
+
+* Bring back Model#set and Model#update, now the same as Model#set_with_params and Model#update_with_params (jeremyevans)
+
+* Allow model datasets to call to_hash without any arguments, which allows easy creation of identity maps (jeremyevans)
+
+* Add Model.set_sti_key, for easily setting up single table inheritance (jeremyevans)
+
+* Make all associations support a :read_only option, which doesn't add methods that modify the database (jeremyevans)
+
+* Make *_to_many associations support a :limit option, for specifying a limit to the resulting records (and possibly an offset) (jeremyevans)
+
+* Make association block argument and :eager option affect the _dataset method (jeremyevans)
+
+* Add a :one_to_one option to one_to_many associations, which creates a getter and setter similar to many_to_one (a.k.a. has_one) (jeremyevans)
+
+* add_ and remove_ one_to_many association methods now raise an error if the passed object cannot be saved, instead of saving without validation (jeremyevans)
+
+* Add support for :if option on validations, using a symbol (specifying an instance method) or a proc (dtsato)
+
+* Support bitwise operators for NumericExpressions: &, |, ^, ~, <<, >> (jeremyevans)
+
+* No longer raise an error for Dataset#filter(true) or Dataset#filter(false) (jeremyevans)
+
+* Allow Dataset #filter, #or, #exclude and other methods that call them to use both the block and regular arguments (jeremyevans)
+
+* ParseTree support is now officially deprecated, use Sequel.use_parse_tree = false to use the expression (blockless) filters inside blocks (jeremyevans)
+
+* Remove :pool_reuse_connections ConnectionPool/Database option, MySQL users need to be careful with nested queries (jeremyevans)
+
+* Allow Dataset#graph :select option to take an array of columns to select (jeremyevans)
+
+* Allow Dataset#to_hash to be called with only one argument, allowing for easy creation of lookup tables for a single key (jeremyevans)
+
+* Allow join_table to accept a block providing the aliases and previous joins, that allows you to specify arbitrary conditions properly qualified (jeremyevans)
+
+* Support NATURAL, CROSS, and USING joins in join_table (jeremyevans)
+
+* Make sure HAVING comes before ORDER BY, per the SQL standard and at least MySQL, PostgreSQL, and SQLite (juco)
+
+* Add cast_numeric and cast_string methods for use in the Sequel DSL, that have default types and wrap the object in the correct class (jeremyevans)
+
+* Add Symbol#qualify, for adding a table/schema qualifier to a column/table name (jeremyevans)
+
+* Remove Module#metaprivate, since it duplicates the standard Module#private_class_method (jeremyevans)
+
+* Support the SQL CASE expression via Array#case and Hash#case (jeremyevans)
+
+* Support the SQL EXTRACT function: :date.extract(:year) (jeremyevans)
+
+* Convert numeric fields to BigDecimal in PostgreSQL adapter (jeremyevans)
+
+* Add :decimal fields to the schema parser (jeremyevans)
+
+* The expr argument in join table now allows the same argument as filter, so it can take a string or a blockless filter expression (brushbox, jeremyevans)
+
+* No longer assume the expr argument to join_table references the primary key column (jeremyevans)
+
+* Rename the Sequel.time_class setting to Sequel.datetime_class (jeremyevans)
+
+* Add savepoint/nesting support to postgresql transactions (elven)
+
+* Use the specified table alias when joining a dataset, instead of the automatically generated alias (brushbox)
+
+=== 2.0.1 (2008-06-04)
+
+* Make the choice of Time or DateTime optional for typecasting :datetime types, default to Time (jeremyevans)
+
+* Reload database schema for table when calling Model.create_table (jeremyevans)
+
+* Have PostgreSQL money type use BigDecimal instead of Float (jeremyevans)
+
+* Have the PostgreSQL and MySQL adapters use the Sequel.time_class setting for datetime/timestamp types (jeremyevans)
+
+* Add Sequel.time_class and String#to_sequel_time, used for converting time values from the database to either Time (default) or DateTime (jeremyevans)
+
+* Make identifier quoting uppercase by default, to work better with the SQL standard, override in PostgreSQL (jeremyevans) (#232)
+
+* Add StringExpression#+, for simple SQL string concatenation (:x.sql_string + :y) (jeremyevans)
+
+* Make StringMethods.like to a case sensensitive search on MySQL (use ilike for the old behavior) (jeremyevans)
+
+* Add StringMethods.ilike, for case insensitive pattern matching (jeremyevans)
+
+* Refactor ComplexExpression into three subclasses and a few modules, so operators that don't make sense are not defined for the class (jeremyevans)
+
+=== 2.0.0 (2008-06-01)
+
+* Comprehensive update of all documentation (jeremyevans)
+
+* Remove methods deprecated in 1.5.0 (jeremyevans)
+
+* Add typecasting on attribute assignment to Sequel::Model objects, optional but enabled by default (jeremyevans)
+
+* Returning false in one of the before_ hooks now causes the appropriate method(s) to immediately return false (jeremyevans)
+
+* Add remove_all_* association method for *_to_many associations, which removes the association with all currently associated objects (jeremyevans)
+
+* Add Model.lazy_load_schema=, when set to true, it loads the schema on first instantiation (jeremyevans)
+
+* Add before_validation and after_validation hooks, called whenever the model is validated (jeremyevans)
+
+* Add Model.default_foreign_key, a private class method that allows changing the default foreign key that Sequel will use in associations (jeremyevans)
+
+* Cache negative lookup when eagerly loading many_to_one associations (jeremyevans)
+
+* Make all associations support the :select option, not just many_to_many (jeremyevans)
+
+* Allow the use of blocks when eager loading, and add the :eager_block and :allow_eager association options for configuration (jeremyevans)
+  
+* Add the :graph_join_type, :graph_conditions, and :graph_join_table_conditions association options, used when eager graphing (jeremyevans)
+
+* Add AssociationReflection class (subclass of Hash), to make calling a couple of private Model methods unnecessary (jeremyevans)
+
+* Change hook methods so that if a tag/method is specified it overwrites an existing hook block with the same tag/method (jeremyevans)
+
+* Refactor String inflection support, you must use String.inflections instead of Inflector.inflections now (jeremyevans)
+
+* Allow connection to ODBC-MSSQL via a URL (petersumskas) (#230)
+
+* Comprehensive update of all documentation, except for the block filters and adapters (jeremyevans)
+
+* Handle Date and DateTime value literalization correctly in adapters (jeremyevans)
+
+* Literalize DateTime values the same as Time values (jeremyevans)
+
+* MySQL tinyints are now returned as boolean values instead of integers (jeremyevans)
+
+* Set additional MySQL charset options required for creating tables and databases (tmm1)
+
+* Remove methods deprecated in 1.5.0 (jeremyevans)
+
+* Add Module#metaattr_accessor for creating attr_accessors for the metaclass (jeremyevans)
+
+* Add SQL string concatenation support to blockless filters, via Array#sql_string_join (jeremyevans)
+
+* Add Pagination#last_page? and Pagination#first_page? (apeiros)
+
+* Add limited column reflection support, tested on PostgreSQL, MySQL, and SQLite (jeremyevans)
+
+* Allow the use of :schema__table___table_alias syntax for tables, similar to the column support (jeremyevans)
+
+* Merge metaid gem into core_ext.rb and clean it up, so sequel now has no external dependencies (jeremyevans)
+
+* Add Dataset#as, so using a dataset as a column with an alias is not deprecated (jeremyevans)
+
+* Add Dataset#invert, which returns a dataset with inverted HAVING and WHERE clauses (jeremyevans)
+
+* Add blockless filter syntax support (jeremyevans)
+
+* Passing an array to Dataset#order and Dataset#select no longer works, you need to pass multiple arguments (jeremyevans)
+
+* You should use '?' instead of '(?)' when using interpolated strings with array arguments (jeremyevans)
+
+* Dataset.literal now surrounds the literalization of arrays with parentheses (jeremyevans)
+
+* Add echo option (back?) to sequel command line tool, via -E or --echo (jeremyevans)
+
+* Allow databases to have multiple loggers (jeremyevans)
+
+* The sequel command line tool now also accepts a path to a database config YAML file in addition to a URI (mtodd)
+
+* Major update of the postgresql adapter (jdavis, jeremyevans) (#225)
+
+* Make returning inside of a database transaction commit the transaction (ahoward, jeremyevans)
+
+* Dataset#to_table_reference is now protected, and it has a different API (jeremyevans)
+
+* Dataset#join_table and related functions now take an explicit optional table_alias argument, you can no longer include the table alias in the table argument (jeremyevans)
+
+* Aliased and/or qualified columns with embedded spaces can now be specified as symbols (jeremyevans)
+
+* When identifier quoting is enabled, the SQL standard double quote is used by default (jeremyevans)
+
+* When identifier quoting is enabled, quote tables as well as columns (jeremyevans)
+
+* Make identifier quoting optional, enabled by default (jeremyevans)
+
+* Allow Sequel::Database.connect and related methods to take a block that disconnects the database when the block finishes (jeremyevans)
+
+* Add Dataset#unfiltered, for removing filters from dataset (jeremyevans)
+
+* Add add_foreign_key and add_primary_key methods to the AlterTableGenerator (jeremyevans)
+
+* Allow migration files to have more than 3 digits (jeremyevans)
+
+* Add methods directly to Dataset instead of including modules (jeremyevans)
+
+* Make some Dataset instance methods private: invert_order, insert_default_values_sql (jeremyevans)
+
+* Don't add methods that depend on ParseTree unless you can load ParseTree (jeremyevans)
+
+* Don't wipeout the cached columns every time a dataset is cloned, but only on changes to :select, :sql, :from, or :join (jeremyevans)
+
+* Fix Oracle Adapter (yasushi.abe)
+
+* Fixed sqlite uri so that sqlite:// works just like file:// (2 slashes for a relative path, 3 for an absolute) (dlee)
+
+* Raise a Sequel::Error if an invalid limit or offset is used (jeremyevans)
+
+* Refactor and beef up Dataset#first and Dataset#last, with some change in functionality (jeremyevans)
+
+* Add String#to_datetime, for consistency (jeremyevans)
+
+* Fix Range#interval so that it returns 1 less for an exclusive range
+
+* Change SQLite adapter so it doesn't swallow exceptions other than SQLite3::Exception (such as Interrupt) (jeremyevans)
+
+* Change PostgreSQL and MySQL adapters to raise Sequel::Error instead of database specific errors if a database error occurs (jeremyevans)
+
+* Using a memory database with SQLite now defaults to a single connection, so all queries it uses run against the same database (jeremyevans)
+
+* Fix attempting to query MySQL using the same connection being used to concurrently execute another query (jeremyevans)
+
+* Add options to the connection pool to configure reusing connections and converting exceptions (jeremyevans)
+
+* Use the database driver provided string quoting methods for MySQL and SQLite (jeremyevans) (#223)
+
+* Add ColumnAll#==, for checking the equality of two ColumnAlls (jeremyevans)
+
+* Allow an array of arrays instead of a hash when specifying conditions (jeremyevans)
+
+* Add Sequel::DBI::Database#lowercase, for lowercasing column names (jamesearl)
+
+* Remove Dataset#extend_with_destroy, which may break code that uses Dataset#set_model directly and expects the destroy method to be added (jeremyevans)
+
+* Fix some issues when running on Ruby 1.9 (Zverok, jeremyevans)
+
+* Make the DBI adapter work (partially) with PostgreSQL (Seb)
+
 === 1.5.1 (2008-04-30)
 
-* Fix Dataset#eager_graph when not all objects have associated objects (jeremyevans)
+* Fix Dataset#eager_graph when not all objects have associated objects (jeremyevans)
+
+* Have Dataset#graph give a nil value instead of a hash with all nil values if no matching rows exist in the graphed table (jeremyevans)
+
+=== 1.5.0 (2008-04-29)
+
+* Make the validation errors API compatible with Merb (Inviz)
+
+* Add validates_uniqueness_of, for protecting against duplicate entries in the database (neaf, jeremyevans)
+
+* Alias Model#dataset= to Model#set_dataset (tmm1)
+
+* Make some Model class methods private: def_hook_method, hooks, add_hook, plugin_module, plugin_gem (jeremyevans)
+
+* Add the eager! and eager_graph! mutation methods to model datasets (jeremyevans)
+
+* Remove Model.database_opened (jeremyevans)
+
+* Remove Model.super_dataset (jeremyevans)
+
+* Deprecate .create_with_params, .create_with, #set, #update, #update_with, and #new_record from Sequel::Model (jeremyevans)
+
+* Add Model.def_dataset_method, for defining methods on the model that reference methods on the dataset (jeremyevans)
+
+* Deprecate Model.method_missing, add dataset methods to Model via metaprogramming (jeremyevans)
+
+* Remove Model.join, so it is the same as Dataset#join (jeremyevans)
+
+* Use reciprocal associations for all types of associations in the getter/setter/add_/remove_ methods (jeremyevans)
+
+* Fix many_to_one associations to cache negative lookups (jeremyevans)
+
+* Change Model#=== to always be false if the primary key is nil (jeremyevans)
+
+* Add Model#hash, which should be unique for a given class and primary key (or values if primary key is nil) (jeremyevans)
+
+* Add Model#eql? as a alias to Model#== (jeremyevans)
+
+* Make Model#reload clear any cached associations (jeremyevans)
+
+* No longer depend on the assistance gem, merge the Inflector and Validations code (jeremyevans)
+
+* Add Model#set_with_params, which is Model#update_with_params without the save (jeremyevans)
+  
+* Fix Model#destroy so that it returns self, not the result of after_destroy (jeremyevans)
+
+* Define Model column accessors in set_dataset, so they should always be avaiable, deprecate Model#method_missing (jeremyevans)
+
+* Add eager loading of associations via new sequel_core object graphing feature (jeremyevans)
+
+* Fix many_to_many associations with classes inside modules without an explicit join table (jeremyevans)
+
+* Allow creation of new records that don't have primary keys when the cache is on (jeremyevans) (#213)
+
+* Make Model#initialize, Model#set, and Model#update_with_params invulnerable to memory exhaustion (jeremyevans) (#210)
+
+* Add Model.str_columns, which gives a list of columns as frozen strings (jeremyevans)
+
+* Remove pretty_table.rb from sequel, since it is in sequel_core (jeremyevans)
+
+* Set a timeout in the Sqlite adapter, default to 5 seconds (hrvoje.marjanovic) (#218)
+
+* Document that calling Sequel::ODBC::Database#execute manually requires you to manually drop the returned object (jeremyevans) (#217)
+
+* Paginating an already paginated/limited dataset now raises an error (jeremyevans)
+
+* Add support for PostgreSQL partial indexes (dlee)
+
+* Added support for arbitrary index types (including spatial indexes) (dlee)
+
+* Quote column names in SQL generated for SQLite (tmm1)
+
+* Deprecate Object#rollback! (jeremyevans)
+
+* Make some Dataset methods private (qualified_column_name, column_list, table_ref, source_list) (jeremyevans)
+
+* Deprecate Dataset methods #set_options, #set_row_proc, #remove_row_proc, and #clone_merge (jeremyevans)
+
+* Add Symbol#*, a replacement for Symbol#all (jeremyevans)
+
+* Deprecate including ColumnMethods in Object, include it in Symbol, String, and Sequel::SQL::Expression (jeremyevans)
+
+* Deprecate Symbol#method_missing, and #AS, #DESC, #ASC, #ALL, and #all from ColumnMethods (jeremyevans)
+
+* Fix table joining in MySQL (jeremyevans)
+
+* Deprecate Sequel.method_missing and Object#Sequel, add real Sequel.adapter methods (jeremyevans)
+
+* Move dataset methods applicable only to paginated datasets into Sequel::Dataset::Pagination (jeremyevans)
+
+* Make Sequel::Dataset::Sequelizer methods private (jeremyevans)
+
+* Deprecate Dataset#method_missing, add real mutation methods (e.g. filter!) (jeremyevans)
+
+* Fix connecting to an MSSQL server via ODBC using domain user credentials (jeremyevans) (#216)
+
+* No longer depend on the assistance gem, merge in the ConnectionPool and .blank methods (jeremyevans)
+
+* No longer depend on ParseTree, RubyInline, or ruby2ruby, but you still need them if you want to use the block filters (jeremyevans)
+
+* Fix JDBC adapter by issuing index things start at 1 (pdamer)
+
+* Fix connecting to a database via the ADO adapter (now requires options instead of URI) (timuckun, jeremyevans) (#204)
+
+* Support storing microseconds in postgres timestamp fields (schnarch...@rootimage.msu.edu) (#215)
+
+* Allow joining of multiple datasets, by making the table alias different for each dataset joined (jeremyevans)
+
+* SECURITY: Fix backslash escaping of strings (dlee)
+
+* Add ability to create a graph of objects from a query, with the result split into corresponding tables (jeremyevans) (#113)
+
+* Add attr_accessor for dataset row_proc (jeremyevans)
+
+* Don't redefine Dataset#each when adding a transform or row_proc (jeremyevans)
+
+* Remove array_keys.rb from sequel_core, it was partially broken (since the arrays came from hashes), and redefined Dataset#each (jeremyevans)
+
+* Fix MySQL default values insert (matt.binary) (#196)
+
+* Fix ODBC adapter improperly escaping date and timestamp values (leo.borisenko) (#165)
+
+* Fix renaming columns on MySQL with type :varchar (jeremyevans) (#206)
+
+* Add Sequel::SQL::Function#==, for comparing SQL Functions (jeremyevans) (#209)
+
+* Update Informix adapter to work with Ruby/Informix 0.7.0 (gerardo.santana@gmail.com)
+
+* Remove sequel_core's knowledge of Sequel::Model (jeremyevans)
+
+* Use "\n" instead of $/ (since $/ can be redefined in ways we do not want) (jeremyevans)
+
+=== 1.4.0 (2008-04-08) 
+
+* Don't mark a column as changed unless the new value is different from the current value (tamas.denes, jeremyevans) (#203).
+
+* Switch gem name from "sequel_model" to just "sequel", which required large version bump (jeremyevans).
+
+* Add :select option to many_to_many associations, default to selecting only the associated model table and not the join table (jeremyevans) (#208).
+
+* Add :reciprocal one_to_many association option, for setting corresponding many_to_one instance variable (jeremyevans).
+
+* Add eager loading implementation (jeremyevans).
+
+* Change *_to_many associations so that the all associations are considered :cache=>true (jeremyevans).
+
+* Fix associations with block arguments and :cache=>true (jeremyevans).
+
+* Merge 3 mysql patches from the bugtracker (mvyver) (#200, #201, #202).
+
+* Merge 2 postgresql patches from the bugtracker (a...@mellowtone.co.jp) (#211, 212).
+
+* Allow overriding of default posgres spec database via ENV['SEQUEL_PG_SPEC_DB'] (jeremyevans).
+
+* Allow using the Sequel::Model as the first argument in a dataset join selection (jeremyevans) (#170).
+
+* Add simple callback mechanism to make model eager loading implementation easier (jeremyevans).
+
+* Added Sequel::Error::InvalidOperation class for invalid operations (#198).
+
+* Implemented MySQL::Database#server_version (#199).
+
+* Added spec configuration for MySQL socket file.
+
+* Fixed transform with array tuples in postgres adapter.
+
+* Changed spec configuration to Database objects instead of URIs in order to support custom options for spec databases.
+
+* Renamed schema files.
+
+* Fixed Dataset#from to work correctly with SQL functions (#193).
+
+===Previous to 1.4.0, Sequel model and Sequel core versioning differed, see the bottom of this file for the changelog to Sequel model prior to 1.4.0.
+
+=== 1.3 (2008-03-08)
+
+* Added configuration file for running specs (#186).
+
+* Changed Database#drop_index to accept fixed arity (#173).
+
+* Changed column definition sql to put UNSIGNED constraint before unique in order to satisfy MySQL (#171).
+
+* Enhanced MySQL adapter to support load data local infile_, added compress option for mysql connection by default (#172).
+
+* Fixed bug when inserting hashes in array tuples mode.
+
+* Changed SQLite adapter to catch RuntimeError raised when executing a statement and raise an Error::InvalidStatement with the offending SQL and error message (#188).
+
+* Added Error::InvalidStatement class.
+
+* Fixed Dataset#reverse to not raise for unordered dataset (#189).
+
+* Added Dataset#unordered method and changed #order to remove order if nil is specified (#190).
+
+* Fixed reversing order of ASC expression (#164).
+
+* Added support for :null => true option when defining table columns (#192).
+
+* Fixed Symbol#method_missing to accept variable arity (#185).
+
+=== 1.2.1 (2008-02-29)
+
+* Added add_constraint and drop_constraint functionality to Database#alter_table (#182).
+
+* Enhanced Dataset#multi_insert to accept datasets (#179).
+
+* Added MySQL::Database#use method for switching database (#180).
+
+* Enhanced Database.uri_to_options to accept uri strings (#178).
+
+* Added Dataset#columns! method that always makes a roundtrip to the DB (#177).
+
+* Added new Dataset#each_page method that iterates over all pages in the result set (#175).
+
+* Added Dataset#reverse alias to Dataset#reverse_order (#174).
+
+* Fixed Dataset#transform_load and #transform_save to create a trasnformed copy of the supplied hash instead of transforming it in place (#184).
+
+* Implemented MySQL::Dataset#replace (#163).
+
+=== 1.2 (2008-02-15)
+
+* Added support for :varchar[100] like type declarations in #create_table.
+
+* Fixed #rename_column in mysql adapter to support types like varchar(255) (#159).
+
+* Added support for order and limit in DELETE statement in MySQL adapter (#160).
+
+* Added checks to Dataset#multi_insert to prevent work if no values are given (#162).
+
+* Override ruby2ruby implementation of Proc#to_sexp which leaks memory (#161).
+
+* Added log option, help for sequel script (#157).
+
+=== 1.1 (2008-02-15)
+
+* Fixed Dataset#join_table to support joining of datasets (#156).
+
+* Changed Dataset#empty? to use EXISTS condition instead of counting records, for much better performance (#158).
+
+* Implemented insertion of multiple records in a single statement for postgres adapter. This feature is available only in postgres 8.2 and newer.
+
+* Implemented Postgres::Database#server_version.
+
+* Implemented Database#get, short for dataset.get(...).
+
+* Refactored Dataset#multi_insert, added #import alias, added support for calling #multi_insert using array of columns and array of value arrays (thanks David Lee).
+
+* Implemented Dataset#get, a replacement for select(column).first[column].
+
+* Implemented Dataset#grep method, poor man's text search.
+
+=== 1.0.10 (2008-02-13)
+
+* Fixed Datset#group_and_count to work inside a query block (#152).
+
+* Changed datasets with transforms to automatically transform hash filters (#155).
+
+* Changed Marshal stock transform to use Base64 encoding with backward-compatibility to support existing marshaled values (#154).
+
+* Added support for inserting multiple records in a single statement using #multi_insert in MySQL adapter (#153).
+
+* Added support for :slice option (same as :commit_every) in Dataset#multi_insert.
+
+* Changed Dataset#all to accept opts and iteration block.
+
+=== 1.0.9 (2008-02-10)
+
+* Implemented Dataset#inspect and Database#inspect (#151).
+
+* Added full-text searching for odbc_mssql adapter (thanks Joseph Love).
+
+* Added AlterTableGenerator#add_full_text_index method.
+
+* Implemented full_text indexing and searching for PostgreSQL adapter (thanks David Lee).
+
+* Implemented full_text indexing and searching for MySQL adapter (thanks David Lee).
+
+* Fixed Dataset#insert_sql to work with array subscript references (thanks Jim Morris).
+
+=== 1.0.8 (2008-02-08)
+
+* Added support for multiple choices in string matching expressions (#147).
+
+* Renamed Dataset#clone_merge to Dataset#clone, works with or without options for merging (#148).
+
+* Fixed MySQL::Database#<< method to always free the result in order to allow multiple calls in a row (#149). Same also for PostgreSQL adapter.
+
+=== 1.0.7 (2008-02-05)
+
+* Added support for conditional filters (using if else statements) inside block filters (thanks Kee).
+
+=== 1.0.6 (2008-02-05)
+
+* Removed code pollution introduced in revs 814, 817 (really bad patch, IMO).
+
+* Fixed joining datasets using aliased tables (#140).
+
+* Added support additional field types in postgresql adapter (#146).
+
+* Added support for date field types in postgresql adapter (#145).
+
+* Fixed Dataset#count to work correctly for grouped datasets (#144).
+
+* Added Dataset#select_more, Dataset#order_more methods (#129).
+
+=== 1.0.5 (2008-01-25)
+
+* Added support for instantiating models by using the load constructor method.
+
+=== 1.0.4.1 (2008-01-24)
+
+* Fixed bin/sequel to require sequel_model if available.
+
+=== 1.0.4 (2008-01-24)
+
+* Added Dataset#select_all method.
+
+* Changed ODBC::Database to support connection using driver and database name, also added support for untitled columns in ODBC::Dataset (thanks Leonid Borisenko).
+
+* Fixed MySQL adapter to correctly format foreign key definitions (#123).
+
+* Changed MySQL::Dataset to allow HAVING clause on ungrouped datasets, and put HAVING clause before ORDER BY clause (#133).
+
+* Changed Dataset#group_and_count to accept multiple columns (#134).
+
+* Fixed database spec to open YAML file in binary mode (#131).
+
+* Cleaned up gem spec (#132).
+
+* Added Dataset#table_exists? convenience method.
+
+=== 1.0.3 (2008-01-17)
+
+* Added support for UNSIGNED constraint, used in MySQL? (#127).
+
+* Implemented constraint definitions inside Database#create_table.
+
+* Fixed postgres adapter to define PGconn#async_exec as alias to #exec if not defined (for pure-ruby postgres driver).
+
+* Added String#to_date. Updated mysql adapter to use String#to_date for mysql date types (thanks drfreeze).
+
+=== 1.0.2 (2008-01-14)
+
+* Removed ConnectionPool, NumericExtensions. Added dependency on assistance.
+
+=== 1.0.1 (2008-01-12)
+
+* Changed postgres adapter to quote column references using double quotes.
+
+* Applied patch for oracle adapter: fix behavior of limit and offset, transactions, #table_exists?, #tables and additional specs (thanks Liming Lian #122).
+
+* Allow for additional filters on a grouped dataset (#119 and #120)
+
+* Changed mysql adapter to default to localhost if :host option is not specified (#114).
+
+* Refactored Sequelizer to use Proc#to_sexp (method provided by r2r).
+
+* Enhanced Database.connect to accept options with string keys, so it can now accept options loaded from YAML files. Database.connect also automatically converts :username option into :user for compatibility with existing YAML configuration files for AR and DataMapper.
+
+=== 1.0.0.1 (2008-01-03)
+
+* Changed MySQL adapter to support specifying socket option.
+
+* Added support for limiting and paginating datasets with fixed SQL, gotten with DB#fetch (thanks Ruy Diaz). 
+
+* Added new Dataset#from_self method that returns a dataset selecting from the original dataset.
+
+=== 1.0 (2008-01-02)
+
+* Removed deprecated adapter stubs.
+
+* Removed Sequel::Model() stub.
+
+* Changed name to sequel_core.
+
+* 100% code coverage.
+
+* Fixed error behavior when sequel_model is not available.
+
+* Fixed error behavior when parse_tree or ruby2ruby are not available.
+
+=== 0.5.0.2 (2008-01-01)
+
+* Fixed String#to_time to raise error correctly for invalid time stamps.
+
+* Improved code coverage - now at 99.2%.
+
+=== 0.5.0.1 (2007-12-31)
+
+* Added a stub for Sequel::Model that auto-loads sequel_model.
+
+* Changed Sequel.method_missing and Database.adapter_class to raise AdapterNotFound if an adapter could not be loaded.
+
+* Fixed behavior of error trap in sequel command line tool.
+
+=== 0.5 (2007-12-30)
+
+* Removed model code into separate sub-project. Rearranged trunk into core, model and model_plugins.
+
+=== 0.4.5 (2007-12-25)
+
+* Added rdoc for new alter_table functionality (#109).
+
+* Fixed update_sql with array sub-item keys (#110).
+
+* Refactored model specs.
+
+* Added Model#update as alias to #set.
+
+* Refactored validations code. Renamed Model.validations? into Model.has_validations?.
+
+* Added initial Model validations (Thanks Lance Carlson)
+
+* Added Database#set_column_default method (thanks Jim Morris.)
+
+* Removed warning on uninitialized @transform value (thanks Jim Morris).
+
+=== 0.4.4.2 (2007-12-20)
+
+* Fixed parsing errors in Ruby 1.9.
+
+* Fixed sync problem in connection_pool_spec.
+
+* Changed String#to_time to raise Error::InvalidValue if Time.parse fails.
+
+* Refactored sequel error classes.
+
+=== 0.4.4.1 (2007-12-19)
+
+* Fixed schema generation code to use field quoting and support adapter-specific literalization of default values (#108).
+
+=== 0.4.4 (2007-12-17)
+
+* Implemented Database#rename_table (#104).
+
+* Fixed drop_index in mysql adapter (#103).
+
+* Added ALTER TABLE specs for postgres, sqlite and mysql adapters. Added custom alter_table behavior for sqlite and mysql adapters (#101, #102).
+
+* Added direct Database API for altering tables.
+
+* Added Database#alter_table method with support for adding, dropping, renaming, modifying columns and adding and droppping indexes.
+
+* Added #unique schema method for defining unique indexes (thanks Dado).
+
+* Implemented unfolding of #each calls inside sequelizer blocks (thanks Jim Morris).
+
+=== 0.4.3 (2007-12-15)
+
+* Fixed Dataset#update to accept strings (#98).
+
+* Fixed Model.[] to raise for boolean argument (#97).
+
+* Added Database#add_index method (thanks coda.hale).
+
+* Added error reporting for filtering on comparison not in a block (thanks Jim Morris).
+
+* Added support for inline index definition (thanks Dado).
+
+* Added Database#create_table! method for forcibly creating a table (thanks Dado).
+
+* Added support for using Dataset#update with block.
+
+* Changed subscript access to use | operator.
+
+* Fixed subscript access in sequelizer.
+
+* Added support for subscript access using Symbol#/ operator.
+
+=== 0.4.2.2 (2007-12-10)
+
+* Improved code coverage.
+
+* Fixed Dataset#count to work properly with datasets with fixed SQL (when using #fetch).
+
+* Added Model.create_with_params method that filters the given parameters accordring to the model's columns (thanks Aman Gupta). 
+
+=== 0.4.2.1 (2007-12-09)
+
+* Refactored and fixed Dataset#reverse_order to work with field quoting (thanks Christian).
+
+* Fixed problem with field quoting in insert statements.
+
+* Changed sequelizer code to silently fail on any error when requiring parsetree and ruby2ruby.
+
+* Added Database#create_view, #create_or_replace_view and #drop_view methods. Also implemented Dataset#create_view and #create_or_replace_view convenience methods.
+
+* Keep DRY by re-using Model#[]= from method_missing.
+
+* Added Model.fetch alias for DB.fetch.set_model(Model)
+
+=== 0.4.2 (2007-12-07)
+
+* Implemented Model#save_changes.
+
+* Extended Model#save to accept specific columns to update.
+
+* Implemented experimental JDBC adapter.
+
+* Added adapter skeleton as starting point for new adapters.
+
+* Cleaned-up adapters and moved automatic requiring of 'sequel' to adapter stubs.
+
+=== 0.4.1.3 (2007-12-05)
+
+* Better plugin conventions.
+
+* Added experimental OpenBase adapter.
+
+* Fixed Sequel.<xxx> methods to accept options hash as well as database name. Fixed Sequel.connect to accept options hash as well as URI (Wayne).
+
+=== 0.4.1.2 (2007-12-04)
+
+* Added release rake task (using RubyForge).
+
+* Changed Model.is to accept variable arity.
+
+* Implemented plugin loading for model classes.
+
+* Fixed odbc-mssql and odbc adapters (thanks Dusty.)
+
+* Implemented odbc-mssql adapter (thanks Dusty.)
+
+=== 0.4.1.1 (2007-11-27)
+
+* Fixed #first and #last functionality in Informix::Dataset (thanks Gerardo Santana).
+
+=== 0.4.1 (2007-11-25)
+
+* Put adapter files in lib/sequel/adapters. Requiring sequel/<adapter> is now deprecated. Users can now just require 'sequel' and adapters are automagically loaded (#93).
+
+=== 0.4.0 (2007-11-24)
+
+* Reorganized lib directory structure.
+
+* Added support for dbi-xxx URI schemes (#86).
+
+* Fixed problem in Database#uri where setting the password would raise an error (#87).
+
+* Improved Dataset#insert_sql to correctly handle string keys (#92).
+
+* Improved error-handling for worker threads. Errors are saved to an array and are accessible through #errors (#91).
+
+* Dataset#uniq/distinct can now accept a column list for DISTINCT ON clauses.
+
+* Fixed Model.all.
+
+* Fixed literalization of strings with escape sequences in postgres adapter (#90).
+
+* Added support for literalizing BigDecimal values (#89).
+
+* Fixed column qualification for joined datasets (thanks Christian).
+
+* Implemented experimental informix adapter.
+
+=== 0.3.4.1 (2007-11-10)
+
+* Changed Dataset#select_sql to support queries without a FROM clause.
+
+=== 0.3.4 (2007-11-10)
+
+* Fixed MySQL adapter to allow calling stored procedures (thanks Sebastian).
+
+* Changed Dataset#each to always return self.
+
+* Fixed SQL functions without arguments in block filters.
+
+* Implemented super-cool Symbol#cast_as method.
+
+* Fixed error message in command-line tool if failed to load adapter (#85).
+
+* Refactored code relating to column references for better extendibility (#88).
+
+* Tiny fix to Model#run_hooks.
+
+=== 0.3.3 (2007-11-04)
+
+* Revised code to generate SQL statements without trailing semicolons.
+
+* Added Sequel::Worker implementation of a simple worker thread for asynchronous execution.
+
+* Added spec for Oracle adapter.
+
+* Fixed Oracle adapter to format INSERT statements without semicolons (thanks Liming Lian).
+
+* Renamed alias to Array#keys as Array#columns instead of Array#fields.
+
+* Renamed FieldCompositionMethods as ColumnCompositionMethods.
+
+* Implemented Sequel::NumericExtensions to provide stuff like 30.days.ago.
+
+=== 0.3.2 (2007-11-01)
+
+* Added #to_column_name as alias to #to_field_name, #column_title as alias to #field_title.
+
+* Added Dataset#interval method for getting interval between minimum/maximum values for a column.
+
+* Fixed Oracle::Database#execute (#84).
+
+* Added group_and_count as general implementation for count_by_xxx.
+
+* Added count_by magic method.
+
+* Added Dataset#range method for getting the minimum/maximum values for a column.
+
+* Fixed timestamp translation in SQLite adapter (#83).
+
+* Experimental DB2 adapter.
+
+* Added Dataset#set as alias to Dataset#update.
+
+* Removed long deprecated expressions.rb code.
+
+* Better documentation.
+
+* Implemented Dataset magic methods: order_by_xxx, group_by_xxx, filter_by_xxx, all_by_xxx, first_by_xxx, last_by_xxx.
+
+* Changed Model.create and Model.new to accept a block.
+
+=== 0.3.1 (2007-10-30)
+
+* Typo fixes (#79).
+
+* Added require 'yaml' to dataset.rb (#78).
+
+* Changed postgres adapter to use the ruby-postgres library's type conversion if available (#76).
+
+* Fixed string literalization in mysql adapter for strings with comment backslashes in them (#75).
+
+* Fixed ParseTree dependency to work with version 2.0.0 and later (#74).
+
+* foreign_key definitions now accept :key option for specifying the remote key (#73).
+
+* Fixed Model#method_missing to not raise error for columns not in the table but for which a value exists (#77).
+
+* New documentation for Model.
+
+* Implemented Oracle adapter based on ruby-oci8 library.
+
+* Implemented Model#pk_hash. Is it really necessary?
+
+* Deprecated Model#pkey. Implemented better Model#pk method.
+
+* Specs and docs for Model.one_to_one, Model.one_to_many macros.
+
+=== 0.3.0.1 (2007-10-20)
+
+* Changed Database#fetch to return a modified dataset.
+
+=== 0.3 (2007-10-20)
+
+* Added stock transforms to Dataset#transform. Refactored Model.serialize.
+
+* Added Database#logger= method for setting the database logger object.
+
+* Fixed Model.[] to act as shortcut to Model.find when a hash is given (#71).
+
+* Added support for old and new decimal types in MySQL adapter, and updated MYSQL_TYPES with MySQL 5.0 constants (#72).
+
+* Implemented Database#disconnect method for all adapters.
+
+* Fixed small bug in ArrayKeys module.
+
+* Implemented model caching by primary key.
+
+* Separated Model.find and Model.[] functionality. Model.find takes a filter. Model.[] is strictly for finding by primary keys.
+
+* Enhanced Dataset#first to accept a filter block. Model#find can also now accept a filter block.
+
+* Changed Database#[] to act as shortcut to #fetch if a string is given.
+
+* Renamed Database#each to #fetch. If no block is given, the method returns an enumerator.
+
+* Changed Dataset#join methods to correctly literalize values in join conditions (#70).
+
+* Fixed #filter with ranges to correctly literalize field names (#69).
+
+* Implemented Database#each method for quickly retrieving records with arbitrary SQL (thanks Aman Gupta).
+
+* Fixed bug in postgres adapter where a LiteralString would be literalized as a regular String.
+
+* Fixed SQLite insert with subquery (#68).
+
+* Reverted back to hashes as default mode. Added Sequel.use_array_tuples and Sequel.use_hash_tuples methods.
+
+* Fixed problem with arrays with keys when using #delete.
+
+* Implemented ArrayKeys as substitute for ArrayFields.
+
+* Added Dataset#each_hash method.
+
+* Rewrote SQLite::Database#transaction to use sqlite3-ruby library implementation of transactions.
+
+* Fixed Model.destroy_all to work correctly in cases where no before_destroy hook is defined and an after_destroy hook is defined.
+
+* Restored Model.has_hooks? implementation.
+
+* Changed Database#<< to strip comments and whitespace only when an array is given.
+
+* Changed Schema::Generator#primary_key to accept calls with the type argument omitted.
+
+* Hooks can now be prepended or appended by choice.
+
+* Changed Model.subset to define filter method on the underlying dataset instead of the model class.
+
+* Fixed Dataset#transform to work with array fields.
+
+* Added Dataset#to_csv method.
+
+* PrettyTable can now extract column names from arrayfields.
+
+* Converted ado, dbi, odbc adapters to use arrayfields instead of hashes.
+
+* Fixed composite key support.
+
+* Fixed Dataset#insert_sql, update_sql to support array fields.
+
+* Converted sqlite, mysql, postgres adapters to use arrayfields instead of hashes.
+
+* Extended Dataset#from to auto alias sub-queries.
+
+* Extended Dataset#from to accept hash for aliasing tables.
+
+* Added before_update, after_update hooks.
+
+=== 0.2.1.1 (2007-10-07)
+
+* Added Date literalization to sqlite adapter (#60).
+
+* Changed Model.serialize to allow calling it after the class is defined (#59).
+
+* Fixed after_create hooks to allow calling save inside the hook (#58).
+
+* Fixed MySQL quoting of sql functions (#57).
+
+* Implemented rollback! global method for cancelling transactions in progress.
+
+* Fixed =~ operator in Sequelizer.
+
+* Fixed ODBC::Dataset#fetch_rows (thanks Dusty).
+
+* Renamed Model.recreate_table to create_table!. recreate_table is deprecated and will issue a warning (#56).
+
+=== 0.2.1 (2007-09-24)
+
+* Added default implementation of Model.primary_key_hash.
+
+* Fixed Sequel::Model() to set dataset for inherited classes.
+
+* Rewrote Model.serialize to use Dataset#transform.
+
+* Implemented Dataset#transform.
+
+* Added gem spec for Windows (without ParseTree dependency).
+
+* Added support for dynamic strings in Sequelizer (#49).
+
+* Query branch merged into trunk.
+
+* Implemented self-changing methods.
+
+* Add support for ternary operator to Sequelizer.
+
+* Fixed sequelizer to evaluate expressions if they don't involve symbols or literal strings.
+
+* Added protection against using #each, #delete, #insert, #update inside query blocks.
+
+* Improved Model#method_missing to deal with invalid attributes.
+
+* Implemented Dataset#query.
+
+* Added Dataset#group_by as alias for Dataset#group.
+
+* Added Dataset#order_by as alias for Dataset#order.
+
+* More model refactoring. Added support for composite keys.
+
+* Added Dataset#empty? method (#46).
+
+* Fixed Symbol#to_field_name to support names with numbers and upper-case characters (#45).
+
+* Added install_no_doc rake task.
+
+* Partial refactoring of model code.
+
+* Refactored dataset-model association and added Dataset#set_row_filter method.
+
+* Added support for case-sensitive regexps to mysql adapter.
+
+* Changed mysql adapter to support encoding option as well.
+
+* Added charset/encoding option to postgres adapter.
+
+* Implemented Model.serialize (thanks Aman Gupta.)
+
+* Changed Model.create to INSERT DEFAULT VALUES instead of (id) VALUES (null) (brings back #41.)
+
+* Fixed Model.new to work without arguments.
+
+* Added Model.no_primary_key method to allow models without primary keys.
+
+* Added Model#this method (#42 thanks Duane Johnson).
+
+* Fixed Dataset#insert_sql to use DEFAULT VALUES clause if argument is an empty hash.
+
+* Fixed Model.create to work correctly when no argument is passed (#41).
+
+=== 0.2.0.2 (2007-09-07)
+
+* Dataset#insert can now accept subqueries.
+
+* Changed Migrator.apply to return the version.
+
+* Changed Sequel::Model() to cache intermediate classes so descendant classes can be reopened (#39).
+
+* Added :charset option to MySQL adapter (#40).
+
+* Fixed Dataset#exclude to add parens around NOT expression (#38).
+
+* Fixed use of sub-queries with all comparison operators in block filters (#38).
+
+* Fixed arithmetic expressions in block filters to not be literalized.
+
+* Changed Symbol#method_missing to return LiteralString.
+
+* Changed PrettyTable to right-align numbers.
+
+* Fixed Model.create_table (thanks Duane Johnson.)
+
+=== 0.2.0.1 (2007-09-04)
+
+* Improved support for invoking methods with inline procs inside block filters.
+
+=== 0.2.0 (2007-09-02)
+
+* Fixed Model.drop_table (thanks Duane Johnson.)
+
+* Dataset#each can now return rows for arbitrary SQL by specifying :sql option.
+
+* Added spec for postgres adapter.
+
+* Fixed Model.method_missing to work with new SQL generation.
+
+* Fixed #compare_expr to support regexps.
+
+* Fixed postgres, mysql adapters to support regexps.
+
+* More specs for block filters. Updated README.
+
+* Added support for globals and $X macros in block filters.
+
+* Fixed Sequelizer to not fail if ParseTree or Ruby2Ruby gems are missing.
+
+* Renamed String#expr into String#lit (#expr should be deprecated in future versions).
+
+* Renamed Sequel::ExpressionString into LiteralString.
 
-=== 1.5.0 (2008-04-29)
+* Fixed Symbol#[] to return an ExpressionString, so as not to be literalized.
 
-* Make the validation errors API compatible with Merb (Inviz)
+* Renamed Dataset::Expressions to Dataset::Sequelizer.
 
-* Add validates_uniqueness_of, for protecting against duplicate entries in the database (neaf, jeremyevans)
+* Renamed Expressions#format_re_expression to match_expr.
 
-* Alias Model#dataset= to Model#set_dataset (tmm1)
+* Renamed Expressions#format_eq_expression to compare_expr.
 
-* Make some Model class methods private: def_hook_method, hooks, add_hook, plugin_module, plugin_gem (jeremyevans)
+* Added support for Regexp in MySQL adapter.
 
-* Add the eager! and eager_graph! mutation methods to model datasets (jeremyevans)
+* Refactored Regexp expressions into a separate #format_re_expression method.
 
-* Remove Model.database_opened (jeremyevans)
+* Added support for arithmetic in proc filters.
 
-* Remove Model.super_dataset (jeremyevans)
+* Added support for nested proc expressions, more specs.
 
-* Deprecate .create_with_params, .create_with, #set, #update, #update_with, and #new_record from Sequel::Model (jeremyevans)
+* Added support for SQL function using symbols, e.g. :sum[:x].
 
-* Add Model.def_dataset_method, for defining methods on the model that reference methods on the dataset (jeremyevans)
+* Fixed deadlock bug in ConnectionPool.
 
-* Deprecate Model.method_missing, add dataset methods to Model via metaprogramming (jeremyevans)
+* Removed deprecated old expressions.rb.
 
-* Remove Model.join, so it is the same as Dataset#join (jeremyevans)
+* Rewrote Proc filter feature using ParseTree.
 
-* Use reciprocal associations for all types of associations in the getter/setter/add_/remove_ methods (jeremyevans)
+* Added support for additional functions on columns using Symbol#method_missing.
 
-* Fix many_to_one associations to cache negative lookups (jeremyevans)
+* Added support for supplying filter block to DB#[] method, to allow stuff like DB[:nodes] {:path =~ /^icex1/}.
 
-* Change Model#=== to always be false if the primary key is nil (jeremyevans)
+=== 0.1.9.12 (2007-08-26)
 
-* Add Model#hash, which should be unique for a given class and primary key (or values if primary key is nil) (jeremyevans)
+* Added spec for PrettyTable.
 
-* Add Model#eql? as a alias to Model#== (jeremyevans)
+* Added specs for Schema::Generator and Model (#36 thanks technoweenie).
 
-* Make Model#reload clear any cached associations (jeremyevans)
+* Fixed Sequel::Model.set_schema (#36 thanks technoweenie.)
 
-* No longer depend on the assistance gem, merge the Inflector and Validations code (jeremyevans)
+* Added support for no options on Schema::Generator#foreign_key (#36 thanks technoweenie.)
 
-* Add Model#set_with_params, which is Model#update_with_params without the save (jeremyevans)
+* Implemented (restored?) Schema::Generator#primary_key_name (#36 thanks technoweenie.)
 
-* Fix Model#destroy so that it returns self, not the result of after_destroy (jeremyevans)
+* Better spec code coverage.
 
-* Define Model column accessors in set_dataset, so they should always be avaiable, deprecate Model#method_missing (jeremyevans)
+=== 0.1.9.11 (2007-08-24)
 
-* Add eager loading of associations via new sequel_core object graphing feature (jeremyevans)
+* Changed Dataset#set_model to allow supplying additional arguments to the model's initialize method (#35). Thanks Sunny Hirai.
 
-* Fix many_to_many associations with classes inside modules without an explicit join table (jeremyevans)
+=== 0.1.9.10 (2007-08-22)
 
-* Allow creation of new records that don't have primary keys when the cache is on (jeremyevans) (#213)
+* Changed schema generation code to generate separate statements for CREATE TABLE and each CREATE INDEX (#34).
 
-* Make Model#initialize, Model#set, and Model#update_with_params invulnerable to memory exhaustion (jeremyevans) (#210)
+* Refactored Dataset::SQL#field_name for better support of different field quoting standards by specific adapters.
 
-* Add Model.str_columns, which gives a list of columns as frozen strings (jeremyevans)
+* Added #current_page_record_count for paginated datasets.
 
-* Remove pretty_table.rb from sequel, since it is in sequel_core (jeremyevans)
+* Removed Database#literal and included Dataset::SQL instead.
 
-=== 1.4.0 (2008-04-08)
+* Sequel::Dataset:SQL#field_name can now take a hash (as well as #select and any method that uses #field_name) for aliasing column names.  E.g. DB[:test].select(:_qqa => 'Date').sql #=> 'SELECT _qqa AS Date FROM test'.
 
-* Don't mark a column as changed unless the new value is different from the current value (tamas.denes, jeremyevans) (#203).
+* Moved SingleThreadedPool to lib/sequel/connection_pool.rb.
 
-* Switch gem name from "sequel_model" to just "sequel", which required large version bump (jeremyevans).
+* Changed SQLite::Dataset to return affected rows for #delete and #update (#33).
 
-* Add :select option to many_to_many associations, default to selecting only the associated model table and not the join table (jeremyevans) (#208).
+* ADO adapter: Added use of Enumerable for Recordset#Fields, playing it safe and moving to the first row before getting results, and changing the auto_increment constant to work for MSSQL.
 
-* Add :reciprocal one_to_many association option, for setting corresponding many_to_one instance variable (jeremyevans).
+=== 0.1.9.9 (2007-08-18)
 
-* Add eager loading implementation (jeremyevans).
+* New ADO adapter by cdcarter (#31).
 
-* Change *_to_many associations so that the all associations are considered :cache=>true (jeremyevans).
+* Added automatic column aliasing to #avg, #sum, #min and #max (#30).
 
-* Fix associations with block arguments and :cache=>true (jeremyevans).
+* Fixed broken Sequel::DBI::Dataset#fetch_rows (#29 thanks cdcarter.)
+
+=== 0.1.9.8 (2007-08-15)
+
+* Fixed DBI adapter.
+
+=== 0.1.9.7 (2007-08-15)
+
+* Added support for executing batch statements in sqlite adapter.
+
+* Changed #current_page_record_range to return 0..0 for an invalid page.
+
+* Fixed joining of aliased tables.
+
+* Improved Symbol#to_field_name to prevent false positives.
+
+* Implemented Dataset#multi_insert with :commit_every option.
+
+* More docs for Dataset#set_model.
+
+* Implemented automatic creation of convenience methods for each adapter (e.g. Sequel.sqlite etc.)
+
+=== 0.1.9.6 (2007-08-13)
+
+* Refactored schema definition code. Gets rid of famous primary_key problem as well as other issues (e.g. issue #22).
+
+* Added #pagination_record_count, #page_range and #current_page_record_range for paginated datasets.
+
+* Changed MySQL adapter to automatically reconnect (issue #26).
+
+* Changed Sequel() to accept variable arity.
+
+* Added :elements option to column definition, in order to support ENUM and SET types.
+
+=== 0.1.9.5 (2007-08-12)
+
+* Fixed migration docs.
+
+* Removed dependency on PGconn in Schema class.
+
+=== 0.1.9.4 (2007-08-11)
+
+* Added Sequel.dbi convenience method for using DBI connection strings to open DBI databases.
+
+=== 0.1.9.3 (2007-08-10)
+
+* Added support for specifying field size in schema definitions (thanks Florian Aßmann.)
+
+* Added migration code based on work by Florian Aßmann.
+
+* Reintroduced metaid dependency. No need to keep a local copy of it.
+
+=== 0.1.9.2 (2007-07-24)
+
+* Removed metaid dependency. Re-factored requires in lib/sequel.rb.
+
+=== 0.1.9.1 (2007-07-22)
+
+* Improved robustness of MySQL::Dataset#field_name.
+
+* Added Sequel.single_threaded= convenience method.
+
+=== 0.1.9 (2007-07-21)
+
+* Fixed #update_sql and #insert_sql to support field quoting by calling #field_name.
+
+* Implemented automatic data type conversion in mysql adapter.
+
+* Added support for boolean literals in mysql adapter.
+
+* Added support for ORDER and LIMIT clauses in UPDATE statements in mysql adapter.
+
+* Implemented correct field quoting (using back-ticks) in mysql adapter.
+
+* Wrote basic MySQL spec.
+
+* Fixd MySQL::Dataset to return correct data types with symbols as hash keys.
+
+* Removed discunctional MySQL::Database#transaction.
+
+* Added support for single threaded operation.
+
+* Fixed bug in Dataset#format_eq_expression where Range objects would not be literalized correctly.
+
+* Added parens around postgres LIKE expressions using regexps.
+
+=== 0.1.8 (2007-07-10)
+
+* Implemented Dataset#columns for retrieving the columns in the result set.
+
+* Updated Model with changes to how model-associated datasets work.
+
+* Beefed-up specs. Coverage is now at 95.0%.
+
+* Added support for polymorphic datasets.
+
+* The adapter dataset interface was simplified and standardized. Only four methods need be overriden: fetch_rows, update, insert and delete.
+
+* The Dataset class was refactored. The bulk of the dataset code was moved into separate modules.
+
+* Renamed Dataset#hash_column to Dataset#to_hash.
+
+* Added some common pragmas to sqlite adapter.
+
+* Added Postgres::Dataset#analyze for EXPLAIN ANALYZE queries.
+
+* Fixed broken Postgres::Dataset#explain.
+
+=== 0.1.7
+
+* Removed db.synchronize wrapping calls in sqlite adapter.
+
+* Implemented Model.join method to restrict returned columns to the model table (thanks Pedro Gutierrez).
+
+* Implemented Dataset#paginate method.
+
+* Fixed after_destroy hook.
+
+* Improved Dataset#first and #last to accept a filter hash.
+
+* Added Dataset#[]= method.
+
+* Added Sequel() convenience method.
+
+* Fixed Dataset#first to include a LIMIT clause for a single record.
+
+* Small fix to Postgres driver to return a primary_key value for the inserted record if it is specified in the insertion values (thanks Florian Aßmann and Pedro Gutierrez).
+
+* Fixed Symbol#DESC to support qualified notation (thanks Pedro Gutierrez).
+
+=== 0.1.6
+
+* Fixed Model#method_missing to raise for an invalid attribute.
+
+* Fixed PrettyTable to print model objects (thanks snok.)
+
+* Fixed ODBC timestamp conversion to return DateTime rather than Time object (thanks snok.)
+
+* Fixed Model.method_missing (thanks snok.)
+
+* Model.method_missing now creates stubs for calling Model.dataset methods. Methods like Model.each etc are removed.
+
+* Changed default join type to INNER JOIN (thanks snok.)
+
+* Added support for literal expressions, e.g. DB[:items].filter(:col1 => 'col2 - 10'.expr).
+
+* Added Dataset#and.
+
+* SQLite adapter opens a memory DB if no database is specified, e.g. Sequel.open 'sqlite:/'.
+
+* Added Dataset#or, pretty nifty.
+
+=== 0.1.5
+
+* Fixed Dataset#join to support multiple joins. Added #left_outer_join, #right_outer_join, #full_outer_join, #inner_join methods.
+
+=== 0.1.4
+
+* Added String#split_sql.
+
+* Implemented Array#to_sql and String#to_sql. Database#to_sql can now take an array of strings and convert into an SQL string. Comments and excessive white-space are removed.
+
+* Improved Schema generator to support data types as method names:
+  DB.create_table :test do
+    integer :abc
+    text :def
+    ...
+  end
+
+* Implemented ODBC adapter.
+
+=== 0.1.3
+
+* Implemented DBI adapter.
+
+* Refactored database connection code. Now handled through Database#connect.
+
+=== 0.1.2
+
+* The first opened database is automatically assigned to to Model.db.
+
+* Removed SequelConnectionError. Exception class errors are converted to RuntimeError.
+
+* Added support for UNION, INTERSECT and EXCEPT set operations.
+
+* Fixed Dataset#single_record to return nil if no record is found.
+
+* Updated specs to conform to RSpec 1.0.
+
+* Added Model#find_or_create method.
+
+* Fixed MySQL::Dataset#query_single (thanks Dries Harnie.)
+
+* Added Model.subset method. Fixed Model.filter and Model.exclude to accept blocks.
+
+* Added Database#uri method.
+
+* Refactored and removed deprecated code in postgres adapter.
+
+===0.1.1
+
+* More documentation for Dataset.
+
+* Added Dataset#size as alias to Dataset#count.
+
+* Changed Database#<< to call execute (instead of being an alias). Thus it will work for descendants as well.
+
+* Fixed Sequel.open to accept variable arity.
+
+* Refactored Model#refresh, Model.create. Removed Model#reload.
+
+* Refactored Model hooks.
+
+* Cleaned up Dataset API.
+
+=== 0.1.0
+
+* Changed Database#create_table to only accept a block. Nobody's gonna use the other way.
+
+* Removed Dataset#[]= method. Too confusing and not really useful.
+
+* Fixed ConnectionPool#hold to wrap exceptions only once.
+
+* Dataset#where_list Renamed Dataset#expression_list.
+
+* Added support for qualified fields in Proc expressions (e.g. filter {items.id == 1}.)
+
+* Added like? and in? Proc expression operators.
+
+* Added require 'date' in dataset.rb. Is this a 1.8.5 thing?
+
+* Refactored Dataset to use literal strings instead of format strings (slight performance improvement and better readability.)
+
+* Added support for literalizing Date objects.
+
+* Refactored literalization of Time objects.
+
+=== 0.0.20
+
+* Refactored Dataset where clause construction to use expressions.
+
+* Implemented Proc expressions (adapted from a great idea by Sam Smoot.)
+
+* Fixed Model#map.
+
+* Documentation for ConnectionPool.
+
+* Specs for Database.
+
+=== 0.0.19
+
+* More specs for Dataset.
+
+* Fixed Dataset#invert_order to work correctly with strings.
+
+* Fixed Model#== to check equality of values.
+
+* Added Model#exclude and Model#order.
+
+* Fixed Dataset#order and Dataset#group to behave correctly when supplied with qualified field name symbols.
+
+* Removed Database#literal. Shouldn't have been there.
+
+* Added SQLite::Dataset#explain. Returns an array of opcode hashes.
+
+* Specs for ConnectionPool.
+
+=== 0.0.18
+
+* Implemented SequelError and SequelConnectionError classes. ConnectionPool#hold now catches any connection errors and reraises them SequelConnectionError.
+
+* Removed duplication in Database#[].
+
+* :from and :select options are now always arrays (patch by Alex Bradbury.)
+
+* Fixed Dataset#exclude to work correctly (patch and specs by Alex Bradbury.)
+
+=== 0.0.17
+
+* Fixed Postgres::Database#tables to return table names as symbols (caused problem when using Database#table_exists?).
+
+* Fixed Dataset#from to have variable arity, like Dataset#select and Dataset#where (patch by Alex Bradbury.)
+
+* Added support for GROUP BY and HAVING clauses (patches by Alex Bradbury.) Refactored Dataset#filter.
+
+* More specs.
+
+* Refactored Dataset#where for better composability.
+
+* Added Dataset#[]= method.
+
+* Added support for DISTINCT and OFFSET clauses (patches by Alex Bradbury.) Dataset#limit now accepts ranges. Added Dataset#uniq and distinct methods.
+
+=== 0.0.16
+
+* More documentation.
+
+* Added support for subqueries in Dataset#literal.
+
+* Added support for Model.all_by_XXX methods through Model.method_missing.
+
+* Added basic SQL logging to Database.
+
+* Added Enumerable#send_each convenience method.
+
+* Changed Dataset#destroy to return the number of deleted records.
+
+=== 0.0.15
+
+* Improved Dataset#insert_sql to allow arrays as well as hashes.
+
+* Database#drop_table now accepts a list of table names.
+
+* Added Model#id to to return the id column.
+
+=== 0.0.14
+
+* Fixed Model's attribute accessors (hopefully for the last time).
+
+* Changed Model.db and Model.db= to allow different databases for different model classes.
+
+* Fixed bug in aggregate methods (max, min, etc) for datasets using record classes.
+
+=== 0.0.13
+
+* Fixed Model#method_missing to do both find, filter and attribute accessors. duh.
+
+* Fixed bug in Dataset#literal when quoting arrays of strings (thanks Douglas Koszerek.)
+
+=== 0.0.12
+
+* Model#save now correctly performs an INSERT for new objects.
+
+* Added Model#reload for reloading an object from the database.
+
+* Added Dataset#naked method for getting a version of a dataset that fetches records as hashes.
+
+* Implemented attribute accessors for column values ala ActiveRecord models.
+
+* Fixed filtering using nil values (e.g. dataset.filter(:parent_id => nil)).
+
+=== 0.0.11
+
+* Renamed Model.schema to Model.set_schema and Model.get_schema to Model.schema.
+
+* Improved Model class to allow descendants of model clases (thanks Pedro Gutierrez.)
+
+* Removed require 'postgres' in schema.rb (thanks Douglas Koszerek.)
+
+=== 0.0.10
+
+* Added some examples.
+
+* Added Dataset#print method for pretty-printing tables.
+
+=== 0.0.9
+
+* Fixed Postgres::Database#tables and #locks methods.
+
+* Added PGconn#last_insert_id method that should support all 7.x and 8.x versions of Postgresql.
+
+* Added Dataset#exists method for EXISTS where clauses.
+
+* Changed behavior of Dataset#literal to regard symbols as field names.
+
+* Refactored and DRY'd Dataset#literal and overrides therof. Added support for subqueries in where clause.
+
+=== 0.0.8
+
+* Fixed Dataset#reverse_order to provide chainability. This method can be called without arguments to invert the current order or with arguments to provide a descending order.
+
+* Fixed literal representation of literals in SQLite adapter (thanks Christian Neukirchen!) 
+
+* Refactored insert code in Postgres adapter (in preparation for fetching the last insert id for pre-8.1 versions).
+
+=== 0.0.7
+
+* Fixed bug in Model.schema, duh!
+
+=== 0.0.6
+
+* Added Dataset#sql as alias to Dataset#select_sql.
+
+* Dataset#where and Dataset#exclude can now be used for refining dataset conditions, enabling stuff like posts.where(:title => 'abcdef').exclude(:user_id => 3).
+
+* Implemented Dataset#exclude method.
+
+* Added Sequel::Schema#auto_primary_key method for setting an automatic primary key to be added to every table definition. Changed the schema generator to not define a primary key by default.
+
+* Changed Sequel::Database#table_exists? to rely on the tables method if it is available.
+
+* Implemented SQLite::Database#tables.
+
+=== 0.0.5
+
+* Added Dataset#[] method. Refactored Model#find and Model#[].
+
+* Renamed Pool#conn_maker to Pool#connection_proc.
+
+* Added automatic require 'sequel' to all adapters for convenience.
+
+=== 0.0.4
+
+* Added preliminary MySQL support.
+
+* Code cleanup.
+
+=== 0.0.3
+
+* Add Dataset#sum method.
+
+* Added support for exclusive ranges (thanks Christian Neukirchen.)
+
+* Added sequel console for quick'n'dirty access to databases.
+
+* Fixed small bug in Dataset#qualified_field_name for better join support.
+
+=== 0.0.2
+
+* Added Sequel.open as alias to Sequel.connect.
+
+* Refactored Dataset#where_equal_condition into Dataset#where_condition, allowing arrays and ranges, e.g. posts.filter(:stamp => (3.days.ago)..(1.day.ago)), or posts.filter(:category => ['ruby', 'postgres', 'linux']).
+
+* Added Model#[]= method for changing column values and Model#save 
+method for saving them.
+
+* Added Dataset#destroy for deleting each record individually as support for models. Renamed Model#delete to Model#destroy (and Model#destroy_all) ala ActiveRecord.
+
+* Refactored Dataset#first and Dataset#last code. These methods can now accept the number of records to fetch.
+
+=== 0.0.1
+
+* More documentation for Dataset.
+
+* Renamed Database#query to Database#dataset.
+
+* Added Dataset#insert_multiple for inserting multiple records.
+
+* Added Dataset#<< as shorthand for inserting records.
+
+* Added Database#<< method for executing arbitrary SQL.
+
+* Imported Sequel code.
+
+
+== Sequel::Model CHANGELOG 0.1 - 0.5.0.2
 
 === 0.5.0.2 (2008-03-12)
 
diff -pruN 1.5.1-1/COPYING 3.0.0-1/COPYING
--- 1.5.1-1/COPYING	2008-04-16 19:35:51.000000000 +0100
+++ 3.0.0-1/COPYING	2009-01-09 17:44:43.000000000 +0000
@@ -1,5 +1,5 @@
 Copyright (c) 2007-2008 Sharon Rosner
-Copyright (c) 2008 Jeremy Evans
+Copyright (c) 2008-2009 Jeremy Evans
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to
diff -pruN 1.5.1-1/debian/changelog 3.0.0-1/debian/changelog
--- 1.5.1-1/debian/changelog	2009-05-20 19:31:06.000000000 +0100
+++ 3.0.0-1/debian/changelog	2009-05-20 19:20:27.000000000 +0100
@@ -1,3 +1,14 @@
+libsequel-ruby (3.0.0-1) unstable; urgency=low
+
+  * New upstream release.
+  * Depend on rubygems (Closes: #529583).
+  * Conflict with libsequel-core-ruby.
+  * Bumped up standards revision and debhelper compatibility level.
+  * Changed section to "ruby".
+  * Added watch file.
+
+ -- Sebastien Delafond <seb@debian.org>  Wed, 20 May 2009 02:24:59 -0700
+
 libsequel-ruby (1.5.1-1) unstable; urgency=low
 
   * Initial Release (Closes: #478783)
diff -pruN 1.5.1-1/debian/compat 3.0.0-1/debian/compat
--- 1.5.1-1/debian/compat	2009-05-20 19:31:06.000000000 +0100
+++ 3.0.0-1/debian/compat	2009-05-20 19:20:27.000000000 +0100
@@ -1 +1 @@
-4
+6
diff -pruN 1.5.1-1/debian/control 3.0.0-1/debian/control
--- 1.5.1-1/debian/control	2009-05-20 19:31:06.000000000 +0100
+++ 3.0.0-1/debian/control	2009-05-20 19:20:27.000000000 +0100
@@ -1,14 +1,15 @@
 Source: libsequel-ruby
-Section: interpreters
+Section: ruby
 Priority: optional
 Maintainer: Sebastien Delafond <seb@debian.org>
-Build-Depends: debhelper (>> 3.0.0)
-Standards-Version: 3.7.3
+Build-Depends: debhelper (>> 6.0.0)
+Standards-Version: 3.8.1
 
 Package: libsequel-ruby1.8
 Architecture: all
-Depends: libruby1.8, libsequel-core-ruby1.8
-Description: Lightweight database access toolkit for Ruby
+Depends: ruby, ruby1.8, rubygems
+Conflicts: libsequel-core-ruby1.8, libsequel-core-ruby
+Description: Lightweight database access toolkit for Ruby 1.8
  Sequel is a lightweight database access toolkit for Ruby. Sequel
  provides thread safety, connection pooling and a concise DSL for
  constructing database queries and table schemas. Sequel also includes
diff -pruN 1.5.1-1/debian/dirs 3.0.0-1/debian/dirs
--- 1.5.1-1/debian/dirs	2009-05-20 19:31:06.000000000 +0100
+++ 3.0.0-1/debian/dirs	2009-05-20 19:20:27.000000000 +0100
@@ -1 +1,2 @@
 usr/lib/ruby/1.8
+usr/bin
\ No newline at end of file
diff -pruN 1.5.1-1/debian/rules 3.0.0-1/debian/rules
--- 1.5.1-1/debian/rules	2009-05-20 19:31:06.000000000 +0100
+++ 3.0.0-1/debian/rules	2009-05-20 19:20:27.000000000 +0100
@@ -21,6 +21,7 @@ install: build
 	dh_clean -k
 	dh_installdirs
 	cp -r lib/* $(DESTDIR)/usr/lib/ruby/1.8/
+	cp -r bin/* $(DESTDIR)/usr/bin/
 
 # Build architecture-independent files here.
 binary-arch: install
@@ -32,7 +33,7 @@ binary-indep: install
 	dh_testroot
 #	dh_installdebconf	
 	dh_installdocs
-	dh_installdocs -p libsequel-ruby README
+	dh_installdocs -p libsequel-ruby README.rdoc
 #	dh_installmenu
 #	dh_installlogrotate
 #	dh_installemacsen
diff -pruN 1.5.1-1/debian/watch 3.0.0-1/debian/watch
--- 1.5.1-1/debian/watch	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/debian/watch	2009-05-20 19:20:27.000000000 +0100
@@ -0,0 +1,4 @@
+version=3
+
+http://rubyforge.org/frs/?group_id=3168&release_id=34124 \
+  .*sequel-(.*)\.tgz.*
diff -pruN 1.5.1-1/doc/advanced_associations.rdoc 3.0.0-1/doc/advanced_associations.rdoc
--- 1.5.1-1/doc/advanced_associations.rdoc	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/doc/advanced_associations.rdoc	2009-04-13 03:44:43.000000000 +0100
@@ -0,0 +1,649 @@
+= Advanced Associations
+
+Sequel::Model has the most powerful and flexible associations of any ruby ORM.
+
+  "Extraordinary claims require extraordinary proof" - Carl Sagan
+
+== Background: Sequel::Model association options
+
+There are a bunch of advanced association options that are available to
+handle the other-than-bog-standard cases.  First we'll go over some of
+the simpler ones:
+
+All associations take a block that can be used to further filter/modify the
+default dataset.  There's also an :eager_block option if you want to use
+a different block when eager loading via Dataset#eager. Association blocks are
+useful for things like:
+
+  Artist.one_to_many :gold_albums, :class=>:Album do |ds|
+    ds.filter{|o| o.copies_sold > 500000}
+  end
+
+There are a whole bunch of options for changing how the association is eagerly
+loaded via Dataset#eager_graph: :graph_block, :graph_conditions,
+:graph_only_conditions, :graph_join_type (and :graph_join_table_* ones for
+JOINing to the join table in a many_to_many association).
+
+- :graph_join_type - The type of join to do
+- :graph_conditions - Additional conditions to put on join (needs to be a
+  hash or array of all two pairs).  Automatically assumes unqualified symbols
+  as first element of the pair to be columns of the associated model, and
+  unqualified symbols of the second element of the pair to be columns of the
+  current model.
+- :graph_block - A block passed to join_table, allowing you to specify
+  conditions other than equality, or to use OR, or set up any arbitrary
+  condition.  The block is passed the associated table alias, current model
+  alias, and array of previous joins.
+- :graph_only_conditions - Use these conditions instead of the standard
+  association conditions. This is necessary when you don't want to have an
+  equal condition between the foreign key and primary key of the tables.
+  You can also use this to have a JOIN USING (array of symbols), or a NATURAL
+  or CROSS JOIN (nil, with the appropriate :graph_join_type).
+  
+These can be used like this:
+
+  # Makes Artist.eager_graph(:required_albums).all not return artists that
+  # don't have any albums
+  Artist.one_to_many :required_albums, :class=>:Album, :graph_join_type=>:inner
+  
+  # Makes sure all returned albums have the active flag set
+  Artist.one_to_many :active_albums, :class=>:Album, \
+    :graph_conditions=>{:active=>true}
+    
+  # Only returns albums that have sold more than 500,000 copies
+  Artist.one_to_many :gold_albums, :class=>:Album, \
+    :graph_block=>proc{|j,lj,js| :copies_sold.qualify(j) > 500000}
+    
+  # Handles the case where the to tables are associated by a case insensitive name string
+  Artist.one_to_many :albums, :key=>:artist_name, \
+    :graph_only_conditions=>nil, \
+    :graph_block=>proc{|j,lj,js| {:lower.sql_function(artist_name.qualify(j))=>:lower.sql_function(name.qualify(lj))}}
+    
+  # Handles the case where both key columns have the name artist_name, and you want to use
+  # a JOIN USING
+  Artist.one_to_many :albums, :key=>:artist_name, :graph_only_conditions=>[:artist_name]
+  
+Remember, using #eager_graph is generally only necessary when you need to
+filter/order based on columns in an associated table, it is recommended to
+use #eager for eager loading if possible.
+
+For lazy loading (e.g. Model[1].association), the :dataset option can be used
+to specify an arbitrary dataset (one that uses different keys, multiple keys,
+joins to other tables, etc.).
+
+For eager loading via #eager, the :eager_loader option can be used to specify
+how to eagerly load a complex association.  This is an extremely powerful
+option.  Though it can often be verbose (compared to other things in Sequel),
+it allows you complete control over how to eagerly load associations for a
+group of objects.
+
+:eager_loader should be a proc that takes 3 arguments, a key_hash,
+an array of records, and a hash of dependent associations.  Since you
+are given all of the records, you can do things like filter on
+associations that are specified by multiple keys, or do multiple
+queries depending on the content of the records (which would be
+necessary for polymorphic associations).  Inside the :eager_loader
+proc, you should get the related objects and populate the
+associations cache for all objects in the array of records.  The hash
+of dependent associations is available for you to cascade the eager
+loading down multiple levels, but it is up to you to use it.  The
+key_hash is a performance enhancement that is used by the default
+code and is also available to you.  It is a hash with keys being
+foreign/primary key symbols in the current table, and the values
+being hashes where the key is foreign/primary key values and values
+being arrays of current model objects having the foreign/primary key
+value associated with the key.  This is hard to visualize, so I'll
+give an example:
+    
+  album1 = Album.load(:id=>1, :artist_id=>2)
+  album2 = Album.load(:id=>3, :artist_id=>2)
+  Album.many_to_one :artist
+  Album.one_to_many :tracks
+  Album.eager(:band, :tracks).all
+  # The key_hash provided to the :eager_loader proc would be:
+  {:id=>{1=>[album1], 3=>[album2]}, :artist_id=>{2=>[album1, album2]}}
+
+Using these options, you can build associations Sequel doesn't natively support,
+and still be able to use the same eager loading features that you are used to.
+
+== ActiveRecord associations
+
+Sequel supports all of associations that ActiveRecord supports, one way or
+another.  Sometimes this requires more code, as Sequel is a toolkit and not
+a swiss army chainsaw.
+
+=== Association callbacks
+
+Sequel supports the same callbacks that ActiveRecord does: :before_add,
+:before_remove, :after_add, and :after_remove. It also supports a
+callback that ActiveRecord does not, :after_load, which is called
+after the association has been loaded.
+
+Each of these options can be a Symbol specifying an instance method
+that takes one argument (the associated object), or a Proc that takes
+two arguments (the current object and the associated object), or an
+array of Symbols and Procs.  For :after_load with a *_to_many association,
+the associated object argument is an array of associated objects.
+
+If any of the before callbacks return false, the adding/removing
+does not happen and it either raises an error (the default), or
+returns nil (if raise_on_save_failure is false).
+
+All callbacks are also run on many_to_one associations.  If there
+was already an existing object for the association, it calls the
+remove callbacks on the existing object and the add callbacks on the
+new object.  The remove callback calls are placed around the add
+callback calls.
+
+=== Association extensions
+
+All associations come with an association_dataset method that can be further filtered or
+otherwise modified:
+
+  class Author < Sequel::Model
+    one_to_many :authorships
+  end
+  Author.first.authorships_dataset.filter{|o| o.number < 10}.first
+ 
+You can extend a dataset with a module easily with :extend.  You can reference
+the model object that created the association dataset via the dataset's
+model_object method, and the related association reflection via the dataset's
+association_reflection method:
+
+  module FindOrCreate
+    def find_or_create(vals)
+      first(vals) || model.create(vals.merge(association_reflection[:key]=>model_object.id))
+    end
+  end
+  class Author < Sequel::Model
+    one_to_many :authorships, :extend=>FindOrCreate
+  end
+  Author.first.authorships_dataset.find_or_create(:name=>'Blah', :number=>10)
+
+=== has_many :through associations
+
+many_to_many handles the usual case of a has_many :through with a belongs_to in
+the associated model.  It doesn't break on the case where the join table is a
+model table, unlike ActiveRecord's has_and_belongs_to_many.
+
+ActiveRecord:
+
+  class Author < ActiveRecord::Base
+    has_many :authorships
+    has_many :books, :through => :authorships
+  end
+
+  class Authorship < ActiveRecord::Base
+    belongs_to :author
+    belongs_to :book
+  end
+
+  @author = Author.find :first
+  @author.books
+
+Sequel::Model:
+
+  class Author < Sequel::Model
+    one_to_many :authorships
+    many_to_many :books, :join_table=>:authorships
+  end
+
+  class Authorship < Sequel::Model
+    many_to_one :author
+    many_to_one :book
+  end
+
+  @author = Author.first
+  @author.books
+
+If you use an association other than belongs_to in the associated model, things
+are a bit more involved (has_many :through a has_many association):
+
+ActiveRecord:
+
+  class Firm < ActiveRecord::Base
+    has_many :clients
+    has_many :invoices, :through => :clients
+  end
+
+  class Client < ActiveRecord::Base
+    belongs_to :firm
+    has_many :invoices
+  end
+
+  class Invoice < ActiveRecord::Base
+    belongs_to :client
+    has_one :firm, :through => :client
+  end
+
+  Firm.find(:first).invoices
+  
+Sequel::Model:
+
+  class Firm < Sequel::Model
+    one_to_many :clients
+    one_to_many :invoices, :read_only=>true, \
+      :dataset=>proc{Invoice.eager_graph(:client).filter(:client__firm_id=>pk)}, \
+      :after_load=>(proc do |firm, invs|
+        invs.each do |inv|
+          inv.client.associations[:firm] = inv.associations[:firm] = firm
+        end
+      end), \
+      :eager_loader=>(proc do |key_hash, firms, associations|
+        id_map = key_hash[Firm.primary_key]
+        firms.each{|firm| firm.associations[:invoices] = []} 
+        Invoice.eager_graph(:client).filter(:client__firm_id=>id_map.keys).all do |inv|
+          id_map[inv.client.firm_id].each do |firm|
+            inv.client.associations[:firm] = inv.associations[:firm] = firm
+            firm.associations[:invoices] << inv 
+          end
+        end
+      end)
+  end 
+
+  class Client < Sequel::Model
+    many_to_one :firm
+    one_to_many :invoices
+  end
+
+  class Invoice < Sequel::Model
+    many_to_one :client
+    many_to_one :firm, :key=>nil, :read_only=>true, \
+      :dataset=>proc{Firm.eager_graph(:clients).filter(:clients__id=>client_id)}, \
+      :after_load=>(proc do |inv, firm|
+        # Delete the cached associations from firm, because it only has the
+        # client with this invoice, instead of all clients of the firm
+        inv.associations[:client] = firm.associations.delete(:clients).first
+      end), \
+      :eager_loader=>(proc do |key_hash, invoices, associations|
+        id_map = {}
+        invoices.each do |inv|
+          inv.associations[:firm] = nil
+          inv.associations[:client] = nil
+          (id_map[inv.client_id] ||= []) << inv
+        end
+        Firm.eager_graph(:clients).filter(:clients__id=>id_map.keys).all do |firm|
+          # Delete the cached associations from firm, because it only has the
+          # clients related the invoices being eagerly loaded, instead of all
+          # clients of the firm.
+          firm.associations.delete(:clients).each do |client|
+            id_map[client.pk].each do |inv|
+              inv.associations[:firm] = firm
+              inv.associations[:client] = client
+            end
+          end
+        end
+      end)
+  end
+  Firm.find(:first).invoices
+
+It is significantly more code in Sequel Model, but quite a bit of it is setting
+the intermediate associated record (the client) and the reciprocal association
+in the associations cache for each object, which ActiveRecord won't do for you.
+The reason you would want to do this is that then firm.invoices.first.firm or
+firm.invoices.first.client doesn't do another query to get the firm/client.
+
+=== Polymorphic Associations
+
+Polymorphic associations are really a design flaw.  The only advantage
+polymorphic associations offer is that they require fewer join tables.
+
+Proof by Reductio ad absurdum: If fewer join tables are preferable, then surely
+fewer tables and columns are preferrable, so you might as well store all of
+your data in a single column in a single table if you think polymorphic
+associations are a good idea.
+
+Compelling Argument: Polymorphic associations are more complex than normal
+associations, and they break referential integrity, so the only reason you
+should use them is if you are already stuck with an existing design that
+uses them.  You should never use them in new code.
+
+ActiveRecord:
+
+  class Asset < ActiveRecord::Base
+    belongs_to :attachable, :polymorphic => true
+  end
+
+  class Post < ActiveRecord::Base
+    has_many :assets, :as => :attachable
+  end
+
+  class Note < ActiveRecord::Base
+    has_many :assets, :as => :attachable
+  end
+
+  @asset.attachable = @post
+  @asset.attachable = @note
+
+Sequel::Model:
+
+  class Asset < Sequel::Model
+    many_to_one :attachable, :reciprocal=>:assets, \
+      :dataset=>(proc do
+        klass = attachable_type.constantize
+        klass.filter(klass.primary_key=>attachable_id)
+      end), \
+      :eager_loader=>(proc do |key_hash, assets, associations|
+        id_map = {}
+        assets.each do |asset|
+          asset.associations[:attachable] = nil 
+          ((id_map[asset.attachable_type] ||= {})[asset.attachable_id] ||= []) << asset
+        end
+        id_map.each do |klass_name, id_map|
+          klass = klass_name.constantize
+          klass.filter(klass.primary_key=>id_map.keys).all do |attach|
+            id_map[attach.pk].each do |asset|
+              asset.associations[:attachable] = attach
+            end
+          end
+        end
+      end)
+  
+    private
+
+    def _attachable=(attachable)
+      self[:attachable_id] = (attachable.pk if attachable)
+      self[:attachable_type] = (attachable.class.name if attachable)
+    end 
+  end 
+
+  class Post < Sequel::Model
+    one_to_many :assets, :key=>:attachable_id do |ds|
+      ds.filter(:attachable_type=>'Post')
+    end 
+  
+    private
+
+    def _add_asset(asset)
+      asset.attachable_id = pk
+      asset.attachable_type = 'Post'
+      asset.save
+    end 
+    def _remove_asset(asset)
+      asset.attachable_id = nil 
+      asset.attachable_type = nil 
+      asset.save
+    end 
+    def _remove_all_assets
+      Asset.filter(:attachable_id=>pk, :attachable_type=>'Post')\
+        .update(:attachable_id=>nil, :attachable_type=>nil)
+    end
+  end
+
+  class Note < Sequel::Model
+    one_to_many :assets, :key=>:attachable_id do |ds|
+      ds.filter(:attachable_type=>'Note')
+    end
+
+    private
+
+    def _add_asset(asset)
+      asset.attachable_id = pk
+      asset.attachable_type = 'Note'
+      asset.save
+    end
+    def _remove_asset(asset)
+      asset.attachable_id = nil
+      asset.attachable_type = nil
+      asset.save
+    end
+    def _remove_all_assets
+      Asset.filter(:attachable_id=>pk, :attachable_type=>'Note')\
+        .update(:attachable_id=>nil, :attachable_type=>nil)
+    end
+  end
+
+  @asset.attachable = @post
+  @asset.attachable = @note
+
+== More advanced associations
+
+So far, we've only shown that Sequel::Model has associations as powerful as
+ActiveRecord's.  Now we will show how Sequel::Model's associations are more
+powerful.
+
+=== many_to_one/one_to_many not referencing primary key
+
+This can now be handled easily in Sequel using the :primary_key association
+option.  However, this example shows how the association was possible before
+the introduction of that option.
+
+Let's say you have two tables, invoices and clients, where each client is
+associated with many invoices.  However, instead of using the client's primary
+key, the invoice is associated to the client by name (this is bad database
+design, but sometimes you have to play with the cards you are dealt).
+
+  class Client < Sequel::Model
+    one_to_many :invoices, :reciprocal=>:client, \
+      :dataset=>proc{Invoice.filter(:client_name=>name)}, \
+      :eager_loader=>(proc do |key_hash, clients, associations|
+        id_map = {}
+        clients.each do |client|
+          id_map[client.name] = client
+          client.associations[:invoices] = []
+        end
+        Invoice.filter(:client_name=>id_map.keys.sort).all do |inv|
+          inv.associations[:client] = client = id_map[inv.client_name]
+          client.associations[:invoices] << inv 
+        end
+      end)
+
+    private
+
+    def _add_invoice(invoice)
+      invoice.client_name = name
+      invoice.save
+    end 
+    def _remove_invoice(invoice)
+      invoice.client_name = nil 
+      invoice.save
+    end 
+    def _remove_all_invoices
+      Invoice.filter(:client_name=>name).update(:client_name=>nil)
+    end 
+  end 
+
+  class Invoice < Sequel::Model
+    many_to_one :client, :key=>:client_name, \
+      :dataset=>proc{Client.filter(:name=>client_name)}, \
+      :eager_loader=>(proc do |key_hash, invoices, associations|
+        id_map = key_hash[:client_name]
+        invoices.each{|inv| inv.associations[:client] = nil}
+        Client.filter(:name=>id_map.keys).all do |client|
+          id_map[client.name].each{|inv| inv.associations[:client] = client}
+        end
+      end)
+
+    private
+
+    def _client=(client)
+      self.client_name = (client.name if client)
+    end
+  end
+
+=== Joining on multiple keys
+
+Let's say you have two tables that are associated with each other with multiple
+keys.  For example:
+
+  # Both of these models have an album_id, number, and disc_number fields.
+  # All FavoriteTracks have an associated track, but not all tracks have an
+  # associated favorite track
+  
+  class Track < Sequel::Model
+    many_to_one :favorite_track, \
+      :dataset=>(proc do
+        FavoriteTrack.filter(:disc_number=>disc_number, :number=>number, :album_id=>album_id)
+      end), \
+      :eager_loader=>(proc do |key_hash, tracks, associations|
+        id_map = {}
+        tracks.each do |t|
+          t.associations[:favorite_track] = nil
+          id_map[[t[:album_id], t[:disc_number], t[:number]]] = t
+        end
+        FavoriteTrack.filter([:album_id, :disc_number, :number]=>id_map.keys).all do |ft|
+          if t = id_map[[ft[:album_id], ft[:disc_number], ft[:number]]]
+            t.associations[:favorite_track] = ft
+          end
+        end
+      end)
+  end
+
+  class FavoriteTrack < Sequel::Model
+    many_to_one :track, \
+      :dataset=>(proc do
+        Track.filter(:disc_number=>disc_number, :number=>number, :album_id=>album_id)
+      end), \
+      :eager_loader=>(proc do |key_hash, ftracks, associations|
+        id_map = {}
+        ftracks.each{|ft| id_map[[ft[:album_id], ft[:disc_number], ft[:number]]] = ft}
+        Track.filter([:album_id, :disc_number, :number]=>id_map.keys).all do |t|
+          id_map[[t[:album_id], t[:disc_number], t[:number]]].associations[:track] = t
+        end
+      end)
+  end
+
+=== Tree - All Ancestors and Descendents
+
+Let's say you want to store a tree relationship in your database, it's pretty
+simple:
+
+  class Node < Sequel::Model
+    many_to_one :parent
+    one_to_many :children, :key=>:parent_id
+  end
+  
+You can easily get a node's parent with node.parent, and a node's children with
+node.children.  You can even eager load the relationship up to a certain depth:
+
+  # Eager load three generations of generations of children for a given node 
+  Node.filter(:id=>1).eager(:children=>{:children=>:children}).all.first
+  # Load parents and grandparents for a group of nodes
+  Node.filter{|o| o.id < 10}.eager(:parent=>:parent).all
+
+What if you want to get all ancestors up to the root node, or all descendents,
+without knowing the depth of the tree?
+
+  class Node < Sequel::Model
+    many_to_one :ancestors, :eager_loader=>(proc do |key_hash, nodes, associations|
+      # Handle cases where the root node has the same parent_id as primary_key
+      # and also when it is NULL
+      non_root_nodes = nodes.reject do |n| 
+        if [nil, n.pk].include?(n.parent_id)
+          # Make sure root nodes have their parent association set to nil
+          n.associations[:parent] = nil 
+          true
+        else
+          false
+        end 
+      end 
+      unless non_root_nodes.empty?
+        id_map = {}
+        # Create an map of parent_ids to nodes that have that parent id
+        non_root_nodes.each{|n| (id_map[n.parent_id] ||= []) << n}
+        # Doesn't cause an infinte loop, because when only the root node
+        # is left, this is not called.
+        Node.filter(Node.primary_key=>id_map.keys).eager(:ancestors).all do |node|
+          # Populate the parent association for each node
+          id_map[node.pk].each{|n| n.associations[:parent] = node}
+        end 
+      end 
+    end)
+    many_to_one :descendants, :eager_loader=>(proc do |key_hash, nodes, associations|
+      id_map = {}
+      nodes.each do |n| 
+        # Initialize an empty array of child associations for each parent node
+        n.associations[:children] = []
+        # Populate identity map of nodes
+        id_map[n.pk] = n 
+      end 
+      # Doesn't cause an infinite loop, because the :eager_loader is not called
+      # if no records are returned.  Exclude id = parent_id to avoid infinite loop
+      # if the root note is one of the returned records and it has parent_id = id
+      # instead of parent_id = NULL.
+      Node.filter(:parent_id=>id_map.keys).exclude(:id=>:parent_id).eager(:descendants).all do |node|
+        # Get the parent from the identity map
+        parent = id_map[node.parent_id]
+        # Set the child's parent association to the parent 
+        node.associations[:parent] = parent
+        # Add the child association to the array of children in the parent
+        parent.associations[:children] << node
+      end 
+    end)
+  end
+
+=== Joining multiple keys to a single key, through a third table
+
+Let's say you have a database, of songs, lyrics, and artists.  Each song
+may or may not have a lyric (most songs are instrumental).  The lyric can be
+associated to an artist in each of four ways: composer, arranger, vocalist,
+or lyricist.  These may all be the same, or they could all be different, and
+none of them are required.  The songs table has a lyric_id field to associate
+it to the lyric, and the lyric table has four fields to associate it to the
+artist (composer_id, arranger_id, vocalist_id, and lyricist_id).
+
+What you want to do is get all songs for a given artist, ordered by the song's
+name, with no duplicates?
+
+  class Artist < Sequel::Model
+    one_to_many :songs, :order=>:songs__name, \
+      :dataset=>proc{Song.select(:songs.*).join(Lyric, :id=>:lyric_id, id=>[:composer_id, :arranger_id, :vocalist_id, :lyricist_id])}, \
+      :eager_loader=>(proc do |key_hash, records, associations|
+        h = key_hash[:id]
+        ids = h.keys
+        records.each{|r| r.associations[:songs] = []}
+        Song.select(:songs.*, :lyrics__composer_id, :lyrics__arranger_id, :lyrics__vocalist_id, :lyrics__lyricist_id)\
+         .join(Lyric, :id=>:lyric_id){{:composer_id=>ids, :arranger_id=>ids, :vocalist_id=>ids, :lyricist_id=>ids}.sql_or}\
+         .order(:songs__name).all do |song|
+          [:composer_id, :arranger_id, :vocalist_id, :lyricist_id].each do |x|
+            recs = h[song.values.delete(x)]
+            recs.each{|r| r.associations[:songs] << song} if recs
+          end
+        end
+        records.each{|r| r.associations[:songs].uniq!}
+      end)
+  end
+
+=== Statistics Associations (Sum of Associated Table Column)
+
+In addition to getting associated records, you can use Sequel's association support
+to get aggregate information for columns in associated tables (sums, averages, etc.).
+
+Let's say you have a database with projects and tickets.  A project can have many
+tickets, and each ticket has a number of hours associated with it.  You can use the
+association support to create a Project association that gives the sum of hours for all
+associated tickets.
+
+
+  class Project < Sequel::Model
+    one_to_many :tickets
+    many_to_one :ticket_hours, :read_only=>true, :key=>:id,
+     :dataset=>proc{Ticket.filter(:project_id=>id).select{sum(hours).as(hours)}},
+     :eager_loader=>(proc do |kh, projects, a|
+      projects.each{|p| p.associations[:ticket_hours] = nil}
+      Ticket.filter(:project_id=>kh[:id].keys).
+       group(:project_id).
+       select{[project_id, sum(hours).as(hours)]}.
+       all do |t|
+        p = kh[:id][t.values.delete(:project_id)].first
+        p.associations[:ticket_hours] = t
+       end
+     end)
+    # The association method returns a Ticket object with a single aggregate
+    # sum-of-hours value, but you want it to return an Integer/Float of just the
+    # sum of hours, so you call super and return just the sum-of-hours value.
+    # This works for both lazy loading and eager loading.
+    def ticket_hours
+      if s = super
+        s[:hours]
+      end
+    end
+  end
+  class Ticket < Sequel::Model
+    many_to_one :project
+  end
+
+Note that it is often better to use a sum cache instead of this approach.  You can implement
+a sum cache using before or after save and delete hooks, or using a database trigger
+(the preferred method if you only have to support one database and that database supports
+triggers).
diff -pruN 1.5.1-1/doc/cheat_sheet.rdoc 3.0.0-1/doc/cheat_sheet.rdoc
--- 1.5.1-1/doc/cheat_sheet.rdoc	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/doc/cheat_sheet.rdoc	2009-03-23 17:06:57.000000000 +0000
@@ -0,0 +1,206 @@
+= Cheat Sheet   
+
+== Open a database
+
+  require 'rubygems'
+  require 'sequel'
+
+  DB = Sequel.sqlite('my_blog.db')
+  DB = Sequel.connect('postgres://user:password@localhost/my_db')
+  DB = Sequel.postgres('my_db', :user => 'user', :password => 'password', :host => 'localhost')
+  DB = Sequel.ado('mydb')
+
+== Open an SQLite memory database
+
+Without a filename argument, the sqlite adapter will setup a new sqlite database in memory.
+
+  DB = Sequel.sqlite
+
+== Logging SQL statements
+
+  require 'logger'
+  DB = Sequel.sqlite '', :loggers => [Logger.new($stdout)]
+  # or
+  DB.loggers << Logger.new(...)
+
+== Using raw SQL
+
+  DB << "CREATE TABLE users (name VARCHAR(255) NOT NULL, age INT(3) NOT NULL)"
+  dataset = DB["SELECT age FROM users WHERE name = ?", name]
+  dataset.map(:age)
+  DB.fetch("SELECT name FROM users") do |row|
+    p r[:name]
+  end
+
+== Create a dataset
+
+  dataset = DB[:items]
+  dataset = DB.from(:items)
+
+== Most dataset methods are chainable
+
+  dataset = DB[:managers].where(:salary => 5000..10000).order(:name, :department)
+
+== Insert rows
+
+  dataset.insert(:name => 'Sharon', :grade => 50)
+
+== Retrieve rows
+
+  dataset.each{|r| p r}
+  dataset.all #=> [{...}, {...}, ...]
+  dataset.first
+
+== Update/Delete rows
+
+  dataset.filter(~:active).delete
+  dataset.filter('price < ?', 100).update(:active => true)
+
+== Datasets are Enumerable
+
+  dataset.map{|r| r[:name]}
+  dataset.map(:name) # same as above
+
+  dataset.inject(0){|sum, r| sum + r[:value]}
+  dataset.sum(:value) # same as above
+
+== Filtering (see also doc/dataset_filtering.rdoc)
+
+  dataset.filter(:name => 'abc')
+  dataset.filter('name = ?', 'abc')
+
+  dataset.filter{|o| o.value > 100}
+  dataset.exclude{|o| o.value <= 100}
+
+  dataset.filter(:value => 50..100)
+  dataset.where{|o| (o.value >= 50) & (o.value <= 100)}
+
+  dataset.where('value IN ?', [50,75,100])
+  dataset.where(:value=>[50,75,100])
+
+  dataset.filter(:name => 'abc').first
+  dataset[:name => 'abc'] # same as above
+
+  # Filter using a subquery
+  dataset.filter{|o| o.price > dataset.select(o.avg(price) + 100)}
+
+=== Advanced filtering using ruby expressions
+
+  DB[:items].filter{|o| o.price < 100}.sql 
+  #=> "SELECT * FROM items WHERE (price < 100)" 
+
+  DB[:items].filter(:name.like('AL%')).sql 
+  #=> "SELECT * FROM items WHERE (name LIKE 'AL%')" 
+
+There's support for nested expressions with AND, OR and NOT:
+
+  DB[:items].filter{|o| (o.x > 5) & (o.y > 10)}.sql 
+  #=> "SELECT * FROM items WHERE ((x > 5) AND (y > 10))" 
+
+  DB[:items].filter({:x => 1, :y => 2}.sql_or & ~{:z => 3}).sql 
+  #=> "SELECT * FROM items WHERE (((x = 1) OR (y = 2)) AND (z != 3))"
+
+You can use arithmetic operators and specify SQL functions:
+
+  DB[:items].filter((:x + :y) > :z).sql 
+  #=> "SELECT * FROM items WHERE ((x + y) > z)" 
+
+  DB[:items].filter{|o| :price - 100 < o.avg(:price)}.sql 
+  #=> "SELECT * FROM items WHERE ((price - 100) < avg(price))" 
+
+== Ordering
+
+  dataset.order(:kind)
+  dataset.reverse_order(:kind)
+  dataset.order(:kind.desc, :name)
+
+== Row ranges
+
+  dataset.limit(30) # LIMIT 30
+  dataset.limit(30, 10) # LIMIT 30 OFFSET 10
+
+== Joins
+
+  DB[:items].left_outer_join(:categories, :id => :category_id).sql #=>
+    "SELECT * FROM items LEFT OUTER JOIN categories ON categories.id = items.category_id"
+
+== Summarizing
+
+  dataset.count #=> record count
+  dataset.max(:price)
+  dataset.min(:price)
+  dataset.avg(:price)
+  dataset.sum(:stock)
+
+  dataset.group(:category).select(:category, :AVG.sql_function(:price))
+
+== SQL Functions / Literals
+
+  dataset.update(:updated_at => :NOW.sql_function)
+  dataset.update(:updated_at => 'NOW()'.lit)
+
+  dataset.update(:updated_at => "DateValue('1/1/2001')".lit)
+  dataset.update(:updated_at => :DateValue.sql_function('1/1/2001'))
+
+== Schema Manipulation
+
+  DB.create_table :items do
+    primary_key :id
+    String :name, :unique => true, :null => false
+    boolean :active, :default => true
+    foreign_key :category_id, :categories
+    Time :created_at
+    
+    index :grade
+  end
+
+  DB.drop_table :items
+
+  DB.create_table :test do
+    String :zipcode
+    enum :system, :elements => ['mac', 'linux', 'windows']
+  end
+
+== Aliasing
+
+  DB[:items].select(:name.as(:item_name))
+  DB[:items].select(:name___item_name)
+  DB[:items___items_table].select(:items_table__name___item_name)
+  # => "SELECT items_table.name AS item_name FROM items AS items_table"
+
+== Transactions
+
+  DB.transaction do
+    dataset.insert(:first_name => 'Inigo', :last_name => 'Montoya')
+    dataset.insert(:first_name => 'Farm', :last_name => 'Boy')
+  end # Either both are inserted or neither are inserted
+
+Database#transaction is re-entrant:
+
+  DB.transaction do # BEGIN issued only here
+    DB.transaction
+      dataset << {:first_name => 'Inigo', :last_name => 'Montoya'}
+    end
+  end # COMMIT issued only here
+
+Transactions are aborted if an error is raised:
+
+  DB.transaction do
+    raise "some error occurred"
+  end # ROLLBACK issued and the error is re-raised
+
+Transactions can also be aborted by raising Sequel::Rollback:
+
+  DB.transaction do
+    raise(Sequel::Rollback) if something_bad_happened
+  end # ROLLBACK issued and no error raised
+
+Miscellaneous:
+
+  dataset.sql #=> "SELECT * FROM items"
+  dataset.delete_sql #=> "DELETE FROM items"
+  dataset.where(:name => 'sequel').exists #=> "EXISTS ( SELECT 1 FROM items WHERE name = 'sequel' )"
+  dataset.columns #=> array of columns in the result set, does a SELECT
+
+  # Works on PostgreSQL, MySQL, SQLite, and JDBC
+  DB.schema(:items) => [[:id, {:type=>:integer, ...}], [:name, {:type=>:string, ...}], ...]
diff -pruN 1.5.1-1/doc/dataset_filtering.rdoc 3.0.0-1/doc/dataset_filtering.rdoc
--- 1.5.1-1/doc/dataset_filtering.rdoc	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/doc/dataset_filtering.rdoc	2009-03-18 21:29:32.000000000 +0000
@@ -0,0 +1,158 @@
+= Dataset Filtering 
+
+Sequel offers unparalleled flexibility when it comes to filtering records. You can specify your conditions as a custom string, as a string with parameters, as a hash of values to compare against, or as ruby code that Sequel translates into SQL expressions.
+
+== Filtering using a custom filter string
+
+If you wish to write your SQL by hand, you can just supply it to the dataset's #filter method:
+
+  items.filter('x < 10').sql
+  #=> "SELECT * FROM items WHERE x < 10"
+
+In order to prevent SQL injection, you can replace literal values with question marks and supply the values as additional arguments:
+
+  items.filter('category = ?', 'ruby').sql
+  #=> "SELECT * FROM items WHERE category = 'ruby'"
+
+=== Specifying SQL functions
+
+Sequel also allows you to specify functions by using the Symbol#sql_function method (and the Symbol#[] method on ruby 1.8):
+
+  items.literal(:avg.sql_function(:price)) #=> "avg(price)"
+
+== Filtering using a hash
+
+If you just need to compare records against values, you can supply a hash:
+
+  items.filter(:category => 'ruby').sql
+  #=> "SELECT * FROM items WHERE (category = 'ruby')"
+
+Sequel can check for null values:
+
+  items.filter(:category => nil).sql
+  #=> "SELECT * FROM items WHERE (category IS NULL)"
+
+Or compare two columns:
+
+  items.filter(:x => :some_table__y).sql
+  #=> "SELECT * FROM items WHERE (x = some_table.y)"
+
+And also compare against multiple values:
+
+  items.filter(:category => ['ruby', 'perl']).sql
+  #=> "SELECT * FROM items WHERE (category IN ('ruby', 'perl'))"
+
+Ranges (both inclusive and exclusive) can also be used:
+
+  items.filter(:price => 100..200).sql
+  #=> "SELECT * FROM items WHERE (price >= 100 AND price <= 200)"
+
+  items.filter(:price => 100...200).sql
+  #=> "SELECT * FROM items WHERE (price >= 100 AND price < 200)"
+
+== Filtering using expressions
+
+Sequel allows you to use ruby expressions directly in the call to filter:
+
+  items.filter(:price * 2 < 50).sql
+  #=> "SELECT * FROM items WHERE ((price * 2) < 50) 
+
+This works for the standard inequality and arithmetic operators (though you can't use the inequality operators directly on a symbol in ruby 1.9):
+
+  items.filter(:price + 100 < 200).sql
+  #=> "SELECT * FROM items WHERE ((price + 100) < 200) 
+
+  items.filter(:price - 100 > 200).sql
+  #=> "SELECT * FROM items WHERE ((price - 100) > 200) 
+
+  items.filter(:price * 100 <= 200).sql
+  #=> "SELECT * FROM items WHERE ((price * 100) <= 200) 
+
+  items.filter(:price / 100 >= 200).sql
+  #=> "SELECT * FROM items WHERE ((price / 100) >= 200) 
+
+You use the overloaded bitwise and (&) and or (|) operators to combine expressions:
+
+  items.filter((:price + 100 < 200) & (:price * 100 <= 200)).sql
+  #=> "SELECT * FROM items WHERE (((price + 100) < 200) AND ((price * 100) <= 200)) 
+
+  items.filter((:price - 100 > 200) | (:price / 100 >= 200)).sql
+  #=> "SELECT * FROM items WHERE (((price - 100) > 200) OR ((price / 100) >= 200)) 
+
+To filter by equality, you use the standard hash, which can be combined with other operators:
+
+  items.filter({:category => 'ruby'} & (:price + 100 < 200)).sql
+  #=> "SELECT * FROM items WHERE ((category = 'ruby') AND ((price + 100) < 200))"
+
+This works with other hash values, such as arrays and ranges:
+
+  items.filter({:category => ['ruby', 'other']} | (:price - 100 > 200)).sql
+  #=> "SELECT * FROM items WHERE ((category IN ('ruby', 'other')) OR ((price - 100) <= 200))"
+
+  items.filter({:price => (100..200)} & :active).sql
+  #=> "SELECT * FROM items WHERE ((price >= 100 AND price <= 200) AND active)"
+
+=== Negating conditions
+
+You can use the negation operator (~) in most cases:
+
+  items.filter(~{:category => 'ruby'}).sql
+  #=> "SELECT * FROM items WHERE (category != 'ruby')"
+
+  items.filter{~:active}.sql
+  #=> "SELECT * FROM items WHERE NOT active"
+
+  items.filter(~(:price / 100 >= 200)).sql
+  #=> "SELECT * FROM items WHERE ((price / 100) < 200) 
+
+=== Comparing against column references
+
+You can also compare against other columns:
+
+  items.filter{|o| o.credit > :debit}.sql
+  #=> "SELECT * FROM items WHERE (credit > debit)
+
+Or against SQL functions:
+
+  items.filter{|o| :price - 100 < o.max(:price)}.sql
+  #=> "SELECT * FROM items WHERE ((price - 100) < max(price))"
+
+== String search functions
+
+You can search SQL strings using the #like method:
+
+  items.filter(:name.like('Acme%')).sql
+  #=> "SELECT * FROM items WHERE (name LIKE 'Acme%')"
+
+You can specify a Regexp as a like argument, but this will probably only work
+on PostgreSQL and MySQL:
+
+  items.filter(:name.like(/Acme.*/)).sql
+  #=> "SELECT * FROM items WHERE (name ~ 'Acme.*')"
+
+Like can also take more than one argument:
+
+  items.filter(:name.like('Acme%', /Beta.*/)).sql
+  #=> "SELECT * FROM items WHERE ((name LIKE 'Acme%') OR (name ~ 'Beta.*'))"
+
+== String concatenation
+
+You can concatenate SQL strings using Array#sql_string_join: 
+
+  items.filter([:name, :comment].sql_string_join.like('%acme%')).sql
+  #=> "SELECT * FROM items WHERE ((name || comment) LIKE 'Acme%')"
+
+Array#sql_string_join also takes a join argument:
+
+  items.filter([:name, :comment].sql_string_join(' ').like('%acme%')).sql
+  #=> "SELECT * FROM items WHERE ((name || ' ' || comment) LIKE 'Acme%')"
+
+== Filtering using sub-queries
+
+One of the best features of Sequel is the ability to use datasets as sub-queries. Sub-queries can be very useful for filtering records, and many times provide a simpler alternative to table joins. Sub-queries can be used in all forms of filters:
+
+  refs = consumer_refs.filter(:logged_in).select(:consumer_id)
+  consumers.filter(:id => refs).sql
+  #=> "SELECT * FROM consumers WHERE (id IN (SELECT consumer_id FROM consumer_refs WHERE logged_in))"
+
+Note that if you compare against a sub-query, you must select a single column in the sub-query.
diff -pruN 1.5.1-1/doc/prepared_statements.rdoc 3.0.0-1/doc/prepared_statements.rdoc
--- 1.5.1-1/doc/prepared_statements.rdoc	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/doc/prepared_statements.rdoc	2009-03-18 21:33:37.000000000 +0000
@@ -0,0 +1,104 @@
+= Prepared Statements and Bound Variables
+
+Sequel has support for prepared statements and bound variables.  No matter which
+database you are using, the Sequel prepared statement/bound variable API remains
+the same.  There is native support for prepared statements/bound variables on
+the following databases:
+
+* PostgreSQL (using the pg driver, requires type specifiers)
+* MySQL (prepared statements only, as the ruby mysql driver doesn't support
+  bound variables)
+* SQLite (a new native prepared statement is used for each call, though)
+* JDBC
+
+Support on other databases is emulated via string interpolation.
+
+== Placeholders
+
+Generally, when using prepared statements (and certainly when using bound
+variables), you need to put placeholders in your SQL to indicate where you
+want your bound arguments to appear.  Database support and syntax vary
+significantly for placeholders (e.g. :name, $1, ?).  Sequel abstracts all of
+that and allows you to specify placeholders by using the :$name format for
+placeholders, e.g.:
+
+  ds = DB[:items].filter(:name=>:$n)
+
+== Bound Variables
+
+Using bound variables for this query is simple:
+
+  ds.call(:select, :n=>'Jim')
+
+This will do the equivalent of selecting records that have the name 'Jim'. It
+returns all records, and can take a block that is passed to Dataset#all.
+
+Deleting or returning the first record works similarly:
+
+  ds.call(:first, :n=>'Jim') # First record with name 'Jim'
+  ds.call(:delete, :n=>'Jim') # Delete records with name 'Jim'
+
+For inserting/updating records, you should also specify a value hash, which
+may itself contain placeholders:
+
+  # Insert record with 'Jim', note that the previous filter is ignored
+  ds.call(:insert, {:n=>'Jim'}, :name=>:$n)
+  # Change name to 'Bob' for all records with name of 'Jim'
+  ds.call(:update, {:n=>'Jim', :new_n=>'Bob'}, :name=>$:new_n)
+
+== Prepared Statements
+
+Prepared statement support is similar to bound variable support, but you
+use Dataset#prepare with a name, and Dataset#call later with the values:
+
+  ds = DB[:items].filter(:name=>:$n)
+  ps = ds.prepare(:select, :select_by_name)
+  ps.call(:n=>'Jim')
+  DB.call(:select_by_name, :n=>'Jim') # same as above
+
+The Dataset#prepare method returns a prepared statement, and also stores a
+copy of the prepared statement in the database for later use.  For insert
+and update queries, the hash to insert/update is passed to prepare:
+
+  ps1 = DB[:items].prepare(:insert, :insert_with_name, :name=>:$n)
+  ps1.call(:n=>'Jim')
+  DB.call(:insert_with_name, :n=>'Jim') # same as above
+  ds = DB[:items].filter(:name=>:$n)
+  ps2 = ds.prepare(:update, :update_name, :name=>:$new_n)
+  ps2.call(:n=>'Jim', :new_n=>'Bob')
+  DB.call(:update_name, :n=>'Jim', :new_n=>'Bob') # same as above
+
+== Database support
+
+=== PostgreSQL
+
+If you are using the ruby-postgres or postgres-pr driver, PostgreSQL uses the
+default emulated support.  If you are using ruby-pg, there is native support,
+but it requires type specifiers most of the time.  This is easy if you have
+direct control over the SQL string, but since Sequel abstracts that, the types
+have to be specified another way.  This is done by adding a __* suffix to the
+placeholder symbol (e.g. :$name__text, which will be compiled to "$1::text"
+in the SQL).  Prepared statements are always server side.
+
+=== SQLite
+
+SQLite supports bound variables and prepared statements exactly the same, since
+a new native prepared statement is created and executed for each call.
+
+=== MySQL
+
+The MySQL ruby driver does not support bound variables, so the the bound
+variable methods fall back to string interpolation.  It uses server side
+prepared statements.
+
+=== JDBC
+
+JDBC supports both prepared statements and bound variables.  Whether these
+are server side or client side depends on the JDBC driver.  For PostgreSQL
+over JDBC, you can add the prepareThreshold=N parameter to the connection
+string, which will use a server side prepared statement after N calls to
+the prepared statement.
+
+=== All Others
+
+Support is emulated using interpolation.
diff -pruN 1.5.1-1/doc/release_notes/1.0.txt 3.0.0-1/doc/release_notes/1.0.txt
--- 1.5.1-1/doc/release_notes/1.0.txt	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/doc/release_notes/1.0.txt	2009-02-19 18:18:10.000000000 +0000
@@ -0,0 +1,38 @@
+=== New code organization
+
+Sequel is now divided into two parts: sequel_core and sequel_model.
+These two parts are distributed as two separate gems. The sequel gem
+bundles sequel_core and sequel_model together. If you don't use
+Sequel::Model in your code, you can just install and use sequel_core.
+
+=== New model hooks implementation
+
+The hooks implementation have been rewritten from scratch, is much
+more robust and offers a few new features:
+
+* More ways to define hooks: hooks can now be defined by supplying a
+block or a method name, or by overriding the hook instance method.
+
+* Inheritable hooks: Hooks can now be inherited, which means that you
+can define general hooks in a model superclass, and use them in
+subclasses. You can also define global hooks on Sequel::Model that
+will be invoked for all model classes.
+
+* Hook chains can be broken by returning false from within the hook.
+
+* New after_initialize hook, invoked after instance initialization.
+
+* The hook invocation order can no longer be changed. Hooks are
+invoked in order of definition, from the top of the class hierarchy
+(that is, from Sequel::Model) down to the specific class.
+
+=== Miscellanea
+
+* Removed deprecated adapter stubs, and all other deprecations in both
+sequel_core and sequel_model.
+
+* Fixed String#to_time to raise error correctly for invalid time
+stamps.
+
+* Fixed error behavior when parse_tree or ruby2ruby are not available.
+
diff -pruN 1.5.1-1/doc/release_notes/1.1.txt 3.0.0-1/doc/release_notes/1.1.txt
--- 1.5.1-1/doc/release_notes/1.1.txt	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/doc/release_notes/1.1.txt	2009-02-19 18:17:42.000000000 +0000
@@ -0,0 +1,143 @@
+=== DRY Sequel models
+
+With the new Sequel release you no longer need to explicitly specify
+the table
+name for each model class, assuming your model name is the singular of
+the
+table name (just like in ActiveRecord or DataMapper):
+
+  class UglyBug < Sequel::Model
+  end
+
+  UglyBug.table_name #=> :ugly_bugs
+
+=== New model validations and support for virtual attributes
+
+Sequel model now include validation functionality which largly follows
+the
+validations offered in ActiveRecord. Validations can be checked
+anytime by
+calling Model#valid?, with validation errors accessible through
+Model#errors:
+
+  class Item < Sequel::Model
+    validates_presence_of :name
+  end
+
+  my_item = Item.new
+  my_item.valid? #=> false
+  my_item.errors.full_messages #=> ["name is not present"]
+
+The Model#save method has been changed to check for validity before
+saving. If
+the model instance is not valid, the #save method returns false
+without saving
+the instance. You can also bypass the validity test by calling
+Model#save!
+instead.
+
+Model classes also now support virtual attributes, letting you assign
+values to
+any attribute (virtual or persistent) at initialization time:
+
+  class User < Sequel::Model
+    attr_accessor :password
+  end
+
+  u = User.new(:password => 'blah', ...)
+  u.password #=> 'blah'
+
+Also, virtual attributes can be validated just like persistent
+attributes.
+
+=== Other changes (long list!)
+
+* Added Model#reload as alias to Model#refresh.
+
+* Changed Model.create to accept a block (#126).
+
+* Fixed Model#initialize to accept nil values (#115).
+
+* Added Model#update_with_params method with support for virtual
+attributes and auto-filtering of unrelated parameters, and changed
+Model.create_with_params to support virtual attributes (#128).
+
+* Fixed Model.dataset to correctly set the dataset if using implicit
+naming or inheriting the superclass dataset (thanks celldee).
+
+* Finalized support for virtual attributes.
+
+* Fixed Model#set to work with string keys (#143).
+
+* Fixed Model.create to correctly initialize instances marked as new
+(#135).
+
+* Fixed Model#initialize to convert string keys into symbol keys. This
+also fixes problem with validating objects initialized with string
+keys (#136).
+
+* Added Dataset#table_exists? convenience method.
+
+* Changed Dataset#group_and_count to accept multiple columns (#134).
+
+* Added Dataset#select_all method.
+
+* Added Dataset#select_more, Dataset#order_more methods (#129).
+
+* Fixed Dataset#count to work correctly for grouped datasets (#144).
+
+* Fixed joining datasets using aliased tables (#140).
+
+* Added support for UNSIGNED constraint, used in MySQL? (#127).
+
+* Implemented constraint definitions inside Database#create_table.
+
+* Enhanced Database.connect to accept options with string keys, so it
+can now accept options loaded from YAML files. Database.connect also
+automatically converts :username option into :user for compatibility
+with existing YAML configuration files for AR and DataMapper.
+
+* Changed ODBC::Database to support connection using driver and
+database name, also added support for untitled columns in
+ODBC::Dataset (thanks Leonid Borisenko).
+
+* Changed MySQL adapter to support specifying socket option.
+
+* Fixed MySQL adapter to correctly format foreign key definitions
+(#123).
+
+* Changed MySQL::Dataset to allow HAVING clause on ungrouped datasets,
+and put HAVING clause before ORDER BY clause (#133).
+
+* Changed mysql adapter to default to localhost if :host option is not
+specified (#114).
+
+* Added String#to_date. Updated mysql adapter to use String#to_date
+for mysql date types (thanks drfreeze).
+
+* Fixed postgres adapter to define PGconn#async_exec as alias to #exec
+if not defined (for pure-ruby postgres driver).
+
+* Changed postgres adapter to quote column references using double
+quotes.
+
+* Applied patch for oracle adapter: fix behavior of limit and offset,
+transactions, #table_exists?, #tables and additional specs (thanks
+Liming Lian #122).
+
+* Added support additional field types in postgresql adapter (#146).
+
+* Added support for date field types in postgresql adapter (#145).
+
+* Added support for limiting and paginating datasets with fixed SQL,
+e.g. using Database#fetch.
+
+* Added new Dataset#from_self method that returns a dataset selecting
+from the original dataset.
+
+* Allow for additional filters on a grouped dataset (#119 and #120)
+
+* Refactored Sequelizer to use Proc#to_sexp (method provided by r2r).
+
+* Fixed bin/sequel to require sequel_model if available.
+
diff -pruN 1.5.1-1/doc/release_notes/1.3.txt 3.0.0-1/doc/release_notes/1.3.txt
--- 1.5.1-1/doc/release_notes/1.3.txt	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/doc/release_notes/1.3.txt	2009-02-19 18:14:47.000000000 +0000
@@ -0,0 +1,101 @@
+=== Better model associations
+
+The latest release of sequel_model includes a new associations
+functionality written by Jeremy Evans which replaces the old relations
+code in previous versions. Please note that this version is not
+completely backward-compatible and you should therefore upgrade with
+caution.
+
+The new implementation supports three kinds of relations: one_to_many,
+many_to_one and many_to_many, which correspond to has_many, belongs_to
+and has_and_belongs_to_many relations in ActiveRecord. In fact, the
+new implementation includes aliases for ActiveRecord assocation macros
+and is basically compatible with ActiveRecord conventions. It also
+supports DRY implicit class name references. Here's a simple example:
+
+  class Author < Sequel::Model
+    has_many :books # equivalent to one_to_many
+  end
+
+  class Book < Sequel::Model
+    belongs_to :author # equivalent to many_to_one
+    has_and_belongs_to_many :categories # equivalent to many_to_many
+  end
+
+  class Category < Sequel::Model
+    has_and_belongs_to_many :books
+  end
+
+These macros will create the following methods:
+
+* Author#books, Author#add_book, Author#remove_book
+* Book#author, Book#categories, Book#add_category,
+Book#remove_category
+* Category#books, Category#add_book, Category#remove_book
+
+Unlike ActiveRecord, one_to_many and many_to_many association methods
+return a dataset:
+
+  a = Author[1234]
+  a.books.sql #=> 'SELECT * FROM books WHERE (author_id = 1234)'
+
+You can also tell Sequel to cache the association result set and
+return it as an array:
+
+  class Author < Sequel::Model
+    has_many :books, :cache => true
+  end
+
+  Author[1234].books.class #=> Array
+
+You can of course bypass the defaults and specify class names and key
+names:
+
+  class Node < Sequel::Model
+    belongs_to :parent, :class => Node
+    belongs_to :session, :key => :producer_id
+  end
+
+Another useful option is :order, which sets the order for the
+association dataset:
+
+  class Author < Sequel::Model
+    has_many :books, :order => :title
+  end
+
+  Author[1234].books.sql #=> 'SELECT * FROM books WHERE (author_id =
+1234) ORDER BY title'
+
+More information about associations can be found in the Sequel
+documentation.
+
+=== Other changes
+
+* Added configuration file for running specs (#186).
+
+* Changed Database#drop_index to accept fixed arity (#173).
+
+* Changed column definition sql to put UNSIGNED constraint before
+unique in order to satisfy MySQL (#171).
+
+* Enhanced MySQL adapter to support load data local infile_, added
+compress option for mysql connection by default (#172).
+
+* Fixed bug when inserting hashes in array tuples mode.
+
+* Changed SQLite adapter to catch RuntimeError raised when executing a
+statement and raise Error::InvalidStatement with the offending SQL and
+error message (#188).
+
+* Fixed Dataset#reverse to not raise for unordered dataset (#189).
+
+* Added Dataset#unordered method and changed #order to remove order if
+nil is specified (#190).
+
+* Fixed reversing order of ASC expression (#164).
+
+* Added support for :null => true option when defining table columns
+(#192).
+
+* Fixed Symbol#method_missing to accept variable arity (#185).
+
diff -pruN 1.5.1-1/doc/release_notes/1.4.0.txt 3.0.0-1/doc/release_notes/1.4.0.txt
--- 1.5.1-1/doc/release_notes/1.4.0.txt	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/doc/release_notes/1.4.0.txt	2009-02-19 18:14:12.000000000 +0000
@@ -0,0 +1,53 @@
+Eager loading for all types of associations:
+
+  Artist.eager(:albums).all
+  Album.eager(:artist, :genre, :tracks).all
+  Album.eager(:artist).eager(:genre).eager(:tracks).all
+  Album.filter(:year=>2008).eager(:artist).all
+
+Eager loading supports cascading to an unlimited depth, and doesn't have
+any aliasing issues:
+
+  Artist.eager(:albums=>:tracks).all
+  Artist.eager(:albums=>{:tracks=>:genre}).all
+
+Unfortunately, eager loading comes at the expense of a small amount of
+backward compatibility.  If you were using uncached associations (the
+default in sequel_model 0.5), they no longer work the same way.  Now,
+all associations act as if :cache=>true (which is now set for all
+associations, so if you wrote a tool that worked with both cached and
+uncached associations, it should still work).
+
+One to many associations now populate the corresponding many to one
+instance variable (even when eagerly loaded):
+
+  # Assuming: Album.one_to_many :tracks
+  album = Album.first
+  # This following code is only one query,
+  # not a query for the album and one for each track
+  album.tracks.each{|t| puts t.album.name}
+
+ActiveRecord style has_many :through associations are now supported via
+many_to_many.  many_to_many will no longer select the entire result set,
+just the columns of the associated table (and not the join table), so it
+works for both has_and_belongs_to_many (simple join table) and has_many
+:through (join table model) scenarios.  If you want to include all or
+part of the join table attributes, see the :select option for
+many_to_many associations.
+
+We reduced the number of gems from three (sequel, sequel_core,
+sequel_model) to two (sequel, sequel_core).  Basically, sequel_model is
+now just sequel, and the old sequel gem metapackage is no longer.  There
+isn't a reason to have a gem metapackage for two gems when one
+(sequel_model) depends on the other (sequel_core).  This required a
+version bump for the model part of sequel from 0.5.0.2 to 1.4.0 (since
+the previous sequel gem version was 1.3).
+
+Sequel 1.4.0 has fixes for 11 trackers issues, including fixes to the
+MySQL and PostgreSQL adapters.
+
+We have switched the source control repository for Sequel from Google
+Code (which uses subversion) to github (which uses git).  If you would
+like to contribute to Sequel, please fork the github repository, make
+your changes, and send a pull request.  As before, posting patches on
+the Google Code issue tracker is fine as well.
diff -pruN 1.5.1-1/doc/release_notes/1.5.0.txt 3.0.0-1/doc/release_notes/1.5.0.txt
--- 1.5.1-1/doc/release_notes/1.5.0.txt	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/doc/release_notes/1.5.0.txt	2009-02-19 18:13:46.000000000 +0000
@@ -0,0 +1,155 @@
+You can now graph a dataset and have the result split into component
+tables:
+
+  DB[:artists].graph(:albums, :artist_id=>:id).first
+  # => {:artists=>{:id=>artists.id, :name=>artists.name}, \
+  # :albums=>{:id=>albums.id, :name=>albums.name,
+:artist_id=>albums.artist_id}}
+
+This aliases columns if necessary so they don't stomp on each other,
+which
+is what usually happens if you just join the tables:
+
+  DB[:artists].left_outer_join(:albums, :artist_id=>:id).first
+  # => {:id=>(albums.id||artists.id),
+:name=>(albums.name||artist.names), \
+        :artist_id=>albums.artist_id}
+
+Models can use graph as well, in which case the values will be model
+objects:
+
+  Artist.graph(Album, :artist_id=>:id)
+  # => {:artists=>#<Artist...>, :albums=>#<Album...>}
+
+Models can now eager load via .eager_graph, which will load all the
+results
+and all associations in a single query.  This is necessary if you want
+to
+filter on columns in associated tables.  It works exactly the same way
+as
+.eager, and supports cascading of associations as well:
+
+  # Artist.one_to_many :albums
+  # Album.one_to_many :tracks
+  # Track.many_to_one :genre
+  Artist.eager_graph(:albums=>{:tracks=>:genre}).filter( \
+    :tracks_name=>"Firewire").all
+
+This will give you all artists have have an album with a track named
+"Firewire", and calling .albums on one of those artists will only return
+albums that have a track named "Firewire", and calling .tracks on one of
+those albums will return only the track(s) named "Firewire".
+
+You can use set_graph_aliases to select specific columns:
+
+  DB[:artists].graph(:albums, :artist_id=>:id).set_graph_aliases( \
+    :artist_name=>[:artists, :name], :album_name=>[:albums,
+:name]).first
+  # => {:artists=>{:name=>artists.name}, :albums=>{:name=>albums.name}}
+
+You can use eager_graph with set_graph_aliases to have eager loading
+with
+control over the SELECT clause.
+
+All associations now update their reciprocal associations whenever the
+association methods are used, so you don't need to refresh the
+association or model to have the reciprocal association updated:
+
+  Album.many_to_one :band
+  Band.one_to_many :albums
+
+  # Note that all of these associations are cached,
+  # so after the first access there are no additional
+  # database queries to fetch associated records.
+
+  # many_to_one setter adds to reciprocal association
+  band1.albums # => []
+  album1.band = band1
+  band1.albums # => [album1]
+  band2.albums # => []
+  album1.band = band2
+  band1.albums # => []
+  band2.albums # => [album1]
+  album1.band = band2
+  band2.albums # => [album1]
+  album1.band = nil
+  band2.albums # => []
+
+  # one_to_many add_* method sets reciprocal association
+  # one_to_many remove_* method removes reciprocal association
+  album1.band # => nil
+  band1.add_album(album1)
+  album1.band # => band1
+  band2.add_album(album1)
+  album1.band # => band2
+  band2.remove_album(album1)
+  album1.band # => nil
+
+  Post.many_to_many :tags
+  Tag.many_to_many :posts
+
+  # many_to_many add_* method adds to reciprocal association
+  # many_to_many remove_* method removes from reciprocal association
+  post1.tags # => []
+  tag1.posts # => []
+  tag1.add_post(post1)
+  post1.tags # => [tag1]
+  tag1.posts # => [post1]
+  tag1.remove_post(post1)
+  post1.tags # => []
+  tag1.posts # => []
+  post1.add_tag(tag1)
+  post1.tags # => [tag1]
+  tag1.posts # => [post1]
+  post1.remove_tag(tag1)
+  post1.tags # => []
+  tag1.posts # => []
+
+The MySQL and PostgreSQL adapters now support index types:
+
+  index :some_column, :type => :hash # or :spatial, :full_text, :rtree,
+etc.
+
+Starting in Sequel 1.5.0, some methods are deprecated.  These methods
+will be
+removed in Sequel 2.0.0.  The deprecation framework is fairly flexible.
+You
+can choose where the messages get sent:
+
+  Sequel::Deprecation.deprecation_message_stream = STDERR # the default
+  Sequel::Deprecation.deprecation_message_stream = \
+    File.new('deprecation.txt', 'wb') # A file
+  Sequel::Deprecation.deprecation_message_stream = nil # ignore the
+messages
+
+You can even have all deprecation messages accompanied by a traceback,
+so you
+can see exactly where in your code you are using a deprecated method:
+
+  Sequel::Deprecation.print_tracebacks = true
+
+All deprecation methods come with an message telling you what
+alternative code
+will work.
+
+In addition to deprecating some methods, we removed the ability to have
+arrays returned instead of hashes.  The array code still had debugging
+messages
+left it in, and we are not aware of anyone using it.  Hashes have been
+returned
+by default since Sequel 0.3.
+
+We have also removed the Numeric date/time extensions (e.g. 3.days.ago).
+The
+existing extensions were incomplete, better ones are provided elsewhere,
+and the extensions were not really related to Sequel's purpose.
+
+Sequel no longer depends on ParseTree, RubyInline, or ruby2ruby.  They
+are
+still required to use the block filters.  Sequel's only gem dependency
+is on
+the tiny metaid.
+
+Sequel 1.5.0 has fixes for 12 tracker issues, including fixes to the
+Informix,
+MySQL, ODBC, ADO, JDBC, Postgres, and SQLite adapters.
diff -pruN 1.5.1-1/doc/release_notes/2.0.0.txt 3.0.0-1/doc/release_notes/2.0.0.txt
--- 1.5.1-1/doc/release_notes/2.0.0.txt	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/doc/release_notes/2.0.0.txt	2009-02-19 18:12:55.000000000 +0000
@@ -0,0 +1,298 @@
+Blockless Filter Expressions
+----------------------------
+
+Before 2.0.0, in order to specify complex SQL expressions, you
+either had to resort to writing the SQL yourself in a string or
+using an expression inside a block that was parsed by ParseTree.
+Because ParseTree was required, only ruby 1.8.* was supported, and
+supporting other ruby versions (ruby 1.9, JRuby, Rubinius) would
+never be possible.
+
+With 2.0.0, you no longer need to use a block to write complex SQL
+expressions. The basics of the blockless filters are the usual
+arithmetic, inequality, and binary operators:
+
+  + = addition
+  - = subtraction
+  * = multiplication
+  / = division
+  > = greater than
+  < = less than
+  >= = greater than or equal to
+  <= = less than or equal to
+  ~ = negation
+  & = AND
+  | = OR
+
+You can use these operators on Symbols, LiteralStrings, and other
+Sequel::SQL::Expressions.  Note that there is no equal operator or
+not equal operator, to specify those, you use a Hash.
+
+Here are some examples:
+
+  # Ruby code => SQL WHERE clause
+  :active => active
+  ~:active => NOT active
+  ~~:active => active
+  ~~~:active => NOT active
+  :is_true[] => is_true()
+  ~:is_true[] => NOT is_true()
+  :x > 100 => (x > 100)
+  :x < 100.01 => (x < 100.01)
+  :x <= 0 => (x <= 0)
+  :x >= 1 => (x >= 1)
+  ~(:x > 100) => (x <= 100)
+  {:x => 100} => (x = 100)
+  {:x => 'a'} => (x = 'a')
+  {:x => nil} => (x IS NULL)
+  ~{:x => 100} => (x != 100)
+  ~{:x => 'a'} => (x != 'a')
+  ~{:x => nil} => (x IS NOT NULL)
+  {:x => /a/} => (x ~ 'blah') # Default, MySQL different
+  ~{:x => /a/} => (x !~ 'blah') # Default, MySQL different
+  :x.like('a') =>  (x LIKE 'a')
+  ~:x.like('a') =>  (x NOT LIKE 'a')
+  :x.like(/a/) => (x ~ 'a') # Default, MySQL different
+  ~:x.like('a', /b/) => ((x NOT LIKE 'a') AND (x !~ 'b')) # Default
+  ~{:x => 1..5} => ((x < 1) OR (x > 5))
+  ~{:x => DB[:items].select(:i)} => (x NOT IN (SELECT i FROM items))
+  ~{:x => [1,2,3]} => (x NOT IN (1, 2, 3))
+  :x + 1 > 100 => ((x + 1) > 100)
+  (:x * :y) < 100.01 => ((x * y) < 100.01)
+  (:x - :y/2) >= 100 => ((x - (y / 2)) >= 100)
+  (((:x - :y)/(:x + :y))*:z) <= 100 => ((((x - y) / (x + y)) * z) <=
+100)
+  ~((((:x - :y)/(:x + :y))*:z) <= 100) => ((((x - y) / (x + y)) * z) >
+100)
+  :x & :y => (x AND y)
+  :x & :y & :z => ((x AND y) AND z)
+  :x & {:y => :z} => (x AND (y = z))
+  {:y => :z} & :x => ((y = z) AND x)
+  {:x => :a} & {:y => :z} => ((x = a) AND (y = z))
+  (:x > 200) & (:y < 200) => ((x > 200) AND (y < 200))
+  :x | :y => (x OR y)
+  :x | :y | :z => ((x OR y) OR z)
+  :x | {:y => :z} => (x OR (y = z))
+  {:y => :z} | :x => ((y = z) OR x)
+  {:x => :a} | {:y => :z} => ((x = a) OR (y = z))
+  (:x > 200) | (:y < 200) => ((x > 200) OR (y < 200))
+  (:x | :y) & :z => ((x OR y) AND z)
+  :x | (:y & :z) => (x OR (y AND z))
+  (:x & :w) | (:y & :z) => ((x AND w) OR (y AND z))
+  ~((:x | :y) & :z) => ((NOT x AND NOT y) OR NOT z)
+  ~((:x & :w) | (:y & :z)) => ((NOT x OR NOT w) AND (NOT y OR NOT z))
+  ~((:x > 200) | (:y & :z)) => ((x <= 200) AND (NOT y OR NOT z))
+  ~('x'.lit + 1 > 100) => ((x + 1) <= 100)
+  'x'.lit.like(/a/) => (x ~ 'a') # (x ~ \'a\')
+
+None of these require blocks, you can use any directly in a call to
+filter:
+
+  DB[:items].filter((:price * :tax) - :discount > 100)
+  # => SELECT * FROM items WHERE (((price * tax) - discount) > 100)
+  DB[:items].filter(:active & ~:archived)
+  # => SELECT * FROM items WHERE (active AND NOT archived)
+
+SQL String Concatenation
+------------------------
+
+Sequel now has support for expressing SQL string concatenation in an
+easy way:
+
+  [:name, :title].sql_string_join(" - ")
+  # SQL:  name || ' - ' || title
+
+You can use this in selecting columns, creating filters, ordering
+datasets, and possibly elsewhere.
+
+Schema Reflection Support/Typecasting on Assignment
+---------------------------------------------------
+
+When used with PostgreSQL, MySQL, or SQLite, Sequel now has the
+ability to get information from the database's schema in regards
+to column types:
+
+  DB.schema(:artist)
+  => [[:id, {:type=>:integer, :db_type=>"integer", :max_chars=>0
+     :numeric_precision=>32, :allow_null=>false,
+     :default=>"nextval('artist_id_seq'::regclass)"}], [:name,
+     {:type=>:string, :default=>nil, :db_type=>"text",
+     :numeric_precision=>0, :allow_null=>true, :max_chars=>0}]]
+
+Models now use this information to typecast values on attribute
+assignment. For example, if you have an integer column named number
+and a text (e.g. varchar) column named title:
+
+  1.5.1:
+    model.number = '1'
+    model.number # => '1'
+    model.title = 1
+    model.title # => 1
+  2.0.0:
+    model.number = '1'
+    model.number # => 1
+    model.title = 1
+    model.title # => '1'
+
+Typecasting can be turned off on a global, per class, and per object
+basis:
+
+  Sequel::Model.typecast_on_assignment = false # Global
+  Album.typecast_on_assignment = false # Per Class
+  Album.new.typecast_on_assignment = false # Per Object
+
+Typecasting is somewhat strict, it does not allow obviously bogus
+data to be used:
+
+  model.number = 'a' # Raises error
+
+This is in contrast to how some other ORMs handle the situation:
+
+  model.number = 'a'
+  model.number # => 0
+
+If Sequel is being used with a web framework and you want to display
+friendly error messages to the user, you should probably turn
+typecasting off and set up the necessary validations in your models.
+
+Model Association Improvements
+------------------------------
+
+Associations can now be eagerly loaded even if they have a block,
+though the block should not rely on being evaluated in the context
+of an instance. This allows you filter on associations when eagerly
+loading:
+
+  Artist.one_to_many :albums_with_10_tracks, :class=>:Album do |ds|
+    ds.filter(:num_tracks => 10)
+  end
+  Artist.filter(:name.like('A%)).eager(:albums_with_10_tracks).all
+  # SELECT * FROM artists WHERE (name LIKE 'A%')
+  # SELECT albums.* FROM albums WHERE ((artist_id IN (...)) AND
+  #   (num_tracks = 10))
+
+Associations now have a remove_all_ method for removing all
+associated objects in a single query:
+
+  Artist.many_to_many :albums
+  Artist[1].remove_all_albums
+  # DELETE FROM albums_artists WHERE artist_id = 1
+
+  Artist.one_to_many :albums
+  Artist[1].remove_all_albums
+  # UPDATE albums SET artist_id = NULL WHERE artist_id = 1
+
+All associations can specify a :select option to change which columns
+are selected.  Previously only many to many associations suppported
+this.
+
+The SQL used when eagerly loading through eager_graph can be
+modified via the :graph_join_type, :graph_conditions, and
+:graph_join_conditions options.
+
+:graph_join_type changes the join type from the default of
+:left_outer.  This can be useful if you do not want any
+albums that don't have an artist in the result set:
+
+  Album.many_to_one :artist, :graph_join_type=>:inner
+  Album.eager_graph(:artist).sql
+  # SELECT ... FROM albums INNER JOIN artists ...
+
+:graph_conditions adds conditions on the join to the table you are
+joining, the eager_graph equivalent of an association block argument
+in eager. It takes either a hash or an array where all elements
+are arrays of length two, similar to join_table, where key symbols
+specify columns in the joined table and value symbols specify
+columns in the last joined or primary table:
+
+  Album.many_to_one :artist, :graph_conditions=>{:active=>true}
+  Album.eager_graph(:artist).sql
+  # SELECT ... FROM albums LEFT OUTER JOIN artists ON ((artists.id =
+  #   albums.artist_id) AND (artists.active = 't'))
+
+:graph_join_table_conditions exists for many to many associations only,
+and operates the same as :graph_conditions, except it specifies a
+condition on the many to many join table instead of the associated
+model's table. This is necessary if the join table is also model
+table with other columns on which you may want to filter:
+
+  Album.many_to_many :genres, :join_table=>:ag, \
+    :graph_join_table_conditions=>{:active=>true}
+  Album.eager_graph(:genres).sql
+  # SELECT ... FROM albums LEFT OUTER JOIN ag ON ((ag.album_id =
+      albums.id) AND (ag.active = 't')) LEFT OUTER JOIN genres ON
+      (genres.id = ag.genre_id)
+
+Other Small Improvements
+------------------------
+
+* Dataset#invert returns a dataset that matches all records not
+  matching the current filter.
+* Dataset#unfiltered returns a dataset that has any filters removed.
+* Dataset#last_page? and Dataset#first_page? for paginated datasets.
+* The sequel command line tool now support an -E or --echo argument
+  that logs all SQL to the standard output.  It also can take a path
+  to a yaml file with database connection options, in addition to a
+  database URL.
+* Databases can now have multiple SQL loggers, so you can log to the
+  standard output as well as a file.
+* SQL identifiers (columns and tables) are now quoted by default (you
+  can turn this off via Sequel.quote_identifiers = false if need be).
+* Sequel.connect now takes an optional block that will disconnect the
+  database when the block finishes.
+* AlterTableGenerator now has add_primary_key and add_foreign_key
+  methods.
+* Running the specs without ParseTree installed skips the specs that
+  require ParseTree.
+* You can use an array of arrays instead of a hash when specifying
+  conditions, which may be necessary in certain situations where
+  you would be using the same hash key more than once.
+* Almost all documentation for Sequel was updated for 2.0.0, so if you
+  found Sequel documentation lacking before, check out the new RDoc
+  pages.
+* There have been many minor refactoring improvements, the code
+  should now be easier to read and follow.
+* Sequel now has no external dependencies.
+* Sequel::Models now have before_validation and after_validation
+  hooks.
+* Sequel::Model hooks that return false cause the methods that call
+  them (such as save) to return false.
+* Sequel::Models can now load their schema on first instantiation,
+  instead of when they are created, via
+  Sequel::Model.lazy_load_schema=.  This is helpful for certain
+  web frameworks that reload all models on every request.
+* Hook methods that use blocks can now include an optional tag,
+  which allows them to work well with web frameworks that load source
+  files every time they are modified.
+
+The PostgreSQL adapter has been rewritten and now supports ruby-pg.
+There have also been improvements in the following adapters: DBI,
+MySQL, SQLite, Oracle, and MSSQL.
+
+All of the methods that have been deprecated in 1.5.0 have now been
+removed.  If you are want to upgrade to Sequel 2.0.0 from version 1.4.0
+or previous, upgrade to 1.5.1 first, fix all of the deprecation
+warnings that show up, and then upgrade to 2.0.0.
+
+There were some backwards incompatible changes made in 2.0.0 beyond the
+removal of deprecated methods.  These are:
+
+* Inflector is no longer used, the inflection methods were moved
+  directly into String (where they belong because inflections only
+  make sense for strings).  So to override singularization
+  or pluralization rules, use String.inflections instead of
+  Inflector.inflections.
+* MySQL tinyints are now returned as boolean values instead of
+  integers.  MySQL doesn't have a boolean type, and usually it
+  is recommended to use tinyint for a boolean column.
+* You can no longer pass an array to Dataset#order or Dataset#select,
+  you need to pass each argument separately (the * operator is your
+  friend).
+* You must use '?' instead of '(?)' when interpolating an array
+  argument into a string (e.g. filter('x IN ?', [1,2,3]))
+* You must pass an explicit table alias argument to join_table and
+  related methods, you can no longer include the table alias
+  inside the table argument.
+* sqlite:// URLs now operate the same as file:// URLs (2 slashes
+  for a relative path, 3 for an absolute path).
+
diff -pruN 1.5.1-1/doc/release_notes/2.10.0.txt 3.0.0-1/doc/release_notes/2.10.0.txt
--- 1.5.1-1/doc/release_notes/2.10.0.txt	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/doc/release_notes/2.10.0.txt	2009-02-19 18:07:12.000000000 +0000
@@ -0,0 +1,328 @@
+New Supported Adapters and Databases
+------------------------------------
+
+* A DataObjects adapter was added that supports PostgreSQL,
+  MySQL, and SQLite.  DataObjects is the underlying database
+  library used by DataMapper, and has potential performance
+  advantages by doing all typecasting in C.
+
+* A Firebird Adapter was added, it requires the modified Ruby
+  Fb adapter found at http://github.com/wishdev/fb.
+
+* An H2 JDBC subadapter was added, based on the code used in JotBot.
+  H2 is an embeddable Java database, and may be preferable to using
+  SQLite on JDBC because SQLite requires native code.
+
+New Core Features
+-----------------
+
+* Sequel now has database independent migrations.  Before, column
+  types in migrations were not translated per database, so it was
+  difficult to set up a migration that worked on multiple databases.
+  Sequel now accepts ruby classes as database types, in addition to
+  symbols and strings.  If a ruby class is used, it is translated
+  to the most appropriate database type.  Here is an example using
+  all supported classes (with Sequel's default database type):
+
+    DB.create_table(:cats) do
+      primary_key :id, :type=>Integer # integer
+      String :a                       # varchar(255)
+      column :b, File                 # blob
+      Fixnum :c                       # integer
+      foreign_key :d, :other_table, :type=>Bignum # bigint
+      Float :e                        # double precision
+      BigDecimal :f                   # numeric
+      Date :g                         # date
+      DateTime :h                     # timestamp
+      Time :i                         # timestamp
+      Numeric :j                      # numeric
+      TrueClass :k                    # boolean
+      FalseClass :l                   # boolean
+    end
+
+  Type translations were tested on the PostgreSQL, MySQL, SQLite,
+  and H2 databases.  The default translations should work OK for
+  most databases, but there will probably be a type or two that
+  doesn't work.  Please send in a patch if Sequel uses a column type
+  that doesn't work on your database.
+
+  Note that existing migrations still work fine, in most cases.  If
+  you were using strings or symbols for types before, they should
+  still work.  See the Backwards Compatibility section below for
+  details.
+
+  Also note that this doesn't relate solely to migrations, as any
+  database schema modification method that accepts types will
+  accept one of the above classes.
+
+* A ton of internal work was done to better support databases that
+  fold unqouted identifiers to uppercase (which is the SQL standard).
+  Sequel now allows you to set a method to call on identifiers going
+  both into and out of the database.  The default is to downcase
+  identifiers coming out, and upcase identifiers going in, though
+  this is overridden by the PostgreSQL, MySQL, and SQLite adapters
+  to not do anything (since they fold to lowercase by default).
+
+  The settings are called identifier_input_method and
+  identifier_output_method, and like most Sequel settings, they can
+  be set globally, per database, or per dataset:
+
+    # Global (use uppercase in ruby and lowercase in the database)
+    Sequel.identifier_input_method = :downcase
+    Sequel.identifier_output_method = :upcase
+    # Per Database (use camelized names in the database, and
+    #               underscored names in ruby)
+    DB.identifier_input_method = :camelize
+    DB.identifier_output_method = :underscore
+    # Per Dataset (obfuscate your database columns!)
+    class String; def rot_13; tr('A-Za-z', 'N-ZA-Mn-za-m') end end
+    ds = DB[:table]
+    ds.identifier_input_method = :rot_13
+    ds.identifier_output_method = :rot_13
+
+* Schema parsing support was added to the JDBC adapter, using the
+  JDBC metadata methods.  This means that models that use the
+  JDBC adapter will typecast data in their column setters and
+  automatically select the correct primary key column(s).  This is
+  currently the only adapter that supports schema parsing when using
+  an MSSQL or Oracle database.
+
+* Database#create_table now takes options, which you can use to
+  specify a MySQL engine, charset, and/or collation.  You can also
+  set a default engine, charset, and collation for MySQL to use:
+
+    Sequel::MySQL.default_engine = 'InnoDB'
+    Sequel::MySQL.default_charset = 'utf8'
+    Sequel::MySQL.default_collate = 'utf8'
+
+  The defaults will be used if the options are not provided.  If a
+  default engine is set, you can specify :engine=>nil to not use it
+  (same goes for charset and collate).
+
+* The Sequel::DatabaseConnectionError exception class was added.  It
+  is raised by the connection pool if there is an error attempting
+  to instantiate a database connection.  Also, if the adapter returns
+  nil instead of raising an error for faulty connection parameters,
+  DatabaseConnectionError will be raised immediately, instead of the
+  connection pool busy waiting until if gives up with a
+  PoolTimeoutError.
+
+* Database#tables is now supported on the JDBC adapter, returning
+  an Array of table name symbols.
+
+* Sequel now converts the following Java types returned by the JDBC
+  adapter into ruby types: Java::JavaSQL::Timestamp,
+  Java::JavaSQL::Time, Java::JavaSQL::Date,
+  Java::JavaMath::BigDecimal, and Java::JavaIo::BufferedReader.
+
+* When using the PostgreSQL adapter with the postgres-pr driver,
+  Sequel will use a custom string escaping routine unless
+  force_standard_strings = false.  This means that using Sequel's
+  defaults, postgres-pr will correctly escape strings now.
+
+* The SQLite adapter now returns float, real, and double precision
+  columns as Floats.
+
+* The SQLite adapter logs beginning, committing, and rolling back
+  transactions.
+
+* Sequel now has an internal version (before, the only way to tell
+  the version was to look at the gem being used).  It is accessible
+  at Sequel.version.
+
+New Model Features
+------------------
+
+* A new validates_not_string validation was added for Sequel Models.
+  It is intended to be used with the raise_on_typecast_failure =
+  false setting.  In this case, for a non-string database column,
+  if there is a string value when the record is going to be
+  saved, it is due to the fact that Sequel was not able to typecast
+  the given data correctly (so it is almost certainly not valid).
+  This should make Sequel easier to use with web applications.
+
+* An :allow_missing validation option was added to all standard
+  validations.  This option skips the validation if the attribute
+  is not in the object's values.  It is different from :allow_nil,
+  which will skip the value if it is present but nil in the values.
+  The intended use case for this option is when the database provides
+  a good default.  If the attribute is not present in values, the
+  database will use its default.  If the attribute is present in
+  the values but equals nil, Sequel will attempt to insert it into
+  the database as a NULL value, instead of using the database's
+  default.  If you don't want Sequel to insert a NULL value in the
+  database, but you want the database to provide the default, this
+  is the option to use.
+
+* validates_each now accepts :allow_nil and :allow_blank options,
+  so it is easier to create custom validations with the same options
+  as the standard validations.
+
+* Before_* hooks now run in the reverse order that they were added.
+  The purpose of hooks is to wrap existing functionality, and making
+  later before_* hooks run before previous before_* hooks is the
+  correct behavior.
+
+* You can now add you own hook types, via Model.add_hook_type.  This
+  is intended for plugin use.  All of the standard hooks are now
+  implemented using this method.
+
+* The value of new? in a after_save hook now reflects the
+  previous state of the model (so true for a create and false for an
+  update), instead of always being false.  This makes it easier
+  to have a complex after_save hook that still needs to
+  differentiate between a newly created record and an updated record,
+  without having to add separate after_create and after_update
+  hooks.
+
+* The value of changed_columns in an after_update hook now reflects
+  the value before the update occurred, instead of usually being
+  empty.  Previously, to have this functionality, you generally had
+  to save the value to an instance variable in a before_update hook
+  so you could reference it in the after_update hook.
+
+Other Improvements
+------------------
+
+* Sequel now longer overwrites the following Symbol instance methods
+  when running on ruby 1.9: [], <, <=, >, and >=.  One of Sequel's
+  principals is that it does not override methods defined by ruby,
+  and now that ruby 1.9 defines the above methods on Symbol, Sequel
+  shouldn't be overwriting them.
+
+  Sequel already provides a way to work around this issue when
+  another library adds the same methods to Symbol that Sequel does.
+  For example, you need to change the following:
+
+    dataset.filter(:number > 1)
+    dataset.filter(:number >= 2)
+    dataset.filter(:name < 'M')
+    dataset.filter(:name <= 'I')
+    dataset.filter(:is_bool[:x])
+
+  To:
+
+    dataset.filter{|o| o.number > 1}
+    dataset.filter{|o| o.number >= 2}
+    dataset.filter{|o| o.name < 'M'}
+    dataset.filter{|o| o.name <= 'I'}
+    dataset.filter{|o| o.is_bool(:x)}
+
+  The argument provided to the block is a Sequel::SQL::VirtualRow.
+  This class uses method_missing so that any methods called on it
+  return Sequel::SQL::Identifiers (if no arguments are provided)
+  or Sequel::SQL::Function (if arguments are provided).
+
+  If you were using one of the above symbol methods outside of a
+  filter, you can to call sql_string, sql_number, or sql_function
+  on the symbol.  So the following would also work:
+
+    dataset.filter(:number.sql_number > 1)
+    dataset.filter(:number.sql_number >= 2)
+    dataset.filter(:name.sql_string < 'M')
+    dataset.filter(:name.sql_number <= 'I')
+    dataset.filter(:is_bool.sql_function(:x))
+
+  Using the block argument makes for a nicer API, though, so I
+  recommend using it when possible.
+
+  Note that if you are running ruby 1.8 or jruby without the --1.9
+  flag, you don't need to worry.  If you are running ruby 1.9 or
+  jruby --1.9, or you plan to at some point in the future, you
+  should inspect your code for existing uses of these methods.
+  Here are a couple command lines that should find most uses:
+
+    # Find :symbol[]
+    egrep -nr ':['\''"]?[a-zA-Z_0-9]*['\''"]?\[' *
+    # Find :symbol (<|>|<=|>=)
+    egrep -nr '[^:]:['\''"]?[a-zA-Z_0-9]*['\''"]? *[<>]=?' *
+
+* Database#quote_identifiers now affects future schema modifications
+  when using the database.  Previous, it only affected future
+  schema modifications if a schema modification method had not yet
+  been called.
+
+* Literalization of Times and DateTimes is now correct when using the
+  MySQL JDBC subadapter.
+
+* Literalization of Blobs is now correct when using the PostgreSQL
+  JDBC subadapter.
+
+* Index and table names are quoted when creating indices in the
+  PostgreSQL adapter.
+
+* Dataset#delete was changed in the SQLite adapter to add a
+  where clause that is always true, instead of doing an explicit
+  count first and the deleting.  This is simpler, though it
+  could potentially have performance implications.
+
+* The sequel command line tool now supports symbol keys and unnested
+  hashes in YAML files, so it should work with Merb's database.yml.
+  It also includes the error class in the case of an error.
+
+* The integration type tests were greatly expanded.  Generally,
+  running the integration tests is a good way to determine how well
+  your database is supported.
+
+* Dataset#quote_identifier now returns LiteralStrings as-is, instead
+  of treating them as regular strings.
+
+* Sequel no longer modifies the MySQL::Result class when using the
+  MySQL adapter.
+
+Backwards Compatibilty
+----------------------
+
+* If you were previously using a database that returned uppercase
+  identifiers, it will probably return lowercase identifiers by
+  default now.  To get back the old behavior:
+
+    DB.identifier_output_method = nil
+
+* The module hierarchy under Sequel::SQL has changed.  Now,
+  modules do not include other modules, and the following modules
+  were removed since they would have been empty after removing
+  the modules they included: Sequel::SQL::SpecificExpressionMethods
+  and Sequel::SQL::GenericExpressionMethods.
+
+* Sequel no longer assumes the public schema by default when
+  connecting to PostgreSQL.  You can still set the default
+  schema to use (even to public).
+
+* The ability to load schema information for all tables at once
+  was removed from the PostgreSQL adapter.  While it worked, it had
+  some issues, and it was difficult to keep it working when some
+  new features were used.  This ability wasn't exposed to the user,
+  and was purely an optimization.  If you have any code like:
+
+    DB.schema
+
+  by itself after the Database object was instantiated, you should
+  remove it.
+
+* The Database#primary_key API changed in the PostgreSQL shared
+  adapter, it now accepts an options hash with :server and :conn keys
+  instead of a server symbol.  Also, quite a few private Database
+  instance methods changed, as well as some constants in the
+  AdapterMethods.
+
+* It is possible that some migrations will break, though it is
+  unlikely.  If you were using any of the classes mentioned above
+  as a method inside a migration, it might be broken.  However,
+  since String, Float, and Integer wouldn't have worked as methods
+  before, it is unlikely that anyone used this.
+
+* The meaning of #String, #Integer, and #Float inside
+  Sequel::SQL::Generator (i.e. inside a Database#create_table
+  block) has changed.  Before, these used to call private Kernel
+  methods, now, they set up columns with the appropriate database
+  type.
+
+* The Database#lowercase method in the DBI adapter was removed,
+  as its use case is now met by the identifier_output_method support.
+
+* Database#uri is now aliased explicitly via a real method, to
+  allow for easier subclassing.
+
+* You can no longer pass nil as the second argument to
+  Database#create_table.
diff -pruN 1.5.1-1/doc/release_notes/2.1.0.txt 3.0.0-1/doc/release_notes/2.1.0.txt
--- 1.5.1-1/doc/release_notes/2.1.0.txt	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/doc/release_notes/2.1.0.txt	2009-02-19 18:12:09.000000000 +0000
@@ -0,0 +1,271 @@
+Model Improvements
+------------------
+
+* one_to_many/many_to_many associations now support a :limit option,
+  adding a limit/offset to the records returned.  This was possible
+  before using a block, so it is just added for convenience.
+
+* Associations now support a :read_only option, which doesn't create
+  methods that modify the database.
+
+* Associations now support a :graph_select option, which allows
+  specifying the columns of associated models to include when using
+  eager_graph.
+
+* one_to_many associations now have a :one_to_one option.  When used
+  it creates a getter and setter method similar to many_to_one.  This
+  fills the same role as ActiveRecord's has_one, but it is
+  implemented as a couple of convenience methods over one_to_many, so
+  it still requires that you specify the association name as a
+  plural.
+
+* Model datasets now have to_hash augmented so that it can be called
+  without any arguments, in which case it yields an identity map (a
+  hash with keys being primary key values and values being model
+  instances).
+
+* The Model.set_sti_key method was added, for easily setting up
+  single table inheritance.  It should be called only in the parent
+  class.
+
+* Calls to def_dataset_method with a block are now cached and
+  reapplied to the new dataset if set_dataset is called afterward,
+  or in a subclass.
+
+* All validation methods can now be made conditional via an :if
+  option, which takes either a symbol (which specifies an instance
+  method) or a proc (which is instance_evaled).
+
+* Model#set and Model#update have been added back, they are now
+  aliases of #set_with_params and #update_with_params.
+
+* Models now have set_only/set_except/update_only/update_except
+  instance methods that take a hash (like you would provide to
+  set or update) and additional arguments specifying which columns
+  to allow or disallow.
+
+* Models now have a set_allowed_columns and set_restricted_columns
+  methods, which operate similarly to ActiveRecord's attr_accessible
+  and attr_protected.  It is recommend that you use the set_only or
+  update_only instead of these methods, though.  You can ignore the
+  allowed or restricted columns by using #set_all or #update_all.
+
+* The primary key column(s) is restricted by default. To allow it to
+  be set via new/set/update, use:
+
+    Sequel::Model.unrestrict_primary_key # Global
+    Artist.unrestrict_primary_key # Per Class
+
+* It is now easy to override the one_to_many/many_to_many association
+  methods that modify the database (add_/remove_/remove_all_), as
+  they have been broken into two methods, one that handles the
+  caching features and a private one (prepended with an _) that
+  handles the database changes (and which you can easily override
+  without worrying about the caching).
+
+Table Joining
+-------------
+
+Dataset#join_table got a nice overhaul.  You can now use any join
+type your database allows:
+
+  DB[:artist].join_table(:natural, :albums)
+  DB[:numbers].join_table(:cross, :numbers)
+
+You can now specify the conditions as
+
+* String: "a.b = c.d" # ON a.b = c.d
+* Expression :x < :y # ON x < y
+* Array of Symbols: [:x, :y, :z] # USING (x, y, z)
+* nil # no conditions, used for NATURAL or CROSS joins
+
+Dataset#join_table also takes a block that yields three arguments:
+
+* join_table_alias - The alias/name of the table currently being
+  joined
+* last_join_table_alias - The alias name of the last table joined
+  (if there was one) or the first FROM table (if not).
+* joins - An array of JoinClause objects for all previous joins in
+  the query.
+
+Using the block you can specify conditions for complex joins without
+needing to know in advance what table aliases will be used.
+
+Expanded SQL Syntax Support
+---------------------------
+
+SQL Case statements are now supported directly using hashes or
+arrays:
+
+  {:x > 1 => 1}.case(0)
+  # CASE WHEN x > 1 THEN 1 ELSE 0 END
+  [[{:x=>1}, 0], [:x < 1, 1], [:x > 1, 2]].case(-1)
+  # CASE WHEN x = 1 THEN 0 WHEN x < 1 THEN 1 WHEN x > 1 THEN 2
+    ELSE -1 END
+
+You should use an array instead of a hash for multiple conditions
+unless all conditions are orthogonal.
+
+The SQL extract function has special syntax:
+
+  EXTRACT(day FROM date)
+
+This syntax is now supported via the following ruby code:
+
+  :date.extract(:day)
+
+Other Notable Changes
+---------------------
+
+* The sequel command line tool can now run migrations.  The -m
+  option specifies the directory holding the migration files,
+  and the -M options specifies the version to which to migrate.
+
+* The PostgreSQL adapter supports nested transactions/savepoints.
+
+* The schema parser now understands decimal fields, and will
+  typecast to BigDecimal.
+
+* PostgreSQL's numeric type is now recognized and returned as
+  BigDecimal.
+
+* HAVING now comes before ORDER BY, which most databases seem to
+  prefer.  If your database wants HAVING after ORDER BY, please
+  let us know.
+
+* Symbol#qualify now exists, to specify the table name for a given
+  symbol, similar to the use of #as to specify an alias.  This is
+  mainly helpful in conjuction with the #join_table block, as that
+  provides the table aliases to use to qualify the columns inside
+  the block.
+
+* BitwiseMethods (&, |, ^, ~, <<, >>) have been added to the
+  NumericExpression class, so you can do the following:
+
+  (x + 1) ^ 10  # SQL: (x + 1) ^ 10
+  ~(x + 1) # SQL: ~(x + 1)
+
+  Usually, &, |, and ~ operate in a logical manner, but for
+  NumericExpressions, they take on their usual bitwise meaning,
+  since logical operations only make sense for booleans.
+
+* #cast_numeric and #cast_string exist for Symbols, Strings, and
+  other Sequel Expressions, which return the results casted and
+  wrapped in either NumericExpression or StringExpression, so you
+  can use the BitwiseMethods (&, |, ^, ~, <<, >>) or
+  StringConcatenationMethods (+) directly.
+
+# Dataset#to_hash can take only one argument, in which case it uses
+  that argument to specify the key, and uses the entire hash for the
+  value.
+
+# Dataset#graph can now take an array of columns to select from the
+  joined table via the :select option.
+
+# Dataset#filter and similar methods now combine the block and
+  regular argument conditions if both are given, instead of ignoring
+  the regular argument conditions.
+
+# Dataset#filter(false) can now be used to make sure that no records
+  are returned.  Dataset#filter(true) also works, but it's a no-op.
+  Before, these raised errors.
+
+# Dataset#count does a subquery for a dataset using DISTINCT, since
+  the otherwise it would yield a count for the query without
+  DISTINCT.
+
+ParseTree Support Officially Deprecated
+---------------------------------------
+
+The support for ParseTree-based block filters has officially been
+deprecated and will be removed in Sequel 2.2.  To use the
+expression filters (which don't require ParseTree) inside blocks,
+use:
+
+  SEQUEL_NO_PARSE_TREE = true
+  require 'sequel'
+  # OR
+  require 'sequel'
+  Sequel.use_parse_tree = false
+
+This is the default if ParseTree cannot be loaded.  If ParseTree
+can be loaded, it remains the default, in order not to immediately
+break existing code.
+
+With this set, you can use the expression filters inside of blocks:
+
+  dataset.filter{((:x + 1) & 10 < :y) & :z}
+
+That doesn't gain you all that much, but there are some methods
+that feed block arguments into filter, such as the following:
+
+  dataset.first(5){((:x + 1) & 10 < :y) & :z}
+
+Which will get you the first 5 records matching the condition.
+
+Backwards Incompatible Changes
+------------------------------
+
+* To change the datetime classe used from Time to DateTime, you
+  now use:
+
+    Sequel.datetime_class = DateTime # instead of Sequel.time_class
+
+* Models now raise errors if you try to access a missing or
+  restricted method via new/set/update, instead of just silently
+  skipping that parameter.  To get the old behavior:
+
+    Sequel::Model.strict_param_setting = false
+
+* The association_dataset method now takes into account the :eager
+  option and the block argument, where it didn't before. It also
+  takes into account the new :limit option.
+
+* Association methods now raise errors in most cases if the model
+  doesn't have a valid primary key.
+
+* Dataset#join_table used to allow a symbol as a conditions argument
+  as a shortcut for a hash:
+
+    DB[:artist].join(:albums, :artist_id)
+    # ON albums.artist_id = artists.id
+
+  With the changes to #join_table, this no longer works.  It would
+  now be interpreted as a boolean column:
+
+    DB[:artist].join(:albums, :artist_id)
+    # ON artists.id
+
+  Use the following slightly longer version for the old behavior:
+
+    DB[:artist].join(:albums, :artist_id=>:id)
+    # ON albums.artist_id = artists.id
+
+* MySQL users need to be careful when upgrading, the following code
+  will once again cause an error:
+
+    DB[:artists].each{|artist| DB[:albums].each{|album| ...}}
+
+  To fix it, change the code to:
+
+    DB[:artists].all{|artist| DB[:albums].each{|album| ...}}
+
+  The issue is the MySQL adapter doesn't release the database
+  connection while running each, and the second call to each gets the
+  same database connection (while the other query is still running),
+  because it is in the same thread.  Using #all for the outside query
+  ensures that the database connection is released before the block is
+  called.
+
+  The reason for this change was that the workaround provided for
+  MySQL could potentially cause issues with transactions for all
+  adapters.
+
+* String#asc and String#desc are no longer defined, as ordering on a
+  plain string column should be a no-op.  They are still defined
+  on LiteralStrings.
+
+* You can no longer abuse the SQL::Function syntax to use a table
+  alias with specified columns (e.g. :table[:col1, :col2, :col3])
+  or to cast to types (e.g. :x.cast_as(:varchar[20])).  Use a
+  LiteralString in both cases.
diff -pruN 1.5.1-1/doc/release_notes/2.11.0.txt 3.0.0-1/doc/release_notes/2.11.0.txt
--- 1.5.1-1/doc/release_notes/2.11.0.txt	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/doc/release_notes/2.11.0.txt	2009-03-02 18:06:36.000000000 +0000
@@ -0,0 +1,215 @@
+Optimizations
+-------------
+
+* Model.[] was optimized to use static SQL in cases where doing so
+  should result in the same output.  This should result in a 30-40%
+  performance increase.  Since this can be the most significant or
+  only method call in a web application action, this has potential
+  to significantly enhance the performance of web application code.
+  
+  In order for this optimization to have an effect, you need to
+  make sure that you are calling set_dataset with a Symbol and
+  not a Dataset object:
+  
+    # Optimized:
+    class Foo < Sequel::Model; end
+    class Foo < Sequel::Model(:foos); end
+    class Foo < Sequel::Model
+      set_dataset :foos
+    end
+    
+    # Not Optimized, but otherwise equivalent:
+    class Foo < Sequel::Model(Model.db[:foos]); end
+    class Foo < Sequel::Model
+      set_dataset db[:foos]
+    end
+  
+* Dataset#literal was refactored for performance reasons to make
+  overriding it in subclasses unnecessary.  The changes made result
+  in a 20-25% performance increase.  Sequel can spend about 10% of
+  it's time in Dataset#literal, so this may be only a 2% overall
+  performance improvement.
+
+New Features
+------------
+
+* Association datasets now know about the model objects that created
+  them, as well as the related association reflection.  This makes
+  association extensions much more powerful.  For example, you can
+  now create generic association extensions such as:
+  
+    module FindOrCreate
+      def find_or_create(vals)
+        first(vals) || association_reflection.associated_class. \
+         create(vals.merge(association_reflection[:key]=> \
+           model_object.id))
+      end
+    end
+    
+  The above will work for any standard one_to_many association:
+  
+    Artist.one_to_many :albums, :extend=>FindOrCreate
+    # Create an album named Foo related to this artist,
+    # unless such an album already exists
+    Artist.first.albums_dataset.find_or_create(:name=>'Foo')
+    
+  Before, the only way to do the above was to use a closure inside
+  the :dataset option proc, which couldn't be done generically
+  for multiple associations.
+  
+* A :conditions association option was added, which allows simple
+  filters to be set up without defining :graph_conditions and
+  an association block:
+  
+    # 2.10.0
+    one_to_many(:japanese_verses, :class=>:Verse, \
+     :graph_conditions=>{:languageid=>3})do |ds|
+       ds.filter(:languageid=>3)
+    end
+    # 2.11.0
+    one_to_many(:japanese_verses, :class=>:Verse, \
+       :conditions=>{:languageid=>3})
+
+* A :clone association option was added, which allows you to clone
+  an existing association.  This is most useful when you are dealing
+  with a legacy schema and had to define the same options redundantly
+  for each type of association. You can now do:
+  
+    many_to_many :deputies, :class=>:Employee, \
+     :join_table=>:employeecurrentaudits, :left_key=>:currentauditid, \
+     :right_key=>:employeeid, :order=>[:firstname, :lastname] do |ds|
+      ds.filter(:active).filter(:capacity=>1)
+    end
+    many_to_many :project_managers, :clone=>:deputies do |ds|
+      ds.filter(:active).filter(:capacity=>2)
+    end
+    many_to_many :team_leaders, :clone=>:deputies do |ds|
+      ds.filter(:active).filter(:capacity=>3)
+    end
+    
+  All of the above would use the same :class, :join_table, :left_key,
+  :right_key, and :order options.  If you don't provide an
+  association block, but you are cloning an association that has one,
+  the cloned association's block is used.  You can use the
+  :block=>nil option to not use a block even if the cloned
+  association has a block.
+
+* Dataset#select, #select_more, #order, #order_more, and #get all
+  take a block that yields a Sequel::SQL::VirtualRow instance,
+  similar to the behavior of filter. This allows for the easier
+  use of SQL functions on Ruby 1.9:
+  
+    # 2.10.0
+    dataset.select(:prettify.sql_function(:name))
+    # 2.11.0
+    dataset.select{|o| o.prettify(:name)}
+
+* String#lit can now accept arguments and return an SQL literal
+  string.  This allows you to do things that were previously hard
+  or at least unnecessarily verbose.  For example, you can now
+  easily use the SQL standard SUBSTRING function:
+  
+    column = :user
+    pattern = params[:pattern]
+    dataset.select{|o| o.substring('? from ?'.lit(column, pattern))}
+    
+* A validates_inclusion_of validation method was added to Model. You
+  can provide a Range or an Array in the :in option to specify the
+  allowed values:
+  
+    validates_inclusion_of :value, :in=>1..5
+    validates_inclusion_of :weekday, :in=>%w'Monday Tuesday ...'
+
+* Dataset#with_sql was added, which returns a copy of the dataset
+  with static SQL.  This is useful if you want to keep the same
+  row_proc/graph/transform/etc., but want to use your own custom
+  SQL.
+
+Other Improvements
+------------------
+
+* You can now use Sequel's database independent types when casting:
+
+    dataset.select(:number.cast(String))
+  
+  Among other things, the default cast types for cast_string and
+  cast_numeric now work in the MySQL adapter.
+
+* Model#set_associated_object was added.  The many_to_one association
+  setter method calls it.  This allows you to easily override the
+  association setters for all many_to_one associations of a class
+  by modifying a single method. 
+  
+* Typecasting invalid date strings now raises a
+  Sequel::Error::InvalidValue instead of an argument error, which
+  means that you can use raise_on_typecast_failure = false and not
+  have an error raised when an invalid date format is used.
+  
+* String#to_sequel_blob was added and should now be used instead
+  of String#to_blob.  sqlite3-ruby defines String#to_blob
+  differently, which could cause problems.
+  
+* Blob columns are now fully supported in the SQLite adapter, with
+  the hex escape syntax being used for input, and returning columns
+  of type Sequel::SQL::Blob on output.
+  
+* The SQLite adapter drop_column support is now significantly more
+  robust.
+  
+* The SQLite adapter now supports rename_column.
+  
+* The MySQL adapter now supports stored procedures with multiple
+  arguments.
+  
+* The MySQL adapter can now not use a compressed connection to the
+  server via the :compress=>false option.
+  
+* The MySQL adapter now sets a default timeout of 30 days to the
+  database connection, you can change it via the :timeout option,
+  which accepts a number of seconds.
+  
+* The MySQL adapter now sets SQL_AUTO_IS_NULL to false by default,
+  you can use the :auto_is_null=>true option to not do this.
+  
+* The MySQL adapter now sets the encoding option on the database
+  connection itself, so it works across reconnects.
+
+* Sequel itself no longer uses String#lit or Symbol#* internally, so
+  it shouldn't break if another library defines them.
+
+* The default index name is now generated correctly if a non-String
+  or Symbol column is used.
+  
+* Some ruby -w warnings have been fixed.
+
+* INSERTs are now sent to the master database instead of the slave
+  database(s) if using a master/slave database configuration and
+  PostgreSQL 8.2+ or Firebird.
+  
+* DateTime literalization has been fixed in the Firebird adapter.
+  
+* Date literalization has been fixed in the H2 JDBC subadapter.
+  
+* Release notes for versions from 1.0 to the present are now included
+  in the Sequel repository and the RDoc documentation, see
+  http://sequel.rubyforge.org/rdoc/files/doc/release_notes/
+
+Backwards Compatibilty
+----------------------
+
+* The optimization of Model.[] may break if you modify the model's
+  dataset behind its back.  Always use Model.set_dataset if you
+  want to change a Model's dataset.
+  
+* Sequel::Dataset::UnsupportedExceptIntersect and
+  Sequel::Dataset::UnsupportedExceptIntersectAll will now only be
+  defined if you are using an adapter that requires them.
+  
+* The private Model#cache_delete_unless_new method has been removed.
+
+* Sequel::SQL::IrregularFunction was removed, as it was a bad hack
+  that is not used by Sequel anymore.  Unless you were instantiating
+  it directly or using a plugin/extension that did, this shouldn't
+  affect you.  Using a Sequel::SQL::Function with a 
+  Sequel::SQL::PlaceholderLiteralString is recommended instead, see
+  the substring example above.
diff -pruN 1.5.1-1/doc/release_notes/2.12.0.txt 3.0.0-1/doc/release_notes/2.12.0.txt
--- 1.5.1-1/doc/release_notes/2.12.0.txt	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/doc/release_notes/2.12.0.txt	2009-04-03 17:57:30.000000000 +0100
@@ -0,0 +1,534 @@
+Overview
+--------
+
+Sequel 2.12 is really just a stepping stone to Sequel 3.0, which will
+be released next month.  All major changes currently planned for 3.0
+have been made in 2.12, but 2.12 contains many features that have
+been deprecated and will be removed or moved into extensions or
+plugins in 3.0.
+
+Deprecation Logging
+-------------------
+
+If you use a deprecated method or feature, Sequel will by default
+print a deprecation message and 10 lines of backtrace to standard
+error to easily allow you to figure out which code needs to be
+updated.  You can change where the deprecation messages go and how
+many lines of backtrace are given using the following:
+
+  # Log deprecation information to a file
+  Sequel::Deprecation.output = File.open('deprecated.txt', 'wb')
+  
+  # Use 5 lines of backtrace when logging deprecation messages
+  Sequel::Deprecation.backtraces = 5
+  
+  # Use all backtrace lines when logging deprecation messages
+  Sequel::Deprecation.backtraces = true
+  
+  # Don't include backtraces in the deprecation logging
+  Sequel.Deprecation.backtraces = false
+  
+  # Turn off all deprecation logging
+  Sequel::Deprecation.output = nil
+
+Deprecated Features Moving to Extensions
+----------------------------------------
+
+* Migrations are being moved into sequel/extensions/migration.  There
+  isn't any reason that they should be loaded in normal use since
+  they are used so rarely.  The sequel command line tool uses this
+  extension to run the migrations.
+  
+* Adding the blank? method to all objects has been moved into
+  sequel/extensions/blank.
+  
+* Dataset#print and Sequel::PrettyTable have been moved into
+  sequel/extensions/pretty_table.
+  
+* Dataset#query and related methods have been moved into
+  sequel/extensions/query.
+  
+* Dataset#paginate and related methods have been moved into
+  sequel/extensions/pagination.
+  
+* String inflection methods (e.g. "people".singularize) have been
+  moved into sequel/extensions/inflector.
+  
+* String date/time conversion methods (e.g. '2000-01-01'.to_date)
+  have been moved into sequel/extensions/string_date_time.
+
+Deprecated Model Features Moving to Plugins
+-------------------------------------------
+
+* Model validation class methods have been moved to a plugin.  Sequel
+  users are encouraged to write their own validate instance method
+  instead. A new validation_helpers plugin has been added to make
+  this easier, it's explained in the New Features section.  If you
+  want to continue using the validation class methods:
+  
+    Sequel::Model.plugin :validation_class_methods
+  
+* Model hook class methods have been moved to a plugin.  Sequel users
+  are encouraged to write their own hook instance methods, and call
+  super to get hooks specified in superclasses or plugins.  If you
+  want to continue using the hook class methods:
+  
+    Sequel::Model.plugin :hook_class_methods
+  
+* Model schema methods (e.g. Model.set_schema, Model.create_table, 
+  Model.drop_table) have been moved to a plugin.  The use of these
+  methods has been discouraged for a long time.  If you want to use
+  them:
+  
+    Sequel::Model.plugin :schema
+
+* Model.set_sti_key has been moved to a plugin.  So you should
+  change:
+
+    MyModel.set_sti_key :key_column
+
+  to:
+  
+    MyModel.plugin :single_table_inheritance, :key_column
+    
+* Model.set_cache has been moved to a plugin.  So you should change:
+
+    MyModel.set_cache cache_store, opts
+    
+  to:
+  
+    MyModel.plugin :caching, cache_store, opts
+    
+* Model.serialize has been moved to a plugin.  So you should change:
+
+    MyModel.serialize :column, :format=>:yaml
+    
+  to:
+  
+    MyModel.plugin :serialization, :yaml, :column
+    
+  Because the previous serialization support depended on dataset
+  transforms, the new serialization support is implemented
+  differently, and behavior may not be identical in all cases.
+  However, this should be a drop in replacement for most users.
+
+Deprecated Features To Be Removed in Sequel 3.0
+-----------------------------------------------
+
+* Dataset#transform is deprecated without any replacement planned.
+  It was announced on the Sequel mailing list that transforms would
+  be removed unless someone said they needed them, and nobody said
+  that they did.
+
+* Dataset#multi_insert and Dataset#import are no longer aliases
+  of each other.  Dataset#multi_insert now takes an array of hashes,
+  and Dataset#import now takes an array of columns and an array
+  of arrays of values.  Using multi_insert with import's API or
+  vice-versa is deprecated.
+  
+* Calling Dataset#[] with no arguments or an integer argument is
+  deprecated.
+  
+* Calling Dataset#map with both an argument and a block is
+  deprecated.
+  
+* Database#multi_threaded? and Database#logger are both deprecated.
+
+* Calling Database#transaction with a symbol to specify which server
+  to use is deprecated.  You should now call it with an option hash
+  with a :server key.
+  
+* Array#extract_options! and Object#is_one_of? are both deprecated.
+
+* The metaprogramming methods taken from metaid are deprecated and
+  have been moved into Sequel::Metaprogramming.  If you want them
+  available to specific objects/classes, just include or extend with
+  Sequel::Metaprogramming.  If you want all objects to have access to
+  the metaprogramming methods, install metaid.  Note that the
+  class_def method from metaid doesn't exist in
+  Sequel::Metaprogramming, since it really isn't different from
+  define_method (except it is public instead of private).
+  
+* Module#class_attr_overridable, #class_attr_reader, and
+  #metaalias are deprecated.
+  
+* Using Model#set or #update when the columns for the model are not
+  set and you provide a hash with symbol keys is deprecated.
+  Basically, you must have setter methods now for any columns used in
+  #set or #update.
+  
+* Model#set_with_params and #update_with_params are deprecated, use
+  #set and #update instead.
+  
+* Model#save! is deprecated, use #save(:validate=>false).
+
+* Model.is and Model.is_a are deprecated, use Model.plugin.
+
+* Model.str_columns, Model#str_columns, #set_values, and
+  #update_values are deprecated.  You should use #set and
+  #update instead of #set_values and #update_values, though they
+  operate differently.
+  
+* Model.delete_all, Model.destroy_all, Model.size, and Model.uniq
+  are deprecated, use .delete, .destroy, .count, and .distinct.
+  
+* Model.belongs_to, Model.has_many, and Model.has_and_belongs_to_many
+  are deprecated, use .many_to_one, .one_to_many, and .many_to_many.
+  
+* Model#dataset is deprecated, use Model.dataset.
+
+* SQL::CastMethods#cast_as is deprecated, use #cast.
+
+* Calling Database#schema without a table argument is deprecated.
+
+* Dataset#uniq is deprecated, use Dataset#distinct.
+
+* Dataset#symbol_to_column_ref is deprecated, use #literal.
+
+* Dataset#quote_column_ref is deprecated, use #quote_identifier.
+
+* Dataset#size is deprecated, use #count.
+
+* Passing options to Dataset#each, #all, #single_record,
+  #single_value, #sql, #select_sql, #update, #update_sql, #delete,
+  #delete_sql, and #exists is deprecated.  Modify the options first
+  using clone or a related method, then call one of the above
+  methods.
+  
+* Dataset#create_view and #create_or_replace_view are deprecated,
+  use the database methods instead.
+
+* Dataset.dataset_classes, #model_classes, #polymorphic_key, and
+  #set_model are deprecated.
+  
+* Database#>> is deprecated.
+
+* String#to_blob and SQL::Blob#to_blob are deprecated, use
+  #to_sequel_blob.
+  
+* The use of Symbol#| to create array subscripts is deprecated,
+  use Symbol#sql_subscript.
+  
+* Symbol#to_column_ref is deprecated, use Dataset#literal.
+
+* String#expr is deprecated, use String#lit.
+
+* Array#to_sql, String#to_sql, and String#split_sql are deprecated.
+
+* Passing an array to Database#<< is deprecated.
+
+* Range#interval is deprecated.
+
+* Enumerable#send_each is deprecated.
+
+* When using ruby 1.8, Hash#key is deprecated.
+
+* Sequel.open is deprecated, use Sequel.connect.
+
+* Sequel.use_parse_tree and Sequel.use_parse_tree= are deprecated.
+
+* All upcase_identifier methods and the :upcase_identifiers database
+  option are deprecated, use identifier_input_method = :upcase
+  instead.
+  
+* Using a virtual row block without an argument is deprecated, see
+  Sequel.virtual_row_instance_eval= under New Features.
+  
+* When using the JDBC adapter, Java::JavaSQL::Timestamp#usec is
+  deprecated.  Sequel has returned Java::JavaSQL::Timestamp as
+  DateTime or Time for a few versions, so this shouldn't affect most
+  people.
+  
+* Sequel will no longer require bigdecimal/util, enumerator, or yaml
+  in 3.0.  If you need them in your code, make sure you require
+  them yourself.  Using features added by requiring these standard
+  libaries will not bring up a deprecation warning, for obvious
+  reasons.
+  
+* Sequel::Error::InvalidTransform, Sequel::Error::NoExistingFilter,
+  and Sequel::Error::InvalidStatement exceptions will be removed in
+  Sequel 3.0.  You will not get a deprecation message if you reference
+  them in 2.12.
+  
+* Sequel::Model::Validation::Errors is deprecated, use
+  Sequel::Model::Errors instead.  Referencing the old name will not
+  bring up a deprecation message.
+
+New Features
+------------
+
+* Sequel.virtual_row_instance_eval= was added, which lets you give
+  Sequel 2.12 the behavior that will be the standard in 3.0.
+  It changes blocks passed to Dataset#filter, #select, or #order that
+  don't accept arguments (or accept any number of arguments) to
+  instance eval the block in the context of a new VirtualRow
+  instance instead of passing a new VirtualRow instance to the block.
+  It allows you to change code that looks like this:
+
+    dataset.filter{|o| (o.number > 10) & (o.name > 'M')}
+
+  to:
+
+    dataset.filter{(number > 10) & (name > 'M')}
+
+  When instance_eval is used, only local variables are available
+  to the block.  Any calls to instance methods will be interpreted
+  as calling VirtualRow#method_missing, which generates identifiers
+  or functions.  When virtual_row_instance_eval is enabled, the
+  following type of code will break:
+
+    # amount is a instance method
+    dataset.filter{:number + amount > 0}
+      
+  Just like this example, the only type of code that should break is
+  when a virtual row block was used when it wasn't necessary (since
+  it doesn't use the VirtualRow argument).
+  
+  When Sequel.virtual_row_instance_eval = false, using a virtual row
+  block that doesn't accept an argument will cause a deprecation
+  message.
+  
+  Here's a regular expression that should catch most places where you
+  are using a virtual row block without an argument.
+
+    egrep -nr '[^A-Za-z0-9_](filter|select|select_more|order|order_more|get|where|having|from|first|and|or|exclude|find|subset|constraint|check)( *(\([^)]*\) *)?){*[^|]' *
+    
+  An RDoc page explaining virtual row blocks was added as well.
+
+* A validation_helpers model plugin was added that allows you to do
+  validations similar to the old class level validations inside
+  the Model#validate instance method.  The API has changed, but it's
+  capable of most of the same validations.  It doesn't handle
+  acceptance_of or confirmation_of validations, as those shouldn't be
+  handled in the model.
+  
+    # Old class level validations
+    validates_format_of :col, :with=>/.../
+    validates_length_of :col, :maximum=>5
+    validates_length_of :col, :minimum=>3
+    validates_length_of :col, :is=>4
+    validates_length_of :col, :within=>3..5
+    validates_not_string :col
+    validates_numericality_of :col
+    validates_numericality_of :col, :only_integer=>true
+    validates_presence_of :col
+    validates_inclusion_of :col, :in=>[3, 4, 5]
+    validates_uniqueness_of :col, :col2
+    validates_uniqueness_of([:col, :col2])
+
+    # New instance level validations
+    def validate
+      validates_format /.../, :col
+      validates_max_length 5, :col
+      validates_min_length 3, :col
+      validates_exact_length 4, :col
+      validates_length_range 3..5, :col
+      validates_not_string :col
+      validates_numeric :col
+      validates_integer :col
+      validates_presence :col
+      validates_includes([3,4,5], :col)
+      validates_unique :col, :col2
+      validates_unique([:col, :col2])
+    end
+    
+  Another change made is to specify the same type of validation on
+  multiple attributes, you must use an array:
+
+    # Old
+    validates_length_of :name, :password, :within=>3..5
+
+    # New
+    def validate
+      validates_length_range 3..5, [:name, :password]
+    end
+    
+  The :message, :allow_blank, :allow_missing, and :allow_nil options
+  are still respected.  The :tag option is not needed as instance level
+  validations work with code reloading without workarounds. The :if
+  option is also not needed for instance level validations:
+
+    # Old
+    validates_presence_of :name, :if=>:new?
+    validates_presence_of :pass, :if=>{flag > 3}
+
+    # New
+    def validate
+      validates_presence(:name) if new?
+      validates_presence(:pass) if flag > 3
+    end
+    
+  The validates_each also doesn't have an equivalent instance method,
+  since it is much easier to just write your own validation:
+  
+    # Old
+    validates_each(:date) do |o,a,v|
+      o.errors.add(a, '...') unless v > Date.today
+    end
+
+    # New
+    def validate
+      errors.add(:date, '...') unless date > Date.today
+    end
+    
+* MySQL adapter datasets now have on_duplicate_key_update and
+  insert_ignore methods which modify the SQL used to support
+  ON DUPLICATE KEY UPDATE and INSERT INGORE syntax in multi_insert
+  and import.
+  
+* If you use the MySQL native adapter, you can set:
+
+    Sequel::MySQL.convert_invalid_date_time = nil
+    
+  to return dates like "0000-00-00" and times like "25:00:00" as
+  nil values instead of raising an error.  You can also set it
+  to :string to return the values as strings.
+
+* You can now use Sequel without modifying any core classes, by
+  defining a SEQUEL_NO_CORE_EXTENSIONS constant or environment
+  variable.  In 2.12, this may still add some deprecated methods to
+  the core classes, but in 3.0 no methods will be added to the core
+  classes if you use this.
+
+* You can now use Sequel::Model without the associations
+  implementation by defining a SEQUEL_NO_ASSOCIATIONS constant or
+  environment variable.
+
+Other Improvements
+------------------
+
+* Model column accessors have been made faster and the overhead of
+  creating them has been reduced significantly.
+
+* ~{:bool_col=>true} now generates an bool_col IS NOT TRUE filter
+  instead of bool_col != TRUE.  This makes it return records with
+  NULL values.  If you only want to have false records, you should
+  use {:bool_col=>false}.  This works better with SQL's 3 valued
+  boolean logic.
+  
+  It is slightly inconsistent with ~{:col=>1}, since that won't
+  return values where col is NULL, but it gives the user the ability
+  to create an IS [NOT] (TRUE|FALSE) filter, which Sequel previously
+  did not support.
+  
+  If you really want the old behavior, you can change it to 
+  ~{true=>:bool_col}.
+  
+* Model.use_transactions was added for setting whether model objects
+  should use transactions when destroying or saving records.  Like
+  most Sequel options, it's settable on a global, per model, and
+  per object basis:
+  
+    Sequel::Model.use_transactions = false
+    MyModel.use_transactions = true
+    my_model.use_transactions = false
+    
+  You can also turn it on or off for specific save calls:
+  
+    my_model.save(:transaction=>true)
+    
+* The Oracle adapter now supports schema parsing.
+    
+* When using Model.db=, all current dataset options are copied to
+  a new dataset created with the new db.
+
+* Model::Errors#count was refactored to improve performance.
+
+* Most exception classes that were located under Sequel::Error are
+  now located directly under Sequel.  The old names are not
+  deprecated (unless mentioned above), but their use is now
+  discouraged.  The exceptions have the same name except that
+  Sequel::Error::PoolTimeoutError changed to Sequel::PoolTimeout.
+
+* Dataset#where now always affects the WHERE clause.  Before, it
+  was just an alias of filter, so it modified the HAVING clause
+  if the dataset already had a HAVING clause.
+
+* The optimization of Model.[] introduced in 2.11.0 broke on
+  databases that didn't support LIMIT.  The optimization now works
+  on those databases.
+
+* All of the the RDoc documentation was reviewed and many updates
+  were made, resulting in significantly improved documentation
+  quality.
+  
+* Model.def_dataset_method now works when the model doesn't have an
+  associated dataset, as it will add the method to a dataset
+  given to set_dataset in the future.
+  
+* Database#get and #select now take a block that is passed to
+  the dataset they create.
+  
+* You can disable the use of INSERT RETURNING in the shared
+  PostgreSQL adapter using disable_insert_returning.  This is mostly
+  useful if you are inserting a large number of records.
+  
+* A bug relating to aliasing columns in the JDBC adapter has been
+  fixed.
+
+* Sequel can now create and drop schema-qualified views.
+  
+* Performance of Dataset#destroy for model datasets was improved.
+
+* The specs now run on Rspec 1.2.
+
+* Internal dependence on the methods that Sequel adds to core
+  classes has been eliminated, any internal use of methods that
+  Sequel adds to the core classes is now considered a bug.
+  
+* A possible bug where Database#rename_table would not remove a
+  cached schema entry has been fixed.
+  
+* The Oracle and MySQL adapters now raise an error as soon as you
+  call distinct on a dataset, instead of waiting until the SQL is
+  generated.
+
+Backwards Compatibilty
+----------------------
+
+* Saving a newly inserted record in an after_create or after_save
+  hook is no longer broken.  It broke in 2.10 as a side effect of
+  allowing the hook to check whether or not the record was a new
+  record.  The code has been changed so that a @was_new instance
+  variable will be defined to true if the record was just created.
+  
+  Similarly, instead of not modifying changed_columns until after
+  the after hooks run, a @columns_updated instance variable will
+  be available in the after hooks that is a hash of exactly what
+  attribute keys and values were used in the update.
+  
+  These changes break compatibility with 2.11.0 and 2.10.0, but
+  restore compatibility with 2.9.0 and previous versions.
+  
+* PostgreSQL no longer uses savepoints for nested transactions by
+  default.  If you want to use a savepoint, you have to pass the
+  :savepoint option to the transaction method.  Using savepoints
+  by default broke expectations when a method raised Rollback to
+  rollback the transaction, and it only rolled back to the last
+  savepoint.
+
+* The anonymous model classes created by Sequel::Model() are now
+  stored in Model::ANONYMOUS_MODEL_CLASSES instead of the @models
+  class instance variable of the main module.
+  
+* The mappings of adapter schemes to classes are now stored in
+  Sequel::ADAPTER_MAP instead of the Database @@adapters class
+  variable.
+
+* Model instances no longer contain a reference to their class's
+  @db_schema.
+
+* Database schema sql methods (e.g. alter_table_sql) are now private.
+
+* Database#[] no longer accepts a block.  It's not possible to call
+  it with a block in general usage, anyway.
+  
+* The Sequel::Schema::SQL module no longer exists, the methods it
+  included were placed directly in the Sequel::Database class.
+
+* The Sequel::SQL::SpecificExpression class has been removed,
+  subclasses now inherit from Sequel::SQL::Expression.
+
+* Sequel now requires its own files with an absolute path.
+
+* The file hierarchy of the sequel library changed significantly.
diff -pruN 1.5.1-1/doc/release_notes/2.2.0.txt 3.0.0-1/doc/release_notes/2.2.0.txt
--- 1.5.1-1/doc/release_notes/2.2.0.txt	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/doc/release_notes/2.2.0.txt	2009-02-19 18:11:22.000000000 +0000
@@ -0,0 +1,253 @@
+The Most Powerful and Flexible Associations of Any Ruby ORM
+-----------------------------------------------------------
+
+Sequel can now support any association type supported by
+ActiveRecord, and many association types ActiveRecord doesn't
+support.
+
+Association callbacks (:before_add, :after_add, :before_remove,
+:after_remove) have been added, and work for all association
+types.  Each of the callback options can be a Symbol specifying an
+instance method that takes one argument (the associated object), or a
+Proc that takes two arguments (the current object and the associated
+object), or an array of Symbols and Procs.  Additionally, an
+:after_load callback is available, which is running after loading the
+associated record(s) from the database.
+
+Association extensions are now supported:
+
+  class FindOrCreate
+    def find_or_create(vals)
+      first(vals) || create(vals)
+    end
+  end
+  class Author < Sequel::Model
+    one_to_many :authorships, :extend=>FindOrCreate
+  end
+  Author.first.authorships_dataset.find_or_create(:name=>'Bob')
+
+Sequel has been able to support most has_many :through style
+associations since 1.3, via many_to_many (since it doesn't break on
+join tables that are also model tables, unlike ActiveRecord's
+has_and_belongs_to_many).  Now it can also support has_many :through
+style associations where it goes through a has_many association.
+
+Sequel can now support polymorphic associations.  Polymorphic
+associations are really a design flaw, so Sequel doesn't support them
+directly, but the tools that Sequel gives you make them pretty easy
+to implement.
+
+Sequel can also support associations that ActiveRecord does not.  For
+example, a belongs_to association where the column referenced in the
+associated table is not the primary key, an association that depends
+on multiple columns in each table, or even situations where the
+association has a column in the primary table that can be referenced
+by any of multiple columns in a second table that has a has_one style
+association with the table you want to associate with.
+
+Some of those associations can be supported for a single object using
+custom SQL in ActiveRecord, but none are supported when eager
+loading or allow further filtering.
+
+Not only can all of these cases be supported with Sequel::Model, all
+can be supported with eager loading, and can allow for further
+filtering. See
+http://sequel.rubyforge.org/files/sequel/doc/advanced_associations_rdoc.html
+for details and example code for all association types covered above.
+
+There have also been many additional options added for controlling
+eager loading via eager_graph.  Every part of the SQL JOINs can now
+be controlled via one of the options, so you can use JOIN USING,
+NATURAL JOIN, or arbitrary JOIN ON conditions.
+
+Finally, just to show off the power that Sequel gives you when eager
+loading, here is example code that will eagerly load all descendants
+and ancestors in a tree structure, without knowing the depth of the
+tree:
+
+  class Node < Sequel::Model
+    set_schema do
+      primary_key :id
+      foreign_key :parent_id, :nodes
+    end
+    create_table
+
+    many_to_one :parent
+    one_to_many :children, :key=>:parent_id
+
+    # Only useful when eager loading
+    many_to_one :ancestors, :eager_loader=>(proc do |key_hash, nodes,
+associations|
+      # Handle cases where the root node has the same parent_id as
+primary_key
+      # and also when it is NULL
+      non_root_nodes = nodes.reject do |n|
+        if [nil, n.pk].include?(n.parent_id)
+          # Make sure root nodes have their parent association set to
+nil
+          n.associations[:parent] = nil
+          true
+        else
+          false
+        end
+      end
+      unless non_root_nodes.empty?
+        id_map = {}
+        # Create an map of parent_ids to nodes that have that parent id
+        non_root_nodes.each{|n| (id_map[n.parent_id] ||= []) << n}
+        # Doesn't cause an infinte loop, because when only the root node
+        # is left, this is not called.
+        Node.filter(Node.primary_key=>id_map.keys).eager(:ancestors).all
+do |node|
+          # Populate the parent association for each node
+          id_map[node.pk].each{|n| n.associations[:parent] = node}
+        end
+      end
+    end)
+    many_to_one :descendants, :eager_loader=>(proc do |key_hash, nodes,
+associations|
+      id_map = {}
+      nodes.each do |n|
+        # Initialize an empty array of child associations for each
+parent node
+        n.associations[:children] = []
+        # Populate identity map of nodes
+        id_map[n.pk] = n
+      end
+      # Doesn't cause an infinite loop, because the :eager_loader is not
+called
+      # if no records are returned.  Exclude id = parent_id to avoid
+infinite loop
+      # if the root note is one of the returned records and it has
+parent_id = id
+      # instead of parent_id = NULL.
+      Node.filter(:parent_id=>id_map.keys).exclude(:id=>:parent_id).eager(:descendants).all
+do |node|
+        # Get the parent from the identity map
+        parent = id_map[node.parent_id]
+        # Set the child's parent association to the parent
+        node.associations[:parent] = parent
+        # Add the child association to the array of children in the
+parent
+        parent.associations[:children] << node
+      end
+    end)
+  end
+
+  nodes = Node.filter(:id < 10).eager(:ancestors, :descendants).all
+
+New Adapter Features
+--------------------
+
+* PostgreSQL bytea fields are now fully supported.
+
+* The PostgreSQL adapter now uses the safer connection-specific
+  string escaping if you are using ruby-pg.
+
+* The SQLite adapter supports drop_column and add_index.
+
+* You can now use URL parameters in the connection string, enabling
+  you to connect to PostgreSQL via a socket using
+  postgres://user:password@blah/database?host=/tmp
+
+Other New Features
+------------------
+
+* Dataset#graph now takes a block which it passes to join_table.
+
+* Symbol#identifier has been added, which can be used if another
+  library defines the same operator(s) on Symbol that Sequel defines.
+
+* Filter blocks now yield a VirtualRow instance, which can yield
+  Identifiers, QualifiedIdentifiers, or Functions.  Like
+  Symbol#identifier, this is useful if another library defines the
+  same operator(s) on Symbol that Sequel defines.
+
+* You can now call Model.to_hash to get an identity map for all
+  rows (before this required Model.dataset.to_hash).
+
+* A model that can get it's column information from the schema will
+  set it in the dataset, potentially saving many queries.
+
+* Model.validates_presence_of now works correctly for boolean
+  columns.
+
+Notable Bug Fixes
+-----------------
+
+* Caching now works with Model subclasses.
+
+* Model validation methods now work with source reloading.
+
+* The PostgreSQL adapter no longer raises an Error if you try to
+  insert a record with the primary key already specified.
+
+* Sequel no longer messes with the native MySQL adapter, so you can
+  use Sequel and ActiveRecord with MySQL in the same process.
+
+* Dataset#count now works correctly for limited dataset.
+
+* PostgreSQL Database#transaction method yields a connection, similar
+  to the other adapters.
+
+* Using a hash argument in #distinct, #order, or #group is treated
+  as an expression instead of a column alias.
+
+* Cloned datasets no longer ignore the existing columns unless it is
+  necessary.
+
+* The :quote_identifiers and :single_threaded Database options now
+  work correctly.
+
+Backwards Incompatible Changes
+------------------------------
+
+* ParseTree support, deprecated in 2.1.0, has been removed in 2.2.0.
+  You should use the expression filter syntax instead, perferably
+  without the block (though it can be used inside a block as well).
+  This usually involves the following types of changes:
+
+    filter{:x == :y} => filter(:x => :y)
+    filter{:x << :y} => filter(:x => :y)
+    filter{:x && :y} => filter(:x & :y) # Don't forget about change
+    filter{:x || :y} => filter(:x | :y) #  in operator precedence
+    filter{:x.like?('%blah%')} => filter(:x.like('%blah%'))
+    filter do => filter((:x > 1) & (:y < 2))
+      :x > 1
+      :y < 2
+    end
+
+* Attempts to save an invalid Model instance will raise an error by
+  default.  To revert to returning a nil value, use:
+
+    Sequel::Model.raise_on_save_failure = false # Global
+    Album.raise_on_save_failure = false # Class
+    album = Album.new
+    album.raise_on_save_failure = false # Instance
+
+  Note that before, save would return false where now it returns nil
+  if you disable raising on save failure.
+
+* Dataset#update no longer takes a block, as it's use of the block
+  depended on ParseTree.  With the introduction of the expression
+  syntax in 2.0.0, it's no longer necessary.  You should use a hash
+  with an expression as the value instead:
+
+    DB[:table].update(:column=>:column + 1)
+
+* validates_presence of now considers false as present instead of
+  absent.  This is so it works with boolean columns.
+
+* Dataset#graph ignores any previously selected columns when it is
+  called for the first time.
+
+* Dataset#columns ignores any filtering, ordering, or distinct
+  clauses.  This shouldn't cause issues unless you were using
+  SQL functions with side effects and expecting them to be called
+  when columns was called (unlikely at best).
+
+One significant point of note is that the 2.2.0 release will be the
+last release with both a sequel_core and sequel gem.  Starting
+with 2.3.0 they will be combined into one sequel gem.  You will still
+be able to get just the sequel_core part by requiring 'sequel_core',
+but they will be packaged together.
diff -pruN 1.5.1-1/doc/release_notes/2.3.0.txt 3.0.0-1/doc/release_notes/2.3.0.txt
--- 1.5.1-1/doc/release_notes/2.3.0.txt	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/doc/release_notes/2.3.0.txt	2009-02-19 18:10:53.000000000 +0000
@@ -0,0 +1,88 @@
+JRuby and Ruby 1.9 Officially Supported
+---------------------------------------
+
+Sequel now officially supports JRuby 1.1.3 and Ruby 1.9 (svn revision
+18194 at least).  Using JRuby with the JDBC adapter, PostgreSQL,
+MySQL, and SQLite now enjoy almost full support, though not
+everything works the same as using the native adapter.  Depending on
+what you are doing, it may make sense to use postgres-pr on JRuby
+instead of PostgreSQL-JDBC.
+
+To use the new JDBC support, the database connection string you give
+Sequel is now passed directly to JDBC, here are a few examples:
+
+  Sequel.connect('jdbc:postgresql://host/database?user=*&password=*')
+  Sequel.connect('jdbc:mysql://host/database?user=*&password=*')
+  Sequel.connect('jdbc:sqlite::memory:')
+  Sequel.connect('jdbc:sqlite:relative/path.db')
+  Sequel.connect('jdbc:sqlite:/absolute/path.db')
+
+Single Gem
+----------
+
+Sequel is now distributed as a single gem named sequel, by combining
+the previous sequel_core and sequel gems.  You can still just
+"require 'sequel_core'" if you don't want the model functionality.
+
+Database Adapter Improvements
+-----------------------------
+
+* Dataset#empty? now works using the MySQL adapter.
+
+* The Oracle adapter now works with a nonstandard database port.
+
+* The JDBC adapter should load JDBC drivers automatically for
+  PostgreSQL, MySQL, SQLite, Oracle, and MSSQL.  For PostgreSQL,
+  MySQL, and SQLite, the jdbc-* gem can be used, for the others, you
+  must have the correct .jar in your CLASSPATH.
+
+* The PostgreSQL adapter no longer raises an error when inserting
+  records into a table without a primary key.
+
+* Database#disconnect now works for the ADO adapter.
+
+* The ADO adapter no longer raises an error if the dataset contains
+  no records.
+
+* The ODBC adapter no longer errors when converting ::ODBC::Time
+  values.
+
+Backwards Incompatible Changes
+------------------------------
+
+* Sequel::Worker has been removed.  There are no known users, and the
+  specs caused problems on JRuby.
+
+* Assigning an empty string to a non-string, non-blob model attribute
+  converts it to nil by default.  You can use
+  "Model.typecast_empty_string_to_nil = false" to get the old
+  behavior.  This should make web development with Sequel
+  significantly easier, hopefully at no expense to other uses.
+
+* Database.uri_to_options is now a private class method.
+
+* Model.create_table! now acts the same as Database.create_table!,
+  dropping the table unconditionally and then creating it.  This was
+  done for consistency.  If you are using Model.create_table! in
+  production code, you should change it to
+  "Model.create_table unless Model.table_exists?", otherwise you risk
+  wiping out your production data.  I recommended you use the
+  migration feature instead of Model.set_schema, as that handles
+  altering existing tables.
+
+Other Notable Changes
+---------------------
+
+* Using validates_length_of more than once on the same attribute with
+  different options without a tag no longer causes the first use to
+  be ignored.  This was a side effect of the validation tags added
+  in 2.2.0.
+
+* Other than the adapters, Sequel now has 100% code coverage (line
+  coverage).
+
+* Model#set* methods now return self.
+
+* An integration test suite was added, testing Sequel against a live
+  database with nothing mocked, which helped greatly when testing the
+  new support for JDBC adapters.
diff -pruN 1.5.1-1/doc/release_notes/2.4.0.txt 3.0.0-1/doc/release_notes/2.4.0.txt
--- 1.5.1-1/doc/release_notes/2.4.0.txt	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/doc/release_notes/2.4.0.txt	2009-02-19 18:10:30.000000000 +0000
@@ -0,0 +1,106 @@
+Prepared Statements/Bound Variables
+===================================
+
+Sequel now supports prepared statements and bound variables.  No
+matter which database you are using, Sequel uses exactly the same API.
+To specify placeholders, you use the :$placeholder syntax:
+
+  ds = DB[:items].filter(:name=>:$n)
+
+To use a bound variable:
+
+  ds.call(:select, :n=>'Jim')
+
+This will do the equivalent of selecting records that have the name
+'Jim'.  In addition to :select, you can use :first or :delete. There
+is also support for bound variables when inserting or updating
+records:
+
+  ds.call(:update, {:n=>'Jim', :new_n=>'Bob'}, :name=>:$new_n)
+
+Which will update all records that have the name 'Jim' to have the
+name 'Bob'.
+
+Prepared statement support is very similar to bound variable support,
+except that the statement is first prepared with a name:
+
+  ps = ds.prepare(:select, :select_by_name)
+
+It is then called later with the bound arguments to use:
+
+  ps.call(:n=>'Jim')
+  DB.call(:select_by_name, :n=>'Jim') # same as above
+
+For inserting or updating, the hash to use when inserting or updating
+is given to prepare:
+
+  ps2 = ds.prepare(:update, :update_name, :name=>:$new_n)
+  ps2.call(:n=>'Jim', :new_n=>'Bob')
+
+There is some level of native support for these features in the
+PostgreSQL, MySQL, SQLite, and JDBC adapters.  For other adapters,
+support is emulated, but it shouldn't be too difficult to add native
+support for them.
+
+For more details see:
+http://sequel.rubyforge.org/rdoc/files/doc/prepared_statements_rdoc.html
+
+Read-Only Slave/Writable Master and Database Sharding
+=====================================================
+
+Sequel now has built in support for master/slave database
+configurations, just by setting an option in Sequel.connect:
+
+  DB=Sequel.connect('postgres://master_server/database', \
+    :servers=>{:read_only=>{:host=>'slave_server'}})
+
+That will use slave_server for SELECT queries and master_server for
+other queries.  It's fairly easy to use multiple slaves or even
+multiple masters, examples are included in the link below.
+
+Sharding support requires some code other than the database
+configuration, but is still fairly simple.  For example, to set up
+a 16 shard configuration based on a hex character:
+
+  servers = {}
+  (('0'..'9').to_a + ('a'..'f').to_a).each do |hex|
+    servers[hex.to_sym] = {:host=>"hash_host_#{hex}"}
+  end
+  DB=Sequel.connect('postgres://hash_host/hashes', :servers=>servers)
+
+To set which shard to use for a query, use the Dataset#server method:
+
+  DB[:hashes].server(:a).filter(:hash=>/31337/)
+
+For more details see:
+http://sequel.rubyforge.org/rdoc/files/doc/sharding_rdoc.html
+
+Other Changes
+=============
+
+* The sequel.rubyforge.org website has a new design thanks to boof.
+  The online RDoc is now located at http://sequel.rubyforge.org/rdoc.
+
+* Support was added for anonymous column names in the ADO adapter.
+
+* Better MSSQL support in the ADO, ODBC, and JDBC adapters.  The
+  odbc_mssql adapter has been removed. If you use MSSQL with ODBC,
+  please use the odbc adapter with a :db_type=>'mssql' option.
+
+* The following Sequel::Error exception subclasses were removed:
+  InvalidExpression, InvalidFilter, InvalidJoinType, and WorkerStop.
+
+* Documentation was added for the PostgreSQL, MySQL, SQLite, and
+  JDBC adapters.
+
+* Various internal interfaces were refactored.  For example, if you
+  use an adapter not included with Sequel, it probably won't work
+  until you update it to the new internal API.
+
+* Many low level methods (such as Database#transaction), now take
+  an optional server argument to indicate which server to use.
+
+* Model plugins that have a DatasetMethods module with non-public
+  methods no longer have Model methods created that call those
+  methods.
+
diff -pruN 1.5.1-1/doc/release_notes/2.5.0.txt 3.0.0-1/doc/release_notes/2.5.0.txt
--- 1.5.1-1/doc/release_notes/2.5.0.txt	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/doc/release_notes/2.5.0.txt	2009-02-19 18:10:02.000000000 +0000
@@ -0,0 +1,137 @@
+New Features
+------------
+
+* The values that are used to insert/update records can now be
+  scoped similar to how filter expressions can be scoped.
+  set_defaults is used to set defaults which can be overridden,
+  and set_overrides is used to set defaults which cannot be
+  overridden:
+
+      DB[:t].set_defaults(:x=>1).insert_sql
+      # => INSERT INTO t (x) VALUES (1)
+      DB[:t].set_defaults(:x=>1).insert_sql(:x=>2)
+      # => INSERT INTO t (x) VALUES (2)
+      DB[:t].set_defaults(:x=>1).insert_sql(:y=>2)
+      # => INSERT INTO t (x, y) VALUES (1, 2)
+      DB[:t].set_overrides(:x=>1).insert_sql(:x=>2)
+      # => INSERT INTO t (x) VALUES (1)
+
+  The difference between set_defaults and set_overrides is that
+  with set_defaults, the last value takes precedence, while with
+  set_overrides, the first value takes precedence.
+
+* The schema generators now support creating and altering tables
+  with composite primary and/or foreign keys:
+
+    DB.create_table(:items) do
+      integer :id
+      text :name
+      primary_key [:id, :name]
+      foreign_key [:id, :name], :other_table, \
+        :key=>[:item_id, :item_name]
+    end
+
+    DB.alter_table(:items) do
+      add_primary_key [:id, :name]
+      add_foreign_key [:id, :name], :other_table, \
+        :key=>[:item_id, :item_name]
+    end
+
+* The AlterTableGenerator now supports unique constraints:
+
+    DB.alter_table(:items) do
+      add_unique_constraint [:aaa, :bbb, :ccc], :name => :con3
+    end
+
+* The schema generators now support ON UPDATE (previously, they only
+  supported ON DELETE):
+
+    DB.create_table(:items) do
+      foreign_key :project_id, :projects, :on_update => :cascade
+    end
+
+* When connecting to a PostgreSQL server version 8.2 and higher,
+  Sequel now uses the INSERT ... RETURNING ... syntax, which should
+  speed up row inserts on PostgreSQL.  In addition, Sequel Models
+  use RETURNING * to speed up model object creation.
+
+* You can now validate multiple attributes at once.  This is useful
+  if the combination of two or more attribute values is important,
+  such as checking the uniqueness of multiple columns.
+  validates_uniqueness_of now supports this directly:
+
+    validates_uniqueness_of [:column1, :column2]
+
+  This protects against the database having multiple rows with the
+  same values for both :column1 and :column2.  This is different
+  from:
+
+    validates_uniqueness_of :column1, :column2
+
+  Which checks that the value of column1 is unique in the table, and
+  that the value of column2 is unique in the table (which is much
+  more restrictive).
+
+Other Improvements
+------------------
+
+* Dataset methods insert_sql, delete_sql, and update_sql respect the
+  :sql option, allowing you to do things such as:
+
+    ds = DB['INSERT INTO t (time) VALUES (CURRENT_TIMESTAMP)']
+    ds.insert
+    ds.insert
+
+* The database adapters (at least MySQL, PostgreSQL, SQLite, and
+  JDBC) generally raise Sequel::DatabaseError for database problems,
+  making it easier to tell what is a true database error versus an
+  error raised by Sequel itself.
+
+* Sequel uses the async features of ruby-pg so that the entire
+  interpreter is not blocked while waiting for the results of
+  queries.
+
+* Sequel now supports the 2008.08.17 version of ruby-pg.
+
+* MSSQL support has been improved when using the ODBC and ADO
+  adapters.
+
+* Index names are quoted and creating or dropping indexes.
+
+* Automatically generated column accessor methods no longer override
+  instance methods specified by plugins.
+
+* Inserting a row with an already specified primary key inside a
+  transaction now works correctly when using PostgreSQL.
+
+* before_save and before_update hooks now work as expected when using
+  save_changes.
+
+* count and paginate now work correctly on graphed datasets.
+
+Backwards Compatibility
+-----------------------
+
+* The SQLite adapter now raises Sequel::DatabaseError instead of
+  Sequel::Error::InvalidStatement whenever an SQLite3::Exception is
+  raised by the SQLite3 driver.
+
+* Date and DateTime conversions now convert 2 digit years.  To revert
+  to the previous behavior:
+
+    Sequel.convert_two_digit_years = false
+
+  Note that Ruby 1.8 and 1.9 handle Date parsing differently, so
+  there is no backwards compatibility change for Ruby 1.9.  However,
+  this also means that the MM/DD/YY date syntax commonly used in the
+  United States is not always parsed correctly on Ruby 1.9, greatly
+  limiting the use of 2 digit year conversion.
+
+* You can no longer abuse the SQL function syntax for specifying
+  database types.  For example, you must change:
+
+    :type=>:varchar[255]
+
+  to:
+
+    :type=>:varchar, :size=>255
diff -pruN 1.5.1-1/doc/release_notes/2.6.0.txt 3.0.0-1/doc/release_notes/2.6.0.txt
--- 1.5.1-1/doc/release_notes/2.6.0.txt	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/doc/release_notes/2.6.0.txt	2009-03-23 17:07:10.000000000 +0000
@@ -0,0 +1,157 @@
+New Features
+------------
+
+* Schema parsing was refactored, resulting in a huge speedup when
+  using MySQL.  MySQL now uses the DESCRIBE statement instead of the
+  INFORMATION_SCHEMA.  PostgreSQL now uses the pg_* system catalogs
+  instead of the INFORMATION schema.
+
+* The schema information now includes the :primary_key field.  Models
+  now use this field to automatically determine the primary key for
+  a table, so it no longer needs to be specified explicitly.  Models
+  even handle the composite primary key case.
+
+* The raise_on_typecast_failure switch was added, with it being true
+  by default (so no change in behavior).  This allows the user to
+  silently ignore errors when typecasting fails, at the global, class,
+  and instance levels.
+
+    Sequel::Model.raise_on_typecast_failure = false # Global
+    Artist.raise_on_typecast_failure = true # Class
+    artist = Artist.new
+    artist.raise_on_typecast_failure = false # Instance
+
+    Album.raise_on_typecast_failure = true
+    Album.new(:numtracks=>'a') # => raises Sequel::Error::InvalidValue
+    Album.raise_on_typecast_failure = false
+    Album.new(:numtracks=>'a') # => #<Album @values={:numtracks=>"a"}>
+
+* Associations' orders are now respected when eager loading via
+  eager_graph.  Sequel will qualify the columns in the order with
+  the alias being used, so you can have overlapping columns when
+  eager loading multiple associations.
+
+    Artist.one_to_many :albums, :order=>:name
+    Album.one_to_many :tracks, :order=>:number
+    Artist.order(:artists__name).eager_graph(:albums=>:tracks).sql
+    # => ... ORDER BY artists.name, albums.name, tracks.number
+
+* The support for CASE expressions has been enhanced by allowing the
+  use of an optional expression:
+
+    {1=>2}.case(0, :x)
+    # => CASE x WHEN 1 THEN 2 ELSE 0 END
+    [[:a, 1], [:b, 2], [:c, 3]].case(4, :y)
+    # => CASE y WHEN a THEN 1 WHEN b THEN 2 WHEN c THEN 3 ELSE 4 END
+
+  Previously, to get something equivalent to this, you had to do:
+
+    {{:x=>1}=>2}.case(0)
+    # => CASE WHEN (x = 1) THEN 2 ELSE 0 END
+    [[{:y=>:a}, 1], [{:y=>:b}, 2], [{:y=>:c}, 3]].case(4)
+    # => CASE WHEN (y = a) THEN 1 WHEN (y = b) THEN 2 WHEN (y = c)
+         THEN 3 ELSE 4 END
+
+* You can now change the NULL/NOT NULL value of an existing column
+  using the set_column_allow_null method.
+
+    # Set NOT NULL
+    DB.alter_table(:artists){set_column_allow_null :name, false}
+    # Set NULL
+    DB.alter_table(:artists){set_column_allow_null :name, true}
+
+* You can now get the schema information for a table in a non-public
+  schema in PostgreSQL using the implicit :schema__table syntax.
+  Before, the :schema option had to be given explicitly to
+  Database#schema.  This allows models to get schema information for
+  tables outside the public schema.
+
+* Transactions are now supported on MSSQL.
+
+* Dataset#tables now returns all tables in the database for MySQL
+  databases accessed via JDBC.
+
+* Database#drop_view can now drop multiple views at once.
+
+Other Improvements
+------------------
+
+* The SQLite adapter now respects the Sequel.datetime_class option
+  for timestamp and datetime columns.
+
+* Adding a unique constraint no longer explicity creates a unique
+  index.  If you want a unique index, use index :unique=>true.
+
+* If no language is specified when creating a full text index on
+  PostgreSQL, the simple language is assumed.
+
+* Errors when typecasting fails are now Sequel::Error::InvalidValue
+  instead of the more generic Sequel::Error.
+
+* Specifying constraints now works correctly for all types of
+  arguments.  Previously, it did not work unless a block or
+  interpolated string were used.
+
+* Loading an association with the same name as a table in the FROM
+  clause no longer causes an error.
+
+* When eagerly loading many_to_one associations where no objects have
+  an associated object, the negative lookup is now cached.
+
+* String keys can now be used with Dataset#multi_insert, just like
+  they can be used for Dataset#insert.
+
+* Dataset#join_table now generates the correct SQL when doing the
+  first join to a dataset where the first source is a dataset, when
+  an unqualified column is used in the conditions.
+
+* Cascading associations after *_to_many associations can now be
+  eagerly loaded via eager_graph.
+
+* Eagerly loading *_to_many associations that are cascaded behind a
+  many_to_one association now have their duplicates removed if a
+  cartesian product join is done.
+
+* The SQLite adapter now uses string literals in all of the AS
+  clauses.  While the SQL standard specifies that identifiers should
+  be used, SQLite documentation explicitly states that string
+  literals are expected (though it generally works with identifiers
+  by converting them implicitly).
+
+* Database methods that modify the schema now remove the cached
+  schema entry.
+
+* The hash keys that Database#schema returns when no table is
+  requested are now always supposed to be symbols.
+
+* The generation of SQL for composite foreign keys on MySQL has been
+  fixed.
+
+* A schema.rdoc file was added to the documentation explaining the
+  various parts of Sequel related to schema generation and
+  modification and how they interact
+  (http://sequel.rubyforge.org/rdoc/files/doc/schema_rdoc.html).
+
+* The RDoc template for the website was changed from the default
+  template to the hanna template.
+
+Backwards Compatibility
+-----------------------
+
+* The :numeric_precision and :max_chars schema entries have been
+  removed.  Use the :db_type entry to determine this information,
+  if available.
+
+* The SQLite adapter used to always return Time instances for
+  timestamp types, even if Sequel.datetime_class was DateTime.  For
+  datetime types it always returned a DateTime instance.  It
+  now returns an instance of Sequel.datetime_class in both cases.
+
+* It's possible that the including of associations' orders when eager
+  loading via eager_graph could cause problems.  You can use the
+  :order_eager_graph=>false option to not use the :order option when
+  eager loading via :eager_graph.
+
+* There were small changes in SQL creation where the AS keyword is
+  now used explicitly.  These should have no effect, but could break
+  tests for explicit SQL.
diff -pruN 1.5.1-1/doc/release_notes/2.7.0.txt 3.0.0-1/doc/release_notes/2.7.0.txt
--- 1.5.1-1/doc/release_notes/2.7.0.txt	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/doc/release_notes/2.7.0.txt	2009-02-19 18:09:03.000000000 +0000
@@ -0,0 +1,166 @@
+Performance Optimizations
+-------------------------
+
+* Fetching a large number of records with the PostgreSQL adapter is
+  significantly faster (up to 3-4 times faster than before).
+
+* Instantiating model objects has been made much faster, as many
+  options (such as raise_on_save_failure) are now lazily loaded, and
+  hook methods are now much faster if no hooks have been defined for
+  that type of hook.
+
+New Association Options
+-----------------------
+
+* The :eager_grapher option has been added allowing you to supply
+  your own block to implement eager loading via eager_graph.
+
+* many_to_one and one_to_many associations now have a :primary_key
+  option, specifying the name of the column that the :key option
+  references.
+
+* many_to_many associations now have :left_primary_key and
+  :right_primary_key options, specifying the columns that :left_key
+  and :right_key reference, respectively.
+
+* many_to_many associations now have a :uniq option, that adds an
+  :after_load callback that makes the returned array of objects
+  unique.
+
+Other New Features
+------------------
+
+* Dataset#set_graph_aliases now allows you to supply a third argument
+  for each column you want graph into the dataset, allowing you to use
+  arbitrary SQL expressions that are graphed into the correct table:
+
+    ds.set_graph_aliases!(:a=>[:b, :c], :d=>[:e, :f, 42])
+    # SELECT b.c AS a, 42 AS d FROM ...
+    ds.first # => {:b=>{:c=>?}, :e=>{:f=>42}}
+
+* Dataset#add_graph_aliases was added, that adds additional graph
+  aliases instead of replacing the existing ones (as
+  #set_graph_aliases does).  It's basically the equivalent of
+  select_more for graphs.
+
+* Dataset#join_table changed it's final argument from a symbol
+  specifying a table name to an option hash (with backwards
+  compatibility kept), and adds support for a :implicit_qualifier
+  option, which it uses instead of the last joined table to
+  qualify columns.
+
+* Association's :after_load callbacks are now called when eager
+  loading via eager (but not when eager loading via eager_graph).
+
+* Any expression can now be used as the argument to Symbol#like,
+  which means that you can pattern match columns to other columns.
+  Before, it always transformed the argument to a string.
+
+    :a.like(:b)
+    # 2.6.0: a LIKE 'b'
+    # 2.7.0: a LIKE b
+
+* Array#sql_array was added, allowing you to specify that an array
+  in ruby be treated like an array in SQL.  This is true anyway,
+  except for arrays of all two pairs, which are treated like hashes,
+  for specifying multiple conditions with the same key:
+
+   DB[:foo].filter([:a,:b] => [[1,2],[3,4]].sql_array)
+   # => SELECT * FROM foo WHERE ((a, b) IN ((1, 2), (3, 4)))
+
+* ComplexExpression#== and #sql? were added, allowing for easier
+  testing.
+
+* Full text searching on PostgreSQL now joins multiple columns with
+  a space, to prevent joining border words, and it works when there
+  is a match in one column but the other column is NULL.
+
+Other Improvements
+------------------
+
+* Instance methods added by creating associations are added to an
+  anonymous module included by the class, so they can be overridden
+  in the class while still allowing the use of super to get the
+  default behavior (this is similar to column accessor methods).
+
+* Many improvements were added to support using multiple schemas in
+  PostgreSQL.
+
+* Model::Validation::Errors objects are now more compatible with
+  Rails, by adding a #count method and making #on return nil if there
+  are no error messages for that attribute.
+
+* Serialized columns in models are no longer typecast.
+
+* Associations are now inherited when a model class is subclassed.
+
+* Many improvements were made that should make adding custom
+  association types easier.
+
+* A corner case in eager_graph where the wrong table name would be
+  used to qualify a column name has been fixed.
+
+* Dataset's cached column information is no longer modified if #each
+  is called with an option that modifies the columns.
+
+* You should now be able to connect to Oracle via the JDBC adapter,
+  and with the same support it has when using the oracle adapter.
+
+* Model.association_reflections is now a public methods, so you can
+  grab a hash of all association reflections at once (keyed by
+  association name symbol).
+
+* The :encoding/:charset option now works in the PostgreSQL adapter
+  if the postgres-pr driver is used.
+
+* The numeric(x,y) type is now interpreted as decimal.
+
+Backwards Compatibilty
+----------------------
+
+* The first argument to Model#initialize must be a hash, you can no
+  longer use nil.  For example, the following code will break if
+  :album is not in params:
+
+    Album.new(params[:album])
+
+  Additionally, Model#initialize does not call the block if the
+  second argument is true.
+
+* The Sequel::Model.lazy_load_schema setting was removed.  It should
+  no longer be necessary now that schema loading is relatively speedy,
+  and schemas can be loaded at startup and cached.
+
+* The PostgreSQL adapter will default to using a unix socket in /tmp
+  if no host is specified.  Before, a TCP/IP socket to localhost was
+  used if no host was specified.  This change makes Sequel operate
+  similarly to the PostgreSQL command line tools.
+
+* The ASSOCIATION_TYPES constant has changed from an array to a hash
+  and it has been moved.  The RECIPROCAL_ASSOCIATIONS constant has
+  been removed.  This is unlikely to matter unless you were using
+  custom association types.
+
+* The PostgreSQL adapter now sets the PostgreSQL DateStyle, in order
+  to implement an optimization.  To turn this off, set
+  Sequel::Postgres.use_iso_date_format = false.
+
+* When using the PostgreSQL adapter, in many places the schema is
+  specified explicitly.  If you do not specify a schema, a default
+  one is used (public by default).  If you use a schema other than
+  public for your work, use Database#default_schema= to set it.  For
+  any table outside of the default schema, you should specify the
+  schema explicitly, even if it is in the PostgreSQL search_path.
+
+* Model::Validation::Errors#on now returns nil instead of [] if there
+  are no errors for an attribute.
+
+* Hooks added to a superclass after a subclass has been created no
+  longer have an effect on the subclass.
+
+* The Postgres.string_to_bool method has been removed.
+
+* PostgreSQL full text searching now always defaults to using the
+  simple dictionary.  If you want to use another dictionary, it must
+  be specified explicitly, both when searching and when creating a
+  full text index.
diff -pruN 1.5.1-1/doc/release_notes/2.8.0.txt 3.0.0-1/doc/release_notes/2.8.0.txt
--- 1.5.1-1/doc/release_notes/2.8.0.txt	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/doc/release_notes/2.8.0.txt	2009-02-19 18:08:37.000000000 +0000
@@ -0,0 +1,171 @@
+New Features
+------------
+
+* Sequel now supports database stored procedures similar to its
+  support for prepared statements.  The API is as follows:
+
+    DB[:table].call_sproc(:select, :mysp, 'param1', 'param2')
+    # or
+    sp = DB[:table].prepare_sproc(:select, :mysp)
+    sp.call('param1', 'param2')
+    sp.call('param3', 'param4')
+
+  This works with Model datasets as well, allowing them to return
+  model objects:
+
+    Album.call_sproc(:select, :new_albums)
+    #=> [#<Album ...>, #<Album ...>]
+
+  You can call a stored procedure directly on the Database object
+  if you want to, but the results and API are adapter dependent,
+  and you definitely shouldn't do it if the stored procedure returns
+  rows:
+
+    DB.call_sproc(:mysp, :args=>['param1', 'param2'])
+
+  Currently, the MySQL and JDBC adapters support stored procedures.
+  Other adapters may support them in a future version.
+
+* The connection pool code can now remove connections if the
+  adapter raises a Sequel::DatabaseDisconnectError indicating that
+  the connection has been lost.  When a query is attempted and
+  the adapter raises this error, the connection pool removes the
+  connection from the pool, and reraises the error.  The Oracle and
+  PostgreSQL adapters currently support this, and other adapters may
+  support it in a future version.
+
+* Whether to upcase or quote identifiers can now be set separately.
+  Previously, upcasing was done when quoting except when using SQLite,
+  PostgreSQL, or MySQL.  Now, you can turn upcasing off while still
+  quoting.  This may be necessary if you are using a MSSQL database
+  that has lower case table names that conflict with reserved words.
+  It also allows you to uppercase identifiers when using SQLite,
+  PostgreSQL, or MySQL, which may be beneficial in certain cases.
+
+  To turn upcasing on or off:
+
+    # Global
+    Sequel.upcase_identifiers = true
+    # Database
+    DB = Sequel.connect("postgres://...", :upcase_identifiers=>true)
+    DB.upcase_identifiers = false
+    # Dataset
+    ds = DB[:items]
+    ds.upcase_identifiers = true
+
+* Options are now supported when altering a columns type:
+
+    DB.alter_table(:items) do
+      set_column_type :score, :integer, :unsigned=>true
+      set_column_type :score, :varchar, :size=>30
+      set_column_type :score, :enum, :elements=>['a', 'b']
+    end
+
+* Standard conforming strings are now turned on by default in the
+  PostgreSQL adapter.  This makes PostgreSQL not interpret backslash
+  escapes.  This is the PostgreSQL recommended setting, which will be
+  the default setting in a future version of PostgreSQL.  If you
+  don't want for force the use of standard strings, use:
+
+    Sequel::Postgres.force_standard_strings = false
+
+  You need to do that after you call Sequel.connect but before you
+  use the database for anything, since that setting is set on
+  initial connection.
+
+* Sequel now raises an error if you attempt to use EXCEPT [ALL] or
+  INTERSECT [ALL] on a database that doesn't support it.
+
+* Sequel now raises an error if you attempt to use DISTINCT ON with
+  MySQL or Oracle, which don't support it.
+
+* A subadapter for the Progress RDBMS was added to the ODBC adapter.
+  To connect to a Progress database, use the :db_type=>'progress'
+  option.  This adapter targets Progress 9.
+
+* The ODBC adapter now supports transactions.
+
+* The MSSQL shared adapter now supports multi_insert (for inserting
+  multiple rows at once), and unicode string literals.
+
+Other Improvements
+------------------
+
+* There were many improvements related to using schemas in databases.
+  Using schema-qualified tables should work in most if not all cases
+  now.  Model associations, getting the schema, joins, and many other
+  parts of Sequel were modified to allow the use of schema-qualifed
+  tables.
+
+* You can now use literal strings with placeholders as well as
+  subselects when using prepared statements.  For example, the
+  following all work now:
+
+    DB[:items].filter("id = ?", :$i).call(:select, :i=>1)
+    DB[:items].filter(:id=>DB[:items].select(:id)\
+      .filter(:id=>:$i)).call(:select, :i=>1)
+    DB["SELECT * FROM items WHERE id = ?", :$i].call(:select, :i=>1)
+
+* Model#initialize received a few more micro-optimizations.
+
+* Model#refresh now clears the changed columns as well as the
+  associations.
+
+* You can now drop columns inside a transaction when using SQLite.
+
+* You can now submit multiple SQL queries at once in the MySQL
+  adapter:
+
+    DB['SELECT 1; SELECT 2'].all
+    #=> [{:"1"=>1, :"2"=>2}]
+
+  This may fix issues if you've seen a MySQL "commands out of sync"
+  message.  Note that this doesn't work if you are connecting to
+  MySQL via JDBC.
+
+* You can now use AliasedExpressions directly in table names given
+  to join_table:
+
+    DB.from(:i.as(:j)).join(:k.as(:l), :a=>:b)
+    #=> ... FROM i AS j INNER JOIN k AS l ON (l.a = j.b)
+
+* Database#rename_table once again works on PostgreSQL.  It was
+  broken in 2.7.0.
+
+* The interval type is now treated as it's own type. It was
+  previously treated as an integer type.
+
+* Subselects are now aliased correctly when using Oracle.
+
+* UNION, INTERSECT, and EXCEPT statements now appear before ORDER
+  and LIMIT on most databases.  If you use these constructs, please
+  test and make sure that they work correctly with your database.
+
+* SQL EXCEPT clause now works on Oracle, which uses MINUS instead.
+
+* Dataset#exists now returns a LiteralString, to make it easier to
+  use.
+
+* The Sequel.odbc_mssql method was removed, as the odbc_mssql adapter
+  was removed in a previous version.  Instead, use:
+
+    Sequel.odbc(..., :db_type=>'mssql')
+
+Backwards Compatibilty
+----------------------
+
+* The hash returned by Database#schema when no table name is provided
+  uses quoted strings instead of symbols as keys.  The hash has a
+  default proc, so using the symbol will return the same value as
+  before, but if you use each to iterate through the hash, the keys
+  will be different.  This was necessary to handle schema-qualified
+  tables.
+
+* Database#table_exists? no longer checks the output of
+  Database#tables.  If the table exists in the schema, it returns
+  true, otherwise, it does a query.  This was necessary because
+  table_exists? accepts multiple formats for table names and
+  Database#tables is an array of symbols.
+
+* When getting the schema on PostgreSQL, the default schema is now
+  used even if the :schema=>nil option is used.
diff -pruN 1.5.1-1/doc/release_notes/2.9.0.txt 3.0.0-1/doc/release_notes/2.9.0.txt
--- 1.5.1-1/doc/release_notes/2.9.0.txt	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/doc/release_notes/2.9.0.txt	2009-02-19 18:08:12.000000000 +0000
@@ -0,0 +1,97 @@
+New Features
+------------
+
+* Compound SQL statement (i.e. UNION, EXCEPT, and INTERSECT) support
+  is much improved.  Chaining compound statement calls now longer
+  wipes out previous compound statements calls of the same type.
+  Also, the ordering of the compound statements is no longer fixed
+  per adapter, it now reflects the order they were called on the
+  object.  For example, the following now work as expected:
+
+    ds1.union(ds2).union(ds3)
+    ds1.except(ds2).except(ds3)
+    ds1.intersect(ds2).intersect(ds3)
+    ds1.union(ds2).except(ds3)
+    ds1.except(ds2).intersect(ds3)
+    ds1.intersect(ds2).union(ds3)
+
+* Exception classes ValidationFailure and BeforeHookFailure were
+  added so it is eaiser to catch a failed validation.  These are
+  both subclasses of Sequel::Error, so there shouldn't be any
+  backwards compatibility issues.  Error messages are also improved,
+  as the ValidationFailure message is a string containing all
+  validation failures and the BeforeHookFailure message contains
+  which hook type caused the failure (i.e. before_save,
+  before_create, or before_validate).
+
+* The sequel command line tool now has a -L option to load
+  all files in the given directory.  This is mainly useful for
+  loading a directory of model files. The files are loaded
+  after the database connection is set up.
+
+* Methods to create and drop database functions, triggers, and
+  procedural languages were added to the PostgreSQL adapter.
+
+Other Improvements
+------------------
+
+* Database#schema now raises an error if you pass a table that
+  doesn't exist.  Before, some adapters would return an empty schema.
+  The bigger problem with this is that it made table_exists? return
+  the wrong value, since it looks at the Database's schema.
+  Generally, this bug would show up in the following code:
+
+    class Blah < Sequel::Model
+    end
+    Blah.table_exists? # True even if blahs is not a table
+
+* AlterTableGenerator#add_foreign_key now works for MySQL.
+
+* Error messages in model association methods that add/remove an
+  associated object are now more descriptive.
+
+* Dataset#destroy for model datasets now works with databases that
+  can't handle nested queries.  However, it now loads all model
+  objects being destroyed before attempting to destroy any of them.
+
+* Dataset#count now works correctly for compound SQL statements
+  (i.e. UNION, EXCEPT, and INTERSECT).
+
+* BigDecimal NaN and (+/-)Infinity values are now literalized
+  correctly.  Database support for this is hit or miss.  Sqlite will
+  work correctly, PostgreSQL raises an error if you try to store an
+  infinite value in a numeric column (though it works for float
+  columns), and MySQL converts all three to 0.
+
+* The SQLite adapter no longer loses primary key information when
+  dropping columns.
+
+* The SQLite adapter now supports dropping indicies.
+
+* A bug in the MSSQL adapter's literalization of LiteralStrings has
+  been fixed.
+
+* The literalization of blobs on PostgreSQL (bytea columns) has been
+  fixed.
+
+* Sequel now raises an error if you attempt to subclass Sequel::Model
+  before setting up a database connection.
+
+* The native postgresql adapter has been changed to only log client
+  messages of level WARNING by default.  You can modify this via:
+
+    Sequel::Postgres.client_min_messages = nil # Use Server Default
+    Sequel::Postgres.client_min_messages = :notice # Use NOTICE level
+
+* Model#inspect now calls Model#inspect_values for easier
+  overloading.
+
+Backwards Compatibilty
+----------------------
+
+* The API to Model#save_failure (a private method) was changed to
+  remove the second argument.
+
+* SQLite columns with type numeric, decimal, or money are now
+  returned as BigDecimal values.  Before, they were probably returned
+  as strings.
diff -pruN 1.5.1-1/doc/release_notes/3.0.0.txt 3.0.0-1/doc/release_notes/3.0.0.txt
--- 1.5.1-1/doc/release_notes/3.0.0.txt	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/doc/release_notes/3.0.0.txt	2009-05-04 17:43:20.000000000 +0100
@@ -0,0 +1,221 @@
+Deprecated Methods/Features Removed
+-----------------------------------
+
+Methods and features that were deprecated in 2.12.0 have been removed
+in 3.0.0.  Many features were moved into plugins or extensions, so in
+many cases you just need to require an extension or use Model.plugin
+and not make any changes to your code.  See the 2.12.0 release notes
+for the list of methods/features deprecated in 2.12.0.
+
+If you are upgrading from a previous 2.x release, please upgrade to
+2.12.0 first, fix your code to remove all deprecation warnings, and
+then upgrade to 3.0.0.
+
+New Adapter
+-----------
+
+* Sequel now has an Amalgalite adapter.  Amalgalite is a ruby
+  extension that embeds SQLite without requiring a separate SQLite
+  installation.  The adapter is functionality complete but
+  significantly slower than the native SQLite adapter.
+
+New Features
+------------
+
+* The JDBC, PostgreSQL, MySQL, and SQLite adapters all now have a
+  Database#indexes method that returns indexes for a given table:
+
+    DB.indexes(:songs)
+    => {:songs_name_index=>{:unique=>true, :columns=>[:name]},
+        :songs_lyricid_index=>{:unique=>false, :columns=>[:lyricid]}}
+
+* A schema_dumper extension was added to Sequel.  It supports dumping
+  the schema of a table (including indexes) as a string that can be
+  evaluated in the context of a Database object to create the table.
+  It also supports dumping all tables in the database as a string
+  containing a Migration subclass that will rebuild the database.
+
+    require 'sequel/extensions/schema_dumper'
+    DB.dump_table_schema(:table)
+    DB.dump_schema_migration
+    DB.dump_schema_migration(:same_db=>true)
+    DB.dump_schema_migration(:indexes=>false)
+    DB.dump_indexes_migration
+
+  The :same_db option causes Sequel to not translate column types
+  to generic column types.  By default, the migration created will
+  use generic types so it will run on other databases.  However, if
+  you only want to support a single database, using the :same_db
+  option will make the migration use the exact database type parsed
+  from the database.
+
+  The :indexes=>false option causes indexes not be included in the
+  migration.  The dump_indexes_migration can be used to create a
+  separate migration with the indexes.  This can be useful if you 
+  plan on loading a lot of data right after creating the tables,
+  since it is faster to add indexes after the data has been added.
+
+* Using options with the generic database types is now supported to
+  a limited extent.  For example, the following code now works:
+
+    DB.create_table(:table) do
+      String :a, :size=>50               # varchar(50)
+      String :b, :text=>true             # text
+      String :c, :fixed=>true, :size=>30 # char(30)
+      Time :ts                           # timestamp
+      Time :t, :only_time=>true          # time
+    end
+
+* Using Dataset#filter and related methods with multiple arguments
+  now works much more intuitively:
+
+    # 2.12.0
+    dataset.filter(:a, :b=>1) # a IS NULL AND (b = 1) IS NULL
+    # 3.0.0
+    dataset.filter(:a, :b=>1) # a AND b = 1
+
+* You can now create temporary tables by passing the :temp=>true
+  option to Database#create_table.
+
+* The Oracle shared adapter now supports emulation of
+  autoincrementing primary keys by creating a sequence and a trigger,
+  similar to how the Firebird adapter works.
+
+* The Database#database_type method was added that returns a symbol
+  specifying the database type being used.  This can be different
+  than Database.adapter_scheme if you are using an adapter like
+  JDBC that allows connecting to multiple different types of
+  databases.
+
+* Database#drop_index and related methods now support an options
+  hash that respects the :name option, so they can now be used to
+  drop an index that doesn't use the default index name.
+
+* The PostgreSQL shared adapter now supports a
+  Database#reset_primary_key_sequence method to reset the
+  primary key sequence for a given table, based on code from
+  ActiveRecord.
+
+* SQL::QualifiedIdentifiers can now be qualified, allowing you to do:
+
+   :column.qualify(:table).qualify(:schema)
+ 
+* Using the :db_type=>'mssql' option with the DBI adapter will now
+  load the MSSQL support.
+
+* The MySQL shared adapter now supports Dataset#full_text_sql, which
+  you can use in queries like the following:
+
+    ds.select(:table.*, ds.full_text_sql(:column, 'value').as(:ft))
+
+Other Improvements
+------------------
+
+* Sequel will now release connections from the connection pool
+  automatically if they are held by a dead thread.  This can happen
+  if you are using MRI 1.8 and you are heavily multithreaded or
+  you call Thread#exit! or similar method explicitly.  Those methods
+  skip the execution of ensure blocks which normally release the
+  connections when the threads exit.
+
+* Model#save will now always use the same server when refreshing data
+  after an insert.  This fixes an issue when Sequel's master/slave
+  database support is used with models.
+
+* SQL Array references are now quoted correctly, so code like this
+  now works:
+
+    :table__column.sql_subscript(1)
+
+* The PostgreSQL shared adapter now handles sequences that need to be
+  quoted correctly (previously these were quoted twice).
+
+* String quoting on Oracle no longer doubles backslashes. 
+
+* Database#count now works correctly when used on MSSQL when using
+  an adapter that doesn't handle unnamed columns.
+
+* Full text searching in the MySQL adapter now works correctly when
+  multiple search terms are used.
+
+* Altering a column's name, type, default, or NULL/NOT NULL status
+  on MySQL now keeps other relevent column information.  For example,
+  if you alter a column's type, it'll keep an existing default. This
+  functionality isn't complete, there may be other column information
+  that is lost.
+
+* Fix creation of an index with a given type on MySQL, since MySQL's
+  documentation lies.
+
+* The schema parser now handles decimal types with size specifiers,
+  fixing use on MySQL.
+
+* Dataset#quote_identifier now works correctly when given an
+  SQL::Identifier.  This allows you to do:
+
+    dataset.select{sum(hours).as(hours)}
+
+Backwards Compatibility
+-----------------------
+
+* Sequel will now use instance_eval on all virtual row blocks without
+  an argument.  This can lead to much nicer code:
+
+    dataset.filter{(number > 10) & (name > 'M')}
+    # WHERE number > 10 AND name > 'M'
+
+  2.12.0 raised a deprecation warning if you used a virtual row block
+  without an argument and you hadn't set
+  Sequel.virtual_row_instance_eval = true.
+
+* Dataset#exclude now inverts the given argument, instead of negating
+  it. This only changes its behavior if it is called with a hash or
+  array of all two pairs that have more than one element.
+
+    # 2.12.0
+    dataset.exclude(:a=>1, :b=>1) # a != 1 AND b != 1
+    # 3.0.0
+    dataset.exclude(:a=>1, :b=>1) # a != 1 OR b != 1
+
+  This was done for consistency, since exclude would only negate a
+  hash if it was given an argument, it would invert the same hash
+  if you used a block:
+
+    # 2.12.0
+    dataset.exclude{{:a=>1, :b=>1}} # a != 1 OR b != 1
+  
+  If you want the previous behavior,
+  change the code to the following:
+
+    dataset.filter({:a=>1, :b=>1}.sql_negate)
+
+* As noted above, the methods/features deprecated in 2.12.0 were
+  removed.
+
+* The private Dataset#select_*_sql methods now only take a single
+  argument, the SQL string being built.
+
+* Dataset#from when called without arguments would previously cause an
+  error to be raised when the SQL string is generated.  Now it causes
+  no FROM clause to be used, similar to how Dataset#select with no
+  arguments causes SELECT * to be used.
+
+* The internals of the generic type support and the schema generators
+  were changed significantly, which could have some fallout in terms
+  of old migrations breaking if they used the generic types and were
+  relying on some undocumented behavior (such as using Integer as a
+  type with the :unsigned option).
+
+* The Firebird adapter no longer translates the text database
+  specific type.  Use the following instead:
+    
+    String :column, :text=>true 
+
+* The MySQL shared adapter used to use the timestamp type for Time,
+  now it uses datetime.  This is because the timestamp type cannot
+  represent everything that the ruby Time class can represent.
+
+* Metaprogramming#metaattr_accessor and #metaattr_reader methods were
+  removed.
+
+* Dataset#irregular_function_sql was removed.
diff -pruN 1.5.1-1/doc/schema.rdoc 3.0.0-1/doc/schema.rdoc
--- 1.5.1-1/doc/schema.rdoc	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/doc/schema.rdoc	2009-03-23 15:14:16.000000000 +0000
@@ -0,0 +1,31 @@
+= Schema Modifications with Sequel
+
+In addition to its support for getting data from databases, Sequel has good support for modifying the schema of databases.  Unfortunately, the documentation from this is spread out in various places in the Sequel RDoc.
+
+The recommended way to set up schema modifications in Sequel is through migrations. Migrations are modifications to the schema that handle modifying the schema in two directions, up and down, with down reversing the changes made by up. Migrations are stored in sequentially numbered files inside a directory (see the Sequel::Migrator documentation for file naming format details).  The migrations can be run on the database using the sequel command line tool (the -m and -M switches).
+
+The format of the individual migration files themselves is explained in the Sequel::Migration documentation.  Each migration file contains a single migration class.  The migration class acts a proxy for the related database (given on the command line if the sequel command line tool is used, or by the db argument to Sequel::Migration#apply if the API is used). The methods that can be used inside Sequel::Migration#up or Sequel::Migration#down are just Sequel::Database instance methods, such as create_table, drop_table, and alter_table.  Most database methods that alter the schema take regular arguments, but create_table and alter_table take a block.  The methods you can use inside the create_table block are documented in Sequel::Schema::Generator, and the methods you can use inside the alter_table block are documented in Sequel::Schema::AlterTableGenerator.
+
+Migrations are not required, you can just call the schema modification methods directly on the database object.  This is often done in test code and examples.  However, it is recommended that you use the migration framework unless the database schema will not be changing in the future, as it provides a way to easily handle modifications to existing database schema.
+
+Sequel has the ability to have database independent migrations using ruby classes as types.  When you use a ruby class as a type, Sequel translates it to the most comparable type in the database you are using.  Here's an example using all supported types:
+
+    DB.create_table(:cats) do
+      primary_key :id, :type=>Integer # integer
+      String :a                       # varchar(255)
+      column :b, File                 # blob
+      Fixnum :c                       # integer
+      foreign_key :d, :other_table, :type=>Bignum # bigint
+      Float :e                        # double precision
+      BigDecimal :f                   # numeric
+      Date :g                         # date
+      DateTime :h                     # timestamp
+      Time :i                         # timestamp
+      Numeric :j                      # numeric
+      TrueClass :k                    # boolean
+      FalseClass :l                   # boolean
+    end
+
+Basically, if you use one of the ruby classes above, it will translate into a database specific type.  If you use a lowercase method, symbol, or string to specify the type, Sequel won't attempt to translate it.
+
+Note that if you use a ruby class, you shouldn't use a :size option.  Using a ruby class means that you want Sequel to pick the database type to use.  If you want to specify the size, you are no longer in database independent territory, and you need to specify the type as well, using a lowercase method, symbol, or string.
diff -pruN 1.5.1-1/doc/sharding.rdoc 3.0.0-1/doc/sharding.rdoc
--- 1.5.1-1/doc/sharding.rdoc	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/doc/sharding.rdoc	2009-03-18 21:49:21.000000000 +0000
@@ -0,0 +1,113 @@
+= Read-Only Slaves/Writable Master and Database Sharding
+
+Sequel has support for read only slave databases
+with a writable master database, as well as database sharding (where you can
+pick a server to use for a given dataset).  Support for both
+features is database independent, and should work for all database adapters
+included with Sequel.
+
+== The :servers Database option
+
+Both features use the :servers Database option.  The :servers option should
+be a hash with symbol keys and values that are either hashes or procs that
+return hashes.  Note that all servers should have the same schema for all
+tables you are accessing, unless you really know what you are doing.
+
+== Master and Slave Database Configurations
+
+=== Single Read-Only Slave, Single Master
+
+To use a single, read-only slave that handles SELECT queries, the following
+is the simplest configuration:
+
+  DB=Sequel.connect('postgres://master_server/database', \
+    :servers=>{:read_only=>{:host=>'slave_server'}})
+
+This will use the slave_server for SELECT queries and master_server for
+other queries.
+
+=== Multiple Read-Only Slaves, Single Master
+
+Let's say you have 4 slave database servers with names slave_server0,
+slave_server1, slave_server2, and slave_server3.
+
+  DB=Sequel.connect('postgres://master_server/database', \
+    :servers=>{:read_only=>proc{|db| :host=>db.get_slave_host}})
+  def DB.get_slave_host
+    @current_host ||= -1
+    "slave_server#{(@current_host+=1)%4}"
+  end
+
+This will use one of the slave servers for SELECT queries and use the
+master_server for other queries.  It's also possible to pick a random host
+instead of using the round robin approach presented above, but that can result
+in less optimal resource usage.
+
+=== Multiple Read-Only Slaves, Multiple Masters
+
+This involves the same basic idea as the multiple slaves, single master, but
+it shows that the master database is named :default.  So for 4 masters and
+4 slaves:
+
+  DB=Sequel.connect('postgres://master_server/database', \
+    :servers=>{:read_only=>proc{|db| :host=>db.get_slave_host}, \
+      :default=>proc{|db| :host=>db.get_master_host}})
+  def DB.get_slave_host
+    @current_slave_host ||= -1
+    "slave_server#{(@current_slave_host+=1)%4}"
+  end
+  def DB.get_master_host
+    @current_master_host ||= -1
+    "master_server#{(@current_master_host+=1)%4}"
+  end
+  
+== Sharding
+
+There is specific support in Sequel for handling master/slave database
+combinations, with the only necessary setup being the database configuration.
+However, since sharding is always going to be implementation dependent, Sequel
+supplies the basic infrastructure, but you have to tell it which server to use
+for each dataset.  Let's assume a simple scenario, a distributed rainbow
+table for SHA-1 hashes, sharding based on the first hex character (for a total
+of 16 shards).  First, you need to configure the database:
+
+  servers = {}
+  (('0'..'9').to_a + ('a'..'f').to_a).each do |hex|
+    servers[hex.to_sym] = {:host=>"hash_host_#{hex}"}
+  end
+  DB=Sequel.connect('postgres://hash_host/hashes', :servers=>servers)
+  
+This configures 17 servers, the 16 shard servers (/hash_host_[0-9a-f]/), and 1
+default server which will be used if no shard is specified ("hash_host").  If
+you want the default server to be one of the shard servers (e.g. hash_host_a),
+it's easiest to do:
+
+  DB=Sequel.connect('postgres://hash_host_a/hashes', :servers=>servers)
+
+That will still set up a second pool of connections for the default server,
+since it considers the default server and shard servers independent.  Note that
+if you always set the shard on a dataset before using it in queries, it will
+not attempt to connect to the default server.  Sequel may use the default
+server in queries it generates itself, such as to get column names or table
+schemas, so you should always have a default server that works.
+
+To set the shard for a given query, you use the Dataset#server method:
+
+  DB[:hashes].server(:a).filter(:hash=>/31337/)
+  
+That will return all matching rows on the hash_host_a shard that have a hash
+column that contains 31337.
+
+Rainbow tables are generally used to find specific hashes, so to save some
+work, you might want to add a method to the dataset that automatically sets
+the shard to use.  This is fairly easy using a Sequel::Model:
+
+  class Rainbow < Sequel::Model(:hashes)
+    def_dataset_method(:plaintext_for_hash) do |hash|
+      raise(ArgumentError, 'Invalid SHA-1 Hash') unless /\A[0-9a-f]{40}\z/.match(hash)
+      row = self.server(hash[0...1].to_sym).first(:hash=>hash)
+      row[:plaintext] if row
+    end
+  end
+  
+  Rainbow.plaintext_for_hash("e580726d31f6e1ad216ffd87279e536d1f74e606")
diff -pruN 1.5.1-1/doc/virtual_rows.rdoc 3.0.0-1/doc/virtual_rows.rdoc
--- 1.5.1-1/doc/virtual_rows.rdoc	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/doc/virtual_rows.rdoc	2009-03-26 19:42:34.000000000 +0000
@@ -0,0 +1,65 @@
+= Virtual Row Blocks
+
+Dataset methods filter, order, and select all take blocks that yield
+instances of Sequel::SQL::VirtualRow.  These are referred to as
+virtual row blocks.  Many other dataset methods pass the blocks
+they are given into one of those three methods, so there are actually
+many Sequel methods that take virtual row blocks.
+
+VirtualRow is a class that returns SQL::Indentifiers,
+SQL::QualifiedIdentifiers, or SQL::Functions depending on how it is
+called.  This is best shown by example:
+
+  ds = DB[:items]
+  ds.filter{|o| o.column > 1}        # column > 1
+  ds.filter{|o| o.table__column > 1} # table.column > 1
+  ds.filter{|o| o.function(1) > 1}   # function(1) > 1
+  
+Basically, the rules are:
+
+* If there are arguments, an SQL::Function is returned with the
+  name of the method used and the arguments given.
+
+* If there are no arguments and the method contains a double
+  underscore, split on the double underscore and return an
+  SQL::QualifiedIdentifier with the table and column.
+  
+* Otherwise, create an SQL::Identifier with the name of the
+  method.
+  
+One of the consequences of these rules is that you cannot
+create an SQL::Function that takes no arguments using a VirtualRow
+instance.
+  
+In Sequel 2.12, the following is deprecated by default, as
+virtual row blocks are required to accept an argument:
+
+  ds.filter{:column > 1}                     # column > 1
+  ds.filter{:table__column > 1}              # table.column > 1
+  ds.filter{:function.sql_function(1) > 1}   # function(1) > 1
+  
+This is to keep backwards compatibility while notifying people
+to change their code.  In Sequel 3.0, you will be able to
+do:
+
+  ds.filter{column > 1}        # column > 1
+  ds.filter{table__column > 1} # table.column > 1
+  ds.filter{function(1) > 1}   # function(1) > 1
+
+Sequel 3.0 will change the virtual row block support to use
+instance_eval if the block doesn't take an argument.  This
+breaks backward compatibility because instance methods called
+inside the block will now have a virtual row instance as a
+receiver instead of the receiver where the block was defined.
+You can still use local variables in the enclosing scope
+inside the block, as they take on their usual meaning.
+
+If you find yourself needing to call instance methods for the
+current receiver inside a virtual row block, you should change
+the block to accept an argument, and use that argument whenever
+you need the virtual row support.
+
+You can get the Sequel 3.0 behavior in Sequel 2.12 using the
+following method:
+
+  Sequel.virtual_row_instance_eval = true
diff -pruN 1.5.1-1/lib/sequel/adapters/ado.rb 3.0.0-1/lib/sequel/adapters/ado.rb
--- 1.5.1-1/lib/sequel/adapters/ado.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/adapters/ado.rb	2009-03-16 20:53:19.000000000 +0000
@@ -0,0 +1,85 @@
+Sequel.require 'adapters/utils/date_format'
+require 'win32ole'
+
+module Sequel
+  # The ADO adapter provides connectivity to ADO databases in Windows. ADO
+  # databases can be opened using a URL with the ado schema:
+  #
+  #   DB = Sequel.connect('ado://mydb')
+  # 
+  # or using the Sequel.ado method:
+  #
+  #   DB = Sequel.ado('mydb')
+  #
+  module ADO
+    class Database < Sequel::Database
+      set_adapter_scheme :ado
+
+      def initialize(opts)
+        super(opts)
+        opts[:driver] ||= 'SQL Server'
+        case opts[:driver]
+        when 'SQL Server'
+          Sequel.require 'adapters/shared/mssql'
+          extend Sequel::MSSQL::DatabaseMethods
+        end
+      end
+
+      def connect(server)
+        opts = server_opts(server)
+        s = "driver=#{opts[:driver]};server=#{opts[:host]};database=#{opts[:database]}#{";uid=#{opts[:user]};pwd=#{opts[:password]}" if opts[:user]}"
+        handle = WIN32OLE.new('ADODB.Connection')
+        handle.Open(s)
+        handle
+      end
+      
+      def dataset(opts = nil)
+        ADO::Dataset.new(self, opts)
+      end
+    
+      def execute(sql, opts={})
+        log_info(sql)
+        synchronize(opts[:server]) do |conn|
+          r = conn.Execute(sql)
+          yield(r) if block_given?
+          r
+        end
+      end
+      alias_method :do, :execute
+
+      private
+
+      def disconnect_connection(conn)
+        conn.Close
+      end
+    end
+    
+    class Dataset < Sequel::Dataset
+      include Dataset::SQLStandardDateFormat
+
+      def fetch_rows(sql)
+        execute(sql) do |s|
+          @columns = s.Fields.extend(Enumerable).map do |column|
+            name = column.Name.empty? ? '(no column name)' : column.Name
+            output_identifier(name)
+          end
+          
+          unless s.eof
+            s.moveFirst
+            s.getRows.transpose.each {|r| yield hash_row(r)}
+          end
+        end
+        self
+      end
+      
+      private
+      
+      def hash_row(row)
+        @columns.inject({}) do |m, c|
+          m[c] = row.shift
+          m
+        end
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/adapters/amalgalite.rb 3.0.0-1/lib/sequel/adapters/amalgalite.rb
--- 1.5.1-1/lib/sequel/adapters/amalgalite.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/adapters/amalgalite.rb	2009-04-28 00:18:19.000000000 +0100
@@ -0,0 +1,208 @@
+require 'amalgalite'
+Sequel.require 'adapters/shared/sqlite'
+
+module Sequel
+  # Top level module for holding all Amalgalite-related modules and classes
+  # for Sequel.
+  module Amalgalite
+    # Type conversion map class for Sequel's use of Amalgamite
+    class SequelTypeMap < ::Amalgalite::TypeMaps::DefaultMap
+      methods_handling_sql_types.delete('string')
+      methods_handling_sql_types.merge!(
+        'datetime' => %w'datetime timestamp',
+        'time' => %w'time',
+        'float' => ['float', 'double', 'real', 'double precision'],
+        'decimal' => %w'numeric decimal money'
+      )
+      
+      # Return blobs as instances of Sequel::SQL::Blob instead of
+      # Amalgamite::Blob
+      def blob(s)
+        SQL::Blob.new(s)
+      end
+      
+      # Return numeric/decimal types as instances of BigDecimal
+      # instead of Float
+      def decimal(s)
+        BigDecimal.new(s)
+      end
+      
+      # Return datetime types as instances of Sequel.datetime_class
+      def datetime(s)
+        Sequel.string_to_datetime(s)
+      end
+      
+      # Don't raise an error if the value is a string and the declared
+      # type doesn't match a known type, just return the value.
+      def result_value_of(declared_type, value)
+        if value.is_a?(::Amalgalite::Blob)
+          SQL::Blob.new(value.source)
+        elsif value.is_a?(String) && declared_type
+          (meth = self.class.sql_to_method(declared_type.downcase)) ? send(meth, value) : value
+        else
+          super
+        end
+      end
+    end
+    
+    # Database class for SQLite databases used with Sequel and the
+    # amalgalite driver.
+    class Database < Sequel::Database
+      include ::Sequel::SQLite::DatabaseMethods
+      
+      set_adapter_scheme :amalgalite
+      
+      # Mimic the file:// uri, by having 2 preceding slashes specify a relative
+      # path, and 3 preceding slashes specify an absolute path.
+      def self.uri_to_options(uri) # :nodoc:
+        { :database => (uri.host.nil? && uri.path == '/') ? nil : "#{uri.host}#{uri.path}" }
+      end
+      private_class_method :uri_to_options
+      
+      # Connect to the database.  Since SQLite is a file based database,
+      # the only options available are :database (to specify the database
+      # name), and :timeout, to specify how long to wait for the database to
+      # be available if it is locked, given in milliseconds (default is 5000).
+      def connect(server)
+        opts = server_opts(server)
+        opts[:database] = ':memory:' if blank_object?(opts[:database])
+        db = ::Amalgalite::Database.new(opts[:database])
+        db.busy_handler(::Amalgalite::BusyTimeout.new(opts.fetch(:timeout, 5000)/50, 50))
+        db.type_map = SequelTypeMap.new
+        db
+      end
+      
+      # Amalgalite is just the SQLite database without a separate SQLite installation.
+      def database_type
+        :sqlite
+      end
+
+      # Return instance of Sequel::Amalgalite::Dataset with the given options.
+      def dataset(opts = nil)
+        Amalgalite::Dataset.new(self, opts)
+      end
+      
+      # Run the given SQL with the given arguments and reload the schema.  Returns nil.
+      def execute_ddl(sql, opts={})
+        _execute(sql, opts){|conn| conn.execute_batch(sql); conn.reload_schema!}
+        nil
+      end
+      
+      # Run the given SQL with the given arguments and return the number of changed rows.
+      def execute_dui(sql, opts={})
+        _execute(sql, opts){|conn| conn.execute_batch(sql); conn.row_changes}
+      end
+      
+      # Run the given SQL with the given arguments and return the last inserted row id.
+      def execute_insert(sql, opts={})
+        _execute(sql, opts){|conn| conn.execute_batch(sql); conn.last_insert_rowid}
+      end
+      
+      # Run the given SQL with the given arguments and yield each row.
+      def execute(sql, opts={})
+        retried = false
+        _execute(sql, opts) do |conn|
+          conn.prepare(sql) do |stmt|
+            begin
+             stmt.result_meta
+            rescue NoMethodError
+              conn.reload_schema!
+              stmt.result_meta
+            end
+            yield stmt
+          end
+        end
+      end
+      
+      # Run the given SQL with the given arguments and return the first value of the first row.
+      def single_value(sql, opts={})
+        _execute(sql, opts){|conn| conn.first_value_from(sql)}
+      end
+      
+      # Use the native driver transaction method if there isn't already a transaction
+      # in progress on the connection, always yielding a connection inside a transaction
+      # transaction.
+      def transaction(opts={})
+        synchronize(opts[:server]) do |conn|
+          return yield(conn) if conn.in_transaction?
+          begin
+            result = nil
+            log_info('Transaction.begin')
+            conn.transaction{result = yield(conn)}
+            result
+          rescue ::Exception => e
+            log_info('Transaction.rollback')
+            transaction_error(e, ::Amalgalite::Error, ::Amalgalite::SQLite3::Error)
+          ensure
+            log_info('Transaction.commit') unless e
+          end
+        end
+      end
+      
+      private
+      
+      # Log the SQL and yield an available connection.  Rescue
+      # any Amalgalite::Errors and turn them into DatabaseErrors.
+      def _execute(sql, opts)
+        begin
+          log_info(sql)
+          synchronize(opts[:server]){|conn| yield conn}
+        rescue ::Amalgalite::Error, ::Amalgalite::SQLite3::Error => e
+          raise_error(e)
+        end
+      end
+      
+      # The Amagalite adapter does not need the pool to convert exceptions.
+      # Also, force the max connections to 1 if a memory database is being
+      # used, as otherwise each connection gets a separate database.
+      def connection_pool_default_options
+        o = super.merge(:pool_convert_exceptions=>false)
+        # Default to only a single connection if a memory database is used,
+        # because otherwise each connection will get a separate database
+        o[:max_connections] = 1 if @opts[:database] == ':memory:' || blank_object?(@opts[:database])
+        o
+      end
+
+      # Disconnect given connections from the database.
+      def disconnect_connection(c)
+        c.close
+      end
+    end
+    
+    # Dataset class for SQLite datasets that use the amalgalite driver.
+    class Dataset < Sequel::Dataset
+      include ::Sequel::SQLite::DatasetMethods
+      
+      EXPLAIN = 'EXPLAIN %s'.freeze
+      
+      # Return an array of strings specifying a query explanation for the
+      # current dataset.
+      def explain
+        res = []
+        @db.result_set(EXPLAIN % select_sql(opts), nil) {|r| res << r}
+        res
+      end
+      
+      # Yield a hash for each row in the dataset.
+      def fetch_rows(sql)
+        execute(sql) do |stmt|
+          stmt.result_meta
+          @columns = cols = stmt.result_fields.map{|c| output_identifier(c)}
+          col_count = cols.size
+          stmt.each do |result|
+            row = {}
+            col_count.times{|i| row[cols[i]] = result[i]}
+            yield row
+          end
+        end
+      end
+
+      private
+      
+      # Quote the string using the adapter instance method.
+      def literal_string(v)
+        "#{db.synchronize{|c| c.quote(v)}}"
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/adapters/db2.rb 3.0.0-1/lib/sequel/adapters/db2.rb
--- 1.5.1-1/lib/sequel/adapters/db2.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/adapters/db2.rb	2009-04-16 20:24:49.000000000 +0100
@@ -0,0 +1,135 @@
+require 'db2/db2cli'
+
+module Sequel
+  module DB2
+    class Database < Sequel::Database
+      set_adapter_scheme :db2
+
+      include DB2CLI
+
+      TEMPORARY = 'GLOBAL TEMPORARY '.freeze
+      
+      rc, @@env = SQLAllocHandle(SQL_HANDLE_ENV, SQL_NULL_HANDLE)
+      #check_error(rc, "Could not allocate DB2 environment")
+
+      def connect(server)
+        opts = server_opts(server)
+        rc, dbc = SQLAllocHandle(SQL_HANDLE_DBC, @@env) 
+        check_error(rc, "Could not allocate database connection")
+        
+        rc = SQLConnect(dbc, opts[:database], opts[:user], opts[:password]) 
+        check_error(rc, "Could not connect to database")
+        
+        dbc
+      end
+      
+      def test_connection(server=nil)
+        synchronize(server){|conn|}
+        true
+      end
+
+      def dataset(opts = nil)
+        DB2::Dataset.new(self, opts)
+      end
+      
+      def execute(sql, opts={})
+        log_info(sql)
+        synchronize(opts[:server]) do |conn|
+          rc, sth = SQLAllocHandle(SQL_HANDLE_STMT, @handle) 
+          check_error(rc, "Could not allocate statement")
+
+          begin
+            rc = SQLExecDirect(sth, sql) 
+            check_error(rc, "Could not execute statement")
+            
+            yield(sth) if block_given?
+
+            rc, rpc = SQLRowCount(sth)
+            check_error(rc, "Could not get RPC") 
+            rpc
+          ensure
+            rc = SQLFreeHandle(SQL_HANDLE_STMT, sth)
+            check_error(rc, "Could not free statement")
+          end
+        end
+      end
+      alias_method :do, :execute
+      
+      private
+      
+      def check_error(rc, msg)
+        case rc
+        when SQL_SUCCESS, SQL_SUCCESS_WITH_INFO
+          nil
+        else
+          raise DatabaseError, msg
+        end
+      end
+
+      def disconnect_connection(conn)
+        rc = SQLDisconnect(conn)
+        check_error(rc, "Could not disconnect from database")
+
+        rc = SQLFreeHandle(SQL_HANDLE_DBC, conn)
+        check_error(rc, "Could not free Database handle")
+      end
+    end
+    
+    class Dataset < Sequel::Dataset
+      MAX_COL_SIZE = 256
+      
+      def fetch_rows(sql)
+        execute(sql) do |sth|
+          @column_info = get_column_info(sth)
+          @columns = @column_info.map {|c| output_identifier(c[:name])}
+          while (rc = SQLFetch(@handle)) != SQL_NO_DATA_FOUND
+            @db.check_error(rc, "Could not fetch row")
+            yield hash_row(sth)
+          end
+        end
+        self
+      end
+      
+      private
+
+      def get_column_info(sth)
+        rc, column_count = SQLNumResultCols(sth)
+        @db.check_error(rc, "Could not get number of result columns")
+
+        (1..column_count).map do |i| 
+          rc, name, buflen, datatype, size, digits, nullable = SQLDescribeCol(sth, i, MAX_COL_SIZE)
+          @b.check_error(rc, "Could not describe column")
+          
+          {:name => name, :db2_type => datatype, :precision => size}
+        end 
+      end
+      
+      def hash_row(sth)
+        row = {}
+        @column_info.each_with_index do |c, i|
+          rc, v = SQLGetData(sth, i+1, c[:db2_type], c[:precision]) 
+          @db.check_error(rc, "Could not get data")
+          
+          row[output_identifier(c[:name])] = convert_type(v)
+        end
+        row
+      end
+      
+      def convert_type(v)
+        case v
+        when DB2CLI::Date 
+          DBI::Date.new(v.year, v.month, v.day)
+        when DB2CLI::Time
+          DBI::Time.new(v.hour, v.minute, v.second)
+        when DB2CLI::Timestamp 
+          DBI::Timestamp.new(v.year, v.month, v.day,
+            v.hour, v.minute, v.second, v.fraction)
+        when DB2CLI::Null
+          nil
+        else  
+          v
+        end
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/adapters/dbi.rb 3.0.0-1/lib/sequel/adapters/dbi.rb
--- 1.5.1-1/lib/sequel/adapters/dbi.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/adapters/dbi.rb	2009-04-22 01:48:42.000000000 +0100
@@ -0,0 +1,110 @@
+require 'dbi'
+
+module Sequel
+  module DBI
+    class Database < Sequel::Database
+      set_adapter_scheme :dbi
+      
+      DBI_ADAPTERS = {
+        :ado => "ADO",
+        :db2 => "DB2",
+        :frontbase => "FrontBase",
+        :interbase => "InterBase",
+        :msql => "Msql",
+        :mysql => "Mysql",
+        :odbc => "ODBC",
+        :oracle => "Oracle",
+        :pg => "pg",
+        :proxy => "Proxy",
+        :sqlite => "SQLite",
+        :sqlrelay => "SQLRelay"
+      }
+      
+      def initialize(opts)
+        super(opts)
+        case opts[:db_type]
+        when 'mssql'
+          Sequel.require 'adapters/shared/mssql'
+          extend Sequel::MSSQL::DatabaseMethods
+        end
+      end
+
+      # Converts a uri to an options hash. These options are then passed
+      # to a newly created database object.
+      def self.uri_to_options(uri) # :nodoc:
+        database = (m = /\/(.*)/.match(uri.path)) && (m[1])
+        if m = /dbi-(.+)/.match(uri.scheme)
+          adapter = DBI_ADAPTERS[m[1].to_sym] || m[1]
+          database = "#{adapter}:dbname=#{database}"
+        end
+        {
+          :user => uri.user,
+          :password => uri.password,
+          :host => uri.host,
+          :port => uri.port,
+          :database => database
+        }
+      end
+
+      private_class_method :uri_to_options
+      
+      def connect(server)
+        opts = server_opts(server)
+        dbname = opts[:database]
+        if dbname !~ /^DBI:/ then
+          dbname = "DBI:#{dbname}"
+          [:host, :port].each{|sym| dbname += ";#{sym}=#{opts[sym]}" unless blank_object?(opts[sym])}
+        end
+        ::DBI.connect(dbname, opts[:user], opts[:password])
+      end
+      
+      def dataset(opts = nil)
+        DBI::Dataset.new(self, opts)
+      end
+    
+      def execute(sql, opts={})
+        log_info(sql)
+        synchronize(opts[:server]) do |conn|
+          r = conn.execute(sql)
+          yield(r) if block_given?
+          r
+        end
+      end
+      
+      def do(sql, opts={})
+        log_info(sql)
+        synchronize(opts[:server]){|conn| conn.do(sql)}
+      end
+      alias_method :execute_dui, :do
+
+      private
+
+      def disconnect_connection(c)
+        c.disconnect
+      end
+    end
+    
+    class Dataset < Sequel::Dataset
+      def fetch_rows(sql, &block)
+        execute(sql) do |s|
+          begin
+            @columns = s.column_names.map{|c| output_identifier(c)}
+            s.fetch{|r| yield hash_row(s, r)}
+          ensure
+            s.finish rescue nil
+          end
+        end
+        self
+      end
+      
+      private
+      
+      def hash_row(stmt, row)
+        @columns.inject({}) do |m, c|
+          m[c] = row.shift
+          m
+        end
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/adapters/do/mysql.rb 3.0.0-1/lib/sequel/adapters/do/mysql.rb
--- 1.5.1-1/lib/sequel/adapters/do/mysql.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/adapters/do/mysql.rb	2009-03-16 20:53:19.000000000 +0000
@@ -0,0 +1,38 @@
+Sequel.require 'adapters/shared/mysql'
+
+module Sequel
+  module DataObjects
+    # Database and Dataset instance methods for MySQL specific
+    # support via DataObjects.
+    module MySQL
+      # Database instance methods for MySQL databases accessed via DataObjects.
+      module DatabaseMethods
+        include Sequel::MySQL::DatabaseMethods
+        
+        # Return instance of Sequel::DataObjects::MySQL::Dataset with the given opts.
+        def dataset(opts=nil)
+          Sequel::DataObjects::MySQL::Dataset.new(self, opts)
+        end
+        
+        private
+        
+        # The database name for the given database.  Need to parse it out
+        # of the connection string, since the DataObjects does no parsing on the
+        # given connection string by default.
+        def database_name
+          (m = /\/(.*)/.match(URI.parse(uri).path)) && m[1]
+        end
+      end
+      
+      # Dataset class for MySQL datasets accessed via DataObjects.
+      class Dataset < DataObjects::Dataset
+        include Sequel::MySQL::DatasetMethods
+        
+        # Use execute_insert to execute the replace_sql.
+        def replace(*args)
+          execute_insert(replace_sql(*args))
+        end
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/adapters/do/postgres.rb 3.0.0-1/lib/sequel/adapters/do/postgres.rb
--- 1.5.1-1/lib/sequel/adapters/do/postgres.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/adapters/do/postgres.rb	2009-03-16 20:53:19.000000000 +0000
@@ -0,0 +1,92 @@
+Sequel.require 'adapters/shared/postgres'
+
+module Sequel
+  Postgres::CONVERTED_EXCEPTIONS << PostgresError
+  
+  module DataObjects
+    # Adapter, Database, and Dataset support for accessing a PostgreSQL
+    # database via DataObjects.
+    module Postgres
+      # Methods to add to the DataObjects adapter/connection to allow it to work
+      # with the shared PostgreSQL code.
+      module AdapterMethods
+        include Sequel::Postgres::AdapterMethods
+        
+        # Give the DataObjects adapter a direct execute method, which creates
+        # a statement with the given sql and executes it.
+        def execute(sql, args=nil)
+          command = create_command(sql)
+          begin
+            if block_given?
+              begin
+                reader = command.execute_reader
+                yield(reader)
+              ensure
+                reader.close if reader
+              end
+            else
+              command.execute_non_query
+            end
+          rescue PostgresError => e
+            raise_error(e)
+          end
+        end
+        
+        private
+        
+        # DataObjects specific method of getting specific values from a result set.
+        def single_value(reader)
+          while(reader.next!) do
+            return reader.values.at(0)
+          end
+        end
+      end
+    
+      # Methods to add to Database instances that access PostgreSQL via
+      # DataObjects.
+      module DatabaseMethods
+        include Sequel::Postgres::DatabaseMethods
+        
+        # Add the primary_keys and primary_key_sequences instance variables,
+        # so we can get the correct return values for inserted rows.
+        def self.extended(db)
+          db.instance_eval do
+            @primary_keys = {}
+            @primary_key_sequences = {}
+          end
+        end
+        
+        # Return instance of Sequel::DataObjects::Postgres::Dataset with the given opts.
+        def dataset(opts=nil)
+          Sequel::DataObjects::Postgres::Dataset.new(self, opts)
+        end
+        
+        # Run the INSERT sql on the database and return the primary key
+        # for the record.
+        def execute_insert(sql, opts={})
+          log_info(sql)
+          synchronize(opts[:server]) do |conn|
+            conn.create_command(sql).execute_non_query
+            insert_result(conn, opts[:table], opts[:values])
+          end
+        end
+        
+        private
+        
+        # Extend the adapter with the DataObjects PostgreSQL AdapterMethods
+        def setup_connection(conn)
+          conn = super(conn)
+          conn.extend(Sequel::DataObjects::Postgres::AdapterMethods)
+          conn.db = self
+          conn.apply_connection_settings
+          conn
+        end
+      end
+      
+      # Dataset subclass used for datasets that connect to PostgreSQL via DataObjects.
+      class Dataset < DataObjects::Dataset
+        include Sequel::Postgres::DatasetMethods
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/adapters/do/sqlite.rb 3.0.0-1/lib/sequel/adapters/do/sqlite.rb
--- 1.5.1-1/lib/sequel/adapters/do/sqlite.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/adapters/do/sqlite.rb	2009-03-16 20:53:19.000000000 +0000
@@ -0,0 +1,31 @@
+Sequel.require 'adapters/shared/sqlite'
+
+module Sequel
+  module DataObjects
+    # Database and Dataset support for SQLite databases accessed via DataObjects.
+    module SQLite
+      # Instance methods for SQLite Database objects accessed via DataObjects.
+      module DatabaseMethods
+        include Sequel::SQLite::DatabaseMethods
+        
+        # Return Sequel::DataObjects::SQLite::Dataset object with the given opts.
+        def dataset(opts=nil)
+          Sequel::DataObjects::SQLite::Dataset.new(self, opts)
+        end
+        
+        private
+        
+        # Default to a single connection for a memory database.
+        def connection_pool_default_options
+          o = super
+          uri == 'sqlite3::memory:' ? o.merge(:max_connections=>1) : o
+        end
+      end
+      
+      # Dataset class for SQLite datasets accessed via DataObjects.
+      class Dataset < DataObjects::Dataset
+        include Sequel::SQLite::DatasetMethods
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/adapters/do.rb 3.0.0-1/lib/sequel/adapters/do.rb
--- 1.5.1-1/lib/sequel/adapters/do.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/adapters/do.rb	2009-04-09 19:03:43.000000000 +0100
@@ -0,0 +1,193 @@
+require 'data_objects'
+
+module Sequel
+  # Module holding the DataObjects support for Sequel.  DataObjects is a
+  # ruby library with a standard API for accessing databases.
+  #
+  # The DataObjects adapter currently supports PostgreSQL, MySQL, and
+  # SQLite:
+  #
+  # *  Sequel.connect('do:sqlite3::memory:')
+  # *  Sequel.connect('do:postgres://user:password@host/database')
+  # *  Sequel.connect('do:mysql://user:password@host/database')
+  module DataObjects
+    # Contains procs keyed on sub adapter type that extend the
+    # given database object so it supports the correct database type.
+    DATABASE_SETUP = {:postgres=>proc do |db|
+        require 'do_postgres'
+        Sequel.require 'adapters/do/postgres'
+        db.converted_exceptions << PostgresError
+        db.extend(Sequel::DataObjects::Postgres::DatabaseMethods)
+      end,
+      :mysql=>proc do |db|
+        require 'do_mysql'
+        Sequel.require 'adapters/do/mysql'
+        db.converted_exceptions << MysqlError
+        db.extend(Sequel::DataObjects::MySQL::DatabaseMethods)
+      end,
+      :sqlite3=>proc do |db|
+        require 'do_sqlite3'
+        Sequel.require 'adapters/do/sqlite'
+        db.converted_exceptions << Sqlite3Error
+        db.extend(Sequel::DataObjects::SQLite::DatabaseMethods)
+      end
+    }
+      
+    # DataObjects uses it's own internal connection pooling in addition to the
+    # pooling that Sequel uses.  You should make sure that you don't set
+    # the connection pool size to more than 8 for a
+    # Sequel::DataObjects::Database object, or hack DataObjects (or Extlib) to
+    # use a pool size at least as large as the pool size being used by Sequel.
+    class Database < Sequel::Database
+      set_adapter_scheme :do
+      
+      # Convert the given exceptions to Sequel:Errors, necessary
+      # because DO raises errors specific to database types in
+      # certain cases.
+      attr_accessor :converted_exceptions
+      
+      # Call the DATABASE_SETUP proc directly after initialization,
+      # so the object always uses sub adapter specific code.  Also,
+      # raise an error immediately if the connection doesn't have a
+      # uri, since DataObjects requires one.
+      def initialize(opts)
+        @opts = opts
+        @converted_exceptions = []
+        raise(Error, "No connection string specified") unless uri
+        if prok = DATABASE_SETUP[subadapter.to_sym]
+          prok.call(self)
+        end
+        super(opts)
+      end
+      
+      # Setup a DataObjects::Connection to the database.
+      def connect(server)
+        setup_connection(::DataObjects::Connection.new(uri(server_opts(server))))
+      end
+      
+      # Return a Sequel::DataObjects::Dataset object for this database.
+      def dataset(opts = nil)
+        DataObjects::Dataset.new(self, opts)
+      end
+    
+      # Execute the given SQL.  If a block is given, the DataObjects::Reader
+      # created is yielded to it. A block should not be provided unless a
+      # a SELECT statement is being used (or something else that returns rows).
+      # Otherwise, the return value is the insert id if opts[:type] is :insert,
+      # or the number of affected rows, otherwise.
+      def execute(sql, opts={})
+        log_info(sql)
+        synchronize(opts[:server]) do |conn|
+          begin
+            command = conn.create_command(sql)
+            res = block_given? ? command.execute_reader : command.execute_non_query
+          rescue Exception => e
+            raise_error(e, :classes=>@converted_exceptions)
+          end
+          if block_given?
+            begin
+              yield(res)
+            ensure
+             res.close if res
+            end
+          elsif opts[:type] == :insert
+            res.insert_id
+          else
+            res.affected_rows
+          end
+        end
+      end
+      
+      # Execute the SQL on the this database, returning the number of affected
+      # rows.
+      def execute_dui(sql, opts={})
+        execute(sql, opts)
+      end
+      
+      # Execute the SQL on this database, returning the primary key of the
+      # table being inserted to.
+      def execute_insert(sql, opts={})
+        execute(sql, opts.merge(:type=>:insert))
+      end
+      
+      # Return the subadapter type for this database, i.e. sqlite3 for
+      # do:sqlite3::memory:.
+      def subadapter
+        uri.split(":").first
+      end
+      
+      # Use DataObject's transaction support for transactions.  This
+      # only supports single level transactions, and it always prepares
+      # transactions and commits them immediately after.  It's wasteful,
+      # but required by DataObject's API.
+      def transaction(opts={})
+        th = Thread.current
+        synchronize(opts[:server]) do |conn|
+          return yield(conn) if @transactions.include?(th)
+          t = ::DataObjects::Transaction.create_for_uri(uri)
+          t.instance_variable_get(:@connection).close
+          t.instance_variable_set(:@connection, conn)
+          begin
+            log_info("Transaction.begin")
+            t.begin
+            @transactions << th
+            yield(conn)
+          rescue Exception => e
+            log_info("Transaction.rollback")
+            t.rollback
+            transaction_error(e)
+          ensure
+            unless e
+              log_info("Transaction.commit")
+              t.prepare
+              t.commit 
+            end
+            @transactions.delete(th)
+          end
+        end
+      end
+      
+      # Return the DataObjects URI for the Sequel URI, removing the do:
+      # prefix.
+      def uri(opts={})
+        opts = @opts.merge(opts)
+        (opts[:uri] || opts[:url]).sub(/\Ado:/, '')
+      end
+
+      private
+
+      # Close the given database connection.
+      def disconnect_connection(c)
+        c.close
+      end
+      
+      # Allow extending the given connection when it is first created.
+      # By default, just returns the connection.
+      def setup_connection(conn)
+        conn
+      end
+      
+      # The DataObjects adapter should convert exceptions by default.
+      def connection_pool_default_options
+        super.merge(:pool_convert_exceptions=>false)
+      end
+    end
+    
+    # Dataset class for Sequel::DataObjects::Database objects.
+    class Dataset < Sequel::Dataset
+      # Execute the SQL on the database and yield the rows as hashes
+      # with symbol keys.
+      def fetch_rows(sql)
+        execute(sql) do |reader|
+          cols = @columns = reader.fields.map{|f| output_identifier(f)}
+          while(reader.next!) do
+            h = {}
+            cols.zip(reader.values).each{|k, v| h[k] = v}
+            yield h
+          end
+        end
+        self
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/adapters/firebird.rb 3.0.0-1/lib/sequel/adapters/firebird.rb
--- 1.5.1-1/lib/sequel/adapters/firebird.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/adapters/firebird.rb	2009-04-27 23:53:01.000000000 +0100
@@ -0,0 +1,305 @@
+require 'fb'
+Sequel.require 'adapters/utils/unsupported'
+
+module Sequel
+  # The Sequel Firebird adapter requires the ruby fb driver located at
+  # http://github.com/wishdev/fb.
+  module Firebird
+    class Database < Sequel::Database
+      set_adapter_scheme :firebird
+
+      AUTO_INCREMENT = ''.freeze
+      TEMPORARY = 'GLOBAL TEMPORARY '.freeze
+
+      # Add the primary_keys and primary_key_sequences instance variables,
+      # so we can get the correct return values for inserted rows.
+      def initialize(*args)
+        super
+        @primary_keys = {}
+        @primary_key_sequences = {}
+      end
+
+      def connect(server)
+        opts = server_opts(server)
+
+        db = Fb::Database.new(
+          :database => "#{opts[:host]}:#{opts[:database]}",
+          :username => opts[:user],
+          :password => opts[:password])
+        conn = db.connect
+        conn.downcase_names = true
+        conn
+      end
+
+      # Creates a table with the columns given in the provided block:
+      #
+      #   DB.create_table :posts do
+      #     primary_key :id, :serial
+      #     column :title, :text
+      #     column :content, :text
+      #     index :title
+      #   end
+      #
+      # See Schema::Generator.
+      # Firebird gets an override because of the mess of creating a
+      # generator for auto-incrementing primary keys.
+      def create_table(name, options={}, &block)
+        options = {:generator=>options} if options.is_a?(Schema::Generator)
+        generator = options[:generator] || Schema::Generator.new(self, &block)
+        drop_statement, create_statements = create_table_sql_list(name, generator, options)
+        (execute_ddl(drop_statement) rescue nil) if drop_statement
+        (create_statements + index_sql_list(name, generator.indexes)).each{|sql| execute_ddl(sql)}
+      end
+
+      def create_trigger(*args)
+        self << create_trigger_sql(*args)
+      end
+
+      def dataset(opts = nil)
+        Firebird::Dataset.new(self, opts)
+      end
+
+      def drop_sequence(name)
+        self << drop_sequence_sql(name)
+      end
+
+      def execute(sql, opts={})
+        log_info(sql)
+        begin
+          synchronize(opts[:server]) do |conn|
+            r = conn.execute(sql)
+            yield(r) if block_given?
+            r
+          end
+        rescue => e
+          log_info(e.message)
+          raise_error(e, :classes=>[Fb::Error])
+        end
+      end
+
+      # Return primary key for the given table.
+      def primary_key(table, server=nil)
+        synchronize(server){|conn| primary_key_for_table(conn, table)}
+      end
+
+      # Returns primary key for the given table.  This information is
+      # cached, and if the primary key for a table is changed, the
+      # @primary_keys instance variable should be reset manually.
+      def primary_key_for_table(conn, table)
+        @primary_keys[quote_identifier(table)] ||= conn.table_primary_key(quote_identifier(table))
+      end
+
+      def restart_sequence(*args)
+        self << restart_sequence_sql(*args)
+      end
+
+      def sequences(opts={})
+        ds = self[:"rdb$generators"].server(opts[:server]).filter(:"rdb$system_flag" => 0).select(:"rdb$generator_name")
+        block_given? ? yield(ds) : ds.map{|r| ds.send(:output_identifier, r[:"rdb$generator_name"])}
+      end
+
+      def tables(opts={})
+        ds = self[:"rdb$relations"].server(opts[:server]).filter(:"rdb$view_blr" => nil, Sequel::SQL::Function.new(:COALESCE, :"rdb$system_flag", 0) => 0).select(:"rdb$relation_name")
+        block_given? ? yield(ds) : ds.map{|r| ds.send(:output_identifier, r[:"rdb$relation_name"])}
+      end
+
+      def transaction(opts={})
+        synchronize(opts[:server]) do |conn|
+          return yield(conn) if @transactions.include?(Thread.current)
+          log_info("Transaction.begin")
+          conn.transaction
+          begin
+            @transactions << Thread.current
+            yield(conn)
+          rescue ::Exception => e
+            log_info("Transaction.rollback")
+            conn.rollback
+            transaction_error(e, Fb::Error)
+          ensure
+            unless e
+              log_info("Transaction.commit")
+              conn.commit
+            end
+            @transactions.delete(Thread.current)
+          end
+        end
+      end
+
+      private
+
+      # Use Firebird specific syntax for add column
+      def alter_table_sql(table, op)
+        case op[:op]
+        when :add_column
+          "ALTER TABLE #{quote_schema_table(table)} ADD #{column_definition_sql(op)}"
+        when :drop_column
+          "ALTER TABLE #{quote_schema_table(table)} DROP #{column_definition_sql(op)}"
+        when :rename_column
+          "ALTER TABLE #{quote_schema_table(table)} ALTER #{quote_identifier(op[:name])} TO #{quote_identifier(op[:new_name])}"
+        when :set_column_type
+          "ALTER TABLE #{quote_schema_table(table)} ALTER #{quote_identifier(op[:name])} TYPE #{type_literal(op)}"
+        else
+          super(table, op)
+        end
+      end
+
+      def auto_increment_sql()
+        AUTO_INCREMENT
+      end
+      
+      def create_sequence_sql(name, opts={})
+        "CREATE SEQUENCE #{quote_identifier(name)}"
+      end
+
+      def create_table_sql_list(name, generator, options={})
+        statements = [create_table_sql(name, generator, options)]
+        drop_seq_statement = nil
+        generator.columns.each do |c|
+          if c[:auto_increment]
+            c[:sequence_name] ||= "seq_#{name}_#{c[:name]}"
+            unless c[:create_sequence] == false
+              drop_seq_statement = drop_sequence_sql(c[:sequence_name])
+              statements << create_sequence_sql(c[:sequence_name])
+              statements << restart_sequence_sql(c[:sequence_name], {:restart_position => c[:sequence_start_position]}) if c[:sequence_start_position]
+            end
+            unless c[:create_trigger] == false
+              c[:trigger_name] ||= "BI_#{name}_#{c[:name]}"
+              c[:quoted_name] = quote_identifier(c[:name])
+              trigger_definition = <<-END
+              begin
+                if ((new.#{c[:quoted_name]} is null) or (new.#{c[:quoted_name]} = 0)) then
+                begin
+                  new.#{c[:quoted_name]} = next value for #{c[:sequence_name]};
+                end
+              end
+              END
+              statements << create_trigger_sql(name, c[:trigger_name], trigger_definition, {:events => [:insert]})
+            end
+          end
+        end
+        [drop_seq_statement, statements]
+      end
+
+      def create_trigger_sql(table, name, definition, opts={})
+        events = opts[:events] ? Array(opts[:events]) : [:insert, :update, :delete]
+        whence = opts[:after] ? 'AFTER' : 'BEFORE'
+        inactive = opts[:inactive] ? 'INACTIVE' : 'ACTIVE'
+        position = opts[:position] ? opts[:position] : 0
+        sql = <<-end_sql
+          CREATE TRIGGER #{quote_identifier(name)} for #{quote_identifier(table)}
+          #{inactive} #{whence} #{events.map{|e| e.to_s.upcase}.join(' OR ')} position #{position}
+          as #{definition}
+        end_sql
+        sql
+      end
+
+      def disconnect_connection(c)
+        c.close
+      end
+
+      def drop_sequence_sql(name)
+        "DROP SEQUENCE #{quote_identifier(name)}"
+      end
+
+      def restart_sequence_sql(name, opts={})
+        seq_name = quote_identifier(name)
+        "ALTER SEQUENCE #{seq_name} RESTART WITH #{opts[:restart_position]}"
+      end
+
+      def type_literal_generic_string(column)
+        column[:text] ? :"BLOB SUBTYPE TEXT" : super
+      end
+    end
+
+    # Dataset class for Firebird datasets
+    class Dataset < Sequel::Dataset
+      include UnsupportedIntersectExcept
+
+      BOOL_TRUE = '1'.freeze
+      BOOL_FALSE = '0'.freeze
+      NULL = LiteralString.new('NULL').freeze
+      COMMA_SEPARATOR = ', '.freeze
+      FIREBIRD_TIMESTAMP_FORMAT = "TIMESTAMP '%Y-%m-%d %H:%M:%S".freeze
+      SELECT_CLAUSE_ORDER = %w'distinct limit columns from join where group having compounds order'.freeze
+
+      # Yield all rows returned by executing the given SQL and converting
+      # the types.
+      def fetch_rows(sql, &block)
+        execute(sql) do |s|
+          begin
+            @columns = s.fields.map{|c| output_identifier(c.name)}
+            s.fetchall(:symbols_hash).each do |r|
+              h = {}
+              r.each{|k,v| h[output_identifier(k)] = v}
+              yield h
+            end
+          ensure
+            s.close
+          end
+        end
+        self
+      end
+
+      # Insert given values into the database.
+      def insert(*values)
+        if !@opts[:sql]
+          clone(default_server_opts(:sql=>insert_returning_pk_sql(*values))).single_value
+        else
+          execute_insert(insert_sql(*values), :table=>opts[:from].first,
+            :values=>values.size == 1 ? values.first : values)
+        end
+      end
+
+      # Use the RETURNING clause to return the primary key of the inserted record, if it exists
+      def insert_returning_pk_sql(*values)
+        pk = db.primary_key(opts[:from].first)
+        insert_returning_sql(pk ? Sequel::SQL::Identifier.new(pk) : NULL, *values)
+      end
+
+      # Use the RETURNING clause to return the columns listed in returning.
+      def insert_returning_sql(returning, *values)
+        "#{insert_sql(*values)} RETURNING #{column_list(Array(returning))}"
+      end
+
+      # Insert a record returning the record inserted
+      def insert_select(*values)
+        naked.clone(default_server_opts(:sql=>insert_returning_sql(nil, *values))).single_record
+      end
+
+      # The order of clauses in the SELECT SQL statement
+      def select_clause_order
+        SELECT_CLAUSE_ORDER
+      end
+      
+      def select_limit_sql(sql)
+        sql << " FIRST #{@opts[:limit]}" if @opts[:limit]
+        sql << " SKIP #{@opts[:offset]}" if @opts[:offset]
+      end
+
+      private
+
+      def hash_row(stmt, row)
+        @columns.inject({}) do |m, c|
+          m[c] = row.shift
+          m
+        end
+      end
+
+      def literal_datetime(v)
+        "#{v.strftime(FIREBIRD_TIMESTAMP_FORMAT)}.#{sprintf("%04d",(v.sec_fraction * 864000000))}'"
+      end
+
+      def literal_false
+        BOOL_FALSE
+      end
+
+      def literal_time(v)
+        "#{v.strftime(FIREBIRD_TIMESTAMP_FORMAT)}.#{sprintf("%04d",v.usec / 100)}'"
+      end
+
+      def literal_true
+        BOOL_TRUE
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/adapters/informix.rb 3.0.0-1/lib/sequel/adapters/informix.rb
--- 1.5.1-1/lib/sequel/adapters/informix.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/adapters/informix.rb	2009-04-27 23:53:18.000000000 +0100
@@ -0,0 +1,77 @@
+Sequel.require 'adapters/utils/unsupported'
+require 'informix'
+
+module Sequel
+  module Informix
+    class Database < Sequel::Database
+      set_adapter_scheme :informix
+      
+      TEMPORARY = 'TEMP '.freeze
+      
+      def connect(server)
+        opts = server_opts(server)
+        ::Informix.connect(opts[:database], opts[:user], opts[:password])
+      end
+    
+      def dataset(opts = nil)
+        Sequel::Informix::Dataset.new(self, opts)
+      end
+      
+      # Returns number of rows affected
+      def execute_dui(sql, opts={})
+        log_info(sql)
+        synchronize(opts[:server]){|c| c.immediate(sql)}
+      end
+      alias_method :do, :execute_dui
+      
+      def execute(sql, opts={})
+        log_info(sql)
+        synchronize(opts[:server]){|c| yield c.cursor(sql)}
+      end
+      alias_method :query, :execute
+      
+      private
+
+      def disconnect_connection(c)
+        c.close
+      end
+    end
+    
+    class Dataset < Sequel::Dataset
+      include UnsupportedIntersectExcept
+
+      SELECT_CLAUSE_ORDER = %w'limit distinct columns from join where having group compounds order'.freeze
+
+      def fetch_rows(sql, &block)
+        execute(sql) do |cursor|
+          begin
+            col_map = nil
+            cursor.open.each_hash do |h|
+              unless col_map
+                col_map = {}
+                @columns = h.keys.map{|k| col_map[k] = output_identifier(k)}
+              end
+              h2 = {}
+              h.each{|k,v| h2[col_map[k]||k] = v}
+              yield h2
+            end
+          ensure
+            cursor.drop
+          end
+        end
+        self
+      end
+
+      private
+
+      def select_clause_order
+        SELECT_CLAUSE_ORDER
+      end
+
+      def select_limit_sql(sql)
+        sql << " SKIP #{@opts[:offset]}" if @opts[:offset]
+        sql << " FIRST #{@opts[:limit]}" if @opts[:limit]
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/adapters/jdbc/h2.rb 3.0.0-1/lib/sequel/adapters/jdbc/h2.rb
--- 1.5.1-1/lib/sequel/adapters/jdbc/h2.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/adapters/jdbc/h2.rb	2009-04-28 00:32:31.000000000 +0100
@@ -0,0 +1,73 @@
+Sequel.require %w'date_format unsupported', 'adapters/utils'
+
+module Sequel
+  module JDBC
+    # Database and Dataset support for H2 databases accessed via JDBC.
+    module H2
+      # Instance methods for H2 Database objects accessed via JDBC.
+      module DatabaseMethods
+        PRIMARY_KEY_INDEX_RE = /\Aprimary_key/i.freeze
+      
+        # H2 uses the :h2 database type.
+        def database_type
+          :h2
+        end
+
+        # Return Sequel::JDBC::H2::Dataset object with the given opts.
+        def dataset(opts=nil)
+          Sequel::JDBC::H2::Dataset.new(self, opts)
+        end
+        
+        # H2 uses an IDENTITY type
+        def serial_primary_key_options
+          {:primary_key => true, :type => :identity}
+        end
+        
+        private
+        
+        # H2 needs to add a primary key column as a constraint
+        def alter_table_sql(table, op)
+          case op[:op]
+          when :add_column
+            if op.delete(:primary_key)
+              sql = super(table, op)
+              [sql, "ALTER TABLE #{quote_schema_table(table)} ADD PRIMARY KEY (#{quote_identifier(op[:name])})"]
+            else
+              super(table, op)
+            end
+          else
+            super(table, op)
+          end
+        end
+        
+        # Default to a single connection for a memory database.
+        def connection_pool_default_options
+          o = super
+          uri == 'jdbc:h2:mem:' ? o.merge(:max_connections=>1) : o
+        end
+      
+        # Use IDENTITY() to get the last inserted id.
+        def last_insert_id(conn, opts={})
+          stmt = conn.createStatement
+          begin
+            rs = stmt.executeQuery('SELECT IDENTITY();')
+            rs.next
+            rs.getInt(1)
+          ensure
+            stmt.close
+          end 
+        end
+        
+        def primary_key_index_re
+          PRIMARY_KEY_INDEX_RE
+        end
+      end
+      
+      # Dataset class for H2 datasets accessed via JDBC.
+      class Dataset < JDBC::Dataset
+        include Dataset::SQLStandardDateFormat
+        include Dataset::UnsupportedIsTrue
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/adapters/jdbc/mysql.rb 3.0.0-1/lib/sequel/adapters/jdbc/mysql.rb
--- 1.5.1-1/lib/sequel/adapters/jdbc/mysql.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/adapters/jdbc/mysql.rb	2009-03-16 20:53:19.000000000 +0000
@@ -0,0 +1,56 @@
+Sequel.require 'adapters/shared/mysql'
+
+module Sequel
+  module JDBC
+    # Database and Dataset instance methods for MySQL specific
+    # support via JDBC.
+    module MySQL
+      # Database instance methods for MySQL databases accessed via JDBC.
+      module DatabaseMethods
+        include Sequel::MySQL::DatabaseMethods
+        
+        # Return instance of Sequel::JDBC::MySQL::Dataset with the given opts.
+        def dataset(opts=nil)
+          Sequel::JDBC::MySQL::Dataset.new(self, opts)
+        end
+        
+        private
+        
+        # The database name for the given database.  Need to parse it out
+        # of the connection string, since the JDBC does no parsing on the
+        # given connection string by default.
+        def database_name
+          u = URI.parse(uri.sub(/\Ajdbc:/, ''))
+          (m = /\/(.*)/.match(u.path)) && m[1]
+        end
+        
+        # Get the last inserted id using LAST_INSERT_ID().
+        def last_insert_id(conn, opts={})
+          stmt = conn.createStatement
+          begin
+            rs = stmt.executeQuery('SELECT LAST_INSERT_ID()')
+            rs.next
+            rs.getInt(1)
+          ensure
+            stmt.close
+          end
+        end
+      end
+      
+      # Dataset class for MySQL datasets accessed via JDBC.
+      class Dataset < JDBC::Dataset
+        include Sequel::MySQL::DatasetMethods
+        
+        # Use execute_insert to execute the insert_sql.
+        def insert(*values)
+          execute_insert(insert_sql(*values))
+        end
+        
+        # Use execute_insert to execute the replace_sql.
+        def replace(*args)
+          execute_insert(replace_sql(*args))
+        end
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/adapters/jdbc/oracle.rb 3.0.0-1/lib/sequel/adapters/jdbc/oracle.rb
--- 1.5.1-1/lib/sequel/adapters/jdbc/oracle.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/adapters/jdbc/oracle.rb	2009-03-16 20:53:19.000000000 +0000
@@ -0,0 +1,23 @@
+Sequel.require 'adapters/shared/oracle'
+
+module Sequel
+  module JDBC
+    # Database and Dataset support for Oracle databases accessed via JDBC.
+    module Oracle
+      # Instance methods for Oracle Database objects accessed via JDBC.
+      module DatabaseMethods
+        include Sequel::Oracle::DatabaseMethods
+        
+        # Return Sequel::JDBC::Oracle::Dataset object with the given opts.
+        def dataset(opts=nil)
+          Sequel::JDBC::Oracle::Dataset.new(self, opts)
+        end
+      end
+      
+      # Dataset class for Oracle datasets accessed via JDBC.
+      class Dataset < JDBC::Dataset
+        include Sequel::Oracle::DatasetMethods
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/adapters/jdbc/postgresql.rb 3.0.0-1/lib/sequel/adapters/jdbc/postgresql.rb
--- 1.5.1-1/lib/sequel/adapters/jdbc/postgresql.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/adapters/jdbc/postgresql.rb	2009-03-16 20:53:19.000000000 +0000
@@ -0,0 +1,101 @@
+Sequel.require 'adapters/shared/postgres'
+
+module Sequel
+  Postgres::CONVERTED_EXCEPTIONS << NativeException
+  
+  module JDBC
+    # Adapter, Database, and Dataset support for accessing a PostgreSQL
+    # database via JDBC.
+    module Postgres
+      # Methods to add to the JDBC adapter/connection to allow it to work
+      # with the shared PostgreSQL code.
+      module AdapterMethods
+        include Sequel::Postgres::AdapterMethods
+        
+        # Give the JDBC adapter a direct execute method, which creates
+        # a statement with the given sql and executes it.
+        def execute(sql, args=nil)
+          method = block_given? ? :executeQuery : :execute
+          stmt = createStatement
+          begin
+            rows = stmt.send(method, sql)
+            yield(rows) if block_given?
+          rescue NativeException => e
+            raise_error(e)
+          ensure
+            stmt.close
+          end
+        end
+        
+        private
+        
+        # JDBC specific method of getting specific values from a result set.
+        def single_value(r)
+          unless r.nil?
+            r.next
+            r.getString(1) unless r.getRow == 0
+          end
+        end
+      end
+    
+      # Methods to add to Database instances that access PostgreSQL via
+      # JDBC.
+      module DatabaseMethods
+        include Sequel::Postgres::DatabaseMethods
+        
+        # Add the primary_keys and primary_key_sequences instance variables,
+        # so we can get the correct return values for inserted rows.
+        def self.extended(db)
+          db.instance_eval do
+            @primary_keys = {}
+            @primary_key_sequences = {}
+          end
+        end
+        
+        # Return instance of Sequel::JDBC::Postgres::Dataset with the given opts.
+        def dataset(opts=nil)
+          Sequel::JDBC::Postgres::Dataset.new(self, opts)
+        end
+        
+        # Run the INSERT sql on the database and return the primary key
+        # for the record.
+        def execute_insert(sql, opts={})
+          super(sql, {:type=>:insert}.merge(opts))
+        end
+        
+        private
+        
+        # Extend the adapter with the JDBC PostgreSQL AdapterMethods
+        def setup_connection(conn)
+          conn = super(conn)
+          conn.extend(Sequel::JDBC::Postgres::AdapterMethods)
+          conn.db = self
+          conn.apply_connection_settings
+          conn
+        end
+        
+        # Call insert_result with the table and values specified in the opts.
+        def last_insert_id(conn, opts)
+          insert_result(conn, opts[:table], opts[:values])
+        end
+      end
+      
+      # Dataset subclass used for datasets that connect to PostgreSQL via JDBC.
+      class Dataset < JDBC::Dataset
+        include Sequel::Postgres::DatasetMethods
+
+        # Add the shared PostgreSQL prepared statement methods
+        def prepare(*args)
+          ps = super
+          ps.extend(::Sequel::Postgres::DatasetMethods::PreparedStatementMethods)
+          ps
+        end
+        
+        # Literalize strings similar to the native postgres adapter
+        def literal_string(v)
+          db.synchronize{|c| "'#{c.escape_string(v)}'"}
+        end
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/adapters/jdbc/sqlite.rb 3.0.0-1/lib/sequel/adapters/jdbc/sqlite.rb
--- 1.5.1-1/lib/sequel/adapters/jdbc/sqlite.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/adapters/jdbc/sqlite.rb	2009-03-16 20:53:19.000000000 +0000
@@ -0,0 +1,43 @@
+Sequel.require 'adapters/shared/sqlite'
+
+module Sequel
+  module JDBC
+    # Database and Dataset support for SQLite databases accessed via JDBC.
+    module SQLite
+      # Instance methods for SQLite Database objects accessed via JDBC.
+      module DatabaseMethods
+        include Sequel::SQLite::DatabaseMethods
+        
+        # Return Sequel::JDBC::SQLite::Dataset object with the given opts.
+        def dataset(opts=nil)
+          Sequel::JDBC::SQLite::Dataset.new(self, opts)
+        end
+        
+        private
+        
+        # Use last_insert_rowid() to get the last inserted id.
+        def last_insert_id(conn, opts={})
+          stmt = conn.createStatement
+          begin
+            rs = stmt.executeQuery('SELECT last_insert_rowid()')
+            rs.next
+            rs.getInt(1)
+          ensure
+            stmt.close
+          end
+        end
+        
+        # Default to a single connection for a memory database.
+        def connection_pool_default_options
+          o = super
+          uri == 'jdbc:sqlite::memory:' ? o.merge(:max_connections=>1) : o
+        end
+      end
+      
+      # Dataset class for SQLite datasets accessed via JDBC.
+      class Dataset < JDBC::Dataset
+        include Sequel::SQLite::DatasetMethods
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/adapters/jdbc.rb 3.0.0-1/lib/sequel/adapters/jdbc.rb
--- 1.5.1-1/lib/sequel/adapters/jdbc.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/adapters/jdbc.rb	2009-04-27 21:25:44.000000000 +0100
@@ -0,0 +1,490 @@
+require 'java'
+Sequel.require 'adapters/utils/stored_procedures'
+
+module Sequel
+  # Houses Sequel's JDBC support when running on JRuby.
+  # Support for individual database types is done using sub adapters.
+  # PostgreSQL, MySQL, SQLite, Oracle, and MSSQL all have relatively good support,
+  # close the the level supported by the native adapter.
+  # PostgreSQL, MySQL, SQLite can load necessary support using
+  # the jdbc-* gem, if it is installed, though they will work if you
+  # have the correct .jar in your CLASSPATH.  Oracle and MSSQL should
+  # load the necessary support if you have the .jar in your CLASSPATH.
+  # For all other databases, the Java class should be loaded manually
+  # before calling Sequel.connect.
+  #
+  # Note that when using a JDBC adapter, the best way to use Sequel
+  # is via Sequel.connect, NOT Sequel.jdbc.  Use the JDBC connection
+  # string when connecting, which will be in a different format than
+  # the native connection string.  The connection string should start
+  # with 'jdbc:'.  For PostgreSQL, use 'jdbc:postgresql:', and for
+  # SQLite you do not need 2 preceding slashes for the database name
+  # (use no preceding slashes for a relative path, and one preceding
+  # slash for an absolute path).
+  module JDBC
+    # Make it accesing the java.lang hierarchy more ruby friendly.
+    module JavaLang
+      include_package 'java.lang'
+    end
+    
+    # Make it accesing the java.sql hierarchy more ruby friendly.
+    module JavaSQL
+      include_package 'java.sql'
+    end
+    
+    # Contains procs keyed on sub adapter type that extend the
+    # given database object so it supports the correct database type.
+    DATABASE_SETUP = {:postgresql=>proc do |db|
+        Sequel.require 'adapters/jdbc/postgresql'
+        db.extend(Sequel::JDBC::Postgres::DatabaseMethods)
+        JDBC.load_gem('postgres')
+        org.postgresql.Driver
+      end,
+      :mysql=>proc do |db|
+        Sequel.require 'adapters/jdbc/mysql'
+        db.extend(Sequel::JDBC::MySQL::DatabaseMethods)
+        JDBC.load_gem('mysql')
+        com.mysql.jdbc.Driver
+      end,
+      :sqlite=>proc do |db|
+        Sequel.require 'adapters/jdbc/sqlite'
+        db.extend(Sequel::JDBC::SQLite::DatabaseMethods)
+        JDBC.load_gem('sqlite3')
+        org.sqlite.JDBC
+      end,
+      :oracle=>proc do |db|
+        Sequel.require 'adapters/jdbc/oracle'
+        db.extend(Sequel::JDBC::Oracle::DatabaseMethods)
+        Java::oracle.jdbc.driver.OracleDriver
+      end,
+      :sqlserver=>proc do |db|
+        Sequel.require 'adapters/shared/mssql'
+        db.extend(Sequel::MSSQL::DatabaseMethods)
+        com.microsoft.sqlserver.jdbc.SQLServerDriver
+      end,
+      :h2=>proc do |db|
+        Sequel.require 'adapters/jdbc/h2'
+        db.extend(Sequel::JDBC::H2::DatabaseMethods)
+        JDBC.load_gem('h2')
+        org.h2.Driver
+      end
+    }
+    
+    # Allowing loading the necessary JDBC support via a gem, which
+    # works for PostgreSQL, MySQL, and SQLite.
+    def self.load_gem(name)
+      begin
+        require "jdbc/#{name}"
+      rescue LoadError
+        # jdbc gem not used, hopefully the user has the .jar in their CLASSPATH
+      end
+    end
+
+    # JDBC Databases offer a fairly uniform interface that does not change
+    # much based on the sub adapter.
+    class Database < Sequel::Database
+      set_adapter_scheme :jdbc
+      
+      # The type of database we are connecting to
+      attr_reader :database_type
+      
+      # Call the DATABASE_SETUP proc directly after initialization,
+      # so the object always uses sub adapter specific code.  Also,
+      # raise an error immediately if the connection doesn't have a
+      # uri, since JDBC requires one.
+      def initialize(opts)
+        @opts = opts
+        raise(Error, "No connection string specified") unless uri
+        if match = /\Ajdbc:([^:]+)/.match(uri) and prok = DATABASE_SETUP[match[1].to_sym]
+          prok.call(self)
+        end
+        super(opts)
+      end
+      
+      # Execute the given stored procedure with the give name. If a block is
+      # given, the stored procedure should return rows.
+      def call_sproc(name, opts = {})
+        args = opts[:args] || []
+        sql = "{call #{name}(#{args.map{'?'}.join(',')})}"
+        synchronize(opts[:server]) do |conn|
+          cps = conn.prepareCall(sql)
+
+          i = 0
+          args.each{|arg| set_ps_arg(cps, arg, i+=1)}
+
+          begin
+            if block_given?
+              yield cps.executeQuery
+            else
+              case opts[:type]
+              when :insert
+                cps.executeUpdate
+                last_insert_id(conn, opts)
+              else
+                cps.executeUpdate
+              end
+            end
+          rescue NativeException, JavaSQL::SQLException => e
+            raise_error(e)
+          ensure
+            cps.close
+          end
+        end
+      end
+      
+      # Connect to the database using JavaSQL::DriverManager.getConnection.
+      def connect(server)
+        setup_connection(JavaSQL::DriverManager.getConnection(uri(server_opts(server))))
+      end
+      
+      # Return instances of JDBC::Dataset with the given opts.
+      def dataset(opts = nil)
+        JDBC::Dataset.new(self, opts)
+      end
+      
+      # Execute the given SQL.  If a block is given, if should be a SELECT
+      # statement or something else that returns rows.
+      def execute(sql, opts={}, &block)
+        return call_sproc(sql, opts, &block) if opts[:sproc]
+        return execute_prepared_statement(sql, opts, &block) if [Symbol, Dataset].any?{|c| sql.is_a?(c)}
+        log_info(sql)
+        synchronize(opts[:server]) do |conn|
+          stmt = conn.createStatement
+          begin
+            if block_given?
+              yield stmt.executeQuery(sql)
+            else
+              case opts[:type]
+              when :ddl
+                stmt.execute(sql)
+              when :insert
+                stmt.executeUpdate(sql)
+                last_insert_id(conn, opts)
+              else
+                stmt.executeUpdate(sql)
+              end
+            end
+          rescue NativeException, JavaSQL::SQLException => e
+            raise_error(e)
+          ensure
+            stmt.close
+          end
+        end
+      end
+      alias execute_dui execute
+      
+      # Execute the given DDL SQL, which should not return any
+      # values or rows.
+      def execute_ddl(sql, opts={})
+        execute(sql, {:type=>:ddl}.merge(opts))
+      end
+      
+      # Execute the given INSERT SQL, returning the last inserted
+      # row id.
+      def execute_insert(sql, opts={})
+        execute(sql, {:type=>:insert}.merge(opts))
+      end
+      
+      # Return a hash containing index information. Hash keys are index name symbols.
+      # Values are subhashes with two keys, :columns and :unique.  The value of :columns
+      # is an array of symbols of column names.  The value of :unique is true or false
+      # depending on if the index is unique.
+      def indexes(table)
+        indexes = {}
+        m = output_identifier_meth
+        metadata(:getIndexInfo, nil, nil, input_identifier_meth.call(table), false, true) do |r|
+          next unless name = r[:column_name]
+          next if respond_to?(:primary_key_index_re, true) and r[:index_name] =~ primary_key_index_re 
+          i = indexes[m.call(r[:index_name])] ||= {:columns=>[], :unique=>!r[:non_unique]}
+          i[:columns] << m.call(name)
+        end
+        indexes
+      end 
+
+      # All tables in this database
+      def tables
+        ts = []
+        m = output_identifier_meth
+        metadata(:getTables, nil, nil, nil, ['TABLE'].to_java(:string)){|h| ts << m.call(h[:table_name])}
+        ts
+      end
+      
+      # Default transaction method that should work on most JDBC
+      # databases.  Does not use the JDBC transaction methods, uses
+      # SQL BEGIN/ROLLBACK/COMMIT statements instead.
+      def transaction(opts={})
+        synchronize(opts[:server]) do |conn|
+          return yield(conn) if @transactions.include?(Thread.current)
+          stmt = conn.createStatement
+          begin
+            log_info(begin_transaction_sql)
+            stmt.execute(begin_transaction_sql)
+            @transactions << Thread.current
+            yield(conn)
+          rescue Exception => e
+            log_info(rollback_transaction_sql)
+            stmt.execute(rollback_transaction_sql)
+            transaction_error(e)
+          ensure
+            unless e
+              log_info(commit_transaction_sql)
+              stmt.execute(commit_transaction_sql)
+            end
+            stmt.close
+            @transactions.delete(Thread.current)
+          end
+        end
+      end
+      
+      # The uri for this connection.  You can specify the uri
+      # using the :uri, :url, or :database options.  You don't
+      # need to worry about this if you use Sequel.connect
+      # with the JDBC connectrion strings.
+      def uri(opts={})
+        opts = @opts.merge(opts)
+        ur = opts[:uri] || opts[:url] || opts[:database]
+        ur =~ /^\Ajdbc:/ ? ur : "jdbc:#{ur}"
+      end
+      
+      private
+      
+      # The JDBC adapter should not need the pool to convert exceptions.
+      def connection_pool_default_options
+        super.merge(:pool_convert_exceptions=>false)
+      end
+      
+      # Close given adapter connections
+      def disconnect_connection(c)
+        c.close
+      end
+      
+      # Execute the prepared statement.  If the provided name is a
+      # dataset, use that as the prepared statement, otherwise use
+      # it as a key to look it up in the prepared_statements hash.
+      # If the connection we are using has already prepared an identical
+      # statement, use that statement instead of creating another.
+      # Otherwise, prepare a new statement for the connection, bind the
+      # variables, and execute it.
+      def execute_prepared_statement(name, opts={})
+        args = opts[:arguments]
+        if Dataset === name
+          ps = name
+          name = ps.prepared_statement_name
+        else
+          ps = prepared_statements[name]
+        end
+        sql = ps.prepared_sql
+        synchronize(opts[:server]) do |conn|
+          if name and cps = conn.prepared_statements[name] and cps[0] == sql
+            cps = cps[1]
+          else
+            if cps
+              log_info("Closing #{name}")
+              cps[1].close
+            end
+            log_info("Preparing#{" #{name}:" if name} #{sql}")
+            cps = conn.prepareStatement(sql)
+            conn.prepared_statements[name] = [sql, cps] if name
+          end
+          i = 0
+          args.each{|arg| set_ps_arg(cps, arg, i+=1)}
+          log_info("Executing#{" #{name}" if name}", args)
+          begin
+            if block_given?
+              yield cps.executeQuery
+            else
+              case opts[:type]
+              when :ddl
+                cps.execute
+              when :insert
+                cps.executeUpdate
+                last_insert_id(conn, opts)
+              else
+                cps.executeUpdate
+              end
+            end
+          rescue NativeException, JavaSQL::SQLException => e
+            raise_error(e)
+          ensure
+            cps.close unless name
+          end
+        end
+      end
+      
+      # By default, there is no support for determining the last inserted
+      # id, so return nil.  This method should be overridden in
+      # sub adapters.
+      def last_insert_id(conn, opts)
+        nil
+      end
+      
+      # Yield the metadata for this database
+      def metadata(*args, &block)
+        synchronize{|c| metadata_dataset.send(:process_result_set, c.getMetaData.send(*args), &block)}
+      end
+      
+      # Java being java, you need to specify the type of each argument
+      # for the prepared statement, and bind it individually.  This
+      # guesses which JDBC method to use, and hopefully JRuby will convert
+      # things properly for us.
+      def set_ps_arg(cps, arg, i)
+        case arg
+        when Integer
+          cps.setInt(i, arg)
+        when String
+          cps.setString(i, arg)
+        when Date, Java::JavaSql::Date
+          cps.setDate(i, arg)
+        when Time, DateTime, Java::JavaSql::Timestamp
+          cps.setTimestamp(i, arg)
+        when Float
+          cps.setDouble(i, arg)
+        when nil
+          cps.setNull(i, JavaSQL::Types::NULL)
+        end
+      end
+      
+      # Add a prepared_statements accessor to the connection,
+      # and set it to an empty hash.  This is used to store
+      # adapter specific prepared statements.
+      def setup_connection(conn)
+        class << conn
+          attr_accessor :prepared_statements
+        end
+        conn.prepared_statements = {}
+        conn
+      end
+      
+      # All tables in this database
+      def schema_parse_table(table, opts={})
+        m = output_identifier_meth
+        im = input_identifier_meth
+        ds = dataset
+        schema, table = schema_and_table(table)
+        schema = im.call(schema) if schema
+        table = im.call(table)
+        pks, ts = [], []
+        metadata(:getPrimaryKeys, nil, schema, table) do |h|
+          pks << h[:column_name]
+        end
+        metadata(:getColumns, nil, schema, table, nil) do |h|
+          ts << [m.call(h[:column_name]), {:type=>schema_column_type(h[:type_name]), :db_type=>h[:type_name], :default=>(h[:column_def] == '' ? nil : h[:column_def]), :allow_null=>(h[:nullable] != 0), :primary_key=>pks.include?(h[:column_name])}]
+        end
+        ts
+      end
+    end
+    
+    class Dataset < Sequel::Dataset
+      include StoredProcedures
+      
+      # Use JDBC PreparedStatements instead of emulated ones.  Statements
+      # created using #prepare are cached at the connection level to allow
+      # reuse.  This also supports bind variables by using unnamed
+      # prepared statements created using #call.
+      module PreparedStatementMethods
+        include Sequel::Dataset::UnnumberedArgumentMapper
+        
+        private
+        
+        # Execute the prepared SQL using the stored type and
+        # arguments derived from the hash passed to call.
+        def execute(sql, opts={}, &block)
+          super(self, {:arguments=>bind_arguments, :type=>sql_query_type}.merge(opts), &block)
+        end
+        
+        # Same as execute, explicit due to intricacies of alias and super.
+        def execute_dui(sql, opts={}, &block)
+          super(self, {:arguments=>bind_arguments, :type=>sql_query_type}.merge(opts), &block)
+        end
+        
+        # Same as execute, explicit due to intricacies of alias and super.
+        def execute_insert(sql, opts={}, &block)
+          super(self, {:arguments=>bind_arguments, :type=>sql_query_type}.merge(opts), &block)
+        end
+      end
+      
+      # Use JDBC CallableStatements to execute stored procedures.  Only supported
+      # if the underlying database has stored procedure support.
+      module StoredProcedureMethods
+        include Sequel::Dataset::StoredProcedureMethods
+        
+        private
+        
+        # Execute the database stored procedure with the stored arguments.
+        def execute(sql, opts={}, &block)
+          super(@sproc_name, {:args=>@sproc_args, :sproc=>true, :type=>sql_query_type}.merge(opts), &block)
+        end
+        
+        # Same as execute, explicit due to intricacies of alias and super.
+        def execute_dui(sql, opts={}, &block)
+          super(@sproc_name, {:args=>@sproc_args, :sproc=>true, :type=>sql_query_type}.merge(opts), &block)
+        end
+        
+        # Same as execute, explicit due to intricacies of alias and super.
+        def execute_insert(sql, opts={}, &block)
+          super(@sproc_name, {:args=>@sproc_args, :sproc=>true, :type=>sql_query_type}.merge(opts), &block)
+        end
+      end
+      
+      # Correctly return rows from the database and return them as hashes.
+      def fetch_rows(sql, &block)
+        execute(sql){|result| process_result_set(result, &block)}
+        self
+      end
+      
+      # Create a named prepared statement that is stored in the
+      # database (and connection) for reuse.
+      def prepare(type, name=nil, values=nil)
+        ps = to_prepared_statement(type, values)
+        ps.extend(PreparedStatementMethods)
+        if name
+          ps.prepared_statement_name = name
+          db.prepared_statements[name] = ps
+        end
+        ps
+      end
+      
+      private
+      
+      # Convert the type.  Used for converting Java types to ruby types.
+      def convert_type(v)
+        case v
+        when Java::JavaSQL::Timestamp, Java::JavaSQL::Time
+          Sequel.string_to_datetime(v.to_string)
+        when Java::JavaSQL::Date
+          Sequel.string_to_date(v.to_string)
+        when Java::JavaIo::BufferedReader
+          lines = []
+          while(line = v.read_line) do lines << line end
+          lines.join("\n")
+        when Java::JavaMath::BigDecimal
+          BigDecimal.new(v.to_string)
+        else
+          v
+        end
+      end
+      
+      # Extend the dataset with the JDBC stored procedure methods.
+      def prepare_extend_sproc(ds)
+        ds.extend(StoredProcedureMethods)
+      end
+      
+      # Split out from fetch rows to allow processing of JDBC result sets
+      # that don't come from issuing an SQL string.
+      def process_result_set(result)
+        # get column names
+        meta = result.getMetaData
+        column_count = meta.getColumnCount
+        @columns = []
+        column_count.times{|i| @columns << output_identifier(meta.getColumnLabel(i+1))}
+
+        # get rows
+        while result.next
+          row = {}
+          @columns.each_with_index{|v, i| row[v] = convert_type(result.getObject(i+1))}
+          yield row
+        end
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/adapters/mysql.rb 3.0.0-1/lib/sequel/adapters/mysql.rb
--- 1.5.1-1/lib/sequel/adapters/mysql.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/adapters/mysql.rb	2009-04-09 20:47:29.000000000 +0100
@@ -0,0 +1,379 @@
+require 'mysql'
+Sequel.require %w'shared/mysql utils/stored_procedures', 'adapters'
+
+module Sequel
+  # Module for holding all MySQL-related classes and modules for Sequel.
+  #
+  # A class level convert_invalid_date_time accessor exists if
+  # the native adapter is used.  If set to nil or :nil, the adapter treats dates
+  # like 0000-00-00 and times like 838:00:00 as nil values.  If set to :string,
+  # it returns the strings as is.  If is false by default, which means that
+  # invalid dates and times will raise errors.
+  module MySQL
+    # Mapping of type numbers to conversion procs
+    MYSQL_TYPES = {}
+
+    # Use only a single proc for each type to save on memory
+    MYSQL_TYPE_PROCS = {
+      [0, 246]  => lambda{|v| BigDecimal.new(v)},                       # decimal
+      [1, 2, 3, 8, 9, 13, 247, 248]  => lambda{|v| v.to_i},             # integer
+      [4, 5]  => lambda{|v| v.to_f},                                    # float
+      [10, 14]  => lambda{|v| convert_date_time(:string_to_date, v)},   # date
+      [7, 12] => lambda{|v| convert_date_time(:string_to_datetime, v)}, # datetime
+      [11]  => lambda{|v| convert_date_time(:string_to_time, v)},       # time
+      [249, 250, 251, 252]  => lambda{|v| Sequel::SQL::Blob.new(v)}     # blob
+    }
+    MYSQL_TYPE_PROCS.each do |k,v|
+      k.each{|n| MYSQL_TYPES[n] = v}
+    end
+    
+    @convert_invalid_date_time = false
+
+    class << self
+      attr_accessor :convert_invalid_date_time
+    end
+
+    # If convert_invalid_date_time is nil, :nil, or :string and
+    # the conversion raises an InvalidValue exception, return v
+    # if :string and nil otherwise.
+    def self.convert_date_time(meth, v)
+      begin
+        Sequel.send(meth, v)
+      rescue InvalidValue
+        case @convert_invalid_date_time
+        when nil, :nil
+          nil
+        when :string
+          v
+        else 
+          raise
+        end
+      end
+    end
+  
+    # Database class for MySQL databases used with Sequel.
+    class Database < Sequel::Database
+      include Sequel::MySQL::DatabaseMethods
+      
+      set_adapter_scheme :mysql
+      
+      # Support stored procedures on MySQL
+      def call_sproc(name, opts={}, &block)
+        args = opts[:args] || [] 
+        execute("CALL #{name}#{args.empty? ? '()' : literal(args)}", opts.merge(:sproc=>false), &block)
+      end
+      
+      # Connect to the database.  In addition to the usual database options,
+      # the following options have effect:
+      #
+      # * :auto_is_null - Set to true to use MySQL default behavior of having
+      #   a filter for an autoincrement column equals NULL to return the last
+      #   inserted row.
+      # * :charset - Same as :encoding (:encoding takes precendence)
+      # * :compress - Set to false to not compress results from the server
+      # * :encoding - Set all the related character sets for this
+      #   connection (connection, client, database, server, and results).
+      # * :socket - Use a unix socket file instead of connecting via TCP/IP.
+      # * :timeout - Set the timeout in seconds before the server will
+      #   disconnect this connection.
+      def connect(server)
+        opts = server_opts(server)
+        conn = Mysql.init
+        conn.options(Mysql::OPT_LOCAL_INFILE, "client")
+        if encoding = opts[:encoding] || opts[:charset]
+          # set charset _before_ the connect. using an option instead of "SET (NAMES|CHARACTER_SET_*)" works across reconnects
+          conn.options(Mysql::SET_CHARSET_NAME, encoding)
+        end
+        conn.real_connect(
+          opts[:host] || 'localhost',
+          opts[:user],
+          opts[:password],
+          opts[:database],
+          opts[:port],
+          opts[:socket],
+          Mysql::CLIENT_MULTI_RESULTS +
+          Mysql::CLIENT_MULTI_STATEMENTS +
+          (opts[:compress] == false ? 0 : Mysql::CLIENT_COMPRESS)
+        )
+
+        # increase timeout so mysql server doesn't disconnect us
+        conn.query("set @@wait_timeout = #{opts[:timeout] || 2592000}")
+
+        # By default, MySQL 'where id is null' selects the last inserted id
+        conn.query("set SQL_AUTO_IS_NULL=0") unless opts[:auto_is_null]
+
+        conn.query_with_result = false
+        class << conn
+          attr_accessor :prepared_statements
+        end
+        conn.prepared_statements = {}
+        conn.reconnect = true
+        conn
+      end
+      
+      # Returns instance of Sequel::MySQL::Dataset with the given options.
+      def dataset(opts = nil)
+        MySQL::Dataset.new(self, opts)
+      end
+      
+      # Executes the given SQL using an available connection, yielding the
+      # connection if the block is given.
+      def execute(sql, opts={}, &block)
+        return call_sproc(sql, opts, &block) if opts[:sproc]
+        return execute_prepared_statement(sql, opts, &block) if Symbol === sql
+        begin
+          synchronize(opts[:server]){|conn| _execute(conn, sql, opts, &block)}
+        rescue Mysql::Error => e
+          raise_error(e)
+        end
+      end
+      
+      # Return the version of the MySQL server two which we are connecting.
+      def server_version(server=nil)
+        @server_version ||= (synchronize(server){|conn| conn.server_version if conn.respond_to?(:server_version)} || super)
+      end
+      
+      # Support single level transactions on MySQL.
+      def transaction(opts={})
+        synchronize(opts[:server]) do |conn|
+          return yield(conn) if @transactions.include?(Thread.current)
+          log_info(begin_transaction_sql)
+          conn.query(begin_transaction_sql)
+          begin
+            @transactions << Thread.current
+            yield(conn)
+          rescue ::Exception => e
+            log_info(rollback_transaction_sql)
+            conn.query(rollback_transaction_sql)
+            transaction_error(e, Mysql::Error)
+          ensure
+            unless e
+              log_info(commit_transaction_sql)
+              conn.query(commit_transaction_sql)
+            end
+            @transactions.delete(Thread.current)
+          end
+        end
+      end
+
+      private
+      
+      # Execute the given SQL on the given connection.  If the :type
+      # option is :select, yield the result of the query, otherwise
+      # yield the connection if a block is given.
+      def _execute(conn, sql, opts)
+        log_info(sql)
+        conn.query(sql)
+        if opts[:type] == :select
+          loop do
+            begin
+              r = conn.use_result
+            rescue Mysql::Error
+              nil
+            else
+              begin
+                yield r
+              ensure
+                r.free
+              end
+            end
+            break unless conn.respond_to?(:next_result) && conn.next_result
+          end
+        else
+          yield conn if block_given?
+        end
+      end
+      
+      # MySQL doesn't need the connection pool to convert exceptions.
+      def connection_pool_default_options
+        super.merge(:pool_convert_exceptions=>false)
+      end
+      
+      # The database name when using the native adapter is always stored in
+      # the :database option.
+      def database_name
+        @opts[:database]
+      end
+      
+      # Closes given database connection.
+      def disconnect_connection(c)
+        c.close
+      end
+      
+      # Executes a prepared statement on an available connection.  If the
+      # prepared statement already exists for the connection and has the same
+      # SQL, reuse it, otherwise, prepare the new statement.  Because of the
+      # usual MySQL stupidity, we are forced to name arguments via separate
+      # SET queries.  Use @sequel_arg_N (for N starting at 1) for these
+      # arguments.
+      def execute_prepared_statement(ps_name, opts, &block)
+        args = opts[:arguments]
+        ps = prepared_statements[ps_name]
+        sql = ps.prepared_sql
+        synchronize(opts[:server]) do |conn|
+          unless conn.prepared_statements[ps_name] == sql
+            conn.prepared_statements[ps_name] = sql
+            s = "PREPARE #{ps_name} FROM '#{::Mysql.quote(sql)}'"
+            log_info(s)
+            conn.query(s)
+          end
+          i = 0
+          args.each do |arg|
+            s = "SET @sequel_arg_#{i+=1} = #{literal(arg)}"
+            log_info(s)
+            conn.query(s)
+          end
+          _execute(conn, "EXECUTE #{ps_name}#{" USING #{(1..i).map{|j| "@sequel_arg_#{j}"}.join(', ')}" unless i == 0}", opts, &block)
+        end
+      end
+    end
+    
+    # Dataset class for MySQL datasets accessed via the native driver.
+    class Dataset < Sequel::Dataset
+      include Sequel::MySQL::DatasetMethods
+      include StoredProcedures
+      
+      # Methods to add to MySQL prepared statement calls without using a
+      # real database prepared statement and bound variables.
+      module CallableStatementMethods
+        # Extend given dataset with this module so subselects inside subselects in
+        # prepared statements work.
+        def subselect_sql(ds)
+          ps = ds.to_prepared_statement(:select)
+          ps.extend(CallableStatementMethods)
+          ps.prepared_args = prepared_args
+          ps.prepared_sql
+        end
+      end
+      
+      # Methods for MySQL prepared statements using the native driver.
+      module PreparedStatementMethods
+        include Sequel::Dataset::UnnumberedArgumentMapper
+        
+        private
+        
+        # Execute the prepared statement with the bind arguments instead of
+        # the given SQL.
+        def execute(sql, opts={}, &block)
+          super(prepared_statement_name, {:arguments=>bind_arguments}.merge(opts), &block)
+        end
+        
+        # Same as execute, explicit due to intricacies of alias and super.
+        def execute_dui(sql, opts={}, &block)
+          super(prepared_statement_name, {:arguments=>bind_arguments}.merge(opts), &block)
+        end
+      end
+      
+      # Methods for MySQL stored procedures using the native driver.
+      module StoredProcedureMethods
+        include Sequel::Dataset::StoredProcedureMethods
+        
+        private
+        
+        # Execute the database stored procedure with the stored arguments.
+        def execute(sql, opts={}, &block)
+          super(@sproc_name, {:args=>@sproc_args, :sproc=>true}.merge(opts), &block)
+        end
+        
+        # Same as execute, explicit due to intricacies of alias and super.
+        def execute_dui(sql, opts={}, &block)
+          super(@sproc_name, {:args=>@sproc_args, :sproc=>true}.merge(opts), &block)
+        end
+      end
+      
+      # MySQL is different in that it supports prepared statements but not bound
+      # variables outside of prepared statements.  The default implementation
+      # breaks the use of subselects in prepared statements, so extend the
+      # temporary prepared statement that this creates with a module that
+      # fixes it.
+      def call(type, bind_arguments={}, values=nil)
+        ps = to_prepared_statement(type, values)
+        ps.extend(CallableStatementMethods)
+        ps.call(bind_arguments)
+      end
+      
+      # Delete rows matching this dataset
+      def delete
+        execute_dui(delete_sql){|c| c.affected_rows}
+      end
+      
+      # Yield all rows matching this dataset
+      def fetch_rows(sql)
+        execute(sql) do |r|
+          column_types = []
+          @columns = r.fetch_fields.map{|f| column_types << f.type; output_identifier(f.name)}
+          while row = r.fetch_row
+            h = {}
+            @columns.each_with_index {|f, i| h[f] = convert_type(row[i], column_types[i])}
+            yield h
+          end
+        end
+        self
+      end
+      
+      # Insert a new value into this dataset
+      def insert(*values)
+        execute_dui(insert_sql(*values)){|c| c.insert_id}
+      end
+      
+      # Store the given type of prepared statement in the associated database
+      # with the given name.
+      def prepare(type, name=nil, values=nil)
+        ps = to_prepared_statement(type, values)
+        ps.extend(PreparedStatementMethods)
+        if name
+          ps.prepared_statement_name = name
+          db.prepared_statements[name] = ps
+        end
+        ps
+      end
+      
+      # Replace (update or insert) the matching row.
+      def replace(*args)
+        execute_dui(replace_sql(*args)){|c| c.insert_id}
+      end
+      
+      # Update the matching rows.
+      def update(values={})
+        execute_dui(update_sql(values)){|c| c.affected_rows}
+      end
+      
+      private
+      
+      # Convert the type of v using the method in MYSQL_TYPES[type].
+      def convert_type(v, type)
+        if v
+          if type == 1 && Sequel.convert_tinyint_to_bool
+            # We special case tinyint here to avoid adding
+            # a method to an ancestor of Fixnum
+            v.to_i == 0 ? false : true
+          else
+            (b = MYSQL_TYPES[type]) ? b.call(v) : v
+          end
+        else
+          nil
+        end
+      end
+      
+      # Set the :type option to :select if it hasn't been set.
+      def execute(sql, opts={}, &block)
+        super(sql, {:type=>:select}.merge(opts), &block)
+      end
+      
+      # Set the :type option to :dui if it hasn't been set.
+      def execute_dui(sql, opts={}, &block)
+        super(sql, {:type=>:dui}.merge(opts), &block)
+      end
+      
+      # Handle correct quoting of strings using ::MySQL.quote.
+      def literal_string(v)
+        "'#{::Mysql.quote(v)}'"
+      end
+      
+      # Extend the dataset with the MySQL stored procedure methods.
+      def prepare_extend_sproc(ds)
+        ds.extend(StoredProcedureMethods)
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/adapters/odbc.rb 3.0.0-1/lib/sequel/adapters/odbc.rb
--- 1.5.1-1/lib/sequel/adapters/odbc.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/adapters/odbc.rb	2009-04-09 19:04:21.000000000 +0100
@@ -0,0 +1,180 @@
+require 'odbc'
+
+module Sequel
+  module ODBC
+    class Database < Sequel::Database
+      set_adapter_scheme :odbc
+
+      GUARDED_DRV_NAME = /^\{.+\}$/.freeze
+      DRV_NAME_GUARDS = '{%s}'.freeze
+
+      def initialize(opts)
+        super(opts)
+        case opts[:db_type]
+        when 'mssql'
+          Sequel.require 'adapters/shared/mssql'
+          extend Sequel::MSSQL::DatabaseMethods
+        when 'progress'
+          Sequel.require 'adapters/shared/progress'
+          extend Sequel::Progress::DatabaseMethods
+        end
+      end
+
+      def connect(server)
+        opts = server_opts(server)
+        if opts.include? :driver
+          drv = ::ODBC::Driver.new
+          drv.name = 'Sequel ODBC Driver130'
+          opts.each do |param, value|
+            if :driver == param and not (value =~ GUARDED_DRV_NAME)
+              value = DRV_NAME_GUARDS % value
+            end
+            drv.attrs[param.to_s.capitalize] = value
+          end
+          db = ::ODBC::Database.new
+          conn = db.drvconnect(drv)
+        else
+          conn = ::ODBC::connect(opts[:database], opts[:user], opts[:password])
+        end
+        conn.autocommit = true
+        conn
+      end      
+
+      def dataset(opts = nil)
+        ODBC::Dataset.new(self, opts)
+      end
+    
+      # ODBC returns native statement objects, which must be dropped if
+      # you call execute manually, or you will get warnings.  See the
+      # fetch_rows method source code for an example of how to drop
+      # the statements.
+      def execute(sql, opts={})
+        log_info(sql)
+        synchronize(opts[:server]) do |conn|
+          r = conn.run(sql)
+          yield(r) if block_given?
+          r
+        end
+      end
+      
+      def execute_dui(sql, opts={})
+        log_info(sql)
+        synchronize(opts[:server]){|conn| conn.do(sql)}
+      end
+      alias_method :do, :execute_dui
+
+      # Support single level transactions on ODBC
+      def transaction(opts={})
+        synchronize(opts[:server]) do |conn|
+          return yield(conn) if @transactions.include?(Thread.current)
+          log_info(begin_transaction_sql)
+          conn.do(begin_transaction_sql)
+          begin
+            @transactions << Thread.current
+            yield(conn)
+          rescue ::Exception => e
+            log_info(rollback_transaction_sql)
+            conn.do(rollback_transaction_sql)
+            transaction_error(e)
+          ensure
+            unless e
+              log_info(commit_transaction_sql)
+              conn.do(commit_transaction_sql)
+            end
+            @transactions.delete(Thread.current)
+          end
+        end
+      end
+
+      private
+
+      def disconnect_connection(c)
+        c.disconnect
+      end
+    end
+    
+    class Dataset < Sequel::Dataset
+      BOOL_TRUE = '1'.freeze
+      BOOL_FALSE = '0'.freeze
+      ODBC_TIMESTAMP_FORMAT = "{ts '%Y-%m-%d %H:%M:%S'}".freeze
+      ODBC_TIMESTAMP_AFTER_SECONDS =
+        ODBC_TIMESTAMP_FORMAT.index( '%S' ).succ - ODBC_TIMESTAMP_FORMAT.length
+      ODBC_DATE_FORMAT = "{d '%Y-%m-%d'}".freeze
+      UNTITLED_COLUMN = 'untitled_%d'.freeze
+
+      def fetch_rows(sql, &block)
+        execute(sql) do |s|
+          begin
+            untitled_count = 0
+            @columns = s.columns(true).map do |c|
+              if (n = c.name).empty?
+                n = UNTITLED_COLUMN % (untitled_count += 1)
+              end
+              output_identifier(n)
+            end
+            rows = s.fetch_all
+            rows.each {|row| yield hash_row(row)} if rows
+          ensure
+            s.drop unless s.nil? rescue nil
+          end
+        end
+        self
+      end
+      
+      private
+
+      def convert_odbc_value(v)
+        # When fetching a result set, the Ruby ODBC driver converts all ODBC 
+        # SQL types to an equivalent Ruby type; with the exception of
+        # SQL_TYPE_DATE, SQL_TYPE_TIME and SQL_TYPE_TIMESTAMP.
+        #
+        # The conversions below are consistent with the mappings in
+        # ODBCColumn#mapSqlTypeToGenericType and Column#klass.
+        case v
+        when ::ODBC::TimeStamp
+          DateTime.new(v.year, v.month, v.day, v.hour, v.minute, v.second)
+        when ::ODBC::Time
+          now = DateTime.now
+          Time.gm(now.year, now.month, now.day, v.hour, v.minute, v.second)
+        when ::ODBC::Date
+          Date.new(v.year, v.month, v.day)
+        else
+          v
+        end
+      end
+      
+      def hash_row(row)
+        hash = {}
+        row.each_with_index do |v, idx|
+          hash[@columns[idx]] = convert_odbc_value(v)
+        end
+        hash
+      end
+      
+      def literal_date(v)
+        v.strftime(ODBC_DATE_FORMAT)
+      end
+      
+      def literal_datetime(v)
+        formatted = v.strftime(ODBC_TIMESTAMP_FORMAT)
+        usec = v.sec_fraction * 86400000000
+        formatted.insert(ODBC_TIMESTAMP_AFTER_SECONDS, ".#{(usec.to_f/1000).round}") if usec >= 1000
+        formatted
+      end
+      
+      def literal_false
+        BOOL_FALSE
+      end
+      
+      def literal_true
+        BOOL_TRUE
+      end
+
+      def literal_time(v)
+        formatted = v.strftime(ODBC_TIMESTAMP_FORMAT)
+        formatted.insert(ODBC_TIMESTAMP_AFTER_SECONDS, ".#{(v.usec.to_f/1000).round}") if usec >= 1000
+        formatted
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/adapters/openbase.rb 3.0.0-1/lib/sequel/adapters/openbase.rb
--- 1.5.1-1/lib/sequel/adapters/openbase.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/adapters/openbase.rb	2009-03-16 20:53:19.000000000 +0000
@@ -0,0 +1,57 @@
+require 'openbase'
+
+module Sequel
+  module OpenBase
+    class Database < Sequel::Database
+      set_adapter_scheme :openbase
+      
+      def connect(server)
+        opts = server_opts(server)
+        OpenBase.new(
+          opts[:database],
+          opts[:host] || 'localhost',
+          opts[:user],
+          opts[:password]
+        )
+      end
+      
+      def dataset(opts = nil)
+        OpenBase::Dataset.new(self, opts)
+      end
+    
+      def execute(sql, opts={})
+        log_info(sql)
+        synchronize(opts[:server]) do |conn|
+          r = conn.execute(sql)
+          yield(r) if block_given?
+          r
+        end
+      end
+      alias_method :do, :execute
+
+      private
+
+      def disconnect_connection(c)
+        c.disconnect
+      end
+    end
+    
+    class Dataset < Sequel::Dataset
+      def fetch_rows(sql)
+        execute(sql) do |result|
+          begin
+            @columns = result.column_infos.map{|c| output_identifier(c.name)}
+            result.each do |r|
+              row = {}
+              r.each_with_index {|v, i| row[@columns[i]] = v}
+              yield row
+            end
+          ensure
+            # result.close
+          end
+        end
+        self
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/adapters/oracle.rb 3.0.0-1/lib/sequel/adapters/oracle.rb
--- 1.5.1-1/lib/sequel/adapters/oracle.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/adapters/oracle.rb	2009-04-09 19:04:27.000000000 +0100
@@ -0,0 +1,136 @@
+require 'oci8'
+Sequel.require 'adapters/shared/oracle'
+
+module Sequel
+  module Oracle
+    class Database < Sequel::Database
+      include DatabaseMethods
+      set_adapter_scheme :oracle
+
+      # ORA-00028: your session has been killed
+      # ORA-01012: not logged on
+      # ORA-03113: end-of-file on communication channel
+      # ORA-03114: not connected to ORACLE
+      CONNECTION_ERROR_CODES = [ 28, 1012, 3113, 3114 ]      
+      
+      def connect(server)
+        opts = server_opts(server)
+        if opts[:database]
+          dbname = opts[:host] ? \
+            "//#{opts[:host]}#{":#{opts[:port]}" if opts[:port]}/#{opts[:database]}" : opts[:database]
+        else
+          dbname = opts[:host]
+        end
+        conn = OCI8.new(opts[:user], opts[:password], dbname, opts[:privilege])
+        conn.autocommit = true
+        conn.non_blocking = true
+        conn
+      end
+      
+      def dataset(opts = nil)
+        Oracle::Dataset.new(self, opts)
+      end
+
+      def schema_parse_table(table, opts={})
+        ds = dataset
+        ds.identifier_output_method = :downcase
+        schema, table = schema_and_table(table)
+        table_schema = []
+        metadata = transaction(opts){|conn| conn.describe_table(table.to_s)}
+        metadata.columns.each do |column|
+          table_schema << [
+            column.name.downcase.to_sym,
+            {
+              :type => column.data_type,
+              :db_type => column.type_string.split(' ')[0],
+              :type_string => column.type_string,
+              :charset_form => column.charset_form,
+              :char_used => column.char_used?,
+              :char_size => column.char_size,
+              :data_size => column.data_size,
+              :precision => column.precision,
+              :scale => column.scale,
+              :fsprecision => column.fsprecision,
+              :lfprecision => column.lfprecision,
+              :allow_null => column.nullable?
+            }
+          ]
+        end
+        table_schema
+      end
+
+      def execute(sql, opts={})
+        log_info(sql)
+        synchronize(opts[:server]) do |conn|
+          begin
+            r = conn.exec(sql)
+            yield(r) if block_given?
+            r
+          rescue OCIException => e
+            if CONNECTION_ERROR_CODES.include?(e.code)  
+              raise(Sequel::DatabaseDisconnectError)              
+            else
+              raise
+            end
+          end
+        end
+      end
+      alias do execute
+      
+      def transaction(opts={})
+        synchronize(opts[:server]) do |conn|
+          return yield(conn) if @transactions.include?(Thread.current)
+          conn.autocommit = false
+          begin
+            @transactions << Thread.current
+            yield(conn)
+          rescue => e
+            conn.rollback
+            raise e unless Rollback === e
+          ensure
+            conn.commit unless e
+            conn.autocommit = true
+            @transactions.delete(Thread.current)
+          end
+        end
+      end
+
+      private
+
+      def disconnect_connection(c)
+        c.logoff
+      end
+    end
+    
+    class Dataset < Sequel::Dataset
+      include DatasetMethods
+
+      def fetch_rows(sql, &block)
+        execute(sql) do |cursor|
+          begin
+            @columns = cursor.get_col_names.map{|c| output_identifier(c)}
+            while r = cursor.fetch
+              row = {}
+              r.each_with_index {|v, i| row[@columns[i]] = v unless @columns[i] == :raw_rnum_}
+              yield row
+            end
+          ensure
+            cursor.close
+          end
+        end
+        self
+      end
+
+      private
+
+      def literal_other(v)
+        case v
+        when OraDate
+          literal_time(Time.local(*v.to_a))
+        else
+          super
+        end
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/adapters/postgres.rb 3.0.0-1/lib/sequel/adapters/postgres.rb
--- 1.5.1-1/lib/sequel/adapters/postgres.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/adapters/postgres.rb	2009-03-23 23:34:01.000000000 +0000
@@ -0,0 +1,453 @@
+Sequel.require 'adapters/shared/postgres'
+
+begin 
+  require 'pg' 
+  SEQUEL_POSTGRES_USES_PG = true
+rescue LoadError => e 
+  SEQUEL_POSTGRES_USES_PG = false
+  begin
+    require 'postgres'
+    # Attempt to get uniform behavior for the PGconn object no matter
+    # if pg, postgres, or postgres-pr is used.
+    class PGconn
+      unless method_defined?(:escape_string)
+        if self.respond_to?(:escape)
+          # If there is no escape_string instead method, but there is an
+          # escape class method, use that instead.
+          def escape_string(str)
+            Sequel::Postgres.force_standard_strings ? str.gsub("'", "''") : self.class.escape(str)
+          end
+        else
+          # Raise an error if no valid string escaping method can be found.
+          def escape_string(obj)
+            raise Sequel::Error, "string escaping not supported with this postgres driver.  Try using ruby-pg, ruby-postgres, or postgres-pr."
+          end
+        end
+      end
+      unless method_defined?(:escape_bytea)
+        if self.respond_to?(:escape_bytea)
+          # If there is no escape_bytea instance method, but there is an
+          # escape_bytea class method, use that instead.
+          def escape_bytea(obj)
+            self.class.escape_bytea(obj)
+          end
+        else
+          begin
+            require 'postgres-pr/typeconv/conv'
+            require 'postgres-pr/typeconv/bytea'
+            extend Postgres::Conversion
+            # If we are using postgres-pr, use the encode_bytea method from
+            # that.
+            def escape_bytea(obj)
+              self.class.encode_bytea(obj)
+            end
+            instance_eval{alias unescape_bytea decode_bytea}
+          rescue
+            # If no valid bytea escaping method can be found, create one that
+            # raises an error
+            def escape_bytea(obj)
+              raise Sequel::Error, "bytea escaping not supported with this postgres driver.  Try using ruby-pg, ruby-postgres, or postgres-pr."
+            end
+            # If no valid bytea unescaping method can be found, create one that
+            # raises an error
+            def self.unescape_bytea(obj)
+              raise Sequel::Error, "bytea unescaping not supported with this postgres driver.  Try using ruby-pg, ruby-postgres, or postgres-pr."
+            end
+          end
+        end
+      end
+      alias_method :finish, :close unless method_defined?(:finish)
+      alias_method :async_exec, :exec unless method_defined?(:async_exec)
+      unless method_defined?(:block)
+        def block(timeout=nil)
+        end
+      end
+      unless defined?(CONNECTION_OK)
+        CONNECTION_OK = -1
+      end
+    end
+    class PGresult 
+      alias_method :nfields, :num_fields unless method_defined?(:nfields) 
+      alias_method :ntuples, :num_tuples unless method_defined?(:ntuples) 
+      alias_method :ftype, :type unless method_defined?(:ftype) 
+      alias_method :fname, :fieldname unless method_defined?(:fname) 
+      alias_method :cmd_tuples, :cmdtuples unless method_defined?(:cmd_tuples) 
+    end 
+  rescue LoadError 
+    raise e 
+  end 
+end
+
+module Sequel
+  module Postgres
+    CONVERTED_EXCEPTIONS << PGError
+    
+    # Hash with integer keys and proc values for converting PostgreSQL types.
+    PG_TYPES = {}
+
+    # Use a single proc for each type to conserve memory
+    PG_TYPE_PROCS  = {
+      [16] =>  lambda{|s| s == 't'}, # boolean
+      [17] => lambda{|s| ::Sequel::SQL::Blob.new(Adapter.unescape_bytea(s))}, # bytea
+      [20, 21, 22, 23, 26] => lambda{|s| s.to_i}, # integer
+      [700, 701] => lambda{|s| s.to_f}, # float
+      [790, 1700] => lambda{|s| BigDecimal.new(s)}, # numeric
+      [1082] => lambda{|s| @use_iso_date_format ? Date.new(*s.split("-").map{|x| x.to_i}) : Sequel.string_to_date(s)}, # date
+      [1083, 1266] => lambda{|s| Sequel.string_to_time(s)}, # time
+      [1114, 1184] => lambda{|s| Sequel.string_to_datetime(s)}, # timestamp
+    }
+    PG_TYPE_PROCS.each do |k,v|
+      k.each{|n| PG_TYPES[n] = v}
+    end
+    
+    @use_iso_date_format = true
+
+    class << self
+      # As an optimization, Sequel sets the date style to ISO, so that PostgreSQL provides
+      # the date in a known format that Sequel can parse faster.  This can be turned off
+      # if you require a date style other than ISO.
+      attr_accessor :use_iso_date_format
+    end
+    
+    # PGconn subclass for connection specific methods used with the
+    # pg, postgres, or postgres-pr driver.
+    class Adapter < ::PGconn
+      include Sequel::Postgres::AdapterMethods
+      self.translate_results = false if respond_to?(:translate_results=)
+      
+      # Apply connection settings for this connection.  Current sets
+      # the date style to ISO in order make Date object creation in ruby faster,
+      # if Postgres.use_iso_date_format is true.
+      def apply_connection_settings
+        super
+        if Postgres.use_iso_date_format
+          sql = "SET DateStyle = 'ISO'"
+          @db.log_info(sql)
+          execute(sql)
+        end
+      end
+
+      # Execute the given SQL with this connection.  If a block is given,
+      # yield the results, otherwise, return the number of changed rows.
+      def execute(sql, args=nil)
+        q = nil
+        begin
+          q = args ? async_exec(sql, args) : async_exec(sql)
+        rescue PGError
+          begin
+            s = status
+          rescue PGError
+            raise(Sequel::DatabaseDisconnectError)
+          end
+          status_ok = (s == Adapter::CONNECTION_OK)
+          status_ok ? raise : raise(Sequel::DatabaseDisconnectError)
+        ensure
+          block if status_ok
+        end
+        begin
+          block_given? ? yield(q) : q.cmd_tuples
+        ensure
+          q.clear
+        end
+      end
+
+      # Reapply the connection settings if the connection is reset.
+      def reset(*args, &block)
+        super(*args, &block)
+        apply_connection_settings
+      end
+      
+      if SEQUEL_POSTGRES_USES_PG
+        # Hash of prepared statements for this connection.  Keys are
+        # string names of the server side prepared statement, and values
+        # are SQL strings.
+        def prepared_statements
+          @prepared_statements ||= {}
+        end
+      end
+      
+      private
+      
+      # Return the requested values for the given row.
+      def single_value(r)
+        r.getvalue(0, 0) unless r.nil? || (r.ntuples == 0)
+      end
+    end
+    
+    # Database class for PostgreSQL databases used with Sequel and the
+    # pg, postgres, or postgres-pr driver.
+    class Database < Sequel::Database
+      include Sequel::Postgres::DatabaseMethods
+      
+      set_adapter_scheme :postgres
+      
+      # Add the primary_keys and primary_key_sequences instance variables,
+      # so we can get the correct return values for inserted rows.
+      def initialize(*args)
+        super
+        @primary_keys = {}
+        @primary_key_sequences = {}
+      end
+
+      # Connects to the database.  In addition to the standard database
+      # options, using the :encoding or :charset option changes the
+      # client encoding for the connection.
+      def connect(server)
+        opts = server_opts(server)
+        conn = Adapter.connect(
+          (opts[:host] unless blank_object?(opts[:host])),
+          opts[:port] || 5432,
+          nil, '',
+          opts[:database],
+          opts[:user],
+          opts[:password]
+        )
+        if encoding = opts[:encoding] || opts[:charset]
+          if conn.respond_to?(:set_client_encoding)
+            conn.set_client_encoding(encoding)
+          else
+            conn.async_exec("set client_encoding to '#{encoding}'")
+          end
+        end
+        conn.db = self
+        conn.apply_connection_settings
+        conn
+      end
+      
+      # Return instance of Sequel::Postgres::Dataset with the given options.
+      def dataset(opts = nil)
+        Postgres::Dataset.new(self, opts)
+      end
+      
+      # Execute the given SQL with the given args on an available connection.
+      def execute(sql, opts={}, &block)
+        return execute_prepared_statement(sql, opts, &block) if Symbol === sql
+        begin
+          log_info(sql, opts[:arguments])
+          synchronize(opts[:server]){|conn| conn.execute(sql, opts[:arguments], &block)}
+        rescue => e
+          log_info(e.message)
+          raise_error(e, :classes=>CONVERTED_EXCEPTIONS)
+        end
+      end
+      
+      # Insert the values into the table and return the primary key (if
+      # automatically generated).
+      def execute_insert(sql, opts={})
+        return execute(sql, opts) if Symbol === sql
+        begin 
+          log_info(sql, opts[:arguments])
+          synchronize(opts[:server]) do |conn|
+            conn.execute(sql, opts[:arguments])
+            insert_result(conn, opts[:table], opts[:values])
+          end
+        rescue => e
+          log_info(e.message)
+          raise_error(e, :classes=>CONVERTED_EXCEPTIONS)
+        end
+      end
+      
+      private
+
+      # PostgreSQL doesn't need the connection pool to convert exceptions.
+      def connection_pool_default_options
+        super.merge(:pool_convert_exceptions=>false)
+      end
+      
+      # Disconnect given connection
+      def disconnect_connection(conn)
+        begin
+          conn.finish
+        rescue PGError
+        end
+      end
+      
+      # Execute the prepared statement with the given name on an available
+      # connection, using the given args.  If the connection has not prepared
+      # a statement with the given name yet, prepare it.  If the connection
+      # has prepared a statement with the same name and different SQL,
+      # deallocate that statement first and then prepare this statement.
+      # If a block is given, yield the result, otherwise, return the number
+      # of rows changed.  If the :insert option is passed, return the value
+      # of the primary key for the last inserted row.
+      def execute_prepared_statement(name, opts={})
+        ps = prepared_statements[name]
+        sql = ps.prepared_sql
+        ps_name = name.to_s
+        args = opts[:arguments]
+        synchronize(opts[:server]) do |conn|
+          unless conn.prepared_statements[ps_name] == sql
+            if conn.prepared_statements.include?(ps_name)
+              s = "DEALLOCATE #{ps_name}"
+              log_info(s)
+              conn.execute(s) unless conn.prepared_statements[ps_name] == sql
+            end
+            conn.prepared_statements[ps_name] = sql
+            log_info("PREPARE #{ps_name} AS #{sql}")
+            conn.prepare(ps_name, sql)
+          end
+          log_info("EXECUTE #{ps_name}", args)
+          q = conn.exec_prepared(ps_name, args)
+          if opts[:table] && opts[:values]
+            insert_result(conn, opts[:table], opts[:values])
+          else
+            begin
+              block_given? ? yield(q) : q.cmd_tuples
+            ensure
+              q.clear
+            end
+          end
+        end
+      end
+    end
+    
+    # Dataset class for PostgreSQL datasets that use the pg, postgres, or
+    # postgres-pr driver.
+    class Dataset < Sequel::Dataset
+      include Sequel::Postgres::DatasetMethods
+      
+      # Yield all rows returned by executing the given SQL and converting
+      # the types.
+      def fetch_rows(sql)
+        cols = []
+        execute(sql) do |res|
+          res.nfields.times do |fieldnum|
+            cols << [fieldnum, PG_TYPES[res.ftype(fieldnum)], output_identifier(res.fname(fieldnum))]
+          end
+          @columns = cols.map{|c| c.at(2)}
+          res.ntuples.times do |recnum|
+            converted_rec = {}
+            cols.each do |fieldnum, type_proc, fieldsym|
+              value = res.getvalue(recnum, fieldnum)
+              converted_rec[fieldsym] = (value && type_proc) ? type_proc.call(value) : value
+            end
+            yield converted_rec
+          end
+        end
+      end
+
+      if SEQUEL_POSTGRES_USES_PG
+        
+        PREPARED_ARG_PLACEHOLDER = LiteralString.new('$').freeze
+        
+        # PostgreSQL specific argument mapper used for mapping the named
+        # argument hash to a array with numbered arguments.  Only used with
+        # the pg driver.
+        module ArgumentMapper
+          include Sequel::Dataset::ArgumentMapper
+          
+          protected
+          
+          # Return an array of strings for each of the hash values, inserting
+          # them to the correct position in the array.
+          def map_to_prepared_args(hash)
+            @prepared_args.map{|k| hash[k.to_sym].to_s}
+          end
+
+          private
+          
+          # PostgreSQL most of the time requires type information for each of
+          # arguments to a prepared statement.  Handle this by allowing the
+          # named argument to have a __* suffix, with the * being the type.
+          # In the generated SQL, cast the bound argument to that type to
+          # elminate ambiguity (and PostgreSQL from raising an exception).
+          def prepared_arg(k)
+            y, type = k.to_s.split("__")
+            if i = @prepared_args.index(y)
+              i += 1
+            else
+              @prepared_args << y
+              i = @prepared_args.length
+            end
+            LiteralString.new("#{prepared_arg_placeholder}#{i}#{"::#{type}" if type}")
+          end
+        end
+
+        # Allow use of bind arguments for PostgreSQL using the pg driver.
+        module BindArgumentMethods
+          include ArgumentMapper
+          
+          private
+          
+          # Execute the given SQL with the stored bind arguments.
+          def execute(sql, opts={}, &block)
+            super(sql, {:arguments=>bind_arguments}.merge(opts), &block)
+          end
+          
+          # Same as execute, explicit due to intricacies of alias and super.
+          def execute_dui(sql, opts={}, &block)
+            super(sql, {:arguments=>bind_arguments}.merge(opts), &block)
+          end
+          
+          # Same as execute, explicit due to intricacies of alias and super.
+          def execute_insert(sql, opts={}, &block)
+            super(sql, {:arguments=>bind_arguments}.merge(opts), &block)
+          end
+        end
+        
+        # Allow use of server side prepared statements for PostgreSQL using the
+        # pg driver.
+        module PreparedStatementMethods
+          include BindArgumentMethods
+          include ::Sequel::Postgres::DatasetMethods::PreparedStatementMethods
+          
+          private
+          
+          # Execute the stored prepared statement name and the stored bind
+          # arguments instead of the SQL given.
+          def execute(sql, opts={}, &block)
+            super(prepared_statement_name, opts, &block)
+          end
+          
+          # Same as execute, explicit due to intricacies of alias and super.
+          def execute_dui(sql, opts={}, &block)
+            super(prepared_statement_name, opts, &block)
+          end
+          
+          # Same as execute, explicit due to intricacies of alias and super.
+          def execute_insert(sql, opts={}, &block)
+            super(prepared_statement_name, opts, &block)
+          end
+        end
+        
+        # Execute the given type of statement with the hash of values.
+        def call(type, hash, values=nil, &block)
+          ps = to_prepared_statement(type, values)
+          ps.extend(BindArgumentMethods)
+          ps.call(hash, &block)
+        end
+        
+        # Prepare the given type of statement with the given name, and store
+        # it in the database to be called later.
+        def prepare(type, name=nil, values=nil)
+          ps = to_prepared_statement(type, values)
+          ps.extend(PreparedStatementMethods)
+          if name
+            ps.prepared_statement_name = name
+            db.prepared_statements[name] = ps
+          end
+          ps
+        end
+        
+        private
+        
+        # PostgreSQL uses $N for placeholders instead of ?, so use a $
+        # as the placeholder.
+        def prepared_arg_placeholder
+          PREPARED_ARG_PLACEHOLDER
+        end
+      end
+      
+      private
+
+      def literal_blob(v)
+        db.synchronize{|c| "'#{c.escape_bytea(v)}'"}
+      end
+
+      def literal_string(v)
+        db.synchronize{|c| "'#{c.escape_string(v)}'"}
+      end
+
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/adapters/shared/mssql.rb 3.0.0-1/lib/sequel/adapters/shared/mssql.rb
--- 1.5.1-1/lib/sequel/adapters/shared/mssql.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/adapters/shared/mssql.rb	2009-04-28 00:19:06.000000000 +0100
@@ -0,0 +1,104 @@
+Sequel.require 'adapters/utils/unsupported'
+
+module Sequel
+  module MSSQL
+    module DatabaseMethods
+      AUTO_INCREMENT = 'IDENTITY(1,1)'.freeze
+      SQL_BEGIN = "BEGIN TRANSACTION".freeze
+      SQL_COMMIT = "COMMIT TRANSACTION".freeze
+      SQL_ROLLBACK = "ROLLBACK TRANSACTION".freeze
+      TEMPORARY = "#".freeze
+      
+      # Microsoft SQL Server uses the :mssql type.
+      def database_type
+        :mssql
+      end
+
+      def dataset(opts = nil)
+        ds = super
+        ds.extend(DatasetMethods)
+        ds
+      end
+
+      private
+
+      def auto_increment_sql
+        AUTO_INCREMENT
+      end
+
+      # SQL to BEGIN a transaction.
+      def begin_transaction_sql
+        SQL_BEGIN
+      end
+
+      # SQL to COMMIT a transaction.
+      def commit_transaction_sql
+        SQL_COMMIT
+      end
+      
+      # SQL to ROLLBACK a transaction.
+      def rollback_transaction_sql
+        SQL_ROLLBACK
+      end
+
+      # SQL fragment for marking a table as temporary
+      def temporary_table_sql
+        TEMPORARY
+      end
+    end
+  
+    module DatasetMethods
+      include Dataset::UnsupportedIntersectExcept
+
+      SELECT_CLAUSE_ORDER = %w'limit distinct columns from with join where group order having compounds'.freeze
+
+      def complex_expression_sql(op, args)
+        case op
+        when :'||'
+          super(:+, args)
+        else
+          super(op, args)
+        end
+      end
+      
+      def full_text_search(cols, terms, opts = {})
+        filter("CONTAINS (#{literal(cols)}, #{literal(terms)})")
+      end
+      
+      def multi_insert_sql(columns, values)
+        values = values.map {|r| "SELECT #{expression_list(r)}" }.join(" UNION ALL ")
+        ["INSERT INTO #{source_list(@opts[:from])} (#{identifier_list(columns)}) #{values}"]
+      end
+
+      # Allows you to do .nolock on a query
+      def nolock
+        clone(:with => "(NOLOCK)")
+      end
+
+      def quoted_identifier(name)
+        "[#{name}]"
+      end
+
+      private
+
+      def literal_string(v)
+        "N#{super}"
+      end
+
+      def select_clause_order
+        SELECT_CLAUSE_ORDER
+      end
+
+      # MSSQL uses TOP for limit, with no offset support
+      def select_limit_sql(sql)
+        raise(Error, "OFFSET not supported") if @opts[:offset]
+        sql << " TOP #{@opts[:limit]}" if @opts[:limit]
+      end
+
+      # MSSQL uses the WITH statement to lock tables
+      def select_with_sql(sql)
+        sql << " WITH #{@opts[:with]}" if @opts[:with]
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/adapters/shared/mysql.rb 3.0.0-1/lib/sequel/adapters/shared/mysql.rb
--- 1.5.1-1/lib/sequel/adapters/shared/mysql.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/adapters/shared/mysql.rb	2009-04-28 00:19:45.000000000 +0100
@@ -0,0 +1,390 @@
+Sequel.require 'adapters/utils/unsupported'
+
+module Sequel
+  class Database
+    # Keep default column_references_sql for add_foreign_key support
+    alias default_column_references_sql column_references_sql
+  end
+  module MySQL
+    class << self
+      # Set the default options used for CREATE TABLE
+      attr_accessor :default_charset, :default_collate, :default_engine
+    end
+
+    # Methods shared by Database instances that connect to MySQL,
+    # currently supported by the native and JDBC adapters.
+    module DatabaseMethods
+      AUTO_INCREMENT = 'AUTO_INCREMENT'.freeze
+      CAST_TYPES = {String=>:CHAR, Integer=>:SIGNED, Time=>:DATETIME, DateTime=>:DATETIME, Numeric=>:DECIMAL, BigDecimal=>:DECIMAL, File=>:BINARY}
+      PRIMARY = 'PRIMARY'.freeze
+      
+      # MySQL's cast rules are restrictive in that you can't just cast to any possible
+      # database type.
+      def cast_type_literal(type)
+        CAST_TYPES[type] || super
+      end
+
+      # MySQL uses the :mysql database type
+      def database_type
+        :mysql
+      end
+
+      # Return a hash containing index information. Hash keys are index name symbols.
+      # Values are subhashes with two keys, :columns and :unique.  The value of :columns
+      # is an array of symbols of column names.  The value of :unique is true or false
+      # depending on if the index is unique.
+      def indexes(table)
+        indexes = {}
+        m = output_identifier_meth
+        im = input_identifier_meth
+        metadata_dataset.with_sql("SHOW INDEX FROM ?", SQL::Identifier.new(im.call(table))).each do |r|
+          name = r[:Key_name]
+          next if name == PRIMARY
+          i = indexes[m.call(name)] ||= {:columns=>[], :unique=>r[:Non_unique] != 1}
+          i[:columns] << m.call(r[:Column_name])
+        end
+        indexes
+      end
+
+      # Get version of MySQL server, used for determined capabilities.
+      def server_version
+        m = /(\d+)\.(\d+)\.(\d+)/.match(get(SQL::Function.new(:version)))
+        @server_version ||= (m[1].to_i * 10000) + (m[2].to_i * 100) + m[3].to_i
+      end
+      
+      # Return an array of symbols specifying table names in the current database.
+      #
+      # Options:
+      # * :server - Set the server to use
+      def tables(opts={})
+        m = output_identifier_meth
+        metadata_dataset.with_sql('SHOW TABLES').server(opts[:server]).map{|r| m.call(r.values.first)}
+      end
+      
+      # Changes the database in use by issuing a USE statement.  I would be
+      # very careful if I used this.
+      def use(db_name)
+        disconnect
+        @opts[:database] = db_name if self << "USE #{db_name}"
+        @schemas = nil
+        self
+      end
+      
+      private
+      
+      # Use MySQL specific syntax for rename column, set column type, and
+      # drop index cases.
+      def alter_table_sql(table, op)
+        case op[:op]
+        when :add_column
+          if related = op.delete(:table)
+            sql = super(table, op)
+            op[:table] = related
+            [sql, "ALTER TABLE #{quote_schema_table(table)} ADD FOREIGN KEY (#{quote_identifier(op[:name])})#{default_column_references_sql(op)}"]
+          else
+            super(table, op)
+          end
+        when :rename_column, :set_column_type, :set_column_null, :set_column_default
+          o = op[:op]
+          opts = schema(table).find{|x| x.first == op[:name]}
+          old_opts = opts ? opts.last : {}
+          name = o == :rename_column ? op[:new_name] : op[:name]
+          type = o == :set_column_type ? op[:type] : old_opts[:db_type]
+          null = o == :set_column_null ? op[:null] : old_opts[:allow_null]
+          default = o == :set_column_default ? op[:default] : (old_opts[:default].lit if old_opts[:default])
+          "ALTER TABLE #{quote_schema_table(table)} CHANGE COLUMN #{quote_identifier(op[:name])} #{column_definition_sql(:name=>name, :type=>type, :null=>null, :default=>default)}"
+        when :drop_index
+          "#{drop_index_sql(table, op)} ON #{quote_schema_table(table)}"
+        else
+          super(table, op)
+        end
+      end
+      
+      # Use MySQL specific AUTO_INCREMENT text.
+      def auto_increment_sql
+        AUTO_INCREMENT
+      end
+      
+      # Handle MySQL specific syntax for column references
+      def column_references_sql(column)
+        "#{", FOREIGN KEY (#{quote_identifier(column[:name])})" unless column[:type] == :check}#{super(column)}"
+      end
+      
+      # Use MySQL specific syntax for engine type and character encoding
+      def create_table_sql(name, generator, options = {})
+        engine = options.include?(:engine) ? options[:engine] : Sequel::MySQL.default_engine
+        charset = options.include?(:charset) ? options[:charset] : Sequel::MySQL.default_charset
+        collate = options.include?(:collate) ? options[:collate] : Sequel::MySQL.default_collate
+        "#{super}#{" ENGINE=#{engine}" if engine}#{" DEFAULT CHARSET=#{charset}" if charset}#{" DEFAULT COLLATE=#{collate}" if collate}"
+      end
+
+      # MySQL folds unquoted identifiers to lowercase, so it shouldn't need to upcase identifiers on input.
+      def identifier_input_method_default
+        nil
+      end
+      
+      # MySQL folds unquoted identifiers to lowercase, so it shouldn't need to upcase identifiers on output.
+      def identifier_output_method_default
+        nil
+      end
+
+      # Handle MySQL specific index SQL syntax
+      def index_definition_sql(table_name, index)
+        index_name = quote_identifier(index[:name] || default_index_name(table_name, index[:columns]))
+        index_type = case index[:type]
+        when :full_text
+          "FULLTEXT "
+        when :spatial
+          "SPATIAL "
+        else
+          using = " USING #{index[:type]}" unless index[:type] == nil
+          "UNIQUE " if index[:unique]
+        end
+        "CREATE #{index_type}INDEX #{index_name}#{using} ON #{quote_schema_table(table_name)} #{literal(index[:columns])}"
+      end
+      
+      # MySQL treats integer primary keys as autoincrementing.
+      def schema_autoincrementing_primary_key?(schema)
+        super and schema[:db_type] =~ /int/io
+      end
+
+      # Use the MySQL specific DESCRIBE syntax to get a table description.
+      def schema_parse_table(table_name, opts)
+        m = output_identifier_meth
+        im = input_identifier_meth
+        metadata_dataset.with_sql("DESCRIBE ?", SQL::Identifier.new(im.call(table_name))).map do |row|
+          row.delete(:Extra)
+          row[:allow_null] = row.delete(:Null) == 'YES'
+          row[:default] = row.delete(:Default)
+          row[:primary_key] = row.delete(:Key) == 'PRI'
+          row[:default] = nil if blank_object?(row[:default])
+          row[:db_type] = row.delete(:Type)
+          row[:type] = schema_column_type(row[:db_type])
+          [m.call(row.delete(:Field)), row]
+        end
+      end
+
+      # MySQL has both datetime and timestamp classes, most people are going
+      # to want datetime
+      def type_literal_generic_datetime(column)
+        :datetime
+      end
+
+      # MySQL has both datetime and timestamp classes, most people are going
+      # to want datetime
+      def type_literal_generic_time(column)
+        column[:only_time] ? :time : :datetime
+      end
+
+      # MySQL doesn't have a true boolean class, so it uses tinyint
+      # MySQL doesn't have a true boolean class, so it uses tinyint
+      def type_literal_generic_trueclass(column)
+        :tinyint
+      end
+    end
+  
+    # Dataset methods shared by datasets that use MySQL databases.
+    module DatasetMethods
+      include Dataset::UnsupportedIntersectExcept
+
+      BOOL_TRUE = '1'.freeze
+      BOOL_FALSE = '0'.freeze
+      TIMESTAMP_FORMAT = "'%Y-%m-%d %H:%M:%S'".freeze
+      COMMA_SEPARATOR = ', '.freeze
+      
+      # MySQL specific syntax for LIKE/REGEXP searches, as well as
+      # string concatenation.
+      def complex_expression_sql(op, args)
+        case op
+        when :~, :'!~', :'~*', :'!~*', :LIKE, :'NOT LIKE', :ILIKE, :'NOT ILIKE'
+          "(#{literal(args.at(0))} #{'NOT ' if [:'NOT LIKE', :'NOT ILIKE', :'!~', :'!~*'].include?(op)}#{[:~, :'!~', :'~*', :'!~*'].include?(op) ? 'REGEXP' : 'LIKE'} #{'BINARY ' if [:~, :'!~', :LIKE, :'NOT LIKE'].include?(op)}#{literal(args.at(1))})"
+        when :'||'
+          if args.length > 1
+            "CONCAT(#{args.collect{|a| literal(a)}.join(', ')})"
+          else
+            literal(args.at(0))
+          end
+        else
+          super(op, args)
+        end
+      end
+      
+      # MySQL supports ORDER and LIMIT clauses in DELETE statements.
+      def delete_sql
+        sql = super
+        sql << " ORDER BY #{expression_list(opts[:order])}" if opts[:order]
+        sql << " LIMIT #{opts[:limit]}" if opts[:limit]
+        sql
+      end
+
+      # MySQL doesn't support DISTINCT ON
+      def distinct(*columns)
+        raise(Error, "DISTINCT ON not supported by MySQL") unless columns.empty?
+        super
+      end
+
+      # Adds full text filter
+      def full_text_search(cols, terms, opts = {})
+        filter(full_text_sql(cols, terms, opts))
+      end
+      
+      # MySQL specific full text search syntax.
+      def full_text_sql(cols, term, opts = {})
+        "MATCH #{literal(Array(cols))} AGAINST (#{literal(Array(term).join(' '))}#{" IN BOOLEAN MODE" if opts[:boolean]})"
+      end
+
+      # MySQL allows HAVING clause on ungrouped datasets.
+      def having(*cond, &block)
+        _filter(:having, *cond, &block)
+      end
+      
+      # MySQL doesn't use the SQL standard DEFAULT VALUES.
+      def insert_default_values_sql
+        "INSERT INTO #{source_list(@opts[:from])} () VALUES ()"
+      end
+
+      # Transforms an CROSS JOIN to an INNER JOIN if the expr is not nil.
+      # Raises an error on use of :full_outer type, since MySQL doesn't support it.
+      def join_table(type, table, expr=nil, table_alias={})
+        type = :inner if (type == :cross) && !expr.nil?
+        raise(Sequel::Error, "MySQL doesn't support FULL OUTER JOIN") if type == :full_outer
+        super(type, table, expr, table_alias)
+      end
+      
+      # Transforms :natural_inner to NATURAL LEFT JOIN and straight to
+      # STRAIGHT_JOIN.
+      def join_type_sql(join_type)
+        case join_type
+        when :straight then 'STRAIGHT_JOIN'
+        when :natural_inner then 'NATURAL LEFT JOIN'
+        else super
+        end
+      end
+      
+      # Sets up multi_insert or import to use INSERT IGNORE.
+      # Useful if you have a unique key and want to just skip
+      # inserting rows that violate the unique key restriction.
+      #
+      # Example:
+      #
+      # dataset.insert_ignore.multi_insert(
+      #  [{:name => 'a', :value => 1}, {:name => 'b', :value => 2}]
+      # )
+      #
+      # INSERT IGNORE INTO tablename (name, value) VALUES (a, 1), (b, 2)
+      #
+      def insert_ignore
+        clone(:insert_ignore=>true)
+      end
+      
+      # Sets up multi_insert or import to use ON DUPLICATE KEY UPDATE
+      # If you pass no arguments, ALL fields will be
+      # updated with the new values.  If you pass the fields you
+      # want then ONLY those field will be updated.
+      #
+      # Useful if you have a unique key and want to update
+      # inserting rows that violate the unique key restriction.
+      #
+      # Examples:
+      #
+      # dataset.on_duplicate_key_update.multi_insert(
+      #  [{:name => 'a', :value => 1}, {:name => 'b', :value => 2}]
+      # )
+      #
+      # INSERT INTO tablename (name, value) VALUES (a, 1), (b, 2)
+      # ON DUPLICATE KEY UPDATE name=VALUES(name), value=VALUES(value)
+      #
+      # dataset.on_duplicate_key_update(:value).multi_insert(
+      #  [{:name => 'a', :value => 1}, {:name => 'b', :value => 2}]
+      # )
+      #
+      # INSERT INTO tablename (name, value) VALUES (a, 1), (b, 2)
+      # ON DUPLICATE KEY UPDATE value=VALUES(value)
+      #
+      def on_duplicate_key_update(*args)
+        clone(:on_duplicate_key_update => args)
+      end
+      
+      # MySQL specific syntax for inserting multiple values at once.
+      def multi_insert_sql(columns, values)
+        if update_cols = opts[:on_duplicate_key_update]
+          update_cols = columns if update_cols.empty?
+          update_string = update_cols.map{|c| "#{quote_identifier(c)}=VALUES(#{quote_identifier(c)})"}.join(COMMA_SEPARATOR)
+        end
+        values = values.map {|r| literal(Array(r))}.join(COMMA_SEPARATOR)
+        ["INSERT#{' IGNORE' if opts[:insert_ignore]} INTO #{source_list(@opts[:from])} (#{identifier_list(columns)}) VALUES #{values}#{" ON DUPLICATE KEY UPDATE #{update_string}" if update_string}"]
+      end
+      
+      # MySQL uses the nonstandard ` (backtick) for quoting identifiers.
+      def quoted_identifier(c)
+        "`#{c}`"
+      end
+      
+      # MySQL specific syntax for REPLACE (aka UPSERT, or update if exists,
+      # insert if it doesn't).
+      def replace_sql(*values)
+        from = source_list(@opts[:from])
+        if values.empty?
+          "REPLACE INTO #{from} DEFAULT VALUES"
+        else
+          values = values[0] if values.size == 1
+          
+          case values
+          when Array
+            if values.empty?
+              "REPLACE INTO #{from} DEFAULT VALUES"
+            else
+              "REPLACE INTO #{from} VALUES #{literal(values)}"
+            end
+          when Hash
+            if values.empty?
+              "REPLACE INTO #{from} DEFAULT VALUES"
+            else
+              fl, vl = [], []
+              values.each {|k, v| fl << literal(k.is_a?(String) ? k.to_sym : k); vl << literal(v)}
+              "REPLACE INTO #{from} (#{fl.join(COMMA_SEPARATOR)}) VALUES (#{vl.join(COMMA_SEPARATOR)})"
+            end
+          when Dataset
+            "REPLACE INTO #{from} #{literal(values)}"
+          else
+            if values.respond_to?(:values)
+              replace_sql(values.values)
+            else  
+              "REPLACE INTO #{from} VALUES (#{literal(values)})"
+            end
+          end
+        end
+      end
+      
+      # MySQL supports ORDER and LIMIT clauses in UPDATE statements.
+      def update_sql(values)
+        sql = super
+        sql << " ORDER BY #{expression_list(opts[:order])}" if opts[:order]
+        sql << " LIMIT #{opts[:limit]}" if opts[:limit]
+        sql
+      end
+
+      private
+
+      # Use MySQL Timestamp format
+      def literal_datetime(v)
+        v.strftime(TIMESTAMP_FORMAT)
+      end
+
+      # Use 0 for false on MySQL
+      def literal_false
+        BOOL_FALSE
+      end
+
+      # Use MySQL Timestamp format
+      def literal_time(v)
+        v.strftime(TIMESTAMP_FORMAT)
+      end
+
+      # Use 1 for true on MySQL
+      def literal_true
+        BOOL_TRUE
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/adapters/shared/oracle.rb 3.0.0-1/lib/sequel/adapters/shared/oracle.rb
--- 1.5.1-1/lib/sequel/adapters/shared/oracle.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/adapters/shared/oracle.rb	2009-04-30 16:36:00.000000000 +0100
@@ -0,0 +1,151 @@
+Sequel.require %w'date_format unsupported', 'adapters/utils'
+
+module Sequel
+  module Oracle
+    module DatabaseMethods
+      TEMPORARY = 'GLOBAL TEMPORARY '.freeze
+      AUTOINCREMENT = ''.freeze
+
+      def create_sequence(name, opts={})
+        self << create_sequence_sql(name, opts)
+      end
+
+      def create_table(name, options={}, &block)
+        options = {:generator=>options} if options.is_a?(Schema::Generator)
+        generator = options[:generator] || Schema::Generator.new(self, &block)
+        drop_statement, create_statements = create_table_sql_list(name, generator, options)
+        (execute_ddl(drop_statement) rescue nil) if drop_statement
+        (create_statements + index_sql_list(name, generator.indexes)).each{|sql| execute_ddl(sql)}
+      end
+
+      def create_trigger(*args)
+        self << create_trigger_sql(*args)
+      end
+
+      def drop_sequence(name)
+        self << drop_sequence_sql(name)
+      end
+
+      # Oracle uses the :oracle database type
+      def database_type
+        :oracle
+      end
+
+      def tables(opts={})
+        ds = from(:tab).server(opts[:server]).select(:tname).filter(:tabtype => 'TABLE')
+        ds.map{|r| ds.send(:output_identifier, r[:tname])}
+      end
+
+      def table_exists?(name)
+        from(:tab).filter(:tname =>dataset.send(:input_identifier, name), :tabtype => 'TABLE').count > 0
+      end
+
+      private
+
+      def auto_increment_sql
+        AUTOINCREMENT
+      end
+
+      # SQL fragment for showing a table is temporary
+      def temporary_table_sql
+        TEMPORARY
+      end
+
+      def create_sequence_sql(name, opts={})
+        "CREATE SEQUENCE #{quote_identifier(name)} start with #{opts [:start_with]||1} increment by #{opts[:increment_by]||1} nomaxvalue"
+      end
+
+      def create_table_sql_list(name, generator, options={})
+        statements = [create_table_sql(name, generator, options)]
+        drop_seq_statement = nil
+        generator.columns.each do |c|
+          if c[:auto_increment]
+            c[:sequence_name] ||= "seq_#{name}_#{c[:name]}"
+            unless c[:create_sequence] == false
+              drop_seq_statement = drop_sequence_sql(c[:sequence_name])
+              statements << create_sequence_sql(c[:sequence_name], c)
+            end
+            unless c[:create_trigger] == false
+              c[:trigger_name] ||= "BI_#{name}_#{c[:name]}"
+              trigger_definition = <<-end_sql
+              BEGIN
+                IF :NEW.#{quote_identifier(c[:name])} IS NULL THEN
+                  SELECT #{c[:sequence_name]}.nextval INTO :NEW.#{quote_identifier(c[:name])} FROM dual;
+                END IF;
+              END;
+              end_sql
+              statements << create_trigger_sql(name, c[:trigger_name], trigger_definition, {:events => [:insert]})
+            end
+          end
+        end
+        [drop_seq_statement, statements]
+      end
+
+      def create_trigger_sql(table, name, definition, opts={})
+        events = opts[:events] ? Array(opts[:events]) : [:insert, :update, :delete]
+        sql = <<-end_sql
+          CREATE#{' OR REPLACE' if opts[:replace]} TRIGGER #{quote_identifier(name)}
+          #{opts[:after] ? 'AFTER' : 'BEFORE'} #{events.map{|e| e.to_s.upcase}.join(' OR ')} ON #{quote_schema_table(table)}
+          REFERENCING NEW AS NEW FOR EACH ROW
+          #{definition}
+        end_sql
+        sql
+      end
+
+      def drop_sequence_sql(name)
+        "DROP SEQUENCE #{quote_identifier(name)}"
+      end
+    end
+    
+    module DatasetMethods
+      include Dataset::UnsupportedIntersectExceptAll
+      include Dataset::SQLStandardDateFormat
+
+      SELECT_CLAUSE_ORDER = %w'distinct columns from join where group having compounds order limit'.freeze
+
+      # Oracle doesn't support DISTINCT ON
+      def distinct(*columns)
+        raise(Error, "DISTINCT ON not supported by Oracle") unless columns.empty?
+        super
+      end
+
+      # Oracle uses MINUS instead of EXCEPT, and doesn't support EXCEPT ALL
+      def except(dataset, all = false)
+        raise(Sequel::Error, "EXCEPT ALL not supported") if all
+        compound_clone(:minus, dataset, all)
+      end
+
+      def empty?
+        db[:dual].where(exists).get(1) == nil
+      end
+
+      private
+
+      # Oracle doesn't support the use of AS when aliasing a dataset.  It doesn't require
+      # the use of AS anywhere, so this disables it in all cases.
+      def as_sql(expression, aliaz)
+        "#{expression} #{quote_identifier(aliaz)}"
+      end
+
+      # Oracle uses the SQL standard of only doubling ' inside strings.
+      def literal_string(v)
+        "'#{v.gsub("'", "''")}'"
+      end
+
+      def select_clause_order
+        SELECT_CLAUSE_ORDER
+      end
+
+      # Oracle requires a subselect to do limit and offset
+      def select_limit_sql(sql)
+        if limit = @opts[:limit]
+          if (offset = @opts[:offset]) && (offset > 0)
+            sql.replace("SELECT * FROM (SELECT raw_sql_.*, ROWNUM raw_rnum_ FROM(#{sql}) raw_sql_ WHERE ROWNUM <= #{limit + offset}) WHERE raw_rnum_ > #{offset}")
+          else
+            sql.replace("SELECT * FROM (#{sql}) WHERE ROWNUM <= #{limit}")
+          end
+        end
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/adapters/shared/postgres.rb 3.0.0-1/lib/sequel/adapters/shared/postgres.rb
--- 1.5.1-1/lib/sequel/adapters/shared/postgres.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/adapters/shared/postgres.rb	2009-04-28 00:20:43.000000000 +0100
@@ -0,0 +1,807 @@
+module Sequel
+  # Top level module for holding all PostgreSQL-related modules and classes
+  # for Sequel.  There are a few module level accessors that are added via
+  # metaprogramming.  These are:
+  # * client_min_messages (only available when using the native adapter) -
+  #   Change the minimum level of messages that PostgreSQL will send to the
+  #   the client.  The PostgreSQL default is NOTICE, the Sequel default is
+  #   WARNING.  Set to nil to not change the server default.
+  # * force_standard_strings - Set to false to not force the use of
+  #   standard strings
+  # * use_iso_date_format (only available when using the native adapter) -
+  #   Set to false to not change the date format to
+  #   ISO.  This disables one of Sequel's optimizations.
+  #
+  # Changes in these settings only affect future connections.  To make
+  # sure that they are applied, they should generally be called right
+  # after the Database object is instantiated and before a connection
+  # is actually made. For example, to use whatever the server defaults are:
+  #
+  #   DB = Sequel.postgres(...)
+  #   Sequel::Postgres.client_min_messages = nil
+  #   Sequel::Postgres.force_standard_strings = false
+  #   Sequel::Postgres.use_iso_date_format = false
+  #   # A connection to the server is not made until here
+  #   DB[:t].all
+  #
+  # The reason they can't be done earlier is that the Sequel::Postgres
+  # module is not loaded until a Database object which uses PostgreSQL
+  # is created.
+  module Postgres
+    # Array of exceptions that need to be converted.  JDBC
+    # uses NativeExceptions, the native adapter uses PGError.
+    CONVERTED_EXCEPTIONS = []
+    
+    @client_min_messages = :warning
+    @force_standard_strings = true
+    
+    class << self
+      # By default, Sequel sets the minimum level of log messages sent to the client
+      # to WARNING, where PostgreSQL uses a default of NOTICE.  This is to avoid a lot
+      # of mostly useless messages when running migrations, such as a couple of lines
+      # for every serial primary key field.
+      attr_accessor :client_min_messages
+
+      # By default, Sequel forces the use of standard strings, so that
+      # '\\' is interpreted as \\ and not \.  While PostgreSQL defaults
+      # to interpreting plain strings as extended strings, this will change
+      # in a future version of PostgreSQL.  Sequel assumes that SQL standard
+      # strings will be used.
+      attr_accessor :force_standard_strings
+    end
+
+    # Methods shared by adapter/connection instances.
+    module AdapterMethods
+      attr_writer :db
+      
+      SELECT_CURRVAL = "SELECT currval('%s')".freeze
+      SELECT_CUSTOM_SEQUENCE = proc do |schema, table| <<-end_sql
+        SELECT '"' || name.nspname || '".' || CASE  
+            WHEN split_part(def.adsrc, '''', 2) ~ '.' THEN  
+              substr(split_part(def.adsrc, '''', 2),  
+                     strpos(split_part(def.adsrc, '''', 2), '.')+1) 
+            ELSE split_part(def.adsrc, '''', 2)  
+          END
+        FROM pg_class t
+        JOIN pg_namespace  name ON (t.relnamespace = name.oid)
+        JOIN pg_attribute  attr ON (t.oid = attrelid)
+        JOIN pg_attrdef    def  ON (adrelid = attrelid AND adnum = attnum)
+        JOIN pg_constraint cons ON (conrelid = adrelid AND adnum = conkey[1])
+        WHERE cons.contype = 'p'
+          AND def.adsrc ~* 'nextval'
+          #{"AND name.nspname = '#{schema}'" if schema}
+          AND t.relname = '#{table}'
+      end_sql
+      end
+      SELECT_PK = proc do |schema, table| <<-end_sql
+        SELECT pg_attribute.attname
+        FROM pg_class, pg_attribute, pg_index, pg_namespace
+        WHERE pg_class.oid = pg_attribute.attrelid
+          AND pg_class.relnamespace  = pg_namespace.oid
+          AND pg_class.oid = pg_index.indrelid
+          AND pg_index.indkey[0] = pg_attribute.attnum
+          AND pg_index.indisprimary = 't'
+          #{"AND pg_namespace.nspname = '#{schema}'" if schema}
+          AND pg_class.relname = '#{table}'
+      end_sql
+      end
+      SELECT_SERIAL_SEQUENCE = proc do |schema, table| <<-end_sql
+        SELECT  '"' || name.nspname || '".' || seq.relname || ''
+        FROM pg_class seq, pg_attribute attr, pg_depend dep,
+          pg_namespace name, pg_constraint cons
+        WHERE seq.oid = dep.objid
+          AND seq.relnamespace  = name.oid
+          AND seq.relkind = 'S'
+          AND attr.attrelid = dep.refobjid
+          AND attr.attnum = dep.refobjsubid
+          AND attr.attrelid = cons.conrelid
+          AND attr.attnum = cons.conkey[1]
+          AND cons.contype = 'p'
+          #{"AND name.nspname = '#{schema}'" if schema}
+          AND seq.relname = '#{table}'
+      end_sql
+      end
+      
+      # Depth of the current transaction on this connection, used
+      # to implement multi-level transactions with savepoints.
+      attr_accessor :transaction_depth
+      
+      # Apply connection settings for this connection. Currently, turns
+      # standard_conforming_strings ON if Postgres.force_standard_strings
+      # is true.
+      def apply_connection_settings
+        if Postgres.force_standard_strings
+          sql = "SET standard_conforming_strings = ON"
+          @db.log_info(sql)
+          # This setting will only work on PostgreSQL 8.2 or greater
+          # and we don't know the server version at this point, so
+          # try it unconditionally and rescue any errors.
+          execute(sql) rescue nil
+        end
+        if cmm = Postgres.client_min_messages
+          sql = "SET client_min_messages = '#{cmm.to_s.upcase}'"
+          @db.log_info(sql)
+          execute(sql)
+        end
+      end
+
+      # Get the last inserted value for the given sequence.
+      def last_insert_id(sequence)
+        sql = SELECT_CURRVAL % sequence
+        @db.log_info(sql)
+        execute(sql) do |r|
+          val = single_value(r)
+          return val.to_i if val
+        end
+      end
+      
+      # Get the primary key for the given table.
+      def primary_key(schema, table)
+        sql = SELECT_PK[schema, table]
+        @db.log_info(sql)
+        execute(sql) do |r|
+          return single_value(r)
+        end
+      end
+      
+      # Get the primary key and sequence for the given table.
+      def sequence(schema, table)
+        sql = SELECT_SERIAL_SEQUENCE[schema, table]
+        @db.log_info(sql)
+        execute(sql) do |r|
+          seq = single_value(r)
+          return seq if seq
+        end
+        
+        sql = SELECT_CUSTOM_SEQUENCE[schema, table]
+        @db.log_info(sql)
+        execute(sql) do |r|
+          return single_value(r)
+        end
+      end
+    end
+    
+    # Methods shared by Database instances that connect to PostgreSQL.
+    module DatabaseMethods
+      PREPARED_ARG_PLACEHOLDER = LiteralString.new('$').freeze
+      RE_CURRVAL_ERROR = /currval of sequence "(.*)" is not yet defined in this session|relation "(.*)" does not exist/.freeze
+      SQL_BEGIN = 'BEGIN'.freeze
+      SQL_SAVEPOINT = 'SAVEPOINT autopoint_%d'.freeze
+      SQL_COMMIT = 'COMMIT'.freeze
+      SQL_ROLLBACK_TO_SAVEPOINT = 'ROLLBACK TO SAVEPOINT autopoint_%d'.freeze
+      SQL_ROLLBACK = 'ROLLBACK'.freeze
+      SQL_RELEASE_SAVEPOINT = 'RELEASE SAVEPOINT autopoint_%d'.freeze
+      SYSTEM_TABLE_REGEXP = /^pg|sql/.freeze
+
+      # Creates the function in the database.  Arguments:
+      # * name : name of the function to create
+      # * definition : string definition of the function, or object file for a dynamically loaded C function.
+      # * opts : options hash:
+      #   * :args : function arguments, can be either a symbol or string specifying a type or an array of 1-3 elements:
+      #     * element 1 : argument data type
+      #     * element 2 : argument name
+      #     * element 3 : argument mode (e.g. in, out, inout)
+      #   * :behavior : Should be IMMUTABLE, STABLE, or VOLATILE.  PostgreSQL assumes VOLATILE by default.
+      #   * :cost : The estimated cost of the function, used by the query planner.
+      #   * :language : The language the function uses.  SQL is the default.
+      #   * :link_symbol : For a dynamically loaded see function, the function's link symbol if different from the definition argument.
+      #   * :returns : The data type returned by the function.  If you are using OUT or INOUT argument modes, this is ignored.
+      #     Otherwise, if this is not specified, void is used by default to specify the function is not supposed to return a value.
+      #   * :rows : The estimated number of rows the function will return.  Only use if the function returns SETOF something.
+      #   * :security_definer : Makes the privileges of the function the same as the privileges of the user who defined the function instead of
+      #     the privileges of the user who runs the function.  There are security implications when doing this, see the PostgreSQL documentation.
+      #   * :set : Configuration variables to set while the function is being run, can be a hash or an array of two pairs.  search_path is
+      #     often used here if :security_definer is used.
+      #   * :strict : Makes the function return NULL when any argument is NULL.
+      def create_function(name, definition, opts={})
+        self << create_function_sql(name, definition, opts)
+      end
+      
+      # Create the procedural language in the database. Arguments:
+      # * name : Name of the procedural language (e.g. plpgsql)
+      # * opts : options hash:
+      #   * :handler : The name of a previously registered function used as a call handler for this language.
+      #   * :trusted : Marks the language being created as trusted, allowing unprivileged users to create functions using this language.
+      #   * :validator : The name of previously registered function used as a validator of functions defined in this language.
+      def create_language(name, opts={})
+        self << create_language_sql(name, opts)
+      end
+      
+      # Create a trigger in the database.  Arguments:
+      # * table : the table on which this trigger operates
+      # * name : the name of this trigger
+      # * function : the function to call for this trigger, which should return type trigger.
+      # * opts : options hash:
+      #   * :after : Calls the trigger after execution instead of before.
+      #   * :args : An argument or array of arguments to pass to the function.
+      #   * :each_row : Calls the trigger for each row instead of for each statement.
+      #   * :events : Can be :insert, :update, :delete, or an array of any of those. Calls the trigger whenever that type of statement is used.  By default,
+      #     the trigger is called for insert, update, or delete.
+      def create_trigger(table, name, function, opts={})
+        self << create_trigger_sql(table, name, function, opts)
+      end
+      
+      # PostgreSQL uses the :postgres database type.
+      def database_type
+        :postgres
+      end
+
+      # Drops the function from the database. Arguments:
+      # * name : name of the function to drop
+      # * opts : options hash:
+      #   * :args : The arguments for the function.  See create_function_sql.
+      #   * :cascade : Drop other objects depending on this function.
+      #   * :if_exists : Don't raise an error if the function doesn't exist.
+      def drop_function(name, opts={})
+        self << drop_function_sql(name, opts)
+      end
+      
+      # Drops a procedural language from the database.  Arguments:
+      # * name : name of the procedural language to drop
+      # * opts : options hash:
+      #   * :cascade : Drop other objects depending on this function.
+      #   * :if_exists : Don't raise an error if the function doesn't exist.
+      def drop_language(name, opts={})
+        self << drop_language_sql(name, opts)
+      end
+      
+      # Remove the cached entries for primary keys and sequences when dropping a table.
+      def drop_table(*names)
+        names.each do |name|
+          name = quote_schema_table(name)
+          @primary_keys.delete(name)
+          @primary_key_sequences.delete(name)
+        end
+        super
+      end
+
+      # Drops a trigger from the database.  Arguments:
+      # * table : table from which to drop the trigger
+      # * name : name of the trigger to drop
+      # * opts : options hash:
+      #   * :cascade : Drop other objects depending on this function.
+      #   * :if_exists : Don't raise an error if the function doesn't exist.
+      def drop_trigger(table, name, opts={})
+        self << drop_trigger_sql(table, name, opts)
+      end
+      
+      # Return a hash containing index information. Hash keys are index name symbols.
+      # Values are subhashes with two keys, :columns and :unique.  The value of :columns
+      # is an array of symbols of column names.  The value of :unique is true or false
+      # depending on if the index is unique.
+      def indexes(table)
+        m = output_identifier_meth
+        im = input_identifier_meth
+        schema, table = schema_and_table(table)
+        ds = metadata_dataset.
+          from(:pg_class___tab).
+          join(:pg_index___ind, :indrelid=>:oid, im.call(table)=>:relname).
+          join(:pg_class___indc, :oid=>:indexrelid).
+          join(:pg_attribute___att, :attrelid=>:tab__oid, :attnum=>SQL::Function.new(:ANY, :ind__indkey)).
+          filter(:indc__relkind=>'i', :ind__indisprimary=>false).
+          exclude(0=>SQL::Function.new(:ANY, :ind__indkey)).
+          order(:indc__relname, (0...32).map{|x| [SQL::Subscript.new(:ind__indkey, [x]), x]}.case(32, :att__attnum)).
+          select(:indc__relname___name, :ind__indisunique___unique, :att__attname___column)
+        
+        ds.join!(:pg_namespace___nsp, :oid=>:tab__relnamespace, :nspname=>schema) if schema
+        
+        indexes = {}
+        ds.each do |r|
+          i = indexes[m.call(r[:name])] ||= {:columns=>[], :unique=>r[:unique]}
+          i[:columns] << m.call(r[:column])
+        end
+        indexes
+      end
+
+      # Dataset containing all current database locks 
+      def locks
+        dataset.from(:pg_class).join(:pg_locks, :relation=>:relfilenode).select(:pg_class__relname, Sequel::SQL::ColumnAll.new(:pg_locks))
+      end
+      
+      # Return primary key for the given table.
+      def primary_key(table, opts={})
+        quoted_table = quote_schema_table(table)
+        return @primary_keys[quoted_table] if @primary_keys.include?(quoted_table)
+        @primary_keys[quoted_table] = if conn = opts[:conn]
+          conn.primary_key(*schema_and_table(table))
+        else
+          synchronize(opts[:server]){|con| con.primary_key(*schema_and_table(table))}
+        end
+      end
+      
+      # Return the sequence providing the default for the primary key for the given table.
+      def primary_key_sequence(table, opts={})
+        quoted_table = quote_schema_table(table)
+        return @primary_key_sequences[quoted_table] if @primary_key_sequences.include?(quoted_table)
+        @primary_key_sequences[quoted_table] = if conn = opts[:conn]
+          conn.sequence(*schema_and_table(table))
+        else
+          synchronize(opts[:server]){|con| con.sequence(*schema_and_table(table))}
+        end
+      end
+      
+      # Reset the primary key sequence for the given table, baseing it on the
+      # maximum current value of the table's primary key.
+      def reset_primary_key_sequence(table)
+        pk = SQL::Identifier.new(primary_key(table))
+        seq = primary_key_sequence(table)
+        db = self
+        seq_ds = db.from(seq.lit)
+        get{setval(seq, db[table].select{coalesce(max(pk)+seq_ds.select{:increment_by}, seq_ds.select(:min_value))}, false)}
+      end
+
+      # PostgreSQL uses SERIAL psuedo-type instead of AUTOINCREMENT for
+      # managing incrementing primary keys.
+      def serial_primary_key_options
+        {:primary_key => true, :type => :serial}
+      end
+      
+      # The version of the PostgreSQL server, used for determining capability.
+      def server_version(server=nil)
+        return @server_version if @server_version
+        @server_version = synchronize(server) do |conn|
+          (conn.server_version rescue nil) if conn.respond_to?(:server_version)
+        end
+        unless @server_version
+          m = /PostgreSQL (\d+)\.(\d+)\.(\d+)/.match(get(SQL::Function.new(:version)))
+          @server_version = (m[1].to_i * 10000) + (m[2].to_i * 100) + m[3].to_i
+        end
+        @server_version
+      end
+      
+      # Whether the given table exists in the database
+      #
+      # Options:
+      # * :schema - The schema to search (default_schema by default)
+      # * :server - The server to use
+      def table_exists?(table, opts={})
+        im = input_identifier_meth
+        schema, table = schema_and_table(table)
+        opts[:schema] ||= schema
+        tables(opts){|ds| !ds.first(:relname=>im.call(table)).nil?}
+      end
+      
+      # Array of symbols specifying table names in the current database.
+      # The dataset used is yielded to the block if one is provided,
+      # otherwise, an array of symbols of table names is returned.  
+      #
+      # Options:
+      # * :schema - The schema to search (default_schema by default)
+      # * :server - The server to use
+      def tables(opts={})
+        ds = metadata_dataset.from(:pg_class).filter(:relkind=>'r').select(:relname).exclude(SQL::StringExpression.like(:relname, SYSTEM_TABLE_REGEXP)).server(opts[:server])
+        ds.join!(:pg_namespace, :oid=>:relnamespace, :nspname=>(opts[:schema]||default_schema).to_s) if opts[:schema] || default_schema
+        m = output_identifier_meth
+        block_given? ? yield(ds) : ds.map{|r| m.call(r[:relname])}
+      end
+      
+      # PostgreSQL supports multi-level transactions using save points.
+      # To use a savepoint instead of reusing the current transaction,
+      # use the :savepoint=>true option.
+      def transaction(opts={})
+        synchronize(opts[:server]) do |conn|
+          return yield(conn) if @transactions.include?(Thread.current) and !opts[:savepoint]
+          conn.transaction_depth ||= 0
+          if conn.transaction_depth > 0
+            log_info(SQL_SAVEPOINT % conn.transaction_depth)
+            conn.execute(SQL_SAVEPOINT % conn.transaction_depth)
+          else
+            log_info(SQL_BEGIN)
+            conn.execute(SQL_BEGIN)
+          end
+          begin
+            conn.transaction_depth += 1
+            @transactions << Thread.current
+            yield conn
+          rescue ::Exception => e
+            if conn.transaction_depth > 1
+              log_info(SQL_ROLLBACK_TO_SAVEPOINT % [conn.transaction_depth - 1])
+              conn.execute(SQL_ROLLBACK_TO_SAVEPOINT % [conn.transaction_depth - 1])
+            else
+              log_info(SQL_ROLLBACK)
+              conn.execute(SQL_ROLLBACK) rescue nil
+              @transactions.delete(Thread.current)
+            end
+            transaction_error(e, *CONVERTED_EXCEPTIONS)
+          ensure
+            unless e
+              begin
+                if conn.transaction_depth > 1
+                  log_info(SQL_RELEASE_SAVEPOINT % [conn.transaction_depth - 1])
+                  conn.execute(SQL_RELEASE_SAVEPOINT % [conn.transaction_depth - 1])
+                else
+                  log_info(SQL_COMMIT)
+                  conn.execute(SQL_COMMIT)
+                  @transactions.delete(Thread.current)
+                end
+              rescue => e
+                log_info(e.message)
+                raise_error(e, :classes=>CONVERTED_EXCEPTIONS)
+              end
+            end
+            conn.transaction_depth -= 1
+          end
+        end
+      end
+
+      private
+      
+      # SQL statement to create database function.
+      def create_function_sql(name, definition, opts={})
+        args = opts[:args]
+        if !opts[:args].is_a?(Array) || !opts[:args].any?{|a| Array(a).length == 3 and %w'OUT INOUT'.include?(a[2].to_s)}
+          returns = opts[:returns] || 'void'
+        end
+        language = opts[:language] || 'SQL'
+        <<-END
+        CREATE#{' OR REPLACE' if opts[:replace]} FUNCTION #{name}#{sql_function_args(args)}
+        #{"RETURNS #{returns}" if returns}
+        LANGUAGE #{language}
+        #{opts[:behavior].to_s.upcase if opts[:behavior]}
+        #{'STRICT' if opts[:strict]}
+        #{'SECURITY DEFINER' if opts[:security_definer]}
+        #{"COST #{opts[:cost]}" if opts[:cost]}
+        #{"ROWS #{opts[:rows]}" if opts[:rows]}
+        #{opts[:set].map{|k,v| " SET #{k} = #{v}"}.join("\n") if opts[:set]}
+        AS #{literal(definition.to_s)}#{", #{literal(opts[:link_symbol].to_s)}" if opts[:link_symbol]}
+        END
+      end
+      
+      # SQL for creating a procedural language.
+      def create_language_sql(name, opts={})
+        "CREATE#{' TRUSTED' if opts[:trusted]} LANGUAGE #{name}#{" HANDLER #{opts[:handler]}" if opts[:handler]}#{" VALIDATOR #{opts[:validator]}" if opts[:validator]}"
+      end
+      
+      # SQL for creating a database trigger. 
+      def create_trigger_sql(table, name, function, opts={})
+        events = opts[:events] ? Array(opts[:events]) : [:insert, :update, :delete]
+        whence = opts[:after] ? 'AFTER' : 'BEFORE'
+        "CREATE TRIGGER #{name} #{whence} #{events.map{|e| e.to_s.upcase}.join(' OR ')} ON #{quote_schema_table(table)}#{' FOR EACH ROW' if opts[:each_row]} EXECUTE PROCEDURE #{function}(#{Array(opts[:args]).map{|a| literal(a)}.join(', ')})"
+      end
+      
+      # SQL for dropping a function from the database. 
+      def drop_function_sql(name, opts={})
+        "DROP FUNCTION#{' IF EXISTS' if opts[:if_exists]} #{name}#{sql_function_args(opts[:args])}#{' CASCADE' if opts[:cascade]}"
+      end
+      
+      # SQL for dropping a procedural language from the database.
+      def drop_language_sql(name, opts={})
+        "DROP LANGUAGE#{' IF EXISTS' if opts[:if_exists]} #{name}#{' CASCADE' if opts[:cascade]}"
+      end
+
+      # Always CASCADE the table drop
+      def drop_table_sql(name)
+        "DROP TABLE #{quote_schema_table(name)} CASCADE"
+      end
+      
+      # SQL for dropping a trigger from the database.
+      def drop_trigger_sql(table, name, opts={})
+        "DROP TRIGGER#{' IF EXISTS' if opts[:if_exists]} #{name} ON #{quote_schema_table(table)}#{' CASCADE' if opts[:cascade]}"
+      end
+
+      # PostgreSQL folds unquoted identifiers to lowercase, so it shouldn't need to upcase identifiers on input.
+      def identifier_input_method_default
+        nil
+      end
+      
+      # PostgreSQL folds unquoted identifiers to lowercase, so it shouldn't need to upcase identifiers on output.
+      def identifier_output_method_default
+        nil
+      end
+
+      # PostgreSQL specific index SQL.
+      def index_definition_sql(table_name, index)
+        cols = index[:columns]
+        index_name = index[:name] || default_index_name(table_name, cols)
+        expr = literal(Array(cols))
+        unique = "UNIQUE " if index[:unique]
+        index_type = index[:type]
+        filter = index[:where] || index[:filter]
+        filter = " WHERE #{filter_expr(filter)}" if filter
+        case index_type
+        when :full_text
+          expr = "(to_tsvector(#{literal(index[:language] || 'simple')}, #{dataset.send(:full_text_string_join, cols)}))"
+          index_type = :gin
+        when :spatial
+          index_type = :gist
+        end
+        "CREATE #{unique}INDEX #{quote_identifier(index_name)} ON #{quote_schema_table(table_name)} #{"USING #{index_type} " if index_type}#{expr}#{filter}"
+      end
+      
+      # The result of the insert for the given table and values.  If values
+      # is an array, assume the first column is the primary key and return
+      # that.  If values is a hash, lookup the primary key for the table.  If
+      # the primary key is present in the hash, return its value.  Otherwise,
+      # look up the sequence for the table's primary key.  If one exists,
+      # return the last value the of the sequence for the connection.
+      def insert_result(conn, table, values)
+        case values
+        when Hash
+          return nil unless pk = primary_key(table, :conn=>conn)
+          if pk and pkv = values[pk.to_sym]
+            pkv
+          else
+            begin
+              if seq = primary_key_sequence(table, :conn=>conn)
+                conn.last_insert_id(seq)
+              end
+            rescue Exception => e
+              raise_error(e, :classes=>CONVERTED_EXCEPTIONS) unless RE_CURRVAL_ERROR.match(e.message)
+            end
+          end
+        when Array
+          values.first
+        else
+          nil
+        end
+      end
+      
+      # Use a dollar sign instead of question mark for the argument
+      # placeholder.
+      def prepared_arg_placeholder
+        PREPARED_ARG_PLACEHOLDER
+      end
+      
+      # SQL DDL statement for renaming a table. PostgreSQL doesn't allow you to change a table's schema in
+      # a rename table operation, so speciying a new schema in new_name will not have an effect.
+      def rename_table_sql(name, new_name)
+        "ALTER TABLE #{quote_schema_table(name)} RENAME TO #{quote_identifier(schema_and_table(new_name).last)}"
+      end 
+
+      # PostgreSQL's autoincrementing primary keys are of type integer or bigint
+      # using a nextval function call as a default.
+      def schema_autoincrementing_primary_key?(schema)
+        super and schema[:db_type] =~ /\A(?:integer|bigint)\z/io and schema[:default]=~/\Anextval/io
+      end
+
+      # The dataset used for parsing table schemas, using the pg_* system catalogs.
+      def schema_parse_table(table_name, opts)
+        m = output_identifier_meth
+        m2 = input_identifier_meth
+        ds = metadata_dataset.select(:pg_attribute__attname___name,
+            SQL::Function.new(:format_type, :pg_type__oid, :pg_attribute__atttypmod).as(:db_type),
+            SQL::Function.new(:pg_get_expr, :pg_attrdef__adbin, :pg_class__oid).as(:default),
+            SQL::BooleanExpression.new(:NOT, :pg_attribute__attnotnull).as(:allow_null),
+            SQL::Function.new(:COALESCE, SQL::BooleanExpression.from_value_pairs(:pg_attribute__attnum => SQL::Function.new(:ANY, :pg_index__indkey)), false).as(:primary_key)).
+          from(:pg_class).
+          join(:pg_attribute, :attrelid=>:oid).
+          join(:pg_type, :oid=>:atttypid).
+          left_outer_join(:pg_attrdef, :adrelid=>:pg_class__oid, :adnum=>:pg_attribute__attnum).
+          left_outer_join(:pg_index, :indrelid=>:pg_class__oid, :indisprimary=>true).
+          filter(:pg_attribute__attisdropped=>false).
+          filter{|o| o.pg_attribute__attnum > 0}.
+          filter(:pg_class__relname=>m2.call(table_name)).
+          order(:pg_attribute__attnum)
+        ds.join!(:pg_namespace, :oid=>:pg_class__relnamespace, :nspname=>(opts[:schema] || default_schema).to_s) if opts[:schema] || default_schema
+        ds.map do |row|
+          row[:default] = nil if blank_object?(row[:default])
+          row[:type] = schema_column_type(row[:db_type])
+          [m.call(row.delete(:name)), row]
+        end
+      end
+
+      # Turns an array of argument specifiers into an SQL fragment used for function arguments.  See create_function_sql.
+      def sql_function_args(args)
+        "(#{Array(args).map{|a| Array(a).reverse.join(' ')}.join(', ')})"
+      end
+      
+      # PostgreSQL uses the bytea data type for blobs
+      def type_literal_generic_file(column)
+        :bytea
+      end
+
+      # PostgreSQL prefers the text datatype.  If a fixed size is requested,
+      # the char type is used.  If the text type is specifically
+      # disallowed or there is a size specified, use the varchar type.
+      # Otherwise use the type type.
+      def type_literal_generic_string(column)
+        if column[:fixed]
+          "char(#{column[:size]||255})"
+        elsif column[:text] == false or column[:size]
+          "varchar(#{column[:size]||255})"
+        else
+          :text
+        end
+      end
+    end
+    
+    # Instance methods for datasets that connect to a PostgreSQL database.
+    module DatasetMethods
+      ACCESS_SHARE = 'ACCESS SHARE'.freeze
+      ACCESS_EXCLUSIVE = 'ACCESS EXCLUSIVE'.freeze
+      BOOL_FALSE = 'false'.freeze
+      BOOL_TRUE = 'true'.freeze
+      COMMA_SEPARATOR = ', '.freeze
+      EXCLUSIVE = 'EXCLUSIVE'.freeze
+      EXPLAIN = 'EXPLAIN '.freeze
+      EXPLAIN_ANALYZE = 'EXPLAIN ANALYZE '.freeze
+      FOR_SHARE = ' FOR SHARE'.freeze
+      FOR_UPDATE = ' FOR UPDATE'.freeze
+      LOCK = 'LOCK TABLE %s IN %s MODE'.freeze
+      NULL = LiteralString.new('NULL').freeze
+      PG_TIMESTAMP_FORMAT = "TIMESTAMP '%Y-%m-%d %H:%M:%S".freeze
+      QUERY_PLAN = 'QUERY PLAN'.to_sym
+      ROW_EXCLUSIVE = 'ROW EXCLUSIVE'.freeze
+      ROW_SHARE = 'ROW SHARE'.freeze
+      SELECT_CLAUSE_ORDER = %w'distinct columns from join where group having compounds order limit lock'.freeze
+      SHARE = 'SHARE'.freeze
+      SHARE_ROW_EXCLUSIVE = 'SHARE ROW EXCLUSIVE'.freeze
+      SHARE_UPDATE_EXCLUSIVE = 'SHARE UPDATE EXCLUSIVE'.freeze
+      
+      # Shared methods for prepared statements when used with PostgreSQL databases.
+      module PreparedStatementMethods
+        # Override insert action to use RETURNING if the server supports it.
+        def prepared_sql
+          return @prepared_sql if @prepared_sql
+          super
+          if @prepared_type == :insert and !@opts[:disable_insert_returning] and server_version >= 80200
+            @prepared_sql = insert_returning_pk_sql(@prepared_modify_values)
+            meta_def(:insert_returning_pk_sql){|*args| prepared_sql}
+          end
+          @prepared_sql
+        end
+      end
+
+      # Add the disable_insert_returning! mutation method
+      def self.extended(obj)
+        obj.def_mutation_method(:disable_insert_returning)
+      end
+
+      # Add the disable_insert_returning! mutation method
+      def self.included(mod)
+        mod.def_mutation_method(:disable_insert_returning)
+      end
+
+      # Return the results of an ANALYZE query as a string
+      def analyze(opts = nil)
+        analysis = []
+        fetch_rows(EXPLAIN_ANALYZE + select_sql(opts)) do |r|
+          analysis << r[QUERY_PLAN]
+        end
+        analysis.join("\r\n")
+      end
+      
+      # Disable the use of INSERT RETURNING, even if the server supports it
+      def disable_insert_returning
+        clone(:disable_insert_returning=>true)
+      end
+
+      # Return the results of an EXPLAIN query as a string
+      def explain(opts = nil)
+        analysis = []
+        fetch_rows(EXPLAIN + select_sql(opts)) do |r|
+          analysis << r[QUERY_PLAN]
+        end
+        analysis.join("\r\n")
+      end
+      
+      # Return a cloned dataset with a :share lock type.
+      def for_share
+        clone(:lock => :share)
+      end
+      
+      # Return a cloned dataset with a :update lock type.
+      def for_update
+        clone(:lock => :update)
+      end
+      
+      # PostgreSQL specific full text search syntax, using tsearch2 (included
+      # in 8.3 by default, and available for earlier versions as an add-on).
+      def full_text_search(cols, terms, opts = {})
+        lang = opts[:language] || 'simple'
+        filter("to_tsvector(#{literal(lang)}, #{full_text_string_join(cols)}) @@ to_tsquery(#{literal(lang)}, #{literal(Array(terms).join(' | '))})")
+      end
+      
+      # Insert given values into the database.
+      def insert(*values)
+        if !@opts[:sql] and !@opts[:disable_insert_returning] and server_version >= 80200
+          clone(default_server_opts(:sql=>insert_returning_pk_sql(*values))).single_value
+        else
+          execute_insert(insert_sql(*values), :table=>opts[:from].first,
+            :values=>values.size == 1 ? values.first : values)
+        end
+      end
+
+      # Use the RETURNING clause to return the columns listed in returning.
+      def insert_returning_sql(returning, *values)
+        "#{insert_sql(*values)} RETURNING #{column_list(Array(returning))}"
+      end
+
+      # Insert a record returning the record inserted
+      def insert_select(*values)
+        naked.clone(default_server_opts(:sql=>insert_returning_sql(nil, *values))).single_record if server_version >= 80200
+      end
+
+      # Locks the table with the specified mode.
+      def lock(mode, server=nil)
+        sql = LOCK % [source_list(@opts[:from]), mode]
+        @db.synchronize(server) do
+          if block_given? # perform locking inside a transaction and yield to block
+            @db.transaction(server){@db.execute(sql, :server=>server); yield}
+          else
+            @db.execute(sql, :server=>server) # lock without a transaction
+            self
+          end
+        end
+      end
+      
+      # For PostgreSQL version > 8.2, allow inserting multiple rows at once.
+      def multi_insert_sql(columns, values)
+        return super if server_version < 80200
+        
+        # postgresql 8.2 introduces support for multi-row insert
+        values = values.map {|r| literal(Array(r))}.join(COMMA_SEPARATOR)
+        ["INSERT INTO #{source_list(@opts[:from])} (#{identifier_list(columns)}) VALUES #{values}"]
+      end
+      
+      private
+      
+      # Use the RETURNING clause to return the primary key of the inserted record, if it exists
+      def insert_returning_pk_sql(*values)
+        pk = db.primary_key(opts[:from].first)
+        insert_returning_sql(pk ? Sequel::SQL::Identifier.new(pk) : NULL, *values)
+      end
+      
+      def literal_blob(v)
+        "'#{v.gsub(/[\000-\037\047\134\177-\377]/){|b| "\\#{("%o" % b[0..1].unpack("C")[0]).rjust(3, '0')}"}}'"
+      end
+
+      def literal_datetime(v)
+        "#{v.strftime(PG_TIMESTAMP_FORMAT)}.#{sprintf("%06d", (v.sec_fraction * 86400000000).to_i)}'"
+      end
+
+      def literal_false
+        BOOL_FALSE
+      end
+
+      def literal_string(v)
+        "'#{v.gsub("'", "''")}'"
+      end
+
+      def literal_true
+        BOOL_TRUE
+      end
+
+      def literal_time(v)
+        "#{v.strftime(PG_TIMESTAMP_FORMAT)}.#{sprintf("%06d",v.usec)}'"
+      end
+
+      # PostgreSQL is smart and can use parantheses around all datasets to get
+      # the correct answers.
+      def select_compounds_sql(sql)
+        return unless @opts[:compounds]
+        @opts[:compounds].each do |type, dataset, all|
+          sql.replace("(#{sql} #{type.to_s.upcase}#{' ALL' if all} #{subselect_sql(dataset)})")
+        end
+      end
+
+      # The order of clauses in the SELECT SQL statement
+      def select_clause_order
+        SELECT_CLAUSE_ORDER
+      end
+
+      # Support lock mode, allowing FOR SHARE and FOR UPDATE queries.
+      def select_lock_sql(sql)
+        case @opts[:lock]
+        when :update
+          sql << FOR_UPDATE
+        when :share
+          sql << FOR_SHARE
+        end
+      end
+      
+      # The version of the database server
+      def server_version
+        db.server_version(@opts[:server])
+      end
+
+      # Concatenate the expressions with a space in between
+      def full_text_string_join(cols)
+        cols = Array(cols).map{|x| SQL::Function.new(:COALESCE, x, '')}
+        cols = cols.zip([' '] * cols.length).flatten
+        cols.pop
+        literal(SQL::StringExpression.new(:'||', *cols))
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/adapters/shared/progress.rb 3.0.0-1/lib/sequel/adapters/shared/progress.rb
--- 1.5.1-1/lib/sequel/adapters/shared/progress.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/adapters/shared/progress.rb	2009-04-28 00:21:46.000000000 +0100
@@ -0,0 +1,39 @@
+Sequel.require %w'date_format unsupported', 'adapters/utils'
+
+module Sequel
+  module Progress
+    module DatabaseMethods
+
+      # Progress uses the :progress database type.
+      def database_type
+        :progress
+      end
+
+      def dataset(opts = nil)
+        ds = super
+        ds.extend(DatasetMethods)
+        ds
+      end
+    end
+  
+    module DatasetMethods
+      include Dataset::UnsupportedIntersectExcept
+      include Dataset::SQLStandardDateFormat
+
+      SELECT_CLAUSE_ORDER = %w'limit distinct columns from join where group order having compounds'.freeze
+
+      private
+
+      def select_clause_order
+        SELECT_CLAUSE_ORDER
+      end
+
+      # Progress uses TOP for limit, but it is only supported in Progress 10.
+      # The Progress adapter targets Progress 9, so it silently ignores the option.
+      def select_limit_sql(sql)
+        raise(Error, "OFFSET not supported") if @opts[:offset]
+        #sql << " TOP #{@opts[:limit]}" if @opts[:limit]
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/adapters/shared/sqlite.rb 3.0.0-1/lib/sequel/adapters/shared/sqlite.rb
--- 1.5.1-1/lib/sequel/adapters/shared/sqlite.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/adapters/shared/sqlite.rb	2009-04-28 00:31:53.000000000 +0100
@@ -0,0 +1,293 @@
+Sequel.require 'adapters/utils/unsupported'
+
+module Sequel
+  module SQLite
+    module DatabaseMethods
+      AUTO_VACUUM = [:none, :full, :incremental].freeze
+      PRIMARY_KEY_INDEX_RE = /\Asqlite_autoindex_/.freeze
+      SYNCHRONOUS = [:off, :normal, :full].freeze
+      TABLES_FILTER = "type = 'table' AND NOT name = 'sqlite_sequence'"
+      TEMP_STORE = [:default, :file, :memory].freeze
+      
+      # Run all alter_table commands in a transaction.  This is technically only
+      # needed for drop column.
+      def alter_table(name, generator=nil, &block)
+        transaction{super}
+      end
+
+      # A symbol signifying the value of the auto_vacuum PRAGMA.
+      def auto_vacuum
+        AUTO_VACUUM[pragma_get(:auto_vacuum).to_i]
+      end
+      
+      # Set the auto_vacuum PRAGMA using the given symbol (:none, :full, or
+      # :incremental).
+      def auto_vacuum=(value)
+        value = AUTO_VACUUM.index(value) || (raise Error, "Invalid value for auto_vacuum option. Please specify one of :none, :full, :incremental.")
+        pragma_set(:auto_vacuum, value)
+      end
+      
+      # SQLite uses the :sqlite database type.
+      def database_type
+        :sqlite
+      end
+
+      # Return a hash containing index information. Hash keys are index name symbols.
+      # Values are subhashes with two keys, :columns and :unique.  The value of :columns
+      # is an array of symbols of column names.  The value of :unique is true or false
+      # depending on if the index is unique.
+      def indexes(table)
+        m = output_identifier_meth
+        im = input_identifier_meth
+        indexes = {}
+        begin
+          metadata_dataset.with_sql("PRAGMA index_list(?)", im.call(table)).each do |r|
+            next if r[:name] =~ PRIMARY_KEY_INDEX_RE
+            indexes[m.call(r[:name])] = {:unique=>r[:unique].to_i==1}
+          end
+        rescue Sequel::DatabaseError
+          nil
+        else
+          indexes.each do |k, v|
+            v[:columns] = metadata_dataset.with_sql("PRAGMA index_info(?)", im.call(k)).map(:name).map{|x| m.call(x)}
+          end
+        end
+        indexes
+      end
+
+      # Get the value of the given PRAGMA.
+      def pragma_get(name)
+        self["PRAGMA #{name}"].single_value
+      end
+      
+      # Set the value of the given PRAGMA to value.
+      def pragma_set(name, value)
+        execute_ddl("PRAGMA #{name} = #{value}")
+      end
+      
+      # A symbol signifying the value of the synchronous PRAGMA.
+      def synchronous
+        SYNCHRONOUS[pragma_get(:synchronous).to_i]
+      end
+      
+      # Set the synchronous PRAGMA using the given symbol (:off, :normal, or :full).
+      def synchronous=(value)
+        value = SYNCHRONOUS.index(value) || (raise Error, "Invalid value for synchronous option. Please specify one of :off, :normal, :full.")
+        pragma_set(:synchronous, value)
+      end
+      
+      # Array of symbols specifying the table names in the current database.
+      #
+      # Options:
+      # * :server - Set the server to use.
+      def tables(opts={})
+        m = output_identifier_meth
+        metadata_dataset.from(:sqlite_master).server(opts[:server]).filter(TABLES_FILTER).map{|r| m.call(r[:name])}
+      end
+      
+      # A symbol signifying the value of the temp_store PRAGMA.
+      def temp_store
+        TEMP_STORE[pragma_get(:temp_store).to_i]
+      end
+      
+      # Set the temp_store PRAGMA using the given symbol (:default, :file, or :memory).
+      def temp_store=(value)
+        value = TEMP_STORE.index(value) || (raise Error, "Invalid value for temp_store option. Please specify one of :default, :file, :memory.")
+        pragma_set(:temp_store, value)
+      end
+      
+      private
+
+      # SQLite supports limited table modification.  You can add a column
+      # or an index.  Dropping columns is supported by copying the table into
+      # a temporary table, dropping the table, and creating a new table without
+      # the column inside of a transaction.
+      def alter_table_sql(table, op)
+        case op[:op]
+        when :add_column, :add_index, :drop_index
+          super
+        when :drop_column
+          qt = quote_schema_table(table)
+          bt = quote_identifier(backup_table_name(qt.gsub('`', '')))
+          columns_str = dataset.send(:identifier_list, columns_for(table, :except => op[:name]))
+          defined_columns_str = column_list_sql(defined_columns_for(table, :except => op[:name]))
+          ["CREATE TEMPORARY TABLE #{bt}(#{defined_columns_str})",
+           "INSERT INTO #{bt} SELECT #{columns_str} FROM #{qt}",
+           "DROP TABLE #{qt}",
+           "CREATE TABLE #{qt}(#{defined_columns_str})",
+           "INSERT INTO #{qt} SELECT #{columns_str} FROM #{bt}",
+           "DROP TABLE #{bt}"]
+        when :rename_column
+          qt = quote_schema_table(table)
+          bt = quote_identifier(backup_table_name(qt.gsub('`', '')))
+          old_columns = dataset.send(:identifier_list, columns_for(table))
+          new_columns_arr = columns_for(table)
+
+          # Replace the old column in place. This is extremely important.
+          new_columns_arr[new_columns_arr.index(op[:name])] = op[:new_name]
+          
+          new_columns = dataset.send(:identifier_list, new_columns_arr)
+          
+          def_old_columns = column_list_sql(defined_columns_for(table))
+
+          def_new_columns_arr = defined_columns_for(table).map do |c|
+            c[:name] = op[:new_name].to_s if c[:name] == op[:name].to_s
+            c
+          end
+          
+          def_new_columns = column_list_sql(def_new_columns_arr)
+
+          [
+           "CREATE TEMPORARY TABLE #{bt}(#{def_old_columns})",
+           "INSERT INTO #{bt}(#{old_columns}) SELECT #{old_columns} FROM #{qt}",
+           "DROP TABLE #{qt}",
+           "CREATE TABLE #{qt}(#{def_new_columns})",
+           "INSERT INTO #{qt}(#{new_columns}) SELECT #{old_columns} FROM #{bt}",
+           "DROP TABLE #{bt}"
+          ]
+
+        else
+          raise Error, "Unsupported ALTER TABLE operation"
+        end
+      end
+      
+      # The array of column symbols in the table, except for ones given in opts[:except]
+      def backup_table_name(table, opts={})
+        (opts[:times]||1000).times do |i|
+          table_name = "#{table}_backup#{i}"
+          return table_name unless table_exists?(table_name)
+        end
+      end
+
+      # The array of column symbols in the table, except for ones given in opts[:except]
+      def columns_for(table, opts={})
+        cols = schema_parse_table(table, {}).map{|c| c[0]}
+        cols = cols - Array(opts[:except])
+        cols
+      end
+
+      # Allow use without a generator, needed for the alter table hackery that Sequel allows.
+      def column_list_sql(generator)
+        generator.is_a?(Schema::Generator) ? super : generator.map{|c| column_definition_sql(c)}.join(', ')
+      end
+
+      # The array of column schema hashes, except for the ones given in opts[:except]
+      def defined_columns_for(table, opts={})
+        cols = parse_pragma(table, {})
+        cols.each{|c| c[:default] = LiteralString.new(c[:default]) if c[:default]}
+        if opts[:except]
+          nono= Array(opts[:except]).compact.map{|n| n.to_s}
+          cols.reject!{|c| nono.include? c[:name] }
+        end
+        cols
+      end
+      
+      # SQLite folds unquoted identifiers to lowercase, so it shouldn't need to upcase identifiers on input.
+      def identifier_input_method_default
+        nil
+      end
+      
+      # SQLite folds unquoted identifiers to lowercase, so it shouldn't need to upcase identifiers on output.
+      def identifier_output_method_default
+        nil
+      end
+
+      # Parse the output of the table_info pragma
+      def parse_pragma(table_name, opts)
+        metadata_dataset.with_sql("PRAGMA table_info(?)", input_identifier_meth.call(table_name)).map do |row|
+          row.delete(:cid)
+          row[:allow_null] = row.delete(:notnull).to_i == 0
+          row[:default] = row.delete(:dflt_value)
+          row[:primary_key] = row.delete(:pk).to_i == 1
+          row[:default] = nil if blank_object?(row[:default])
+          row[:db_type] = row.delete(:type)
+          row[:type] = schema_column_type(row[:db_type])
+          row
+        end
+      end
+      
+      # SQLite treats integer primary keys as autoincrementing (alias of rowid).
+      def schema_autoincrementing_primary_key?(schema)
+        super and schema[:db_type].downcase == 'integer'
+      end
+
+      # SQLite supports schema parsing using the table_info PRAGMA, so
+      # parse the output of that into the format Sequel expects.
+      def schema_parse_table(table_name, opts)
+        m = output_identifier_meth
+        parse_pragma(table_name, opts).map do |row|
+          [m.call(row.delete(:name)), row]
+        end
+      end
+      
+      # SQLite uses the integer data type even for bignums.  This is because they
+      # are both stored internally as text, and converted when returned from
+      # the database.  Using an integer type instead of bigint makes it more likely
+      # that software will automatically return the column as an integer.
+      def type_literal_generic_bignum(column)
+        :integer
+      end
+    end
+    
+    # Instance methods for datasets that connect to an SQLite database
+    module DatasetMethods
+      include Dataset::UnsupportedIntersectExceptAll
+      include Dataset::UnsupportedIsTrue
+
+      # SQLite does not support pattern matching via regular expressions.
+      # SQLite is case insensitive (depending on pragma), so use LIKE for
+      # ILIKE.
+      def complex_expression_sql(op, args)
+        case op
+        when :~, :'!~', :'~*', :'!~*'
+          raise Error, "SQLite does not support pattern matching via regular expressions"
+        when :LIKE, :'NOT LIKE', :ILIKE, :'NOT ILIKE'
+          # SQLite is case insensitive for ASCII, and non case sensitive for other character sets
+          "#{'NOT ' if [:'NOT LIKE', :'NOT ILIKE'].include?(op)}(#{literal(args.at(0))} LIKE #{literal(args.at(1))})"
+        else
+          super(op, args)
+        end
+      end
+      
+      # SQLite performs a TRUNCATE style DELETE if no filter is specified.
+      # Since we want to always return the count of records, add a condition
+      # that is always true and then delete.
+      def delete
+        @opts[:where] ? super : filter(1=>1).delete
+      end
+      
+      # Insert the values into the database.
+      def insert(*values)
+        execute_insert(insert_sql(*values))
+      end
+      
+      # Allow inserting of values directly from a dataset.
+      def insert_sql(*values)
+        if (values.size == 1) && values.first.is_a?(Sequel::Dataset)
+          "INSERT INTO #{source_list(@opts[:from])} #{values.first.sql};"
+        else
+          super(*values)
+        end
+      end
+      
+      # SQLite uses the nonstandard ` (backtick) for quoting identifiers.
+      def quoted_identifier(c)
+        "`#{c}`"
+      end
+      
+      private
+
+      def literal_blob(v)
+        blob = ''
+        v.each_byte{|x| blob << sprintf('%02x', x)}
+        "X'#{blob}'"
+      end
+      
+      # SQLite uses string literals instead of identifiers in AS clauses.
+      def as_sql(expression, aliaz)
+        aliaz = aliaz.value if aliaz.is_a?(SQL::Identifier)
+        "#{expression} AS #{literal(aliaz.to_s)}"
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/adapters/sqlite.rb 3.0.0-1/lib/sequel/adapters/sqlite.rb
--- 1.5.1-1/lib/sequel/adapters/sqlite.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/adapters/sqlite.rb	2009-04-15 22:25:10.000000000 +0100
@@ -0,0 +1,240 @@
+require 'sqlite3'
+Sequel.require 'adapters/shared/sqlite'
+
+module Sequel
+  # Top level module for holding all SQLite-related modules and classes
+  # for Sequel.
+  module SQLite
+    # Database class for SQLite databases used with Sequel and the
+    # ruby-sqlite3 driver.
+    class Database < Sequel::Database
+      UNIX_EPOCH_TIME_FORMAT = /\A\d+\z/.freeze
+      include ::Sequel::SQLite::DatabaseMethods
+      
+      set_adapter_scheme :sqlite
+      
+      # Mimic the file:// uri, by having 2 preceding slashes specify a relative
+      # path, and 3 preceding slashes specify an absolute path.
+      def self.uri_to_options(uri) # :nodoc:
+        { :database => (uri.host.nil? && uri.path == '/') ? nil : "#{uri.host}#{uri.path}" }
+      end
+      
+      private_class_method :uri_to_options
+      
+      # Connect to the database.  Since SQLite is a file based database,
+      # the only options available are :database (to specify the database
+      # name), and :timeout, to specify how long to wait for the database to
+      # be available if it is locked, given in milliseconds (default is 5000).
+      def connect(server)
+        opts = server_opts(server)
+        opts[:database] = ':memory:' if blank_object?(opts[:database])
+        db = ::SQLite3::Database.new(opts[:database])
+        db.busy_timeout(opts.fetch(:timeout, 5000))
+        db.type_translation = true
+        
+        # Handle datetime's with Sequel.datetime_class
+        prok = proc do |t,v|
+          v = Time.at(v.to_i).iso8601 if UNIX_EPOCH_TIME_FORMAT.match(v)
+          Sequel.string_to_datetime(v)
+        end
+        db.translator.add_translator("timestamp", &prok)
+        db.translator.add_translator("datetime", &prok)
+        
+        # Handle numeric values with BigDecimal
+        prok = proc{|t,v| BigDecimal.new(v) rescue v}
+        db.translator.add_translator("numeric", &prok)
+        db.translator.add_translator("decimal", &prok)
+        db.translator.add_translator("money", &prok)
+        
+        # Handle floating point values with Float
+        prok = proc{|t,v| Float(v) rescue v}
+        db.translator.add_translator("float", &prok)
+        db.translator.add_translator("real", &prok)
+        db.translator.add_translator("double precision", &prok)
+        
+        # Handle blob values with Sequel::SQL::Blob
+        db.translator.add_translator("blob"){|t,v| ::Sequel::SQL::Blob.new(v)}
+        
+        db
+      end
+      
+      # Return instance of Sequel::SQLite::Dataset with the given options.
+      def dataset(opts = nil)
+        SQLite::Dataset.new(self, opts)
+      end
+      
+      # Run the given SQL with the given arguments and return the number of changed rows.
+      def execute_dui(sql, opts={})
+        _execute(sql, opts){|conn| conn.execute_batch(sql, opts[:arguments]); conn.changes}
+      end
+      
+      # Run the given SQL with the given arguments and return the last inserted row id.
+      def execute_insert(sql, opts={})
+        _execute(sql, opts){|conn| conn.execute(sql, opts[:arguments]); conn.last_insert_row_id}
+      end
+      
+      # Run the given SQL with the given arguments and yield each row.
+      def execute(sql, opts={}, &block)
+        _execute(sql, opts){|conn| conn.query(sql, opts[:arguments], &block)}
+      end
+      
+      # Run the given SQL with the given arguments and return the first value of the first row.
+      def single_value(sql, opts={})
+        _execute(sql, opts){|conn| conn.get_first_value(sql, opts[:arguments])}
+      end
+      
+      # Use the native driver transaction method if there isn't already a transaction
+      # in progress on the connection, always yielding a connection inside a transaction
+      # transaction.
+      def transaction(opts={})
+        synchronize(opts[:server]) do |conn|
+          return yield(conn) if conn.transaction_active?
+          begin
+            result = nil
+            log_info('Transaction.begin')
+            conn.transaction{result = yield(conn)}
+            result
+          rescue ::Exception => e
+            log_info('Transaction.rollback')
+            transaction_error(e, SQLite3::Exception)
+          ensure
+            log_info('Transaction.commit') unless e
+          end
+        end
+      end
+      
+      private
+      
+      # Log the SQL and the arguments, and yield an available connection.  Rescue
+      # any SQLite3::Exceptions and turn them into DatabaseErrors.
+      def _execute(sql, opts)
+        begin
+          log_info(sql, opts[:arguments])
+          synchronize(opts[:server]){|conn| yield conn}
+        rescue SQLite3::Exception => e
+          raise_error(e)
+        end
+      end
+      
+      # The SQLite adapter does not need the pool to convert exceptions.
+      # Also, force the max connections to 1 if a memory database is being
+      # used, as otherwise each connection gets a separate database.
+      def connection_pool_default_options
+        o = super.merge(:pool_convert_exceptions=>false)
+        # Default to only a single connection if a memory database is used,
+        # because otherwise each connection will get a separate database
+        o[:max_connections] = 1 if @opts[:database] == ':memory:' || blank_object?(@opts[:database])
+        o
+      end
+
+      # Disconnect given connections from the database.
+      def disconnect_connection(c)
+        c.close
+      end
+    end
+    
+    # Dataset class for SQLite datasets that use the ruby-sqlite3 driver.
+    class Dataset < Sequel::Dataset
+      include ::Sequel::SQLite::DatasetMethods
+      
+      EXPLAIN = 'EXPLAIN %s'.freeze
+      PREPARED_ARG_PLACEHOLDER = ':'.freeze
+      
+      # SQLite already supports named bind arguments, so use directly.
+      module ArgumentMapper
+        include Sequel::Dataset::ArgumentMapper
+        
+        protected
+        
+        # Return a hash with the same values as the given hash,
+        # but with the keys converted to strings.
+        def map_to_prepared_args(hash)
+          args = {}
+          hash.each{|k,v| args[k.to_s] = v}
+          args
+        end
+        
+        private
+        
+        # SQLite uses a : before the name of the argument for named
+        # arguments.
+        def prepared_arg(k)
+          LiteralString.new("#{prepared_arg_placeholder}#{k}")
+        end
+      end
+      
+      # SQLite prepared statement uses a new prepared statement each time
+      # it is called, but it does use the bind arguments.
+      module PreparedStatementMethods
+        include ArgumentMapper
+        
+        private
+        
+        # Run execute_select on the database with the given SQL and the stored
+        # bind arguments.
+        def execute(sql, opts={}, &block)
+          super(sql, {:arguments=>bind_arguments}.merge(opts), &block)
+        end
+        
+        # Same as execute, explicit due to intricacies of alias and super.
+        def execute_dui(sql, opts={}, &block)
+          super(sql, {:arguments=>bind_arguments}.merge(opts), &block)
+        end
+        
+        # Same as execute, explicit due to intricacies of alias and super.
+        def execute_insert(sql, opts={}, &block)
+          super(sql, {:arguments=>bind_arguments}.merge(opts), &block)
+        end
+      end
+      
+      # Prepare an unnamed statement of the given type and call it with the
+      # given values.
+      def call(type, hash, values=nil, &block)
+        prepare(type, nil, values).call(hash, &block)
+      end
+      
+      # Return an array of strings specifying a query explanation for the
+      # current dataset.
+      def explain
+        res = []
+        @db.result_set(EXPLAIN % select_sql(opts), nil) {|r| res << r}
+        res
+      end
+      
+      # Yield a hash for each row in the dataset.
+      def fetch_rows(sql)
+        execute(sql) do |result|
+          @columns = result.columns.map{|c| output_identifier(c)}
+          column_count = @columns.size
+          result.each do |values|
+            row = {}
+            column_count.times {|i| row[@columns[i]] = values[i]}
+            yield row
+          end
+        end
+      end
+      
+      # Prepare the given type of query with the given name and store
+      # it in the database.  Note that a new native prepared statement is
+      # created on each call to this prepared statement.
+      def prepare(type, name=nil, values=nil)
+        ps = to_prepared_statement(type, values)
+        ps.extend(PreparedStatementMethods)
+        db.prepared_statements[name] = ps if name
+        ps
+      end
+      
+      private
+      
+      # Quote the string using the adapter class method.
+      def literal_string(v)
+        "'#{::SQLite3::Database.quote(v)}'"
+      end
+
+      # SQLite uses a : before the name of the argument as a placeholder.
+      def prepared_arg_placeholder
+        PREPARED_ARG_PLACEHOLDER
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/adapters/utils/date_format.rb 3.0.0-1/lib/sequel/adapters/utils/date_format.rb
--- 1.5.1-1/lib/sequel/adapters/utils/date_format.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/adapters/utils/date_format.rb	2009-03-16 20:53:19.000000000 +0000
@@ -0,0 +1,21 @@
+# Module containing overrides for Sequel's standard date/time literalization
+# to use the SQL standrd.  The SQL standard is used by fewer databases than
+# the defacto standard (which is just a normal string).
+module Sequel::Dataset::SQLStandardDateFormat
+  private
+
+  # Use SQL standard syntax for Date
+  def literal_date(v)
+    v.strftime("DATE '%Y-%m-%d'") 
+  end
+    
+  # Use SQL standard syntax for DateTime
+  def literal_datetime(v)
+    v.strftime("TIMESTAMP '%Y-%m-%d %H:%M:%S'")
+  end
+
+  # Use SQL standard syntax for Time
+  def literal_time(v)
+    v.strftime("TIMESTAMP '%Y-%m-%d %H:%M:%S'")
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/adapters/utils/stored_procedures.rb 3.0.0-1/lib/sequel/adapters/utils/stored_procedures.rb
--- 1.5.1-1/lib/sequel/adapters/utils/stored_procedures.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/adapters/utils/stored_procedures.rb	2009-03-16 20:53:19.000000000 +0000
@@ -0,0 +1,75 @@
+module Sequel 
+  class Dataset
+    module StoredProcedureMethods
+      SQL_QUERY_TYPE = Hash.new{|h,k| h[k] = k}
+      SQL_QUERY_TYPE[:first] = SQL_QUERY_TYPE[:all] = :select
+      
+      # The name of the stored procedure to call
+      attr_accessor :sproc_name
+      
+      # Call the prepared statement
+      def call(*args, &block)
+        @sproc_args = args
+        case @sproc_type
+        when :select, :all
+          all(&block)
+        when :first
+          first
+        when :insert
+          insert
+        when :update
+          update
+        when :delete
+          delete
+        end
+      end
+      
+      # Programmer friendly string showing this is a stored procedure,
+      # showing the name of the procedure.
+      def inspect
+        "<#{self.class.name}/StoredProcedure name=#{@sproc_name}>"
+      end
+      
+      # Set the type of the sproc and override the corresponding _sql
+      # method to return the empty string (since the result will be
+      # ignored anyway).
+      def sproc_type=(type)
+        @sproc_type = type
+        meta_def("#{sql_query_type}_sql"){|*a| ''}
+      end
+      
+      private
+      
+      # The type of query (:select, :insert, :delete, :update).
+      def sql_query_type
+        SQL_QUERY_TYPE[@sproc_type]
+      end
+    end
+  
+    module StoredProcedures
+      # For the given type (:select, :first, :insert, :update, or :delete),
+      # run the database stored procedure with the given name with the given
+      # arguments.
+      def call_sproc(type, name, *args)
+        prepare_sproc(type, name).call(*args)
+      end
+      
+      # Transform this dataset into a stored procedure that you can call
+      # multiple times with new arguments.
+      def prepare_sproc(type, name)
+        sp = clone
+        prepare_extend_sproc(sp)
+        sp.sproc_type = type
+        sp.sproc_name = name
+        sp
+      end
+      
+      private
+      
+      # Extend the dataset with the stored procedure methods.
+      def prepare_extend_sproc(ds)
+        ds.extend(StoredProcedureMethods)
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/adapters/utils/unsupported.rb 3.0.0-1/lib/sequel/adapters/utils/unsupported.rb
--- 1.5.1-1/lib/sequel/adapters/utils/unsupported.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/adapters/utils/unsupported.rb	2009-04-27 23:55:44.000000000 +0100
@@ -0,0 +1,62 @@
+class Sequel::Dataset
+  # This module should be included in the dataset class for all databases that
+  # don't support INTERSECT or EXCEPT.
+  module UnsupportedIntersectExcept
+    # Raise an Error if EXCEPT is used
+    def except(ds, all=false)
+      raise(Sequel::Error, "EXCEPT not supported")
+    end
+
+    # Raise an Error if INTERSECT is used
+    def intersect(ds, all=false)
+      raise(Sequel::Error, "INTERSECT not supported")
+    end
+    
+    private
+    
+    # Since EXCEPT and INTERSECT are not supported, and order shouldn't matter
+    # when UNION is used, don't worry about parantheses.  This may potentially
+    # give incorrect results if UNION ALL is used.
+    def select_compounds_sql(sql)
+      return unless @opts[:compounds]
+      @opts[:compounds].each do |type, dataset, all|
+        sql << " #{type.to_s.upcase}#{' ALL' if all} #{subselect_sql(dataset)}"
+      end
+    end
+  end
+
+  # This module should be included in the dataset class for all databases that
+  # don't support INTERSECT ALL or EXCEPT ALL.
+  module UnsupportedIntersectExceptAll
+    # Raise an Error if EXCEPT is used
+    def except(ds, all=false)
+      raise(Sequel::Error, "EXCEPT ALL not supported") if all
+      super(ds)
+    end
+
+    # Raise an Error if INTERSECT is used
+    def intersect(ds, all=false)
+      raise(Sequel::Error, "INTERSECT ALL not supported") if all
+      super(ds)
+    end
+  end
+
+  # This module should be included in the dataset class for all databases that
+  # don't support IS [NOT] (TRUE|FALSE)
+  module UnsupportedIsTrue
+    # Use an = construct instead of IS and an != OR IS NULL construct instead of IS NOT.
+    def complex_expression_sql(op, args)
+      case op
+      when :IS, :'IS NOT'
+        isnot = op != :IS
+        return super if (v1 = args.at(1)).nil?
+        v0 = literal(args.at(0))
+        s = "(#{v0} #{'!' if isnot}= #{literal(v1)})"
+        s = "(#{s} OR (#{v0} IS NULL))" if isnot
+        s
+      else
+        super(op, args)
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/connection_pool.rb 3.0.0-1/lib/sequel/connection_pool.rb
--- 1.5.1-1/lib/sequel/connection_pool.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/connection_pool.rb	2009-05-01 23:06:32.000000000 +0100
@@ -0,0 +1,253 @@
+# A ConnectionPool manages access to database connections by keeping
+# multiple connections and giving threads exclusive access to each
+# connection.
+class Sequel::ConnectionPool
+  # The proc used to create a new database connection.
+  attr_accessor :connection_proc
+
+  # The proc used to disconnect a database connection.
+  attr_accessor :disconnection_proc
+
+  # The maximum number of connections.
+  attr_reader :max_size
+  
+  # The mutex that protects access to the other internal vairables.  You must use
+  # this if you want to manipulate the variables safely.
+  attr_reader :mutex
+  
+  # Constructs a new pool with a maximum size. If a block is supplied, it
+  # is used to create new connections as they are needed.
+  #
+  #   pool = ConnectionPool.new(:max_connections=>10) {MyConnection.new(opts)}
+  #
+  # The connection creation proc can be changed at any time by assigning a 
+  # Proc to pool#connection_proc.
+  #
+  #   pool = ConnectionPool.new(:max_connections=>10)
+  #   pool.connection_proc = proc {MyConnection.new(opts)}
+  #
+  # The connection pool takes the following options:
+  #
+  # * :disconnection_proc - The proc called when removing connections from the pool.
+  # * :max_connections - The maximum number of connections the connection pool
+  #   will open (default 4)
+  # * :pool_convert_exceptions - Whether to convert non-StandardError based exceptions
+  #   to RuntimeError exceptions (default true)
+  # * :pool_sleep_time - The amount of time to sleep before attempting to acquire
+  #   a connection again (default 0.001)
+  # * :pool_timeout - The amount of seconds to wait to acquire a connection
+  #   before raising a PoolTimeoutError (default 5)
+  # * :servers - A hash of servers to use.  Keys should be symbols.  If not
+  #   present, will use a single :default server.  The server name symbol will
+  #   be passed to the connection_proc.
+  def initialize(opts = {}, &block)
+    @max_size = opts[:max_connections] || 4
+    @mutex = Mutex.new
+    @connection_proc = block
+    @disconnection_proc = opts[:disconnection_proc]
+    @servers = [:default]
+    @servers += opts[:servers].keys - @servers if opts[:servers] 
+    @available_connections = Hash.new{|h,k| h[:default]}
+    @allocated = Hash.new{|h,k| h[:default]}
+    @servers.each do |s|
+      @available_connections[s] = []
+      @allocated[s] = {}
+    end
+    @timeout = opts[:pool_timeout] || 5
+    @sleep_time = opts[:pool_sleep_time] || 0.001
+    @convert_exceptions = opts.include?(:pool_convert_exceptions) ? opts[:pool_convert_exceptions] : true
+  end
+  
+  # A hash of connections currently being used for the given server, key is the
+  # Thread, value is the connection.
+  def allocated(server=:default)
+    @allocated[server]
+  end
+  
+  # An array of connections opened but not currently used, for the given
+  # server.
+  def available_connections(server=:default)
+    @available_connections[server]
+  end
+  
+  # The total number of connections opened for the given server, should
+  # be equal to available_connections.length + allocated.length
+  def created_count(server=:default)
+    @allocated[server].length + @available_connections[server].length
+  end
+  alias size created_count
+  
+  # Chooses the first available connection to the given server, or if none are
+  # available, creates a new connection.  Passes the connection to the supplied
+  # block:
+  # 
+  #   pool.hold {|conn| conn.execute('DROP TABLE posts')}
+  # 
+  # Pool#hold is re-entrant, meaning it can be called recursively in
+  # the same thread without blocking.
+  #
+  # If no connection is immediately available and the pool is already using the maximum
+  # number of connections, Pool#hold will block until a connection
+  # is available or the timeout expires.  If the timeout expires before a
+  # connection can be acquired, a Sequel::PoolTimeout is 
+  # raised.
+  def hold(server=:default)
+    begin
+      t = Thread.current
+      time = Time.new
+      timeout = time + @timeout
+      sleep_time = @sleep_time
+      if conn = owned_connection(t, server)
+        return yield(conn)
+      end
+      begin
+        until conn = acquire(t, server)
+          raise(::Sequel::PoolTimeout) if Time.new > timeout
+          sleep sleep_time
+        end
+        yield conn
+      rescue Sequel::DatabaseDisconnectError => dde
+        remove(t, conn, server) if conn
+        raise
+      ensure
+        @mutex.synchronize{release(t, server)} if conn && !dde
+      end
+    rescue StandardError => e
+      raise e
+    rescue Exception => e
+      raise(@convert_exceptions ? RuntimeError.new(e.message) : e)
+    end
+  end
+  
+  # Removes all connection currently available on all servers, optionally
+  # yielding each connection to the given block. This method has the effect of 
+  # disconnecting from the database, assuming that no connections are currently
+  # being used. Once a connection is requested using #hold, the connection pool
+  # creates new connections to the database.
+  def disconnect(&block)
+    block ||= @disconnection_proc
+    @mutex.synchronize do
+      @available_connections.each do |server, conns|
+        conns.each{|c| block.call(c)} if block
+        conns.clear
+      end
+    end
+  end
+  
+  private
+
+  # Assigns a connection to the supplied thread for the given server, if one
+  # is available.
+  def acquire(thread, server)
+    @mutex.synchronize do
+      if conn = available(server)
+        allocated(server)[thread] = conn
+      end
+    end
+  end
+  
+  # Returns an available connection to the given server. If no connection is
+  # available, tries to create a new connection.
+  def available(server)
+    available_connections(server).pop || make_new(server)
+  end
+  
+  # Creates a new connection to the given server if the size of the pool for
+  # the server is less than the maximum size of the pool.
+  def make_new(server)
+    if (n = created_count(server)) >= @max_size
+      allocated(server).keys.reject{|t| t.alive?}.each{|t| release(t, server)}
+      n = nil
+    end
+    if (n || created_count(server)) < @max_size
+      raise(Sequel::Error, "No connection proc specified") unless @connection_proc
+      begin
+        conn = @connection_proc.call(server)
+      rescue Exception=>exception
+        e = Sequel::DatabaseConnectionError.new("#{exception.class} #{exception.message}")
+        e.set_backtrace(exception.backtrace)
+        raise e
+      end
+      raise(Sequel::DatabaseConnectionError, "Connection parameters not valid") unless conn
+      conn
+    end
+  end
+  
+  # Returns the connection owned by the supplied thread for the given server,
+  # if any.
+  def owned_connection(thread, server)
+    @mutex.synchronize{@allocated[server][thread]}
+  end
+  
+  # Releases the connection assigned to the supplied thread and server.
+  # You must already have the mutex before you call this.
+  def release(thread, server)
+    available_connections(server) << allocated(server).delete(thread)
+  end
+
+  # Removes the currently allocated connection from the connection pool.
+  def remove(thread, conn, server)
+    @mutex.synchronize do
+      allocated(server).delete(thread)
+      @disconnection_proc.call(conn) if @disconnection_proc
+    end
+  end
+end
+
+# A SingleThreadedPool acts as a replacement for a ConnectionPool for use
+# in single-threaded applications. ConnectionPool imposes a substantial
+# performance penalty, so SingleThreadedPool is used to gain some speed.
+class Sequel::SingleThreadedPool
+  # The proc used to create a new database connection
+  attr_writer :connection_proc
+  
+  # The proc used to disconnect a database connection.
+  attr_accessor :disconnection_proc
+
+  # Initializes the instance with the supplied block as the connection_proc.
+  #
+  # The single threaded pool takes the following options:
+  #
+  # * :disconnection_proc - The proc called when removing connections from the pool.
+  # * :pool_convert_exceptions - Whether to convert non-StandardError based exceptions
+  #   to RuntimeError exceptions (default true)
+  # * :servers - A hash of servers to use.  Keys should be symbols.  If not
+  #   present, will use a single :default server.  The server name symbol will
+  #   be passed to the connection_proc.
+  def initialize(opts={}, &block)
+    @connection_proc = block
+    @disconnection_proc = opts[:disconnection_proc]
+    @conns = {}
+    @convert_exceptions = opts.include?(:pool_convert_exceptions) ? opts[:pool_convert_exceptions] : true
+  end
+  
+  # The connection for the given server.
+  def conn(server=:default)
+    @conns[server]
+  end
+  
+  # Yields the connection to the supplied block for the given server.
+  # This method simulates the ConnectionPool#hold API.
+  def hold(server=:default)
+    begin
+      begin
+        yield(c = (@conns[server] ||= @connection_proc.call(server)))
+      rescue Sequel::DatabaseDisconnectError => dde
+        @conns.delete(server)
+        @disconnection_proc.call(c) if @disconnection_proc
+        raise
+      end
+    rescue Exception => e
+      # if the error is not a StandardError it is converted into RuntimeError.
+      raise(@convert_exceptions && !e.is_a?(StandardError) ? RuntimeError.new(e.message) : e)
+    end
+  end
+  
+  # Disconnects from the database. Once a connection is requested using
+  # #hold, the connection is reestablished.
+  def disconnect(&block)
+    block ||= @disconnection_proc
+    @conns.values.each{|conn| block.call(conn) if block}
+    @conns = {}
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/core.rb 3.0.0-1/lib/sequel/core.rb
--- 1.5.1-1/lib/sequel/core.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/core.rb	2009-04-09 19:06:06.000000000 +0100
@@ -0,0 +1,202 @@
+%w'bigdecimal date thread time uri'.each{|f| require f}
+
+# Top level module for Sequel
+#
+# There are some class methods that are added via metaprogramming, one for
+# each supported adapter.  For example:
+#
+#   DB = Sequel.sqlite # Memory database
+#   DB = Sequel.sqlite('blog.db')
+#   DB = Sequel.postgres('database_name', :user=>'user', 
+#          :password=>'password', :host=>'host', :port=>5432, 
+#          :max_connections=>10)
+#
+# If a block is given to these methods, it is passed the opened Database
+# object, which is closed (disconnected) when the block exits, just
+# like a block passed to connect.  For example:
+#
+#   Sequel.sqlite('blog.db'){|db| puts db[:users].count}  
+#
+# Sequel converts the column type tinyint to a boolean by default,
+# you can override the conversion to use tinyint as an integer:
+#
+#   Sequel.convert_tinyint_to_bool = false
+#
+# Sequel converts two digit years in Dates and DateTimes by default,
+# so 01/02/03 is interpreted at January 2nd, 2003, and 12/13/99 is interpreted
+# as December 13, 1999. You can override this to treat those dates as
+# January 2nd, 0003 and December 13, 0099, respectively, by setting: 
+#
+#   Sequel.convert_two_digit_years = false
+#
+# Sequel can use either Time or DateTime for times returned from the
+# database.  It defaults to Time.  To change it to DateTime, use:
+#
+#   Sequel.datetime_class = DateTime
+#
+# You can set the SEQUEL_NO_CORE_EXTENSIONS constant or environment variable to have
+# Sequel not extend the core classes.
+module Sequel
+  @convert_tinyint_to_bool = true
+  @convert_two_digit_years = true
+  @datetime_class = Time
+  @virtual_row_instance_eval = true
+  
+  class << self
+    attr_accessor :convert_tinyint_to_bool, :convert_two_digit_years, :datetime_class, :virtual_row_instance_eval
+  end
+
+  # Returns true if the passed object could be a specifier of conditions, false otherwise.
+  # Currently, Sequel considers hashes and arrays of all two pairs as
+  # condition specifiers.
+  def self.condition_specifier?(obj)
+    case obj
+    when Hash
+      true
+    when Array
+      !obj.empty? && obj.all?{|i| (Array === i) && (i.length == 2)}
+    else
+      false
+    end
+  end
+
+  # Creates a new database object based on the supplied connection string
+  # and optional arguments.  The specified scheme determines the database
+  # class used, and the rest of the string specifies the connection options.
+  # For example:
+  #
+  #   DB = Sequel.connect('sqlite:/') # Memory database
+  #   DB = Sequel.connect('sqlite://blog.db') # ./blog.db
+  #   DB = Sequel.connect('sqlite:///blog.db') # /blog.db
+  #   DB = Sequel.connect('postgres://user:password@host:port/database_name')
+  #   DB = Sequel.connect('sqlite:///blog.db', :max_connections=>10)
+  #
+  # If a block is given, it is passed the opened Database object, which is
+  # closed when the block exits.  For example:
+  #
+  #   Sequel.connect('sqlite://blog.db'){|db| puts db[:users].count}  
+  def self.connect(*args, &block)
+    Database.connect(*args, &block)
+  end
+  
+  # Set the method to call on identifiers going into the database.  This affects
+  # the literalization of identifiers by calling this method on them before they are input.
+  # Sequel upcases identifiers in all SQL strings for most databases, so to turn that off:
+  #
+  #   Sequel.identifier_input_method = nil
+  # 
+  # to downcase instead:
+  #
+  #   Sequel.identifier_input_method = :downcase
+  #
+  # Other String instance methods work as well.
+  def self.identifier_input_method=(value)
+    Database.identifier_input_method = value
+  end
+  
+  # Set the method to call on identifiers coming out of the database.  This affects
+  # the literalization of identifiers by calling this method on them when they are
+  # retrieved from the database.  Sequel downcases identifiers retrieved for most
+  # databases, so to turn that off:
+  #
+  #   Sequel.identifier_output_method = nil
+  # 
+  # to upcase instead:
+  #
+  #   Sequel.identifier_output_method = :upcase
+  #
+  # Other String instance methods work as well.
+  def self.identifier_output_method=(value)
+    Database.identifier_output_method = value
+  end
+  
+  # Set whether to quote identifiers for all databases by default. By default,
+  # Sequel quotes identifiers in all SQL strings, so to turn that off:
+  #
+  #   Sequel.quote_identifiers = false
+  def self.quote_identifiers=(value)
+    Database.quote_identifiers = value
+  end
+
+  # Require all given files which should be in the same or a subdirectory of
+  # this file.  If a subdir is given, assume all files are in that subdir.
+  def self.require(files, subdir=nil)
+    Array(files).each{|f| super("#{File.dirname(__FILE__)}/#{"#{subdir}/" if subdir}#{f}")}
+  end
+  
+  # Set whether to set the single threaded mode for all databases by default. By default,
+  # Sequel uses a threadsafe connection pool, which isn't as fast as the
+  # single threaded connection pool.  If your program will only have one thread,
+  # and speed is a priority, you may want to set this to true:
+  #
+  #   Sequel.single_threaded = true
+  def self.single_threaded=(value)
+    Database.single_threaded = value
+  end
+
+  # Converts the given string into a Date object.
+  def self.string_to_date(s)
+    begin
+      Date.parse(s, Sequel.convert_two_digit_years)
+    rescue => e
+      raise InvalidValue, "Invalid Date value '#{self}' (#{e.message})"
+    end
+  end
+
+  # Converts the given string into a Time or DateTime object, depending on the
+  # value of Sequel.datetime_class.
+  def self.string_to_datetime(s)
+    begin
+      if datetime_class == DateTime
+        DateTime.parse(s, convert_two_digit_years)
+      else
+        datetime_class.parse(s)
+      end
+    rescue => e
+      raise InvalidValue, "Invalid #{datetime_class} value '#{self}' (#{e.message})"
+    end
+  end
+
+  # Converts the given string into a Time object.
+  def self.string_to_time(s)
+    begin
+      Time.parse(s)
+    rescue => e
+      raise InvalidValue, "Invalid Time value '#{self}' (#{e.message})"
+    end
+  end
+
+  ### Private Class Methods ###
+
+  # Helper method that the database adapter class methods that are added to Sequel via
+  # metaprogramming use to parse arguments.
+  def self.adapter_method(adapter, *args, &block) # :nodoc:
+    raise(::Sequel::Error, "Wrong number of arguments, 0-2 arguments valid") if args.length > 2
+    opts = {:adapter=>adapter.to_sym}
+    opts[:database] = args.shift if args.length >= 1 && !(args[0].is_a?(Hash))
+    if Hash === (arg = args[0])
+      opts.merge!(arg)
+    elsif !arg.nil?
+      raise ::Sequel::Error, "Wrong format of arguments, either use (), (String), (Hash), or (String, Hash)"
+    end
+    connect(opts, &block)
+  end
+
+  # Method that adds a database adapter class method to Sequel that calls
+  # Sequel.adapter_method.
+  def self.def_adapter_method(*adapters) # :nodoc:
+    adapters.each do |adapter|
+      instance_eval("def #{adapter}(*args, &block); adapter_method('#{adapter}', *args, &block) end")
+    end
+  end
+
+  private_class_method :adapter_method, :def_adapter_method
+  
+  require(%w"metaprogramming sql connection_pool exceptions dataset database version")
+  require(%w"schema_generator schema_methods schema_sql", 'database')
+  require(%w"convenience graph prepared_statements sql", 'dataset')
+  require('core_sql') if !defined?(::SEQUEL_NO_CORE_EXTENSIONS) && !ENV.has_key?('SEQUEL_NO_CORE_EXTENSIONS')
+
+  # Add the database adapter class methods to Sequel via metaprogramming
+  def_adapter_method(*Database::ADAPTERS)
+end
diff -pruN 1.5.1-1/lib/sequel/core_sql.rb 3.0.0-1/lib/sequel/core_sql.rb
--- 1.5.1-1/lib/sequel/core_sql.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/core_sql.rb	2009-03-19 16:37:47.000000000 +0000
@@ -0,0 +1,248 @@
+# Sequel extends the Array class to add methods to implement the SQL DSL.
+# Most of these methods require that the array not be empty and that it
+# must consist solely of other arrays that have exactly two elements.
+class Array
+  # Return a Sequel::SQL::BooleanExpression created from this array, not matching all of the
+  # conditions.
+  #
+  #   ~[[:a, true]] # SQL: a IS NOT TRUE
+  #   ~[[:a, 1], [:b, [2, 3]]] # SQL: a != 1 OR b NOT IN (2, 3)
+  def ~
+    sql_expr_if_all_two_pairs(:OR, true)
+  end
+
+  # True if the array is not empty and all of its elements are
+  # arrays of size 2, false otherwise.  This is used to determine if the array
+  # could be a specifier of conditions, used similarly to a hash
+  # but allowing for duplicate keys and a specific order.
+  #
+  #    [].to_a.all_two_pairs? # => false
+  #    [:a].to_a.all_two_pairs? # => false
+  #    [[:b]].to_a.all_two_pairs? # => false
+  #    [[:a, 1]].to_a.all_two_pairs? # => true
+  def all_two_pairs?
+    !empty? && all?{|i| (Array === i) && (i.length == 2)}
+  end
+
+  # Return a Sequel::SQL::CaseExpression with this array as the conditions and the given
+  # default value and expression.
+  #
+  #   [[{:a=>[2,3]}, 1]].case(0) # SQL: CASE WHEN a IN (2, 3) THEN 1 ELSE 0 END
+  #   [[:a, 1], [:b, 2]].case(:d, :c) # SQL: CASE c WHEN a THEN 1 WHEN b THEN 2 ELSE d END
+  def case(default, expression = nil)
+    ::Sequel::SQL::CaseExpression.new(self, default, expression)
+  end
+
+  # Return a Sequel::SQL::Array created from this array.  Used if this array contains
+  # all two pairs and you want it treated as an SQL array instead of a ordered hash-like
+  # conditions.
+  #
+  #   [[1, 2], [3, 4]] # SQL: 1 = 2 AND 3 = 4
+  #   [[1, 2], [3, 4]].sql_array # SQL: ((1, 2), (3, 4))
+  def sql_array
+    ::Sequel::SQL::SQLArray.new(self)
+  end
+
+  # Return a Sequel::SQL::BooleanExpression created from this array, matching all of the
+  # conditions.  Rarely do you need to call this explicitly, as Sequel generally
+  # assumes that arrays of all two pairs specify this type of condition.
+  #
+  #   [[:a, true]].sql_expr # SQL: a IS TRUE
+  #   [[:a, 1], [:b, [2, 3]]].sql_expr # SQL: a = 1 AND b IN (2, 3)
+  def sql_expr
+    sql_expr_if_all_two_pairs
+  end
+
+  # Return a Sequel::SQL::BooleanExpression created from this array, matching none
+  # of the conditions.
+  #
+  #   [[:a, true]].sql_negate # SQL: a IS NOT TRUE
+  #   [[:a, 1], [:b, [2, 3]]].sql_negate # SQL: a != 1 AND b NOT IN (2, 3)
+  def sql_negate
+    sql_expr_if_all_two_pairs(:AND, true)
+  end
+
+  # Return a Sequel::SQL::BooleanExpression created from this array, matching any of the
+  # conditions.
+  #
+  #   [[:a, true]].sql_or # SQL: a IS TRUE
+  #   [[:a, 1], [:b, [2, 3]]].sql_or # SQL: a = 1 OR b IN (2, 3)
+  def sql_or
+    sql_expr_if_all_two_pairs(:OR)
+  end
+
+  # Return a Sequel::SQL::BooleanExpression representing an SQL string made up of the
+  # concatenation of this array's elements.  If an argument is passed
+  # it is used in between each element of the array in the SQL
+  # concatenation.  This does not require that the array be made up of all two pairs.
+  #
+  #   [:a].sql_string_join # SQL: a
+  #   [:a, :b].sql_string_join # SQL: a || b
+  #   [:a, 'b'].sql_string_join # SQL: a || 'b'
+  #   ['a', :b].sql_string_join(' ') # SQL: 'a' || ' ' || b
+  def sql_string_join(joiner=nil)
+    if joiner
+      args = zip([joiner]*length).flatten
+      args.pop
+    else
+      args = self
+    end
+    args = args.collect{|a| [Symbol, ::Sequel::SQL::Expression, ::Sequel::LiteralString, TrueClass, FalseClass, NilClass].any?{|c| a.is_a?(c)} ? a : a.to_s}
+    ::Sequel::SQL::StringExpression.new(:'||', *args)
+  end
+
+  private
+
+  # Raise an error if this array is not made up of all two pairs, otherwise create a Sequel::SQL::BooleanExpression from this array.
+  def sql_expr_if_all_two_pairs(*args)
+    raise(Sequel::Error, 'Not all elements of the array are arrays of size 2, so it cannot be converted to an SQL expression') unless all_two_pairs?
+    ::Sequel::SQL::BooleanExpression.from_value_pairs(self, *args)
+  end
+end
+
+# Sequel extends the Hash class to add methods to implement the SQL DSL.
+class Hash
+  # Return a Sequel::SQL::BooleanExpression created from this hash, matching
+  # all of the conditions in this hash and the condition specified by
+  # the given argument.
+  #
+  #   {:a=>1} & :b # SQL: a = 1 AND b
+  #   {:a=>true} & ~:b # SQL: a IS TRUE AND NOT b
+  def &(ce)
+    ::Sequel::SQL::BooleanExpression.new(:AND, self, ce)
+  end
+
+  # Return a Sequel::SQL::BooleanExpression created from this hash, matching
+  # all of the conditions in this hash or the condition specified by
+  # the given argument.
+  #
+  #   {:a=>1} | :b # SQL: a = 1 OR b
+  #   {:a=>true} | ~:b # SQL: a IS TRUE OR NOT b
+  def |(ce)
+    ::Sequel::SQL::BooleanExpression.new(:OR, self, ce)
+  end
+
+  # Return a Sequel::SQL::BooleanExpression created from this hash, not matching all of the
+  # conditions.
+  #
+  #   ~{:a=>true} # SQL: a IS NOT TRUE
+  #   ~{:a=>1, :b=>[2, 3]} # SQL: a != 1 OR b NOT IN (2, 3)
+  def ~
+    ::Sequel::SQL::BooleanExpression.from_value_pairs(self, :OR, true)
+  end
+
+  # Return a Sequel::SQL::CaseExpression with this hash as the conditions and the given
+  # default value.  Note that the order of the conditions will be arbitrary, so all
+  # conditions should be orthogonal.
+  #
+  #   {{:a=>[2,3]}=>1}.case(0) # SQL: CASE WHEN a IN (2, 3) THEN 1 ELSE 0 END
+  #   {:a=>1, {:b=>2}].case(:d, :c) # SQL: CASE c WHEN a THEN 1 WHEN b THEN 2 ELSE d END
+  #                                 #  or: CASE c WHEN b THEN 2 WHEN a THEN 1 ELSE d END
+  def case(default, expression = nil)
+    ::Sequel::SQL::CaseExpression.new(to_a, default, expression)
+  end
+
+  # Return a Sequel::SQL::BooleanExpression created from this hash, matching all of the
+  # conditions.  Rarely do you need to call this explicitly, as Sequel generally
+  # assumes that hashes specify this type of condition.
+  #
+  #   {:a=>true}.sql_expr # SQL: a IS TRUE
+  #   {:a=>1, :b=>[2, 3]}.sql_expr # SQL: a = 1 AND b IN (2, 3)
+  def sql_expr
+    ::Sequel::SQL::BooleanExpression.from_value_pairs(self)
+  end
+
+  # Return a Sequel::SQL::BooleanExpression created from this hash, matching none
+  # of the conditions.
+  #
+  #   {:a=>true}.sql_negate # SQL: a IS NOT TRUE
+  #   {:a=>1, :b=>[2, 3]}.sql_negate # SQL: a != 1 AND b NOT IN (2, 3)
+  def sql_negate
+    ::Sequel::SQL::BooleanExpression.from_value_pairs(self, :AND, true)
+  end
+
+  # Return a Sequel::SQL::BooleanExpression created from this hash, matching any of the
+  # conditions.
+  #
+  #   {:a=>true}.sql_or # SQL: a IS TRUE
+  #   {:a=>1, :b=>[2, 3]}.sql_or # SQL: a = 1 OR b IN (2, 3)
+  def sql_or
+    ::Sequel::SQL::BooleanExpression.from_value_pairs(self, :OR)
+  end
+end
+
+# Sequel extends the String class to add methods to implement the SQL DSL.
+class String
+  include Sequel::SQL::AliasMethods
+  include Sequel::SQL::CastMethods
+
+  # Converts a string into a Sequel::LiteralString, in order to override string
+  # literalization, e.g.:
+  #
+  #   DB[:items].filter(:abc => 'def').sql #=>
+  #     "SELECT * FROM items WHERE (abc = 'def')"
+  #
+  #   DB[:items].filter(:abc => 'def'.lit).sql #=>
+  #     "SELECT * FROM items WHERE (abc = def)"
+  #
+  # You can also provide arguments, to create a Sequel::SQL::PlaceholderLiteralString:
+  #
+  #    DB[:items].select{|o| o.count('DISTINCT ?'.lit(:a))}.sql #=>
+  #      "SELECT count(DISTINCT a) FROM items"
+  def lit(*args)
+    args.empty? ? Sequel::LiteralString.new(self) : Sequel::SQL::PlaceholderLiteralString.new(self, args)
+  end
+  
+  # Returns a Sequel::SQL::Blob that holds the same data as this string. Blobs provide proper
+  # escaping of binary data.
+  def to_sequel_blob
+    ::Sequel::SQL::Blob.new(self)
+  end
+end
+
+# Sequel extends the Symbol class to add methods to implement the SQL DSL.
+class Symbol
+  include Sequel::SQL::QualifyingMethods
+  include Sequel::SQL::IdentifierMethods
+  include Sequel::SQL::AliasMethods
+  include Sequel::SQL::CastMethods
+  include Sequel::SQL::OrderMethods
+  include Sequel::SQL::BooleanMethods
+  include Sequel::SQL::NumericMethods
+  include Sequel::SQL::StringMethods
+  include Sequel::SQL::ComplexExpressionMethods
+  include Sequel::SQL::InequalityMethods if RUBY_VERSION < '1.9.0'
+
+  # If no argument is given, returns a Sequel::SQL::ColumnAll object specifying all
+  # columns for this table.
+  # If an argument is given, returns a Sequel::SQL::NumericExpression using the *
+  # (multiplication) operator with this and the given argument.
+  # 
+  #   :table.* # SQL: table.*
+  #   :column * 2 # SQL: column * 2
+  def *(ce=(arg=false;nil))
+    return super(ce) unless arg == false
+    Sequel::SQL::ColumnAll.new(self);
+  end
+
+  # Returns a Sequel::SQL::Function  with this as the function name,
+  # and the given arguments. This is aliased as Symbol#[] if the RUBY_VERSION
+  # is less than 1.9.0. Ruby 1.9 defines Symbol#[], and Sequel
+  # doesn't override methods defined by ruby itself.
+  #
+  #   :now.sql_function # SQL: now()
+  #   :sum.sql_function(:a) # SQL: sum(a)
+  #   :concat.sql_function(:a, :b) # SQL: concat(a, b)
+  def sql_function(*args)
+    Sequel::SQL::Function.new(self, *args)
+  end
+  alias_method(:[], :sql_function) if RUBY_VERSION < '1.9.0'
+
+  # Return an SQL array subscript with the given arguments.
+  #
+  #   :array.sql_subscript(1) # SQL: array[1]
+  #   :array.sql_subscript(1, 2) # SQL: array[1, 2]
+  def sql_subscript(*sub)
+    Sequel::SQL::Subscript.new(self, sub.flatten)
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/database/schema_generator.rb 3.0.0-1/lib/sequel/database/schema_generator.rb
--- 1.5.1-1/lib/sequel/database/schema_generator.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/database/schema_generator.rb	2009-04-27 02:57:30.000000000 +0100
@@ -0,0 +1,319 @@
+module Sequel
+  # The Schema module holds the schema generators and the SQL code relating
+  # to SQL DDL (Data Definition Language).
+  module Schema
+    # Schema::Generator is an internal class that the user is not expected
+    # to instantiate directly.  Instances are created by Database#create_table.
+    # It is used to specify table creation parameters.  It takes a Database
+    # object and a block of column/index/constraint specifications, and
+    # gives the Database a table description, which the database uses to
+    # create a table.
+    #
+    # Schema::Generator has some methods but also includes method_missing,
+    # allowing users to specify column type as a method instead of using
+    # the column method, which makes for a nicer DSL.
+    class Generator
+      # Classes specifying generic types that Sequel will convert to database-specific types.
+      GENERIC_TYPES=[String, Integer, Fixnum, Bignum, Float, Numeric, BigDecimal,
+      Date, DateTime, Time, File, TrueClass, FalseClass]
+      
+      # Return the columns created by this generator
+      attr_reader :columns
+
+      # Return the constraints created by this generator
+      attr_reader :constraints
+
+      # Return the indexes created by this generator
+      attr_reader :indexes
+
+      # Set the database in which to create the table, and evaluate the block
+      # in the context of this object.
+      def initialize(db, &block)
+        @db = db
+        @columns = []
+        @indexes = []
+        @constraints = []
+        @primary_key = nil
+        instance_eval(&block) if block
+        @columns.unshift(@primary_key) if @primary_key && !has_column?(primary_key_name)
+      end
+      
+      # Add a method for each of the given types that creates a column
+      # with that type as a constant.  Types given should either already
+      # be constants/classes or a capitalized string/symbol with the same name
+      # as a constant/class.
+      def self.add_type_method(*types)
+        types.each do |type|
+          class_eval("def #{type}(name, opts={}); column(name, #{type}, opts); end", __FILE__, __LINE__)
+        end
+      end
+      
+      # Add a unnamed constraint to the DDL, specified by the given block
+      # or args.
+      def check(*args, &block)
+        constraint(nil, *args, &block)
+      end
+
+      # Add a column with the given name, type, and opts to the DDL. 
+      #
+      # You can also create columns via method missing, so the following are
+      # equivalent:
+      #
+      #   column :number, :integer
+      #   integer :number
+      #
+      # The following options are supported:
+      #
+      # * :default - The default value for the column.
+      # * :index - Create an index on this column.
+      # * :key - For foreign key columns, the column in the associated table
+      #   that this column references.  Unnecessary if this column
+      #   references the primary key of the associated table.
+      # * :null - Mark the column as allowing NULL values (if true),
+      #   or not allowing NULL values (if false).  If unspecified, will default
+      #   to whatever the database default is.
+      # * :on_delete - Specify the behavior of this column when being deleted.
+      #   See Dataset#on_delete_clause for options.
+      # * :on_update - Specify the behavior of this column when being updated.
+      #   See Schema::SQL#on_delete_clause for options.
+      # * :size - The size of the column, generally used with string
+      #   columns to specify the maximum number of characters the column will hold.
+      # * :unique - Mark the column as unique, generally has the same effect as
+      #   creating a unique index on the column.
+      # * :unsigned - Make the column type unsigned, only useful for integer
+      #   columns.
+      def column(name, type, opts = {})
+        columns << {:name => name, :type => type}.merge(opts)
+        index(name) if opts[:index]
+      end
+      
+      # Adds a named constraint (or unnamed if name is nil) to the DDL,
+      # with the given block or args.
+      def constraint(name, *args, &block)
+        constraints << {:name => name, :type => :check, :check => block || args}
+      end
+      
+      # Add a foreign key in the table that references another table to the DDL. See column
+      # for available options.
+      def foreign_key(name, table=nil, opts = {})
+        opts = case table
+        when Hash
+          table.merge(opts)
+        when Symbol
+          opts.merge(:table=>table)
+        when NilClass
+          opts
+        else
+          raise(Error, "The second argument to foreign_key should be a Hash, Symbol, or nil")
+        end
+        return composite_foreign_key(name, opts) if name.is_a?(Array)
+        column(name, Integer, opts)
+      end
+
+      # Add a full text index on the given columns to the DDL.
+      def full_text_index(columns, opts = {})
+        index(columns, opts.merge(:type => :full_text))
+      end
+      
+      # True if the DDL includes the creation of a column with the given name.
+      def has_column?(name)
+        columns.any?{|c| c[:name] == name}
+      end
+      
+      # Add an index on the given column(s) with the given options to the DDL.
+      # The available options are:
+      #
+      # * :type - The type of index to use (only supported by some databases)
+      # * :unique - Make the index unique, so duplicate values are not allowed.
+      # * :where - Create a partial index (only supported by some databases)
+      def index(columns, opts = {})
+        indexes << {:columns => Array(columns)}.merge(opts)
+      end
+      
+      # Add a column with the given type, name, and opts to the DDL.  See column for available
+      # options.
+      def method_missing(type, name = nil, opts = {})
+        name ? column(name, type, opts) : super
+      end
+      
+      # Add primary key information to the DDL. Takes between one and three
+      # arguments. The last one is an options hash as for Generator#column.
+      # The first one distinguishes two modes: an array of existing column
+      # names adds a composite primary key constraint. A single symbol adds a
+      # new column of that name and makes it the primary key. In that case the
+      # optional middle argument denotes the type.
+      # 
+      # Examples:
+      #   primary_key(:id)
+      #   primary_key(:zip_code, :null => false)
+      #   primary_key([:street_number, :house_number])
+      #   primary_key(:id, :string, :auto_increment => false)
+      def primary_key(name, *args)
+        return composite_primary_key(name, *args) if name.is_a?(Array)
+        @primary_key = @db.serial_primary_key_options.merge({:name => name})
+        
+        if opts = args.pop
+          opts = {:type => opts} unless opts.is_a?(Hash)
+          if type = args.pop
+            opts.merge!(:type => type)
+          end
+          @primary_key.merge!(opts)
+        end
+        @primary_key
+      end
+
+      # The name of the primary key for this table, if it has a primary key.
+      def primary_key_name
+        @primary_key[:name] if @primary_key
+      end
+      
+      # Add a spatial index on the given columns to the DDL.
+      def spatial_index(columns, opts = {})
+        index(columns, opts.merge(:type => :spatial))
+      end
+
+      # Add a unique constraint on the given columns to the DDL.
+      def unique(columns, opts = {})
+        constraints << {:type => :unique, :columns => Array(columns)}.merge(opts)
+      end
+
+      private
+
+      # Add a composite primary key constraint
+      def composite_primary_key(columns, *args)
+        opts = args.pop || {}
+        constraints << {:type => :primary_key, :columns => columns}.merge(opts)
+      end
+
+      # Add a composite foreign key constraint
+      def composite_foreign_key(columns, opts)
+        constraints << {:type => :foreign_key, :columns => columns}.merge(opts)
+      end
+      
+      add_type_method(*GENERIC_TYPES)
+    end
+  
+    # Schema::AlterTableGenerator is an internal class that the user is not expected
+    # to instantiate directly.  Instances are created by Database#alter_table.
+    # It is used to specify table alteration parameters.  It takes a Database
+    # object and a block of operations to perform on the table, and
+    # gives the Database a table an array of operations, which the database uses to
+    # alter a table's description.
+    class AlterTableGenerator
+      # An array of DDL operations to perform
+      attr_reader :operations
+      
+      # Set the Database object to which to apply the DDL, and evaluate the
+      # block in the context of this object.
+      def initialize(db, &block)
+        @db = db
+        @operations = []
+        instance_eval(&block) if block
+      end
+      
+      # Add a column with the given name, type, and opts to the DDL for the table.
+      # See Generator#column for the available options.
+      def add_column(name, type, opts = {})
+        @operations << {:op => :add_column, :name => name, :type => type}.merge(opts)
+      end
+      
+      # Add a constraint with the given name and args to the DDL for the table.
+      # See Generator#constraint.
+      def add_constraint(name, *args, &block)
+        @operations << {:op => :add_constraint, :name => name, :type => :check, :check => block || args}
+      end
+
+      def add_unique_constraint(columns, opts = {})
+        @operations << {:op => :add_constraint, :type => :unique, :columns => Array(columns)}.merge(opts)
+      end
+
+      # Add a foreign key with the given name and referencing the given table
+      # to the DDL for the table.  See Generator#column for the available options.
+      #
+      # You can also pass an array of column names for creating composite foreign
+      # keys. In this case, it will assume the columns exists and will only add
+      # the constraint.
+      #
+      # NOTE: If you need to add a foreign key constraint to an existing column
+      # use the composite key syntax even if it is only one column.
+      def add_foreign_key(name, table, opts = {})
+        return add_composite_foreign_key(name, table, opts) if name.is_a?(Array)
+        add_column(name, Integer, {:table=>table}.merge(opts))
+      end
+      
+      # Add a full text index on the given columns to the DDL for the table.
+      # See Generator#index for available options.
+      def add_full_text_index(columns, opts = {})
+        add_index(columns, {:type=>:full_text}.merge(opts))
+      end
+      
+      # Add an index on the given columns to the DDL for the table.  See
+      # Generator#index for available options.
+      def add_index(columns, opts = {})
+        @operations << {:op => :add_index, :columns => Array(columns)}.merge(opts)
+      end
+      
+      # Add a primary key to the DDL for the table.  See Generator#column
+      # for the available options.
+      def add_primary_key(name, opts = {})
+        return add_composite_primary_key(name, opts) if name.is_a?(Array)
+        opts = @db.serial_primary_key_options.merge(opts)
+        add_column(name, opts.delete(:type), opts)
+      end
+      
+      # Add a spatial index on the given columns to the DDL for the table.
+      # See Generator#index for available options.
+      def add_spatial_index(columns, opts = {})
+        add_index(columns, {:type=>:spatial}.merge(opts))
+      end
+      
+      # Remove a column from the DDL for the table.
+      def drop_column(name)
+        @operations << {:op => :drop_column, :name => name}
+      end
+      
+      # Remove a constraint from the DDL for the table.
+      def drop_constraint(name)
+        @operations << {:op => :drop_constraint, :name => name}
+      end
+      
+      # Remove an index from the DDL for the table.
+      def drop_index(columns, options={})
+        @operations << {:op => :drop_index, :columns => Array(columns)}.merge(options)
+      end
+
+      # Modify a column's name in the DDL for the table.
+      def rename_column(name, new_name, opts = {})
+        @operations << {:op => :rename_column, :name => name, :new_name => new_name}.merge(opts)
+      end
+      
+      # Modify a column's default value in the DDL for the table.
+      def set_column_default(name, default)
+        @operations << {:op => :set_column_default, :name => name, :default => default}
+      end
+
+      # Modify a column's type in the DDL for the table.
+      def set_column_type(name, type, opts={})
+        @operations << {:op => :set_column_type, :name => name, :type => type}.merge(opts)
+      end
+      
+      # Modify a column's NOT NULL constraint.
+      def set_column_allow_null(name, allow_null)
+        @operations << {:op => :set_column_null, :name => name, :null => allow_null}
+      end
+
+      private
+
+      # Add a composite primary key constraint
+      def add_composite_primary_key(columns, opts)
+        @operations << {:op => :add_constraint, :type => :primary_key, :columns => columns}.merge(opts)
+      end
+
+      # Add a composite foreign key constraint
+      def add_composite_foreign_key(columns, table, opts)
+        @operations << {:op => :add_constraint, :type => :foreign_key, :columns => columns, :table => table}.merge(opts)
+      end
+    end
+  end
+end
+
diff -pruN 1.5.1-1/lib/sequel/database/schema_methods.rb 3.0.0-1/lib/sequel/database/schema_methods.rb
--- 1.5.1-1/lib/sequel/database/schema_methods.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/database/schema_methods.rb	2009-04-27 02:56:12.000000000 +0100
@@ -0,0 +1,169 @@
+module Sequel
+  class Database
+    # Adds a column to the specified table. This method expects a column name,
+    # a datatype and optionally a hash with additional constraints and options:
+    #
+    #   DB.add_column :items, :name, :text, :unique => true, :null => false
+    #   DB.add_column :items, :category, :text, :default => 'ruby'
+    #
+    # See alter_table.
+    def add_column(table, *args)
+      alter_table(table) {add_column(*args)}
+    end
+    
+    # Adds an index to a table for the given columns:
+    # 
+    #   DB.add_index :posts, :title
+    #   DB.add_index :posts, [:author, :title], :unique => true
+    #
+    # See alter_table.
+    def add_index(table, *args)
+      alter_table(table) {add_index(*args)}
+    end
+    
+    # Alters the given table with the specified block. Example:
+    #
+    #   DB.alter_table :items do
+    #     add_column :category, :text, :default => 'ruby'
+    #     drop_column :category
+    #     rename_column :cntr, :counter
+    #     set_column_type :value, :float
+    #     set_column_default :value, :float
+    #     add_index [:group, :category]
+    #     drop_index [:group, :category]
+    #   end
+    #
+    # Note that #add_column accepts all the options available for column
+    # definitions using create_table, and #add_index accepts all the options
+    # available for index definition.
+    #
+    # See Schema::AlterTableGenerator.
+    def alter_table(name, generator=nil, &block)
+      remove_cached_schema(name)
+      generator ||= Schema::AlterTableGenerator.new(self, &block)
+      alter_table_sql_list(name, generator.operations).flatten.each {|sql| execute_ddl(sql)}
+    end
+    
+    # Creates a table with the columns given in the provided block:
+    #
+    #   DB.create_table :posts do
+    #     primary_key :id
+    #     column :title, :text
+    #     String :content
+    #     index :title
+    #   end
+    #
+    # Options:
+    # * :temp - Create the table as a temporary table.
+    #
+    # See Schema::Generator.
+    def create_table(name, options={}, &block)
+      options = {:generator=>options} if options.is_a?(Schema::Generator)
+      generator = options[:generator] || Schema::Generator.new(self, &block)
+      execute_ddl(create_table_sql(name, generator, options))
+      index_sql_list(name, generator.indexes).each{|sql| execute_ddl(sql)}
+    end
+    
+    # Forcibly creates a table, attempting to drop it unconditionally (and catching any errors), then creating it.
+    def create_table!(name, options={}, &block)
+      drop_table(name) rescue nil
+      create_table(name, options, &block)
+    end
+    
+    # Creates a view, replacing it if it already exists:
+    #
+    #   DB.create_or_replace_view(:cheap_items, "SELECT * FROM items WHERE price < 100")
+    #   DB.create_or_replace_view(:ruby_items, DB[:items].filter(:category => 'ruby'))
+    def create_or_replace_view(name, source)
+      remove_cached_schema(name)
+      source = source.sql if source.is_a?(Dataset)
+      execute_ddl("CREATE OR REPLACE VIEW #{quote_schema_table(name)} AS #{source}")
+    end
+    
+    # Creates a view based on a dataset or an SQL string:
+    #
+    #   DB.create_view(:cheap_items, "SELECT * FROM items WHERE price < 100")
+    #   DB.create_view(:ruby_items, DB[:items].filter(:category => 'ruby'))
+    def create_view(name, source)
+      source = source.sql if source.is_a?(Dataset)
+      execute_ddl("CREATE VIEW #{quote_schema_table(name)} AS #{source}")
+    end
+    
+    # Removes a column from the specified table:
+    #
+    #   DB.drop_column :items, :category
+    #
+    # See alter_table.
+    def drop_column(table, *args)
+      alter_table(table) {drop_column(*args)}
+    end
+    
+    # Removes an index for the given table and column/s:
+    #
+    #   DB.drop_index :posts, :title
+    #   DB.drop_index :posts, [:author, :title]
+    #
+    # See alter_table.
+    def drop_index(table, columns, options={})
+      alter_table(table){drop_index(columns, options)}
+    end
+    
+    # Drops one or more tables corresponding to the given names:
+    #
+    #   DB.drop_table(:posts, :comments)
+    def drop_table(*names)
+      names.each do |n|
+        remove_cached_schema(n)
+        execute_ddl(drop_table_sql(n))
+      end
+    end
+    
+    # Drops one or more views corresponding to the given names:
+    #
+    #   DB.drop_view(:cheap_items)
+    def drop_view(*names)
+      names.each do |n|
+        remove_cached_schema(n)
+        execute_ddl("DROP VIEW #{quote_schema_table(n)}")
+      end
+    end
+
+    # Renames a table:
+    #
+    #   DB.tables #=> [:items]
+    #   DB.rename_table :items, :old_items
+    #   DB.tables #=> [:old_items]
+    def rename_table(name, new_name)
+      remove_cached_schema(name)
+      execute_ddl(rename_table_sql(name, new_name))
+    end
+    
+    # Renames a column in the specified table. This method expects the current
+    # column name and the new column name:
+    #
+    #   DB.rename_column :items, :cntr, :counter
+    #
+    # See alter_table.
+    def rename_column(table, *args)
+      alter_table(table) {rename_column(*args)}
+    end
+    
+    # Sets the default value for the given column in the given table:
+    #
+    #   DB.set_column_default :items, :category, 'perl!'
+    #
+    # See alter_table.
+    def set_column_default(table, *args)
+      alter_table(table) {set_column_default(*args)}
+    end
+    
+    # Set the data type for the given column in the given table:
+    #
+    #   DB.set_column_type :items, :price, :float
+    #
+    # See alter_table.
+    def set_column_type(table, *args)
+      alter_table(table) {set_column_type(*args)}
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/database/schema_sql.rb 3.0.0-1/lib/sequel/database/schema_sql.rb
--- 1.5.1-1/lib/sequel/database/schema_sql.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/database/schema_sql.rb	2009-04-21 21:11:03.000000000 +0100
@@ -0,0 +1,310 @@
+module Sequel
+  class Database
+    AUTOINCREMENT = 'AUTOINCREMENT'.freeze
+    CASCADE = 'CASCADE'.freeze
+    COMMA_SEPARATOR = ', '.freeze
+    NO_ACTION = 'NO ACTION'.freeze
+    NOT_NULL = ' NOT NULL'.freeze
+    NULL = ' NULL'.freeze
+    PRIMARY_KEY = ' PRIMARY KEY'.freeze
+    RESTRICT = 'RESTRICT'.freeze
+    SET_DEFAULT = 'SET DEFAULT'.freeze
+    SET_NULL = 'SET NULL'.freeze
+    TEMPORARY = 'TEMPORARY '.freeze
+    UNDERSCORE = '_'.freeze
+    UNIQUE = ' UNIQUE'.freeze
+    UNSIGNED = ' UNSIGNED'.freeze
+
+    # Default serial primary key options.
+    def serial_primary_key_options
+      {:primary_key => true, :type => Integer, :auto_increment => true}
+    end
+
+    private
+
+    # The SQL to execute to modify the DDL for the given table name.  op
+    # should be one of the operations returned by the AlterTableGenerator.
+    def alter_table_sql(table, op)
+      quoted_name = quote_identifier(op[:name]) if op[:name]
+      alter_table_op = case op[:op]
+      when :add_column
+        "ADD COLUMN #{column_definition_sql(op)}"
+      when :drop_column
+        "DROP COLUMN #{quoted_name}"
+      when :rename_column
+        "RENAME COLUMN #{quoted_name} TO #{quote_identifier(op[:new_name])}"
+      when :set_column_type
+        "ALTER COLUMN #{quoted_name} TYPE #{type_literal(op)}"
+      when :set_column_default
+        "ALTER COLUMN #{quoted_name} SET DEFAULT #{literal(op[:default])}"
+      when :set_column_null
+        "ALTER COLUMN #{quoted_name} #{op[:null] ? 'DROP' : 'SET'} NOT NULL"
+      when :add_index
+        return index_definition_sql(table, op)
+      when :drop_index
+        return drop_index_sql(table, op)
+      when :add_constraint
+        "ADD #{constraint_definition_sql(op)}"
+      when :drop_constraint
+        "DROP CONSTRAINT #{quoted_name}"
+      else
+        raise Error, "Unsupported ALTER TABLE operation"
+      end
+      "ALTER TABLE #{quote_schema_table(table)} #{alter_table_op}"
+    end
+
+    # Array of SQL DDL modification statements for the given table,
+    # corresponding to the DDL changes specified by the operations.
+    def alter_table_sql_list(table, operations)
+      operations.map{|op| alter_table_sql(table, op)}
+    end
+    
+    # The SQL string specify the autoincrement property, generally used by
+    # primary keys.
+    def auto_increment_sql
+      AUTOINCREMENT
+    end
+    
+    # SQL DDL fragment containing the column creation SQL for the given column.
+    def column_definition_sql(column)
+      sql = "#{quote_identifier(column[:name])} #{type_literal(column)}"
+      sql << UNIQUE if column[:unique]
+      sql << NOT_NULL if column[:null] == false
+      sql << NULL if column[:null] == true
+      sql << " DEFAULT #{literal(column[:default])}" if column.include?(:default)
+      sql << PRIMARY_KEY if column[:primary_key]
+      sql << " #{auto_increment_sql}" if column[:auto_increment]
+      sql << column_references_sql(column) if column[:table]
+      sql
+    end
+    
+    # SQL DDL fragment containing the column creation
+    # SQL for all given columns, used inside a CREATE TABLE block.
+    def column_list_sql(generator)
+      (generator.columns.map{|c| column_definition_sql(c)} + generator.constraints.map{|c| constraint_definition_sql(c)}).join(COMMA_SEPARATOR)
+    end
+
+    # SQL DDL fragment for column foreign key references
+    def column_references_sql(column)
+      sql = " REFERENCES #{quote_schema_table(column[:table])}"
+      sql << "(#{Array(column[:key]).map{|x| quote_identifier(x)}.join(COMMA_SEPARATOR)})" if column[:key]
+      sql << " ON DELETE #{on_delete_clause(column[:on_delete])}" if column[:on_delete]
+      sql << " ON UPDATE #{on_delete_clause(column[:on_update])}" if column[:on_update]
+      sql
+    end
+  
+    # SQL DDL fragment specifying a constraint on a table.
+    def constraint_definition_sql(constraint)
+      sql = constraint[:name] ? "CONSTRAINT #{quote_identifier(constraint[:name])} " : ""
+      case constraint[:type]
+      when :check
+        check = constraint[:check]
+        sql << "CHECK #{filter_expr((check.is_a?(Array) && check.length == 1) ? check.first : check)}"
+      when :primary_key
+        sql << "PRIMARY KEY #{literal(constraint[:columns])}"
+      when :foreign_key
+        sql << "FOREIGN KEY #{literal(constraint[:columns])}"
+        sql << column_references_sql(constraint)
+      when :unique
+        sql << "UNIQUE #{literal(constraint[:columns])}"
+      else
+        raise Error, "Invalid constriant type #{constraint[:type]}, should be :check, :primary_key, :foreign_key, or :unique"
+      end
+      sql
+    end
+
+    # DDL statement for creating a table with the given name, columns, and options
+    def create_table_sql(name, generator, options)
+      "CREATE #{temporary_table_sql if options[:temp]}TABLE #{quote_schema_table(name)} (#{column_list_sql(generator)})"
+    end
+
+    # Default index name for the table and columns, may be too long
+    # for certain databases.
+    def default_index_name(table_name, columns)
+      schema, table = schema_and_table(table_name)
+      "#{"#{schema}_" if schema and schema != default_schema}#{table}_#{columns.map{|c| [String, Symbol].any?{|cl| c.is_a?(cl)} ? c : literal(c).gsub(/\W/, '_')}.join(UNDERSCORE)}_index"
+    end
+  
+    # The SQL to drop an index for the table.
+    def drop_index_sql(table, op)
+      "DROP INDEX #{quote_identifier(op[:name] || default_index_name(table, op[:columns]))}"
+    end
+
+    # SQL DDL statement to drop the table with the given name.
+    def drop_table_sql(name)
+      "DROP TABLE #{quote_schema_table(name)}"
+    end
+    
+    # Proxy the filter_expr call to the dataset, used for creating constraints.
+    def filter_expr(*args, &block)
+      schema_utility_dataset.literal(schema_utility_dataset.send(:filter_expr, *args, &block))
+    end
+
+    # SQL DDL statement for creating an index for the table with the given name
+    # and index specifications.
+    def index_definition_sql(table_name, index)
+      index_name = index[:name] || default_index_name(table_name, index[:columns])
+      if index[:type]
+        raise Error, "Index types are not supported for this database"
+      elsif index[:where]
+        raise Error, "Partial indexes are not supported for this database"
+      else
+        "CREATE #{'UNIQUE ' if index[:unique]}INDEX #{quote_identifier(index_name)} ON #{quote_identifier(table_name)} #{literal(index[:columns])}"
+      end
+    end
+  
+    # Array of SQL DDL statements, one for each index specification,
+    # for the given table.
+    def index_sql_list(table_name, indexes)
+      indexes.map{|i| index_definition_sql(table_name, i)}
+    end
+
+    # SQL DDL ON DELETE fragment to use, based on the given action.
+    # The following actions are recognized:
+    # 
+    # * :cascade - Delete rows referencing this row.
+    # * :no_action (default) - Raise an error if other rows reference this
+    #   row, allow deferring of the integrity check.
+    # * :restrict - Raise an error if other rows reference this row,
+    #   but do not allow deferring the integrity check.
+    # * :set_default - Set columns referencing this row to their default value.
+    # * :set_null - Set columns referencing this row to NULL.
+    def on_delete_clause(action)
+      case action
+      when :restrict
+        RESTRICT
+      when :cascade
+        CASCADE
+      when :set_null
+        SET_NULL
+      when :set_default
+        SET_DEFAULT
+      else
+        NO_ACTION
+      end
+    end
+    
+    # Proxy the quote_schema_table method to the dataset
+    def quote_schema_table(table)
+      schema_utility_dataset.quote_schema_table(table)
+    end
+    
+    # Proxy the quote_identifier method to the dataset, used for quoting tables and columns.
+    def quote_identifier(v)
+      schema_utility_dataset.quote_identifier(v)
+    end
+    
+    # SQL DDL statement for renaming a table.
+    def rename_table_sql(name, new_name)
+      "ALTER TABLE #{quote_schema_table(name)} RENAME TO #{quote_schema_table(new_name)}"
+    end
+
+    # SQL DDL fragment for temporary table
+    def temporary_table_sql
+      self.class.const_get(:TEMPORARY)
+    end
+
+    # SQL fragment specifying the type of a given column.
+    def type_literal(column)
+      column[:type].is_a?(Class) ? type_literal_generic(column) : type_literal_specific(column)
+    end
+    
+    # SQL fragment specifying the full type of a column,
+    # consider the type with possible modifiers.
+    def type_literal_generic(column)
+      meth = "type_literal_generic_#{column[:type].name.to_s.downcase}"
+      if respond_to?(meth, true)
+        send(meth, column)
+      else
+        raise Error, "Unsupported ruby class used as database type: #{column[:type]}"
+      end
+    end
+
+    # Alias for type_literal_generic_numeric, to make overriding in a subclass easier.
+    def type_literal_generic_bigdecimal(column)
+      type_literal_generic_numeric(column)
+    end
+
+    # Sequel uses the bigint type by default for Bignums.
+    def type_literal_generic_bignum(column)
+      :bigint
+    end
+
+    # Sequel uses the date type by default for Dates.
+    def type_literal_generic_date(column)
+      :date
+    end
+
+    # Sequel uses the timestamp type by default for DateTimes.
+    def type_literal_generic_datetime(column)
+      :timestamp
+    end
+
+    # Alias for type_literal_generic_trueclass, to make overriding in a subclass easier.
+    def type_literal_generic_falseclass(column)
+      type_literal_generic_trueclass(column)
+    end
+
+    # Sequel uses the blob type by default for Files.
+    def type_literal_generic_file(column)
+      :blob
+    end
+
+    # Alias for type_literal_generic_integer, to make overriding in a subclass easier.
+    def type_literal_generic_fixnum(column)
+      type_literal_generic_integer(column)
+    end
+
+    # Sequel uses the double precision type by default for Floats.
+    def type_literal_generic_float(column)
+      :"double precision"
+    end
+
+    # Sequel uses the integer type by default for integers
+    def type_literal_generic_integer(column)
+      :integer
+    end
+
+    # Sequel uses the numeric type by default for Numerics and BigDecimals.
+    # If a size is given, it is used, otherwise, it will default to whatever
+    # the database default is for an unsized value.
+    def type_literal_generic_numeric(column)
+      column[:size] ? "numeric(#{Array(column[:size]).join(', ')})" : :numeric
+    end
+
+    # Sequel uses the varchar type by default for Strings.  If a
+    # size isn't present, Sequel assumes a size of 255.  If the
+    # :fixed option is used, Sequel uses the char type.  If the
+    # :text option is used, Sequel uses the :text type.
+    def type_literal_generic_string(column)
+      if column[:text]
+        :text
+      elsif column[:fixed]
+        "char(#{column[:size]||255})"
+      else
+        "varchar(#{column[:size]||255})"
+      end
+    end
+    
+    # Sequel uses the timestamp type by default for Time values.
+    # If the :only_time option is used, the time type is used.
+    def type_literal_generic_time(column)
+      column[:only_time] ? :time : :timestamp
+    end
+
+    # Sequel uses the boolean type by default for TrueClass and FalseClass.
+    def type_literal_generic_trueclass(column)
+      :boolean
+    end
+
+    # SQL fragment for the given type of a column if the column is not one of the
+    # generic types specified with a ruby class.
+    def type_literal_specific(column)
+      type = column[:type]
+      type = "double precision" if type.to_s == 'double'
+      column[:size] ||= 255 if type.to_s == 'varchar'
+      elements = column[:size] || column[:elements]
+      "#{type}#{literal(Array(elements)) if elements}#{UNSIGNED if column[:unsigned]}"
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/database.rb 3.0.0-1/lib/sequel/database.rb
--- 1.5.1-1/lib/sequel/database.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/database.rb	2009-04-28 00:16:33.000000000 +0100
@@ -0,0 +1,822 @@
+module Sequel
+  # Hash of adapters that have been used. The key is the adapter scheme
+  # symbol, and the value is the Database subclass.
+  ADAPTER_MAP = {}
+    
+  # Array of all databases to which Sequel has connected.  If you are
+  # developing an application that can connect to an arbitrary number of 
+  # databases, delete the database objects from this or they will not get
+  # garbage collected.
+  DATABASES = []
+
+  # A Database object represents a virtual connection to a database.
+  # The Database class is meant to be subclassed by database adapters in order
+  # to provide the functionality needed for executing queries.
+  class Database
+    extend Metaprogramming
+    include Metaprogramming
+
+    # Array of supported database adapters
+    ADAPTERS = %w'ado db2 dbi do firebird informix jdbc mysql odbc openbase oracle postgres sqlite'.collect{|x| x.to_sym}
+
+    SQL_BEGIN = 'BEGIN'.freeze
+    SQL_COMMIT = 'COMMIT'.freeze
+    SQL_ROLLBACK = 'ROLLBACK'.freeze
+
+    # The identifier input method to use by default
+    @@identifier_input_method = nil
+
+    # The identifier output method to use by default
+    @@identifier_output_method = nil
+
+    # Whether to use the single threaded connection pool by default
+    @@single_threaded = false
+
+    # Whether to quote identifiers (columns and tables) by default
+    @@quote_identifiers = nil
+
+    # The default schema to use, generally should be nil.
+    attr_accessor :default_schema
+
+    # Array of SQL loggers to use for this database
+    attr_accessor :loggers
+    
+    # The options for this database
+    attr_reader :opts
+    
+    # The connection pool for this database
+    attr_reader :pool
+
+    # The prepared statement objects for this database, keyed by name
+    attr_reader :prepared_statements
+    
+    # Constructs a new instance of a database connection with the specified
+    # options hash.
+    #
+    # Sequel::Database is an abstract class that is not useful by itself.
+    #
+    # Takes the following options:
+    # * :default_schema : The default schema to use, should generally be nil
+    # * :disconnection_proc: A proc used to disconnect the connection.
+    # * :identifier_input_method: A string method symbol to call on identifiers going into the database
+    # * :identifier_output_method: A string method symbol to call on identifiers coming from the database
+    # * :loggers : An array of loggers to use.
+    # * :quote_identifiers : Whether to quote identifiers
+    # * :single_threaded : Whether to use a single-threaded connection pool
+    #
+    # All options given are also passed to the ConnectionPool.  If a block
+    # is given, it is used as the connection_proc for the ConnectionPool.
+    def initialize(opts = {}, &block)
+      @opts ||= opts
+      
+      @single_threaded = opts.include?(:single_threaded) ? opts[:single_threaded] : @@single_threaded
+      @schemas = nil
+      @default_schema = opts.include?(:default_schema) ? opts[:default_schema] : default_schema_default
+      @prepared_statements = {}
+      @transactions = []
+      @identifier_input_method = nil
+      @identifier_output_method = nil
+      @quote_identifiers = nil
+      @pool = (@single_threaded ? SingleThreadedPool : ConnectionPool).new(connection_pool_default_options.merge(opts), &block)
+      @pool.connection_proc = proc{|server| connect(server)} unless block
+      @pool.disconnection_proc = proc{|conn| disconnect_connection(conn)} unless opts[:disconnection_proc]
+
+      @loggers = Array(opts[:logger]) + Array(opts[:loggers])
+      ::Sequel::DATABASES.push(self)
+    end
+    
+    ### Class Methods ###
+
+    # The Database subclass for the given adapter scheme.
+    # Raises Sequel::AdapterNotFound if the adapter
+    # could not be loaded.
+    def self.adapter_class(scheme)
+      scheme = scheme.to_s.gsub('-', '_').to_sym
+      
+      unless klass = ADAPTER_MAP[scheme]
+        # attempt to load the adapter file
+        begin
+          Sequel.require "adapters/#{scheme}"
+        rescue LoadError => e
+          raise AdapterNotFound, "Could not load #{scheme} adapter:\n  #{e.message}"
+        end
+        
+        # make sure we actually loaded the adapter
+        unless klass = ADAPTER_MAP[scheme]
+          raise AdapterNotFound, "Could not load #{scheme} adapter"
+        end
+      end
+      klass
+    end
+        
+    # Returns the scheme for the Database class.
+    def self.adapter_scheme
+      @scheme
+    end
+    
+    # Connects to a database.  See Sequel.connect.
+    def self.connect(conn_string, opts = {}, &block)
+      if conn_string.is_a?(String)
+        if match = /\A(jdbc|do):/o.match(conn_string)
+          c = adapter_class(match[1].to_sym)
+          opts = {:uri=>conn_string}.merge(opts)
+        else
+          uri = URI.parse(conn_string)
+          scheme = uri.scheme
+          scheme = :dbi if scheme =~ /\Adbi-/
+          c = adapter_class(scheme)
+          uri_options = {}
+          uri.query.split('&').collect{|s| s.split('=')}.each{|k,v| uri_options[k.to_sym] = v} unless uri.query.to_s.strip.empty?
+          opts = c.send(:uri_to_options, uri).merge(uri_options).merge(opts)
+        end
+      else
+        opts = conn_string.merge(opts)
+        c = adapter_class(opts[:adapter] || opts['adapter'])
+      end
+      # process opts a bit
+      opts = opts.inject({}) do |m, kv| k, v = *kv
+        k = :user if k.to_s == 'username'
+        m[k.to_sym] = v
+        m
+      end
+      if block
+        begin
+          yield(db = c.new(opts))
+        ensure
+          db.disconnect if db
+          ::Sequel::DATABASES.delete(db)
+        end
+        nil
+      else
+        c.new(opts)
+      end
+    end
+    
+    # The method to call on identifiers going into the database
+    def self.identifier_input_method
+      @@identifier_input_method
+    end
+    
+    # Set the method to call on identifiers going into the database
+    # See Sequel.identifier_input_method=.
+    def self.identifier_input_method=(v)
+      @@identifier_input_method = v || ""
+    end
+    
+    # The method to call on identifiers coming from the database
+    def self.identifier_output_method
+      @@identifier_output_method
+    end
+    
+    # Set the method to call on identifiers coming from the database
+    # See Sequel.identifier_output_method=.
+    def self.identifier_output_method=(v)
+      @@identifier_output_method = v || ""
+    end
+
+    # Sets the default quote_identifiers mode for new databases.
+    # See Sequel.quote_identifiers=.
+    def self.quote_identifiers=(value)
+      @@quote_identifiers = value
+    end
+
+    # Sets the default single_threaded mode for new databases.
+    # See Sequel.single_threaded=.
+    def self.single_threaded=(value)
+      @@single_threaded = value
+    end
+
+    ### Private Class Methods ###
+
+    # Sets the adapter scheme for the Database class. Call this method in
+    # descendants of Database to allow connection using a URL. For example the
+    # following:
+    #
+    #   class Sequel::MyDB::Database < Sequel::Database
+    #     set_adapter_scheme :mydb
+    #     ...
+    #   end
+    #
+    # would allow connection using:
+    #
+    #   Sequel.connect('mydb://user:password@dbserver/mydb')
+    def self.set_adapter_scheme(scheme) # :nodoc:
+      @scheme = scheme
+      ADAPTER_MAP[scheme.to_sym] = self
+    end
+    
+    # Converts a uri to an options hash. These options are then passed
+    # to a newly created database object. 
+    def self.uri_to_options(uri) # :nodoc:
+      { :user => uri.user,
+        :password => uri.password,
+        :host => uri.host,
+        :port => uri.port,
+        :database => (m = /\/(.*)/.match(uri.path)) && (m[1]) }
+    end
+    
+    private_class_method :set_adapter_scheme, :uri_to_options
+    
+    ### Instance Methods ###
+
+    # Executes the supplied SQL statement string.
+    def <<(sql)
+      execute_ddl(sql)
+    end
+    
+    # Returns a dataset from the database. If the first argument is a string,
+    # the method acts as an alias for Database#fetch, returning a dataset for
+    # arbitrary SQL:
+    #
+    #   DB['SELECT * FROM items WHERE name = ?', my_name].all
+    #
+    # Otherwise, acts as an alias for Database#from, setting the primary
+    # table for the dataset:
+    #
+    #   DB[:items].sql #=> "SELECT * FROM items"
+    def [](*args)
+      (String === args.first) ? fetch(*args) : from(*args)
+    end
+    
+    # Call the prepared statement with the given name with the given hash
+    # of arguments.
+    def call(ps_name, hash={})
+      prepared_statements[ps_name].call(hash)
+    end
+    
+    # Cast the given type to a literal type
+    def cast_type_literal(type)
+      type_literal(:type=>type)
+    end
+
+    # Connects to the database. This method should be overridden by descendants.
+    def connect
+      raise NotImplementedError, "#connect should be overridden by adapters"
+    end
+    
+    # The database type for this database object, the same as the adapter scheme
+    # by default.  Should be overridden in adapters (especially shared adapters)
+    # to be the correct type, so that even if two separate Database objects are
+    # using different adapters you can tell that they are using the same database
+    # type.  Even better, you can tell that two Database objects that are using
+    # the same adapter are connecting to different database types (think JDBC or
+    # DataObjects).
+    def database_type
+      self.class.adapter_scheme
+    end
+    
+    # Returns a blank dataset for this database
+    def dataset
+      ds = Sequel::Dataset.new(self)
+    end
+    
+    # Disconnects all available connections from the connection pool.  Any
+    # connections currently in use will not be disconnected.
+    def disconnect
+      pool.disconnect
+    end
+
+    # Executes the given SQL on the database. This method should be overridden in descendants.
+    # This method should not be called directly by user code.
+    def execute(sql, opts={})
+      raise NotImplementedError, "#execute should be overridden by adapters"
+    end
+    
+    # Method that should be used when submitting any DDL (Data Definition
+    # Language) SQL.  By default, calls execute_dui.
+    # This method should not be called directly by user code.
+    def execute_ddl(sql, opts={}, &block)
+      execute_dui(sql, opts, &block)
+    end
+
+    # Method that should be used when issuing a DELETE, UPDATE, or INSERT
+    # statement.  By default, calls execute.
+    # This method should not be called directly by user code.
+    def execute_dui(sql, opts={}, &block)
+      execute(sql, opts, &block)
+    end
+
+    # Method that should be used when issuing a INSERT
+    # statement.  By default, calls execute_dui.
+    # This method should not be called directly by user code.
+    def execute_insert(sql, opts={}, &block)
+      execute_dui(sql, opts, &block)
+    end
+
+    # Fetches records for an arbitrary SQL statement. If a block is given,
+    # it is used to iterate over the records:
+    #
+    #   DB.fetch('SELECT * FROM items'){|r| p r}
+    #
+    # The method returns a dataset instance:
+    #
+    #   DB.fetch('SELECT * FROM items').all
+    #
+    # Fetch can also perform parameterized queries for protection against SQL
+    # injection:
+    #
+    #   DB.fetch('SELECT * FROM items WHERE name = ?', my_name).all
+    def fetch(sql, *args, &block)
+      ds = dataset.with_sql(sql, *args)
+      ds.each(&block) if block
+      ds
+    end
+    
+    # Returns a new dataset with the from method invoked. If a block is given,
+    # it is used as a filter on the dataset.
+    def from(*args, &block)
+      ds = dataset.from(*args)
+      block ? ds.filter(&block) : ds
+    end
+    
+    # Returns a single value from the database, e.g.:
+    #
+    #   # SELECT 1
+    #   DB.get(1) #=> 1 
+    #
+    #   # SELECT version()
+    #   DB.get(:version.sql_function) #=> ...
+    def get(*args, &block)
+      dataset.get(*args, &block)
+    end
+    
+    # The method to call on identifiers going into the database
+    def identifier_input_method
+      case @identifier_input_method
+      when nil
+        @identifier_input_method = @opts.include?(:identifier_input_method) ? @opts[:identifier_input_method] : (@@identifier_input_method.nil? ? identifier_input_method_default : @@identifier_input_method)
+        @identifier_input_method == "" ? nil : @identifier_input_method
+      when ""
+        nil
+      else
+        @identifier_input_method
+      end
+    end
+    
+    # Set the method to call on identifiers going into the database
+    def identifier_input_method=(v)
+      reset_schema_utility_dataset
+      @identifier_input_method = v || ""
+    end
+    
+    # The method to call on identifiers coming from the database
+    def identifier_output_method
+      case @identifier_output_method
+      when nil
+        @identifier_output_method = @opts.include?(:identifier_output_method) ? @opts[:identifier_output_method] : (@@identifier_output_method.nil? ? identifier_output_method_default : @@identifier_output_method)
+        @identifier_output_method == "" ? nil : @identifier_output_method
+      when ""
+        nil
+      else
+        @identifier_output_method
+      end
+    end
+    
+    # Set the method to call on identifiers coming from the database
+    def identifier_output_method=(v)
+      reset_schema_utility_dataset
+      @identifier_output_method = v || ""
+    end
+    
+    # Returns a string representation of the database object including the
+    # class name and the connection URI (or the opts if the URI
+    # cannot be constructed).
+    def inspect
+      "#<#{self.class}: #{(uri rescue opts).inspect}>" 
+    end
+
+    # Proxy the literal call to the dataset, used for default values.
+    def literal(v)
+      schema_utility_dataset.literal(v)
+    end
+
+    # Log a message at level info to all loggers.  All SQL logging
+    # goes through this method.
+    def log_info(message, args=nil)
+      message = "#{message}; #{args.inspect}" if args
+      @loggers.each{|logger| logger.info(message)}
+    end
+
+    # Remove any existing loggers and just use the given logger.
+    def logger=(logger)
+      @loggers = Array(logger)
+    end
+
+    # Whether to quote identifiers (columns and tables) for this database
+    def quote_identifiers=(v)
+      reset_schema_utility_dataset
+      @quote_identifiers = v
+    end
+    
+    # Returns true if the database quotes identifiers.
+    def quote_identifiers?
+      return @quote_identifiers unless @quote_identifiers.nil?
+      @quote_identifiers = @opts.include?(:quote_identifiers) ? @opts[:quote_identifiers] : (@@quote_identifiers.nil? ? quote_identifiers_default : @@quote_identifiers)
+    end
+    
+    # Returns a new dataset with the select method invoked.
+    def select(*args, &block)
+      dataset.select(*args, &block)
+    end
+    
+    # Parse the schema from the database.
+    # Returns the schema for the given table as an array with all members being arrays of length 2,
+    # the first member being the column name, and the second member being a hash of column information.
+    # Available options are:
+    #
+    # * :reload - Get fresh information from the database, instead of using
+    #   cached information.  If table_name is blank, :reload should be used
+    #   unless you are sure that schema has not been called before with a
+    #   table_name, otherwise you may only getting the schemas for tables
+    #   that have been requested explicitly.
+    # * :schema - An explicit schema to use.  It may also be implicitly provided
+    #   via the table name.
+    def schema(table, opts={})
+      raise(Error, 'schema parsing is not implemented on this database') unless respond_to?(:schema_parse_table, true)
+
+      sch, table_name = schema_and_table(table)
+      quoted_name = quote_schema_table(table)
+      opts = opts.merge(:schema=>sch) if sch && !opts.include?(:schema)
+
+      @schemas.delete(quoted_name) if opts[:reload] && @schemas
+      return @schemas[quoted_name] if @schemas && @schemas[quoted_name]
+
+      @schemas ||= Hash.new do |h,k|
+        quote_name = quote_schema_table(k)
+        h[quote_name] if h.include?(quote_name)
+      end
+
+      cols = schema_parse_table(table_name, opts)
+      raise(Error, 'schema parsing returned no columns, table probably doesn\'t exist') if cols.nil? || cols.empty?
+      @schemas[quoted_name] = cols
+    end
+
+    # Returns true if the database is using a single-threaded connection pool.
+    def single_threaded?
+      @single_threaded
+    end
+    
+    # Acquires a database connection, yielding it to the passed block.
+    def synchronize(server=nil, &block)
+      @pool.hold(server || :default, &block)
+    end
+
+    # Returns true if a table with the given name exists.  This requires a query
+    # to the database unless this database object already has the schema for
+    # the given table name.
+    def table_exists?(name)
+      if @schemas && @schemas[name]
+        true
+      else
+        begin 
+          from(name).first
+          true
+        rescue
+          false
+        end
+      end
+    end
+    
+    # Attempts to acquire a database connection.  Returns true if successful.
+    # Will probably raise an error if unsuccessful.
+    def test_connection(server=nil)
+      synchronize(server){|conn|}
+      true
+    end
+
+    # A simple implementation of SQL transactions. Nested transactions are not 
+    # supported - calling #transaction within a transaction will reuse the 
+    # current transaction. Should be overridden for databases that support nested 
+    # transactions.
+    def transaction(opts={})
+      synchronize(opts[:server]) do |conn|
+        return yield(conn) if @transactions.include?(Thread.current)
+        log_info(begin_transaction_sql)
+        conn.execute(begin_transaction_sql)
+        begin
+          @transactions << Thread.current
+          yield(conn)
+        rescue Exception => e
+          log_info(rollback_transaction_sql)
+          conn.execute(rollback_transaction_sql)
+          transaction_error(e)
+        ensure
+          unless e
+            log_info(commit_transaction_sql)
+            conn.execute(commit_transaction_sql)
+          end
+          @transactions.delete(Thread.current)
+        end
+      end
+    end
+    
+    # Typecast the value to the given column_type. Calls
+    # typecast_value_#{column_type} if the method exists,
+    # otherwise returns the value.
+    # This method should raise Sequel::InvalidValue if assigned value
+    # is invalid.
+    def typecast_value(column_type, value)
+      return nil if value.nil?
+      meth = "typecast_value_#{column_type}"
+      begin
+        respond_to?(meth, true) ? send(meth, value) : value
+      rescue ArgumentError, TypeError => exp
+        e = Sequel::InvalidValue.new("#{exp.class} #{exp.message}")
+        e.set_backtrace(exp.backtrace)
+        raise e
+      end
+    end
+    
+    # Returns the URI identifying the database.
+    # This method can raise an error if the database used options
+    # instead of a connection string.
+    def uri
+      uri = URI::Generic.new(
+        self.class.adapter_scheme.to_s,
+        nil,
+        @opts[:host],
+        @opts[:port],
+        nil,
+        "/#{@opts[:database]}",
+        nil,
+        nil,
+        nil
+      )
+      uri.user = @opts[:user]
+      uri.password = @opts[:password] if uri.user
+      uri.to_s
+    end
+    
+    # Explicit alias of uri for easier subclassing.
+    def url
+      uri
+    end
+    
+    private
+    
+    # SQL to BEGIN a transaction.
+    def begin_transaction_sql
+      SQL_BEGIN
+    end
+
+    # Returns true when the object is considered blank.
+    # The only objects that are blank are nil, false,
+    # strings with all whitespace, and ones that respond
+    # true to empty?
+    def blank_object?(obj)
+      return obj.blank? if obj.respond_to?(:blank?)
+      case obj
+      when NilClass, FalseClass
+        true
+      when Numeric, TrueClass
+        false
+      when String
+        obj.strip.empty?
+      else
+        obj.respond_to?(:empty?) ? obj.empty? : false
+      end
+    end
+
+    # SQL to COMMIT a transaction.
+    def commit_transaction_sql
+      SQL_COMMIT
+    end
+
+    # The default options for the connection pool.
+    def connection_pool_default_options
+      {}
+    end
+    
+    # The default value for default_schema.
+    def default_schema_default
+      nil
+    end
+
+    # The method to apply to identifiers going into the database by default.
+    # Should be overridden in subclasses for databases that fold unquoted
+    # identifiers to lower case instead of uppercase, such as
+    # MySQL, PostgreSQL, and SQLite.
+    def identifier_input_method_default
+      :upcase
+    end
+    
+    # The method to apply to identifiers coming the database by default.
+    # Should be overridden in subclasses for databases that fold unquoted
+    # identifiers to lower case instead of uppercase, such as
+    # MySQL, PostgreSQL, and SQLite.
+    def identifier_output_method_default
+      :downcase
+    end
+    
+    # Return a Method object for the dataset's output_identifier_method.
+    # Used in metadata parsing to make sure the returned information is in the
+    # correct format.
+    def input_identifier_meth
+      dataset.method(:input_identifier)
+    end
+
+    # Return a dataset that uses the default identifier input and output methods
+    # for this database.  Used when parsing metadata so that column symbols are
+    # returned as expected.
+    def metadata_dataset
+      return @metadata_dataset if @metadata_dataset
+      ds = dataset
+      ds.identifier_input_method = identifier_input_method_default
+      ds.identifier_output_method = identifier_output_method_default
+      @metadata_dataset = ds
+    end
+
+    # Return a Method object for the dataset's output_identifier_method.
+    # Used in metadata parsing to make sure the returned information is in the
+    # correct format.
+    def output_identifier_meth
+      dataset.method(:output_identifier)
+    end
+
+    # Whether to quote identifiers by default for this database, true
+    # by default.
+    def quote_identifiers_default
+      true
+    end
+
+    # SQL to ROLLBACK a transaction.
+    def rollback_transaction_sql
+      SQL_ROLLBACK
+    end
+    
+    # Convert the given exception to a DatabaseError, keeping message
+    # and traceback.
+    def raise_error(exception, opts={})
+      if !opts[:classes] || Array(opts[:classes]).any?{|c| exception.is_a?(c)}
+        e = DatabaseError.new("#{exception.class} #{exception.message}")
+        e.set_backtrace(exception.backtrace)
+        raise e
+      else
+        raise exception
+      end
+    end
+    
+    # Remove the cached schema for the given schema name
+    def remove_cached_schema(table)
+      @schemas.delete(quote_schema_table(table)) if @schemas
+    end
+
+    # Remove the cached schema_utility_dataset, because the identifier
+    # quoting has changed.
+    def reset_schema_utility_dataset
+      @schema_utility_dataset = nil
+    end
+
+    # Split the schema information from the table
+    def schema_and_table(table_name)
+      schema_utility_dataset.schema_and_table(table_name)
+    end
+
+    # Return true if the given column schema represents an autoincrementing primary key.
+    def schema_autoincrementing_primary_key?(schema)
+      !!schema[:primary_key]
+    end
+
+    # Match the database's column type to a ruby type via a
+    # regular expression.  The following ruby types are supported:
+    # integer, string, date, datetime, boolean, and float.
+    def schema_column_type(db_type)
+      case db_type
+      when /\Atinyint/io
+        Sequel.convert_tinyint_to_bool ? :boolean : :integer
+      when /\Ainterval\z/io
+        :interval
+      when /\A(character( varying)?|varchar|text)/io
+        :string
+      when /\A(int(eger)?|bigint|smallint)/io
+        :integer
+      when /\Adate\z/io
+        :date
+      when /\A(datetime|timestamp( with(out)? time zone)?)\z/io
+        :datetime
+      when /\Atime( with(out)? time zone)?\z/io
+        :time
+      when /\Aboolean\z/io
+        :boolean
+      when /\A(real|float|double( precision)?)\z/io
+        :float
+      when /\A(((numeric|decimal)(\(\d+,\d+\))?)|money)\z/io
+        :decimal
+      when /bytea|blob/io
+        :blob
+      end
+    end
+
+    # The dataset to use for proxying certain schema methods.
+    def schema_utility_dataset
+      @schema_utility_dataset ||= dataset
+    end
+
+    # Return the options for the given server by merging the generic
+    # options for all server with the specific options for the given
+    # server specified in the :servers option.
+    def server_opts(server)
+      opts = if @opts[:servers] && server_options = @opts[:servers][server]
+        case server_options
+        when Hash
+          @opts.merge(server_options)
+        when Proc
+          @opts.merge(server_options.call(self))
+        else
+          raise Error, 'Server opts should be a hash or proc'
+        end
+      else
+        @opts.dup
+      end
+      opts.delete(:servers)
+      opts
+    end
+
+    # Raise a database error unless the exception is an Rollback.
+    def transaction_error(e, *classes)
+      raise_error(e, :classes=>classes) unless Rollback === e
+    end
+
+    # Typecast the value to an SQL::Blob
+    def typecast_value_blob(value)
+      value.is_a?(Sequel::SQL::Blob) ? value : Sequel::SQL::Blob.new(value)
+    end
+
+    # Typecast the value to true, false, or nil
+    def typecast_value_boolean(value)
+      case value
+      when false, 0, "0", /\Af(alse)?\z/i
+        false
+      else
+        blank_object?(value) ? nil : true
+      end
+    end
+
+    # Typecast the value to a Date
+    def typecast_value_date(value)
+      case value
+      when Date
+        value
+      when DateTime, Time
+        Date.new(value.year, value.month, value.day)
+      when String
+        Sequel.string_to_date(value)
+      else
+        raise InvalidValue, "invalid value for Date: #{value.inspect}"
+      end
+    end
+
+    # Typecast the value to a DateTime or Time depending on Sequel.datetime_class
+    def typecast_value_datetime(value)
+      raise(Sequel::InvalidValue, "invalid value for Datetime: #{value.inspect}") unless [DateTime, Date, Time, String].any?{|c| value.is_a?(c)}
+      if Sequel.datetime_class === value
+        # Already the correct class, no need to convert
+       value
+      else
+        # First convert it to standard ISO 8601 time, then
+        # parse that string using the time class.
+        Sequel.string_to_datetime(Time === value ? value.iso8601 : value.to_s)
+      end
+    end
+
+    # Typecast the value to a BigDecimal
+    def typecast_value_decimal(value)
+      case value
+      when BigDecimal
+        value
+      when String, Numeric
+        BigDecimal.new(value.to_s)
+      else
+        raise InvalidValue, "invalid value for BigDecimal: #{value.inspect}"
+      end
+    end
+
+    # Typecast the value to a Float
+    def typecast_value_float(value)
+      Float(value)
+    end
+
+    # Typecast the value to an Integer
+    def typecast_value_integer(value)
+      Integer(value)
+    end
+
+    # Typecast the value to a String
+    def typecast_value_string(value)
+      value.to_s
+    end
+
+    # Typecast the value to a Time
+    def typecast_value_time(value)
+      case value
+      when Time
+        value
+      when String
+        Sequel.string_to_time(value)
+      else
+        raise Sequel::InvalidValue, "invalid value for Time: #{value.inspect}"
+      end
+    end
+  end
+end
+
diff -pruN 1.5.1-1/lib/sequel/dataset/convenience.rb 3.0.0-1/lib/sequel/dataset/convenience.rb
--- 1.5.1-1/lib/sequel/dataset/convenience.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/dataset/convenience.rb	2009-04-09 21:29:42.000000000 +0100
@@ -0,0 +1,252 @@
+module Sequel
+  class Dataset
+    COMMA_SEPARATOR = ', '.freeze
+    COUNT_OF_ALL_AS_COUNT = SQL::Function.new(:count, LiteralString.new('*'.freeze)).as(:count)
+    ARRAY_ACCESS_ERROR_MSG = 'You cannot call Dataset#[] with an integer or with no arguments.'.freeze
+    MAP_ERROR_MSG = 'Using Dataset#map with an argument and a block is not allowed'.freeze
+    GET_ERROR_MSG = 'must provide argument or block to Dataset#get, not both'.freeze
+    IMPORT_ERROR_MSG = 'Using Sequel::Dataset#import an empty column array is not allowed'.freeze
+
+    # Returns the first record matching the conditions. Examples:
+    #
+    #   ds[:id=>1] => {:id=1}
+    def [](*conditions)
+      raise(Error, ARRAY_ACCESS_ERROR_MSG) if (conditions.length == 1 and conditions.is_a?(Integer)) or conditions.length == 0
+      first(*conditions)
+    end
+
+    # Update all records matching the conditions
+    # with the values specified. Examples:
+    #
+    #   ds[:id=>1] = {:id=>2} # SQL: UPDATE ... SET id = 2 WHERE id = 1
+    def []=(conditions, values)
+      filter(conditions).update(values)
+    end
+
+    # Returns the average value for the given column.
+    def avg(column)
+      get{|o| o.avg(column)}
+    end
+    
+    # Returns true if no records exist in the dataset, false otherwise
+    def empty?
+      get(1).nil?
+    end
+    
+    # If a integer argument is
+    # given, it is interpreted as a limit, and then returns all 
+    # matching records up to that limit.  If no argument is passed,
+    # it returns the first matching record.  If any other type of
+    # argument(s) is passed, it is given to filter and the
+    # first matching record is returned. If a block is given, it is used
+    # to filter the dataset before returning anything.  Examples:
+    # 
+    #   ds.first => {:id=>7}
+    #   ds.first(2) => [{:id=>6}, {:id=>4}]
+    #   ds.order(:id).first(2) => [{:id=>1}, {:id=>2}]
+    #   ds.first(:id=>2) => {:id=>2}
+    #   ds.first("id = 3") => {:id=>3}
+    #   ds.first("id = ?", 4) => {:id=>4}
+    #   ds.first{|o| o.id > 2} => {:id=>5}
+    #   ds.order(:id).first{|o| o.id > 2} => {:id=>3}
+    #   ds.first{|o| o.id > 2} => {:id=>5}
+    #   ds.first("id > ?", 4){|o| o.id < 6} => {:id=>5}
+    #   ds.order(:id).first(2){|o| o.id < 2} => [{:id=>1}]
+    def first(*args, &block)
+      ds = block ? filter(&block) : self
+
+      if args.empty?
+        ds.single_record
+      else
+        args = (args.size == 1) ? args.first : args
+        if Integer === args
+          ds.limit(args).all
+        else
+          ds.filter(args).single_record
+        end
+      end
+    end
+
+    # Return the column value for the first matching record in the dataset.
+    # Raises an error if both an argument and block is given.
+    #
+    #   ds.get(:id)
+    #   ds.get{|o| o.sum(:id)}
+    def get(column=nil, &block)
+      if column
+        raise(Error, GET_ERROR_MSG) if block
+        select(column).single_value
+      else
+        select(&block).single_value
+      end
+    end
+
+    # Returns a dataset grouped by the given column with count by group,
+    # order by the count of records.  Examples:
+    #
+    #   ds.group_and_count(:name) => [{:name=>'a', :count=>1}, ...]
+    #   ds.group_and_count(:first_name, :last_name) => [{:first_name=>'a', :last_name=>'b', :count=>1}, ...]
+    def group_and_count(*columns)
+      group(*columns).select(*(columns + [COUNT_OF_ALL_AS_COUNT])).order(:count)
+    end
+    
+    # Inserts multiple records into the associated table. This method can be
+    # to efficiently insert a large amounts of records into a table. Inserts
+    # are automatically wrapped in a transaction.
+    # 
+    # This method is called with a columns array and an array of value arrays:
+    #
+    #   dataset.import([:x, :y], [[1, 2], [3, 4]])
+    #
+    # This method also accepts a dataset instead of an array of value arrays:
+    #
+    #   dataset.import([:x, :y], other_dataset.select(:a___x, :b___y))
+    #
+    # The method also accepts a :slice or :commit_every option that specifies
+    # the number of records to insert per transaction. This is useful especially
+    # when inserting a large number of records, e.g.:
+    #
+    #   # this will commit every 50 records
+    #   dataset.import([:x, :y], [[1, 2], [3, 4], ...], :slice => 50)
+    def import(columns, values, opts={})
+      return @db.transaction{execute_dui("INSERT INTO #{quote_schema_table(@opts[:from].first)} (#{identifier_list(columns)}) VALUES #{literal(values)}")} if values.is_a?(Dataset)
+
+      return if values.empty?
+      raise(Error, IMPORT_ERROR_MSG) if columns.empty?
+      
+      if slice_size = opts[:commit_every] || opts[:slice]
+        offset = 0
+        loop do
+          @db.transaction(opts){multi_insert_sql(columns, values[offset, slice_size]).each{|st| execute_dui(st)}}
+          offset += slice_size
+          break if offset >= values.length
+        end
+      else
+        statements = multi_insert_sql(columns, values)
+        @db.transaction{statements.each{|st| execute_dui(st)}}
+      end
+    end
+    
+    # Returns the interval between minimum and maximum values for the given 
+    # column.
+    def interval(column)
+      get{|o| o.max(column) - o.min(column)}
+    end
+
+    # Reverses the order and then runs first.  Note that this
+    # will not necessarily give you the last record in the dataset,
+    # unless you have an unambiguous order.  If there is not
+    # currently an order for this dataset, raises an Error.
+    def last(*args, &block)
+      raise(Error, 'No order specified') unless @opts[:order]
+      reverse.first(*args, &block)
+    end
+    
+    # Maps column values for each record in the dataset (if a column name is
+    # given), or performs the stock mapping functionality of Enumerable. 
+    # Raises an error if both an argument and block are given. Examples:
+    #
+    #   ds.map(:id) => [1, 2, 3, ...]
+    #   ds.map{|r| r[:id] * 2} => [2, 4, 6, ...]
+    def map(column=nil, &block)
+      if column
+        raise(Error, MAP_ERROR_MSG) if block
+        super(){|r| r[column]}
+      else
+        super(&block)
+      end
+    end
+
+    # Returns the maximum value for the given column.
+    def max(column)
+      get{|o| o.max(column)}
+    end
+
+    # Returns the minimum value for the given column.
+    def min(column)
+      get{|o| o.min(column)}
+    end
+
+    # This is a front end for import that allows you to submit an array of
+    # hashes instead of arrays of columns and values:
+    # 
+    #   dataset.multi_insert([{:x => 1}, {:x => 2}])
+    #
+    # Be aware that all hashes should have the same keys if you use this calling method,
+    # otherwise some columns could be missed or set to null instead of to default
+    # values.
+    #
+    # You can also use the :slice or :commit_every option that import accepts.
+    def multi_insert(hashes, opts={})
+      return if hashes.empty?
+      columns = hashes.first.keys
+      import(columns, hashes.map{|h| columns.map{|c| h[c]}}, opts)
+    end
+
+    # Returns a Range object made from the minimum and maximum values for the
+    # given column.
+    def range(column)
+      if r = select{|o| [o.min(column).as(:v1), o.max(column).as(:v2)]}.first
+        (r[:v1]..r[:v2])
+      end
+    end
+    
+    # Returns the first record in the dataset.
+    def single_record
+      clone(:limit=>1).each{|r| return r}
+      nil
+    end
+
+    # Returns the first value of the first record in the dataset.
+    # Returns nil if dataset is empty.
+    def single_value
+      if r = naked.clone(:graph=>false).single_record
+        r.values.first
+      end
+    end
+    
+    # Returns the sum for the given column.
+    def sum(column)
+      get{|o| o.sum(column)}
+    end
+
+    # Returns true if the table exists.  Will raise an error
+    # if the dataset has fixed SQL or selects from another dataset
+    # or more than one table.
+    def table_exists?
+      raise(Sequel::Error, "this dataset has fixed SQL") if @opts[:sql]
+      raise(Sequel::Error, "this dataset selects from multiple sources") if @opts[:from].size != 1
+      t = @opts[:from].first
+      raise(Sequel::Error, "this dataset selects from a sub query") if t.is_a?(Dataset)
+      @db.table_exists?(t)
+    end
+
+    # Returns a string in CSV format containing the dataset records. By 
+    # default the CSV representation includes the column titles in the
+    # first line. You can turn that off by passing false as the 
+    # include_column_titles argument.
+    #
+    # This does not use a CSV library or handle quoting of values in
+    # any way.  If any values in any of the rows could include commas or line
+    # endings, you shouldn't use this.
+    def to_csv(include_column_titles = true)
+      n = naked
+      cols = n.columns
+      csv = ''
+      csv << "#{cols.join(COMMA_SEPARATOR)}\r\n" if include_column_titles
+      n.each{|r| csv << "#{cols.collect{|c| r[c]}.join(COMMA_SEPARATOR)}\r\n"}
+      csv
+    end
+    
+    # Returns a hash with one column used as key and another used as value.
+    # If rows have duplicate values for the key column, the latter row(s)
+    # will overwrite the value of the previous row(s). If the value_column
+    # is not given or nil, uses the entire hash as the value.
+    def to_hash(key_column, value_column = nil)
+      inject({}) do |m, r|
+        m[r[key_column]] = value_column ? r[value_column] : r
+        m
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/dataset/graph.rb 3.0.0-1/lib/sequel/dataset/graph.rb
--- 1.5.1-1/lib/sequel/dataset/graph.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/dataset/graph.rb	2009-04-09 20:56:55.000000000 +0100
@@ -0,0 +1,226 @@
+module Sequel
+  class Dataset
+    # Allows you to join multiple datasets/tables and have the result set
+    # split into component tables.
+    #
+    # This differs from the usual usage of join, which returns the result set
+    # as a single hash.  For example:
+    #
+    #   # CREATE TABLE artists (id INTEGER, name TEXT);
+    #   # CREATE TABLE albums (id INTEGER, name TEXT, artist_id INTEGER);
+    #   DB[:artists].left_outer_join(:albums, :artist_id=>:id).first
+    #   => {:id=>albums.id, :name=>albums.name, :artist_id=>albums.artist_id}
+    #   DB[:artists].graph(:albums, :artist_id=>:id).first
+    #   => {:artists=>{:id=>artists.id, :name=>artists.name}, :albums=>{:id=>albums.id, :name=>albums.name, :artist_id=>albums.artist_id}}
+    #
+    # Using a join such as left_outer_join, the attribute names that are shared between
+    # the tables are combined in the single return hash.  You can get around that by
+    # using .select with correct aliases for all of the columns, but it is simpler to
+    # use graph and have the result set split for you.  In addition, graph respects
+    # any row_proc of the current dataset and the datasets you use with graph.
+    #
+    # If you are graphing a table and all columns for that table are nil, this
+    # indicates that no matching rows existed in the table, so graph will return nil
+    # instead of a hash with all nil values:
+    #
+    #   # If the artist doesn't have any albums
+    #   DB[:artists].graph(:albums, :artist_id=>:id).first
+    #   => {:artists=>{:id=>artists.id, :name=>artists.name}, :albums=>nil}
+    #
+    # Arguments:
+    # * dataset -  Can be a symbol (specifying a table), another dataset,
+    #   or an object that responds to .dataset and return a symbol or a dataset
+    # * join_conditions - Any condition(s) allowed by join_table.
+    # * options -  A hash of graph options.  The following options are currently used:
+    #   * :implicit_qualifier - The qualifier of implicit conditions, see #join_table.
+    #   * :join_type - The type of join to use (passed to join_table).  Defaults to
+    #     :left_outer.
+    #   * :select - An array of columns to select.  When not used, selects
+    #     all columns in the given dataset.  When set to false, selects no
+    #     columns and is like simply joining the tables, though graph keeps
+    #     some metadata about join that makes it important to use graph instead
+    #     of join.
+    #   * :table_alias - The alias to use for the table.  If not specified, doesn't
+    #     alias the table.  You will get an error if the the alias (or table) name is
+    #     used more than once.
+    # * block - A block that is passed to join_table.
+    def graph(dataset, join_conditions = nil, options = {}, &block)
+      # Allow the use of a model, dataset, or symbol as the first argument
+      # Find the table name/dataset based on the argument
+      dataset = dataset.dataset if dataset.respond_to?(:dataset)
+      case dataset
+      when Symbol
+        table = dataset
+        dataset = @db[dataset]
+      when ::Sequel::Dataset
+        table = dataset.first_source
+      else
+        raise Error, "The dataset argument should be a symbol, dataset, or model"
+      end
+
+      # Raise Sequel::Error with explanation that the table alias has been used
+      raise_alias_error = lambda do
+        raise(Error, "this #{options[:table_alias] ? 'alias' : 'table'} has already been been used, please specify " \
+          "#{options[:table_alias] ? 'a different alias' : 'an alias via the :table_alias option'}") 
+      end
+
+      # Only allow table aliases that haven't been used
+      table_alias = options[:table_alias] || table
+      raise_alias_error.call if @opts[:graph] && @opts[:graph][:table_aliases] && @opts[:graph][:table_aliases].include?(table_alias)
+
+      # Join the table early in order to avoid cloning the dataset twice
+      ds = join_table(options[:join_type] || :left_outer, table, join_conditions, :table_alias=>table_alias, :implicit_qualifier=>options[:implicit_qualifier], &block)
+      opts = ds.opts
+
+      # Whether to include the table in the result set
+      add_table = options[:select] == false ? false : true
+      # Whether to add the columns to the list of column aliases
+      add_columns = !ds.opts.include?(:graph_aliases)
+
+      # Setup the initial graph data structure if it doesn't exist
+      unless graph = opts[:graph]
+        master = ds.first_source
+        raise_alias_error.call if master == table_alias
+        # Master hash storing all .graph related information
+        graph = opts[:graph] = {}
+        # Associates column aliases back to tables and columns
+        column_aliases = graph[:column_aliases] = {}
+        # Associates table alias (the master is never aliased)
+        table_aliases = graph[:table_aliases] = {master=>self}
+        # Keep track of the alias numbers used
+        ca_num = graph[:column_alias_num] = Hash.new(0)
+        # All columns in the master table are never
+        # aliased, but are not included if set_graph_aliases
+        # has been used.
+        if add_columns
+          select = opts[:select] = []
+          columns.each do |column|
+            column_aliases[column] = [master, column]
+            select.push(SQL::QualifiedIdentifier.new(master, column))
+          end
+        end
+      end
+
+      # Add the table alias to the list of aliases
+      # Even if it isn't been used in the result set,
+      # we add a key for it with a nil value so we can check if it
+      # is used more than once
+      table_aliases = graph[:table_aliases]
+      table_aliases[table_alias] = add_table ? dataset : nil
+
+      # Add the columns to the selection unless we are ignoring them
+      if add_table && add_columns
+        select = opts[:select]
+        column_aliases = graph[:column_aliases]
+        ca_num = graph[:column_alias_num]
+        # Which columns to add to the result set
+        cols = options[:select] || dataset.columns
+        # If the column hasn't been used yet, don't alias it.
+        # If it has been used, try table_column.
+        # If that has been used, try table_column_N 
+        # using the next value of N that we know hasn't been
+        # used
+        cols.each do |column|
+          col_alias, identifier = if column_aliases[column]
+            column_alias = :"#{table_alias}_#{column}"
+            if column_aliases[column_alias]
+              column_alias_num = ca_num[column_alias]
+              column_alias = :"#{column_alias}_#{column_alias_num}" 
+              ca_num[column_alias] += 1
+            end
+            [column_alias, SQL::QualifiedIdentifier.new(table_alias, column).as(column_alias)]
+          else
+            [column, SQL::QualifiedIdentifier.new(table_alias, column)]
+          end
+          column_aliases[col_alias] = [table_alias, column]
+          select.push(identifier)
+        end
+      end
+      ds
+    end
+
+    # This allows you to manually specify the graph aliases to use
+    # when using graph.  You can use it to only select certain
+    # columns, and have those columns mapped to specific aliases
+    # in the result set.  This is the equivalent of .select for a
+    # graphed dataset, and must be used instead of .select whenever
+    # graphing is used. Example:
+    #
+    #   DB[:artists].graph(:albums, :artist_id=>:id).set_graph_aliases(:artist_name=>[:artists, :name], :album_name=>[:albums, :name], :forty_two=>[:albums, :fourtwo, 42]).first
+    #   => {:artists=>{:name=>artists.name}, :albums=>{:name=>albums.name, :fourtwo=>42}}
+    #
+    # Arguments:
+    # * graph_aliases - Should be a hash with keys being symbols of
+    #   column aliases, and values being arrays with two or three elements.
+    #   The first element of the array should be the table alias symbol,
+    #   and the second should be the actual column name symbol. If the array
+    #   has a third element, it is used as the value returned, instead of
+    #   table_alias.column_name.
+    def set_graph_aliases(graph_aliases)
+      ds = select(*graph_alias_columns(graph_aliases))
+      ds.opts[:graph_aliases] = graph_aliases
+      ds
+    end
+
+    # Adds the give graph aliases to the list of graph aliases to use,
+    # unlike #set_graph_aliases, which replaces the list.  See
+    # #set_graph_aliases.
+    def add_graph_aliases(graph_aliases)
+      ds = select_more(*graph_alias_columns(graph_aliases))
+      ds.opts[:graph_aliases] = (ds.opts[:graph_aliases] || {}).merge(graph_aliases)
+      ds
+    end
+
+    private
+
+    # Transform the hash of graph aliases to an array of columns
+    def graph_alias_columns(graph_aliases)
+      graph_aliases.collect do |col_alias, tc| 
+        identifier = tc[2] || SQL::QualifiedIdentifier.new(tc[0], tc[1])
+        identifier = SQL::AliasedExpression.new(identifier, col_alias) if tc[2] or tc[1] != col_alias
+        identifier
+      end
+    end
+
+    # Fetch the rows, split them into component table parts,
+    # tranform and run the row_proc on each part (if applicable),
+    # and yield a hash of the parts.
+    def graph_each
+      # Reject tables with nil datasets, as they are excluded from
+      # the result set
+      datasets = @opts[:graph][:table_aliases].to_a.reject{|ta,ds| ds.nil?}
+      # Get just the list of table aliases into a local variable, for speed
+      table_aliases = datasets.collect{|ta,ds| ta}
+      # Get an array of arrays, one for each dataset, with
+      # the necessary information about each dataset, for speed
+      datasets = datasets.collect{|ta, ds| [ta, ds, ds.row_proc]}
+      # Use the manually set graph aliases, if any, otherwise
+      # use the ones automatically created by .graph
+      column_aliases = @opts[:graph_aliases] || @opts[:graph][:column_aliases]
+      fetch_rows(select_sql) do |r|
+        graph = {}
+        # Create the sub hashes, one per table
+        table_aliases.each{|ta| graph[ta]={}}
+        # Split the result set based on the column aliases
+        # If there are columns in the result set that are
+        # not in column_aliases, they are ignored
+        column_aliases.each do |col_alias, tc|
+          ta, column = tc
+          graph[ta][column] = r[col_alias]
+        end
+        # For each dataset run the row_proc if applicable
+        datasets.each do |ta,ds,rp|
+          g = graph[ta]
+          graph[ta] = if g.values.any?{|x| !x.nil?}
+            rp ? rp.call(g) : g
+          else
+            nil
+          end
+        end
+
+        yield graph
+      end
+      self
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/dataset/prepared_statements.rb 3.0.0-1/lib/sequel/dataset/prepared_statements.rb
--- 1.5.1-1/lib/sequel/dataset/prepared_statements.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/dataset/prepared_statements.rb	2009-03-16 20:53:19.000000000 +0000
@@ -0,0 +1,220 @@
+module Sequel 
+  class Dataset
+    PREPARED_ARG_PLACEHOLDER = LiteralString.new('?').freeze
+    
+    # Default implementation of the argument mapper to allow
+    # native database support for bind variables and prepared
+    # statements (as opposed to the emulated ones used by default).
+    module ArgumentMapper
+      SQL_QUERY_TYPE = Hash.new{|h,k| h[k] = k}
+      SQL_QUERY_TYPE[:first] = SQL_QUERY_TYPE[:all] = :select
+      
+      # The name of the prepared statement, if any.
+      attr_accessor :prepared_statement_name
+      
+      # The bind arguments to use for running this prepared statement
+      attr_accessor :bind_arguments
+      
+      # Set the bind arguments based on the hash and call super.
+      def call(hash, &block)
+        ds = clone
+        ds.prepared_sql
+        ds.bind_arguments = ds.map_to_prepared_args(hash)
+        ds.prepared_args = hash
+        ds.run(&block)
+      end
+        
+      # Override the given *_sql method based on the type, and
+      # cache the result of the sql.
+      def prepared_sql
+        return @prepared_sql if @prepared_sql
+        @prepared_args ||= []
+        @prepared_sql = super
+        meta_def("#{sql_query_type}_sql"){|*args| prepared_sql}
+        @prepared_sql
+      end
+      
+      private
+      
+      # The type of query (:select, :insert, :delete, :update).
+      def sql_query_type
+        SQL_QUERY_TYPE[@prepared_type]
+      end
+    end
+
+    # Backbone of the prepared statement support.  Grafts bind variable
+    # support into datasets by hijacking #literal and using placeholders.
+    # By default, emulates prepared statements and bind variables by
+    # taking the hash of bind variables and directly substituting them
+    # into the query, which works on all databases, as it is no different
+    # from using the dataset without bind variables.
+    module PreparedStatementMethods
+      PLACEHOLDER_RE = /\A\$(.*)\z/
+      
+      # The type of prepared statement, should be one of :select, :first,
+      # :insert, :update, or :delete
+      attr_accessor :prepared_type
+      
+      # The bind variable hash to use when substituting
+      attr_accessor :prepared_args
+      
+      # The argument to supply to insert and update, which may use
+      # placeholders specified by prepared_args
+      attr_accessor :prepared_modify_values
+      
+      # Sets the prepared_args to the given hash and runs the
+      # prepared statement.
+      def call(hash, &block)
+        ds = clone
+        ds.prepared_args = hash
+        ds.run(&block)
+      end
+      
+      # Returns the SQL for the prepared statement, depending on
+      # the type of the statement and the prepared_modify_values.
+      def prepared_sql
+        case @prepared_type
+        when :select, :all
+          select_sql
+        when :first
+          clone(:limit=>1).select_sql
+        when :insert
+          insert_sql(@prepared_modify_values)
+        when :update
+          update_sql(@prepared_modify_values)
+        when :delete
+          delete_sql
+        end
+      end
+      
+      # Changes the values of symbols if they start with $ and
+      # prepared_args is present.  If so, they are considered placeholders,
+      # and they are substituted using prepared_arg.
+      def literal(v)
+        case v
+        when Symbol
+          if match = PLACEHOLDER_RE.match(v.to_s) and @prepared_args
+            super(prepared_arg(match[1].to_sym))
+          else
+            super
+          end
+        else
+          super
+        end
+      end
+      
+      # Programmer friendly string showing this is a prepared statement,
+      # with the prepared SQL it represents (which in general won't have
+      # substituted variables).
+      def inspect
+        "<#{self.class.name}/PreparedStatement #{prepared_sql.inspect}>"
+      end
+      
+      protected
+      
+      # Run the method based on the type of prepared statement, with
+      # :select running #all to get all of the rows, and the other
+      # types running the method with the same name as the type.
+      def run(&block)
+        case @prepared_type
+        when :select, :all
+          all(&block)
+        when :first
+          first
+        when :insert
+          insert(@prepared_modify_values)
+        when :update
+          update(@prepared_modify_values)
+        when :delete
+          delete
+        end
+      end
+      
+      private
+      
+      # Returns the value of the prepared_args hash for the given key.
+      def prepared_arg(k)
+        @prepared_args[k]
+      end
+
+      # Use a clone of the dataset extended with prepared statement
+      # support and using the same argument hash so that you can use
+      # bind variables/prepared arguments in subselects.
+      def subselect_sql(ds)
+        ps = ds.prepare(:select)
+        ps.prepared_args = prepared_args
+        ps.prepared_sql
+      end
+    end
+    
+    # Default implementation for an argument mapper that uses
+    # unnumbered SQL placeholder arguments.  Keeps track of which
+    # arguments have been used, and allows arguments to
+    # be used more than once.
+    module UnnumberedArgumentMapper
+      include ArgumentMapper
+      
+      protected
+      
+      # Returns a single output array mapping the values of the input hash.
+      # Keys in the input hash that are used more than once in the query
+      # have multiple entries in the output array.
+      def map_to_prepared_args(hash)
+        @prepared_args.map{|v| hash[v]}
+      end
+      
+      private
+      
+      # Associates the argument with name k with the next position in
+      # the output array.
+      def prepared_arg(k)
+        @prepared_args << k
+        prepared_arg_placeholder
+      end
+    end
+    
+    # For the given type (:select, :insert, :update, or :delete),
+    # run the sql with the bind variables
+    # specified in the hash.  values is a hash of passed to
+    # insert or update (if one of those types is used),
+    # which may contain placeholders.
+    def call(type, bind_variables={}, values=nil)
+      prepare(type, nil, values).call(bind_variables)
+    end
+    
+    # Prepare an SQL statement for later execution. This returns
+    # a clone of the dataset extended with PreparedStatementMethods,
+    # on which you can call call with the hash of bind variables to
+    # do substitution.  The prepared statement is also stored in
+    # the associated database.  The following usage is identical:
+    #
+    #   ps = prepare(:select, :select_by_name)
+    #   ps.call(:name=>'Blah')
+    #   db.call(:select_by_name, :name=>'Blah')
+    def prepare(type, name=nil, values=nil)
+      ps = to_prepared_statement(type, values)
+      db.prepared_statements[name] = ps if name
+      ps
+    end
+    
+    protected
+    
+    # Return a cloned copy of the current dataset extended with
+    # PreparedStatementMethods, setting the type and modify values.
+    def to_prepared_statement(type, values=nil)
+      ps = clone
+      ps.extend(PreparedStatementMethods)
+      ps.prepared_type = type
+      ps.prepared_modify_values = values
+      ps
+    end
+
+    private
+    
+    # The argument placeholder.  Most databases used unnumbered
+    # arguments with question marks, so that is the default.
+    def prepared_arg_placeholder
+      PREPARED_ARG_PLACEHOLDER
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/dataset/sql.rb 3.0.0-1/lib/sequel/dataset/sql.rb
--- 1.5.1-1/lib/sequel/dataset/sql.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/dataset/sql.rb	2009-04-27 23:51:15.000000000 +0100
@@ -0,0 +1,1136 @@
+module Sequel
+  class Dataset
+    AND_SEPARATOR = " AND ".freeze
+    BOOL_FALSE = "'f'".freeze
+    BOOL_TRUE = "'t'".freeze
+    COLUMN_REF_RE1 = /\A([\w ]+)__([\w ]+)___([\w ]+)\z/.freeze
+    COLUMN_REF_RE2 = /\A([\w ]+)___([\w ]+)\z/.freeze
+    COLUMN_REF_RE3 = /\A([\w ]+)__([\w ]+)\z/.freeze
+    COUNT_FROM_SELF_OPTS = [:distinct, :group, :sql, :limit, :compounds]
+    IS_LITERALS = {nil=>'NULL'.freeze, true=>'TRUE'.freeze, false=>'FALSE'.freeze}.freeze
+    IS_OPERATORS = ::Sequel::SQL::ComplexExpression::IS_OPERATORS
+    N_ARITY_OPERATORS = ::Sequel::SQL::ComplexExpression::N_ARITY_OPERATORS
+    NULL = "NULL".freeze
+    QUESTION_MARK = '?'.freeze
+    STOCK_COUNT_OPTS = {:select => [SQL::AliasedExpression.new(LiteralString.new("COUNT(*)").freeze, :count)], :order => nil}.freeze
+    SELECT_CLAUSE_ORDER = %w'distinct columns from join where group having compounds order limit'.freeze
+    TWO_ARITY_OPERATORS = ::Sequel::SQL::ComplexExpression::TWO_ARITY_OPERATORS
+    WILDCARD = '*'.freeze
+
+    # Adds an further filter to an existing filter using AND. If no filter 
+    # exists an error is raised. This method is identical to #filter except
+    # it expects an existing filter.
+    #
+    #   ds.filter(:a).and(:b) # SQL: WHERE a AND b
+    def and(*cond, &block)
+      raise(InvalidOperation, "No existing filter found.") unless @opts[:having] || @opts[:where]
+      filter(*cond, &block)
+    end
+
+    # SQL fragment for the aliased expression
+    def aliased_expression_sql(ae)
+      as_sql(literal(ae.expression), ae.aliaz)
+    end
+
+    # SQL fragment for the SQL array.
+    def array_sql(a)
+      a.empty? ? '(NULL)' : "(#{expression_list(a)})"     
+    end
+
+    # SQL fragment for specifying given CaseExpression.
+    def case_expression_sql(ce)
+      sql = '(CASE '
+      sql << "#{literal(ce.expression)} " if ce.expression
+      ce.conditions.collect{ |c,r|
+        sql << "WHEN #{literal(c)} THEN #{literal(r)} "
+      }
+      sql << "ELSE #{literal(ce.default)} END)"
+    end
+
+    # SQL fragment for the SQL CAST expression.
+    def cast_sql(expr, type)
+      "CAST(#{literal(expr)} AS #{db.cast_type_literal(type)})"
+    end
+
+    # SQL fragment for specifying all columns in a given table.
+    def column_all_sql(ca)
+      "#{quote_schema_table(ca.table)}.*"
+    end
+
+    # SQL fragment for complex expressions
+    def complex_expression_sql(op, args)
+      case op
+      when *IS_OPERATORS
+        v = IS_LITERALS[args.at(1)] || raise(Error, 'Invalid argument used for IS operator')
+        "(#{literal(args.at(0))} #{op} #{v})"
+      when *TWO_ARITY_OPERATORS
+        "(#{literal(args.at(0))} #{op} #{literal(args.at(1))})"
+      when *N_ARITY_OPERATORS
+        "(#{args.collect{|a| literal(a)}.join(" #{op} ")})"
+      when :NOT
+        "NOT #{literal(args.at(0))}"
+      when :NOOP
+        literal(args.at(0))
+      when :'B~'
+        "~#{literal(args.at(0))}"
+      else
+        raise(Sequel::Error, "invalid operator #{op}")
+      end
+    end
+
+    # Returns the number of records in the dataset.
+    def count
+      options_overlap(COUNT_FROM_SELF_OPTS) ? from_self.count : clone(STOCK_COUNT_OPTS).single_value.to_i
+    end
+
+    # Formats a DELETE statement using the given options and dataset options.
+    # 
+    #   dataset.filter{|o| o.price >= 100}.delete_sql #=>
+    #     "DELETE FROM items WHERE (price >= 100)"
+    def delete_sql
+      opts = @opts
+
+      return static_sql(opts[:sql]) if opts[:sql]
+
+      if opts[:group]
+        raise InvalidOperation, "Grouped datasets cannot be deleted from"
+      elsif opts[:from].is_a?(Array) && opts[:from].size > 1
+        raise InvalidOperation, "Joined datasets cannot be deleted from"
+      end
+
+      sql = "DELETE FROM #{source_list(opts[:from])}"
+
+      if where = opts[:where]
+        sql << " WHERE #{literal(where)}"
+      end
+
+      sql
+    end
+
+    # Returns a copy of the dataset with the SQL DISTINCT clause.
+    # The DISTINCT clause is used to remove duplicate rows from the
+    # output.  If arguments are provided, uses a DISTINCT ON clause,
+    # in which case it will only be distinct on those columns, instead
+    # of all returned columns.
+    #
+    #  dataset.distinct # SQL: SELECT DISTINCT * FROM items
+    #  dataset.order(:id).distinct(:id) # SQL: SELECT DISTINCT ON (id) * FROM items ORDER BY id
+    def distinct(*args)
+      clone(:distinct => args)
+    end
+
+    # Adds an EXCEPT clause using a second dataset object. If all is true the
+    # clause used is EXCEPT ALL, which may return duplicate rows.
+    #
+    #   DB[:items].except(DB[:other_items]).sql
+    #   #=> "SELECT * FROM items EXCEPT SELECT * FROM other_items"
+    def except(dataset, all = false)
+      compound_clone(:except, dataset, all)
+    end
+
+    # Performs the inverse of Dataset#filter.
+    #
+    #   dataset.exclude(:category => 'software').sql #=>
+    #     "SELECT * FROM items WHERE (category != 'software')"
+    def exclude(*cond, &block)
+      clause = (@opts[:having] ? :having : :where)
+      cond = cond.first if cond.size == 1
+      cond = filter_expr(cond, &block)
+      cond = SQL::BooleanExpression.invert(cond)
+      cond = SQL::BooleanExpression.new(:AND, @opts[clause], cond) if @opts[clause]
+      clone(clause => cond)
+    end
+
+    # Returns an EXISTS clause for the dataset as a LiteralString.
+    #
+    #   DB.select(1).where(DB[:items].exists).sql
+    #   #=> "SELECT 1 WHERE EXISTS (SELECT * FROM items)"
+    def exists
+      LiteralString.new("EXISTS (#{select_sql})")
+    end
+
+    # Returns a copy of the dataset with the given conditions imposed upon it.  
+    # If the query already has a HAVING clause, then the conditions are imposed in the 
+    # HAVING clause. If not, then they are imposed in the WHERE clause.
+    # 
+    # filter accepts the following argument types:
+    #
+    # * Hash - list of equality/inclusion expressions
+    # * Array - depends:
+    #   * If first member is a string, assumes the rest of the arguments
+    #     are parameters and interpolates them into the string.
+    #   * If all members are arrays of length two, treats the same way
+    #     as a hash, except it allows for duplicate keys to be
+    #     specified.
+    # * String - taken literally
+    # * Symbol - taken as a boolean column argument (e.g. WHERE active)
+    # * Sequel::SQL::BooleanExpression - an existing condition expression,
+    #   probably created using the Sequel expression filter DSL.
+    #
+    # filter also takes a block, which should return one of the above argument
+    # types, and is treated the same way.  This block yields a virtual row object,
+    # which is easy to use to create identifiers and functions.
+    #
+    # If both a block and regular argument
+    # are provided, they get ANDed together.
+    #
+    # Examples:
+    #
+    #   dataset.filter(:id => 3).sql #=>
+    #     "SELECT * FROM items WHERE (id = 3)"
+    #   dataset.filter('price < ?', 100).sql #=>
+    #     "SELECT * FROM items WHERE price < 100"
+    #   dataset.filter([[:id, (1,2,3)], [:id, 0..10]]).sql #=>
+    #     "SELECT * FROM items WHERE ((id IN (1, 2, 3)) AND ((id >= 0) AND (id <= 10)))"
+    #   dataset.filter('price < 100').sql #=>
+    #     "SELECT * FROM items WHERE price < 100"
+    #   dataset.filter(:active).sql #=>
+    #     "SELECT * FROM items WHERE :active
+    #   dataset.filter{|o| o.price < 100}.sql #=>
+    #     "SELECT * FROM items WHERE (price < 100)"
+    # 
+    # Multiple filter calls can be chained for scoping:
+    #
+    #   software = dataset.filter(:category => 'software')
+    #   software.filter{|o| o.price < 100}.sql #=>
+    #     "SELECT * FROM items WHERE ((category = 'software') AND (price < 100))"
+    #
+    # See doc/dataset_filters.rdoc for more examples and details.
+    def filter(*cond, &block)
+      _filter(@opts[:having] ? :having : :where, *cond, &block)
+    end
+
+    # The first source (primary table) for this dataset.  If the dataset doesn't
+    # have a table, raises an error.  If the table is aliased, returns the aliased name.
+    def first_source
+      source = @opts[:from]
+      if source.nil? || source.empty?
+        raise Error, 'No source specified for query'
+      end
+      case s = source.first
+      when Hash
+        s.values.first
+      when Symbol
+        sch, table, aliaz = split_symbol(s)
+        aliaz ? aliaz.to_sym : s
+      else
+        s
+      end
+    end
+
+    # Returns a copy of the dataset with the source changed.
+    #
+    #   dataset.from # SQL: SELECT *
+    #   dataset.from(:blah) # SQL: SELECT * FROM blah
+    #   dataset.from(:blah, :foo) # SQL: SELECT * FROM blah, foo
+    def from(*source)
+      clone(:from=>source.empty? ? nil : source)
+    end
+    
+    # Returns a dataset selecting from the current dataset.
+    #
+    #   ds = DB[:items].order(:name)
+    #   ds.sql #=> "SELECT * FROM items ORDER BY name"
+    #   ds.from_self.sql #=> "SELECT * FROM (SELECT * FROM items ORDER BY name)"
+    def from_self
+      fs = {}
+      @opts.keys.each{|k| fs[k] = nil} 
+      fs[:from] = [self]
+      clone(fs)
+    end
+
+    # SQL fragment specifying an SQL function call
+    def function_sql(f)
+      args = f.args
+      "#{f.f}#{args.empty? ? '()' : literal(args)}"
+    end
+
+    # Pattern match any of the columns to any of the terms.  The terms can be
+    # strings (which use LIKE) or regular expressions (which are only supported
+    # in some databases).  See Sequel::SQL::StringExpression.like.  Note that the
+    # total number of pattern matches will be cols.length * terms.length,
+    # which could cause performance issues.
+    #
+    #   dataset.grep(:a, '%test%') # SQL: SELECT * FROM items WHERE a LIKE '%test%'
+    #   dataset.grep([:a, :b], %w'%test% foo') # SQL: SELECT * FROM items WHERE a LIKE '%test%' OR a LIKE 'foo' OR b LIKE '%test%' OR b LIKE 'foo' 
+    def grep(cols, terms)
+      filter(SQL::BooleanExpression.new(:OR, *Array(cols).collect{|c| SQL::StringExpression.like(c, *terms)}))
+    end
+
+    # Returns a copy of the dataset with the results grouped by the value of 
+    # the given columns.
+    #
+    #   dataset.group(:id) # SELECT * FROM items GROUP BY id
+    #   dataset.group(:id, :name) # SELECT * FROM items GROUP BY id, name
+    def group(*columns)
+      clone(:group => columns)
+    end
+    alias group_by group
+
+    # Returns a copy of the dataset with the HAVING conditions changed. Raises 
+    # an error if the dataset has not been grouped. See #filter for argument types.
+    #
+    #   dataset.group(:sum).having(:sum=>10) # SQL: SELECT * FROM items GROUP BY sum HAVING sum = 10 
+    def having(*cond, &block)
+      raise(InvalidOperation, "Can only specify a HAVING clause on a grouped dataset") unless @opts[:group]
+      _filter(:having, *cond, &block)
+    end
+    
+    # Inserts multiple values. If a block is given it is invoked for each
+    # item in the given array before inserting it.  See #multi_insert as
+    # a possible faster version that inserts multiple records in one
+    # SQL statement.
+    def insert_multiple(array, &block)
+      if block
+        array.each {|i| insert(block[i])}
+      else
+        array.each {|i| insert(i)}
+      end
+    end
+
+    # Formats an INSERT statement using the given values. If a hash is given,
+    # the resulting statement includes column names. If no values are given, 
+    # the resulting statement includes a DEFAULT VALUES clause.
+    #
+    #   dataset.insert_sql #=> 'INSERT INTO items DEFAULT VALUES'
+    #   dataset.insert_sql(1,2,3) #=> 'INSERT INTO items VALUES (1, 2, 3)'
+    #   dataset.insert_sql(:a => 1, :b => 2) #=>
+    #     'INSERT INTO items (a, b) VALUES (1, 2)'
+    def insert_sql(*values)
+      return static_sql(@opts[:sql]) if @opts[:sql]
+
+      from = source_list(@opts[:from])
+      case values.size
+      when 0
+        values = {}
+      when 1
+        vals = values.at(0)
+        if [Hash, Dataset, Array].any?{|c| vals.is_a?(c)}
+          values = vals
+        elsif vals.respond_to?(:values)
+          values = vals.values
+        end
+      end
+
+      case values
+      when Array
+        if values.empty?
+          insert_default_values_sql
+        else
+          "INSERT INTO #{from} VALUES #{literal(values)}"
+        end
+      when Hash
+        values = @opts[:defaults].merge(values) if @opts[:defaults]
+        values = values.merge(@opts[:overrides]) if @opts[:overrides]
+        if values.empty?
+          insert_default_values_sql
+        else
+          fl, vl = [], []
+          values.each do |k, v|
+            fl << literal(String === k ? k.to_sym : k)
+            vl << literal(v)
+          end
+          "INSERT INTO #{from} (#{fl.join(COMMA_SEPARATOR)}) VALUES (#{vl.join(COMMA_SEPARATOR)})"
+        end
+      when Dataset
+        "INSERT INTO #{from} #{literal(values)}"
+      end
+    end
+    
+    # Adds an INTERSECT clause using a second dataset object. If all is true 
+    # the clause used is INTERSECT ALL, which may return duplicate rows.
+    #
+    #   DB[:items].intersect(DB[:other_items]).sql
+    #   #=> "SELECT * FROM items INTERSECT SELECT * FROM other_items"
+    def intersect(dataset, all = false)
+      compound_clone(:intersect, dataset, all)
+    end
+
+    # Inverts the current filter
+    #
+    #   dataset.filter(:category => 'software').invert.sql #=>
+    #     "SELECT * FROM items WHERE (category != 'software')"
+    def invert
+      having, where = @opts[:having], @opts[:where]
+      raise(Error, "No current filter") unless having || where
+      o = {}
+      o[:having] = SQL::BooleanExpression.invert(having) if having
+      o[:where] = SQL::BooleanExpression.invert(where) if where
+      clone(o)
+    end
+
+    # SQL fragment specifying a JOIN clause without ON or USING.
+    def join_clause_sql(jc)
+      table = jc.table
+      table_alias = jc.table_alias
+      table_alias = nil if table == table_alias
+      tref = table_ref(table)
+      " #{join_type_sql(jc.join_type)} #{table_alias ? as_sql(tref, table_alias) : tref}"
+    end
+
+    # SQL fragment specifying a JOIN clause with ON.
+    def join_on_clause_sql(jc)
+      "#{join_clause_sql(jc)} ON #{literal(filter_expr(jc.on))}"
+    end
+
+    # SQL fragment specifying a JOIN clause with USING.
+    def join_using_clause_sql(jc)
+      "#{join_clause_sql(jc)} USING (#{column_list(jc.using)})"
+    end
+
+    # Returns a joined dataset.  Uses the following arguments:
+    #
+    # * type - The type of join to do (e.g. :inner)
+    # * table - Depends on type:
+    #   * Dataset - a subselect is performed with an alias of tN for some value of N
+    #   * Model (or anything responding to :table_name) - table.table_name
+    #   * String, Symbol: table
+    # * expr - specifies conditions, depends on type:
+    #   * Hash, Array with all two pairs - Assumes key (1st arg) is column of joined table (unless already
+    #     qualified), and value (2nd arg) is column of the last joined or primary table (or the
+    #     :implicit_qualifier option).
+    #     To specify multiple conditions on a single joined table column, you must use an array.
+    #     Uses a JOIN with an ON clause.
+    #   * Array - If all members of the array are symbols, considers them as columns and 
+    #     uses a JOIN with a USING clause.  Most databases will remove duplicate columns from
+    #     the result set if this is used.
+    #   * nil - If a block is not given, doesn't use ON or USING, so the JOIN should be a NATURAL
+    #     or CROSS join. If a block is given, uses a ON clause based on the block, see below.
+    #   * Everything else - pretty much the same as a using the argument in a call to filter,
+    #     so strings are considered literal, symbols specify boolean columns, and blockless
+    #     filter expressions can be used. Uses a JOIN with an ON clause.
+    # * options - a hash of options, with any of the following keys:
+    #   * :table_alias - the name of the table's alias when joining, necessary for joining
+    #     to the same table more than once.  No alias is used by default.
+    #   * :implicit_qualifer - The name to use for qualifying implicit conditions.  By default,
+    #     the last joined or primary table is used.
+    # * block - The block argument should only be given if a JOIN with an ON clause is used,
+    #   in which case it yields the table alias/name for the table currently being joined,
+    #   the table alias/name for the last joined (or first table), and an array of previous
+    #   SQL::JoinClause.
+    def join_table(type, table, expr=nil, options={}, &block)
+      if [Symbol, String].any?{|c| options.is_a?(c)}
+        table_alias = options
+        last_alias = nil 
+      else
+        table_alias = options[:table_alias]
+        last_alias = options[:implicit_qualifier]
+      end
+      if Dataset === table
+        if table_alias.nil?
+          table_alias_num = (@opts[:num_dataset_sources] || 0) + 1
+          table_alias = "t#{table_alias_num}"
+        end
+        table_name = table_alias
+      else
+        table = table.table_name if table.respond_to?(:table_name)
+        table_name = table_alias || table
+      end
+
+      join = if expr.nil? and !block_given?
+        SQL::JoinClause.new(type, table, table_alias)
+      elsif Array === expr and !expr.empty? and expr.all?{|x| Symbol === x}
+        raise(Sequel::Error, "can't use a block if providing an array of symbols as expr") if block_given?
+        SQL::JoinUsingClause.new(expr, type, table, table_alias)
+      else
+        last_alias ||= @opts[:last_joined_table] || (first_source.is_a?(Dataset) ? 't1' : first_source)
+        if Sequel.condition_specifier?(expr)
+          expr = expr.collect do |k, v|
+            k = qualified_column_name(k, table_name) if k.is_a?(Symbol)
+            v = qualified_column_name(v, last_alias) if v.is_a?(Symbol)
+            [k,v]
+          end
+        end
+        if block_given?
+          expr2 = yield(table_name, last_alias, @opts[:join] || [])
+          expr = expr ? SQL::BooleanExpression.new(:AND, expr, expr2) : expr2
+        end
+        SQL::JoinOnClause.new(expr, type, table, table_alias)
+      end
+
+      opts = {:join => (@opts[:join] || []) + [join], :last_joined_table => table_name}
+      opts[:num_dataset_sources] = table_alias_num if table_alias_num
+      clone(opts)
+    end
+
+    # If given an integer, the dataset will contain only the first l results.
+    # If given a range, it will contain only those at offsets within that
+    # range. If a second argument is given, it is used as an offset.
+    #
+    #   dataset.limit(10) # SQL: SELECT * FROM items LIMIT 10
+    #   dataset.limit(10, 20) # SQL: SELECT * FROM items LIMIT 10 OFFSET 20
+    def limit(l, o = nil)
+      return from_self.limit(l, o) if @opts[:sql]
+
+      if Range === l
+        o = l.first
+        l = l.last - l.first + (l.exclude_end? ? 0 : 1)
+      end
+      l = l.to_i
+      raise(Error, 'Limits must be greater than or equal to 1') unless l >= 1
+      opts = {:limit => l}
+      if o
+        o = o.to_i
+        raise(Error, 'Offsets must be greater than or equal to 0') unless o >= 0
+        opts[:offset] = o
+      end
+      clone(opts)
+    end
+    
+    # Returns a literal representation of a value to be used as part
+    # of an SQL expression. 
+    # 
+    #   dataset.literal("abc'def\\") #=> "'abc''def\\\\'"
+    #   dataset.literal(:items__id) #=> "items.id"
+    #   dataset.literal([1, 2, 3]) => "(1, 2, 3)"
+    #   dataset.literal(DB[:items]) => "(SELECT * FROM items)"
+    #   dataset.literal(:x + 1 > :y) => "((x + 1) > y)"
+    #
+    # If an unsupported object is given, an exception is raised.
+    def literal(v)
+      case v
+      when String
+        return v if v.is_a?(LiteralString)
+        v.is_a?(SQL::Blob) ? literal_blob(v) : literal_string(v)
+      when Symbol
+        literal_symbol(v)
+      when Integer
+        literal_integer(v)
+      when Hash
+        literal_hash(v)
+      when SQL::Expression
+        literal_expression(v)
+      when Float
+        literal_float(v)
+      when BigDecimal
+        literal_big_decimal(v)
+      when NilClass
+        NULL
+      when TrueClass
+        literal_true
+      when FalseClass
+        literal_false
+      when Array
+        literal_array(v)
+      when Time
+        literal_time(v)
+      when DateTime
+        literal_datetime(v)
+      when Date
+        literal_date(v)
+      when Dataset
+        literal_dataset(v)
+      else
+        literal_other(v)
+      end
+    end
+
+    # Returns an array of insert statements for inserting multiple records.
+    # This method is used by #multi_insert to format insert statements and
+    # expects a keys array and and an array of value arrays.
+    #
+    # This method should be overridden by descendants if the support
+    # inserting multiple records in a single SQL statement.
+    def multi_insert_sql(columns, values)
+      s = "INSERT INTO #{source_list(@opts[:from])} (#{identifier_list(columns)}) VALUES "
+      values.map{|r| s + literal(r)}
+    end
+    
+    # Adds an alternate filter to an existing filter using OR. If no filter 
+    # exists an error is raised.
+    #
+    #   dataset.filter(:a).or(:b) # SQL: SELECT * FROM items WHERE a OR b
+    def or(*cond, &block)
+      clause = (@opts[:having] ? :having : :where)
+      raise(InvalidOperation, "No existing filter found.") unless @opts[clause]
+      cond = cond.first if cond.size == 1
+      clone(clause => SQL::BooleanExpression.new(:OR, @opts[clause], filter_expr(cond, &block)))
+    end
+
+    # Returns a copy of the dataset with the order changed. If a nil is given
+    # the returned dataset has no order. This can accept multiple arguments
+    # of varying kinds, and even SQL functions.  If a block is given, it is treated
+    # as a virtual row block, similar to filter.
+    #
+    #   ds.order(:name).sql #=> 'SELECT * FROM items ORDER BY name'
+    #   ds.order(:a, :b).sql #=> 'SELECT * FROM items ORDER BY a, b'
+    #   ds.order('a + b'.lit).sql #=> 'SELECT * FROM items ORDER BY a + b'
+    #   ds.order(:a + :b).sql #=> 'SELECT * FROM items ORDER BY (a + b)'
+    #   ds.order(:name.desc).sql #=> 'SELECT * FROM items ORDER BY name DESC'
+    #   ds.order(:name.asc).sql #=> 'SELECT * FROM items ORDER BY name ASC'
+    #   ds.order{|o| o.sum(:name)}.sql #=> 'SELECT * FROM items ORDER BY sum(name)'
+    #   ds.order(nil).sql #=> 'SELECT * FROM items'
+    def order(*columns, &block)
+      columns += Array(virtual_row_block_call(block)) if block
+      clone(:order => (columns.compact.empty?) ? nil : columns)
+    end
+    alias_method :order_by, :order
+    
+    # Returns a copy of the dataset with the order columns added
+    # to the existing order.
+    #
+    #   ds.order(:a).order(:b).sql #=> 'SELECT * FROM items ORDER BY b'
+    #   ds.order(:a).order_more(:b).sql #=> 'SELECT * FROM items ORDER BY a, b'
+    def order_more(*columns, &block)
+      order(*Array(@opts[:order]).concat(columns), &block)
+    end
+    
+    # SQL fragment for the ordered expression, used in the ORDER BY
+    # clause.
+    def ordered_expression_sql(oe)
+      "#{literal(oe.expression)} #{oe.descending ? 'DESC' : 'ASC'}"
+    end
+
+    # SQL fragment for a literal string with placeholders
+    def placeholder_literal_string_sql(pls)
+      args = pls.args.dup
+      s = pls.str.gsub(QUESTION_MARK){literal(args.shift)}
+      s = "(#{s})" if pls.parens
+      s
+    end
+
+    # SQL fragment for the qualifed identifier, specifying
+    # a table and a column (or schema and table).
+    def qualified_identifier_sql(qcr)
+      [qcr.table, qcr.column].map{|x| [SQL::QualifiedIdentifier, SQL::Identifier, Symbol].any?{|c| x.is_a?(c)} ? literal(x) : quote_identifier(x)}.join('.')
+    end
+
+    # Adds quoting to identifiers (columns and tables). If identifiers are not
+    # being quoted, returns name as a string.  If identifiers are being quoted
+    # quote the name with quoted_identifier.
+    def quote_identifier(name)
+      return name if name.is_a?(LiteralString)
+      name = name.value if name.is_a?(SQL::Identifier)
+      name = input_identifier(name)
+      name = quoted_identifier(name) if quote_identifiers?
+      name
+    end
+
+    # Separates the schema from the table and returns a string with them
+    # quoted (if quoting identifiers)
+    def quote_schema_table(table)
+      schema, table = schema_and_table(table)
+      "#{"#{quote_identifier(schema)}." if schema}#{quote_identifier(table)}"
+    end
+
+    # This method quotes the given name with the SQL standard double quote. 
+    # should be overridden by subclasses to provide quoting not matching the
+    # SQL standard, such as backtick (used by MySQL and SQLite).
+    def quoted_identifier(name)
+      "\"#{name.to_s.gsub('"', '""')}\""
+    end
+
+    # Returns a copy of the dataset with the order reversed. If no order is
+    # given, the existing order is inverted.
+    def reverse_order(*order)
+      order(*invert_order(order.empty? ? @opts[:order] : order))
+    end
+    alias reverse reverse_order
+
+    # Split the schema information from the table
+    def schema_and_table(table_name)
+      sch = db.default_schema if db
+      case table_name
+      when Symbol
+        s, t, a = split_symbol(table_name)
+        [s||sch, t]
+      when SQL::QualifiedIdentifier
+        [table_name.table, table_name.column]
+      when SQL::Identifier
+        [sch, table_name.value]
+      when String
+        [sch, table_name]
+      else
+        raise Error, 'table_name should be a Symbol, SQL::QualifiedIdentifier, SQL::Identifier, or String'
+      end
+    end
+
+    # Returns a copy of the dataset with the columns selected changed
+    # to the given columns. This also takes a virtual row block,
+    # similar to filter.
+    #
+    #   dataset.select(:a) # SELECT a FROM items
+    #   dataset.select(:a, :b) # SELECT a, b FROM items
+    #   dataset.select{|o| o.a, o.sum(:b)} # SELECT a, sum(b) FROM items
+    def select(*columns, &block)
+      columns += Array(virtual_row_block_call(block)) if block
+      clone(:select => columns)
+    end
+    
+    # Returns a copy of the dataset selecting the wildcard.
+    #
+    #   dataset.select(:a).select_all # SELECT * FROM items
+    def select_all
+      clone(:select => nil)
+    end
+
+    # Returns a copy of the dataset with the given columns added
+    # to the existing selected columns.
+    #
+    #   dataset.select(:a).select(:b) # SELECT b FROM items
+    #   dataset.select(:a).select_more(:b) # SELECT a, b FROM items
+    def select_more(*columns, &block)
+      select(*Array(@opts[:select]).concat(columns), &block)
+    end
+    
+    # Formats a SELECT statement
+    #
+    #   dataset.select_sql # => "SELECT * FROM items"
+    def select_sql
+      return static_sql(@opts[:sql]) if @opts[:sql]
+      sql = 'SELECT'
+      select_clause_order.each{|x| send(:"select_#{x}_sql", sql)}
+      sql
+    end
+
+    # Same as select_sql, not aliased directly to make subclassing simpler.
+    def sql
+      select_sql
+    end
+
+    # SQL fragment for specifying subscripts (SQL arrays)
+    def subscript_sql(s)
+      "#{literal(s.f)}[#{s.sub.join(COMMA_SEPARATOR)}]"
+    end
+
+    # Returns a copy of the dataset with no filters (HAVING or WHERE clause) applied.
+    # 
+    #   dataset.group(:a).having(:a=>1).where(:b).unfiltered # SELECT * FROM items
+    def unfiltered
+      clone(:where => nil, :having => nil)
+    end
+
+    # Adds a UNION clause using a second dataset object. If all is true the
+    # clause used is UNION ALL, which may return duplicate rows.
+    #
+    #   DB[:items].union(DB[:other_items]).sql
+    #   #=> "SELECT * FROM items UNION SELECT * FROM other_items"
+    def union(dataset, all = false)
+      compound_clone(:union, dataset, all)
+    end
+
+    # Returns a copy of the dataset with no order.
+    # 
+    #   dataset.order(:a).unordered # SELECT * FROM items
+    def unordered
+      order(nil)
+    end
+
+    # Formats an UPDATE statement using the given values.
+    #
+    #   dataset.update_sql(:price => 100, :category => 'software') #=>
+    #     "UPDATE items SET price = 100, category = 'software'"
+    #
+    # Raises an error if the dataset is grouped or includes more
+    # than one table.
+    def update_sql(values = {})
+      opts = @opts
+
+      return static_sql(opts[:sql]) if opts[:sql]
+
+      if opts[:group]
+        raise InvalidOperation, "A grouped dataset cannot be updated"
+      elsif (opts[:from].size > 1) or opts[:join]
+        raise InvalidOperation, "A joined dataset cannot be updated"
+      end
+      
+      sql = "UPDATE #{source_list(@opts[:from])} SET "
+      set = if values.is_a?(Hash)
+        values = opts[:defaults].merge(values) if opts[:defaults]
+        values = values.merge(opts[:overrides]) if opts[:overrides]
+        # get values from hash
+        values.map do |k, v|
+          "#{[String, Symbol].any?{|c| k.is_a?(c)} ? quote_identifier(k) : literal(k)} = #{literal(v)}"
+        end.join(COMMA_SEPARATOR)
+      else
+        # copy values verbatim
+        values
+      end
+      sql << set
+      if where = opts[:where]
+        sql << " WHERE #{literal(where)}"
+      end
+
+      sql
+    end
+
+    # Add a condition to the WHERE clause.  See #filter for argument types.
+    #
+    #   dataset.group(:a).having(:a).filter(:b) # SELECT * FROM items GROUP BY a HAVING a AND b
+    #   dataset.group(:a).having(:a).where(:b) # SELECT * FROM items WHERE b GROUP BY a HAVING a
+    def where(*cond, &block)
+      _filter(:where, *cond, &block)
+    end
+
+    # Returns a copy of the dataset with the static SQL used.  This is useful if you want
+    # to keep the same row_proc/graph, but change the SQL used to custom SQL.
+    #
+    #   dataset.with_sql('SELECT * FROM foo') # SELECT * FROM foo
+    def with_sql(sql, *args)
+      sql = SQL::PlaceholderLiteralString.new(sql, args) unless args.empty?
+      clone(:sql=>sql)
+    end
+
+    [:inner, :full_outer, :right_outer, :left_outer].each do |jtype|
+      class_eval("def #{jtype}_join(*args, &block); join_table(:#{jtype}, *args, &block) end")
+    end
+    alias join inner_join
+
+    protected
+
+    # Returns a table reference for use in the FROM clause.  Returns an SQL subquery
+    # frgament with an optional table alias.
+    def to_table_reference(table_alias=nil)
+      s = "(#{sql})"
+      table_alias ? as_sql(s, table_alias) : s
+    end
+
+    private
+
+    # Internal filter method so it works on either the having or where clauses.
+    def _filter(clause, *cond, &block)
+      cond = cond.first if cond.size == 1
+      cond = filter_expr(cond, &block)
+      cond = SQL::BooleanExpression.new(:AND, @opts[clause], cond) if @opts[clause]
+      clone(clause => cond)
+    end
+
+    # SQL fragment for specifying an alias.  expression should already be literalized.
+    def as_sql(expression, aliaz)
+      "#{expression} AS #{quote_identifier(aliaz)}"
+    end
+
+    # Converts an array of column names into a comma seperated string of 
+    # column names. If the array is empty, a wildcard (*) is returned.
+    def column_list(columns)
+      if columns.nil? || columns.empty?
+        WILDCARD
+      else
+        m = columns.map do |i|
+          i.is_a?(Hash) ? i.map{|k, v| as_sql(literal(k), v)} : literal(i)
+        end
+        m.join(COMMA_SEPARATOR)
+      end
+    end
+    
+    # Add the dataset to the list of compounds
+    def compound_clone(type, dataset, all)
+      clone(:compounds=>Array(@opts[:compounds]).map{|x| x.dup} + [[type, dataset, all]])
+    end
+    
+    # Converts an array of expressions into a comma separated string of
+    # expressions.
+    def expression_list(columns)
+      columns.map{|i| literal(i)}.join(COMMA_SEPARATOR)
+    end
+    
+    # SQL fragment based on the expr type.  See #filter.
+    def filter_expr(expr = nil, &block)
+      expr = nil if expr == []
+      if expr && block
+        return SQL::BooleanExpression.new(:AND, filter_expr(expr), filter_expr(block))
+      elsif block
+        expr = block
+      end
+      case expr
+      when Hash
+        SQL::BooleanExpression.from_value_pairs(expr)
+      when Array
+        if String === expr[0]
+          SQL::PlaceholderLiteralString.new(expr.shift, expr, true)
+        elsif Sequel.condition_specifier?(expr)
+          SQL::BooleanExpression.from_value_pairs(expr)
+        else
+          SQL::BooleanExpression.new(:AND, *expr.map{|x| filter_expr(x)})
+        end
+      when Proc
+        filter_expr(virtual_row_block_call(expr))
+      when SQL::NumericExpression, SQL::StringExpression
+        raise(Error, "Invalid SQL Expression type: #{expr.inspect}") 
+      when Symbol, SQL::Expression
+        expr
+      when TrueClass, FalseClass
+        SQL::BooleanExpression.new(:NOOP, expr)
+      when String
+        LiteralString.new("(#{expr})")
+      else
+        raise(Error, 'Invalid filter argument')
+      end
+    end
+
+    # SQL fragment specifying a list of identifiers
+    def identifier_list(columns)
+      columns.map{|i| quote_identifier(i)}.join(COMMA_SEPARATOR)
+    end
+
+    # SQL statement for formatting an insert statement with default values
+    def insert_default_values_sql
+      "INSERT INTO #{source_list(@opts[:from])} DEFAULT VALUES"
+    end
+
+    # Inverts the given order by breaking it into a list of column references
+    # and inverting them.
+    #
+    #   dataset.invert_order([:id.desc]]) #=> [:id]
+    #   dataset.invert_order(:category, :price.desc]) #=>
+    #     [:category.desc, :price]
+    def invert_order(order)
+      return nil unless order
+      new_order = []
+      order.map do |f|
+        case f
+        when SQL::OrderedExpression
+          SQL::OrderedExpression.new(f.expression, !f.descending)
+        else
+          SQL::OrderedExpression.new(f)
+        end
+      end
+    end
+    
+    # SQL fragment specifying a JOIN type, converts underscores to
+    # spaces and upcases.
+    def join_type_sql(join_type)
+      "#{join_type.to_s.gsub('_', ' ').upcase} JOIN"
+    end
+
+    # SQL fragment for Array.  Treats as an expression if an array of all two pairs, or as a SQL array otherwise.
+    def literal_array(v)
+      Sequel.condition_specifier?(v) ? literal_expression(SQL::BooleanExpression.from_value_pairs(v)) : array_sql(v)
+    end
+
+    # SQL fragment for BigDecimal
+    def literal_big_decimal(v)
+      d = v.to_s("F")
+      v.nan? || v.infinite? ?  "'#{d}'" : d
+    end
+
+    # SQL fragment for SQL::Blob
+    def literal_blob(v)
+      literal_string(v)
+    end
+
+    # SQL fragment for Dataset.  Does a subselect inside parantheses.
+    def literal_dataset(v)
+      "(#{subselect_sql(v)})"
+    end
+
+    # SQL fragment for Date, using the ISO8601 format.
+    def literal_date(v)
+      "'#{v}'"
+    end
+
+    # SQL fragment for DateTime, using the ISO8601 format.
+    def literal_datetime(v)
+      "'#{v}'"
+    end
+
+    # SQL fragment for SQL::Expression, result depends on the specific type of expression.
+    def literal_expression(v)
+      v.to_s(self)
+    end
+
+    # SQL fragment for false
+    def literal_false
+      BOOL_FALSE
+    end
+
+    # SQL fragment for Float
+    def literal_float(v)
+      v.to_s
+    end
+
+    # SQL fragment for Hash, treated as an expression
+    def literal_hash(v)
+      literal_expression(SQL::BooleanExpression.from_value_pairs(v))
+    end
+
+    # SQL fragment for Integer
+    def literal_integer(v)
+      v.to_s
+    end
+
+    # SQL fragmento for a type of object not handled by Dataset#literal.  Raises an error.  If a database specific type is allowed, this should be overriden in a subclass.
+    def literal_other(v)
+      raise Error, "can't express #{v.inspect} as a SQL literal"
+    end
+
+    # SQL fragment for String.  Doubles \ and ' by default.
+    def literal_string(v)
+      "'#{v.gsub(/\\/, "\\\\\\\\").gsub(/'/, "''")}'"
+    end
+
+    # Converts a symbol into a column name. This method supports underscore
+    # notation in order to express qualified (two underscores) and aliased
+    # (three underscores) columns:
+    #
+    #   dataset.literal(:abc) #=> "abc"
+    #   dataset.literal(:abc___a) #=> "abc AS a"
+    #   dataset.literal(:items__abc) #=> "items.abc"
+    #   dataset.literal(:items__abc___a) #=> "items.abc AS a"
+    def literal_symbol(v)
+      c_table, column, c_alias = split_symbol(v)
+      qc = "#{"#{quote_identifier(c_table)}." if c_table}#{quote_identifier(column)}"
+      c_alias ? as_sql(qc, c_alias) : qc
+    end
+
+    # SQL fragment for Time, uses the ISO8601 format.
+    def literal_time(v)
+      "'#{v.iso8601}'"
+    end
+
+    # SQL fragment for true.
+    def literal_true
+      BOOL_TRUE
+    end
+
+    # Returns a qualified column name (including a table name) if the column
+    # name isn't already qualified.
+    def qualified_column_name(column, table)
+      if Symbol === column 
+        c_table, column, c_alias = split_symbol(column)
+        unless c_table
+          case table
+          when Symbol
+            schema, table, t_alias = split_symbol(table)
+            t_alias ||= Sequel::SQL::QualifiedIdentifier.new(schema, table) if schema
+          when Sequel::SQL::AliasedExpression
+            t_alias = table.aliaz
+          end
+          c_table = t_alias || table
+        end
+        ::Sequel::SQL::QualifiedIdentifier.new(c_table, column)
+      else
+        column
+      end
+    end
+
+    # The order of methods to call to build the SELECT SQL statement
+    def select_clause_order
+      SELECT_CLAUSE_ORDER
+    end
+
+    # Modify the sql to add the columns selected
+    def select_columns_sql(sql)
+      sql << " #{column_list(@opts[:select])}"
+    end
+
+    # Modify the sql to add the DISTINCT modifier
+    def select_distinct_sql(sql)
+      if distinct = @opts[:distinct]
+        sql << " DISTINCT#{" ON (#{expression_list(distinct)})" unless distinct.empty?}"
+      end
+    end
+
+    # Modify the sql to add a dataset to the via an EXCEPT, INTERSECT, or UNION clause.
+    # This uses a subselect for the compound datasets used, because using parantheses doesn't
+    # work on all databases.  I consider this an ugly hack, but can't I think of a better default.
+    def select_compounds_sql(sql)
+      return unless @opts[:compounds]
+      @opts[:compounds].each do |type, dataset, all|
+        compound_sql = subselect_sql(dataset)
+        compound_sql = "SELECT * FROM (#{compound_sql})" if dataset.opts[:compounds]
+        sql.replace("#{sql} #{type.to_s.upcase}#{' ALL' if all} #{compound_sql}")
+      end
+    end
+
+    # Modify the sql to add the list of tables to select FROM
+    def select_from_sql(sql)
+      sql << " FROM #{source_list(@opts[:from])}" if @opts[:from]
+    end
+
+    # Modify the sql to add the expressions to GROUP BY
+    def select_group_sql(sql)
+      sql << " GROUP BY #{expression_list(@opts[:group])}" if @opts[:group]
+    end
+
+    # Modify the sql to add the filter criteria in the HAVING clause
+    def select_having_sql(sql)
+      sql << " HAVING #{literal(@opts[:having])}" if @opts[:having]
+    end
+
+    # Modify the sql to add the list of tables to JOIN to
+    def select_join_sql(sql)
+      @opts[:join].each{|j| sql << literal(j)} if @opts[:join]
+    end
+
+    # Modify the sql to limit the number of rows returned and offset
+    def select_limit_sql(sql)
+      sql << " LIMIT #{@opts[:limit]}" if @opts[:limit]
+      sql << " OFFSET #{@opts[:offset]}" if @opts[:offset]
+    end
+
+    # Modify the sql to add the expressions to ORDER BY
+    def select_order_sql(sql)
+      sql << " ORDER BY #{expression_list(@opts[:order])}" if @opts[:order]
+    end
+
+    # Modify the sql to add the filter criteria in the WHERE clause
+    def select_where_sql(sql)
+      sql << " WHERE #{literal(@opts[:where])}" if @opts[:where]
+    end
+
+    # Converts an array of source names into into a comma separated list.
+    def source_list(source)
+      if source.nil? || source.empty?
+        raise Error, 'No source specified for query'
+      end
+      auto_alias_count = @opts[:num_dataset_sources] || 0
+      m = source.map do |s|
+        case s
+        when Dataset
+          auto_alias_count += 1
+          s.to_table_reference("t#{auto_alias_count}")
+        else
+          table_ref(s)
+        end
+      end
+      m.join(COMMA_SEPARATOR)
+    end
+    
+    # Splits the symbol into three parts.  Each part will
+    # either be a string or nil.
+    #
+    # For columns, these parts are the table, column, and alias.
+    # For tables, these parts are the schema, table, and alias.
+    def split_symbol(sym)
+      s = sym.to_s
+      if m = COLUMN_REF_RE1.match(s)
+        m[1..3]
+      elsif m = COLUMN_REF_RE2.match(s)
+        [nil, m[1], m[2]]
+      elsif m = COLUMN_REF_RE3.match(s)
+        [m[1], m[2], nil]
+      else
+        [nil, s, nil]
+      end
+    end
+
+    # SQL to use if this dataset uses static SQL.  Since static SQL
+    # can be a PlaceholderLiteralString in addition to a String,
+    # we literalize nonstrings.
+    def static_sql(sql)
+      sql.is_a?(String) ? sql : literal(sql)
+    end
+
+    # SQL fragment for a subselect using the given database's SQL.
+    def subselect_sql(ds)
+      ds.sql
+    end
+
+    # SQL fragment specifying a table name.
+    def table_ref(t)
+      case t
+      when Symbol
+        literal_symbol(t)
+      when Dataset
+        t.to_table_reference
+      when Hash
+        t.map{|k, v| as_sql(table_ref(k), v)}.join(COMMA_SEPARATOR)
+      when String
+        quote_identifier(t)
+      else
+        literal(t)
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/dataset.rb 3.0.0-1/lib/sequel/dataset.rb
--- 1.5.1-1/lib/sequel/dataset.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/dataset.rb	2009-04-09 20:55:38.000000000 +0100
@@ -0,0 +1,326 @@
+module Sequel
+  # A dataset represents an SQL query, or more generally, an abstract
+  # set of rows in the database.  Datasets
+  # can be used to create, retrieve, update and delete records.
+  # 
+  # Query results are always retrieved on demand, so a dataset can be kept
+  # around and reused indefinitely (datasets never cache results):
+  #
+  #   my_posts = DB[:posts].filter(:author => 'david') # no records are retrieved
+  #   my_posts.all # records are retrieved
+  #   my_posts.all # records are retrieved again
+  #
+  # Most dataset methods return modified copies of the dataset (functional style), so you can
+  # reuse different datasets to access data:
+  #
+  #   posts = DB[:posts]
+  #   davids_posts = posts.filter(:author => 'david')
+  #   old_posts = posts.filter('stamp < ?', Date.today - 7)
+  #   davids_old_posts = davids_posts.filter('stamp < ?', Date.today - 7)
+  #
+  # Datasets are Enumerable objects, so they can be manipulated using any
+  # of the Enumerable methods, such as map, inject, etc.
+  #
+  # === Methods added via metaprogramming
+  #
+  # Some methods are added via metaprogramming:
+  #
+  # * ! methods - These methods are the same as their non-! counterparts,
+  #   but they modify the receiver instead of returning a modified copy
+  #   of the dataset.
+  # * inner_join, full_outer_join, right_outer_join, left_outer_join - 
+  #   This methods are shortcuts to join_table with the join type
+  #   already specified.
+  class Dataset
+    extend Metaprogramming
+    include Metaprogramming
+    include Enumerable
+    
+    # The dataset options that require the removal of cached columns
+    # if changed.
+    COLUMN_CHANGE_OPTS = [:select, :sql, :from, :join].freeze
+
+    # All methods that should have a ! method added that modifies
+    # the receiver.
+    MUTATION_METHODS = %w'add_graph_aliases and distinct exclude exists
+    filter from from_self full_outer_join graph
+    group group_and_count group_by having inner_join intersect invert join
+    left_outer_join limit naked or order order_by order_more paginate query reject
+    reverse reverse_order right_outer_join select select_all select_more
+    set_defaults set_graph_aliases set_overrides sort sort_by
+    unfiltered union unordered where with_sql'.collect{|x| x.to_sym}
+
+    NOTIMPL_MSG = "This method must be overridden in Sequel adapters".freeze
+
+    # The database that corresponds to this dataset
+    attr_accessor :db
+    
+    # Set the method to call on identifiers going into the database for this dataset
+    attr_accessor :identifier_input_method
+    
+    # Set the method to call on identifiers coming the database for this dataset
+    attr_accessor :identifier_output_method
+
+    # The hash of options for this dataset, keys are symbols.
+    attr_accessor :opts
+
+    # Whether to quote identifiers for this dataset
+    attr_writer :quote_identifiers
+    
+    # The row_proc for this database, should be a Proc that takes
+    # a single hash argument and returns the object you want
+    # each to return.
+    attr_accessor :row_proc
+
+    # Constructs a new Dataset instance with an associated database and 
+    # options. Datasets are usually constructed by invoking the Database#[] method:
+    #
+    #   DB[:posts]
+    #
+    # Sequel::Dataset is an abstract class that is not useful by itself. Each
+    # database adaptor should provide a subclass of Sequel::Dataset, and have
+    # the Database#dataset method return an instance of that class.
+    def initialize(db, opts = nil)
+      @db = db
+      @quote_identifiers = db.quote_identifiers? if db.respond_to?(:quote_identifiers?)
+      @identifier_input_method = db.identifier_input_method if db.respond_to?(:identifier_input_method)
+      @identifier_output_method = db.identifier_output_method if db.respond_to?(:identifier_output_method)
+      @opts = opts || {}
+      @row_proc = nil
+    end
+    
+    ### Class Methods ###
+
+    # Setup mutation (e.g. filter!) methods.  These operate the same as the
+    # non-! methods, but replace the options of the current dataset with the
+    # options of the resulting dataset.
+    def self.def_mutation_method(*meths)
+      meths.each do |meth|
+        class_eval("def #{meth}!(*args, &block); mutation_method(:#{meth}, *args, &block) end")
+      end
+    end
+
+    ### Instance Methods ###
+
+    # Alias for insert, but not aliased directly so subclasses
+    # don't have to override both methods.
+    def <<(*args)
+      insert(*args)
+    end
+
+    # Return the dataset as an aliased expression with the given alias. You can
+    # use this as a FROM or JOIN dataset, or as a column if this dataset
+    # returns a single row and column.
+    def as(aliaz)
+      ::Sequel::SQL::AliasedExpression.new(self, aliaz)
+    end
+
+    # Returns an array with all records in the dataset. If a block is given,
+    # the array is iterated over after all items have been loaded.
+    def all(&block)
+      a = []
+      each{|r| a << r}
+      post_load(a)
+      a.each(&block) if block
+      a
+    end
+  
+    # Returns a new clone of the dataset with with the given options merged.
+    # If the options changed include options in COLUMN_CHANGE_OPTS, the cached
+    # columns are deleted.
+    def clone(opts = {})
+      c = super()
+      c.opts = @opts.merge(opts)
+      c.instance_variable_set(:@columns, nil) if opts.keys.any?{|o| COLUMN_CHANGE_OPTS.include?(o)}
+      c
+    end
+    
+    # Returns the columns in the result set in order.
+    # If the columns are currently cached, returns the cached value. Otherwise,
+    # a SELECT query is performed to get a single row. Adapters are expected
+    # to fill the columns cache with the column information when a query is performed.
+    # If the dataset does not have any rows, this may be an empty array depending on how
+    # the adapter is programmed.
+    #
+    # If you are looking for all columns for a single table and maybe some information about
+    # each column (e.g. type), see Database#schema.
+    def columns
+      return @columns if @columns
+      ds = unfiltered.unordered.clone(:distinct => nil, :limit => 1)
+      ds.each{break}
+      @columns = ds.instance_variable_get(:@columns)
+      @columns || []
+    end
+    
+    # Remove the cached list of columns and do a SELECT query to find
+    # the columns.
+    def columns!
+      @columns = nil
+      columns
+    end
+    
+    # Add a mutation method to this dataset instance.
+    def def_mutation_method(*meths)
+      meths.each do |meth|
+        instance_eval("def #{meth}!(*args, &block); mutation_method(:#{meth}, *args, &block) end")
+      end
+    end
+
+    # Deletes the records in the dataset.  The returned value is generally the
+    # number of records deleted, but that is adapter dependent.
+    def delete
+      execute_dui(delete_sql)
+    end
+    
+    # Iterates over the records in the dataset as they are yielded from the
+    # database adapter, and returns self.
+    def each(&block)
+      if @opts[:graph]
+        graph_each(&block)
+      else
+        if row_proc = @row_proc
+          fetch_rows(select_sql){|r| yield row_proc.call(r)}
+        else
+          fetch_rows(select_sql, &block)
+        end
+      end
+      self
+    end
+
+    # Executes a select query and fetches records, passing each record to the
+    # supplied block.  The yielded records should be hashes with symbol keys.
+    def fetch_rows(sql, &block)
+      raise NotImplementedError, NOTIMPL_MSG
+    end
+  
+    # Inserts values into the associated table.  The returned value is generally
+    # the value of the primary key for the inserted row, but that is adapter dependent.
+    def insert(*values)
+      execute_insert(insert_sql(*values))
+    end
+  
+    # Returns a string representation of the dataset including the class name 
+    # and the corresponding SQL select statement.
+    def inspect
+      "#<#{self.class}: #{sql.inspect}>"
+    end
+
+    # Returns a naked dataset clone - i.e. a dataset that returns records as
+    # hashes instead of calling the row proc.
+    def naked
+      ds = clone
+      ds.row_proc = nil
+      ds
+    end
+    
+    # Whether this dataset quotes identifiers.
+    def quote_identifiers?
+      @quote_identifiers
+    end
+    
+    # Set the server for this dataset to use.  Used to pick a specific database
+    # shard to run a query against, or to override the default (which is SELECT uses
+    # :read_only database and all other queries use the :default database).
+    def server(servr)
+      clone(:server=>servr)
+    end
+
+    # Alias for set, but not aliased directly so subclasses
+    # don't have to override both methods.
+    def set(*args)
+      update(*args)
+    end
+
+    # Set the default values for insert and update statements.  The values passed
+    # to insert or update are merged into this hash.
+    def set_defaults(hash)
+      clone(:defaults=>(@opts[:defaults]||{}).merge(hash))
+    end
+
+    # Set values that override hash arguments given to insert and update statements.
+    # This hash is merged into the hash provided to insert or update.
+    def set_overrides(hash)
+      clone(:overrides=>hash.merge(@opts[:overrides]||{}))
+    end
+
+    # Updates values for the dataset.  The returned value is generally the
+    # number of rows updated, but that is adapter dependent.
+    def update(values={})
+      execute_dui(update_sql(values))
+    end
+  
+    # Add the mutation methods via metaprogramming
+    def_mutation_method(*MUTATION_METHODS)
+
+    protected
+
+    # Return true if the dataset has a non-nil value for any key in opts.
+    def options_overlap(opts)
+      !(@opts.collect{|k,v| k unless v.nil?}.compact & opts).empty?
+    end
+
+    private
+    
+    # Set the server to use to :default unless it is already set in the passed opts
+    def default_server_opts(opts)
+      {:server=>@opts[:server] || :default}.merge(opts)
+    end
+
+    # Execute the given SQL on the database using execute.
+    def execute(sql, opts={}, &block)
+      @db.execute(sql, {:server=>@opts[:server] || :read_only}.merge(opts), &block)
+    end
+    
+    # Execute the given SQL on the database using execute_dui.
+    def execute_dui(sql, opts={}, &block)
+      @db.execute_dui(sql, default_server_opts(opts), &block)
+    end
+    
+    # Execute the given SQL on the database using execute_insert.
+    def execute_insert(sql, opts={}, &block)
+      @db.execute_insert(sql, default_server_opts(opts), &block)
+    end
+    
+    # Modify the identifier returned from the database based on the
+    # identifier_output_method.
+    def input_identifier(v)
+      (i = identifier_input_method) ? v.to_s.send(i) : v.to_s
+    end
+
+    # Modify the receiver with the results of sending the meth, args, and block
+    # to the receiver and merging the options of the resulting dataset into
+    # the receiver's options.
+    def mutation_method(meth, *args, &block)
+      copy = send(meth, *args, &block)
+      @opts.merge!(copy.opts)
+      self
+    end
+    
+    # Modify the identifier returned from the database based on the
+    # identifier_output_method.
+    def output_identifier(v)
+      (i = identifier_output_method) ? v.to_s.send(i).to_sym : v.to_sym
+    end
+
+    # This is run inside .all, after all of the records have been loaded
+    # via .each, but before any block passed to all is called.  It is called with
+    # a single argument, an array of all returned records.  Does nothing by
+    # default, added to make the model eager loading code simpler.
+    def post_load(all_records)
+    end
+
+    # If a block argument is passed to a method that uses a VirtualRow,
+    # yield a new VirtualRow instance to the block if it accepts a single
+    # argument.  Otherwise, evaluate the block in the context of a new
+    # VirtualRow instance.
+    def virtual_row_block_call(block)
+      return unless block
+      case block.arity
+      when -1, 0
+        SQL::VirtualRow.new.instance_eval(&block)
+      else
+        block.call(SQL::VirtualRow.new)
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/exceptions.rb 3.0.0-1/lib/sequel/exceptions.rb
--- 1.5.1-1/lib/sequel/exceptions.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/exceptions.rb	2009-03-23 17:00:41.000000000 +0000
@@ -0,0 +1,48 @@
+module Sequel
+  # The default exception class for exceptions raised by Sequel.
+  # All exception classes defined by Sequel are descendants of this class.
+  class Error < ::StandardError
+  end  
+    
+  # Raised when the adapter requested doesn't exist or can't be loaded.
+  class AdapterNotFound < Error ; end
+
+  # Generic error raised by the database adapters, indicating a
+  # problem originating from the database server.  Usually raised
+  # because incorrect SQL syntax is used.
+  class DatabaseError < Error; end
+  
+  # Error raised when the Sequel is unable to connect to the database with the
+  # connection parameters it was given.
+  class DatabaseConnectionError < DatabaseError; end
+
+  # Error that should be raised by adapters when they determine that the connection
+  # to the database has been lost.  Instructs the connection pool code to 
+  # remove that connection from the pool so that other connections can be acquired
+  # automatically.
+  class DatabaseDisconnectError < DatabaseError; end
+
+  # Raised on an invalid operation, such as trying to update or delete
+  # a joined or grouped dataset.
+  class InvalidOperation < Error; end
+                                     
+  # Raised when attempting an invalid type conversion.
+  class InvalidValue < Error ; end
+                                     
+  # Raised when the connection pool cannot acquire a database connection
+  # before the timeout.
+  class PoolTimeout < Error ; end
+                                     
+  # Exception that you should raise to signal a rollback of the current transaction.
+  # The transaction block will catch this exception, rollback the current transaction,
+  # and won't reraise it.
+  class Rollback < Error ; end
+
+  class Error
+    AdapterNotFound = Sequel::AdapterNotFound
+    InvalidOperation = Sequel::InvalidOperation
+    InvalidValue = Sequel::InvalidValue
+    PoolTimeoutError = Sequel::PoolTimeout
+    Rollback = Sequel::Rollback
+  end  
+end
diff -pruN 1.5.1-1/lib/sequel/extensions/blank.rb 3.0.0-1/lib/sequel/extensions/blank.rb
--- 1.5.1-1/lib/sequel/extensions/blank.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/extensions/blank.rb	2009-03-16 20:53:19.000000000 +0000
@@ -0,0 +1,42 @@
+class FalseClass
+  # false is always blank
+  def blank?
+    true
+  end
+end
+
+# Helpers from Metaid and a bit more
+class Object
+  # Objects are blank if they respond true to empty?
+  def blank?
+    respond_to?(:empty?) && empty?
+  end
+end
+
+class NilClass
+  # nil is always blank
+  def blank?
+    true
+  end
+end
+
+class Numeric
+  # Numerics are never blank (not even 0)
+  def blank?
+    false
+  end
+end
+
+class String
+  # Strings are blank if they are empty or include only whitespace
+  def blank?
+    strip.empty?
+  end
+end
+
+class TrueClass
+  # true is never blank
+  def blank?
+    false
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/extensions/inflector.rb 3.0.0-1/lib/sequel/extensions/inflector.rb
--- 1.5.1-1/lib/sequel/extensions/inflector.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/extensions/inflector.rb	2009-04-09 21:35:14.000000000 +0100
@@ -0,0 +1,283 @@
+# Add inflection methods to String, which allows the easy transformation of
+# words from singular to plural,class names to table names, modularized class
+# names to ones without, and class names to foreign keys.
+
+class String
+  # This module acts as a singleton returned/yielded by String.inflections,
+  # which is used to override or specify additional inflection rules. Examples:
+  #
+  #   String.inflections do |inflect|
+  #     inflect.plural /^(ox)$/i, '\1\2en'
+  #     inflect.singular /^(ox)en/i, '\1'
+  #
+  #     inflect.irregular 'octopus', 'octopi'
+  #
+  #     inflect.uncountable "equipment"
+  #   end
+  #
+  # New rules are added at the top. So in the example above, the irregular rule for octopus will now be the first of the
+  # pluralization and singularization rules that is runs. This guarantees that your rules run before any of the rules that may
+  # already have been loaded.
+  module Inflections
+    @plurals, @singulars, @uncountables = [], [], []
+
+    class << self
+      attr_reader :plurals, :singulars, :uncountables
+    end
+
+    # Clears the loaded inflections within a given scope (default is :all). Give the scope as a symbol of the inflection type,
+    # the options are: :plurals, :singulars, :uncountables
+    #
+    # Examples:
+    #   clear :all
+    #   clear :plurals
+    def self.clear(scope = :all)
+      case scope
+      when :all
+        @plurals, @singulars, @uncountables = [], [], []
+      else
+        instance_variable_set("@#{scope}", [])
+      end
+    end
+
+    # Specifies a new irregular that applies to both pluralization and singularization at the same time. This can only be used
+    # for strings, not regular expressions. You simply pass the irregular in singular and plural form.
+    #
+    # Examples:
+    #   irregular 'octopus', 'octopi'
+    #   irregular 'person', 'people'
+    def self.irregular(singular, plural)
+      plural(Regexp.new("(#{singular[0,1]})#{singular[1..-1]}$", "i"), '\1' + plural[1..-1])
+      singular(Regexp.new("(#{plural[0,1]})#{plural[1..-1]}$", "i"), '\1' + singular[1..-1])
+    end
+
+    # Specifies a new pluralization rule and its replacement. The rule can either be a string or a regular expression.
+    # The replacement should always be a string that may include references to the matched data from the rule.
+    #
+    # Example:
+    #   plural(/(x|ch|ss|sh)$/i, '\1es')
+    def self.plural(rule, replacement)
+      @plurals.insert(0, [rule, replacement])
+    end
+
+    # Specifies a new singularization rule and its replacement. The rule can either be a string or a regular expression.
+    # The replacement should always be a string that may include references to the matched data from the rule.
+    #
+    # Example:
+    #   singular(/([^aeiouy]|qu)ies$/i, '\1y') 
+    def self.singular(rule, replacement)
+      @singulars.insert(0, [rule, replacement])
+    end
+
+    # Add uncountable words that shouldn't be attempted inflected.
+    #
+    # Examples:
+    #   uncountable "money"
+    #   uncountable "money", "information"
+    #   uncountable %w( money information rice )
+    def self.uncountable(*words)
+      (@uncountables << words).flatten!
+    end
+
+    # Setup the default inflections
+    plural(/$/, 's')
+    plural(/s$/i, 's')
+    plural(/(ax|test)is$/i, '\1es')
+    plural(/(octop|vir)us$/i, '\1i')
+    plural(/(alias|status)$/i, '\1es')
+    plural(/(bu)s$/i, '\1ses')
+    plural(/(buffal|tomat)o$/i, '\1oes')
+    plural(/([ti])um$/i, '\1a')
+    plural(/sis$/i, 'ses')
+    plural(/(?:([^f])fe|([lr])f)$/i, '\1\2ves')
+    plural(/(hive)$/i, '\1s')
+    plural(/([^aeiouy]|qu)y$/i, '\1ies')
+    plural(/(x|ch|ss|sh)$/i, '\1es')
+    plural(/(matr|vert|ind)ix|ex$/i, '\1ices')
+    plural(/([m|l])ouse$/i, '\1ice')
+    plural(/^(ox)$/i, '\1en')
+    plural(/(quiz)$/i, '\1zes')
+
+    singular(/s$/i, '')
+    singular(/(n)ews$/i, '\1ews')
+    singular(/([ti])a$/i, '\1um')
+    singular(/((a)naly|(b)a|(d)iagno|(p)arenthe|(p)rogno|(s)ynop|(t)he)ses$/i, '\1\2sis')
+    singular(/(^analy)ses$/i, '\1sis')
+    singular(/([^f])ves$/i, '\1fe')
+    singular(/(hive)s$/i, '\1')
+    singular(/(tive)s$/i, '\1')
+    singular(/([lr])ves$/i, '\1f')
+    singular(/([^aeiouy]|qu)ies$/i, '\1y')
+    singular(/(s)eries$/i, '\1eries')
+    singular(/(m)ovies$/i, '\1ovie')
+    singular(/(x|ch|ss|sh)es$/i, '\1')
+    singular(/([m|l])ice$/i, '\1ouse')
+    singular(/(bus)es$/i, '\1')
+    singular(/(o)es$/i, '\1')
+    singular(/(shoe)s$/i, '\1')
+    singular(/(cris|ax|test)es$/i, '\1is')
+    singular(/(octop|vir)i$/i, '\1us')
+    singular(/(alias|status)es$/i, '\1')
+    singular(/^(ox)en/i, '\1')
+    singular(/(vert|ind)ices$/i, '\1ex')
+    singular(/(matr)ices$/i, '\1ix')
+    singular(/(quiz)zes$/i, '\1')
+
+    irregular('person', 'people')
+    irregular('man', 'men')
+    irregular('child', 'children')
+    irregular('sex', 'sexes')
+    irregular('move', 'moves')
+
+    uncountable(%w(equipment information rice money species series fish sheep))
+  end
+
+  # Yield the Inflections module if a block is given, and return
+  # the Inflections module.
+  def self.inflections
+    yield Inflections if block_given?
+    Inflections
+  end
+
+  # By default, camelize converts the string to UpperCamelCase. If the argument to camelize
+  # is set to :lower then camelize produces lowerCamelCase.
+  #
+  # camelize will also convert '/' to '::' which is useful for converting paths to namespaces
+  #
+  # Examples
+  #   "active_record".camelize #=> "ActiveRecord"
+  #   "active_record".camelize(:lower) #=> "activeRecord"
+  #   "active_record/errors".camelize #=> "ActiveRecord::Errors"
+  #   "active_record/errors".camelize(:lower) #=> "activeRecord::Errors"
+  def camelize(first_letter_in_uppercase = :upper)
+    s = gsub(/\/(.?)/){|x| "::#{x[-1..-1].upcase unless x == '/'}"}.gsub(/(^|_)(.)/){|x| x[-1..-1].upcase}
+    s[0...1] = s[0...1].downcase unless first_letter_in_uppercase == :upper
+    s
+  end
+  alias_method :camelcase, :camelize
+
+  # Singularizes and camelizes the string.  Also strips out all characters preceding
+  # and including a period (".").
+  #
+  # Examples
+  #   "egg_and_hams".classify #=> "EggAndHam"
+  #   "post".classify #=> "Post"
+  #   "schema.post".classify #=> "Post"
+  def classify
+    sub(/.*\./, '').singularize.camelize
+  end
+
+  # Constantize tries to find a declared constant with the name specified
+  # in the string. It raises a NameError when the name is not in CamelCase
+  # or is not initialized.
+  #
+  # Examples
+  #   "Module".constantize #=> Module
+  #   "Class".constantize #=> Class
+  def constantize
+    raise(NameError, "#{inspect} is not a valid constant name!") unless m = /\A(?:::)?([A-Z]\w*(?:::[A-Z]\w*)*)\z/.match(self)
+    Object.module_eval("::#{m[1]}", __FILE__, __LINE__)
+  end
+
+  # Replaces underscores with dashes in the string.
+  #
+  # Example
+  #   "puni_puni".dasherize #=> "puni-puni"
+  def dasherize
+    gsub(/_/, '-')
+  end
+
+  # Removes the module part from the expression in the string
+  #
+  # Examples
+  #   "ActiveRecord::CoreExtensions::String::Inflections".demodulize #=> "Inflections"
+  #   "Inflections".demodulize #=> "Inflections"
+  def demodulize
+    gsub(/^.*::/, '')
+  end
+
+  # Creates a foreign key name from a class name.
+  # +use_underscore+ sets whether the method should put '_' between the name and 'id'.
+  #
+  # Examples
+  #   "Message".foreign_key #=> "message_id"
+  #   "Message".foreign_key(false) #=> "messageid"
+  #   "Admin::Post".foreign_key #=> "post_id"
+  def foreign_key(use_underscore = true)
+    "#{demodulize.underscore}#{'_' if use_underscore}id"
+  end
+
+  # Capitalizes the first word and turns underscores into spaces and strips _id.
+  # Like titleize, this is meant for creating pretty output.
+  #
+  # Examples
+  #   "employee_salary" #=> "Employee salary"
+  #   "author_id" #=> "Author"
+  def humanize
+    gsub(/_id$/, "").gsub(/_/, " ").capitalize
+  end
+
+  # Returns the plural form of the word in the string.
+  #
+  # Examples
+  #   "post".pluralize #=> "posts"
+  #   "octopus".pluralize #=> "octopi"
+  #   "sheep".pluralize #=> "sheep"
+  #   "words".pluralize #=> "words"
+  #   "the blue mailman".pluralize #=> "the blue mailmen"
+  #   "CamelOctopus".pluralize #=> "CamelOctopi"
+  def pluralize
+    result = dup
+    Inflections.plurals.each{|(rule, replacement)| break if result.gsub!(rule, replacement)} unless Inflections.uncountables.include?(downcase)
+    result
+  end
+
+  # The reverse of pluralize, returns the singular form of a word in a string.
+  #
+  # Examples
+  #   "posts".singularize #=> "post"
+  #   "octopi".singularize #=> "octopus"
+  #   "sheep".singluarize #=> "sheep"
+  #   "word".singluarize #=> "word"
+  #   "the blue mailmen".singularize #=> "the blue mailman"
+  #   "CamelOctopi".singularize #=> "CamelOctopus"
+  def singularize
+    result = dup
+    Inflections.singulars.each{|(rule, replacement)| break if result.gsub!(rule, replacement)} unless Inflections.uncountables.include?(downcase)
+    result
+  end
+
+  # Underscores and pluralizes the string.
+  #
+  # Examples
+  #   "RawScaledScorer".tableize #=> "raw_scaled_scorers"
+  #   "egg_and_ham".tableize #=> "egg_and_hams"
+  #   "fancyCategory".tableize #=> "fancy_categories"
+  def tableize
+    underscore.pluralize
+  end
+
+  # Capitalizes all the words and replaces some characters in the string to create
+  # a nicer looking title. Titleize is meant for creating pretty output.
+  #
+  # titleize is also aliased as as titlecase
+  #
+  # Examples
+  #   "man from the boondocks".titleize #=> "Man From The Boondocks"
+  #   "x-men: the last stand".titleize #=> "X Men: The Last Stand"
+  def titleize
+    underscore.humanize.gsub(/\b([a-z])/){|x| x[-1..-1].upcase}
+  end
+  alias_method :titlecase, :titleize
+
+  # The reverse of camelize. Makes an underscored form from the expression in the string.
+  # Also changes '::' to '/' to convert namespaces to paths.
+  #
+  # Examples
+  #   "ActiveRecord".underscore #=> "active_record"
+  #   "ActiveRecord::Errors".underscore #=> active_record/errors
+  def underscore
+    gsub(/::/, '/').gsub(/([A-Z]+)([A-Z][a-z])/,'\1_\2').
+      gsub(/([a-z\d])([A-Z])/,'\1_\2').tr("-", "_").downcase
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/extensions/migration.rb 3.0.0-1/lib/sequel/extensions/migration.rb
--- 1.5.1-1/lib/sequel/extensions/migration.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/extensions/migration.rb	2009-03-23 17:41:33.000000000 +0000
@@ -0,0 +1,212 @@
+module Sequel
+  # The Migration class describes a database migration that can be reversed.
+  # The migration looks very similar to ActiveRecord (Rails) migrations, e.g.:
+  #
+  #   class CreateSessions < Sequel::Migration
+  #     def up
+  #       create_table :sessions do
+  #         primary_key :id
+  #         String :session_id, :size => 32, :unique => true
+  #         DateTime :created_at
+  #         text :data
+  #       end
+  #     end
+  # 
+  #     def down
+  #       # You can use raw SQL if you need to
+  #       self << 'DROP TABLE sessions'
+  #     end
+  #   end
+  #
+  #   class AlterItems < Sequel::Migration
+  #     def up
+  #       alter_table :items do
+  #         add_column :category, String, :default => 'ruby'
+  #       end
+  #     end
+  # 
+  #     def down
+  #       alter_table :items do
+  #         drop_column :category
+  #       end  
+  #     end
+  #   end
+  #
+  # To apply a migration to a database, you can invoke the #apply with
+  # the target database instance and the direction :up or :down, e.g.:
+  #
+  #   DB = Sequel.connect('sqlite://mydb')
+  #   CreateSessions.apply(DB, :up)
+  #
+  # See Sequel::Schema::Generator for the syntax to use for creating tables,
+  # and Sequel::Schema::AlterTableGenerator for the syntax to use when
+  # altering existing tables.  Migrations act as a proxy for the database
+  # given in #apply, so inside #down and #up, you can act as though self
+  # refers to the database.  So you can use any of the Sequel::Database
+  # instance methods directly.
+  class Migration
+    # Creates a new instance of the migration and sets the @db attribute.
+    def initialize(db)
+      @db = db
+    end
+    
+    # Applies the migration to the supplied database in the specified
+    # direction.
+    def self.apply(db, direction)
+      obj = new(db)
+      case direction
+      when :up
+        obj.up
+      when :down
+        obj.down
+      else
+        raise ArgumentError, "Invalid migration direction specified (#{direction.inspect})"
+      end
+    end
+
+    # Returns the list of Migration descendants.
+    def self.descendants
+      @descendants ||= []
+    end
+    
+    # Adds the new migration class to the list of Migration descendants.
+    def self.inherited(base)
+      descendants << base
+    end
+    
+    # The default down action does nothing
+    def down
+    end
+    
+    # Intercepts method calls intended for the database and sends them along.
+    def method_missing(method_sym, *args, &block)
+      @db.send(method_sym, *args, &block)
+    end
+
+    # The default up action does nothing
+    def up
+    end
+  end
+
+  # The Migrator module performs migrations based on migration files in a 
+  # specified directory. The migration files should be named using the
+  # following pattern (in similar fashion to ActiveRecord migrations):
+  # 
+  #   <version>_<title>.rb
+  #
+  # For example, the following files are considered migration files:
+  #   
+  #   001_create_sessions.rb
+  #   002_add_data_column.rb
+  #   ...
+  #
+  # The migration files should contain one or more migration classes based
+  # on Sequel::Migration.
+  #
+  # Migrations are generally run via the sequel command line tool,
+  # using the -m and -M switches.  The -m switch specifies the migration
+  # directory, and the -M switch specifies the version to which to migrate.
+  # 
+  # You can apply migrations using the Migrator API, as well (this is necessary
+  # if you want to specify the version from which to migrate in addition to the version
+  # to which to migrate).
+  # To apply a migration, the #apply method must be invoked with the database
+  # instance, the directory of migration files and the target version. If
+  # no current version is supplied, it is read from the database. The migrator
+  # automatically creates a schema_info table in the database to keep track
+  # of the current migration version. If no migration version is stored in the
+  # database, the version is considered to be 0. If no target version is 
+  # specified, the database is migrated to the latest version available in the
+  # migration directory.
+  #
+  # For example, to migrate the database to the latest version:
+  #
+  #   Sequel::Migrator.apply(DB, '.')
+  #
+  # To migrate the database from version 1 to version 5:
+  #
+  #   Sequel::Migrator.apply(DB, '.', 5, 1)
+  module Migrator
+    MIGRATION_FILE_PATTERN = /\A\d+_.+\.rb\z/.freeze
+
+    # Migrates the supplied database in the specified directory from the
+    # current version to the target version. If no current version is
+    # supplied, it is extracted from a schema_info table. The schema_info
+    # table is automatically created and maintained by the apply function.
+    def self.apply(db, directory, target = nil, current = nil)
+      # determine current and target version and direction
+      current ||= get_current_migration_version(db)
+      target ||= latest_migration_version(directory)
+      raise Error, "No current version available" if current.nil?
+      raise Error, "No target version available" if target.nil?
+
+      direction = current < target ? :up : :down
+      
+      classes = migration_classes(directory, target, current, direction)
+      
+      db.transaction do
+        classes.each {|c| c.apply(db, direction)}
+        set_current_migration_version(db, target)
+      end
+      
+      target
+    end
+
+    # Gets the current migration version stored in the database. If no version
+    # number is stored, 0 is returned.
+    def self.get_current_migration_version(db)
+      r = schema_info_dataset(db).first
+      r ? r[:version] : 0
+    end
+
+    # Returns the latest version available in the specified directory.
+    def self.latest_migration_version(directory)
+      l = migration_files(directory).last
+      l ? File.basename(l).to_i : nil
+    end
+
+    # Returns a list of migration classes filtered for the migration range and
+    # ordered according to the migration direction.
+    def self.migration_classes(directory, target, current, direction)
+      range = direction == :up ?
+        (current + 1)..target : (target + 1)..current
+
+      # Remove class definitions
+      Migration.descendants.each do |c|
+        Object.send(:remove_const, c.to_s) rescue nil
+      end
+      Migration.descendants.clear # remove any defined migration classes
+
+      # load migration files
+      migration_files(directory, range).each {|fn| load(fn)}
+      
+      # get migration classes
+      classes = Migration.descendants
+      classes.reverse! if direction == :down
+      classes
+    end
+    
+    # Returns any found migration files in the supplied directory.
+    def self.migration_files(directory, range = nil)
+      files = []
+      Dir.new(directory).each do |file|
+        files[file.to_i] = File.join(directory, file) if MIGRATION_FILE_PATTERN.match(file)
+      end
+      filtered = range ? files[range] : files
+      filtered ? filtered.compact : []
+    end
+    
+    # Returns the dataset for the schema_info table. If no such table
+    # exists, it is automatically created.
+    def self.schema_info_dataset(db)
+      db.create_table(:schema_info) {integer :version} unless db.table_exists?(:schema_info)
+      db[:schema_info]
+    end
+    
+    # Sets the current migration  version stored in the database.
+    def self.set_current_migration_version(db, version)
+      dataset = schema_info_dataset(db)
+      dataset.send(dataset.first ? :update : :<<, :version => version)
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/extensions/pagination.rb 3.0.0-1/lib/sequel/extensions/pagination.rb
--- 1.5.1-1/lib/sequel/extensions/pagination.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/extensions/pagination.rb	2009-03-04 22:22:23.000000000 +0000
@@ -0,0 +1,96 @@
+module Sequel
+  class Dataset
+    # Returns a paginated dataset. The returned dataset is limited to
+    # the page size at the correct offset, and extended with the Pagination
+    # module.  If a record count is not provided, does a count of total
+    # number of records for this dataset.
+    def paginate(page_no, page_size, record_count=nil)
+      raise(Error, "You cannot paginate a dataset that already has a limit") if @opts[:limit]
+      paginated = limit(page_size, (page_no - 1) * page_size)
+      paginated.extend(Pagination)
+      paginated.set_pagination_info(page_no, page_size, record_count || count)
+    end
+      
+    # Yields a paginated dataset for each page and returns the receiver. Does
+    # a count to find the total number of records for this dataset.
+    def each_page(page_size, &block)
+      raise(Error, "You cannot paginate a dataset that already has a limit") if @opts[:limit]
+      record_count = count
+      total_pages = (record_count / page_size.to_f).ceil
+      (1..total_pages).each{|page_no| yield paginate(page_no, page_size, record_count)}
+      self
+    end
+
+    # Holds methods that only relate to paginated datasets. Paginated dataset
+    # have pages starting at 1 (page 1 is offset 0, page 1 is offset page_size).
+    module Pagination
+      # The number of records per page (the final page may have fewer than
+      # this number of records).
+      attr_accessor :page_size
+
+      # The number of pages in the dataset before pagination, of which
+      # this paginated dataset is one.
+      attr_accessor :page_count
+
+      # The current page of the dataset, starting at 1 and not 0.
+      attr_accessor :current_page
+
+      # The total number of records in the dataset before pagination.
+      attr_accessor :pagination_record_count
+
+      # Returns the record range for the current page
+      def current_page_record_range
+        return (0..0) if @current_page > @page_count
+        
+        a = 1 + (@current_page - 1) * @page_size
+        b = a + @page_size - 1
+        b = @pagination_record_count if b > @pagination_record_count
+        a..b
+      end
+
+      # Returns the number of records in the current page
+      def current_page_record_count
+        return 0 if @current_page > @page_count
+        
+        a = 1 + (@current_page - 1) * @page_size
+        b = a + @page_size - 1
+        b = @pagination_record_count if b > @pagination_record_count
+        b - a + 1
+      end
+
+      # Returns true if the current page is the first page
+      def first_page?
+        @current_page == 1
+      end
+
+      # Returns true if the current page is the last page
+      def last_page?
+        @current_page == @page_count
+      end
+
+      # Returns the next page number or nil if the current page is the last page
+      def next_page
+        current_page < page_count ? (current_page + 1) : nil
+      end
+      
+      # Returns the page range
+      def page_range
+        1..page_count
+      end
+      
+      # Returns the previous page number or nil if the current page is the first
+      def prev_page
+        current_page > 1 ? (current_page - 1) : nil
+      end
+
+      # Sets the pagination info for this paginated dataset, and returns self.
+      def set_pagination_info(page_no, page_size, record_count)
+        @current_page = page_no
+        @page_size = page_size
+        @pagination_record_count = record_count
+        @page_count = (record_count / page_size.to_f).ceil
+        self
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/extensions/pretty_table.rb 3.0.0-1/lib/sequel/extensions/pretty_table.rb
--- 1.5.1-1/lib/sequel/extensions/pretty_table.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/extensions/pretty_table.rb	2009-03-10 03:16:39.000000000 +0000
@@ -0,0 +1,78 @@
+module Sequel
+  class Dataset
+    # Pretty prints the records in the dataset as plain-text table.
+    def print(*cols)
+      Sequel::PrettyTable.print(naked.all, cols.empty? ? columns : cols)
+    end
+  end
+
+  module PrettyTable
+    # Prints nice-looking plain-text tables via puts
+    # 
+    #   +--+-------+
+    #   |id|name   |
+    #   |--+-------|
+    #   |1 |fasdfas|
+    #   |2 |test   |
+    #   +--+-------+
+    def self.print(records, columns = nil) # records is an array of hashes
+      columns ||= records.first.keys.sort_by{|x|x.to_s}
+      sizes = column_sizes(records, columns)
+      sep_line = separator_line(columns, sizes)
+
+      puts sep_line
+      puts header_line(columns, sizes)
+      puts sep_line
+      records.each {|r| puts data_line(columns, sizes, r)}
+      puts sep_line
+    end
+
+    ### Private Module Methods ###
+
+    # Hash of the maximum size of the value for each column 
+    def self.column_sizes(records, columns) # :nodoc:
+      sizes = Hash.new {0}
+      columns.each do |c|
+        s = c.to_s.size
+        sizes[c.to_sym] = s if s > sizes[c.to_sym]
+      end
+      records.each do |r|
+        columns.each do |c|
+          s = r[c].to_s.size
+          sizes[c.to_sym] = s if s > sizes[c.to_sym]
+        end
+      end
+      sizes
+    end
+    
+    # String for each data line
+    def self.data_line(columns, sizes, record) # :nodoc:
+      '|' << columns.map {|c| format_cell(sizes[c], record[c])}.join('|') << '|'
+    end
+    
+    # Format the value so it takes up exactly size characters
+    def self.format_cell(size, v) # :nodoc:
+      case v
+      when Bignum, Fixnum
+        "%#{size}d" % v
+      when Float
+        "%#{size}g" % v
+      else
+        "%-#{size}s" % v.to_s
+      end
+    end
+    
+    # String for header line
+    def self.header_line(columns, sizes) # :nodoc:
+      '|' << columns.map {|c| "%-#{sizes[c]}s" % c.to_s}.join('|') << '|'
+    end
+
+    # String for separtor line
+    def self.separator_line(columns, sizes) # :nodoc:
+      '+' << columns.map {|c| '-' * sizes[c]}.join('+') << '+'
+    end
+
+    private_class_method :column_sizes, :data_line, :format_cell, :header_line, :separator_line
+  end
+end
+
diff -pruN 1.5.1-1/lib/sequel/extensions/query.rb 3.0.0-1/lib/sequel/extensions/query.rb
--- 1.5.1-1/lib/sequel/extensions/query.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/extensions/query.rb	2009-03-09 23:33:23.000000000 +0000
@@ -0,0 +1,48 @@
+module Sequel
+  class Database
+    # Return a dataset modified by the query block
+    def query(&block)
+      dataset.query(&block)
+    end
+  end
+
+  class Dataset
+    # Translates a query block into a dataset. Query blocks can be useful
+    # when expressing complex SELECT statements, e.g.:
+    #
+    #   dataset = DB[:items].query do
+    #     select :x, :y, :z
+    #     filter{|o| (o.x > 1) & (o.y > 2)}
+    #     order :z.desc
+    #   end
+    #
+    # Which is the same as:
+    #
+    #  dataset = DB[:items].select(:x, :y, :z).filter{|o| (o.x > 1) & (o.y > 2)}.order(:z.desc)
+    #
+    # Note that inside a call to query, you cannot call each, insert, update,
+    # or delete (or any method that calls those), or Sequel will raise an
+    # error.
+    def query(&block)
+      copy = clone({})
+      copy.extend(QueryBlockCopy)
+      copy.instance_eval(&block)
+      clone(copy.opts)
+    end
+
+    # Module used by Dataset#query that has the effect of making all
+    # dataset methods into !-style methods that modify the receiver.
+    module QueryBlockCopy
+      %w'each insert update delete'.each do |meth|
+        define_method(meth){|*args| raise Error, "##{meth} cannot be invoked inside a query block."}
+      end
+
+      # Merge the given options into the receiver's options and return the receiver
+      # instead of cloning the receiver.
+      def clone(opts = nil)
+        @opts.merge!(opts)
+        self
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/extensions/schema_dumper.rb 3.0.0-1/lib/sequel/extensions/schema_dumper.rb
--- 1.5.1-1/lib/sequel/extensions/schema_dumper.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/extensions/schema_dumper.rb	2009-04-29 18:02:11.000000000 +0100
@@ -0,0 +1,238 @@
+module Sequel
+  class Database
+    # Dump indexes for all tables as a migration.  This complements
+    # the :indexes=>false option to dump_schema_migration.
+    def dump_indexes_migration
+      ts = tables
+      <<END_MIG
+Class.new(Sequel::Migration) do
+  def up
+#{ts.map{|t| dump_table_indexes(t, :add_index)}.reject{|x| x == ''}.join("\n\n").gsub(/^/o, '    ')}
+  end
+  
+  def down
+#{ts.map{|t| dump_table_indexes(t, :drop_index)}.reject{|x| x == ''}.join("\n\n").gsub(/^/o, '    ')}
+  end
+end
+END_MIG
+    end
+
+    # Return a string that contains a Sequel::Migration subclass that when
+    # run would recreate the database structure. Options:
+    # * :same_db - Don't attempt to translate database types to ruby types.
+    #   If this isn't set to true, all database types will be translated to
+    #   ruby types, but there is no guarantee that the migration generated
+    #   will yield the same type.  Without this set, types that aren't
+    #   recognized will be translated to a string-like type.
+    # * :indexes - If set to false, don't dump indexes (they can be added
+    #   later via dump_index_migration).
+    def dump_schema_migration(options={})
+      ts = tables
+      <<END_MIG
+Class.new(Sequel::Migration) do
+  def up
+#{ts.map{|t| dump_table_schema(t, options)}.join("\n\n").gsub(/^/o, '    ')}
+  end
+  
+  def down
+    drop_table(#{ts.inspect[1...-1]})
+  end
+end
+END_MIG
+    end
+
+    # Return a string with a create table block that will recreate the given
+    # table's schema.  Takes the same options as dump_schema_migration.
+    def dump_table_schema(table, options={})
+      s = schema(table).dup
+      pks = s.find_all{|x| x.last[:primary_key] == true}.map{|x| x.first}
+      options = options.merge(:single_pk=>true) if pks.length == 1
+      m = method(:column_schema_to_generator_opts)
+      im = method(:index_to_generator_opts)
+      indexes = indexes(table).sort_by{|k,v| k.to_s} if options[:indexes] != false and respond_to?(:indexes)
+      gen = Schema::Generator.new(self) do
+        s.each{|name, info| send(*m.call(name, info, options))}
+        primary_key(pks) if !@primary_key && pks.length > 0
+        indexes.each{|iname, iopts| send(:index, iopts[:columns], im.call(table, iname, iopts))} if indexes
+      end
+      commands = [gen.dump_columns, gen.dump_constraints, gen.dump_indexes].reject{|x| x == ''}.join("\n\n")
+      "create_table(#{table.inspect}) do\n#{commands.gsub(/^/o, '  ')}\nend"
+    end
+
+    private
+
+    # Convert the given default, which should be a database specific string, into
+    # a ruby object.  If it can't be converted, return the string with the inspect
+    # method modified so that .lit is always appended after it.
+    def column_schema_to_ruby_default(default, type)
+      case default 
+      when /false/
+        false
+      when 'true'
+        true
+      when /\A\d+\z/
+        default.to_i
+      else
+        def default.inspect
+          "#{super}.lit"
+        end
+        default
+      end
+    end
+
+    # Convert the given name and parsed database schema into an array with a method
+    # name and arguments to it to pass to a Schema::Generator to recreate the column.
+    def column_schema_to_generator_opts(name, schema, options)
+      if options[:single_pk] && schema_autoincrementing_primary_key?(schema)
+        [:primary_key, name]
+      else
+        col_opts = options[:same_db] ? {:type=>schema[:db_type]} : column_schema_to_ruby_type(schema)
+        type = col_opts.delete(:type)
+        col_opts.delete(:size) if col_opts[:size].nil?
+        col_opts[:default] = column_schema_to_ruby_default(schema[:default], type) if schema[:default]
+        col_opts[:null] = false if schema[:allow_null] == false
+        [:column, name, type, col_opts]
+      end
+    end
+
+    # Convert the column schema information to a hash of column options, one of which must
+    # be :type.  The other options added should modify that type (e.g. :size).  If a
+    # database type is not recognized, return it as a String type.
+    def column_schema_to_ruby_type(schema)
+      case t = schema[:db_type].downcase
+      when /\A(?:medium|small)?int(?:eger)?(?:\((?:\d+)\))?\z/o
+        {:type=>Integer}
+      when /\Atinyint(?:\((?:\d+)\))?\z/o
+        {:type=>(Sequel.convert_tinyint_to_bool ? TrueClass : Integer)}
+      when /\Abigint(?:\((?:\d+)\))?\z/o
+        {:type=>Bignum}
+      when /\A(?:real|float|double(?: precision)?)\z/o
+        {:type=>Float}
+      when 'boolean'
+        {:type=>TrueClass}
+      when /\A(?:(?:tiny|medium|long)?text|clob)\z/o
+        {:type=>String, :text=>true}
+      when 'date'
+        {:type=>Date}
+      when 'datetime'
+        {:type=>DateTime}
+      when /\Atimestamp(?: with(?:out)? time zone)?\z/o
+        {:type=>DateTime}
+      when /\Atime(?: with(?:out)? time zone)?\z/o
+        {:type=>Time, :only_time=>true}
+      when /\Achar(?:acter)?(?:\((\d+)\))?\z/o
+        {:type=>String, :size=>($1.to_i if $1), :fixed=>true}
+      when /\A(?:varchar|character varying|bpchar|string)(?:\((\d+)\))?\z/o
+        s = ($1.to_i if $1)
+        {:type=>String, :size=>(s == 255 ? nil : s)}
+      when 'money'
+        {:type=>BigDecimal, :size=>[19,2]}
+      when /\A(?:decimal|numeric|number)(?:\((\d+)(?:,\s*(\d+))?\))?\z/o
+        s = [($1.to_i if $1), ($2.to_i if $2)].compact
+        {:type=>BigDecimal, :size=>(s.empty? ? nil : s)}
+      when /\A(?:bytea|(?:tiny|medium|long)?blob|(?:var)?binary)(?:\((\d+)\))?\z/o
+        {:type=>File, :size=>($1.to_i if $1)}
+      when 'year'
+        {:type=>Integer}
+      else
+        {:type=>String}
+      end
+    end
+
+    # Return a string that containing add_index/drop_index method calls for
+    # creating the index migration.
+    def dump_table_indexes(table, meth)
+      return '' unless respond_to?(:indexes)
+      im = method(:index_to_generator_opts)
+      indexes = indexes(table).sort_by{|k,v| k.to_s} 
+      gen = Schema::Generator.new(self) do
+        indexes.each{|iname, iopts| send(:index, iopts[:columns], im.call(table, iname, iopts))}
+      end
+      gen.dump_indexes(meth=>table)
+    end
+
+    # Convert the parsed index information into options to the Generators index method. 
+    def index_to_generator_opts(table, name, index_opts)
+      h = {}
+      h[:name] = name unless default_index_name(table, index_opts[:columns]) == name.to_s
+      h[:unique] = true if index_opts[:unique]
+      h
+    end
+  end
+
+  module Schema
+    class Generator
+      # Dump this generator's columns to a string that could be evaled inside
+      # another instance to represent the same columns
+      def dump_columns
+        strings = []
+        cols = columns.dup
+        if pkn = primary_key_name
+          cols.delete_if{|x| x[:name] == pkn}
+          pk = @primary_key.dup
+          pkname = pk.delete(:name)
+          @db.serial_primary_key_options.each{|k,v| pk.delete(k) if v == pk[k]}
+          strings << "primary_key #{pkname.inspect}#{opts_inspect(pk)}"
+        end
+        cols.each do |c|
+          c = c.dup
+          name = c.delete(:name)
+          type = c.delete(:type)
+          opts = opts_inspect(c)
+          strings << if type.is_a?(Class)
+            "#{type.name} #{name.inspect}#{opts}"
+          else
+            "column #{name.inspect}, #{type.inspect}#{opts}"
+          end
+        end
+        strings.join("\n")
+      end
+
+      # Dump this generator's constraints to a string that could be evaled inside
+      # another instance to represent the same constraints
+      def dump_constraints
+        constraints.map do |c|
+          c = c.dup
+          type = c.delete(:type)
+          case type
+          when :check
+            raise(Error, "can't dump check/constraint specified with Proc") if c[:check].is_a?(Proc)
+            name = c.delete(:name)
+            if !name and c[:check].length == 1 and c[:check].first.is_a?(Hash)
+              "check #{c[:check].first.inspect[1...-1]}"
+            else
+              "#{name ? "constraint #{name.inspect}," : 'check'} #{c[:check].map{|x| x.inspect}.join(', ')}"
+            end
+          else
+            cols = c.delete(:columns)
+            "#{type} #{cols.inspect}#{opts_inspect(c)}"
+          end
+        end.join("\n")
+      end
+
+      # Dump this generator's indexes to a string that could be evaled inside
+      # another instance to represent the same indexes. Options:
+      # * :add_index - Use add_index instead of index, so the methods
+      #   can be called outside of a generator but inside a migration.
+      #   The value of this option should be the table name to use.
+      # * :drop_index - Same as add_index, but create drop_index statements.
+      def dump_indexes(options={})
+        indexes.map do |c|
+          c = c.dup
+          cols = c.delete(:columns)
+          if table = options[:add_index] || options[:drop_index]
+            "#{options[:drop_index] ? 'drop' : 'add'}_index #{table.inspect}, #{cols.inspect}#{opts_inspect(c)}"
+          else
+            "index #{cols.inspect}#{opts_inspect(c)}"
+          end
+        end.join("\n")
+      end
+
+      private
+
+      def opts_inspect(opts)
+        ", #{opts.inspect[1...-1]}" if opts.length > 0
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/extensions/string_date_time.rb 3.0.0-1/lib/sequel/extensions/string_date_time.rb
--- 1.5.1-1/lib/sequel/extensions/string_date_time.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/extensions/string_date_time.rb	2009-03-23 17:02:02.000000000 +0000
@@ -0,0 +1,47 @@
+# This file contains the previous extensions to String for date/time
+# conversions.  These are provided mainly for backward compatibility,
+# Sequel now uses a module level method instead of extending string
+# to handle the internal conversions.
+
+class String
+  # Converts a string into a Date object.
+  def to_date
+    begin
+      Date.parse(self, Sequel.convert_two_digit_years)
+    rescue => e
+      raise Sequel::InvalidValue, "Invalid Date value '#{self}' (#{e.message})"
+    end
+  end
+
+  # Converts a string into a DateTime object.
+  def to_datetime
+    begin
+      DateTime.parse(self, Sequel.convert_two_digit_years)
+    rescue => e
+      raise Sequel::InvalidValue, "Invalid DateTime value '#{self}' (#{e.message})"
+    end
+  end
+
+  # Converts a string into a Time or DateTime object, depending on the
+  # value of Sequel.datetime_class
+  def to_sequel_time
+    begin
+      if Sequel.datetime_class == DateTime
+        DateTime.parse(self, Sequel.convert_two_digit_years)
+      else
+        Sequel.datetime_class.parse(self)
+      end
+    rescue => e
+      raise Sequel::InvalidValue, "Invalid #{Sequel.datetime_class} value '#{self}' (#{e.message})"
+    end
+  end
+
+  # Converts a string into a Time object.
+  def to_time
+    begin
+      Time.parse(self)
+    rescue => e
+      raise Sequel::InvalidValue, "Invalid Time value '#{self}' (#{e.message})"
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/metaprogramming.rb 3.0.0-1/lib/sequel/metaprogramming.rb
--- 1.5.1-1/lib/sequel/metaprogramming.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/metaprogramming.rb	2009-04-15 18:13:44.000000000 +0100
@@ -0,0 +1,23 @@
+module Sequel
+  # Contains methods that ease metaprogramming, used by some of Sequel's classes.
+  module Metaprogramming
+    # Add methods to the object's metaclass
+    def meta_def(name, &block)
+      meta_eval{define_method(name, &block)}
+    end 
+  
+    private
+  
+    # Evaluate the block in the context of the object's metaclass
+    def meta_eval(&block)
+      metaclass.instance_eval(&block)
+    end 
+  
+    # The hidden singleton lurks behind everyone
+    def metaclass
+      class << self
+        self
+      end 
+    end 
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/model/associations.rb 3.0.0-1/lib/sequel/model/associations.rb
--- 1.5.1-1/lib/sequel/model/associations.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/model/associations.rb	2009-04-01 21:12:03.000000000 +0100
@@ -0,0 +1,1300 @@
+module Sequel
+  class Model
+    # Associations are used in order to specify relationships between model classes
+    # that reflect relations between tables in the database using foreign keys.
+    module Associations
+      # Map of association type symbols to association reflection classes.
+      ASSOCIATION_TYPES = {}
+    
+      # Set an empty association reflection hash in the model
+      def self.apply(model)
+        model.instance_variable_set(:@association_reflections, {})
+      end
+
+      # AssociationReflection is a Hash subclass that keeps information on Sequel::Model associations. It
+      # provides methods to reduce internal code duplication.  It should not
+      # be instantiated by the user.
+      class AssociationReflection < Hash
+        include Sequel::Inflections
+    
+        # Name symbol for the _add internal association method
+        def _add_method
+          :"_add_#{singularize(self[:name])}"
+        end
+      
+        # Name symbol for the _dataset association method
+        def _dataset_method
+          :"_#{self[:name]}_dataset"
+        end
+      
+        # Name symbol for the _remove_all internal association method
+        def _remove_all_method
+          :"_remove_all_#{self[:name]}"
+        end
+      
+        # Name symbol for the _remove internal association method
+        def _remove_method
+          :"_remove_#{singularize(self[:name])}"
+        end
+      
+        # Name symbol for the _setter association method
+        def _setter_method
+          :"_#{self[:name]}="
+        end
+      
+        # Name symbol for the add association method
+        def add_method
+          :"add_#{singularize(self[:name])}"
+        end
+      
+        # Name symbol for association method, the same as the name of the association.
+        def association_method
+          self[:name]
+        end
+      
+        # The class associated to the current model class via this association
+        def associated_class
+          self[:class] ||= constantize(self[:class_name])
+        end
+      
+        # Name symbol for the dataset association method
+        def dataset_method
+          :"#{self[:name]}_dataset"
+        end
+      
+        # Name symbol for the _helper internal association method
+        def dataset_helper_method
+          :"_#{self[:name]}_dataset_helper"
+        end
+      
+        # Whether the dataset needs a primary key to function, true by default.
+        def dataset_need_primary_key?
+          true
+        end
+    
+        # Whether to eagerly graph a lazy dataset, true by default.  If this
+        # is false, the association won't respect the :eager_graph option
+        # when loading the association for a single record.
+        def eager_graph_lazy_dataset?
+          true
+        end
+    
+        # Whether the associated object needs a primary key to be added/removed,
+        # false by default.
+        def need_associated_primary_key?
+          false
+        end
+    
+        # Returns the reciprocal association variable, if one exists. The reciprocal
+        # association is the association in the associated class that is the opposite
+        # of the current association.  For example, Album.many_to_one :artist and
+        # Artist.one_to_many :albums are reciprocal associations.  This information is
+        # to populate reciprocal associations.  For example, when you do this_artist.add_album(album)
+        # it sets album.artist to this_artist.
+        def reciprocal
+          return self[:reciprocal] if include?(:reciprocal)
+          r_type = reciprocal_type
+          key = self[:key]
+          associated_class.all_association_reflections.each do |assoc_reflect|
+            if assoc_reflect[:type] == r_type && assoc_reflect[:key] == key
+              return self[:reciprocal] = assoc_reflect[:name]
+            end
+          end
+          self[:reciprocal] = nil
+        end
+    
+        # Whether the reciprocal of this association returns an array of objects instead of a single object,
+        # true by default.
+        def reciprocal_array?
+          true
+        end
+    
+        # Name symbol for the remove_all_ association method
+        def remove_all_method
+          :"remove_all_#{self[:name]}"
+        end
+      
+        # Name symbol for the remove_ association method
+        def remove_method
+          :"remove_#{singularize(self[:name])}"
+        end
+      
+        # Whether this association returns an array of objects instead of a single object,
+        # true by default.
+        def returns_array?
+          true
+        end
+    
+        # The columns to select when loading the association, nil by default.
+        def select
+          self[:select]
+        end
+    
+        # Whether to set the reciprocal association to self when loading associated
+        # records, false by default.
+        def set_reciprocal_to_self?
+          false
+        end
+    
+        # Name symbol for the setter association method
+        def setter_method
+          :"#{self[:name]}="
+        end
+      end
+    
+      class ManyToOneAssociationReflection < AssociationReflection
+        ASSOCIATION_TYPES[:many_to_one] = self
+    
+        # Whether the dataset needs a primary key to function, false for many_to_one associations.
+        def dataset_need_primary_key?
+          false
+        end
+    
+        # Default foreign key name symbol for foreign key in current model's table that points to
+        # the given association's table's primary key.
+        def default_key
+          :"#{self[:name]}_id"
+        end
+      
+        # Whether to eagerly graph a lazy dataset, true for many_to_one associations
+        # only if the key is nil.
+        def eager_graph_lazy_dataset?
+          self[:key].nil?
+        end
+    
+        # The key to use for the key hash when eager loading
+        def eager_loader_key
+          self[:key]
+        end
+    
+        # The column in the associated table that the key in the current table references.
+        def primary_key
+         self[:primary_key] ||= associated_class.primary_key
+        end
+      
+        # Whether this association returns an array of objects instead of a single object,
+        # false for a many_to_one association.
+        def returns_array?
+          false
+        end
+    
+        private
+    
+        # The reciprocal type of a many_to_one association is a one_to_many association.
+        def reciprocal_type
+          :one_to_many
+        end
+      end
+    
+      class OneToManyAssociationReflection < AssociationReflection
+        ASSOCIATION_TYPES[:one_to_many] = self
+    
+        # Default foreign key name symbol for key in associated table that points to
+        # current table's primary key.
+        def default_key
+          :"#{underscore(demodulize(self[:model].name))}_id"
+        end
+    
+        # The column in the current table that the key in the associated table references.
+        def primary_key
+         self[:primary_key] ||= self[:model].primary_key
+        end
+        alias eager_loader_key primary_key
+      
+        # One to many associations set the reciprocal to self when loading associated records.
+        def set_reciprocal_to_self?
+          true
+        end
+    
+        # Whether the reciprocal of this association returns an array of objects instead of a single object,
+        # false for a one_to_many association.
+        def reciprocal_array?
+          false
+        end
+    
+        private
+    
+        # The reciprocal type of a one_to_many association is a many_to_one association.
+        def reciprocal_type
+          :many_to_one
+        end
+      end
+    
+      class ManyToManyAssociationReflection < AssociationReflection
+        ASSOCIATION_TYPES[:many_to_many] = self
+    
+        # Default name symbol for the join table.
+        def default_join_table
+          [self[:class_name], self[:model].name].map{|i| underscore(pluralize(demodulize(i)))}.sort.join('_').to_sym
+        end
+      
+        # Default foreign key name symbol for key in join table that points to
+        # current table's primary key (or :left_primary_key column).
+        def default_left_key
+          :"#{underscore(demodulize(self[:model].name))}_id"
+        end
+    
+        # Default foreign key name symbol for foreign key in join table that points to
+        # the association's table's primary key (or :right_primary_key column).
+        def default_right_key
+          :"#{singularize(self[:name])}_id"
+        end
+      
+        # The key to use for the key hash when eager loading
+        def eager_loader_key
+          self[:left_primary_key]
+        end
+    
+        # Whether the associated object needs a primary key to be added/removed,
+        # true for many_to_many associations.
+        def need_associated_primary_key?
+          true
+        end
+    
+        # Returns the reciprocal association symbol, if one exists.
+        def reciprocal
+          return self[:reciprocal] if include?(:reciprocal)
+          left_key = self[:left_key]
+          right_key = self[:right_key]
+          join_table = self[:join_table]
+          associated_class.all_association_reflections.each do |assoc_reflect|
+            if assoc_reflect[:type] == :many_to_many && assoc_reflect[:left_key] == right_key \
+               && assoc_reflect[:right_key] == left_key && assoc_reflect[:join_table] == join_table
+              return self[:reciprocal] = assoc_reflect[:name]
+            end
+          end
+          self[:reciprocal] = nil
+        end
+    
+        # The primary key column to use in the associated table.
+        def right_primary_key
+          self[:right_primary_key] ||= associated_class.primary_key
+        end
+    
+        # The columns to select when loading the association, associated_class.table_name.* by default.
+        def select
+         return self[:select] if include?(:select)
+         self[:select] ||= Sequel::SQL::ColumnAll.new(associated_class.table_name)
+        end
+      end
+  
+      # This module contains methods added to all association datasets
+      module AssociationDatasetMethods
+        # The model object that created the association dataset
+        attr_accessor :model_object
+    
+        # The association reflection related to the association dataset
+        attr_accessor :association_reflection
+      end
+      
+      # Each kind of association adds a number of instance methods to the model class which
+      # are specialized according to the association type and optional parameters
+      # given in the definition. Example:
+      # 
+      #   class Project < Sequel::Model
+      #     many_to_one :portfolio
+      #     one_to_many :milestones
+      #     # or: many_to_many :milestones 
+      #   end
+      # 
+      # The project class now has the following instance methods:
+      # * portfolio - Returns the associated portfolio.
+      # * portfolio=(obj) - Sets the associated portfolio to the object,
+      #   but the change is not persisted until you save the record.
+      # * portfolio_dataset - Returns a dataset that would return the associated
+      #   portfolio, only useful in fairly specific circumstances.
+      # * milestones - Returns an array of associated milestones
+      # * add_milestone(obj) - Associates the passed milestone with this object.
+      # * remove_milestone(obj) - Removes the association with the passed milestone.
+      # * remove_all_milestones - Removes associations with all associated milestones.
+      # * milestones_dataset - Returns a dataset that would return the associated
+      #   milestones, allowing for further filtering/limiting/etc.
+      #
+      # If you want to override the behavior of the add_/remove_/remove_all_ methods,
+      # there are private instance methods created that a prepended with an
+      # underscore (e.g. _add_milestone).  The private instance methods can be
+      # easily overridden, but you shouldn't override the public instance methods without
+      # calling super, as they deal with callbacks and caching.
+      #
+      # By default the classes for the associations are inferred from the association
+      # name, so for example the Project#portfolio will return an instance of 
+      # Portfolio, and Project#milestones will return an array of Milestone 
+      # instances.
+      #
+      # Association definitions are also reflected by the class, e.g.:
+      #
+      #   Project.associations
+      #   => [:portfolio, :milestones]
+      #   Project.association_reflection(:portfolio)
+      #   => {:type => :many_to_one, :name => :portfolio, :class_name => "Portfolio"}
+      module ClassMethods
+        # All association reflections defined for this model (default: none).
+        attr_reader :association_reflections
+
+        # Array of all association reflections for this model class
+        def all_association_reflections
+          association_reflections.values
+        end
+        
+        # Associates a related model with the current model. The following types are
+        # supported:
+        #
+        # * :many_to_one - Foreign key in current model's table points to 
+        #   associated model's primary key.  Each associated model object can
+        #   be associated with more than one current model objects.  Each current
+        #   model object can be associated with only one associated model object.
+        # * :one_to_many - Foreign key in associated model's table points to this
+        #   model's primary key.   Each current model object can be associated with
+        #   more than one associated model objects.  Each associated model object
+        #   can be associated with only one current model object.
+        # * :many_to_many - A join table is used that has a foreign key that points
+        #   to this model's primary key and a foreign key that points to the
+        #   associated model's primary key.  Each current model object can be
+        #   associated with many associated model objects, and each associated
+        #   model object can be associated with many current model objects.
+        #
+        # A one to one relationship can be set up with a many_to_one association
+        # on the table with the foreign key, and a one_to_many association with the
+        # :one_to_one option specified on the table without the foreign key.  The
+        # two associations will operate similarly, except that the many_to_one
+        # association setter doesn't update the database until you call save manually.
+        # Also, in most cases you need to specify the plural association name when using
+        # one_to_many with the :one_to_one option.
+        # 
+        # The following options can be supplied:
+        # * *ALL types*:
+        #   - :after_add - Symbol, Proc, or array of both/either specifying a callback to call
+        #     after a new item is added to the association.
+        #   - :after_load - Symbol, Proc, or array of both/either specifying a callback to call
+        #     after the associated record(s) have been retrieved from the database.  Not called
+        #     when eager loading via eager_graph, but called when eager loading via eager.
+        #   - :after_remove - Symbol, Proc, or array of both/either specifying a callback to call
+        #     after an item is removed from the association.
+        #   - :allow_eager - If set to false, you cannot load the association eagerly
+        #     via eager or eager_graph
+        #   - :before_add - Symbol, Proc, or array of both/either specifying a callback to call
+        #     before a new item is added to the association.
+        #   - :before_remove - Symbol, Proc, or array of both/either specifying a callback to call
+        #     before an item is removed from the association.
+        #   - :class - The associated class or its name. If not
+        #     given, uses the association's name, which is camelized (and
+        #     singularized unless the type is :many_to_one)
+        #   - :clone - Merge the current options and block into the options and block used in defining
+        #     the given association.  Can be used to DRY up a bunch of similar associations that
+        #     all share the same options such as :class and :key, while changing the order and block used.
+        #   - :conditions - The conditions to use to filter the association, can be any argument passed to filter.
+        #   - :dataset - A proc that is instance_evaled to get the base dataset
+        #     to use for the _dataset method (before the other options are applied).
+        #   - :eager - The associations to eagerly load via #eager when loading the associated object(s).
+        #     For many_to_one associations, this is ignored unless this association is
+        #     being eagerly loaded, as it doesn't save queries unless multiple objects
+        #     can be loaded at once.
+        #   - :eager_block - If given, use the block instead of the default block when
+        #     eagerly loading.  To not use a block when eager loading (when one is used normally),
+        #     set to nil.
+        #   - :eager_graph - The associations to eagerly load via #eager_graph when loading the associated object(s).
+        #     For many_to_one associations, this is ignored unless this association is
+        #     being eagerly loaded, as it doesn't save queries unless multiple objects
+        #     can be loaded at once.
+        #   - :eager_grapher - A proc to use to implement eager loading via eager graph, overriding the default.
+        #     Takes three arguments, a dataset, an alias to use for the table to graph for this association,
+        #     and the alias that was used for the current table (since you can cascade associations),
+        #     Should return a copy of the dataset with the association graphed into it.
+        #   - :eager_loader - A proc to use to implement eager loading, overriding the default.  Takes three arguments,
+        #     a key hash (used solely to enhance performance), an array of records,
+        #     and a hash of dependent associations.  The associated records should
+        #     be queried from the database and the associations cache for each
+        #     record should be populated for this to work correctly.
+        #   - :extend - A module or array of modules to extend the dataset with.
+        #   - :graph_block - The block to pass to join_table when eagerly loading
+        #     the association via eager_graph.
+        #   - :graph_conditions - The additional conditions to use on the SQL join when eagerly loading
+        #     the association via eager_graph.  Should be a hash or an array of all two pairs. If not
+        #     specified, the :conditions option is used if it is a hash or array of all two pairs.
+        #   - :graph_join_type - The type of SQL join to use when eagerly loading the association via
+        #     eager_graph.  Defaults to :left_outer.
+        #   - :graph_only_conditions - The conditions to use on the SQL join when eagerly loading
+        #     the association via eager_graph, instead of the default conditions specified by the
+        #     foreign/primary keys.  This option causes the :graph_conditions option to be ignored.
+        #   - :graph_select - A column or array of columns to select from the associated table
+        #     when eagerly loading the association via eager_graph. Defaults to all
+        #     columns in the associated table.
+        #   - :limit - Limit the number of records to the provided value.  Use
+        #     an array with two arguments for the value to specify a limit and an offset.
+        #   - :order - the column(s) by which to order the association dataset.  Can be a
+        #     singular column or an array.
+        #   - :order_eager_graph - Whether to add the order to the dataset's order when graphing
+        #     via eager graph.  Defaults to true, so set to false to disable.
+        #   - :read_only - Do not add a setter method (for many_to_one or one_to_many with :one_to_one),
+        #     or add_/remove_/remove_all_ methods (for one_to_many, many_to_many)
+        #   - :reciprocal - the symbol name of the reciprocal association,
+        #     if it exists.  By default, sequel will try to determine it by looking at the
+        #     associated model's assocations for a association that matches
+        #     the current association's key(s).  Set to nil to not use a reciprocal.
+        #   - :select - the attributes to select.  Defaults to the associated class's
+        #     table_name.* in a many_to_many association, which means it doesn't include the attributes from the
+        #     join table.  If you want to include the join table attributes, you can
+        #     use this option, but beware that the join table attributes can clash with
+        #     attributes from the model table, so you should alias any attributes that have
+        #     the same name in both the join table and the associated table.
+        # * :many_to_one:
+        #   - :key - foreign_key in current model's table that references
+        #     associated model's primary key, as a symbol.  Defaults to :"#{name}_id".
+        #   - :primary_key - column in the associated table that :key option references, as a symbol.
+        #     Defaults to the primary key of the associated table.
+        # * :one_to_many:
+        #   - :key - foreign key in associated model's table that references
+        #     current model's primary key, as a symbol.  Defaults to
+        #     :"#{self.name.underscore}_id".
+        #   - :one_to_one: Create a getter and setter similar to those of many_to_one
+        #     associations.  The getter returns a singular matching record, or raises an
+        #     error if multiple records match.  The setter updates the record given and removes
+        #     associations with all other records. When this option is used, the other
+        #     association methods usually added are either removed or made private,
+        #     so using this is similar to using many_to_one, in terms of the methods
+        #     it adds, the main difference is that the foreign key is in the associated
+        #     table instead of the current table.  Note that using this option still requires
+        #     you to use a plural name when creating and using the association (e.g. for reflections, eager loading, etc.).
+        #   - :primary_key - column in the current table that :key option references, as a symbol.
+        #     Defaults to primary key of the current table.
+        # * :many_to_many:
+        #   - :graph_join_table_block - The block to pass to join_table for
+        #     the join table when eagerly loading the association via eager_graph.
+        #   - :graph_join_table_conditions - The additional conditions to use on the SQL join for
+        #     the join table when eagerly loading the association via eager_graph. Should be a hash
+        #     or an array of all two pairs.
+        #   - :graph_join_table_join_type - The type of SQL join to use for the join table when eagerly
+        #     loading the association via eager_graph.  Defaults to the :graph_join_type option or
+        #     :left_outer.
+        #   - :graph_join_table_only_conditions - The conditions to use on the SQL join for the join
+        #     table when eagerly loading the association via eager_graph, instead of the default
+        #     conditions specified by the foreign/primary keys.  This option causes the 
+        #     :graph_join_table_conditions option to be ignored.
+        #   - :join_table - name of table that includes the foreign keys to both
+        #     the current model and the associated model, as a symbol.  Defaults to the name
+        #     of current model and name of associated model, pluralized,
+        #     underscored, sorted, and joined with '_'.
+        #   - :left_key - foreign key in join table that points to current model's
+        #     primary key, as a symbol. Defaults to :"#{self.name.underscore}_id".
+        #   - :left_primary_key - column in current table that :left_key points to, as a symbol.
+        #     Defaults to primary key of current table.
+        #   - :right_key - foreign key in join table that points to associated
+        #     model's primary key, as a symbol.  Defaults to Defaults to :"#{name.to_s.singularize}_id".
+        #   - :right_primary_key - column in associated table that :right_key points to, as a symbol.
+        #     Defaults to primary key of the associated table.
+        #   - :uniq - Adds a after_load callback that makes the array of objects unique.
+        def associate(type, name, opts = {}, &block)
+          raise(Error, 'invalid association type') unless assoc_class = ASSOCIATION_TYPES[type]
+          raise(Error, 'Model.associate name argument must be a symbol') unless Symbol === name
+      
+          # merge early so we don't modify opts
+          orig_opts = opts.dup
+          orig_opts = association_reflection(opts[:clone])[:orig_opts].merge(orig_opts) if opts[:clone]
+          opts = orig_opts.merge(:type => type, :name => name, :cache => true, :model => self)
+          opts[:block] = block if block
+          opts = assoc_class.new.merge!(opts)
+          opts[:eager_block] = block unless opts.include?(:eager_block)
+          opts[:graph_join_type] ||= :left_outer
+          opts[:order_eager_graph] = true unless opts.include?(:order_eager_graph)
+          conds = opts[:conditions]
+          opts[:graph_conditions] = conds if !opts.include?(:graph_conditions) and Sequel.condition_specifier?(conds)
+          opts[:graph_conditions] = opts[:graph_conditions] ? opts[:graph_conditions].to_a : []
+          opts[:graph_select] = Array(opts[:graph_select]) if opts[:graph_select]
+          [:before_add, :before_remove, :after_add, :after_remove, :after_load, :extend].each do |cb_type|
+            opts[cb_type] = Array(opts[cb_type])
+          end
+      
+          # find class
+          case opts[:class]
+            when String, Symbol
+              # Delete :class to allow late binding
+              opts[:class_name] ||= opts.delete(:class).to_s
+            when Class
+              opts[:class_name] ||= opts[:class].name
+          end
+      
+          send(:"def_#{type}", opts)
+      
+          orig_opts.delete(:clone)
+          orig_opts.merge!(:class_name=>opts[:class_name], :class=>opts[:class], :block=>block)
+          opts[:orig_opts] = orig_opts
+          # don't add to association_reflections until we are sure there are no errors
+          association_reflections[name] = opts
+        end
+        
+        # The association reflection hash for the association of the given name.
+        def association_reflection(name)
+          association_reflections[name]
+        end
+        
+        # Array of association name symbols
+        def associations
+          association_reflections.keys
+        end
+
+        # Modify and return eager loading dataset based on association options
+        def eager_loading_dataset(opts, ds, select, associations)
+          ds = ds.select(*select) if select
+          ds = ds.filter(opts[:conditions]) if opts[:conditions]
+          ds = ds.order(*opts[:order]) if opts[:order]
+          ds = ds.eager(opts[:eager]) if opts[:eager]
+          ds = ds.eager_graph(opts[:eager_graph]) if opts[:eager_graph]
+          ds = ds.eager(associations) unless Array(associations).empty?
+          ds = opts[:eager_block].call(ds) if opts[:eager_block]
+          ds
+        end
+
+        # Copy the association reflections to the subclass
+        def inherited(subclass)
+          super
+          subclass.instance_variable_set(:@association_reflections, @association_reflections.dup)
+        end
+      
+        # Shortcut for adding a many_to_many association, see associate
+        def many_to_many(*args, &block)
+          associate(:many_to_many, *args, &block)
+        end
+        
+        # Shortcut for adding a many_to_one association, see associate
+        def many_to_one(*args, &block)
+          associate(:many_to_one, *args, &block)
+        end
+        
+        # Shortcut for adding a one_to_many association, see associate
+        def one_to_many(*args, &block)
+          associate(:one_to_many, *args, &block)
+        end
+        
+        private
+      
+        # Add a method to the module included in the class, so the method
+        # can be easily overridden in the class itself while allowing for
+        # super to be called.
+        def association_module_def(name, &block)
+          overridable_methods_module.module_eval{define_method(name, &block)}
+        end
+      
+        # Add a private method to the module included in the class.
+        def association_module_private_def(name, &block)
+          association_module_def(name, &block)
+          overridable_methods_module.send(:private, name)
+        end
+      
+        # Add the add_ instance method 
+        def def_add_method(opts)
+          association_module_def(opts.add_method){|o| add_associated_object(opts, o)}
+        end
+      
+        # Adds methods to the module included in the class related to getting the
+        # dataset and associated object(s).
+        def def_association_dataset_methods(opts)
+          # If a block is given, define a helper method for it, because it takes
+          # an argument.  This is unnecessary in Ruby 1.9, as that has instance_exec.
+          association_module_private_def(opts.dataset_helper_method, &opts[:block]) if opts[:block]
+          association_module_private_def(opts._dataset_method, &opts[:dataset])
+          association_module_def(opts.dataset_method){_dataset(opts)}
+          association_module_def(opts.association_method){|*reload| load_associated_objects(opts, reload[0])}
+        end
+      
+        # Adds many_to_many association instance methods
+        def def_many_to_many(opts)
+          name = opts[:name]
+          model = self
+          left = (opts[:left_key] ||= opts.default_left_key)
+          right = (opts[:right_key] ||= opts.default_right_key)
+          left_pk = (opts[:left_primary_key] ||= self.primary_key)
+          opts[:class_name] ||= camelize(singularize(name))
+          join_table = (opts[:join_table] ||= opts.default_join_table)
+          left_key_alias = opts[:left_key_alias] ||= :x_foreign_key_x
+          left_key_select = opts[:left_key_select] ||= SQL::QualifiedIdentifier.new(join_table, left).as(opts[:left_key_alias])
+          graph_jt_conds = opts[:graph_join_table_conditions] = opts[:graph_join_table_conditions] ? opts[:graph_join_table_conditions].to_a : []
+          opts[:graph_join_table_join_type] ||= opts[:graph_join_type]
+          opts[:after_load].unshift(:array_uniq!) if opts[:uniq]
+          opts[:dataset] ||= proc{opts.associated_class.inner_join(join_table, [[right, opts.right_primary_key], [left, send(left_pk)]])}
+          database = db
+          
+          opts[:eager_loader] ||= proc do |key_hash, records, associations|
+            h = key_hash[left_pk]
+            records.each{|object| object.associations[name] = []}
+            model.eager_loading_dataset(opts, opts.associated_class.inner_join(join_table, [[right, opts.right_primary_key], [left, h.keys]]), Array(opts.select) + Array(left_key_select), associations).all do |assoc_record|
+              next unless objects = h[assoc_record.values.delete(left_key_alias)]
+              objects.each{|object| object.associations[name].push(assoc_record)}
+            end
+          end
+          
+          join_type = opts[:graph_join_type]
+          select = opts[:graph_select]
+          use_only_conditions = opts.include?(:graph_only_conditions)
+          only_conditions = opts[:graph_only_conditions]
+          conditions = opts[:graph_conditions]
+          graph_block = opts[:graph_block]
+          use_jt_only_conditions = opts.include?(:graph_join_table_only_conditions)
+          jt_only_conditions = opts[:graph_join_table_only_conditions]
+          jt_join_type = opts[:graph_join_table_join_type]
+          jt_graph_block = opts[:graph_join_table_block]
+          opts[:eager_grapher] ||= proc do |ds, assoc_alias, table_alias|
+            ds = ds.graph(join_table, use_jt_only_conditions ? jt_only_conditions : [[left, left_pk]] + graph_jt_conds, :select=>false, :table_alias=>ds.send(:eager_unique_table_alias, ds, join_table), :join_type=>jt_join_type, :implicit_qualifier=>table_alias, &jt_graph_block)
+            ds.graph(opts.associated_class, use_only_conditions ? only_conditions : [[opts.right_primary_key, right]] + conditions, :select=>select, :table_alias=>assoc_alias, :join_type=>join_type, &graph_block)
+          end
+      
+          def_association_dataset_methods(opts)
+      
+          return if opts[:read_only]
+      
+          association_module_private_def(opts._add_method) do |o|
+            database.dataset.from(join_table).insert(left=>send(left_pk), right=>o.send(opts.right_primary_key))
+          end
+          association_module_private_def(opts._remove_method) do |o|
+            database.dataset.from(join_table).filter([[left, send(left_pk)], [right, o.send(opts.right_primary_key)]]).delete
+          end
+          association_module_private_def(opts._remove_all_method) do
+            database.dataset.from(join_table).filter(left=>send(left_pk)).delete
+          end
+      
+          def_add_method(opts)
+          def_remove_methods(opts)
+        end
+        
+        # Adds many_to_one association instance methods
+        def def_many_to_one(opts)
+          name = opts[:name]
+          model = self
+          opts[:key] = opts.default_key unless opts.include?(:key)
+          key = opts[:key]
+          opts[:class_name] ||= camelize(name)
+          opts[:dataset] ||= proc do
+            klass = opts.associated_class
+            klass.filter(SQL::QualifiedIdentifier.new(klass.table_name, opts.primary_key)=>send(key))
+          end
+          opts[:eager_loader] ||= proc do |key_hash, records, associations|
+            h = key_hash[key]
+            keys = h.keys
+            # Default the cached association to nil, so any object that doesn't have it
+            # populated will have cached the negative lookup.
+            records.each{|object| object.associations[name] = nil}
+            # Skip eager loading if no objects have a foreign key for this association
+            unless keys.empty?
+              klass = opts.associated_class
+              model.eager_loading_dataset(opts, klass.filter(SQL::QualifiedIdentifier.new(klass.table_name, opts.primary_key)=>keys), opts.select, associations).all do |assoc_record|
+                next unless objects = h[assoc_record.send(opts.primary_key)]
+                objects.each{|object| object.associations[name] = assoc_record}
+              end
+            end
+          end
+      
+          join_type = opts[:graph_join_type]
+          select = opts[:graph_select]
+          use_only_conditions = opts.include?(:graph_only_conditions)
+          only_conditions = opts[:graph_only_conditions]
+          conditions = opts[:graph_conditions]
+          graph_block = opts[:graph_block]
+          opts[:eager_grapher] ||= proc do |ds, assoc_alias, table_alias|
+            ds.graph(opts.associated_class, use_only_conditions ? only_conditions : [[opts.primary_key, key]] + conditions, :select=>select, :table_alias=>assoc_alias, :join_type=>join_type, :implicit_qualifier=>table_alias, &graph_block)
+          end
+      
+          def_association_dataset_methods(opts)
+          
+          return if opts[:read_only]
+      
+          association_module_private_def(opts._setter_method){|o| send(:"#{key}=", (o.send(opts.primary_key) if o))}
+          association_module_def(opts.setter_method){|o| set_associated_object(opts, o)}
+        end
+        
+        # Adds one_to_many association instance methods
+        def def_one_to_many(opts)
+          name = opts[:name]
+          model = self
+          key = (opts[:key] ||= opts.default_key)
+          primary_key = (opts[:primary_key] ||= self.primary_key)
+          opts[:class_name] ||= camelize(singularize(name))
+          opts[:dataset] ||= proc do
+            klass = opts.associated_class
+            klass.filter(SQL::QualifiedIdentifier.new(klass.table_name, key) => send(primary_key))
+          end
+          opts[:eager_loader] ||= proc do |key_hash, records, associations|
+            h = key_hash[primary_key]
+            records.each{|object| object.associations[name] = []}
+            reciprocal = opts.reciprocal
+            klass = opts.associated_class
+            model.eager_loading_dataset(opts, klass.filter(SQL::QualifiedIdentifier.new(klass.table_name, key)=>h.keys), opts.select, associations).all do |assoc_record|
+              next unless objects = h[assoc_record[key]]
+              objects.each do |object| 
+                object.associations[name].push(assoc_record)
+                assoc_record.associations[reciprocal] = object if reciprocal
+              end
+            end
+          end
+          
+          join_type = opts[:graph_join_type]
+          select = opts[:graph_select]
+          use_only_conditions = opts.include?(:graph_only_conditions)
+          only_conditions = opts[:graph_only_conditions]
+          conditions = opts[:graph_conditions]
+          graph_block = opts[:graph_block]
+          opts[:eager_grapher] ||= proc do |ds, assoc_alias, table_alias|
+            ds = ds.graph(opts.associated_class, use_only_conditions ? only_conditions : [[key, primary_key]] + conditions, :select=>select, :table_alias=>assoc_alias, :join_type=>join_type, :implicit_qualifier=>table_alias, &graph_block)
+            # We only load reciprocals for one_to_many associations, as other reciprocals don't make sense
+            ds.opts[:eager_graph][:reciprocals][assoc_alias] = opts.reciprocal
+            ds
+          end
+      
+          def_association_dataset_methods(opts)
+          
+          unless opts[:read_only]
+            association_module_private_def(opts._add_method) do |o|
+              o.send(:"#{key}=", send(primary_key))
+              o.save || raise(Sequel::Error, "invalid associated object, cannot save")
+            end
+            def_add_method(opts)
+      
+            unless opts[:one_to_one]
+              association_module_private_def(opts._remove_method) do |o|
+                o.send(:"#{key}=", nil)
+                o.save || raise(Sequel::Error, "invalid associated object, cannot save")
+              end
+              association_module_private_def(opts._remove_all_method) do
+                opts.associated_class.filter(key=>send(primary_key)).update(key=>nil)
+              end
+              def_remove_methods(opts)
+            end
+          end
+          if opts[:one_to_one]
+            overridable_methods_module.send(:private, opts.association_method, opts.dataset_method)
+            n = singularize(name).to_sym
+            raise(Sequel::Error, "one_to_many association names should still be plural even when using the :one_to_one option") if n == name
+            association_module_def(n) do |*o|
+              objs = send(name, *o)
+              raise(Sequel::Error, "multiple values found for a one-to-one relationship") if objs.length > 1
+              objs.first
+            end
+            unless opts[:read_only]
+              overridable_methods_module.send(:private, opts.add_method)
+              association_module_def(:"#{n}=") do |o|
+                klass = opts.associated_class
+                update_database = lambda do 
+                  send(opts.add_method, o)
+                  klass.filter(Sequel::SQL::BooleanExpression.new(:AND, {key=>send(primary_key)}, SQL::BooleanExpression.new(:'!=', klass.primary_key, o.pk))).update(key=>nil)
+                end
+                use_transactions ? db.transaction(opts){update_database.call} : update_database.call
+              end
+            end
+          end
+        end
+        
+        # Add the remove_ and remove_all instance methods
+        def def_remove_methods(opts)
+          association_module_def(opts.remove_method){|o| remove_associated_object(opts, o)}
+          association_module_def(opts.remove_all_method){remove_all_associated_objects(opts)}
+        end
+      end
+
+      # Private instance methods used to implement the associations support.
+      module InstanceMethods
+        private
+
+        # Backbone behind association dataset methods
+        def _dataset(opts)
+          raise(Sequel::Error, "model object #{model} does not have a primary key") if opts.dataset_need_primary_key? && !pk
+          ds = send(opts._dataset_method)
+          ds.extend(AssociationDatasetMethods)
+          ds.model_object = self
+          ds.association_reflection = opts
+          opts[:extend].each{|m| ds.extend(m)}
+          ds = ds.select(*opts.select) if opts.select
+          ds = ds.filter(opts[:conditions]) if opts[:conditions]
+          ds = ds.order(*opts[:order]) if opts[:order]
+          ds = ds.limit(*opts[:limit]) if opts[:limit]
+          ds = ds.eager(*opts[:eager]) if opts[:eager]
+          ds = ds.eager_graph(opts[:eager_graph]) if opts[:eager_graph] && opts.eager_graph_lazy_dataset?
+          ds = send(opts.dataset_helper_method, ds) if opts[:block]
+          ds
+        end
+
+        # Add the given associated object to the given association
+        def add_associated_object(opts, o)
+          raise(Sequel::Error, "model object #{model} does not have a primary key") unless pk
+          raise(Sequel::Error, "associated object #{o.model} does not have a primary key") if opts.need_associated_primary_key? && !o.pk
+          return if run_association_callbacks(opts, :before_add, o) == false
+          send(opts._add_method, o)
+          associations[opts[:name]].push(o) if associations.include?(opts[:name])
+          add_reciprocal_object(opts, o)
+          run_association_callbacks(opts, :after_add, o)
+          o
+        end
+
+        # Add/Set the current object to/as the given object's reciprocal association.
+        def add_reciprocal_object(opts, o)
+          return unless reciprocal = opts.reciprocal
+          if opts.reciprocal_array?
+            if array = o.associations[reciprocal] and !array.include?(self)
+              array.push(self)
+            end
+          else
+            o.associations[reciprocal] = self
+          end
+        end
+        
+        # Call uniq! on the given array. This is used by the :uniq option,
+        # and is an actual method for memory reasons.
+        def array_uniq!(a)
+          a.uniq!
+        end
+
+        # Load the associated objects using the dataset
+        def load_associated_objects(opts, reload=false)
+          name = opts[:name]
+          if associations.include?(name) and !reload
+            associations[name]
+          else
+            objs = if opts.returns_array?
+              send(opts.dataset_method).all
+            else
+              if !opts[:key]
+                send(opts.dataset_method).all.first
+              elsif send(opts[:key])
+                send(opts.dataset_method).first
+              end
+            end
+            run_association_callbacks(opts, :after_load, objs)
+            objs.each{|o| add_reciprocal_object(opts, o)} if opts.set_reciprocal_to_self?
+            associations[name] = objs
+          end
+        end
+
+        # Remove all associated objects from the given association
+        def remove_all_associated_objects(opts)
+          raise(Sequel::Error, "model object #{model} does not have a primary key") unless pk
+          send(opts._remove_all_method)
+          ret = associations[opts[:name]].each{|o| remove_reciprocal_object(opts, o)} if associations.include?(opts[:name])
+          associations[opts[:name]] = []
+          ret
+        end
+
+        # Remove the given associated object from the given association
+        def remove_associated_object(opts, o)
+          raise(Sequel::Error, "model object #{model} does not have a primary key") unless pk
+          raise(Sequel::Error, "associated object #{o.model} does not have a primary key") if opts.need_associated_primary_key? && !o.pk
+          return if run_association_callbacks(opts, :before_remove, o) == false
+          send(opts._remove_method, o)
+          associations[opts[:name]].delete_if{|x| o === x} if associations.include?(opts[:name])
+          remove_reciprocal_object(opts, o)
+          run_association_callbacks(opts, :after_remove, o)
+          o
+        end
+
+        # Remove/unset the current object from/as the given object's reciprocal association.
+        def remove_reciprocal_object(opts, o)
+          return unless reciprocal = opts.reciprocal
+          if opts.reciprocal_array?
+            if array = o.associations[reciprocal]
+              array.delete_if{|x| self === x}
+            end
+          else
+            o.associations[reciprocal] = nil
+          end
+        end
+
+        # Run the callback for the association with the object.
+        def run_association_callbacks(reflection, callback_type, object)
+          raise_error = raise_on_save_failure || !reflection.returns_array?
+          stop_on_false = [:before_add, :before_remove].include?(callback_type)
+          reflection[callback_type].each do |cb|
+            res = case cb
+            when Symbol
+              send(cb, object)
+            when Proc
+              cb.call(self, object)
+            else
+              raise Error, "callbacks should either be Procs or Symbols"
+            end
+            if res == false and stop_on_false
+              raise(BeforeHookFailed, "Unable to modify association for record: one of the #{callback_type} hooks returned false") if raise_error
+              return false
+            end
+          end
+        end
+
+        # Set the given object as the associated object for the given association
+        def set_associated_object(opts, o)
+          raise(Sequel::Error, "model object #{model} does not have a primary key") if o && !o.pk
+          old_val = send(opts.association_method)
+          return o if old_val == o
+          return if old_val and run_association_callbacks(opts, :before_remove, old_val) == false
+          return if o and run_association_callbacks(opts, :before_add, o) == false
+          send(opts._setter_method, o)
+          associations[opts[:name]] = o
+          remove_reciprocal_object(opts, old_val) if old_val
+          if o
+            add_reciprocal_object(opts, o)
+            run_association_callbacks(opts, :after_add, o)
+          end
+          run_association_callbacks(opts, :after_remove, old_val) if old_val
+          o
+        end
+      end
+
+      # Eager loading makes it so that you can load all associated records for a
+      # set of objects in a single query, instead of a separate query for each object.
+      #
+      # Two separate implementations are provided.  #eager should be used most of the
+      # time, as it loads associated records using one query per association.  However,
+      # it does not allow you the ability to filter based on columns in associated tables.  #eager_graph loads
+      # all records in one query.  Using #eager_graph you can filter based on columns in associated
+      # tables.  However, #eager_graph can be slower than #eager, especially if multiple
+      # *_to_many associations are joined.
+      #
+      # You can cascade the eager loading (loading associations' associations)
+      # with no limit to the depth of the cascades.  You do this by passing a hash to #eager or #eager_graph
+      # with the keys being associations of the current model and values being
+      # associations of the model associated with the current model via the key.
+      #  
+      # The arguments can be symbols or hashes with symbol keys (for cascaded
+      # eager loading). Examples:
+      #
+      #   Album.eager(:artist).all
+      #   Album.eager_graph(:artist).all
+      #   Album.eager(:artist, :genre).all
+      #   Album.eager_graph(:artist, :genre).all
+      #   Album.eager(:artist).eager(:genre).all
+      #   Album.eager_graph(:artist).eager(:genre).all
+      #   Artist.eager(:albums=>:tracks).all
+      #   Artist.eager_graph(:albums=>:tracks).all
+      #   Artist.eager(:albums=>{:tracks=>:genre}).all
+      #   Artist.eager_graph(:albums=>{:tracks=>:genre}).all
+      module DatasetMethods
+        # Add the #eager! and #eager_graph! mutation methods to the dataset.
+        def self.extended(obj)
+          obj.def_mutation_method(:eager, :eager_graph)
+        end
+      
+        # The preferred eager loading method.  Loads all associated records using one
+        # query for each association.
+        #
+        # The basic idea for how it works is that the dataset is first loaded normally.
+        # Then it goes through all associations that have been specified via eager.
+        # It loads each of those associations separately, then associates them back
+        # to the original dataset via primary/foreign keys.  Due to the necessity of
+        # all objects being present, you need to use .all to use eager loading, as it
+        # can't work with .each.
+        #
+        # This implementation avoids the complexity of extracting an object graph out
+        # of a single dataset, by building the object graph out of multiple datasets,
+        # one for each association.  By using a separate dataset for each association,
+        # it avoids problems such as aliasing conflicts and creating cartesian product
+        # result sets if multiple *_to_many eager associations are requested.
+        #
+        # One limitation of using this method is that you cannot filter the dataset
+        # based on values of columns in an associated table, since the associations are loaded
+        # in separate queries.  To do that you need to load all associations in the
+        # same query, and extract an object graph from the results of that query. If you
+        # need to filter based on columns in associated tables, look at #eager_graph
+        # or join the tables you need to filter on manually. 
+        #
+        # Each association's order, if defined, is respected. Eager also works
+        # on a limited dataset, but does not use any :limit options for associations.
+        # If the association uses a block or has an :eager_block argument, it is used.
+        def eager(*associations)
+          opt = @opts[:eager]
+          opt = opt ? opt.dup : {}
+          associations.flatten.each do |association|
+            case association
+              when Symbol
+                check_association(model, association)
+                opt[association] = nil
+              when Hash
+                association.keys.each{|assoc| check_association(model, assoc)}
+                opt.merge!(association)
+              else raise(Sequel::Error, 'Associations must be in the form of a symbol or hash')
+            end
+          end
+          clone(:eager=>opt)
+        end
+      
+        # The secondary eager loading method.  Loads all associations in a single query. This
+        # method should only be used if you need to filter based on columns in associated tables.
+        #
+        # This method builds an object graph using Dataset#graph.  Then it uses the graph
+        # to build the associations, and finally replaces the graph with a simple array
+        # of model objects.
+        #
+        # Be very careful when using this with multiple *_to_many associations, as you can
+        # create large cartesian products.  If you must graph multiple *_to_many associations,
+        # make sure your filters are specific if you have a large database.
+        # 
+        # Each association's order, if definied, is respected. #eager_graph probably
+        # won't work correctly on a limited dataset, unless you are
+        # only graphing many_to_one associations.
+        # 
+        # Does not use the block defined for the association, since it does a single query for
+        # all objects.  You can use the :graph_* association options to modify the SQL query.
+        #
+        # Like eager, you need to call .all on the dataset for the eager loading to work.  If you just
+        # call each, you will get a normal graphed result back (a hash with model object values).
+        def eager_graph(*associations)
+          table_name = model.table_name
+          ds = if @opts[:eager_graph]
+            self
+          else
+            # Each of the following have a symbol key for the table alias, with the following values: 
+            # :reciprocals - the reciprocal instance variable to use for this association
+            # :requirements - array of requirements for this association
+            # :alias_association_type_map - the type of association for this association
+            # :alias_association_name_map - the name of the association for this association
+            clone(:eager_graph=>{:requirements=>{}, :master=>model.table_name, :alias_association_type_map=>{}, :alias_association_name_map=>{}, :reciprocals=>{}})
+          end
+          ds.eager_graph_associations(ds, model, table_name, [], *associations)
+        end
+        
+        protected
+      
+        # Call graph on the association with the correct arguments,
+        # update the eager_graph data structure, and recurse into
+        # eager_graph_associations if there are any passed in associations
+        # (which would be dependencies of the current association)
+        #
+        # Arguments:
+        # * ds - Current dataset
+        # * model - Current Model
+        # * ta - table_alias used for the parent association
+        # * requirements - an array, used as a stack for requirements
+        # * r - association reflection for the current association
+        # * *associations - any associations dependent on this one
+        def eager_graph_association(ds, model, ta, requirements, r, *associations)
+          klass = r.associated_class
+          assoc_name = r[:name]
+          assoc_table_alias = ds.eager_unique_table_alias(ds, assoc_name)
+          ds = r[:eager_grapher].call(ds, assoc_table_alias, ta)
+          ds = ds.order_more(*Array(r[:order]).map{|c| eager_graph_qualify_order(assoc_table_alias, c)}) if r[:order] and r[:order_eager_graph]
+          eager_graph = ds.opts[:eager_graph]
+          eager_graph[:requirements][assoc_table_alias] = requirements.dup
+          eager_graph[:alias_association_name_map][assoc_table_alias] = assoc_name
+          eager_graph[:alias_association_type_map][assoc_table_alias] = r.returns_array?
+          ds = ds.eager_graph_associations(ds, r.associated_class, assoc_table_alias, requirements + [assoc_table_alias], *associations) unless associations.empty?
+          ds
+        end
+      
+        # Check the associations are valid for the given model.
+        # Call eager_graph_association on each association.
+        #
+        # Arguments:
+        # * ds - Current dataset
+        # * model - Current Model
+        # * ta - table_alias used for the parent association
+        # * requirements - an array, used as a stack for requirements
+        # * *associations - the associations to add to the graph
+        def eager_graph_associations(ds, model, ta, requirements, *associations)
+          return ds if associations.empty?
+          associations.flatten.each do |association|
+            ds = case association
+            when Symbol
+              ds.eager_graph_association(ds, model, ta, requirements, check_association(model, association))
+            when Hash
+              association.each do |assoc, assoc_assocs|
+                ds = ds.eager_graph_association(ds, model, ta, requirements, check_association(model, assoc), assoc_assocs)
+              end
+              ds
+            else raise(Sequel::Error, 'Associations must be in the form of a symbol or hash')
+            end
+          end
+          ds
+        end
+      
+        # Build associations out of the array of returned object graphs.
+        def eager_graph_build_associations(record_graphs)
+          eager_graph = @opts[:eager_graph]
+          master = eager_graph[:master]
+          requirements = eager_graph[:requirements]
+          alias_map = eager_graph[:alias_association_name_map]
+          type_map = eager_graph[:alias_association_type_map]
+          reciprocal_map = eager_graph[:reciprocals]
+      
+          # Make dependency map hash out of requirements array for each association.
+          # This builds a tree of dependencies that will be used for recursion
+          # to ensure that all parts of the object graph are loaded into the
+          # appropriate subordinate association.
+          dependency_map = {}
+          # Sort the associations by requirements length, so that
+          # requirements are added to the dependency hash before their
+          # dependencies.
+          requirements.sort_by{|a| a[1].length}.each do |ta, deps|
+            if deps.empty?
+              dependency_map[ta] = {}
+            else
+              deps = deps.dup
+              hash = dependency_map[deps.shift]
+              deps.each do |dep|
+                hash = hash[dep]
+              end
+              hash[ta] = {}
+            end
+          end
+      
+          # This mapping is used to make sure that duplicate entries in the
+          # result set are mapped to a single record.  For example, using a
+          # single one_to_many association with 10 associated records,
+          # the main object will appear in the object graph 10 times.
+          # We map by primary key, if available, or by the object's entire values,
+          # if not. The mapping must be per table, so create sub maps for each table
+          # alias.
+          records_map = {master=>{}}
+          alias_map.keys.each{|ta| records_map[ta] = {}}
+      
+          # This will hold the final record set that we will be replacing the object graph with.
+          records = []
+          record_graphs.each do |record_graph|
+            primary_record = record_graph[master]
+            key = primary_record.pk || primary_record.values.sort_by{|x| x[0].to_s}
+            if cached_pr = records_map[master][key]
+              primary_record = cached_pr
+            else
+              records_map[master][key] = primary_record
+              # Only add it to the list of records to return if it is a new record
+              records.push(primary_record)
+            end
+            # Build all associations for the current object and it's dependencies
+            eager_graph_build_associations_graph(dependency_map, alias_map, type_map, reciprocal_map, records_map, primary_record, record_graph)
+          end
+      
+          # Remove duplicate records from all associations if this graph could possibly be a cartesian product
+          eager_graph_make_associations_unique(records, dependency_map, alias_map, type_map) if type_map.values.select{|v| v}.length > 1
+          
+          # Replace the array of object graphs with an array of model objects
+          record_graphs.replace(records)
+        end
+      
+        # Creates a unique table alias that hasn't already been used in the query.
+        # Will either be the table_alias itself or table_alias_N for some integer
+        # N (starting at 0 and increasing until an unused one is found).
+        def eager_unique_table_alias(ds, table_alias)
+          used_aliases = ds.opts[:from]
+          graph = ds.opts[:graph]
+          used_aliases += graph[:table_aliases].keys if graph
+          if used_aliases.include?(table_alias)
+            i = 0
+            loop do
+              ta = :"#{table_alias}_#{i}"
+              return ta unless used_aliases.include?(ta)
+              i += 1
+            end
+          end
+          table_alias
+        end
+      
+        private
+      
+        # Make sure the association is valid for this model, and return the related AssociationReflection.
+        def check_association(model, association)
+          raise(Sequel::Error, 'Invalid association') unless reflection = model.association_reflection(association)
+          raise(Sequel::Error, "Eager loading is not allowed for #{model.name} association #{association}") if reflection[:allow_eager] == false
+          reflection
+        end
+      
+        # Build associations for the current object.  This is called recursively
+        # to build all dependencies.
+        def eager_graph_build_associations_graph(dependency_map, alias_map, type_map, reciprocal_map, records_map, current, record_graph)
+          return if dependency_map.empty?
+          # Don't clobber the instance variable array for *_to_many associations if it has already been setup
+          dependency_map.keys.each do |ta|
+            assoc_name = alias_map[ta]
+            current.associations[assoc_name] = type_map[ta] ? [] : nil unless current.associations.include?(assoc_name)
+          end
+          dependency_map.each do |ta, deps|
+            next unless rec = record_graph[ta]
+            key = rec.pk || rec.values.sort_by{|x| x[0].to_s}
+            if cached_rec = records_map[ta][key]
+              rec = cached_rec
+            else
+              records_map[ta][rec.pk] = rec
+            end
+            assoc_name = alias_map[ta]
+            case type_map[ta]
+            when false
+              current.associations[assoc_name] = rec
+            else
+              current.associations[assoc_name].push(rec) 
+              if reciprocal = reciprocal_map[ta]
+                rec.associations[reciprocal] = current
+              end
+            end
+            # Recurse into dependencies of the current object
+            eager_graph_build_associations_graph(deps, alias_map, type_map, reciprocal_map, records_map, rec, record_graph)
+          end
+        end
+      
+        # If the result set is the result of a cartesian product, then it is possible that
+        # there are multiple records for each association when there should only be one.
+        # In that case, for each object in all associations loaded via #eager_graph, run
+        # uniq! on the association to make sure no duplicate records show up.
+        # Note that this can cause legitimate duplicate records to be removed.
+        def eager_graph_make_associations_unique(records, dependency_map, alias_map, type_map)
+          records.each do |record|
+            dependency_map.each do |ta, deps|
+              list = if !type_map[ta]
+                item = record.send(alias_map[ta])
+                [item] if item
+              else
+                list = record.send(alias_map[ta])
+                list.uniq!
+              end
+              # Recurse into dependencies
+              eager_graph_make_associations_unique(list, deps, alias_map, type_map) if list
+            end
+          end
+        end
+      
+        # Qualify the given expression if necessary.  The only expressions which are qualified are
+        # unqualified symbols and identifiers, either of which may by sorted.
+        def eager_graph_qualify_order(table_alias, expression)
+          case expression
+          when Symbol
+            table, column, aliaz = split_symbol(expression)
+            raise(Sequel::Error, "Can't use an aliased expression in the :order option") if aliaz
+            table ? expression : Sequel::SQL::QualifiedIdentifier.new(table_alias, expression)
+          when Sequel::SQL::Identifier
+            Sequel::SQL::QualifiedIdentifier.new(table_alias, expression)
+          when Sequel::SQL::OrderedExpression
+            Sequel::SQL::OrderedExpression.new(eager_graph_qualify_order(table_alias, expression.expression), expression.descending)
+          else
+            expression
+          end
+        end
+      
+        # Eagerly load all specified associations 
+        def eager_load(a)
+          return if a.empty?
+          # All associations to eager load
+          eager_assoc = @opts[:eager]
+          # Key is foreign/primary key name symbol
+          # Value is hash with keys being foreign/primary key values (generally integers)
+          #  and values being an array of current model objects with that
+          #  specific foreign/primary key
+          key_hash = {}
+          # Reflections for all associations to eager load
+          reflections = eager_assoc.keys.collect{|assoc| model.association_reflection(assoc)}
+      
+          # Populate keys to monitor
+          reflections.each{|reflection| key_hash[reflection.eager_loader_key] ||= Hash.new{|h,k| h[k] = []}}
+          
+          # Associate each object with every key being monitored
+          a.each do |rec|
+            key_hash.each do |key, id_map|
+              id_map[rec[key]] << rec if rec[key]
+            end
+          end
+          
+          reflections.each do |r|
+            r[:eager_loader].call(key_hash, a, eager_assoc[r[:name]])
+            a.each{|object| object.send(:run_association_callbacks, r, :after_load, object.associations[r[:name]])} unless r[:after_load].empty?
+          end 
+        end
+      
+        # Build associations from the graph if #eager_graph was used, 
+        # and/or load other associations if #eager was used.
+        def post_load(all_records)
+          eager_graph_build_associations(all_records) if @opts[:eager_graph]
+          eager_load(all_records) if @opts[:eager]
+        end
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/model/base.rb 3.0.0-1/lib/sequel/model/base.rb
--- 1.5.1-1/lib/sequel/model/base.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/model/base.rb	2009-04-29 00:13:39.000000000 +0100
@@ -0,0 +1,939 @@
+module Sequel
+  class Model
+    extend Enumerable
+    extend Inflections
+    extend Metaprogramming
+    include Metaprogramming
+
+    # Class methods for Sequel::Model that implement basic model functionality.
+    #
+    # * All of the method names in Model::DATASET_METHODS have class methods created that call
+    #   the Model's dataset with the method of the same name with the given arguments.
+    module ClassMethods
+      # Which columns should be the only columns allowed in a call to set
+      # (default: not set, so all columns not otherwise restricted).
+      attr_reader :allowed_columns
+  
+      # Array of modules that extend this model's dataset.  Stored
+      # so that if the model's dataset is changed, it will be extended
+      # with all of these modules.
+      attr_reader :dataset_method_modules
+
+      # Hash of dataset methods with method name keys and proc values that are
+      # stored so when the dataset changes, methods defined with def_dataset_method
+      # will be applied to the new dataset.
+      attr_reader :dataset_methods
+  
+      # The primary key for the class.  Sequel can determine this automatically for
+      # many databases, but not all, so you may need to set it manually.  If not
+      # determined automatically, the default is :id.
+      attr_reader :primary_key
+  
+      # Whether to raise an error instead of returning nil on a failure
+      # to save/create/save_changes/etc due to a validation failure or
+      # a before_* hook returning false.
+      attr_accessor :raise_on_save_failure
+  
+      # Whether to raise an error when unable to typecast data for a column
+      # (default: true).  This should be set to false if you want to use
+      # validations to display nice error messages to the user (e.g. most
+      # web applications).  You can use the validates_not_string validations
+      # (from either the validation_helpers or validation_class_methods standard
+      # plugins) in connection with option to check for typecast failures for
+      # columns that aren't blobs or strings.
+      attr_accessor :raise_on_typecast_failure
+  
+      # Which columns are specifically restricted in a call to set/update/new/etc.
+      # (default: not set).  Some columns are restricted regardless of
+      # this setting, such as the primary key column and columns in Model::RESTRICTED_SETTER_METHODS.
+      attr_reader :restricted_columns
+  
+      # Should be the literal primary key column name if this Model's table has a simple primary key, or
+      # nil if the model has a compound primary key or no primary key.
+      attr_reader :simple_pk
+  
+      # Should be the literal table name if this Model's dataset is a simple table (no select, order, join, etc.),
+      # or nil otherwise.  This and simple_pk are used for an optimization in Model.[].
+      attr_reader :simple_table
+  
+      # Whether new/set/update and their variants should raise an error
+      # if an invalid key is used.  A key is invalid if no setter method exists
+      # for that key or the access to the setter method is restricted (e.g. due to it
+      # being a primary key field).  If set to false, silently skip
+      # any key where the setter method doesn't exist or access to it is restricted.
+      attr_accessor :strict_param_setting
+  
+      # Whether to typecast the empty string ('') to nil for columns that
+      # are not string or blob.  In most cases the empty string would be the
+      # way to specify a NULL SQL value in string form (nil.to_s == ''),
+      # and an empty string would not usually be typecast correctly for other
+      # types, so the default is true.
+      attr_accessor :typecast_empty_string_to_nil
+  
+      # Whether to typecast attribute values on assignment (default: true).
+      # If set to false, no typecasting is done, so it will be left up to the
+      # database to typecast the value correctly.
+      attr_accessor :typecast_on_assignment
+  
+      # Whether to use a transaction by default when saving/deleting records (default: true).
+      # If you are sending database queries in before_* or after_* hooks, you shouldn't change
+      # the default setting without a good reason.
+      attr_accessor :use_transactions
+  
+      # Returns the first record from the database matching the conditions.
+      # If a hash is given, it is used as the conditions.  If another
+      # object is given, it finds the first record whose primary key(s) match
+      # the given argument(s).  
+      def [](*args)
+        args = args.first if (args.size == 1)
+        return dataset[args] if args.is_a?(Hash)
+        if t = simple_table and p = simple_pk
+          with_sql("SELECT * FROM #{t} WHERE #{p} = #{dataset.literal(args)}").first
+        else
+          dataset[primary_key_hash(args)]
+        end
+      end
+      
+      # Returns the columns in the result set in their original order.
+      # Generally, this will use the columns determined via the database
+      # schema, but in certain cases (e.g. models that are based on a joined
+      # dataset) it will use Dataset#columns to find the columns, which
+      # may be empty if the Dataset has no records.
+      def columns
+        @columns || set_columns(dataset.naked.columns)
+      end
+    
+      # Creates instance using new with the given values and block, and saves it.
+      def create(values = {}, &block)
+        new(values, &block).save
+      end
+  
+      # Returns the dataset associated with the Model class.  Raises
+      # an error if there is no associated dataset for this class.
+      def dataset
+        @dataset || raise(Error, "No dataset associated with #{self}")
+      end
+    
+      # Returns the database associated with the Model class.
+      # If this model doesn't have a database associated with it,
+      # assumes the superclass's database, or the first object in
+      # Sequel::DATABASES.  If no Sequel::Database object has
+      # been created, raises an error.
+      def db
+        return @db if @db
+        @db = self == Model ? DATABASES.first : superclass.db
+        raise(Error, "No database associated with #{self}") unless @db
+        @db
+      end
+      
+      # Sets the database associated with the Model class. If the
+      # model has an associated dataset, sets the model's dataset
+      # to a dataset on the new database with the same options
+      # used by the current dataset.
+      def db=(db)
+        @db = db
+        set_dataset(db.dataset(@dataset.opts)) if @dataset
+      end
+      
+      # Returns the cached schema information if available or gets it
+      # from the database.
+      def db_schema
+        @db_schema ||= get_db_schema
+      end
+  
+      # If a block is given, define a method on the dataset (if the model has an associated dataset)  with the given argument name using
+      # the given block as well as a method on the model that calls the
+      # dataset method.  Stores the method name and block so that it can be reapplied if the model's
+      # dataset changes.
+      #
+      # If a block is not given, define a method on the model for each argument
+      # that calls the dataset method of the same argument name.
+      def def_dataset_method(*args, &block)
+        raise(Error, "No arguments given") if args.empty?
+        if block_given?
+          raise(Error, "Defining a dataset method using a block requires only one argument") if args.length > 1
+          meth = args.first
+          @dataset_methods[meth] = block
+          dataset.meta_def(meth, &block) if @dataset
+        end
+        args.each{|arg| instance_eval("def #{arg}(*args, &block); dataset.#{arg}(*args, &block) end", __FILE__, __LINE__) unless respond_to?(arg)}
+      end
+      
+      # Finds a single record according to the supplied filter, e.g.:
+      #
+      #   Ticket.find :author => 'Sharon' # => record
+      #
+      # You are encouraged to use Model.[] or Model.first instead of this method.
+      def find(*args, &block)
+        filter(*args, &block).first
+      end
+      
+      # Like find but invokes create with given conditions when record does not
+      # exist.
+      def find_or_create(cond)
+        find(cond) || create(cond)
+      end
+    
+      # If possible, set the dataset for the model subclass as soon as it
+      # is created.  Also, make sure the inherited class instance variables
+      # are copied into the subclass.
+      def inherited(subclass)
+        ivs = subclass.instance_variables.collect{|x| x.to_s}
+        EMPTY_INSTANCE_VARIABLES.each{|iv| subclass.instance_variable_set(iv, nil) unless ivs.include?(iv.to_s)}
+        INHERITED_INSTANCE_VARIABLES.each do |iv, dup|
+          next if ivs.include?(iv.to_s)
+          sup_class_value = instance_variable_get(iv)
+          sup_class_value = sup_class_value.dup if dup == :dup && sup_class_value
+          subclass.instance_variable_set(iv, sup_class_value)
+        end
+        unless ivs.include?("@dataset")
+          db
+          begin
+            if self == Model
+              subclass.set_dataset(subclass.implicit_table_name) unless subclass.name.empty?
+            elsif ds = instance_variable_get(:@dataset)
+              subclass.set_dataset(ds.clone, :inherited=>true)
+            end
+          rescue
+            nil
+          end
+        end
+      end
+    
+      # Returns the implicit table name for the model class.
+      def implicit_table_name
+        pluralize(underscore(demodulize(name))).to_sym
+      end
+  
+      # Initializes a model instance as an existing record. This constructor is
+      # used by Sequel to initialize model instances when fetching records.
+      # load requires that values be a hash where all keys are symbols. It
+      # probably should not be used by external code.
+      def load(values)
+        new(values, true)
+      end
+  
+      # Mark the model as not having a primary key. Not having a primary key
+      # can cause issues, among which is that you won't be able to update records.
+      def no_primary_key
+        @simple_pk = @primary_key = nil
+      end
+  
+      # Returns primary key attribute hash.  If using a composite primary key
+      # value such be an array with values for each primary key in the correct
+      # order.  For a standard primary key, value should be an object with a
+      # compatible type for the key.  If the model does not have a primary key,
+      # raises an Error.
+      def primary_key_hash(value)
+        raise(Error, "#{self} does not have a primary key") unless key = @primary_key
+        case key
+        when Array
+          hash = {}
+          key.each_with_index{|k,i| hash[k] = value[i]}
+          hash
+        else
+          {key => value}
+        end
+      end
+  
+      # Restrict the setting of the primary key(s) inside new/set/update.  Because
+      # this is the default, this only make sense to use in a subclass where the
+      # parent class has used unrestrict_primary_key.
+      def restrict_primary_key
+        @restrict_primary_key = true
+      end
+  
+      # Whether or not setting the primary key inside new/set/update is
+      # restricted, true by default.
+      def restrict_primary_key?
+        @restrict_primary_key
+      end
+  
+      # Set the columns to allow in new/set/update.  Using this means that
+      # any columns not listed here will not be modified.  If you have any virtual
+      # setter methods (methods that end in =) that you want to be used in
+      # new/set/update, they need to be listed here as well (without the =).
+      #
+      # It may be better to use (set|update)_only instead of this in places where
+      # only certain columns may be allowed.
+      def set_allowed_columns(*cols)
+        @allowed_columns = cols
+      end
+  
+      # Sets the dataset associated with the Model class. ds can be a Symbol
+      # (specifying a table name in the current database), or a Dataset.
+      # If a dataset is used, the model's database is changed to the given
+      # dataset.  If a symbol is used, a dataset is created from the current
+      # database with the table name given. Other arguments raise an Error.
+      #
+      # This changes the row_proc of the given dataset to return
+      # model objects, extends the dataset with the dataset_method_modules,
+      # and defines methods on the dataset using the dataset_methods.
+      # It also attempts to determine the database schema for the model,
+      # based on the given dataset.
+      def set_dataset(ds, opts={})
+        inherited = opts[:inherited]
+        @dataset = case ds
+        when Symbol
+          @simple_table = db.literal(ds)
+          db[ds]
+        when Dataset
+          @simple_table = nil
+          @db = ds.db
+          ds
+        else
+          raise(Error, "Model.set_dataset takes a Symbol or a Sequel::Dataset")
+        end
+        @dataset.row_proc = Proc.new{|r| load(r)}
+        if inherited
+          @simple_table = superclass.simple_table
+          @columns = @dataset.columns rescue nil
+        else
+          @dataset_method_modules.each{|m| @dataset.extend(m)} if @dataset_method_modules
+          @dataset_methods.each{|meth, block| @dataset.meta_def(meth, &block)} if @dataset_methods
+        end
+        @dataset.model = self if @dataset.respond_to?(:model=)
+        @db_schema = (inherited ? superclass.db_schema : get_db_schema) rescue nil
+        self
+      end
+      alias dataset= set_dataset
+    
+      # Sets the primary key for this model. You can use either a regular 
+      # or a composite primary key.
+      #
+      # Example:
+      #   class Tagging < Sequel::Model
+      #     # composite key
+      #     set_primary_key :taggable_id, :tag_id
+      #   end
+      #
+      #   class Person < Sequel::Model
+      #     # regular key
+      #     set_primary_key :person_id
+      #   end
+      #
+      # You can set it to nil to not have a primary key, but that
+      # cause certain things not to work, see no_primary_key.
+      def set_primary_key(*key)
+        @simple_pk = key.length == 1 ? db.literal(key.first) : nil 
+        @primary_key = (key.length == 1) ? key[0] : key.flatten
+      end
+  
+      # Set the columns to restrict in new/set/update.  Using this means that
+      # attempts to call setter methods for the columns listed here will cause an
+      # exception or be silently skipped (based on the strict_param_setting setting.
+      # If you have any virtual # setter methods (methods that end in =) that you
+      # want not to be used in new/set/update, they need to be listed here as well (without the =).
+      #
+      # It may be better to use (set|update)_except instead of this in places where
+      # only certain columns may be allowed.
+      def set_restricted_columns(*cols)
+        @restricted_columns = cols
+      end
+  
+      # Defines a method that returns a filtered dataset.  Subsets
+      # create dataset methods, so they can be chained for scoping.
+      # For example:
+      #
+      #   Topic.subset(:joes, :username.like('%joe%'))
+      #   Topic.subset(:popular){|o| o.num_posts > 100}
+      #   Topic.subset(:recent){|o| o.created_on > Date.today - 7}
+      #
+      # Allows you to do:
+      #
+      #   Topic.joes.recent.popular
+      #
+      # to get topics with a username that includes joe that
+      # have more than 100 posts and were created less than
+      # 7 days ago.
+      #
+      # Both the args given and the block are passed to Dataset#filter.
+      def subset(name, *args, &block)
+        def_dataset_method(name){filter(*args, &block)}
+      end
+      
+      # Returns name of primary table for the dataset.
+      def table_name
+        dataset.opts[:from].first
+      end
+  
+      # Allow the setting of the primary key(s) inside new/set/update.
+      def unrestrict_primary_key
+        @restrict_primary_key = false
+      end
+  
+      private
+      
+      # Create a column accessor for a column with a method name that is hard to use in ruby code.
+      def def_bad_column_accessor(column)
+        overridable_methods_module.module_eval do
+          define_method(column){self[column]}
+          define_method("#{column}="){|v| self[column] = v}
+        end
+      end
+  
+      # Create the column accessors.  For columns that can be used as method names directly in ruby code,
+      # use a string to define the method for speed.  For other columns names, use a block.
+      def def_column_accessor(*columns)
+        columns, bad_columns = columns.partition{|x| NORMAL_METHOD_NAME_REGEXP.match(x.to_s)}
+        bad_columns.each{|x| def_bad_column_accessor(x)}
+        im = instance_methods.collect{|x| x.to_s}
+        columns.each do |column|
+          meth = "#{column}="
+          overridable_methods_module.module_eval("def #{column}; self[:#{column}] end") unless im.include?(column.to_s)
+          overridable_methods_module.module_eval("def #{meth}(v); self[:#{column}] = v end") unless im.include?(meth)
+        end
+      end
+  
+      # Get the schema from the database, fall back on checking the columns
+      # via the database if that will return inaccurate results or if
+      # it raises an error.
+      def get_db_schema(reload = false)
+        set_columns(nil)
+        return nil unless @dataset
+        schema_hash = {}
+        ds_opts = dataset.opts
+        single_table = ds_opts[:from] && (ds_opts[:from].length == 1) \
+          && !ds_opts.include?(:join) && !ds_opts.include?(:sql)
+        get_columns = proc{columns rescue []}
+        if single_table && (schema_array = (db.schema(table_name, :reload=>reload) rescue nil))
+          schema_array.each{|k,v| schema_hash[k] = v}
+          if ds_opts.include?(:select)
+            # Dataset only selects certain columns, delete the other
+            # columns from the schema
+            cols = get_columns.call
+            schema_hash.delete_if{|k,v| !cols.include?(k)}
+            cols.each{|c| schema_hash[c] ||= {}}
+          else
+            # Dataset is for a single table with all columns,
+            # so set the columns based on the order they were
+            # returned by the schema.
+            cols = schema_array.collect{|k,v| k}
+            set_columns(cols)
+            # Set the primary key(s) based on the schema information
+            pks = schema_array.collect{|k,v| k if v[:primary_key]}.compact
+            pks.length > 0 ? set_primary_key(*pks) : no_primary_key
+            # Also set the columns for the dataset, so the dataset
+            # doesn't have to do a query to get them.
+            dataset.instance_variable_set(:@columns, cols)
+          end
+        else
+          # If the dataset uses multiple tables or custom sql or getting
+          # the schema raised an error, just get the columns and
+          # create an empty schema hash for it.
+          get_columns.call.each{|c| schema_hash[c] = {}}
+        end
+        schema_hash
+      end
+  
+      # Module that the class includes that holds methods the class adds for column accessors and
+      # associations so that the methods can be overridden with super
+      def overridable_methods_module
+        include(@overridable_methods_module = Module.new) unless @overridable_methods_module
+        @overridable_methods_module
+      end
+  
+      # Set the columns for this model and create accessor methods for each column.
+      def set_columns(new_columns)
+        @columns = new_columns
+        def_column_accessor(*new_columns) if new_columns
+        @columns
+      end
+
+      # Add model methods that call dataset methods
+      DATASET_METHODS.each{|arg| class_eval("def #{arg}(*args, &block); dataset.#{arg}(*args, &block) end", __FILE__, __LINE__)}
+  
+      # Returns a copy of the model's dataset with custom SQL
+      alias fetch with_sql
+    end
+
+    # Sequel::Model instance methods that implement basic model functionality.
+    #
+    # * All of the methods in HOOKS create instance methods that are called
+    #   by Sequel when the appropriate action occurs.  For example, when destroying
+    #   a model object, Sequel will call before_destroy, do the destroy,
+    #   and then call after_destroy.
+    # * The following instance_methods all call the class method of the same
+    #   name: columns, dataset, db, primary_key, db_schema.
+    # * The following instance methods allow boolean flags to be set on a per-object
+    #   basis: raise_on_save_failure, raise_on_typecast_failure, strict_param_setting, 
+    #   typecast_empty_string_to_nil, typecast_on_assignment, use_transactions.
+    #   If they are not used, the object will default to whatever the model setting is.
+    module InstanceMethods
+      HOOKS.each{|h| class_eval("def #{h}; end", __FILE__, __LINE__)}
+
+      # Define instance method(s) that calls class method(s) of the
+      # same name, caching the result in an instance variable.  Define
+      # standard attr_writer method for modifying that instance variable
+      def self.class_attr_overridable(*meths) # :nodoc:
+        meths.each{|meth| class_eval("def #{meth}; !defined?(@#{meth}) ? (@#{meth} = self.class.#{meth}) : @#{meth} end")}
+        attr_writer(*meths) 
+      end 
+    
+      # Define instance method(s) that calls class method(s) of the
+      # same name. Replaces the construct:
+      #   
+      #   define_method(meth){self.class.send(meth)}
+      def self.class_attr_reader(*meths) # :nodoc:
+        meths.each{|meth| class_eval("def #{meth}; model.#{meth} end")}
+      end
+
+      private_class_method :class_attr_overridable, :class_attr_reader
+
+      class_attr_reader :columns, :db, :primary_key, :db_schema
+      class_attr_overridable :raise_on_save_failure, :raise_on_typecast_failure, :strict_param_setting, :typecast_empty_string_to_nil, :typecast_on_assignment, :use_transactions
+      
+      # The hash of attribute values.  Keys are symbols with the names of the
+      # underlying database columns.
+      attr_reader :values
+
+      # Creates new instance and passes the given values to set.
+      # If a block is given, yield the instance to the block unless
+      # from_db is true.
+      # This method runs the after_initialize hook after
+      # it has optionally yielded itself to the block.
+      #
+      # Arguments:
+      # * values - should be a hash to pass to set. 
+      # * from_db - should only be set by Model.load, forget it
+      #   exists.
+      def initialize(values = {}, from_db = false)
+        if from_db
+          @new = false
+          @values = values
+        else
+          @values = {}
+          @new = true
+          set(values)
+          changed_columns.clear 
+          yield self if block_given?
+        end
+        after_initialize
+      end
+      
+      # Returns value of the column's attribute.
+      def [](column)
+        @values[column]
+      end
+  
+      # Sets value of the column's attribute and marks the column as changed.
+      # If the column already has the same value, this is a no-op. Note that
+      # changing a columns value and then changing it back will cause the
+      # column to appear in changed_columns.  Similarly, providing a
+      # value that is different from the column's current value but is the
+      # same after typecasting will also cause changed_columns to include the
+      # column.
+      def []=(column, value)
+        # If it is new, it doesn't have a value yet, so we should
+        # definitely set the new value.
+        # If the column isn't in @values, we can't assume it is
+        # NULL in the database, so assume it has changed.
+        if new? || !@values.include?(column) || value != @values[column]
+          changed_columns << column unless changed_columns.include?(column)
+          @values[column] = typecast_value(column, value)
+        end
+      end
+  
+      # Compares model instances by values.
+      def ==(obj)
+        (obj.class == model) && (obj.values == @values)
+      end
+      alias eql? ==
+  
+      # If pk is not nil, true only if the objects have the same class and pk.
+      # If pk is nil, false.
+      def ===(obj)
+        pk.nil? ? false : (obj.class == model) && (obj.pk == pk)
+      end
+  
+      # class is defined in Object, but it is also a keyword,
+      # and since a lot of instance methods call class methods,
+      # this alias makes it so you can use model instead of
+      # self.class.
+      alias_method :model, :class
+  
+      # The current cached associations.  A hash with the keys being the
+      # association name symbols and the values being the associated object
+      # or nil (many_to_one), or the array of associated objects (*_to_many).
+      def associations
+        @associations ||= {}
+      end
+  
+      # The columns that have been updated.  This isn't completely accurate,
+      # see Model#[]=.
+      def changed_columns
+        @changed_columns ||= []
+      end
+  
+      # Deletes and returns self.  Does not run destroy hooks.
+      # Look into using destroy instead.
+      def delete
+        this.delete
+        self
+      end
+      
+      # Like delete but runs hooks before and after delete.
+      # If before_destroy returns false, returns false without
+      # deleting the object the the database. Otherwise, deletes
+      # the item from the database and returns self.  Uses a transaction
+      # if use_transactions is true.
+      def destroy
+        use_transactions ? db.transaction{_destroy} : _destroy
+      end
+
+      # Iterates through all of the current values using each.
+      #
+      # Example:
+      #   Ticket.find(7).each { |k, v| puts "#{k} => #{v}" }
+      def each(&block)
+        @values.each(&block)
+      end
+  
+      # Returns the validation errors associated with this object.
+      def errors
+        @errors ||= Errors.new
+      end 
+
+      # Returns true when current instance exists, false otherwise.
+      # Generally an object that isn't new will exist unless it has
+      # been deleted.
+      def exists?
+        this.count > 0
+      end
+      
+      # Value that should be unique for objects with the same class and pk (if pk is not nil), or
+      # the same class and values (if pk is nil).
+      def hash
+        [model, pk.nil? ? @values.sort_by{|k,v| k.to_s} : pk].hash
+      end
+  
+      # Returns value for the :id attribute, even if the primary key is
+      # not id. To get the primary key value, use #pk.
+      def id
+        @values[:id]
+      end
+  
+      # Returns a string representation of the model instance including
+      # the class name and values.
+      def inspect
+        "#<#{model.name} @values=#{inspect_values}>"
+      end
+  
+      # Returns the keys in values.  May not include all column names.
+      def keys
+        @values.keys
+      end
+  
+      # Returns true if the current instance represents a new record.
+      def new?
+        @new
+      end
+      
+      # Returns the primary key value identifying the model instance.
+      # Raises an error if this model does not have a primary key.
+      # If the model has a composite primary key, returns an array of values.
+      def pk
+        raise(Error, "No primary key is associated with this model") unless key = primary_key
+        case key
+        when Array
+          key.collect{|k| @values[k]}
+        else
+          @values[key]
+        end
+      end
+      
+      # Returns a hash identifying the model instance.  It should be true that:
+      # 
+      #  Model[model_instance.pk_hash] === model_instance
+      def pk_hash
+        model.primary_key_hash(pk)
+      end
+      
+      # Reloads attributes from database and returns self. Also clears all
+      # cached association and changed_columns information.  Raises an Error if the record no longer
+      # exists in the database.
+      def refresh
+        _refresh(this)
+      end
+
+      # Alias of refresh, but not aliased directly to make overriding in a plugin easier.
+      def reload
+        refresh
+      end
+  
+      # Creates or updates the record, after making sure the record
+      # is valid.  If the record is not valid, or before_save,
+      # before_create (if new?), or before_update (if !new?) return
+      # false, returns nil unless raise_on_save_failure is true (if it
+      # is true, it raises an error).
+      # Otherwise, returns self. You can provide an optional list of
+      # columns to update, in which case it only updates those columns.
+      #
+      # Takes the following options:
+      #
+      # * :changed - save all changed columns, instead of all columns or the columns given
+      # * :transaction - set to false not to use a transaction
+      # * :validate - set to false not to validate the model before saving
+      def save(*columns)
+        opts = columns.last.is_a?(Hash) ? columns.pop : {}
+        return save_failure(:invalid) if opts[:validate] != false and !valid?
+        use_transaction = opts.include?(:transaction) ? opts[:transaction] : use_transactions
+        use_transaction ? db.transaction(opts){_save(columns, opts)} : _save(columns, opts)
+      end
+
+      # Saves only changed columns or does nothing if no columns are marked as 
+      # chanaged.  If no columns have been changed, returns nil.  If unable to
+      # save, returns false unless raise_on_save_failure is true.
+      def save_changes
+        save(:changed=>true) || false unless changed_columns.empty?
+      end
+  
+      # Updates the instance with the supplied values with support for virtual
+      # attributes, raising an exception if a value is used that doesn't have
+      # a setter method (or ignoring it if strict_param_setting = false).
+      # Does not save the record.
+      def set(hash)
+        set_restricted(hash, nil, nil)
+      end
+  
+      # Set all values using the entries in the hash, ignoring any setting of
+      # allowed_columns or restricted columns in the model.
+      def set_all(hash)
+        set_restricted(hash, false, false)
+      end
+  
+      # Set all values using the entries in the hash, except for the keys
+      # given in except.
+      def set_except(hash, *except)
+        set_restricted(hash, false, except.flatten)
+      end
+  
+      # Set the values using the entries in the hash, only if the key
+      # is included in only.
+      def set_only(hash, *only)
+        set_restricted(hash, only.flatten, false)
+      end
+  
+      # Returns (naked) dataset that should return only this instance.
+      def this
+        @this ||= model.dataset.filter(pk_hash).limit(1).naked
+      end
+      
+      # Runs set with the passed hash and runs save_changes (which runs any callback methods).
+      def update(hash)
+        update_restricted(hash, nil, nil)
+      end
+  
+      # Update all values using the entries in the hash, ignoring any setting of
+      # allowed_columns or restricted columns in the model.
+      def update_all(hash)
+        update_restricted(hash, false, false)
+      end
+  
+      # Update all values using the entries in the hash, except for the keys
+      # given in except.
+      def update_except(hash, *except)
+        update_restricted(hash, false, except.flatten)
+      end
+  
+      # Update the values using the entries in the hash, only if the key
+      # is included in only.
+      def update_only(hash, *only)
+        update_restricted(hash, only.flatten, false)
+      end
+      
+      # Validates the object.  If the object is invalid, errors should be added
+      # to the errors attribute.  By default, does nothing, as all models
+      # are valid by default.
+      def validate
+      end
+
+      # Validates the object and returns true if no errors are reported.
+      def valid?
+        errors.clear
+        if before_validation == false
+          save_failure(:validation)
+          return false
+        end
+        validate
+        after_validation
+        errors.empty?
+      end
+
+      private
+  
+      # Internal destroy method, separted from destroy to
+      # allow running inside a transaction
+      def _destroy
+        return save_failure(:destroy) if before_destroy == false
+        delete
+        after_destroy
+        self
+      end
+      
+      # Refresh using a particular dataset, used inside save to make sure the same server
+      # is used for reading newly inserted values from the database
+      def _refresh(dataset)
+        @values = dataset.first || raise(Error, "Record not found")
+        changed_columns.clear
+        associations.clear
+        self
+      end
+      
+      # Internal version of save, split from save to allow running inside
+      # it's own transaction.
+      def _save(columns, opts)
+        return save_failure(:save) if before_save == false
+        if new?
+          return save_failure(:create) if before_create == false
+          ds = model.dataset
+          if ds.respond_to?(:insert_select) and h = ds.insert_select(@values)
+            @values = h
+            @this = nil
+          else
+            iid = ds.insert(@values)
+            # if we have a regular primary key and it's not set in @values,
+            # we assume it's the last inserted id
+            if (pk = primary_key) && !(Array === pk) && !@values[pk]
+              @values[pk] = iid
+            end
+            @this = nil if pk
+          end
+          @new = false
+          @was_new = true
+          after_create
+          after_save
+          @was_new = nil
+          if pk
+            ds = this
+            ds = ds.server(:default) unless ds.opts[:server]
+            _refresh(ds)
+          end
+        else
+          return save_failure(:update) if before_update == false
+          if columns.empty?
+            @columns_updated = opts[:changed] ? @values.reject{|k,v| !changed_columns.include?(k)} : @values
+            changed_columns.clear
+          else # update only the specified columns
+            @columns_updated = @values.reject{|k, v| !columns.include?(k)}
+            changed_columns.reject!{|c| columns.include?(c)}
+          end
+          this.update(@columns_updated)
+          after_update
+          after_save
+          @columns_updated = nil
+        end
+        self
+      end
+      
+      # Default inspection output for the values hash, overwrite to change what #inspect displays.
+      def inspect_values
+        @values.inspect
+      end
+  
+      # Raise an error if raise_on_save_failure is true, return nil otherwise.
+      def save_failure(type)
+        if raise_on_save_failure
+          if type == :invalid
+            raise ValidationFailed, errors.full_messages.join(', ')
+          else
+            raise BeforeHookFailed, "one of the before_#{type} hooks returned false"
+          end
+        end
+      end
+  
+      # Set the columns, filtered by the only and except arrays.
+      def set_restricted(hash, only, except)
+        meths = setter_methods(only, except)
+        strict = strict_param_setting
+        hash.each do |k,v|
+          m = "#{k}="
+          if meths.include?(m)
+            send(m, v)
+          elsif strict
+            raise Error, "method #{m} doesn't exist or access is restricted to it"
+          end
+        end
+        self
+      end
+  
+      # Returns all methods that can be used for attribute
+      # assignment (those that end with =), modified by the only
+      # and except arguments:
+      #
+      # * only
+      #   * false - Don't modify the results
+      #   * nil - if the model has allowed_columns, use only these, otherwise, don't modify
+      #   * Array - allow only the given methods to be used
+      # * except
+      #   * false - Don't modify the results
+      #   * nil - if the model has restricted_columns, remove these, otherwise, don't modify
+      #   * Array - remove the given methods
+      #
+      # only takes precedence over except, and if only is not used, certain methods are always
+      # restricted (RESTRICTED_SETTER_METHODS).  The primary key is restricted by default as
+      # well, see Model.unrestrict_primary_key to change this.
+      def setter_methods(only, except)
+        only = only.nil? ? model.allowed_columns : only
+        except = except.nil? ? model.restricted_columns : except
+        if only
+          only.map{|x| "#{x}="}
+        else
+          meths = methods.collect{|x| x.to_s}.grep(SETTER_METHOD_REGEXP) - RESTRICTED_SETTER_METHODS
+          meths -= Array(primary_key).map{|x| "#{x}="} if primary_key && model.restrict_primary_key?
+          meths -= except.map{|x| "#{x}="} if except
+          meths
+        end
+      end
+  
+      # Typecast the value to the column's type if typecasting.  Calls the database's
+      # typecast_value method, so database adapters can override/augment the handling
+      # for database specific column types.
+      def typecast_value(column, value)
+        return value unless typecast_on_assignment && db_schema && (col_schema = db_schema[column])
+        value = nil if value == '' and typecast_empty_string_to_nil and col_schema[:type] and ![:string, :blob].include?(col_schema[:type])
+        raise(InvalidValue, "nil/NULL is not allowed for the #{column} column") if raise_on_typecast_failure && value.nil? && (col_schema[:allow_null] == false)
+        begin
+          model.db.typecast_value(col_schema[:type], value)
+        rescue InvalidValue
+          raise_on_typecast_failure ? raise : value
+        end
+      end
+  
+      # Set the columns, filtered by the only and except arrays.
+      def update_restricted(hash, only, except)
+        set_restricted(hash, only, except)
+        save_changes
+      end
+    end
+
+    # Dataset methods are methods that the model class extends its dataset with in
+    # the call to set_dataset.
+    module DatasetMethods
+      # The model class associated with this dataset
+      attr_accessor :model
+
+      # Destroy each row in the dataset by instantiating it and then calling
+      # destroy on the resulting model object.  This isn't as fast as deleting
+      # the dataset, which does a single SQL call, but this runs any destroy
+      # hooks on each object in the dataset.
+      def destroy
+        @db.transaction{all{|r| r.destroy}.length}
+      end
+
+      # This allows you to call to_hash without any arguments, which will
+      # result in a hash with the primary key value being the key and the
+      # model object being the value.
+      def to_hash(key_column=nil, value_column=nil)
+        if key_column
+          super
+        else
+          raise(Sequel::Error, "No primary key for model") unless model and pk = model.primary_key
+          super(pk, value_column) 
+        end
+      end
+    end
+
+    plugin self
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/model/errors.rb 3.0.0-1/lib/sequel/model/errors.rb
--- 1.5.1-1/lib/sequel/model/errors.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/model/errors.rb	2009-03-30 19:23:54.000000000 +0100
@@ -0,0 +1,39 @@
+module Sequel
+  class Model
+    # Errors represents validation errors, a simple hash subclass
+    # with a few convenience methods.
+    class Errors < ::Hash
+      ATTRIBUTE_JOINER = ' and '
+
+      # Assign an array of messages for each attribute on access
+      def initialize
+        super{|h,k| h[k] = []}
+      end
+
+      # Adds an error for the given attribute.
+      def add(att, msg)
+        self[att] << msg
+      end
+
+      # Return the total number of error messages.
+      def count
+        values.inject(0){|m, v| m + v.length}
+      end
+      
+      # Returns an array of fully-formatted error messages.
+      def full_messages
+        inject([]) do |m, kv| 
+          att, errors = *kv
+          errors.each {|e| m << "#{Array(att).join(ATTRIBUTE_JOINER)} #{e}"}
+          m
+        end
+      end
+      
+      # Returns the array of errors for the given attribute, or nil
+      # if there are no errors for the attribute.
+      def on(att)
+        self[att] if include?(att)
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/model/exceptions.rb 3.0.0-1/lib/sequel/model/exceptions.rb
--- 1.5.1-1/lib/sequel/model/exceptions.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/model/exceptions.rb	2009-03-30 19:24:48.000000000 +0100
@@ -0,0 +1,7 @@
+module Sequel
+  # This exception will be raised when raise_on_save_failure is set and validation fails
+  class ValidationFailed < Error; end
+
+  # This exception will be raised when raise_on_save_failure is set and a before hook returns false
+  class BeforeHookFailed < Error; end
+end
diff -pruN 1.5.1-1/lib/sequel/model/inflections.rb 3.0.0-1/lib/sequel/model/inflections.rb
--- 1.5.1-1/lib/sequel/model/inflections.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/model/inflections.rb	2009-04-09 19:33:55.000000000 +0100
@@ -0,0 +1,205 @@
+module Sequel
+  # Yield the Inflections module if a block is given, and return
+  # the Inflections module.
+  def self.inflections
+    yield Inflections if block_given?
+    Inflections
+  end
+
+  # This module acts as a singleton returned/yielded by Sequel.inflections,
+  # which is used to override or specify additional inflection rules
+  # for Sequel. Examples:
+  #
+  #   Sequel.inflections do |inflect|
+  #     inflect.plural /^(ox)$/i, '\1\2en'
+  #     inflect.singular /^(ox)en/i, '\1'
+  #
+  #     inflect.irregular 'octopus', 'octopi'
+  #
+  #     inflect.uncountable "equipment"
+  #   end
+  #
+  # New rules are added at the top. So in the example above, the irregular rule for octopus will now be the first of the
+  # pluralization and singularization rules that is runs. This guarantees that your rules run before any of the rules that may
+  # already have been loaded.
+  module Inflections
+    CAMELIZE_CONVERT_REGEXP = /(^|_)(.)/.freeze
+    CAMELIZE_MODULE_REGEXP = /\/(.?)/.freeze
+    DASH = '-'.freeze
+    DEMODULIZE_CONVERT_REGEXP = /^.*::/.freeze
+    EMPTY_STRING= ''.freeze
+    SLASH = '/'.freeze
+    VALID_CONSTANT_NAME_REGEXP = /\A(?:::)?([A-Z]\w*(?:::[A-Z]\w*)*)\z/.freeze
+    UNDERSCORE = '_'.freeze
+    UNDERSCORE_CONVERT_REGEXP1 = /([A-Z]+)([A-Z][a-z])/.freeze
+    UNDERSCORE_CONVERT_REGEXP2 = /([a-z\d])([A-Z])/.freeze
+    UNDERSCORE_CONVERT_REPLACE = '\1_\2'.freeze
+    UNDERSCORE_MODULE_REGEXP = /::/.freeze
+
+    @plurals, @singulars, @uncountables = [], [], []
+
+    class << self
+      attr_reader :plurals, :singulars, :uncountables
+    end
+
+    # Clears the loaded inflections within a given scope (default is :all). Give the scope as a symbol of the inflection type,
+    # the options are: :plurals, :singulars, :uncountables
+    #
+    # Examples:
+    #   clear :all
+    #   clear :plurals
+    def self.clear(scope = :all)
+      case scope
+      when :all
+        @plurals, @singulars, @uncountables = [], [], []
+      else
+        instance_variable_set("@#{scope}", [])
+      end
+    end
+
+    # Specifies a new irregular that applies to both pluralization and singularization at the same time. This can only be used
+    # for strings, not regular expressions. You simply pass the irregular in singular and plural form.
+    #
+    # Examples:
+    #   irregular 'octopus', 'octopi'
+    #   irregular 'person', 'people'
+    def self.irregular(singular, plural)
+      plural(Regexp.new("(#{singular[0,1]})#{singular[1..-1]}$", "i"), '\1' + plural[1..-1])
+      singular(Regexp.new("(#{plural[0,1]})#{plural[1..-1]}$", "i"), '\1' + singular[1..-1])
+    end
+
+    # Specifies a new pluralization rule and its replacement. The rule can either be a string or a regular expression.
+    # The replacement should always be a string that may include references to the matched data from the rule.
+    #
+    # Example:
+    #   plural(/(x|ch|ss|sh)$/i, '\1es')
+    def self.plural(rule, replacement)
+      @plurals.insert(0, [rule, replacement])
+    end
+
+    # Specifies a new singularization rule and its replacement. The rule can either be a string or a regular expression.
+    # The replacement should always be a string that may include references to the matched data from the rule.
+    #
+    # Example:
+    #   singular(/([^aeiouy]|qu)ies$/i, '\1y') 
+    def self.singular(rule, replacement)
+      @singulars.insert(0, [rule, replacement])
+    end
+
+    # Add uncountable words that shouldn't be attempted inflected.
+    #
+    # Examples:
+    #   uncountable "money"
+    #   uncountable "money", "information"
+    #   uncountable %w( money information rice )
+    def self.uncountable(*words)
+      (@uncountables << words).flatten!
+    end
+
+    # Setup the default inflections
+    plural(/$/, 's')
+    plural(/s$/i, 's')
+    plural(/(ax|test)is$/i, '\1es')
+    plural(/(octop|vir)us$/i, '\1i')
+    plural(/(alias|status)$/i, '\1es')
+    plural(/(bu)s$/i, '\1ses')
+    plural(/(buffal|tomat)o$/i, '\1oes')
+    plural(/([ti])um$/i, '\1a')
+    plural(/sis$/i, 'ses')
+    plural(/(?:([^f])fe|([lr])f)$/i, '\1\2ves')
+    plural(/(hive)$/i, '\1s')
+    plural(/([^aeiouy]|qu)y$/i, '\1ies')
+    plural(/(x|ch|ss|sh)$/i, '\1es')
+    plural(/(matr|vert|ind)ix|ex$/i, '\1ices')
+    plural(/([m|l])ouse$/i, '\1ice')
+    plural(/^(ox)$/i, '\1en')
+    plural(/(quiz)$/i, '\1zes')
+
+    singular(/s$/i, '')
+    singular(/(n)ews$/i, '\1ews')
+    singular(/([ti])a$/i, '\1um')
+    singular(/((a)naly|(b)a|(d)iagno|(p)arenthe|(p)rogno|(s)ynop|(t)he)ses$/i, '\1\2sis')
+    singular(/(^analy)ses$/i, '\1sis')
+    singular(/([^f])ves$/i, '\1fe')
+    singular(/(hive)s$/i, '\1')
+    singular(/(tive)s$/i, '\1')
+    singular(/([lr])ves$/i, '\1f')
+    singular(/([^aeiouy]|qu)ies$/i, '\1y')
+    singular(/(s)eries$/i, '\1eries')
+    singular(/(m)ovies$/i, '\1ovie')
+    singular(/(x|ch|ss|sh)es$/i, '\1')
+    singular(/([m|l])ice$/i, '\1ouse')
+    singular(/(bus)es$/i, '\1')
+    singular(/(o)es$/i, '\1')
+    singular(/(shoe)s$/i, '\1')
+    singular(/(cris|ax|test)es$/i, '\1is')
+    singular(/(octop|vir)i$/i, '\1us')
+    singular(/(alias|status)es$/i, '\1')
+    singular(/^(ox)en/i, '\1')
+    singular(/(vert|ind)ices$/i, '\1ex')
+    singular(/(matr)ices$/i, '\1ix')
+    singular(/(quiz)zes$/i, '\1')
+
+    irregular('person', 'people')
+    irregular('man', 'men')
+    irregular('child', 'children')
+    irregular('sex', 'sexes')
+    irregular('move', 'moves')
+
+    uncountable(%w(equipment information rice money species series fish sheep))
+
+    private
+
+    # Convert the given string to CamelCase.  Will also convert '/' to '::' which is useful for converting paths to namespaces.
+    def camelize(s)
+      s = s.to_s
+      return s.camelize if s.respond_to?(:camelize)
+      s = s.gsub(CAMELIZE_MODULE_REGEXP){|x| "::#{x[-1..-1].upcase unless x == SLASH}"}.gsub(CAMELIZE_CONVERT_REGEXP){|x| x[-1..-1].upcase}
+      s
+    end
+  
+    # Tries to find a declared constant with the name specified
+    # in the string. It raises a NameError when the name is not in CamelCase
+    # or is not initialized.
+    def constantize(s)
+      s = s.to_s
+      return s.constantize if s.respond_to?(:constantize)
+      raise(NameError, "#{inspect} is not a valid constant name!") unless m = VALID_CONSTANT_NAME_REGEXP.match(s.to_s)
+      Object.module_eval("::#{m[1]}", __FILE__, __LINE__)
+    end
+  
+    # Removes the module part from the expression in the string
+    def demodulize(s)
+      s = s.to_s
+      return s.demodulize if s.respond_to?(:demodulize)
+      s.gsub(DEMODULIZE_CONVERT_REGEXP, EMPTY_STRING)
+    end
+  
+    # Returns the plural form of the word in the string.
+    def pluralize(s)
+      s = s.to_s
+      return s.pluralize if s.respond_to?(:pluralize)
+      result = s.dup
+      Inflections.plurals.each{|(rule, replacement)| break if result.gsub!(rule, replacement)} unless Inflections.uncountables.include?(s.downcase)
+      result
+    end
+  
+    # The reverse of pluralize, returns the singular form of a word in a string.
+    def singularize(s)
+      s = s.to_s
+      return s.singularize if s.respond_to?(:singularize)
+      result = s.dup
+      Inflections.singulars.each{|(rule, replacement)| break if result.gsub!(rule, replacement)} unless Inflections.uncountables.include?(s.downcase)
+      result
+    end
+  
+    # The reverse of camelize. Makes an underscored form from the expression in the string.
+    # Also changes '::' to '/' to convert namespaces to paths.
+    def underscore(s)
+      s = s.to_s
+      return s.underscore if s.respond_to?(:underscore)
+      s.gsub(UNDERSCORE_MODULE_REGEXP, SLASH).gsub(UNDERSCORE_CONVERT_REGEXP1, UNDERSCORE_CONVERT_REPLACE).
+        gsub(UNDERSCORE_CONVERT_REGEXP2, UNDERSCORE_CONVERT_REPLACE).tr(DASH, UNDERSCORE).downcase
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/model/plugins.rb 3.0.0-1/lib/sequel/model/plugins.rb
--- 1.5.1-1/lib/sequel/model/plugins.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/model/plugins.rb	2009-03-30 01:36:57.000000000 +0100
@@ -0,0 +1,76 @@
+module Sequel
+  # Empty namespace that plugins should use to store themselves,
+  # so they can be loaded via Model.plugin.
+  #
+  # Plugins should be modules with one of the following conditions:
+  # * A singleton method named apply, which takes a model and 
+  #   additional arguments.
+  # * A module inside the plugin module named InstanceMethods,
+  #   which will be included in the model class.
+  # * A module inside the plugin module named ClassMethods,
+  #   which will extend the model class.
+  # * A module inside the plugin module named DatasetMethods,
+  #   which will extend the model's dataset.
+  module Plugins
+  end
+  
+  class Model
+    # Loads a plugin for use with the model class, passing optional arguments
+    # to the plugin.  If the plugin is a module, load it directly.  Otherwise,
+    # require the plugin from either sequel/plugins/#{plugin} or
+    # sequel_#{plugin}, and then attempt to load the module using a
+    # the camelized plugin name under Sequel::Plugins.
+    def self.plugin(plugin, *args)
+      arg = args.first
+      block = lambda{arg}
+      m = plugin.is_a?(Module) ? plugin : plugin_module(plugin)
+      if m.respond_to?(:apply)
+        m.apply(self, *args)
+      end
+      if m.const_defined?("InstanceMethods")
+        define_method(:"#{plugin}_opts", &block)
+        include(m::InstanceMethods)
+      end
+      if m.const_defined?("ClassMethods")
+        meta_def(:"#{plugin}_opts", &block)
+        extend(m::ClassMethods)
+      end
+      if m.const_defined?("DatasetMethods")
+        if @dataset
+          dataset.meta_def(:"#{plugin}_opts", &block)
+          dataset.extend(m::DatasetMethods)
+        end
+        dataset_method_modules << m::DatasetMethods
+        def_dataset_method(*m::DatasetMethods.public_instance_methods.reject{|x| NORMAL_METHOD_NAME_REGEXP !~ x.to_s})
+      end
+    end
+    
+    module ClassMethods
+      private
+  
+      # Returns the new style location for the plugin name.
+      def plugin_gem_location(plugin)
+        "sequel/plugins/#{plugin}"
+      end
+
+      # Returns the old style location for the plugin name.
+      def plugin_gem_location_old(plugin)
+        "sequel_#{plugin}"
+      end
+  
+      # Returns the module for the specified plugin. If the module is not 
+      # defined, the corresponding plugin gem is automatically loaded.
+      def plugin_module(plugin)
+        module_name = plugin.to_s.gsub(/(^|_)(.)/){|x| x[-1..-1].upcase}
+        if not Sequel::Plugins.const_defined?(module_name)
+          begin
+            require plugin_gem_location(plugin)
+          rescue LoadError
+            require plugin_gem_location_old(plugin)
+          end
+        end
+        Sequel::Plugins.const_get(module_name)
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/model.rb 3.0.0-1/lib/sequel/model.rb
--- 1.5.1-1/lib/sequel/model.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/model.rb	2009-04-09 20:57:20.000000000 +0100
@@ -0,0 +1,109 @@
+require 'sequel/core'
+
+module Sequel
+  # Lets you create a Model subclass with its dataset already set.
+  # source can be an existing dataset or a symbol (in which case
+  # it will create a dataset using the default database with 
+  # the given symbol as the table name).
+  #
+  # The purpose of this method is to set the dataset automatically
+  # for a model class, if the table name doesn't match the implicit
+  # name.  This is neater than using set_dataset inside the class,
+  # doesn't require a bogus query for the schema, and allows
+  # it to work correctly in a system that uses code reloading.
+  #
+  # Example:
+  #   class Comment < Sequel::Model(:something)
+  #     table_name # => :something
+  #   end
+  def self.Model(source)
+    Model::ANONYMOUS_MODEL_CLASSES[source] ||= Class.new(Model).set_dataset(source)
+  end
+
+  # Sequel::Model is an object relational mapper built on top of Sequel core.  Each
+  # model class is backed by a dataset instance, and many dataset methods can be
+  # called directly on the class.  Model datasets return rows as model instances,
+  # which have fairly standard ORM instance behavior.
+  #
+  # Sequel::Model is built completely out of plugins, the only method not part of a
+  # plugin is the plugin method itself.  Plugins can override any class, instance, or
+  # dataset method defined by a previous plugin and call super to get the default
+  # behavior.
+  #
+  # You can set the SEQUEL_NO_ASSOCIATIONS constant or environment variable to
+  # make Sequel not load the associations plugin by default.
+  class Model
+    # Map that stores model classes created with Sequel::Model(), to allow the reopening
+    # of classes when dealing with code reloading.
+    ANONYMOUS_MODEL_CLASSES = {}
+
+    # Class methods added to model that call the method of the same name on the dataset
+    DATASET_METHODS = %w'<< all avg count delete distinct eager eager_graph
+       each each_page empty? except exclude filter first from from_self
+       full_outer_join get graph group group_and_count group_by having import
+       inner_join insert insert_multiple intersect interval join join_table
+       last left_outer_join limit map multi_insert naked order order_by
+       order_more paginate print query range reverse_order right_outer_join
+       select select_all select_more server set set_graph_aliases 
+       single_value to_csv to_hash union unfiltered unordered 
+       update where with_sql'.map{|x| x.to_sym}
+  
+    # Class instance variables to set to nil when a subclass is created, for -w compliance
+    EMPTY_INSTANCE_VARIABLES = [:@overridable_methods_module, :@db]
+
+    # Empty instance methods to create that the user can override to get hook/callback behavior.
+    # Just like any other method defined by Sequel, if you override one of these, you should
+    # call super to get the default behavior (while empty by default, they can also be defined
+    # by plugins).  
+    HOOKS = [:after_initialize, :before_create, :after_create, :before_update,
+      :after_update, :before_save, :after_save, :before_destroy, :after_destroy,
+      :before_validation, :after_validation]
+
+    # Class instance variables that are inherited in subclasses.  If the value is :dup, dup is called
+    # on the superclass's instance variable when creating the instance variable in the subclass.
+    # If the value is nil, the superclass's instance variable is used directly in the subclass.
+    INHERITED_INSTANCE_VARIABLES = {:@allowed_columns=>:dup, :@dataset_methods=>:dup, 
+      :@dataset_method_modules=>:dup, :@primary_key=>nil, :@use_transactions=>nil,
+      :@raise_on_save_failure=>nil, :@restricted_columns=>:dup, :@restrict_primary_key=>nil,
+      :@simple_pk=>nil, :@simple_table=>nil, :@strict_param_setting=>nil,
+      :@typecast_empty_string_to_nil=>nil, :@typecast_on_assignment=>nil,
+      :@raise_on_typecast_failure=>nil}
+
+    # Regexp that determines if a method name is normal in the sense that
+    # it could be called directly in ruby code without using send.  Used to
+    # avoid problems when using eval with a string to define methods.
+    NORMAL_METHOD_NAME_REGEXP = /\A[A-Za-z_][A-Za-z0-9_]*\z/
+
+    # The setter methods (methods ending with =) that are never allowed
+    # to be called automatically via set/update/new/etc..
+    RESTRICTED_SETTER_METHODS = %w"== === []= taguri= typecast_empty_string_to_nil= typecast_on_assignment= strict_param_setting= raise_on_save_failure= raise_on_typecast_failure="
+
+    # Regular expression that determines if the method is a valid setter name
+    # (i.e. it ends with =).
+    SETTER_METHOD_REGEXP = /=\z/
+
+    @allowed_columns = nil
+    @db = nil
+    @db_schema = nil
+    @dataset_method_modules = []
+    @dataset_methods = {}
+    @overridable_methods_module = nil
+    @primary_key = :id
+    @raise_on_save_failure = true
+    @raise_on_typecast_failure = true
+    @restrict_primary_key = true
+    @restricted_columns = nil
+    @simple_pk = nil
+    @simple_table = nil
+    @strict_param_setting = true
+    @typecast_empty_string_to_nil = true
+    @typecast_on_assignment = true
+    @use_transactions = true
+  end
+
+  require %w"inflections plugins base exceptions errors", "model"
+  if !defined?(::SEQUEL_NO_ASSOCIATIONS) && !ENV.has_key?('SEQUEL_NO_ASSOCIATIONS')
+    require 'associations', 'model'
+    Model.plugin Model::Associations
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/plugins/caching.rb 3.0.0-1/lib/sequel/plugins/caching.rb
--- 1.5.1-1/lib/sequel/plugins/caching.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/plugins/caching.rb	2009-04-21 21:54:43.000000000 +0100
@@ -0,0 +1,116 @@
+module Sequel
+  module Plugins
+    # Sequel's built-in caching plugin supports caching to any object that
+    # implements the Ruby-Memcache API.  You can add caching for any model
+    # or for all models via:
+    #
+    #   Model.plugin :caching, store   # Cache all models
+    #   MyModel.plugin :caching, store # Just cache MyModel
+    #
+    # The cache store should implement the Ruby-Memcache API:
+    #
+    #    cache_store.set(key, obj, time) # Associate the obj with the given key
+    #                                    # in the cache for the time (specified
+    #                                    # in seconds)
+    #    cache_store.get(key) => obj # Returns object set with same key
+    #    cache_store.get(key2) => nil # nil returned if there isn't an object
+    #                                 # currently in the cache with that key
+    module Caching
+      # Set the cache_store and cache_ttl attributes for the given model.
+      # If the :ttl option is not given, 3600 seconds is the default.
+      def self.apply(model, store, opts={})
+        model.instance_eval do
+          @cache_store = store
+          @cache_ttl = opts[:ttl] || 3600
+        end
+      end
+
+      module ClassMethods
+        # The cache store object for the model, which should implement the
+        # Ruby-Memcache API
+        attr_reader :cache_store
+        
+        # The time to live for the cache store, in seconds.
+        attr_reader :cache_ttl
+    
+        # Check the cache before a database lookup unless a hash is supplied.
+        def [](*args)
+          args = args.first if (args.size == 1)
+          return super(args) if args.is_a?(Hash)
+          ck = cache_key(args)
+          if obj = @cache_store.get(ck)
+            return obj
+          end
+          if obj = super(args)
+            @cache_store.set(ck, obj, @cache_ttl)
+          end 
+          obj
+        end
+
+        # Set the time to live for the cache store, in seconds (default is 3600, # so 1 hour).
+        def set_cache_ttl(ttl)
+          @cache_ttl = ttl
+        end
+        
+        # Copy the cache_store and cache_ttl to the subclass.
+        def inherited(subclass)
+          super
+          store = @cache_store
+          ttl = @cache_ttl
+          subclass.instance_eval do
+            @cache_store = store
+            @cache_ttl = ttl
+          end
+        end
+
+        private
+    
+        # Delete the entry with the matching key from the cache
+        def cache_delete(key)
+          @cache_store.delete(key)
+          nil
+        end
+    
+        # Return a key string for the pk
+        def cache_key(pk)
+          "#{self}:#{Array(pk).join(',')}"
+        end
+      end
+
+      module InstanceMethods
+        # Remove the object from the cache when updating
+        def before_update
+          return false if super == false
+          cache_delete
+        end
+
+        # Return a key unique to the underlying record for caching, based on the
+        # primary key value(s) for the object.  If the model does not have a primary
+        # key, raise an Error.
+        def cache_key
+          raise(Error, "No primary key is associated with this model") unless key = primary_key
+          pk = case key
+          when Array
+            key.collect{|k| @values[k]}
+          else
+            @values[key] || (raise Error, 'no primary key for this record')
+          end
+          model.send(:cache_key, pk)
+        end
+    
+        # Remove the object from the cache when deleting
+        def delete
+          cache_delete
+          super
+        end
+
+        private
+    
+        # Delete this object from the cache
+        def cache_delete
+          model.send(:cache_delete, cache_key)
+        end
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/plugins/hook_class_methods.rb 3.0.0-1/lib/sequel/plugins/hook_class_methods.rb
--- 1.5.1-1/lib/sequel/plugins/hook_class_methods.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/plugins/hook_class_methods.rb	2009-04-09 21:34:30.000000000 +0100
@@ -0,0 +1,122 @@
+module Sequel
+  module Plugins
+    # Sequel's built-in hook class methods plugin is designed for backwards
+    # compatibility.  Its use is not encouraged, it is recommended to use
+    # instance methods and super instead of this plugin.  What this plugin
+    # allows you to do is, for example:
+    #
+    #   # Block only, can cause duplicate hooks if code is reloaded
+    #   before_save{self.created_at = Time.now}
+    #   # Block with tag, safe for reloading
+    #   before_save(:set_created_at){self.created_at = Time.now}
+    #   # Tag only, safe for reloading, calls instance method
+    #   before_save(:set_created_at)
+    #
+    # Pretty much anything you can do with a hook class method, you can also
+    # do with an instance method instead:
+    #
+    #    def before_save
+    #      return false if super == false
+    #      self.created_at = Time.now
+    #    end
+    module HookClassMethods
+      # Set up the hooks instance variable in the model.
+      def self.apply(model)
+        hooks = model.instance_variable_set(:@hooks, {})
+        Model::HOOKS.each{|h| hooks[h] = []}
+      end
+
+      module ClassMethods
+        Model::HOOKS.each{|h| class_eval("def #{h}(method = nil, &block); add_hook(:#{h}, method, &block) end", __FILE__, __LINE__)}
+
+        # This adds a new hook type. It will define both a class
+        # method that you can use to add hooks, as well as an instance method
+        # that you can use to call all hooks of that type.  The class method
+        # can be called with a symbol or a block or both.  If a block is given and
+        # and symbol is not, it adds the hook block to the hook type.  If a block
+        # and symbol are both given, it replaces the hook block associated with
+        # that symbol for a given hook type, or adds it if there is no hook block
+        # with that symbol for that hook type.  If no block is given, it assumes
+        # the symbol specifies an instance method to call and adds it to the hook
+        # type.
+        #
+        # If any hook block returns false, the instance method will return false
+        # immediately without running the rest of the hooks of that type.
+        #
+        # It is recommended that you always provide a symbol to this method,
+        # for descriptive purposes.  It's only necessary to do so when you 
+        # are using a system that reloads code.
+        # 
+        # Example of usage:
+        #
+        #  class MyModel
+        #   define_hook :before_move_to
+        #   before_move_to(:check_move_allowed){|o| o.allow_move?}
+        #   def move_to(there)
+        #     return if before_move_to == false
+        #     # move MyModel object to there
+        #   end
+        #  end
+        def add_hook_type(*hooks)
+          Model::HOOKS.concat(hooks)
+          hooks.each do |hook|
+            @hooks[hook] = []
+            instance_eval("def #{hook}(method = nil, &block); add_hook(:#{hook}, method, &block) end", __FILE__, __LINE__)
+            class_eval("def #{hook}; run_hooks(:#{hook}); end", __FILE__, __LINE__)
+          end
+        end
+    
+        # Returns true if there are any hook blocks for the given hook.
+        def has_hooks?(hook)
+          !@hooks[hook].empty?
+        end
+    
+        # Yield every block related to the given hook.
+        def hook_blocks(hook)
+          @hooks[hook].each{|k,v| yield v}
+        end
+
+        # Make a copy of the current class's hooks for the subclass.
+        def inherited(subclass)
+          super
+          hooks = subclass.instance_variable_set(:@hooks, {}) 
+          instance_variable_get(:@hooks).each{|k,v| hooks[k] = v.dup}
+        end
+    
+        private
+    
+        # Add a hook block to the list of hook methods.
+        # If a non-nil tag is given and it already is in the list of hooks,
+        # replace it with the new block.
+        def add_hook(hook, tag, &block)
+          unless block
+            (raise Error, 'No hook method specified') unless tag
+            block = proc {send tag}
+          end
+          h = @hooks[hook]
+          if tag && (old = h.find{|x| x[0] == tag})
+            old[1] = block
+          else
+            if hook.to_s =~ /^before/
+              h.unshift([tag,block])
+            else
+              h << [tag, block]
+            end
+          end
+        end
+      end
+
+      module InstanceMethods
+        Model::HOOKS.each{|h| class_eval("def #{h}; return false if super == false; run_hooks(:#{h}); end", __FILE__, __LINE__)}
+
+        private
+
+        # Runs all hook blocks of given hook type on this object.
+        # Stops running hook blocks and returns false if any hook block returns false.
+        def run_hooks(hook)
+          model.hook_blocks(hook){|block| return false if instance_eval(&block) == false}
+        end
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/plugins/schema.rb 3.0.0-1/lib/sequel/plugins/schema.rb
--- 1.5.1-1/lib/sequel/plugins/schema.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/plugins/schema.rb	2009-03-08 05:34:59.000000000 +0000
@@ -0,0 +1,53 @@
+module Sequel
+  module Plugins
+    module Schema
+      module ClassMethods
+        # Creates table, using the column information from set_schema.
+        def create_table
+          db.create_table(table_name, :generator=>@schema)
+          @db_schema = get_db_schema(true)
+          columns
+        end
+        
+        # Drops the table if it exists and then runs create_table.  Should probably
+        # not be used except in testing.
+        def create_table!
+          drop_table rescue nil
+          create_table
+        end
+        
+        # Drops table.
+        def drop_table
+          db.drop_table(table_name)
+        end
+    
+        # Returns table schema created with set_schema for direct descendant of Model.
+        # Does not retreive schema information from the database, see db_schema if you
+        # want that.
+        def schema
+          @schema || (superclass.schema unless superclass == Model)
+        end
+    
+        # Defines a table schema (see Schema::Generator for more information).
+        #
+        # This is only needed if you want to use the create_table/create_table! methods.
+        # Will also set the dataset if you provide a name, as well as setting
+        # the primary key if you defined one in the passed block.
+        #
+        # In general, it is a better idea to use migrations for production code, as
+        # migrations allow changes to existing schema.  set_schema is mostly useful for
+        # test code or simple examples.
+        def set_schema(name = nil, &block)
+          set_dataset(db[name]) if name
+          @schema = Sequel::Schema::Generator.new(db, &block)
+          set_primary_key(@schema.primary_key_name) if @schema.primary_key_name
+        end
+        
+        # Returns true if table exists, false otherwise.
+        def table_exists?
+          db.table_exists?(table_name)
+        end
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/plugins/serialization.rb 3.0.0-1/lib/sequel/plugins/serialization.rb
--- 1.5.1-1/lib/sequel/plugins/serialization.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/plugins/serialization.rb	2009-03-26 18:43:51.000000000 +0000
@@ -0,0 +1,117 @@
+module Sequel
+  module Plugins
+    # Sequel's built in Serialization plugin allows you to keep serialized
+    # ruby objects in the database, while giving you deserialized objects
+    # when you call an accessor.
+    # 
+    # This plugin works by keeping the serialized value in the values, and
+    # adding a @deserialized_values hash.  The reader method for serialized columns
+    # will check the @deserialized_values for the value, return it if present,
+    # or deserialized the entry in @values and return it.  The writer method will
+    # set the @deserialized_values entry.  This plugin adds a before_save hook
+    # that serializes all @deserialized_values to @values.
+    #
+    # You can use either marshal or yaml as the serialization format.
+    # If you use yaml, you should require yaml yourself.
+    #
+    # Because of how this plugin works, it must be used inside each model class
+    # that needs serialization, after any set_dataset method calls in that class.
+    # Otherwise, it is possible that the default column accessors will take
+    # precedence.
+    module Serialization
+      # Set up the column readers to do deserialization and the column writers
+      # to save the value in deserialized_values.
+      def self.apply(model, format, *columns)
+        raise(Error, "Unsupported serialization format (#{format}), should be :marshal or :yaml") unless [:marshal, :yaml].include?(format)
+        raise(Error, "No columns given.  The serialization plugin requires you specify which columns to serialize") if columns.empty?
+        model.instance_eval do
+          @serialization_format = format
+          @serialized_columns = columns
+          InstanceMethods.module_eval do
+            columns.each do |column|
+              define_method(column) do 
+                if deserialized_values.has_key?(column)
+                  deserialized_values[column]
+                else
+                  deserialized_values[column] = deserialize_value(@values[column])
+                end
+              end
+              define_method("#{column}=") do |v| 
+                changed_columns << column unless changed_columns.include?(column)
+                deserialized_values[column] = v
+              end
+            end
+          end
+        end
+      end
+
+      module ClassMethods
+        # The serialization format to use, should be :marshal or :yaml
+        attr_reader :serialization_format
+
+        # The columns to serialize
+        attr_reader :serialized_columns
+
+        # Copy the serialization format and columns to serialize into the subclass.
+        def inherited(subclass)
+          super
+          sf = serialization_format
+          sc = serialized_columns
+          subclass.instance_eval do
+            @serialization_format = sf
+            @serialized_columns = sc
+          end
+        end
+      end
+
+      module InstanceMethods
+        # Hash of deserialized values, used as a cache.
+        attr_reader :deserialized_values
+
+        # Set @deserialized_values to the empty hash
+        def initialize(*args, &block)
+          @deserialized_values = {}
+          super
+        end
+
+        # Serialize all deserialized values
+        def before_save
+          super
+          deserialized_values.each do |k,v|
+            @values[k] = serialize_value(v)
+          end
+        end
+        
+        # Empty the deserialized values when refreshing.
+        def refresh
+          @deserialized_values = {}
+          super
+        end
+
+        private
+
+        # Deserialize the column from either marshal or yaml format
+        def deserialize_value(v)
+          return v if v.nil?
+          case model.serialization_format 
+          when :marshal
+            Marshal.load(v.unpack('m')[0]) rescue Marshal.load(v)
+          when :yaml
+            YAML.load v if v
+          end
+        end
+
+        # Serialize the column to either marshal or yaml format
+        def serialize_value(v)
+          return v if v.nil?
+          case model.serialization_format 
+          when :marshal
+            [Marshal.dump(v)].pack('m')
+          when :yaml
+            v.to_yaml
+          end
+        end
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/plugins/single_table_inheritance.rb 3.0.0-1/lib/sequel/plugins/single_table_inheritance.rb
--- 1.5.1-1/lib/sequel/plugins/single_table_inheritance.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/plugins/single_table_inheritance.rb	2009-03-09 03:19:48.000000000 +0000
@@ -0,0 +1,63 @@
+module Sequel
+  module Plugins
+    # Sequel's built in Single Table Inheritance plugin makes subclasses
+    # of this model only load rows where the given key field matches the
+    # subclass's name.  If the key given has a NULL value or there are
+    # any problems looking up the class, uses the current class.
+    #   
+    # You should only use this in the parent class, not in the subclasses.
+    #   
+    # You shouldn't call set_dataset in the model after applying this
+    # plugin, otherwise subclasses might use the wrong dataset.
+    #   
+    # The filters and row_proc that sti_key sets up in subclasses may not work correctly if
+    # those subclasses have further subclasses.  For those middle subclasses,
+    # you may need to call set_dataset manually with the correct filter and
+    # row_proc.
+    module SingleTableInheritance
+      # Set the sti_key and sti_dataset for the model, and change the
+      # dataset's row_proc so that the dataset yields objects of varying classes,
+      # where the class used has the same name as the key field.
+      def self.apply(model, key)
+        m = model.method(:constantize)
+        model.instance_eval do
+          @sti_key = key 
+          @sti_dataset = dataset
+          dataset.row_proc = lambda{|r| (m.call(r[key]) rescue model).load(r)}
+        end
+      end
+
+      module ClassMethods
+        # The base dataset for STI, to which filters are added to get
+        # only the models for the specific STI subclass.
+        attr_reader :sti_dataset
+
+        # The column name holding the STI key for this model
+        attr_reader :sti_key
+
+        # Copy the sti_key and sti_dataset to the subclasses, and filter the
+        # subclass's dataset so it is restricted to rows where the key column
+        # matches the subclass's name.
+        def inherited(subclass)
+          super
+          sk = sti_key
+          sd = sti_dataset
+          subclass.set_dataset(sd.filter(sk=>subclass.name.to_s), :inherited=>true)
+          subclass.instance_eval do
+            @sti_key = sk
+            @sti_dataset = sd
+            @simple_table = nil
+          end
+        end
+      end
+
+      module InstanceMethods
+        # Set the sti_key column to the name of the model.
+        def before_create
+          return false if super == false
+          send("#{model.sti_key}=", model.name.to_s)
+        end
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/plugins/validation_class_methods.rb 3.0.0-1/lib/sequel/plugins/validation_class_methods.rb
--- 1.5.1-1/lib/sequel/plugins/validation_class_methods.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/plugins/validation_class_methods.rb	2009-03-30 20:08:40.000000000 +0100
@@ -0,0 +1,384 @@
+module Sequel
+  require 'extensions/blank'
+
+  module Plugins
+    module ValidationClassMethods
+      module ClassMethods
+        # The Generator class is used to generate validation definitions using 
+        # the validates {} idiom.
+        class Generator
+          # Initializes a new generator.
+          def initialize(receiver ,&block)
+            @receiver = receiver
+            instance_eval(&block)
+          end
+      
+          # Delegates method calls to the receiver by calling receiver.validates_xxx.
+          def method_missing(m, *args, &block)
+            @receiver.send(:"validates_#{m}", *args, &block)
+          end
+        end
+    
+        # Returns true if validations are defined.
+        def has_validations?
+          !validations.empty?
+        end
+    
+        # Instructs the model to skip validations defined in superclasses
+        def skip_superclass_validations
+          @skip_superclass_validations = true
+        end
+        
+        # Instructs the model to skip validations defined in superclasses
+        def skip_superclass_validations?
+          defined?(@skip_superclass_validations) && @skip_superclass_validations
+        end
+
+        # Defines validations by converting a longhand block into a series of 
+        # shorthand definitions. For example:
+        #
+        #   class MyClass < Sequel::Model
+        #     validates do
+        #       length_of :name, :minimum => 6
+        #       length_of :password, :minimum => 8
+        #     end
+        #   end
+        #
+        # is equivalent to:
+        #   class MyClass < Sequel::Model
+        #     validates_length_of :name, :minimum => 6
+        #     validates_length_of :password, :minimum => 8
+        #   end
+        def validates(&block)
+          Generator.new(self, &block)
+        end
+    
+        # Validates the given instance.
+        def validate(o)
+          superclass.validate(o) if superclass.respond_to?(:validate) && !skip_superclass_validations?
+          validations.each do |att, procs|
+            v = case att
+            when Array
+              att.collect{|a| o.send(a)}
+            else
+              o.send(att)
+            end
+            procs.each {|tag, p| p.call(o, att, v)}
+          end
+        end
+        
+        # Validates acceptance of an attribute.  Just checks that the value
+        # is equal to the :accept option. This method is unique in that
+        # :allow_nil is assumed to be true instead of false.
+        #
+        # Possible Options:
+        # * :accept - The value required for the object to be valid (default: '1')
+        # * :message - The message to use (default: 'is not accepted')
+        def validates_acceptance_of(*atts)
+          opts = {
+            :message => 'is not accepted',
+            :allow_nil => true,
+            :accept => '1',
+            :tag => :acceptance,
+          }.merge!(extract_options!(atts))
+          atts << opts
+          validates_each(*atts) do |o, a, v|
+            o.errors.add(a, opts[:message]) unless v == opts[:accept]
+          end
+        end
+    
+        # Validates confirmation of an attribute. Checks that the object has
+        # a _confirmation value matching the current value.  For example:
+        #
+        #   validates_confirmation_of :blah
+        #
+        # Just makes sure that object.blah = object.blah_confirmation.  Often used for passwords
+        # or email addresses on web forms.
+        #
+        # Possible Options:
+        # * :message - The message to use (default: 'is not confirmed')
+        def validates_confirmation_of(*atts)
+          opts = {
+            :message => 'is not confirmed',
+            :tag => :confirmation,
+          }.merge!(extract_options!(atts))
+          atts << opts
+          validates_each(*atts) do |o, a, v|
+            o.errors.add(a, opts[:message]) unless v == o.send(:"#{a}_confirmation")
+          end
+        end
+    
+        # Adds a validation for each of the given attributes using the supplied
+        # block. The block must accept three arguments: instance, attribute and 
+        # value, e.g.:
+        #
+        #   validates_each :name, :password do |object, attribute, value|
+        #     object.errors.add(attribute, 'is not nice') unless value.nice?
+        #   end
+        #
+        # Possible Options:
+        # * :allow_blank - Whether to skip the validation if the value is blank. 
+        # * :allow_missing - Whether to skip the validation if the attribute isn't a key in the
+        #   values hash.  This is different from allow_nil, because Sequel only sends the attributes
+        #   in the values when doing an insert or update.  If the attribute is not present, Sequel
+        #   doesn't specify it, so the database will use the table's default value.  This is different
+        #   from having an attribute in values with a value of nil, which Sequel will send as NULL.
+        #   If your database table has a non NULL default, this may be a good option to use.  You
+        #   don't want to use allow_nil, because if the attribute is in values but has a value nil,
+        #   Sequel will attempt to insert a NULL value into the database, instead of using the
+        #   database's default.
+        # * :allow_nil - Whether to skip the validation if the value is nil.
+        # * :if - A symbol (indicating an instance_method) or proc (which is instance_evaled)
+        #   skipping this validation if it returns nil or false.
+        # * :tag - The tag to use for this validation.
+        def validates_each(*atts, &block)
+          opts = extract_options!(atts)
+          blk = if (i = opts[:if]) || (am = opts[:allow_missing]) || (an = opts[:allow_nil]) || (ab = opts[:allow_blank])
+            proc do |o,a,v|
+              next if i && !validation_if_proc(o, i)
+              next if an && Array(v).all?{|x| x.nil?}
+              next if ab && Array(v).all?{|x| x.blank?}
+              next if am && Array(a).all?{|x| !o.values.has_key?(x)}
+              block.call(o,a,v)
+            end
+          else
+            block
+          end
+          tag = opts[:tag]
+          atts.each do |a| 
+            a_vals = validations[a]
+            if tag && (old = a_vals.find{|x| x[0] == tag})
+              old[1] = blk
+            else
+              a_vals << [tag, blk]
+            end
+          end
+        end
+    
+        # Validates the format of an attribute, checking the string representation of the
+        # value against the regular expression provided by the :with option.
+        #
+        # Possible Options:
+        # * :message - The message to use (default: 'is invalid')
+        # * :with - The regular expression to validate the value with (required).
+        def validates_format_of(*atts)
+          opts = {
+            :message => 'is invalid',
+            :tag => :format,
+          }.merge!(extract_options!(atts))
+          
+          unless opts[:with].is_a?(Regexp)
+            raise ArgumentError, "A regular expression must be supplied as the :with option of the options hash"
+          end
+          
+          atts << opts
+          validates_each(*atts) do |o, a, v|
+            o.errors.add(a, opts[:message]) unless v.to_s =~ opts[:with]
+          end
+        end
+    
+        # Validates the length of an attribute.
+        #
+        # Possible Options:
+        # * :is - The exact size required for the value to be valid (no default)
+        # * :maximum - The maximum size allowed for the value (no default)
+        # * :message - The message to use (no default, overrides :too_long, :too_short, and :wrong_length
+        #   options if present)
+        # * :minimum - The minimum size allowed for the value (no default)
+        # * :too_long - The message to use use if it the value is too long (default: 'is too long')
+        # * :too_short - The message to use use if it the value is too short (default: 'is too short')
+        # * :within - The array/range that must include the size of the value for it to be valid (no default)
+        # * :wrong_length - The message to use use if it the value is not valid (default: 'is the wrong length')
+        def validates_length_of(*atts)
+          opts = {
+            :too_long     => 'is too long',
+            :too_short    => 'is too short',
+            :wrong_length => 'is the wrong length'
+          }.merge!(extract_options!(atts))
+          
+          opts[:tag] ||= ([:length] + [:maximum, :minimum, :is, :within].reject{|x| !opts.include?(x)}).join('-').to_sym
+          atts << opts
+          validates_each(*atts) do |o, a, v|
+            if m = opts[:maximum]
+              o.errors.add(a, opts[:message] || opts[:too_long]) unless v && v.size <= m
+            end
+            if m = opts[:minimum]
+              o.errors.add(a, opts[:message] || opts[:too_short]) unless v && v.size >= m
+            end
+            if i = opts[:is]
+              o.errors.add(a, opts[:message] || opts[:wrong_length]) unless v && v.size == i
+            end
+            if w = opts[:within]
+              o.errors.add(a, opts[:message] || opts[:wrong_length]) unless v && w.include?(v.size)
+            end
+          end
+        end
+    
+        # Validates whether an attribute is not a string.  This is generally useful
+        # in conjunction with raise_on_typecast_failure = false, where you are
+        # passing in string values for non-string attributes (such as numbers and dates).
+        # If typecasting fails (invalid number or date), the value of the attribute will
+        # be a string in an invalid format, and if typecasting succeeds, the value will
+        # not be a string.
+        #
+        # Possible Options:
+        # * :message - The message to use (default: 'is a string' or 'is not a valid (integer|datetime|etc.)' if the type is known)
+        def validates_not_string(*atts)
+          opts = {
+            :tag => :not_string,
+          }.merge!(extract_options!(atts))
+          atts << opts
+          validates_each(*atts) do |o, a, v|
+            if v.is_a?(String)
+              unless message = opts[:message]
+                message = if sch = o.db_schema[a] and typ = sch[:type]
+                  "is not a valid #{typ}"
+                else
+                  "is a string"
+                end
+              end
+              o.errors.add(a, message)
+            end
+          end
+        end
+    
+        # Validates whether an attribute is a number.
+        #
+        # Possible Options:
+        # * :message - The message to use (default: 'is not a number')
+        # * :only_integer - Whether only integers are valid values (default: false)
+        def validates_numericality_of(*atts)
+          opts = {
+            :message => 'is not a number',
+            :tag => :numericality,
+          }.merge!(extract_options!(atts))
+          atts << opts
+          validates_each(*atts) do |o, a, v|
+            begin
+              if opts[:only_integer]
+                Kernel.Integer(v.to_s)
+              else
+                Kernel.Float(v.to_s)
+              end
+            rescue
+              o.errors.add(a, opts[:message])
+            end
+          end
+        end
+    
+        # Validates the presence of an attribute.  Requires the value not be blank,
+        # with false considered present instead of absent.
+        #
+        # Possible Options:
+        # * :message - The message to use (default: 'is not present')
+        def validates_presence_of(*atts)
+          opts = {
+            :message => 'is not present',
+            :tag => :presence,
+          }.merge!(extract_options!(atts))
+          atts << opts
+          validates_each(*atts) do |o, a, v|
+            o.errors.add(a, opts[:message]) if v.blank? && v != false
+          end
+        end
+        
+        # Validates that an attribute is within a specified range or set of values.
+        #
+        # Possible Options:
+        # * :in - An array or range of values to check for validity (required)
+        # * :message - The message to use (default: 'is not in range or set: <specified range>')
+        def validates_inclusion_of(*atts)
+          opts = extract_options!(atts)
+          unless opts[:in] && opts[:in].respond_to?(:include?) 
+            raise ArgumentError, "The :in parameter is required, and respond to include?"
+          end
+          opts[:message] ||= "is not in range or set: #{opts[:in].inspect}"
+          atts << opts
+          validates_each(*atts) do |o, a, v|
+            o.errors.add(a, opts[:message]) unless opts[:in].include?(v)
+          end
+        end
+    
+        # Validates only if the fields in the model (specified by atts) are
+        # unique in the database.  Pass an array of fields instead of multiple
+        # fields to specify that the combination of fields must be unique,
+        # instead of that each field should have a unique value.
+        #
+        # This means that the code:
+        #   validates_uniqueness_of([:column1, :column2])
+        # validates the grouping of column1 and column2 while
+        #   validates_uniqueness_of(:column1, :column2)
+        # validates them separately.
+        #
+        # You should also add a unique index in the
+        # database, as this suffers from a fairly obvious race condition.
+        #
+        # Possible Options:
+        # * :message - The message to use (default: 'is already taken')
+        def validates_uniqueness_of(*atts)
+          opts = {
+            :message => 'is already taken',
+            :tag => :uniqueness,
+          }.merge!(extract_options!(atts))
+    
+          atts << opts
+          validates_each(*atts) do |o, a, v|
+            error_field = a
+            a = Array(a)
+            v = Array(v)
+            ds = o.class.filter(a.zip(v))
+            num_dups = ds.count
+            allow = if num_dups == 0
+              # No unique value in the database
+              true
+            elsif num_dups > 1
+              # Multiple "unique" values in the database!!
+              # Someone didn't add a unique index
+              false
+            elsif o.new?
+              # New record, but unique value already exists in the database
+              false
+            elsif ds.first === o
+              # Unique value exists in database, but for the same record, so the update won't cause a duplicate record
+              true
+            else
+              false
+            end
+            o.errors.add(error_field, opts[:message]) unless allow
+          end
+        end
+    
+        # Returns the validations hash for the class.
+        def validations
+          @validations ||= Hash.new {|h, k| h[k] = []}
+        end
+        
+        private
+    
+        # Removes and returns the last member of the array if it is a hash. Otherwise,
+        # an empty hash is returned This method is useful when writing methods that
+        # take an options hash as the last parameter.
+        def extract_options!(array)
+          array.last.is_a?(Hash) ? array.pop : {}
+        end
+
+        # Handle the :if option for validations
+        def validation_if_proc(o, i)
+          case i
+          when Symbol then o.send(i)
+          when Proc then o.instance_eval(&i)
+          when nil then true
+          else raise(::Sequel::Error, "invalid value for :if validation option")
+          end
+        end
+      end
+    
+      module InstanceMethods
+        # Validates the object.
+        def validate
+          model.validate(self)
+        end
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/plugins/validation_helpers.rb 3.0.0-1/lib/sequel/plugins/validation_helpers.rb
--- 1.5.1-1/lib/sequel/plugins/validation_helpers.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/plugins/validation_helpers.rb	2009-04-02 18:58:54.000000000 +0100
@@ -0,0 +1,150 @@
+module Sequel
+  module Plugins
+    module ValidationHelpers
+      # ValidationHelpers contains instance method equivalents for most of the previous
+      # default validations.  The names and APIs have changed, though.  
+      #
+      # The validates_unique validation has a unique API, but the other validations have
+      # the API explained here:
+      #
+      # Arguments:
+      # * atts - Single attribute symbol or an array of attribute symbols specifying the
+      #   attribute(s) to validate.
+      # Options:
+      # * :allow_blank - Whether to skip the validation if the value is blank.  You should
+      #   make sure all objects respond to blank if you use this option, which you can do by
+      #   requiring 'sequel/extensions/blank'
+      # * :allow_missing - Whether to skip the validation if the attribute isn't a key in the
+      #   values hash.  This is different from allow_nil, because Sequel only sends the attributes
+      #   in the values when doing an insert or update.  If the attribute is not present, Sequel
+      #   doesn't specify it, so the database will use the table's default value.  This is different
+      #   from having an attribute in values with a value of nil, which Sequel will send as NULL.
+      #   If your database table has a non NULL default, this may be a good option to use.  You
+      #   don't want to use allow_nil, because if the attribute is in values but has a value nil,
+      #   Sequel will attempt to insert a NULL value into the database, instead of using the
+      #   database's default.
+      # * :allow_nil - Whether to skip the validation if the value is nil.
+      # * :message - The message to use
+      module InstanceMethods 
+        # Check that the attribute values are the given exact length.
+        def validates_exact_length(exact, atts, opts={})
+          validatable_attributes(atts, opts){|a,v,m| (m || "is not #{exact} characters") unless v && v.length == exact}
+        end
+
+        # Check the string representation of the attribute value(s) against the regular expression with.
+        def validates_format(with, atts, opts={})
+          validatable_attributes(atts, opts){|a,v,m| (m || 'is invalid') unless v.to_s =~ with}
+        end
+    
+        # Check attribute value(s) is included in the given set.
+        def validates_includes(set, atts, opts={})
+          validatable_attributes(atts, opts){|a,v,m| (m || "is not in range or set: #{set.inspect}") unless set.include?(v)}
+        end
+    
+        # Check attribute value(s) string representation is a valid integer.
+        def validates_integer(atts, opts={})
+          validatable_attributes(atts, opts) do |a,v,m|
+            begin
+              Kernel.Integer(v.to_s)
+              nil
+            rescue
+              m || 'is not a number'
+            end
+          end
+        end
+
+        # Check that the attribute values length is in the specified range.
+        def validates_length_range(range, atts, opts={})
+          validatable_attributes(atts, opts){|a,v,m| (m || "is outside the allowed range") unless v && range.include?(v.length)}
+        end
+    
+        # Check that the attribute values are not longer than the given max length.
+        def validates_max_length(max, atts, opts={})
+          validatable_attributes(atts, opts){|a,v,m| (m || "is longer than #{max} characters") unless v && v.length <= max}
+        end
+
+        # Check that the attribute values are not shorter than the given min length.
+        def validates_min_length(min, atts, opts={})
+          validatable_attributes(atts, opts){|a,v,m| (m || "is shorter than #{min} characters") unless v && v.length >= min}
+        end
+
+        # Check that the attribute value(s) is not a string.  This is generally useful
+        # in conjunction with raise_on_typecast_failure = false, where you are
+        # passing in string values for non-string attributes (such as numbers and dates).
+        # If typecasting fails (invalid number or date), the value of the attribute will
+        # be a string in an invalid format, and if typecasting succeeds, the value will
+        # not be a string.
+        def validates_not_string(atts, opts={})
+          validatable_attributes(atts, opts) do |a,v,m|
+            next unless v.is_a?(String)
+            next m if m
+            (sch = db_schema[a] and typ = sch[:type]) ?  "is not a valid #{typ}" : "is a string"
+          end
+        end
+    
+        # Check attribute value(s) string representation is a valid float.
+        def validates_numeric(atts, opts={})
+          validatable_attributes(atts, opts) do |a,v,m|
+            begin
+              Kernel.Float(v.to_s)
+              nil
+            rescue
+              m || 'is not a number'
+            end
+          end
+        end
+    
+        # Check attribute value(s) is not considered blank by the database, but allow false values.
+        def validates_presence(atts, opts={})
+          validatable_attributes(atts, opts){|a,v,m| (m || "is not present") if model.db.send(:blank_object?, v) && v != false}
+        end
+        
+        # Checks that there are no duplicate values in the database for the given
+        # attributes.  Pass an array of fields instead of multiple
+        # fields to specify that the combination of fields must be unique,
+        # instead of that each field should have a unique value.
+        #
+        # This means that the code:
+        #   validates_unique([:column1, :column2])
+        # validates the grouping of column1 and column2 while
+        #   validates_unique(:column1, :column2)
+        # validates them separately.
+        #
+        # You should also add a unique index in the
+        # database, as this suffers from a fairly obvious race condition.
+        #
+        # This validation does not respect the :allow_* options that the other validations accept,
+        # since it can deals with multiple attributes at once.
+        #
+        # Possible Options:
+        # * :message - The message to use (default: 'is already taken')
+        def validates_unique(*atts)
+          message = (atts.pop[:message] if atts.last.is_a?(Hash)) || 'is already taken'
+          atts.each do |a|
+            ds = model.filter(Array(a).map{|x| [x, send(x)]})
+            errors.add(a, message) unless (new? ? ds : ds.exclude(pk_hash)).count == 0
+          end
+        end
+        
+        private
+
+        # Skip validating any attribute that matches one of the allow_* options.
+        # Otherwise, yield the attribute, value, and passed option :message to
+        # the block.  If the block returns anything except nil or false, add it as
+        # an error message for that attributes.
+        def validatable_attributes(atts, opts)
+          am, an, ab, m = opts.values_at(:allow_missing, :allow_nil, :allow_blank, :message)
+          Array(atts).each do |a|
+            next if am && !values.has_key?(a)
+            v = send(a)
+            next if an && v.nil?
+            next if ab && v.respond_to?(:blank?) && v.blank?
+            if message = yield(a, v, m)
+              errors.add(a, message)
+            end
+          end
+        end
+      end
+    end
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/sql.rb 3.0.0-1/lib/sequel/sql.rb
--- 1.5.1-1/lib/sequel/sql.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/sql.rb	2009-04-22 05:08:37.000000000 +0100
@@ -0,0 +1,795 @@
+module Sequel
+  # The SQL module holds classes whose instances represent SQL fragments.
+  # It also holds modules that are included in core ruby classes that
+  # make Sequel a friendly DSL.
+  module SQL
+    ### Parent Classes ###
+
+    # Classes/Modules aren't an alphabetical order due to the fact that
+    # some reference constants defined in others at load time.
+
+    # Base class for all SQL fragments
+    class Expression
+      # Create a to_s instance method that takes a dataset, and calls
+      # the method provided on the dataset with args as the argument (self by default).
+      # Used to DRY up some code.
+      def self.to_s_method(meth, args=:self) # :nodoc:
+        class_eval("def to_s(ds); ds.#{meth}(#{args}) end", __FILE__, __LINE__)
+      end
+      private_class_method :to_s_method
+
+      # Returns self, because SQL::Expression already acts like
+      # LiteralString.
+      def lit
+        self
+      end
+    end
+
+    # Represents a complex SQL expression, with a given operator and one
+    # or more attributes (which may also be ComplexExpressions, forming
+    # a tree).  This class is the backbone of the blockless filter support in
+    # Sequel.
+    #
+    # This is an abstract class that is not that useful by itself.  The
+    # subclasses BooleanExpression, NumericExpression, and StringExpression
+    # define the behavior of the DSL via operators.
+    class ComplexExpression < Expression
+      # A hash of the opposite for each operator symbol, used for inverting
+      # objects.
+      OPERTATOR_INVERSIONS = {:AND => :OR, :OR => :AND, :< => :>=, :> => :<=,
+        :<= => :>, :>= => :<, :'=' => :'!=' , :'!=' => :'=', :LIKE => :'NOT LIKE',
+        :'NOT LIKE' => :LIKE, :~ => :'!~', :'!~' => :~, :IN => :'NOT IN',
+        :'NOT IN' => :IN, :IS => :'IS NOT', :'IS NOT' => :IS, :'~*' => :'!~*',
+        :'!~*' => :'~*', :NOT => :NOOP, :NOOP => :NOT, :ILIKE => :'NOT ILIKE',
+        :'NOT ILIKE'=>:ILIKE}
+
+      # Mathematical Operators used in NumericMethods
+      MATHEMATICAL_OPERATORS = [:+, :-, :/, :*]
+
+      # Mathematical Operators used in NumericMethods
+      BITWISE_OPERATORS = [:&, :|, :^, :<<, :>>]
+
+      # Inequality Operators used in InequalityMethods
+      INEQUALITY_OPERATORS = [:<, :>, :<=, :>=]
+
+      # Hash of ruby operator symbols to SQL operators, used in BooleanMethods
+      BOOLEAN_OPERATOR_METHODS = {:& => :AND, :| =>:OR}
+
+      # Operators that use IS, used for special casing to override literal true/false values
+      IS_OPERATORS = [:IS, :'IS NOT']
+
+      # Operator symbols that take exactly two arguments
+      TWO_ARITY_OPERATORS = [:'=', :'!=', :LIKE, :'NOT LIKE', \
+        :~, :'!~', :'~*', :'!~*', :IN, :'NOT IN', :ILIKE, :'NOT ILIKE'] + \
+        INEQUALITY_OPERATORS + BITWISE_OPERATORS + IS_OPERATORS 
+
+      # Operator symbols that take one or more arguments
+      N_ARITY_OPERATORS = [:AND, :OR, :'||'] + MATHEMATICAL_OPERATORS
+
+      # Operator symbols that take one argument
+      ONE_ARITY_OPERATORS = [:NOT, :NOOP, :'B~']
+
+      # An array of args for this object
+      attr_reader :args
+
+      # The operator symbol for this object
+      attr_reader :op
+      
+      # Set the operator symbol and arguments for this object to the ones given.
+      # Convert all args that are hashes or arrays with all two pairs to BooleanExpressions.
+      # Raise an error if the operator doesn't allow boolean input and a boolean argument is given.
+      # Raise an error if the wrong number of arguments for a given operator is used.
+      def initialize(op, *args)
+        args.map!{|a| Sequel.condition_specifier?(a) ? SQL::BooleanExpression.from_value_pairs(a) : a}
+        case op
+        when *N_ARITY_OPERATORS
+          raise(Error, "The #{op} operator requires at least 1 argument") unless args.length >= 1
+        when *TWO_ARITY_OPERATORS
+          raise(Error, "The #{op} operator requires precisely 2 arguments") unless args.length == 2
+        when *ONE_ARITY_OPERATORS
+          raise(Error, "The #{op} operator requires a single argument") unless args.length == 1
+        else
+          raise(Error, "Invalid operator #{op}")
+        end
+        @op = op
+        @args = args
+      end
+
+      # Returns true if the receiver is the same expression as the
+      # the +other+ expression.
+      def eql?(other)
+        other.is_a?(self.class) && @op.eql?(other.op) && @args.eql?(other.args)
+      end
+      alias == eql?
+      
+      to_s_method :complex_expression_sql, '@op, @args'
+    end
+
+    # The base class for expressions that can be used in multiple places in
+    # the SQL query.  
+    class GenericExpression < Expression
+    end
+    
+    ### Modules ###
+
+    # Methods that create aliased identifiers
+    module AliasMethods
+      # Create an SQL column alias of the receiving column or expression to the given alias.
+      def as(aliaz)
+        AliasedExpression.new(self, aliaz)
+      end
+    end
+
+    # This defines the bitwise methods: &, |, ^, ~, <<, and >>.  Because these
+    # methods overlap with the standard BooleanMethods methods, and they only
+    # make sense for numbers, they are only included in NumericExpression.
+    module BitwiseMethods
+      ComplexExpression::BITWISE_OPERATORS.each do |o|
+        define_method(o) do |ce|
+          case ce
+          when NumericExpression 
+            NumericExpression.new(o, self, ce)
+          when ComplexExpression
+            raise(Sequel::Error, "cannot apply #{o} to a non-numeric expression")
+          else  
+            NumericExpression.new(o, self, ce)
+          end
+        end
+      end
+
+      # Do the bitwise compliment of the self
+      def ~
+        NumericExpression.new(:'B~', self)
+      end
+    end
+
+    # This module includes the methods that are defined on objects that can be 
+    # used in a boolean context in SQL (Symbol, LiteralString, SQL::Function,
+    # and SQL::BooleanExpression).
+    #
+    # This defines the ~ (NOT), & (AND), and | (OR) methods.
+    module BooleanMethods
+      # Create a new BooleanExpression with NOT, representing the inversion of whatever self represents.
+      def ~
+        BooleanExpression.invert(self)
+      end
+      
+      ComplexExpression::BOOLEAN_OPERATOR_METHODS.each do |m, o|
+        define_method(m) do |ce|
+          case ce
+          when BooleanExpression
+            BooleanExpression.new(o, self, ce)
+          when ComplexExpression
+            raise(Sequel::Error, "cannot apply #{o} to a non-boolean expression")
+          else  
+            BooleanExpression.new(o, self, ce)
+          end
+        end
+      end
+    end
+
+    # Holds methods that are used to cast objects to differen SQL types.
+    module CastMethods 
+      # Cast the reciever to the given SQL type.  You can specify a ruby class as a type,
+      # and it is handled similarly to using a database independent type in the schema methods.
+      def cast(sql_type)
+        Cast.new(self, sql_type)
+      end
+
+      # Cast the reciever to the given SQL type (or the database's default integer type if none given),
+      # and return the result as a NumericExpression. 
+      def cast_numeric(sql_type = nil)
+        cast(sql_type || Integer).sql_number
+      end
+
+      # Cast the reciever to the given SQL type (or the database's default string type if none given),
+      # and return the result as a StringExpression, so you can use +
+      # directly on the result for SQL string concatenation.
+      def cast_string(sql_type = nil)
+        cast(sql_type || String).sql_string
+      end
+    end
+    
+    # Adds methods that allow you to treat an object as an instance of a specific
+    # ComplexExpression subclass.  This is useful if another library
+    # overrides the methods defined by Sequel.
+    #
+    # For example, if Symbol#/ is overridden to produce a string (for
+    # example, to make file system path creation easier), the
+    # following code will not do what you want:
+    #
+    #   :price/10 > 100
+    #
+    # In that case, you need to do the following:
+    #
+    #   :price.sql_number/10 > 100
+    module ComplexExpressionMethods
+      # Extract a datetime_part (e.g. year, month) from self:
+      #
+      #   :date.extract(:year) # SQL:  extract(year FROM "date")
+      #
+      # Also has the benefit of returning the result as a
+      # NumericExpression instead of a generic ComplexExpression.
+      #
+      # The extract function is in the SQL standard, but it doesn't
+      # doesn't use the standard function calling convention.
+      def extract(datetime_part)
+        Function.new(:extract, PlaceholderLiteralString.new("#{datetime_part} FROM ?", [self])).sql_number
+      end
+
+      # Return a BooleanExpression representation of self.
+      def sql_boolean
+        BooleanExpression.new(:NOOP, self)
+      end
+
+      # Return a NumericExpression representation of self.
+      def sql_number
+        NumericExpression.new(:NOOP, self)
+      end
+
+      # Return a StringExpression representation of self.
+      def sql_string
+        StringExpression.new(:NOOP, self)
+      end
+    end
+
+    # Includes a method that returns Identifiers.
+    module IdentifierMethods
+      # Return self wrapped as an identifier.
+      def identifier
+        Identifier.new(self)
+      end
+    end
+
+    # This module includes the methods that are defined on objects that can be 
+    # used in a numeric or string context in SQL (Symbol (except on ruby 1.9), LiteralString, 
+    # SQL::Function, and SQL::StringExpression).
+    #
+    # This defines the >, <, >=, and <= methods.
+    module InequalityMethods
+      ComplexExpression::INEQUALITY_OPERATORS.each do |o|
+        define_method(o) do |ce|
+          case ce
+          when BooleanExpression, TrueClass, FalseClass, NilClass, Hash, Array
+            raise(Error, "cannot apply #{o} to a boolean expression")
+          else  
+            BooleanExpression.new(o, self, ce)
+          end
+        end
+      end
+    end
+
+    # This module augments the default initalize method for the 
+    # ComplexExpression subclass it is included in, so that
+    # attempting to use boolean input when initializing a NumericExpression
+    # or StringExpression results in an error.
+    module NoBooleanInputMethods
+      # Raise an Error if one of the args would be boolean in an SQL
+      # context, otherwise call super.
+      def initialize(op, *args)
+        args.each do |a|
+          case a
+          when BooleanExpression, TrueClass, FalseClass, NilClass, Hash, Array
+            raise(Error, "cannot apply #{op} to a boolean expression")
+          end
+        end
+        super
+      end
+    end
+
+    # This module includes the methods that are defined on objects that can be 
+    # used in a numeric context in SQL (Symbol, LiteralString, SQL::Function,
+    # and SQL::NumericExpression).
+    #
+    # This defines the +, -, *, and / methods.
+    module NumericMethods
+      ComplexExpression::MATHEMATICAL_OPERATORS.each do |o|
+        define_method(o) do |ce|
+          case ce
+          when NumericExpression
+            NumericExpression.new(o, self, ce)
+          when ComplexExpression
+            raise(Sequel::Error, "cannot apply #{o} to a non-numeric expression")
+          else  
+            NumericExpression.new(o, self, ce)
+          end
+        end
+      end
+    end
+
+    # Methods that create OrderedExpressions, used for sorting by columns
+    # or more complex expressions.
+    module OrderMethods
+      # Mark the receiving SQL column as sorting in a descending fashion.
+      def desc
+        OrderedExpression.new(self)
+      end
+      
+      # Mark the receiving SQL column as sorting in an ascending fashion (generally a no-op).
+      def asc
+        OrderedExpression.new(self, false)
+      end
+    end
+
+    # Methods that created QualifiedIdentifiers, used for qualifying column
+    # names with a table or table names with a schema.
+    module QualifyingMethods
+      # Qualify the current object with the given table/schema.
+      def qualify(ts)
+        QualifiedIdentifier.new(ts, self)
+      end
+    end
+
+    # This module includes the methods that are defined on objects that can be 
+    # used in a string context in SQL (Symbol, LiteralString, SQL::Function,
+    # and SQL::StringExpression).
+    #
+    # This defines the like (LIKE) and ilike methods, used for pattern matching.
+    # like is case sensitive, ilike is case insensitive.
+    module StringMethods
+      # Create a BooleanExpression case insensitive pattern match of self
+      # with the given patterns.  See StringExpression.like.
+      def ilike(*ces)
+        StringExpression.like(self, *(ces << {:case_insensitive=>true}))
+      end
+
+      # Create a BooleanExpression case sensitive pattern match of self with
+      # the given patterns.  See StringExpression.like.
+      def like(*ces)
+        StringExpression.like(self, *ces)
+      end
+    end
+
+    # This module is included in StringExpression and can be included elsewhere
+    # to allow the use of the + operator to represent concatenation of SQL
+    # Strings:
+    #
+    #   :x.sql_string + :y => # SQL: x || y
+    module StringConcatenationMethods
+      def +(ce)
+        StringExpression.new(:'||', self, ce)
+      end
+    end
+
+    class ComplexExpression
+      include AliasMethods
+      include CastMethods
+      include OrderMethods
+    end
+
+    class GenericExpression
+      include AliasMethods
+      include CastMethods
+      include OrderMethods
+      include ComplexExpressionMethods
+      include BooleanMethods
+      include NumericMethods
+      include StringMethods
+      include InequalityMethods
+    end
+
+    ### Classes ###
+
+    # Represents an aliasing of an expression/column to a given name.
+    class AliasedExpression < Expression
+      # The expression to alias
+      attr_reader :expression
+
+      # The alias to use for the expression, not alias since that is
+      # a keyword in ruby.
+      attr_reader :aliaz
+
+      # Create an object with the given expression and alias.
+      def initialize(expression, aliaz)
+        @expression, @aliaz = expression, aliaz
+      end
+
+      to_s_method :aliased_expression_sql
+    end
+
+    # Blob is used to represent binary data in the Ruby environment that is
+    # stored as a blob type in the database. Sequel represents binary data as a Blob object because 
+    # certain database engines require binary data to be escaped.
+    class Blob < ::String
+      # Returns self
+      def to_sequel_blob
+        self
+      end
+    end
+
+    # Subclass of ComplexExpression where the expression results
+    # in a boolean value in SQL.
+    class BooleanExpression < ComplexExpression
+      include BooleanMethods
+      
+      # Take pairs of values (e.g. a hash or array of arrays of two pairs)
+      # and converts it to a BooleanExpression.  The operator and args
+      # used depends on the case of the right (2nd) argument:
+      #
+      # * 0..10 - left >= 0 AND left <= 10
+      # * [1,2] - left IN (1,2)
+      # * nil - left IS NULL
+      # * /as/ - left ~ 'as'
+      # * :blah - left = blah
+      # * 'blah' - left = 'blah'
+      #
+      # If multiple arguments are given, they are joined with the op given (AND
+      # by default, OR possible).  If negate is set to true,
+      # all subexpressions are inverted before used.  Therefore, the following
+      # expressions are equivalent:
+      #
+      #   ~from_value_pairs(hash)
+      #   from_value_pairs(hash, :OR, true)
+      def self.from_value_pairs(pairs, op=:AND, negate=false)
+        pairs = pairs.collect do |l,r|
+          ce = case r
+          when Range
+            new(:AND, new(:>=, l, r.begin), new(r.exclude_end? ? :< : :<=, l, r.end))
+          when Array, ::Sequel::Dataset, SQLArray
+            new(:IN, l, r)
+          when NilClass, TrueClass, FalseClass
+            new(:IS, l, r)
+          when Regexp
+            StringExpression.like(l, r)
+          else
+            new(:'=', l, r)
+          end
+          negate ? invert(ce) : ce
+        end
+        pairs.length == 1 ? pairs.at(0) : new(op, *pairs)
+      end
+      
+      # Invert the expression, if possible.  If the expression cannot
+      # be inverted, raise an error.  An inverted expression should match everything that the
+      # uninverted expression did not match, and vice-versa, except for possible issues with
+      # SQL NULL (i.e. 1 == NULL is NULL and 1 != NULL is also NULL).
+      def self.invert(ce)
+        case ce
+        when BooleanExpression
+          case op = ce.op
+          when :AND, :OR
+            BooleanExpression.new(OPERTATOR_INVERSIONS[op], *ce.args.collect{|a| BooleanExpression.invert(a)})
+          else
+            BooleanExpression.new(OPERTATOR_INVERSIONS[op], *ce.args.dup)
+          end
+        when ComplexExpression
+          raise(Sequel::Error, "operator #{ce.op} cannot be inverted")
+        else
+          BooleanExpression.new(:NOT, ce)
+        end
+      end
+    end
+
+    # Represents an SQL CASE expression, used for conditions.
+    class CaseExpression < GenericExpression
+      # An array of all two pairs with the first element specifying the
+      # condition and the second element specifying the result.
+      attr_reader :conditions
+
+      # The default value if no conditions are true
+      attr_reader :default
+
+      # The expression to test the conditions against
+      attr_reader :expression
+
+      # Create an object with the given conditions and
+      # default value.
+      def initialize(conditions, default, expression = nil)
+        raise(Sequel::Error, 'CaseExpression conditions must be a hash or array of all two pairs') unless Sequel.condition_specifier?(conditions)
+        @conditions, @default, @expression = conditions.to_a, default, expression
+      end
+
+      to_s_method :case_expression_sql
+    end
+
+    # Represents a cast of an SQL expression to a specific type.
+    class Cast < GenericExpression
+      # The expression to cast
+      attr_reader :expr
+
+      # The type to which to cast the expression
+      attr_reader :type
+      
+      # Set the attributes to the given arguments
+      def initialize(expr, type)
+        @expr = expr
+        @type = type
+      end
+
+      to_s_method :cast_sql, '@expr, @type'
+    end
+
+    # Represents all columns in a given table, table.* in SQL
+    class ColumnAll < Expression
+      # The table containing the columns being selected
+      attr_reader :table
+
+      # Create an object with the given table
+      def initialize(table)
+        @table = table
+      end
+
+      # ColumnAll expressions are considered equivalent if they
+      # have the same class and string representation
+      def ==(x)
+        x.class == self.class and @table == x.table
+      end
+
+      to_s_method :column_all_sql
+    end
+
+    # Represents an SQL function call.
+    class Function < GenericExpression
+      # The array of arguments to pass to the function (may be blank)
+      attr_reader :args
+
+      # The SQL function to call
+      attr_reader :f
+      
+      # Set the attributes to the given arguments
+      def initialize(f, *args)
+        @f, @args = f, args
+      end
+
+      # Functions are considered equivalent if they
+      # have the same class, function, and arguments.
+      def ==(x)
+         x.class == self.class && @f == x.f && @args == x.args
+      end
+
+      to_s_method :function_sql
+    end
+    
+    # Represents an identifier (column or table). Can be used
+    # to specify a Symbol with multiple underscores should not be
+    # split, or for creating an identifier without using a symbol.
+    class Identifier < GenericExpression
+      include QualifyingMethods
+
+      # The table and column to reference
+      attr_reader :value
+
+      # Set the value to the given argument
+      def initialize(value)
+        @value = value
+      end
+      
+      to_s_method :quote_identifier, '@value'
+    end
+    
+    # Represents an SQL JOIN clause, used for joining tables.
+    class JoinClause < Expression
+      # The type of join to do
+      attr_reader :join_type
+
+      # The actual table to join
+      attr_reader :table
+
+      # The table alias to use for the join, if any
+      attr_reader :table_alias
+
+      # Create an object with the given join_type, table, and table alias
+      def initialize(join_type, table, table_alias = nil)
+        @join_type, @table, @table_alias = join_type, table, table_alias
+      end
+
+      to_s_method :join_clause_sql
+    end
+
+    # Represents an SQL JOIN table ON conditions clause.
+    class JoinOnClause < JoinClause
+      # The conditions for the join
+      attr_reader :on
+
+      # Create an object with the ON conditions and call super with the
+      # remaining args.
+      def initialize(on, *args)
+        @on = on
+        super(*args)
+      end
+
+      to_s_method :join_on_clause_sql
+    end
+
+    # Represents an SQL JOIN table USING (columns) clause.
+    class JoinUsingClause < JoinClause
+      # The columns that appear in both tables that should be equal 
+      # for the conditions to match.
+      attr_reader :using
+
+      # Create an object with the given USING conditions and call super
+      # with the remaining args.
+      def initialize(using, *args)
+        @using = using
+        super(*args)
+      end
+
+      to_s_method :join_using_clause_sql
+    end
+
+    # Represents a literal string with placeholders and arguments.
+    # This is necessary to ensure delayed literalization of the arguments
+    # required for the prepared statement support
+    class PlaceholderLiteralString < Expression
+      # The arguments that will be subsituted into the placeholders.
+      attr_reader :args
+
+      # The literal string containing placeholders
+      attr_reader :str
+
+      # Whether to surround the expression with parantheses
+      attr_reader :parens
+
+      # Create an object with the given string, placeholder arguments, and parens flag.
+      def initialize(str, args, parens=false)
+        @str = str
+        @args = args
+        @parens = parens
+      end
+
+      to_s_method :placeholder_literal_string_sql
+    end
+
+    # Subclass of ComplexExpression where the expression results
+    # in a numeric value in SQL.
+    class NumericExpression < ComplexExpression
+      include BitwiseMethods 
+      include NumericMethods
+      include InequalityMethods
+      include NoBooleanInputMethods
+    end
+
+    # Represents a column/expression to order the result set by.
+    class OrderedExpression < Expression
+      # The expression to order the result set by.
+      attr_reader :expression
+
+      # Whether the expression should order the result set in a descending manner
+      attr_reader :descending
+
+      # Set the expression and descending attributes to the given values.
+      def initialize(expression, descending = true)
+        @expression, @descending = expression, descending
+      end
+
+      to_s_method :ordered_expression_sql
+    end
+
+    # Represents a qualified (column with table or table with schema) reference. 
+    class QualifiedIdentifier < GenericExpression
+      include QualifyingMethods
+
+      # The column to reference
+      attr_reader :column
+
+      # The table to reference
+      attr_reader :table
+
+      # Set the table and column to the given arguments
+      def initialize(table, column)
+        @table, @column = table, column
+      end
+      
+      to_s_method :qualified_identifier_sql
+    end
+    
+    # Subclass of ComplexExpression where the expression results
+    # in a text/string/varchar value in SQL.
+    class StringExpression < ComplexExpression
+      include StringMethods
+      include StringConcatenationMethods
+      include InequalityMethods
+      include NoBooleanInputMethods
+      
+      # Creates a SQL pattern match exprssion. left (l) is the SQL string we
+      # are matching against, and ces are the patterns we are matching.
+      # The match succeeds if any of the patterns match (SQL OR).  Patterns
+      # can be given as strings or regular expressions.  Strings will cause
+      # the SQL LIKE operator to be used, and should be supported by most
+      # databases.  Regular expressions will probably only work on MySQL
+      # and PostgreSQL, and SQL regular expression syntax is not fully compatible
+      # with ruby regular expression syntax, so be careful if using regular
+      # expressions.
+      # 
+      # The pattern match will be case insensitive if the last argument is a hash
+      # with a key of :case_insensitive that is not false or nil. Also,
+      # if a case insensitive regular expression is used (//i), that particular
+      # pattern which will always be case insensitive.
+      def self.like(l, *ces)
+        case_insensitive = (ces.last.is_a?(Hash) ? ces.pop : {})[:case_insensitive]
+        ces.collect! do |ce|
+          op, expr = Regexp === ce ? [ce.casefold? || case_insensitive ? :'~*' : :~, ce.source] : [case_insensitive ? :ILIKE : :LIKE, ce]
+          BooleanExpression.new(op, l, expr)
+        end
+        ces.length == 1 ? ces.at(0) : BooleanExpression.new(:OR, *ces)
+      end
+    end
+
+    # Represents an SQL array.  Added so it is possible to deal with a
+    # ruby array of all two pairs as an SQL array instead of an ordered
+    # hash-like conditions specifier.
+    class SQLArray < Expression
+      # Create an object with the given array.
+      def initialize(array)
+        @array = array
+      end
+
+      to_s_method :array_sql, '@array'
+    end
+
+    # Represents an SQL array access, with multiple possible arguments.
+    class Subscript < GenericExpression
+      # The SQL array column
+      attr_reader :f
+
+      # The array of subscripts to use (should be an array of numbers)
+      attr_reader :sub
+
+      # Set the array column and subscripts to the given arguments
+      def initialize(f, sub)
+        @f, @sub = f, sub
+      end
+
+      # Create a new subscript appending the given subscript(s)
+      # the the current array of subscripts.
+      def |(sub)
+        Subscript.new(@f, @sub + Array(sub))
+      end
+      
+      to_s_method :subscript_sql
+    end
+
+    if RUBY_VERSION >= '1.9.0'
+      class VirtualRow < BasicObject
+      end
+    else
+      class VirtualRow
+        (instance_methods - %w"__id__ __send__ instance_eval == equal?").each{|m| undef_method(m)}
+      end
+    end
+
+    # The purpose of this class is to allow the easy creation of SQL identifiers and functions
+    # without relying on methods defined on Symbol.  This is useful if another library defines
+    # the methods defined by Sequel, or if you are running on ruby 1.9.
+    #
+    # An instance of this class is yielded to the block supplied to filter, order, and select.
+    # If Sequel.virtual_row_instance_eval is true and the block doesn't take an argument,
+    # the block is instance_evaled in the context of a new instance of this class.
+    #
+    # Examples:
+    #
+    #   ds = DB[:t]
+    #   ds.filter{|r| r.name < 2} # SELECT * FROM t WHERE (name < 2)
+    #   ds.filter{|r| r.table__column + 1 < 2} # SELECT * FROM t WHERE ((table.column + 1) < 2)
+    #   ds.filter{|r| r.is_active(1, 'arg2')} # SELECT * FROM t WHERE is_active(1, 'arg2')
+    class VirtualRow
+      # Can return Identifiers, QualifiedIdentifiers, or Functions:
+      #
+      # * Function - returned if any arguments are supplied, using the method name
+      #   as the function name, and the arguments as the function arguments.
+      # * QualifiedIdentifier - returned if the method name contains __, with the
+      #   table being the part before __, and the column being the part after.
+      # * Identifier - returned otherwise, using the method name.
+      def method_missing(m, *args)
+        if args.empty?
+          table, column = m.to_s.split('__', 2)
+          column ? QualifiedIdentifier.new(table, column) : Identifier.new(m)
+        else
+          Function.new(m, *args)
+        end
+      end
+    end
+  end
+
+  # LiteralString is used to represent literal SQL expressions. A 
+  # LiteralString is copied verbatim into an SQL statement. Instances of
+  # LiteralString can be created by calling String#lit.
+  class LiteralString < ::String
+    include SQL::OrderMethods
+    include SQL::ComplexExpressionMethods
+    include SQL::BooleanMethods
+    include SQL::NumericMethods
+    include SQL::StringMethods
+    include SQL::InequalityMethods
+  end
+end
diff -pruN 1.5.1-1/lib/sequel/version.rb 3.0.0-1/lib/sequel/version.rb
--- 1.5.1-1/lib/sequel/version.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel/version.rb	2009-05-04 17:36:50.000000000 +0100
@@ -0,0 +1,12 @@
+module Sequel
+  MAJOR = 3
+  MINOR = 0
+  TINY  = 0
+  
+  VERSION = [MAJOR, MINOR, TINY].join('.')
+  
+  # The version of Sequel you are using, as a string (e.g. "2.11.0")
+  def self.version
+    VERSION
+  end
+end
diff -pruN 1.5.1-1/lib/sequel_core.rb 3.0.0-1/lib/sequel_core.rb
--- 1.5.1-1/lib/sequel_core.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/lib/sequel_core.rb	2009-03-16 20:53:19.000000000 +0000
@@ -0,0 +1 @@
+require 'sequel/core'
diff -pruN 1.5.1-1/lib/sequel_model/associations.rb 3.0.0-1/lib/sequel_model/associations.rb
--- 1.5.1-1/lib/sequel_model/associations.rb	2008-04-24 22:14:31.000000000 +0100
+++ 3.0.0-1/lib/sequel_model/associations.rb	1970-01-01 01:00:00.000000000 +0100
@@ -1,388 +0,0 @@
-# Associations are used in order to specify relationships between model classes
-# that reflect relations between tables in the database using foreign keys.
-#
-# Each kind of association adds a number of methods to the model class which
-# are specialized according to the association type and optional parameters
-# given in the definition. Example:
-# 
-#   class Project < Sequel::Model
-#     many_to_one :portfolio
-#     one_to_many :milestones
-#   end
-# 
-# The project class now has the following instance methods:
-# * portfolio - Returns the associated portfolio
-# * portfolio=(obj) - Sets the associated portfolio to the object,
-#   but the change is not persisted until you save the record.
-# * milestones - Returns an array of associated milestones
-# * milestones_dataset - Returns a dataset that would return the associated
-#   milestones, allowing for further filtering/limiting/etc.
-# * add_milestone(obj) - Associates the passed milestone with this object
-# * remove_milestone(obj) - Removes the association with the passed milestone
-#
-# By default the classes for the associations are inferred from the association
-# name, so for example the Project#portfolio will return an instance of 
-# Portfolio, and Project#milestones will return an array of Milestone 
-# instances, in similar fashion to how ActiveRecord infers class names.
-#
-# Association definitions are also reflected by the class, e.g.:
-#
-#   >> Project.associations
-#   => [:portfolio, :milestones]
-#   >> Project.association_reflection(:portfolio)
-#   => {:type => :many_to_one, :name => :portfolio, :class_name => "Portfolio"}
-#
-# Associations can be defined by either using the associate method, or by
-# calling one of the three methods: many_to_one, one_to_many, many_to_many.
-# Sequel::Model also provides aliases for these methods that conform to
-# ActiveRecord conventions: belongs_to, has_many, has_and_belongs_to_many.
-# For example, the following three statements are equivalent:
-#
-#   associate :one_to_many, :attributes
-#   one_to_many :attributes
-#   has_many :attributes
-module Sequel::Model::Associations
-  RECIPROCAL_ASSOCIATIONS = {:many_to_one=>:one_to_many, :one_to_many=>:many_to_one, :many_to_many=>:many_to_many}
-
-  # Array of all association reflections
-  def all_association_reflections
-    association_reflections.values
-  end
-  
-  # Associates a related model with the current model. The following types are
-  # supported:
-  #
-  # * :many_to_one - Foreign key in current model's table points to 
-  #   associated model's primary key.  Each associated model object can
-  #   be associated with more than one current model objects.  Each current
-  #   model object can be associated with only one associated model object.
-  #   Similar to ActiveRecord/DataMapper's belongs_to.
-  # * :one_to_many - Foreign key in associated model's table points to this
-  #   model's primary key.   Each current model object can be associated with
-  #   more than one associated model objects.  Each associated model object
-  #   can be associated with only one current model object.
-  #   Similar to ActiveRecord/DataMapper's has_many.
-  # * :many_to_many - A join table is used that has a foreign key that points
-  #   to this model's primary key and a foreign key that points to the
-  #   associated model's primary key.  Each current model object can be
-  #   associated with many associated model objects, and each associated
-  #   model object can be associated with many current model objects.
-  #   Similar to ActiveRecord/DataMapper's has_and_belongs_to_many.
-  #
-  # The following options can be supplied:
-  # * *ALL types*:
-  #   - :class - The associated class or its name. If not
-  #     given, uses the association's name, which is camelized (and
-  #     singularized unless the type is :many_to_one)
-  #   - :eager - The associations to eagerly load when loading the associated object.
-  #     For many_to_one associations, this is ignored unless this association is
-  #     being eagerly loaded, as it doesn't save queries unless multiple objects
-  #     can be loaded at once.
-  #   - :reciprocal - the symbol name of the instance variable of the reciprocal association,
-  #     if it exists.  By default, sequel will try to determine it by looking at the
-  #     associated model's assocations for a association that matches
-  #     the current association's key(s).  Set to nil to not use a reciprocal.
-  # * :one_to_many/:many_to_many:
-  #   - :order - the column(s) by which to order the association dataset.  Can be a
-  #     singular column or an array.
-  # * :many_to_one:
-  #   - :key - foreign_key in current model's table that references
-  #     associated model's primary key, as a symbol.  Defaults to :"#{name}_id".
-  # * :one_to_many:
-  #   - :key - foreign key in associated model's table that references
-  #     current model's primary key, as a symbol.  Defaults to
-  #     :"#{self.name.underscore}_id".
-  # * :many_to_many:
-  #   - :join_table - name of table that includes the foreign keys to both
-  #     the current model and the associated model, as a symbol.  Defaults to the name
-  #     of current model and name of associated model, pluralized,
-  #     underscored, sorted, and joined with '_'.
-  #   - :left_key - foreign key in join table that points to current model's
-  #     primary key, as a symbol.
-  #   - :right_key - foreign key in join table that points to associated
-  #     model's primary key, as a symbol.
-  #   - :select - the attributes to select.  Defaults to the associated class's
-  #     table_name.*, which means it doesn't include the attributes from the
-  #     join table.  If you want to include the join table attributes, you can
-  #     use this option, but beware that the join table attributes can clash with
-  #     attributes from the model table, so you should alias any attributes that have
-  #     the same name in both the join table and the associated table.
-  def associate(type, name, opts = {}, &block)
-    # check arguments
-    raise ArgumentError unless [:many_to_one, :one_to_many, :many_to_many].include?(type) && Symbol === name
-
-    # merge early so we don't modify opts
-    opts = opts.merge(:type => type, :name => name, :block => block, :cache => true)
-
-    # deprecation
-    if opts[:from]
-      Sequel::Deprecation.deprecate("The :from option to Sequel::Model.associate is deprecated and will be removed in Sequel 2.0.  Use the :class option.")
-      opts[:class] = opts[:from]
-    end
-
-    # find class
-    case opts[:class]
-      when String, Symbol
-        # Delete :class to allow late binding
-        opts[:class_name] ||= opts.delete(:class).to_s
-      when Class
-        opts[:class_name] ||= opts[:class].name
-    end
-
-    send(:"def_#{type}", name, opts)
-
-    # don't add to association_reflections until we are sure there are no errors
-    association_reflections[name] = opts
-  end
-  
-  # The association reflection hash for the association of the given name.
-  def association_reflection(name)
-    association_reflections[name]
-  end
-  
-  # Array of association name symbols
-  def associations
-    association_reflections.keys
-  end
-
-  # Shortcut for adding a one_to_many association, see associate
-  def one_to_many(*args, &block)
-    associate(:one_to_many, *args, &block)
-  end
-  alias_method :has_many, :one_to_many
-  
-  # Shortcut for adding a many_to_one association, see associate
-  def many_to_one(*args, &block)
-    associate(:many_to_one, *args, &block)
-  end
-  alias_method :belongs_to, :many_to_one
-  
-  # Shortcut for adding a many_to_many association, see associate
-  def many_to_many(*args, &block)
-    associate(:many_to_many, *args, &block)
-  end
-  alias_method :has_and_belongs_to_many, :many_to_many
-  
-  private
-  # The class related to the given association reflection
-  def associated_class(opts)
-    opts[:class] ||= opts[:class_name].constantize
-  end
-
-  # Name symbol for add association method
-  def association_add_method_name(name)
-    :"add_#{name.to_s.singularize}"
-  end
-
-  # Name symbol of association instance variable
-  def association_ivar(name)
-    :"@#{name}"
-  end
-  
-  # Name symbol for remove_method_name
-  def association_remove_method_name(name)
-    :"remove_#{name.to_s.singularize}"
-  end
-  
-  # Hash storing the association reflections.  Keys are association name
-  # symbols, values are association reflection hashes.
-  def association_reflections
-    @association_reflections ||= {}
-  end
-  
-  # Defines an association
-  def def_association_dataset_methods(name, opts, &block)
-    dataset_method = :"#{name}_dataset"
-    helper_method = :"#{name}_helper"
-    dataset_block = opts[:block]
-    ivar = association_ivar(name)
-    
-    # define a method returning the association dataset (with optional order)
-    if order = opts[:order]
-      class_def(dataset_method) {instance_eval(&block).order(order)}
-    else
-      class_def(dataset_method, &block)
-    end
-    
-    # If a block is given, define a helper method for it, because it takes
-    # an argument.  This is unnecessary in Ruby 1.9, as that has instance_exec.
-    if dataset_block
-      class_def(helper_method, &dataset_block)
-    end
-    
-    class_def(name) do |*reload|
-      if !reload[0] && obj = instance_variable_get(ivar)
-        obj
-      else
-        ds = send(dataset_method)
-        # if the a dataset block was specified, we need to call it and use
-        # the result as the dataset to fetch records from.
-        if dataset_block
-          ds = send(helper_method, ds)
-        end
-        if eager = opts[:eager]
-          ds = ds.eager(eager)
-        end
-        objs = ds.all
-        # Only one_to_many associations should set the reciprocal object
-        if (opts[:type] == :one_to_many) && (reciprocal = model.send(:reciprocal_association, opts))
-          objs.each{|o| o.instance_variable_set(reciprocal, self)}
-        end
-        instance_variable_set(ivar, objs)
-      end
-    end
-  end
-  
-  # Defines an association getter method, caching the block result in an 
-  # instance variable. The defined method takes an optional reload parameter
-  # that can be set to true in order to bypass the cache.
-  def def_association_getter(name, &block)
-    ivar = association_ivar(name)
-    class_def(name) do |*reload|
-      if !reload[0] && obj = instance_variable_get(ivar)
-        obj == :null ? nil : obj
-      else
-        obj = instance_eval(&block)
-        instance_variable_set(ivar, obj || :null)
-        obj
-      end
-    end
-  end
-
-  # Adds many_to_many association instance methods
-  def def_many_to_many(name, opts)
-    assoc_class = method(:associated_class) # late binding of association dataset
-    recip_assoc = method(:reciprocal_association) # late binding of the reciprocal association
-    ivar = association_ivar(name)
-    left = (opts[:left_key] ||= default_remote_key)
-    right = (opts[:right_key] ||= :"#{name.to_s.singularize}_id")
-    opts[:class_name] ||= name.to_s.singularize.camelize
-    join_table = (opts[:join_table] ||= default_join_table_name(opts))
-    database = db
-    
-    def_association_dataset_methods(name, opts) do
-      klass = assoc_class[opts]
-      key = (opts[:right_primary_key] ||= :"#{klass.table_name}__#{klass.primary_key}")
-      selection = (opts[:select] ||= klass.table_name.*)
-      klass.select(selection).inner_join(join_table, right => key, left => pk)
-    end
-
-    class_def(association_add_method_name(name)) do |o|
-      database[join_table].insert(left => pk, right => o.pk)
-      if arr = instance_variable_get(ivar)
-        arr.push(o)
-      end
-      if (reciprocal = recip_assoc[opts]) && (list = o.instance_variable_get(reciprocal)) \
-         && !(list.include?(self))
-        list.push(self)
-      end
-      o
-    end
-    class_def(association_remove_method_name(name)) do |o|
-      database[join_table].filter(left => pk, right => o.pk).delete
-      if arr = instance_variable_get(ivar)
-        arr.delete(o)
-      end
-      if (reciprocal = recip_assoc[opts]) && (list = o.instance_variable_get(reciprocal))
-        list.delete(self)
-      end
-      o
-    end
-  end
-  
-  # Adds many_to_one association instance methods
-  def def_many_to_one(name, opts)
-    assoc_class = method(:associated_class) # late binding of association dataset
-    recip_assoc = method(:reciprocal_association) # late binding of the reciprocal association
-    ivar = association_ivar(name)
-    
-    key = (opts[:key] ||= :"#{name}_id")
-    opts[:class_name] ||= name.to_s.camelize
-    
-    def_association_getter(name) {(fk = send(key)) ? assoc_class[opts][fk] : nil}
-    class_def(:"#{name}=") do |o|
-      old_val = instance_variable_get(ivar) if reciprocal = recip_assoc[opts]
-      instance_variable_set(ivar, o)
-      send(:"#{key}=", (o.pk if o))
-      if reciprocal && (old_val != o)
-        if old_val && (list = old_val.instance_variable_get(reciprocal))
-          list.delete(self)
-        end
-        if o && (list = o.instance_variable_get(reciprocal)) && !(list.include?(self))
-          list.push(self) 
-        end
-      end
-      o
-    end
-  end
-  
-  # Adds one_to_many association instance methods
-  def def_one_to_many(name, opts)
-    assoc_class = method(:associated_class) # late binding of association dataset
-    recip_assoc = method(:reciprocal_association) # late binding of the reciprocal association
-    ivar = association_ivar(name)
-    key = (opts[:key] ||= default_remote_key)
-    opts[:class_name] ||= name.to_s.singularize.camelize
-    
-    def_association_dataset_methods(name, opts) {assoc_class[opts].filter(key => pk)}
-    
-    class_def(association_add_method_name(name)) do |o|
-      o.send(:"#{key}=", pk)
-      o.save!
-      if arr = instance_variable_get(ivar)
-        arr.push(o)
-      end
-      if reciprocal = recip_assoc[opts]
-        o.instance_variable_set(reciprocal, self)
-      end
-      o
-    end
-    class_def(association_remove_method_name(name)) do |o|
-      o.send(:"#{key}=", nil)
-      o.save!
-      if arr = instance_variable_get(ivar)
-        arr.delete(o)
-      end
-      if reciprocal = recip_assoc[opts]
-        o.instance_variable_set(reciprocal, :null)
-      end
-      o
-    end
-  end
-  
-  # Name symbol for default join table
-  def default_join_table_name(opts)
-    ([opts[:class_name].demodulize, name.demodulize]. \
-      map{|i| i.pluralize.underscore}.sort.join('_')).to_sym
-  end
-  
-  # Name symbol for default foreign key
-  def default_remote_key
-    :"#{name.demodulize.underscore}_id"
-  end
-
-  # Sets the reciprocal association variable in the reflection, if one exists
-  def reciprocal_association(reflection)
-    return reflection[:reciprocal] if reflection.include?(:reciprocal)
-    reciprocal_type = ::Sequel::Model::Associations::RECIPROCAL_ASSOCIATIONS[reflection[:type]]
-    if reciprocal_type == :many_to_many
-      left_key = reflection[:left_key]
-      right_key = reflection[:right_key]
-      join_table = reflection[:join_table]
-      associated_class(reflection).all_association_reflections.each do |assoc_reflect|
-        if assoc_reflect[:type] == :many_to_many && assoc_reflect[:left_key] == right_key \
-           && assoc_reflect[:right_key] == left_key && assoc_reflect[:join_table] == join_table
-          return reflection[:reciprocal] = association_ivar(assoc_reflect[:name]).to_s.freeze
-        end
-      end
-    else
-      key = reflection[:key]
-      associated_class(reflection).all_association_reflections.each do |assoc_reflect|
-        if assoc_reflect[:type] == reciprocal_type && assoc_reflect[:key] == key
-          return reflection[:reciprocal] = association_ivar(assoc_reflect[:name])
-        end
-      end
-    end
-    reflection[:reciprocal] = nil
-  end
-end
diff -pruN 1.5.1-1/lib/sequel_model/base.rb 3.0.0-1/lib/sequel_model/base.rb
--- 1.5.1-1/lib/sequel_model/base.rb	2008-04-28 17:14:45.000000000 +0100
+++ 3.0.0-1/lib/sequel_model/base.rb	1970-01-01 01:00:00.000000000 +0100
@@ -1,165 +0,0 @@
-module Sequel
-  class Model
-    # If possible, set the dataset for the model subclass as soon as it
-    # is created.
-    def self.inherited(subclass)
-      begin
-        if subclass.superclass == Model
-          unless subclass.name.empty?
-            subclass.set_dataset(Model.db[subclass.implicit_table_name])
-          end
-        elsif ds = subclass.superclass.instance_variable_get(:@dataset)
-          subclass.set_dataset(ds.clone)
-        end
-      rescue StandardError
-      end
-    end
-
-    # Returns the database associated with the Model class.
-    def self.db
-      return @db if @db
-      @db = self == Model ? ::Sequel::DATABASES.first : superclass.db
-      raise(Error, "No database associated with #{self}") unless @db
-      @db
-    end
-    
-    # Sets the database associated with the Model class.
-    def self.db=(db)
-      @db = db
-      if @dataset
-        set_dataset(db[table_name])
-      end
-    end
-    
-    # Returns the implicit table name for the model class.
-    def self.implicit_table_name
-      name.demodulize.underscore.pluralize.to_sym
-    end
-
-    # Returns the dataset associated with the Model class.
-    def self.dataset
-      @dataset || raise(Error, "No dataset associated with #{self}")
-    end
-
-    # If a block is given, define a method on the dataset with the given argument name using
-    # the given block as well as a method on the model that calls the
-    # dataset method.
-    #
-    # If a block is not given, define a method on the model for each argument
-    # that calls the dataset method of the same argument name.
-    def self.def_dataset_method(*args, &block)
-      raise(Error, "No arguments given") if args.empty?
-      if block_given?
-        raise(Error, "Defining a dataset method using a block requires only one argument") if args.length > 1
-        dataset.meta_def(args.first, &block)
-      end
-      args.each{|arg| instance_eval("def #{arg}(*args, &block); dataset.#{arg}(*args, &block) end", __FILE__, __LINE__)}
-    end
-    
-    # Returns the columns in the result set in their original order.
-    #
-    # See Dataset#columns for more information.
-    def self.columns
-      return @columns if @columns
-      @columns = dataset.naked.columns or
-      raise Error, "Could not fetch columns for #{self}"
-      def_column_accessor(*@columns)
-      @str_columns = nil
-      @columns
-    end
-
-    # Returns the columns as a list of frozen strings.
-    def self.str_columns
-      @str_columns ||= columns.map{|c| c.to_s.freeze}
-    end
-
-    # Sets the dataset associated with the Model class.
-    def self.set_dataset(ds)
-      @db = ds.db
-      @dataset = ds
-      @dataset.set_model(self)
-      @dataset.extend(Associations::EagerLoading)
-      @dataset.transform(@transform) if @transform
-      begin
-        @columns = nil
-        columns
-      rescue StandardError
-      end
-    end
-    class << self; alias :dataset= :set_dataset; end
-
-    class << self
-      private
-        def def_column_accessor(*columns)
-          Thread.exclusive do
-            columns.each do |column|
-              im = instance_methods
-              meth = "#{column}="
-              define_method(column){self[column]} unless im.include?(column.to_s)
-              unless im.include?(meth)
-                define_method(meth) do |*v|
-                  len = v.length
-                  raise(ArgumentError, "wrong number of arguments (#{len} for 1)") unless len == 1
-                  self[column] = v.first 
-                end
-              end
-            end
-          end
-        end
-    end
-    
-    # Returns the database assoiated with the object's Model class.
-    def db
-      @db ||= model.db
-    end
-
-    # Returns the dataset assoiated with the object's Model class.
-    #
-    # See Dataset for more information.
-    def dataset
-      model.dataset
-    end
-    
-    # Returns the columns associated with the object's Model class.
-    def columns
-      model.columns
-    end
-
-    # Returns the str_columns associated with the object's Model class.
-    def str_columns
-      model.str_columns
-    end
-
-    # Serializes column with YAML or through marshalling.
-    def self.serialize(*columns)
-      format = columns.pop[:format] if Hash === columns.last
-      format ||= :yaml
-      
-      @transform = columns.inject({}) do |m, c|
-        m[c] = format
-        m
-      end
-      @dataset.transform(@transform) if @dataset
-    end
-  end
-
-  # Lets you create a Model class with its table name already set or reopen
-  # an existing Model.
-  #
-  # Makes given dataset inherited.
-  #
-  # === Example:
-  #   class Comment < Sequel::Model(:something)
-  #     table_name # => :something
-  #
-  #     # ...
-  #
-  #   end
-  @models = {}
-  def self.Model(source)
-    return @models[source] if @models[source]
-    klass = Class.new(Sequel::Model)
-    klass.set_dataset(source.is_a?(Dataset) ? source : Model.db[source])
-    @models[source] = klass
-  end
-end
diff -pruN 1.5.1-1/lib/sequel_model/caching.rb 3.0.0-1/lib/sequel_model/caching.rb
--- 1.5.1-1/lib/sequel_model/caching.rb	2008-04-28 23:27:06.000000000 +0100
+++ 3.0.0-1/lib/sequel_model/caching.rb	1970-01-01 01:00:00.000000000 +0100
@@ -1,42 +0,0 @@
-module Sequel
-  class Model
-    def self.set_cache(store, opts = {})
-      @cache_store = store
-      if (ttl = opts[:ttl])
-        set_cache_ttl(ttl)
-      end
-      
-      meta_def(:[]) do |*args|
-        if (args.size == 1) && (Hash === (h = args.first))
-          return dataset[h]
-        end
-        
-        unless obj = @cache_store.get(cache_key_from_values(args))
-          obj = dataset[primary_key_hash((args.size == 1) ? args.first : args)]
-          @cache_store.set(cache_key_from_values(args), obj, cache_ttl)
-        end
-        obj
-      end
-      
-      class_def(:update_values) {|v| store.delete(cache_key); super}
-      class_def(:save) {store.delete(cache_key) unless new?; super}
-      class_def(:delete) {store.delete(cache_key); super}
-    end
-    
-    def self.set_cache_ttl(ttl)
-      @cache_ttl = ttl
-    end
-    
-    def self.cache_store
-      @cache_store
-    end
-    
-    def self.cache_ttl
-      @cache_ttl ||= 3600
-    end
-    
-    def self.cache_key_from_values(values)
-      "#{self}:#{values.join(',')}"
-    end
-  end
-end
diff -pruN 1.5.1-1/lib/sequel_model/deprecated.rb 3.0.0-1/lib/sequel_model/deprecated.rb
--- 1.5.1-1/lib/sequel_model/deprecated.rb	2008-04-24 22:14:31.000000000 +0100
+++ 3.0.0-1/lib/sequel_model/deprecated.rb	1970-01-01 01:00:00.000000000 +0100
@@ -1,81 +0,0 @@
-module Sequel
-  class Model
-    include Sequel::Deprecation
-    extend Sequel::Deprecation
-
-    # Check the Model.associate method to remove the :from option
-
-    def self.is_dataset_magic_method?(m) #:nodoc:
-      method_name = m.to_s
-      Sequel::Dataset::MAGIC_METHODS.each_key do |r|
-        return true if method_name =~ r
-      end
-      false
-    end
-
-    def self.method_missing(m, *args, &block) #:nodoc:
-      Thread.exclusive do
-        if dataset.respond_to?(m) || is_dataset_magic_method?(m)
-          instance_eval("def #{m}(*args, &block); deprecate('Sequel::Model.method_missing', 'Please define Sequel::Model.#{m} or use def_dataset_method :#{m}'); dataset.#{m}(*args, &block); end")
-        end
-      end
-      respond_to?(m) ? send(m, *args, &block) : super(m, *args)
-    end
-
-    def method_missing(m, *args, &block) #:nodoc:
-      if m.to_s =~ /=\z/
-        attribute = m.to_s.chop
-        values.keys.each do |k|
-          next unless k.to_s == attribute
-          deprecate("Sequel::Model#method_missing", "Use model[:#{attribute}] = ...")
-          return self[attribute.to_sym] = args.first
-        end
-        super
-      else
-        attribute = m.to_s
-        values.keys.each do |k|
-          next unless k.to_s == attribute
-          deprecate("Sequel::Model#method_missing", "Use model[:#{attribute}]")
-          return self[attribute.to_sym]
-        end
-        super
-      end
-    end
-
-    def self.create_with_params(params) #:nodoc:
-      deprecate("Sequel::Model.create_with_params", "Use .create")
-      create(params)
-    end
-
-    def self.create_with(params) #:nodoc:
-      deprecate("Sequel::Model.create_with", "Use .create")
-      create(params)
-    end
-
-    def update_with(params) #:nodoc:
-      deprecate("Sequel::Model#update_with", "Use #update_with_params")
-      update_with_params(params)
-    end
-
-    def new_record? #:nodoc:
-      deprecate("Sequel::Model#new_record?", "Use #new?")
-      new?
-    end
-
-    def set(values) #:nodoc:
-      deprecate("Sequel::Model#set", "Use #update_values")
-      update_values(values)
-    end
-
-    def update(values) #:nodoc:
-      deprecate("Sequel::Model#update", "Use #update_values")
-      update_values(values)
-    end
-
-    # deprecated, please use many_to_one instead
-    def self.one_to_one(*args, &block) #:nodoc:
-      deprecate("Sequel::Model.one_to_one", "Use many_to_one")
-      many_to_one(*args, &block)
-    end
-  end
-end
diff -pruN 1.5.1-1/lib/sequel_model/eager_loading.rb 3.0.0-1/lib/sequel_model/eager_loading.rb
--- 1.5.1-1/lib/sequel_model/eager_loading.rb	2008-04-30 03:44:46.000000000 +0100
+++ 3.0.0-1/lib/sequel_model/eager_loading.rb	1970-01-01 01:00:00.000000000 +0100
@@ -1,429 +0,0 @@
-# Eager loading makes it so that you can load all associated records for a
-# set of objects in a single query, instead of a separate query for each object.
-#
-# Two separate implementations are provided.  .eager should be used most of the
-# time, as it loads associated records using one query per association.  However,
-# it does not allow you the ability to filter based on columns in associated tables.  .eager_graph loads
-# all records in one query.  Using .eager_graph you can filter based on columns in associated
-# tables.  However, .eager_graph can be much slower than .eager, especially if multiple
-# *_to_many associations are joined.
-#
-# You can cascade the eager loading (loading associations' associations)
-# with no limit to the depth of the cascades.  You do this by passing a hash to .eager or .eager_graph
-# with the keys being associations of the current model and values being
-# associations of the model associated with the current model via the key.
-#  
-# You cannot eagerly load an association with a block argument, as the block argument is
-# evaluated in terms of a specific instance of the model, and no specific instance exists.
-#
-# The arguments can be symbols or hashes with symbol keys (for cascaded
-# eager loading). Examples:
-#
-#  Album.eager(:artist).all
-#  Album.eager_graph(:artist).all
-#  Album.eager(:artist, :genre).all
-#  Album.eager_graph(:artist, :genre).all
-#  Album.eager(:artist).eager(:genre).all
-#  Album.eager_graph(:artist).eager(:genre).all
-#  Artist.eager(:albums=>:tracks).all
-#  Artist.eager_graph(:albums=>:tracks).all
-#  Artist.eager(:albums=>{:tracks=>:genre}).all
-#  Artist.eager_graph(:albums=>{:tracks=>:genre}).all
-module Sequel::Model::Associations::EagerLoading
-  # Add the .eager! and .eager_graph! mutation methods to the dataset.
-  def self.extended(obj)
-    obj.def_mutation_method(:eager, :eager_graph)
-  end
-
-  # The preferred eager loading method.  Loads all associated records using one
-  # query for each association.
-  #
-  # The basic idea for how it works is that the dataset is first loaded normally.
-  # Then it goes through all associations that have been specified via .eager.
-  # It loads each of those associations separately, then associates them back
-  # to the original dataset via primary/foreign keys.  Due to the necessity of
-  # all objects being present, you need to use .all to use eager loading, as it
-  # can't work with .each.
-  #
-  # This implementation avoids the complexity of extracting an object graph out
-  # of a single dataset, by building the object graph out of multiple datasets,
-  # one for each association.  By using a separate dataset for each association,
-  # it avoids problems such as aliasing conflicts and creating cartesian product
-  # result sets if multiple *_to_many eager associations are requested.
-  #
-  # One limitation of using this method is that you cannot filter the dataset
-  # based on values of columns in an associated table, since the associations are loaded
-  # in separate queries.  To do that you need to load all associations in the
-  # same query, and extract an object graph from the results of that query. If you
-  # need to filter based on columns in associated tables, look at .eager_graph
-  # or join the tables you need to filter on manually. 
-  #
-  # Each association's order, if definied, is respected. Eager also works
-  # on a limited dataset.
-  def eager(*associations)
-    model = check_model
-    opt = @opts[:eager]
-    opt = opt ? opt.dup : {}
-    associations.flatten.each do |association|
-      case association
-        when Symbol
-          check_association(model, association)
-          opt[association] = nil
-        when Hash
-          association.keys.each{|assoc| check_association(model, assoc)}
-          opt.merge!(association)
-        else raise(ArgumentError, 'Associations must be in the form of a symbol or hash')
-      end
-    end
-    clone(:eager=>opt)
-  end
-
-  # The secondary eager loading method.  Loads all associations in a single query. This
-  # method should only be used if you need to filter based on columns in associated tables.
-  #
-  # This method builds an object graph using the .graph method.  Then it uses the graph
-  # to build the associations, and finally replaces the graph with a simple array
-  # of model objects.
-  #
-  # Be very careful when using this with multiple *_to_many associations, as you can
-  # create large cartesian products.  If you must graph multiple *_to_many associations,
-  # make sure your filters are specific if you have a large database.
-  # 
-  # This does not respect each association's order, as all associations are loaded in
-  # a single query.  If you want to order the results, you must manually call .order.
-  #
-  # eager_graph probably won't work the way you suspect with limit, unless you are
-  # only graphing many_to_one associations.
-  def eager_graph(*associations)
-    model = check_model
-    table_name = model.table_name
-    ds = if @opts[:eager_graph]
-      self
-    else
-      # Each of the following have a symbol key for the table alias, with the following values: 
-      # :requirements - array of requirements for this association
-      # :alias_association_type_map - the type of association for this association
-      # :alias_association_name_map - the name of the association for this association
-      clone(:eager_graph=>{:requirements=>{}, :master=>model.table_name, :alias_association_type_map=>{}, :alias_association_name_map=>{}, :reciprocals=>{}})
-    end
-    ds.eager_graph_associations(ds, model, table_name, [], *associations)
-  end
-  
-  protected
-    # Call graph on the association with the correct arguments,
-    # update the eager_graph data structure, and recurse into
-    # eager_graph_associations if there are any passed in associations
-    # (which would be dependencies of the current association)
-    #
-    # Arguments:
-    # * ds - Current dataset
-    # * model - Current Model
-    # * ta - table_alias used for the parent association
-    # * requirements - an array, used as a stack for requirements
-    # * r - association reflection for the current association
-    # * *associations - any associations dependent on this one
-    def eager_graph_association(ds, model, ta, requirements, r, *associations)
-      klass = model.send(:associated_class, r)
-      assoc_name = r[:name]
-      assoc_table_alias = ds.eager_unique_table_alias(ds, assoc_name)
-      ds = case assoc_type = r[:type]
-      when :many_to_one
-        ds.graph(klass, {klass.primary_key=>:"#{ta}__#{r[:key]}"}, :table_alias=>assoc_table_alias)
-      when :one_to_many
-        ds = ds.graph(klass, {r[:key]=>:"#{ta}__#{model.primary_key}"}, :table_alias=>assoc_table_alias)
-        # We only load reciprocals for one_to_many associations, as other reciprocals don't make sense
-        ds.opts[:eager_graph][:reciprocals][assoc_table_alias] = model.send(:reciprocal_association, r)
-        ds
-      when :many_to_many
-        ds = ds.graph(r[:join_table], {r[:left_key]=>:"#{ta}__#{model.primary_key}"}, :select=>false, :table_alias=>ds.eager_unique_table_alias(ds, r[:join_table]))
-        ds.graph(klass, {klass.primary_key=>r[:right_key]}, :table_alias=>assoc_table_alias)
-      end
-      eager_graph = ds.opts[:eager_graph]
-      eager_graph[:requirements][assoc_table_alias] = requirements.dup
-      eager_graph[:alias_association_name_map][assoc_table_alias] = assoc_name
-      eager_graph[:alias_association_type_map][assoc_table_alias] = assoc_type
-      ds = ds.eager_graph_associations(ds, klass, assoc_table_alias, requirements + [assoc_table_alias], *associations) unless associations.empty?
-      ds
-    end
-  
-    # Check the associations are valid for the given model.
-    # Call eager_graph_association on each association.
-    #
-    # Arguments:
-    # * ds - Current dataset
-    # * model - Current Model
-    # * ta - table_alias used for the parent association
-    # * requirements - an array, used as a stack for requirements
-    # * *associations - the associations to add to the graph
-    def eager_graph_associations(ds, model, ta, requirements, *associations)
-      return ds if associations.empty?
-      associations.flatten.each do |association|
-        ds = case association
-        when Symbol
-          ds.eager_graph_association(ds, model, ta, requirements, check_association(model, association))
-        when Hash
-          association.each do |assoc, assoc_assocs|
-            ds = ds.eager_graph_association(ds, model, ta, requirements, check_association(model, assoc), assoc_assocs)
-          end
-          ds
-        else raise(ArgumentError, 'Associations must be in the form of a symbol or hash')
-        end
-      end
-      ds
-    end
-
-    # Build associations out of the array of returned object graphs.
-    def eager_graph_build_associations(record_graphs)
-      # Dup the tables that will be used, so that self is not modified.
-      eager_graph = @opts[:eager_graph]
-      master = eager_graph[:master]
-      requirements = eager_graph[:requirements]
-      alias_map = eager_graph[:alias_association_name_map]
-      type_map = eager_graph[:alias_association_type_map]
-      reciprocal_map = eager_graph[:reciprocals]
-
-      # Make dependency map hash out of requirements array for each association.
-      # This builds a tree of dependencies that will be used for recursion
-      # to ensure that all parts of the object graph are loaded into the
-      # appropriate subordinate association.
-      dependency_map = {}
-      # Sort the associations be requirements length, so that
-      # requirements are added to the dependency hash before their
-      # dependencies.
-      requirements.sort_by{|a| a[1].length}.each do |ta, deps|
-        if deps.empty?
-          dependency_map[ta] = {}
-        else
-          deps = deps.dup
-          hash = dependency_map[deps.shift]
-          deps.each do |dep|
-            hash = hash[dep]
-          end
-          hash[ta] = {}
-        end
-      end
-
-      # This mapping is used to make sure that duplicate entries in the
-      # result set are mapped to a single record.  For example, using a
-      # single one_to_many association with 10 associated records,
-      # the main object will appear in the object graph 10 times.
-      # We map by primary key, if available, or by the object's entire values,
-      # if not. The mapping must be per table, so create sub maps for each table
-      # alias.
-      records_map = {master=>{}}
-      alias_map.keys.each{|ta| records_map[ta] = {}}
-
-      # This will hold the final record set that we will be replacing the object graph with.
-      records = []
-      record_graphs.each do |record_graph|
-        primary_record = record_graph[master]
-        key = primary_record.pk || primary_record.values.sort_by{|x| x[0].to_s}
-        if cached_pr = records_map[master][key]
-          primary_record = cached_pr
-        else
-          records_map[master][key] = primary_record
-          # Only add it to the list of records to return if it is a new record
-          records.push(primary_record)
-        end
-        # Build all associations for the current object and it's dependencies
-        eager_graph_build_associations_graph(dependency_map, alias_map, type_map, reciprocal_map, records_map, primary_record, record_graph)
-      end
-
-      # Remove duplicate records from all associations if this graph could possibly be a cartesian product
-      eager_graph_make_associations_unique(records, dependency_map, alias_map, type_map) if type_map.reject{|k,v| v == :many_to_one}.length > 1
-      
-      # Replace the array of object graphs with an array of model objects
-      record_graphs.replace(records)
-    end
-
-    # Creates a unique table alias that hasn't already been used in the query.
-    # Will either be the table_alias itself or table_alias_N for some integer
-    # N (starting at 0 and increasing until an unused one is found).
-    def eager_unique_table_alias(ds, table_alias)
-      if (graph = ds.opts[:graph]) && (table_aliases = graph[:table_aliases]) && (table_aliases.include?(table_alias))
-        i = 0
-        loop do
-          ta = :"#{table_alias}_#{i}"
-          return ta unless table_aliases[ta]
-          i += 1
-        end
-      else
-        table_alias
-      end
-    end
-  
-  private
-    # Make sure a standard (non-polymorphic model) is used for this dataset, and return the model
-    def check_model
-      raise(ArgumentError, 'No model for this dataset') unless @opts[:models] && model = @opts[:models][nil]
-      model
-    end
-
-    # Make sure the association is valid for this model, and return the association's reflection
-    def check_association(model, association)
-      raise(ArgumentError, 'Invalid association') unless reflection = model.association_reflection(association)
-      raise(ArgumentError, 'Cannot eagerly load associations with block arguments') if reflection[:block]
-      reflection
-    end
-  
-    # Build associations for the current object.  This is called recursively
-    # to build object's dependencies.
-    def eager_graph_build_associations_graph(dependency_map, alias_map, type_map, reciprocal_map, records_map, current, record_graph)
-      return if dependency_map.empty?
-      # Don't clobber the instance variable array for *_to_many associations if it has already been setup
-      dependency_map.keys.each do |ta|
-        current.instance_variable_set("@#{alias_map[ta]}", type_map[ta] == :many_to_one ? :null : []) unless current.instance_variable_get("@#{alias_map[ta]}")
-      end
-      dependency_map.each do |ta, deps|
-        next unless rec = record_graph[ta]
-        key = rec.pk || rec.values.sort_by{|x| x[0].to_s}
-        if cached_rec = records_map[ta][key]
-          rec = cached_rec
-        else
-          records_map[ta][rec.pk] = rec
-        end
-        ivar = "@#{alias_map[ta]}"
-        case assoc_type = type_map[ta]
-        when :many_to_one
-          current.instance_variable_set(ivar, rec)
-        else
-          list = current.instance_variable_get(ivar)
-          list.push(rec) 
-          if (assoc_type == :one_to_many) && (reciprocal = reciprocal_map[ta])
-            rec.instance_variable_set(reciprocal, current)
-          end
-        end
-        # Recurse into dependencies of the current object
-        eager_graph_build_associations_graph(deps, alias_map, type_map, reciprocal_map, records_map, rec, record_graph)
-      end
-    end
-
-    # If the result set is the result of a cartesian product, then it is possible that
-    # there a multiple records for each association when there should only be one.
-    def eager_graph_make_associations_unique(records, dependency_map, alias_map, type_map)
-      records.each do |record|
-        dependency_map.each do |ta, deps|
-          list = if type_map[ta] == :many_to_one
-            item = record.send(alias_map[ta])
-            [item] if item
-          else
-            list = record.send(alias_map[ta])
-            list.uniq!
-            # Recurse into dependencies
-            list.each{|rec| eager_graph_make_associations_unique(rec, deps, alias_map, type_map)}
-          end
-        end
-      end
-    end
-
-    # Eagerly load all specified associations 
-    def eager_load(a)
-      return if a.empty?
-      # Current model class
-      model = @opts[:models][nil]
-      # All associations to eager load
-      eager_assoc = @opts[:eager]
-      # Key is foreign/primary key name symbol
-      # Value is hash with keys being foreign/primary key values (generally integers)
-      #  and values being an array of current model objects with that
-      #  specific foreign/primary key
-      key_hash = {}
-      # array of attribute_values keys to monitor
-      keys = []
-      # Reflections for all associations to eager load
-      reflections = eager_assoc.keys.collect{|assoc| model.association_reflection(assoc)}
-
-      # Populate keys to monitor
-      reflections.each do |reflection|
-        key = reflection[:type] == :many_to_one ? reflection[:key] : model.primary_key
-        next if key_hash[key]
-        key_hash[key] = {}
-        keys << key
-      end
-      
-      # Associate each object with every key being monitored
-      a.each do |r|
-        keys.each do |key|
-          ((key_hash[key][r[key]] ||= []) << r) if r[key]
-        end
-      end
-      
-      # Iterate through eager associations and assign instance variables
-      # for the association for all model objects
-      reflections.each do |reflection|
-        assoc_class = model.send(:associated_class, reflection)
-        assoc_name = reflection[:name]
-        # Proc for setting cascaded eager loading
-        cascade = Proc.new do |d|
-          if c = eager_assoc[assoc_name]
-            d = d.eager(c)
-          end
-          if c = reflection[:eager]
-            d = d.eager(c)
-          end
-          d
-        end
-        case rtype = reflection[:type]
-          when :many_to_one
-            key = reflection[:key]
-            h = key_hash[key]
-            keys = h.keys
-            # No records have the foreign key set for this association, so skip it
-            next unless keys.length > 0
-            ds = assoc_class.filter(assoc_class.primary_key=>keys)
-            ds = cascade.call(ds)
-            ds.all do |assoc_object|
-              h[assoc_object.pk].each do |object|
-                object.instance_variable_set(:"@#{assoc_name}", assoc_object)
-              end
-            end
-          when :one_to_many, :many_to_many
-            if rtype == :one_to_many
-              fkey = key = reflection[:key]
-              h = key_hash[model.primary_key]
-              reciprocal = model.send(:reciprocal_association, reflection)
-              ds = assoc_class.filter(key=>h.keys)
-            else
-              assoc_table = assoc_class.table_name
-              left = reflection[:left_key]
-              right = reflection[:right_key]
-              right_pk = (reflection[:right_primary_key] || :"#{assoc_table}__#{assoc_class.primary_key}")
-              join_table = reflection[:join_table]
-              fkey = (reflection[:left_key_alias] ||= :"x_foreign_key_x")
-              table_selection = (reflection[:select] ||= assoc_table.*)
-              key_selection = (reflection[:left_key_select] ||= :"#{join_table}__#{left}___#{fkey}")
-              h = key_hash[model.primary_key]
-              ds = assoc_class.select(table_selection, key_selection).inner_join(join_table, right=>right_pk, left=>h.keys)
-            end
-            if order = reflection[:order]
-              ds = ds.order(order)
-            end
-            ds = cascade.call(ds)
-            ivar = :"@#{assoc_name}"
-            h.values.each do |object_array|
-              object_array.each do |object|
-                object.instance_variable_set(ivar, [])
-              end
-            end
-            ds.all do |assoc_object|
-              fk = if rtype == :many_to_many
-                assoc_object.values.delete(fkey)
-              else
-                assoc_object[fkey]
-              end
-              h[fk].each do |object|
-                object.instance_variable_get(ivar) << assoc_object
-                assoc_object.instance_variable_set(reciprocal, object) if reciprocal
-              end
-            end
-        end
-      end
-    end
-
-    # Build associations from the graph if .eager_graph was used, 
-    # and/or load other associations if .eager was used.
-    def post_load(all_records)
-      eager_graph_build_associations(all_records) if @opts[:eager_graph]
-      eager_load(all_records) if @opts[:eager]
-    end
-end
diff -pruN 1.5.1-1/lib/sequel_model/hooks.rb 3.0.0-1/lib/sequel_model/hooks.rb
--- 1.5.1-1/lib/sequel_model/hooks.rb	2008-04-22 20:16:00.000000000 +0100
+++ 3.0.0-1/lib/sequel_model/hooks.rb	1970-01-01 01:00:00.000000000 +0100
@@ -1,59 +0,0 @@
-module Sequel
-  class Model
-    HOOKS = [
-      :after_initialize,
-      :before_create,
-      :after_create,
-      :before_update,
-      :after_update,
-      :before_save,
-      :after_save,
-      :before_destroy,
-      :after_destroy
-    ]
-    
-    # Some fancy code generation here in order to define the hook class methods...
-    HOOK_METHOD_STR = %Q{
-      def self.%s(method = nil, &block)
-        unless block
-          (raise SequelError, 'No hook method specified') unless method
-          block = proc {send method}
-        end
-        add_hook(%s, &block)
-      end
-    }
-    
-    class << self
-      # Returns true if the model class or any of its ancestors have defined
-      # hooks for the given hook key. Notice that this method cannot detect 
-      # hooks defined using overridden methods.
-      def has_hooks?(key)
-        has = hooks[key] && !hooks[key].empty?
-        has || ((self != Model) && superclass.has_hooks?(key))
-      end
-      private
-        def def_hook_method(m) #:nodoc:
-          instance_eval(HOOK_METHOD_STR % [m.to_s, m.inspect])
-        end
-        
-        # Returns the hooks hash for the model class.
-        def hooks #:nodoc:
-          @hooks ||= Hash.new {|h, k| h[k] = []}
-        end
-        
-        def add_hook(hook, &block) #:nodoc:
-          chain = hooks[hook]
-          chain << block
-          define_method(hook) do 
-            return false if super == false
-            chain.each {|h| break false if instance_eval(&h) == false}
-          end
-        end
-    end
-
-    HOOKS.each do |h|
-      define_method(h){}
-      def_hook_method(h)
-    end
-  end
-end
diff -pruN 1.5.1-1/lib/sequel_model/inflections.rb 3.0.0-1/lib/sequel_model/inflections.rb
--- 1.5.1-1/lib/sequel_model/inflections.rb	2008-04-22 20:16:00.000000000 +0100
+++ 3.0.0-1/lib/sequel_model/inflections.rb	1970-01-01 01:00:00.000000000 +0100
@@ -1,112 +0,0 @@
-Inflector.inflections do |inflect|
-  inflect.plural(/$/, 's')
-  inflect.plural(/s$/i, 's')
-  inflect.plural(/(ax|test)is$/i, '\1es')
-  inflect.plural(/(octop|vir)us$/i, '\1i')
-  inflect.plural(/(alias|status)$/i, '\1es')
-  inflect.plural(/(bu)s$/i, '\1ses')
-  inflect.plural(/(buffal|tomat)o$/i, '\1oes')
-  inflect.plural(/([ti])um$/i, '\1a')
-  inflect.plural(/sis$/i, 'ses')
-  inflect.plural(/(?:([^f])fe|([lr])f)$/i, '\1\2ves')
-  inflect.plural(/(hive)$/i, '\1s')
-  inflect.plural(/([^aeiouy]|qu)y$/i, '\1ies')
-  inflect.plural(/(x|ch|ss|sh)$/i, '\1es')
-  inflect.plural(/(matr|vert|ind)ix|ex$/i, '\1ices')
-  inflect.plural(/([m|l])ouse$/i, '\1ice')
-  inflect.plural(/^(ox)$/i, '\1en')
-  inflect.plural(/(quiz)$/i, '\1zes')
-
-  inflect.singular(/s$/i, '')
-  inflect.singular(/(n)ews$/i, '\1ews')
-  inflect.singular(/([ti])a$/i, '\1um')
-  inflect.singular(/((a)naly|(b)a|(d)iagno|(p)arenthe|(p)rogno|(s)ynop|(t)he)ses$/i, '\1\2sis')
-  inflect.singular(/(^analy)ses$/i, '\1sis')
-  inflect.singular(/([^f])ves$/i, '\1fe')
-  inflect.singular(/(hive)s$/i, '\1')
-  inflect.singular(/(tive)s$/i, '\1')
-  inflect.singular(/([lr])ves$/i, '\1f')
-  inflect.singular(/([^aeiouy]|qu)ies$/i, '\1y')
-  inflect.singular(/(s)eries$/i, '\1eries')
-  inflect.singular(/(m)ovies$/i, '\1ovie')
-  inflect.singular(/(x|ch|ss|sh)es$/i, '\1')
-  inflect.singular(/([m|l])ice$/i, '\1ouse')
-  inflect.singular(/(bus)es$/i, '\1')
-  inflect.singular(/(o)es$/i, '\1')
-  inflect.singular(/(shoe)s$/i, '\1')
-  inflect.singular(/(cris|ax|test)es$/i, '\1is')
-  inflect.singular(/(octop|vir)i$/i, '\1us')
-  inflect.singular(/(alias|status)es$/i, '\1')
-  inflect.singular(/^(ox)en/i, '\1')
-  inflect.singular(/(vert|ind)ices$/i, '\1ex')
-  inflect.singular(/(matr)ices$/i, '\1ix')
-  inflect.singular(/(quiz)zes$/i, '\1')
-
-  inflect.irregular('person', 'people')
-  inflect.irregular('man', 'men')
-  inflect.irregular('child', 'children')
-  inflect.irregular('sex', 'sexes')
-  inflect.irregular('move', 'moves')
-
-  inflect.uncountable(%w(equipment information rice money species series fish sheep))
-end
-
-module Inflections #:nodoc:
-
-  def pluralize
-    Inflector.pluralize(self)
-  end
-
-  def singularize
-    Inflector.singularize(self)
-  end
-
-  def camelize(first_letter = :upper)
-    case first_letter
-      when :upper then Inflector.camelize(self, true)
-      when :lower then Inflector.camelize(self, false)
-    end
-  end
-  alias_method :camelcase, :camelize
-
-  def titleize
-    Inflector.titleize(self)
-  end
-  alias_method :titlecase, :titleize
-
-  def underscore
-    Inflector.underscore(self)
-  end
-
-  def dasherize
-    Inflector.dasherize(self)
-  end
-
-  def demodulize
-    Inflector.demodulize(self)
-  end
-
-  def tableize
-    Inflector.tableize(self)
-  end
-
-  def classify
-    Inflector.classify(self)
-  end
-
-  def humanize
-    Inflector.humanize(self)
-  end
-
-  def foreign_key(separate_class_name_and_id_with_underscore = true)
-    Inflector.foreign_key(self, separate_class_name_and_id_with_underscore)
-  end
-
-  def constantize
-    Inflector.constantize(self)
-  end
-end
-
-class String
-  include Inflections
-end
\ No newline at end of file
diff -pruN 1.5.1-1/lib/sequel_model/inflector.rb 3.0.0-1/lib/sequel_model/inflector.rb
--- 1.5.1-1/lib/sequel_model/inflector.rb	2008-04-22 20:16:00.000000000 +0100
+++ 3.0.0-1/lib/sequel_model/inflector.rb	1970-01-01 01:00:00.000000000 +0100
@@ -1,279 +0,0 @@
-require 'singleton'
-
-# The Inflector transforms words from singular to plural, class names to table names, modularized class names to ones without,
-# and class names to foreign keys. The default inflections for pluralization, singularization, and uncountable words are kept
-# in inflections.rb.
-module Inflector
-  # A singleton instance of this class is yielded by Inflector.inflections, which can then be used to specify additional
-  # inflection rules. Examples:
-  #
-  #   Inflector.inflections do |inflect|
-  #     inflect.plural /^(ox)$/i, '\1\2en'
-  #     inflect.singular /^(ox)en/i, '\1'
-  #
-  #     inflect.irregular 'octopus', 'octopi'
-  #
-  #     inflect.uncountable "equipment"
-  #   end
-  #
-  # New rules are added at the top. So in the example above, the irregular rule for octopus will now be the first of the
-  # pluralization and singularization rules that is runs. This guarantees that your rules run before any of the rules that may
-  # already have been loaded.
-  class Inflections
-    include Singleton
-
-    attr_reader :plurals, :singulars, :uncountables
-
-    def initialize
-      @plurals, @singulars, @uncountables = [], [], []
-    end
-
-    # Specifies a new pluralization rule and its replacement. The rule can either be a string or a regular expression.
-    # The replacement should always be a string that may include references to the matched data from the rule.
-    def plural(rule, replacement)
-      @plurals.insert(0, [rule, replacement])
-    end
-
-    # Specifies a new singularization rule and its replacement. The rule can either be a string or a regular expression.
-    # The replacement should always be a string that may include references to the matched data from the rule.
-    def singular(rule, replacement)
-      @singulars.insert(0, [rule, replacement])
-    end
-
-    # Specifies a new irregular that applies to both pluralization and singularization at the same time. This can only be used
-    # for strings, not regular expressions. You simply pass the irregular in singular and plural form.
-    #
-    # Examples:
-    #   irregular 'octopus', 'octopi'
-    #   irregular 'person', 'people'
-    def irregular(singular, plural)
-      plural(Regexp.new("(#{singular[0,1]})#{singular[1..-1]}$", "i"), '\1' + plural[1..-1])
-      singular(Regexp.new("(#{plural[0,1]})#{plural[1..-1]}$", "i"), '\1' + singular[1..-1])
-    end
-
-    # Add uncountable words that shouldn't be attempted inflected.
-    #
-    # Examples:
-    #   uncountable "money"
-    #   uncountable "money", "information"
-    #   uncountable %w( money information rice )
-    def uncountable(*words)
-      (@uncountables << words).flatten!
-    end
-
-    # Clears the loaded inflections within a given scope (default is :all). Give the scope as a symbol of the inflection type,
-    # the options are: :plurals, :singulars, :uncountables
-    #
-    # Examples:
-    #   clear :all
-    #   clear :plurals
-    def clear(scope = :all)
-      case scope
-        when :all
-          @plurals, @singulars, @uncountables = [], [], []
-        else
-          instance_variable_set "@#{scope}", []
-      end
-    end
-  end
-
-  extend self
-
-  def inflections
-    if block_given?
-      yield Inflections.instance
-    else
-      Inflections.instance
-    end
-  end
-
-  # Returns the plural form of the word in the string.
-  #
-  # Examples
-  #   "post".pluralize #=> "posts"
-  #   "octopus".pluralize #=> "octopi"
-  #   "sheep".pluralize #=> "sheep"
-  #   "words".pluralize #=> "words"
-  #   "the blue mailman".pluralize #=> "the blue mailmen"
-  #   "CamelOctopus".pluralize #=> "CamelOctopi"
-  def pluralize(word)
-    result = word.to_s.dup
-
-    if inflections.uncountables.include?(result.downcase)
-      result
-    else
-      inflections.plurals.each { |(rule, replacement)| break if result.gsub!(rule, replacement) }
-      result
-    end
-  end
-
-  # The reverse of pluralize, returns the singular form of a word in a string.
-  #
-  # Examples
-  #   "posts".singularize #=> "post"
-  #   "octopi".singularize #=> "octopus"
-  #   "sheep".singluarize #=> "sheep"
-  #   "word".singluarize #=> "word"
-  #   "the blue mailmen".singularize #=> "the blue mailman"
-  #   "CamelOctopi".singularize #=> "CamelOctopus"
-  def singularize(word)
-    result = word.to_s.dup
-
-    if inflections.uncountables.include?(result.downcase)
-      result
-    else
-      inflections.singulars.each { |(rule, replacement)| break if result.gsub!(rule, replacement) }
-      result
-    end
-  end
-
-  # By default, camelize converts strings to UpperCamelCase. If the argument to camelize
-  # is set to ":lower" then camelize produces lowerCamelCase.
-  #
-  # camelize will also convert '/' to '::' which is useful for converting paths to namespaces
-  #
-  # Examples
-  #   "active_record".camelize #=> "ActiveRecord"
-  #   "active_record".camelize(:lower) #=> "activeRecord"
-  #   "active_record/errors".camelize #=> "ActiveRecord::Errors"
-  #   "active_record/errors".camelize(:lower) #=> "activeRecord::Errors"
-  def camelize(lower_case_and_underscored_word, first_letter_in_uppercase = true)
-    if first_letter_in_uppercase
-      lower_case_and_underscored_word.to_s.gsub(/\/(.?)/) { "::" + $1.upcase }.gsub(/(^|_)(.)/) { $2.upcase }
-    else
-      lower_case_and_underscored_word.first + camelize(lower_case_and_underscored_word)[1..-1]
-    end
-  end
-
-  # Capitalizes all the words and replaces some characters in the string to create
-  # a nicer looking title. Titleize is meant for creating pretty output. It is not
-  # used in the Rails internals.
-  #
-  # titleize is also aliased as as titlecase
-  #
-  # Examples
-  #   "man from the boondocks".titleize #=> "Man From The Boondocks"
-  #   "x-men: the last stand".titleize #=> "X Men: The Last Stand"
-  def titleize(word)
-    humanize(underscore(word)).gsub(/\b([a-z])/) { $1.capitalize }
-  end
-
-  # The reverse of +camelize+. Makes an underscored form from the expression in the string.
-  #
-  # Changes '::' to '/' to convert namespaces to paths.
-  #
-  # Examples
-  #   "ActiveRecord".underscore #=> "active_record"
-  #   "ActiveRecord::Errors".underscore #=> active_record/errors
-  def underscore(camel_cased_word)
-    camel_cased_word.to_s.gsub(/::/, '/').
-      gsub(/([A-Z]+)([A-Z][a-z])/,'\1_\2').
-      gsub(/([a-z\d])([A-Z])/,'\1_\2').
-      tr("-", "_").
-      downcase
-  end
-
-  # Replaces underscores with dashes in the string.
-  #
-  # Example
-  #   "puni_puni" #=> "puni-puni"
-  def dasherize(underscored_word)
-    underscored_word.gsub(/_/, '-')
-  end
-
-  # Capitalizes the first word and turns underscores into spaces and strips _id.
-  # Like titleize, this is meant for creating pretty output.
-  #
-  # Examples
-  #   "employee_salary" #=> "Employee salary"
-  #   "author_id" #=> "Author"
-  def humanize(lower_case_and_underscored_word)
-    lower_case_and_underscored_word.to_s.gsub(/_id$/, "").gsub(/_/, " ").capitalize
-  end
-
-  # Removes the module part from the expression in the string
-  #
-  # Examples
-  #   "ActiveRecord::CoreExtensions::String::Inflections".demodulize #=> "Inflections"
-  #   "Inflections".demodulize #=> "Inflections"
-  def demodulize(class_name_in_module)
-    class_name_in_module.to_s.gsub(/^.*::/, '')
-  end
-
-  # Create the name of a table like Rails does for models to table names. This method
-  # uses the pluralize method on the last word in the string.
-  #
-  # Examples
-  #   "RawScaledScorer".tableize #=> "raw_scaled_scorers"
-  #   "egg_and_ham".tableize #=> "egg_and_hams"
-  #   "fancyCategory".tableize #=> "fancy_categories"
-  def tableize(class_name)
-    pluralize(underscore(class_name))
-  end
-
-  # Create a class name from a table name like Rails does for table names to models.
-  # Note that this returns a string and not a Class. (To convert to an actual class
-  # follow classify with constantize.)
-  #
-  # Examples
-  #   "egg_and_hams".classify #=> "EggAndHam"
-  #   "post".classify #=> "Post"
-  def classify(table_name)
-    # strip out any leading schema name
-    camelize(singularize(table_name.to_s.sub(/.*\./, '')))
-  end
-
-  # Creates a foreign key name from a class name.
-  # +separate_class_name_and_id_with_underscore+ sets whether
-  # the method should put '_' between the name and 'id'.
-  #
-  # Examples
-  #   "Message".foreign_key #=> "message_id"
-  #   "Message".foreign_key(false) #=> "messageid"
-  #   "Admin::Post".foreign_key #=> "post_id"
-  def foreign_key(class_name, separate_class_name_and_id_with_underscore = true)
-    underscore(demodulize(class_name)) + (separate_class_name_and_id_with_underscore ? "_id" : "id")
-  end
-
-  # Constantize tries to find a declared constant with the name specified
-  # in the string. It raises a NameError when the name is not in CamelCase
-  # or is not initialized.
-  #
-  # Examples
-  #   "Module".constantize #=> Module
-  #   "Class".constantize #=> Class
-  def constantize(camel_cased_word)
-    unless /\A(?:::)?([A-Z]\w*(?:::[A-Z]\w*)*)\z/ =~ camel_cased_word
-      raise NameError, "#{camel_cased_word.inspect} is not a valid constant name!"
-    end
-
-    Object.module_eval("::#{$1}", __FILE__, __LINE__)
-  end
-
-  # Ordinalize turns a number into an ordinal string used to denote the
-  # position in an ordered sequence such as 1st, 2nd, 3rd, 4th.
-  #
-  # Examples
-  #   ordinalize(1)     # => "1st"
-  #   ordinalize(2)     # => "2nd"
-  #   ordinalize(1002)  # => "1002nd"
-  #   ordinalize(1003)  # => "1003rd"
-  def ordinalize(number)
-    if (11..13).include?(number.to_i % 100)
-      "#{number}th"
-    else
-      case number.to_i % 10
-        when 1
-          "#{number}st"
-        when 2
-          "#{number}nd"
-        when 3
-          "#{number}rd"
-        else
-          "#{number}th"
-      end
-    end
-  end
-end
-
-require File.dirname(__FILE__) + '/inflections'
\ No newline at end of file
diff -pruN 1.5.1-1/lib/sequel_model/plugins.rb 3.0.0-1/lib/sequel_model/plugins.rb
--- 1.5.1-1/lib/sequel_model/plugins.rb	2008-04-22 20:16:00.000000000 +0100
+++ 3.0.0-1/lib/sequel_model/plugins.rb	1970-01-01 01:00:00.000000000 +0100
@@ -1,48 +0,0 @@
-module Sequel
-  module Plugins; end
-  
-  class Model
-    class << self
-      # Loads a plugin for use with the model class, passing optional arguments
-      # to the plugin.
-      def is(plugin, *args)
-        m = plugin_module(plugin)
-        if m.respond_to?(:apply)
-          m.apply(self, *args)
-        end
-        if m.const_defined?("InstanceMethods")
-          class_def(:"#{plugin}_opts") {args.first}
-          include(m::InstanceMethods)
-        end
-        if m.const_defined?("ClassMethods")
-          meta_def(:"#{plugin}_opts") {args.first}
-          metaclass.send(:include, m::ClassMethods)
-        end
-        if m.const_defined?("DatasetMethods")
-          unless @dataset
-            raise Sequel::Error, "Plugin cannot be applied because the model class has no dataset"
-          end
-          dataset.meta_def(:"#{plugin}_opts") {args.first}
-          dataset.metaclass.send(:include, m::DatasetMethods)
-        end
-      end
-      alias_method :is_a, :is
-    
-      private
-        # Returns the module for the specified plugin. If the module is not 
-        # defined, the corresponding plugin gem is automatically loaded.
-        def plugin_module(plugin)
-          module_name = plugin.to_s.gsub(/(^|_)(.)/) {$2.upcase}
-          if not Sequel::Plugins.const_defined?(module_name)
-            require plugin_gem(plugin)
-          end
-          Sequel::Plugins.const_get(module_name)
-        end
-  
-        # Returns the gem name for the given plugin.
-        def plugin_gem(plugin)
-          "sequel_#{plugin}"
-        end
-    end
-  end
-end
diff -pruN 1.5.1-1/lib/sequel_model/record.rb 3.0.0-1/lib/sequel_model/record.rb
--- 1.5.1-1/lib/sequel_model/record.rb	2008-04-28 23:26:16.000000000 +0100
+++ 3.0.0-1/lib/sequel_model/record.rb	1970-01-01 01:00:00.000000000 +0100
@@ -1,348 +0,0 @@
-module Sequel
-  class Model
-    attr_reader :values
-    attr_reader :changed_columns
-    
-    # Returns value of attribute.
-    def [](column)
-      @values[column]
-    end
-    # Sets value of attribute and marks the column as changed.
-    def []=(column, value)
-      # If it is new, it doesn't have a value yet, so we should
-      # definitely set the new value.
-      # If the column isn't in @values, we can't assume it is
-      # NULL in the database, so assume it has changed.
-      if new? || !@values.include?(column) || value != @values[column]
-        @changed_columns << column unless @changed_columns.include?(column)
-        @values[column] = value
-      end
-    end
-
-    # Enumerates through all attributes.
-    #
-    # === Example:
-    #   Ticket.find(7).each { |k, v| puts "#{k} => #{v}" }
-    def each(&block)
-      @values.each(&block)
-    end
-    # Returns attribute names.
-    def keys
-      @values.keys
-    end
-
-    # Returns value for <tt>:id</tt> attribute.
-    def id
-      @values[:id]
-    end
-
-    # Compares model instances by values.
-    def ==(obj)
-      (obj.class == model) && (obj.values == @values)
-    end
-    alias_method :eql?, :"=="
-
-    # If pk is not nil, true only if the objects have the same class and pk.
-    # If pk is nil, false.
-    def ===(obj)
-      pk.nil? ? false : (obj.class == model) && (obj.pk == pk)
-    end
-
-    # Unique for objects with the same class and pk (if pk is not nil), or
-    # the same class and values (if pk is nil).
-    def hash
-      [model, pk.nil? ? @values.sort_by{|k,v| k.to_s} : pk].hash
-    end
-
-    # Returns key for primary key.
-    def self.primary_key
-      :id
-    end
-    
-    # Returns primary key attribute hash.
-    def self.primary_key_hash(value)
-      {:id => value}
-    end
-    
-    # Sets primary key, regular and composite are possible.
-    #
-    # == Example:
-    #   class Tagging < Sequel::Model
-    #     # composite key
-    #     set_primary_key :taggable_id, :tag_id
-    #   end
-    #
-    #   class Person < Sequel::Model
-    #     # regular key
-    #     set_primary_key :person_id
-    #   end
-    #
-    # <i>You can even set it to nil!</i>
-    def self.set_primary_key(*key)
-      # if k is nil, we go to no_primary_key
-      if key.empty? || (key.size == 1 && key.first == nil)
-        return no_primary_key
-      end
-      
-      # backwards compat
-      key = (key.length == 1) ? key[0] : key.flatten
-
-      # redefine primary_key
-      meta_def(:primary_key) {key}
-      
-      unless key.is_a? Array # regular primary key
-        class_def(:this) do
-          @this ||= dataset.filter(key => @values[key]).limit(1).naked
-        end
-        class_def(:pk) do
-          @pk ||= @values[key]
-        end
-        class_def(:pk_hash) do
-          @pk ||= {key => @values[key]}
-        end
-        class_def(:cache_key) do
-          pk = @values[key] || (raise Error, 'no primary key for this record')
-          @cache_key ||= "#{self.class}:#{pk}"
-        end
-        meta_def(:primary_key_hash) do |v|
-          {key => v}
-        end
-      else # composite key
-        exp_list = key.map {|k| "#{k.inspect} => @values[#{k.inspect}]"}
-        block = eval("proc {@this ||= self.class.dataset.filter(#{exp_list.join(',')}).limit(1).naked}")
-        class_def(:this, &block)
-        
-        exp_list = key.map {|k| "@values[#{k.inspect}]"}
-        block = eval("proc {@pk ||= [#{exp_list.join(',')}]}")
-        class_def(:pk, &block)
-        
-        exp_list = key.map {|k| "#{k.inspect} => @values[#{k.inspect}]"}
-        block = eval("proc {@this ||= {#{exp_list.join(',')}}}")
-        class_def(:pk_hash, &block)
-        
-        exp_list = key.map {|k| '#{@values[%s]}' % k.inspect}.join(',')
-        block = eval('proc {@cache_key ||= "#{self.class}:%s"}' % exp_list)
-        class_def(:cache_key, &block)
-
-        meta_def(:primary_key_hash) do |v|
-          key.inject({}) {|m, i| m[i] = v.shift; m}
-        end
-      end
-    end
-    
-    def self.no_primary_key #:nodoc:
-      meta_def(:primary_key) {nil}
-      meta_def(:primary_key_hash) {|v| raise Error, "#{self} does not have a primary key"}
-      class_def(:this)      {raise Error, "No primary key is associated with this model"}
-      class_def(:pk)        {raise Error, "No primary key is associated with this model"}
-      class_def(:pk_hash)   {raise Error, "No primary key is associated with this model"}
-      class_def(:cache_key) {raise Error, "No primary key is associated with this model"}
-    end
-    
-    # Creates new instance with values set to passed-in Hash, saves it
-    # (running any callbacks), and returns the instance.
-    def self.create(values = {}, &block)
-      obj = new(values, &block)
-      obj.save
-      obj
-    end
-    
-    # Updates the instance with the supplied values with support for virtual
-    # attributes, ignoring any values for which no setter method is available.
-    # Does not save the record.
-    #
-    # If no columns have been set for this model (very unlikely), assume symbol
-    # keys are valid column names, and assign the column value based on that.
-    def set_with_params(hash)
-      columns_not_set = !model.instance_variable_get(:@columns)
-      meths = setter_methods
-      hash.each do |k,v|
-        m = "#{k}="
-        if meths.include?(m)
-          send(m, v)
-        elsif columns_not_set && (Symbol === k)
-          self[k] = v
-        end
-      end
-    end
-
-    # Runs set_with_params and saves the changes (which runs any callback methods).
-    def update_with_params(values)
-      set_with_params(values)
-      save_changes
-    end
-
-    # Returns (naked) dataset bound to current instance.
-    def this
-      @this ||= self.class.dataset.filter(:id => @values[:id]).limit(1).naked
-    end
-    
-    # Returns a key unique to the underlying record for caching
-    def cache_key
-      pk = @values[:id] || (raise Error, 'no primary key for this record')
-      @cache_key ||= "#{self.class}:#{pk}"
-    end
-
-    # Returns primary key column(s) for object's Model class.
-    def primary_key
-      @primary_key ||= self.class.primary_key
-    end
-    
-    # Returns the primary key value identifying the model instance. If the
-    # model's primary key is changed (using #set_primary_key or #no_primary_key)
-    # this method is redefined accordingly.
-    def pk
-      @pk ||= @values[:id]
-    end
-    
-    # Returns a hash identifying the model instance. Stock implementation.
-    def pk_hash
-      @pk_hash ||= {:id => @values[:id]}
-    end
-    
-    # Creates new instance with values set to passed-in Hash.
-    #
-    # This method guesses whether the record exists when
-    # <tt>new_record</tt> is set to false.
-    def initialize(values = nil, from_db = false, &block)
-      values ||=  {}
-      @changed_columns = []
-      if from_db
-        @new = false
-        @values = values
-      else
-        @values = {}
-        @new = true
-        set_with_params(values)
-      end
-      @changed_columns.clear 
-      
-      yield self if block
-      after_initialize
-    end
-    
-    # Initializes a model instance as an existing record. This constructor is 
-    # used by Sequel to initialize model instances when fetching records.
-    def self.load(values)
-      new(values, true)
-    end
-    
-    # Returns true if the current instance represents a new record.
-    def new?
-      @new
-    end
-    
-    # Returns true when current instance exists, false otherwise.
-    def exists?
-      this.count > 0
-    end
-    
-    # Creates or updates the associated record. This method can also
-    # accept a list of specific columns to update.
-    def save(*columns)
-      before_save
-      if @new
-        before_create
-        iid = model.dataset.insert(@values)
-        # if we have a regular primary key and it's not set in @values,
-        # we assume it's the last inserted id
-        if (pk = primary_key) && !(Array === pk) && !@values[pk]
-          @values[pk] = iid
-        end
-        if pk
-          @this = nil # remove memoized this dataset
-          refresh
-        end
-        @new = false
-        after_create
-      else
-        before_update
-        if columns.empty?
-          this.update(@values)
-          @changed_columns = []
-        else # update only the specified columns
-          this.update(@values.reject {|k, v| !columns.include?(k)})
-          @changed_columns.reject! {|c| columns.include?(c)}
-        end
-        after_update
-      end
-      after_save
-      self
-    end
-    
-    # Saves only changed columns or does nothing if no columns are marked as 
-    # chanaged.
-    def save_changes
-      save(*@changed_columns) unless @changed_columns.empty?
-    end
-
-    # Sets the value attributes without saving the record.  Returns
-    # the values changed.  Raises an error if the keys are not symbols
-    # or strings or a string key was passed that was not a valid column.
-    # This is a low level method that does not respect virtual attributes.  It
-    # should probably be avoided.  Look into using set_with_params instead.
-    def set_values(values)
-      s = str_columns
-      vals = values.inject({}) do |m, kv| 
-        k, v = kv
-        k = case k
-        when Symbol
-          k
-        when String
-          # Prevent denial of service via memory exhaustion by only 
-          # calling to_sym if the symbol already exists.
-          raise(::Sequel::Error, "all string keys must be a valid columns") unless s.include?(k)
-          k.to_sym
-        else
-          raise(::Sequel::Error, "Only symbols and strings allows as keys")
-        end
-        m[k] = v
-        m
-      end
-      vals.each {|k, v| @values[k] = v}
-      vals
-    end
-
-    # Sets the values attributes with set_values and then updates
-    # the record in the database using those values.  This is a
-    # low level method that does not run the usual save callbacks.
-    # It should probably be avoided.  Look into using update_with_params instead.
-    def update_values(values)
-      this.update(set_values(values))
-    end
-    
-    # Reloads values from database and returns self.
-    def refresh
-      @values = this.first || raise(Error, "Record not found")
-      model.all_association_reflections.each do |r|
-        instance_variable_set("@#{r[:name]}", nil)
-      end
-      self
-    end
-    alias_method :reload, :refresh
-
-    # Like delete but runs hooks before and after delete.
-    def destroy
-      db.transaction do
-        before_destroy
-        delete
-        after_destroy
-      end
-      self
-    end
-    
-    # Deletes and returns self.  Does not run callbacks.
-    # Look into using destroy instead.
-    def delete
-      this.delete
-      self
-    end
-    
-    private
-      # Returns all methods that can be used for attribute
-      # assignment (those that end with =)
-      def setter_methods
-        methods.grep(/=\z/)
-      end
-  end
-end
diff -pruN 1.5.1-1/lib/sequel_model/schema.rb 3.0.0-1/lib/sequel_model/schema.rb
--- 1.5.1-1/lib/sequel_model/schema.rb	2008-04-28 23:26:16.000000000 +0100
+++ 3.0.0-1/lib/sequel_model/schema.rb	1970-01-01 01:00:00.000000000 +0100
@@ -1,50 +0,0 @@
-module Sequel
-  class Model
-    # Defines a table schema (see Schema::Generator for more information).
-    #
-    # This is only needed if you want to use the create_table or drop_table
-    # methods.
-    def self.set_schema(name = nil, &block)
-      if name
-        set_dataset(db[name])
-      end
-      @schema = Schema::Generator.new(db, &block)
-      if @schema.primary_key_name
-        set_primary_key @schema.primary_key_name
-      end
-    end
-    
-    # Returns table schema for direct descendant of Model.
-    def self.schema
-      @schema || ((superclass != Model) && (superclass.schema))
-    end
-
-    # Returns name of table.
-    def self.table_name
-      dataset.opts[:from].first
-    end
-    
-    # Returns true if table exists, false otherwise.
-    def self.table_exists?
-      db.table_exists?(table_name)
-    end
-    
-    # Creates table.
-    def self.create_table
-      db.create_table_sql_list(table_name, *schema.create_info).each {|s| db << s} 
-      @columns = nil
-      columns
-    end
-    
-    # Drops table.
-    def self.drop_table
-      db.execute db.drop_table_sql(table_name)
-    end
-    
-    # Like create_table but invokes drop_table when table_exists? is true.
-    def self.create_table!
-      drop_table if table_exists?
-      create_table
-    end
-  end
-end
diff -pruN 1.5.1-1/lib/sequel_model/validations.rb 3.0.0-1/lib/sequel_model/validations.rb
--- 1.5.1-1/lib/sequel_model/validations.rb	2008-04-28 23:26:16.000000000 +0100
+++ 3.0.0-1/lib/sequel_model/validations.rb	1970-01-01 01:00:00.000000000 +0100
@@ -1,313 +0,0 @@
-class Array
-  # Removes and returns the last member of the array if it is a hash. Otherwise,
-  # an empty hash is returned This method is useful when writing methods that
-  # take an options hash as the last parameter. For example:
-  #
-  #   def validate_each(*args, &block)
-  #     opts = args.extract_options!
-  #     ...
-  #   end
-  def extract_options!
-    last.is_a?(Hash) ? pop : {}
-  end
-end
-
-# The Validations module provides validation capabilities as a mixin. When
-# included into a class, it enhances the class with class and instance
-# methods for defining validations and validating class instances.
-# 
-# The Validation emulates the validation capabilities of ActiveRecord, and
-# provides methods for validating acceptance, confirmation, presence, format, 
-# length and numericality of attributes.
-#
-# To use validations, you need to include the Validation module in your
-# class:
-#
-#   class MyClass
-#     include Validation
-#     validates_length_of :password, :minimum => 6
-#   end
-module Validation
-  # Includes the Validation class methods into the including class.
-  def self.included(c)
-    c.extend ClassMethods
-  end
-  
-  # Returns the validation errors associated with the object.
-  def errors
-    @errors ||= Errors.new
-  end
-
-  # Validates the object.
-  def validate
-    errors.clear
-    self.class.validate(self)
-  end
-
-  # Validates the object and returns true if no errors are reported.
-  def valid?
-    validate
-    errors.empty?
-  end
-
-  # Validation::Errors represents validation errors.
-  class Errors
-    # Initializes a new instance of validation errors.
-    def initialize
-      @errors = Hash.new {|h, k| h[k] = []}
-    end
-    
-    # Returns true if no errors are stored.
-    def empty?
-      @errors.empty?
-    end
-    
-    # Clears all errors.
-    def clear
-      @errors.clear
-    end
-    
-    # Returns size of errors array
-    def size
-      @errors.size
-    end
-    
-    # Iterates over errors
-    def each(&block)
-      @errors.each(&block)
-    end
-    
-    # Returns the errors for the given attribute.
-    def on(att)
-      @errors[att]
-    end
-    alias_method :[], :on
-    
-    # Adds an error for the given attribute.
-    def add(att, msg)
-      @errors[att] << msg
-    end
-    
-    # Returns an array of fully-formatted error messages.
-    def full_messages
-      @errors.inject([]) do |m, kv| att, errors = *kv
-        errors.each {|e| m << "#{att} #{e}"}
-        m
-      end
-    end
-  end
-  
-  # The Generator class is used to generate validation definitions using 
-  # the validates {} idiom.
-  class Generator
-    # Initializes a new generator.
-    def initialize(receiver ,&block)
-      @receiver = receiver
-      instance_eval(&block)
-    end
-
-    # Delegates method calls to the receiver by calling receiver.validates_xxx.
-    def method_missing(m, *args, &block)
-      @receiver.send(:"validates_#{m}", *args, &block)
-    end
-  end
-  
-  # Validation class methods.
-  module ClassMethods
-    # Defines validations by converting a longhand block into a series of 
-    # shorthand definitions. For example:
-    #
-    #   class MyClass
-    #     include Validation
-    #     validates do
-    #       length_of :name, :minimum => 6
-    #       length_of :password, :minimum => 8
-    #     end
-    #   end
-    #
-    # is equivalent to:
-    #   class MyClass
-    #     include Validation
-    #     validates_length_of :name, :minimum => 6
-    #     validates_length_of :password, :minimum => 8
-    #   end
-    def validates(&block)
-      Generator.new(self, &block)
-    end
-
-    # Returns the validations hash for the class.
-    def validations
-      @validations ||= Hash.new {|h, k| h[k] = []}
-    end
-
-    # Returns true if validations are defined.
-    def has_validations?
-      !validations.empty?
-    end
-
-    # Validates the given instance.
-    def validate(o)
-      if superclass.respond_to?(:validate) && !@skip_superclass_validations
-        superclass.validate(o)
-      end
-      validations.each do |att, procs|
-        v = o.send(att)
-        procs.each {|p| p[o, att, v]}
-      end
-    end
-    
-    def skip_superclass_validations
-      @skip_superclass_validations = true
-    end
-    
-    # Adds a validation for each of the given attributes using the supplied
-    # block. The block must accept three arguments: instance, attribute and 
-    # value, e.g.:
-    #
-    #   validates_each :name, :password do |object, attribute, value|
-    #     object.errors[attribute] << 'is not nice' unless value.nice?
-    #   end
-    def validates_each(*atts, &block)
-      atts.each {|a| validations[a] << block}
-    end
-
-    # Validates acceptance of an attribute.
-    def validates_acceptance_of(*atts)
-      opts = {
-        :message => 'is not accepted',
-        :allow_nil => true,
-        :accept => '1'
-      }.merge!(atts.extract_options!)
-      
-      validates_each(*atts) do |o, a, v|
-        next if (v.nil? && opts[:allow_nil]) || (v.blank? && opts[:allow_blank])
-        o.errors[a] << opts[:message] unless v == opts[:accept]
-      end
-    end
-
-    # Validates confirmation of an attribute.
-    def validates_confirmation_of(*atts)
-      opts = {
-        :message => 'is not confirmed',
-      }.merge!(atts.extract_options!)
-      
-      validates_each(*atts) do |o, a, v|
-        next if (v.nil? && opts[:allow_nil]) || (v.blank? && opts[:allow_blank])
-        c = o.send(:"#{a}_confirmation")
-        o.errors[a] << opts[:message] unless v == c
-      end
-    end
-
-    # Validates the format of an attribute.
-    def validates_format_of(*atts)
-      opts = {
-        :message => 'is invalid',
-      }.merge!(atts.extract_options!)
-      
-      unless opts[:with].is_a?(Regexp)
-        raise ArgumentError, "A regular expression must be supplied as the :with option of the options hash"
-      end
-      
-      validates_each(*atts) do |o, a, v|
-        next if (v.nil? && opts[:allow_nil]) || (v.blank? && opts[:allow_blank])
-        o.errors[a] << opts[:message] unless v.to_s =~ opts[:with]
-      end
-    end
-
-    # Validates the length of an attribute.
-    def validates_length_of(*atts)
-      opts = {
-        :too_long     => 'is too long',
-        :too_short    => 'is too short',
-        :wrong_length => 'is the wrong length'
-      }.merge!(atts.extract_options!)
-      
-      validates_each(*atts) do |o, a, v|
-        next if (v.nil? && opts[:allow_nil]) || (v.blank? && opts[:allow_blank])
-        if m = opts[:maximum]
-          o.errors[a] << (opts[:message] || opts[:too_long]) unless v && v.size <= m
-        end
-        if m = opts[:minimum]
-          o.errors[a] << (opts[:message] || opts[:too_short]) unless v && v.size >= m
-        end
-        if i = opts[:is]
-          o.errors[a] << (opts[:message] || opts[:wrong_length]) unless v && v.size == i
-        end
-        if w = opts[:within]
-          o.errors[a] << (opts[:message] || opts[:wrong_length]) unless v && w.include?(v.size)
-        end
-      end
-    end
-
-    NUMBER_RE = /^\d*\.{0,1}\d+$/
-    INTEGER_RE = /\A[+-]?\d+\Z/
-
-    # Validates whether an attribute is a number.
-    def validates_numericality_of(*atts)
-      opts = {
-        :message => 'is not a number',
-      }.merge!(atts.extract_options!)
-      
-      re = opts[:only_integer] ? INTEGER_RE : NUMBER_RE
-      
-      validates_each(*atts) do |o, a, v|
-        next if (v.nil? && opts[:allow_nil]) || (v.blank? && opts[:allow_blank])
-        o.errors[a] << opts[:message] unless v.to_s =~ re
-      end
-    end
-
-    # Validates the presence of an attribute.
-    def validates_presence_of(*atts)
-      opts = {
-        :message => 'is not present',
-      }.merge!(atts.extract_options!)
-      
-      validates_each(*atts) do |o, a, v|
-        o.errors[a] << opts[:message] unless v && !v.blank?
-      end
-    end
-
-    # Validates only if the fields in the model (specified by atts) are
-    # unique in the database.  You should also add a unique index in the
-    # database, as this suffers from a fairly obvious race condition.
-    def validates_uniqueness_of(*atts)
-      opts = {
-        :message => 'is already taken',
-      }.merge!(atts.extract_options!)
-
-      validates_each(*atts) do |o, a, v|
-        next if v.blank? 
-        num_dups = o.class.filter(a => v).count
-        allow = if num_dups == 0
-          # No unique value in the database
-          true
-        elsif num_dups > 1
-          # Multiple "unique" values in the database!!
-          # Someone didn't add a unique index
-          false
-        elsif o.new?
-          # New record, but unique value already exists in the database
-          false
-        elsif o.class[a => v].pk == o.pk
-          # Unique value exists in database, but for the same record, so the update won't cause a duplicate record
-          true
-        else
-          false
-        end
-        o.errors[a] << opts[:message] unless allow
-      end
-    end
-  end
-end
-
-module Sequel
-  class Model
-    include Validation
-    
-    alias_method :save!, :save
-    def save(*args)
-      return false unless valid?
-      save!(*args)
-    end
-  end
-end
diff -pruN 1.5.1-1/lib/sequel_model.rb 3.0.0-1/lib/sequel_model.rb
--- 1.5.1-1/lib/sequel_model.rb	2008-04-24 22:14:31.000000000 +0100
+++ 3.0.0-1/lib/sequel_model.rb	2009-03-16 20:53:19.000000000 +0000
@@ -1,82 +1 @@
-require 'sequel_core'
-
-module Sequel
-  class Model
-    alias_method :model, :class
-  end
-end
-
-files = %w[
-  inflector inflections base hooks record schema associations 
-  caching plugins validations eager_loading deprecated
-]
-dir = File.join(File.dirname(__FILE__), "sequel_model")
-files.each {|f| require(File.join(dir, f))}
-
-module Sequel
-  class Model
-    extend Enumerable
-    extend Associations
-    # Returns a string representation of the model instance including
-    # the class name and values.
-    def inspect
-      "#<%s @values=%s>" % [self.class.name, @values.inspect]
-    end
-    
-    # Defines a method that returns a filtered dataset.
-    def self.subset(name, *args, &block)
-      def_dataset_method(name){filter(*args, &block)}
-    end
-    
-    # Finds a single record according to the supplied filter, e.g.:
-    #
-    #   Ticket.find :author => 'Sharon' # => record
-    #   Ticket.find {:price == 17} # => Dataset
-    #
-    def self.find(*args, &block)
-      dataset.filter(*args, &block).first
-    end
-    
-    # TODO: doc
-    def self.[](*args)
-      args = args.first if (args.size == 1)
-      if args === true || args === false
-        raise Error::InvalidFilter, "Did you mean to supply a hash?"
-      end
-      dataset[(Hash === args) ? args : primary_key_hash(args)]
-    end
-    
-    # TODO: doc
-    def self.fetch(*args)
-      db.fetch(*args).set_model(self)
-    end
-
-    # Like find but invokes create with given conditions when record does not
-    # exists.
-    def self.find_or_create(cond)
-      find(cond) || create(cond)
-    end
-
-    # Deletes all records in the model's table.
-    def self.delete_all
-      dataset.delete
-    end
-
-    # Like delete_all, but invokes before_destroy and after_destroy hooks if used.
-    def self.destroy_all
-      dataset.destroy
-    end
-
-    # Add dataset methods via metaprogramming
-    DATASET_METHODS = %w'all avg count delete distinct eager eager_graph each each_page 
-       empty? except exclude filter first from_self full_outer_join graph 
-       group group_and_count group_by having import inner_join insert 
-       insert_multiple intersect interval invert_order join join_table last 
-       left_outer_join limit multi_insert naked order order_by order_more 
-       paginate print query range reverse_order right_outer_join select 
-       select_all select_more set set_graph_aliases single_value size to_csv 
-       transform union uniq unordered update where'
-
-    def_dataset_method *DATASET_METHODS
-  end
-end
+require 'sequel/model'
diff -pruN 1.5.1-1/lib/sequel.rb 3.0.0-1/lib/sequel.rb
--- 1.5.1-1/lib/sequel.rb	2008-04-16 19:35:51.000000000 +0100
+++ 3.0.0-1/lib/sequel.rb	2009-03-16 20:53:19.000000000 +0000
@@ -1 +1 @@
-require 'sequel_model'
+require 'sequel/model'
diff -pruN 1.5.1-1/Rakefile 3.0.0-1/Rakefile
--- 1.5.1-1/Rakefile	2008-04-30 18:09:15.000000000 +0100
+++ 3.0.0-1/Rakefile	2009-04-21 21:52:15.000000000 +0100
@@ -1,142 +1,199 @@
 require "rake"
 require "rake/clean"
 require "rake/gempackagetask"
-require "rake/rdoctask"
-require "fileutils"
-include FileUtils
-
-##############################################################################
-# Configuration
-##############################################################################
-NAME = "sequel"
-VERS = "1.5.1"
-SEQUEL_CORE_VERS= "1.5.1"
-CLEAN.include ["**/.*.sw?", "pkg/*", ".config", "doc/*", "coverage/*"]
-RDOC_OPTS = ["--quiet", "--line-numbers", "--inline-source"]
-
-##############################################################################
-# RDoc
-##############################################################################
-Rake::RDocTask.new do |rdoc|
-  rdoc.rdoc_dir = "doc/rdoc"
-  rdoc.options += RDOC_OPTS
-  rdoc.main = "README"
-  rdoc.title = "Sequel: The Database Toolkit for Ruby: Model Classes"
-  rdoc.rdoc_files.add ["README", "COPYING", "lib/**/*.rb"]
-end
+begin
+  require "hanna/rdoctask"
+rescue LoadError
+  require "rake/rdoctask"
+end
+
+NAME = 'sequel'
+VERS = lambda do
+  require "lib/sequel/version"
+  Sequel.version
+end
+CLEAN.include ["**/.*.sw?", "pkg", ".config", "rdoc", "coverage", "www/public/*.html", "www/public/rdoc*"]
+RDOC_DEFAULT_OPTS = ["--quiet", "--line-numbers", "--inline-source", '--title', 'Sequel: The Database Toolkit for Ruby']
+RDOC_OPTS = RDOC_DEFAULT_OPTS + ['--main', 'README.rdoc']
 
-##############################################################################
-# Gem packaging
-##############################################################################
-desc "Packages up Sequel."
-task :package => [:clean]
+# Gem Packaging and Release
 
 spec = Gem::Specification.new do |s|
   s.name = NAME
   s.rubyforge_project = 'sequel'
-  s.version = VERS
+  s.version = VERS.call
   s.platform = Gem::Platform::RUBY
   s.has_rdoc = true
-  s.extra_rdoc_files = ["README", "CHANGELOG", "COPYING"]
-  s.rdoc_options += RDOC_OPTS + ["--exclude", "^(examples|extras)\/"]
-  s.summary = "The Database Toolkit for Ruby: Model Classes"
+  s.extra_rdoc_files = ["README.rdoc", "CHANGELOG", "COPYING"] + Dir["doc/*.rdoc"] + Dir['doc/release_notes/*.txt']
+  s.rdoc_options += RDOC_OPTS 
+  s.summary = "The Database Toolkit for Ruby"
   s.description = s.summary
   s.author = "Jeremy Evans"
   s.email = "code@jeremyevans.net"
   s.homepage = "http://sequel.rubyforge.org"
   s.required_ruby_version = ">= 1.8.4"
-
-  case RUBY_PLATFORM
-  when /java/
-    s.platform = "jruby"
-  else
-    s.platform = Gem::Platform::RUBY
-  end
-
-  s.add_dependency("sequel_core", "= #{SEQUEL_CORE_VERS}")
-  
-  s.files = %w(COPYING README Rakefile) + Dir.glob("{doc,spec,lib}/**/*")
-
+  s.files = %w(COPYING CHANGELOG README.rdoc Rakefile) + Dir["{bin,doc,spec,lib}/**/*"]
   s.require_path = "lib"
+  s.bindir = 'bin'
+  s.executables << 'sequel'
 end
 
+desc "Packages sequel"
+task :package=>[:clean]
 Rake::GemPackageTask.new(spec) do |p|
   p.need_tar = true
   p.gem_spec = spec
 end
 
-##############################################################################
-# installation & removal
-##############################################################################
 desc "Install sequel gem"
-task :install do
-  sh %{rake package}
-  sh %{sudo gem install pkg/#{NAME}-#{VERS}}
-end
-
-desc "Install sequel gem without docs"
-task :install_no_docs do
-  sh %{rake package}
-  sh %{sudo gem install pkg/#{NAME}-#{VERS} --no-rdoc --no-ri}
+task :install=>[:package] do
+  sh %{sudo gem install pkg/#{NAME}-#{VERS.call} --local}
 end
 
 desc "Uninstall sequel gem"
-task :uninstall => [:clean] do
+task :uninstall=>[:clean] do
   sh %{sudo gem uninstall #{NAME}}
 end
 
-##############################################################################
-# gem and rdoc release
-##############################################################################
 desc "Upload sequel gem to rubyforge"
-task :release => [:package] do
+task :release=>[:package] do
   sh %{rubyforge login}
-  sh %{rubyforge add_release sequel #{NAME} #{VERS} pkg/#{NAME}-#{VERS}.tgz}
-  sh %{rubyforge add_file sequel #{NAME} #{VERS} pkg/#{NAME}-#{VERS}.gem}
+  sh %{rubyforge add_release sequel #{NAME} #{VERS.call} pkg/#{NAME}-#{VERS.call}.tgz}
+  sh %{rubyforge add_file sequel #{NAME} #{VERS.call} pkg/#{NAME}-#{VERS.call}.gem} 
 end
 
-##############################################################################
-# specs
-##############################################################################
-require "spec/rake/spectask"
-sqcdir = File.join(File.dirname(File.dirname(__FILE__)), 'sequel_core', 'lib') 
-fixRUBYLIB = Proc.new{ENV['RUBYLIB'] ? (ENV['RUBYLIB'] += ":#{sqcdir}") : (ENV['RUBYLIB'] = sqcdir)}
-
-desc "Run specs with coverage"
-Spec::Rake::SpecTask.new("spec") do |t|
-  fixRUBYLIB.call
-  t.spec_files = FileList["spec/**/*_spec.rb"]
-  t.spec_opts  = File.read("spec/spec.opts").split("\n")
-  t.rcov_opts  = File.read("spec/rcov.opts").split("\n")
-  t.rcov = true
-end
-
-desc "Run specs without coverage"
-task :default => [:spec_no_cov]
-Spec::Rake::SpecTask.new("spec_no_cov") do |t|
-  fixRUBYLIB.call
-  t.spec_files = FileList["spec/**/*_spec.rb"]
-  t.spec_opts  = File.read("spec/spec.opts").split("\n")
+### RDoc
+
+Rake::RDocTask.new do |rdoc|
+  rdoc.rdoc_dir = "rdoc"
+  rdoc.options += RDOC_OPTS
+  rdoc.rdoc_files.add %w"README.rdoc CHANGELOG COPYING lib/**/*.rb doc/*.rdoc doc/release_notes/*.txt"
+end
+
+### Website
+
+desc "Make local version of website"
+task :website do
+  sh %{www/make_www.rb}
+end
+
+desc "Make rdoc for website"
+task :website_rdoc=>[:website_rdoc_main, :website_rdoc_adapters, :website_rdoc_plugins]
+
+Rake::RDocTask.new(:website_rdoc_main) do |rdoc|
+  rdoc.rdoc_dir = "www/public/rdoc"
+  rdoc.options += RDOC_OPTS
+  rdoc.rdoc_files.add %w"README.rdoc CHANGELOG COPYING lib/*.rb lib/sequel/*.rb lib/sequel/{dataset,database,model}/*.rb doc/*.rdoc doc/release_notes/*.txt"
+end
+
+Rake::RDocTask.new(:website_rdoc_adapters) do |rdoc|
+  rdoc.rdoc_dir = "www/public/rdoc-adapters"
+  rdoc.options += RDOC_DEFAULT_OPTS + %w'--main Sequel'
+  rdoc.rdoc_files.add %w"lib/sequel/adapters/**/*.rb"
+end
+
+Rake::RDocTask.new(:website_rdoc_plugins) do |rdoc|
+  rdoc.rdoc_dir = "www/public/rdoc-plugins"
+  rdoc.options += RDOC_DEFAULT_OPTS + %w'--main Sequel'
+  rdoc.rdoc_files.add %w"lib/sequel/{extensions,plugins}/**/*.rb"
+end
+
+desc "Update Non-RDoc section of sequel.rubyforge.org"
+task :website_rf_base=>[:website] do
+  sh %{rsync -rt www/public/*.html rubyforge.org:/var/www/gforge-projects/sequel/}
+end
+
+desc "Update sequel.rubyforge.org"
+task :website_rf=>[:website, :website_rdoc] do
+  sh %{rsync -rvt www/public/* rubyforge.org:/var/www/gforge-projects/sequel/}
+end
+
+### Specs
+
+begin
+  require "spec/rake/spectask"
+
+  spec_opts = lambda do
+    lib_dir = File.join(File.dirname(__FILE__), 'lib')
+    ENV['RUBYLIB'] ? (ENV['RUBYLIB'] += ":#{lib_dir}") : (ENV['RUBYLIB'] = lib_dir)
+    File.read("spec/spec.opts").split("\n")
+  end
+
+  rcov_opts = lambda do
+    [true, File.read("spec/rcov.opts").split("\n")]
+  end
+  
+  desc "Run core and model specs with coverage"
+  Spec::Rake::SpecTask.new("spec_coverage") do |t|
+    t.spec_files = Dir["spec/{core,model}/*_spec.rb"]
+    t.spec_opts  = spec_opts.call
+    t.rcov, t.rcov_opts = rcov_opts.call
+  end
+  
+  desc "Run core and model specs"
+  task :default => [:spec]
+  Spec::Rake::SpecTask.new("spec") do |t|
+    t.spec_files = Dir["spec/{core,model}/*_spec.rb"]
+    t.spec_opts  = spec_opts.call
+  end
+  
+  desc "Run core specs"
+  Spec::Rake::SpecTask.new("spec_core") do |t|
+    t.spec_files = Dir["spec/core/*_spec.rb"]
+    t.spec_opts  = spec_opts.call
+  end
+  
+  desc "Run model specs"
+  Spec::Rake::SpecTask.new("spec_model") do |t|
+    t.spec_files = Dir["spec/model/*_spec.rb"]
+    t.spec_opts  = spec_opts.call
+  end
+  
+  desc "Run extension/plugin specs"
+  Spec::Rake::SpecTask.new("spec_plugin") do |t|
+    t.spec_files = Dir["spec/extensions/*_spec.rb"]
+    t.spec_opts  = spec_opts.call
+  end
+  
+  desc "Run extention/plugin specs with coverage"
+  Spec::Rake::SpecTask.new("spec_plugin_cov") do |t|
+    t.spec_files = Dir["spec/extensions/*_spec.rb"]
+    t.spec_opts  = spec_opts.call
+    t.rcov, t.rcov_opts = rcov_opts.call
+  end
+  
+  desc "Run integration tests"
+  Spec::Rake::SpecTask.new("integration") do |t|
+    t.spec_files = FileList["spec/integration/*_test.rb"]
+    t.spec_opts  = spec_opts.call
+  end
+  
+  %w'postgres sqlite mysql informix oracle ado'.each do |adapter|
+    desc "Run #{adapter} specs without coverage"
+    Spec::Rake::SpecTask.new("spec_#{adapter}") do |t|
+      t.spec_files = ["spec/adapters/#{adapter}_spec.rb"]
+      t.spec_opts  = spec_opts.call
+    end
+  end
+rescue LoadError
 end
 
 desc "check documentation coverage"
 task :dcov do
-  sh "find lib -name '*.rb' | xargs dcov"
+  sh %{find lib -name '*.rb' | xargs dcov}
 end
 
-##############################################################################
-# Statistics
-##############################################################################
-
-STATS_DIRECTORIES = [
-  %w(Code   lib/),
-  %w(Spec   spec/)
-].collect { |name, dir| [ name, "./#{dir}" ] }.select { |name, dir| File.directory?(dir) }
+### Statistics
 
 desc "Report code statistics (KLOCs, etc) from the application"
 task :stats do
+  STATS_DIRECTORIES = [%w(Code lib/), %w(Spec spec)].map{|name, dir| [ name, "./#{dir}" ] }.select { |name, dir| File.directory?(dir)}
   require "extra/stats"
   verbose = true
   CodeStatistics.new(*STATS_DIRECTORIES).to_s
 end
 
+desc "Print Sequel version"
+task :version do
+  puts VERS.call
+end
diff -pruN 1.5.1-1/README 3.0.0-1/README
--- 1.5.1-1/README	2008-04-29 16:06:52.000000000 +0100
+++ 3.0.0-1/README	1970-01-01 01:00:00.000000000 +0100
@@ -1,355 +0,0 @@
-== Sequel Models
-
-Models in Sequel are based on the Active Record pattern described by Martin Fowler (http://www.martinfowler.com/eaaCatalog/activeRecord.html). A model class corresponds to a table or a dataset, and an instance of that class wraps a single record in the model's underlying dataset.
-
-Model classes are defined as regular Ruby classes:
-
-  DB = Sequel.connect('sqlite:/blog.db')
-  class Post < Sequel::Model
-  end
-
-Just like in DataMapper or ActiveRecord, Sequel model classes assume that the table name is a plural of the class name:
-
-  Post.table_name #=> :posts
-
-You can, however, explicitly set the table name or even the dataset used:
-
-  class Post < Sequel::Model(:my_posts)
-  end
-
-  # or:
-
-  Post.set_dataset :my_posts
-
-  # or:
-
-  Post.set_dataset DB[:my_posts].where(:category => 'ruby')
-
-=== Resources
-
-* {Project page}[http://code.google.com/p/ruby-sequel/]
-* {Source code}[http://github.com/jeremyevans/sequel]
-* {Bug tracking}[http://code.google.com/p/ruby-sequel/issues/list]
-* {Google group}[http://groups.google.com/group/sequel-talk]
-* {RubyForge page}[http://rubyforge.org/projects/sequel/]
-* {API RDoc}[http://sequel.rubyforge.org]
-
-To check out the source code:
-
-  git clone git://github.com/jeremyevans/sequel.git
-
-=== Contact
-
-If you have any comments or suggestions please post to the Google group.
-
-=== Installation
-
-  sudo gem install sequel
-
-=== Model instances
-
-Model instance are identified by a primary key. By default, Sequel assumes the primary key column to be :id. The Model#[] method can be used to fetch records by their primary key:
-
-  post = Post[123]
-
-The Model#pk method is used to retrieve the record's primary key value:
-
-  post.pk #=> 123
-
-Sequel models allow you to use any column as a primary key, and even composite keys made from multiple columns:
-
-  class Post < Sequel::Model
-    set_primary_key [:category, :title]
-  end
-
-  post = Post['ruby', 'hello world']
-  post.pk #=> ['ruby', 'hello world']
-
-You can also define a model class that does not have a primary key, but then you lose the ability to update records.
-
-A model instance can also be fetched by specifying a condition:
-
-  post = Post[:title => 'hello world']
-  post = Post.find{:num_comments < 10}
-
-=== Iterating over records
-
-A model class lets you iterate over specific records by acting as a proxy to the underlying dataset. This means that you can use the entire Dataset API to create customized queries that return model instances, e.g.:
-
-  Post.filter(:category => 'ruby').each{|post| p post}
-
-You can also manipulate the records in the dataset:
-
-  Post.filter{:num_comments < 7}.delete
-  Post.filter{:title =~ /ruby/}.update(:category => 'ruby')
-
-=== Accessing record values
-
-A model instances stores its values as a hash:
-
-  post.values #=> {:id => 123, :category => 'ruby', :title => 'hello world'}
-
-You can read the record values as object attributes (assuming the attribute names are valid columns in the model's dataset):
-
-  post.id #=> 123
-  post.title #=> 'hello world'
-
-You can also change record values:
-
-  post.title = 'hey there'
-  post.save
-
-Another way to change values by using the #update_with_params method:
-
-  post.update_with_params(:title => 'hey there')
-
-=== Creating new records
-
-New records can be created by calling Model.create:
-
-  post = Post.create(:title => 'hello world')
-
-Another way is to construct a new instance and save it:
-
-  post = Post.new
-  post.title = 'hello world'
-  post.save
-
-You can also supply a block to Model.new and Model.create:
-
-  post = Post.create {|p| p.title = 'hello world'}
-
-  post = Post.new do |p|
-    p.title = 'hello world'
-    p.save
-  end
-
-=== Hooks
-
-You can execute custom code when creating, updating, or deleting records by using hooks. The before_create and after_create hooks wrap record creation. The before_update and after_update wrap record updating. The before_save and after_save wrap record creation and updating. The before_destroy and after_destroy wrap destruction.
-
-Hooks are defined by supplying a block:
-
-  class Post < Sequel::Model
-    after_create do
-      self.created_at = Time.now
-    end
-
-    after_destroy do
-      author.update_post_count
-    end
-  end
-
-=== Deleting records
-
-You can delete individual records by calling #delete or #destroy. The only difference between the two methods is that #destroy invokes before_destroy and after_destroy hooks, while #delete does not:
-
-  post.delete #=> bypasses hooks
-  post.destroy #=> runs hooks
-
-Records can also be deleted en-masse by invoking Model.delete and Model.destroy. As stated above, you can specify filters for the deleted records:
-
-  Post.filter(:category => 32).delete #=> bypasses hooks
-  Post.filter(:category => 32).destroy #=> runs hooks
-
-Please note that if Model.destroy is called, each record is deleted 
-separately, but Model.delete deletes all relevant records with a single 
-SQL statement.
-
-=== Associations
-
-Associations are used in order to specify relationships between model classes that reflect relations between tables in the database using foreign keys.
-
-  class Post < Sequel::Model
-    many_to_one :author
-    one_to_many :comments
-    many_to_many :tags
-  end
-
-You can also use the ActiveRecord names for these associations:
-
-  class Post < Sequel::Model
-    belongs_to :author
-    has_many :comments
-    has_and_belongs_to_many :tags
-  end
-
-many_to_one/belongs_to creates a getter and setter for each model object:
-
-  class Post < Sequel::Model
-    many_to_one :author
-  end
-
-  post = Post.create(:name => 'hi!')
-  post.author = Author[:name => 'Sharon']
-  post.author
-
-one_to_many/has_many and many_to_many/has_and_belongs_to_many create a getter method, a method for adding an object to the association, and a method for removing an object from the association:
-
-  class Post < Sequel::Model
-    one_to_many :comments
-    many_to_many :tags
-  end
-
-  post = Post.create(:name => 'hi!')
-  post.comments
-  comment = Comment.create(:text=>'hi')
-  post.add_comment(comment)
-  post.remove_comment(comment)
-  tag = Tag.create(:tag=>'interesting')
-  post.add_tag(tag)
-  post.remove_tag(tag)
-
-=== Eager Loading
-
-Associations can be eagerly loaded via .eager and the :eager association option. Eager loading is used when loading a group of objects. It loads all associated objects for all of the current objects in one query, instead of using a separate query to get the associated objects for each current object. Eager loading requires that you retrieve all model objects at once via .all (instead of individually by .each). Eager loading can be cascaded, loading association's associated objects.
-
-  class Person < Sequel::Model
-    one_to_many :posts, :eager=>[:tags]
-  end
-
-  class Post < Sequel::Model
-    many_to_one :person
-    one_to_many :replies
-    many_to_many :tags
-  end
-
-  class Tag < Sequel::Model
-    many_to_many :posts
-    many_to_many :replies
-  end
-
-  class Reply < Sequel::Model
-    many_to_one :person
-    many_to_one :post
-    many_to_many :tags
-  end
-
-  # Eager loading via .eager
-  Post.eager(:person).all
-
-  # eager is a dataset method, so it works with filters/orders/limits/etc.
-  Post.filter("topic > 'M'").order(:date).limit(5).eager(:person).all
-  
-  person = Person.first
-  # Eager loading via :eager (will eagerly load the tags for this person's posts)
-  person.posts
-  
-  # These are equivalent
-  Post.eager(:person, :tags).all
-  Post.eager(:person).eager(:tags).all
-  
-  # Cascading via .eager
-  Tag.eager(:posts=>:replies).all
-  
-  # Will also grab all associated posts' tags (because of :eager)
-  Reply.eager(:person=>:posts).all
-  
-  # No depth limit (other than memory/stack), and will also grab posts' tags
-  # Loads all people, their posts, their posts' tags, replies to those posts,
-  # the person for each reply, the tag for each reply, and all posts and
-  # replies that have that tag.  Uses a total of 8 queries.
-  Person.eager(:posts=>{:replies=>[:person, {:tags=>{:posts, :replies}}]}).all
-
-In addition to using eager, you can also use eager_graph, which will use a single query to get the object and all associated objects.  This may be necessary if you want to filter the result set based on columns in associated tables.  It works with cascading as well, the syntax is exactly the same.  Note that using eager_graph to eagerly load multiple *_to_many associations will cause the result set to be a cartesian product, so you should be very careful with your filters when using it in that case.
-
-=== Caching model instances with memcached
-
-Sequel models can be cached using memcached based on their primary keys. The use of memcached can significantly reduce database load by keeping model instances in memory. The set_cache method is used to specify caching:
-
-  require 'memcache'
-  CACHE = MemCache.new 'localhost:11211', :namespace => 'blog'
-
-  class Author < Sequel::Model
-    set_cache CACHE, :ttl => 3600
-  end
-
-  Author[333] # database hit
-  Author[333] # cache hit
-
-=== Extending the underlying dataset
-
-The obvious way to add table-wide logic is to define class methods to the model class definition. That way you can define subsets of the underlying dataset, change the ordering, or perform actions on multiple records:
-
-  class Post < Sequel::Model
-    def self.posts_with_few_comments
-      filter{:num_comments < 30}
-    end
-
-    def self.clean_posts_with_few_comments
-      posts_with_few_comments.delete
-    end
-  end
-
-You can also implement table-wide logic by defining methods on the dataset:
-
-  class Post < Sequel::Model
-    def_dataset_method(:posts_with_few_comments) do
-      filter{:num_comments < 30}
-    end
-
-    def_dataset_method(:clean_posts_with_few_comments) do
-      posts_with_few_comments.delete
-    end
-  end
-
-This is the recommended way of implementing table-wide operations, and allows you to have access to your model API from filtered datasets as well:
-
-  Post.filter(:category => 'ruby').clean_old_posts
-
-Sequel models also provide a short hand notation for filters:
-
-  class Post < Sequel::Model
-    subset(:posts_with_few_comments){:num_comments < 30}
-    subset :invisible, :visible => false
-  end
-
-=== Defining the underlying schema
-
-Model classes can also be used as a place to define your table schema and control it. The schema DSL is exactly the same provided by Sequel::Schema::Generator:
-
-  class Post < Sequel::Model
-    set_schema do
-      primary_key :id
-      text :title
-      text :category
-      foreign_key :author_id, :table => :authors
-    end
-  end
-
-You can then create the underlying table, drop it, or recreate it:
-
-  Post.table_exists?
-  Post.create_table
-  Post.drop_table
-  Post.create_table! # drops the table if it exists and then recreates it
-
-=== Basic Model Validations
-
-To assign default validations to a sequel model:
-
-  class MyModel < Sequel::Model
-    validates do
-      format_of...
-      presence_of...
-      acceptance_of...
-      confirmation_of...
-      length_of...
-      numericality_of...
-      format_of...
-      each...
-    end
-  end
-
-You may also perform the usual 'longhand' way to assign default model validates directly within the model class itself:
-
-  class MyModel < Sequel::Model
-    validates_format_of...
-    validates_presence_of...
-    validates_acceptance_of...
-    validates_confirmation_of...
-    validates_length_of...
-    validates_numericality_of...
-    validates_format_of...
-    validates_each...
-  end
diff -pruN 1.5.1-1/README.rdoc 3.0.0-1/README.rdoc
--- 1.5.1-1/README.rdoc	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/README.rdoc	2009-04-15 22:25:10.000000000 +0100
@@ -0,0 +1,652 @@
+== Sequel: The Database Toolkit for Ruby
+
+Sequel is a lightweight database access toolkit for Ruby.
+
+* Sequel provides thread safety, connection pooling and a concise DSL
+  for constructing database queries and table schemas.
+* Sequel also includes a lightweight but comprehensive ORM layer for
+  mapping records to Ruby objects and handling associated records.
+* Sequel supports advanced database features such as prepared
+  statements, bound variables, stored procedures, master/slave
+  configurations, and database sharding.
+* Sequel makes it easy to deal with multiple records without having
+  to break your teeth on SQL.
+* Sequel currently has adapters for ADO, Amalgalite, DataObjects,
+  DB2, DBI, Firebird, Informix, JDBC, MySQL, ODBC, OpenBase, Oracle,
+  PostgreSQL and SQLite3.
+
+== Resources
+
+* {Website}[http://sequel.rubyforge.org]
+* {Source code}[http://github.com/jeremyevans/sequel]
+* {Bug tracking}[http://code.google.com/p/ruby-sequel/issues/list]
+* {Google group}[http://groups.google.com/group/sequel-talk]
+* {RDoc}[http://sequel.rubyforge.org/rdoc]
+
+To check out the source code:
+  
+  git clone git://github.com/jeremyevans/sequel.git
+  
+=== Contact
+
+If you have any comments or suggestions please post to the Google group.
+
+== Installation
+
+  sudo gem install sequel
+  
+== A Short Example
+
+  require 'rubygems'
+  require 'sequel'
+  
+  DB = Sequel.sqlite # memory database
+  
+  DB.create_table :items do
+    primary_key :id
+    String name
+    Float price
+  end
+  
+  items = DB[:items] # Create a dataset
+  
+  # Populate the table
+  items.insert(:name => 'abc', :price => rand * 100)
+  items.insert(name => 'def', :price => rand * 100)
+  items.insert(:name => 'ghi', :price => rand * 100)
+  
+  # Print out the number of records
+  puts "Item count: #{items.count}"
+  
+  # Print out the average price
+  puts "The average price is: #{items.avg(:price)}"
+
+== The Sequel Console
+
+Sequel includes an IRB console for quick access to databases. You can use it like this:
+
+  sequel sqlite://test.db # test.db in current directory
+
+You get an IRB session with the database object stored in DB.
+
+== An Introduction
+
+Sequel is designed to take the hassle away from connecting to databases and manipulating them. Sequel deals with all the boring stuff like maintaining connections, formatting SQL correctly and fetching records so you can concentrate on your application.
+
+Sequel uses the concept of datasets to retrieve data. A Dataset object encapsulates an SQL query and supports chainability, letting you fetch data using a convenient Ruby DSL that is both concise and flexible.
+
+For example, the following one-liner returns the average GDP for the five biggest countries in the middle east region:
+
+  DB[:countries].filter(:region => 'Middle East').reverse_order(:area).limit(5).avg(:GDP)
+  
+Which is equivalent to:
+
+  SELECT avg(GDP) FROM countries WHERE region = 'Middle East' ORDER BY area DESC LIMIT 5
+
+Since datasets retrieve records only when needed, they can be stored and later reused. Records are fetched as hashes (or custom model objects), and are accessed using an Enumerable interface:
+
+  middle_east = DB[:countries].filter(:region => 'Middle East')
+  middle_east.order(:name).each{|r| puts r[:name]}
+  
+Sequel also offers convenience methods for extracting data from Datasets, such as an extended map method:
+
+  middle_east.map(:name) #=> ['Egypt', 'Greece', 'Israel', ...]
+  
+Or getting results as a transposed hash, with one column as key and another as value:
+
+  middle_east.to_hash(:name, :area) #=> {'Israel' => 20000, 'Greece' => 120000, ...}
+
+== Getting Started
+
+=== Connecting to a database
+
+To connect to a database you simply provide Sequel with a URL:
+
+  require 'sequel'
+  DB = Sequel.connect('sqlite://blog.db')
+  
+The connection URL can also include such stuff as the user name and password:
+
+  DB = Sequel.connect('postgres://cico:12345@localhost:5432/mydb')
+
+You can also specify optional parameters, such as the connection pool size, or loggers for logging SQL queries:
+
+  DB = Sequel.connect("postgres://postgres:postgres@localhost/my_db",
+    :max_connections => 10, :loggers => [Logger.new('log/db.log']))
+
+You can specify a block to connect, which will disconnect from the database after it completes:
+
+  Sequel.connect('postgres://cico:12345@localhost:5432/mydb'){|db| db[:posts].delete}
+
+=== Arbitrary SQL queries
+
+  DB << "create table t (a text, b text)"
+  DB << "insert into t values ('a', 'b')"
+
+You can also create datasets based on raw SQL:
+
+  dataset = DB['select id from items']
+  dataset.count # will return the number of records in the result set
+  dataset.map(:id) # will return an array containing all values of the id column in the result set
+
+You can also fetch records with raw SQL through the dataset:
+
+  DB['select * from items'].each do |row|
+    p row
+  end
+
+You can use placeholders in your SQL string as well:
+
+  DB['select * from items where name = ?', name].each do |row|
+    p row
+  end
+
+=== Getting Dataset Instances
+
+Datasets are the primary way records are retrieved and manipulated.  They are generally created via the Database#from or Database#[] methods:
+
+  posts = DB.from(:posts)
+  posts = DB[:posts] # same
+
+Datasets will only fetch records when you tell them to. They can be manipulated to filter records, change ordering, join tables, etc..
+
+=== Retrieving Records
+
+You can retrieve all records by using the all method:
+
+  posts.all
+
+The all method returns an array of hashes, where each hash corresponds to a record.
+
+You can also iterate through records one at a time:
+
+  posts.each{|row| p row}
+
+Or perform more advanced stuff:
+
+  names_and_dates = posts.map{|r| [r[:name], r[:date]]}
+  old_posts, recent_posts = posts.partition{|r| r[:date] < Date.today - 7}
+  
+You can also retrieve the first record in a dataset:
+
+  posts.first
+  
+Or retrieve a single record with a specific value:
+
+  posts[:id => 1]
+  
+If the dataset is ordered, you can also ask for the last record:
+
+  posts.order(:stamp).last
+  
+=== Filtering Records
+
+An easy way to filter records is to provide a hash of values to match:
+
+  my_posts = posts.filter(:category => 'ruby', :author => 'david')
+  
+You can also specify ranges:
+
+  my_posts = posts.filter(:stamp => (Date.today - 14)..(Date.today - 7))
+  
+Or arrays of values:
+
+  my_posts = posts.filter(:category => ['ruby', 'postgres', 'linux'])
+  
+Sequel also accepts expressions:
+  
+  my_posts = posts.filter{|o| o.stamp > Date.today << 1}
+  
+Some adapters will also let you specify Regexps:
+
+  my_posts = posts.filter(:category => /ruby/i)
+  
+You can also use an inverse filter:
+
+  my_posts = posts.exclude(:category => /ruby/i)
+
+You can also specify a custom WHERE clause using a string:
+
+  posts.filter('stamp IS NOT NULL')
+
+You can use parameters in your string, as well:
+
+  posts.filter('(stamp < ?) AND (author != ?)', Date.today - 3, author_name)
+  posts.filter{|o| (o.stamp < Date.today - 3) & ~{:author => author_name}} # same as above
+
+Datasets can also be used as subqueries:
+
+  DB[:items].filter('price > ?', DB[:items].select{|o| o.avg(:price) + 100})
+
+After filtering you can retrieve the matching records by using any of the retrieval methods:
+
+  my_posts.each{|row| p row}
+  
+See the doc/dataset_filtering.rdoc file for more details.
+
+=== Summarizing Records
+
+Counting records is easy:
+  posts.filter(:category => /ruby/i).count
+
+And you can also query maximum/minimum values:
+  max = DB[:history].max(:value)
+  min = DB[:history].min(:value)
+  
+Or calculate a sum or average:
+  sum = DB[:items].sum(:price)
+  avg = DB[:items].avg(:price)
+  
+=== Ordering Records
+
+Ordering datasets is simple:
+
+  posts.order(:stamp) # ORDER BY stamp
+  posts.order(:stamp, :name) # ORDER BY stamp, name
+
+Chaining order doesn't work the same as filter:
+
+  posts.order(:stamp).order(:name) # ORDER BY name
+
+The order_more method chains this way, though:
+
+  posts.order(:stamp).order_more(:name) # ORDER BY stamp, name
+  
+You can also specify descending order:
+
+  posts.order(:stamp.desc) # ORDER BY stamp DESC
+
+=== Selecting Columns
+
+Selecting specific columns to be returned is also simple:
+
+  posts.select(:stamp) # SELECT stamp FROM posts
+  posts.select(:stamp, :name) # SELECT stamp, name FROM posts
+
+Chaining select works like order, not filter:
+
+  posts.select(:stamp).select(:name) # SELECT name FROM posts
+
+As you might expect, there is an order_more equivalent for select:
+
+  posts.select(:stamp).select_more(:name) # SELECT stamp, name FROM posts
+  
+=== Deleting Records
+
+Deleting records from the table is done with delete:
+
+  posts.filter('stamp < ?', Date.today - 3).delete
+  
+Be very careful when deleting, as delete affects all rows in the dataset.
+Filter first, delete second, unless you want to empty the table.
+
+=== Inserting Records
+
+Inserting records into the table is done with insert:
+
+  posts.insert(:category => 'ruby', :author => 'david')
+  
+=== Updating Records
+
+Updating records in the table is done with update:
+
+  posts.filter('stamp < ?', Date.today - 7).update(:state => 'archived')
+
+You can reference table columns when choosing what values to set:
+
+  posts.filter{|o| o.stamp < Date.today - 7}.update(:backup_number => :backup_number + 1)
+
+As with delete, this affects all rows in the dataset, so filter first,
+update second, unless you want to update all rows.
+
+=== Joining Tables
+
+Sequel makes it easy to join tables:
+
+  order_items = DB[:items].join(:order_items, :item_id => :id).
+    filter(:order_items__order_id => 1234)
+  
+This is equivalent to the SQL:
+
+  SELECT * FROM items INNER JOIN order_items
+  ON order_items.item_id = items.id 
+  WHERE order_items.order_id = 1234
+
+You can then do anything you like with the dataset:
+
+  order_total = order_items.sum(:price)
+  
+Which is equivalent to the SQL:
+
+  SELECT sum(price) FROM items INNER JOIN order_items
+  ON order_items.item_id = items.id
+  WHERE order_items.order_id = 1234
+  
+=== Graphing Datasets
+
+When retrieving records from joined datasets, you get the results in a single hash, which is subject to clobbering if you have columns with the same name in multiple tables:
+
+  DB[:items].join(:order_items, :item_id => :id).first
+  => {:id=>order_items.id), :item_id=>order_items.item_id}
+
+Using graph, you can split the result hashes into subhashes, one per join:
+
+  DB[:items].graph(:order_items, :item_id => :id).first
+  => {:items=>{:id=>items.id}, :order_items=>{:id=>order_items.id, :item_id=>order_items.item_id}}
+
+== An aside: column references in Sequel
+
+Sequel expects column names to be specified using symbols. In addition, returned hashes always use symbols as their keys. This allows you to freely mix literal values and column references. For example, the two following lines produce equivalent SQL:
+
+  items.filter(:x => 1) #=> "SELECT * FROM items WHERE (x = 1)" 
+  items.filter(1 => :x) #=> "SELECT * FROM items WHERE (1 = x)" 
+
+Ruby strings are generally treated as SQL strings:
+
+  items.filter(:x => 'x') #=> "SELECT * FROM items WHERE (x = 'x')" 
+
+=== Qualifying column names
+
+Column references can be qualified by using the double underscore special notation :table__column:
+
+  items.literal(:items__price) #=> "items.price"
+
+=== Column aliases
+
+You can also alias columns by using the triple undersecore special notation :column___alias or :table__column___alias:
+
+  items.literal(:price___p) #=> "price AS p"
+  items.literal(:items__price___p) #=> "items.price AS p"
+
+Another way to alias columns is to use the #as method:
+
+  items.literal(:price.as(:p)) #=> "price AS p"
+
+== Sequel Models
+
+A model class wraps a dataset, and an instance of that class wraps a single record in the dataset.
+
+Model classes are defined as regular Ruby classes:
+
+  DB = Sequel.connect('sqlite://blog.db')
+  class Post < Sequel::Model
+  end
+
+Just like in DataMapper or ActiveRecord, Sequel model classes assume that the table name is a plural of the class name:
+
+  Post.table_name #=> :posts
+
+You can, however, explicitly set the table name or even the dataset used:
+
+  class Post < Sequel::Model(:my_posts)
+  end
+  # or:
+  Post.set_dataset :my_posts
+
+If you use a symbol, it assumes you are referring to the table with the same name.  You can also give it a dataset:
+
+  Post.set_dataset DB[:my_posts].filter(:category => 'ruby')
+  Post.set_dataset DB[:my_posts].select(:id, :name).order(:date)
+
+=== Model instances
+
+Model instance are identified by a primary key. By default, Sequel assumes the primary key column to be :id, unless it can get the primary key information from the database. The Model.[] method can be used to fetch records by their primary key:
+
+  post = Post[123]
+
+The Model#pk method is used to retrieve the record's primary key value:
+
+  post.pk #=> 123
+
+Sequel models allow you to use any column as a primary key, and even composite keys made from multiple columns:
+
+  class Post < Sequel::Model
+    set_primary_key [:category, :title]
+  end
+
+  post = Post['ruby', 'hello world']
+  post.pk #=> ['ruby', 'hello world']
+
+You can also define a model class that does not have a primary key, but then you lose the ability to update records.
+
+A model instance can also be fetched by specifying a condition:
+
+  post = Post[:title => 'hello world']
+  post = Post.find{|o| o.num_comments < 10}
+
+=== Iterating over records
+
+A model class lets you iterate over subsets of records by proxying many methods to the underlying dataset. This means that you can use most of the Dataset API to create customized queries that return model instances, e.g.:
+
+  Post.filter(:category => 'ruby').each{|post| p post}
+
+You can also manipulate the records in the dataset:
+
+  Post.filter{|o| o.num_comments < 7}.delete
+  Post.filter(:title.like(/ruby/)).update(:category => 'ruby')
+
+=== Accessing record values
+
+A model instances stores its values as a hash:
+
+  post.values #=> {:id => 123, :category => 'ruby', :title => 'hello world'}
+
+You can read the record values as object attributes (assuming the attribute names are valid columns in the model's dataset):
+
+  post.id #=> 123
+  post.title #=> 'hello world'
+
+You can also change record values:
+
+  post.title = 'hey there'
+  post.save
+
+Another way to change values by using the #update method:
+
+  post.update(:title => 'hey there')
+
+=== Creating new records
+
+New records can be created by calling Model.create:
+
+  post = Post.create(:title => 'hello world')
+
+Another way is to construct a new instance and save it:
+
+  post = Post.new
+  post.title = 'hello world'
+  post.save
+
+You can also supply a block to Model.new and Model.create:
+
+  post = Post.create{|p| p.title = 'hello world'}
+
+  post = Post.new do |p|
+    p.title = 'hello world'
+    p.save
+  end
+
+=== Hooks
+
+You can execute custom code when creating, updating, or deleting records by defining hook methods. The before_create and after_create hook methods wrap record creation. The before_update and after_update hook methods wrap record updating. The before_save and after_save hook methods wrap record creation and updating. The before_destroy and after_destroy hook methods wrap destruction. The before_validation and after_validation hook methods wrap validation. Example:
+
+  class Post < Sequel::Model
+    def after_create
+      author.increase_post_count
+    end
+
+    def after_destroy
+      author.decrease_post_count
+    end
+  end
+
+For the example above, you should probably use a database trigger if you can.  Hooks can be used for data integrity, but they will only enforce that integrity when you are using the model.  If you plan on allowing any other access to the database, it's best to use database triggers for data integrity.
+
+=== Deleting records
+
+You can delete individual records by calling #delete or #destroy. The only difference between the two methods is that #destroy invokes before_destroy and after_destroy hook methods, while #delete does not:
+
+  post.delete #=> bypasses hooks
+  post.destroy #=> runs hooks
+
+Records can also be deleted en-masse by invoking Model.delete and Model.destroy. As stated above, you can specify filters for the deleted records:
+
+  Post.filter(:category => 32).delete #=> bypasses hooks
+  Post.filter(:category => 32).destroy #=> runs hooks
+
+Please note that if Model.destroy is called, each record is deleted 
+separately, but Model.delete deletes all relevant records with a single 
+SQL statement.
+
+=== Associations
+
+Associations are used in order to specify relationships between model classes that reflect relations between tables in the database using foreign keys.
+
+  class Post < Sequel::Model
+    many_to_one :author
+    one_to_many :comments
+    many_to_many :tags
+  end
+
+many_to_one creates a getter and setter for each model object:
+
+  class Post < Sequel::Model
+    many_to_one :author
+  end
+
+  post = Post.create(:name => 'hi!')
+  post.author = Author[:name => 'Sharon']
+  post.author
+
+one_to_many and many_to_many create a getter method, a method for adding an object to the association, a method for removing an object from the association, and a method for removing all associated objected from the association:
+
+  class Post < Sequel::Model
+    one_to_many :comments
+    many_to_many :tags
+  end
+
+  post = Post.create(:name => 'hi!')
+  post.comments
+  comment = Comment.create(:text=>'hi')
+  post.add_comment(comment)
+  post.remove_comment(comment)
+  post.remove_all_comments
+  tag = Tag.create(:tag=>'interesting')
+  post.add_tag(tag)
+  post.remove_tag(tag)
+  post.remove_all_tags
+
+All associations add a dataset method that can be used to further filter or reorder the returned objects, or modify all of them:
+
+  # Delete all of this post's comments from the database
+  Post.comments_dataset.destroy
+
+  # Return all tags related to this post with no subscribers, ordered by the tag's name
+  Post.tags_dataset.filter(:subscribers=>0).order(:name).all
+
+=== Eager Loading
+
+Associations can be eagerly loaded via .eager and the :eager association option. Eager loading is used when loading a group of objects. It loads all associated objects for all of the current objects in one query, instead of using a separate query to get the associated objects for each current object. Eager loading requires that you retrieve all model objects at once via .all (instead of individually by .each). Eager loading can be cascaded, loading association's associated objects.
+
+  class Person < Sequel::Model
+    one_to_many :posts, :eager=>[:tags]
+  end
+
+  class Post < Sequel::Model
+    many_to_one :person
+    one_to_many :replies
+    many_to_many :tags
+  end
+
+  class Tag < Sequel::Model
+    many_to_many :posts
+    many_to_many :replies
+  end
+
+  class Reply < Sequel::Model
+    many_to_one :person
+    many_to_one :post
+    many_to_many :tags
+  end
+
+  # Eager loading via .eager
+  Post.eager(:person).all
+
+  # eager is a dataset method, so it works with filters/orders/limits/etc.
+  Post.filter{|o| o.topic > 'M'}.order(:date).limit(5).eager(:person).all
+  
+  person = Person.first
+  # Eager loading via :eager (will eagerly load the tags for this person's posts)
+  person.posts
+  
+  # These are equivalent
+  Post.eager(:person, :tags).all
+  Post.eager(:person).eager(:tags).all
+  
+  # Cascading via .eager
+  Tag.eager(:posts=>:replies).all
+  
+  # Will also grab all associated posts' tags (because of :eager)
+  Reply.eager(:person=>:posts).all
+  
+  # No depth limit (other than memory/stack), and will also grab posts' tags
+  # Loads all people, their posts, their posts' tags, replies to those posts,
+  # the person for each reply, the tag for each reply, and all posts and
+  # replies that have that tag.  Uses a total of 8 queries.
+  Person.eager(:posts=>{:replies=>[:person, {:tags=>{:posts, :replies}}]}).all
+
+In addition to using eager, you can also use eager_graph, which will use a single query to get the object and all associated objects.  This may be necessary if you want to filter or order the result set based on columns in associated tables.  It works with cascading as well, the syntax is exactly the same.  Note that using eager_graph to eagerly load multiple *_to_many associations will cause the result set to be a cartesian product, so you should be very careful with your filters when using it in that case.
+
+=== Extending the underlying dataset
+
+The obvious way to add table-wide logic is to define class methods to the model class definition. That way you can define subsets of the underlying dataset, change the ordering, or perform actions on multiple records:
+
+  class Post < Sequel::Model
+    def self.posts_with_few_comments
+      filter{|o| o.num_comments < 30}
+    end
+
+    def self.clean_posts_with_few_comments
+      posts_with_few_comments.delete
+    end
+  end
+
+You can also implement table-wide logic by defining methods on the dataset:
+
+  class Post < Sequel::Model
+    def_dataset_method(:posts_with_few_comments) do
+      filter{|o| o.num_comments < 30}
+    end
+
+    def_dataset_method(:clean_posts_with_few_comments) do
+      posts_with_few_comments.delete
+    end
+  end
+
+This is the recommended way of implementing table-wide operations, and allows you to have access to your model API from filtered datasets as well:
+
+  Post.filter(:category => 'ruby').clean_posts_with_few_comments
+
+Sequel models also provide a short hand notation for filters:
+
+  class Post < Sequel::Model
+    subset(:posts_with_few_comments){|o| o.num_comments < 30}
+    subset :invisible, ~:visible
+  end
+
+=== Model Validations
+
+You can define a validate method for your model, which save
+will check before attempting to save the model in the database.
+If an attribute of the model isn't valid, you should add a error
+message for that attribute to the model object's errors. If an
+object has any errors added by the validate method, save will
+raise an error or return false depending on how it is configured.
+
+  class Post < Sequel::Model
+    def validate
+      errors[:name] << "can't be empty" if name.empty?
+      errors[:written_on] << "should be in the past" if written_on >= Time.now
+    end
+  end
+
diff -pruN 1.5.1-1/spec/adapters/ado_spec.rb 3.0.0-1/spec/adapters/ado_spec.rb
--- 1.5.1-1/spec/adapters/ado_spec.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/adapters/ado_spec.rb	2008-10-19 02:59:23.000000000 +0100
@@ -0,0 +1,46 @@
+require File.join(File.dirname(__FILE__), 'spec_helper.rb')
+
+unless defined?(ADO_DB)
+  ADO_DB = Sequel.ado(:host => 'MY_SQL_SERVER', :database => 'MyDB', :user => 'my_pwd', :password => 'my_usr')
+end
+
+context "An ADO dataset" do
+  before(:each) do
+    ADO_DB.create_table!(:items) { text :name }
+  end
+  
+  specify "should not raise exceptions when working with empty datasets" do
+    lambda {
+      ADO_DB[:items].all
+    }.should_not raise_error
+  end
+end
+
+context "An MSSQL dataset" do
+  before(:each) do
+    ADO_DB.create_table!(:items) { text :name }
+  end
+
+  specify "should assign a default name to anonymous columns" do
+    col = ADO_DB.fetch('SELECT COUNT(*) FROM items').columns[0]
+    col.to_s.should == '(no column name)'
+  end
+
+  specify "should support counting" do
+    ADO_DB[:items] << {:name => 'my name' }
+    ADO_DB[:items].count.should == 1
+  end
+
+  specify "should support first" do
+    ADO_DB[:items] << {:name => 'x' }
+    ADO_DB[:items] << {:name => 'y' }
+    ADO_DB[:items].first[:name].should == 'x'
+  end
+
+  specify "should support limit" do
+    3.times do
+      ADO_DB[:items] << {:name => 'my name' }
+    end
+    ADO_DB[:items].limit(2).all.size.should == 2
+  end
+end
diff -pruN 1.5.1-1/spec/adapters/firebird_spec.rb 3.0.0-1/spec/adapters/firebird_spec.rb
--- 1.5.1-1/spec/adapters/firebird_spec.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/adapters/firebird_spec.rb	2009-04-16 23:29:45.000000000 +0100
@@ -0,0 +1,403 @@
+require File.join(File.dirname(__FILE__), 'spec_helper.rb')
+
+unless defined?(FIREBIRD_DB)
+  FIREBIRD_URL = 'firebird://sysdba:masterkey@localhost/reality_spec' unless defined? FIREBIRD_URL
+  FIREBIRD_DB = Sequel.connect(ENV['SEQUEL_FB_SPEC_DB']||FIREBIRD_URL)
+end
+
+FIREBIRD_DB.create_table! :test do
+  varchar :name,  :size => 50
+  integer :val,   :index => true
+end
+
+FIREBIRD_DB.create_table! :test2 do
+  integer :val
+  timestamp :time_stamp
+end
+
+FIREBIRD_DB.create_table! :test3 do
+  integer :val
+  timestamp :time_stamp
+end
+
+FIREBIRD_DB.create_table! :test5 do
+  primary_key :xid
+  integer :val
+end
+
+FIREBIRD_DB.create_table! :test6 do
+  primary_key :xid
+  blob :val
+  String :val2
+  varchar :val3, :size=>200
+  text :val4
+end
+
+context "A Firebird database" do
+  before do
+    @db = FIREBIRD_DB
+  end
+
+  specify "should provide disconnect functionality" do
+    @db.tables
+    @db.pool.size.should == 1
+    @db.disconnect
+    @db.pool.size.should == 0
+  end
+
+  specify "should raise Sequel::Error on error" do
+    proc{@db << "SELECT 1 + 'a'"}.should raise_error(Sequel::Error)
+  end
+end
+
+context "A Firebird dataset" do
+  before do
+    @d = FIREBIRD_DB[:test]
+    @d.delete # remove all records
+  end
+
+  specify "should return the correct record count" do
+    @d.count.should == 0
+    @d << {:name => 'abc', :val => 123}
+    @d << {:name => 'abc', :val  => 456}
+    @d << {:name => 'def', :val => 789}
+    @d.count.should == 3
+  end
+
+  specify "should return the correct records" do
+    @d.to_a.should == []
+    @d << {:name => 'abc', :val => 123}
+    @d << {:name => 'abc', :val => 456}
+    @d << {:name => 'def', :val => 789}
+
+    @d.order(:val).to_a.should == [
+      {:name => 'abc', :val => 123},
+      {:name => 'abc', :val => 456},
+      {:name => 'def', :val => 789}
+    ]
+  end
+
+  specify "should update records correctly" do
+    @d << {:name => 'abc', :val => 123}
+    @d << {:name => 'abc', :val => 456}
+    @d << {:name => 'def', :val => 789}
+    @d.filter(:name => 'abc').update(:val => 530)
+
+    # the third record should stay the same
+    # floating-point precision bullshit
+    @d[:name => 'def'][:val].should == 789
+    @d.filter(:val => 530).count.should == 2
+  end
+
+  specify "should delete records correctly" do
+    @d << {:name => 'abc', :val => 123}
+    @d << {:name => 'abc', :val => 456}
+    @d << {:name => 'def', :val => 789}
+    @d.filter(:name => 'abc').delete
+
+    @d.count.should == 1
+    @d.first[:name].should == 'def'
+  end
+
+  specify "should be able to literalize booleans" do
+    proc {@d.literal(true)}.should_not raise_error
+    proc {@d.literal(false)}.should_not raise_error
+  end
+
+  specify "should quote columns and tables using double quotes if quoting identifiers" do
+    @d.quote_identifiers = true
+    @d.select(:name).sql.should == \
+      'SELECT "NAME" FROM "TEST"'
+
+    @d.select('COUNT(*)'.lit).sql.should == \
+      'SELECT COUNT(*) FROM "TEST"'
+
+    @d.select(:max[:val]).sql.should == \
+      'SELECT max("VAL") FROM "TEST"'
+
+    @d.select(:now[]).sql.should == \
+    'SELECT now() FROM "TEST"'
+
+    @d.select(:max[:items__val]).sql.should == \
+      'SELECT max("ITEMS"."VAL") FROM "TEST"'
+
+    @d.order(:name.desc).sql.should == \
+      'SELECT * FROM "TEST" ORDER BY "NAME" DESC'
+
+    @d.select('TEST.NAME AS item_:name'.lit).sql.should == \
+      'SELECT TEST.NAME AS item_:name FROM "TEST"'
+
+    @d.select('"NAME"'.lit).sql.should == \
+      'SELECT "NAME" FROM "TEST"'
+
+    @d.select('max(TEST."NAME") AS "max_:name"'.lit).sql.should == \
+      'SELECT max(TEST."NAME") AS "max_:name" FROM "TEST"'
+
+    @d.select(:test[:ABC, 'hello']).sql.should == \
+      "SELECT test(\"ABC\", 'hello') FROM \"TEST\""
+
+    @d.select(:test[:ABC__DEF, 'hello']).sql.should == \
+      "SELECT test(\"ABC\".\"DEF\", 'hello') FROM \"TEST\""
+
+    @d.select(:test[:ABC__DEF, 'hello'].as(:X2)).sql.should == \
+      "SELECT test(\"ABC\".\"DEF\", 'hello') AS \"X2\" FROM \"TEST\""
+
+    @d.insert_sql(:val => 333).should =~ \
+      /\AINSERT INTO "TEST" \("VAL"\) VALUES \(333\)( RETURNING NULL)?\z/
+
+    @d.insert_sql(:X => :Y).should =~ \
+      /\AINSERT INTO "TEST" \("X"\) VALUES \("Y"\)( RETURNING NULL)?\z/
+  end
+
+  specify "should quote fields correctly when reversing the order if quoting identifiers" do
+    @d.quote_identifiers = true
+    @d.reverse_order(:name).sql.should == \
+      'SELECT * FROM "TEST" ORDER BY "NAME" DESC'
+
+    @d.reverse_order(:name.desc).sql.should == \
+      'SELECT * FROM "TEST" ORDER BY "NAME" ASC'
+
+    @d.reverse_order(:name, :test.desc).sql.should == \
+      'SELECT * FROM "TEST" ORDER BY "NAME" DESC, "TEST" ASC'
+
+    @d.reverse_order(:name.desc, :test).sql.should == \
+      'SELECT * FROM "TEST" ORDER BY "NAME" ASC, "TEST" DESC'
+  end
+
+  specify "should support transactions" do
+    FIREBIRD_DB.transaction do
+      @d << {:name => 'abc', :val => 1}
+    end
+
+    @d.count.should == 1
+  end
+
+  specify "should have #transaction yield the connection" do
+    FIREBIRD_DB.transaction do |conn|
+      conn.should_not == nil
+    end
+  end
+
+  specify "should correctly rollback transactions" do
+    proc do
+      FIREBIRD_DB.transaction do
+        @d << {:name => 'abc', :val => 1}
+        raise RuntimeError, 'asdf'
+      end
+    end.should raise_error(RuntimeError)
+
+    @d.count.should == 0
+  end
+
+  specify "should handle returning inside of the block by committing" do
+    def FIREBIRD_DB.ret_commit
+      transaction do
+        self[:test] << {:name => 'abc'}
+        return
+        self[:test] << {:name => 'd'}
+      end
+    end
+    @d.count.should == 0
+    FIREBIRD_DB.ret_commit
+    @d.count.should == 1
+    FIREBIRD_DB.ret_commit
+    @d.count.should == 2
+    proc do
+      FIREBIRD_DB.transaction do
+        raise RuntimeError, 'asdf'
+      end
+    end.should raise_error(RuntimeError)
+
+    @d.count.should == 2
+  end
+
+  specify "should quote and upcase reserved keywords" do
+    @d = FIREBIRD_DB[:testing]
+    @d.quote_identifiers = true
+    @d.select(:select).sql.should == \
+      'SELECT "SELECT" FROM "TESTING"'
+  end
+end
+
+context "A Firebird dataset with a timestamp field" do
+  before do
+    @d = FIREBIRD_DB[:test3]
+    @d.delete
+  end
+
+  specify "should store milliseconds in time fields" do
+    t = Time.now
+    @d << {:val=>1, :time_stamp=>t}
+    @d.literal(@d[:val =>'1'][:time_stamp]).should == @d.literal(t)
+    @d[:val=>'1'][:time_stamp].usec.should == t.usec - t.usec % 100
+  end
+end
+
+context "A Firebird database" do
+  before do
+    @db = FIREBIRD_DB
+  end
+
+  specify "should allow us to name the sequences" do
+    g = Sequel::Schema::Generator.new(FIREBIRD_DB) do
+      primary_key :id, :sequence_name => "seq_test"
+    end
+    FIREBIRD_DB.send(:create_table_sql_list, :posts, *g.create_info).should == [[
+      "CREATE TABLE POSTS (ID integer PRIMARY KEY )",
+      "CREATE SEQUENCE SEQ_TEST",
+      "          CREATE TRIGGER BI_POSTS_ID for POSTS\n          ACTIVE BEFORE INSERT position 0\n          as               begin\n                if ((new.ID is null) or (new.ID = 0)) then\n                begin\n                  new.ID = next value for seq_test;\n                end\n              end\n\n"
+    ], "DROP SEQUENCE SEQ_TEST" ]
+  end
+
+  specify "should allow us to set the starting position for the sequences" do
+    g = Sequel::Schema::Generator.new(FIREBIRD_DB) do
+      primary_key :id, :sequence_start_position => 999
+    end
+    FIREBIRD_DB.send(:create_table_sql_list, :posts, *g.create_info).should == [[
+      "CREATE TABLE POSTS (ID integer PRIMARY KEY )",
+      "CREATE SEQUENCE SEQ_POSTS_ID",
+      "ALTER SEQUENCE SEQ_POSTS_ID RESTART WITH 999",
+      "          CREATE TRIGGER BI_POSTS_ID for POSTS\n          ACTIVE BEFORE INSERT position 0\n          as               begin\n                if ((new.ID is null) or (new.ID = 0)) then\n                begin\n                  new.ID = next value for seq_posts_id;\n                end\n              end\n\n"
+    ], "DROP SEQUENCE SEQ_POSTS_ID" ]
+  end
+
+  specify "should allow us to name and set the starting position for the sequences" do
+    g = Sequel::Schema::Generator.new(FIREBIRD_DB) do
+      primary_key :id, :sequence_name => "seq_test", :sequence_start_position => 999
+    end
+    FIREBIRD_DB.send(:create_table_sql_list, :posts, *g.create_info).should == [[
+      "CREATE TABLE POSTS (ID integer PRIMARY KEY )",
+      "CREATE SEQUENCE SEQ_TEST",
+      "ALTER SEQUENCE SEQ_TEST RESTART WITH 999",
+      "          CREATE TRIGGER BI_POSTS_ID for POSTS\n          ACTIVE BEFORE INSERT position 0\n          as               begin\n                if ((new.ID is null) or (new.ID = 0)) then\n                begin\n                  new.ID = next value for seq_test;\n                end\n              end\n\n"
+    ], "DROP SEQUENCE SEQ_TEST" ]
+  end
+
+  specify "should allow us to name the triggers" do
+    g = Sequel::Schema::Generator.new(FIREBIRD_DB) do
+      primary_key :id, :trigger_name => "trig_test"
+    end
+    FIREBIRD_DB.send(:create_table_sql_list, :posts, *g.create_info).should == [[
+      "CREATE TABLE POSTS (ID integer PRIMARY KEY )",
+      "CREATE SEQUENCE SEQ_POSTS_ID",
+      "          CREATE TRIGGER TRIG_TEST for POSTS\n          ACTIVE BEFORE INSERT position 0\n          as               begin\n                if ((new.ID is null) or (new.ID = 0)) then\n                begin\n                  new.ID = next value for seq_posts_id;\n                end\n              end\n\n"
+    ], "DROP SEQUENCE SEQ_POSTS_ID" ]
+  end
+
+  specify "should allow us to not create the sequence" do
+    g = Sequel::Schema::Generator.new(FIREBIRD_DB) do
+      primary_key :id, :create_sequence => false
+    end
+    FIREBIRD_DB.send(:create_table_sql_list, :posts, *g.create_info).should == [[
+      "CREATE TABLE POSTS (ID integer PRIMARY KEY )",
+      "          CREATE TRIGGER BI_POSTS_ID for POSTS\n          ACTIVE BEFORE INSERT position 0\n          as               begin\n                if ((new.ID is null) or (new.ID = 0)) then\n                begin\n                  new.ID = next value for seq_posts_id;\n                end\n              end\n\n"
+    ], nil]
+  end
+
+  specify "should allow us to not create the trigger" do
+    g = Sequel::Schema::Generator.new(FIREBIRD_DB) do
+      primary_key :id, :create_trigger => false
+    end
+    FIREBIRD_DB.send(:create_table_sql_list, :posts, *g.create_info).should == [[
+      "CREATE TABLE POSTS (ID integer PRIMARY KEY )",
+      "CREATE SEQUENCE SEQ_POSTS_ID",
+    ], "DROP SEQUENCE SEQ_POSTS_ID"]
+  end
+
+  specify "should allow us to not create either the sequence nor the trigger" do
+    g = Sequel::Schema::Generator.new(FIREBIRD_DB) do
+      primary_key :id, :create_sequence => false, :create_trigger => false
+    end
+    FIREBIRD_DB.send(:create_table_sql_list, :posts, *g.create_info).should == [[
+      "CREATE TABLE POSTS (ID integer PRIMARY KEY )"
+    ], nil]
+  end
+
+  specify "should support column operations" do
+    @db.create_table!(:test2){varchar :name, :size => 50; integer :val}
+    @db[:test2] << {}
+    @db[:test2].columns.should == [:name, :val]
+
+    @db.add_column :test2, :xyz, :varchar, :size => 50
+    @db[:test2].columns.should == [:name, :val, :xyz]
+
+    @db[:test2].columns.should == [:name, :val, :xyz]
+    @db.drop_column :test2, :xyz
+
+    @db[:test2].columns.should == [:name, :val]
+
+    @db[:test2].delete
+    @db.add_column :test2, :xyz, :varchar, :default => '000', :size => 50#, :create_domain => 'xyz_varchar'
+    @db[:test2] << {:name => 'mmm', :val => 111, :xyz => 'qqqq'}
+
+    @db[:test2].columns.should == [:name, :val, :xyz]
+    @db.rename_column :test2, :xyz, :zyx
+    @db[:test2].columns.should == [:name, :val, :zyx]
+    @db[:test2].first[:zyx].should == 'qqqq'
+
+    @db.add_column :test2, :xyz, :decimal, :elements => [12, 2]
+    @db[:test2].delete
+    @db[:test2] << {:name => 'mmm', :val => 111, :xyz => 56.4}
+    @db.set_column_type :test2, :xyz, :varchar, :size => 50
+
+    @db[:test2].first[:xyz].should == "56.40"
+  end
+
+  specify "should allow us to retrieve the primary key for a table" do
+    @db.create_table!(:test2){primary_key :id}
+    @db.primary_key(:test2).should == ["id"]
+  end
+end
+
+context "Postgres::Dataset#insert" do
+  before do
+    @ds = FIREBIRD_DB[:test5]
+    @ds.delete
+  end
+
+  specify "should using call insert_returning_sql" do
+#    @ds.should_receive(:single_value).once.with(:sql=>'INSERT INTO TEST5 (VAL) VALUES (10) RETURNING XID', :server=> :default)
+    @ds.should_receive(:single_value).once
+    @ds.insert(:val=>10)
+  end
+
+  specify "should have insert_returning_sql use the RETURNING keyword" do
+    @ds.insert_returning_sql(:XID, :val=>10).should == "INSERT INTO TEST5 (VAL) VALUES (10) RETURNING XID"
+    @ds.insert_returning_sql('*'.lit, :val=>10).should == "INSERT INTO TEST5 (VAL) VALUES (10) RETURNING *"
+    @ds.insert_returning_sql('NULL'.lit, :val=>10).should == "INSERT INTO TEST5 (VAL) VALUES (10) RETURNING NULL"
+  end
+
+  specify "should correctly return the inserted record's primary key value" do
+    value1 = 10
+    id1 = @ds.insert(:val=>value1)
+    @ds.first(:XID=>id1)[:val].should == value1
+    value2 = 20
+    id2 = @ds.insert(:val=>value2)
+    @ds.first(:XID=>id2)[:val].should == value2
+  end
+
+  specify "should return nil if the table has no primary key" do
+    ds = FIREBIRD_DB[:test]
+    ds.delete
+    ds.insert(:name=>'a').should == nil
+  end
+end
+
+context "Postgres::Dataset#insert" do
+  before do
+    @ds = FIREBIRD_DB[:test6]
+    @ds.delete
+  end
+
+  specify "should insert and retrieve a blob successfully" do
+    value1 = "\1\2\2\2\2222\2\2\2"
+    value2 = "abcd"
+    value3 = "efgh"
+    value4 = "ijkl"
+    id1 = @ds.insert(:val=>value1, :val2=>value2, :val3=>value3, :val4=>value4)
+    @ds.first(:XID=>id1)[:val].should == value1
+    @ds.first(:XID=>id1)[:val2].should == value2
+    @ds.first(:XID=>id1)[:val3].should == value3
+    @ds.first(:XID=>id1)[:val4].should == value4
+  end
+end
diff -pruN 1.5.1-1/spec/adapters/informix_spec.rb 3.0.0-1/spec/adapters/informix_spec.rb
--- 1.5.1-1/spec/adapters/informix_spec.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/adapters/informix_spec.rb	2009-03-25 15:18:58.000000000 +0000
@@ -0,0 +1,96 @@
+require File.join(File.dirname(__FILE__), 'spec_helper.rb')
+
+unless defined?(INFORMIX_DB)
+  INFORMIX_DB = Sequel.connect('informix://localhost/mydb')
+end
+
+if INFORMIX_DB.table_exists?(:test)
+  INFORMIX_DB.drop_table :test
+end
+INFORMIX_DB.create_table :test do
+  text :name
+  integer :value
+  
+  index :value
+end
+
+context "A Informix database" do
+  specify "should provide disconnect functionality" do
+    INFORMIX_DB.execute("select user from dual")
+    INFORMIX_DB.pool.size.should == 1
+    INFORMIX_DB.disconnect
+    INFORMIX_DB.pool.size.should == 0
+  end
+end
+
+context "A Informix dataset" do
+  before do
+    @d = INFORMIX_DB[:test]
+    @d.delete # remove all records
+  end
+  
+  specify "should return the correct record count" do
+    @d.count.should == 0
+    @d << {:name => 'abc', :value => 123}
+    @d << {:name => 'abc', :value => 456}
+    @d << {:name => 'def', :value => 789}
+    @d.count.should == 3
+  end
+  
+  specify "should return the correct records" do
+    @d.to_a.should == []
+    @d << {:name => 'abc', :value => 123}
+    @d << {:name => 'abc', :value => 456}
+    @d << {:name => 'def', :value => 789}
+
+    @d.order(:value).to_a.should == [
+      {:name => 'abc', :value => 123},
+      {:name => 'abc', :value => 456},
+      {:name => 'def', :value => 789}
+    ]
+  end
+  
+  specify "should update records correctly" do
+    @d << {:name => 'abc', :value => 123}
+    @d << {:name => 'abc', :value => 456}
+    @d << {:name => 'def', :value => 789}
+    @d.filter(:name => 'abc').update(:value => 530)
+    
+    # the third record should stay the same
+    # floating-point precision bullshit
+    @d[:name => 'def'][:value].should == 789
+    @d.filter(:value => 530).count.should == 2
+  end
+  
+  specify "should delete records correctly" do
+    @d << {:name => 'abc', :value => 123}
+    @d << {:name => 'abc', :value => 456}
+    @d << {:name => 'def', :value => 789}
+    @d.filter(:name => 'abc').delete
+    
+    @d.count.should == 1
+    @d.first[:name].should == 'def'
+  end
+  
+  specify "should be able to literalize booleans" do
+    proc {@d.literal(true)}.should_not raise_error
+    proc {@d.literal(false)}.should_not raise_error
+  end
+  
+  specify "should support transactions" do
+    INFORMIX_DB.transaction do
+      @d << {:name => 'abc', :value => 1}
+    end
+
+    @d.count.should == 1
+  end
+  
+  specify "should support #first and #last" do
+    @d << {:name => 'abc', :value => 123}
+    @d << {:name => 'abc', :value => 456}
+    @d << {:name => 'def', :value => 789}
+    
+    @d.order(:value).first.should == {:name => 'abc', :value => 123}
+    @d.order(:value).last.should == {:name => 'def', :value => 789}
+  end
+end
diff -pruN 1.5.1-1/spec/adapters/mysql_spec.rb 3.0.0-1/spec/adapters/mysql_spec.rb
--- 1.5.1-1/spec/adapters/mysql_spec.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/adapters/mysql_spec.rb	2009-04-20 18:39:19.000000000 +0100
@@ -0,0 +1,874 @@
+require File.join(File.dirname(__FILE__), 'spec_helper.rb')
+
+unless defined?(MYSQL_USER)
+  MYSQL_USER = 'root'
+end
+unless defined?(MYSQL_DB)
+  MYSQL_URL = (ENV['SEQUEL_MY_SPEC_DB']||"mysql://#{MYSQL_USER}@localhost/sandbox") unless defined? MYSQL_URL
+  MYSQL_DB = Sequel.connect(MYSQL_URL)
+end
+unless defined?(MYSQL_SOCKET_FILE)
+  MYSQL_SOCKET_FILE = '/tmp/mysql.sock'
+end
+
+MYSQL_URI = URI.parse(MYSQL_DB.uri)
+
+MYSQL_DB.create_table! :test2 do
+  text :name
+  integer :value
+end
+def MYSQL_DB.sqls
+  (@sqls ||= [])
+end
+logger = Object.new
+def logger.method_missing(m, msg)
+  MYSQL_DB.sqls << msg
+end
+MYSQL_DB.logger = logger
+MYSQL_DB.drop_table(:items) rescue nil
+MYSQL_DB.drop_table(:dolls) rescue nil
+MYSQL_DB.drop_table(:booltest) rescue nil
+
+if MYSQL_DB.class.adapter_scheme == :do
+  SQL_BEGIN = 'Transaction.begin'
+  SQL_ROLLBACK = 'Transaction.rollback'
+  SQL_COMMIT = 'Transaction.commit'
+else
+  SQL_BEGIN = 'BEGIN'
+  SQL_ROLLBACK = 'ROLLBACK'
+  SQL_COMMIT = 'COMMIT'
+end
+
+context "MySQL", '#create_table' do
+  before do
+    @db = MYSQL_DB
+    MYSQL_DB.sqls.clear
+  end
+  after do
+    @db.drop_table(:dolls) rescue nil
+  end
+  
+  specify "should allow to specify options for MySQL" do
+    @db.create_table(:dolls, :engine => 'MyISAM', :charset => 'latin2'){text :name}
+    @db.sqls.should == ["CREATE TABLE dolls (name text) ENGINE=MyISAM DEFAULT CHARSET=latin2"]
+  end
+  
+  specify "should create a temporary table" do
+    @db.create_table(:tmp_dolls, :temp => true, :engine => 'MyISAM', :charset => 'latin2'){text :name}
+    @db.sqls.should == ["CREATE TEMPORARY TABLE tmp_dolls (name text) ENGINE=MyISAM DEFAULT CHARSET=latin2"]
+  end
+end
+
+context "A MySQL database" do
+  before do
+    @db = MYSQL_DB
+    @db.create_table(:booltest){TrueClass :value}
+  end
+  after do
+    Sequel.convert_tinyint_to_bool = true
+    @db.drop_table(:booltest)
+  end
+
+  specify "should provide the server version" do
+    @db.server_version.should >= 40000
+  end
+
+  specify "should correctly parse the schema" do
+    @db.schema(:booltest, :reload=>true).should == [[:value, {:type=>:boolean, :allow_null=>true, :primary_key=>false, :default=>nil, :db_type=>"tinyint(4)"}]]
+    
+    Sequel.convert_tinyint_to_bool = false
+    @db.schema(:booltest, :reload=>true).should == [[:value, {:type=>:integer, :allow_null=>true, :primary_key=>false, :default=>nil, :db_type=>"tinyint(4)"}]]
+  end
+  
+  specify "should accept and return tinyints as bools or integers when configured to do so" do
+    MYSQL_DB[:booltest].delete
+    MYSQL_DB[:booltest] << {:value=>true}
+    MYSQL_DB[:booltest].all.should == [{:value=>true}]
+    MYSQL_DB[:booltest].delete
+    MYSQL_DB[:booltest] << {:value=>false}
+    MYSQL_DB[:booltest].all.should == [{:value=>false}]
+    
+    Sequel.convert_tinyint_to_bool = false
+    MYSQL_DB[:booltest].delete
+    MYSQL_DB[:booltest] << {:value=>true}
+    MYSQL_DB[:booltest].all.should == [{:value=>1}]
+    MYSQL_DB[:booltest].delete
+    MYSQL_DB[:booltest] << {:value=>false}
+    MYSQL_DB[:booltest].all.should == [{:value=>0}]
+    
+    MYSQL_DB[:booltest].delete
+    MYSQL_DB[:booltest] << {:value=>1}
+    MYSQL_DB[:booltest].all.should == [{:value=>1}]
+    MYSQL_DB[:booltest].delete
+    MYSQL_DB[:booltest] << {:value=>0}
+    MYSQL_DB[:booltest].all.should == [{:value=>0}]
+  end
+end
+
+context "A MySQL dataset" do
+  before do
+    MYSQL_DB.create_table(:items){String :name; Integer :value}
+    @d = MYSQL_DB[:items]
+    MYSQL_DB.sqls.clear
+  end
+  after do
+    MYSQL_DB.drop_table(:items)
+  end
+  
+  specify "should quote columns and tables using back-ticks if quoting identifiers" do
+    @d.quote_identifiers = true
+    @d.select(:name).sql.should == \
+      'SELECT `name` FROM `items`'
+      
+    @d.select('COUNT(*)'.lit).sql.should == \
+      'SELECT COUNT(*) FROM `items`'
+
+    @d.select(:max.sql_function(:value)).sql.should == \
+      'SELECT max(`value`) FROM `items`'
+      
+    @d.select(:NOW.sql_function).sql.should == \
+    'SELECT NOW() FROM `items`'
+
+    @d.select(:max.sql_function(:items__value)).sql.should == \
+      'SELECT max(`items`.`value`) FROM `items`'
+
+    @d.order(:name.desc).sql.should == \
+      'SELECT * FROM `items` ORDER BY `name` DESC'
+
+    @d.select('items.name AS item_name'.lit).sql.should == \
+      'SELECT items.name AS item_name FROM `items`'
+      
+    @d.select('`name`'.lit).sql.should == \
+      'SELECT `name` FROM `items`'
+
+    @d.select('max(items.`name`) AS `max_name`'.lit).sql.should == \
+      'SELECT max(items.`name`) AS `max_name` FROM `items`'
+      
+    @d.select(:test.sql_function(:abc, 'hello')).sql.should == \
+      "SELECT test(`abc`, 'hello') FROM `items`"
+
+    @d.select(:test.sql_function(:abc__def, 'hello')).sql.should == \
+      "SELECT test(`abc`.`def`, 'hello') FROM `items`"
+
+    @d.select(:test.sql_function(:abc__def, 'hello').as(:x2)).sql.should == \
+      "SELECT test(`abc`.`def`, 'hello') AS `x2` FROM `items`"
+
+    @d.insert_sql(:value => 333).should == \
+      'INSERT INTO `items` (`value`) VALUES (333)'
+
+    @d.insert_sql(:x => :y).should == \
+      'INSERT INTO `items` (`x`) VALUES (`y`)'
+  end
+  
+  specify "should quote fields correctly when reversing the order" do
+    @d.quote_identifiers = true
+    @d.reverse_order(:name).sql.should == \
+      'SELECT * FROM `items` ORDER BY `name` DESC'
+
+    @d.reverse_order(:name.desc).sql.should == \
+      'SELECT * FROM `items` ORDER BY `name` ASC'
+
+    @d.reverse_order(:name, :test.desc).sql.should == \
+      'SELECT * FROM `items` ORDER BY `name` DESC, `test` ASC'
+
+    @d.reverse_order(:name.desc, :test).sql.should == \
+      'SELECT * FROM `items` ORDER BY `name` ASC, `test` DESC'
+  end
+  
+  specify "should support ORDER clause in UPDATE statements" do
+    @d.order(:name).update_sql(:value => 1).should == \
+      'UPDATE items SET value = 1 ORDER BY name'
+  end
+  
+  specify "should support LIMIT clause in UPDATE statements" do
+    @d.limit(10).update_sql(:value => 1).should == \
+      'UPDATE items SET value = 1 LIMIT 10'
+  end
+  
+  specify "should support regexps" do
+    @d << {:name => 'abc', :value => 1}
+    @d << {:name => 'bcd', :value => 2}
+    @d.filter(:name => /bc/).count.should == 2
+    @d.filter(:name => /^bc/).count.should == 1
+  end
+  
+  specify "should correctly literalize strings with comment backslashes in them" do
+    @d.delete
+    proc {@d << {:name => ':\\'}}.should_not raise_error
+    
+    @d.first[:name].should == ':\\'
+  end
+end
+
+context "MySQL datasets" do
+  before do
+    @d = MYSQL_DB[:orders]
+  end
+  
+  specify "should correctly quote column references" do
+    @d.quote_identifiers = true
+    market = 'ICE'
+    ack_stamp = Time.now - 15 * 60 # 15 minutes ago
+    @d.select(:market, :minute.sql_function(:from_unixtime.sql_function(:ack)).as(:minute)).
+      where{|o|(:ack.sql_number > ack_stamp) & {:market => market}}.
+      group_by(:minute.sql_function(:from_unixtime.sql_function(:ack))).sql.should == \
+      "SELECT `market`, minute(from_unixtime(`ack`)) AS `minute` FROM `orders` WHERE ((`ack` > #{@d.literal(ack_stamp)}) AND (`market` = 'ICE')) GROUP BY minute(from_unixtime(`ack`))"
+  end
+end
+
+context "MySQL join expressions" do
+  before do
+    @ds = MYSQL_DB[:nodes]
+    @ds.db.meta_def(:server_version) {50014}
+  end
+
+  specify "should raise error for :full_outer join requests." do
+    lambda{@ds.join_table(:full_outer, :nodes)}.should raise_error(Sequel::Error)
+  end
+  specify "should support natural left joins" do
+    @ds.join_table(:natural_left, :nodes).sql.should == \
+      'SELECT * FROM nodes NATURAL LEFT JOIN nodes'
+  end
+  specify "should support natural right joins" do
+    @ds.join_table(:natural_right, :nodes).sql.should == \
+      'SELECT * FROM nodes NATURAL RIGHT JOIN nodes'
+  end
+  specify "should support natural left outer joins" do
+    @ds.join_table(:natural_left_outer, :nodes).sql.should == \
+      'SELECT * FROM nodes NATURAL LEFT OUTER JOIN nodes'
+  end
+  specify "should support natural right outer joins" do
+    @ds.join_table(:natural_right_outer, :nodes).sql.should == \
+      'SELECT * FROM nodes NATURAL RIGHT OUTER JOIN nodes'
+  end
+  specify "should support natural inner joins" do
+    @ds.join_table(:natural_inner, :nodes).sql.should == \
+      'SELECT * FROM nodes NATURAL LEFT JOIN nodes'
+  end
+  specify "should support cross joins" do
+    @ds.join_table(:cross, :nodes).sql.should == \
+      'SELECT * FROM nodes CROSS JOIN nodes'
+  end
+  specify "should support cross joins as inner joins if conditions are used" do
+    @ds.join_table(:cross, :nodes, :id=>:id).sql.should == \
+      'SELECT * FROM nodes INNER JOIN nodes ON (nodes.id = nodes.id)'
+  end
+  specify "should support straight joins (force left table to be read before right)" do
+    @ds.join_table(:straight, :nodes).sql.should == \
+      'SELECT * FROM nodes STRAIGHT_JOIN nodes'
+  end
+  specify "should support natural joins on multiple tables." do
+    @ds.join_table(:natural_left_outer, [:nodes, :branches]).sql.should == \
+      'SELECT * FROM nodes NATURAL LEFT OUTER JOIN (nodes, branches)'
+  end
+  specify "should support straight joins on multiple tables." do
+    @ds.join_table(:straight, [:nodes,:branches]).sql.should == \
+      'SELECT * FROM nodes STRAIGHT_JOIN (nodes, branches)'
+  end
+end
+
+context "Joined MySQL dataset" do
+  before do
+    @ds = MYSQL_DB[:nodes]
+  end
+  
+  specify "should quote fields correctly" do
+    @ds.quote_identifiers = true
+    @ds.join(:attributes, :node_id => :id).sql.should == \
+      "SELECT * FROM `nodes` INNER JOIN `attributes` ON (`attributes`.`node_id` = `nodes`.`id`)"
+  end
+  
+  specify "should allow a having clause on ungrouped datasets" do
+    proc {@ds.having('blah')}.should_not raise_error
+
+    @ds.having('blah').sql.should == \
+      "SELECT * FROM nodes HAVING (blah)"
+  end
+  
+  specify "should put a having clause before an order by clause" do
+    @ds.order(:aaa).having(:bbb => :ccc).sql.should == \
+      "SELECT * FROM nodes HAVING (bbb = ccc) ORDER BY aaa"
+  end
+end
+
+context "A MySQL database" do
+  before do
+    @db = MYSQL_DB
+  end
+
+  specify "should support add_column operations" do
+    @db.add_column :test2, :xyz, :text
+    
+    @db[:test2].columns.should == [:name, :value, :xyz]
+    @db[:test2] << {:name => 'mmm', :value => 111, :xyz => '000'}
+    @db[:test2].first[:xyz].should == '000'
+  end
+  
+  specify "should support drop_column operations" do
+    @db[:test2].columns.should == [:name, :value, :xyz]
+    @db.drop_column :test2, :xyz
+    
+    @db[:test2].columns.should == [:name, :value]
+  end
+  
+  specify "should support rename_column operations" do
+    @db[:test2].delete
+    @db.add_column :test2, :xyz, :text
+    @db[:test2] << {:name => 'mmm', :value => 111, :xyz => 'qqqq'}
+
+    @db[:test2].columns.should == [:name, :value, :xyz]
+    @db.rename_column :test2, :xyz, :zyx, :type => :text
+    @db[:test2].columns.should == [:name, :value, :zyx]
+    @db[:test2].first[:zyx].should == 'qqqq'
+  end
+  
+  specify "should support rename_column operations with types like varchar(255)" do
+    @db[:test2].delete
+    @db.add_column :test2, :tre, :text
+    @db[:test2] << {:name => 'mmm', :value => 111, :tre => 'qqqq'}
+
+    @db[:test2].columns.should == [:name, :value, :zyx, :tre]
+    @db.rename_column :test2, :tre, :ert, :type => :varchar, :size=>255
+    @db[:test2].columns.should == [:name, :value, :zyx, :ert]
+    @db[:test2].first[:ert].should == 'qqqq'
+  end
+  
+  specify "should support set_column_type operations" do
+    @db.add_column :test2, :xyz, :float
+    @db[:test2].delete
+    @db[:test2] << {:name => 'mmm', :value => 111, :xyz => 56.78}
+    @db.set_column_type :test2, :xyz, :integer
+    
+    @db[:test2].first[:xyz].should == 57
+  end
+  
+  specify "should support add_index" do
+    @db.add_index :test2, :value
+  end
+  
+  specify "should support drop_index" do
+    @db.drop_index :test2, :value
+  end
+  
+  specify "should support add_foreign_key" do
+    @db.alter_table :test2 do
+      add_foreign_key :value2, :test2, :key=>:value
+    end
+    @db[:test2].columns.should == [:name, :value, :zyx, :ert, :xyz, :value2]
+  end
+end  
+
+context "A MySQL database with table options" do
+  before do
+    @options = {:engine=>'MyISAM', :charset=>'latin1', :collate => 'latin1_swedish_ci'}
+    
+    Sequel::MySQL.default_engine = 'InnoDB'
+    Sequel::MySQL.default_charset = 'utf8'
+    Sequel::MySQL.default_collate = 'utf8_general_ci'    
+    
+    @db = MYSQL_DB
+    @db.drop_table(:items) rescue nil
+    
+    MYSQL_DB.sqls.clear
+  end
+  after do
+    @db.drop_table(:items) rescue nil
+    Sequel::MySQL.default_engine = nil
+    Sequel::MySQL.default_charset = nil
+    Sequel::MySQL.default_collate = nil
+  end
+  
+  specify "should allow to pass custom options (engine, charset, collate) for table creation" do
+    @db.create_table(:items, @options){Integer :size; text :name}
+    @db.sqls.should == ["CREATE TABLE items (size integer, name text) ENGINE=MyISAM DEFAULT CHARSET=latin1 DEFAULT COLLATE=latin1_swedish_ci"]
+  end
+  
+  specify "should use default options if specified (engine, charset, collate) for table creation" do
+    @db.create_table(:items){Integer :size; text :name}
+    @db.sqls.should == ["CREATE TABLE items (size integer, name text) ENGINE=InnoDB DEFAULT CHARSET=utf8 DEFAULT COLLATE=utf8_general_ci"]
+  end
+  
+  specify "should not use default if option has a nil value" do
+    @db.create_table(:items, :engine=>nil, :charset=>nil, :collate=>nil){Integer :size; text :name}
+    @db.sqls.should == ["CREATE TABLE items (size integer, name text)"]
+  end
+end
+
+context "A MySQL database" do
+  before do
+    @db = MYSQL_DB
+    @db.drop_table(:items) rescue nil
+    MYSQL_DB.sqls.clear
+  end
+  after do
+    @db.drop_table(:items) rescue nil
+  end
+  
+  specify "should support defaults for boolean columns" do
+    @db.create_table(:items){TrueClass :active1, :default=>true; FalseClass :active2, :default => false}
+    @db.sqls.should == ["CREATE TABLE items (active1 tinyint DEFAULT 1, active2 tinyint DEFAULT 0)"]
+  end
+  
+  specify "should correctly format CREATE TABLE statements with foreign keys" do
+    @db.create_table(:items){Integer :id; foreign_key :p_id, :items, :key => :id, :null => false, :on_delete => :cascade}
+    @db.sqls.should == ["CREATE TABLE items (id integer, p_id integer NOT NULL, FOREIGN KEY (p_id) REFERENCES items(id) ON DELETE CASCADE)"]
+  end
+  
+  specify "should correctly format ALTER TABLE statements with foreign keys" do
+    @db.create_table(:items){Integer :id}
+    @db.alter_table(:items){add_foreign_key :p_id, :users, :key => :id, :null => false, :on_delete => :cascade}
+    @db.sqls.should == ["CREATE TABLE items (id integer)", "ALTER TABLE items ADD COLUMN p_id integer NOT NULL", "ALTER TABLE items ADD FOREIGN KEY (p_id) REFERENCES users(id) ON DELETE CASCADE"]
+  end
+  
+  specify "should have rename_column support keep existing options" do
+    @db.create_table(:items){Integer :id, :null=>false, :default=>5}
+    @db.alter_table(:items){rename_column :id, :nid}
+    @db.sqls.should == ["CREATE TABLE items (id integer NOT NULL DEFAULT 5)", "DESCRIBE items", "ALTER TABLE items CHANGE COLUMN id nid int(11) NOT NULL DEFAULT 5"]
+    @db[:items].insert
+    @db[:items].all.should == [{:nid=>5}]
+    proc{@db[:items].insert(:nid=>nil)}.should raise_error(Sequel::DatabaseError)
+  end
+  
+  specify "should have set_column_type support keep existing options" do
+    @db.create_table(:items){Integer :id, :null=>false, :default=>5}
+    @db.alter_table(:items){set_column_type :id, Bignum}
+    @db.sqls.should == ["CREATE TABLE items (id integer NOT NULL DEFAULT 5)", "DESCRIBE items", "ALTER TABLE items CHANGE COLUMN id id bigint NOT NULL DEFAULT 5"]
+    @db[:items].insert
+    @db[:items].all.should == [{:id=>5}]
+    proc{@db[:items].insert(:id=>nil)}.should raise_error(Sequel::DatabaseError)
+    @db[:items].delete
+    @db[:items].insert(2**40)
+    @db[:items].all.should == [{:id=>2**40}]
+  end
+  
+  specify "should have set_column_default support keep existing options" do
+    @db.create_table(:items){Integer :id, :null=>false, :default=>5}
+    @db.alter_table(:items){set_column_default :id, 6}
+    @db.sqls.should == ["CREATE TABLE items (id integer NOT NULL DEFAULT 5)", "DESCRIBE items", "ALTER TABLE items CHANGE COLUMN id id int(11) NOT NULL DEFAULT 6"]
+    @db[:items].insert
+    @db[:items].all.should == [{:id=>6}]
+    proc{@db[:items].insert(:id=>nil)}.should raise_error(Sequel::DatabaseError)
+  end
+  
+  specify "should have set_column_allow_null support keep existing options" do
+    @db.create_table(:items){Integer :id, :null=>false, :default=>5}
+    @db.alter_table(:items){set_column_allow_null :id, true}
+    @db.sqls.should == ["CREATE TABLE items (id integer NOT NULL DEFAULT 5)", "DESCRIBE items", "ALTER TABLE items CHANGE COLUMN id id int(11) NULL DEFAULT 5"]
+    @db[:items].insert
+    @db[:items].all.should == [{:id=>5}]
+    proc{@db[:items].insert(:id=>nil)}.should_not
+  end
+  
+  specify "should accept repeated raw sql statements using Database#<<" do
+    @db.create_table(:items){String :name; Integer :value}
+    @db << 'DELETE FROM items'
+    @db[:items].count.should == 0
+    
+    @db << "INSERT INTO items (name, value) VALUES ('tutu', 1234)"
+    @db[:items].first.should == {:name => 'tutu', :value => 1234}
+    
+    @db << 'DELETE FROM items'
+    @db[:items].first.should == nil
+  end
+  
+  specify "should handle multiple select statements at once" do
+    @db.create_table(:items){String :name; Integer :value}
+    @db[:items].delete
+    @db[:items].insert(:name => 'tutu', :value => 1234)
+    @db["SELECT * FROM items; SELECT * FROM items"].all.should == \
+      [{:name => 'tutu', :value => 1234}, {:name => 'tutu', :value => 1234}]
+  end
+end  
+
+# Socket tests should only be run if the MySQL server is on localhost
+if %w'localhost 127.0.0.1 ::1'.include?(MYSQL_URI.host) and MYSQL_DB.class.adapter_scheme == :mysql
+  context "A MySQL database" do
+    specify "should accept a socket option" do
+      db = Sequel.mysql(MYSQL_DB.opts[:database], :host => 'localhost', :user => MYSQL_DB.opts[:user], :password => MYSQL_DB.opts[:password], :socket => MYSQL_SOCKET_FILE)
+      proc {db.test_connection}.should_not raise_error
+    end
+    
+    specify "should accept a socket option without host option" do
+      db = Sequel.mysql(MYSQL_DB.opts[:database], :user => MYSQL_DB.opts[:user], :password => MYSQL_DB.opts[:password], :socket => MYSQL_SOCKET_FILE)
+      proc {db.test_connection}.should_not raise_error
+    end
+    
+    specify "should fail to connect with invalid socket" do
+      db = Sequel.mysql(MYSQL_DB.opts[:database], :user => MYSQL_DB.opts[:user], :password => MYSQL_DB.opts[:password], :socket =>'blah')
+      proc {db.test_connection}.should raise_error
+    end
+  end
+end
+
+context "A grouped MySQL dataset" do
+  before do
+    MYSQL_DB[:test2].delete
+    MYSQL_DB[:test2] << {:name => '11', :value => 10}
+    MYSQL_DB[:test2] << {:name => '11', :value => 20}
+    MYSQL_DB[:test2] << {:name => '11', :value => 30}
+    MYSQL_DB[:test2] << {:name => '12', :value => 10}
+    MYSQL_DB[:test2] << {:name => '12', :value => 20}
+    MYSQL_DB[:test2] << {:name => '13', :value => 10}
+  end
+  
+  specify "should return the correct count for raw sql query" do
+    ds = MYSQL_DB["select name FROM test2 WHERE name = '11' GROUP BY name"]
+    ds.count.should == 1
+  end
+  
+  specify "should return the correct count for a normal dataset" do
+    ds = MYSQL_DB[:test2].select(:name).where(:name => '11').group(:name)
+    ds.count.should == 1
+  end
+end
+
+context "A MySQL database" do
+  before do
+    @db = MYSQL_DB
+    @db.drop_table(:posts) rescue nil
+    @db.sqls.clear
+  end
+  after do
+    @db.drop_table(:posts) rescue nil
+  end
+  
+  specify "should support fulltext indexes and full_text_search" do
+    @db.create_table(:posts){text :title; text :body; full_text_index :title; full_text_index [:title, :body]}
+    @db.sqls.should == [
+      "CREATE TABLE posts (title text, body text)",
+      "CREATE FULLTEXT INDEX posts_title_index ON posts (title)",
+      "CREATE FULLTEXT INDEX posts_title_body_index ON posts (title, body)"
+    ]
+    
+    @db[:posts].insert(:title=>'ruby rails', :body=>'y')
+    @db[:posts].insert(:title=>'sequel', :body=>'ruby')
+    @db[:posts].insert(:title=>'ruby scooby', :body=>'x')
+    @db.sqls.clear
+
+    @db[:posts].full_text_search(:title, 'rails').all.should == [{:title=>'ruby rails', :body=>'y'}]
+    @db[:posts].full_text_search([:title, :body], ['sequel', 'ruby']).all.should == [{:title=>'sequel', :body=>'ruby'}]
+    @db[:posts].full_text_search(:title, '+ruby -rails', :boolean => true).all.should == [{:title=>'ruby scooby', :body=>'x'}]
+    @db.sqls.should == [
+      "SELECT * FROM posts WHERE (MATCH (title) AGAINST ('rails'))",
+      "SELECT * FROM posts WHERE (MATCH (title, body) AGAINST ('sequel ruby'))",
+      "SELECT * FROM posts WHERE (MATCH (title) AGAINST ('+ruby -rails' IN BOOLEAN MODE))"]
+  end
+
+  specify "should support spatial indexes" do
+    @db.create_table(:posts){point :geom, :null=>false; spatial_index [:geom]}
+    @db.sqls.should == [
+      "CREATE TABLE posts (geom point NOT NULL)",
+      "CREATE SPATIAL INDEX posts_geom_index ON posts (geom)"
+    ]
+  end
+
+  specify "should support indexes with index type" do
+    @db.create_table(:posts){Integer :id; index :id, :type => :btree}
+    @db.sqls.should == [
+      "CREATE TABLE posts (id integer)",
+      "CREATE INDEX posts_id_index USING btree ON posts (id)"
+    ]
+  end
+
+  specify "should support unique indexes with index type" do
+    @db.create_table(:posts){Integer :id; index :id, :type => :btree, :unique => true}
+    @db.sqls.should == [
+      "CREATE TABLE posts (id integer)",
+      "CREATE UNIQUE INDEX posts_id_index USING btree ON posts (id)"
+    ]
+  end
+end
+
+context "MySQL::Dataset#insert and related methods" do
+  before do
+    MYSQL_DB.create_table(:items){String :name; Integer :value}
+    @d = MYSQL_DB[:items]
+    MYSQL_DB.sqls.clear
+  end
+  after do
+    MYSQL_DB.drop_table(:items)
+  end
+
+  specify "#insert should insert record with default values when no arguments given" do
+    @d.insert
+    
+    MYSQL_DB.sqls.should == [
+      "INSERT INTO items () VALUES ()"
+    ]
+    
+    @d.all.should == [
+      {:name => nil, :value => nil}
+    ]
+  end
+
+  specify "#insert  should insert record with default values when empty hash given" do
+    @d.insert({})
+    
+    MYSQL_DB.sqls.should == [
+      "INSERT INTO items () VALUES ()"
+    ]
+    
+    @d.all.should == [
+      {:name => nil, :value => nil}
+    ]
+  end
+
+  specify "#insert should insert record with default values when empty array given" do
+    @d.insert []
+    
+    MYSQL_DB.sqls.should == [
+      "INSERT INTO items () VALUES ()"
+    ]
+    
+    @d.all.should == [
+      {:name => nil, :value => nil}
+    ]
+  end
+
+  specify "#multi_insert should insert multiple records in a single statement" do
+    @d.multi_insert([{:name => 'abc'}, {:name => 'def'}])
+    
+    MYSQL_DB.sqls.should == [
+      SQL_BEGIN,
+      "INSERT INTO items (name) VALUES ('abc'), ('def')",
+      SQL_COMMIT
+    ]
+
+    @d.all.should == [
+      {:name => 'abc', :value => nil}, {:name => 'def', :value => nil}
+    ]
+  end
+
+  specify "#multi_insert should split the list of records into batches if :commit_every option is given" do
+    @d.multi_insert([{:value => 1}, {:value => 2}, {:value => 3}, {:value => 4}],
+      :commit_every => 2)
+
+    MYSQL_DB.sqls.should == [
+      SQL_BEGIN,
+      "INSERT INTO items (value) VALUES (1), (2)",
+      SQL_COMMIT,
+      SQL_BEGIN,
+      "INSERT INTO items (value) VALUES (3), (4)",
+      SQL_COMMIT
+    ]
+    
+    @d.all.should == [
+      {:name => nil, :value => 1}, 
+      {:name => nil, :value => 2},
+      {:name => nil, :value => 3}, 
+      {:name => nil, :value => 4}
+    ]
+  end
+
+  specify "#multi_insert should split the list of records into batches if :slice option is given" do
+    @d.multi_insert([{:value => 1}, {:value => 2}, {:value => 3}, {:value => 4}],
+      :slice => 2)
+
+    MYSQL_DB.sqls.should == [
+      SQL_BEGIN,
+      "INSERT INTO items (value) VALUES (1), (2)",
+      SQL_COMMIT,
+      SQL_BEGIN,
+      "INSERT INTO items (value) VALUES (3), (4)",
+      SQL_COMMIT
+    ]
+    
+    @d.all.should == [
+      {:name => nil, :value => 1}, 
+      {:name => nil, :value => 2},
+      {:name => nil, :value => 3}, 
+      {:name => nil, :value => 4}
+    ]
+  end
+  
+  specify "#import should support inserting using columns and values arrays" do
+    @d.import([:name, :value], [['abc', 1], ['def', 2]])
+
+    MYSQL_DB.sqls.should == [
+      SQL_BEGIN,
+      "INSERT INTO items (name, value) VALUES ('abc', 1), ('def', 2)",
+      SQL_COMMIT
+    ]
+    
+    @d.all.should == [
+      {:name => 'abc', :value => 1}, 
+      {:name => 'def', :value => 2}
+    ]
+  end
+  
+  specify "#insert_ignore should add the IGNORE keyword when inserting" do
+    @d.insert_ignore.multi_insert([{:name => 'abc'}, {:name => 'def'}])
+    
+    MYSQL_DB.sqls.should == [
+      SQL_BEGIN,
+      "INSERT IGNORE INTO items (name) VALUES ('abc'), ('def')",
+      SQL_COMMIT
+    ]
+
+    @d.all.should == [
+      {:name => 'abc', :value => nil}, {:name => 'def', :value => nil}
+    ]
+  end
+  
+  specify "#on_duplicate_key_update should add the ON DUPLICATE KEY UPDATE and ALL columns when no args given" do
+    @d.on_duplicate_key_update.import([:name,:value], 
+      [['abc', 1], ['def',2]]
+    )
+    
+    MYSQL_DB.sqls.should == [
+      SQL_BEGIN,
+      "INSERT INTO items (name, value) VALUES ('abc', 1), ('def', 2) ON DUPLICATE KEY UPDATE name=VALUES(name), value=VALUES(value)",
+      SQL_COMMIT
+    ]
+
+    @d.all.should == [
+      {:name => 'abc', :value => 1}, {:name => 'def', :value => 2}
+    ]
+  end
+  
+  specify "#on_duplicate_key_update should add the ON DUPLICATE KEY UPDATE and columns specified when args are given" do
+    @d.on_duplicate_key_update(:value).import([:name,:value], 
+      [['abc', 1], ['def',2]]
+    )
+    
+    MYSQL_DB.sqls.should == [
+      SQL_BEGIN,
+      "INSERT INTO items (name, value) VALUES ('abc', 1), ('def', 2) ON DUPLICATE KEY UPDATE value=VALUES(value)",
+      SQL_COMMIT
+    ]
+
+    @d.all.should == [
+      {:name => 'abc', :value => 1}, {:name => 'def', :value => 2}
+    ]
+  end
+  
+end
+
+context "MySQL::Dataset#replace" do
+  before do
+    MYSQL_DB.create_table(:items){Integer :id, :unique=>true; Integer :value}
+    @d = MYSQL_DB[:items]
+    MYSQL_DB.sqls.clear
+  end
+  after do
+    MYSQL_DB.drop_table(:items)
+  end
+  
+  specify "should create a record if the condition is not met" do
+    @d.replace(:id => 111, :value => 333)
+    @d.all.should == [{:id => 111, :value => 333}]
+  end
+  
+  specify "should update a record if the condition is met" do
+    @d << {:id => 111}
+    @d.all.should == [{:id => 111, :value => nil}]
+    @d.replace(:id => 111, :value => 333)
+    @d.all.should == [{:id => 111, :value => 333}]
+  end
+end
+
+context "MySQL::Dataset#complex_expression_sql" do
+  before do
+    @d = MYSQL_DB.dataset
+  end
+
+  specify "should handle pattern matches correctly" do
+    @d.literal(:x.like('a')).should == "(x LIKE BINARY 'a')"
+    @d.literal(~:x.like('a')).should == "(x NOT LIKE BINARY 'a')"
+    @d.literal(:x.ilike('a')).should == "(x LIKE 'a')"
+    @d.literal(~:x.ilike('a')).should == "(x NOT LIKE 'a')"
+    @d.literal(:x.like(/a/)).should == "(x REGEXP BINARY 'a')"
+    @d.literal(~:x.like(/a/)).should == "(x NOT REGEXP BINARY 'a')"
+    @d.literal(:x.like(/a/i)).should == "(x REGEXP 'a')"
+    @d.literal(~:x.like(/a/i)).should == "(x NOT REGEXP 'a')"
+  end
+
+  specify "should handle string concatenation with CONCAT if more than one record" do
+    @d.literal([:x, :y].sql_string_join).should == "CONCAT(x, y)"
+    @d.literal([:x, :y].sql_string_join(' ')).should == "CONCAT(x, ' ', y)"
+    @d.literal([:x.sql_function(:y), 1, 'z'.lit].sql_string_join(:y.sql_subscript(1))).should == "CONCAT(x(y), y[1], '1', y[1], z)"
+  end
+
+  specify "should handle string concatenation as simple string if just one record" do
+    @d.literal([:x].sql_string_join).should == "x"
+    @d.literal([:x].sql_string_join(' ')).should == "x"
+  end
+end
+
+unless MYSQL_DB.class.adapter_scheme == :do
+  context "MySQL Stored Procedures" do
+    before do
+      MYSQL_DB.create_table(:items){Integer :id; Integer :value}
+      @d = MYSQL_DB[:items]
+      MYSQL_DB.sqls.clear
+    end
+    after do
+      MYSQL_DB.drop_table(:items)
+      MYSQL_DB.execute('DROP PROCEDURE test_sproc')
+    end
+    
+    specify "should be callable on the database object" do
+      MYSQL_DB.execute('CREATE PROCEDURE test_sproc() BEGIN DELETE FROM items; END')
+      MYSQL_DB[:items].delete
+      MYSQL_DB[:items].insert(:value=>1)
+      MYSQL_DB[:items].count.should == 1
+      MYSQL_DB.call_sproc(:test_sproc)
+      MYSQL_DB[:items].count.should == 0
+    end
+    
+    specify "should be callable on the dataset object" do
+      MYSQL_DB.execute('CREATE PROCEDURE test_sproc(a INTEGER) BEGIN SELECT *, a AS b FROM items; END')
+      MYSQL_DB[:items].delete
+      @d = MYSQL_DB[:items]
+      @d.call_sproc(:select, :test_sproc, 3).should == []
+      @d.insert(:value=>1)
+      @d.call_sproc(:select, :test_sproc, 4).should == [{:id=>nil, :value=>1, :b=>4}]
+      @d.row_proc = proc{|r| r.keys.each{|k| r[k] *= 2 if r[k].is_a?(Integer)}; r}
+      @d.call_sproc(:select, :test_sproc, 3).should == [{:id=>nil, :value=>2, :b=>6}]
+    end
+    
+    specify "should be callable on the dataset object with multiple arguments" do
+      MYSQL_DB.execute('CREATE PROCEDURE test_sproc(a INTEGER, c INTEGER) BEGIN SELECT *, a AS b, c AS d FROM items; END')
+      MYSQL_DB[:items].delete
+      @d = MYSQL_DB[:items]
+      @d.call_sproc(:select, :test_sproc, 3, 4).should == []
+      @d.insert(:value=>1)
+      @d.call_sproc(:select, :test_sproc, 4, 5).should == [{:id=>nil, :value=>1, :b=>4, :d=>5}]
+      @d.row_proc = proc{|r| r.keys.each{|k| r[k] *= 2 if r[k].is_a?(Integer)}; r}
+      @d.call_sproc(:select, :test_sproc, 3, 4).should == [{:id=>nil, :value=>2, :b=>6, :d => 8}]
+    end
+  end
+end
+
+if MYSQL_DB.class.adapter_scheme == :mysql
+  context "MySQL bad date/time conversions" do
+    after do
+      Sequel::MySQL.convert_invalid_date_time = false
+    end
+  
+    specify "should raise an exception when a bad date/time is used and convert_invalid_date_time is false" do
+      Sequel::MySQL.convert_invalid_date_time = false
+      proc{MYSQL_DB["SELECT CAST('0000-00-00' AS date)"].single_value}.should raise_error(Sequel::InvalidValue)
+      proc{MYSQL_DB["SELECT CAST('0000-00-00 00:00:00' AS datetime)"].single_value}.should raise_error(Sequel::InvalidValue)
+      proc{MYSQL_DB["SELECT CAST('25:00:00' AS time)"].single_value}.should raise_error(Sequel::InvalidValue)
+    end
+  
+    specify "should not use a nil value bad date/time is used and convert_invalid_date_time is nil or :nil" do
+      Sequel::MySQL.convert_invalid_date_time = nil
+      MYSQL_DB["SELECT CAST('0000-00-00' AS date)"].single_value.should == nil
+      MYSQL_DB["SELECT CAST('0000-00-00 00:00:00' AS datetime)"].single_value.should == nil
+      MYSQL_DB["SELECT CAST('25:00:00' AS time)"].single_value.should == nil
+      Sequel::MySQL.convert_invalid_date_time = :nil
+      MYSQL_DB["SELECT CAST('0000-00-00' AS date)"].single_value.should == nil
+      MYSQL_DB["SELECT CAST('0000-00-00 00:00:00' AS datetime)"].single_value.should == nil
+      MYSQL_DB["SELECT CAST('25:00:00' AS time)"].single_value.should == nil
+    end
+  
+    specify "should not use a nil value bad date/time is used and convert_invalid_date_time is :string" do
+      Sequel::MySQL.convert_invalid_date_time = :string
+      MYSQL_DB["SELECT CAST('0000-00-00' AS date)"].single_value.should == '0000-00-00'
+      MYSQL_DB["SELECT CAST('0000-00-00 00:00:00' AS datetime)"].single_value.should == '0000-00-00 00:00:00'
+      MYSQL_DB["SELECT CAST('25:00:00' AS time)"].single_value.should == '25:00:00'
+    end
+  end
+end
diff -pruN 1.5.1-1/spec/adapters/oracle_spec.rb 3.0.0-1/spec/adapters/oracle_spec.rb
--- 1.5.1-1/spec/adapters/oracle_spec.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/adapters/oracle_spec.rb	2009-04-16 20:55:34.000000000 +0100
@@ -0,0 +1,285 @@
+require File.join(File.dirname(__FILE__), 'spec_helper.rb')
+
+unless defined?(ORACLE_DB)
+  ORACLE_DB = Sequel.connect('oracle://hr:hr@localhost/XE')
+end
+
+if ORACLE_DB.table_exists?(:items)
+  ORACLE_DB.drop_table :items
+end
+ORACLE_DB.create_table :items do
+  varchar2 :name, :size => 50
+  number :value, :size => 38
+  date :date_created
+  index :value
+end
+
+if ORACLE_DB.table_exists?(:books)
+  ORACLE_DB.drop_table :books
+end
+ORACLE_DB.create_table :books do
+  number :id, :size => 38
+  varchar2 :title, :size => 50
+  number :category_id, :size => 38
+end
+
+if ORACLE_DB.table_exists?(:categories)
+  ORACLE_DB.drop_table :categories
+end
+ORACLE_DB.create_table :categories do
+  number :id, :size => 38
+  varchar2 :cat_name, :size => 50
+end
+
+context "An Oracle database" do
+  specify "should provide disconnect functionality" do
+    ORACLE_DB.execute("select user from dual")
+    ORACLE_DB.pool.size.should == 1
+    ORACLE_DB.disconnect
+    ORACLE_DB.pool.size.should == 0
+  end
+  
+  specify "should provide schema information" do
+    books_schema = [
+      [:id, {:char_size=>0, :type=>:number, :allow_null=>true, :type_string=>"NUMBER(38)", :data_size=>22, :precision=>38, :char_used=>false, :scale=>0, :charset_form=>nil, :fsprecision=>38, :lfprecision=>0, :db_type=>"NUMBER(38)"}], 
+      [:title, {:char_size=>50, :type=>:varchar2, :allow_null=>true, :type_string=>"VARCHAR2(50)", :data_size=>50, :precision=>0, :char_used=>false, :scale=>0, :charset_form=>:implicit, :fsprecision=>0, :lfprecision=>0, :db_type=>"VARCHAR2(50)"}], 
+      [:category_id, {:char_size=>0, :type=>:number, :allow_null=>true, :type_string=>"NUMBER(38)", :data_size=>22, :precision=>38, :char_used=>false, :scale=>0, :charset_form=>nil, :fsprecision=>38, :lfprecision=>0, :db_type=>"NUMBER(38)"}]]
+    categories_schema = [
+      [:id, {:char_size=>0, :type=>:number, :allow_null=>true, :type_string=>"NUMBER(38)", :data_size=>22, :precision=>38, :char_used=>false, :scale=>0, :charset_form=>nil, :fsprecision=>38, :lfprecision=>0, :db_type=>"NUMBER(38)"}], 
+      [:cat_name, {:char_size=>50, :type=>:varchar2, :allow_null=>true, :type_string=>"VARCHAR2(50)", :data_size=>50, :precision=>0, :char_used=>false, :scale=>0, :charset_form=>:implicit, :fsprecision=>0, :lfprecision=>0, :db_type=>"VARCHAR2(50)"}]]
+    items_schema = [
+      [:name, {:char_size=>50, :type=>:varchar2, :allow_null=>true, :type_string=>"VARCHAR2(50)", :data_size=>50, :precision=>0, :char_used=>false, :scale=>0, :charset_form=>:implicit, :fsprecision=>0, :lfprecision=>0, :db_type=>"VARCHAR2(50)"}], 
+      [:value, {:char_size=>0, :type=>:number, :allow_null=>true, :type_string=>"NUMBER(38)", :data_size=>22, :precision=>38, :char_used=>false, :scale=>0, :charset_form=>nil, :fsprecision=>38, :lfprecision=>0, :db_type=>"NUMBER(38)"}],
+      [:date_created, {:charset_form=>nil, :type=>:date, :type_string=>"DATE", :fsprecision=>0, :data_size=>7, :lfprecision=>0, :precision=>0, :db_type=>"DATE", :char_used=>false, :char_size=>0, :scale=>0, :allow_null=>true}]]
+     
+    {:books => books_schema, :categories => categories_schema, :items => items_schema}.each_pair do |table, expected_schema|
+      schema = ORACLE_DB.schema(table)
+      schema.should_not be_nil
+      expected_schema.should == schema
+    end
+  end
+  
+  specify "should create a temporary table" do
+    ORACLE_DB.create_table :test_tmp, :temporary => true do
+      primary_key :id, :integer, :null => false
+      column :name, :text
+      index :name, :unique => true
+    end
+    
+    ORACLE_DB.sqls.should == [
+      'CREATE GLOBAL TEMPORARY TABLE test_tmp (id integer NOT NULL PRIMARY KEY AUTOINCREMENT, name text)',
+      'CREATE UNIQUE INDEX test_tmp_name_index ON test_tmp (name)'
+    ]
+  end
+end
+
+context "An Oracle dataset" do
+  before do
+    @d = ORACLE_DB[:items]
+    @d.delete # remove all records
+  end
+  
+  specify "should return the correct record count" do
+    @d.count.should == 0
+    @d << {:name => 'abc', :value => 123}
+    @d << {:name => 'abc', :value => 456}
+    @d << {:name => 'def', :value => 789}
+    @d.count.should == 3
+  end
+  
+  specify "should return the correct records" do
+    @d.to_a.should == []
+    @d << {:name => 'abc', :value => 123}
+    @d << {:name => 'abc', :value => 456}
+    @d << {:name => 'def', :value => 789}
+
+    @d.order(:value).to_a.should == [
+      {:date_created=>nil, :name => 'abc', :value => 123},
+      {:date_created=>nil, :name => 'abc', :value => 456},
+      {:date_created=>nil, :name => 'def', :value => 789}
+    ]
+
+    @d.select(:name).distinct.order_by(:name).to_a.should == [
+      {:name => 'abc'},
+      {:name => 'def'}
+    ]
+           
+    @d.order(:value.desc).limit(1).to_a.should == [
+      {:date_created=>nil, :name => 'def', :value => 789}                                        
+    ]
+
+    @d.filter(:name => 'abc').to_a.should == [
+      {:date_created=>nil, :name => 'abc', :value => 123},
+      {:date_created=>nil, :name => 'abc', :value => 456} 
+    ]
+    
+    @d.order(:value.desc).filter(:name => 'abc').to_a.should == [
+      {:date_created=>nil, :name => 'abc', :value => 456},
+      {:date_created=>nil, :name => 'abc', :value => 123} 
+    ]
+
+    @d.filter(:name => 'abc').limit(1).to_a.should == [
+      {:date_created=>nil, :name => 'abc', :value => 123}                                        
+    ]
+        
+    @d.filter(:name => 'abc').order(:value.desc).limit(1).to_a.should == [
+      {:date_created=>nil, :name => 'abc', :value => 456}                                        
+    ]
+    
+    @d.filter(:name => 'abc').order(:value).limit(1).to_a.should == [
+      {:date_created=>nil, :name => 'abc', :value => 123}                                        
+    ]
+        
+    @d.order(:value).limit(1).to_a.should == [
+      {:date_created=>nil, :name => 'abc', :value => 123}                                        
+    ]
+
+    @d.order(:value).limit(1, 1).to_a.should == [
+      {:date_created=>nil, :name => 'abc', :value => 456}
+    ]
+
+    @d.order(:value).limit(1, 2).to_a.should == [
+      {:date_created=>nil, :name => 'def', :value => 789}
+    ]    
+    
+    @d.avg(:value).to_i.should == (789+123+456)/3
+    
+    @d.max(:value).to_i.should == 789
+    
+    @d.select(:name, :AVG.sql_function(:value)).filter(:name => 'abc').group(:name).to_a.should == [
+      {:name => 'abc', :"avg(value)" => (456+123)/2.0}
+    ]
+
+    @d.select(:AVG.sql_function(:value)).group(:name).order(:name).limit(1).to_a.should == [
+      {:"avg(value)" => (456+123)/2.0}
+    ]
+        
+    @d.select(:name, :AVG.sql_function(:value)).group(:name).order(:name).to_a.should == [
+      {:name => 'abc', :"avg(value)" => (456+123)/2.0},
+      {:name => 'def', :"avg(value)" => 789*1.0}
+    ]
+    
+    @d.select(:name, :AVG.sql_function(:value)).group(:name).order(:name).to_a.should == [
+      {:name => 'abc', :"avg(value)" => (456+123)/2.0},
+      {:name => 'def', :"avg(value)" => 789*1.0}
+    ]
+
+    @d.select(:name, :AVG.sql_function(:value)).group(:name).having(:name => ['abc', 'def']).order(:name).to_a.should == [
+      {:name => 'abc', :"avg(value)" => (456+123)/2.0},
+      {:name => 'def', :"avg(value)" => 789*1.0}
+    ]
+    
+    @d.select(:name, :value).filter(:name => 'abc').union(@d.select(:name, :value).filter(:name => 'def')).order(:value).to_a.should == [
+      {:name => 'abc', :value => 123},
+      {:name => 'abc', :value => 456},
+      {:name => 'def', :value => 789}
+    ]
+     
+  end
+  
+  specify "should update records correctly" do
+    @d << {:name => 'abc', :value => 123}
+    @d << {:name => 'abc', :value => 456}
+    @d << {:name => 'def', :value => 789}
+    @d.filter(:name => 'abc').update(:value => 530)
+    
+    # the third record should stay the same
+    # floating-point precision bullshit
+    @d[:name => 'def'][:value].should == 789
+    @d.filter(:value => 530).count.should == 2
+  end
+
+  specify "should translate values correctly" do
+    @d << {:name => 'abc', :value => 456}
+    @d << {:name => 'def', :value => 789}
+    @d.filter(:value > 500).update(:date_created => "to_timestamp('2009-09-09', 'YYYY-MM-DD')".lit)
+    
+    @d[:name => 'def'][:date_created].should == Time.parse('2009-09-09')
+  end
+  
+  specify "should delete records correctly" do
+    @d << {:name => 'abc', :value => 123}
+    @d << {:name => 'abc', :value => 456}
+    @d << {:name => 'def', :value => 789}
+    @d.filter(:name => 'abc').delete
+    
+    @d.count.should == 1
+    @d.first[:name].should == 'def'
+  end
+  
+  specify "should be able to literalize booleans" do
+    proc {@d.literal(true)}.should_not raise_error
+    proc {@d.literal(false)}.should_not raise_error
+  end
+  
+  specify "should support transactions" do
+    ORACLE_DB.transaction do
+      @d << {:name => 'abc', :value => 1}
+    end
+
+    @d.count.should == 1
+  end
+end
+
+context "Joined Oracle dataset" do
+  before do
+    @d1 = ORACLE_DB[:books]
+    @d1.delete # remove all records
+    @d1 << {:id => 1, :title => 'aaa', :category_id => 100}
+    @d1 << {:id => 2, :title => 'bbb', :category_id => 100}
+    @d1 << {:id => 3, :title => 'ccc', :category_id => 101}
+    @d1 << {:id => 4, :title => 'ddd', :category_id => 102}
+    
+    @d2 = ORACLE_DB[:categories]
+    @d2.delete # remove all records
+    @d2 << {:id => 100, :cat_name => 'ruby'}
+    @d2 << {:id => 101, :cat_name => 'rails'}
+  end
+  
+  specify "should return correct result" do
+    @d1.join(:categories, :id => :category_id).select(:books__id, :title, :cat_name).order(:books__id).to_a.should == [
+      {:id => 1, :title => 'aaa', :cat_name => 'ruby'},
+      {:id => 2, :title => 'bbb', :cat_name => 'ruby'},
+      {:id => 3, :title => 'ccc', :cat_name => 'rails'}
+    ]
+
+    @d1.join(:categories, :id => :category_id).select(:books__id, :title, :cat_name).order(:books__id).limit(2, 1).to_a.should == [
+      {:id => 2, :title => 'bbb', :cat_name => 'ruby'},
+      {:id => 3, :title => 'ccc', :cat_name => 'rails'},
+    ]
+   
+    @d1.left_outer_join(:categories, :id => :category_id).select(:books__id, :title, :cat_name).order(:books__id).to_a.should == [
+      {:id => 1, :title => 'aaa', :cat_name => 'ruby'},
+      {:id => 2, :title => 'bbb', :cat_name => 'ruby'},
+      {:id => 3, :title => 'ccc', :cat_name => 'rails'},
+      {:id => 4, :title => 'ddd', :cat_name => nil} 
+    ]
+    
+    @d1.left_outer_join(:categories, :id => :category_id).select(:books__id, :title, :cat_name).order(:books__id.desc).limit(2, 0).to_a.should == [      
+      {:id => 4, :title => 'ddd', :cat_name => nil}, 
+      {:id => 3, :title => 'ccc', :cat_name => 'rails'}
+    ]      
+  end  
+end
+
+context "Oracle aliasing" do
+  before do
+    @d1 = ORACLE_DB[:books]
+    @d1.delete # remove all records
+    @d1 << {:id => 1, :title => 'aaa', :category_id => 100}
+    @d1 << {:id => 2, :title => 'bbb', :category_id => 100}
+    @d1 << {:id => 3, :title => 'bbb', :category_id => 100}
+  end
+
+  specify "should allow columns to be renamed" do
+    @d1.select(:title.as(:name)).order_by(:id).to_a.should == [
+      { :name => 'aaa' },
+      { :name => 'bbb' },
+      { :name => 'bbb' },
+    ]
+  end
+
+  specify "nested queries should work" do
+    @d1.select(:title).group_by(:title).count.should == 2
+  end
+end
diff -pruN 1.5.1-1/spec/adapters/postgres_spec.rb 3.0.0-1/spec/adapters/postgres_spec.rb
--- 1.5.1-1/spec/adapters/postgres_spec.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/adapters/postgres_spec.rb	2009-04-27 23:47:32.000000000 +0100
@@ -0,0 +1,599 @@
+require File.join(File.dirname(__FILE__), 'spec_helper.rb')
+
+unless defined?(POSTGRES_DB)
+  POSTGRES_URL = 'postgres://postgres:postgres@localhost:5432/reality_spec' unless defined? POSTGRES_URL
+  POSTGRES_DB = Sequel.connect(ENV['SEQUEL_PG_SPEC_DB']||POSTGRES_URL)
+end
+
+def POSTGRES_DB.sqls
+  (@sqls ||= [])
+end
+logger = Object.new
+def logger.method_missing(m, msg)
+  POSTGRES_DB.sqls << msg
+end
+POSTGRES_DB.logger = logger
+
+#POSTGRES_DB.instance_variable_set(:@server_version, 80100)
+POSTGRES_DB.create_table! :test do
+  text :name
+  integer :value, :index => true
+end
+POSTGRES_DB.create_table! :test2 do
+  text :name
+  integer :value
+end
+POSTGRES_DB.create_table! :test3 do
+  integer :value
+  timestamp :time
+end
+POSTGRES_DB.create_table! :test4 do
+  varchar :name, :size => 20
+  bytea :value
+end
+
+context "A PostgreSQL database" do
+  before do
+    @db = POSTGRES_DB
+  end
+  
+  specify "should provide the server version" do
+    @db.server_version.should > 70000
+  end
+
+  specify "should correctly parse the schema" do
+    @db.schema(:test3, :reload=>true).should == [
+      [:value, {:type=>:integer, :allow_null=>true, :default=>nil, :db_type=>"integer", :primary_key=>false}],
+      [:time, {:type=>:datetime, :allow_null=>true, :default=>nil, :db_type=>"timestamp without time zone", :primary_key=>false}]
+    ]
+    @db.schema(:test4, :reload=>true).should == [
+      [:name, {:type=>:string, :allow_null=>true, :default=>nil, :db_type=>"character varying(20)", :primary_key=>false}],
+      [:value, {:type=>:blob, :allow_null=>true, :default=>nil, :db_type=>"bytea", :primary_key=>false}]
+    ]
+  end
+end
+
+context "A PostgreSQL dataset" do
+  before do
+    @d = POSTGRES_DB[:test]
+    @d.delete # remove all records
+  end
+  
+  specify "should quote columns and tables using double quotes if quoting identifiers" do
+    @d.quote_identifiers = true
+    @d.select(:name).sql.should == \
+      'SELECT "name" FROM "test"'
+      
+    @d.select('COUNT(*)'.lit).sql.should == \
+      'SELECT COUNT(*) FROM "test"'
+
+    @d.select(:max.sql_function(:value)).sql.should == \
+      'SELECT max("value") FROM "test"'
+      
+    @d.select(:NOW.sql_function).sql.should == \
+    'SELECT NOW() FROM "test"'
+
+    @d.select(:max.sql_function(:items__value)).sql.should == \
+      'SELECT max("items"."value") FROM "test"'
+
+    @d.order(:name.desc).sql.should == \
+      'SELECT * FROM "test" ORDER BY "name" DESC'
+
+    @d.select('test.name AS item_name'.lit).sql.should == \
+      'SELECT test.name AS item_name FROM "test"'
+      
+    @d.select('"name"'.lit).sql.should == \
+      'SELECT "name" FROM "test"'
+
+    @d.select('max(test."name") AS "max_name"'.lit).sql.should == \
+      'SELECT max(test."name") AS "max_name" FROM "test"'
+      
+    @d.select(:test.sql_function(:abc, 'hello')).sql.should == \
+      "SELECT test(\"abc\", 'hello') FROM \"test\""
+
+    @d.select(:test.sql_function(:abc__def, 'hello')).sql.should == \
+      "SELECT test(\"abc\".\"def\", 'hello') FROM \"test\""
+
+    @d.select(:test.sql_function(:abc__def, 'hello').as(:x2)).sql.should == \
+      "SELECT test(\"abc\".\"def\", 'hello') AS \"x2\" FROM \"test\""
+
+    @d.insert_sql(:value => 333).should =~ \
+      /\AINSERT INTO "test" \("value"\) VALUES \(333\)( RETURNING NULL)?\z/
+
+    @d.insert_sql(:x => :y).should =~ \
+      /\AINSERT INTO "test" \("x"\) VALUES \("y"\)( RETURNING NULL)?\z/
+
+    @d.disable_insert_returning.insert_sql(:value => 333).should =~ \
+      /\AINSERT INTO "test" \("value"\) VALUES \(333\)\z/
+  end
+  
+  specify "should quote fields correctly when reversing the order if quoting identifiers" do
+    @d.quote_identifiers = true
+    @d.reverse_order(:name).sql.should == \
+      'SELECT * FROM "test" ORDER BY "name" DESC'
+
+    @d.reverse_order(:name.desc).sql.should == \
+      'SELECT * FROM "test" ORDER BY "name" ASC'
+
+    @d.reverse_order(:name, :test.desc).sql.should == \
+      'SELECT * FROM "test" ORDER BY "name" DESC, "test" ASC'
+
+    @d.reverse_order(:name.desc, :test).sql.should == \
+      'SELECT * FROM "test" ORDER BY "name" ASC, "test" DESC'
+  end
+  
+  specify "should support nested transactions through savepoints using the savepoint option" do
+    POSTGRES_DB.transaction do
+      @d << {:name => '1'}
+      POSTGRES_DB.transaction(:savepoint=>true) do
+        @d << {:name => '2'}
+        POSTGRES_DB.transaction do
+          @d << {:name => '3'}
+          raise Sequel::Rollback
+        end
+      end
+      @d << {:name => '4'}
+      POSTGRES_DB.transaction do
+        @d << {:name => '6'}
+        POSTGRES_DB.transaction(:savepoint=>true) do
+          @d << {:name => '7'}
+          raise Sequel::Rollback
+        end
+      end
+      @d << {:name => '5'}
+    end
+
+    @d.order(:name).map(:name).should == %w{1 4 5 6}
+  end
+
+  specify "should support regexps" do
+    @d << {:name => 'abc', :value => 1}
+    @d << {:name => 'bcd', :value => 2}
+    @d.filter(:name => /bc/).count.should == 2
+    @d.filter(:name => /^bc/).count.should == 1
+  end
+end
+
+context "A PostgreSQL dataset with a timestamp field" do
+  before do
+    @d = POSTGRES_DB[:test3]
+    @d.delete
+  end
+
+  specify "should store milliseconds in time fields" do
+    t = Time.now
+    @d << {:value=>1, :time=>t}
+    @d.literal(@d[:value =>'1'][:time]).should == @d.literal(t)
+    @d[:value=>'1'][:time].usec.should == t.usec
+  end
+end
+
+context "A PostgreSQL database" do
+  before do
+    @db = POSTGRES_DB
+  end
+
+  specify "should support column operations" do
+    @db.create_table!(:test2){text :name; integer :value}
+    @db[:test2] << {}
+    @db[:test2].columns.should == [:name, :value]
+
+    @db.add_column :test2, :xyz, :text, :default => '000'
+    @db[:test2].columns.should == [:name, :value, :xyz]
+    @db[:test2] << {:name => 'mmm', :value => 111}
+    @db[:test2].first[:xyz].should == '000'
+  
+    @db[:test2].columns.should == [:name, :value, :xyz]
+    @db.drop_column :test2, :xyz
+    
+    @db[:test2].columns.should == [:name, :value]
+  
+    @db[:test2].delete
+    @db.add_column :test2, :xyz, :text, :default => '000'
+    @db[:test2] << {:name => 'mmm', :value => 111, :xyz => 'qqqq'}
+
+    @db[:test2].columns.should == [:name, :value, :xyz]
+    @db.rename_column :test2, :xyz, :zyx
+    @db[:test2].columns.should == [:name, :value, :zyx]
+    @db[:test2].first[:zyx].should == 'qqqq'
+  
+    @db.add_column :test2, :xyz, :float
+    @db[:test2].delete
+    @db[:test2] << {:name => 'mmm', :value => 111, :xyz => 56.78}
+    @db.set_column_type :test2, :xyz, :integer
+    
+    @db[:test2].first[:xyz].should == 57
+  end
+end  
+
+context "A PostgreSQL database" do
+  before do
+    @db = POSTGRES_DB
+    @db.drop_table(:posts) rescue nil
+    @db.sqls.clear
+  end
+  after do
+    @db.drop_table(:posts) rescue nil
+  end
+  
+  specify "should support resetting the primary key sequence" do
+    @db.create_table(:posts){primary_key :a}
+    @db[:posts].insert(:a=>20).should == 20
+    @db[:posts].insert.should == 1
+    @db[:posts].insert.should == 2
+    @db[:posts].insert(:a=>10).should == 10
+    @db.reset_primary_key_sequence(:posts).should == 21
+    @db[:posts].insert.should == 21
+    @db[:posts].order(:a).map(:a).should == [1, 2, 10, 20, 21]
+  end
+  
+  specify "should support fulltext indexes and searching" do
+    @db.create_table(:posts){text :title; text :body; full_text_index [:title, :body]; full_text_index :title, :language => 'french'}
+    @db.sqls.should == [
+      "CREATE TABLE posts (title text, body text)",
+      "CREATE INDEX posts_title_body_index ON posts USING gin (to_tsvector('simple', (COALESCE(title, '') || ' ' || COALESCE(body, ''))))",
+      "CREATE INDEX posts_title_index ON posts USING gin (to_tsvector('french', (COALESCE(title, ''))))"
+    ]
+
+    @db[:posts].insert(:title=>'ruby rails', :body=>'yowsa')
+    @db[:posts].insert(:title=>'sequel', :body=>'ruby')
+    @db[:posts].insert(:title=>'ruby scooby', :body=>'x')
+    @db.sqls.clear
+
+    @db[:posts].full_text_search(:title, 'rails').all.should == [{:title=>'ruby rails', :body=>'yowsa'}]
+    @db[:posts].full_text_search([:title, :body], ['yowsa', 'rails']).all.should == [:title=>'ruby rails', :body=>'yowsa']
+    @db[:posts].full_text_search(:title, 'scooby', :language => 'french').all.should == [{:title=>'ruby scooby', :body=>'x'}]
+    @db.sqls.should == [
+      "SELECT * FROM posts WHERE (to_tsvector('simple', (COALESCE(title, ''))) @@ to_tsquery('simple', 'rails'))",
+      "SELECT * FROM posts WHERE (to_tsvector('simple', (COALESCE(title, '') || ' ' || COALESCE(body, ''))) @@ to_tsquery('simple', 'yowsa | rails'))",
+      "SELECT * FROM posts WHERE (to_tsvector('french', (COALESCE(title, ''))) @@ to_tsquery('french', 'scooby'))"]
+  end
+
+  specify "should support spatial indexes" do
+    @db.create_table(:posts){box :geom; spatial_index [:geom]}
+    @db.sqls.should == [
+      "CREATE TABLE posts (geom box)",
+      "CREATE INDEX posts_geom_index ON posts USING gist (geom)"
+    ]
+  end
+  
+  specify "should support indexes with index type" do
+    @db.create_table(:posts){varchar :title, :size => 5; index :title, :type => 'hash'}
+    @db.sqls.should == [
+      "CREATE TABLE posts (title varchar(5))",
+      "CREATE INDEX posts_title_index ON posts USING hash (title)"
+    ]
+  end
+  
+  specify "should support unique indexes with index type" do
+    @db.create_table(:posts){varchar :title, :size => 5; index :title, :type => 'btree', :unique => true}
+    @db.sqls.should == [
+      "CREATE TABLE posts (title varchar(5))",
+      "CREATE UNIQUE INDEX posts_title_index ON posts USING btree (title)"
+    ]
+  end
+  
+  specify "should support partial indexes" do
+    @db.create_table(:posts){varchar :title, :size => 5; index :title, :where => {:title => '5'}}
+    @db.sqls.should == [
+      "CREATE TABLE posts (title varchar(5))",
+      "CREATE INDEX posts_title_index ON posts (title) WHERE (title = '5')"
+    ]
+  end
+  
+  specify "should support identifiers for table names in indicies" do
+    @db.create_table(Sequel::SQL::Identifier.new(:posts)){varchar :title, :size => 5; index :title, :where => {:title => '5'}}
+    @db.sqls.should == [
+      "CREATE TABLE posts (title varchar(5))",
+      "CREATE INDEX posts_title_index ON posts (title) WHERE (title = '5')"
+    ]
+  end
+end
+
+context "Postgres::Dataset#import" do
+  before do
+    @db = POSTGRES_DB
+    @db.create_table!(:test){Integer :x; Integer :y}
+    @db.sqls.clear
+    @ds = @db[:test]
+  end
+  after do
+    @db.drop_table(:test) rescue nil
+  end
+  
+  specify "#import should return separate insert statements if server_version < 80200" do
+    @ds.meta_def(:server_version){80199}
+    
+    @ds.import([:x, :y], [[1, 2], [3, 4]])
+    
+    @db.sqls.should == [
+      'BEGIN',
+      'INSERT INTO test (x, y) VALUES (1, 2)',
+      'INSERT INTO test (x, y) VALUES (3, 4)',
+      'COMMIT'
+    ]
+    @ds.all.should == [{:x=>1, :y=>2}, {:x=>3, :y=>4}]
+  end
+  
+  specify "#import should a single insert statement if server_version >= 80200" do
+    @ds.meta_def(:server_version){80200}
+    
+    @ds.import([:x, :y], [[1, 2], [3, 4]])
+    
+    @db.sqls.should == [
+      'BEGIN',
+      'INSERT INTO test (x, y) VALUES (1, 2), (3, 4)',
+      'COMMIT'
+    ]
+    @ds.all.should == [{:x=>1, :y=>2}, {:x=>3, :y=>4}]
+  end
+end
+
+context "Postgres::Dataset#insert" do
+  before do
+    @db = POSTGRES_DB
+    @db.create_table!(:test5){primary_key :xid; Integer :value}
+    @db.sqls.clear
+    @ds = @db[:test5]
+  end
+  after do
+    @db.drop_table(:test5) rescue nil
+  end
+
+  specify "should work regardless of how it is used" do
+    @ds.insert(:value=>10).should == 1
+    @ds.disable_insert_returning.insert(:value=>20).should == 2
+    @ds.meta_def(:server_version){80100}
+    @ds.insert(:value=>13).should == 3
+    
+    @db.sqls.reject{|x| x =~ /pg_class/}.should == [
+      'INSERT INTO test5 (value) VALUES (10) RETURNING xid',
+      'INSERT INTO test5 (value) VALUES (20)',
+      "SELECT currval('\"public\".test5_xid_seq')",
+      'INSERT INTO test5 (value) VALUES (13)',
+      "SELECT currval('\"public\".test5_xid_seq')"
+    ]
+    @ds.all.should == [{:xid=>1, :value=>10}, {:xid=>2, :value=>20}, {:xid=>3, :value=>13}]
+  end
+  
+  specify "should call execute_insert if server_version < 80200" do
+    @ds.meta_def(:server_version){80100}
+    @ds.should_receive(:execute_insert).once.with('INSERT INTO test5 (value) VALUES (10)', :table=>:test5, :values=>{:value=>10})
+    @ds.insert(:value=>10)
+  end
+
+  specify "should call execute_insert if disabling insert returning" do
+    @ds.disable_insert_returning!
+    @ds.should_receive(:execute_insert).once.with('INSERT INTO test5 (value) VALUES (10)', :table=>:test5, :values=>{:value=>10})
+    @ds.insert(:value=>10)
+  end
+
+  specify "should use INSERT RETURNING if server_version >= 80200" do
+    @ds.meta_def(:server_version){80201}
+    @ds.should_receive(:clone).once.with(:server=>:default, :sql=>'INSERT INTO test5 (value) VALUES (10) RETURNING xid').and_return(@ds)
+    @ds.should_receive(:single_value).once
+    @ds.insert(:value=>10)
+  end
+
+  specify "should have insert_returning_sql use the RETURNING keyword" do
+    @ds.insert_returning_sql(:xid, :value=>10).should == "INSERT INTO test5 (value) VALUES (10) RETURNING xid"
+    @ds.insert_returning_sql('*'.lit, :value=>10).should == "INSERT INTO test5 (value) VALUES (10) RETURNING *"
+  end
+
+  specify "should have insert_select return nil if server_version < 80200" do
+    @ds.meta_def(:server_version){80100}
+    @ds.insert_select(:value=>10).should == nil
+  end
+
+  specify "should have insert_select insert the record and return the inserted record if server_version < 80200" do
+    @ds.meta_def(:server_version){80201}
+    h = @ds.insert_select(:value=>10)
+    h[:value].should == 10
+    @ds.first(:xid=>h[:xid])[:value].should == 10
+  end
+
+  specify "should correctly return the inserted record's primary key value" do
+    value1 = 10
+    id1 = @ds.insert(:value=>value1)
+    @ds.first(:xid=>id1)[:value].should == value1
+    value2 = 20
+    id2 = @ds.insert(:value=>value2)
+    @ds.first(:xid=>id2)[:value].should == value2
+  end
+
+  specify "should return nil if the table has no primary key" do
+    ds = POSTGRES_DB[:test4]
+    ds.delete
+    ds.insert(:name=>'a').should == nil
+  end
+end
+
+context "Postgres::Database schema qualified tables" do
+  before do
+    POSTGRES_DB << "CREATE SCHEMA schema_test"
+    POSTGRES_DB.instance_variable_set(:@primary_keys, {})
+    POSTGRES_DB.instance_variable_set(:@primary_key_sequences, {})
+  end
+  after do
+    POSTGRES_DB.quote_identifiers = false
+    POSTGRES_DB << "DROP SCHEMA schema_test CASCADE"
+    POSTGRES_DB.default_schema = :public
+  end
+  
+  specify "should be able to create, drop, select and insert into tables in a given schema" do
+    POSTGRES_DB.create_table(:schema_test__schema_test){primary_key :i}
+    POSTGRES_DB[:schema_test__schema_test].first.should == nil
+    POSTGRES_DB[:schema_test__schema_test].insert(:i=>1).should == 1
+    POSTGRES_DB[:schema_test__schema_test].first.should == {:i=>1}
+    POSTGRES_DB.from('schema_test.schema_test'.lit).first.should == {:i=>1}
+    POSTGRES_DB.drop_table(:schema_test__schema_test)
+    POSTGRES_DB.create_table(:schema_test.qualify(:schema_test)){integer :i}
+    POSTGRES_DB[:schema_test__schema_test].first.should == nil
+    POSTGRES_DB.from('schema_test.schema_test'.lit).first.should == nil
+    POSTGRES_DB.drop_table(:schema_test.qualify(:schema_test))
+  end
+  
+  specify "#tables should include only tables in the public schema if no schema is given" do
+    POSTGRES_DB.create_table(:schema_test__schema_test){integer :i}
+    POSTGRES_DB.tables.should_not include(:schema_test)
+  end
+  
+  specify "#tables should return tables in the schema provided by the :schema argument" do
+    POSTGRES_DB.create_table(:schema_test__schema_test){integer :i}
+    POSTGRES_DB.tables(:schema=>:schema_test).should == [:schema_test]
+  end
+  
+  specify "#table_exists? should assume the public schema if no schema is provided" do
+    POSTGRES_DB.create_table(:schema_test__schema_test){integer :i}
+    POSTGRES_DB.table_exists?(:schema_test).should == false
+  end
+  
+  specify "#table_exists? should see if the table is in a given schema" do
+    POSTGRES_DB.create_table(:schema_test__schema_test){integer :i}
+    POSTGRES_DB.table_exists?(:schema_test__schema_test).should == true
+  end
+  
+  specify "should be able to get primary keys for tables in a given schema" do
+    POSTGRES_DB.create_table(:schema_test__schema_test){primary_key :i}
+    POSTGRES_DB.primary_key(:schema_test__schema_test).should == 'i'
+  end
+  
+  specify "should be able to get serial sequences for tables in a given schema" do
+    POSTGRES_DB.create_table(:schema_test__schema_test){primary_key :i}
+    POSTGRES_DB.primary_key_sequence(:schema_test__schema_test).should == '"schema_test".schema_test_i_seq'
+  end
+  
+  specify "should be able to get serial sequences for tables that have spaces in the name in a given schema" do
+    POSTGRES_DB.quote_identifiers = true
+    POSTGRES_DB.create_table(:"schema_test__schema test"){primary_key :i}
+    POSTGRES_DB.primary_key_sequence(:"schema_test__schema test").should == '"schema_test"."schema test_i_seq"'
+  end
+  
+  specify "should be able to get custom sequences for tables in a given schema" do
+    POSTGRES_DB << "CREATE SEQUENCE schema_test.kseq"
+    POSTGRES_DB.create_table(:schema_test__schema_test){integer :j; primary_key :k, :type=>:integer, :default=>"nextval('schema_test.kseq'::regclass)".lit}
+    POSTGRES_DB.primary_key_sequence(:schema_test__schema_test).should == '"schema_test".kseq'
+  end
+  
+  specify "should be able to get custom sequences for tables that have spaces in the name in a given schema" do
+    POSTGRES_DB.quote_identifiers = true
+    POSTGRES_DB << "CREATE SEQUENCE schema_test.\"ks eq\""
+    POSTGRES_DB.create_table(:"schema_test__schema test"){integer :j; primary_key :k, :type=>:integer, :default=>"nextval('schema_test.\"ks eq\"'::regclass)".lit}
+    POSTGRES_DB.primary_key_sequence(:"schema_test__schema test").should == '"schema_test"."ks eq"'
+  end
+  
+  specify "#default_schema= should change the default schema used from public" do
+    POSTGRES_DB.create_table(:schema_test__schema_test){primary_key :i}
+    POSTGRES_DB.default_schema = :schema_test
+    POSTGRES_DB.table_exists?(:schema_test).should == true
+    POSTGRES_DB.tables.should == [:schema_test]
+    POSTGRES_DB.primary_key(:schema_test__schema_test).should == 'i'
+    POSTGRES_DB.primary_key_sequence(:schema_test__schema_test).should == '"schema_test".schema_test_i_seq'
+  end
+end
+
+if POSTGRES_DB.server_version >= 80300
+
+  POSTGRES_DB.create_table! :test6 do
+    text :title
+    text :body
+    full_text_index [:title, :body]
+  end
+
+  context "PostgreSQL tsearch2" do
+    before do
+      @ds = POSTGRES_DB[:test6]
+    end
+    after do
+      POSTGRES_DB[:test6].delete
+    end
+
+    specify "should search by indexed column" do
+      record =  {:title => "oopsla conference", :body => "test"}
+      @ds << record
+      @ds.full_text_search(:title, "oopsla").all.should include(record)
+    end
+
+    specify "should join multiple coumns with spaces to search by last words in row" do
+      record = {:title => "multiple words", :body => "are easy to search"}
+      @ds << record
+      @ds.full_text_search([:title, :body], "words").all.should include(record)
+    end
+
+    specify "should return rows with a NULL in one column if a match in another column" do
+      record = {:title => "multiple words", :body =>nil}
+      @ds << record
+      @ds.full_text_search([:title, :body], "words").all.should include(record)
+    end
+  end
+end
+
+context "Postgres::Database functions, languages, and triggers" do
+  before do
+    @d = POSTGRES_DB
+  end
+  after do
+    @d.drop_function('tf', :if_exists=>true, :cascade=>true)
+    @d.drop_function('tf', :if_exists=>true, :cascade=>true, :args=>%w'integer integer')
+    @d.drop_language(:plpgsql, :if_exists=>true, :cascade=>true)
+    @d.drop_table(:test) rescue nil
+  end
+
+  specify "#create_function and #drop_function should create and drop functions" do
+    proc{@d['SELECT tf()'].all}.should raise_error(Sequel::DatabaseError)
+    args = ['tf', 'SELECT 1', {:returns=>:integer}]
+    @d.send(:create_function_sql, *args).should =~ /\A\s*CREATE FUNCTION tf\(\)\s+RETURNS integer\s+LANGUAGE SQL\s+AS 'SELECT 1'\s*\z/
+    @d.create_function(*args)
+    rows = @d['SELECT tf()'].all.should == [{:tf=>1}]
+    @d.send(:drop_function_sql, 'tf').should == 'DROP FUNCTION tf()'
+    @d.drop_function('tf')
+    proc{@d['SELECT tf()'].all}.should raise_error(Sequel::DatabaseError)
+  end
+  
+  specify "#create_function and #drop_function should support options" do
+    args = ['tf', 'SELECT $1 + $2', {:args=>[[:integer, :a], :integer], :replace=>true, :returns=>:integer, :language=>'SQL', :behavior=>:immutable, :strict=>true, :security_definer=>true, :cost=>2, :set=>{:search_path => 'public'}}]
+    @d.send(:create_function_sql,*args).should =~ /\A\s*CREATE OR REPLACE FUNCTION tf\(a integer, integer\)\s+RETURNS integer\s+LANGUAGE SQL\s+IMMUTABLE\s+STRICT\s+SECURITY DEFINER\s+COST 2\s+SET search_path = public\s+AS 'SELECT \$1 \+ \$2'\s*\z/
+    @d.create_function(*args)
+    # Make sure replace works
+    @d.create_function(*args)
+    rows = @d['SELECT tf(1, 2)'].all.should == [{:tf=>3}]
+    args = ['tf', {:if_exists=>true, :cascade=>true, :args=>[[:integer, :a], :integer]}]
+    @d.send(:drop_function_sql,*args).should == 'DROP FUNCTION IF EXISTS tf(a integer, integer) CASCADE'
+    @d.drop_function(*args)
+    # Make sure if exists works
+    @d.drop_function(*args)
+  end
+  
+  specify "#create_language and #drop_language should create and drop languages" do
+    @d.send(:create_language_sql, :plpgsql).should == 'CREATE LANGUAGE plpgsql'
+    @d.create_language(:plpgsql)
+    proc{@d.create_language(:plpgsql)}.should raise_error(Sequel::DatabaseError)
+    @d.send(:drop_language_sql, :plpgsql).should == 'DROP LANGUAGE plpgsql'
+    @d.drop_language(:plpgsql)
+    proc{@d.drop_language(:plpgsql)}.should raise_error(Sequel::DatabaseError)
+    @d.send(:create_language_sql, :plpgsql, :trusted=>true, :handler=>:a, :validator=>:b).should == 'CREATE TRUSTED LANGUAGE plpgsql HANDLER a VALIDATOR b'
+    @d.send(:drop_language_sql, :plpgsql, :if_exists=>true, :cascade=>true).should == 'DROP LANGUAGE IF EXISTS plpgsql CASCADE'
+    # Make sure if exists works
+    @d.drop_language(:plpgsql, :if_exists=>true, :cascade=>true)
+  end
+  
+  specify "#create_trigger and #drop_trigger should create and drop triggers" do
+    @d.create_language(:plpgsql)
+    @d.create_function(:tf, 'BEGIN IF NEW.value IS NULL THEN RAISE EXCEPTION \'Blah\'; END IF; RETURN NEW; END;', :language=>:plpgsql, :returns=>:trigger)
+    @d.send(:create_trigger_sql, :test, :identity, :tf, :each_row=>true).should == 'CREATE TRIGGER identity BEFORE INSERT OR UPDATE OR DELETE ON public.test FOR EACH ROW EXECUTE PROCEDURE tf()'
+    @d.create_table(:test){String :name; Integer :value}
+    @d.create_trigger(:test, :identity, :tf, :each_row=>true)
+    @d[:test].insert(:name=>'a', :value=>1)
+    @d[:test].filter(:name=>'a').all.should == [{:name=>'a', :value=>1}]
+    proc{@d[:test].filter(:name=>'a').update(:value=>nil)}.should raise_error(Sequel::DatabaseError)
+    @d[:test].filter(:name=>'a').all.should == [{:name=>'a', :value=>1}]
+    @d[:test].filter(:name=>'a').update(:value=>3)
+    @d[:test].filter(:name=>'a').all.should == [{:name=>'a', :value=>3}]
+    @d.send(:drop_trigger_sql, :test, :identity).should == 'DROP TRIGGER identity ON public.test'
+    @d.drop_trigger(:test, :identity)
+    @d.send(:create_trigger_sql, :test, :identity, :tf, :after=>true, :events=>:insert, :args=>[1, 'a']).should == 'CREATE TRIGGER identity AFTER INSERT ON public.test EXECUTE PROCEDURE tf(1, \'a\')'
+    @d.send(:drop_trigger_sql, :test, :identity, :if_exists=>true, :cascade=>true).should == 'DROP TRIGGER IF EXISTS identity ON public.test CASCADE'
+    # Make sure if exists works
+    @d.drop_trigger(:test, :identity, :if_exists=>true, :cascade=>true)
+  end
+end
diff -pruN 1.5.1-1/spec/adapters/spec_helper.rb 3.0.0-1/spec/adapters/spec_helper.rb
--- 1.5.1-1/spec/adapters/spec_helper.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/adapters/spec_helper.rb	2009-03-16 20:53:19.000000000 +0000
@@ -0,0 +1,10 @@
+require 'rubygems'
+unless Object.const_defined?('Sequel')
+  $:.unshift(File.join(File.dirname(__FILE__), "../../lib/"))
+  require 'sequel/core'
+  Sequel.quote_identifiers = false
+end
+begin
+  require File.join(File.dirname(__FILE__), '../spec_config.rb')
+rescue LoadError
+end
diff -pruN 1.5.1-1/spec/adapters/sqlite_spec.rb 3.0.0-1/spec/adapters/sqlite_spec.rb
--- 1.5.1-1/spec/adapters/sqlite_spec.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/adapters/sqlite_spec.rb	2009-04-13 02:21:28.000000000 +0100
@@ -0,0 +1,392 @@
+require File.join(File.dirname(__FILE__), 'spec_helper.rb')
+
+unless defined?(SQLITE_DB)
+  SQLITE_URL = 'sqlite:/' unless defined? SQLITE_URL
+  SQLITE_DB = Sequel.connect(ENV['SEQUEL_SQLITE_SPEC_DB']||SQLITE_URL)
+end
+
+context "An SQLite database" do
+  before do
+    @db = SQLITE_DB
+  end
+  after do
+    Sequel.datetime_class = Time
+  end
+
+  if SQLITE_DB.auto_vacuum == :none
+    specify "should support getting pragma values" do
+      @db.pragma_get(:auto_vacuum).to_s.should == '0'
+    end
+    
+    specify "should support setting pragma values" do
+      @db.pragma_set(:auto_vacuum, '1')
+      @db.pragma_get(:auto_vacuum).to_s.should == '1'
+      @db.pragma_set(:auto_vacuum, '2')
+      @db.pragma_get(:auto_vacuum).to_s.should == '2'
+    end
+    
+    specify "should support getting and setting the auto_vacuum pragma" do
+      @db.auto_vacuum = :full
+      @db.auto_vacuum.should == :full
+      @db.auto_vacuum = :incremental
+      @db.auto_vacuum.should == :incremental
+      
+      proc {@db.auto_vacuum = :invalid}.should raise_error(Sequel::Error)
+    end
+  end
+  
+  specify "should provide a list of existing tables" do
+    @db.drop_table(:testing) rescue nil
+    @db.tables.should be_a_kind_of(Array)
+    @db.tables.should_not include(:testing)
+    @db.create_table! :testing do
+      text :name
+    end
+    @db.tables.should include(:testing)
+  end
+
+  specify "should support getting and setting the synchronous pragma" do
+    @db.synchronous = :off
+    @db.synchronous.should == :off
+    @db.synchronous = :normal
+    @db.synchronous.should == :normal
+    @db.synchronous = :full
+    @db.synchronous.should == :full
+    
+    proc {@db.synchronous = :invalid}.should raise_error(Sequel::Error)
+  end
+  
+  specify "should support getting and setting the temp_store pragma" do
+    @db.temp_store = :default
+    @db.temp_store.should == :default
+    @db.temp_store = :file
+    @db.temp_store.should == :file
+    @db.temp_store = :memory
+    @db.temp_store.should == :memory
+    
+    proc {@db.temp_store = :invalid}.should raise_error(Sequel::Error)
+  end
+  
+  specify "should support timestamps and datetimes and respect datetime_class" do
+    @db.create_table!(:time){timestamp :t; datetime :d}
+    t1 = Time.at(1)
+    @db[:time] << {:t => t1, :d => t1.to_i}
+    @db[:time] << {:t => t1.to_i, :d => t1}
+    @db[:time].map(:t).should == [t1, t1]
+    @db[:time].map(:d).should == [t1, t1]
+    Sequel.datetime_class = DateTime
+    t2 = Sequel.string_to_datetime(t1.iso8601)
+    @db[:time].map(:t).should == [t2, t2]
+    @db[:time].map(:d).should == [t2, t2]
+  end
+  
+  specify "should support sequential primary keys" do
+    @db.create_table!(:with_pk) {primary_key :id; text :name}
+    @db[:with_pk] << {:name => 'abc'}
+    @db[:with_pk] << {:name => 'def'}
+    @db[:with_pk] << {:name => 'ghi'}
+    @db[:with_pk].order(:name).all.should == [
+      {:id => 1, :name => 'abc'},
+      {:id => 2, :name => 'def'},
+      {:id => 3, :name => 'ghi'}
+    ]
+  end
+  
+  specify "should correctly parse the schema" do
+    @db.create_table!(:time2) {timestamp :t}
+    @db.schema(:time2, :reload=>true).should == [[:t, {:type=>:datetime, :allow_null=>true, :default=>nil, :db_type=>"timestamp", :primary_key=>false}]]
+  end
+end
+
+context "An SQLite dataset" do
+  before do
+    @d = SQLITE_DB[:items]
+  end
+  
+  specify "should handle string pattern matches correctly" do
+    @d.literal(:x.like('a')).should == "(x LIKE 'a')"
+    @d.literal(~:x.like('a')).should == "NOT (x LIKE 'a')"
+    @d.literal(:x.ilike('a')).should == "(x LIKE 'a')"
+    @d.literal(~:x.ilike('a')).should == "NOT (x LIKE 'a')"
+  end
+
+  specify "should raise errors if given a regexp pattern match" do
+    proc{@d.literal(:x.like(/a/))}.should raise_error(Sequel::Error)
+    proc{@d.literal(~:x.like(/a/))}.should raise_error(Sequel::Error)
+    proc{@d.literal(:x.like(/a/i))}.should raise_error(Sequel::Error)
+    proc{@d.literal(~:x.like(/a/i))}.should raise_error(Sequel::Error)
+  end
+end
+
+context "An SQLite numeric column" do
+  specify "should handle and return BigDecimal values" do
+    SQLITE_DB.create_table!(:d){numeric :d}
+    d = SQLITE_DB[:d]
+    d.insert(:d=>BigDecimal.new('80.0'))
+    d.insert(:d=>BigDecimal.new('NaN'))
+    d.insert(:d=>BigDecimal.new('Infinity'))
+    d.insert(:d=>BigDecimal.new('-Infinity'))
+    ds = d.all
+    ds.shift.should == {:d=>BigDecimal.new('80.0')}
+    ds.map{|x| x[:d].to_s}.should == %w'NaN Infinity -Infinity'
+  end
+end
+
+context "An SQLite dataset AS clause" do
+  specify "should use a string literal for :col___alias" do
+    SQLITE_DB.literal(:c___a).should == "c AS 'a'"
+  end
+
+  specify "should use a string literal for :table__col___alias" do
+    SQLITE_DB.literal(:t__c___a).should == "t.c AS 'a'"
+  end
+
+  specify "should use a string literal for :column.as(:alias)" do
+    SQLITE_DB.literal(:c.as(:a)).should == "c AS 'a'"
+  end
+
+  specify "should use a string literal in the SELECT clause" do
+    SQLITE_DB[:t].select(:c___a).sql.should == "SELECT c AS 'a' FROM t"
+  end
+
+  specify "should use a string literal in the FROM clause" do
+    SQLITE_DB[:t___a].sql.should == "SELECT * FROM t AS 'a'"
+  end
+
+  specify "should use a string literal in the JOIN clause" do
+    SQLITE_DB[:t].join_table(:natural, :j, nil, :a).sql.should == "SELECT * FROM t NATURAL JOIN j AS 'a'"
+  end
+end
+
+context "SQLite::Dataset#delete" do
+  before do
+    SQLITE_DB.create_table! :items do
+      primary_key :id
+      String :name
+      Float :value
+    end
+    @d = SQLITE_DB[:items]
+    @d.delete # remove all records
+    @d << {:name => 'abc', :value => 1.23}
+    @d << {:name => 'def', :value => 4.56}
+    @d << {:name => 'ghi', :value => 7.89}
+  end
+  
+  specify "should return the number of records affected when filtered" do
+    @d.count.should == 3
+    @d.filter(:value.sql_number < 3).delete.should == 1
+    @d.count.should == 2
+
+    @d.filter(:value.sql_number < 3).delete.should == 0
+    @d.count.should == 2
+  end
+  
+  specify "should return the number of records affected when unfiltered" do
+    @d.count.should == 3
+    @d.delete.should == 3
+    @d.count.should == 0
+
+    @d.delete.should == 0
+  end
+end
+
+context "SQLite::Dataset#update" do
+  before do
+    SQLITE_DB.create_table! :items do
+      primary_key :id
+      String :name
+      Float :value
+    end
+    @d = SQLITE_DB[:items]
+    @d.delete # remove all records
+    @d << {:name => 'abc', :value => 1.23}
+    @d << {:name => 'def', :value => 4.56}
+    @d << {:name => 'ghi', :value => 7.89}
+  end
+  
+  specify "should return the number of records affected" do
+    @d.filter(:name => 'abc').update(:value => 2).should == 1
+    
+    @d.update(:value => 10).should == 3
+    
+    @d.filter(:name => 'xxx').update(:value => 23).should == 0
+  end
+end
+
+context "SQLite dataset" do
+  before do
+    SQLITE_DB.create_table! :test do
+      primary_key :id
+      String :name
+      Float :value
+    end
+    SQLITE_DB.create_table! :items do
+      primary_key :id
+      String :name
+      Float :value
+    end
+    @d = SQLITE_DB[:items]
+    @d << {:name => 'abc', :value => 1.23}
+    @d << {:name => 'def', :value => 4.56}
+    @d << {:name => 'ghi', :value => 7.89}
+  end
+  after do
+    SQLITE_DB.drop_table(:test, :items)
+  end
+  
+  specify "should be able to insert from a subquery" do
+    SQLITE_DB[:test] << @d
+    SQLITE_DB[:test].count.should == 3
+    SQLITE_DB[:test].select(:name, :value).order(:value).to_a.should == \
+      @d.select(:name, :value).order(:value).to_a
+  end
+end
+
+context "A SQLite database" do
+  before do
+    @db = SQLITE_DB
+    @db.create_table! :test2 do
+      text :name
+      integer :value
+    end
+  end
+
+  specify "should support add_column operations" do
+    @db.add_column :test2, :xyz, :text
+    
+    @db[:test2].columns.should == [:name, :value, :xyz]
+    @db[:test2] << {:name => 'mmm', :value => 111, :xyz=>'000'}
+    @db[:test2].first.should == {:name => 'mmm', :value => 111, :xyz=>'000'}
+  end
+  
+  specify "should support drop_column operations" do
+    @db.drop_column :test2, :value
+    @db[:test2].columns.should == [:name]
+    @db[:test2] << {:name => 'mmm'}
+    @db[:test2].first.should == {:name => 'mmm'}
+  end
+  
+  specify "should support drop_column operations in a transaction" do
+    @db.transaction{@db.drop_column :test2, :value}
+    @db[:test2].columns.should == [:name]
+    @db[:test2] << {:name => 'mmm'}
+    @db[:test2].first.should == {:name => 'mmm'}
+  end
+
+  specify "should keep column attributes when dropping a column" do
+    @db.create_table! :test3 do
+      primary_key :id
+      text :name
+      integer :value
+    end
+
+    # This lame set of additions and deletions are to test that the primary keys
+    # don't get messed up when we recreate the database.
+    @db[:test3] << { :name => "foo", :value => 1}
+    @db[:test3] << { :name => "foo", :value => 2}
+    @db[:test3] << { :name => "foo", :value => 3}
+    @db[:test3].filter(:id => 2).delete
+    
+    @db.drop_column :test3, :value
+
+    @db['PRAGMA table_info(?)', :test3][:id][:pk].to_i.should == 1
+    @db[:test3].select(:id).all.should == [{:id => 1}, {:id => 3}]
+  end
+
+  specify "should support rename_column operations" do
+    @db[:test2].delete
+    @db.add_column :test2, :xyz, :text
+    @db[:test2] << {:name => 'mmm', :value => 111, :xyz => 'qqqq'}
+
+    @db[:test2].columns.should == [:name, :value, :xyz]
+    @db.rename_column :test2, :xyz, :zyx, :type => :text
+    @db[:test2].columns.should == [:name, :value, :zyx]
+    @db[:test2].first[:zyx].should == 'qqqq'
+    @db[:test2].count.should eql(1)
+  end
+  
+  specify "should preserve defaults when dropping or renaming columns" do
+    @db.create_table! :test3 do
+      String :s, :default=>'a'
+      Integer :i
+    end
+
+    @db[:test3].insert
+    @db[:test3].first[:s].should == 'a'
+    @db[:test3].delete
+    @db.drop_column :test3, :i
+    @db[:test3].insert
+    @db[:test3].first[:s].should == 'a'
+    @db[:test3].delete
+    @db.rename_column :test3, :s, :t
+    @db[:test3].insert
+    @db[:test3].first[:t].should == 'a'
+    @db[:test3].delete
+  end
+  
+  specify "should handle quoted tables when dropping or renaming columns" do
+    @db.quote_identifiers = true
+    table_name = "T T"
+    @db.drop_table(table_name) rescue nil
+    @db.create_table! table_name do
+      Integer :"s s"
+      Integer :"i i"
+    end
+
+    @db.from(table_name).insert(:"s s"=>1, :"i i"=>2)
+    @db.from(table_name).all.should == [{:"s s"=>1, :"i i"=>2}]
+    @db.drop_column table_name, :"i i"
+    @db.from(table_name).all.should == [{:"s s"=>1}]
+    @db.rename_column table_name, :"s s", :"t t"
+    @db.from(table_name).all.should == [{:"t t"=>1}]
+  end
+  
+  specify "should choose a temporary table name that isn't already used when dropping or renaming columns" do
+    @db.create_table! :test3 do
+      Integer :h
+      Integer :i
+    end
+    @db.create_table! :test3_backup0 do
+      Integer :j
+    end
+    @db.create_table! :test3_backup1 do
+      Integer :k
+    end
+
+    @db[:test3].columns.should == [:h, :i]
+    @db[:test3_backup0].columns.should == [:j]
+    @db[:test3_backup1].columns.should == [:k]
+    sqls = @db.drop_column(:test3, :i)
+    sqls.any?{|x| x =~ /test3_backup2/}.should == true
+    sqls.any?{|x| x =~ /test3_backup[01]/}.should == false
+    @db[:test3].columns.should == [:h]
+    @db[:test3_backup0].columns.should == [:j]
+    @db[:test3_backup1].columns.should == [:k]
+
+    @db.create_table! :test3_backup2 do
+      Integer :l
+    end
+
+    sqls = @db.rename_column(:test3, :h, :i)
+    sqls.any?{|x| x =~ /test3_backup3/}.should == true
+    sqls.any?{|x| x =~ /test3_backup[012]/}.should == false
+    @db[:test3].columns.should == [:i]
+    @db[:test3_backup0].columns.should == [:j]
+    @db[:test3_backup1].columns.should == [:k]
+    @db[:test3_backup2].columns.should == [:l]
+  end
+  
+  specify "should not support set_column_type operations" do
+    proc {@db.set_column_type :test2, :value, :integer}.should raise_error(Sequel::Error)
+  end
+  
+  specify "should support add_index" do
+    @db.add_index :test2, :value, :unique => true
+    @db.add_index :test2, [:name, :value]
+  end
+  
+  specify "should support drop_index" do
+    @db.add_index :test2, :value, :unique => true
+    @db.drop_index :test2, :value
+  end
+end  
diff -pruN 1.5.1-1/spec/associations_spec.rb 3.0.0-1/spec/associations_spec.rb
--- 1.5.1-1/spec/associations_spec.rb	2008-04-22 20:16:28.000000000 +0100
+++ 3.0.0-1/spec/associations_spec.rb	1970-01-01 01:00:00.000000000 +0100
@@ -1,878 +0,0 @@
-require File.join(File.dirname(__FILE__), "spec_helper")
-
-describe Sequel::Model, "associate" do
-  it "should use explicit class if given a class, symbol, or string" do
-    MODEL_DB.reset
-    klass = Class.new(Sequel::Model(:nodes))
-    class ParParent < Sequel::Model
-    end
-    
-    klass.associate :many_to_one, :par_parent0, :class=>ParParent
-    klass.associate :one_to_many, :par_parent1s, :class=>'ParParent'
-    klass.associate :many_to_many, :par_parent2s, :class=>:ParParent
-    
-    klass.send(:associated_class, klass.association_reflection(:"par_parent0")).should == ParParent
-    klass.send(:associated_class, klass.association_reflection(:"par_parent1s")).should == ParParent
-    klass.send(:associated_class, klass.association_reflection(:"par_parent2s")).should == ParParent
-  end
-end
-
-describe Sequel::Model, "many_to_one" do
-
-  before(:each) do
-    MODEL_DB.reset
-
-    @c2 = Class.new(Sequel::Model(:nodes)) do
-      columns :id, :parent_id, :par_parent_id, :blah
-      def self.create_new(values)
-        obj = self.new(values)
-        obj.instance_variable_set(:@new, false)
-        obj
-      end
-    end
-
-    @dataset = @c2.dataset
-  end
-
-  it "should use implicit key if omitted" do
-    @c2.many_to_one :parent, :class => @c2
-
-    d = @c2.new(:id => 1, :parent_id => 234)
-    p = d.parent
-    p.class.should == @c2
-    p.values.should == {:x => 1, :id => 1}
-
-    MODEL_DB.sqls.should == ["SELECT * FROM nodes WHERE (id = 234) LIMIT 1"]
-  end
-  
-  it "should use implicit class if omitted" do
-    class ParParent < Sequel::Model
-    end
-    
-    @c2.many_to_one :par_parent
-    
-    d = @c2.new(:id => 1, :par_parent_id => 234)
-    p = d.par_parent
-    p.class.should == ParParent
-    
-    MODEL_DB.sqls.should == ["SELECT * FROM par_parents WHERE (id = 234) LIMIT 1"]
-  end
-
-  it "should use class inside module if given as a string" do
-    module Par 
-      class Parent < Sequel::Model
-      end
-    end
-    
-    @c2.many_to_one :par_parent, :class=>"Par::Parent"
-    
-    d = @c2.new(:id => 1, :par_parent_id => 234)
-    p = d.par_parent
-    p.class.should == Par::Parent
-    
-    MODEL_DB.sqls.should == ["SELECT * FROM parents WHERE (id = 234) LIMIT 1"]
-  end
-
-  it "should use explicit key if given" do
-    @c2.many_to_one :parent, :class => @c2, :key => :blah
-
-    d = @c2.new(:id => 1, :blah => 567)
-    p = d.parent
-    p.class.should == @c2
-    p.values.should == {:x => 1, :id => 1}
-
-    MODEL_DB.sqls.should == ["SELECT * FROM nodes WHERE (id = 567) LIMIT 1"]
-  end
-
-  it "should return nil if key value is nil" do
-    @c2.many_to_one :parent, :class => @c2
-
-    d = @c2.new(:id => 1)
-    d.parent.should == nil
-  end
-
-  it "should cache negative lookup" do
-    @c2.many_to_one :parent, :class => @c2
-    ds = @c2.dataset
-    def ds.fetch_rows(sql, &block)
-      MODEL_DB.sqls << sql
-    end
-
-    d = @c2.new(:id => 1, :parent_id=>555)
-    MODEL_DB.sqls.should == []
-    d.parent.should == nil
-    MODEL_DB.sqls.should == ['SELECT * FROM nodes WHERE (id = 555) LIMIT 1']
-    d.parent.should == nil
-    MODEL_DB.sqls.should == ['SELECT * FROM nodes WHERE (id = 555) LIMIT 1']
-  end
-
-  it "should define a setter method" do
-    @c2.many_to_one :parent, :class => @c2
-
-    d = @c2.new(:id => 1)
-    d.parent = @c2.new(:id => 4321)
-    d.values.should == {:id => 1, :parent_id => 4321}
-
-    d.parent = nil
-    d.values.should == {:id => 1, :parent_id => nil}
-
-    e = @c2.new(:id => 6677)
-    d.parent = e
-    d.values.should == {:id => 1, :parent_id => 6677}
-  end
-  
-  it "should not persist changes until saved" do
-    @c2.many_to_one :parent, :class => @c2
-
-    d = @c2.create_new(:id => 1)
-    MODEL_DB.reset
-    d.parent = @c2.new(:id => 345)
-    MODEL_DB.sqls.should == []
-    d.save_changes
-    MODEL_DB.sqls.should == ['UPDATE nodes SET parent_id = 345 WHERE (id = 1)']
-  end
-
-  it "should set cached instance variable when accessed" do
-    @c2.many_to_one :parent, :class => @c2
-
-    d = @c2.create_new(:id => 1)
-    MODEL_DB.reset
-    d.parent_id = 234
-    d.instance_variable_get("@parent").should == nil
-    ds = @c2.dataset
-    def ds.fetch_rows(sql, &block); MODEL_DB.sqls << sql; yield({:id=>234}) end
-    e = d.parent 
-    MODEL_DB.sqls.should == ["SELECT * FROM nodes WHERE (id = 234) LIMIT 1"]
-    d.instance_variable_get("@parent").should == e
-  end
-
-  it "should set cached instance variable when assigned" do
-    @c2.many_to_one :parent, :class => @c2
-
-    d = @c2.create(:id => 1)
-    MODEL_DB.reset
-    d.instance_variable_get("@parent").should == nil
-    d.parent = @c2.new(:id => 234)
-    e = d.parent 
-    d.instance_variable_get("@parent").should == e
-    MODEL_DB.sqls.should == []
-  end
-
-  it "should use cached instance variable if available" do
-    @c2.many_to_one :parent, :class => @c2
-
-    d = @c2.create(:id => 1, :parent_id => 234)
-    MODEL_DB.reset
-    d.instance_variable_set(:@parent, 42)
-    d.parent.should == 42
-    MODEL_DB.sqls.should == []
-  end
-
-  it "should not use cached instance variable if asked to reload" do
-    @c2.many_to_one :parent, :class => @c2
-
-    d = @c2.create(:id => 1)
-    MODEL_DB.reset
-    d.parent_id = 234
-    d.instance_variable_set(:@parent, 42)
-    d.parent(true).should_not == 42 
-    MODEL_DB.sqls.should == ["SELECT * FROM nodes WHERE (id = 234) LIMIT 1"]
-  end
-  
-  it "should have the setter add to the reciprocal one_to_many cached association list if it exists" do
-    @c2.many_to_one :parent, :class => @c2
-    @c2.one_to_many :children, :class => @c2, :key=>:parent_id
-    ds = @c2.dataset
-    def ds.fetch_rows(sql, &block)
-      MODEL_DB.sqls << sql
-    end
-
-    d = @c2.new(:id => 1)
-    e = @c2.new(:id => 2)
-    MODEL_DB.sqls.should == []
-    d.parent = e
-    e.children.should_not(include(d))
-    MODEL_DB.sqls.should == ['SELECT * FROM nodes WHERE (parent_id = 2)']
-
-    MODEL_DB.reset
-    d = @c2.new(:id => 1)
-    e = @c2.new(:id => 2)
-    e.children.should_not(include(d))
-    MODEL_DB.sqls.should == ['SELECT * FROM nodes WHERE (parent_id = 2)']
-    d.parent = e
-    e.children.should(include(d))
-    MODEL_DB.sqls.should == ['SELECT * FROM nodes WHERE (parent_id = 2)']
-  end
-
-  it "should have the setter remove the object from the previous associated object's reciprocal one_to_many cached association list if it exists" do
-    @c2.many_to_one :parent, :class => @c2
-    @c2.one_to_many :children, :class => @c2, :key=>:parent_id
-    ds = @c2.dataset
-    def ds.fetch_rows(sql, &block)
-      MODEL_DB.sqls << sql
-    end
-
-    d = @c2.new(:id => 1)
-    e = @c2.new(:id => 2)
-    f = @c2.new(:id => 3)
-    e.children.should_not(include(d))
-    f.children.should_not(include(d))
-    MODEL_DB.reset
-    d.parent = e
-    e.children.should(include(d))
-    d.parent = f
-    f.children.should(include(d))
-    e.children.should_not(include(d))
-    d.parent = nil
-    f.children.should_not(include(d))
-    MODEL_DB.sqls.should == []
-  end
-
-  it "should have belongs_to alias" do
-    @c2.belongs_to :parent, :class => @c2
-
-    d = @c2.create_new(:id => 1)
-    MODEL_DB.reset
-    d.parent_id = 234
-    d.instance_variable_get("@parent").should == nil
-    ds = @c2.dataset
-    def ds.fetch_rows(sql, &block); MODEL_DB.sqls << sql; yield({:id=>234}) end
-    e = d.parent 
-    MODEL_DB.sqls.should == ["SELECT * FROM nodes WHERE (id = 234) LIMIT 1"]
-    d.instance_variable_get("@parent").should == e
-  end
-end
-
-describe Sequel::Model, "one_to_many" do
-
-  before(:each) do
-    MODEL_DB.reset
-
-    @c1 = Class.new(Sequel::Model(:attributes)) do
-      columns :id, :node_id
-    end
-
-    @c2 = Class.new(Sequel::Model(:nodes)) do
-      attr_accessor :xxx
-      
-      def self.name; 'Node'; end
-      def self.to_s; 'Node'; end
-      columns :id
-    end
-    @dataset = @c2.dataset
-    
-    @c2.dataset.extend(Module.new {
-      def fetch_rows(sql)
-        @db << sql
-        yield Hash.new
-      end
-    })
-
-    @c1.dataset.extend(Module.new {
-      def fetch_rows(sql)
-        @db << sql
-        yield Hash.new
-      end
-    })
-  end
-
-  it "should use implicit key if omitted" do
-    @c2.one_to_many :attributes, :class => @c1 
-
-    n = @c2.new(:id => 1234)
-    a = n.attributes_dataset
-    a.should be_a_kind_of(Sequel::Dataset)
-    a.sql.should == 'SELECT * FROM attributes WHERE (node_id = 1234)'
-  end
-  
-  it "should use implicit class if omitted" do
-    class HistoricalValue < Sequel::Model
-    end
-    
-    @c2.one_to_many :historical_values
-    
-    n = @c2.new(:id => 1234)
-    v = n.historical_values_dataset
-    v.should be_a_kind_of(Sequel::Dataset)
-    v.sql.should == 'SELECT * FROM historical_values WHERE (node_id = 1234)'
-    v.model_classes.should == {nil => HistoricalValue}
-  end
-  
-  it "should use class inside a module if given as a string" do
-    module Historical
-      class Value < Sequel::Model
-      end
-    end
-    
-    @c2.one_to_many :historical_values, :class=>'Historical::Value'
-    
-    n = @c2.new(:id => 1234)
-    v = n.historical_values_dataset
-    v.should be_a_kind_of(Sequel::Dataset)
-    v.sql.should == 'SELECT * FROM values WHERE (node_id = 1234)'
-    v.model_classes.should == {nil => Historical::Value}
-  end
-
-  it "should use explicit key if given" do
-    @c2.one_to_many :attributes, :class => @c1, :key => :nodeid
-    
-    n = @c2.new(:id => 1234)
-    a = n.attributes_dataset
-    a.should be_a_kind_of(Sequel::Dataset)
-    a.sql.should == 'SELECT * FROM attributes WHERE (nodeid = 1234)'
-  end
-
-  it "should define an add_ method" do
-    @c2.one_to_many :attributes, :class => @c1
-    
-    n = @c2.new(:id => 1234)
-    a = @c1.new(:id => 2345)
-    a.save!
-    MODEL_DB.reset
-    a.should == n.add_attribute(a)
-    MODEL_DB.sqls.should == ['UPDATE attributes SET node_id = 1234 WHERE (id = 2345)']
-  end
-
-  it "should define a remove_ method" do
-    @c2.one_to_many :attributes, :class => @c1
-    
-    n = @c2.new(:id => 1234)
-    a = @c1.new(:id => 2345)
-    a.save!
-    MODEL_DB.reset
-    a.should == n.remove_attribute(a)
-    MODEL_DB.sqls.should == ['UPDATE attributes SET node_id = NULL WHERE (id = 2345)']
-  end
-  
-  it "should support an order option" do
-    @c2.one_to_many :attributes, :class => @c1, :order => :kind
-
-    n = @c2.new(:id => 1234)
-    n.attributes_dataset.sql.should == "SELECT * FROM attributes WHERE (node_id = 1234) ORDER BY kind"
-  end
-  
-  it "should support an array for the order option" do
-    @c2.one_to_many :attributes, :class => @c1, :order => [:kind1, :kind2]
-
-    n = @c2.new(:id => 1234)
-    n.attributes_dataset.sql.should == "SELECT * FROM attributes WHERE (node_id = 1234) ORDER BY kind1, kind2"
-  end
-  
-  it "should return array with all members of the association" do
-    @c2.one_to_many :attributes, :class => @c1
-    
-    n = @c2.new(:id => 1234)
-    atts = n.attributes
-    atts.should be_a_kind_of(Array)
-    atts.size.should == 1
-    atts.first.should be_a_kind_of(@c1)
-    atts.first.values.should == {}
-    
-    MODEL_DB.sqls.should == ['SELECT * FROM attributes WHERE (node_id = 1234)']
-  end
-  
-  it "should accept a block" do
-    @c2.one_to_many :attributes, :class => @c1 do |ds|
-      ds.filter(:xxx => @xxx)
-    end
-    
-    n = @c2.new(:id => 1234)
-    atts = n.attributes
-    atts.should be_a_kind_of(Array)
-    atts.size.should == 1
-    atts.first.should be_a_kind_of(@c1)
-    atts.first.values.should == {}
-    
-    MODEL_DB.sqls.should == ['SELECT * FROM attributes WHERE (node_id = 1234) AND (xxx IS NULL)']
-  end
-  
-  it "should support order option with block" do
-    @c2.one_to_many :attributes, :class => @c1, :order => :kind do |ds|
-      ds.filter(:xxx => @xxx)
-    end
-    
-    n = @c2.new(:id => 1234)
-    atts = n.attributes
-    atts.should be_a_kind_of(Array)
-    atts.size.should == 1
-    atts.first.should be_a_kind_of(@c1)
-    atts.first.values.should == {}
-    
-    MODEL_DB.sqls.should == ['SELECT * FROM attributes WHERE (node_id = 1234) AND (xxx IS NULL) ORDER BY kind']
-  end
-  
-  it "should set cached instance variable when accessed" do
-    @c2.one_to_many :attributes, :class => @c1
-
-    n = @c2.new(:id => 1234)
-    MODEL_DB.reset
-    n.instance_variables.include?("@attributes").should == false
-    atts = n.attributes
-    atts.should == n.instance_variable_get("@attributes")
-    MODEL_DB.sqls.should == ['SELECT * FROM attributes WHERE (node_id = 1234)']
-  end
-
-  it "should use cached instance variable if available" do
-    @c2.one_to_many :attributes, :class => @c1
-
-    n = @c2.new(:id => 1234)
-    MODEL_DB.reset
-    n.instance_variable_set(:@attributes, 42)
-    n.attributes.should == 42
-    MODEL_DB.sqls.should == []
-  end
-
-  it "should not use cached instance variable if asked to reload" do
-    @c2.one_to_many :attributes, :class => @c1
-
-    n = @c2.new(:id => 1234)
-    MODEL_DB.reset
-    n.instance_variable_set(:@attributes, 42)
-    n.attributes(true).should_not == 42
-    MODEL_DB.sqls.should == ['SELECT * FROM attributes WHERE (node_id = 1234)']
-  end
-
-  it "should add item to cached instance variable if it exists when calling add_" do
-    @c2.one_to_many :attributes, :class => @c1
-
-    n = @c2.new(:id => 1234)
-    att = @c1.new(:id => 345)
-    MODEL_DB.reset
-    a = []
-    n.instance_variable_set(:@attributes, a)
-    n.add_attribute(att)
-    a.should == [att]
-  end
-
-  it "should set object to item's reciprocal cached association variable when calling add_" do
-    @c2.one_to_many :attributes, :class => @c1
-    @c1.many_to_one :node, :class => @c2
-
-    n = @c2.new(:id => 1234)
-    att = @c1.new(:id => 345)
-    n.add_attribute(att)
-    att.node.should == n
-  end
-
-  it "should remove item from cached instance variable if it exists when calling remove_" do
-    @c2.one_to_many :attributes, :class => @c1
-
-    n = @c2.new(:id => 1234)
-    att = @c1.new(:id => 345)
-    MODEL_DB.reset
-    a = [att]
-    n.instance_variable_set(:@attributes, a)
-    n.remove_attribute(att)
-    a.should == []
-  end
-
-  it "should remove item's reciprocal cached association variable when calling remove_" do
-    @c2.one_to_many :attributes, :class => @c1
-    @c1.many_to_one :node, :class => @c2
-
-    n = @c2.new(:id => 1234)
-    att = @c1.new(:id => 345)
-    att.instance_variable_set(:@node, n)
-    att.node.should == n
-    n.remove_attribute(att)
-    att.node.should == nil
-  end
-
-  it "should have has_many alias" do
-    @c2.has_many :attributes, :class => @c1
-    
-    n = @c2.new(:id => 1234)
-    atts = n.attributes
-    atts.should be_a_kind_of(Array)
-    atts.size.should == 1
-    atts.first.should be_a_kind_of(@c1)
-    atts.first.values.should == {}
-    
-    MODEL_DB.sqls.should == ['SELECT * FROM attributes WHERE (node_id = 1234)']
-  end
-  
-  it "should populate the reciprocal many_to_one instance variable when loading the one_to_many association" do
-    @c2.one_to_many :attributes, :class => @c1, :key => :node_id
-    @c1.many_to_one :node, :class => @c2, :key => :node_id
-    
-    n = @c2.new(:id => 1234)
-    atts = n.attributes
-    MODEL_DB.sqls.should == ['SELECT * FROM attributes WHERE (node_id = 1234)']
-    atts.should be_a_kind_of(Array)
-    atts.size.should == 1
-    atts.first.should be_a_kind_of(@c1)
-    atts.first.values.should == {}
-    atts.first.node.should == n
-    
-    MODEL_DB.sqls.length.should == 1
-  end
-  
-  it "should use an explicit reciprocal instance variable if given" do
-    @c2.one_to_many :attributes, :class => @c1, :key => :node_id, :reciprocal=>'@wxyz'
-    
-    n = @c2.new(:id => 1234)
-    atts = n.attributes
-    MODEL_DB.sqls.should == ['SELECT * FROM attributes WHERE (node_id = 1234)']
-    atts.should be_a_kind_of(Array)
-    atts.size.should == 1
-    atts.first.should be_a_kind_of(@c1)
-    atts.first.values.should == {}
-    atts.first.instance_variable_get('@wxyz').should == n
-    
-    MODEL_DB.sqls.length.should == 1
-  end
-end
-
-describe Sequel::Model, "many_to_many" do
-
-  before(:each) do
-    MODEL_DB.reset
-
-    @c1 = Class.new(Sequel::Model(:attributes)) do
-      def self.name; 'Attribute'; end
-      def self.to_s; 'Attribute'; end
-      columns :id
-    end
-
-    @c2 = Class.new(Sequel::Model(:nodes)) do
-      attr_accessor :xxx
-      
-      def self.name; 'Node'; end
-      def self.to_s; 'Node'; end
-      columns :id
-    end
-    @dataset = @c2.dataset
-
-    [@c1, @c2].each do |c|
-      c.dataset.extend(Module.new {
-        def fetch_rows(sql)
-          @db << sql
-          yield Hash.new
-        end
-      })
-    end
-  end
-
-  it "should use implicit key values and join table if omitted" do
-    @c2.many_to_many :attributes, :class => @c1 
-
-    n = @c2.new(:id => 1234)
-    a = n.attributes_dataset
-    a.should be_a_kind_of(Sequel::Dataset)
-    ['SELECT attributes.* FROM attributes INNER JOIN attributes_nodes ON (attributes_nodes.attribute_id = attributes.id) AND (attributes_nodes.node_id = 1234)',
-     'SELECT attributes.* FROM attributes INNER JOIN attributes_nodes ON (attributes_nodes.node_id = 1234) AND (attributes_nodes.attribute_id = attributes.id)'
-    ].should(include(a.sql))
-  end
-  
-  it "should use implicit class if omitted" do
-    class Tag < Sequel::Model
-    end
-    
-    @c2.many_to_many :tags
-
-    n = @c2.new(:id => 1234)
-    a = n.tags_dataset
-    a.should be_a_kind_of(Sequel::Dataset)
-    ['SELECT tags.* FROM tags INNER JOIN nodes_tags ON (nodes_tags.tag_id = tags.id) AND (nodes_tags.node_id = 1234)',
-     'SELECT tags.* FROM tags INNER JOIN nodes_tags ON (nodes_tags.node_id = 1234) AND (nodes_tags.tag_id = tags.id)'
-    ].should(include(a.sql))
-  end
-  
-  it "should use class inside module if given as a string" do
-    module Historical
-      class Tag < Sequel::Model
-      end
-    end
-    
-    @c2.many_to_many :tags, :class=>'::Historical::Tag'
-
-    n = @c2.new(:id => 1234)
-    a = n.tags_dataset
-    a.should be_a_kind_of(Sequel::Dataset)
-    ['SELECT tags.* FROM tags INNER JOIN nodes_tags ON (nodes_tags.tag_id = tags.id) AND (nodes_tags.node_id = 1234)',
-     'SELECT tags.* FROM tags INNER JOIN nodes_tags ON (nodes_tags.node_id = 1234) AND (nodes_tags.tag_id = tags.id)'
-    ].should(include(a.sql))
-  end
-  
-  it "should use explicit key values and join table if given" do
-    @c2.many_to_many :attributes, :class => @c1, :left_key => :nodeid, :right_key => :attributeid, :join_table => :attribute2node
-    
-    n = @c2.new(:id => 1234)
-    a = n.attributes_dataset
-    a.should be_a_kind_of(Sequel::Dataset)
-    ['SELECT attributes.* FROM attributes INNER JOIN attribute2node ON (attribute2node.nodeid = 1234) AND (attribute2node.attributeid = attributes.id)',
-     'SELECT attributes.* FROM attributes INNER JOIN attribute2node ON (attribute2node.attributeid = attributes.id) AND (attribute2node.nodeid = 1234)'
-    ].should(include(a.sql))
-  end
-  
-  it "should support an order option" do
-    @c2.many_to_many :attributes, :class => @c1, :order => :blah
-
-    n = @c2.new(:id => 1234)
-    a = n.attributes_dataset
-    a.should be_a_kind_of(Sequel::Dataset)
-    ['SELECT attributes.* FROM attributes INNER JOIN attributes_nodes ON (attributes_nodes.attribute_id = attributes.id) AND (attributes_nodes.node_id = 1234) ORDER BY blah',
-     'SELECT attributes.* FROM attributes INNER JOIN attributes_nodes ON (attributes_nodes.node_id = 1234) AND (attributes_nodes.attribute_id = attributes.id) ORDER BY blah'
-    ].should(include(a.sql))
-  end
-  
-  it "should support an array for the order option" do
-    @c2.many_to_many :attributes, :class => @c1, :order => [:blah1, :blah2]
-
-    n = @c2.new(:id => 1234)
-    a = n.attributes_dataset
-    a.should be_a_kind_of(Sequel::Dataset)
-    ['SELECT attributes.* FROM attributes INNER JOIN attributes_nodes ON (attributes_nodes.attribute_id = attributes.id) AND (attributes_nodes.node_id = 1234) ORDER BY blah1, blah2',
-     'SELECT attributes.* FROM attributes INNER JOIN attributes_nodes ON (attributes_nodes.node_id = 1234) AND (attributes_nodes.attribute_id = attributes.id) ORDER BY blah1, blah2'
-    ].should(include(a.sql))
-  end
-  
-  it "should support a select option" do
-    @c2.many_to_many :attributes, :class => @c1, :select => :blah
-
-    n = @c2.new(:id => 1234)
-    a = n.attributes_dataset
-    a.should be_a_kind_of(Sequel::Dataset)
-    ['SELECT blah FROM attributes INNER JOIN attributes_nodes ON (attributes_nodes.attribute_id = attributes.id) AND (attributes_nodes.node_id = 1234)',
-     'SELECT blah FROM attributes INNER JOIN attributes_nodes ON (attributes_nodes.node_id = 1234) AND (attributes_nodes.attribute_id = attributes.id)'
-    ].should(include(a.sql))
-  end
-  
-  it "should support an array for the select option" do
-    @c2.many_to_many :attributes, :class => @c1, :select => [:attributes.*, :attribute_nodes__blah2]
-
-    n = @c2.new(:id => 1234)
-    a = n.attributes_dataset
-    a.should be_a_kind_of(Sequel::Dataset)
-    ['SELECT attributes.*, attribute_nodes.blah2 FROM attributes INNER JOIN attributes_nodes ON (attributes_nodes.attribute_id = attributes.id) AND (attributes_nodes.node_id = 1234)',
-     'SELECT attributes.*, attribute_nodes.blah2 FROM attributes INNER JOIN attributes_nodes ON (attributes_nodes.node_id = 1234) AND (attributes_nodes.attribute_id = attributes.id)'
-    ].should(include(a.sql))
-  end
-  
-  it "should accept a block" do
-    @c2.many_to_many :attributes, :class => @c1 do |ds|
-      ds.filter(:xxx => @xxx)
-    end
-
-    n = @c2.new(:id => 1234)
-    n.xxx = 555
-    a = n.attributes
-    a.should be_a_kind_of(Array)
-    a.size.should == 1
-    a.first.should be_a_kind_of(@c1)
-    ['SELECT attributes.* FROM attributes INNER JOIN attributes_nodes ON (attributes_nodes.attribute_id = attributes.id) AND (attributes_nodes.node_id = 1234) WHERE (xxx = 555)',
-     'SELECT attributes.* FROM attributes INNER JOIN attributes_nodes ON (attributes_nodes.node_id = 1234) AND (attributes_nodes.attribute_id = attributes.id) WHERE (xxx = 555)'
-    ].should(include(MODEL_DB.sqls.first))
-  end
-
-  it "should allow the order option while accepting a block" do
-    @c2.many_to_many :attributes, :class => @c1, :order=>[:blah1, :blah2] do |ds|
-      ds.filter(:xxx => @xxx)
-    end
-
-    n = @c2.new(:id => 1234)
-    n.xxx = 555
-    a = n.attributes
-    a.should be_a_kind_of(Array)
-    a.size.should == 1
-    a.first.should be_a_kind_of(@c1)
-    ['SELECT attributes.* FROM attributes INNER JOIN attributes_nodes ON (attributes_nodes.attribute_id = attributes.id) AND (attributes_nodes.node_id = 1234) WHERE (xxx = 555) ORDER BY blah1, blah2',
-     'SELECT attributes.* FROM attributes INNER JOIN attributes_nodes ON (attributes_nodes.node_id = 1234) AND (attributes_nodes.attribute_id = attributes.id) WHERE (xxx = 555) ORDER BY blah1, blah2'
-    ].should(include(MODEL_DB.sqls.first))
-  end
-
-  it "should define an add_ method" do
-    @c2.many_to_many :attributes, :class => @c1
-    
-    n = @c2.new(:id => 1234)
-    a = @c1.new(:id => 2345)
-    a.should == n.add_attribute(a)
-    ['INSERT INTO attributes_nodes (node_id, attribute_id) VALUES (1234, 2345)',
-     'INSERT INTO attributes_nodes (attribute_id, node_id) VALUES (2345, 1234)'
-    ].should(include(MODEL_DB.sqls.first))
-  end
-
-  it "should define a remove_ method" do
-    @c2.many_to_many :attributes, :class => @c1
-    
-    n = @c2.new(:id => 1234)
-    a = @c1.new(:id => 2345)
-    a.should == n.remove_attribute(a)
-    ['DELETE FROM attributes_nodes WHERE (node_id = 1234) AND (attribute_id = 2345)',
-     'DELETE FROM attributes_nodes WHERE (attribute_id = 2345) AND (node_id = 1234)'
-    ].should(include(MODEL_DB.sqls.first))
-  end
-
-  it "should provide an array with all members of the association" do
-    @c2.many_to_many :attributes, :class => @c1
-    
-    n = @c2.new(:id => 1234)
-    atts = n.attributes
-    atts.should be_a_kind_of(Array)
-    atts.size.should == 1
-    atts.first.should be_a_kind_of(@c1)
-
-    ['SELECT attributes.* FROM attributes INNER JOIN attributes_nodes ON (attributes_nodes.attribute_id = attributes.id) AND (attributes_nodes.node_id = 1234)',
-     'SELECT attributes.* FROM attributes INNER JOIN attributes_nodes ON (attributes_nodes.node_id = 1234) AND (attributes_nodes.attribute_id = attributes.id)'
-    ].should(include(MODEL_DB.sqls.first))
-  end
-
-  it "should set cached instance variable when accessed" do
-    @c2.many_to_many :attributes, :class => @c1
-
-    n = @c2.new(:id => 1234)
-    MODEL_DB.reset
-    n.instance_variables.include?("@attributes").should == false
-    atts = n.attributes
-    atts.should == n.instance_variable_get("@attributes")
-    MODEL_DB.sqls.length.should == 1
-  end
-
-  it "should use cached instance variable if available" do
-    @c2.many_to_many :attributes, :class => @c1
-
-    n = @c2.new(:id => 1234)
-    MODEL_DB.reset
-    n.instance_variable_set(:@attributes, 42)
-    n.attributes.should == 42
-    MODEL_DB.sqls.should == []
-  end
-
-  it "should not use cached instance variable if asked to reload" do
-    @c2.many_to_many :attributes, :class => @c1
-
-    n = @c2.new(:id => 1234)
-    MODEL_DB.reset
-    n.instance_variable_set(:@attributes, 42)
-    n.attributes(true).should_not == 42
-    MODEL_DB.sqls.length.should == 1
-  end
-
-  it "should add item to cached instance variable if it exists when calling add_" do
-    @c2.many_to_many :attributes, :class => @c1
-
-    n = @c2.new(:id => 1234)
-    att = @c1.new(:id => 345)
-    MODEL_DB.reset
-    a = []
-    n.instance_variable_set(:@attributes, a)
-    n.add_attribute(att)
-    a.should == [att]
-  end
-
-  it "should add item to reciprocal cached instance variable if it exists when calling add_" do
-    @c2.many_to_many :attributes, :class => @c1
-    @c1.many_to_many :nodes, :class => @c2
-
-    n = @c2.new(:id => 1234)
-    att = @c1.new(:id => 345)
-    att.instance_variable_set(:@nodes, [])
-    n.add_attribute(att)
-    att.nodes.should == [n]
-  end
-
-  it "should remove item from cached instance variable if it exists when calling remove_" do
-    @c2.many_to_many :attributes, :class => @c1
-
-    n = @c2.new(:id => 1234)
-    att = @c1.new(:id => 345)
-    MODEL_DB.reset
-    a = [att]
-    n.instance_variable_set(:@attributes, a)
-    n.remove_attribute(att)
-    a.should == []
-  end
-
-  it "should remove item from reciprocal cached instance variable if it exists when calling remove_" do
-    @c2.many_to_many :attributes, :class => @c1
-    @c1.many_to_many :nodes, :class => @c2
-
-    n = @c2.new(:id => 1234)
-    att = @c1.new(:id => 345)
-    att.instance_variable_set(:@nodes, [n])
-    n.remove_attribute(att)
-    att.nodes.should == []
-  end
-
-  it "should have has_and_belongs_to_many alias" do
-    @c2.has_and_belongs_to_many :attributes, :class => @c1 
-
-    n = @c2.new(:id => 1234)
-    a = n.attributes_dataset
-    a.should be_a_kind_of(Sequel::Dataset)
-    ['SELECT attributes.* FROM attributes INNER JOIN attributes_nodes ON (attributes_nodes.attribute_id = attributes.id) AND (attributes_nodes.node_id = 1234)',
-     'SELECT attributes.* FROM attributes INNER JOIN attributes_nodes ON (attributes_nodes.node_id = 1234) AND (attributes_nodes.attribute_id = attributes.id)'
-    ].should(include(a.sql))
-  end
-  
-end
-
-describe Sequel::Model, "all_association_reflections" do
-  before(:each) do
-    MODEL_DB.reset
-    @c1 = Class.new(Sequel::Model(:nodes)) do
-      def self.name; 'Node'; end
-      def self.to_s; 'Node'; end
-    end
-  end
-  
-  it "should include all association reflection hashes" do
-    @c1.all_association_reflections.should == []
-    @c1.associate :many_to_one, :parent, :class => @c1
-    @c1.all_association_reflections.should == [{
-      :type => :many_to_one, :name => :parent, :class_name => 'Node', 
-      :class => @c1, :key => :parent_id, :block => nil, :cache => true
-    }]
-    @c1.associate :one_to_many, :children, :class => @c1
-    @c1.all_association_reflections.sort_by{|x|x[:name].to_s}.should == [{
-      :type => :one_to_many, :name => :children, :class_name => 'Node', 
-      :class => @c1, :key => :node_id, :block => nil, :cache => true}, {
-      :type => :many_to_one, :name => :parent, :class_name => 'Node',
-      :class => @c1, :key => :parent_id, :block => nil, :cache => true}]
-  end
-end
-
-describe Sequel::Model, "association_reflection" do
-  before(:each) do
-    MODEL_DB.reset
-    @c1 = Class.new(Sequel::Model(:nodes)) do
-      def self.name; 'Node'; end
-      def self.to_s; 'Node'; end
-    end
-  end
-
-  it "should return nil for nonexistent association" do
-    @c1.association_reflection(:blah).should == nil
-  end
-
-  it "should return association reflection hash if association exists" do
-    @c1.associate :many_to_one, :parent, :class => @c1
-    @c1.association_reflection(:parent).should == {
-      :type => :many_to_one, :name => :parent, :class_name => 'Node', 
-      :class => @c1, :key => :parent_id, :block => nil, :cache => true
-    }
-    @c1.associate :one_to_many, :children, :class => @c1
-    @c1.association_reflection(:children).should == {
-      :type => :one_to_many, :name => :children, :class_name => 'Node', 
-      :class => @c1, :key => :node_id, :block => nil, :cache => true
-    }
-  end
-end
-
-describe Sequel::Model, "associations" do
-  before(:each) do
-    MODEL_DB.reset
-    @c1 = Class.new(Sequel::Model(:nodes)) do
-      def self.name; 'Node'; end
-      def self.to_s; 'Node'; end
-    end
-  end
-
-  it "should include all association names" do
-    @c1.associations.should == []
-    @c1.associate :many_to_one, :parent, :class => @c1
-    @c1.associations.should == [:parent]
-    @c1.associate :one_to_many, :children, :class => @c1
-    @c1.associations.sort_by{|x|x.to_s}.should == [:children, :parent]
-  end
-end
diff -pruN 1.5.1-1/spec/base_spec.rb 3.0.0-1/spec/base_spec.rb
--- 1.5.1-1/spec/base_spec.rb	2008-04-22 20:16:28.000000000 +0100
+++ 3.0.0-1/spec/base_spec.rb	1970-01-01 01:00:00.000000000 +0100
@@ -1,258 +0,0 @@
-require File.join(File.dirname(__FILE__), "spec_helper")
-
-describe "Model attribute setters" do
-
-  before(:each) do
-    MODEL_DB.reset
-
-    @c = Class.new(Sequel::Model(:items)) do
-      columns :id, :x, :y
-    end
-  end
-
-  it "should mark the column value as changed" do
-    o = @c.new
-    o.changed_columns.should == []
-
-    o.x = 2
-    o.changed_columns.should == [:x]
-
-    o.y = 3
-    o.changed_columns.should == [:x, :y]
-
-    o.changed_columns.clear
-
-    o[:x] = 2
-    o.changed_columns.should == [:x]
-
-    o[:y] = 3
-    o.changed_columns.should == [:x, :y]
-  end
-
-end
-
-describe "Model#serialize" do
-
-  before(:each) do
-    MODEL_DB.reset
-  end
-
-  it "should translate values to YAML when creating records" do
-    @c = Class.new(Sequel::Model(:items)) do
-      no_primary_key
-      serialize :abc
-      columns :abc
-    end
-
-    @c.create(:abc => 1)
-    @c.create(:abc => "hello")
-
-    MODEL_DB.sqls.should == [ \
-      "INSERT INTO items (abc) VALUES ('--- 1\n')", \
-      "INSERT INTO items (abc) VALUES ('--- hello\n')", \
-    ]
-  end
-
-  it "should support calling after the class is defined" do
-    @c = Class.new(Sequel::Model(:items)) do
-      no_primary_key
-      columns :def
-    end
-
-    @c.serialize :def
-
-    @c.create(:def => 1)
-    @c.create(:def => "hello")
-
-    MODEL_DB.sqls.should == [ \
-      "INSERT INTO items (def) VALUES ('--- 1\n')", \
-      "INSERT INTO items (def) VALUES ('--- hello\n')", \
-    ]
-  end
-
-  it "should support using the Marshal format" do
-    @c = Class.new(Sequel::Model(:items)) do
-      no_primary_key
-      serialize :abc, :format => :marshal
-      columns :abc
-    end
-
-    @c.create(:abc => 1)
-    @c.create(:abc => "hello")
-
-    MODEL_DB.sqls.should == [ \
-      "INSERT INTO items (abc) VALUES ('BAhpBg==\n')", \
-      "INSERT INTO items (abc) VALUES ('BAgiCmhlbGxv\n')", \
-    ]
-  end
-
-  it "should translate values to and from YAML using accessor methods" do
-    @c = Class.new(Sequel::Model(:items)) do
-      serialize :abc, :def
-      columns :abc, :def
-    end
-
-    ds = @c.dataset
-    ds.extend(Module.new {
-      attr_accessor :raw
-
-      def fetch_rows(sql, &block)
-        block.call(@raw)
-      end
-
-      @@sqls = nil
-
-      def insert(*args)
-        @@sqls = insert_sql(*args)
-      end
-
-      def update(*args)
-        @@sqls = update_sql(*args)
-      end
-
-      def sqls
-        @@sqls
-      end
-
-      def columns
-        [:id, :abc, :def]
-      end
-    }
-    )
-
-    ds.raw = {:id => 1, :abc => "--- 1\n", :def => "--- hello\n"}
-    o = @c.first
-    o.id.should == 1
-    o.abc.should == 1
-    o.def.should == "hello"
-
-    o.update_values(:abc => 23)
-    ds.sqls.should == "UPDATE items SET abc = '#{23.to_yaml}' WHERE (id = 1)"
-
-    ds.raw = {:id => 1, :abc => "--- 1\n", :def => "--- hello\n"}
-    o = @c.create(:abc => [1, 2, 3])
-    ds.sqls.should == "INSERT INTO items (abc) VALUES ('#{[1, 2, 3].to_yaml}')"
-  end
-
-end
-
-describe Sequel::Model, "dataset" do
-  before do
-    @a = Class.new(Sequel::Model(:items))
-    @b = Class.new(Sequel::Model)
-    
-    class Elephant < Sequel::Model(:ele1)
-    end
-    
-    class Maggot < Sequel::Model
-    end
-
-    class ShoeSize < Sequel::Model
-    end
-    
-    class BootSize < ShoeSize
-    end
-  end
-  
-  specify "should default to the plural of the class name" do
-    Maggot.dataset.sql.should == 'SELECT * FROM maggots'
-    ShoeSize.dataset.sql.should == 'SELECT * FROM shoe_sizes'
-  end
-  
-  specify "should return the dataset for the superclass if available" do
-    BootSize.dataset.sql.should == 'SELECT * FROM shoe_sizes'
-  end
-  
-  specify "should return the correct dataset if set explicitly" do
-    Elephant.dataset.sql.should == 'SELECT * FROM ele1'
-    @a.dataset.sql.should == 'SELECT * FROM items'
-  end
-  
-  specify "should raise if no dataset is explicitly set and the class is anonymous" do
-    proc {@b.dataset}.should raise_error(Sequel::Error)
-  end
-  
-  specify "should disregard namespaces for the table name" do
-    module BlahBlah
-      class MwaHaHa < Sequel::Model
-      end
-    end
-    
-    BlahBlah::MwaHaHa.dataset.sql.should == 'SELECT * FROM mwa_ha_has'
-  end
-end
-
-describe Sequel::Model, "def_dataset_method" do
-  setup do
-    @c = Class.new(Sequel::Model(:items)) do
-      @dataset = Object.new
-    end
-  end
-  
-  it "should add a method to the dataset and model if called with a block argument" do
-    @c.instance_eval do
-      def_dataset_method(:return_3){3}
-    end
-    @c.return_3.should == 3
-    @c.dataset.return_3.should == 3
-  end
-
-  it "should add all passed methods to the model if called without a block argument" do
-    @c.instance_eval do
-      def_dataset_method(:return_3, :return_4)
-    end
-    proc{@c.return_3}.should raise_error(NoMethodError)
-    proc{@c.return_4}.should raise_error(NoMethodError)
-    @c.dataset.instance_eval do
-      def return_3; 3; end
-      def return_4; 4; end
-    end
-    @c.return_3.should == 3
-    @c.return_4.should == 4
-  end
-end
-
-describe "A model class with implicit table name" do
-  setup do
-    class Donkey < Sequel::Model
-    end
-  end
-  
-  specify "should have a dataset associated with the model class" do
-    Donkey.dataset.model_classes.should == {nil => Donkey}
-  end
-end
-
-describe "A model inheriting from a model" do
-  setup do
-    class Feline < Sequel::Model
-    end
-    
-    class Leopard < Feline
-    end
-  end
-  
-  specify "should have a dataset associated with itself" do
-    Feline.dataset.model_classes.should == {nil => Feline}
-    Leopard.dataset.model_classes.should == {nil => Leopard}
-  end
-end
-
-describe "Model.db=" do
-  setup do
-    $db1 = Sequel::Database.new
-    $db2 = Sequel::Database.new
-    
-    class BlueBlue < Sequel::Model
-      set_dataset $db1[:blue]
-    end
-  end
-  
-  specify "should affect the underlying dataset" do
-    BlueBlue.db = $db2
-    
-    BlueBlue.dataset.db.should === $db2
-    BlueBlue.dataset.db.should_not === $db1
-  end
-end
-
diff -pruN 1.5.1-1/spec/caching_spec.rb 3.0.0-1/spec/caching_spec.rb
--- 1.5.1-1/spec/caching_spec.rb	2008-04-22 20:16:28.000000000 +0100
+++ 3.0.0-1/spec/caching_spec.rb	1970-01-01 01:00:00.000000000 +0100
@@ -1,153 +0,0 @@
-require File.join(File.dirname(__FILE__), "spec_helper")
-
-describe Sequel::Model, "caching" do
-
-  before(:each) do
-    MODEL_DB.reset
-    
-    @cache_class = Class.new(Hash) do
-      attr_accessor :ttl
-      def set(k, v, ttl); self[k] = v; @ttl = ttl; end
-      def get(k); self[k]; end
-    end
-    cache = @cache_class.new
-    @cache = cache
-    
-    @c = Class.new(Sequel::Model(:items)) do
-      set_cache cache
-      
-      columns :name, :id
-    end
-    
-    $cache_dataset_row = {:name => 'sharon', :id => 1}
-    @dataset = @c.dataset
-    $sqls = []
-    @dataset.extend(Module.new {
-      def fetch_rows(sql)
-        $sqls << sql
-        yield $cache_dataset_row
-      end
-      
-      def update(values)
-        $sqls << update_sql(values)
-        $cache_dataset_row.merge!(values)
-      end
-      
-      def delete
-        $sqls << delete_sql
-      end
-    })
-  end
-  
-  it "should set the model's cache store" do
-    @c.cache_store.should be(@cache)
-  end
-  
-  it "should have a default ttl of 3600" do
-    @c.cache_ttl.should == 3600
-  end
-  
-  it "should take a ttl option" do
-    @c.set_cache @cache, :ttl => 1234
-    @c.cache_ttl.should == 1234
-  end
-  
-  it "should offer a set_cache_ttl method for setting the ttl" do
-    @c.cache_ttl.should == 3600
-    @c.set_cache_ttl 1234
-    @c.cache_ttl.should == 1234
-  end
-  
-  it "should generate a cache key appropriate to the class" do
-    m = @c.new
-    m.values[:id] = 1
-    m.cache_key.should == "#{m.class}:1"
-    
-    # custom primary key
-    @c.set_primary_key :ttt
-    m = @c.new
-    m.values[:ttt] = 333
-    m.cache_key.should == "#{m.class}:333"
-    
-    # composite primary key
-    @c.set_primary_key [:a, :b, :c]
-    m = @c.new
-    m.values[:a] = 123
-    m.values[:c] = 456
-    m.values[:b] = 789
-    m.cache_key.should == "#{m.class}:123,789,456"
-  end
-  
-  it "should raise error if attempting to generate cache_key and primary key value is null" do
-    m = @c.new
-    proc {m.cache_key}.should raise_error(Sequel::Error)
-    
-    m.values[:id] = 1
-    proc {m.cache_key}.should_not raise_error(Sequel::Error)
-  end
-  
-  it "should not raise error if trying to save a new record" do
-    proc {@c.new(:name=>'blah').save}.should_not raise_error
-    proc {@c.create(:name=>'blah')}.should_not raise_error
-  end
-  
-  it "should set the cache when reading from the database" do
-    $sqls.should == []
-    @cache.should be_empty
-    
-    m = @c[1]
-    $sqls.should == ['SELECT * FROM items WHERE (id = 1) LIMIT 1']
-    m.values.should == $cache_dataset_row
-    @cache[m.cache_key].should == m
-    
-    # read from cache
-    m2 = @c[1]
-    $sqls.should == ['SELECT * FROM items WHERE (id = 1) LIMIT 1']
-    m2.should == m
-    m2.values.should == $cache_dataset_row
-  end
-  
-  it "should delete the cache when writing to the database" do
-    # fill the cache
-    m = @c[1]
-    @cache[m.cache_key].should == m
-    
-    m.update_values(:name => 'tutu')
-    @cache.has_key?(m.cache_key).should be_false
-    $sqls.last.should == "UPDATE items SET name = 'tutu' WHERE (id = 1)"
-    
-    m = @c[1]
-    @cache[m.cache_key].should == m
-    m.name = 'hey'
-    m.save
-    @cache.has_key?(m.cache_key).should be_false
-    $sqls.last.should == "UPDATE items SET name = 'hey', id = 1 WHERE (id = 1)"
-  end
-  
-  it "should delete the cache when deleting the record" do
-    # fill the cache
-    m = @c[1]
-    @cache[m.cache_key].should == m
-    
-    m.delete
-    @cache.has_key?(m.cache_key).should be_false
-    $sqls.last.should == "DELETE FROM items WHERE (id = 1)"
-  end
-  
-  it "should support #[] as a shortcut to #find with hash" do
-    m = @c[:id => 3]
-    @cache[m.cache_key].should be_nil
-    $sqls.last.should == "SELECT * FROM items WHERE (id = 3) LIMIT 1"
-    
-    m = @c[1]
-    @cache[m.cache_key].should == m
-    $sqls.should == ["SELECT * FROM items WHERE (id = 3) LIMIT 1", \
-      "SELECT * FROM items WHERE (id = 1) LIMIT 1"]
-    
-    @c[:id => 4]
-    $sqls.should == ["SELECT * FROM items WHERE (id = 3) LIMIT 1", \
-      "SELECT * FROM items WHERE (id = 1) LIMIT 1", \
-      "SELECT * FROM items WHERE (id = 4) LIMIT 1"]
-  end
-  
-end
diff -pruN 1.5.1-1/spec/core/connection_pool_spec.rb 3.0.0-1/spec/core/connection_pool_spec.rb
--- 1.5.1-1/spec/core/connection_pool_spec.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/core/connection_pool_spec.rb	2009-05-01 23:06:32.000000000 +0100
@@ -0,0 +1,537 @@
+require File.join(File.dirname(__FILE__), 'spec_helper')
+CONNECTION_POOL_DEFAULTS = {:pool_timeout=>5, :pool_sleep_time=>0.001,
+  :pool_reuse_connections=>:allow, :pool_convert_exceptions=>true, :max_connections=>4}
+
+context "An empty ConnectionPool" do
+  before do
+    @cpool = Sequel::ConnectionPool.new(CONNECTION_POOL_DEFAULTS)
+  end
+
+  specify "should have no available connections" do
+    @cpool.available_connections.should == []
+  end
+
+  specify "should have no allocated connections" do
+    @cpool.allocated.should == {}
+  end
+
+  specify "should have a created_count of zero" do
+    @cpool.created_count.should == 0
+  end
+end
+
+context "A connection pool handling connections" do
+  before do
+    @max_size = 2
+    @cpool = Sequel::ConnectionPool.new(CONNECTION_POOL_DEFAULTS.merge(:disconnection_proc=>proc{|c| @max_size=3},  :max_connections=>@max_size)) {:got_connection}
+  end
+
+  specify "#hold should increment #created_count" do
+    @cpool.hold do
+      @cpool.created_count.should == 1
+      @cpool.hold {@cpool.hold {@cpool.created_count.should == 1}}
+      Thread.new{@cpool.hold {@cpool.created_count.should == 2}}.join
+    end
+  end
+
+  specify "#hold should add the connection to the #allocated array" do
+    @cpool.hold do
+      @cpool.allocated.size.should == 1
+  
+      @cpool.allocated.should == {Thread.current=>:got_connection}
+    end
+  end
+
+  specify "#hold should yield a new connection" do
+    @cpool.hold {|conn| conn.should == :got_connection}
+  end
+
+  specify "a connection should be de-allocated after it has been used in #hold" do
+    @cpool.hold {}
+    @cpool.allocated.size.should == 0
+  end
+
+  specify "#hold should return the value of its block" do
+    @cpool.hold {:block_return}.should == :block_return
+  end
+
+  if RUBY_VERSION < '1.9.0' and (!defined?(RUBY_ENGINE) or RUBY_ENGINE != 'jruby')
+    specify "#hold should remove dead threads from the pool if it reaches its max_size" do
+      Thread.new{@cpool.hold{Thread.current.exit!}}.join
+      @cpool.allocated.keys.map{|t| t.alive?}.should == [false]
+
+      Thread.new{@cpool.hold{Thread.current.exit!}}.join
+      @cpool.allocated.keys.map{|t| t.alive?}.should == [false, false]
+
+      Thread.new{@cpool.hold{}}.join
+      @cpool.allocated.should == {}
+    end
+  end
+
+  specify "#make_new should not make more than max_size connections" do
+    50.times{Thread.new{@cpool.hold{sleep 0.01}}}
+    @cpool.created_count.should == @max_size
+  end
+
+  specify ":disconnection_proc option should set the disconnection proc to use" do
+    @max_size.should == 2
+    proc{@cpool.hold{raise Sequel::DatabaseDisconnectError}}.should raise_error(Sequel::DatabaseDisconnectError)
+    @max_size.should == 3
+  end
+
+  specify "#disconnection_proc= should set the disconnection proc to use" do
+    a = 1
+    @cpool.disconnection_proc = proc{|c| a += 1}
+    proc{@cpool.hold{raise Sequel::DatabaseDisconnectError}}.should raise_error(Sequel::DatabaseDisconnectError)
+    a.should == 2
+  end
+
+  specify "#hold should remove the connection if a DatabaseDisconnectError is raised" do
+    @cpool.created_count.should == 0
+    @cpool.hold{Thread.new{@cpool.hold{}}; sleep 0.01}
+    @cpool.created_count.should == 2
+    proc{@cpool.hold{raise Sequel::DatabaseDisconnectError}}.should raise_error(Sequel::DatabaseDisconnectError)
+    @cpool.created_count.should == 1
+    proc{@cpool.hold{raise Sequel::DatabaseDisconnectError}}.should raise_error(Sequel::DatabaseDisconnectError)
+    @cpool.created_count.should == 0
+    proc{@cpool.hold{raise Sequel::DatabaseDisconnectError}}.should raise_error(Sequel::DatabaseDisconnectError)
+    @cpool.created_count.should == 0
+  end
+end
+
+context "A connection pool handling connection errors" do 
+  specify "#hold should raise a Sequel::DatabaseConnectionError if an exception is raised by the connection_proc" do
+    cpool = Sequel::ConnectionPool.new(CONNECTION_POOL_DEFAULTS){raise Interrupt}
+    proc{cpool.hold{:block_return}}.should raise_error(Sequel::DatabaseConnectionError)
+    cpool.created_count.should == 0
+  end
+
+  specify "#hold should raise a Sequel::DatabaseConnectionError if nil is returned by the connection_proc" do
+    cpool = Sequel::ConnectionPool.new(CONNECTION_POOL_DEFAULTS){nil}
+    proc{cpool.hold{:block_return}}.should raise_error(Sequel::DatabaseConnectionError)
+    cpool.created_count.should == 0
+  end
+end
+
+class DummyConnection
+  @@value = 0
+  def initialize
+    @@value += 1
+  end
+  
+  def value
+    @@value
+  end
+end
+
+context "ConnectionPool#hold" do
+  before do
+    @pool = Sequel::ConnectionPool.new(CONNECTION_POOL_DEFAULTS) {DummyConnection.new}
+  end
+  
+  specify "should pass the result of the connection maker proc to the supplied block" do
+    res = nil
+    @pool.hold {|c| res = c}
+    res.should be_a_kind_of(DummyConnection)
+    res.value.should == 1
+    @pool.hold {|c| res = c}
+    res.should be_a_kind_of(DummyConnection)
+    res.value.should == 1 # the connection maker is invoked only once
+  end
+  
+  specify "should be re-entrant by the same thread" do
+    cc = nil
+    @pool.hold {|c| @pool.hold {|c| @pool.hold {|c| cc = c}}}
+    cc.should be_a_kind_of(DummyConnection)
+  end
+  
+  specify "should catch exceptions and reraise them" do
+    proc {@pool.hold {|c| c.foobar}}.should raise_error(NoMethodError)
+  end
+  
+  specify "should handle Exception errors (normally not caught by rescue)" do
+    err = nil
+    begin
+      @pool.hold {raise Exception}
+    rescue => e
+      err = e
+    end
+    err.should be_a_kind_of(RuntimeError)
+  end
+end
+
+context "ConnectionPool#connection_proc" do
+  before do
+    @pool = Sequel::ConnectionPool.new(CONNECTION_POOL_DEFAULTS)
+  end
+  
+  specify "should be nil if no block is supplied to the pool" do
+    @pool.connection_proc.should be_nil
+    proc {@pool.hold {}}.should raise_error
+  end
+  
+  specify "should be mutable" do
+    @pool.connection_proc = proc {'herro'}
+    res = nil
+    proc {@pool.hold {|c| res = c}}.should_not raise_error
+    res.should == 'herro'
+  end
+end
+
+context "A connection pool with a max size of 1" do
+  before do
+    @invoked_count = 0
+    @pool = Sequel::ConnectionPool.new(CONNECTION_POOL_DEFAULTS.merge(:max_connections=>1)) {@invoked_count += 1; 'herro'}
+  end
+  
+  specify "should let only one thread access the connection at any time" do
+    cc,c1, c2 = nil
+    
+    t1 = Thread.new {@pool.hold {|c| cc = c; c1 = c.dup; while c == 'herro';sleep 0.01;end}}
+    sleep 0.02
+    cc.should == 'herro'
+    c1.should == 'herro'
+    
+    t2 = Thread.new {@pool.hold {|c| c2 = c.dup; while c == 'hello';sleep 0.01;end}}
+    sleep 0.02
+    
+    # connection held by t1
+    t1.should be_alive
+    t2.should be_alive
+    
+    cc.should == 'herro'
+    c1.should == 'herro'
+    c2.should be_nil
+    
+    @pool.available_connections.should be_empty
+    @pool.allocated.should == {t1=>cc}
+    
+    cc.gsub!('rr', 'll')
+    sleep 0.05
+    
+    # connection held by t2
+    t1.should_not be_alive
+    t2.should be_alive
+
+    c2.should == 'hello'
+
+    @pool.available_connections.should be_empty
+    @pool.allocated.should == {t2=>cc}
+    
+    cc.gsub!('ll', 'rr')
+    sleep 0.05
+    
+    #connection released
+    t2.should_not be_alive
+    
+    cc.should == 'herro'
+    
+    @invoked_count.should == 1
+    @pool.size.should == 1
+    @pool.available_connections.should == [cc]
+    @pool.allocated.should be_empty
+  end
+  
+  specify "should let the same thread reenter #hold" do
+    c1, c2, c3 = nil
+    @pool.hold do |c|
+      c1 = c
+      @pool.hold do |c|
+        c2 = c
+        @pool.hold do |c|
+          c3 = c
+        end
+      end
+    end
+    c1.should == 'herro'
+    c2.should == 'herro'
+    c3.should == 'herro'
+    
+    @invoked_count.should == 1
+    @pool.size.should == 1
+    @pool.available_connections.size.should == 1
+    @pool.allocated.should be_empty
+  end
+end
+
+context "A connection pool with a max size of 5" do
+  before do
+    @invoked_count = 0
+    @pool = Sequel::ConnectionPool.new(CONNECTION_POOL_DEFAULTS.merge(:max_connections=>5)) {@invoked_count += 1}
+  end
+  
+  specify "should let five threads simultaneously access separate connections" do
+    cc = {}
+    threads = []
+    stop = nil
+    
+    5.times {|i| threads << Thread.new {@pool.hold {|c| cc[i] = c; while !stop;sleep 0.01;end}}; sleep 0.01}
+    sleep 0.02
+    threads.each {|t| t.should be_alive}
+    cc.size.should == 5
+    @invoked_count.should == 5
+    @pool.size.should == 5
+    @pool.available_connections.should be_empty
+    i = 0
+    h = {}
+    threads.each{|t| h[t] = (i+=1)}
+    @pool.allocated.should == h
+    
+    threads[0].raise "your'e dead"
+    sleep 0.01
+    threads[3].raise "your'e dead too"
+    
+    sleep 0.01
+    
+    @pool.available_connections.should == [1, 4]
+    @pool.allocated.should == {threads[1]=>2, threads[2]=>3, threads[4]=>5}
+    
+    stop = true
+    sleep 0.02
+    
+    @pool.available_connections.size.should == 5
+    @pool.allocated.should be_empty
+  end
+  
+  specify "should block threads until a connection becomes available" do
+    cc = {}
+    threads = []
+    stop = nil
+    
+    5.times {|i| threads << Thread.new {@pool.hold {|c| cc[i] = c; while !stop;sleep 0.01;end}}; sleep 0.01}
+    sleep 0.02
+    threads.each {|t| t.should be_alive}
+    @pool.available_connections.should be_empty
+
+    3.times {|i| threads << Thread.new {@pool.hold {|c| cc[i + 5] = c}}}
+    
+    sleep 0.02
+    threads[5].should be_alive
+    threads[6].should be_alive
+    threads[7].should be_alive
+    cc.size.should == 5
+    cc[5].should be_nil
+    cc[6].should be_nil
+    cc[7].should be_nil
+    
+    stop = true
+    sleep 0.05
+    
+    threads.each {|t| t.should_not be_alive}
+    
+    @pool.size.should == 5
+    @invoked_count.should == 5
+    @pool.available_connections.size.should == 5
+    @pool.allocated.should be_empty
+  end
+end
+
+context "ConnectionPool#disconnect" do
+  before do
+    @count = 0
+    @pool = Sequel::ConnectionPool.new(CONNECTION_POOL_DEFAULTS.merge(:max_connections=>5)) {{:id => @count += 1}}
+  end
+  
+  specify "should invoke the given block for each available connection" do
+    threads = []
+    stop = nil
+    5.times {|i| threads << Thread.new {@pool.hold {|c| while !stop;sleep 0.01;end}}; sleep 0.01}
+    while @pool.size < 5
+      sleep 0.02
+    end
+    stop = true
+    sleep 0.1
+    threads.each {|t| t.join}
+    
+    @pool.size.should == 5
+    @pool.available_connections.size.should == 5
+    @pool.available_connections.each {|c| c[:id].should_not be_nil}
+    conns = []
+    @pool.disconnect {|c| conns << c}
+    conns.size.should == 5
+  end
+  
+  specify "should remove all available connections" do
+    threads = []
+    stop = nil
+    5.times {|i| threads << Thread.new {@pool.hold {|c| while !stop;sleep 0.01;end}}; sleep 0.01}
+    while @pool.size < 5
+      sleep 0.02
+    end
+    stop = true
+    sleep 0.1
+    threads.each {|t| t.join}
+    
+    @pool.size.should == 5
+    @pool.disconnect
+    @pool.size.should == 0
+  end
+
+  specify "should not touch connections in use" do
+    threads = []
+    stop = nil
+    5.times {|i| threads << Thread.new {@pool.hold {|c| while !stop;sleep 0.01;end}}; sleep 0.01}
+    while @pool.size < 5
+      sleep 0.02
+    end
+    stop = true
+    sleep 0.1
+    threads.each {|t| t.join}
+    
+    @pool.size.should == 5
+    
+    @pool.hold do |conn|
+      @pool.available_connections.size.should == 4
+      @pool.available_connections.each {|c| c.should_not be(conn)}
+      conns = []
+      @pool.disconnect {|c| conns << c}
+      conns.size.should == 4
+    end
+    @pool.size.should == 1
+  end
+end
+
+context "A connection pool with multiple servers" do
+  before do
+    @invoked_counts = Hash.new(0)
+    @pool = Sequel::ConnectionPool.new(CONNECTION_POOL_DEFAULTS.merge(:servers=>{:read_only=>{}})){|server| "#{server}#{@invoked_counts[server] += 1}"}
+  end
+  
+  specify "should use the :default server by default" do
+    @pool.size.should == 0
+    @pool.hold do |c|
+      c.should == "default1"
+      @pool.allocated.should == {Thread.current=>"default1"}
+    end
+    @pool.available_connections.should == ["default1"]
+    @pool.size.should == 1
+    @invoked_counts.should == {:default=>1}
+  end
+
+  specify "should use the requested server if server is given" do
+    @pool.size(:read_only).should == 0
+    @pool.hold(:read_only) do |c|
+      c.should == "read_only1"
+      @pool.allocated(:read_only).should == {Thread.current=>"read_only1"}
+    end
+    @pool.available_connections(:read_only).should == ["read_only1"]
+    @pool.size(:read_only).should == 1
+    @invoked_counts.should == {:read_only=>1}
+  end
+  
+  specify "#hold should only yield connections for the server requested" do
+    @pool.hold(:read_only) do |c|
+      c.should == "read_only1"
+      @pool.allocated(:read_only).should == {Thread.current=>"read_only1"}
+      @pool.hold do |d|
+        d.should == "default1"
+        @pool.hold do |e|
+          e.should == d
+          @pool.hold(:read_only){|b| b.should == c}
+        end
+        @pool.allocated.should == {Thread.current=>"default1"}
+      end
+    end
+    @invoked_counts.should == {:read_only=>1, :default=>1}
+  end
+  
+  specify "#disconnect should disconnect from all servers" do
+    @pool.hold(:read_only){}
+    @pool.hold{}
+    conns = []
+    @pool.size.should == 1
+    @pool.size(:read_only).should == 1
+    @pool.disconnect{|c| conns << c}
+    conns.sort.should == %w'default1 read_only1'
+    @pool.size.should == 0
+    @pool.size(:read_only).should == 0
+    @pool.hold(:read_only){|c| c.should == 'read_only2'}
+    @pool.hold{|c| c.should == 'default2'}
+  end
+end
+
+context "SingleThreadedPool" do
+  before do
+    @pool = Sequel::SingleThreadedPool.new(CONNECTION_POOL_DEFAULTS){1234}
+  end
+  
+  specify "should provide a #hold method" do
+    conn = nil
+    @pool.hold {|c| conn = c}
+    conn.should == 1234
+  end
+  
+  specify "should provide a #disconnect method" do
+    @pool.hold {|c|}
+    @pool.conn.should == 1234
+    conn = nil
+    @pool.disconnect {|c| conn = c}
+    conn.should == 1234
+    @pool.conn.should be_nil
+  end
+end
+
+context "A single threaded pool with multiple servers" do
+  before do
+    @max_size=2
+    @pool = Sequel::SingleThreadedPool.new(CONNECTION_POOL_DEFAULTS.merge(:disconnection_proc=>proc{|c| @max_size=3}, :servers=>{:read_only=>{}})){|server| server}
+  end
+  
+  specify "should use the :default server by default" do
+    @pool.hold{|c| c.should == :default}
+    @pool.conn.should == :default
+  end
+
+  specify "should use the requested server if server is given" do
+    @pool.hold(:read_only){|c| c.should == :read_only}
+    @pool.conn(:read_only).should == :read_only
+  end
+  
+  specify "#hold should only yield connections for the server requested" do
+    @pool.hold(:read_only) do |c|
+      c.should == :read_only
+      @pool.hold do |d|
+        d.should == :default
+        @pool.hold do |e|
+          e.should == d
+          @pool.hold(:read_only){|b| b.should == c}
+        end
+      end
+    end
+    @pool.conn.should == :default
+    @pool.conn(:read_only).should == :read_only
+  end
+  
+  specify "#disconnect should disconnect from all servers" do
+    @pool.hold(:read_only){}
+    @pool.hold{}
+    conns = []
+    @pool.conn.should == :default
+    @pool.conn(:read_only).should == :read_only
+    @pool.disconnect{|c| conns << c}
+    conns.sort_by{|x| x.to_s}.should == [:default, :read_only]
+    @pool.conn.should == nil
+    @pool.conn(:read_only).should == nil
+  end
+
+  specify ":disconnection_proc option should set the disconnection proc to use" do
+    @max_size.should == 2
+    proc{@pool.hold{raise Sequel::DatabaseDisconnectError}}.should raise_error(Sequel::DatabaseDisconnectError)
+    @max_size.should == 3
+  end
+
+  specify "#disconnection_proc= should set the disconnection proc to use" do
+    a = 1
+    @pool.disconnection_proc = proc{|c| a += 1}
+    proc{@pool.hold{raise Sequel::DatabaseDisconnectError}}.should raise_error(Sequel::DatabaseDisconnectError)
+    a.should == 2
+  end
+
+  specify "#hold should remove the connection if a DatabaseDisconnectError is raised" do
+    @pool.instance_variable_get(:@conns).length.should == 0
+    @pool.hold{}
+    @pool.instance_variable_get(:@conns).length.should == 1
+    proc{@pool.hold{raise Sequel::DatabaseDisconnectError}}.should raise_error(Sequel::DatabaseDisconnectError)
+    @pool.instance_variable_get(:@conns).length.should == 0
+  end
+end
diff -pruN 1.5.1-1/spec/core/core_sql_spec.rb 3.0.0-1/spec/core/core_sql_spec.rb
--- 1.5.1-1/spec/core/core_sql_spec.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/core/core_sql_spec.rb	2009-04-27 04:27:30.000000000 +0100
@@ -0,0 +1,372 @@
+require File.join(File.dirname(__FILE__), 'spec_helper')
+
+context "Array#all_two_pairs?" do
+  specify "should return false if empty" do
+    [].all_two_pairs?.should == false
+  end
+
+  specify "should return false if any of the elements is not an array" do
+    [1].all_two_pairs?.should == false
+    [[1,2],1].all_two_pairs?.should == false
+  end
+
+  specify "should return false if any of the elements has a length other than two" do
+    [[1,2],[]].all_two_pairs?.should == false
+    [[1,2],[1]].all_two_pairs?.should == false
+    [[1,2],[1,2,3]].all_two_pairs?.should == false
+  end
+
+  specify "should return true if all of the elements are arrays with a length of two" do
+    [[1,2]].all_two_pairs?.should == true
+    [[1,2],[1,2]].all_two_pairs?.should == true
+    [[1,2],[1,2],[1,2]].all_two_pairs?.should == true
+  end
+end
+  
+context "Array#case and Hash#case" do
+  before do
+    @d = Sequel::Dataset.new(nil)
+  end
+
+  specify "should return SQL CASE expression" do
+    @d.literal({:x=>:y}.case(:z)).should == '(CASE WHEN x THEN y ELSE z END)'
+    @d.literal({:x=>:y}.case(:z, :exp)).should == '(CASE exp WHEN x THEN y ELSE z END)'
+    ['(CASE WHEN x THEN y WHEN a THEN b ELSE z END)',
+     '(CASE WHEN a THEN b WHEN x THEN y ELSE z END)'].should(include(@d.literal({:x=>:y, :a=>:b}.case(:z))))
+    @d.literal([[:x, :y]].case(:z)).should == '(CASE WHEN x THEN y ELSE z END)'
+    @d.literal([[:x, :y], [:a, :b]].case(:z)).should == '(CASE WHEN x THEN y WHEN a THEN b ELSE z END)'
+    @d.literal([[:x, :y], [:a, :b]].case(:z, :exp)).should == '(CASE exp WHEN x THEN y WHEN a THEN b ELSE z END)'
+    @d.literal([[:x, :y], [:a, :b]].case(:z, :exp__w)).should == '(CASE exp.w WHEN x THEN y WHEN a THEN b ELSE z END)'
+  end
+
+  specify "should raise an error if an array that isn't all two pairs is used" do
+    proc{[:b].case(:a)}.should raise_error(Sequel::Error)
+    proc{[:b, :c].case(:a)}.should raise_error(Sequel::Error)
+    proc{[[:b, :c], :d].case(:a)}.should raise_error(Sequel::Error)
+  end
+
+  specify "should raise an error if an empty array/hash is used" do
+    proc{[].case(:a)}.should raise_error(Sequel::Error)
+    proc{{}.case(:a)}.should raise_error(Sequel::Error)
+  end
+end
+
+context "Array#sql_array" do
+  before do
+    @d = Sequel::Dataset.new(nil)
+  end
+
+  specify "should treat the array as an SQL array instead of conditions" do
+    @d.literal([[:x, 1], [:y, 2]]).should == '((x = 1) AND (y = 2))'
+    @d.literal([[:x, 1], [:y, 2]].sql_array).should == '((x, 1), (y, 2))'
+    @d.filter([:a, :b]=>[[:x, 1], [:y, 2]].sql_array).sql.should == 'SELECT * WHERE ((a, b) IN ((x, 1), (y, 2)))'
+  end
+end
+
+context "String#lit" do
+  specify "should return an LiteralString object" do
+    'xyz'.lit.should be_a_kind_of(Sequel::LiteralString)
+    'xyz'.lit.to_s.should == 'xyz'
+  end
+  
+  specify "should inhibit string literalization" do
+    Sequel::Database.new[:t].update_sql(:stamp => "NOW()".lit).should == \
+      "UPDATE t SET stamp = NOW()"
+  end
+
+  specify "should return a PlaceholderLiteralString object if args are given" do
+    a = 'DISTINCT ?'.lit(:a)
+    a.should be_a_kind_of(Sequel::SQL::PlaceholderLiteralString)
+    ds = MockDatabase.new.dataset
+    ds.literal(a).should == 'DISTINCT a'
+    ds.quote_identifiers = true
+    ds.literal(a).should == 'DISTINCT "a"'
+  end
+end
+
+context "String#to_sequel_blob" do
+  specify "should return a Blob object" do
+    'xyz'.to_sequel_blob.should be_a_kind_of(::Sequel::SQL::Blob)
+    'xyz'.to_sequel_blob.should == 'xyz'
+  end
+
+  specify "should retain binary data" do
+    "\1\2\3\4".to_sequel_blob.should == "\1\2\3\4"
+  end
+end
+
+context "#desc" do
+  before do
+    @ds = Sequel::Dataset.new(nil)
+  end
+  
+  specify "should format a DESC clause for a column ref" do
+    :test.desc.to_s(@ds).should == 'test DESC'
+    
+    :items__price.desc.to_s(@ds).should == 'items.price DESC'
+  end
+
+  specify "should format a DESC clause for a function" do
+    :avg.sql_function(:test).desc.to_s(@ds).should == 'avg(test) DESC'
+  end
+end
+
+context "#asc" do
+  before do
+    @ds = Sequel::Dataset.new(nil)
+  end
+  
+  specify "should format a ASC clause for a column ref" do
+    :test.asc.to_s(@ds).should == 'test ASC'
+    
+    :items__price.asc.to_s(@ds).should == 'items.price ASC'
+  end
+
+  specify "should format a ASC clause for a function" do
+    :avg.sql_function(:test).asc.to_s(@ds).should == 'avg(test) ASC'
+  end
+end
+
+context "#as" do
+  before do
+    @ds = Sequel::Dataset.new(nil)
+  end
+  
+  specify "should format a AS clause for a column ref" do
+    :test.as(:t).to_s(@ds).should == 'test AS t'
+    
+    :items__price.as(:p).to_s(@ds).should == 'items.price AS p'
+  end
+
+  specify "should format a AS clause for a function" do
+    :avg.sql_function(:test).as(:avg).to_s(@ds).should == 'avg(test) AS avg'
+  end
+  
+  specify "should format a AS clause for a literal value" do
+    'abc'.as(:abc).to_s(@ds).should == "'abc' AS abc"
+  end
+end
+
+context "Column references" do
+  before do
+    @c = Class.new(Sequel::Dataset) do
+      def quoted_identifier(c); "`#{c}`"; end
+    end
+    @ds = @c.new(MockDatabase.new)
+    @ds.quote_identifiers = true
+  end
+  
+  specify "should be quoted properly" do
+    @ds.literal(:xyz).should == "`xyz`"
+    @ds.literal(:xyz__abc).should == "`xyz`.`abc`"
+
+    @ds.literal(:xyz.as(:x)).should == "`xyz` AS `x`"
+    @ds.literal(:xyz__abc.as(:x)).should == "`xyz`.`abc` AS `x`"
+
+    @ds.literal(:xyz___x).should == "`xyz` AS `x`"
+    @ds.literal(:xyz__abc___x).should == "`xyz`.`abc` AS `x`"
+  end
+  
+  specify "should be quoted properly in SQL functions" do
+    @ds.literal(:avg.sql_function(:xyz)).should == "avg(`xyz`)"
+    @ds.literal(:avg.sql_function(:xyz, 1)).should == "avg(`xyz`, 1)"
+    @ds.literal(:avg.sql_function(:xyz).as(:a)).should == "avg(`xyz`) AS `a`"
+  end
+
+  specify "should be quoted properly in ASC/DESC clauses" do
+    @ds.literal(:xyz.asc).should == "`xyz` ASC"
+    @ds.literal(:avg.sql_function(:xyz, 1).desc).should == "avg(`xyz`, 1) DESC"
+  end
+  
+  specify "should be quoted properly in a cast function" do
+    @ds.literal(:x.cast(:integer)).should == "CAST(`x` AS integer)"
+    @ds.literal(:x__y.cast('varchar(20)')).should == "CAST(`x`.`y` AS varchar(20))"
+  end
+end
+
+context "Blob" do
+  specify "#to_sequel_blob should return self" do
+    blob = "x".to_sequel_blob
+    blob.to_sequel_blob.object_id.should == blob.object_id
+  end
+end
+
+if RUBY_VERSION < '1.9.0'
+  context "Symbol#[]" do
+    specify "should format an SQL Function" do
+      ds = Sequel::Dataset.new(nil)
+      ds.literal(:xyz[]).should == 'xyz()'
+      ds.literal(:xyz[1]).should == 'xyz(1)'
+      ds.literal(:xyz[1, 2, :abc[3]]).should == 'xyz(1, 2, abc(3))'
+    end
+  end
+end
+
+context "Symbol#*" do
+  before do
+    @ds = Sequel::Dataset.new(nil)
+  end
+  
+  specify "should format a qualified wildcard if no argument" do
+    :xyz.*.to_s(@ds).should == 'xyz.*'
+    :abc.*.to_s(@ds).should == 'abc.*'
+  end
+
+  specify "should format a filter expression if an argument" do
+    :xyz.*(3).to_s(@ds).should == '(xyz * 3)'
+    :abc.*(5).to_s(@ds).should == '(abc * 5)'
+  end
+
+  specify "should support qualified symbols if no argument" do
+    :xyz__abc.*.to_s(@ds).should == 'xyz.abc.*'
+  end
+end
+
+context "Symbol" do
+  before do
+    @ds = Sequel::Dataset.new(nil)
+    @ds.quote_identifiers = true
+    @ds.identifier_input_method = :upcase
+  end
+
+  specify "#identifier should format an identifier" do
+    @ds.literal(:xyz__abc.identifier).should == '"XYZ__ABC"'
+  end
+
+  specify "#qualify should format a qualified column" do
+    @ds.literal(:xyz.qualify(:abc)).should == '"ABC"."XYZ"'
+  end
+
+  specify "#qualify should work on QualifiedIdentifiers" do
+    @ds.literal(:xyz.qualify(:abc).qualify(:def)).should == '"DEF"."ABC"."XYZ"'
+  end
+
+  specify "should be able to qualify an identifier" do
+    @ds.literal(:xyz.identifier.qualify(:xyz__abc)).should == '"XYZ"."ABC"."XYZ"'
+  end
+
+  specify "should be able to specify a schema.table.column" do
+    @ds.literal(:column.qualify(:table.qualify(:schema))).should == '"SCHEMA"."TABLE"."COLUMN"'
+    @ds.literal(:column.qualify(:table__name.identifier.qualify(:schema))).should == '"SCHEMA"."TABLE__NAME"."COLUMN"'
+  end
+end
+
+context "Dataset#literal" do
+  before do
+    @ds = MockDataset.new(nil)
+  end
+  
+  specify "should convert qualified symbol notation into dot notation" do
+    @ds.literal(:abc__def).should == 'abc.def'
+  end
+  
+  specify "should convert AS symbol notation into SQL AS notation" do
+    @ds.literal(:xyz___x).should == 'xyz AS x'
+    @ds.literal(:abc__def___x).should == 'abc.def AS x'
+  end
+  
+  specify "should support names with digits" do
+    @ds.literal(:abc2).should == 'abc2'
+    @ds.literal(:xx__yy3).should == 'xx.yy3'
+    @ds.literal(:ab34__temp3_4ax).should == 'ab34.temp3_4ax'
+    @ds.literal(:x1___y2).should == 'x1 AS y2'
+    @ds.literal(:abc2__def3___ggg4).should == 'abc2.def3 AS ggg4'
+  end
+  
+  specify "should support upper case and lower case" do
+    @ds.literal(:ABC).should == 'ABC'
+    @ds.literal(:Zvashtoy__aBcD).should == 'Zvashtoy.aBcD'
+  end
+
+  specify "should support spaces inside column names" do
+    @ds.quote_identifiers = true
+    @ds.literal(:"AB C").should == '"AB C"'
+    @ds.literal(:"Zvas htoy__aB cD").should == '"Zvas htoy"."aB cD"'
+    @ds.literal(:"aB cD___XX XX").should == '"aB cD" AS "XX XX"'
+    @ds.literal(:"Zva shtoy__aB cD___XX XX").should == '"Zva shtoy"."aB cD" AS "XX XX"'
+  end
+end
+
+context "Symbol" do
+  before do
+    @ds = Sequel::Dataset.new(MockDatabase.new)
+  end
+  
+  specify "should support upper case outer functions" do
+    :COUNT.sql_function('1').to_s(@ds).should == "COUNT('1')"
+  end
+  
+  specify "should inhibit string literalization" do
+    db = Sequel::Database.new
+    ds = db[:t]
+    ds.select(:COUNT.sql_function('1')).sql.should == "SELECT COUNT('1') FROM t"
+  end
+  
+  specify "should support cast method" do
+    :abc.cast(:integer).to_s(@ds).should == "CAST(abc AS integer)"
+  end
+
+  specify "should support sql array accesses via sql_subscript" do
+    @ds.literal(:abc.sql_subscript(1)).should == "abc[1]"
+    @ds.literal(:abc__def.sql_subscript(1)).should == "abc.def[1]"
+    @ds.literal(:abc.sql_subscript(1)|2).should == "abc[1, 2]"
+  end
+
+  specify "should support cast_numeric and cast_string" do
+    x = :abc.cast_numeric
+    x.should be_a_kind_of(Sequel::SQL::NumericExpression)
+    x.to_s(@ds).should == "CAST(abc AS integer)"
+
+    x = :abc.cast_numeric(:real)
+    x.should be_a_kind_of(Sequel::SQL::NumericExpression)
+    x.to_s(@ds).should == "CAST(abc AS real)"
+
+    x = :abc.cast_string
+    x.should be_a_kind_of(Sequel::SQL::StringExpression)
+    x.to_s(@ds).should == "CAST(abc AS varchar(255))"
+
+    x = :abc.cast_string(:varchar)
+    x.should be_a_kind_of(Sequel::SQL::StringExpression)
+    x.to_s(@ds).should == "CAST(abc AS varchar(255))"
+  end
+  
+  specify "should allow database independent types when casting" do
+    m = MockDatabase.new
+    m.instance_eval do
+       def cast_type_literal(type)
+         return :foo if type == Integer
+         return :bar if type == String
+         type
+       end
+    end
+    @ds2 = Sequel::Dataset.new(m)
+    :abc.cast(String).to_s(@ds).should == "CAST(abc AS varchar(255))"
+    :abc.cast(String).to_s(@ds2).should == "CAST(abc AS bar)"
+    :abc.cast(String).to_s(@ds2).should == "CAST(abc AS bar)"
+    :abc.cast_string.to_s(@ds2).should == "CAST(abc AS bar)"
+    :abc.cast_string(Integer).to_s(@ds2).should == "CAST(abc AS foo)"
+    :abc.cast_numeric.to_s(@ds2).should == "CAST(abc AS foo)"
+    :abc.cast_numeric(String).to_s(@ds2).should == "CAST(abc AS bar)"
+  end
+
+  specify "should support SQL EXTRACT function via #extract " do
+    :abc.extract(:year).to_s(@ds).should == "extract(year FROM abc)"
+  end
+end
+
+context "Sequel::SQL::Function#==" do
+  specify "should be true for functions with the same name and arguments, false otherwise" do
+    a = :date.sql_function(:t)
+    b = :date.sql_function(:t)
+    a.should == b
+    (a == b).should == true
+    c = :date.sql_function(:c)
+    a.should_not == c
+    (a == c).should == false
+    d = :time.sql_function(:c)
+    a.should_not == d
+    c.should_not == d
+    (a == d).should == false
+    (c == d).should == false
+  end
+end
diff -pruN 1.5.1-1/spec/core/database_spec.rb 3.0.0-1/spec/core/database_spec.rb
--- 1.5.1-1/spec/core/database_spec.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/core/database_spec.rb	2009-04-28 00:06:14.000000000 +0100
@@ -0,0 +1,1142 @@
+require File.join(File.dirname(__FILE__), 'spec_helper')
+
+context "A new Database" do
+  before do
+    @db = Sequel::Database.new(1 => 2, :logger => 3)
+  end
+  after do
+    Sequel.quote_identifiers = false
+    Sequel.identifier_input_method = nil
+    Sequel.identifier_output_method = nil
+  end
+  
+  specify "should receive options" do
+    @db.opts.should == {1 => 2, :logger => 3}  
+  end
+  
+  specify "should set the logger from opts[:logger] and opts[:loggers]" do
+    @db.loggers.should == [3]
+    Sequel::Database.new(1 => 2, :loggers => 3).loggers.should == [3]
+    Sequel::Database.new(1 => 2, :loggers => [3]).loggers.should == [3]
+    Sequel::Database.new(1 => 2, :logger => 4, :loggers => 3).loggers.should == [4,3]
+    Sequel::Database.new(1 => 2, :logger => [4], :loggers => [3]).loggers.should == [4,3]
+  end
+  
+  specify "should create a connection pool" do
+    @db.pool.should be_a_kind_of(Sequel::ConnectionPool)
+    @db.pool.max_size.should == 4
+    
+    Sequel::Database.new(:max_connections => 10).pool.max_size.should == 10
+  end
+  
+  specify "should pass the supplied block to the connection pool" do
+    cc = nil
+    d = Sequel::Database.new {1234}
+    d.synchronize {|c| cc = c}
+    cc.should == 1234
+  end
+
+  specify "should respect the :single_threaded option" do
+    db = Sequel::Database.new(:single_threaded=>true)
+    db.pool.should be_a_kind_of(Sequel::SingleThreadedPool)
+    db = Sequel::Database.new(:single_threaded=>false)
+    db.pool.should be_a_kind_of(Sequel::ConnectionPool)
+  end
+
+  specify "should respect the :quote_identifiers option" do
+    db = Sequel::Database.new(:quote_identifiers=>false)
+    db.quote_identifiers?.should == false
+    db = Sequel::Database.new(:quote_identifiers=>true)
+    db.quote_identifiers?.should == true
+  end
+
+  specify "should upcase on input and downcase on output by default" do
+    db = Sequel::Database.new
+    db.send(:identifier_input_method_default).should == :upcase
+    db.send(:identifier_output_method_default).should == :downcase
+  end
+
+  specify "should respect the :identifier_input_method option" do
+    Sequel.identifier_input_method = nil
+    Sequel::Database.identifier_input_method.should == ""
+    db = Sequel::Database.new(:identifier_input_method=>nil)
+    db.identifier_input_method.should == nil
+    db.identifier_input_method = :downcase
+    db.identifier_input_method.should == :downcase
+    db = Sequel::Database.new(:identifier_input_method=>:upcase)
+    db.identifier_input_method.should == :upcase
+    db.identifier_input_method = nil
+    db.identifier_input_method.should == nil
+    Sequel.identifier_input_method = :downcase
+    Sequel::Database.identifier_input_method.should == :downcase
+    db = Sequel::Database.new(:identifier_input_method=>nil)
+    db.identifier_input_method.should == nil
+    db.identifier_input_method = :upcase
+    db.identifier_input_method.should == :upcase
+    db = Sequel::Database.new(:identifier_input_method=>:upcase)
+    db.identifier_input_method.should == :upcase
+    db.identifier_input_method = nil
+    db.identifier_input_method.should == nil
+  end
+  
+  specify "should respect the :identifier_output_method option" do
+    Sequel.identifier_output_method = nil
+    Sequel::Database.identifier_output_method.should == ""
+    db = Sequel::Database.new(:identifier_output_method=>nil)
+    db.identifier_output_method.should == nil
+    db.identifier_output_method = :downcase
+    db.identifier_output_method.should == :downcase
+    db = Sequel::Database.new(:identifier_output_method=>:upcase)
+    db.identifier_output_method.should == :upcase
+    db.identifier_output_method = nil
+    db.identifier_output_method.should == nil
+    Sequel.identifier_output_method = :downcase
+    Sequel::Database.identifier_output_method.should == :downcase
+    db = Sequel::Database.new(:identifier_output_method=>nil)
+    db.identifier_output_method.should == nil
+    db.identifier_output_method = :upcase
+    db.identifier_output_method.should == :upcase
+    db = Sequel::Database.new(:identifier_output_method=>:upcase)
+    db.identifier_output_method.should == :upcase
+    db.identifier_output_method = nil
+    db.identifier_output_method.should == nil
+  end
+
+  specify "should use the default Sequel.quote_identifiers value" do
+    Sequel.quote_identifiers = true
+    Sequel::Database.new({}).quote_identifiers?.should == true
+    Sequel.quote_identifiers = false
+    Sequel::Database.new({}).quote_identifiers?.should == false
+    Sequel::Database.quote_identifiers = true
+    Sequel::Database.new({}).quote_identifiers?.should == true
+    Sequel::Database.quote_identifiers = false
+    Sequel::Database.new({}).quote_identifiers?.should == false
+  end
+
+  specify "should use the default Sequel.identifier_input_method value" do
+    Sequel.identifier_input_method = :downcase
+    Sequel::Database.new({}).identifier_input_method.should == :downcase
+    Sequel.identifier_input_method = :upcase
+    Sequel::Database.new({}).identifier_input_method.should == :upcase
+    Sequel::Database.identifier_input_method = :downcase
+    Sequel::Database.new({}).identifier_input_method.should == :downcase
+    Sequel::Database.identifier_input_method = :upcase
+    Sequel::Database.new({}).identifier_input_method.should == :upcase
+  end
+  
+  specify "should use the default Sequel.identifier_output_method value" do
+    Sequel.identifier_output_method = :downcase
+    Sequel::Database.new({}).identifier_output_method.should == :downcase
+    Sequel.identifier_output_method = :upcase
+    Sequel::Database.new({}).identifier_output_method.should == :upcase
+    Sequel::Database.identifier_output_method = :downcase
+    Sequel::Database.new({}).identifier_output_method.should == :downcase
+    Sequel::Database.identifier_output_method = :upcase
+    Sequel::Database.new({}).identifier_output_method.should == :upcase
+  end
+
+  specify "should respect the quote_indentifiers_default method if Sequel.quote_identifiers = nil" do
+    Sequel.quote_identifiers = nil
+    Sequel::Database.new({}).quote_identifiers?.should == true
+    x = Class.new(Sequel::Database){def quote_identifiers_default; false end}
+    x.new({}).quote_identifiers?.should == false
+    y = Class.new(Sequel::Database){def quote_identifiers_default; true end}
+    y.new({}).quote_identifiers?.should == true
+  end
+  
+  specify "should respect the identifier_input_method_default method" do
+    class Sequel::Database
+      @@identifier_input_method = nil
+    end
+    x = Class.new(Sequel::Database){def identifier_input_method_default; :downcase end}
+    x.new({}).identifier_input_method.should == :downcase
+    y = Class.new(Sequel::Database){def identifier_input_method_default; :camelize end}
+    y.new({}).identifier_input_method.should == :camelize
+  end
+  
+  specify "should respect the identifier_output_method_default method if Sequel.identifier_output_method is not called" do
+    class Sequel::Database
+      @@identifier_output_method = nil
+    end
+    x = Class.new(Sequel::Database){def identifier_output_method_default; :upcase end}
+    x.new({}).identifier_output_method.should == :upcase
+    y = Class.new(Sequel::Database){def identifier_output_method_default; :underscore end}
+    y.new({}).identifier_output_method.should == :underscore
+  end
+
+  specify "should just use a :uri option for jdbc with the full connection string" do
+    Sequel::Database.should_receive(:adapter_class).once.with(:jdbc).and_return(Sequel::Database)
+    db = Sequel.connect('jdbc:test://host/db_name')
+    db.should be_a_kind_of(Sequel::Database)
+    db.opts[:uri].should == 'jdbc:test://host/db_name'
+  end
+
+  specify "should just use a :uri option for do with the full connection string" do
+    Sequel::Database.should_receive(:adapter_class).once.with(:do).and_return(Sequel::Database)
+    db = Sequel.connect('do:test://host/db_name')
+    db.should be_a_kind_of(Sequel::Database)
+    db.opts[:uri].should == 'do:test://host/db_name'
+  end
+end
+
+context "Database#disconnect" do
+  specify "should call pool.disconnect" do
+    d = Sequel::Database.new
+    p = d.pool
+    p.should_receive(:disconnect).once.and_return(2)
+    d.disconnect.should == 2
+  end
+end
+
+context "Database#connect" do
+  specify "should raise NotImplementedError" do
+    proc {Sequel::Database.new.connect}.should raise_error(NotImplementedError)
+  end
+end
+
+context "Database#uri" do
+  before do
+    @c = Class.new(Sequel::Database) do
+      set_adapter_scheme :mau
+    end
+    
+    @db = Sequel.connect('mau://user:pass@localhost:9876/maumau')
+  end
+  
+  specify "should return the connection URI for the database" do
+    @db.uri.should == 'mau://user:pass@localhost:9876/maumau'
+  end
+  
+  specify "should be aliased as #url" do
+    @db.url.should == 'mau://user:pass@localhost:9876/maumau'
+  end
+end
+
+context "Database.adapter_scheme" do
+  specify "should return the database schema" do
+    Sequel::Database.adapter_scheme.should be_nil
+
+    @c = Class.new(Sequel::Database) do
+      set_adapter_scheme :mau
+    end
+    
+    @c.adapter_scheme.should == :mau
+  end
+end
+
+context "Database#dataset" do
+  before do
+    @db = Sequel::Database.new
+    @ds = @db.dataset
+  end
+  
+  specify "should provide a blank dataset through #dataset" do
+    @ds.should be_a_kind_of(Sequel::Dataset)
+    @ds.opts.should == {}
+    @ds.db.should be(@db)
+  end
+  
+  specify "should provide a #from dataset" do
+    d = @db.from(:mau)
+    d.should be_a_kind_of(Sequel::Dataset)
+    d.sql.should == 'SELECT * FROM mau'
+    
+    e = @db[:miu]
+    e.should be_a_kind_of(Sequel::Dataset)
+    e.sql.should == 'SELECT * FROM miu'
+  end
+  
+  specify "should provide a filtered #from dataset if a block is given" do
+    d = @db.from(:mau) {:x.sql_number > 100}
+    d.should be_a_kind_of(Sequel::Dataset)
+    d.sql.should == 'SELECT * FROM mau WHERE (x > 100)'
+  end
+  
+  specify "should provide a #select dataset" do
+    d = @db.select(:a, :b, :c).from(:mau)
+    d.should be_a_kind_of(Sequel::Dataset)
+    d.sql.should == 'SELECT a, b, c FROM mau'
+  end
+  
+  specify "should allow #select to take a block" do
+    d = @db.select(:a, :b){c}.from(:mau)
+    d.should be_a_kind_of(Sequel::Dataset)
+    d.sql.should == 'SELECT a, b, c FROM mau'
+  end
+end
+
+context "Database#execute" do
+  specify "should raise NotImplementedError" do
+    proc {Sequel::Database.new.execute('blah blah')}.should raise_error(NotImplementedError)
+    proc {Sequel::Database.new << 'blah blah'}.should raise_error(NotImplementedError)
+  end
+end
+
+context "Database#<<" do
+  before do
+    @c = Class.new(Sequel::Database) do
+      define_method(:execute) {|sql, opts| sql}
+    end
+    @db = @c.new({})
+  end
+  
+  specify "should pass the supplied sql to #execute" do
+    (@db << "DELETE FROM items").should == "DELETE FROM items"
+  end
+  
+  specify "should not remove comments and whitespace from strings" do
+    s = "INSERT INTO items VALUES ('---abc')"
+    (@db << s).should == s
+  end
+end
+
+context "Database#synchronize" do
+  before do
+    @db = Sequel::Database.new(:max_connections => 1)
+    @db.pool.connection_proc = proc {12345}
+  end
+  
+  specify "should wrap the supplied block in pool.hold" do
+    stop = false
+    c1, c2 = nil
+    t1 = Thread.new {@db.synchronize {|c| c1 = c; while !stop;sleep 0.1;end}}
+    while !c1;end
+    c1.should == 12345
+    t2 = Thread.new {@db.synchronize {|c| c2 = c}}
+    sleep 0.2
+    @db.pool.available_connections.should be_empty
+    c2.should be_nil
+    stop = true
+    t1.join
+    sleep 0.1
+    c2.should == 12345
+    t2.join
+  end
+end
+
+context "Database#test_connection" do
+  before do
+    @db = Sequel::Database.new
+    @test = nil
+    @db.pool.connection_proc = proc {@test = rand(100)}
+  end
+  
+  specify "should call pool#hold" do
+    @db.test_connection
+    @test.should_not be_nil
+  end
+  
+  specify "should return true if successful" do
+    @db.test_connection.should be_true
+  end
+end
+
+class DummyDataset < Sequel::Dataset
+  def first
+    raise if @opts[:from] == [:a]
+    true
+  end
+end
+
+class DummyDatabase < Sequel::Database
+  attr_reader :sqls
+  
+  def execute(sql, opts={})
+    @sqls ||= []
+    @sqls << sql
+  end
+  
+  def transaction; yield; end
+
+  def dataset
+    DummyDataset.new(self)
+  end
+end
+
+context "Database#create_table" do
+  before do
+    @db = DummyDatabase.new
+  end
+  
+  specify "should construct proper SQL" do
+    @db.create_table :test do
+      primary_key :id, :integer, :null => false
+      column :name, :text
+      index :name, :unique => true
+    end
+    @db.sqls.should == [
+      'CREATE TABLE test (id integer NOT NULL PRIMARY KEY AUTOINCREMENT, name text)',
+      'CREATE UNIQUE INDEX test_name_index ON test (name)'
+    ]
+  end
+  
+  specify "should create a temporary table" do
+    @db.create_table :test_tmp, :temp => true do
+      primary_key :id, :integer, :null => false
+      column :name, :text
+      index :name, :unique => true
+    end
+    
+    @db.sqls.should == [
+      'CREATE TEMPORARY TABLE test_tmp (id integer NOT NULL PRIMARY KEY AUTOINCREMENT, name text)',
+      'CREATE UNIQUE INDEX test_tmp_name_index ON test_tmp (name)'
+    ]
+  end
+end
+
+context "Database#alter_table" do
+  before do
+    @db = DummyDatabase.new
+  end
+  
+  specify "should construct proper SQL" do
+    @db.alter_table :xyz do
+      add_column :aaa, :text, :null => false, :unique => true
+      drop_column :bbb
+      rename_column :ccc, :ddd
+      set_column_type :eee, :integer
+      set_column_default :hhh, 'abcd'
+      
+      add_index :fff, :unique => true
+      drop_index :ggg
+    end
+    
+    @db.sqls.should == [
+      'ALTER TABLE xyz ADD COLUMN aaa text UNIQUE NOT NULL',
+      'ALTER TABLE xyz DROP COLUMN bbb',
+      'ALTER TABLE xyz RENAME COLUMN ccc TO ddd',
+      'ALTER TABLE xyz ALTER COLUMN eee TYPE integer',
+      "ALTER TABLE xyz ALTER COLUMN hhh SET DEFAULT 'abcd'",
+      
+      'CREATE UNIQUE INDEX xyz_fff_index ON xyz (fff)',
+      'DROP INDEX xyz_ggg_index'
+    ]
+  end
+end
+
+context "Database#add_column" do
+  before do
+    @db = DummyDatabase.new
+  end
+  
+  specify "should construct proper SQL" do
+    @db.add_column :test, :name, :text, :unique => true
+    @db.sqls.should == [
+      'ALTER TABLE test ADD COLUMN name text UNIQUE'
+    ]
+  end
+end
+
+context "Database#drop_column" do
+  before do
+    @db = DummyDatabase.new
+  end
+  
+  specify "should construct proper SQL" do
+    @db.drop_column :test, :name
+    @db.sqls.should == [
+      'ALTER TABLE test DROP COLUMN name'
+    ]
+  end
+end
+
+context "Database#rename_column" do
+  before do
+    @db = DummyDatabase.new
+  end
+  
+  specify "should construct proper SQL" do
+    @db.rename_column :test, :abc, :def
+    @db.sqls.should == [
+      'ALTER TABLE test RENAME COLUMN abc TO def'
+    ]
+  end
+end
+
+context "Database#set_column_type" do
+  before do
+    @db = DummyDatabase.new
+  end
+  
+  specify "should construct proper SQL" do
+    @db.set_column_type :test, :name, :integer
+    @db.sqls.should == [
+      'ALTER TABLE test ALTER COLUMN name TYPE integer'
+    ]
+  end
+end
+
+context "Database#set_column_default" do
+  before do
+    @db = DummyDatabase.new
+  end
+  
+  specify "should construct proper SQL" do
+    @db.set_column_default :test, :name, 'zyx'
+    @db.sqls.should == [
+      "ALTER TABLE test ALTER COLUMN name SET DEFAULT 'zyx'"
+    ]
+  end
+end
+
+context "Database#add_index" do
+  before do
+    @db = DummyDatabase.new
+  end
+  
+  specify "should construct proper SQL" do
+    @db.add_index :test, :name, :unique => true
+    @db.sqls.should == [
+      'CREATE UNIQUE INDEX test_name_index ON test (name)'
+    ]
+  end
+  
+  specify "should accept multiple columns" do
+    @db.add_index :test, [:one, :two]
+    @db.sqls.should == [
+      'CREATE INDEX test_one_two_index ON test (one, two)'
+    ]
+  end
+end
+
+context "Database#drop_index" do
+  before do
+    @db = DummyDatabase.new
+  end
+  
+  specify "should construct proper SQL" do
+    @db.drop_index :test, :name
+    @db.sqls.should == [
+      'DROP INDEX test_name_index'
+    ]
+  end
+  
+end
+
+class Dummy2Database < Sequel::Database
+  attr_reader :sql
+  def execute(sql); @sql = sql; end
+  def transaction; yield; end
+end
+
+context "Database#drop_table" do
+  before do
+    @db = DummyDatabase.new
+  end
+  
+  specify "should construct proper SQL" do
+    @db.drop_table :test
+    @db.sqls.should == ['DROP TABLE test']
+  end
+  
+  specify "should accept multiple table names" do
+    @db.drop_table :a, :bb, :ccc
+    @db.sqls.should == [
+      'DROP TABLE a',
+      'DROP TABLE bb',
+      'DROP TABLE ccc'
+    ]
+  end
+end
+
+context "Database#rename_table" do
+  before do
+    @db = DummyDatabase.new
+  end
+  
+  specify "should construct proper SQL" do
+    @db.rename_table :abc, :xyz
+    @db.sqls.should == ['ALTER TABLE abc RENAME TO xyz']
+  end
+end
+
+context "Database#table_exists?" do
+  before do
+    @db = DummyDatabase.new
+    @db.instance_variable_set(:@schemas, {:a=>[]})
+    @db2 = DummyDatabase.new
+  end
+  
+  specify "should use schema information if available" do
+    @db.table_exists?(:a).should be_true
+  end
+  
+  specify "should otherwise try to select the first record from the table's dataset" do
+    @db2.table_exists?(:a).should be_false
+    @db2.table_exists?(:b).should be_true
+  end
+end
+
+class Dummy3Database < Sequel::Database
+  attr_reader :sql, :transactions
+  def execute(sql, opts={}); @sql ||= []; @sql << sql; end
+
+  class DummyConnection
+    def initialize(db); @db = db; end
+    def execute(sql); @db.execute(sql); end
+  end
+end
+
+context "Database#transaction" do
+  before do
+    @db = Dummy3Database.new
+    @db.pool.connection_proc = proc {Dummy3Database::DummyConnection.new(@db)}
+  end
+  
+  specify "should wrap the supplied block with BEGIN + COMMIT statements" do
+    @db.transaction {@db.execute 'DROP TABLE test;'}
+    @db.sql.should == ['BEGIN', 'DROP TABLE test;', 'COMMIT']
+  end
+  
+  specify "should handle returning inside of the block by committing" do
+    def @db.ret_commit
+      transaction do
+        execute 'DROP TABLE test;'
+        return
+        execute 'DROP TABLE test2;';
+      end
+    end
+    @db.ret_commit
+    @db.sql.should == ['BEGIN', 'DROP TABLE test;', 'COMMIT']
+  end
+  
+  specify "should issue ROLLBACK if an exception is raised, and re-raise" do
+    @db.transaction {@db.execute 'DROP TABLE test'; raise RuntimeError} rescue nil
+    @db.sql.should == ['BEGIN', 'DROP TABLE test', 'ROLLBACK']
+    
+    proc {@db.transaction {raise RuntimeError}}.should raise_error(RuntimeError)
+  end
+  
+  specify "should issue ROLLBACK if Sequel::Rollback is called in the transaction" do
+    @db.transaction do
+      @db.drop_table(:a)
+      raise Sequel::Rollback
+      @db.drop_table(:b)
+    end
+    
+    @db.sql.should == ['BEGIN', 'DROP TABLE a', 'ROLLBACK']
+  end
+  
+  specify "should be re-entrant" do
+    stop = false
+    cc = nil
+    t = Thread.new do
+      @db.transaction {@db.transaction {@db.transaction {|c|
+        cc = c
+        while !stop; sleep 0.1; end
+      }}}
+    end
+    while cc.nil?; sleep 0.1; end
+    cc.should be_a_kind_of(Dummy3Database::DummyConnection)
+    @db.transactions.should == [t]
+    stop = true
+    t.join
+    @db.transactions.should be_empty
+  end
+end
+
+context "A Database adapter with a scheme" do
+  before do
+    class CCC < Sequel::Database
+      if defined?(DISCONNECTS)
+        DISCONNECTS.clear
+      else
+        DISCONNECTS = []
+      end
+      set_adapter_scheme :ccc
+      def disconnect
+        DISCONNECTS << self
+      end
+    end
+  end
+
+  specify "should be registered in the ADAPTER_MAP" do
+    Sequel::ADAPTER_MAP[:ccc].should == CCC
+  end
+  
+  specify "should give the database_type as the adapter scheme by default" do
+    CCC.new.database_type.should == :ccc
+  end
+  
+  specify "should be instantiated when its scheme is specified" do
+    c = Sequel::Database.connect('ccc://localhost/db')
+    c.should be_a_kind_of(CCC)
+    c.opts[:host].should == 'localhost'
+    c.opts[:database].should == 'db'
+  end
+  
+  specify "should be accessible through Sequel.connect" do
+    c = Sequel.connect 'ccc://localhost/db'
+    c.should be_a_kind_of(CCC)
+    c.opts[:host].should == 'localhost'
+    c.opts[:database].should == 'db'
+  end
+
+  specify "should be accessible through Sequel.connect via a block" do
+    x = nil
+    y = nil
+    z = nil
+
+    p = proc do |c|
+      c.should be_a_kind_of(CCC)
+      c.opts[:host].should == 'localhost'
+      c.opts[:database].should == 'db'
+      z = y
+      y = x
+      x = c
+    end
+    Sequel::Database.connect('ccc://localhost/db', &p).should == nil
+    CCC::DISCONNECTS.should == [x]
+
+    Sequel.connect('ccc://localhost/db', &p).should == nil
+    CCC::DISCONNECTS.should == [y, x]
+
+    Sequel.send(:def_adapter_method, :ccc)
+    Sequel.ccc('db', :host=>'localhost', &p).should == nil
+    CCC::DISCONNECTS.should == [z, y, x]
+  end
+
+  specify "should be accessible through Sequel.<adapter>" do
+    Sequel.send(:def_adapter_method, :ccc)
+
+    # invalid parameters
+    proc {Sequel.ccc('abc', 'def')}.should raise_error(Sequel::Error)
+    
+    c = Sequel.ccc('mydb')
+    c.should be_a_kind_of(CCC)
+    c.opts.should == {:adapter=>:ccc, :database => 'mydb'}
+    
+    c = Sequel.ccc('mydb', :host => 'localhost')
+    c.should be_a_kind_of(CCC)
+    c.opts.should == {:adapter=>:ccc, :database => 'mydb', :host => 'localhost'}
+    
+    c = Sequel.ccc
+    c.should be_a_kind_of(CCC)
+    c.opts.should == {:adapter=>:ccc}
+    
+    c = Sequel.ccc(:database => 'mydb', :host => 'localhost')
+    c.should be_a_kind_of(CCC)
+    c.opts.should == {:adapter=>:ccc, :database => 'mydb', :host => 'localhost'}
+  end
+  
+  specify "should be accessible through Sequel.connect with options" do
+    c = Sequel.connect(:adapter => :ccc, :database => 'mydb')
+    c.should be_a_kind_of(CCC)
+    c.opts.should == {:adapter => :ccc, :database => 'mydb'}
+  end
+
+  specify "should be accessible through Sequel.connect with URL parameters" do
+    c = Sequel.connect 'ccc://localhost/db?host=/tmp&user=test'
+    c.should be_a_kind_of(CCC)
+    c.opts[:host].should == '/tmp'
+    c.opts[:database].should == 'db'
+    c.opts[:user].should == 'test'
+  end
+
+end
+
+context "An unknown database scheme" do
+  specify "should raise an error in Sequel::Database.connect" do
+    proc {Sequel::Database.connect('ddd://localhost/db')}.should raise_error(Sequel::AdapterNotFound)
+  end
+
+  specify "should raise an error in Sequel.connect" do
+    proc {Sequel.connect('ddd://localhost/db')}.should raise_error(Sequel::AdapterNotFound)
+  end
+end
+
+context "A broken adapter (lib is there but the class is not)" do
+  before do
+    @fn = File.join(File.dirname(__FILE__), '../../lib/sequel/adapters/blah.rb')
+    File.open(@fn,'a'){}
+  end
+  
+  after do
+    File.delete(@fn)
+  end
+  
+  specify "should raise an error" do
+    proc {Sequel.connect('blah://blow')}.should raise_error(Sequel::AdapterNotFound)
+  end
+end
+
+context "A single threaded database" do
+  after do
+    Sequel::Database.single_threaded = false
+  end
+  
+  specify "should use a SingleThreadedPool instead of a ConnectionPool" do
+    db = Sequel::Database.new(:single_threaded => true)
+    db.pool.should be_a_kind_of(Sequel::SingleThreadedPool)
+  end
+  
+  specify "should be constructable using :single_threaded => true option" do
+    db = Sequel::Database.new(:single_threaded => true)
+    db.pool.should be_a_kind_of(Sequel::SingleThreadedPool)
+  end
+  
+  specify "should be constructable using Database.single_threaded = true" do
+    Sequel::Database.single_threaded = true
+    db = Sequel::Database.new
+    db.pool.should be_a_kind_of(Sequel::SingleThreadedPool)
+  end
+
+  specify "should be constructable using Sequel.single_threaded = true" do
+    Sequel.single_threaded = true
+    db = Sequel::Database.new
+    db.pool.should be_a_kind_of(Sequel::SingleThreadedPool)
+  end
+end
+
+context "A single threaded database" do
+  before do
+    conn = 1234567
+    @db = Sequel::Database.new(:single_threaded => true) do
+      conn += 1
+    end
+  end
+  
+  specify "should invoke connection_proc only once" do
+    @db.pool.hold {|c| c.should == 1234568}
+    @db.pool.hold {|c| c.should == 1234568}
+  end
+  
+  specify "should convert an Exception into a RuntimeError" do
+    db = Sequel::Database.new(:single_threaded => true) do
+      raise Exception
+    end
+    
+    proc {db.pool.hold {|c|}}.should raise_error(RuntimeError)
+  end
+end
+
+context "A database" do
+  before do
+    Sequel::Database.single_threaded = false
+  end
+  
+  after do
+    Sequel::Database.single_threaded = false
+  end
+  
+  specify "should have single_threaded? respond to true if in single threaded mode" do
+    db = Sequel::Database.new(:single_threaded => true)
+    db.should be_single_threaded
+    
+    db = Sequel::Database.new(:max_options => 1)
+    db.should_not be_single_threaded
+    
+    db = Sequel::Database.new
+    db.should_not be_single_threaded
+    
+    Sequel::Database.single_threaded = true
+    
+    db = Sequel::Database.new
+    db.should be_single_threaded
+    
+    db = Sequel::Database.new(:max_options => 4)
+    db.should be_single_threaded
+  end
+  
+  specify "should be able to set loggers via the logger= and loggers= methods" do
+    db = Sequel::Database.new
+    s = "I'm a logger"
+    db.logger = s
+    db.loggers.should == [s]
+    db.logger = nil
+    db.loggers.should == []
+
+    db.loggers = [s]
+    db.loggers.should == [s]
+    db.loggers = []
+    db.loggers.should == []
+
+    t = "I'm also a logger"
+    db.loggers = [s, t]
+    db.loggers.should == [s,t]
+  end
+end
+
+context "Database#fetch" do
+  before do
+    @db = Sequel::Database.new
+    c = Class.new(Sequel::Dataset) do
+      def fetch_rows(sql); yield({:sql => sql}); end
+    end
+    @db.meta_def(:dataset) {c.new(self)}
+  end
+  
+  specify "should create a dataset and invoke its fetch_rows method with the given sql" do
+    sql = nil
+    @db.fetch('select * from xyz') {|r| sql = r[:sql]}
+    sql.should == 'select * from xyz'
+  end
+  
+  specify "should format the given sql with any additional arguments" do
+    sql = nil
+    @db.fetch('select * from xyz where x = ? and y = ?', 15, 'abc') {|r| sql = r[:sql]}
+    sql.should == "select * from xyz where x = 15 and y = 'abc'"
+    
+    # and Aman Gupta's example
+    @db.fetch('select name from table where name = ? or id in ?',
+    'aman', [3,4,7]) {|r| sql = r[:sql]}
+    sql.should == "select name from table where name = 'aman' or id in (3, 4, 7)"
+  end
+  
+  specify "should return the dataset if no block is given" do
+    @db.fetch('select * from xyz').should be_a_kind_of(Sequel::Dataset)
+    
+    @db.fetch('select a from b').map {|r| r[:sql]}.should == ['select a from b']
+
+    @db.fetch('select c from d').inject([]) {|m, r| m << r; m}.should == \
+      [{:sql => 'select c from d'}]
+  end
+  
+  specify "should return a dataset that always uses the given sql for SELECTs" do
+    ds = @db.fetch('select * from xyz')
+    ds.select_sql.should == 'select * from xyz'
+    ds.sql.should == 'select * from xyz'
+    
+    ds.filter!(:price.sql_number < 100)
+    ds.select_sql.should == 'select * from xyz'
+    ds.sql.should == 'select * from xyz'
+  end
+end
+
+
+context "Database#[]" do
+  before do
+    @db = Sequel::Database.new
+  end
+  
+  specify "should return a dataset when symbols are given" do
+    ds = @db[:items]
+    ds.class.should == Sequel::Dataset
+    ds.opts[:from].should == [:items]
+  end
+  
+  specify "should return an enumerator when a string is given" do
+    c = Class.new(Sequel::Dataset) do
+      def fetch_rows(sql); yield({:sql => sql}); end
+    end
+    @db.meta_def(:dataset) {c.new(self)}
+
+    sql = nil
+    @db['select * from xyz where x = ? and y = ?', 15, 'abc'].each {|r| sql = r[:sql]}
+    sql.should == "select * from xyz where x = 15 and y = 'abc'"
+  end
+end
+
+context "Database#create_view" do
+  before do
+    @db = DummyDatabase.new
+  end
+  
+  specify "should construct proper SQL with raw SQL" do
+    @db.create_view :test, "SELECT * FROM xyz"
+    @db.sqls.should == ['CREATE VIEW test AS SELECT * FROM xyz']
+    @db.sqls.clear
+    @db.create_view :test.identifier, "SELECT * FROM xyz"
+    @db.sqls.should == ['CREATE VIEW test AS SELECT * FROM xyz']
+  end
+  
+  specify "should construct proper SQL with dataset" do
+    @db.create_view :test, @db[:items].select(:a, :b).order(:c)
+    @db.sqls.should == ['CREATE VIEW test AS SELECT a, b FROM items ORDER BY c']
+    @db.sqls.clear
+    @db.create_view :test.qualify(:sch), @db[:items].select(:a, :b).order(:c)
+    @db.sqls.should == ['CREATE VIEW sch.test AS SELECT a, b FROM items ORDER BY c']
+  end
+end
+
+context "Database#create_or_replace_view" do
+  before do
+    @db = DummyDatabase.new
+  end
+  
+  specify "should construct proper SQL with raw SQL" do
+    @db.create_or_replace_view :test, "SELECT * FROM xyz"
+    @db.sqls.should == ['CREATE OR REPLACE VIEW test AS SELECT * FROM xyz']
+    @db.sqls.clear
+    @db.create_or_replace_view :sch__test, "SELECT * FROM xyz"
+    @db.sqls.should == ['CREATE OR REPLACE VIEW sch.test AS SELECT * FROM xyz']
+  end
+
+  specify "should construct proper SQL with dataset" do
+    @db.create_or_replace_view :test, @db[:items].select(:a, :b).order(:c)
+    @db.sqls.should == ['CREATE OR REPLACE VIEW test AS SELECT a, b FROM items ORDER BY c']
+    @db.sqls.clear
+    @db.create_or_replace_view :test.identifier, @db[:items].select(:a, :b).order(:c)
+    @db.sqls.should == ['CREATE OR REPLACE VIEW test AS SELECT a, b FROM items ORDER BY c']
+  end
+end
+
+context "Database#drop_view" do
+  before do
+    @db = DummyDatabase.new
+  end
+  
+  specify "should construct proper SQL" do
+    @db.drop_view :test
+    @db.drop_view :test.identifier
+    @db.drop_view :sch__test
+    @db.drop_view :test.qualify(:sch)
+    @db.sqls.should == ['DROP VIEW test', 'DROP VIEW test', 'DROP VIEW sch.test', 'DROP VIEW sch.test']
+  end
+end
+
+context "Database#alter_table_sql" do
+  before do
+    @db = DummyDatabase.new
+  end
+  
+  specify "should raise error for an invalid op" do
+    proc {@db.send(:alter_table_sql, :mau, :op => :blah)}.should raise_error(Sequel::Error)
+  end
+end
+
+context "Database#inspect" do
+  before do
+    @db = DummyDatabase.new
+    
+    @db.meta_def(:uri) {'blah://blahblah/blah'}
+  end
+  
+  specify "should include the class name and the connection url" do
+    @db.inspect.should == '#<DummyDatabase: "blah://blahblah/blah">'
+  end
+end
+
+context "Database#get" do
+  before do
+    @c = Class.new(DummyDatabase) do
+      def dataset
+        ds = super
+        def ds.get(*args, &block)
+          @db.execute select(*args, &block).sql
+          args
+        end
+        ds
+      end
+    end
+    
+    @db = @c.new
+  end
+  
+  specify "should use Dataset#get to get a single value" do
+    @db.get(1).should == [1]
+    @db.sqls.last.should == 'SELECT 1'
+    
+    @db.get(:version.sql_function)
+    @db.sqls.last.should == 'SELECT version()'
+  end
+
+  specify "should accept a block" do
+    @db.get{1}
+    @db.sqls.last.should == 'SELECT 1'
+    
+    @db.get{version(1)}
+    @db.sqls.last.should == 'SELECT version(1)'
+  end
+end
+
+context "Database#call" do
+  specify "should call the prepared statement with the given name" do
+    db = MockDatabase.new
+    db[:items].prepare(:select, :select_all)
+    db.call(:select_all).should == [{:id => 1, :x => 1}]
+    db[:items].filter(:n=>:$n).prepare(:select, :select_n)
+    db.call(:select_n, :n=>1).should == [{:id => 1, :x => 1}]
+    db.sqls.should == ['SELECT * FROM items', 'SELECT * FROM items WHERE (n = 1)']
+  end
+end
+
+context "Database#server_opts" do
+  specify "should return the general opts if no :servers option is used" do
+    opts = {:host=>1, :database=>2}
+    MockDatabase.new(opts).send(:server_opts, :server1).should == {:host=>1, :database=>2}
+  end
+  
+  specify "should return the general opts if entry for the server is present in the :servers option" do
+    opts = {:host=>1, :database=>2, :servers=>{}}
+    MockDatabase.new(opts).send(:server_opts, :server1).should == {:host=>1, :database=>2}
+  end
+  
+  specify "should return the general opts merged with the specific opts if given as a hash" do
+    opts = {:host=>1, :database=>2, :servers=>{:server1=>{:host=>3}}}
+    MockDatabase.new(opts).send(:server_opts, :server1).should == {:host=>3, :database=>2}
+  end
+  
+  specify "should return the sgeneral opts merged with the specific opts if given as a proc" do
+    opts = {:host=>1, :database=>2, :servers=>{:server1=>proc{|db| {:host=>4}}}}
+    MockDatabase.new(opts).send(:server_opts, :server1).should == {:host=>4, :database=>2}
+  end
+  
+  specify "should raise an error if the specific opts is not a proc or hash" do
+    opts = {:host=>1, :database=>2, :servers=>{:server1=>2}}
+    proc{MockDatabase.new(opts).send(:server_opts, :server1)}.should raise_error(Sequel::Error)
+  end
+end
+
+context "Database#raise_error" do
+  specify "should reraise if the exception class is not in opts[:classes]" do
+    e = Class.new(StandardError)
+    proc{MockDatabase.new.send(:raise_error, e.new(''), :classes=>[])}.should raise_error(e)
+  end
+
+  specify "should convert the exception to a DatabaseError if the exception class is not in opts[:classes]" do
+    proc{MockDatabase.new.send(:raise_error, Interrupt.new(''), :classes=>[Interrupt])}.should raise_error(Sequel::DatabaseError)
+  end
+
+  specify "should convert the exception to a DatabaseError if opts[:classes] if not present" do
+    proc{MockDatabase.new.send(:raise_error, Interrupt.new(''))}.should raise_error(Sequel::DatabaseError)
+  end
+end
+
+context "Database#typecast_value" do
+  before do
+    @db = Sequel::Database.new
+  end
+  specify "should raise an InvalidValue when given an invalid value" do
+    proc{@db.typecast_value(:integer, "13a")}.should raise_error(Sequel::InvalidValue)
+    proc{@db.typecast_value(:float, "4.e2")}.should raise_error(Sequel::InvalidValue)
+    proc{@db.typecast_value(:decimal, :invalid_value)}.should raise_error(Sequel::InvalidValue)
+    proc{@db.typecast_value(:date, Object.new)}.should raise_error(Sequel::InvalidValue)
+    proc{@db.typecast_value(:date, 'a')}.should raise_error(Sequel::InvalidValue)
+    proc{@db.typecast_value(:time, Date.new)}.should raise_error(Sequel::InvalidValue)
+    proc{@db.typecast_value(:datetime, 4)}.should raise_error(Sequel::InvalidValue)
+  end
+end
+
+context "Database#blank_object?" do
+  specify "should return whether the object is considered blank" do
+    db = Sequel::Database.new
+    c = lambda{|meth, value| Class.new{define_method(meth){value}}.new}
+
+    db.send(:blank_object?, "").should == true
+    db.send(:blank_object?, "  ").should == true
+    db.send(:blank_object?, nil).should == true
+    db.send(:blank_object?, false).should == true
+    db.send(:blank_object?, []).should == true
+    db.send(:blank_object?, {}).should == true
+    db.send(:blank_object?, c[:empty?, true]).should == true
+    db.send(:blank_object?, c[:blank?, true]).should == true
+
+    db.send(:blank_object?, " a ").should == false
+    db.send(:blank_object?, 1).should == false
+    db.send(:blank_object?, 1.0).should == false
+    db.send(:blank_object?, true).should == false
+    db.send(:blank_object?, [1]).should == false
+    db.send(:blank_object?, {1.0=>2.0}).should == false
+    db.send(:blank_object?, c[:empty?, false]).should == false
+    db.send(:blank_object?, c[:blank?, false]).should == false 
+  end
+end
+
+context "Database#schema_autoincrementing_primary_key?" do
+  specify "should whether the parsed schema row indicates a primary key" do
+    m = Sequel::Database.new.method(:schema_autoincrementing_primary_key?)
+    m.call(:primary_key=>true).should == true
+    m.call(:primary_key=>false).should == false
+  end
+end
diff -pruN 1.5.1-1/spec/core/dataset_spec.rb 3.0.0-1/spec/core/dataset_spec.rb
--- 1.5.1-1/spec/core/dataset_spec.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/core/dataset_spec.rb	2009-04-27 02:05:37.000000000 +0100
@@ -0,0 +1,2750 @@
+require File.join(File.dirname(__FILE__), "spec_helper")
+
+context "Dataset" do
+  before do
+    @dataset = Sequel::Dataset.new("db")
+  end
+  
+  specify "should accept database and opts in initialize" do
+    db = "db"
+    opts = {:from => :test}
+    d = Sequel::Dataset.new(db, opts)
+    d.db.should be(db)
+    d.opts.should be(opts)
+    
+    d = Sequel::Dataset.new(db)
+    d.db.should be(db)
+    d.opts.should be_a_kind_of(Hash)
+    d.opts.should == {}
+  end
+  
+  specify "should provide clone for chainability" do
+    d1 = @dataset.clone(:from => [:test])
+    d1.class.should == @dataset.class
+    d1.should_not == @dataset
+    d1.db.should be(@dataset.db)
+    d1.opts[:from].should == [:test]
+    @dataset.opts[:from].should be_nil
+    
+    d2 = d1.clone(:order => [:name])
+    d2.class.should == @dataset.class
+    d2.should_not == d1
+    d2.should_not == @dataset
+    d2.db.should be(@dataset.db)
+    d2.opts[:from].should == [:test]
+    d2.opts[:order].should == [:name]
+    d1.opts[:order].should be_nil
+  end
+  
+  specify "should include Enumerable" do
+    Sequel::Dataset.included_modules.should include(Enumerable)
+  end
+  
+  specify "should get quote_identifiers default from database" do
+    db = Sequel::Database.new(:quote_identifiers=>true)
+    db[:a].quote_identifiers?.should == true
+    db = Sequel::Database.new(:quote_identifiers=>false)
+    db[:a].quote_identifiers?.should == false
+  end
+
+  specify "should get identifier_input_method default from database" do
+    db = Sequel::Database.new(:identifier_input_method=>:upcase)
+    db[:a].identifier_input_method.should == :upcase
+    db = Sequel::Database.new(:identifier_input_method=>:downcase)
+    db[:a].identifier_input_method.should == :downcase
+  end
+
+  specify "should get identifier_output_method default from database" do
+    db = Sequel::Database.new(:identifier_output_method=>:upcase)
+    db[:a].identifier_output_method.should == :upcase
+    db = Sequel::Database.new(:identifier_output_method=>:downcase)
+    db[:a].identifier_output_method.should == :downcase
+  end
+end
+
+context "Dataset" do
+  before do
+    @dataset = Sequel::Dataset.new("db")
+  end
+  
+  specify "should have quote_identifiers= method which changes literalization of identifiers" do
+    @dataset.quote_identifiers = true
+    @dataset.literal(:a).should == '"a"'
+    @dataset.quote_identifiers = false
+    @dataset.literal(:a).should == 'a'
+  end
+  
+  specify "should have identifier_input_method= method which changes literalization of identifiers" do
+    @dataset.identifier_input_method = :upcase
+    @dataset.literal(:a).should == 'A'
+    @dataset.identifier_input_method = :downcase
+    @dataset.literal(:A).should == 'a'
+    @dataset.identifier_input_method = :reverse
+    @dataset.literal(:at_b).should == 'b_ta'
+  end
+  
+  specify "should have identifier_output_method= method which changes identifiers returned from the database" do
+    @dataset.send(:output_identifier, "at_b_C").should == :at_b_C
+    @dataset.identifier_output_method = :upcase
+    @dataset.send(:output_identifier, "at_b_C").should == :AT_B_C
+    @dataset.identifier_output_method = :downcase
+    @dataset.send(:output_identifier, "at_b_C").should == :at_b_c
+    @dataset.identifier_output_method = :reverse
+    @dataset.send(:output_identifier, "at_b_C").should == :C_b_ta
+  end
+end
+
+context "Dataset#clone" do
+  before do
+    @dataset = Sequel::Dataset.new(nil).from(:items)
+  end
+  
+  specify "should create an exact copy of the dataset" do
+    @dataset.row_proc = Proc.new{|r| r}
+    @clone = @dataset.clone
+
+    @clone.should_not === @dataset
+    @clone.class.should == @dataset.class
+    @clone.opts.should == @dataset.opts
+    @clone.row_proc.should == @dataset.row_proc
+  end
+  
+  specify "should deep-copy the dataset opts" do
+    @clone = @dataset.clone
+
+    @clone.opts.should_not equal(@dataset.opts)
+    @dataset.filter!(:a => 'b')
+    @clone.opts[:filter].should be_nil
+  end
+  
+  specify "should return a clone self" do
+    clone = @dataset.clone({})
+    clone.class.should == @dataset.class
+    clone.db.should == @dataset.db
+    clone.opts.should == @dataset.opts
+  end
+  
+  specify "should merge the specified options" do
+    clone = @dataset.clone(1 => 2)
+    clone.opts.should == {1 => 2, :from => [:items]}
+  end
+  
+  specify "should overwrite existing options" do
+    clone = @dataset.clone(:from => [:other])
+    clone.opts.should == {:from => [:other]}
+  end
+  
+  specify "should create a clone with a deep copy of options" do
+    clone = @dataset.clone(:from => [:other])
+    @dataset.opts[:from].should == [:items]
+    clone.opts[:from].should == [:other]
+  end
+  
+  specify "should return an object with the same modules included" do
+    m = Module.new do
+      def __xyz__; "xyz"; end
+    end
+    @dataset.extend(m)
+    @dataset.clone({}).should respond_to(:__xyz__)
+  end
+end
+
+context "A simple dataset" do
+  before do
+    @dataset = Sequel::Dataset.new(nil).from(:test)
+  end
+  
+  specify "should format a select statement" do
+    @dataset.select_sql.should == 'SELECT * FROM test'
+  end
+  
+  specify "should format a delete statement" do
+    @dataset.delete_sql.should == 'DELETE FROM test'
+  end
+  
+  specify "should format an insert statement with default values" do
+    @dataset.insert_sql.should == 'INSERT INTO test DEFAULT VALUES'
+  end
+  
+  specify "should format an insert statement with hash" do
+    @dataset.insert_sql(:name => 'wxyz', :price => 342).
+      should match(/INSERT INTO test \(name, price\) VALUES \('wxyz', 342\)|INSERT INTO test \(price, name\) VALUES \(342, 'wxyz'\)/)
+
+      @dataset.insert_sql({}).should == "INSERT INTO test DEFAULT VALUES"
+  end
+
+  specify "should format an insert statement with string keys" do
+    @dataset.insert_sql('name' => 'wxyz', 'price' => 342).
+      should match(/INSERT INTO test \(name, price\) VALUES \('wxyz', 342\)|INSERT INTO test \(price, name\) VALUES \(342, 'wxyz'\)/)
+  end
+  
+  specify "should format an insert statement with an object that respond_to? :values" do
+    dbb = Sequel::Database.new
+    
+    v = Object.new
+    def v.values; {:a => 1}; end
+    
+    @dataset.insert_sql(v).should == "INSERT INTO test (a) VALUES (1)"
+    
+    def v.values; {}; end
+    @dataset.insert_sql(v).should == "INSERT INTO test DEFAULT VALUES"
+  end
+  
+  specify "should format an insert statement with an arbitrary value" do
+    @dataset.insert_sql(123).should == "INSERT INTO test VALUES (123)"
+  end
+  
+  specify "should format an insert statement with sub-query" do
+    @sub = Sequel::Dataset.new(nil).from(:something).filter(:x => 2)
+    @dataset.insert_sql(@sub).should == \
+      "INSERT INTO test (SELECT * FROM something WHERE (x = 2))"
+  end
+  
+  specify "should format an insert statement with array" do
+    @dataset.insert_sql('a', 2, 6.5).should ==
+      "INSERT INTO test VALUES ('a', 2, 6.5)"
+  end
+  
+  specify "should format an update statement" do
+    @dataset.update_sql(:name => 'abc').should ==
+      "UPDATE test SET name = 'abc'"
+  end
+
+  specify "should be able to return rows for arbitrary SQL" do
+    @dataset.clone(:sql => 'xxx yyy zzz').select_sql.should ==
+      "xxx yyy zzz"
+  end
+
+  specify "should use the :sql option for all sql methods" do
+    sql = "X"
+    ds = Sequel::Dataset.new(nil, :sql=>sql)
+    ds.sql.should == sql
+    ds.select_sql.should == sql
+    ds.insert_sql.should == sql
+    ds.delete_sql.should == sql
+    ds.update_sql.should == sql
+  end
+end
+
+context "A dataset with multiple tables in its FROM clause" do
+  before do
+    @dataset = Sequel::Dataset.new(nil).from(:t1, :t2)
+  end
+
+  specify "should raise on #update_sql" do
+    proc {@dataset.update_sql(:a=>1)}.should raise_error(Sequel::InvalidOperation)
+  end
+
+  specify "should raise on #delete_sql" do
+    proc {@dataset.delete_sql}.should raise_error(Sequel::InvalidOperation)
+  end
+
+  specify "should generate a select query FROM all specified tables" do
+    @dataset.select_sql.should == "SELECT * FROM t1, t2"
+  end
+end
+
+context "Dataset#exists" do
+  before do
+    @ds1 = Sequel::Dataset.new(nil).from(:test)
+    @ds2 = @ds1.filter(:price.sql_number < 100)
+    @ds3 = @ds1.filter(:price.sql_number > 50)
+  end
+  
+  specify "should work in filters" do
+    @ds1.filter(@ds2.exists).sql.should ==
+      'SELECT * FROM test WHERE (EXISTS (SELECT * FROM test WHERE (price < 100)))'
+    @ds1.filter(@ds2.exists & @ds3.exists).sql.should ==
+      'SELECT * FROM test WHERE (EXISTS (SELECT * FROM test WHERE (price < 100)) AND EXISTS (SELECT * FROM test WHERE (price > 50)))'
+  end
+
+  specify "should work in select" do
+    @ds1.select(@ds2.exists.as(:a), @ds3.exists.as(:b)).sql.should ==
+      'SELECT EXISTS (SELECT * FROM test WHERE (price < 100)) AS a, EXISTS (SELECT * FROM test WHERE (price > 50)) AS b FROM test'
+  end
+end
+
+context "Dataset#where" do
+  before do
+    @dataset = Sequel::Dataset.new(nil).from(:test)
+    @d1 = @dataset.where(:region => 'Asia')
+    @d2 = @dataset.where('region = ?', 'Asia')
+    @d3 = @dataset.where("a = 1")
+  end
+  
+  specify "should work with hashes" do
+    @dataset.where(:name => 'xyz', :price => 342).select_sql.
+      should match(/WHERE \(\(name = 'xyz'\) AND \(price = 342\)\)|WHERE \(\(price = 342\) AND \(name = 'xyz'\)\)/)
+  end
+  
+  specify "should work with arrays (ala ActiveRecord)" do
+    @dataset.where('price < ? AND id in ?', 100, [1, 2, 3]).select_sql.should ==
+      "SELECT * FROM test WHERE (price < 100 AND id in (1, 2, 3))"
+  end
+  
+  specify "should work with strings (custom SQL expressions)" do
+    @dataset.where('(a = 1 AND b = 2)').select_sql.should ==
+      "SELECT * FROM test WHERE ((a = 1 AND b = 2))"
+  end
+  
+  specify "should affect select, delete and update statements" do
+    @d1.select_sql.should == "SELECT * FROM test WHERE (region = 'Asia')"
+    @d1.delete_sql.should == "DELETE FROM test WHERE (region = 'Asia')"
+    @d1.update_sql(:GDP => 0).should == "UPDATE test SET GDP = 0 WHERE (region = 'Asia')"
+    
+    @d2.select_sql.should == "SELECT * FROM test WHERE (region = 'Asia')"
+    @d2.delete_sql.should == "DELETE FROM test WHERE (region = 'Asia')"
+    @d2.update_sql(:GDP => 0).should == "UPDATE test SET GDP = 0 WHERE (region = 'Asia')"
+    
+    @d3.select_sql.should == "SELECT * FROM test WHERE (a = 1)"
+    @d3.delete_sql.should == "DELETE FROM test WHERE (a = 1)"
+    @d3.update_sql(:GDP => 0).should == "UPDATE test SET GDP = 0 WHERE (a = 1)"
+    
+  end
+  
+  specify "should be composable using AND operator (for scoping)" do
+    # hashes are merged, no problem
+    @d1.where(:size => 'big').select_sql.should == 
+      "SELECT * FROM test WHERE ((region = 'Asia') AND (size = 'big'))"
+    
+    # hash and string
+    @d1.where('population > 1000').select_sql.should ==
+      "SELECT * FROM test WHERE ((region = 'Asia') AND (population > 1000))"
+    @d1.where('(a > 1) OR (b < 2)').select_sql.should ==
+    "SELECT * FROM test WHERE ((region = 'Asia') AND ((a > 1) OR (b < 2)))"
+    
+    # hash and array
+    @d1.where('GDP > ?', 1000).select_sql.should == 
+      "SELECT * FROM test WHERE ((region = 'Asia') AND (GDP > 1000))"
+    
+    # array and array
+    @d2.where('GDP > ?', 1000).select_sql.should ==
+      "SELECT * FROM test WHERE ((region = 'Asia') AND (GDP > 1000))"
+    
+    # array and hash
+    @d2.where(:name => ['Japan', 'China']).select_sql.should ==
+      "SELECT * FROM test WHERE ((region = 'Asia') AND (name IN ('Japan', 'China')))"
+      
+    # array and string
+    @d2.where('GDP > ?').select_sql.should ==
+      "SELECT * FROM test WHERE ((region = 'Asia') AND (GDP > ?))"
+    
+    # string and string
+    @d3.where('b = 2').select_sql.should ==
+      "SELECT * FROM test WHERE ((a = 1) AND (b = 2))"
+    
+    # string and hash
+    @d3.where(:c => 3).select_sql.should == 
+      "SELECT * FROM test WHERE ((a = 1) AND (c = 3))"
+      
+    # string and array
+    @d3.where('d = ?', 4).select_sql.should ==
+      "SELECT * FROM test WHERE ((a = 1) AND (d = 4))"
+  end
+      
+  specify "should be composable using AND operator (for scoping) with block" do
+    @d3.where{:e.sql_number < 5}.select_sql.should ==
+      "SELECT * FROM test WHERE ((a = 1) AND (e < 5))"
+  end
+  
+  specify "should raise if the dataset is grouped" do
+    proc {@dataset.group(:t).where(:a => 1)}.should_not raise_error
+    @dataset.group(:t).where(:a => 1).sql.should ==
+      "SELECT * FROM test WHERE (a = 1) GROUP BY t"
+  end
+  
+  specify "should accept ranges" do
+    @dataset.filter(:id => 4..7).sql.should ==
+      'SELECT * FROM test WHERE ((id >= 4) AND (id <= 7))'
+    @dataset.filter(:id => 4...7).sql.should ==
+      'SELECT * FROM test WHERE ((id >= 4) AND (id < 7))'
+
+    @dataset.filter(:table__id => 4..7).sql.should ==
+      'SELECT * FROM test WHERE ((table.id >= 4) AND (table.id <= 7))'
+    @dataset.filter(:table__id => 4...7).sql.should ==
+      'SELECT * FROM test WHERE ((table.id >= 4) AND (table.id < 7))'
+  end
+
+  specify "should accept nil" do
+    @dataset.filter(:owner_id => nil).sql.should ==
+      'SELECT * FROM test WHERE (owner_id IS NULL)'
+  end
+
+  specify "should accept a subquery" do
+    @dataset.filter('gdp > ?', @d1.select(:avg.sql_function(:gdp))).sql.should ==
+      "SELECT * FROM test WHERE (gdp > (SELECT avg(gdp) FROM test WHERE (region = 'Asia')))"
+
+    @dataset.filter(:id => @d1.select(:id)).sql.should ==
+      "SELECT * FROM test WHERE (id IN (SELECT id FROM test WHERE (region = 'Asia')))"
+  end
+  
+  specify "should accept a subquery for an EXISTS clause" do
+    a = @dataset.filter(:price.sql_number < 100)
+    @dataset.filter(a.exists).sql.should ==
+      'SELECT * FROM test WHERE (EXISTS (SELECT * FROM test WHERE (price < 100)))'
+  end
+  
+  specify "should accept proc expressions" do
+    d = @d1.select(:avg.sql_function(:gdp))
+    @dataset.filter {:gdp.sql_number > d}.sql.should ==
+      "SELECT * FROM test WHERE (gdp > (SELECT avg(gdp) FROM test WHERE (region = 'Asia')))"
+    
+    @dataset.filter {:a.sql_number < 1}.sql.should ==
+      'SELECT * FROM test WHERE (a < 1)'
+
+    @dataset.filter {(:a.sql_number >= 1) & (:b.sql_number <= 2)}.sql.should ==
+      'SELECT * FROM test WHERE ((a >= 1) AND (b <= 2))'
+      
+    @dataset.filter {:c.like 'ABC%'}.sql.should ==
+      "SELECT * FROM test WHERE (c LIKE 'ABC%')"
+
+    @dataset.filter {:c.like 'ABC%'}.sql.should ==
+      "SELECT * FROM test WHERE (c LIKE 'ABC%')"
+
+    @dataset.filter {:c.like 'ABC%', '%XYZ'}.sql.should ==
+      "SELECT * FROM test WHERE ((c LIKE 'ABC%') OR (c LIKE '%XYZ'))"
+  end
+  
+  specify "should work for grouped datasets" do
+    @dataset.group(:a).filter(:b => 1).sql.should ==
+      'SELECT * FROM test WHERE (b = 1) GROUP BY a'
+  end
+
+  specify "should accept true and false as arguments" do
+    @dataset.filter(true).sql.should ==
+      "SELECT * FROM test WHERE 't'"
+    @dataset.filter(false).sql.should ==
+      "SELECT * FROM test WHERE 'f'"
+  end
+
+  specify "should allow the use of multiple arguments" do
+    @dataset.filter(:a, :b).sql.should ==
+      'SELECT * FROM test WHERE (a AND b)'
+    @dataset.filter(:a, :b=>1).sql.should ==
+      'SELECT * FROM test WHERE (a AND (b = 1))'
+    @dataset.filter(:a, :c.sql_number > 3, :b=>1).sql.should ==
+      'SELECT * FROM test WHERE (a AND (c > 3) AND (b = 1))'
+  end
+
+  specify "should allow the use of blocks and arguments simultaneously" do
+    @dataset.filter(:zz.sql_number < 3){:yy.sql_number > 3}.sql.should ==
+      'SELECT * FROM test WHERE ((zz < 3) AND (yy > 3))'
+  end
+
+  specify "should yield a VirtualRow to the block" do
+    x = nil
+    @dataset.filter{|r| x = r; false}
+    x.should be_a_kind_of(Sequel::SQL::VirtualRow)
+    @dataset.filter{|r| ((r.name < 'b') & {r.table__id => 1}) | r.is_active(r.blah, r.xx, r.x__y_z)}.sql.should ==
+      "SELECT * FROM test WHERE (((name < 'b') AND (table.id = 1)) OR is_active(blah, xx, x.y_z))"
+  end
+
+  specify "should instance_eval the block in the context of a VirtualRow if the block doesn't request an argument" do
+    x = nil
+    @dataset.filter{x = self; false}
+    x.should be_a_kind_of(Sequel::SQL::VirtualRow)
+    @dataset.filter{((name < 'b') & {table__id => 1}) | is_active(blah, xx, x__y_z)}.sql.should ==
+      "SELECT * FROM test WHERE (((name < 'b') AND (table.id = 1)) OR is_active(blah, xx, x.y_z))"
+  end
+
+  specify "should raise an error if an invalid argument is used" do
+    proc{@dataset.filter(1)}.should raise_error(Sequel::Error)
+  end
+
+  specify "should raise an error if a NumericExpression or StringExpression is used" do
+    proc{@dataset.filter(:x + 1)}.should raise_error(Sequel::Error)
+    proc{@dataset.filter(:x.sql_string)}.should raise_error(Sequel::Error)
+  end
+end
+
+context "Dataset#or" do
+  before do
+    @dataset = Sequel::Dataset.new(nil).from(:test)
+    @d1 = @dataset.where(:x => 1)
+  end
+  
+  specify "should raise if no filter exists" do
+    proc {@dataset.or(:a => 1)}.should raise_error(Sequel::Error)
+  end
+  
+  specify "should add an alternative expression to the where clause" do
+    @d1.or(:y => 2).sql.should == 
+      'SELECT * FROM test WHERE ((x = 1) OR (y = 2))'
+  end
+  
+  specify "should accept all forms of filters" do
+    @d1.or('y > ?', 2).sql.should ==
+      'SELECT * FROM test WHERE ((x = 1) OR (y > 2))'
+    @d1.or(:yy.sql_number > 3).sql.should ==
+      'SELECT * FROM test WHERE ((x = 1) OR (yy > 3))'
+  end    
+
+  specify "should accept blocks passed to filter" do
+    @d1.or{:yy.sql_number > 3}.sql.should ==
+      'SELECT * FROM test WHERE ((x = 1) OR (yy > 3))'
+  end
+  
+  specify "should correctly add parens to give predictable results" do
+    @d1.filter(:y => 2).or(:z => 3).sql.should == 
+      'SELECT * FROM test WHERE (((x = 1) AND (y = 2)) OR (z = 3))'
+
+    @d1.or(:y => 2).filter(:z => 3).sql.should == 
+      'SELECT * FROM test WHERE (((x = 1) OR (y = 2)) AND (z = 3))'
+  end
+
+  specify "should allow the use of blocks and arguments simultaneously" do
+    @d1.or(:zz.sql_number < 3){:yy.sql_number > 3}.sql.should ==
+      'SELECT * FROM test WHERE ((x = 1) OR ((zz < 3) AND (yy > 3)))'
+  end
+end
+
+context "Dataset#and" do
+  before do
+    @dataset = Sequel::Dataset.new(nil).from(:test)
+    @d1 = @dataset.where(:x => 1)
+  end
+  
+  specify "should raise if no filter exists" do
+    proc {@dataset.and(:a => 1)}.should raise_error(Sequel::Error)
+    proc {@dataset.where(:a => 1).group(:t).and(:b => 2)}.should_not raise_error(Sequel::Error)
+    @dataset.where(:a => 1).group(:t).and(:b => 2).sql ==
+      "SELECT * FROM test WHERE (a = 1) AND (b = 2) GROUP BY t"
+  end
+  
+  specify "should add an alternative expression to the where clause" do
+    @d1.and(:y => 2).sql.should == 
+      'SELECT * FROM test WHERE ((x = 1) AND (y = 2))'
+  end
+  
+  specify "should accept all forms of filters" do
+    # probably not exhaustive, but good enough
+    @d1.and('y > ?', 2).sql.should ==
+      'SELECT * FROM test WHERE ((x = 1) AND (y > 2))'
+    @d1.and(:yy.sql_number > 3).sql.should ==
+      'SELECT * FROM test WHERE ((x = 1) AND (yy > 3))'
+  end
+      
+  specify "should accept blocks passed to filter" do
+    @d1.and {:yy.sql_number > 3}.sql.should ==
+      'SELECT * FROM test WHERE ((x = 1) AND (yy > 3))'
+  end
+  
+  specify "should correctly add parens to give predictable results" do
+    @d1.or(:y => 2).and(:z => 3).sql.should == 
+      'SELECT * FROM test WHERE (((x = 1) OR (y = 2)) AND (z = 3))'
+
+    @d1.and(:y => 2).or(:z => 3).sql.should == 
+      'SELECT * FROM test WHERE (((x = 1) AND (y = 2)) OR (z = 3))'
+  end
+end
+
+context "Dataset#exclude" do
+  before do
+    @dataset = Sequel::Dataset.new(nil).from(:test)
+  end
+
+  specify "should correctly negate the expression when one condition is given" do
+    @dataset.exclude(:region=>'Asia').select_sql.should ==
+      "SELECT * FROM test WHERE (region != 'Asia')"
+  end
+
+  specify "should take multiple conditions as a hash and express the logic correctly in SQL" do
+    @dataset.exclude(:region => 'Asia', :name => 'Japan').select_sql.
+      should match(Regexp.union(/WHERE \(\(region != 'Asia'\) OR \(name != 'Japan'\)\)/,
+                                /WHERE \(\(name != 'Japan'\) OR \(region != 'Asia'\)\)/))
+  end
+
+  specify "should parenthesize a single string condition correctly" do
+    @dataset.exclude("region = 'Asia' AND name = 'Japan'").select_sql.should ==
+      "SELECT * FROM test WHERE NOT (region = 'Asia' AND name = 'Japan')"
+  end
+
+  specify "should parenthesize an array condition correctly" do
+    @dataset.exclude('region = ? AND name = ?', 'Asia', 'Japan').select_sql.should ==
+      "SELECT * FROM test WHERE NOT (region = 'Asia' AND name = 'Japan')"
+  end
+
+  specify "should correctly parenthesize when it is used twice" do
+    @dataset.exclude(:region => 'Asia').exclude(:name => 'Japan').select_sql.should ==
+      "SELECT * FROM test WHERE ((region != 'Asia') AND (name != 'Japan'))"
+  end
+  
+  specify "should support proc expressions" do
+    @dataset.exclude{:id.sql_number < 6}.sql.should == 
+      'SELECT * FROM test WHERE (id >= 6)'
+  end
+  
+  specify "should allow the use of blocks and arguments simultaneously" do
+    @dataset.exclude(:id => (7..11)){:id.sql_number < 6}.sql.should == 
+      'SELECT * FROM test WHERE (((id < 7) OR (id > 11)) OR (id >= 6))'
+    @dataset.exclude([:id, 1], [:x, 3]){:id.sql_number < 6}.sql.should == 
+      'SELECT * FROM test WHERE (((id != 1) OR (x != 3)) OR (id >= 6))'
+  end
+end
+
+context "Dataset#invert" do
+  before do
+    @d = Sequel::Dataset.new(nil).from(:test)
+  end
+
+  specify "should raise error if the dataset is not filtered" do
+    proc{@d.invert}.should raise_error(Sequel::Error)
+  end
+
+  specify "should invert current filter if dataset is filtered" do
+    @d.filter(:x).invert.sql.should == 'SELECT * FROM test WHERE NOT x'
+  end
+
+  specify "should invert both having and where if both are preset" do
+    @d.filter(:x).group(:x).having(:x).invert.sql.should == 'SELECT * FROM test WHERE NOT x GROUP BY x HAVING NOT x'
+  end
+end
+
+context "Dataset#having" do
+  before do
+    @dataset = Sequel::Dataset.new(nil).from(:test)
+    @grouped = @dataset.group(:region).select(:region, :sum.sql_function(:population), :avg.sql_function(:gdp))
+    @d1 = @grouped.having('sum(population) > 10')
+    @d2 = @grouped.having(:region => 'Asia')
+    @columns = "region, sum(population), avg(gdp)"
+  end
+
+  specify "should raise if the dataset is not grouped" do
+    proc {@dataset.having('avg(gdp) > 10')}.should raise_error(Sequel::InvalidOperation)
+  end
+
+  specify "should affect select statements" do
+    @d1.select_sql.should ==
+      "SELECT #{@columns} FROM test GROUP BY region HAVING (sum(population) > 10)"
+  end
+
+  specify "should support proc expressions" do
+    @grouped.having {:sum.sql_function(:population) > 10}.sql.should == 
+      "SELECT #{@columns} FROM test GROUP BY region HAVING (sum(population) > 10)"
+  end
+
+  specify "should work with and on the having clause" do
+    @grouped.having( :a.sql_number > 1 ).and( :b.sql_number < 2 ).sql.should ==
+      "SELECT #{@columns} FROM test GROUP BY region HAVING ((a > 1) AND (b < 2))"
+  end
+end
+
+context "a grouped dataset" do
+  before do
+    @dataset = Sequel::Dataset.new(nil).from(:test).group(:type_id)
+  end
+
+  specify "should raise when trying to generate an update statement" do
+    proc {@dataset.update_sql(:id => 0)}.should raise_error
+  end
+
+  specify "should raise when trying to generate a delete statement" do
+    proc {@dataset.delete_sql}.should raise_error
+  end
+
+  specify "should specify the grouping in generated select statement" do
+    @dataset.select_sql.should ==
+      "SELECT * FROM test GROUP BY type_id"
+  end
+  
+  specify "should format the right statement for counting (as a subquery)" do
+    db = MockDatabase.new
+    db[:test].select(:name).group(:name).count
+    db.sqls.should == ["SELECT COUNT(*) AS count FROM (SELECT name FROM test GROUP BY name) AS t1 LIMIT 1"]
+  end
+end
+
+context "Dataset#group_by" do
+  before do
+    @dataset = Sequel::Dataset.new(nil).from(:test).group_by(:type_id)
+  end
+
+  specify "should raise when trying to generate an update statement" do
+    proc {@dataset.update_sql(:id => 0)}.should raise_error
+  end
+
+  specify "should raise when trying to generate a delete statement" do
+    proc {@dataset.delete_sql}.should raise_error
+  end
+
+  specify "should specify the grouping in generated select statement" do
+    @dataset.select_sql.should ==
+      "SELECT * FROM test GROUP BY type_id"
+    @dataset.group_by(:a, :b).select_sql.should ==
+      "SELECT * FROM test GROUP BY a, b"
+  end
+
+  specify "should specify the grouping in generated select statement" do
+    @dataset.group_by(:type_id=>nil).select_sql.should ==
+      "SELECT * FROM test GROUP BY (type_id IS NULL)"
+  end
+
+  specify "should be aliased as #group" do
+    @dataset.group(:type_id=>nil).select_sql.should ==
+      "SELECT * FROM test GROUP BY (type_id IS NULL)"
+  end
+end
+
+context "Dataset#as" do
+  specify "should set up an alias" do
+    dataset = Sequel::Dataset.new(nil).from(:test)
+    dataset.select(dataset.limit(1).select(:name).as(:n)).sql.should == \
+      'SELECT (SELECT name FROM test LIMIT 1) AS n FROM test'
+  end
+end
+
+context "Dataset#literal" do
+  before do
+    @dataset = Sequel::Dataset.new(nil).from(:test)
+  end
+  
+  specify "should escape strings properly" do
+    @dataset.literal('abc').should == "'abc'"
+    @dataset.literal('a"x"bc').should == "'a\"x\"bc'"
+    @dataset.literal("a'bc").should == "'a''bc'"
+    @dataset.literal("a''bc").should == "'a''''bc'"
+    @dataset.literal("a\\bc").should == "'a\\\\bc'"
+    @dataset.literal("a\\\\bc").should == "'a\\\\\\\\bc'"
+    @dataset.literal("a\\'bc").should == "'a\\\\''bc'"
+  end
+  
+  specify "should escape blobs as strings by default" do
+    @dataset.literal('abc'.to_sequel_blob).should == "'abc'"
+  end
+
+  specify "should literalize numbers properly" do
+    @dataset.literal(1).should == "1"
+    @dataset.literal(1.5).should == "1.5"
+  end
+  
+  specify "should literalize nil as NULL" do
+    @dataset.literal(nil).should == "NULL"
+  end
+  
+  specify "should literalize an array properly" do
+    @dataset.literal([]).should == "(NULL)"
+    @dataset.literal([1, 'abc', 3]).should == "(1, 'abc', 3)"
+    @dataset.literal([1, "a'b''c", 3]).should == "(1, 'a''b''''c', 3)"
+  end
+  
+  specify "should literalize symbols as column references" do
+    @dataset.literal(:name).should == "name"
+    @dataset.literal(:items__name).should == "items.name"
+  end
+  
+  specify "should raise an error for unsupported types" do
+    proc {@dataset.literal({})}.should raise_error
+  end
+  
+  specify "should literalize datasets as subqueries" do
+    d = @dataset.from(:test)
+    d.literal(d).should == "(#{d.sql})"
+  end
+  
+  specify "should literalize Time properly" do
+    t = Time.now
+    s = t.strftime("'%Y-%m-%dT%H:%M:%S%z'").gsub(/(\d\d')\z/, ':\1')
+    @dataset.literal(t).should == s
+  end
+  
+  specify "should literalize DateTime properly" do
+    t = DateTime.now
+    s = t.strftime("'%Y-%m-%dT%H:%M:%S%z'").gsub(/(\d\d')\z/, ':\1')
+    @dataset.literal(t).should == s
+  end
+  
+  specify "should literalize Date properly" do
+    d = Date.today
+    s = d.strftime("'%Y-%m-%d'")
+    @dataset.literal(d).should == s
+  end
+  
+  specify "should not modify literal strings" do
+    @dataset.literal('col1 + 2'.lit).should == 'col1 + 2'
+    
+    @dataset.update_sql(:a => 'a + 2'.lit).should == 
+      'UPDATE test SET a = a + 2'
+  end
+
+  specify "should literalize BigDecimal instances correctly" do
+    @dataset.literal(BigDecimal.new("80")).should == "80.0"
+    @dataset.literal(BigDecimal.new("NaN")).should == "'NaN'"
+    @dataset.literal(BigDecimal.new("Infinity")).should == "'Infinity'"
+    @dataset.literal(BigDecimal.new("-Infinity")).should == "'-Infinity'"
+  end
+
+  specify "should raise an Error if the object can't be literalized" do
+    proc{@dataset.literal(Object.new)}.should raise_error(Sequel::Error)
+  end
+end
+
+context "Dataset#from" do
+  before do
+    @dataset = Sequel::Dataset.new(nil)
+  end
+
+  specify "should accept a Dataset" do
+    proc {@dataset.from(@dataset)}.should_not raise_error
+  end
+
+  specify "should format a Dataset as a subquery if it has had options set" do
+    @dataset.from(@dataset.from(:a).where(:a=>1)).select_sql.should ==
+      "SELECT * FROM (SELECT * FROM a WHERE (a = 1)) AS t1"
+  end
+  
+  specify "should automatically alias sub-queries" do
+    @dataset.from(@dataset.from(:a).group(:b)).select_sql.should ==
+      "SELECT * FROM (SELECT * FROM a GROUP BY b) AS t1"
+      
+    d1 = @dataset.from(:a).group(:b)
+    d2 = @dataset.from(:c).group(:d)
+    
+    @dataset.from(d1, d2).sql.should == 
+      "SELECT * FROM (SELECT * FROM a GROUP BY b) AS t1, (SELECT * FROM c GROUP BY d) AS t2"
+  end
+  
+  specify "should accept a hash for aliasing" do
+    @dataset.from(:a => :b).sql.should ==
+      "SELECT * FROM a AS b"
+      
+    @dataset.from(:a => 'b').sql.should ==
+      "SELECT * FROM a AS b"
+
+    @dataset.from(@dataset.from(:a).group(:b) => :c).sql.should ==
+      "SELECT * FROM (SELECT * FROM a GROUP BY b) AS c"
+  end
+
+  specify "should always use a subquery if given a dataset" do
+    @dataset.from(@dataset.from(:a)).select_sql.should ==
+      "SELECT * FROM (SELECT * FROM a) AS t1"
+  end
+  
+  specify "should remove all FROM tables if called with no arguments" do
+    @dataset.from.sql.should == 'SELECT *'
+  end
+  
+  specify "should accept sql functions" do
+    @dataset.from(:abc.sql_function(:def)).select_sql.should ==
+      "SELECT * FROM abc(def)"
+    
+    @dataset.from(:a.sql_function(:i)).select_sql.should ==
+      "SELECT * FROM a(i)"
+  end
+
+  specify "should accept :schema__table___alias symbol format" do
+    @dataset.from(:abc__def).select_sql.should ==
+      "SELECT * FROM abc.def"
+    @dataset.from(:abc__def___d).select_sql.should ==
+      "SELECT * FROM abc.def AS d"
+    @dataset.from(:abc___def).select_sql.should ==
+      "SELECT * FROM abc AS def"
+  end
+end
+
+context "Dataset#select" do
+  before do
+    @d = Sequel::Dataset.new(nil).from(:test)
+  end
+
+  specify "should accept variable arity" do
+    @d.select(:name).sql.should == 'SELECT name FROM test'
+    @d.select(:a, :b, :test__c).sql.should == 'SELECT a, b, test.c FROM test'
+  end
+  
+  specify "should accept symbols and literal strings" do
+    @d.select('aaa'.lit).sql.should == 'SELECT aaa FROM test'
+    @d.select(:a, 'b'.lit).sql.should == 'SELECT a, b FROM test'
+    @d.select(:test__cc, 'test.d AS e'.lit).sql.should == 
+      'SELECT test.cc, test.d AS e FROM test'
+    @d.select('test.d AS e'.lit, :test__cc).sql.should == 
+      'SELECT test.d AS e, test.cc FROM test'
+
+    # symbol helpers      
+    @d.select(:test.*).sql.should ==
+      'SELECT test.* FROM test'
+    @d.select(:test__name.as(:n)).sql.should ==
+      'SELECT test.name AS n FROM test'
+    @d.select(:test__name___n).sql.should ==
+      'SELECT test.name AS n FROM test'
+  end
+
+  specify "should use the wildcard if no arguments are given" do
+    @d.select.sql.should == 'SELECT * FROM test'
+  end
+  
+  specify "should accept a hash for AS values" do
+    @d.select(:name => 'n', :__ggh => 'age').sql.should =~
+      /SELECT ((name AS n, __ggh AS age)|(__ggh AS age, name AS n)) FROM test/
+  end
+
+  specify "should overrun the previous select option" do
+    @d.select!(:a, :b, :c).select.sql.should == 'SELECT * FROM test'
+    @d.select!(:price).select(:name).sql.should == 'SELECT name FROM test'
+  end
+  
+  specify "should accept arbitrary objects and literalize them correctly" do
+    @d.select(1, :a, 't').sql.should == "SELECT 1, a, 't' FROM test"
+
+    @d.select(nil, :sum.sql_function(:t), :x___y).sql.should == "SELECT NULL, sum(t), x AS y FROM test"
+
+    @d.select(nil, 1, :x => :y).sql.should == "SELECT NULL, 1, x AS y FROM test"
+  end
+
+  specify "should accept a block that yields a virtual row" do
+    @d.select{|o| o.a}.sql.should == 'SELECT a FROM test'
+    @d.select{a(1)}.sql.should == 'SELECT a(1) FROM test'
+    @d.select{|o| o.a(1, 2)}.sql.should == 'SELECT a(1, 2) FROM test'
+    @d.select{[a, a(1, 2)]}.sql.should == 'SELECT a, a(1, 2) FROM test'
+  end
+
+  specify "should merge regular arguments with argument returned from block" do
+    @d.select(:b){a}.sql.should == 'SELECT b, a FROM test'
+    @d.select(:b, :c){|o| o.a(1)}.sql.should == 'SELECT b, c, a(1) FROM test'
+    @d.select(:b){[a, a(1, 2)]}.sql.should == 'SELECT b, a, a(1, 2) FROM test'
+    @d.select(:b, :c){|o| [o.a, o.a(1, 2)]}.sql.should == 'SELECT b, c, a, a(1, 2) FROM test'
+  end
+end
+
+context "Dataset#select_all" do
+  before do
+    @d = Sequel::Dataset.new(nil).from(:test)
+  end
+
+  specify "should select the wildcard" do
+    @d.select_all.sql.should == 'SELECT * FROM test'
+  end
+  
+  specify "should overrun the previous select option" do
+    @d.select!(:a, :b, :c).select_all.sql.should == 'SELECT * FROM test'
+  end
+end
+
+context "Dataset#select_more" do
+  before do
+    @d = Sequel::Dataset.new(nil).from(:test)
+  end
+  
+  specify "should act like #select for datasets with no selection" do
+    @d.select_more(:a, :b).sql.should == 'SELECT a, b FROM test'
+    @d.select_all.select_more(:a, :b).sql.should == 'SELECT a, b FROM test'
+    @d.select(:blah).select_all.select_more(:a, :b).sql.should == 'SELECT a, b FROM test'
+  end
+
+  specify "should add to the currently selected columns" do
+    @d.select(:a).select_more(:b).sql.should == 'SELECT a, b FROM test'
+    @d.select(:a.*).select_more(:b.*).sql.should == 'SELECT a.*, b.* FROM test'
+  end
+
+  specify "should accept a block that yields a virtual row" do
+    @d.select(:a).select_more{|o| o.b}.sql.should == 'SELECT a, b FROM test'
+    @d.select(:a.*).select_more(:b.*){b(1)}.sql.should == 'SELECT a.*, b.*, b(1) FROM test'
+  end
+end
+
+context "Dataset#order" do
+  before do
+    @dataset = Sequel::Dataset.new(nil).from(:test)
+  end
+  
+  specify "should include an ORDER BY clause in the select statement" do
+    @dataset.order(:name).sql.should == 
+      'SELECT * FROM test ORDER BY name'
+  end
+  
+  specify "should accept multiple arguments" do
+    @dataset.order(:name, :price.desc).sql.should ==
+      'SELECT * FROM test ORDER BY name, price DESC'
+  end
+  
+  specify "should overrun a previous ordering" do
+    @dataset.order(:name).order(:stamp).sql.should ==
+      'SELECT * FROM test ORDER BY stamp'
+  end
+  
+  specify "should accept a literal string" do
+    @dataset.order('dada ASC'.lit).sql.should ==
+      'SELECT * FROM test ORDER BY dada ASC'
+  end
+  
+  specify "should accept a hash as an expression" do
+    @dataset.order(:name=>nil).sql.should ==
+      'SELECT * FROM test ORDER BY (name IS NULL)'
+  end
+  
+  specify "should accept a nil to remove ordering" do
+    @dataset.order(:bah).order(nil).sql.should == 
+      'SELECT * FROM test'
+  end
+
+  specify "should accept a block that yields a virtual row" do
+    @dataset.order{|o| o.a}.sql.should == 'SELECT * FROM test ORDER BY a'
+    @dataset.order{a(1)}.sql.should == 'SELECT * FROM test ORDER BY a(1)'
+    @dataset.order{|o| o.a(1, 2)}.sql.should == 'SELECT * FROM test ORDER BY a(1, 2)'
+    @dataset.order{[a, a(1, 2)]}.sql.should == 'SELECT * FROM test ORDER BY a, a(1, 2)'
+  end
+
+  specify "should merge regular arguments with argument returned from block" do
+    @dataset.order(:b){a}.sql.should == 'SELECT * FROM test ORDER BY b, a'
+    @dataset.order(:b, :c){|o| o.a(1)}.sql.should == 'SELECT * FROM test ORDER BY b, c, a(1)'
+    @dataset.order(:b){[a, a(1, 2)]}.sql.should == 'SELECT * FROM test ORDER BY b, a, a(1, 2)'
+    @dataset.order(:b, :c){|o| [o.a, o.a(1, 2)]}.sql.should == 'SELECT * FROM test ORDER BY b, c, a, a(1, 2)'
+  end
+end
+
+context "Dataset#unfiltered" do
+  before do
+    @dataset = Sequel::Dataset.new(nil).from(:test)
+  end
+  
+  specify "should remove filtering from the dataset" do
+    @dataset.filter(:score=>1).unfiltered.sql.should ==
+      'SELECT * FROM test'
+  end
+end
+
+context "Dataset#unordered" do
+  before do
+    @dataset = Sequel::Dataset.new(nil).from(:test)
+  end
+  
+  specify "should remove ordering from the dataset" do
+    @dataset.order(:name).unordered.sql.should ==
+      'SELECT * FROM test'
+  end
+end
+
+context "Dataset#with_sql" do
+  before do
+    @dataset = Sequel::Dataset.new(nil).from(:test)
+  end
+  
+  specify "should remove use static sql" do
+    @dataset.with_sql('SELECT 1 FROM test').sql.should == 'SELECT 1 FROM test'
+  end
+  
+  specify "should keep row_proc" do
+    @dataset.with_sql('SELECT 1 FROM test').row_proc.should == @dataset.row_proc
+  end
+end
+
+context "Dataset#order_by" do
+  before do
+    @dataset = Sequel::Dataset.new(nil).from(:test)
+  end
+  
+  specify "should include an ORDER BY clause in the select statement" do
+    @dataset.order_by(:name).sql.should == 
+      'SELECT * FROM test ORDER BY name'
+  end
+  
+  specify "should accept multiple arguments" do
+    @dataset.order_by(:name, :price.desc).sql.should ==
+      'SELECT * FROM test ORDER BY name, price DESC'
+  end
+  
+  specify "should overrun a previous ordering" do
+    @dataset.order_by(:name).order(:stamp).sql.should ==
+      'SELECT * FROM test ORDER BY stamp'
+  end
+  
+  specify "should accept a string" do
+    @dataset.order_by('dada ASC'.lit).sql.should ==
+      'SELECT * FROM test ORDER BY dada ASC'
+  end
+
+  specify "should accept a nil to remove ordering" do
+    @dataset.order_by(:bah).order_by(nil).sql.should == 
+      'SELECT * FROM test'
+  end
+end
+
+context "Dataset#order_more" do
+  before do
+    @dataset = Sequel::Dataset.new(nil).from(:test)
+  end
+  
+  specify "should include an ORDER BY clause in the select statement" do
+    @dataset.order_more(:name).sql.should == 
+      'SELECT * FROM test ORDER BY name'
+  end
+  
+  specify "should add to a previous ordering" do
+    @dataset.order(:name).order_more(:stamp.desc).sql.should ==
+      'SELECT * FROM test ORDER BY name, stamp DESC'
+  end
+
+  specify "should accept a block that yields a virtual row" do
+    @dataset.order(:a).order_more{|o| o.b}.sql.should == 'SELECT * FROM test ORDER BY a, b'
+    @dataset.order(:a, :b).order_more(:c, :d){[e, f(1, 2)]}.sql.should == 'SELECT * FROM test ORDER BY a, b, c, d, e, f(1, 2)'
+  end
+end
+
+context "Dataset#reverse_order" do
+  before do
+    @dataset = Sequel::Dataset.new(nil).from(:test)
+  end
+  
+  specify "should use DESC as default order" do
+    @dataset.reverse_order(:name).sql.should == 
+      'SELECT * FROM test ORDER BY name DESC'
+  end
+  
+  specify "should invert the order given" do
+    @dataset.reverse_order(:name.desc).sql.should ==
+      'SELECT * FROM test ORDER BY name ASC'
+  end
+  
+  specify "should invert the order for ASC expressions" do
+    @dataset.reverse_order(:name.asc).sql.should ==
+      'SELECT * FROM test ORDER BY name DESC'
+  end
+  
+  specify "should accept multiple arguments" do
+    @dataset.reverse_order(:name, :price.desc).sql.should ==
+      'SELECT * FROM test ORDER BY name DESC, price ASC'
+  end
+
+  specify "should reverse a previous ordering if no arguments are given" do
+    @dataset.order(:name).reverse_order.sql.should ==
+      'SELECT * FROM test ORDER BY name DESC'
+    @dataset.order(:clumsy.desc, :fool).reverse_order.sql.should ==
+      'SELECT * FROM test ORDER BY clumsy ASC, fool DESC'
+  end
+  
+  specify "should return an unordered dataset for a dataset with no order" do
+    @dataset.unordered.reverse_order.sql.should == 
+      'SELECT * FROM test'
+  end
+  
+  specify "should have #reverse alias" do
+    @dataset.order(:name).reverse.sql.should ==
+      'SELECT * FROM test ORDER BY name DESC'
+  end
+end
+
+context "Dataset#limit" do
+  before do
+    @dataset = Sequel::Dataset.new(nil).from(:test)
+  end
+  
+  specify "should include a LIMIT clause in the select statement" do
+    @dataset.limit(10).sql.should == 
+      'SELECT * FROM test LIMIT 10'
+  end
+  
+  specify "should accept ranges" do
+    @dataset.limit(3..7).sql.should ==
+      'SELECT * FROM test LIMIT 5 OFFSET 3'
+      
+    @dataset.limit(3...7).sql.should ==
+      'SELECT * FROM test LIMIT 4 OFFSET 3'
+  end
+  
+  specify "should include an offset if a second argument is given" do
+    @dataset.limit(6, 10).sql.should ==
+      'SELECT * FROM test LIMIT 6 OFFSET 10'
+  end
+  
+  specify "should work with fixed sql datasets" do
+    @dataset.opts[:sql] = 'select * from cccc'
+    @dataset.limit(6, 10).sql.should ==
+      'SELECT * FROM (select * from cccc) AS t1 LIMIT 6 OFFSET 10'
+  end
+  
+  specify "should raise an error if an invalid limit or offset is used" do
+    proc{@dataset.limit(-1)}.should raise_error(Sequel::Error)
+    proc{@dataset.limit(0)}.should raise_error(Sequel::Error)
+    proc{@dataset.limit(1)}.should_not raise_error(Sequel::Error)
+    proc{@dataset.limit(1, -1)}.should raise_error(Sequel::Error)
+    proc{@dataset.limit(1, 0)}.should_not raise_error(Sequel::Error)
+    proc{@dataset.limit(1, 1)}.should_not raise_error(Sequel::Error)
+  end
+end
+
+context "Dataset#naked" do
+  before do
+    @d1 = Sequel::Dataset.new(nil, {1 => 2, 3 => 4})
+    @d2 = @d1.clone
+    @d2.row_proc = Proc.new{|r| r}
+  end
+  
+  specify "should remove any existing row_proc" do
+    naked = @d2.naked
+    naked.row_proc.should be_nil
+  end
+end
+
+context "Dataset#qualified_column_name" do
+  before do
+    @dataset = Sequel::Dataset.new(nil).from(:test)
+  end
+  
+  specify "should return the literal value if not given a symbol" do
+    @dataset.literal(@dataset.send(:qualified_column_name, 'ccc__b', :items)).should == "'ccc__b'"
+    @dataset.literal(@dataset.send(:qualified_column_name, 3, :items)).should == '3'
+    @dataset.literal(@dataset.send(:qualified_column_name, 'a'.lit, :items)).should == 'a'
+  end
+  
+  specify "should qualify the column with the supplied table name if given an unqualified symbol" do
+    @dataset.literal(@dataset.send(:qualified_column_name, :b1, :items)).should == 'items.b1'
+  end
+
+  specify "should not changed the qualifed column's table if given a qualified symbol" do
+    @dataset.literal(@dataset.send(:qualified_column_name, :ccc__b, :items)).should == 'ccc.b'
+  end
+end
+
+class DummyDataset < Sequel::Dataset
+  VALUES = [
+    {:a => 1, :b => 2},
+    {:a => 3, :b => 4},
+    {:a => 5, :b => 6}
+  ]
+  def fetch_rows(sql, &block)
+    VALUES.each(&block)
+  end
+end
+
+context "Dataset#map" do
+  before do
+    @d = DummyDataset.new(nil).from(:items)
+  end
+  
+  specify "should provide the usual functionality if no argument is given" do
+    @d.map {|n| n[:a] + n[:b]}.should == [3, 7, 11]
+  end
+  
+  specify "should map using #[column name] if column name is given" do
+    @d.map(:a).should == [1, 3, 5]
+  end
+  
+  specify "should return the complete dataset values if nothing is given" do
+    @d.map.to_a.should == DummyDataset::VALUES
+  end
+end
+
+context "Dataset#to_hash" do
+  before do
+    @d = DummyDataset.new(nil).from(:items)
+  end
+  
+  specify "should provide a hash with the first column as key and the second as value" do
+    @d.to_hash(:a, :b).should == {1 => 2, 3 => 4, 5 => 6}
+    @d.to_hash(:b, :a).should == {2 => 1, 4 => 3, 6 => 5}
+  end
+  
+  specify "should provide a hash with the first column as key and the entire hash as value if the value column is blank or nil" do
+    @d.to_hash(:a).should == {1 => {:a => 1, :b => 2}, 3 => {:a => 3, :b => 4}, 5 => {:a => 5, :b => 6}}
+    @d.to_hash(:b).should == {2 => {:a => 1, :b => 2}, 4 => {:a => 3, :b => 4}, 6 => {:a => 5, :b => 6}}
+  end
+end
+
+context "Dataset#distinct" do
+  before do
+    @db = MockDatabase.new
+    @dataset = @db[:test].select(:name)
+  end
+  
+  specify "should include DISTINCT clause in statement" do
+    @dataset.distinct.sql.should == 'SELECT DISTINCT name FROM test'
+  end
+  
+  specify "should accept an expression list" do
+    @dataset.distinct(:a, :b).sql.should == 'SELECT DISTINCT ON (a, b) name FROM test'
+    @dataset.distinct(:stamp.cast(:integer), :node_id=>nil).sql.should == 'SELECT DISTINCT ON (CAST(stamp AS integer), (node_id IS NULL)) name FROM test'
+  end
+
+  specify "should do a subselect for count" do
+    @dataset.distinct.count
+    @db.sqls.should == ['SELECT COUNT(*) AS count FROM (SELECT DISTINCT name FROM test) AS t1 LIMIT 1']
+  end
+end
+
+context "Dataset#count" do
+  before do
+    @c = Class.new(Sequel::Dataset) do
+      def self.sql
+        @@sql
+      end
+      
+      def fetch_rows(sql)
+        @columns = [sql =~ /SELECT COUNT/i ? :count : :a]
+        @@sql = sql
+        yield({@columns.first=>1})
+      end
+    end
+    @dataset = @c.new(nil).from(:test)
+  end
+  
+  specify "should format SQL properly" do
+    @dataset.count.should == 1
+    @c.sql.should == 'SELECT COUNT(*) AS count FROM test LIMIT 1'
+  end
+  
+  specify "should include the where clause if it's there" do
+    @dataset.filter(:abc.sql_number < 30).count.should == 1
+    @c.sql.should == 'SELECT COUNT(*) AS count FROM test WHERE (abc < 30) LIMIT 1'
+  end
+  
+  specify "should count properly for datasets with fixed sql" do
+    @dataset.opts[:sql] = "select abc from xyz"
+    @dataset.count.should == 1
+    @c.sql.should == "SELECT COUNT(*) AS count FROM (select abc from xyz) AS t1 LIMIT 1"
+  end
+
+  specify "should count properly when using UNION, INTERSECT, or EXCEPT" do
+    @dataset.union(@dataset).count.should == 1
+    @c.sql.should == "SELECT COUNT(*) AS count FROM (SELECT * FROM test UNION SELECT * FROM test) AS t1 LIMIT 1"
+    @dataset.intersect(@dataset).count.should == 1
+    @c.sql.should == "SELECT COUNT(*) AS count FROM (SELECT * FROM test INTERSECT SELECT * FROM test) AS t1 LIMIT 1"
+    @dataset.except(@dataset).count.should == 1
+    @c.sql.should == "SELECT COUNT(*) AS count FROM (SELECT * FROM test EXCEPT SELECT * FROM test) AS t1 LIMIT 1"
+  end
+
+  specify "should return limit if count is greater than it" do
+    @dataset.limit(5).count.should == 1
+    @c.sql.should == "SELECT COUNT(*) AS count FROM (SELECT * FROM test LIMIT 5) AS t1 LIMIT 1"
+  end
+  
+  it "should work on a graphed_dataset" do
+    @dataset.should_receive(:columns).twice.and_return([:a])
+    @dataset.graph(@dataset, [:a], :table_alias=>:test2).count.should == 1
+    @c.sql.should == 'SELECT COUNT(*) AS count FROM test LEFT OUTER JOIN test AS test2 USING (a) LIMIT 1'
+  end
+
+  specify "should not cache the columns value" do
+    ds = @dataset.from(:blah)
+    ds.columns.should == [:a]
+    ds.count.should == 1
+    @c.sql.should == 'SELECT COUNT(*) AS count FROM blah LIMIT 1'
+    ds.columns.should == [:a]
+  end
+end
+
+
+context "Dataset#group_and_count" do
+  before do
+    @c = Class.new(Sequel::Dataset) do
+      def self.sql
+        @@sql
+      end
+      
+      def fetch_rows(sql)
+        @@sql = sql
+        yield({1 => 1})
+      end
+    end
+    @ds = @c.new(nil).from(:test)
+  end
+  
+  specify "should format SQL properly" do
+    @ds.group_and_count(:name).sql.should == 
+      "SELECT name, count(*) AS count FROM test GROUP BY name ORDER BY count"
+  end
+
+  specify "should accept multiple columns for grouping" do
+    @ds.group_and_count(:a, :b).sql.should == 
+      "SELECT a, b, count(*) AS count FROM test GROUP BY a, b ORDER BY count"
+  end
+end
+
+context "Dataset#empty?" do
+  specify "should return true if records exist in the dataset" do
+    @c = Class.new(Sequel::Dataset) do
+      def self.sql
+        @@sql
+      end
+      
+      def fetch_rows(sql)
+        @@sql = sql
+        yield({1 => 1}) unless sql =~ /WHERE 'f'/
+      end
+    end
+    @c.new(nil).from(:test).should_not be_empty
+    @c.sql.should == 'SELECT 1 FROM test LIMIT 1'
+    @c.new(nil).from(:test).filter(false).should be_empty
+    @c.sql.should == "SELECT 1 FROM test WHERE 'f' LIMIT 1"
+  end
+end
+
+context "Dataset#join_table" do
+  before do
+    @d = MockDataset.new(nil).from(:items)
+    @d.quote_identifiers = true
+  end
+  
+  specify "should format the JOIN clause properly" do
+    @d.join_table(:left_outer, :categories, :category_id => :id).sql.should ==
+      'SELECT * FROM "items" LEFT OUTER JOIN "categories" ON ("categories"."category_id" = "items"."id")'
+  end
+  
+  specify "should handle multiple conditions on the same join table column" do
+    @d.join_table(:left_outer, :categories, [[:category_id, :id], [:category_id, 0..100]]).sql.should ==
+      'SELECT * FROM "items" LEFT OUTER JOIN "categories" ON (("categories"."category_id" = "items"."id") AND (("categories"."category_id" >= 0) AND ("categories"."category_id" <= 100)))'
+  end
+  
+  specify "should include WHERE clause if applicable" do
+    @d.filter(:price.sql_number < 100).join_table(:right_outer, :categories, :category_id => :id).sql.should ==
+      'SELECT * FROM "items" RIGHT OUTER JOIN "categories" ON ("categories"."category_id" = "items"."id") WHERE ("price" < 100)'
+  end
+  
+  specify "should include ORDER BY clause if applicable" do
+    @d.order(:stamp).join_table(:full_outer, :categories, :category_id => :id).sql.should ==
+      'SELECT * FROM "items" FULL OUTER JOIN "categories" ON ("categories"."category_id" = "items"."id") ORDER BY "stamp"'
+  end
+  
+  specify "should support multiple joins" do
+    @d.join_table(:inner, :b, :items_id=>:id).join_table(:left_outer, :c, :b_id => :b__id).sql.should ==
+      'SELECT * FROM "items" INNER JOIN "b" ON ("b"."items_id" = "items"."id") LEFT OUTER JOIN "c" ON ("c"."b_id" = "b"."id")'
+  end
+  
+  specify "should support left outer joins" do
+    @d.join_table(:left_outer, :categories, :category_id=>:id).sql.should ==
+      'SELECT * FROM "items" LEFT OUTER JOIN "categories" ON ("categories"."category_id" = "items"."id")'
+
+    @d.left_outer_join(:categories, :category_id=>:id).sql.should ==
+      'SELECT * FROM "items" LEFT OUTER JOIN "categories" ON ("categories"."category_id" = "items"."id")'
+  end
+
+  specify "should support right outer joins" do
+    @d.join_table(:right_outer, :categories, :category_id=>:id).sql.should ==
+      'SELECT * FROM "items" RIGHT OUTER JOIN "categories" ON ("categories"."category_id" = "items"."id")'
+
+    @d.right_outer_join(:categories, :category_id=>:id).sql.should ==
+      'SELECT * FROM "items" RIGHT OUTER JOIN "categories" ON ("categories"."category_id" = "items"."id")'
+  end
+
+  specify "should support full outer joins" do
+    @d.join_table(:full_outer, :categories, :category_id=>:id).sql.should ==
+      'SELECT * FROM "items" FULL OUTER JOIN "categories" ON ("categories"."category_id" = "items"."id")'
+
+    @d.full_outer_join(:categories, :category_id=>:id).sql.should ==
+      'SELECT * FROM "items" FULL OUTER JOIN "categories" ON ("categories"."category_id" = "items"."id")'
+  end
+
+  specify "should support inner joins" do
+    @d.join_table(:inner, :categories, :category_id=>:id).sql.should ==
+      'SELECT * FROM "items" INNER JOIN "categories" ON ("categories"."category_id" = "items"."id")'
+
+    @d.inner_join(:categories, :category_id=>:id).sql.should ==
+      'SELECT * FROM "items" INNER JOIN "categories" ON ("categories"."category_id" = "items"."id")'
+  end
+  
+  specify "should default to a plain join if nil is used for the type" do
+    @d.join_table(nil, :categories, :category_id=>:id).sql.should ==
+      'SELECT * FROM "items"  JOIN "categories" ON ("categories"."category_id" = "items"."id")'
+  end
+
+  specify "should use an inner join for Dataset#join" do
+    @d.join(:categories, :category_id=>:id).sql.should ==
+      'SELECT * FROM "items" INNER JOIN "categories" ON ("categories"."category_id" = "items"."id")'
+  end
+  
+  specify "should support aliased tables using the deprecated argument" do
+    @d.from('stats').join('players', {:id => :player_id}, 'p').sql.should ==
+      'SELECT * FROM "stats" INNER JOIN "players" AS "p" ON ("p"."id" = "stats"."player_id")'
+  end
+
+  specify "should support aliased tables using the :table_alias option" do
+    @d.from('stats').join('players', {:id => :player_id}, :table_alias=>:p).sql.should ==
+      'SELECT * FROM "stats" INNER JOIN "players" AS "p" ON ("p"."id" = "stats"."player_id")'
+  end
+  
+  specify "should support using an alias for the FROM when doing the first join with unqualified condition columns" do
+    ds = MockDataset.new(nil).from(:foo => :f)
+    ds.quote_identifiers = true
+    ds.join_table(:inner, :bar, :id => :bar_id).sql.should ==
+      'SELECT * FROM "foo" AS "f" INNER JOIN "bar" ON ("bar"."id" = "f"."bar_id")'
+  end
+  
+  specify "should support implicit schemas in from table symbols" do
+    @d.from(:s__t).join(:u__v, {:id => :player_id}).sql.should ==
+      'SELECT * FROM "s"."t" INNER JOIN "u"."v" ON ("u"."v"."id" = "s"."t"."player_id")'
+  end
+
+  specify "should support implicit aliases in from table symbols" do
+    @d.from(:t___z).join(:v___y, {:id => :player_id}).sql.should ==
+      'SELECT * FROM "t" AS "z" INNER JOIN "v" AS "y" ON ("y"."id" = "z"."player_id")'
+    @d.from(:s__t___z).join(:u__v___y, {:id => :player_id}).sql.should ==
+      'SELECT * FROM "s"."t" AS "z" INNER JOIN "u"."v" AS "y" ON ("y"."id" = "z"."player_id")'
+  end
+  
+  specify "should support AliasedExpressions" do
+    @d.from(:s.as(:t)).join(:u.as(:v), {:id => :player_id}).sql.should ==
+      'SELECT * FROM "s" AS "t" INNER JOIN "u" AS "v" ON ("v"."id" = "t"."player_id")'
+  end
+
+  specify "should support the :implicit_qualifier option" do
+    @d.from('stats').join('players', {:id => :player_id}, :implicit_qualifier=>:p).sql.should ==
+      'SELECT * FROM "stats" INNER JOIN "players" ON ("players"."id" = "p"."player_id")'
+  end
+  
+  specify "should allow for arbitrary conditions in the JOIN clause" do
+    @d.join_table(:left_outer, :categories, :status => 0).sql.should ==
+      'SELECT * FROM "items" LEFT OUTER JOIN "categories" ON ("categories"."status" = 0)'
+    @d.join_table(:left_outer, :categories, :categorizable_type => "Post").sql.should ==
+      'SELECT * FROM "items" LEFT OUTER JOIN "categories" ON ("categories"."categorizable_type" = \'Post\')'
+    @d.join_table(:left_outer, :categories, :timestamp => "CURRENT_TIMESTAMP".lit).sql.should ==
+      'SELECT * FROM "items" LEFT OUTER JOIN "categories" ON ("categories"."timestamp" = CURRENT_TIMESTAMP)'
+    @d.join_table(:left_outer, :categories, :status => [1, 2, 3]).sql.should ==
+      'SELECT * FROM "items" LEFT OUTER JOIN "categories" ON ("categories"."status" IN (1, 2, 3))'
+  end
+  
+  specify "should raise error for a table without a source" do
+    proc {Sequel::Dataset.new(nil).join('players', :id => :player_id)}. \
+      should raise_error(Sequel::Error)
+  end
+
+  specify "should support joining datasets" do
+    ds = Sequel::Dataset.new(nil).from(:categories)
+    
+    @d.join_table(:left_outer, ds, :item_id => :id).sql.should ==
+      'SELECT * FROM "items" LEFT OUTER JOIN (SELECT * FROM categories) AS "t1" ON ("t1"."item_id" = "items"."id")'
+      
+    ds.filter!(:active => true)
+
+    @d.join_table(:left_outer, ds, :item_id => :id).sql.should ==
+      'SELECT * FROM "items" LEFT OUTER JOIN (SELECT * FROM categories WHERE (active IS TRUE)) AS "t1" ON ("t1"."item_id" = "items"."id")'
+  end
+  
+  specify "should support joining datasets and aliasing the join" do
+    ds = Sequel::Dataset.new(nil).from(:categories)
+    
+    @d.join_table(:left_outer, ds, {:ds__item_id => :id}, :ds).sql.should ==
+      'SELECT * FROM "items" LEFT OUTER JOIN (SELECT * FROM categories) AS "ds" ON ("ds"."item_id" = "items"."id")'      
+  end
+  
+  specify "should support joining multiple datasets" do
+    ds = Sequel::Dataset.new(nil).from(:categories)
+    ds2 = Sequel::Dataset.new(nil).from(:nodes).select(:name)
+    ds3 = Sequel::Dataset.new(nil).from(:attributes).filter("name = 'blah'")
+
+    @d.join_table(:left_outer, ds, :item_id => :id).join_table(:inner, ds2, :node_id=>:id).join_table(:right_outer, ds3, :attribute_id=>:id).sql.should ==
+      'SELECT * FROM "items" LEFT OUTER JOIN (SELECT * FROM categories) AS "t1" ON ("t1"."item_id" = "items"."id") ' \
+      'INNER JOIN (SELECT name FROM nodes) AS "t2" ON ("t2"."node_id" = "t1"."id") ' \
+      'RIGHT OUTER JOIN (SELECT * FROM attributes WHERE (name = \'blah\')) AS "t3" ON ("t3"."attribute_id" = "t2"."id")'
+  end
+
+  specify "should support joining objects that respond to :table_name" do
+    ds = Object.new
+    def ds.table_name; :categories end
+    
+    @d.join(ds, :item_id => :id).sql.should ==
+      'SELECT * FROM "items" INNER JOIN "categories" ON ("categories"."item_id" = "items"."id")'
+  end
+  
+  specify "should support using a SQL String as the join condition" do
+    @d.join(:categories, %{c.item_id = items.id}, :c).sql.should ==
+      'SELECT * FROM "items" INNER JOIN "categories" AS "c" ON (c.item_id = items.id)'
+  end
+  
+  specify "should support using a boolean column as the join condition" do
+    @d.join(:categories, :active).sql.should ==
+      'SELECT * FROM "items" INNER JOIN "categories" ON "active"'
+  end
+
+  specify "should support using an expression as the join condition" do
+    @d.join(:categories, :number.sql_number > 10).sql.should ==
+      'SELECT * FROM "items" INNER JOIN "categories" ON ("number" > 10)'
+  end
+
+  specify "should support natural and cross joins using nil" do
+    @d.join_table(:natural, :categories).sql.should ==
+      'SELECT * FROM "items" NATURAL JOIN "categories"'
+    @d.join_table(:cross, :categories, nil).sql.should ==
+      'SELECT * FROM "items" CROSS JOIN "categories"'
+    @d.join_table(:natural, :categories, nil, :c).sql.should ==
+      'SELECT * FROM "items" NATURAL JOIN "categories" AS "c"'
+  end
+
+  specify "should support joins with a USING clause if an array of symbols is used" do
+    @d.join(:categories, [:id]).sql.should ==
+      'SELECT * FROM "items" INNER JOIN "categories" USING ("id")'
+    @d.join(:categories, [:id1, :id2]).sql.should ==
+      'SELECT * FROM "items" INNER JOIN "categories" USING ("id1", "id2")'
+  end
+
+  specify "should raise an error if using an array of symbols with a block" do
+    proc{@d.join(:categories, [:id]){|j,lj,js|}}.should raise_error(Sequel::Error)
+  end
+
+  specify "should support using a block that receieves the join table/alias, last join table/alias, and array of previous joins" do
+    @d.join(:categories) do |join_alias, last_join_alias, joins| 
+      join_alias.should == :categories
+      last_join_alias.should == :items
+      joins.should == []
+    end
+
+    @d.from(:items=>:i).join(:categories, nil, :c) do |join_alias, last_join_alias, joins| 
+      join_alias.should == :c
+      last_join_alias.should == :i
+      joins.should == []
+    end
+
+    @d.from(:items___i).join(:categories, nil, :c) do |join_alias, last_join_alias, joins| 
+      join_alias.should == :c
+      last_join_alias.should == :i
+      joins.should == []
+    end
+
+    @d.join(:blah).join(:categories, nil, :c) do |join_alias, last_join_alias, joins| 
+      join_alias.should == :c
+      last_join_alias.should == :blah
+      joins.should be_a_kind_of(Array)
+      joins.length.should == 1
+      joins.first.should be_a_kind_of(Sequel::SQL::JoinClause)
+      joins.first.join_type.should == :inner
+    end
+
+    @d.join_table(:natural, :blah, nil, :b).join(:categories, nil, :c) do |join_alias, last_join_alias, joins| 
+      join_alias.should == :c
+      last_join_alias.should == :b
+      joins.should be_a_kind_of(Array)
+      joins.length.should == 1
+      joins.first.should be_a_kind_of(Sequel::SQL::JoinClause)
+      joins.first.join_type.should == :natural
+    end
+
+    @d.join(:blah).join(:categories).join(:blah2) do |join_alias, last_join_alias, joins| 
+      join_alias.should == :blah2
+      last_join_alias.should == :categories
+      joins.should be_a_kind_of(Array)
+      joins.length.should == 2
+      joins.first.should be_a_kind_of(Sequel::SQL::JoinClause)
+      joins.first.table.should == :blah
+      joins.last.should be_a_kind_of(Sequel::SQL::JoinClause)
+      joins.last.table.should == :categories
+    end
+  end
+
+  specify "should use the block result as the only condition if no condition is given" do
+    @d.join(:categories){|j,lj,js| {:b.qualify(j)=>:c.qualify(lj)}}.sql.should ==
+      'SELECT * FROM "items" INNER JOIN "categories" ON ("categories"."b" = "items"."c")'
+    @d.join(:categories){|j,lj,js| :b.qualify(j) > :c.qualify(lj)}.sql.should ==
+      'SELECT * FROM "items" INNER JOIN "categories" ON ("categories"."b" > "items"."c")'
+  end
+
+  specify "should combine the block conditions and argument conditions if both given" do
+    @d.join(:categories, :a=>:d){|j,lj,js| {:b.qualify(j)=>:c.qualify(lj)}}.sql.should ==
+      'SELECT * FROM "items" INNER JOIN "categories" ON (("categories"."a" = "items"."d") AND ("categories"."b" = "items"."c"))'
+    @d.join(:categories, :a=>:d){|j,lj,js| :b.qualify(j) > :c.qualify(lj)}.sql.should ==
+      'SELECT * FROM "items" INNER JOIN "categories" ON (("categories"."a" = "items"."d") AND ("categories"."b" > "items"."c"))'
+  end
+end
+
+context "Dataset#[]=" do
+  before do
+    c = Class.new(Sequel::Dataset) do
+      def last_sql
+        @@last_sql
+      end
+      
+      def update(*args)
+        @@last_sql = update_sql(*args)
+      end
+    end
+    
+    @d = c.new(nil).from(:items)
+  end
+  
+  specify "should perform an update on the specified filter" do
+    @d[:a => 1] = {:x => 3}
+    @d.last_sql.should == 'UPDATE items SET x = 3 WHERE (a = 1)'
+  end
+end
+
+context "Dataset#set" do
+  before do
+    c = Class.new(Sequel::Dataset) do
+      def last_sql
+        @@last_sql
+      end
+      
+      def update(*args, &block)
+        @@last_sql = update_sql(*args, &block)
+      end
+    end
+    
+    @d = c.new(nil).from(:items)
+  end
+  
+  specify "should act as alias to #update" do
+    @d.set({:x => 3})
+    @d.last_sql.should == 'UPDATE items SET x = 3'
+  end
+end
+
+
+context "Dataset#insert_multiple" do
+  before do
+    c = Class.new(Sequel::Dataset) do
+      attr_reader :inserts
+      def insert(arg)
+        @inserts ||= []
+        @inserts << arg
+      end
+    end
+    
+    @d = c.new(nil)
+  end
+  
+  specify "should insert all items in the supplied array" do
+    @d.insert_multiple [:aa, 5, 3, {1 => 2}]
+    @d.inserts.should == [:aa, 5, 3, {1 => 2}]
+  end
+  
+  specify "should pass array items through the supplied block if given" do
+    a = ["inevitable", "hello", "the ticking clock"]
+    @d.insert_multiple(a) {|i| i.gsub('l', 'r')}
+    @d.inserts.should == ["inevitabre", "herro", "the ticking crock"]
+  end
+end
+
+context "Dataset aggregate methods" do
+  before do
+    c = Class.new(Sequel::Dataset) do
+      def fetch_rows(sql)
+        yield({1 => sql})
+      end
+    end
+    @d = c.new(nil).from(:test)
+  end
+  
+  specify "should include min" do
+    @d.min(:a).should == 'SELECT min(a) FROM test LIMIT 1'
+  end
+  
+  specify "should include max" do
+    @d.max(:b).should == 'SELECT max(b) FROM test LIMIT 1'
+  end
+  
+  specify "should include sum" do
+    @d.sum(:c).should == 'SELECT sum(c) FROM test LIMIT 1'
+  end
+  
+  specify "should include avg" do
+    @d.avg(:d).should == 'SELECT avg(d) FROM test LIMIT 1'
+  end
+  
+  specify "should accept qualified columns" do
+    @d.avg(:test__bc).should == 'SELECT avg(test.bc) FROM test LIMIT 1'
+  end
+end
+
+context "Dataset#range" do
+  before do
+    c = Class.new(Sequel::Dataset) do
+      @@sql = nil
+      
+      def last_sql; @@sql; end
+      
+      def fetch_rows(sql)
+        @@sql = sql
+        yield(:v1 => 1, :v2 => 10)
+      end
+    end
+    @d = c.new(nil).from(:test)
+  end
+  
+  specify "should generate a correct SQL statement" do
+    @d.range(:stamp)
+    @d.last_sql.should == "SELECT min(stamp) AS v1, max(stamp) AS v2 FROM test LIMIT 1"
+
+    @d.filter(:price.sql_number > 100).range(:stamp)
+    @d.last_sql.should == "SELECT min(stamp) AS v1, max(stamp) AS v2 FROM test WHERE (price > 100) LIMIT 1"
+  end
+  
+  specify "should return a range object" do
+    @d.range(:tryme).should == (1..10)
+  end
+end
+
+context "Dataset#interval" do
+  before do
+    c = Class.new(Sequel::Dataset) do
+      @@sql = nil
+      
+      def last_sql; @@sql; end
+      
+      def fetch_rows(sql)
+        @@sql = sql
+        yield(:v => 1234)
+      end
+    end
+    @d = c.new(nil).from(:test)
+  end
+  
+  specify "should generate a correct SQL statement" do
+    @d.interval(:stamp)
+    @d.last_sql.should == "SELECT (max(stamp) - min(stamp)) FROM test LIMIT 1"
+
+    @d.filter(:price.sql_number > 100).interval(:stamp)
+    @d.last_sql.should == "SELECT (max(stamp) - min(stamp)) FROM test WHERE (price > 100) LIMIT 1"
+  end
+  
+  specify "should return an integer" do
+    @d.interval(:tryme).should == 1234
+  end
+end
+
+context "Dataset #first and #last" do
+  before do
+    @c = Class.new(Sequel::Dataset) do
+      def each(&block)
+        s = select_sql
+        x = [:a,1,:b,2,s]
+        i = /LIMIT (\d+)/.match(s)[1].to_i.times{yield x}
+      end
+    end
+    @d = @c.new(nil).from(:test)
+  end
+  
+  specify "should return a single record if no argument is given" do
+    @d.order(:a).first.should == [:a,1,:b,2, 'SELECT * FROM test ORDER BY a LIMIT 1']
+    @d.order(:a).last.should == [:a,1,:b,2, 'SELECT * FROM test ORDER BY a DESC LIMIT 1']
+  end
+
+  specify "should return the first/last matching record if argument is not an Integer" do
+    @d.order(:a).first(:z => 26).should == [:a,1,:b,2, 'SELECT * FROM test WHERE (z = 26) ORDER BY a LIMIT 1']
+    @d.order(:a).first('z = ?', 15).should == [:a,1,:b,2, 'SELECT * FROM test WHERE (z = 15) ORDER BY a LIMIT 1']
+    @d.order(:a).last(:z => 26).should == [:a,1,:b,2, 'SELECT * FROM test WHERE (z = 26) ORDER BY a DESC LIMIT 1']
+    @d.order(:a).last('z = ?', 15).should == [:a,1,:b,2, 'SELECT * FROM test WHERE (z = 15) ORDER BY a DESC LIMIT 1']
+  end
+  
+  specify "should set the limit and return an array of records if the given number is > 1" do
+    i = rand(10) + 10
+    r = @d.order(:a).first(i).should == [[:a,1,:b,2, "SELECT * FROM test ORDER BY a LIMIT #{i}"]] * i
+    i = rand(10) + 10
+    r = @d.order(:a).last(i).should == [[:a,1,:b,2, "SELECT * FROM test ORDER BY a DESC LIMIT #{i}"]] * i
+  end
+  
+  specify "should return the first matching record if a block is given without an argument" do
+    @d.first{:z.sql_number > 26}.should == [:a,1,:b,2, 'SELECT * FROM test WHERE (z > 26) LIMIT 1']
+    @d.order(:name).last{:z.sql_number > 26}.should == [:a,1,:b,2, 'SELECT * FROM test WHERE (z > 26) ORDER BY name DESC LIMIT 1']
+  end
+  
+  specify "should combine block and standard argument filters if argument is not an Integer" do
+    @d.first(:y=>25){:z.sql_number > 26}.should == [:a,1,:b,2, 'SELECT * FROM test WHERE ((z > 26) AND (y = 25)) LIMIT 1']
+    @d.order(:name).last('y = ?', 16){:z.sql_number > 26}.should == [:a,1,:b,2, 'SELECT * FROM test WHERE ((z > 26) AND (y = 16)) ORDER BY name DESC LIMIT 1']
+  end
+  
+  specify "should filter and return an array of records if an Integer argument is provided and a block is given" do
+    i = rand(10) + 10
+    r = @d.order(:a).first(i){:z.sql_number > 26}.should == [[:a,1,:b,2, "SELECT * FROM test WHERE (z > 26) ORDER BY a LIMIT #{i}"]] * i
+    i = rand(10) + 10
+    r = @d.order(:a).last(i){:z.sql_number > 26}.should == [[:a,1,:b,2, "SELECT * FROM test WHERE (z > 26) ORDER BY a DESC LIMIT #{i}"]] * i
+  end
+  
+  specify "#last should raise if no order is given" do
+    proc {@d.last}.should raise_error(Sequel::Error)
+    proc {@d.last(2)}.should raise_error(Sequel::Error)
+    proc {@d.order(:a).last}.should_not raise_error
+    proc {@d.order(:a).last(2)}.should_not raise_error
+  end
+  
+  specify "#last should invert the order" do
+    @d.order(:a).last.pop.should == 'SELECT * FROM test ORDER BY a DESC LIMIT 1'
+    @d.order(:b.desc).last.pop.should == 'SELECT * FROM test ORDER BY b ASC LIMIT 1'
+    @d.order(:c, :d).last.pop.should == 'SELECT * FROM test ORDER BY c DESC, d DESC LIMIT 1'
+    @d.order(:e.desc, :f).last.pop.should == 'SELECT * FROM test ORDER BY e ASC, f DESC LIMIT 1'
+  end
+end
+
+context "Dataset compound operations" do
+  before do
+    @a = Sequel::Dataset.new(nil).from(:a).filter(:z => 1)
+    @b = Sequel::Dataset.new(nil).from(:b).filter(:z => 2)
+  end
+  
+  specify "should support UNION and UNION ALL" do
+    @a.union(@b).sql.should == \
+      "SELECT * FROM a WHERE (z = 1) UNION SELECT * FROM b WHERE (z = 2)"
+    @b.union(@a, true).sql.should == \
+      "SELECT * FROM b WHERE (z = 2) UNION ALL SELECT * FROM a WHERE (z = 1)"
+  end
+
+  specify "should support INTERSECT and INTERSECT ALL" do
+    @a.intersect(@b).sql.should == \
+      "SELECT * FROM a WHERE (z = 1) INTERSECT SELECT * FROM b WHERE (z = 2)"
+    @b.intersect(@a, true).sql.should == \
+      "SELECT * FROM b WHERE (z = 2) INTERSECT ALL SELECT * FROM a WHERE (z = 1)"
+  end
+
+  specify "should support EXCEPT and EXCEPT ALL" do
+    @a.except(@b).sql.should == \
+      "SELECT * FROM a WHERE (z = 1) EXCEPT SELECT * FROM b WHERE (z = 2)"
+    @b.except(@a, true).sql.should == \
+      "SELECT * FROM b WHERE (z = 2) EXCEPT ALL SELECT * FROM a WHERE (z = 1)"
+  end
+    
+  specify "should handle chained compound operations" do
+    @a.union(@b).union(@a, true).sql.should == \
+      "SELECT * FROM a WHERE (z = 1) UNION SELECT * FROM b WHERE (z = 2) UNION ALL SELECT * FROM a WHERE (z = 1)"
+    @a.intersect(@b, true).intersect(@a).sql.should == \
+      "SELECT * FROM a WHERE (z = 1) INTERSECT ALL SELECT * FROM b WHERE (z = 2) INTERSECT SELECT * FROM a WHERE (z = 1)"
+    @a.except(@b).except(@a, true).sql.should == \
+      "SELECT * FROM a WHERE (z = 1) EXCEPT SELECT * FROM b WHERE (z = 2) EXCEPT ALL SELECT * FROM a WHERE (z = 1)"
+    @a.union(@b, true).intersect(@a).except(@b, true).union(@a).intersect(@b, true).except(@a).sql.should == \
+      "SELECT * FROM a WHERE (z = 1) UNION ALL SELECT * FROM b WHERE (z = 2) INTERSECT SELECT * FROM a WHERE (z = 1) EXCEPT ALL SELECT * FROM b WHERE (z = 2) UNION SELECT * FROM a WHERE (z = 1) INTERSECT ALL SELECT * FROM b WHERE (z = 2) EXCEPT SELECT * FROM a WHERE (z = 1)"
+  end
+  
+  specify "should use a subselect when using a compound operation with a dataset that already has a compound operation" do
+    @a.union(@b.union(@a, true)).sql.should == \
+      "SELECT * FROM a WHERE (z = 1) UNION SELECT * FROM (SELECT * FROM b WHERE (z = 2) UNION ALL SELECT * FROM a WHERE (z = 1))"
+    @a.intersect(@b.intersect(@a), true).sql.should == \
+      "SELECT * FROM a WHERE (z = 1) INTERSECT ALL SELECT * FROM (SELECT * FROM b WHERE (z = 2) INTERSECT SELECT * FROM a WHERE (z = 1))"
+    @a.except(@b.except(@a, true)).sql.should == \
+      "SELECT * FROM a WHERE (z = 1) EXCEPT SELECT * FROM (SELECT * FROM b WHERE (z = 2) EXCEPT ALL SELECT * FROM a WHERE (z = 1))"
+    @a.union(@b.intersect(@a.except(@b, true)), true).union(@a.intersect(@b.except(@a), true)).sql.should == \
+      "SELECT * FROM a WHERE (z = 1) UNION ALL SELECT * FROM (SELECT * FROM b WHERE (z = 2) INTERSECT SELECT * FROM (SELECT * FROM a WHERE (z = 1) EXCEPT ALL SELECT * FROM b WHERE (z = 2))) UNION SELECT * FROM (SELECT * FROM a WHERE (z = 1) INTERSECT ALL SELECT * FROM (SELECT * FROM b WHERE (z = 2) EXCEPT SELECT * FROM a WHERE (z = 1)))"
+  end
+end
+
+context "Dataset#[]" do
+  before do
+    @c = Class.new(Sequel::Dataset) do
+      @@last_dataset = nil
+      
+      def self.last_dataset
+        @@last_dataset
+      end
+
+      def single_record
+        @@last_dataset = opts ? clone(opts) : self
+        {1 => 2, 3 => 4}
+      end
+    end
+    @d = @c.new(nil).from(:test)
+  end
+  
+  specify "should return a single record filtered according to the given conditions" do
+    @d[:name => 'didi'].should == {1 => 2, 3 => 4}
+    @c.last_dataset.literal(@c.last_dataset.opts[:where]).should == "(name = 'didi')"
+
+    @d[:id => 5..45].should == {1 => 2, 3 => 4}
+    @c.last_dataset.literal(@c.last_dataset.opts[:where]).should == "((id >= 5) AND (id <= 45))"
+  end
+end
+
+context "Dataset#single_record" do
+  before do
+    @c = Class.new(Sequel::Dataset) do
+      def fetch_rows(sql)
+        yield sql
+      end
+    end
+    @cc = Class.new(@c) do
+      def fetch_rows(sql); end
+    end
+    
+    @d = @c.new(nil).from(:test)
+    @e = @cc.new(nil).from(:test)
+  end
+  
+  specify "should call each with a limit of 1 and return the record" do
+    @d.single_record.should == 'SELECT * FROM test LIMIT 1'
+  end
+  
+  specify "should return nil if no record is present" do
+    @e.single_record.should be_nil
+  end
+end
+
+context "Dataset#single_value" do
+  before do
+    @c = Class.new(Sequel::Dataset) do
+      def fetch_rows(sql)
+        yield({1 => sql})
+      end
+    end
+    @cc = Class.new(@c) do
+      def fetch_rows(sql); end
+    end
+    
+    @d = @c.new(nil).from(:test)
+    @e = @cc.new(nil).from(:test)
+  end
+  
+  specify "should call each and return the first value of the first record" do
+    @d.single_value.should == 'SELECT * FROM test LIMIT 1'
+  end
+  
+  specify "should return nil if no records" do
+    @e.single_value.should be_nil
+  end
+  
+  it "should work on a graphed_dataset" do
+    @d.should_receive(:columns).twice.and_return([:a])
+    @d.graph(@d, [:a], :table_alias=>:test2).single_value.should == 'SELECT test.a, test2.a AS test2_a FROM test LEFT OUTER JOIN test AS test2 USING (a) LIMIT 1'
+  end
+end
+
+context "Dataset#get" do
+  before do
+    @c = Class.new(Sequel::Dataset) do
+      attr_reader :last_sql
+      
+      def fetch_rows(sql)
+        @last_sql = sql
+        yield(:name => sql)
+      end
+    end
+    
+    @d = @c.new(nil).from(:test)
+  end
+  
+  specify "should select the specified column and fetch its value" do
+    @d.get(:name).should == "SELECT name FROM test LIMIT 1"
+    @d.get(:abc).should == "SELECT abc FROM test LIMIT 1" # the first available value is returned always
+  end
+  
+  specify "should work with filters" do
+    @d.filter(:id => 1).get(:name).should == "SELECT name FROM test WHERE (id = 1) LIMIT 1"
+  end
+  
+  specify "should work with aliased fields" do
+    @d.get(:x__b.as(:name)).should == "SELECT x.b AS name FROM test LIMIT 1"
+  end
+  
+  specify "should accept a block that yields a virtual row" do
+    @d.get{|o| o.x__b.as(:name)}.should == "SELECT x.b AS name FROM test LIMIT 1"
+    @d.get{x(1).as(:name)}.should == "SELECT x(1) AS name FROM test LIMIT 1"
+  end
+  
+  specify "should raise an error if both a regular argument and block argument are used" do
+    proc{@d.get(:name){|o| o.x__b.as(:name)}}.should raise_error(Sequel::Error)
+  end
+end
+
+context "Dataset#set_row_proc" do
+  before do
+    @c = Class.new(Sequel::Dataset) do
+      def fetch_rows(sql, &block)
+        # yield a hash with kind as the 1 bit of a number
+        (1..10).each {|i| block.call({:kind => i[0]})}
+      end
+    end
+    @dataset = @c.new(nil).from(:items)
+  end
+  
+  specify "should cause dataset to pass all rows through the filter" do
+    @dataset.row_proc = proc{|h| h[:der] = h[:kind] + 2; h}
+    
+    rows = @dataset.all
+    rows.size.should == 10
+    
+    rows.each {|r| r[:der].should == (r[:kind] + 2)}
+  end
+  
+  specify "should be copied over when dataset is cloned" do
+    @dataset.row_proc = proc{|h| h[:der] = h[:kind] + 2; h}
+    
+    @dataset.filter(:a => 1).first.should == {:kind => 1, :der => 3}
+  end
+end
+
+context "Dataset#<<" do
+  before do
+    @d = Sequel::Dataset.new(nil)
+    @d.meta_def(:insert) do |*args|
+      1234567890
+    end
+  end
+  
+  specify "should call #insert" do
+    (@d << {:name => 1}).should == 1234567890
+  end
+end
+
+context "Dataset#columns" do
+  before do
+    @dataset = DummyDataset.new(nil).from(:items)
+    @dataset.meta_def(:columns=) {|c| @columns = c}
+    i = 'a' 
+    @dataset.meta_def(:each){@columns = select_sql + i; i = i.next}
+  end
+  
+  specify "should return the value of @columns if @columns is not nil" do
+    @dataset.columns = [:a, :b, :c]
+    @dataset.columns.should == [:a, :b, :c]
+  end
+  
+  specify "should attempt to get a single record and return @columns if @columns is nil" do
+    @dataset.columns = nil
+    @dataset.columns.should == 'SELECT * FROM items LIMIT 1a'
+    @dataset.opts[:from] = [:nana]
+    @dataset.columns.should == 'SELECT * FROM items LIMIT 1a'
+  end
+  
+  specify "should ignore any filters, orders, or DISTINCT clauses" do
+    @dataset.filter!(:b=>100).order!(:b).distinct!(:b)
+    @dataset.columns = nil
+    @dataset.columns.should == 'SELECT * FROM items LIMIT 1a'
+  end
+end
+
+context "Dataset#columns!" do
+  before do
+    @dataset = DummyDataset.new(nil).from(:items)
+    i = 'a' 
+    @dataset.meta_def(:each){@columns = select_sql + i; i = i.next}
+  end
+  
+  specify "should always attempt to get a record and return @columns" do
+    @dataset.columns!.should == 'SELECT * FROM items LIMIT 1a'
+    @dataset.columns!.should == 'SELECT * FROM items LIMIT 1b'
+    @dataset.opts[:from] = [:nana]
+    @dataset.columns!.should == 'SELECT * FROM nana LIMIT 1c'
+  end
+end
+
+context "Dataset#import" do
+  before do
+    @dbc = Class.new(Sequel::Database) do
+      attr_reader :sqls
+      
+      def execute(sql, opts={})
+        @sqls ||= []
+        @sqls << sql
+      end
+      alias execute_dui execute
+      
+      def transaction(opts={})
+        @sqls ||= []
+        @sqls << 'BEGIN'
+        yield
+        @sqls << 'COMMIT'
+      end
+    end
+    @db = @dbc.new
+    
+    @ds = Sequel::Dataset.new(@db).from(:items)
+    
+    @list = [{:name => 'abc'}, {:name => 'def'}, {:name => 'ghi'}]
+  end
+  
+  specify "should accept string keys as column names" do
+    @ds.import(['x', 'y'], [[1, 2], [3, 4]])
+    @db.sqls.should == [
+      'BEGIN',
+      "INSERT INTO items (x, y) VALUES (1, 2)",
+      "INSERT INTO items (x, y) VALUES (3, 4)",
+      'COMMIT'
+    ]
+  end
+
+  specify "should accept a columns array and a values array" do
+    @ds.import([:x, :y], [[1, 2], [3, 4]])
+    @db.sqls.should == [
+      'BEGIN',
+      "INSERT INTO items (x, y) VALUES (1, 2)",
+      "INSERT INTO items (x, y) VALUES (3, 4)",
+      'COMMIT'
+    ]
+  end
+
+  specify "should accept a columns array and a dataset" do
+    @ds2 = Sequel::Dataset.new(@db).from(:cats).filter(:purr => true).select(:a, :b)
+    
+    @ds.import([:x, :y], @ds2)
+    @db.sqls.should == [
+      'BEGIN',
+      "INSERT INTO items (x, y) VALUES (SELECT a, b FROM cats WHERE (purr IS TRUE))",
+      'COMMIT'
+    ]
+  end
+
+  specify "should accept a columns array and a values array with :commit_every option" do
+    @ds.import([:x, :y], [[1, 2], [3, 4], [5, 6]], :commit_every => 3)
+    @db.sqls.should == [
+      'BEGIN',
+      "INSERT INTO items (x, y) VALUES (1, 2)",
+      "INSERT INTO items (x, y) VALUES (3, 4)",
+      "INSERT INTO items (x, y) VALUES (5, 6)",
+      'COMMIT',
+    ]
+  end
+  specify "should accept a columns array and a values array with slice option" do
+    @ds.import([:x, :y], [[1, 2], [3, 4], [5, 6]], :slice => 2)
+    @db.sqls.should == [
+      'BEGIN',
+      "INSERT INTO items (x, y) VALUES (1, 2)",
+      "INSERT INTO items (x, y) VALUES (3, 4)",
+      'COMMIT',
+      'BEGIN',
+      "INSERT INTO items (x, y) VALUES (5, 6)",
+      'COMMIT'
+    ]
+  end
+end
+
+context "Dataset#multi_insert" do
+  before do
+    @dbc = Class.new do
+      attr_reader :sqls
+      
+      def execute(sql, opts={})
+        @sqls ||= []
+        @sqls << sql
+      end
+      alias execute_dui execute
+      
+      def transaction(opts={})
+        @sqls ||= []
+        @sqls << 'BEGIN'
+        yield
+        @sqls << 'COMMIT'
+      end
+    end
+    @db = @dbc.new
+    
+    @ds = Sequel::Dataset.new(@db).from(:items)
+    
+    @list = [{:name => 'abc'}, {:name => 'def'}, {:name => 'ghi'}]
+  end
+  
+  specify "should issue multiple insert statements inside a transaction" do
+    @ds.multi_insert(@list)
+    @db.sqls.should == [
+      'BEGIN',
+      "INSERT INTO items (name) VALUES ('abc')",
+      "INSERT INTO items (name) VALUES ('def')",
+      "INSERT INTO items (name) VALUES ('ghi')",
+      'COMMIT'
+    ]
+  end
+  
+  specify "should handle different formats for tables" do
+    @ds = @ds.from(:sch__tab)
+    @ds.multi_insert(@list)
+    @db.sqls.should == [
+      'BEGIN',
+      "INSERT INTO sch.tab (name) VALUES ('abc')",
+      "INSERT INTO sch.tab (name) VALUES ('def')",
+      "INSERT INTO sch.tab (name) VALUES ('ghi')",
+      'COMMIT'
+    ]
+    @db.sqls.clear
+
+    @ds = @ds.from(:tab.qualify(:sch))
+    @ds.multi_insert(@list)
+    @db.sqls.should == [
+      'BEGIN',
+      "INSERT INTO sch.tab (name) VALUES ('abc')",
+      "INSERT INTO sch.tab (name) VALUES ('def')",
+      "INSERT INTO sch.tab (name) VALUES ('ghi')",
+      'COMMIT'
+    ]
+    @db.sqls.clear
+    @ds = @ds.from(:sch__tab.identifier)
+    @ds.multi_insert(@list)
+    @db.sqls.should == [
+      'BEGIN',
+      "INSERT INTO sch__tab (name) VALUES ('abc')",
+      "INSERT INTO sch__tab (name) VALUES ('def')",
+      "INSERT INTO sch__tab (name) VALUES ('ghi')",
+      'COMMIT'
+    ]
+  end
+  
+  specify "should accept the :commit_every option for committing every x records" do
+    @ds.multi_insert(@list, :commit_every => 1)
+    @db.sqls.should == [
+      'BEGIN',
+      "INSERT INTO items (name) VALUES ('abc')",
+      'COMMIT',
+      'BEGIN',
+      "INSERT INTO items (name) VALUES ('def')",
+      'COMMIT',
+      'BEGIN',
+      "INSERT INTO items (name) VALUES ('ghi')",
+      'COMMIT'
+    ]
+  end
+
+  specify "should accept the :slice option for committing every x records" do
+    @ds.multi_insert(@list, :slice => 2)
+    @db.sqls.should == [
+      'BEGIN',
+      "INSERT INTO items (name) VALUES ('abc')",
+      "INSERT INTO items (name) VALUES ('def')",
+      'COMMIT',
+      'BEGIN',
+      "INSERT INTO items (name) VALUES ('ghi')",
+      'COMMIT'
+    ]
+  end
+  
+  specify "should accept string keys as column names" do
+    @ds.multi_insert([{'x'=>1, 'y'=>2}, {'x'=>3, 'y'=>4}])
+    @db.sqls.should == [
+      'BEGIN',
+      "INSERT INTO items (x, y) VALUES (1, 2)",
+      "INSERT INTO items (x, y) VALUES (3, 4)",
+      'COMMIT'
+    ]
+  end
+
+  specify "should not do anything if no hashes are provided" do
+    @ds.multi_insert([])
+    @db.sqls.should be_nil
+  end
+end
+
+context "Dataset" do
+  before do
+    @d = Sequel::Dataset.new(nil).from(:x)
+  end
+
+  specify "should support self-changing select!" do
+    @d.select!(:y)
+    @d.sql.should == "SELECT y FROM x"
+  end
+  
+  specify "should support self-changing from!" do
+    @d.from!(:y)
+    @d.sql.should == "SELECT * FROM y"
+  end
+
+  specify "should support self-changing order!" do
+    @d.order!(:y)
+    @d.sql.should == "SELECT * FROM x ORDER BY y"
+  end
+  
+  specify "should support self-changing filter!" do
+    @d.filter!(:y => 1)
+    @d.sql.should == "SELECT * FROM x WHERE (y = 1)"
+  end
+
+  specify "should support self-changing filter! with block" do
+    @d.filter!{:y.sql_number < 2}
+    @d.sql.should == "SELECT * FROM x WHERE (y < 2)"
+  end
+  
+  specify "should raise for ! methods that don't return a dataset" do
+    proc {@d.opts!}.should raise_error(NameError)
+  end
+  
+  specify "should raise for missing methods" do
+    proc {@d.xuyz}.should raise_error(NameError)
+    proc {@d.xyz!}.should raise_error(NameError)
+    proc {@d.xyz?}.should raise_error(NameError)
+  end
+  
+  specify "should support chaining of bang methods" do
+      @d.order!(:y)
+      @d.filter!(:y => 1)
+      @d.sql.should == "SELECT * FROM x WHERE (y = 1) ORDER BY y"
+  end
+end
+
+context "Dataset#to_csv" do
+  before do
+    @c = Class.new(Sequel::Dataset) do
+      attr_accessor :data
+      attr_accessor :columns
+      
+      def fetch_rows(sql, &block)
+        @data.each(&block)
+      end
+      
+      # naked should return self here because to_csv wants a naked result set.
+      def naked
+        self
+      end
+    end
+    
+    @ds = @c.new(nil).from(:items)
+    @ds.columns = [:a, :b, :c]
+    @ds.data = [ {:a=>1, :b=>2, :c=>3}, {:a=>4, :b=>5, :c=>6}, {:a=>7, :b=>8, :c=>9} ]
+  end
+  
+  specify "should format a CSV representation of the records" do
+    @ds.to_csv.should ==
+      "a, b, c\r\n1, 2, 3\r\n4, 5, 6\r\n7, 8, 9\r\n"
+  end
+
+  specify "should exclude column titles if so specified" do
+    @ds.to_csv(false).should ==
+      "1, 2, 3\r\n4, 5, 6\r\n7, 8, 9\r\n"
+  end
+end
+
+context "Dataset#update_sql" do
+  before do
+    @ds = Sequel::Dataset.new(nil).from(:items)
+  end
+  
+  specify "should accept strings" do
+    @ds.update_sql("a = b").should == "UPDATE items SET a = b"
+  end
+  
+  specify "should accept hash with string keys" do
+    @ds.update_sql('c' => 'd').should == "UPDATE items SET c = 'd'"
+  end
+
+  specify "should accept array subscript references" do
+    @ds.update_sql((:day.sql_subscript(1)) => 'd').should == "UPDATE items SET day[1] = 'd'"
+  end
+end
+
+context "Dataset#insert_sql" do
+  before do
+    @ds = Sequel::Dataset.new(nil).from(:items)
+  end
+  
+  specify "should accept hash with symbol keys" do
+    @ds.insert_sql(:c => 'd').should == "INSERT INTO items (c) VALUES ('d')"
+  end
+
+  specify "should accept hash with string keys" do
+    @ds.insert_sql('c' => 'd').should == "INSERT INTO items (c) VALUES ('d')"
+  end
+
+  specify "should accept array subscript references" do
+    @ds.insert_sql((:day.sql_subscript(1)) => 'd').should == "INSERT INTO items (day[1]) VALUES ('d')"
+  end
+end
+
+class DummyMummyDataset < Sequel::Dataset
+  def first
+    raise if @opts[:from] == [:a]
+    true
+  end
+end
+
+class DummyMummyDatabase < Sequel::Database
+  attr_reader :sqls
+  
+  def execute(sql)
+    @sqls ||= []
+    @sqls << sql
+  end
+  
+  def transaction; yield; end
+
+  def dataset
+    DummyMummyDataset.new(self)
+  end
+end
+
+context "Dataset#table_exists?" do
+  before do
+    @db = DummyMummyDatabase.new
+    @db.instance_variable_set(:@schemas, {:a=>[]})
+    @db2 = DummyMummyDatabase.new
+  end
+  
+  specify "should use the database schema if available" do
+    @db[:a].table_exists?.should be_true
+  end
+  
+  specify "should otherwise try to select the first record from the table's dataset" do
+    @db2[:a].table_exists?.should be_false
+    @db2[:b].table_exists?.should be_true
+  end
+  
+  specify "should raise Sequel::Error if dataset references more than one table" do
+    proc {@db.from(:a, :b).table_exists?}.should raise_error(Sequel::Error)
+  end
+
+  specify "should raise Sequel::Error if dataset is from a subquery" do
+    proc {@db.from(@db[:a]).table_exists?}.should raise_error(Sequel::Error)
+  end
+
+  specify "should raise Sequel::Error if dataset has fixed sql" do
+    proc {@db['select * from blah'].table_exists?}.should raise_error(Sequel::Error)
+  end
+end
+
+context "Dataset#inspect" do
+  before do
+    @ds = Sequel::Dataset.new(nil).from(:blah)
+  end
+  
+  specify "should include the class name and the corresponding SQL statement" do
+    @ds.inspect.should == '#<%s: %s>' % [@ds.class.to_s, @ds.sql.inspect]
+  end
+end
+
+context "Dataset#all" do
+  before do
+    @c = Class.new(Sequel::Dataset) do
+      def fetch_rows(sql, &block)
+        block.call({:x => 1, :y => 2})
+        block.call({:x => 3, :y => 4})
+        block.call(sql)
+      end
+    end
+    @dataset = @c.new(nil).from(:items)
+  end
+
+  specify "should return an array with all records" do
+    @dataset.all.should == [
+      {:x => 1, :y => 2},
+      {:x => 3, :y => 4},
+      "SELECT * FROM items"
+    ]
+  end
+  
+  specify "should iterate over the array if a block is given" do
+    a = []
+    
+    @dataset.all do |r|
+      a << (r.is_a?(Hash) ? r[:x] : r)
+    end
+    
+    a.should == [1, 3, "SELECT * FROM items"]
+  end
+end
+
+context "Dataset#grep" do
+  before do
+    @ds = Sequel::Dataset.new(nil).from(:posts)
+  end
+  
+  specify "should format a SQL filter correctly" do
+    @ds.grep(:title, 'ruby').sql.should ==
+      "SELECT * FROM posts WHERE ((title LIKE 'ruby'))"
+  end
+
+  specify "should support multiple columns" do
+    @ds.grep([:title, :body], 'ruby').sql.should ==
+      "SELECT * FROM posts WHERE ((title LIKE 'ruby') OR (body LIKE 'ruby'))"
+  end
+  
+  specify "should support multiple search terms" do
+    @ds.grep(:title, ['abc', 'def']).sql.should == 
+      "SELECT * FROM posts WHERE (((title LIKE 'abc') OR (title LIKE 'def')))"
+  end
+  
+  specify "should support multiple columns and search terms" do
+    @ds.grep([:title, :body], ['abc', 'def']).sql.should ==
+      "SELECT * FROM posts WHERE (((title LIKE 'abc') OR (title LIKE 'def')) OR ((body LIKE 'abc') OR (body LIKE 'def')))"
+  end
+  
+  specify "should support regexps though the database may not support it" do
+    @ds.grep(:title, /ruby/).sql.should ==
+      "SELECT * FROM posts WHERE ((title ~ 'ruby'))"
+
+    @ds.grep(:title, [/^ruby/, 'ruby']).sql.should ==
+      "SELECT * FROM posts WHERE (((title ~ '^ruby') OR (title LIKE 'ruby')))"
+  end
+
+  specify "should support searching against other columns" do
+    @ds.grep(:title, :body).sql.should ==
+      "SELECT * FROM posts WHERE ((title LIKE body))"
+  end
+end
+
+context "Dataset default #fetch_rows, #insert, #update, and #delete, #execute" do
+  before do
+    @db = Sequel::Database.new
+    @ds = @db[:items]
+  end
+
+  specify "#fetch_rows should raise a NotImplementedError" do
+    proc{@ds.fetch_rows(''){}}.should raise_error(NotImplementedError)
+  end
+
+  specify "#delete should execute delete SQL" do
+    @db.should_receive(:execute).once.with('DELETE FROM items', :server=>:default)
+    @ds.delete
+  end
+
+  specify "#insert should execute insert SQL" do
+    @db.should_receive(:execute).once.with('INSERT INTO items DEFAULT VALUES', :server=>:default)
+    @ds.insert([])
+  end
+
+  specify "#update should execute update SQL" do
+    @db.should_receive(:execute).once.with('UPDATE items SET number = 1', :server=>:default)
+    @ds.update(:number=>1)
+  end
+  
+  specify "#execute should execute the SQL on the database" do
+    @db.should_receive(:execute).once.with('SELECT 1', :server=>:read_only)
+    @ds.send(:execute, 'SELECT 1')
+  end
+end
+
+context "Dataset prepared statements and bound variables " do
+  before do
+    @db = Sequel::Database.new
+    @db.meta_def(:sqls){@sqls||=[]}
+    def @db.execute(sql, opts={})
+      sqls << sql
+    end
+    def @db.dataset
+      ds = super()
+      def ds.fetch_rows(sql, &block)
+        execute(sql)
+      end
+      ds
+    end
+    @ds = @db[:items]
+  end
+  
+  specify "#call should take a type and bind hash and interpolate it" do
+    @ds.filter(:num=>:$n).call(:select, :n=>1)
+    @ds.filter(:num=>:$n).call(:first, :n=>1)
+    @ds.filter(:num=>:$n).call(:delete, :n=>1)
+    @ds.filter(:num=>:$n).call(:update, {:n=>1, :n2=>2}, :num=>:$n2)
+    @ds.call(:insert, {:n=>1}, :num=>:$n)
+    @db.sqls.should == ['SELECT * FROM items WHERE (num = 1)',
+      'SELECT * FROM items WHERE (num = 1) LIMIT 1',
+      'DELETE FROM items WHERE (num = 1)',
+      'UPDATE items SET num = 2 WHERE (num = 1)',
+      'INSERT INTO items (num) VALUES (1)']
+  end
+    
+  specify "#prepare should take a type and name and store it in the database for later use with call" do
+    pss = []
+    pss << @ds.filter(:num=>:$n).prepare(:select, :sn)
+    pss << @ds.filter(:num=>:$n).prepare(:first, :fn)
+    pss << @ds.filter(:num=>:$n).prepare(:delete, :dn)
+    pss << @ds.filter(:num=>:$n).prepare(:update, :un, :num=>:$n2)
+    pss << @ds.prepare(:insert, :in, :num=>:$n)
+    @db.prepared_statements.keys.sort_by{|k| k.to_s}.should == [:dn, :fn, :in, :sn, :un]
+    [:sn, :fn, :dn, :un, :in].each_with_index{|x, i| @db.prepared_statements[x].should == pss[i]}
+    @db.call(:sn, :n=>1)
+    @db.call(:fn, :n=>1)
+    @db.call(:dn, :n=>1)
+    @db.call(:un, :n=>1, :n2=>2)
+    @db.call(:in, :n=>1)
+    @db.sqls.should == ['SELECT * FROM items WHERE (num = 1)',
+      'SELECT * FROM items WHERE (num = 1) LIMIT 1',
+      'DELETE FROM items WHERE (num = 1)',
+      'UPDATE items SET num = 2 WHERE (num = 1)',
+      'INSERT INTO items (num) VALUES (1)']
+  end
+    
+  specify "#inspect should indicate it is a prepared statement with the prepared SQL" do
+    @ds.filter(:num=>:$n).prepare(:select, :sn).inspect.should == \
+      '<Sequel::Dataset/PreparedStatement "SELECT * FROM items WHERE (num = $n)">'
+  end
+    
+  specify "should handle literal strings" do
+    @ds.filter("num = ?", :$n).call(:select, :n=>1)
+    @db.sqls.should == ['SELECT * FROM items WHERE (num = 1)']
+  end
+    
+  specify "should handle datasets using static sql and placeholders" do
+    @db["SELECT * FROM items WHERE (num = ?)", :$n].call(:select, :n=>1)
+    @db.sqls.should == ['SELECT * FROM items WHERE (num = 1)']
+  end
+    
+  specify "should handle subselects" do
+    @ds.filter(:$b).filter(:num=>@ds.select(:num).filter(:num=>:$n)).filter(:$c).call(:select, :n=>1, :b=>0, :c=>2)
+    @db.sqls.should == ['SELECT * FROM items WHERE ((0 AND (num IN (SELECT num FROM items WHERE (num = 1)))) AND 2)']
+  end
+    
+  specify "should handle subselects in subselects" do
+    @ds.filter(:$b).filter(:num=>@ds.select(:num).filter(:num=>@ds.select(:num).filter(:num=>:$n))).call(:select, :n=>1, :b=>0)
+    @db.sqls.should == ['SELECT * FROM items WHERE (0 AND (num IN (SELECT num FROM items WHERE (num IN (SELECT num FROM items WHERE (num = 1))))))']
+  end
+    
+  specify "should handle subselects with literal strings" do
+    @ds.filter(:$b).filter(:num=>@ds.select(:num).filter("num = ?", :$n)).call(:select, :n=>1, :b=>0)
+    @db.sqls.should == ['SELECT * FROM items WHERE (0 AND (num IN (SELECT num FROM items WHERE (num = 1))))']
+  end
+    
+  specify "should handle subselects with static sql and placeholders" do
+    @ds.filter(:$b).filter(:num=>@db["SELECT num FROM items WHERE (num = ?)", :$n]).call(:select, :n=>1, :b=>0)
+    @db.sqls.should == ['SELECT * FROM items WHERE (0 AND (num IN (SELECT num FROM items WHERE (num = 1))))']
+  end
+end
+
+context Sequel::Dataset::UnnumberedArgumentMapper do
+  before do
+    @db = Sequel::Database.new
+    @db.meta_def(:sqls){@sqls||=[]}
+    def @db.execute(sql, opts={})
+      sqls << [sql, *opts[:arguments]]
+    end
+    @ds = @db[:items].filter(:num=>:$n)
+    def @ds.fetch_rows(sql, &block)
+      execute(sql)
+    end
+    def @ds.execute(sql, opts={}, &block)
+      @db.execute(sql, {:arguments=>bind_arguments}.merge(opts))
+    end
+    @ps = @ds.prepare(:select, :sn)
+    @ps.extend(Sequel::Dataset::UnnumberedArgumentMapper)
+  end
+
+  specify "#inspect should show the actual SQL submitted to the database" do
+    @ps.inspect.should == '<Sequel::Dataset/PreparedStatement "SELECT * FROM items WHERE (num = ?)">'
+  end
+  
+  specify "should submitted the SQL to the database with placeholders and bind variables" do
+    @ps.call(:n=>1)
+    @db.sqls.should == [["SELECT * FROM items WHERE (num = ?)", 1]]
+  end
+end
+
+context "Sequel::Dataset#server" do
+  specify "should set the server to use for the dataset" do
+    @db = Sequel::Database.new
+    @ds = @db[:items].server(:s)
+    sqls = []
+    @db.meta_def(:execute) do |sql, opts|
+      sqls << [sql, opts[:server]]
+    end
+    def @ds.fetch_rows(sql, &block)
+      execute(sql)
+    end
+    @ds.all
+    @ds.server(:i).insert(:a=>1)
+    @ds.server(:d).delete
+    @ds.server(:u).update(:a=>:a+1)
+    sqls.should == [['SELECT * FROM items', :s],
+      ['INSERT INTO items (a) VALUES (1)', :i],
+      ['DELETE FROM items', :d],
+      ['UPDATE items SET a = (a + 1)', :u]]
+  end
+end
+
+context "Sequel::Dataset #set_defaults" do
+  before do
+    @ds = Sequel::Dataset.new(nil).from(:items).set_defaults(:x=>1)
+  end
+
+  specify "should set the default values for inserts" do
+    @ds.insert_sql.should == "INSERT INTO items (x) VALUES (1)"
+    @ds.insert_sql(:x=>2).should == "INSERT INTO items (x) VALUES (2)"
+    @ds.insert_sql(:y=>2).should =~ /INSERT INTO items \([xy], [xy]\) VALUES \([21], [21]\)/
+    @ds.set_defaults(:y=>2).insert_sql.should =~ /INSERT INTO items \([xy], [xy]\) VALUES \([21], [21]\)/
+    @ds.set_defaults(:x=>2).insert_sql.should == "INSERT INTO items (x) VALUES (2)"
+  end
+
+  specify "should set the default values for updates" do
+    @ds.update_sql.should == "UPDATE items SET x = 1"
+    @ds.update_sql(:x=>2).should == "UPDATE items SET x = 2"
+    @ds.update_sql(:y=>2).should =~ /UPDATE items SET (x = 1|y = 2), (x = 1|y = 2)/
+    @ds.set_defaults(:y=>2).update_sql.should =~ /UPDATE items SET (x = 1|y = 2), (x = 1|y = 2)/
+    @ds.set_defaults(:x=>2).update_sql.should == "UPDATE items SET x = 2"
+  end
+end
+
+context "Sequel::Dataset #set_overrides" do
+  before do
+    @ds = Sequel::Dataset.new(nil).from(:items).set_overrides(:x=>1)
+  end
+
+  specify "should override the given values for inserts" do
+    @ds.insert_sql.should == "INSERT INTO items (x) VALUES (1)"
+    @ds.insert_sql(:x=>2).should == "INSERT INTO items (x) VALUES (1)"
+    @ds.insert_sql(:y=>2).should =~ /INSERT INTO items \([xy], [xy]\) VALUES \([21], [21]\)/
+    @ds.set_overrides(:y=>2).insert_sql.should =~ /INSERT INTO items \([xy], [xy]\) VALUES \([21], [21]\)/
+    @ds.set_overrides(:x=>2).insert_sql.should == "INSERT INTO items (x) VALUES (1)"
+  end
+
+  specify "should override the given values for updates" do
+    @ds.update_sql.should == "UPDATE items SET x = 1"
+    @ds.update_sql(:x=>2).should == "UPDATE items SET x = 1"
+    @ds.update_sql(:y=>2).should =~ /UPDATE items SET (x = 1|y = 2), (x = 1|y = 2)/
+    @ds.set_overrides(:y=>2).update_sql.should =~ /UPDATE items SET (x = 1|y = 2), (x = 1|y = 2)/
+    @ds.set_overrides(:x=>2).update_sql.should == "UPDATE items SET x = 1"
+  end
+end
diff -pruN 1.5.1-1/spec/core/expression_filters_spec.rb 3.0.0-1/spec/core/expression_filters_spec.rb
--- 1.5.1-1/spec/core/expression_filters_spec.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/core/expression_filters_spec.rb	2009-04-15 18:10:24.000000000 +0100
@@ -0,0 +1,375 @@
+require File.join(File.dirname(__FILE__), 'spec_helper')
+
+context "Blockless Ruby Filters" do
+  before do
+    db = Sequel::Database.new
+    db.quote_identifiers = false
+    @d = db[:items]
+    def @d.l(*args, &block)
+      literal(filter_expr(*args, &block))
+    end
+    def @d.lit(*args)
+      literal(*args)
+    end
+  end
+  
+  it "should support boolean columns directly" do
+    @d.l(:x).should == 'x'
+  end
+  
+  it "should support NOT via Symbol#~" do
+    @d.l(~:x).should == 'NOT x'
+  end
+  
+  it "should support qualified columns" do
+    @d.l(:x__y).should == 'x.y'
+    @d.l(~:x__y).should == 'NOT x.y'
+  end
+
+  it "should support NOT with SQL functions" do
+    @d.l(~:is_blah.sql_function).should == 'NOT is_blah()'
+    @d.l(~:is_blah.sql_function(:x)).should == 'NOT is_blah(x)'
+    @d.l(~:is_blah.sql_function(:x__y)).should == 'NOT is_blah(x.y)'
+    @d.l(~:is_blah.sql_function(:x, :x__y)).should == 'NOT is_blah(x, x.y)'
+  end
+
+  it "should handle multiple ~" do
+    @d.l(~~:x).should == 'x'
+    @d.l(~~~:x).should == 'NOT x'
+    @d.l(~~(:x & :y)).should == '(x AND y)'
+    @d.l(~~(:x | :y)).should == '(x OR y)'
+  end
+
+  it "should support = via Hash" do
+    @d.l(:x => 100).should == '(x = 100)'
+    @d.l(:x => 'a').should == '(x = \'a\')'
+    @d.l(:x => true).should == '(x IS TRUE)'
+    @d.l(:x => false).should == '(x IS FALSE)'
+    @d.l(:x => nil).should == '(x IS NULL)'
+    @d.l(:x => [1,2,3]).should == '(x IN (1, 2, 3))'
+  end
+  
+  it "should support != via Hash#~" do
+    @d.l(~{:x => 100}).should == '(x != 100)'
+    @d.l(~{:x => 'a'}).should == '(x != \'a\')'
+    @d.l(~{:x => true}).should == '(x IS NOT TRUE)'
+    @d.l(~{:x => false}).should == '(x IS NOT FALSE)'
+    @d.l(~{:x => nil}).should == '(x IS NOT NULL)'
+  end
+  
+  it "should support ~ via Hash and Regexp (if supported by database)" do
+    @d.l(:x => /blah/).should == '(x ~ \'blah\')'
+  end
+  
+  it "should support !~ via Hash#~ and Regexp" do
+    @d.l(~{:x => /blah/}).should == '(x !~ \'blah\')'
+  end
+  
+  it "should support LIKE via Symbol#like" do
+    @d.l(:x.like('a')).should == '(x LIKE \'a\')'
+    @d.l(:x.like(/a/)).should == '(x ~ \'a\')'
+    @d.l(:x.like('a', 'b')).should == '((x LIKE \'a\') OR (x LIKE \'b\'))'
+    @d.l(:x.like(/a/, /b/i)).should == '((x ~ \'a\') OR (x ~* \'b\'))'
+    @d.l(:x.like('a', /b/)).should == '((x LIKE \'a\') OR (x ~ \'b\'))'
+  end
+
+  it "should support NOT LIKE via Symbol#like and Symbol#~" do
+    @d.l(~:x.like('a')).should == '(x NOT LIKE \'a\')'
+    @d.l(~:x.like(/a/)).should == '(x !~ \'a\')'
+    @d.l(~:x.like('a', 'b')).should == '((x NOT LIKE \'a\') AND (x NOT LIKE \'b\'))'
+    @d.l(~:x.like(/a/, /b/i)).should == '((x !~ \'a\') AND (x !~* \'b\'))'
+    @d.l(~:x.like('a', /b/)).should == '((x NOT LIKE \'a\') AND (x !~ \'b\'))'
+  end
+
+  it "should support ILIKE via Symbol#ilike" do
+    @d.l(:x.ilike('a')).should == '(x ILIKE \'a\')'
+    @d.l(:x.ilike(/a/)).should == '(x ~* \'a\')'
+    @d.l(:x.ilike('a', 'b')).should == '((x ILIKE \'a\') OR (x ILIKE \'b\'))'
+    @d.l(:x.ilike(/a/, /b/i)).should == '((x ~* \'a\') OR (x ~* \'b\'))'
+    @d.l(:x.ilike('a', /b/)).should == '((x ILIKE \'a\') OR (x ~* \'b\'))'
+  end
+
+  it "should support NOT ILIKE via Symbol#ilike and Symbol#~" do
+    @d.l(~:x.ilike('a')).should == '(x NOT ILIKE \'a\')'
+    @d.l(~:x.ilike(/a/)).should == '(x !~* \'a\')'
+    @d.l(~:x.ilike('a', 'b')).should == '((x NOT ILIKE \'a\') AND (x NOT ILIKE \'b\'))'
+    @d.l(~:x.ilike(/a/, /b/i)).should == '((x !~* \'a\') AND (x !~* \'b\'))'
+    @d.l(~:x.ilike('a', /b/)).should == '((x NOT ILIKE \'a\') AND (x !~* \'b\'))'
+  end
+
+  it "should support negating ranges via Hash#~ and Range" do
+    @d.l(~{:x => 1..5}).should == '((x < 1) OR (x > 5))'
+    @d.l(~{:x => 1...5}).should == '((x < 1) OR (x >= 5))'
+  end
+  
+  it "should support negating NOT IN via Hash#~ and Dataset or Array" do
+    @d.l(~{:x => @d.select(:i)}).should == '(x NOT IN (SELECT i FROM items))'
+    @d.l(~{:x => [1,2,3]}).should == '(x NOT IN (1, 2, 3))'
+  end
+
+  it "should support + - * / via Symbol#+,-,*,/" do
+    @d.l(:x + 1 > 100).should == '((x + 1) > 100)'
+    @d.l((:x * :y) < 100.01).should == '((x * y) < 100.01)'
+    @d.l((:x - :y/2) >= 100000000000000000000000000000000000).should == '((x - (y / 2)) >= 100000000000000000000000000000000000)'
+    @d.l((((:x - :y)/(:x + :y))*:z) <= 100).should == '((((x - y) / (x + y)) * z) <= 100)'
+    @d.l(~((((:x - :y)/(:x + :y))*:z) <= 100)).should == '((((x - y) / (x + y)) * z) > 100)'
+  end
+  
+  it "should not allow negation of string expressions" do
+    proc{~:x.sql_string}.should raise_error
+    proc{~([:x, :y].sql_string_join)}.should raise_error
+  end
+
+  it "should not allow mathematical or string operations on true, false, or nil" do
+    proc{:x + 1}.should_not raise_error
+    proc{:x - true}.should raise_error(Sequel::Error)
+    proc{:x / false}.should raise_error(Sequel::Error)
+    proc{:x * nil}.should raise_error(Sequel::Error)
+    proc{[:x, nil].sql_string_join}.should raise_error(Sequel::Error)
+  end
+
+  it "should not allow mathematical or string operations on boolean complex expressions" do
+    proc{:x + (:y + 1)}.should_not raise_error
+    proc{:x - (~:y)}.should raise_error(Sequel::Error)
+    proc{:x / (:y & :z)}.should raise_error(Sequel::Error)
+    proc{:x * (:y | :z)}.should raise_error(Sequel::Error)
+    proc{:x + :y.like('a')}.should raise_error(Sequel::Error)
+    proc{:x - :y.like(/a/)}.should raise_error(Sequel::Error)
+    proc{:x * :y.like(/a/i)}.should raise_error(Sequel::Error)
+    proc{:x + ~:y.like('a')}.should raise_error(Sequel::Error)
+    proc{:x - ~:y.like(/a/)}.should raise_error(Sequel::Error)
+    proc{:x * ~:y.like(/a/i)}.should raise_error(Sequel::Error)
+    proc{[:x, ~:y.like(/a/i)].sql_string_join}.should raise_error(Sequel::Error)
+  end
+
+  it "should support AND conditions via &" do
+    @d.l(:x & :y).should == '(x AND y)'
+    @d.l(:x.sql_boolean & :y).should == '(x AND y)'
+    @d.l(:x & :y & :z).should == '((x AND y) AND z)'
+    @d.l(:x & {:y => :z}).should == '(x AND (y = z))'
+    @d.l({:y => :z} & :x).should == '((y = z) AND x)'
+    @d.l({:x => :a} & {:y => :z}).should == '((x = a) AND (y = z))'
+    @d.l((:x + 200 < 0) & (:y - 200 < 0)).should == '(((x + 200) < 0) AND ((y - 200) < 0))'
+    @d.l(:x & ~:y).should == '(x AND NOT y)'
+    @d.l(~:x & :y).should == '(NOT x AND y)'
+    @d.l(~:x & ~:y).should == '(NOT x AND NOT y)'
+  end
+  
+  it "should support OR conditions via |" do
+    @d.l(:x | :y).should == '(x OR y)'
+    @d.l(:x.sql_boolean | :y).should == '(x OR y)'
+    @d.l(:x | :y | :z).should == '((x OR y) OR z)'
+    @d.l(:x | {:y => :z}).should == '(x OR (y = z))'
+    @d.l({:y => :z} | :x).should == '((y = z) OR x)'
+    @d.l({:x => :a} | {:y => :z}).should == '((x = a) OR (y = z))'
+    @d.l((:x.sql_number > 200) | (:y.sql_number < 200)).should == '((x > 200) OR (y < 200))'
+  end
+  
+  it "should support & | combinations" do
+    @d.l((:x | :y) & :z).should == '((x OR y) AND z)'
+    @d.l(:x | (:y & :z)).should == '(x OR (y AND z))'
+    @d.l((:x & :w) | (:y & :z)).should == '((x AND w) OR (y AND z))'
+  end
+  
+  it "should support & | with ~" do
+    @d.l(~((:x | :y) & :z)).should == '((NOT x AND NOT y) OR NOT z)'
+    @d.l(~(:x | (:y & :z))).should == '(NOT x AND (NOT y OR NOT z))'
+    @d.l(~((:x & :w) | (:y & :z))).should == '((NOT x OR NOT w) AND (NOT y OR NOT z))'
+    @d.l(~((:x.sql_number > 200) | (:y & :z))).should == '((x <= 200) AND (NOT y OR NOT z))'
+  end
+  
+  it "should support LiteralString" do
+    @d.l('x'.lit).should == '(x)'
+    @d.l(~'x'.lit).should == 'NOT x'
+    @d.l(~~'x'.lit).should == 'x'
+    @d.l(~(('x'.lit | :y) & :z)).should == '((NOT x AND NOT y) OR NOT z)'
+    @d.l(~(:x | 'y'.lit)).should == '(NOT x AND NOT y)'
+    @d.l(~('x'.lit & 'y'.lit)).should == '(NOT x OR NOT y)'
+    @d.l({'y'.lit => 'z'.lit} & 'x'.lit).should == '((y = z) AND x)'
+    @d.l(('x'.lit > 200) & ('y'.lit < 200)).should == '((x > 200) AND (y < 200))'
+    @d.l(~('x'.lit + 1 > 100)).should == '((x + 1) <= 100)'
+    @d.l('x'.lit.like(/a/)).should == '(x ~ \'a\')'
+    @d.l('x'.lit + 1 > 100).should == '((x + 1) > 100)'
+    @d.l(('x'.lit * :y) < 100.01).should == '((x * y) < 100.01)'
+    @d.l(('x'.lit - :y/2) >= 100000000000000000000000000000000000).should == '((x - (y / 2)) >= 100000000000000000000000000000000000)'
+    @d.l(('z'.lit * (('x'.lit / :y)/(:x + :y))) <= 100).should == '((z * ((x / y) / (x + y))) <= 100)'
+    @d.l(~(((('x'.lit - :y)/(:x + :y))*:z) <= 100)).should == '((((x - y) / (x + y)) * z) > 100)'
+  end
+
+  it "should support hashes by ANDing the conditions" do
+    @d.l(:x => 100, :y => 'a')[1...-1].split(' AND ').sort.should == ['(x = 100)', '(y = \'a\')']
+    @d.l(:x => true, :y => false)[1...-1].split(' AND ').sort.should == ['(x IS TRUE)', '(y IS FALSE)']
+    @d.l(:x => nil, :y => [1,2,3])[1...-1].split(' AND ').sort.should == ['(x IS NULL)', '(y IN (1, 2, 3))']
+  end
+  
+  it "should support sql_expr on hashes" do
+    @d.l({:x => 100, :y => 'a'}.sql_expr)[1...-1].split(' AND ').sort.should == ['(x = 100)', '(y = \'a\')']
+    @d.l({:x => true, :y => false}.sql_expr)[1...-1].split(' AND ').sort.should == ['(x IS TRUE)', '(y IS FALSE)']
+    @d.l({:x => nil, :y => [1,2,3]}.sql_expr)[1...-1].split(' AND ').sort.should == ['(x IS NULL)', '(y IN (1, 2, 3))']
+  end
+  
+  it "should support sql_negate on hashes" do
+    @d.l({:x => 100, :y => 'a'}.sql_negate)[1...-1].split(' AND ').sort.should == ['(x != 100)', '(y != \'a\')']
+    @d.l({:x => true, :y => false}.sql_negate)[1...-1].split(' AND ').sort.should == ['(x IS NOT TRUE)', '(y IS NOT FALSE)']
+    @d.l({:x => nil, :y => [1,2,3]}.sql_negate)[1...-1].split(' AND ').sort.should == ['(x IS NOT NULL)', '(y NOT IN (1, 2, 3))']
+  end
+  
+  it "should support ~ on hashes" do
+    @d.l(~{:x => 100, :y => 'a'})[1...-1].split(' OR ').sort.should == ['(x != 100)', '(y != \'a\')']
+    @d.l(~{:x => true, :y => false})[1...-1].split(' OR ').sort.should == ['(x IS NOT TRUE)', '(y IS NOT FALSE)']
+    @d.l(~{:x => nil, :y => [1,2,3]})[1...-1].split(' OR ').sort.should == ['(x IS NOT NULL)', '(y NOT IN (1, 2, 3))']
+  end
+  
+  it "should support sql_or on hashes" do
+    @d.l({:x => 100, :y => 'a'}.sql_or)[1...-1].split(' OR ').sort.should == ['(x = 100)', '(y = \'a\')']
+    @d.l({:x => true, :y => false}.sql_or)[1...-1].split(' OR ').sort.should == ['(x IS TRUE)', '(y IS FALSE)']
+    @d.l({:x => nil, :y => [1,2,3]}.sql_or)[1...-1].split(' OR ').sort.should == ['(x IS NULL)', '(y IN (1, 2, 3))']
+  end
+  
+  it "should support arrays with all two pairs the same as hashes" do
+    @d.l([[:x, 100],[:y, 'a']]).should == '((x = 100) AND (y = \'a\'))'
+    @d.l([[:x, true], [:y, false]]).should == '((x IS TRUE) AND (y IS FALSE))'
+    @d.l([[:x, nil], [:y, [1,2,3]]]).should == '((x IS NULL) AND (y IN (1, 2, 3)))'
+  end
+  
+  it "should support sql_expr on arrays with all two pairs" do
+    @d.l([[:x, 100],[:y, 'a']].sql_expr).should == '((x = 100) AND (y = \'a\'))'
+    @d.l([[:x, true], [:y, false]].sql_expr).should == '((x IS TRUE) AND (y IS FALSE))'
+    @d.l([[:x, nil], [:y, [1,2,3]]].sql_expr).should == '((x IS NULL) AND (y IN (1, 2, 3)))'
+  end
+  
+  it "should support sql_negate on arrays with all two pairs" do
+    @d.l([[:x, 100],[:y, 'a']].sql_negate).should == '((x != 100) AND (y != \'a\'))'
+    @d.l([[:x, true], [:y, false]].sql_negate).should == '((x IS NOT TRUE) AND (y IS NOT FALSE))'
+    @d.l([[:x, nil], [:y, [1,2,3]]].sql_negate).should == '((x IS NOT NULL) AND (y NOT IN (1, 2, 3)))'
+  end
+  
+  it "should support ~ on arrays with all two pairs" do
+    @d.l(~[[:x, 100],[:y, 'a']]).should == '((x != 100) OR (y != \'a\'))'
+    @d.l(~[[:x, true], [:y, false]]).should == '((x IS NOT TRUE) OR (y IS NOT FALSE))'
+    @d.l(~[[:x, nil], [:y, [1,2,3]]]).should == '((x IS NOT NULL) OR (y NOT IN (1, 2, 3)))'
+  end
+  
+  it "should support sql_or on arrays with all two pairs" do
+    @d.l([[:x, 100],[:y, 'a']].sql_or).should == '((x = 100) OR (y = \'a\'))'
+    @d.l([[:x, true], [:y, false]].sql_or).should == '((x IS TRUE) OR (y IS FALSE))'
+    @d.l([[:x, nil], [:y, [1,2,3]]].sql_or).should == '((x IS NULL) OR (y IN (1, 2, 3)))'
+  end
+  
+  it "should support Array#sql_string_join for concatenation of SQL strings" do
+    @d.lit([:x].sql_string_join).should == '(x)'
+    @d.lit([:x].sql_string_join(', ')).should == '(x)'
+    @d.lit([:x, :y].sql_string_join).should == '(x || y)'
+    @d.lit([:x, :y].sql_string_join(', ')).should == "(x || ', ' || y)"
+    @d.lit([:x.sql_function(1), :y.sql_subscript(1)].sql_string_join).should == '(x(1) || y[1])'
+    @d.lit([:x.sql_function(1), 'y.z'.lit].sql_string_join(', ')).should == "(x(1) || ', ' || y.z)"
+    @d.lit([:x, 1, :y].sql_string_join).should == "(x || '1' || y)"
+    @d.lit([:x, 1, :y].sql_string_join(', ')).should == "(x || ', ' || '1' || ', ' || y)"
+    @d.lit([:x, 1, :y].sql_string_join(:y__z)).should == "(x || y.z || '1' || y.z || y)"
+    @d.lit([:x, 1, :y].sql_string_join(1)).should == "(x || '1' || '1' || '1' || y)"
+    @d.lit([:x, :y].sql_string_join('y.x || x.y'.lit)).should == "(x || y.x || x.y || y)"
+    @d.lit([[:x, :y].sql_string_join, [:a, :b].sql_string_join].sql_string_join).should == "((x || y) || (a || b))"
+  end
+
+  it "should support StringExpression#+ for concatenation of SQL strings" do
+    @d.lit(:x.sql_string + :y).should == '(x || y)'
+    @d.lit([:x].sql_string_join + :y).should == '((x) || y)'
+    @d.lit([:x, :z].sql_string_join(' ') + :y).should == "((x || ' ' || z) || y)"
+  end
+
+  it "should be supported inside blocks" do
+    @d.l{[[:x, nil], [:y, [1,2,3]]].sql_or}.should == '((x IS NULL) OR (y IN (1, 2, 3)))'
+    @d.l{~[[:x, nil], [:y, [1,2,3]]]}.should == '((x IS NOT NULL) OR (y NOT IN (1, 2, 3)))'
+    @d.l{~(((('x'.lit - :y)/(:x + :y))*:z) <= 100)}.should == '((((x - y) / (x + y)) * z) > 100)'
+    @d.l{{:x => :a} & {:y => :z}}.should == '((x = a) AND (y = z))'
+  end
+
+  it "should support &, |, ^, ~, <<, and >> for NumericExpressions" do
+    @d.l(:x.sql_number & 1 > 100).should == '((x & 1) > 100)'
+    @d.l(:x.sql_number | 1 > 100).should == '((x | 1) > 100)'
+    @d.l(:x.sql_number ^ 1 > 100).should == '((x ^ 1) > 100)'
+    @d.l(~:x.sql_number > 100).should == '(~x > 100)'
+    @d.l(:x.sql_number << 1 > 100).should == '((x << 1) > 100)'
+    @d.l(:x.sql_number >> 1 > 100).should == '((x >> 1) > 100)'
+    @d.l((:x + 1) & 1 > 100).should == '(((x + 1) & 1) > 100)'
+    @d.l((:x + 1) | 1 > 100).should == '(((x + 1) | 1) > 100)'
+    @d.l((:x + 1) ^ 1 > 100).should == '(((x + 1) ^ 1) > 100)'
+    @d.l(~(:x + 1) > 100).should == '(~(x + 1) > 100)'
+    @d.l((:x + 1) << 1 > 100).should == '(((x + 1) << 1) > 100)'
+    @d.l((:x + 1) >> 1 > 100).should == '(((x + 1) >> 1) > 100)'
+    @d.l((:x + 1) & (:x + 2) > 100).should == '(((x + 1) & (x + 2)) > 100)'
+  end
+
+  it "should raise an error if use a Bitwise method on a ComplexExpression that isn't a NumericExpression" do
+    proc{(:x + 1) & (:x & 2)}.should raise_error(Sequel::Error)
+  end
+
+  it "should raise an error if use a Boolean method on a ComplexExpression that isn't a BooleanExpression" do
+    proc{:x & (:x + 2)}.should raise_error(Sequel::Error)
+  end
+
+  it "should raise an error if attempting to invert a ComplexExpression that isn't a BooleanExpression" do
+    proc{Sequel::SQL::BooleanExpression.invert(:x + 2)}.should raise_error(Sequel::Error)
+  end
+
+  it "should return self on .lit" do
+    y = :x + 1
+    y.lit.should == y
+  end
+
+  it "should raise an error if trying to create an invalid complex expression" do
+    proc{Sequel::SQL::ComplexExpression.new(:BANG, 1, 2)}.should raise_error(Sequel::Error)
+  end
+
+  it "should raise an error if trying to literalize an invalid complex expression" do
+    ce = :x + 1
+    ce.instance_variable_set(:@op, :BANG)
+    proc{@d.lit(ce)}.should raise_error(Sequel::Error)
+  end
+
+  it "should support equality comparison of two expressions" do
+    e1 = ~:comment.like('%:hidden:%')
+    e2 = ~:comment.like('%:hidden:%')
+    e1.should == e2
+  end
+  
+  if RUBY_VERSION < '1.9.0'
+    it "should not allow inequality operations on true, false, or nil" do
+      proc{:x > 1}.should_not raise_error
+      proc{:x < true}.should raise_error(Sequel::Error)
+      proc{:x >= false}.should raise_error(Sequel::Error)
+      proc{:x <= nil}.should raise_error(Sequel::Error)
+    end
+
+    it "should not allow inequality operations on boolean complex expressions" do
+      proc{:x > (:y > 5)}.should raise_error(Sequel::Error)
+      proc{:x < (:y < 5)}.should raise_error(Sequel::Error)
+      proc{:x >= (:y >= 5)}.should raise_error(Sequel::Error)
+      proc{:x <= (:y <= 5)}.should raise_error(Sequel::Error)
+      proc{:x > {:y => nil}}.should raise_error(Sequel::Error)
+      proc{:x < ~{:y => nil}}.should raise_error(Sequel::Error)
+      proc{:x >= {:y => 5}}.should raise_error(Sequel::Error)
+      proc{:x <= ~{:y => 5}}.should raise_error(Sequel::Error)
+      proc{:x >= {:y => [1,2,3]}}.should raise_error(Sequel::Error)
+      proc{:x <= ~{:y => [1,2,3]}}.should raise_error(Sequel::Error)
+    end
+    
+    it "should support >, <, >=, and <= via Symbol#>,<,>=,<=" do
+      @d.l(:x > 100).should == '(x > 100)'
+      @d.l(:x < 100.01).should == '(x < 100.01)'
+      @d.l(:x >= 100000000000000000000000000000000000).should == '(x >= 100000000000000000000000000000000000)'
+      @d.l(:x <= 100).should == '(x <= 100)'
+    end
+    
+    it "should support negation of >, <, >=, and <= via Symbol#~" do
+      @d.l(~(:x > 100)).should == '(x <= 100)'
+      @d.l(~(:x < 100.01)).should == '(x >= 100.01)'
+      @d.l(~(:x >= 100000000000000000000000000000000000)).should == '(x < 100000000000000000000000000000000000)'
+      @d.l(~(:x <= 100)).should == '(x > 100)'
+    end
+    
+    it "should support double negation via ~" do
+      @d.l(~~(:x > 100)).should == '(x > 100)'
+    end
+  end
+end
diff -pruN 1.5.1-1/spec/core/object_graph_spec.rb 3.0.0-1/spec/core/object_graph_spec.rb
--- 1.5.1-1/spec/core/object_graph_spec.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/core/object_graph_spec.rb	2009-03-26 16:21:05.000000000 +0000
@@ -0,0 +1,272 @@
+require File.join(File.dirname(__FILE__), 'spec_helper')
+
+describe Sequel::Dataset, " graphing" do
+  before do
+    dbc = Class.new
+    @db = dbc.new
+    @ds1 = Sequel::Dataset.new(@db).from(:points)
+    @ds2 = Sequel::Dataset.new(@db).from(:lines)
+    @ds3 = Sequel::Dataset.new(@db).from(:graphs)
+    dss = {:points=>@ds1, :lines=>@ds2, :graphs=>@ds3}
+    dbc.send(:define_method, :[]){|ds| dss[ds]} 
+    def @ds1.columns; [:id, :x, :y] end
+    def @ds2.columns; [:id, :x, :y, :graph_id] end
+    def @ds3.columns; [:id, :name, :x, :y, :lines_x] end
+  end
+
+  it "#graph should not modify the current dataset's opts" do
+    o1 = @ds1.opts
+    o2 = o1.dup
+    ds1 = @ds1.graph(@ds2, :x=>:id)
+    @ds1.opts.should == o1
+    @ds1.opts.should == o2
+    ds1.opts.should_not == o1
+  end
+
+  it "#graph should accept a dataset as the dataset" do
+    ds = @ds1.graph(@ds2, :x=>:id)
+    ds.sql.should == 'SELECT points.id, points.x, points.y, lines.id AS lines_id, lines.x AS lines_x, lines.y AS lines_y, lines.graph_id FROM points LEFT OUTER JOIN lines ON (lines.x = points.id)'
+  end
+
+  it "#graph should accept a symbol table name as the dataset" do
+    ds = @ds1.graph(:lines, :x=>:id)
+    ds.sql.should == 'SELECT points.id, points.x, points.y, lines.id AS lines_id, lines.x AS lines_x, lines.y AS lines_y, lines.graph_id FROM points LEFT OUTER JOIN lines ON (lines.x = points.id)'
+  end
+
+  it "#graph should accept an object that responds to dataset as the dataset" do
+    oc = Class.new
+    o = oc.new
+    ds = @ds2
+    oc.send(:define_method, :dataset){ds} 
+    ds = @ds1.graph(o, :x=>:id)
+    ds.sql.should == 'SELECT points.id, points.x, points.y, lines.id AS lines_id, lines.x AS lines_x, lines.y AS lines_y, lines.graph_id FROM points LEFT OUTER JOIN lines ON (lines.x = points.id)'
+    ds = :lines
+    oc.send(:define_method, :dataset){ds} 
+    ds = @ds1.graph(o, :x=>:id)
+    ds.sql.should == 'SELECT points.id, points.x, points.y, lines.id AS lines_id, lines.x AS lines_x, lines.y AS lines_y, lines.graph_id FROM points LEFT OUTER JOIN lines ON (lines.x = points.id)'
+  end
+
+  it "#graph should raise an error if a symbol, dataset, or model is not used" do
+    proc{@ds1.graph(Object.new, :x=>:id)}.should raise_error(Sequel::Error)
+  end
+
+  it "#graph should accept a :table_alias option" do
+    ds = @ds1.graph(:lines, {:x=>:id}, :table_alias=>:planes)
+    ds.sql.should == 'SELECT points.id, points.x, points.y, planes.id AS planes_id, planes.x AS planes_x, planes.y AS planes_y, planes.graph_id FROM points LEFT OUTER JOIN lines AS planes ON (planes.x = points.id)'
+  end
+
+  it "#graph should accept a :implicit_qualifier option" do
+    ds = @ds1.graph(:lines, {:x=>:id}, :implicit_qualifier=>:planes)
+    ds.sql.should == 'SELECT points.id, points.x, points.y, lines.id AS lines_id, lines.x AS lines_x, lines.y AS lines_y, lines.graph_id FROM points LEFT OUTER JOIN lines ON (lines.x = planes.id)'
+  end
+
+  it "#graph should accept a :join_type option" do
+    ds = @ds1.graph(:lines, {:x=>:id}, :join_type=>:inner)
+    ds.sql.should == 'SELECT points.id, points.x, points.y, lines.id AS lines_id, lines.x AS lines_x, lines.y AS lines_y, lines.graph_id FROM points INNER JOIN lines ON (lines.x = points.id)'
+  end
+
+  it "#graph should not select any columns from the graphed table if :select option is false" do
+    ds = @ds1.graph(:lines, {:x=>:id}, :select=>false).graph(:graphs, :id=>:graph_id)
+    ds.sql.should == 'SELECT points.id, points.x, points.y, graphs.id AS graphs_id, graphs.name, graphs.x AS graphs_x, graphs.y AS graphs_y, graphs.lines_x FROM points LEFT OUTER JOIN lines ON (lines.x = points.id) LEFT OUTER JOIN graphs ON (graphs.id = lines.graph_id)'
+  end
+
+  it "#graph should use the given columns if :select option is used" do
+    ds = @ds1.graph(:lines, {:x=>:id}, :select=>[:x, :graph_id]).graph(:graphs, :id=>:graph_id)
+    ds.sql.should == 'SELECT points.id, points.x, points.y, lines.x AS lines_x, lines.graph_id, graphs.id AS graphs_id, graphs.name, graphs.x AS graphs_x, graphs.y AS graphs_y, graphs.lines_x AS graphs_lines_x FROM points LEFT OUTER JOIN lines ON (lines.x = points.id) LEFT OUTER JOIN graphs ON (graphs.id = lines.graph_id)'
+  end
+
+  it "#graph should pass all join_conditions to join_table" do
+    ds = @ds1.graph(@ds2, [[:x, :id], [:y, :id]])
+    ds.sql.should == 'SELECT points.id, points.x, points.y, lines.id AS lines_id, lines.x AS lines_x, lines.y AS lines_y, lines.graph_id FROM points LEFT OUTER JOIN lines ON ((lines.x = points.id) AND (lines.y = points.id))'
+  end
+
+  it "#graph should accept a block instead of conditions and pass it to join_table" do
+    ds = @ds1.graph(@ds2){|ja, lja, js| [[:x.qualify(ja), :id.qualify(lja)], [:y.qualify(ja), :id.qualify(lja)]]}
+    ds.sql.should == 'SELECT points.id, points.x, points.y, lines.id AS lines_id, lines.x AS lines_x, lines.y AS lines_y, lines.graph_id FROM points LEFT OUTER JOIN lines ON ((lines.x = points.id) AND (lines.y = points.id))'
+  end
+
+  it "#graph should not add columns if graph is called after set_graph_aliases" do
+    ds = @ds1.set_graph_aliases([[:x,[:points, :x]], [:y,[:lines, :y]]])
+    ds.sql.should == 'SELECT points.x, lines.y FROM points'
+    ds = ds.graph(:lines, :x=>:id)
+    ds.sql.should == 'SELECT points.x, lines.y FROM points LEFT OUTER JOIN lines ON (lines.x = points.id)'
+  end
+
+  it "#graph should allow graphing of multiple datasets" do
+    ds = @ds1.graph(@ds2, :x=>:id).graph(@ds3, :id=>:graph_id)
+    ds.sql.should == 'SELECT points.id, points.x, points.y, lines.id AS lines_id, lines.x AS lines_x, lines.y AS lines_y, lines.graph_id, graphs.id AS graphs_id, graphs.name, graphs.x AS graphs_x, graphs.y AS graphs_y, graphs.lines_x AS graphs_lines_x FROM points LEFT OUTER JOIN lines ON (lines.x = points.id) LEFT OUTER JOIN graphs ON (graphs.id = lines.graph_id)'
+  end
+
+  it "#graph should allow graphing of the same dataset multiple times" do
+    ds = @ds1.graph(@ds2, :x=>:id).graph(@ds2, {:y=>:points__id}, :table_alias=>:graph)
+    ds.sql.should == 'SELECT points.id, points.x, points.y, lines.id AS lines_id, lines.x AS lines_x, lines.y AS lines_y, lines.graph_id, graph.id AS graph_id_0, graph.x AS graph_x, graph.y AS graph_y, graph.graph_id AS graph_graph_id FROM points LEFT OUTER JOIN lines ON (lines.x = points.id) LEFT OUTER JOIN lines AS graph ON (graph.y = points.id)'
+  end
+
+  it "#graph should raise an error if the table/table alias has already been used" do
+    proc{@ds1.graph(@ds1, :x=>:id)}.should raise_error(Sequel::Error)
+    proc{@ds1.graph(@ds2, :x=>:id)}.should_not raise_error
+    proc{@ds1.graph(@ds2, :x=>:id).graph(@ds2, :x=>:id)}.should raise_error(Sequel::Error)
+    proc{@ds1.graph(@ds2, :x=>:id).graph(@ds2, {:x=>:id}, :table_alias=>:blah)}.should_not raise_error
+  end
+
+  it "#set_graph_aliases and #add_graph_aliases should not modify the current dataset's opts" do
+    o1 = @ds1.opts
+    o2 = o1.dup
+    ds1 = @ds1.set_graph_aliases(:x=>[:graphs,:id])
+    @ds1.opts.should == o1
+    @ds1.opts.should == o2
+    ds1.opts.should_not == o1
+    o3 = ds1.opts
+    o4 = o3.dup
+    ds2 = ds1.add_graph_aliases(:y=>[:blah,:id])
+    ds1.opts.should == o3
+    ds1.opts.should == o3
+    ds2.opts.should_not == o2
+  end
+
+  it "#set_graph_aliases should specify the graph mapping" do
+    ds = @ds1.graph(:lines, :x=>:id)
+    ds.sql.should == 'SELECT points.id, points.x, points.y, lines.id AS lines_id, lines.x AS lines_x, lines.y AS lines_y, lines.graph_id FROM points LEFT OUTER JOIN lines ON (lines.x = points.id)'
+    ds = ds.set_graph_aliases(:x=>[:points, :x], :y=>[:lines, :y])
+    ['SELECT points.x, lines.y FROM points LEFT OUTER JOIN lines ON (lines.x = points.id)',
+    'SELECT lines.y, points.x FROM points LEFT OUTER JOIN lines ON (lines.x = points.id)'
+    ].should(include(ds.sql))
+  end
+
+  it "#add_graph_aliases should add columns to the graph mapping" do
+    @ds1.graph(:lines, :x=>:id).set_graph_aliases(:x=>[:points, :q]).add_graph_aliases(:y=>[:lines, :r]).sql.should == 'SELECT points.q AS x, lines.r AS y FROM points LEFT OUTER JOIN lines ON (lines.x = points.id)'
+  end
+
+  it "#set_graph_aliases should allow a third entry to specify an expression to use other than the default" do
+    ds = @ds1.graph(:lines, :x=>:id).set_graph_aliases(:x=>[:points, :x, 1], :y=>[:lines, :y, :random.sql_function])
+    ['SELECT 1 AS x, random() AS y FROM points LEFT OUTER JOIN lines ON (lines.x = points.id)',
+    'SELECT random() AS y, 1 AS x FROM points LEFT OUTER JOIN lines ON (lines.x = points.id)'
+    ].should(include(ds.sql))
+  end
+
+  it "#set_graph_aliases should only alias columns if necessary" do
+    ds = @ds1.set_graph_aliases(:x=>[:points, :x], :y=>[:lines, :y])
+    ['SELECT points.x, lines.y FROM points',
+    'SELECT lines.y, points.x FROM points'
+    ].should(include(ds.sql))
+
+    ds = @ds1.set_graph_aliases(:x1=>[:points, :x], :y=>[:lines, :y])
+    ['SELECT points.x AS x1, lines.y FROM points',
+    'SELECT lines.y, points.x AS x1 FROM points'
+    ].should(include(ds.sql))
+  end
+
+  it "#graph_each should split the result set into component tables" do
+    ds = @ds1.graph(@ds2, :x=>:id)
+    def ds.fetch_rows(sql, &block)
+      yield({:id=>1,:x=>2,:y=>3,:lines_id=>4,:lines_x=>5,:lines_y=>6,:graph_id=>7})
+    end
+    results = ds.all
+    results.length.should == 1
+    results.first.should == {:points=>{:id=>1, :x=>2, :y=>3}, :lines=>{:id=>4, :x=>5, :y=>6, :graph_id=>7}}
+
+    ds = @ds1.graph(@ds2, :x=>:id).graph(@ds3, :id=>:graph_id)
+    def ds.fetch_rows(sql, &block)
+      yield({:id=>1,:x=>2,:y=>3,:lines_id=>4,:lines_x=>5,:lines_y=>6,:graph_id=>7, :graphs_id=>8, :name=>9, :graphs_x=>10, :graphs_y=>11, :graphs_lines_x=>12})
+    end
+    results = ds.all
+    results.length.should == 1
+    results.first.should == {:points=>{:id=>1, :x=>2, :y=>3}, :lines=>{:id=>4, :x=>5, :y=>6, :graph_id=>7}, :graphs=>{:id=>8, :name=>9, :x=>10, :y=>11, :lines_x=>12}}
+
+    ds = @ds1.graph(@ds2, :x=>:id).graph(@ds2, {:y=>:points__id}, :table_alias=>:graph)
+    def ds.fetch_rows(sql, &block)
+      yield({:id=>1,:x=>2,:y=>3,:lines_id=>4,:lines_x=>5,:lines_y=>6,:graph_id=>7, :graph_id_0=>8, :graph_x=>9, :graph_y=>10, :graph_graph_id=>11})
+    end
+    results = ds.all
+    results.length.should == 1
+    results.first.should == {:points=>{:id=>1, :x=>2, :y=>3}, :lines=>{:id=>4, :x=>5, :y=>6, :graph_id=>7}, :graph=>{:id=>8, :x=>9, :y=>10, :graph_id=>11}}
+  end
+
+  it "#graph_each should give a nil value instead of a hash when all values for a table are nil" do
+    ds = @ds1.graph(@ds2, :x=>:id)
+    def ds.fetch_rows(sql, &block)
+      yield({:id=>1,:x=>2,:y=>3,:lines_id=>nil,:lines_x=>nil,:lines_y=>nil,:graph_id=>nil})
+    end
+    results = ds.all
+    results.length.should == 1
+    results.first.should == {:points=>{:id=>1, :x=>2, :y=>3}, :lines=>nil}
+
+    ds = @ds1.graph(@ds2, :x=>:id).graph(@ds3, :id=>:graph_id)
+    def ds.fetch_rows(sql, &block)
+      yield({:id=>1,:x=>2,:y=>3,:lines_id=>4,:lines_x=>5,:lines_y=>6,:graph_id=>7, :graphs_id=>nil, :name=>nil, :graphs_x=>nil, :graphs_y=>nil, :graphs_lines_x=>nil})
+      yield({:id=>2,:x=>4,:y=>5,:lines_id=>nil,:lines_x=>nil,:lines_y=>nil,:graph_id=>nil, :graphs_id=>nil, :name=>nil, :graphs_x=>nil, :graphs_y=>nil, :graphs_lines_x=>nil})
+      yield({:id=>3,:x=>5,:y=>6,:lines_id=>4,:lines_x=>5,:lines_y=>6,:graph_id=>7, :graphs_id=>7, :name=>8, :graphs_x=>9, :graphs_y=>10, :graphs_lines_x=>11})
+      yield({:id=>3,:x=>5,:y=>6,:lines_id=>7,:lines_x=>5,:lines_y=>8,:graph_id=>9, :graphs_id=>9, :name=>10, :graphs_x=>10, :graphs_y=>11, :graphs_lines_x=>12})
+    end
+    results = ds.all
+    results.length.should == 4
+    results[0].should == {:points=>{:id=>1, :x=>2, :y=>3}, :lines=>{:id=>4, :x=>5, :y=>6, :graph_id=>7}, :graphs=>nil}
+    results[1].should == {:points=>{:id=>2, :x=>4, :y=>5}, :lines=>nil, :graphs=>nil}
+    results[2].should == {:points=>{:id=>3, :x=>5, :y=>6}, :lines=>{:id=>4, :x=>5, :y=>6, :graph_id=>7}, :graphs=>{:id=>7, :name=>8, :x=>9, :y=>10, :lines_x=>11}}
+    results[3].should == {:points=>{:id=>3, :x=>5, :y=>6}, :lines=>{:id=>7, :x=>5, :y=>8, :graph_id=>9}, :graphs=>{:id=>9, :name=>10, :x=>10, :y=>11, :lines_x=>12}}
+  end
+
+  it "#graph_each should not give a nil value instead of a hash when any value for a table is false" do
+    ds = @ds1.graph(@ds2, :x=>:id)
+    def ds.fetch_rows(sql, &block)
+      block.call(:id=>1,:x=>2,:y=>3,:lines_id=>nil,:lines_x=>false,:lines_y=>nil,:graph_id=>nil)
+    end
+    ds.all.should == [{:points=>{:id=>1, :x=>2, :y=>3}, :lines=>{:id=>nil, :x=>false, :y=>nil, :graph_id=>nil}}]
+  end
+
+  it "#graph_each should not included tables graphed with the :select => false option in the result set" do
+    ds = @ds1.graph(:lines, {:x=>:id}, :select=>false).graph(:graphs, :id=>:graph_id)
+    def ds.fetch_rows(sql, &block)
+      yield({:id=>1,:x=>2,:y=>3,:graphs_id=>8, :name=>9, :graphs_x=>10, :graphs_y=>11, :lines_x=>12})
+    end
+    results = ds.all
+    results.length.should == 1
+    results.first.should == {:points=>{:id=>1, :x=>2, :y=>3}, :graphs=>{:id=>8, :name=>9, :x=>10, :y=>11, :lines_x=>12}}
+  end
+
+  it "#graph_each should only include the columns selected with #set_graph_aliases and #add_graph_aliases, if called" do
+    ds = @ds1.graph(:lines, :x=>:id).set_graph_aliases(:x=>[:points, :x], :y=>[:lines, :y])
+    def ds.fetch_rows(sql, &block)
+      yield({:x=>2,:y=>3})
+    end
+    results = ds.all
+    results.length.should == 1
+    results.first.should == {:points=>{:x=>2}, :lines=>{:y=>3}}
+
+    ds = @ds1.graph(:lines, :x=>:id).set_graph_aliases(:x=>[:points, :x])
+    def ds.fetch_rows(sql, &block)
+      yield({:x=>2})
+    end
+    results = ds.all
+    results.length.should == 1
+    results.first.should == {:points=>{:x=>2}, :lines=>nil}
+
+    ds = ds.add_graph_aliases(:q=>[:points, :r, 18])
+    def ds.fetch_rows(sql, &block)
+      yield({:x=>2, :q=>18})
+    end
+    ds.all.should == [{:points=>{:x=>2, :r=>18}, :lines=>nil}]
+  end
+
+  it "#graph_each should correctly map values when #set_graph_aliases is used with a third argument for each entry" do
+    ds = @ds1.graph(:lines, :x=>:id).set_graph_aliases(:x=>[:points, :z1, 2], :y=>[:lines, :z2, :random.sql_function])
+    def ds.fetch_rows(sql, &block)
+      yield({:x=>2,:y=>3})
+    end
+    results = ds.all
+    results.length.should == 1
+    results.first.should == {:points=>{:z1=>2}, :lines=>{:z2=>3}}
+  end
+
+  it "#graph_each should run the row_proc for graphed datasets" do
+    @ds1.row_proc = proc{|h| h.keys.each{|k| h[k] *= 2}; h}
+    @ds2.row_proc = proc{|h| h.keys.each{|k| h[k] *= 3}; h}
+    ds = @ds1.graph(@ds2, :x=>:id)
+    def ds.fetch_rows(sql, &block)
+      yield({:id=>1,:x=>2,:y=>3,:lines_id=>4,:lines_x=>5,:lines_y=>6,:graph_id=>7})
+    end
+    results = ds.all
+    results.length.should == 1
+    results.first.should == {:points=>{:id=>2, :x=>4, :y=>6}, :lines=>{:id=>12, :x=>15, :y=>18, :graph_id=>21}}
+  end
+end
diff -pruN 1.5.1-1/spec/core/schema_generator_spec.rb 3.0.0-1/spec/core/schema_generator_spec.rb
--- 1.5.1-1/spec/core/schema_generator_spec.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/core/schema_generator_spec.rb	2009-04-27 02:58:47.000000000 +0100
@@ -0,0 +1,159 @@
+require File.join(File.dirname(__FILE__), 'spec_helper')
+
+describe Sequel::Schema::Generator do
+  before do
+    @generator = Sequel::Schema::Generator.new(SchemaDummyDatabase.new) do
+      string :title
+      column :body, :text
+      foreign_key :parent_id
+      primary_key :id
+      check 'price > 100'
+      constraint(:xxx) {:yyy == :zzz}
+      index :title
+      index [:title, :body], :unique => true
+      foreign_key :node_id, :nodes
+      primary_key [:title, :parent_id], :name => :cpk
+      foreign_key [:node_id, :prop_id], :nodes_props, :name => :cfk
+    end
+    @columns, @indexes, @constraints = @generator.columns, @generator.indexes, @generator.constraints
+  end
+  
+  it "should primary key column first" do
+    @columns.first[:name].should == :id
+    @columns.first[:primary_key].should == true
+    @columns[3][:name].should == :parent_id
+    @columns[3][:primary_key].should == nil
+  end
+  
+  it "counts definitions correctly" do
+    @columns.size.should == 5
+    @indexes.size.should == 2
+    @constraints.size.should == 4
+  end
+  
+  it "retrieves primary key name" do
+    @generator.primary_key_name.should == :id
+  end
+
+  it "keeps columns in order" do
+    @columns[1][:name].should == :title
+    @columns[1][:type].should == :string
+    @columns[2][:name].should == :body
+    @columns[2][:type].should == :text
+  end
+  
+  it "creates foreign key column" do
+    @columns[3][:name].should == :parent_id
+    @columns[3][:type].should == Integer
+    @columns[4][:name].should == :node_id
+    @columns[4][:type].should == Integer
+  end
+  
+  it "uses table for foreign key columns, if specified" do
+    @columns[3][:table].should == nil
+    @columns[4][:table].should == :nodes
+    @constraints[3][:table].should == :nodes_props
+  end
+  
+  it "finds columns" do
+    [:title, :body, :parent_id, :id].each do |col|
+      @generator.has_column?(col).should be_true
+    end
+    @generator.has_column?(:foo).should_not be_true
+  end
+  
+  it "creates constraints" do
+    @constraints[0][:name].should == nil
+    @constraints[0][:type].should == :check
+    @constraints[0][:check].should == ['price > 100']
+
+    @constraints[1][:name].should == :xxx
+    @constraints[1][:type].should == :check
+    @constraints[1][:check].should be_a_kind_of(Proc)
+
+    @constraints[2][:name].should == :cpk
+    @constraints[2][:type].should == :primary_key
+    @constraints[2][:columns].should == [ :title, :parent_id ]
+
+    @constraints[3][:name].should == :cfk
+    @constraints[3][:type].should == :foreign_key
+    @constraints[3][:columns].should == [ :node_id, :prop_id ]
+    @constraints[3][:table].should == :nodes_props
+  end
+  
+  it "creates indexes" do
+    @indexes[0][:columns].should include(:title)
+    @indexes[1][:columns].should include(:title)
+    @indexes[1][:columns].should include(:body)
+  end
+end
+
+describe Sequel::Schema::AlterTableGenerator do
+  before do
+    @generator = Sequel::Schema::AlterTableGenerator.new(SchemaDummyDatabase.new) do
+      add_column :aaa, :text
+      drop_column :bbb
+      rename_column :ccc, :ho
+      set_column_type :ddd, :float
+      set_column_default :eee, 1
+      add_index [:fff, :ggg]
+      drop_index :hhh
+      drop_index :hhh, :name=>:blah_blah
+      add_full_text_index :blah
+      add_spatial_index :geom
+      add_index :blah, :type => :hash
+      add_index :blah, :where => {:something => true}
+      add_constraint :con1, 'fred > 100'
+      drop_constraint :con2
+      add_unique_constraint [:aaa, :bbb, :ccc], :name => :con3
+      add_primary_key :id
+      add_foreign_key :node_id, :nodes
+      add_primary_key [:aaa, :bbb]
+      add_foreign_key [:node_id, :prop_id], :nodes_props
+    end
+  end
+  
+  specify "should generate operation records" do
+    @generator.operations.should == [
+      {:op => :add_column, :name => :aaa, :type => :text},
+      {:op => :drop_column, :name => :bbb},
+      {:op => :rename_column, :name => :ccc, :new_name => :ho},
+      {:op => :set_column_type, :name => :ddd, :type => :float},
+      {:op => :set_column_default, :name => :eee, :default => 1},
+      {:op => :add_index, :columns => [:fff, :ggg]},
+      {:op => :drop_index, :columns => [:hhh]},
+      {:op => :drop_index, :columns => [:hhh], :name=>:blah_blah},
+      {:op => :add_index, :columns => [:blah], :type => :full_text},
+      {:op => :add_index, :columns => [:geom], :type => :spatial},
+      {:op => :add_index, :columns => [:blah], :type => :hash},
+      {:op => :add_index, :columns => [:blah], :where => {:something => true}},
+      {:op => :add_constraint, :type => :check, :name => :con1, :check => ['fred > 100']},
+      {:op => :drop_constraint, :name => :con2},
+      {:op => :add_constraint, :type => :unique, :name => :con3, :columns => [:aaa, :bbb, :ccc]},
+      {:op => :add_column, :name => :id, :type => Integer, :primary_key=>true, :auto_increment=>true},
+      {:op => :add_column, :name => :node_id, :type => Integer, :table=>:nodes},
+      {:op => :add_constraint, :type => :primary_key, :columns => [:aaa, :bbb]},
+      {:op => :add_constraint, :type => :foreign_key, :columns => [:node_id, :prop_id], :table => :nodes_props}
+    ]
+  end
+end
+
+describe "Sequel::Schema::Generator generic type methods" do
+  it "should store the type class in :type for each column" do
+    Sequel::Schema::Generator.new(SchemaDummyDatabase.new) do
+      String :a
+      Integer :b
+      Fixnum :c
+      Bignum :d
+      Float :e
+      BigDecimal :f
+      Date :g
+      DateTime :h
+      Time :i
+      Numeric :j
+      File :k
+      TrueClass :l
+      FalseClass :m
+    end.columns.map{|c| c[:type]}.should == [String, Integer, Fixnum, Bignum, Float, BigDecimal, Date, DateTime, Time, Numeric, File, TrueClass, FalseClass]
+  end
+end
diff -pruN 1.5.1-1/spec/core/schema_spec.rb 3.0.0-1/spec/core/schema_spec.rb
--- 1.5.1-1/spec/core/schema_spec.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/core/schema_spec.rb	2009-04-27 02:59:44.000000000 +0100
@@ -0,0 +1,788 @@
+require File.join(File.dirname(__FILE__), 'spec_helper')
+
+context "DB#create_table" do
+  before do
+    @db = SchemaDummyDatabase.new
+  end
+  
+  specify "should accept the table name" do
+    @db.create_table(:cats) {}
+    @db.sqls.should == ['CREATE TABLE cats ()']
+  end
+
+  specify "should accept the table name in multiple formats" do
+    @db.create_table(:cats__cats) {}
+    @db.create_table("cats__cats1") {}
+    @db.create_table(:cats__cats2.identifier) {}
+    @db.create_table(:cats.qualify(:cats3)) {}
+    @db.sqls.should == ['CREATE TABLE cats.cats ()', 'CREATE TABLE cats__cats1 ()', 'CREATE TABLE cats__cats2 ()', 'CREATE TABLE cats3.cats ()']
+  end
+
+  specify "should raise an error if the table name argument is not valid" do
+    proc{@db.create_table(1) {}}.should raise_error(Sequel::Error)
+    proc{@db.create_table(:cats.as(:c)) {}}.should raise_error(Sequel::Error)
+  end
+
+  specify "should accept multiple columns" do
+    @db.create_table(:cats) do
+      column :id, :integer
+      column :name, :text
+    end
+    @db.sqls.should == ['CREATE TABLE cats (id integer, name text)']
+  end
+  
+  specify "should accept method calls as data types" do
+    @db.create_table(:cats) do
+      integer :id
+      text :name
+    end
+    @db.sqls.should == ['CREATE TABLE cats (id integer, name text)']
+  end
+  
+  specify "should transform types given as ruby classes to database-specific types" do
+    @db.create_table(:cats) do
+      String :a
+      Integer :b
+      Fixnum :c
+      Bignum :d
+      Float :e
+      BigDecimal :f
+      Date :g
+      DateTime :h
+      Time :i
+      Numeric :j
+      File :k
+      TrueClass :l
+      FalseClass :m
+      column :n, Fixnum
+      primary_key :o, :type=>String
+      foreign_key :p, :f, :type=>Date
+    end
+    @db.sqls.should == ['CREATE TABLE cats (o varchar(255) PRIMARY KEY AUTOINCREMENT, a varchar(255), b integer, c integer, d bigint, e double precision, f numeric, g date, h timestamp, i timestamp, j numeric, k blob, l boolean, m boolean, n integer, p date REFERENCES f)']
+  end
+
+  specify "should allow the use of modifiers with ruby class types" do
+    @db.create_table(:cats) do
+      String :a, :size=>50
+      String :b, :text=>true
+      String :c, :fixed=>true, :size=>40
+      Time :d, :only_time=>true
+      BigDecimal :e, :size=>[11,2]
+    end
+    @db.sqls.should == ['CREATE TABLE cats (a varchar(50), b text, c char(40), d time, e numeric(11, 2))']
+  end
+
+  specify "should raise an error if you use a ruby class that isn't handled" do
+    proc{@db.create_table(:cats){column :a, Class}}.should raise_error(Sequel::Error)
+  end
+
+  specify "should accept primary key definition" do
+    @db.create_table(:cats) do
+      primary_key :id
+    end
+    @db.sqls.should == ['CREATE TABLE cats (id integer PRIMARY KEY AUTOINCREMENT)']
+
+    @db.sqls.clear
+    @db.create_table(:cats) do
+      primary_key :id, :serial, :auto_increment => false
+    end
+    @db.sqls.should == ['CREATE TABLE cats (id serial PRIMARY KEY)']
+
+    @db.sqls.clear
+    @db.create_table(:cats) do
+      primary_key :id, :type => :serial, :auto_increment => false
+    end
+    @db.sqls.should == ['CREATE TABLE cats (id serial PRIMARY KEY)']
+  end
+
+  specify "should accept and literalize default values" do
+    @db.create_table(:cats) do
+      integer :id, :default => 123
+      text :name, :default => "abc'def"
+    end
+    @db.sqls.should == ["CREATE TABLE cats (id integer DEFAULT 123, name text DEFAULT 'abc''def')"]
+  end
+  
+  specify "should accept not null definition" do
+    @db.create_table(:cats) do
+      integer :id
+      text :name, :null => false
+    end
+    @db.sqls.should == ["CREATE TABLE cats (id integer, name text NOT NULL)"]
+  end
+  
+  specify "should accept null definition" do
+    @db.create_table(:cats) do
+      integer :id
+      text :name, :null => true
+    end
+    @db.sqls.should == ["CREATE TABLE cats (id integer, name text NULL)"]
+  end
+  
+  specify "should accept unique definition" do
+    @db.create_table(:cats) do
+      integer :id
+      text :name, :unique => true
+    end
+    @db.sqls.should == ["CREATE TABLE cats (id integer, name text UNIQUE)"]
+  end
+  
+  specify "should accept unsigned definition" do
+    @db.create_table(:cats) do
+      integer :value, :unsigned => true
+    end
+    @db.sqls.should == ["CREATE TABLE cats (value integer UNSIGNED)"]
+  end
+  
+  specify "should accept [SET|ENUM](...) types" do
+    @db.create_table(:cats) do
+      set :color, :elements => ['black', 'tricolor', 'grey']
+    end
+    @db.sqls.should == ["CREATE TABLE cats (color set('black', 'tricolor', 'grey'))"]
+  end
+  
+  specify "should accept varchar size" do
+    @db.create_table(:cats) do
+      varchar :name
+    end
+    @db.sqls.should == ["CREATE TABLE cats (name varchar(255))"]
+    @db.sqls.clear
+    @db.create_table(:cats) do
+      varchar :name, :size => 51
+    end
+    @db.sqls.should == ["CREATE TABLE cats (name varchar(51))"]
+  end
+  
+  specify "should use double precision for double type" do
+    @db.create_table(:cats) do
+      double :name
+    end
+    @db.sqls.should == ["CREATE TABLE cats (name double precision)"]
+  end
+
+  specify "should accept foreign keys without options" do
+    @db.create_table(:cats) do
+      foreign_key :project_id
+    end
+    @db.sqls.should == ["CREATE TABLE cats (project_id integer)"]
+  end
+
+  specify "should accept foreign keys with options" do
+    @db.create_table(:cats) do
+      foreign_key :project_id, :table => :projects
+    end
+    @db.sqls.should == ["CREATE TABLE cats (project_id integer REFERENCES projects)"]
+  end
+
+  specify "should accept foreign keys with separate table argument" do
+    @db.create_table(:cats) do
+      foreign_key :project_id, :projects, :default=>3
+    end
+    @db.sqls.should == ["CREATE TABLE cats (project_id integer DEFAULT 3 REFERENCES projects)"]
+  end
+  
+  specify "should raise an error if the table argument to foreign_key isn't a hash, symbol, or nil" do
+    proc{@db.create_table(:cats){foreign_key :project_id, Object.new, :default=>3}}.should raise_error(Sequel::Error)
+  end
+  
+  specify "should accept foreign keys with arbitrary keys" do
+    @db.create_table(:cats) do
+      foreign_key :project_id, :table => :projects, :key => :id
+    end
+    @db.sqls.should == ["CREATE TABLE cats (project_id integer REFERENCES projects(id))"]
+
+    @db.sqls.clear
+    @db.create_table(:cats) do
+      foreign_key :project_id, :table => :projects, :key => :zzz
+    end
+    @db.sqls.should == ["CREATE TABLE cats (project_id integer REFERENCES projects(zzz))"]
+  end
+  
+  specify "should accept foreign keys with ON DELETE clause" do
+    @db.create_table(:cats) do
+      foreign_key :project_id, :table => :projects, :on_delete => :restrict
+    end
+    @db.sqls.should == ["CREATE TABLE cats (project_id integer REFERENCES projects ON DELETE RESTRICT)"]
+
+    @db.sqls.clear
+    @db.create_table(:cats) do
+      foreign_key :project_id, :table => :projects, :on_delete => :cascade
+    end
+    @db.sqls.should == ["CREATE TABLE cats (project_id integer REFERENCES projects ON DELETE CASCADE)"]
+
+    @db.sqls.clear
+    @db.create_table(:cats) do
+      foreign_key :project_id, :table => :projects, :on_delete => :no_action
+    end
+    @db.sqls.should == ["CREATE TABLE cats (project_id integer REFERENCES projects ON DELETE NO ACTION)"]
+    @db.sqls.clear
+
+    @db.sqls.clear
+    @db.create_table(:cats) do
+      foreign_key :project_id, :table => :projects, :on_delete => :set_null
+    end
+    @db.sqls.should == ["CREATE TABLE cats (project_id integer REFERENCES projects ON DELETE SET NULL)"]
+    @db.sqls.clear
+
+    @db.sqls.clear
+    @db.create_table(:cats) do
+      foreign_key :project_id, :table => :projects, :on_delete => :set_default
+    end
+    @db.sqls.should == ["CREATE TABLE cats (project_id integer REFERENCES projects ON DELETE SET DEFAULT)"]
+    @db.sqls.clear
+  end
+
+  specify "should accept foreign keys with ON UPDATE clause" do
+    @db.create_table(:cats) do
+      foreign_key :project_id, :table => :projects, :on_update => :restrict
+    end
+    @db.sqls.should == ["CREATE TABLE cats (project_id integer REFERENCES projects ON UPDATE RESTRICT)"]
+
+    @db.sqls.clear
+    @db.create_table(:cats) do
+      foreign_key :project_id, :table => :projects, :on_update => :cascade
+    end
+    @db.sqls.should == ["CREATE TABLE cats (project_id integer REFERENCES projects ON UPDATE CASCADE)"]
+
+    @db.sqls.clear
+    @db.create_table(:cats) do
+      foreign_key :project_id, :table => :projects, :on_update => :no_action
+    end
+    @db.sqls.should == ["CREATE TABLE cats (project_id integer REFERENCES projects ON UPDATE NO ACTION)"]
+    @db.sqls.clear
+
+    @db.sqls.clear
+    @db.create_table(:cats) do
+      foreign_key :project_id, :table => :projects, :on_update => :set_null
+    end
+    @db.sqls.should == ["CREATE TABLE cats (project_id integer REFERENCES projects ON UPDATE SET NULL)"]
+    @db.sqls.clear
+
+    @db.sqls.clear
+    @db.create_table(:cats) do
+      foreign_key :project_id, :table => :projects, :on_update => :set_default
+    end
+    @db.sqls.should == ["CREATE TABLE cats (project_id integer REFERENCES projects ON UPDATE SET DEFAULT)"]
+    @db.sqls.clear
+  end
+  
+  specify "should accept inline index definition" do
+    @db.create_table(:cats) do
+      integer :id, :index => true
+    end
+    @db.sqls.should == ["CREATE TABLE cats (id integer)", "CREATE INDEX cats_id_index ON cats (id)"]
+  end
+  
+  specify "should accept inline index definition for foreign keys" do
+    @db.create_table(:cats) do
+      foreign_key :project_id, :table => :projects, :on_delete => :cascade, :index => true
+    end
+    @db.sqls.should == ["CREATE TABLE cats (project_id integer REFERENCES projects ON DELETE CASCADE)",
+      "CREATE INDEX cats_project_id_index ON cats (project_id)"]
+  end
+  
+  specify "should accept index definitions" do
+    @db.create_table(:cats) do
+      integer :id
+      index :id
+    end
+    @db.sqls.should == ["CREATE TABLE cats (id integer)", "CREATE INDEX cats_id_index ON cats (id)"]
+  end
+
+  specify "should accept unique index definitions" do
+    @db.create_table(:cats) do
+      text :name
+      unique :name
+    end
+    @db.sqls.should == ["CREATE TABLE cats (name text, UNIQUE (name))"]
+  end
+
+  specify "should raise on full-text index definitions" do
+    proc {
+      @db.create_table(:cats) do
+        text :name
+        full_text_index :name
+      end
+    }.should raise_error(Sequel::Error)
+  end
+
+  specify "should raise on spatial index definitions" do
+    proc {
+      @db.create_table(:cats) do
+        point :geom
+        spatial_index :geom
+      end
+    }.should raise_error(Sequel::Error)
+  end
+
+  specify "should raise on partial index definitions" do
+    proc {
+      @db.create_table(:cats) do
+        text :name
+        index :name, :where => {:something => true}
+      end
+    }.should raise_error(Sequel::Error)
+  end
+
+  specify "should raise index definitions with type" do
+    proc {
+      @db.create_table(:cats) do
+        text :name
+        index :name, :type => :hash
+      end
+    }.should raise_error(Sequel::Error)
+  end
+
+  specify "should accept multiple index definitions" do
+    @db.create_table(:cats) do
+      integer :id
+      index :id
+      index :name
+    end
+    @db.sqls.should == ["CREATE TABLE cats (id integer)", "CREATE INDEX cats_id_index ON cats (id)", "CREATE INDEX cats_name_index ON cats (name)"]
+  end
+  
+  specify "should accept functional indexes" do
+    @db.create_table(:cats) do
+      integer :id
+      index :lower.sql_function(:name)
+    end
+    @db.sqls.should == ["CREATE TABLE cats (id integer)", "CREATE INDEX cats_lower_name__index ON cats (lower(name))"]
+  end
+  
+  specify "should accept indexes with identifiers" do
+    @db.create_table(:cats) do
+      integer :id
+      index :lower__name.identifier
+    end
+    @db.sqls.should == ["CREATE TABLE cats (id integer)", "CREATE INDEX cats_lower__name_index ON cats (lower__name)"]
+  end
+  
+  specify "should accept custom index names" do
+    @db.create_table(:cats) do
+      integer :id
+      index :id, :name => 'abc'
+    end
+    @db.sqls.should == ["CREATE TABLE cats (id integer)", "CREATE INDEX abc ON cats (id)"]
+  end
+
+  specify "should accept unique index definitions" do
+    @db.create_table(:cats) do
+      integer :id
+      index :id, :unique => true
+    end
+    @db.sqls.should == ["CREATE TABLE cats (id integer)", "CREATE UNIQUE INDEX cats_id_index ON cats (id)"]
+  end
+  
+  specify "should accept composite index definitions" do
+    @db.create_table(:cats) do
+      integer :id
+      index [:id, :name], :unique => true
+    end
+    @db.sqls.should == ["CREATE TABLE cats (id integer)", "CREATE UNIQUE INDEX cats_id_name_index ON cats (id, name)"]
+  end
+  
+  specify "should accept unnamed constraint definitions with blocks" do
+    @db.create_table(:cats) do
+      integer :score
+      check {(:x.sql_number > 0) & (:y.sql_number < 1)}
+    end
+    @db.sqls.should == ["CREATE TABLE cats (score integer, CHECK ((x > 0) AND (y < 1)))"]
+  end
+
+  specify "should accept unnamed constraint definitions" do
+    @db.create_table(:cats) do
+      check 'price < ?', 100
+    end
+    @db.sqls.should == ["CREATE TABLE cats (CHECK (price < 100))"]
+  end
+
+  specify "should accept hash constraints" do
+    @db.create_table(:cats) do
+      check :price=>100
+    end
+    @db.sqls.should == ["CREATE TABLE cats (CHECK (price = 100))"]
+  end
+
+  specify "should accept named constraint definitions" do
+    @db.create_table(:cats) do
+      integer :score
+      constraint :valid_score, 'score <= 100'
+    end
+    @db.sqls.should == ["CREATE TABLE cats (score integer, CONSTRAINT valid_score CHECK (score <= 100))"]
+  end
+
+  specify "should accept named constraint definitions with block" do
+    @db.create_table(:cats) do
+      constraint(:blah_blah) {(:x.sql_number > 0) & (:y.sql_number < 1)}
+    end
+    @db.sqls.should == ["CREATE TABLE cats (CONSTRAINT blah_blah CHECK ((x > 0) AND (y < 1)))"]
+  end
+
+  specify "should raise an error if an invalid constraint type is used" do
+    proc{@db.create_table(:cats){unique [:a, :b], :type=>:bb}}.should raise_error(Sequel::Error)
+  end
+
+  specify "should accept composite primary keys" do
+    @db.create_table(:cats) do
+      integer :a
+      integer :b
+      primary_key [:a, :b]
+    end
+    @db.sqls.should == ["CREATE TABLE cats (a integer, b integer, PRIMARY KEY (a, b))"]
+  end
+
+  specify "should accept named composite primary keys" do
+    @db.create_table(:cats) do
+      integer :a
+      integer :b
+      primary_key [:a, :b], :name => :cpk
+    end
+    @db.sqls.should == ["CREATE TABLE cats (a integer, b integer, CONSTRAINT cpk PRIMARY KEY (a, b))"]
+  end
+
+  specify "should accept composite foreign keys" do
+    @db.create_table(:cats) do
+      integer :a
+      integer :b
+      foreign_key [:a, :b], :abc
+    end
+    @db.sqls.should == ["CREATE TABLE cats (a integer, b integer, FOREIGN KEY (a, b) REFERENCES abc)"]
+  end
+
+  specify "should accept named composite foreign keys" do
+    @db.create_table(:cats) do
+      integer :a
+      integer :b
+      foreign_key [:a, :b], :abc, :name => :cfk
+    end
+    @db.sqls.should == ["CREATE TABLE cats (a integer, b integer, CONSTRAINT cfk FOREIGN KEY (a, b) REFERENCES abc)"]
+  end
+
+  specify "should accept composite foreign keys with arbitrary keys" do
+    @db.create_table(:cats) do
+      integer :a
+      integer :b
+      foreign_key [:a, :b], :abc, :key => [:real_a, :real_b]
+    end
+    @db.sqls.should == ["CREATE TABLE cats (a integer, b integer, FOREIGN KEY (a, b) REFERENCES abc(real_a, real_b))"]
+    @db.sqls.clear
+
+    @db.create_table(:cats) do
+      integer :a
+      integer :b
+      foreign_key [:a, :b], :abc, :key => [:z, :x]
+    end
+    @db.sqls.should == ["CREATE TABLE cats (a integer, b integer, FOREIGN KEY (a, b) REFERENCES abc(z, x))"]
+  end
+
+  specify "should accept composite foreign keys with on delete and on update clauses" do
+    @db.create_table(:cats) do
+      integer :a
+      integer :b
+      foreign_key [:a, :b], :abc, :on_delete => :cascade
+    end
+    @db.sqls.should == ["CREATE TABLE cats (a integer, b integer, FOREIGN KEY (a, b) REFERENCES abc ON DELETE CASCADE)"]
+
+    @db.sqls.clear
+    @db.create_table(:cats) do
+      integer :a
+      integer :b
+      foreign_key [:a, :b], :abc, :on_update => :no_action
+    end
+    @db.sqls.should == ["CREATE TABLE cats (a integer, b integer, FOREIGN KEY (a, b) REFERENCES abc ON UPDATE NO ACTION)"]
+
+    @db.sqls.clear
+    @db.create_table(:cats) do
+      integer :a
+      integer :b
+      foreign_key [:a, :b], :abc, :on_delete => :restrict, :on_update => :set_default
+    end
+    @db.sqls.should == ["CREATE TABLE cats (a integer, b integer, FOREIGN KEY (a, b) REFERENCES abc ON DELETE RESTRICT ON UPDATE SET DEFAULT)"]
+
+    @db.sqls.clear
+    @db.create_table(:cats) do
+      integer :a
+      integer :b
+      foreign_key [:a, :b], :abc, :key => [:x, :y], :on_delete => :set_null, :on_update => :set_null
+    end
+    @db.sqls.should == ["CREATE TABLE cats (a integer, b integer, FOREIGN KEY (a, b) REFERENCES abc(x, y) ON DELETE SET NULL ON UPDATE SET NULL)"]
+  end
+end
+
+context "DB#create_table!" do
+  before do
+    @db = SchemaDummyDatabase.new
+  end
+  
+  specify "should drop the table and then create it" do
+    @db.create_table!(:cats) {}
+    @db.sqls.should == ['DROP TABLE cats', 'CREATE TABLE cats ()']
+  end
+end
+
+context "DB#drop_table" do
+  before do
+    @db = SchemaDummyDatabase.new
+  end
+
+  specify "should generate a DROP TABLE statement" do
+    @db.drop_table :cats
+    @db.sqls.should == ['DROP TABLE cats']
+  end
+end
+
+context "DB#alter_table" do
+  before do
+    @db = SchemaDummyDatabase.new
+  end
+  
+  specify "should allow adding not null constraint" do
+    @db.alter_table(:cats) do
+      set_column_allow_null :score, false
+    end
+    @db.sqls.should == ["ALTER TABLE cats ALTER COLUMN score SET NOT NULL"]
+  end
+  
+  specify "should allow droping not null constraint" do
+    @db.alter_table(:cats) do
+      set_column_allow_null :score, true
+    end
+    @db.sqls.should == ["ALTER TABLE cats ALTER COLUMN score DROP NOT NULL"]
+  end
+
+  specify "should support add_column" do
+    @db.alter_table(:cats) do
+      add_column :score, :integer
+    end
+    @db.sqls.should == ["ALTER TABLE cats ADD COLUMN score integer"]
+  end
+
+  specify "should support add_constraint" do
+    @db.alter_table(:cats) do
+      add_constraint :valid_score, 'score <= 100'
+    end
+    @db.sqls.should == ["ALTER TABLE cats ADD CONSTRAINT valid_score CHECK (score <= 100)"]
+  end
+
+  specify "should support add_constraint with block" do
+    @db.alter_table(:cats) do
+      add_constraint(:blah_blah) {(:x.sql_number > 0) & (:y.sql_number < 1)}
+    end
+    @db.sqls.should == ["ALTER TABLE cats ADD CONSTRAINT blah_blah CHECK ((x > 0) AND (y < 1))"]
+  end
+
+  specify "should support add_unique_constraint" do
+    @db.alter_table(:cats) do
+      add_unique_constraint [:a, :b]
+    end
+    @db.sqls.should == ["ALTER TABLE cats ADD UNIQUE (a, b)"]
+
+    @db.sqls.clear
+    @db.alter_table(:cats) do
+      add_unique_constraint [:a, :b], :name => :ab_uniq
+    end
+    @db.sqls.should == ["ALTER TABLE cats ADD CONSTRAINT ab_uniq UNIQUE (a, b)"]
+  end
+
+  specify "should support add_foreign_key" do
+    @db.alter_table(:cats) do
+      add_foreign_key :node_id, :nodes
+    end
+    @db.sqls.should == ["ALTER TABLE cats ADD COLUMN node_id integer REFERENCES nodes"]
+  end
+
+  specify "should support add_foreign_key with composite foreign keys" do
+    @db.alter_table(:cats) do
+      add_foreign_key [:node_id, :prop_id], :nodes_props
+    end
+    @db.sqls.should == ["ALTER TABLE cats ADD FOREIGN KEY (node_id, prop_id) REFERENCES nodes_props"]
+
+    @db.sqls.clear
+    @db.alter_table(:cats) do
+      add_foreign_key [:node_id, :prop_id], :nodes_props, :name => :cfk
+    end
+    @db.sqls.should == ["ALTER TABLE cats ADD CONSTRAINT cfk FOREIGN KEY (node_id, prop_id) REFERENCES nodes_props"]
+
+    @db.sqls.clear
+    @db.alter_table(:cats) do
+      add_foreign_key [:node_id, :prop_id], :nodes_props, :key => [:nid, :pid]
+    end
+    @db.sqls.should == ["ALTER TABLE cats ADD FOREIGN KEY (node_id, prop_id) REFERENCES nodes_props(nid, pid)"]
+
+    @db.sqls.clear
+    @db.alter_table(:cats) do
+      add_foreign_key [:node_id, :prop_id], :nodes_props, :on_delete => :restrict, :on_update => :cascade
+    end
+    @db.sqls.should == ["ALTER TABLE cats ADD FOREIGN KEY (node_id, prop_id) REFERENCES nodes_props ON DELETE RESTRICT ON UPDATE CASCADE"]
+  end
+
+  specify "should support add_index" do
+    @db.alter_table(:cats) do
+      add_index :name
+    end
+    @db.sqls.should == ["CREATE INDEX cats_name_index ON cats (name)"]
+  end
+
+  specify "should support add_primary_key" do
+    @db.alter_table(:cats) do
+      add_primary_key :id
+    end
+    @db.sqls.should == ["ALTER TABLE cats ADD COLUMN id integer PRIMARY KEY AUTOINCREMENT"]
+  end
+
+  specify "should support add_primary_key with composite primary keys" do
+    @db.alter_table(:cats) do
+      add_primary_key [:id, :type]
+    end
+    @db.sqls.should == ["ALTER TABLE cats ADD PRIMARY KEY (id, type)"]
+
+    @db.sqls.clear
+    @db.alter_table(:cats) do
+      add_primary_key [:id, :type], :name => :cpk
+    end
+    @db.sqls.should == ["ALTER TABLE cats ADD CONSTRAINT cpk PRIMARY KEY (id, type)"]
+  end
+
+  specify "should support drop_column" do
+    @db.alter_table(:cats) do
+      drop_column :score
+    end
+    @db.sqls.should == ["ALTER TABLE cats DROP COLUMN score"]
+  end
+
+  specify "should support drop_constraint" do
+    @db.alter_table(:cats) do
+      drop_constraint :valid_score
+    end
+    @db.sqls.should == ["ALTER TABLE cats DROP CONSTRAINT valid_score"]
+  end
+
+  specify "should support drop_index" do
+    @db.alter_table(:cats) do
+      drop_index :name
+    end
+    @db.sqls.should == ["DROP INDEX cats_name_index"]
+  end
+
+  specify "should support drop_index with a given name" do
+    @db.alter_table(:cats) do
+      drop_index :name, :name=>:blah_blah
+    end
+    @db.sqls.should == ["DROP INDEX blah_blah"]
+  end
+
+  specify "should support rename_column" do
+    @db.alter_table(:cats) do
+      rename_column :name, :old_name
+    end
+    @db.sqls.should == ["ALTER TABLE cats RENAME COLUMN name TO old_name"]
+  end
+
+  specify "should support set_column_default" do
+    @db.alter_table(:cats) do
+      set_column_default :score, 3
+    end
+    @db.sqls.should == ["ALTER TABLE cats ALTER COLUMN score SET DEFAULT 3"]
+  end
+
+  specify "should support set_column_type" do
+    @db.alter_table(:cats) do
+      set_column_type :score, :real
+    end
+    @db.sqls.should == ["ALTER TABLE cats ALTER COLUMN score TYPE real"]
+  end
+
+  specify "should support set_column_type with options" do
+    @db.alter_table(:cats) do
+      set_column_type :score, :integer, :unsigned=>true
+      set_column_type :score, :varchar, :size=>30
+      set_column_type :score, :enum, :elements=>['a', 'b']
+    end
+    @db.sqls.should == ["ALTER TABLE cats ALTER COLUMN score TYPE integer UNSIGNED",
+      "ALTER TABLE cats ALTER COLUMN score TYPE varchar(30)",
+      "ALTER TABLE cats ALTER COLUMN score TYPE enum('a', 'b')"]
+  end
+end
+
+context "Schema Parser" do
+  before do
+    @sqls = []
+    @db = Sequel::Database.new
+  end
+  after do
+    Sequel.convert_tinyint_to_bool = true
+  end
+
+  specify "should raise an error if there are no columns" do
+    @db.meta_def(:schema_parse_table) do |t, opts|
+      []
+    end
+    proc{@db.schema(:x)}.should raise_error(Sequel::Error)
+  end
+
+  specify "should parse the schema correctly for a single table" do
+    sqls = @sqls
+    proc{@db.schema(:x)}.should raise_error(Sequel::Error)
+    @db.meta_def(:schema_parse_table) do |t, opts|
+      sqls << t
+      [[:a, {:db_type=>t.to_s}]]
+    end
+    @db.schema(:x).should == [[:a, {:db_type=>"x"}]]
+    @sqls.should == ['x']
+    @db.schema(:x).should == [[:a, {:db_type=>"x"}]]
+    @sqls.should == ['x']
+    @db.schema(:x, :reload=>true).should == [[:a, {:db_type=>"x"}]]
+    @sqls.should == ['x', 'x']
+  end
+
+  specify "should convert various types of table name arguments" do
+    @db.meta_def(:schema_parse_table) do |t, opts|
+      [[t, {:db_type=>t}]]
+    end
+    s1 = @db.schema(:x)
+    s1.should == [['x', {:db_type=>'x'}]]
+    @db.schema(:x).object_id.should == s1.object_id
+    @db.schema(:x.identifier).object_id.should == s1.object_id
+    s2 = @db.schema(:x__y)
+    s2.should == [['y', {:db_type=>'y'}]]
+    @db.schema(:x__y).object_id.should == s2.object_id
+    @db.schema(:y.qualify(:x)).object_id.should == s2.object_id
+  end
+
+  specify "should correctly parse all supported data types" do
+    @db.meta_def(:schema_parse_table) do |t, opts|
+      [[:x, {:type=>schema_column_type(t.to_s)}]]
+    end
+    @db.schema(:tinyint).first.last[:type].should == :boolean
+    Sequel.convert_tinyint_to_bool = false
+    @db.schema(:tinyint, :reload=>true).first.last[:type].should == :integer
+    @db.schema(:interval).first.last[:type].should == :interval
+    @db.schema(:int).first.last[:type].should == :integer
+    @db.schema(:integer).first.last[:type].should == :integer
+    @db.schema(:bigint).first.last[:type].should == :integer
+    @db.schema(:smallint).first.last[:type].should == :integer
+    @db.schema(:character).first.last[:type].should == :string
+    @db.schema(:"character varying").first.last[:type].should == :string
+    @db.schema(:varchar).first.last[:type].should == :string
+    @db.schema(:"varchar(255)").first.last[:type].should == :string
+    @db.schema(:text).first.last[:type].should == :string
+    @db.schema(:date).first.last[:type].should == :date
+    @db.schema(:datetime).first.last[:type].should == :datetime
+    @db.schema(:timestamp).first.last[:type].should == :datetime
+    @db.schema(:"timestamp with time zone").first.last[:type].should == :datetime
+    @db.schema(:"timestamp without time zone").first.last[:type].should == :datetime
+    @db.schema(:time).first.last[:type].should == :time
+    @db.schema(:"time with time zone").first.last[:type].should == :time
+    @db.schema(:"time without time zone").first.last[:type].should == :time
+    @db.schema(:boolean).first.last[:type].should == :boolean
+    @db.schema(:real).first.last[:type].should == :float
+    @db.schema(:float).first.last[:type].should == :float
+    @db.schema(:double).first.last[:type].should == :float
+    @db.schema(:"double precision").first.last[:type].should == :float
+    @db.schema(:numeric).first.last[:type].should == :decimal
+    @db.schema(:decimal).first.last[:type].should == :decimal
+    @db.schema(:money).first.last[:type].should == :decimal
+    @db.schema(:bytea).first.last[:type].should == :blob
+  end
+end
diff -pruN 1.5.1-1/spec/core/spec_helper.rb 3.0.0-1/spec/core/spec_helper.rb
--- 1.5.1-1/spec/core/spec_helper.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/core/spec_helper.rb	2009-04-09 17:59:01.000000000 +0100
@@ -0,0 +1,57 @@
+require 'rubygems'
+unless Object.const_defined?('Sequel')
+  $:.unshift(File.join(File.dirname(__FILE__), "../../lib/"))
+  require 'sequel/core'
+end
+
+Sequel.virtual_row_instance_eval = true
+
+class MockDataset < Sequel::Dataset
+  def insert(*args)
+    @db.execute insert_sql(*args)
+  end
+  
+  def update(*args)
+    @db.execute update_sql(*args)
+  end
+  
+  def fetch_rows(sql)
+    @db.execute(sql)
+    yield({:id => 1, :x => 1})
+  end
+
+  def quoted_identifier(c)
+    "\"#{c}\""
+  end
+end
+
+class MockDatabase < Sequel::Database
+  @@quote_identifiers = false
+  self.identifier_input_method = nil
+  self.identifier_output_method = nil
+  attr_reader :sqls
+  
+  def execute(sql, opts={})
+    @sqls ||= []
+    @sqls << sql
+  end
+
+  def reset
+    @sqls = []
+  end
+
+  def transaction(opts={}); yield; end
+  
+  def dataset; MockDataset.new(self); end
+end
+
+class SchemaDummyDatabase < Sequel::Database
+  attr_reader :sqls
+  self.identifier_input_method = nil
+  self.identifier_output_method = nil
+  
+  def execute(sql, opts={})
+    @sqls ||= []
+    @sqls << sql
+  end
+end
diff -pruN 1.5.1-1/spec/core/version_spec.rb 3.0.0-1/spec/core/version_spec.rb
--- 1.5.1-1/spec/core/version_spec.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/core/version_spec.rb	2009-03-16 20:53:19.000000000 +0000
@@ -0,0 +1,7 @@
+require File.join(File.dirname(__FILE__), 'spec_helper')
+
+context "Sequel.version" do
+  specify "should be in the form X.Y.Z with all being numbers" do
+    Sequel.version.should =~ /\A\d+\.\d+\.\d+\z/
+  end
+end
diff -pruN 1.5.1-1/spec/deprecated_relations_spec.rb 3.0.0-1/spec/deprecated_relations_spec.rb
--- 1.5.1-1/spec/deprecated_relations_spec.rb	2008-04-22 20:16:00.000000000 +0100
+++ 3.0.0-1/spec/deprecated_relations_spec.rb	1970-01-01 01:00:00.000000000 +0100
@@ -1,113 +0,0 @@
-require File.join(File.dirname(__FILE__), "spec_helper")
-
-describe Sequel::Model, "one_to_one" do
-
-  before(:each) do
-    MODEL_DB.reset
-
-    @c1 = Class.new(Sequel::Model(:attributes)) do
-    end
-
-    @c2 = Class.new(Sequel::Model(:nodes)) do
-      columns :id, :parent_id, :blah
-    end
-
-    @dataset = @c2.dataset
-
-    @dataset.extend(Module.new {
-      def fetch_rows(sql)
-        @db << sql
-        yield({:hey => 1})
-      end
-    })
-  end
-
-  it "should use implicit key if omitted" do
-    @c2.one_to_one :parent, :from => @c2
-
-    d = @c2.new(:id => 1, :parent_id => 234)
-    p = d.parent
-    p.class.should == @c2
-    p.values.should == {:hey => 1}
-
-    MODEL_DB.sqls.should == ["SELECT * FROM nodes WHERE (id = 234) LIMIT 1"]
-  end
-
-  it "should use explicit key if given" do
-    @c2.one_to_one :parent, :from => @c2, :key => :blah
-
-    d = @c2.new(:id => 1, :blah => 567)
-    p = d.parent
-    p.class.should == @c2
-    p.values.should == {:hey => 1}
-
-    MODEL_DB.sqls.should == ["SELECT * FROM nodes WHERE (id = 567) LIMIT 1"]
-  end
-
-  it "should return nil if key value is nil" do
-    @c2.one_to_one :parent, :from => @c2
-
-    d = @c2.new(:id => 1)
-    d.parent.should == nil
-  end
-
-  it "should define a setter method" do
-    @c2.one_to_one :parent, :from => @c2
-
-    d = @c2.load(:id => 1)
-    d.parent = @c2.new(:id => 4321)
-    d.values.should == {:id => 1, :parent_id => 4321}
-    d.save_changes
-    MODEL_DB.sqls.last.should == "UPDATE nodes SET parent_id = 4321 WHERE (id = 1)"
-
-    d.parent = nil
-    d.values.should == {:id => 1, :parent_id => nil}
-    d.save_changes
-    MODEL_DB.sqls.last.should == "UPDATE nodes SET parent_id = NULL WHERE (id = 1)"
-
-    e = @c2.new(:id => 6677)
-    d.parent = e
-    d.values.should == {:id => 1, :parent_id => 6677}
-    d.save_changes
-    MODEL_DB.sqls.last.should == "UPDATE nodes SET parent_id = 6677 WHERE (id = 1)"
-  end
-end
-
-describe Sequel::Model, "one_to_many" do
-
-  before(:each) do
-    MODEL_DB.reset
-
-    @c1 = Class.new(Sequel::Model(:attributes)) do
-    end
-
-    @c2 = Class.new(Sequel::Model(:nodes)) do
-      columns :id
-    end
-  end
-
-  it "should define a getter method" do
-    @c2.one_to_many :attributes, :from => @c1, :key => :node_id
-
-    n = @c2.new(:id => 1234)
-    a = n.attributes_dataset
-    a.should be_a_kind_of(Sequel::Dataset)
-    a.sql.should == 'SELECT * FROM attributes WHERE (node_id = 1234)'
-  end
-  
-  it "should support implicit key names" do
-    $c1 = @c1
-    
-    module Music
-      class BlueNote < Sequel::Model
-        one_to_many :attributes, :from => $c1
-        columns :id
-      end
-    end
-    
-    n = Music::BlueNote.new(:id => 1234)
-    a = n.attributes_dataset
-    a.should be_a_kind_of(Sequel::Dataset)
-    a.sql.should == 'SELECT * FROM attributes WHERE (blue_note_id = 1234)'
-  end
-end
diff -pruN 1.5.1-1/spec/eager_loading_spec.rb 3.0.0-1/spec/eager_loading_spec.rb
--- 1.5.1-1/spec/eager_loading_spec.rb	2008-04-30 03:44:46.000000000 +0100
+++ 3.0.0-1/spec/eager_loading_spec.rb	1970-01-01 01:00:00.000000000 +0100
@@ -1,648 +0,0 @@
-require File.join(File.dirname(__FILE__), "spec_helper")
-
-describe Sequel::Model, "#eager" do
-  before(:each) do
-    MODEL_DB.reset
-    
-    class EagerAlbum < Sequel::Model(:albums)
-      columns :id, :band_id
-      many_to_one :band, :class=>'EagerBand', :key=>:band_id
-      one_to_many :tracks, :class=>'EagerTrack', :key=>:album_id
-      many_to_many :genres, :class=>'EagerGenre', :left_key=>:album_id, :right_key=>:genre_id, :join_table=>:ag
-    end
-
-    class EagerBand < Sequel::Model(:bands)
-      columns :id
-      one_to_many :albums, :class=>'EagerAlbum', :key=>:band_id, :eager=>:tracks
-      many_to_many :members, :class=>'EagerBandMember', :left_key=>:band_id, :right_key=>:member_id, :join_table=>:bm
-    end
-    
-    class EagerTrack < Sequel::Model(:tracks)
-      columns :id, :album_id
-      many_to_one :album, :class=>'EagerAlbum', :key=>:album_id
-    end
-    
-    class EagerGenre < Sequel::Model(:genres)
-      columns :id
-      many_to_many :albums, :class=>'EagerAlbum', :left_key=>:genre_id, :right_key=>:album_id, :join_table=>:ag
-    end
-    
-    class EagerBandMember < Sequel::Model(:members)
-      columns :id
-      many_to_many :bands, :class=>'EagerBand', :left_key=>:member_id, :right_key=>:band_id, :join_table=>:bm, :order =>:id
-    end
-    
-    EagerAlbum.dataset.extend(Module.new {
-      def fetch_rows(sql)
-        h = {:id => 1, :band_id=> 2}
-        h.merge!(:x_foreign_key_x=>4) if sql =~ /ag\.genre_id/
-        @db << sql
-        yield h
-      end
-    })
-
-    EagerBand.dataset.extend(Module.new {
-      def fetch_rows(sql)
-        h = {:id => 2}
-        h.merge!(:x_foreign_key_x=>5) if sql =~ /bm\.member_id/
-        @db << sql
-        yield h
-      end
-    })
-    
-    EagerTrack.dataset.extend(Module.new {
-      def fetch_rows(sql)
-        @db << sql
-        yield({:id => 3, :album_id => 1})
-      end
-    })
-    
-    EagerGenre.dataset.extend(Module.new {
-      def fetch_rows(sql)
-        h = {:id => 4}
-        h.merge!(:x_foreign_key_x=>1) if sql =~ /ag\.album_id/
-        @db << sql
-        yield h
-      end
-    })
-    
-    EagerBandMember.dataset.extend(Module.new {
-      def fetch_rows(sql)
-        h = {:id => 5}
-        h.merge!(:x_foreign_key_x=>2) if sql =~ /bm\.band_id/
-        @db << sql
-        yield h
-      end
-    })
-  end
-  
-  it "should eagerly load a single many_to_one association" do
-    a = EagerAlbum.eager(:band).all
-    a.should be_a_kind_of(Array)
-    a.size.should == 1
-    a.first.should be_a_kind_of(EagerAlbum)
-    a.first.values.should == {:id => 1, :band_id => 2}
-    MODEL_DB.sqls.should == ['SELECT * FROM albums', 'SELECT * FROM bands WHERE (id IN (2))']
-    a = a.first
-    a.band.should be_a_kind_of(EagerBand)
-    a.band.values.should == {:id => 2}
-    MODEL_DB.sqls.length.should == 2
-  end
-  
-  it "should eagerly load a single one_to_many association" do
-    a = EagerAlbum.eager(:tracks).all
-    a.should be_a_kind_of(Array)
-    a.size.should == 1
-    a.first.should be_a_kind_of(EagerAlbum)
-    a.first.values.should == {:id => 1, :band_id => 2}
-    MODEL_DB.sqls.should == ['SELECT * FROM albums', 'SELECT * FROM tracks WHERE (album_id IN (1))']
-    a = a.first
-    a.tracks.should be_a_kind_of(Array)
-    a.tracks.size.should == 1
-    a.tracks.first.should be_a_kind_of(EagerTrack)
-    a.tracks.first.values.should == {:id => 3, :album_id=>1}
-    MODEL_DB.sqls.length.should == 2
-  end
-  
-  it "should eagerly load a single many_to_many association" do
-    a = EagerAlbum.eager(:genres).all
-    a.should be_a_kind_of(Array)
-    a.size.should == 1
-    a.first.should be_a_kind_of(EagerAlbum)
-    a.first.values.should == {:id => 1, :band_id => 2}
-    MODEL_DB.sqls.length.should == 2
-    MODEL_DB.sqls[0].should == 'SELECT * FROM albums'
-    ["SELECT genres.*, ag.album_id AS x_foreign_key_x FROM genres INNER JOIN ag ON (ag.genre_id = genres.id) AND (ag.album_id IN (1))",
-     "SELECT genres.*, ag.album_id AS x_foreign_key_x FROM genres INNER JOIN ag ON (ag.album_id IN (1)) AND (ag.genre_id = genres.id)"
-    ].should(include(MODEL_DB.sqls[1]))
-    a = a.first
-    a.genres.should be_a_kind_of(Array)
-    a.genres.size.should == 1
-    a.genres.first.should be_a_kind_of(EagerGenre)
-    a.genres.first.values.should == {:id => 4}
-    MODEL_DB.sqls.length.should == 2
-  end
-  
-  it "should eagerly load multiple associations" do
-    a = EagerAlbum.eager(:genres, :tracks, :band).all
-    a.should be_a_kind_of(Array)
-    a.size.should == 1
-    a.first.should be_a_kind_of(EagerAlbum)
-    a.first.values.should == {:id => 1, :band_id => 2}
-    MODEL_DB.sqls.length.should == 4
-    MODEL_DB.sqls[0].should == 'SELECT * FROM albums'
-    MODEL_DB.sqls[1..-1].should(include('SELECT * FROM bands WHERE (id IN (2))'))
-    MODEL_DB.sqls[1..-1].should(include('SELECT * FROM tracks WHERE (album_id IN (1))'))
-    sqls = MODEL_DB.sqls[1..-1] - ['SELECT * FROM bands WHERE (id IN (2))', 'SELECT * FROM tracks WHERE (album_id IN (1))']
-    ["SELECT genres.*, ag.album_id AS x_foreign_key_x FROM genres INNER JOIN ag ON (ag.genre_id = genres.id) AND (ag.album_id IN (1))",
-     "SELECT genres.*, ag.album_id AS x_foreign_key_x FROM genres INNER JOIN ag ON (ag.album_id IN (1)) AND (ag.genre_id = genres.id)"
-    ].should(include(sqls[0]))
-    a = a.first
-    a.band.should be_a_kind_of(EagerBand)
-    a.band.values.should == {:id => 2}
-    a.tracks.should be_a_kind_of(Array)
-    a.tracks.size.should == 1
-    a.tracks.first.should be_a_kind_of(EagerTrack)
-    a.tracks.first.values.should == {:id => 3, :album_id=>1}
-    a.genres.should be_a_kind_of(Array)
-    a.genres.size.should == 1
-    a.genres.first.should be_a_kind_of(EagerGenre)
-    a.genres.first.values.should == {:id => 4}
-    MODEL_DB.sqls.length.should == 4
-  end
-  
-  it "should allow cascading of eager loading for associations of associated models" do
-    a = EagerTrack.eager(:album=>{:band=>:members}).all
-    a.should be_a_kind_of(Array)
-    a.size.should == 1
-    a.first.should be_a_kind_of(EagerTrack)
-    a.first.values.should == {:id => 3, :album_id => 1}
-    MODEL_DB.sqls.length.should == 4
-    MODEL_DB.sqls[0...-1].should == ['SELECT * FROM tracks', 
-                             'SELECT * FROM albums WHERE (id IN (1))',
-                             'SELECT * FROM bands WHERE (id IN (2))']
-    ["SELECT members.*, bm.band_id AS x_foreign_key_x FROM members INNER JOIN bm ON (bm.member_id = members.id) AND (bm.band_id IN (2))",
-     "SELECT members.*, bm.band_id AS x_foreign_key_x FROM members INNER JOIN bm ON (bm.band_id IN (2)) AND (bm.member_id = members.id)"
-    ].should(include(MODEL_DB.sqls[-1]))
-    a = a.first
-    a.album.should be_a_kind_of(EagerAlbum)
-    a.album.values.should == {:id => 1, :band_id => 2}
-    a.album.band.should be_a_kind_of(EagerBand)
-    a.album.band.values.should == {:id => 2}
-    a.album.band.members.should be_a_kind_of(Array)
-    a.album.band.members.size.should == 1
-    a.album.band.members.first.should be_a_kind_of(EagerBandMember)
-    a.album.band.members.first.values.should == {:id => 5}
-    MODEL_DB.sqls.length.should == 4
-  end
-  
-  it "should cascade eagerly loading when the :eager association option is used" do
-    a = EagerBand.eager(:albums).all
-    a.should be_a_kind_of(Array)
-    a.size.should == 1
-    a.first.should be_a_kind_of(EagerBand)
-    a.first.values.should == {:id => 2}
-    MODEL_DB.sqls.should == ['SELECT * FROM bands', 
-                             'SELECT * FROM albums WHERE (band_id IN (2))',
-                             'SELECT * FROM tracks WHERE (album_id IN (1))']
-    a = a.first
-    a.albums.should be_a_kind_of(Array)
-    a.albums.size.should == 1
-    a.albums.first.should be_a_kind_of(EagerAlbum)
-    a.albums.first.values.should == {:id => 1, :band_id => 2}
-    a = a.albums.first
-    a.tracks.should be_a_kind_of(Array)
-    a.tracks.size.should == 1
-    a.tracks.first.should be_a_kind_of(EagerTrack)
-    a.tracks.first.values.should == {:id => 3, :album_id => 1}
-    MODEL_DB.sqls.length.should == 3
-  end
-  
-  it "should respect :eager when lazily loading an association" do
-    a = EagerBand.all
-    a.should be_a_kind_of(Array)
-    a.size.should == 1
-    a.first.should be_a_kind_of(EagerBand)
-    a.first.values.should == {:id => 2}
-    MODEL_DB.sqls.should == ['SELECT * FROM bands']
-    a = a.first
-    a.albums
-    MODEL_DB.sqls.should == ['SELECT * FROM bands', 
-                             'SELECT * FROM albums WHERE (band_id = 2)',
-                             'SELECT * FROM tracks WHERE (album_id IN (1))']
-    a.albums.should be_a_kind_of(Array)
-    a.albums.size.should == 1
-    a.albums.first.should be_a_kind_of(EagerAlbum)
-    a.albums.first.values.should == {:id => 1, :band_id => 2}
-    a = a.albums.first
-    a.tracks.should be_a_kind_of(Array)
-    a.tracks.size.should == 1
-    a.tracks.first.should be_a_kind_of(EagerTrack)
-    a.tracks.first.values.should == {:id => 3, :album_id => 1}
-    MODEL_DB.sqls.length.should == 3
-  end
-  
-  it "should respect :order when eagerly loading" do
-    a = EagerBandMember.eager(:bands).all
-    a.should be_a_kind_of(Array)
-    a.size.should == 1
-    a.first.should be_a_kind_of(EagerBandMember)
-    a.first.values.should == {:id => 5}
-    MODEL_DB.sqls.length.should == 2
-    MODEL_DB.sqls[0].should == 'SELECT * FROM members'
-    ['SELECT bands.*, bm.member_id AS x_foreign_key_x FROM bands INNER JOIN bm ON (bm.band_id = bands.id) AND (bm.member_id IN (5)) ORDER BY id',
-     'SELECT bands.*, bm.member_id AS x_foreign_key_x FROM bands INNER JOIN bm ON (bm.member_id IN (5)) AND (bm.band_id = bands.id) ORDER BY id'
-    ].should(include(MODEL_DB.sqls[1]))
-    a = a.first
-    a.bands.should be_a_kind_of(Array)
-    a.bands.size.should == 1
-    a.bands.first.should be_a_kind_of(EagerBand)
-    a.bands.first.values.should == {:id => 2}
-    MODEL_DB.sqls.length.should == 2
-  end
-  
-  it "should populate the reciprocal many_to_one association when eagerly loading the one_to_many association" do
-    a = EagerAlbum.eager(:tracks).all
-    a.should be_a_kind_of(Array)
-    a.size.should == 1
-    a.first.should be_a_kind_of(EagerAlbum)
-    a.first.values.should == {:id => 1, :band_id => 2}
-    MODEL_DB.sqls.should == ['SELECT * FROM albums', 'SELECT * FROM tracks WHERE (album_id IN (1))']
-    a = a.first
-    a.tracks.should be_a_kind_of(Array)
-    a.tracks.size.should == 1
-    a.tracks.first.should be_a_kind_of(EagerTrack)
-    a.tracks.first.values.should == {:id => 3, :album_id=>1}
-    a.tracks.first.album.should be_a_kind_of(EagerAlbum)
-    a.tracks.first.album.should == a
-    MODEL_DB.sqls.length.should == 2
-  end
-end
-
-describe Sequel::Model, "#eager_graph" do
-  after(:all) do
-    class MockDataset
-      alias clone orig_clone
-    end
-  end
-
-  before(:all) do
-    class MockDataset
-      alias orig_clone clone
-      def clone(opts = {})
-        c = super()
-        c.opts = @opts.merge(opts)
-        c.instance_variable_set(:@columns, (@columns.dup if @columns))
-        c
-      end
-    end
-
-    class GraphAlbum < Sequel::Model(:albums)
-      dataset.opts[:from] = [:albums]
-      columns :id, :band_id
-      many_to_one :band, :class=>'GraphBand', :key=>:band_id
-      one_to_many :tracks, :class=>'GraphTrack', :key=>:album_id
-      many_to_many :genres, :class=>'GraphGenre', :left_key=>:album_id, :right_key=>:genre_id, :join_table=>:ag
-    end
-
-    class GraphBand < Sequel::Model(:bands)
-      dataset.opts[:from] = [:bands]
-      columns :id, :vocalist_id
-      many_to_one :vocalist, :class=>'GraphBandMember', :key=>:vocalist_id
-      one_to_many :albums, :class=>'GraphAlbum', :key=>:band_id
-      many_to_many :members, :class=>'GraphBandMember', :left_key=>:band_id, :right_key=>:member_id, :join_table=>:bm
-      many_to_many :genres, :class=>'GraphGenre', :left_key=>:band_id, :right_key=>:genre_id, :join_table=>:bg
-    end
-    
-    class GraphTrack < Sequel::Model(:tracks)
-      dataset.opts[:from] = [:tracks]
-      columns :id, :album_id
-      many_to_one :album, :class=>'GraphAlbum', :key=>:album_id
-    end
-    
-    class GraphGenre < Sequel::Model(:genres)
-      dataset.opts[:from] = [:genres]
-      columns :id
-      many_to_many :albums, :class=>'GraphAlbum', :left_key=>:genre_id, :right_key=>:album_id, :join_table=>:ag
-    end
-    
-    class GraphBandMember < Sequel::Model(:members)
-      dataset.opts[:from] = [:members]
-      columns :id
-      many_to_many :bands, :class=>'GraphBand', :left_key=>:member_id, :right_key=>:band_id, :join_table=>:bm
-    end
-  end
-    
-  it "should eagerly load a single many_to_one association" do
-    ds = GraphAlbum.eager_graph(:band)
-    ds.sql.should == 'SELECT albums.id, albums.band_id, band.id AS band_id_0, band.vocalist_id FROM albums LEFT OUTER JOIN bands band ON (band.id = albums.band_id)'
-    def ds.fetch_rows(sql, &block)
-      yield({:id=>1, :band_id=>2, :band_id_0=>2, :vocalist_id=>3})
-    end
-    a = ds.all
-    a.should be_a_kind_of(Array)
-    a.size.should == 1
-    a.first.should be_a_kind_of(GraphAlbum)
-    a.first.values.should == {:id => 1, :band_id => 2}
-    a = a.first
-    a.band.should be_a_kind_of(GraphBand)
-    a.band.values.should == {:id => 2, :vocalist_id=>3}
-  end
-
-  it "should eagerly load a single one_to_many association" do
-    ds = GraphAlbum.eager_graph(:tracks)
-    ds.sql.should == 'SELECT albums.id, albums.band_id, tracks.id AS tracks_id, tracks.album_id FROM albums LEFT OUTER JOIN tracks ON (tracks.album_id = albums.id)'
-    def ds.fetch_rows(sql, &block)
-      yield({:id=>1, :band_id=>2, :tracks_id=>3, :album_id=>1})
-    end
-    a = ds.all
-    a.should be_a_kind_of(Array)
-    a.size.should == 1
-    a.first.should be_a_kind_of(GraphAlbum)
-    a.first.values.should == {:id => 1, :band_id => 2}
-    a = a.first
-    a.tracks.should be_a_kind_of(Array)
-    a.tracks.size.should == 1
-    a.tracks.first.should be_a_kind_of(GraphTrack)
-    a.tracks.first.values.should == {:id => 3, :album_id=>1}
-  end
-
-  it "should eagerly load a single many_to_many association" do
-    ds = GraphAlbum.eager_graph(:genres)
-    ds.sql.should == 'SELECT albums.id, albums.band_id, genres.id AS genres_id FROM albums LEFT OUTER JOIN ag ON (ag.album_id = albums.id) LEFT OUTER JOIN genres ON (genres.id = ag.genre_id)'
-    def ds.fetch_rows(sql, &block)
-      yield({:id=>1, :band_id=>2, :genres_id=>4})
-    end
-    a = ds.all
-    a.should be_a_kind_of(Array)
-    a.size.should == 1
-    a.first.should be_a_kind_of(GraphAlbum)
-    a.first.values.should == {:id => 1, :band_id => 2}
-    a = a.first
-    a.genres.should be_a_kind_of(Array)
-    a.genres.size.should == 1
-    a.genres.first.should be_a_kind_of(GraphGenre)
-    a.genres.first.values.should == {:id => 4}
-  end
-
-  it "should eagerly load multiple associations" do 
-    ds = GraphAlbum.eager_graph(:genres, :tracks, :band)
-    ds.sql.should == 'SELECT albums.id, albums.band_id, genres.id AS genres_id, tracks.id AS tracks_id, tracks.album_id, band.id AS band_id_0, band.vocalist_id FROM albums LEFT OUTER JOIN ag ON (ag.album_id = albums.id) LEFT OUTER JOIN genres ON (genres.id = ag.genre_id) LEFT OUTER JOIN tracks ON (tracks.album_id = albums.id) LEFT OUTER JOIN bands band ON (band.id = albums.band_id)'
-    def ds.fetch_rows(sql, &block)
-      yield({:id=>1, :band_id=>2, :genres_id=>4, :tracks_id=>3, :album_id=>1, :band_id_0=>2, :vocalist_id=>6})
-    end
-    a = ds.all
-    a.should be_a_kind_of(Array)
-    a.size.should == 1
-    a.first.should be_a_kind_of(GraphAlbum)
-    a.first.values.should == {:id => 1, :band_id => 2}
-    a = a.first
-    a.band.should be_a_kind_of(GraphBand)
-    a.band.values.should == {:id => 2, :vocalist_id=>6}
-    a.tracks.should be_a_kind_of(Array)
-    a.tracks.size.should == 1
-    a.tracks.first.should be_a_kind_of(GraphTrack)
-    a.tracks.first.values.should == {:id => 3, :album_id=>1}
-    a.genres.should be_a_kind_of(Array)
-    a.genres.size.should == 1
-    a.genres.first.should be_a_kind_of(GraphGenre)
-    a.genres.first.values.should == {:id => 4}
-  end
-
-  it "should allow cascading of eager loading for associations of associated models" do
-    ds = GraphTrack.eager_graph(:album=>{:band=>:members})
-    ds.sql.should == 'SELECT tracks.id, tracks.album_id, album.id AS album_id_0, album.band_id, band.id AS band_id_0, band.vocalist_id, members.id AS members_id FROM tracks LEFT OUTER JOIN albums album ON (album.id = tracks.album_id) LEFT OUTER JOIN bands band ON (band.id = album.band_id) LEFT OUTER JOIN bm ON (bm.band_id = band.id) LEFT OUTER JOIN members ON (members.id = bm.member_id)'
-    def ds.fetch_rows(sql, &block)
-      yield({:id=>3, :album_id=>1, :album_id_0=>1, :band_id=>2, :members_id=>5, :band_id_0=>2, :vocalist_id=>6})
-    end
-    a = ds.all
-    a.should be_a_kind_of(Array)
-    a.size.should == 1
-    a.first.should be_a_kind_of(GraphTrack)
-    a.first.values.should == {:id => 3, :album_id => 1}
-    a = a.first
-    a.album.should be_a_kind_of(GraphAlbum)
-    a.album.values.should == {:id => 1, :band_id => 2}
-    a.album.band.should be_a_kind_of(GraphBand)
-    a.album.band.values.should == {:id => 2, :vocalist_id=>6}
-    a.album.band.members.should be_a_kind_of(Array)
-    a.album.band.members.size.should == 1
-    a.album.band.members.first.should be_a_kind_of(GraphBandMember)
-    a.album.band.members.first.values.should == {:id => 5}
-  end
-  
-  it "should populate the reciprocal many_to_one association when eagerly loading the one_to_many association" do
-    MODEL_DB.reset
-    ds = GraphAlbum.eager_graph(:tracks)
-    ds.sql.should == 'SELECT albums.id, albums.band_id, tracks.id AS tracks_id, tracks.album_id FROM albums LEFT OUTER JOIN tracks ON (tracks.album_id = albums.id)'
-    def ds.fetch_rows(sql, &block)
-      @db << sql
-      yield({:id=>1, :band_id=>2, :tracks_id=>3, :album_id=>1})
-    end
-    a = ds.all
-    a.should be_a_kind_of(Array)
-    a.size.should == 1
-    a.first.should be_a_kind_of(GraphAlbum)
-    a.first.values.should == {:id => 1, :band_id => 2}
-    a = a.first
-    a.tracks.should be_a_kind_of(Array)
-    a.tracks.size.should == 1
-    a.tracks.first.should be_a_kind_of(GraphTrack)
-    a.tracks.first.values.should == {:id => 3, :album_id=>1}
-    a.tracks.first.album.should be_a_kind_of(GraphAlbum)
-    a.tracks.first.album.should == a
-    MODEL_DB.sqls.length.should == 1
-  end
-
-  it "should eager load multiple associations from the same table" do
-    ds = GraphBand.eager_graph(:vocalist, :members)
-    ds.sql.should == 'SELECT bands.id, bands.vocalist_id, vocalist.id AS vocalist_id_0, members.id AS members_id FROM bands LEFT OUTER JOIN members vocalist ON (vocalist.id = bands.vocalist_id) LEFT OUTER JOIN bm ON (bm.band_id = bands.id) LEFT OUTER JOIN members ON (members.id = bm.member_id)'
-    def ds.fetch_rows(sql, &block)
-      yield({:id=>2, :vocalist_id=>6, :vocalist_id_0=>6, :members_id=>5})
-    end
-    a = ds.all
-    a.should be_a_kind_of(Array)
-    a.size.should == 1
-    a.first.should be_a_kind_of(GraphBand)
-    a.first.values.should == {:id => 2, :vocalist_id => 6}
-    a = a.first
-    a.vocalist.should be_a_kind_of(GraphBandMember)
-    a.vocalist.values.should == {:id => 6}
-    a.members.should be_a_kind_of(Array)
-    a.members.size.should == 1
-    a.members.first.should be_a_kind_of(GraphBandMember)
-    a.members.first.values.should == {:id => 5}
-  end
-
-  it "should give you a graph of tables when called without .all" do 
-    ds = GraphAlbum.eager_graph(:band)
-    ds.sql.should == 'SELECT albums.id, albums.band_id, band.id AS band_id_0, band.vocalist_id FROM albums LEFT OUTER JOIN bands band ON (band.id = albums.band_id)'
-    def ds.fetch_rows(sql, &block)
-      yield({:id=>1, :band_id=>2, :band_id_0=>2, :vocalist_id=>3})
-    end
-    ds.first.should == {:albums=>GraphAlbum.new(:id => 1, :band_id => 2), :band=>GraphBand.new(:id => 2, :vocalist_id=>3)}
-  end
-
-  it "should not drop any associated objects if the graph could not be a cartesian product" do
-    ds = GraphBand.eager_graph(:members, :vocalist)
-    ds.sql.should == 'SELECT bands.id, bands.vocalist_id, members.id AS members_id, vocalist.id AS vocalist_id_0 FROM bands LEFT OUTER JOIN bm ON (bm.band_id = bands.id) LEFT OUTER JOIN members ON (members.id = bm.member_id) LEFT OUTER JOIN members vocalist ON (vocalist.id = bands.vocalist_id)'
-    def ds.fetch_rows(sql, &block)
-      yield({:id=>2, :vocalist_id=>6, :members_id=>5, :vocalist_id_0=>6})
-      yield({:id=>2, :vocalist_id=>6, :members_id=>5, :vocalist_id_0=>6})
-    end
-    a = ds.all
-    a.should be_a_kind_of(Array)
-    a.size.should == 1
-    a.first.should be_a_kind_of(GraphBand)
-    a.first.values.should == {:id => 2, :vocalist_id => 6}
-    a = a.first
-    a.vocalist.should be_a_kind_of(GraphBandMember)
-    a.vocalist.values.should == {:id => 6}
-    a.members.should be_a_kind_of(Array)
-    a.members.size.should == 2
-    a.members.first.should be_a_kind_of(GraphBandMember)
-    a.members.first.values.should == {:id => 5}
-    a.members.last.should be_a_kind_of(GraphBandMember)
-    a.members.last.values.should == {:id => 5}
-  end
-
-  it "should drop duplicate items that occur in sequence if the graph could be a cartesian product" do
-    ds = GraphBand.eager_graph(:members, :genres)
-    ds.sql.should == 'SELECT bands.id, bands.vocalist_id, members.id AS members_id, genres.id AS genres_id FROM bands LEFT OUTER JOIN bm ON (bm.band_id = bands.id) LEFT OUTER JOIN members ON (members.id = bm.member_id) LEFT OUTER JOIN bg ON (bg.band_id = bands.id) LEFT OUTER JOIN genres ON (genres.id = bg.genre_id)'
-    def ds.fetch_rows(sql, &block)
-      yield({:id=>2, :vocalist_id=>6, :members_id=>5, :genres_id=>7})
-      yield({:id=>2, :vocalist_id=>6, :members_id=>5, :genres_id=>8})
-      yield({:id=>2, :vocalist_id=>6, :members_id=>6, :genres_id=>7})
-      yield({:id=>2, :vocalist_id=>6, :members_id=>6, :genres_id=>8})
-    end
-    a = ds.all
-    a.should be_a_kind_of(Array)
-    a.size.should == 1
-    a.first.should be_a_kind_of(GraphBand)
-    a.first.values.should == {:id => 2, :vocalist_id => 6}
-    a = a.first
-    a.members.should be_a_kind_of(Array)
-    a.members.size.should == 2
-    a.members.first.should be_a_kind_of(GraphBandMember)
-    a.members.first.values.should == {:id => 5}
-    a.members.last.should be_a_kind_of(GraphBandMember)
-    a.members.last.values.should == {:id => 6}
-    a.genres.size.should == 2
-    a.genres.first.should be_a_kind_of(GraphGenre)
-    a.genres.first.values.should == {:id => 7}
-    a.genres.last.should be_a_kind_of(GraphGenre)
-    a.genres.last.values.should == {:id => 8}
-  end
-
-  it "should be able to be used in combination with #eager" do
-    MODEL_DB.reset
-    ds = GraphAlbum.eager_graph(:tracks).eager(:genres)
-    def ds.fetch_rows(sql, &block)
-      @db << sql
-      yield({:id=>1, :band_id=>2, :tracks_id=>3, :album_id=>1})
-    end
-    ds2 = GraphGenre.dataset
-    def ds2.fetch_rows(sql, &block)
-      @db << sql
-      yield({:id=>6, :x_foreign_key_x=>1})
-    end
-    a = ds.all
-    a.should be_a_kind_of(Array)
-    a.size.should == 1
-    a.first.should be_a_kind_of(GraphAlbum)
-    a.first.values.should == {:id => 1, :band_id => 2}
-    a = a.first
-    a.tracks.should be_a_kind_of(Array)
-    a.tracks.size.should == 1
-    a.tracks.first.should be_a_kind_of(GraphTrack)
-    a.tracks.first.values.should == {:id=>3, :album_id=>1}
-    a.genres.should be_a_kind_of(Array)
-    a.genres.size.should == 1
-    a.genres.first.should be_a_kind_of(GraphGenre)
-    a.genres.first.values.should == {:id=>6}
-    MODEL_DB.sqls[0].should == 'SELECT albums.id, albums.band_id, tracks.id AS tracks_id, tracks.album_id FROM albums LEFT OUTER JOIN tracks ON (tracks.album_id = albums.id)'
-    ["SELECT genres.*, ag.album_id AS x_foreign_key_x FROM genres INNER JOIN ag ON (ag.genre_id = genres.id) AND (ag.album_id IN (1))",
-     "SELECT genres.*, ag.album_id AS x_foreign_key_x FROM genres INNER JOIN ag ON (ag.album_id IN (1)) AND (ag.genre_id = genres.id)"
-    ].should(include(MODEL_DB.sqls[1]))
-  end
-
-  it "should handle no associated records for a single many_to_one association" do
-    ds = GraphAlbum.eager_graph(:band)
-    ds.sql.should == 'SELECT albums.id, albums.band_id, band.id AS band_id_0, band.vocalist_id FROM albums LEFT OUTER JOIN bands band ON (band.id = albums.band_id)'
-    def ds.fetch_rows(sql, &block)
-      yield({:id=>1, :band_id=>2, :band_id_0=>nil, :vocalist_id=>nil})
-    end
-    a = ds.all
-    a.should be_a_kind_of(Array)
-    a.size.should == 1
-    a.first.should be_a_kind_of(GraphAlbum)
-    a.first.values.should == {:id => 1, :band_id => 2}
-    a.first.instance_variable_get(:@band).should == :null
-  end
-
-  it "should handle no associated records for a single one_to_many association" do
-    ds = GraphAlbum.eager_graph(:tracks)
-    ds.sql.should == 'SELECT albums.id, albums.band_id, tracks.id AS tracks_id, tracks.album_id FROM albums LEFT OUTER JOIN tracks ON (tracks.album_id = albums.id)'
-    def ds.fetch_rows(sql, &block)
-      yield({:id=>1, :band_id=>2, :tracks_id=>nil, :album_id=>nil})
-    end
-    a = ds.all
-    a.should be_a_kind_of(Array)
-    a.size.should == 1
-    a.first.should be_a_kind_of(GraphAlbum)
-    a.first.values.should == {:id => 1, :band_id => 2}
-    a.first.tracks.should == []
-  end
-
-  it "should handle no associated records for a single many_to_many association" do
-    ds = GraphAlbum.eager_graph(:genres)
-    ds.sql.should == 'SELECT albums.id, albums.band_id, genres.id AS genres_id FROM albums LEFT OUTER JOIN ag ON (ag.album_id = albums.id) LEFT OUTER JOIN genres ON (genres.id = ag.genre_id)'
-    def ds.fetch_rows(sql, &block)
-      yield({:id=>1, :band_id=>2, :genres_id=>nil})
-    end
-    a = ds.all
-    a.should be_a_kind_of(Array)
-    a.size.should == 1
-    a.first.should be_a_kind_of(GraphAlbum)
-    a.first.values.should == {:id => 1, :band_id => 2}
-    a.first.genres.should == []
-  end
-
-  it "should handle missing associated records when loading multiple associations" do 
-    ds = GraphAlbum.eager_graph(:genres, :tracks, :band)
-    ds.sql.should == 'SELECT albums.id, albums.band_id, genres.id AS genres_id, tracks.id AS tracks_id, tracks.album_id, band.id AS band_id_0, band.vocalist_id FROM albums LEFT OUTER JOIN ag ON (ag.album_id = albums.id) LEFT OUTER JOIN genres ON (genres.id = ag.genre_id) LEFT OUTER JOIN tracks ON (tracks.album_id = albums.id) LEFT OUTER JOIN bands band ON (band.id = albums.band_id)'
-    def ds.fetch_rows(sql, &block)
-      yield({:id=>1, :band_id=>2, :genres_id=>nil, :tracks_id=>3, :album_id=>1, :band_id_0=>nil, :vocalist_id=>nil})
-      yield({:id=>1, :band_id=>2, :genres_id=>nil, :tracks_id=>4, :album_id=>1, :band_id_0=>nil, :vocalist_id=>nil})
-      yield({:id=>1, :band_id=>2, :genres_id=>nil, :tracks_id=>5, :album_id=>1, :band_id_0=>nil, :vocalist_id=>nil})
-      yield({:id=>1, :band_id=>2, :genres_id=>nil, :tracks_id=>6, :album_id=>1, :band_id_0=>nil, :vocalist_id=>nil})
-    end
-    a = ds.all
-    a.should be_a_kind_of(Array)
-    a.size.should == 1
-    a.first.should be_a_kind_of(GraphAlbum)
-    a.first.values.should == {:id => 1, :band_id => 2}
-    a = a.first
-    a.tracks.should be_a_kind_of(Array)
-    a.tracks.size.should == 4
-    a.tracks.first.should be_a_kind_of(GraphTrack)
-    a.tracks.collect{|x|x[:id]}.should == [3,4,5,6]
-    a.instance_variable_get(:@band).should == :null
-    a.genres.should == []
-  end
-
-  it "should handle missing associated records when cascading eager loading for associations of associated models" do
-    ds = GraphTrack.eager_graph(:album=>{:band=>:members})
-    ds.sql.should == 'SELECT tracks.id, tracks.album_id, album.id AS album_id_0, album.band_id, band.id AS band_id_0, band.vocalist_id, members.id AS members_id FROM tracks LEFT OUTER JOIN albums album ON (album.id = tracks.album_id) LEFT OUTER JOIN bands band ON (band.id = album.band_id) LEFT OUTER JOIN bm ON (bm.band_id = band.id) LEFT OUTER JOIN members ON (members.id = bm.member_id)'
-    def ds.fetch_rows(sql, &block)
-      yield({:id=>2, :album_id=>2, :album_id_0=>nil, :band_id=>nil, :members_id=>nil, :band_id_0=>nil, :vocalist_id=>nil})
-      yield({:id=>3, :album_id=>3, :album_id_0=>3, :band_id=>3, :members_id=>nil, :band_id_0=>nil, :vocalist_id=>nil})
-      yield({:id=>4, :album_id=>4, :album_id_0=>4, :band_id=>2, :members_id=>nil, :band_id_0=>2, :vocalist_id=>6})
-      yield({:id=>5, :album_id=>1, :album_id_0=>1, :band_id=>4, :members_id=>5, :band_id_0=>4, :vocalist_id=>8})
-      yield({:id=>5, :album_id=>1, :album_id_0=>1, :band_id=>4, :members_id=>6, :band_id_0=>4, :vocalist_id=>8})
-    end
-    a = ds.all
-    a.should be_a_kind_of(Array)
-    a.size.should == 4
-    a.first.should be_a_kind_of(GraphTrack)
-    a.collect{|x|x[:id]}.should == [2,3,4,5]
-    a[0].instance_variable_get(:@album).should == :null
-    a[1].album.should be_a_kind_of(GraphAlbum)
-    a[1].album.values.should == {:id => 3, :band_id => 3}
-    a[1].album.instance_variable_get(:@band).should == :null
-    a[2].album.should be_a_kind_of(GraphAlbum)
-    a[2].album.values.should == {:id => 4, :band_id => 2}
-    a[2].album.band.should be_a_kind_of(GraphBand)
-    a[2].album.band.values.should == {:id => 2, :vocalist_id=>6}
-    a[2].album.band.members.should == []
-    a[3].album.should be_a_kind_of(GraphAlbum)
-    a[3].album.values.should == {:id => 1, :band_id => 4}
-    a[3].album.band.should be_a_kind_of(GraphBand)
-    a[3].album.band.values.should == {:id => 4, :vocalist_id=>8}
-    a[3].album.band.members.size.should == 2
-    a[3].album.band.members.first.should be_a_kind_of(GraphBandMember)
-    a[3].album.band.members.first.values.should == {:id => 5}
-    a[3].album.band.members.last.should be_a_kind_of(GraphBandMember)
-    a[3].album.band.members.last.values.should == {:id => 6}
-  end
-end
diff -pruN 1.5.1-1/spec/extensions/blank_spec.rb 3.0.0-1/spec/extensions/blank_spec.rb
--- 1.5.1-1/spec/extensions/blank_spec.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/extensions/blank_spec.rb	2009-03-16 20:53:19.000000000 +0000
@@ -0,0 +1,67 @@
+require File.join(File.dirname(__FILE__), 'spec_helper')
+
+context "Object#blank?" do
+  specify "it should be true if the object responds true to empty?" do
+    [].blank?.should == true
+    {}.blank?.should == true
+    o = Object.new
+    def o.empty?; true; end
+    o.blank?.should == true
+  end
+
+  specify "it should be false if the object doesn't respond true to empty?" do
+    [2].blank?.should == false
+    {1=>2}.blank?.should == false
+    Object.new.blank?.should == false
+  end
+end
+
+context "Numeric#blank?" do
+  specify "it should always be false" do
+    1.blank?.should == false
+    0.blank?.should == false
+    -1.blank?.should == false
+    1.0.blank?.should == false
+    0.0.blank?.should == false
+    -1.0.blank?.should == false
+    10000000000000000.blank?.should == false
+    -10000000000000000.blank?.should == false
+    10000000000000000.0.blank?.should == false
+    -10000000000000000.0.blank?.should == false
+  end
+end
+
+context "NilClass#blank?" do
+  specify "it should always be true" do
+    nil.blank?.should == true
+  end
+end
+
+context "TrueClass#blank?" do
+  specify "it should always be false" do
+    true.blank?.should == false
+  end
+end
+
+context "FalseClass#blank?" do
+  specify "it should always be true" do
+    false.blank?.should == true
+  end
+end
+
+context "String#blank?" do
+  specify "it should be true if the string is empty" do
+    ''.blank?.should == true
+  end
+  specify "it should be true if the string is composed of just whitespace" do
+    ' '.blank?.should == true
+    "\r\n\t".blank?.should == true
+    (' '*4000).blank?.should == true
+    ("\r\n\t"*4000).blank?.should == true
+  end
+  specify "it should be false if the string has any non whitespace characters" do
+    '1'.blank?.should == false
+    ("\r\n\t"*4000 + 'a').blank?.should == false
+    ("\r\na\t"*4000).blank?.should == false
+  end
+end
diff -pruN 1.5.1-1/spec/extensions/caching_spec.rb 3.0.0-1/spec/extensions/caching_spec.rb
--- 1.5.1-1/spec/extensions/caching_spec.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/extensions/caching_spec.rb	2009-03-08 04:21:44.000000000 +0000
@@ -0,0 +1,201 @@
+require File.join(File.dirname(__FILE__), "spec_helper")
+
+describe Sequel::Model, "caching" do
+  before do
+    MODEL_DB.reset
+    
+    @cache_class = Class.new(Hash) do
+      attr_accessor :ttl
+      def set(k, v, ttl); self[k] = v; @ttl = ttl; end
+      def get(k); self[k]; end
+    end
+    cache = @cache_class.new
+    @cache = cache
+    
+    @c = Class.new(Sequel::Model(:items))
+    @c.class_eval do
+      plugin :caching, cache
+      def self.name; 'Item' end
+      
+      columns :name, :id
+    end
+    
+    $cache_dataset_row = {:name => 'sharon', :id => 1}
+    @dataset = @c.dataset
+    $sqls = []
+    @dataset.extend(Module.new {
+      def fetch_rows(sql)
+        $sqls << sql
+        yield $cache_dataset_row
+      end
+      
+      def update(values)
+        $sqls << update_sql(values)
+        $cache_dataset_row.merge!(values)
+      end
+      
+      def delete
+        $sqls << delete_sql
+      end
+    })
+    @c2 = Class.new(@c) do
+      def self.name; 'SubItem' end
+    end
+  end
+  
+  it "should set the model's cache store" do
+    @c.cache_store.should be(@cache)
+    @c2.cache_store.should be(@cache)
+  end
+  
+  it "should have a default ttl of 3600" do
+    @c.cache_ttl.should == 3600
+    @c2.cache_ttl.should == 3600
+  end
+  
+  it "should take a ttl option" do
+    @c.plugin :caching, @cache, :ttl => 1234
+    @c.cache_ttl.should == 1234
+    Class.new(@c).cache_ttl.should == 1234
+  end
+  
+  it "should offer a set_cache_ttl method for setting the ttl" do
+    @c.cache_ttl.should == 3600
+    @c.set_cache_ttl 1234
+    @c.cache_ttl.should == 1234
+    Class.new(@c).cache_ttl.should == 1234
+  end
+  
+  it "should generate a cache key appropriate to the class" do
+    m = @c.new
+    m.values[:id] = 1
+    m.cache_key.should == "#{m.class}:1"
+    m = @c2.new
+    m.values[:id] = 1
+    m.cache_key.should == "#{m.class}:1"
+    
+    # custom primary key
+    @c.set_primary_key :ttt
+    m = @c.new
+    m.values[:ttt] = 333
+    m.cache_key.should == "#{m.class}:333"
+    c = Class.new(@c)
+    m = c.new
+    m.values[:ttt] = 333
+    m.cache_key.should == "#{m.class}:333"
+    
+    # composite primary key
+    @c.set_primary_key [:a, :b, :c]
+    m = @c.new
+    m.values[:a] = 123
+    m.values[:c] = 456
+    m.values[:b] = 789
+    m.cache_key.should == "#{m.class}:123,789,456"
+    c = Class.new(@c)
+    m = c.new
+    m.values[:a] = 123
+    m.values[:c] = 456
+    m.values[:b] = 789
+    m.cache_key.should == "#{m.class}:123,789,456"
+  end
+  
+  it "should raise error if attempting to generate cache_key and primary key value is null" do
+    m = @c.new
+    proc {m.cache_key}.should raise_error(Sequel::Error)
+    m.values[:id] = 1
+    proc {m.cache_key}.should_not raise_error(Sequel::Error)
+
+    m = @c2.new
+    proc {m.cache_key}.should raise_error(Sequel::Error)
+    m.values[:id] = 1
+    proc {m.cache_key}.should_not raise_error(Sequel::Error)
+  end
+  
+  it "should not raise error if trying to save a new record" do
+    proc {@c.new(:name=>'blah').save}.should_not raise_error
+    proc {@c.create(:name=>'blah')}.should_not raise_error
+    proc {@c2.new(:name=>'blah').save}.should_not raise_error
+    proc {@c2.create(:name=>'blah')}.should_not raise_error
+  end
+  
+  it "should set the cache when reading from the database" do
+    $sqls.should == []
+    @cache.should be_empty
+    
+    m = @c[1]
+    $sqls.should == ['SELECT * FROM items WHERE (id = 1) LIMIT 1']
+    m.values.should == $cache_dataset_row
+    @cache[m.cache_key].should == m
+    m2 = @c[1]
+    $sqls.should == ['SELECT * FROM items WHERE (id = 1) LIMIT 1']
+    m2.should == m
+    m2.values.should == $cache_dataset_row
+
+    $sqls.clear
+    m = @c2[1]
+    $sqls.should == ['SELECT * FROM items WHERE (id = 1) LIMIT 1']
+    m.values.should == $cache_dataset_row
+    @cache[m.cache_key].should == m
+    m2 = @c2[1]
+    $sqls.should == ['SELECT * FROM items WHERE (id = 1) LIMIT 1']
+    m2.should == m
+    m2.values.should == $cache_dataset_row
+  end
+  
+  it "should delete the cache when writing to the database" do
+    m = @c[1]
+    @cache[m.cache_key].should == m
+    m.name = 'hey'
+    m.save
+    @cache.has_key?(m.cache_key).should be_false
+    $sqls.last.should == "UPDATE items SET name = 'hey', id = 1 WHERE (id = 1)"
+
+    m = @c2[1]
+    @cache[m.cache_key].should == m
+    m.name = 'hey'
+    m.save
+    @cache.has_key?(m.cache_key).should be_false
+    $sqls.last.should == "UPDATE items SET name = 'hey', id = 1 WHERE (id = 1)"
+  end
+
+  it "should delete the cache when deleting the record" do
+    m = @c[1]
+    @cache[m.cache_key].should == m
+    m.delete
+    @cache.has_key?(m.cache_key).should be_false
+    $sqls.last.should == "DELETE FROM items WHERE (id = 1)"
+
+    m = @c2[1]
+    @cache[m.cache_key].should == m
+    m.delete
+    @cache.has_key?(m.cache_key).should be_false
+    $sqls.last.should == "DELETE FROM items WHERE (id = 1)"
+  end
+  
+  it "should support #[] as a shortcut to #find with hash" do
+    m = @c[:id => 3]
+    @cache[m.cache_key].should be_nil
+    $sqls.last.should == "SELECT * FROM items WHERE (id = 3) LIMIT 1"
+    m = @c[1]
+    @cache[m.cache_key].should == m
+    $sqls.should == ["SELECT * FROM items WHERE (id = 3) LIMIT 1", \
+      "SELECT * FROM items WHERE (id = 1) LIMIT 1"]
+    @c[:id => 4]
+    $sqls.should == ["SELECT * FROM items WHERE (id = 3) LIMIT 1", \
+      "SELECT * FROM items WHERE (id = 1) LIMIT 1", \
+      "SELECT * FROM items WHERE (id = 4) LIMIT 1"]
+
+    $sqls.clear
+    m = @c2[:id => 3]
+    @cache[m.cache_key].should be_nil
+    $sqls.last.should == "SELECT * FROM items WHERE (id = 3) LIMIT 1"
+    m = @c2[1]
+    @cache[m.cache_key].should == m
+    $sqls.should == ["SELECT * FROM items WHERE (id = 3) LIMIT 1", \
+      "SELECT * FROM items WHERE (id = 1) LIMIT 1"]
+    @c2[:id => 4]
+    $sqls.should == ["SELECT * FROM items WHERE (id = 3) LIMIT 1", \
+      "SELECT * FROM items WHERE (id = 1) LIMIT 1", \
+      "SELECT * FROM items WHERE (id = 4) LIMIT 1"]
+  end
+end
diff -pruN 1.5.1-1/spec/extensions/hook_class_methods_spec.rb 3.0.0-1/spec/extensions/hook_class_methods_spec.rb
--- 1.5.1-1/spec/extensions/hook_class_methods_spec.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/extensions/hook_class_methods_spec.rb	2009-03-25 15:20:44.000000000 +0000
@@ -0,0 +1,470 @@
+require File.join(File.dirname(__FILE__), "spec_helper")
+
+describe "Model hooks" do
+  before do
+    MODEL_DB.reset
+  end
+  
+  specify "should be definable using a block" do
+    $adds = []
+    c = Class.new(Sequel::Model)
+    c.class_eval do
+      before_save {$adds << 'hi'}
+    end
+    
+    c.new.before_save
+    $adds.should == ['hi']
+  end
+  
+  specify "should be definable using a method name" do
+    $adds = []
+    c = Class.new(Sequel::Model)
+    c.class_eval do
+      def bye; $adds << 'bye'; end
+      before_save :bye
+    end
+    
+    c.new.before_save
+    $adds.should == ['bye']
+  end
+
+  specify "should be additive" do
+    $adds = []
+    c = Class.new(Sequel::Model)
+    c.class_eval do
+      after_save {$adds << 'hyiyie'}
+      after_save {$adds << 'byiyie'}
+    end
+
+    c.new.after_save
+    $adds.should == ['hyiyie', 'byiyie']
+  end
+  
+  specify "before hooks should run in reverse order" do
+    $adds = []
+    c = Class.new(Sequel::Model)
+    c.class_eval do
+      before_save {$adds << 'hyiyie'}
+      before_save {$adds << 'byiyie'}
+    end
+    
+    c.new.before_save
+    $adds.should == ['byiyie', 'hyiyie']
+  end
+
+  specify "should not be additive if the method or tag already exists" do
+    $adds = []
+    c = Class.new(Sequel::Model)
+    c.class_eval do
+      def bye; $adds << 'bye'; end
+      before_save :bye
+      before_save :bye
+    end
+    
+    c.new.before_save
+    $adds.should == ['bye']
+
+    $adds = []
+    d = Class.new(Sequel::Model)
+    d.class_eval do
+      before_save(:bye){$adds << 'hyiyie'}
+      before_save(:bye){$adds << 'byiyie'}
+    end
+    
+    d.new.before_save
+    $adds.should == ['byiyie']
+
+    $adds = []
+    e = Class.new(Sequel::Model)
+    e.class_eval do
+      def bye; $adds << 'bye'; end
+      before_save :bye
+      before_save(:bye){$adds << 'byiyie'}
+    end
+    
+    e.new.before_save
+    $adds.should == ['byiyie']
+
+    $adds = []
+    e = Class.new(Sequel::Model)
+    e.class_eval do
+      def bye; $adds << 'bye'; end
+      before_save(:bye){$adds << 'byiyie'}
+      before_save :bye
+    end
+    
+    e.new.before_save
+    $adds.should == ['bye']
+  end
+  
+  specify "should be inheritable" do
+    # pending
+    
+    $adds = []
+    a = Class.new(Sequel::Model)
+    a.class_eval do
+      after_save {$adds << '123'}
+    end
+    
+    b = Class.new(a)
+    b.class_eval do
+      after_save {$adds << '456'}
+      after_save {$adds << '789'}
+    end
+    
+    b.new.after_save
+    $adds.should == ['123', '456', '789']
+  end
+  
+  specify "should be overridable in descendant classes" do
+    $adds = []
+    a = Class.new(Sequel::Model)
+    a.class_eval do
+      before_save {$adds << '123'}
+    end
+    
+    b = Class.new(a)
+    b.class_eval do
+      def before_save; $adds << '456'; end
+    end
+    
+    a.new.before_save
+    $adds.should == ['123']
+    $adds = []
+    b.new.before_save
+    $adds.should == ['456']
+  end
+  
+  specify "should stop processing if a hook returns false" do
+    $flag = true
+    $adds = []
+    
+    a = Class.new(Sequel::Model)
+    a.class_eval do
+      after_save {$adds << 'blah'; $flag}
+      after_save {$adds << 'cruel'}
+    end
+    
+    a.new.after_save
+    $adds.should == ['blah', 'cruel']
+
+    # chain should not break on nil
+    $adds = []
+    $flag = nil
+    a.new.after_save
+    $adds.should == ['blah', 'cruel']
+    
+    $adds = []
+    $flag = false
+    a.new.after_save
+    $adds.should == ['blah']
+    
+    b = Class.new(a)
+    b.class_eval do
+      after_save {$adds << 'mau'}
+    end
+    
+    $adds = []
+    b.new.after_save
+    $adds.should == ['blah']
+  end
+end
+
+describe "Model#after_initialize" do
+  specify "should be called after initialization" do
+    $values1 = nil
+    $reached_after_initialized = false
+    
+    a = Class.new(Sequel::Model)
+    a.class_eval do
+      columns :x, :y
+      after_initialize do
+        $values1 = @values.clone
+        $reached_after_initialized = true
+      end
+    end
+    
+    a.new(:x => 1, :y => 2)
+    $values1.should == {:x => 1, :y => 2}
+    $reached_after_initialized.should == true
+  end
+end
+
+describe "Model#before_create && Model#after_create" do
+  before do
+    MODEL_DB.reset
+
+    @c = Class.new(Sequel::Model(:items))
+    @c.class_eval do
+      columns :x
+      no_primary_key
+      
+      after_create {MODEL_DB << "BLAH after"}
+    end
+  end
+  
+  specify "should be called around new record creation" do
+    @c.before_create {MODEL_DB << "BLAH before"}
+    @c.create(:x => 2)
+    MODEL_DB.sqls.should == [
+      'BLAH before',
+      'INSERT INTO items (x) VALUES (2)',
+      'BLAH after'
+    ]
+  end
+
+  specify ".create should cancel the save and raise an error if before_create returns false and raise_on_save_failure is true" do
+    @c.before_create{false}
+    proc{@c.load(:id => 2233).save}.should_not raise_error(Sequel::ValidationFailed)
+    proc{@c.create(:x => 2)}.should raise_error(Sequel::BeforeHookFailed)
+    MODEL_DB.sqls.should == []
+  end
+
+  specify ".create should cancel the save and return nil if before_create returns false and raise_on_save_failure is false" do
+    @c.before_create{false}
+    @c.raise_on_save_failure = false
+    @c.create(:x => 2).should == nil
+    MODEL_DB.sqls.should == []
+  end
+end
+
+describe "Model#before_update && Model#after_update" do
+  before do
+    MODEL_DB.reset
+
+    @c = Class.new(Sequel::Model(:items))
+    @c.class_eval do
+      after_update {MODEL_DB << "BLAH after"}
+    end
+  end
+  
+  specify "should be called around record update" do
+    @c.before_update {MODEL_DB << "BLAH before"}
+    m = @c.load(:id => 2233)
+    m.save
+    MODEL_DB.sqls.should == [
+      'BLAH before',
+      'UPDATE items SET id = 2233 WHERE (id = 2233)',
+      'BLAH after'
+    ]
+  end
+
+  specify "#save should cancel the save and raise an error if before_update returns false and raise_on_save_failure is true" do
+    @c.before_update{false}
+    proc{@c.load(:id => 2233).save}.should_not raise_error(Sequel::ValidationFailed)
+    proc{@c.load(:id => 2233).save}.should raise_error(Sequel::BeforeHookFailed)
+    MODEL_DB.sqls.should == []
+  end
+
+  specify "#save should cancel the save and return nil if before_update returns false and raise_on_save_failure is false" do
+    @c.before_update{false}
+    @c.raise_on_save_failure = false
+    @c.load(:id => 2233).save.should == nil
+    MODEL_DB.sqls.should == []
+  end
+end
+
+describe "Model#before_save && Model#after_save" do
+  before do
+    MODEL_DB.reset
+
+    @c = Class.new(Sequel::Model(:items))
+    @c.class_eval do
+      columns :x
+      after_save {MODEL_DB << "BLAH after"}
+    end
+  end
+  
+  specify "should be called around record update" do
+    @c.before_save {MODEL_DB << "BLAH before"}
+    m = @c.load(:id => 2233)
+    m.save
+    MODEL_DB.sqls.should == [
+      'BLAH before',
+      'UPDATE items SET id = 2233 WHERE (id = 2233)',
+      'BLAH after'
+    ]
+  end
+  
+  specify "should be called around record creation" do
+    @c.before_save {MODEL_DB << "BLAH before"}
+    @c.no_primary_key
+    @c.create(:x => 2)
+    MODEL_DB.sqls.should == [
+      'BLAH before',
+      'INSERT INTO items (x) VALUES (2)',
+      'BLAH after'
+    ]
+  end
+
+  specify "#save should cancel the save and raise an error if before_save returns false and raise_on_save_failure is true" do
+    @c.before_save{false}
+    proc{@c.load(:id => 2233).save}.should_not raise_error(Sequel::ValidationFailed)
+    proc{@c.load(:id => 2233).save}.should raise_error(Sequel::BeforeHookFailed)
+    MODEL_DB.sqls.should == []
+  end
+
+  specify "#save should cancel the save and return nil if before_save returns false and raise_on_save_failure is false" do
+    @c.before_save{false}
+    @c.raise_on_save_failure = false
+    @c.load(:id => 2233).save.should == nil
+    MODEL_DB.sqls.should == []
+  end
+end
+
+describe "Model#before_destroy && Model#after_destroy" do
+  before do
+    MODEL_DB.reset
+
+    @c = Class.new(Sequel::Model(:items))
+    @c.class_eval do
+      after_destroy {MODEL_DB << "BLAH after"}
+      
+      def delete
+        MODEL_DB << "DELETE BLAH"
+      end
+    end
+  end
+  
+  specify "should be called around record destruction" do
+    @c.before_destroy {MODEL_DB << "BLAH before"}
+    m = @c.load(:id => 2233)
+    m.destroy
+    MODEL_DB.sqls.should == [
+      'BLAH before',
+      'DELETE BLAH',
+      'BLAH after'
+    ]
+  end
+
+  specify "#destroy should cancel the destroy and raise an error if before_destroy returns false and raise_on_save_failure is true" do
+    @c.before_destroy{false}
+    proc{@c.load(:id => 2233).destroy}.should raise_error(Sequel::BeforeHookFailed)
+    MODEL_DB.sqls.should == []
+  end
+
+  specify "#destroy should cancel the destroy and return nil if before_destroy returns false and raise_on_save_failure is false" do
+    @c.before_destroy{false}
+    @c.raise_on_save_failure = false
+    @c.load(:id => 2233).destroy.should == nil
+    MODEL_DB.sqls.should == []
+  end
+end
+
+describe "Model#before_validation && Model#after_validation" do
+  before do
+    MODEL_DB.reset
+
+    @c = Class.new(Sequel::Model(:items))
+    @c.class_eval do
+      after_validation{MODEL_DB << "BLAH after"}
+
+      def self.validate(o)
+        o.errors[:id] << 'not valid' unless o[:id] == 2233
+      end
+      columns :id
+    end
+  end
+  
+  specify "should be called around validation" do
+    @c.before_validation{MODEL_DB << "BLAH before"}
+    m = @c.load(:id => 2233)
+    m.should be_valid
+    MODEL_DB.sqls.should == ['BLAH before', 'BLAH after']
+
+    MODEL_DB.sqls.clear
+    m = @c.load(:id => 22)
+    m.should_not be_valid
+    MODEL_DB.sqls.should == ['BLAH before', 'BLAH after']
+  end
+
+  specify "should be called when calling save" do
+    @c.before_validation{MODEL_DB << "BLAH before"}
+    m = @c.load(:id => 2233)
+    m.save.should == m
+    MODEL_DB.sqls.should == ['BLAH before', 'BLAH after', 'UPDATE items SET id = 2233 WHERE (id = 2233)']
+
+    MODEL_DB.sqls.clear
+    m = @c.load(:id => 22)
+    m.raise_on_save_failure = false
+    m.save.should == nil
+    MODEL_DB.sqls.should == ['BLAH before', 'BLAH after']
+  end
+
+  specify "#save should cancel the save and raise an error if before_validation returns false and raise_on_save_failure is true" do
+    @c.before_validation{false}
+    proc{@c.load(:id => 2233).save}.should_not raise_error(Sequel::ValidationFailed)
+    proc{@c.load(:id => 2233).save}.should raise_error(Sequel::BeforeHookFailed)
+    MODEL_DB.sqls.should == []
+  end
+
+  specify "#save should cancel the save and return nil if before_validation returns false and raise_on_save_failure is false" do
+    @c.before_validation{false}
+    @c.raise_on_save_failure = false
+    @c.load(:id => 2233).save.should == nil
+    MODEL_DB.sqls.should == []
+  end
+end
+
+describe "Model.has_hooks?" do
+  before do
+    @c = Class.new(Sequel::Model(:items))
+  end
+  
+  specify "should return false if no hooks are defined" do
+    @c.has_hooks?(:before_save).should be_false
+  end
+  
+  specify "should return true if hooks are defined" do
+    @c.before_save {'blah'}
+    @c.has_hooks?(:before_save).should be_true
+  end
+  
+  specify "should return true if hooks are inherited" do
+    @d = Class.new(@c)
+    @d.has_hooks?(:before_save).should be_false
+  end
+end
+
+describe "Model#add_hook_type" do
+  before do
+    class Foo < Sequel::Model(:items)
+      add_hook_type :before_bar, :after_bar
+
+      def bar
+        return :b if before_bar == false
+        return :a if after_bar == false
+        true
+      end
+    end
+    @f = Class.new(Foo)
+  end
+
+  specify "should have before_bar and after_bar class methods" do
+    @f.should respond_to(:before_bar)
+    @f.should respond_to(:before_bar)
+  end
+
+  specify "should have before_bar and after_bar instance methods" do
+    @f.new.should respond_to(:before_bar)
+    @f.new.should respond_to(:before_bar)
+  end
+
+  specify "it should return true for bar when before_bar and after_bar hooks are returing true" do
+    a = 1
+    @f.before_bar { a += 1}
+    @f.new.bar.should be_true
+    a.should == 2
+    @f.after_bar { a *= 2}
+    @f.new.bar.should be_true
+    a.should == 6
+  end
+
+  specify "it should return nil for bar when before_bar and after_bar hooks are returing false" do
+    @f.new.bar.should be_true
+    @f.after_bar { false }
+    @f.new.bar.should == :a
+    @f.before_bar { false }
+    @f.new.bar.should == :b
+  end
+end
diff -pruN 1.5.1-1/spec/extensions/inflector_spec.rb 3.0.0-1/spec/extensions/inflector_spec.rb
--- 1.5.1-1/spec/extensions/inflector_spec.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/extensions/inflector_spec.rb	2009-04-15 18:22:37.000000000 +0100
@@ -0,0 +1,119 @@
+require File.join(File.dirname(__FILE__), 'spec_helper')
+
+describe String do
+  it "#camelize and #camelcase should transform the word to CamelCase" do
+    "egg_and_hams".camelize.should == "EggAndHams"
+    "egg_and_hams".camelize(false).should == "eggAndHams"
+    "post".camelize.should == "Post"
+    "post".camelcase.should == "Post"
+  end
+
+  it "#constantize should eval the string to get a constant" do
+    "String".constantize.should == String
+    "String::Inflections".constantize.should == String::Inflections
+    proc{"BKSDDF".constantize}.should raise_error
+    proc{"++A++".constantize}.should raise_error
+  end
+  
+  it "#dasherize should transform underscores to dashes" do
+    "egg_and_hams".dasherize.should == "egg-and-hams"
+    "post".dasherize.should == "post"
+  end
+  
+  it "#demodulize should remove any preceding modules" do
+    "String::Inflections::Blah".demodulize.should == "Blah"
+    "String::Inflections".demodulize.should == "Inflections"
+    "String".demodulize.should == "String"
+  end
+  
+  it "#humanize should remove _i, transform underscore to spaces, and capitalize" do
+    "egg_and_hams".humanize.should == "Egg and hams"
+    "post".humanize.should == "Post"
+    "post_id".humanize.should == "Post"
+  end
+  
+  it "#titleize and #titlecase should underscore, humanize, and capitalize all words" do
+    "egg-and: hams".titleize.should == "Egg And: Hams"
+    "post".titleize.should == "Post"
+    "post".titlecase.should == "Post"
+  end
+  
+  it "#underscore should add underscores between CamelCased words, change :: to / and - to _, and downcase" do
+    "EggAndHams".underscore.should == "egg_and_hams"
+    "EGGAndHams".underscore.should == "egg_and_hams"
+    "Egg::And::Hams".underscore.should == "egg/and/hams"
+    "post".underscore.should == "post"
+    "post-id".underscore.should == "post_id"
+  end
+
+  it "#pluralize should transform words from singular to plural" do
+    "post".pluralize.should == "posts"
+    "octopus".pluralize.should =="octopi"
+    "the blue mailman".pluralize.should == "the blue mailmen"
+    "CamelOctopus".pluralize.should == "CamelOctopi"
+  end
+  
+  it "#singularize should transform words from plural to singular" do
+    "posts".singularize.should == "post"
+    "octopi".singularize.should == "octopus"
+    "the blue mailmen".singularize.should == "the blue mailman"
+    "CamelOctopi".singularize.should == "CamelOctopus"
+  end
+  
+  it "#tableize should transform class names to table names" do
+    "RawScaledScorer".tableize.should == "raw_scaled_scorers"
+    "egg_and_ham".tableize.should == "egg_and_hams"
+    "fancyCategory".tableize.should == "fancy_categories"
+  end
+  
+  it "#classify should tranform table names to class names" do
+    "egg_and_hams".classify.should == "EggAndHam"
+    "post".classify.should == "Post"
+  end
+  
+  it "#foreign_key should create a foreign key name from a class name" do
+    "Message".foreign_key.should == "message_id"
+    "Message".foreign_key(false).should == "messageid"
+    "Admin::Post".foreign_key.should == "post_id"
+  end
+end
+
+describe String::Inflections do
+  before do
+    @plurals, @singulars, @uncountables = String.inflections.plurals.dup, String.inflections.singulars.dup, String.inflections.uncountables.dup
+  end
+  after do
+    String.inflections.plurals.replace(@plurals)
+    String.inflections.singulars.replace(@singulars)
+    String.inflections.uncountables.replace(@uncountables)
+  end
+
+  it "should be possible to clear the list of singulars, plurals, and uncountables" do
+    String.inflections.clear(:plurals)
+    String.inflections.plurals.should == []
+    String.inflections.plural('blah', 'blahs')
+    String.inflections.clear
+    String.inflections.plurals.should == []
+    String.inflections.singulars.should == []
+    String.inflections.uncountables.should == []
+  end
+
+  it "should be able to specify new inflection rules" do
+    String.inflections do |i|
+      i.plural(/xx$/i, 'xxx')
+      i.singular(/ttt$/i, 'tt')
+      i.irregular('yy', 'yyy')
+      i.uncountable(%w'zz')
+    end
+    'roxx'.pluralize.should == 'roxxx'
+    'rottt'.singularize.should == 'rott'
+    'yy'.pluralize.should == 'yyy'
+    'yyy'.singularize.should == 'yy'
+    'zz'.pluralize.should == 'zz'
+    'zz'.singularize.should == 'zz'
+  end
+
+  it "should be yielded and returned by String.inflections" do
+    String.inflections{|i| i.should == String::Inflections}.should == String::Inflections
+  end
+end
diff -pruN 1.5.1-1/spec/extensions/migration_spec.rb 3.0.0-1/spec/extensions/migration_spec.rb
--- 1.5.1-1/spec/extensions/migration_spec.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/extensions/migration_spec.rb	2009-03-25 15:40:15.000000000 +0000
@@ -0,0 +1,261 @@
+require File.join(File.dirname(__FILE__), 'spec_helper')
+
+context "Migration classes" do
+  before do
+    Sequel::Migration.descendants.clear
+  end
+
+  specify "should be registred in Migration.descendants" do
+    @class = Class.new(Sequel::Migration)
+    
+    Sequel::Migration.descendants.should == [@class]
+  end
+  
+  specify "should be registered in the right order" do
+    @c1 = Class.new(Sequel::Migration)
+    @c2 = Class.new(Sequel::Migration)
+    @c3 = Class.new(Sequel::Migration)
+    
+    Sequel::Migration.descendants.should == [@c1, @c2, @c3]
+  end
+end
+
+context "Migration#apply" do
+  before do
+    @c = Class.new do
+      define_method(:one) {|x| [1111, x]}
+      define_method(:two) {|x| [2222, x]}
+    end
+    @db = @c.new
+    
+    @migration = Class.new(Sequel::Migration) do
+      define_method(:up) {one(3333)}
+      define_method(:down) {two(4444)}
+    end
+  end
+  
+  specify "should raise for an invalid direction" do
+    proc {@migration.apply(@db, :hahaha)}.should raise_error(ArgumentError)
+  end
+  
+  specify "should apply the up direction correctly" do
+    @migration.apply(@db, :up).should == [1111, 3333]
+  end
+
+  specify "should apply the down direction correctly" do
+    @migration.apply(@db, :down).should == [2222, 4444]
+  end
+end
+
+class DummyMigrationDataset
+  attr_reader :from
+  
+  def initialize(x); @from = x; end
+  
+  @@version = nil
+  
+  def version; @@version; end
+  def version=(x); @@version = x; end
+  def first; @@version ? {:version => @@version} : nil; end
+  def update(h); @@version = h[:version]; end
+  def <<(h); @@version = h[:version]; end
+end
+
+class DummyMigrationDB
+  attr_reader :creates, :drops, :table_created
+  
+  def initialize
+    @creates = []
+    @drops = []
+  end
+  
+  def create(x); @creates << x; end
+  def drop(x); @drops << x; end
+  
+  def [](x); DummyMigrationDataset.new(x); end
+  
+  def create_table(x); raise if @table_created == x; @table_created = x; end
+  def table_exists?(x); @table_created == x; end
+  
+  def transaction; yield; end
+end
+
+MIGRATION_001 = %[
+  class CreateSessions < Sequel::Migration
+    def up
+      create(1111)
+    end
+    
+    def down
+      drop(1111)
+    end
+  end
+]
+
+MIGRATION_002 = %[
+  class CreateNodes < Sequel::Migration
+    def up
+      create(2222)
+    end
+    
+    def down
+      drop(2222)
+    end
+  end
+]
+
+MIGRATION_003 = %[
+  class CreateUsers < Sequel::Migration
+    def up
+      create(3333)
+    end
+    
+    def down
+      drop(3333)
+    end
+  end
+]
+
+MIGRATION_005 = %[
+  class CreateAttributes < Sequel::Migration
+    def up
+      create(5555)
+    end
+    
+    def down
+      drop(5555)
+    end
+  end
+]
+
+context "Sequel::Migrator" do
+  before do
+    @db = DummyMigrationDB.new
+    
+    File.open('001_create_sessions.rb', 'w') {|f| f << MIGRATION_001}
+    File.open('002_create_nodes.rb', 'w') {|f| f << MIGRATION_002}
+    File.open('003_create_users.rb', 'w') {|f| f << MIGRATION_003}
+    File.open('005_create_attributes.rb', 'w') {|f| f << MIGRATION_005}
+    
+    @db[:schema_info].version = nil
+  end
+  
+  after do
+    Object.send(:remove_const, "CreateSessions") if Object.const_defined?("CreateSessions")
+    Object.send(:remove_const, "CreateNodes") if Object.const_defined?("CreateNodes")
+    Object.send(:remove_const, "CreateUsers") if Object.const_defined?("CreateUsers")
+    Object.send(:remove_const, "CreateAttributes") if Object.const_defined?("CreateAttributes")
+
+    File.delete('001_create_sessions.rb')
+    File.delete('002_create_nodes.rb')
+    File.delete('003_create_users.rb')
+    File.delete('005_create_attributes.rb')
+  end
+  
+  specify "should return the list of files for a specified version range" do
+    Sequel::Migrator.migration_files('.', 1..1).should == \
+      ['./001_create_sessions.rb']
+
+    Sequel::Migrator.migration_files('.', 1..3).should == \
+      ['./001_create_sessions.rb', './002_create_nodes.rb', './003_create_users.rb']
+
+    Sequel::Migrator.migration_files('.', 3..5).should == \
+      ['./003_create_users.rb', './005_create_attributes.rb']
+      
+    Sequel::Migrator.migration_files('.', 7..8).should == []
+  end
+  
+  specify "should return the latest version available" do
+    Sequel::Migrator.latest_migration_version('.').should == 5
+  end
+  
+  specify "should load the migration classes for the specified range" do
+    Sequel::Migrator.migration_classes('.', 3, 0, :up).should == \
+      [CreateSessions, CreateNodes, CreateUsers]
+  end
+  
+  specify "should load the migration classes for the specified range" do
+    Sequel::Migrator.migration_classes('.', 0, 5, :down).should == \
+      [CreateAttributes, CreateUsers, CreateNodes, CreateSessions]
+  end
+  
+  specify "should start from current + 1 for the up direction" do
+    Sequel::Migrator.migration_classes('.', 3, 1, :up).should == \
+      [CreateNodes, CreateUsers]
+  end
+  
+  specify "should end on current + 1 for the down direction" do
+    Sequel::Migrator.migration_classes('.', 2, 5, :down).should == \
+      [CreateAttributes, CreateUsers]
+  end
+  
+  specify "should automatically create the schema_info table" do
+    @db.table_exists?(:schema_info).should be_false
+    Sequel::Migrator.schema_info_dataset(@db)
+    @db.table_exists?(:schema_info).should be_true
+    
+    # should not raise if table already exists
+    proc {Sequel::Migrator.schema_info_dataset(@db)}.should_not raise_error
+  end
+  
+  specify "should return a dataset for the schema_info table" do
+    d = Sequel::Migrator.schema_info_dataset(@db)
+    d.from.should == :schema_info
+  end
+  
+  specify "should get the migration version stored in the database" do
+    # default is 0
+    Sequel::Migrator.get_current_migration_version(@db).should == 0
+    
+    Sequel::Migrator.schema_info_dataset(@db) << {:version => 4321}
+
+    Sequel::Migrator.get_current_migration_version(@db).should == 4321
+  end
+  
+  specify "should set the migration version stored in the database" do
+    Sequel::Migrator.get_current_migration_version(@db).should == 0
+    Sequel::Migrator.set_current_migration_version(@db, 6666)
+    Sequel::Migrator.get_current_migration_version(@db).should == 6666
+  end
+  
+  specify "should apply migrations correctly in the up direction" do
+    Sequel::Migrator.apply(@db, '.', 3, 2)
+    @db.creates.should == [3333]
+    
+    Sequel::Migrator.get_current_migration_version(@db).should == 3
+
+    Sequel::Migrator.apply(@db, '.', 5)
+    @db.creates.should == [3333, 5555]
+
+    Sequel::Migrator.get_current_migration_version(@db).should == 5
+  end
+  
+  specify "should apply migrations correctly in the down direction" do
+    Sequel::Migrator.apply(@db, '.', 1, 5)
+    @db.drops.should == [5555, 3333, 2222]
+
+    Sequel::Migrator.get_current_migration_version(@db).should == 1
+  end
+
+  specify "should apply migrations up to the latest version if no target is given" do
+    Sequel::Migrator.apply(@db, '.')
+    @db.creates.should == [1111, 2222, 3333, 5555]
+
+    Sequel::Migrator.get_current_migration_version(@db).should == 5
+  end
+
+  specify "should apply migrations down to 0 version correctly" do
+    Sequel::Migrator.apply(@db, '.', 0, 5)
+    @db.drops.should == [5555, 3333, 2222, 1111]
+
+    Sequel::Migrator.get_current_migration_version(@db).should == 0
+  end
+  
+  specify "should return the target version" do
+    Sequel::Migrator.apply(@db, '.', 3, 2).should == 3
+
+    Sequel::Migrator.apply(@db, '.', 0).should == 0
+
+    Sequel::Migrator.apply(@db, '.').should == 5
+  end
+end
diff -pruN 1.5.1-1/spec/extensions/pagination_spec.rb 3.0.0-1/spec/extensions/pagination_spec.rb
--- 1.5.1-1/spec/extensions/pagination_spec.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/extensions/pagination_spec.rb	2009-03-25 15:37:42.000000000 +0000
@@ -0,0 +1,99 @@
+require File.join(File.dirname(__FILE__), "spec_helper")
+
+context "A paginated dataset" do
+  before do
+    @d = Sequel::Dataset.new(nil)
+    @d.meta_def(:count) {153}
+    
+    @paginated = @d.paginate(1, 20)
+  end
+  
+  specify "should raise an error if the dataset already has a limit" do
+    proc{@d.limit(10).paginate(1,10)}.should raise_error(Sequel::Error)
+    proc{@paginated.paginate(2,20)}.should raise_error(Sequel::Error)
+  end
+  
+  specify "should set the limit and offset options correctly" do
+    @paginated.opts[:limit].should == 20
+    @paginated.opts[:offset].should == 0
+  end
+  
+  specify "should set the page count correctly" do
+    @paginated.page_count.should == 8
+    @d.paginate(1, 50).page_count.should == 4
+  end
+  
+  specify "should set the current page number correctly" do
+    @paginated.current_page.should == 1
+    @d.paginate(3, 50).current_page.should == 3
+  end
+  
+  specify "should return the next page number or nil if we're on the last" do
+    @paginated.next_page.should == 2
+    @d.paginate(4, 50).next_page.should be_nil
+  end
+  
+  specify "should return the previous page number or nil if we're on the last" do
+    @paginated.prev_page.should be_nil
+    @d.paginate(4, 50).prev_page.should == 3
+  end
+  
+  specify "should return the page range" do
+    @paginated.page_range.should == (1..8)
+    @d.paginate(4, 50).page_range.should == (1..4)
+  end
+  
+  specify "should return the record range for the current page" do
+    @paginated.current_page_record_range.should == (1..20)
+    @d.paginate(4, 50).current_page_record_range.should == (151..153)
+    @d.paginate(5, 50).current_page_record_range.should == (0..0)
+  end
+
+  specify "should return the record count for the current page" do
+    @paginated.current_page_record_count.should == 20
+    @d.paginate(3, 50).current_page_record_count.should == 50
+    @d.paginate(4, 50).current_page_record_count.should == 3
+    @d.paginate(5, 50).current_page_record_count.should == 0
+  end
+
+  specify "should know if current page is last page" do
+    @paginated.last_page?.should be_false
+    @d.paginate(2, 20).last_page?.should be_false
+    @d.paginate(5, 30).last_page?.should be_false
+    @d.paginate(6, 30).last_page?.should be_true
+  end
+
+  specify "should know if current page is first page" do
+    @paginated.first_page?.should be_true
+    @d.paginate(1, 20).first_page?.should be_true
+    @d.paginate(2, 20).first_page?.should be_false
+  end
+
+  specify "should work with fixed sql" do
+    ds = @d.clone(:sql => 'select * from blah')
+    ds.meta_def(:count) {150}
+    ds.paginate(2, 50).sql.should == 'SELECT * FROM (select * from blah) AS t1 LIMIT 50 OFFSET 50'
+  end
+end
+
+context "Dataset#each_page" do
+  before do
+    @d = Sequel::Dataset.new(nil).from(:items)
+    @d.meta_def(:count) {153}
+  end
+  
+  specify "should raise an error if the dataset already has a limit" do
+    proc{@d.limit(10).each_page(10){}}.should raise_error(Sequel::Error)
+  end
+  
+  specify "should iterate over each page in the resultset as a paginated dataset" do
+    a = []
+    @d.each_page(50) {|p| a << p}
+    a.map {|p| p.sql}.should == [
+      'SELECT * FROM items LIMIT 50 OFFSET 0',
+      'SELECT * FROM items LIMIT 50 OFFSET 50',
+      'SELECT * FROM items LIMIT 50 OFFSET 100',
+      'SELECT * FROM items LIMIT 50 OFFSET 150',
+    ]
+  end
+end
diff -pruN 1.5.1-1/spec/extensions/pretty_table_spec.rb 3.0.0-1/spec/extensions/pretty_table_spec.rb
--- 1.5.1-1/spec/extensions/pretty_table_spec.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/extensions/pretty_table_spec.rb	2009-03-25 15:40:10.000000000 +0000
@@ -0,0 +1,91 @@
+require File.join(File.dirname(__FILE__), 'spec_helper')
+
+require 'stringio'
+
+context "Dataset#print" do
+  before do
+    @output = StringIO.new
+    @orig_stdout = $stdout
+    $stdout = @output
+    @dataset = Sequel::Dataset.new(nil).from(:items)
+    def @dataset.fetch_rows(sql)
+      yield({:a=>1, :b=>2})
+      yield({:a=>3, :b=>4})
+      yield({:a=>5, :b=>6})
+    end
+  end
+
+  after do
+    $stdout = @orig_stdout
+  end
+
+  specify "should print out a table with the values" do
+    @dataset.print(:a, :b)
+    @output.rewind
+    @output.read.should == \
+      "+-+-+\n|a|b|\n+-+-+\n|1|2|\n|3|4|\n|5|6|\n+-+-+\n"
+  end
+
+  specify "should default to the dataset's columns" do
+    @dataset.meta_def(:columns) {[:a, :b]}
+    @dataset.print
+    @output.rewind
+    @output.read.should == \
+      "+-+-+\n|a|b|\n+-+-+\n|1|2|\n|3|4|\n|5|6|\n+-+-+\n"
+  end
+end
+
+context "PrettyTable" do
+  before do
+    @data1 = [
+      {:x => 3, :y => 4}
+    ]
+    
+    @data2 = [
+      {:a => 23, :b => 45},
+      {:a => 45, :b => 2377}
+    ]
+
+    @data3 = [
+      {:aaa => 1},
+      {:bb => 2},
+      {:c => 3}
+    ]
+
+    @output = StringIO.new
+    @orig_stdout = $stdout
+    $stdout = @output
+  end
+
+  after do
+    $stdout = @orig_stdout
+  end
+  
+  specify "should infer the columns if not given" do
+    Sequel::PrettyTable.print(@data1)
+    @output.rewind
+    @output.read.should =~ \
+      /\n(\|x\|y\|)|(\|y\|x\|)\n/
+  end
+  
+  specify "should calculate the maximum width of each column correctly" do
+    Sequel::PrettyTable.print(@data2, [:a, :b])
+    @output.rewind
+    @output.read.should == \
+      "+--+----+\n|a |b   |\n+--+----+\n|23|  45|\n|45|2377|\n+--+----+\n"
+  end
+
+  specify "should also take header width into account" do
+    Sequel::PrettyTable.print(@data3, [:aaa, :bb, :c])
+    @output.rewind
+    @output.read.should == \
+      "+---+--+-+\n|aaa|bb|c|\n+---+--+-+\n|  1|  | |\n|   | 2| |\n|   |  |3|\n+---+--+-+\n"
+  end
+  
+  specify "should print only the specified columns" do
+    Sequel::PrettyTable.print(@data2, [:a])
+    @output.rewind
+    @output.read.should == \
+      "+--+\n|a |\n+--+\n|23|\n|45|\n+--+\n"
+  end
+end
diff -pruN 1.5.1-1/spec/extensions/query_spec.rb 3.0.0-1/spec/extensions/query_spec.rb
--- 1.5.1-1/spec/extensions/query_spec.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/extensions/query_spec.rb	2009-03-25 15:37:37.000000000 +0000
@@ -0,0 +1,85 @@
+require File.join(File.dirname(__FILE__), "spec_helper")
+
+context "Database#dataset" do
+  before do
+    @db = Sequel::Database.new
+  end
+
+  specify "should delegate to Dataset#query if block is provided" do
+    @d = @db.query {select :x; from :y}
+    @d.should be_a_kind_of(Sequel::Dataset)
+    @d.sql.should == "SELECT x FROM y"
+  end
+end
+
+context "Dataset#query" do
+  before do
+    @d = Sequel::Dataset.new(nil)
+  end
+  
+  specify "should support #from" do
+    q = @d.query {from :xxx}
+    q.class.should == @d.class
+    q.sql.should == "SELECT * FROM xxx"
+  end
+  
+  specify "should support #select" do
+    q = @d.query do
+      select :a, :b___mongo
+      from :yyy
+    end
+    q.class.should == @d.class
+    q.sql.should == "SELECT a, b AS mongo FROM yyy"
+  end
+  
+  specify "should support #where" do
+    q = @d.query do
+      from :zzz
+      where(:x + 2 > :y + 3)
+    end
+    q.class.should == @d.class
+    q.sql.should == "SELECT * FROM zzz WHERE ((x + 2) > (y + 3))"
+
+    q = @d.from(:zzz).query do
+      where((:x.sql_number > 1) & (:y.sql_number > 2))
+    end
+    q.class.should == @d.class
+    q.sql.should == "SELECT * FROM zzz WHERE ((x > 1) AND (y > 2))"
+
+    q = @d.from(:zzz).query do
+      where :x => 33
+    end
+    q.class.should == @d.class
+    q.sql.should == "SELECT * FROM zzz WHERE (x = 33)"
+  end
+  
+  specify "should support #group_by and #having" do
+    q = @d.query do
+      from :abc
+      group_by :id
+      having(:x.sql_number >= 2)
+    end
+    q.class.should == @d.class
+    q.sql.should == "SELECT * FROM abc GROUP BY id HAVING (x >= 2)"
+  end
+  
+  specify "should support #order, #order_by" do
+    q = @d.query do
+      from :xyz
+      order_by :stamp
+    end
+    q.class.should == @d.class
+    q.sql.should == "SELECT * FROM xyz ORDER BY stamp"
+  end
+  
+  specify "should raise on non-chainable method calls" do
+    proc {@d.query {first_source}}.should raise_error(Sequel::Error)
+  end
+  
+  specify "should raise on each, insert, update, delete" do
+    proc {@d.query {each}}.should raise_error(Sequel::Error)
+    proc {@d.query {insert(:x => 1)}}.should raise_error(Sequel::Error)
+    proc {@d.query {update(:x => 1)}}.should raise_error(Sequel::Error)
+    proc {@d.query {delete}}.should raise_error(Sequel::Error)
+  end
+end
diff -pruN 1.5.1-1/spec/extensions/schema_dumper_spec.rb 3.0.0-1/spec/extensions/schema_dumper_spec.rb
--- 1.5.1-1/spec/extensions/schema_dumper_spec.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/extensions/schema_dumper_spec.rb	2009-04-27 22:12:33.000000000 +0100
@@ -0,0 +1,292 @@
+require File.join(File.dirname(__FILE__), 'spec_helper')
+
+describe "Sequel::Schema::Generator dump methods" do
+  before do
+    @d = Sequel::Database.new
+    @g = Sequel::Schema::Generator
+  end
+
+  it "should allow the same table information to be converted to a string for evaling inside of another instance with the same result" do
+    g = @g.new(@d) do
+      Integer :a
+      varchar :b
+      column :dt, DateTime
+      column :vc, :varchar
+      primary_key :c
+      foreign_key :d, :a
+      foreign_key :e
+      foreign_key [:d, :e], :name=>:cfk
+      constraint :blah, "a=1"
+      check :a=>1
+      unique [:e]
+      index :a
+      index [:c, :e]
+      index [:b, :c], :type=>:hash
+      index [:d], :unique=>true
+      spatial_index :a
+      full_text_index [:b, :c]
+    end
+    g2 = @g.new(@d) do
+      instance_eval(g.dump_columns, __FILE__, __LINE__)
+      instance_eval(g.dump_constraints, __FILE__, __LINE__)
+      instance_eval(g.dump_indexes, __FILE__, __LINE__)
+    end
+    g.columns.should == g2.columns
+    g.constraints.should == g2.constraints
+    g.indexes.should == g2.indexes
+  end
+
+  it "should allow dumping indexes as separate add_index and drop_index methods" do
+    g = @g.new(@d) do
+      index :a
+      index [:c, :e], :name=>:blah
+      index [:b, :c], :unique=>true
+    end
+
+    t = <<END_CODE
+add_index :t, [:a]
+add_index :t, [:c, :e], :name=>:blah
+add_index :t, [:b, :c], :unique=>true
+END_CODE
+    g.dump_indexes(:add_index=>:t).should == t.strip
+
+    t = <<END_CODE
+drop_index :t, [:a]
+drop_index :t, [:c, :e], :name=>:blah
+drop_index :t, [:b, :c], :unique=>true
+END_CODE
+    g.dump_indexes(:drop_index=>:t).should == t.strip
+  end
+
+  it "should raise an error if you try to dump a Generator that uses a constraint with a proc" do
+    proc{@g.new(@d){check{a>1}}.dump_constraints}.should raise_error(Sequel::Error)
+  end
+end
+
+describe "Sequel::Database dump methods" do
+  before do
+    @d = Sequel::Database.new
+    @d.meta_def(:tables){[:t1, :t2]}
+    @d.meta_def(:schema) do |t, *o|
+      case t
+      when :t1
+        [[:c1, {:db_type=>'integer', :primary_key=>true, :allow_null=>false}],
+         [:c2, {:db_type=>'varchar(20)', :allow_null=>true}]]
+      when :t2
+        [[:c1, {:db_type=>'integer', :primary_key=>true, :allow_null=>false}],
+         [:c2, {:db_type=>'numeric', :primary_key=>true, :allow_null=>false}]]
+      when :t3
+        [[:c1, {:db_type=>'date', :default=>"'now()'", :allow_null=>true}],
+         [:c2, {:db_type=>'datetime', :allow_null=>false}]]
+      when :t4
+        [[:c1, {:db_type=>'boolean', :default=>"false", :allow_null=>true}],
+         [:c2, {:db_type=>'boolean', :default=>"true", :allow_null=>true}],
+         [:c3, {:db_type=>'varchar', :default=>"'blah'", :allow_null=>true}],
+         [:c4, {:db_type=>'integer', :default=>"35", :allow_null=>true}]]
+      when :t5
+        [[:c1, {:db_type=>'blahblah', :allow_null=>true}]]
+      end
+    end
+  end
+
+  it "should support dumping table schemas as create_table method calls" do
+    @d.dump_table_schema(:t1).should == "create_table(:t1) do\n  primary_key :c1\n  String :c2, :size=>20\nend"
+  end
+
+  it "should use a composite primary_key calls if there is a composite primary key" do
+    @d.dump_table_schema(:t2).should == "create_table(:t2) do\n  Integer :c1, :null=>false\n  BigDecimal :c2, :null=>false\n  \n  primary_key [:c1, :c2]\nend"
+  end
+
+  it "should include index information if available" do
+    @d.meta_def(:indexes) do |t|
+      {:i1=>{:columns=>[:c1], :unique=>false},
+       :t1_c2_c1_index=>{:columns=>[:c2, :c1], :unique=>true}}
+    end
+    @d.dump_table_schema(:t1).should == "create_table(:t1) do\n  primary_key :c1\n  String :c2, :size=>20\n  \n  index [:c1], :name=>:i1\n  index [:c2, :c1], :unique=>true\nend"
+  end
+
+  it "should support dumping the whole database as a migration" do
+    @d.dump_schema_migration.should == <<-END_MIG
+Class.new(Sequel::Migration) do
+  def up
+    create_table(:t1) do
+      primary_key :c1
+      String :c2, :size=>20
+    end
+    
+    create_table(:t2) do
+      Integer :c1, :null=>false
+      BigDecimal :c2, :null=>false
+      
+      primary_key [:c1, :c2]
+    end
+  end
+  
+  def down
+    drop_table(:t1, :t2)
+  end
+end
+END_MIG
+  end
+
+  it "should honor the :same_db option to not convert types" do
+    @d.dump_table_schema(:t1, :same_db=>true).should == "create_table(:t1) do\n  primary_key :c1\n  column :c2, \"varchar(20)\"\nend"
+    @d.dump_schema_migration(:same_db=>true).should == <<-END_MIG
+Class.new(Sequel::Migration) do
+  def up
+    create_table(:t1) do
+      primary_key :c1
+      column :c2, "varchar(20)"
+    end
+    
+    create_table(:t2) do
+      column :c1, "integer", :null=>false
+      column :c2, "numeric", :null=>false
+      
+      primary_key [:c1, :c2]
+    end
+  end
+  
+  def down
+    drop_table(:t1, :t2)
+  end
+end
+END_MIG
+  end
+
+  it "should honor the :indexes => false option to not include indexes" do
+    @d.meta_def(:indexes) do |t|
+      {:i1=>{:columns=>[:c1], :unique=>false},
+       :t1_c2_c1_index=>{:columns=>[:c2, :c1], :unique=>true}}
+    end
+    @d.dump_table_schema(:t1, :indexes=>false).should == "create_table(:t1) do\n  primary_key :c1\n  String :c2, :size=>20\nend"
+    @d.dump_schema_migration(:indexes=>false).should == <<-END_MIG
+Class.new(Sequel::Migration) do
+  def up
+    create_table(:t1) do
+      primary_key :c1
+      String :c2, :size=>20
+    end
+    
+    create_table(:t2) do
+      Integer :c1, :null=>false
+      BigDecimal :c2, :null=>false
+      
+      primary_key [:c1, :c2]
+    end
+  end
+  
+  def down
+    drop_table(:t1, :t2)
+  end
+end
+END_MIG
+  end
+
+  it "should support dumping just indexes as a migration" do
+    @d.meta_def(:tables){[:t1]}
+    @d.meta_def(:indexes) do |t|
+      {:i1=>{:columns=>[:c1], :unique=>false},
+       :t1_c2_c1_index=>{:columns=>[:c2, :c1], :unique=>true}}
+    end
+    @d.dump_indexes_migration.should == <<-END_MIG
+Class.new(Sequel::Migration) do
+  def up
+    add_index :t1, [:c1], :name=>:i1
+    add_index :t1, [:c2, :c1], :unique=>true
+  end
+  
+  def down
+    drop_index :t1, [:c1], :name=>:i1
+    drop_index :t1, [:c2, :c1], :unique=>true
+  end
+end
+END_MIG
+  end
+
+  it "should handle not null values and defaults" do
+    @d.dump_table_schema(:t3).should == "create_table(:t3) do\n  Date :c1, :default=>\"'now()'\".lit\n  DateTime :c2, :null=>false\nend"
+  end
+
+  it "should handle converting common defaults" do
+    @d.dump_table_schema(:t4).should == "create_table(:t4) do\n  TrueClass :c1, :default=>false\n  TrueClass :c2, :default=>true\n  String :c3, :default=>\"'blah'\".lit\n  Integer :c4, :default=>35\nend"
+  end
+
+  it "should convert unknown database types to strings" do
+    @d.dump_table_schema(:t5).should == "create_table(:t5) do\n  String :c1\nend"
+  end
+
+  it "should convert many database types to ruby types" do
+    types = %w"mediumint smallint int integer mediumint(6) smallint(7) int(8) integer(9)
+      tinyint tinyint(1) bigint bigint(20) real float double boolean tinytext mediumtext
+      longtext text clob date datetime timestamp time char character
+      varchar varchar(255) varchar(30) bpchar string money
+      decimal decimal(10,2) numeric numeric(15,3) number bytea tinyblob mediumblob longblob
+      blob varbinary varbinary(10) binary binary(20) year" +
+      ["double precision", "timestamp with time zone", "timestamp without time zone",
+       "time with time zone", "time without time zone", "character varying(20)"]
+    @d.meta_def(:schema) do |t, *o|
+      i = 0
+      types.map{|x| [:"c#{i+=1}", {:db_type=>x, :allow_null=>true}]}
+    end
+    table = <<END_MIG
+create_table(:x) do
+  Integer :c1
+  Integer :c2
+  Integer :c3
+  Integer :c4
+  Integer :c5
+  Integer :c6
+  Integer :c7
+  Integer :c8
+  TrueClass :c9
+  TrueClass :c10
+  Bignum :c11
+  Bignum :c12
+  Float :c13
+  Float :c14
+  Float :c15
+  TrueClass :c16
+  String :c17, :text=>true
+  String :c18, :text=>true
+  String :c19, :text=>true
+  String :c20, :text=>true
+  String :c21, :text=>true
+  Date :c22
+  DateTime :c23
+  DateTime :c24
+  Time :c25, :only_time=>true
+  String :c26, :fixed=>true
+  String :c27, :fixed=>true
+  String :c28
+  String :c29
+  String :c30, :size=>30
+  String :c31
+  String :c32
+  BigDecimal :c33, :size=>[19, 2]
+  BigDecimal :c34
+  BigDecimal :c35, :size=>[10, 2]
+  BigDecimal :c36
+  BigDecimal :c37, :size=>[15, 3]
+  BigDecimal :c38
+  File :c39
+  File :c40
+  File :c41
+  File :c42
+  File :c43
+  File :c44
+  File :c45, :size=>10
+  File :c46
+  File :c47, :size=>20
+  Integer :c48
+  Float :c49
+  DateTime :c50
+  DateTime :c51
+  Time :c52, :only_time=>true
+  Time :c53, :only_time=>true
+  String :c54, :size=>20
+end
+END_MIG
+    @d.dump_table_schema(:x).should == table.chomp
+  end
+end
diff -pruN 1.5.1-1/spec/extensions/schema_spec.rb 3.0.0-1/spec/extensions/schema_spec.rb
--- 1.5.1-1/spec/extensions/schema_spec.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/extensions/schema_spec.rb	2009-03-08 05:23:32.000000000 +0000
@@ -0,0 +1,111 @@
+require File.join(File.dirname(__FILE__), "spec_helper")
+
+describe Sequel::Model, "dataset & schema" do
+  before do
+    @model = Class.new(Sequel::Model(:items))
+  end 
+
+  specify "sets schema with implicit table name" do
+    @model.set_schema do
+      primary_key :ssn, :string
+    end
+    @model.primary_key.should == :ssn
+    @model.table_name.should == :items
+  end
+
+  specify "sets schema with explicit table name" do
+    @model.set_schema :foo do
+      primary_key :id
+    end
+    @model.primary_key.should == :id
+    @model.table_name.should == :foo
+  end
+end
+
+describe Sequel::Model, "table_exists?" do
+
+  before(:each) do
+    MODEL_DB.reset
+    @model = Class.new(Sequel::Model(:items))
+  end
+
+  it "should get the table name and question the model's db if table_exists?" do
+    @model.should_receive(:table_name).and_return(:items)
+    @model.db.should_receive(:table_exists?)
+    @model.table_exists?
+  end
+end
+
+describe Sequel::Model, "create_table and schema" do
+
+  before(:each) do
+    MODEL_DB.reset
+    @model = Class.new(Sequel::Model) do
+      set_schema(:items) do
+        text :name
+        float :price, :null => false
+      end
+    end
+  end
+
+  it "should get the create table SQL list from the db and execute it line by line" do
+    @model.create_table
+    MODEL_DB.sqls.should == ['CREATE TABLE items (name text, price float NOT NULL)']
+  end
+
+  it "should reload the schema from the database" do
+    schem = {:name=>{:type=>:string}, :price=>{:type=>:float}}
+    @model.db.should_receive(:schema).with(:items, :reload=>true).and_return(schem.to_a.sort_by{|x| x[0].to_s})
+    @model.create_table
+    @model.db_schema.should == schem
+    @model.instance_variable_get(:@columns).should == [:name, :price]
+  end
+
+  it "should return the schema generator via schema" do
+    @model.schema.should be_a_kind_of(Sequel::Schema::Generator)
+  end
+
+  it "should use the superclasses schema if it exists" do
+    @submodel = Class.new(@model)
+    @submodel.schema.should be_a_kind_of(Sequel::Schema::Generator)
+  end
+
+  it "should return nil if no schema is present" do
+    @model = Class.new(Sequel::Model)
+    @model.schema.should == nil
+    @submodel = Class.new(@model)
+    @submodel.schema.should == nil
+  end
+end
+
+describe Sequel::Model, "drop_table" do
+
+  before(:each) do
+    MODEL_DB.reset
+    @model = Class.new(Sequel::Model(:items))
+  end
+
+  it "should get the drop table SQL for the associated table and then execute the SQL." do
+    @model.should_receive(:table_name).and_return(:items)
+    @model.db.should_receive(:drop_table_sql).with(:items)
+    @model.db.should_receive(:execute).and_return(:true)
+    @model.drop_table
+  end
+
+end
+
+describe Sequel::Model, "create_table!" do
+
+  before(:each) do
+    MODEL_DB.reset
+    @model = Class.new(Sequel::Model(:items))
+  end
+  
+  it "should drop table if it exists and then create the table" do
+    @model.should_receive(:drop_table).and_return(true)
+    @model.should_receive(:create_table).and_return(true)
+
+    @model.create_table!
+  end
+
+end
diff -pruN 1.5.1-1/spec/extensions/serialization_spec.rb 3.0.0-1/spec/extensions/serialization_spec.rb
--- 1.5.1-1/spec/extensions/serialization_spec.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/extensions/serialization_spec.rb	2009-04-21 21:57:46.000000000 +0100
@@ -0,0 +1,118 @@
+require File.join(File.dirname(__FILE__), "spec_helper")
+
+require 'yaml'
+
+describe "Serialization plugin" do
+  before do
+    @c = Class.new(Sequel::Model(:items)) do
+      include(Module.new do
+        def before_save
+        end
+      end)
+      no_primary_key
+      columns :id, :abc, :def
+    end
+    MODEL_DB.reset
+  end
+
+  it "should allow serializing attributes to yaml" do
+    @c.plugin :serialization, :yaml, :abc
+    @c.create(:abc => 1)
+    @c.create(:abc => "hello")
+
+    MODEL_DB.sqls.should == [ \
+      "INSERT INTO items (abc) VALUES ('--- 1\n')", \
+      "INSERT INTO items (abc) VALUES ('--- hello\n')", \
+    ]
+  end
+
+  it "should allow serializing attributes to marshal" do
+    @c.plugin :serialization, :marshal, :abc
+    @c.create(:abc => 1)
+    @c.create(:abc => "hello")
+    x = [Marshal.dump("hello")].pack('m')
+
+    MODEL_DB.sqls.should == [ \
+      "INSERT INTO items (abc) VALUES ('BAhpBg==\n')", \
+      "INSERT INTO items (abc) VALUES ('#{x}')", \
+    ]
+  end
+
+  it "should translate values to and from yaml serialization format using accessor methods" do
+    @c.set_primary_key :id
+    @c.plugin :serialization, :yaml, :abc, :def
+    vals = nil
+
+    ds = @c.dataset
+    def ds.fetch_rows(sql, &block)
+      block.call(:id => 1, :abc => "--- 1\n", :def => "--- hello\n")
+    end
+
+    o = @c.first
+    o.id.should == 1
+    o.abc.should == 1
+    o.abc.should == 1
+    o.def.should == "hello"
+    o.def.should == "hello"
+
+    o.update(:abc => 23)
+    @c.create(:abc => [1, 2, 3])
+    MODEL_DB.sqls.should == ["UPDATE items SET abc = '--- 23\n' WHERE (id = 1)",
+      "INSERT INTO items (abc) VALUES ('#{[1, 2, 3].to_yaml}')"]
+  end
+
+  it "should translate values to and from marshal serialization format using accessor methods" do
+    @c.set_primary_key :id
+    @c.plugin :serialization, :marshal, :abc, :def
+
+    ds = @c.dataset
+    def ds.fetch_rows(sql, &block)
+      block.call(:id => 1, :abc =>[Marshal.dump(1)].pack('m'), :def =>[Marshal.dump('hello')].pack('m'))
+    end
+
+    o = @c.first
+    o.id.should == 1
+    o.abc.should == 1
+    o.abc.should == 1
+    o.def.should == "hello"
+    o.def.should == "hello"
+
+    o.update(:abc => 23)
+    @c.create(:abc => [1, 2, 3])
+    MODEL_DB.sqls.should == ["UPDATE items SET abc = '#{[Marshal.dump(23)].pack('m')}' WHERE (id = 1)",
+      "INSERT INTO items (abc) VALUES ('#{[Marshal.dump([1, 2, 3])].pack('m')}')"]
+  end
+
+  it "should copy serialization formats and columns to subclasses" do
+    @c.set_primary_key :id
+    @c.plugin :serialization, :yaml, :abc, :def
+
+    ds = @c.dataset
+    def ds.fetch_rows(sql, &block)
+      block.call(:id => 1, :abc => "--- 1\n", :def => "--- hello\n")
+    end
+
+    o = Class.new(@c).first
+    o.id.should == 1
+    o.abc.should == 1
+    o.abc.should == 1
+    o.def.should == "hello"
+    o.def.should == "hello"
+
+    o.update(:abc => 23)
+    Class.new(@c).create(:abc => [1, 2, 3])
+    MODEL_DB.sqls.should == ["UPDATE items SET abc = '--- 23\n' WHERE (id = 1)",
+      "INSERT INTO items (abc) VALUES ('#{[1, 2, 3].to_yaml}')"]
+  end
+
+  it "should clear the deserialized columns when refreshing" do
+    @c.set_primary_key :id
+    @c.plugin :serialization, :yaml, :abc, :def
+    o = @c.load(:id => 1, :abc => "--- 1\n", :def => "--- hello\n")
+    o.abc = 23
+    o.deserialized_values.length.should == 1
+    o.abc.should == 23
+    o.refresh
+    o.deserialized_values.length.should == 0
+  end
+end
diff -pruN 1.5.1-1/spec/extensions/single_table_inheritance_spec.rb 3.0.0-1/spec/extensions/single_table_inheritance_spec.rb
--- 1.5.1-1/spec/extensions/single_table_inheritance_spec.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/extensions/single_table_inheritance_spec.rb	2009-04-29 18:03:41.000000000 +0100
@@ -0,0 +1,58 @@
+require File.join(File.dirname(__FILE__), "spec_helper")
+
+describe Sequel::Model, "#sti_key" do
+  before do
+    class ::StiTest < Sequel::Model
+      def kind=(x); self[:kind] = x; end 
+      def _refresh(x); end 
+      plugin :single_table_inheritance, :kind
+    end 
+    class ::StiTestSub1 < StiTest
+    end 
+    class ::StiTestSub2 < StiTest
+    end 
+    @ds = StiTest.dataset
+    MODEL_DB.reset
+  end 
+
+  specify "should have simple_table = nil" do
+    StiTest.simple_table.should == nil
+    StiTestSub1.simple_table.should == nil
+  end
+  
+  it "should return rows with the correct class based on the polymorphic_key value" do
+    def @ds.fetch_rows(sql)
+      yield({:kind=>'StiTest'})
+      yield({:kind=>'StiTestSub1'})
+      yield({:kind=>'StiTestSub2'})
+    end 
+    StiTest.all.collect{|x| x.class}.should == [StiTest, StiTestSub1, StiTestSub2]
+  end 
+
+  it "should fallback to the main class if polymophic_key value is NULL" do
+    def @ds.fetch_rows(sql)
+      yield({:kind=>nil})
+    end 
+    StiTest.all.collect{|x| x.class}.should == [StiTest]
+  end 
+  
+  it "should fallback to the main class if the given class does not exist" do
+    def @ds.fetch_rows(sql)
+      yield({:kind=>'StiTestSub3'})
+    end
+    StiTest.all.collect{|x| x.class}.should == [StiTest]
+  end
+
+  it "should add a before_create hook that sets the model class name for the key" do
+    StiTest.new.save
+    StiTestSub1.new.save
+    StiTestSub2.new.save
+    MODEL_DB.sqls.should == ["INSERT INTO sti_tests (kind) VALUES ('StiTest')", "INSERT INTO sti_tests (kind) VALUES ('StiTestSub1')", "INSERT INTO sti_tests (kind) VALUES ('StiTestSub2')"]
+  end
+
+  it "should add a filter to model datasets inside subclasses hook to only retreive objects with the matching key" do
+    StiTest.dataset.sql.should == "SELECT * FROM sti_tests"
+    StiTestSub1.dataset.sql.should == "SELECT * FROM sti_tests WHERE (kind = 'StiTestSub1')"
+    StiTestSub2.dataset.sql.should == "SELECT * FROM sti_tests WHERE (kind = 'StiTestSub2')"
+  end
+end
diff -pruN 1.5.1-1/spec/extensions/spec_helper.rb 3.0.0-1/spec/extensions/spec_helper.rb
--- 1.5.1-1/spec/extensions/spec_helper.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/extensions/spec_helper.rb	2009-04-20 23:01:34.000000000 +0100
@@ -0,0 +1,88 @@
+require 'rubygems'
+unless Object.const_defined?('Sequel')
+  $:.unshift(File.join(File.dirname(__FILE__), "../../lib/"))
+  require 'sequel/core'
+end
+unless Sequel.const_defined?('Model')
+  $:.unshift(File.join(File.dirname(__FILE__), "../../lib/"))
+  require 'sequel/model'
+end
+
+Sequel.virtual_row_instance_eval = true
+
+extensions = %w'string_date_time inflector pagination query pretty_table blank migration schema_dumper'
+plugins = {:hook_class_methods=>[], :schema=>[], :validation_class_methods=>[]}
+
+extensions.each{|e| require "sequel/extensions/#{e}"}
+plugins.each{|p, opts| Sequel::Model.plugin(p, *opts)}
+
+class MockDataset < Sequel::Dataset
+  def insert(*args)
+    @db.execute insert_sql(*args)
+  end
+  
+  def update(*args)
+    @db.execute update_sql(*args)
+  end
+  
+  def delete(*args)
+    @db.execute delete_sql(*args)
+  end
+  
+  def fetch_rows(sql)
+    return if sql =~ /information_schema/
+    @db.execute(sql)
+    yield({:id => 1, :x => 1})
+  end
+
+  def quoted_identifier(c)
+    "\"#{c}\""
+  end
+end
+
+class MockDatabase < Sequel::Database
+  @@quote_identifiers = false
+  self.identifier_input_method = nil
+  self.identifier_output_method = nil
+  attr_reader :sqls
+  
+  def execute(sql, opts={})
+    @sqls ||= []
+    @sqls << sql
+  end
+
+  def reset
+    @sqls = []
+  end
+
+  def schema(table_name, opts)
+    if table_name
+      [[:id, {:primary_key=>true}]]
+    else
+      {table_name=>[[:id, {:primary_key=>true}]]}
+    end
+  end
+
+  def transaction(opts={}); yield; end
+  
+  def dataset(opts=nil); MockDataset.new(self, opts); end
+end
+
+class << Sequel::Model
+  alias orig_columns columns
+  def columns(*cols)
+    return if cols.empty?
+    define_method(:columns){cols}
+    @dataset.instance_variable_set(:@columns, cols) if @dataset
+    def_column_accessor(*cols)
+    @columns = cols
+    @db_schema = {}
+    cols.each{|c| @db_schema[c] = {}}
+  end
+  def simple_table
+    nil
+  end
+end
+
+Sequel::Model.db = MODEL_DB = MockDatabase.new
+Sequel::Model.use_transactions = false
diff -pruN 1.5.1-1/spec/extensions/string_date_time_spec.rb 3.0.0-1/spec/extensions/string_date_time_spec.rb
--- 1.5.1-1/spec/extensions/string_date_time_spec.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/extensions/string_date_time_spec.rb	2009-03-23 17:12:44.000000000 +0000
@@ -0,0 +1,93 @@
+require File.join(File.dirname(__FILE__), 'spec_helper')
+
+context "String#to_time" do
+  specify "should convert the string into a Time object" do
+    "2007-07-11".to_time.should == Time.parse("2007-07-11")
+    "06:30".to_time.should == Time.parse("06:30")
+  end
+  
+  specify "should raise InvalidValue for an invalid time" do
+    proc {'0000-00-00'.to_time}.should raise_error(Sequel::InvalidValue)
+  end
+end
+
+context "String#to_date" do
+  after do
+    Sequel.convert_two_digit_years = true
+  end
+
+  specify "should convert the string into a Date object" do
+    "2007-07-11".to_date.should == Date.parse("2007-07-11")
+  end
+  
+  specify "should convert 2 digit years by default" do
+    "July 11, 07".to_date.should == Date.parse("2007-07-11")
+  end
+
+  specify "should not convert 2 digit years if set not to" do
+    Sequel.convert_two_digit_years = false
+    "July 11, 07".to_date.should == Date.parse("0007-07-11")
+  end
+
+  specify "should raise InvalidValue for an invalid date" do
+    proc {'0000-00-00'.to_date}.should raise_error(Sequel::InvalidValue)
+  end
+end
+
+context "String#to_datetime" do
+  after do
+    Sequel.convert_two_digit_years = true
+  end
+
+  specify "should convert the string into a DateTime object" do
+    "2007-07-11 10:11:12a".to_datetime.should == DateTime.parse("2007-07-11 10:11:12a")
+  end
+  
+  specify "should convert 2 digit years by default" do
+    "July 11, 07 10:11:12a".to_datetime.should == DateTime.parse("2007-07-11 10:11:12a")
+  end
+
+  specify "should not convert 2 digit years if set not to" do
+    Sequel.convert_two_digit_years = false
+    "July 11, 07 10:11:12a".to_datetime.should == DateTime.parse("0007-07-11 10:11:12a")
+  end
+
+  specify "should raise InvalidValue for an invalid date" do
+    proc {'0000-00-00'.to_datetime}.should raise_error(Sequel::InvalidValue)
+  end
+end
+
+context "String#to_sequel_time" do
+  after do
+    Sequel.datetime_class = Time
+    Sequel.convert_two_digit_years = true
+  end
+
+  specify "should convert the string into a Time object by default" do
+    "2007-07-11 10:11:12a".to_sequel_time.class.should == Time
+    "2007-07-11 10:11:12a".to_sequel_time.should == Time.parse("2007-07-11 10:11:12a")
+  end
+  
+  specify "should convert the string into a DateTime object if that is set" do
+    Sequel.datetime_class = DateTime
+    "2007-07-11 10:11:12a".to_sequel_time.class.should == DateTime
+    "2007-07-11 10:11:12a".to_sequel_time.should == DateTime.parse("2007-07-11 10:11:12a")
+  end
+  
+  specify "should convert 2 digit years by default if using DateTime class" do
+    Sequel.datetime_class = DateTime
+    "July 11, 07 10:11:12a".to_sequel_time.should == DateTime.parse("2007-07-11 10:11:12a")
+  end
+
+  specify "should not convert 2 digit years if set not to when using DateTime class" do
+    Sequel.datetime_class = DateTime
+    Sequel.convert_two_digit_years = false
+    "July 11, 07 10:11:12a".to_sequel_time.should == DateTime.parse("0007-07-11 10:11:12a")
+  end
+
+  specify "should raise InvalidValue for an invalid time" do
+    proc {'0000-00-00'.to_sequel_time}.should raise_error(Sequel::InvalidValue)
+    Sequel.datetime_class = DateTime
+    proc {'0000-00-00'.to_sequel_time}.should raise_error(Sequel::InvalidValue)
+  end
+end
diff -pruN 1.5.1-1/spec/extensions/validation_class_methods_spec.rb 3.0.0-1/spec/extensions/validation_class_methods_spec.rb
--- 1.5.1-1/spec/extensions/validation_class_methods_spec.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/extensions/validation_class_methods_spec.rb	2009-03-26 23:36:27.000000000 +0000
@@ -0,0 +1,979 @@
+require File.join(File.dirname(__FILE__), "spec_helper")
+
+describe Sequel::Model do
+  before do
+    @c = Class.new(Sequel::Model) do
+      def self.validates_coolness_of(attr)
+        validates_each(attr) {|o, a, v| o.errors[a] << 'is not cool' if v != :cool}
+      end
+    end
+  end
+  
+  specify "should respond to validates, validations, has_validations?" do
+    @c.should respond_to(:validations)
+    @c.should respond_to(:has_validations?)
+  end
+  
+  specify "should acccept validation definitions using validates_each" do
+    @c.validates_each(:xx, :yy) {|o, a, v| o.errors[a] << 'too low' if v < 50}
+    o = @c.new
+    o.should_receive(:xx).once.and_return(40)
+    o.should_receive(:yy).once.and_return(60)
+    o.valid?.should == false
+    o.errors.full_messages.should == ['xx too low']
+  end
+
+  specify "should return true/false for has_validations?" do
+    @c.has_validations?.should == false
+    @c.validates_each(:xx) {1}
+    @c.has_validations?.should == true
+  end
+  
+  specify "should validate multiple attributes at once" do
+    o = @c.new
+    def o.xx
+      1
+    end
+    def o.yy
+      2
+    end
+    vals = nil
+    atts = nil
+    @c.validates_each([:xx, :yy]){|obj,a,v| atts=a; vals=v}
+    o.valid?
+    vals.should == [1,2]
+    atts.should == [:xx, :yy]
+  end
+  
+  specify "should respect allow_missing option when using multiple attributes" do
+    o = @c.new
+    def o.xx
+      self[:xx]
+    end
+    def o.yy
+      self[:yy]
+    end
+    vals = nil
+    atts = nil
+    @c.validates_each([:xx, :yy], :allow_missing=>true){|obj,a,v| atts=a; vals=v}
+
+    o.values[:xx] = 1
+    o.valid?
+    vals.should == [1,nil]
+    atts.should == [:xx, :yy]
+
+    vals = nil
+    atts = nil
+    o.values.clear
+    o.values[:yy] = 2
+    o.valid?
+    vals.should == [nil, 2]
+    atts.should == [:xx, :yy]
+
+    vals = nil
+    atts = nil
+    o.values.clear
+    o.valid?.should == true
+    vals.should == nil
+    atts.should == nil
+  end
+  
+  specify "should overwrite existing validation with the same tag and attribute" do
+    @c.validates_each(:xx, :xx, :tag=>:low) {|o, a, v| o.xxx; o.errors[a] << 'too low' if v < 50}
+    @c.validates_each(:yy, :yy) {|o, a, v| o.yyy; o.errors[a] << 'too low' if v < 50}
+    @c.validates_presence_of(:zz, :zz)
+    @c.validates_length_of(:aa, :aa, :tag=>:blah)
+    o = @c.new
+    def o.zz
+      @a ||= 0
+      @a += 1
+    end
+    def o.aa
+      @b ||= 0
+      @b += 1
+    end
+    o.should_receive(:xx).once.and_return(40)
+    o.should_receive(:yy).once.and_return(60)
+    o.should_receive(:xxx).once
+    o.should_receive(:yyy).twice
+    o.valid?.should == false
+    o.zz.should == 2
+    o.aa.should == 2
+    o.errors.full_messages.should == ['xx too low']
+  end
+
+  specify "should provide a validates method that takes block with validation definitions" do
+    @c.validates do
+      coolness_of :blah
+    end
+    @c.validations[:blah].should_not be_empty
+    o = @c.new
+    o.should_receive(:blah).once.and_return(nil)
+    o.valid?.should == false
+    o.errors.full_messages.should == ['blah is not cool']
+  end
+end
+
+describe Sequel::Model do
+  before do
+    @c = Class.new(Sequel::Model) do
+      columns :score
+      validates_each :score do |o, a, v|
+        o.errors[a] << 'too low' if v < 87
+      end
+    end
+    
+    @o = @c.new
+  end
+  
+  specify "should supply a #valid? method that returns true if validations pass" do
+    @o.score = 50
+    @o.should_not be_valid
+    @o.score = 100
+    @o.should be_valid
+  end
+  
+  specify "should provide an errors object" do
+    @o.score = 100
+    @o.should be_valid
+    @o.errors.should be_empty
+    
+    @o.score = 86
+    @o.should_not be_valid
+    @o.errors[:score].should == ['too low']
+    @o.errors[:blah].should be_empty
+  end
+end
+
+describe Sequel::Plugins::ValidationClassMethods::ClassMethods::Generator do
+  before do
+    $testit = nil
+    
+    @c = Class.new(Sequel::Model) do
+      def self.validates_blah
+        $testit = 1324
+      end
+    end
+  end
+  
+  specify "should instance_eval the block, sending everything to its receiver" do
+    @c.validates do
+      blah
+    end
+    $testit.should == 1324
+  end
+end
+
+describe Sequel::Model do
+  before do
+    @c = Class.new(Sequel::Model) do
+      columns :value
+      
+      def self.filter(*args)
+        o = Object.new
+        def o.count; 2; end
+        o
+      end
+
+      def skip; false; end
+      def dont_skip; true; end
+    end
+    @m = @c.new
+  end
+
+  specify "should validate acceptance_of" do
+    @c.validates_acceptance_of :value
+    @m.should be_valid
+    @m.value = '1'
+    @m.should be_valid
+  end
+  
+  specify "should validate acceptance_of with accept" do
+    @c.validates_acceptance_of :value, :accept => 'true'
+    @m.value = '1'
+    @m.should_not be_valid
+    @m.value = 'true'
+    @m.should be_valid
+  end
+  
+  specify "should validate acceptance_of with allow_nil => false" do
+    @c.validates_acceptance_of :value, :allow_nil => false
+    @m.should_not be_valid
+  end
+
+  specify "should validate acceptance_of with allow_missing => true" do
+    @c.validates_acceptance_of :value, :allow_missing => true
+    @m.should be_valid
+  end
+
+  specify "should validate acceptance_of with allow_missing => true and allow_nil => false" do
+    @c.validates_acceptance_of :value, :allow_missing => true, :allow_nil => false
+    @m.should be_valid
+    @m.value = nil
+    @m.should_not be_valid
+  end
+
+  specify "should validate acceptance_of with if => true" do
+    @c.validates_acceptance_of :value, :if => :dont_skip
+    @m.value = '0'
+    @m.should_not be_valid
+  end
+
+  specify "should validate acceptance_of with if => false" do
+    @c.validates_acceptance_of :value, :if => :skip
+    @m.value = '0'
+    @m.should be_valid
+  end
+
+  specify "should validate acceptance_of with if proc that evaluates to true" do
+    @c.validates_acceptance_of :value, :if => proc{true}
+    @m.value = '0'
+    @m.should_not be_valid
+  end
+
+  specify "should validate acceptance_of with if proc that evaluates to false" do
+    @c.validates_acceptance_of :value, :if => proc{false}
+    @m.value = '0'
+    @m.should be_valid
+  end
+
+  specify "should raise an error if :if option is not a Symbol, Proc, or nil" do
+    @c.validates_acceptance_of :value, :if => 1
+    @m.value = '0'
+    proc{@m.valid?}.should raise_error(Sequel::Error)
+  end
+
+  specify "should validate confirmation_of" do
+    @c.send(:attr_accessor, :value_confirmation)
+    @c.validates_confirmation_of :value
+    
+    @m.value = 'blah'
+    @m.should_not be_valid
+    
+    @m.value_confirmation = 'blah'
+    @m.should be_valid
+  end
+  
+  specify "should validate confirmation_of with if => true" do
+    @c.send(:attr_accessor, :value_confirmation)
+    @c.validates_confirmation_of :value, :if => :dont_skip
+
+    @m.value = 'blah'
+    @m.should_not be_valid
+  end
+
+  specify "should validate confirmation_of with if => false" do
+    @c.send(:attr_accessor, :value_confirmation)
+    @c.validates_confirmation_of :value, :if => :skip
+
+    @m.value = 'blah'
+    @m.should be_valid
+  end
+
+  specify "should validate confirmation_of with allow_missing => true" do
+    @c.send(:attr_accessor, :value_confirmation)
+    @c.validates_acceptance_of :value, :allow_missing => true
+    @m.should be_valid
+    @m.value_confirmation = 'blah'
+    @m.should be_valid
+    @m.value = nil
+    @m.should_not be_valid
+  end
+
+  specify "should validate format_of" do
+    @c.validates_format_of :value, :with => /.+_.+/
+    @m.value = 'abc_'
+    @m.should_not be_valid
+    @m.value = 'abc_def'
+    @m.should be_valid
+  end
+  
+  specify "should raise for validate_format_of without regexp" do
+    proc {@c.validates_format_of :value}.should raise_error(ArgumentError)
+    proc {@c.validates_format_of :value, :with => :blah}.should raise_error(ArgumentError)
+  end
+  
+  specify "should validate format_of with if => true" do
+    @c.validates_format_of :value, :with => /_/, :if => :dont_skip
+
+    @m.value = 'a'
+    @m.should_not be_valid
+  end
+
+  specify "should validate format_of with if => false" do
+    @c.validates_format_of :value, :with => /_/, :if => :skip
+
+    @m.value = 'a'
+    @m.should be_valid
+  end
+  
+  specify "should validate format_of with allow_missing => true" do
+    @c.validates_format_of :value, :allow_missing => true, :with=>/./
+    @m.should be_valid
+    @m.value = nil
+    @m.should_not be_valid
+  end
+
+  specify "should validate length_of with maximum" do
+    @c.validates_length_of :value, :maximum => 5
+    @m.should_not be_valid
+    @m.value = '12345'
+    @m.should be_valid
+    @m.value = '123456'
+    @m.should_not be_valid
+  end
+
+  specify "should validate length_of with minimum" do
+    @c.validates_length_of :value, :minimum => 5
+    @m.should_not be_valid
+    @m.value = '12345'
+    @m.should be_valid
+    @m.value = '1234'
+    @m.should_not be_valid
+  end
+
+  specify "should validate length_of with within" do
+    @c.validates_length_of :value, :within => 2..5
+    @m.should_not be_valid
+    @m.value = '12345'
+    @m.should be_valid
+    @m.value = '1'
+    @m.should_not be_valid
+    @m.value = '123456'
+    @m.should_not be_valid
+  end
+
+  specify "should validate length_of with is" do
+    @c.validates_length_of :value, :is => 3
+    @m.should_not be_valid
+    @m.value = '123'
+    @m.should be_valid
+    @m.value = '12'
+    @m.should_not be_valid
+    @m.value = '1234'
+    @m.should_not be_valid
+  end
+  
+  specify "should validate length_of with allow_nil" do
+    @c.validates_length_of :value, :is => 3, :allow_nil => true
+    @m.should be_valid
+  end
+
+  specify "should validate length_of with if => true" do
+    @c.validates_length_of :value, :is => 3, :if => :dont_skip
+
+    @m.value = 'a'
+    @m.should_not be_valid
+  end
+
+  specify "should validate length_of with if => false" do
+    @c.validates_length_of :value, :is => 3, :if => :skip
+
+    @m.value = 'a'
+    @m.should be_valid
+  end
+
+  specify "should validate length_of with allow_missing => true" do
+    @c.validates_length_of :value, :allow_missing => true, :minimum => 5
+    @m.should be_valid
+    @m.value = nil
+    @m.should_not be_valid
+  end
+
+  specify "should allow multiple calls to validates_length_of with different options without overwriting" do
+    @c.validates_length_of :value, :maximum => 5
+    @c.validates_length_of :value, :minimum => 5
+    @m.should_not be_valid
+    @m.value = '12345'
+    @m.should be_valid
+    @m.value = '123456'
+    @m.should_not be_valid
+    @m.value = '12345'
+    @m.should be_valid
+    @m.value = '1234'
+    @m.should_not be_valid
+  end
+
+  specify "should validate numericality_of" do
+    @c.validates_numericality_of :value
+    @m.value = 'blah'
+    @m.should_not be_valid
+    @m.value = '123'
+    @m.should be_valid
+    @m.value = '123.1231'
+    @m.should be_valid
+    @m.value = '+1'
+    @m.should be_valid
+    @m.value = '-1'
+    @m.should be_valid
+    @m.value = '+1.123'
+    @m.should be_valid
+    @m.value = '-0.123'
+    @m.should be_valid
+    @m.value = '-0.123E10'
+    @m.should be_valid
+    @m.value = '32.123e10'
+    @m.should be_valid
+    @m.value = '+32.123E10'
+    @m.should be_valid
+    @m.should be_valid
+    @m.value = '.0123'
+  end
+
+  specify "should validate numericality_of with only_integer" do
+    @c.validates_numericality_of :value, :only_integer => true
+    @m.value = 'blah'
+    @m.should_not be_valid
+    @m.value = '123'
+    @m.should be_valid
+    @m.value = '123.1231'
+    @m.should_not be_valid
+  end
+  
+  specify "should validate numericality_of with if => true" do
+    @c.validates_numericality_of :value, :if => :dont_skip
+
+    @m.value = 'a'
+    @m.should_not be_valid
+  end
+
+  specify "should validate numericality_of with if => false" do
+    @c.validates_numericality_of :value, :if => :skip
+
+    @m.value = 'a'
+    @m.should be_valid
+  end
+
+  specify "should validate numericality_of with allow_missing => true" do
+    @c.validates_numericality_of :value, :allow_missing => true
+    @m.should be_valid
+    @m.value = nil
+    @m.should_not be_valid
+  end
+
+  specify "should validate presence_of" do
+    @c.validates_presence_of :value
+    @m.should_not be_valid
+    @m.value = ''
+    @m.should_not be_valid
+    @m.value = 1234
+    @m.should be_valid
+    @m.value = nil
+    @m.should_not be_valid
+    @m.value = true
+    @m.should be_valid
+    @m.value = false
+    @m.should be_valid
+  end
+  
+  specify "should validate inclusion_of with an array" do
+    @c.validates_inclusion_of :value, :in => [1,2]
+    @m.should_not be_valid
+    @m.value = 1
+    @m.should be_valid
+    @m.value = 1.5
+    @m.should_not be_valid
+    @m.value = 2
+    @m.should be_valid    
+    @m.value = 3
+    @m.should_not be_valid 
+  end
+  
+  specify "should validate inclusion_of with a range" do
+    @c.validates_inclusion_of :value, :in => 1..4
+    @m.should_not be_valid
+    @m.value = 1
+    @m.should be_valid
+    @m.value = 1.5
+    @m.should be_valid
+    @m.value = 0
+    @m.should_not be_valid
+    @m.value = 5
+    @m.should_not be_valid    
+  end
+  
+  specify "should raise an error if inclusion_of doesn't receive a valid :in option" do
+    lambda {
+      @c.validates_inclusion_of :value
+    }.should raise_error(ArgumentError)
+    
+    lambda {
+      @c.validates_inclusion_of :value, :in => 1
+    }.should raise_error(ArgumentError)
+  end
+  
+  specify "should raise an error if inclusion_of handles :allow_nil too" do
+    @c.validates_inclusion_of :value, :in => 1..4, :allow_nil => true
+    @m.value = nil
+    @m.should be_valid
+    @m.value = 0
+    @m.should_not be_valid
+  end
+
+  specify "should validate presence_of with if => true" do
+    @c.validates_presence_of :value, :if => :dont_skip
+    @m.should_not be_valid
+  end
+
+  specify "should validate presence_of with if => false" do
+    @c.validates_presence_of :value, :if => :skip
+    @m.should be_valid
+  end
+
+  specify "should validate presence_of with allow_missing => true" do
+    @c.validates_presence_of :value, :allow_missing => true
+    @m.should be_valid
+    @m.value = nil
+    @m.should_not be_valid
+  end
+
+  specify "should validate uniqueness_of with if => true" do
+    @c.validates_uniqueness_of :value, :if => :dont_skip
+
+    @m.value = 'a'
+    @m.should_not be_valid
+  end
+
+  specify "should validate uniqueness_of with if => false" do
+    @c.validates_uniqueness_of :value, :if => :skip
+
+    @m.value = 'a'
+    @m.should be_valid
+  end
+  
+  specify "should validate uniqueness_of with allow_missing => true" do
+    @c.validates_uniqueness_of :value, :allow_missing => true
+    @m.should be_valid
+    @m.value = nil
+    @m.should_not be_valid
+  end
+end
+
+context "Superclass validations" do
+  before do
+    @c1 = Class.new(Sequel::Model) do
+      columns :value
+      validates_length_of :value, :minimum => 5
+    end
+    
+    @c2 = Class.new(@c1) do
+      columns :value
+      validates_format_of :value, :with => /^[a-z]+$/
+    end
+  end
+  
+  specify "should be checked when validating" do
+    o = @c2.new
+    o.value = 'ab'
+    o.valid?.should == false
+    o.errors.full_messages.should == [
+      'value is too short'
+    ]
+
+    o.value = '12'
+    o.valid?.should == false
+    o.errors.full_messages.should == [
+      'value is too short',
+      'value is invalid'
+    ]
+
+    o.value = 'abcde'
+    o.valid?.should be_true
+  end
+  
+  specify "should be skipped if skip_superclass_validations is called" do
+    @c2.skip_superclass_validations
+
+    o = @c2.new
+    o.value = 'ab'
+    o.valid?.should be_true
+
+    o.value = '12'
+    o.valid?.should == false
+    o.errors.full_messages.should == [
+      'value is invalid'
+    ]
+
+    o.value = 'abcde'
+    o.valid?.should be_true
+  end
+end
+
+context ".validates with block" do
+  specify "should support calling .each" do
+    @c = Class.new(Sequel::Model) do
+      columns :vvv
+      validates do
+        each :vvv do |o, a, v|
+          o.errors[a] << "is less than zero" if v.to_i < 0
+        end
+      end
+    end
+    
+    o = @c.new
+    o.vvv = 1
+    o.should be_valid
+    o.vvv = -1
+    o.should_not be_valid
+  end
+end
+
+describe Sequel::Model, "Validations" do
+
+  before(:all) do
+    class ::Person < Sequel::Model
+      columns :id,:name,:first_name,:last_name,:middle_name,:initials,:age, :terms
+    end
+
+    class ::Smurf < Person
+    end
+    
+    class ::Cow < Sequel::Model
+      columns :id, :name, :got_milk
+    end
+
+    class ::User < Sequel::Model
+      columns :id, :username, :password
+    end
+    
+    class ::Address < Sequel::Model
+      columns :id, :zip_code
+    end
+  end
+  
+  it "should validate the acceptance of a column" do
+    class ::Cow < Sequel::Model
+      validations.clear
+      validates_acceptance_of :got_milk, :accept => 'blah', :allow_nil => false
+    end
+    
+    @cow = Cow.new
+    @cow.should_not be_valid
+    @cow.errors.full_messages.should == ["got_milk is not accepted"]
+    
+    @cow.got_milk = "blah"
+    @cow.should be_valid
+  end
+  
+  it "should validate the confirmation of a column" do
+    class ::User < Sequel::Model
+      def password_confirmation
+        "test"
+      end
+      
+      validations.clear
+      validates_confirmation_of :password
+    end
+    
+    @user = User.new
+    @user.should_not be_valid
+    @user.errors.full_messages.should == ["password is not confirmed"]
+    
+    @user.password = "test"
+    @user.should be_valid
+  end
+  
+  it "should validate format of column" do
+    class ::Person < Sequel::Model
+      validates_format_of :first_name, :with => /^[a-zA-Z]+$/
+    end
+
+    @person = Person.new :first_name => "Lancelot99"
+    @person.valid?.should be_false
+    @person = Person.new :first_name => "Anita"
+    @person.valid?.should be_true
+  end
+  
+  it "should validate length of column" do
+    class ::Person < Sequel::Model
+      validations.clear
+      validates_length_of :first_name, :maximum => 30
+      validates_length_of :last_name, :minimum => 30
+      validates_length_of :middle_name, :within => 1..5
+      validates_length_of :initials, :is => 2
+    end
+    
+    @person = Person.new(
+      :first_name => "Anamethatiswaytofreakinglongandwayoverthirtycharacters",
+      :last_name => "Alastnameunderthirtychars",
+      :initials => "LGC",
+      :middle_name => "danger"
+    )
+    
+    @person.should_not be_valid
+    @person.errors.full_messages.size.should == 4
+    @person.errors.full_messages.should include(
+      'first_name is too long',
+      'last_name is too short',
+      'middle_name is the wrong length',
+      'initials is the wrong length'
+    )
+    
+    @person.first_name  = "Lancelot"
+    @person.last_name   = "1234567890123456789012345678901"
+    @person.initials    = "LC"
+    @person.middle_name = "Will"
+    @person.should be_valid
+  end
+  
+  it "should validate that a column doesn't have a string value" do
+    p = Class.new(Sequel::Model)
+    p.class_eval do
+      columns :age, :price, :confirmed
+      self.raise_on_typecast_failure = false
+      validates_not_string :age
+      validates_not_string :confirmed
+      validates_not_string :price, :message=>'is not valid'
+      @db_schema = {:age=>{:type=>:integer}}
+    end
+    
+    @person = p.new
+    @person.should be_valid
+
+    @person.confirmed = 't'
+    @person.should_not be_valid
+    @person.errors.full_messages.should == ['confirmed is a string']
+    @person.confirmed = true
+    @person.should be_valid
+
+    @person.age = 'a'
+    @person.should_not be_valid
+    @person.errors.full_messages.should == ['age is not a valid integer']
+    @person.db_schema[:age][:type] = :datetime
+    @person.should_not be_valid
+    @person.errors.full_messages.should == ['age is not a valid datetime']
+    @person.age = 20
+    @person.should be_valid
+
+    @person.price = 'a'
+    @person.should_not be_valid
+    @person.errors.full_messages.should == ['price is not valid']
+    @person.price = 20
+    @person.should be_valid
+  end
+  
+  it "should validate numericality of column" do
+    class ::Person < Sequel::Model
+      validations.clear
+      validates_numericality_of :age
+    end
+    
+    @person = Person.new :age => "Twenty"
+    @person.should_not be_valid
+    @person.errors.full_messages.should == ['age is not a number']
+    
+    @person.age = 20
+    @person.should be_valid
+  end
+  
+  it "should validate the presence of a column" do
+    class ::Cow < Sequel::Model
+      validations.clear
+      validates_presence_of :name
+    end
+    
+    @cow = Cow.new
+    @cow.should_not be_valid
+    @cow.errors.full_messages.should == ['name is not present']
+    
+    @cow.name = "Betsy"
+    @cow.should be_valid
+  end
+ 
+  it "should validate the uniqueness of a column" do
+    class ::User < Sequel::Model
+      validations.clear
+      validates do
+        uniqueness_of :username
+      end
+    end
+    User.dataset.extend(Module.new {
+      def fetch_rows(sql)
+        @db << sql
+        
+        case sql
+        when /COUNT.*username = '0records'/
+          yield({:v => 0})
+        when /COUNT.*username = '2records'/
+          yield({:v => 2})
+        when /COUNT.*username = '1record'/
+          yield({:v => 1})
+        when /username = '1record'/
+          yield({:id => 3, :username => "1record", :password => "test"})
+        end
+      end
+    })
+    
+    @user = User.new(:username => "2records", :password => "anothertest")
+    @user.should_not be_valid
+    @user.errors.full_messages.should == ['username is already taken']
+
+    @user = User.new(:username => "1record", :password => "anothertest")
+    @user.should_not be_valid
+    @user.errors.full_messages.should == ['username is already taken']
+
+    @user = User.load(:id=>4, :username => "1record", :password => "anothertest")
+    @user.should_not be_valid
+    @user.errors.full_messages.should == ['username is already taken']
+
+    @user = User.load(:id=>3, :username => "1record", :password => "anothertest")
+    @user.should be_valid
+    @user.errors.full_messages.should == []
+
+    @user = User.new(:username => "0records", :password => "anothertest")
+    @user.should be_valid
+    @user.errors.full_messages.should == []
+  end
+  
+  it "should validate the uniqueness of multiple columns" do
+    class ::User < Sequel::Model
+      validations.clear
+      validates do
+        uniqueness_of [:username, :password]
+      end
+    end
+    User.dataset.extend(Module.new {
+      def fetch_rows(sql)
+        @db << sql
+        
+        case sql
+        when /COUNT.*username = '0records'/
+          yield({:v => 0})
+        when /COUNT.*username = '2records'/
+          yield({:v => 2})
+        when /COUNT.*username = '1record'/
+          yield({:v => 1})
+        when /username = '1record'/
+          if sql =~ /password = 'anothertest'/
+            yield({:id => 3, :username => "1record", :password => "anothertest"})
+          else
+            yield({:id => 4, :username => "1record", :password => "test"})
+          end
+        end
+      end
+    })
+    
+    @user = User.new(:username => "2records", :password => "anothertest")
+    @user.should_not be_valid
+    @user.errors.full_messages.should == ['username and password is already taken']
+
+    @user = User.new(:username => "1record", :password => "anothertest")
+    @user.should_not be_valid
+    @user.errors.full_messages.should == ['username and password is already taken']
+
+    @user = User.load(:id=>4, :username => "1record", :password => "anothertest")
+    @user.should_not be_valid
+    @user.errors.full_messages.should == ['username and password is already taken']
+
+    @user = User.load(:id=>3, :username => "1record", :password => "test")
+    @user.should_not be_valid
+    @user.errors.full_messages.should == ['username and password is already taken']
+
+    @user = User.load(:id=>3, :username => "1record", :password => "anothertest")
+    @user.should be_valid
+    @user.errors.full_messages.should == []
+
+    @user = User.new(:username => "0records", :password => "anothertest")
+    @user.should be_valid
+    @user.errors.full_messages.should == []
+  end
+  
+  it "should have a validates block that contains multiple validations" do
+    class ::Person < Sequel::Model
+      validations.clear
+      validates do
+        format_of :first_name, :with => /^[a-zA-Z]+$/
+        length_of :first_name, :maximum => 30
+      end
+    end
+
+    Person.validations[:first_name].size.should == 2
+    
+    @person = Person.new :first_name => "Lancelot99"
+    @person.valid?.should be_false
+    
+    @person2 = Person.new :first_name => "Wayne"
+    @person2.valid?.should be_true
+  end
+
+  it "should allow 'longhand' validations direcly within the model." do
+    lambda {
+      class ::Person < Sequel::Model
+        validations.clear
+        validates_length_of :first_name, :maximum => 30
+      end
+    }.should_not raise_error
+    Person.validations.length.should eql(1)
+  end
+
+  it "should define a has_validations? method which returns true if the model has validations, false otherwise" do
+    class ::Person < Sequel::Model
+      validations.clear
+      validates do
+        format_of :first_name, :with => /\w+/
+        length_of :first_name, :maximum => 30
+      end
+    end
+
+    class ::Smurf < Person
+      validations.clear
+    end
+
+    Person.should have_validations
+    Smurf.should_not have_validations
+  end
+
+  it "should validate correctly instances initialized with string keys" do
+    class ::Can < Sequel::Model
+      columns :id, :name
+      
+      validates_length_of :name, :minimum => 4
+    end
+    
+    Can.new('name' => 'ab').should_not be_valid
+    Can.new('name' => 'abcd').should be_valid
+  end
+  
+end
+
+describe "Model#save" do
+  before do
+    @c = Class.new(Sequel::Model(:people)) do
+      columns :id
+
+      validates_each :id do |o, a, v|
+        o.errors[a] << 'blah' unless v == 5
+      end
+    end
+    @m = @c.load(:id => 4)
+    MODEL_DB.reset
+  end
+
+  specify "should save only if validations pass" do
+    @m.raise_on_save_failure = false
+    @m.should_not be_valid
+    @m.save
+    MODEL_DB.sqls.should be_empty
+    
+    @m.id = 5
+    @m.should be_valid
+    @m.save.should_not be_false
+    MODEL_DB.sqls.should == ['UPDATE people SET id = 5 WHERE (id = 5)']
+  end
+  
+  specify "should skip validations if the :validate=>false option is used" do
+    @m.raise_on_save_failure = false
+    @m.should_not be_valid
+    @m.save(:validate=>false)
+    MODEL_DB.sqls.should == ['UPDATE people SET id = 4 WHERE (id = 4)']
+  end
+    
+  specify "should raise error if validations fail and raise_on_save_faiure is true" do
+    proc{@m.save}.should raise_error(Sequel::ValidationFailed)
+  end
+  
+  specify "should return nil if validations fail and raise_on_save_faiure is false" do
+    @m.raise_on_save_failure = false
+    @m.save.should == nil
+  end
+end
diff -pruN 1.5.1-1/spec/extensions/validation_helpers_spec.rb 3.0.0-1/spec/extensions/validation_helpers_spec.rb
--- 1.5.1-1/spec/extensions/validation_helpers_spec.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/extensions/validation_helpers_spec.rb	2009-04-22 17:12:56.000000000 +0100
@@ -0,0 +1,291 @@
+require File.join(File.dirname(__FILE__), "spec_helper")
+
+describe "Sequel::Plugins::ValidationHelpers" do
+  before do
+    @c = Class.new(Sequel::Model) do
+      def self.set_validations(&block)
+        define_method(:validate, &block)
+      end
+      columns :value
+    end
+    @c.plugin :validation_helpers
+    @m = @c.new
+  end
+
+  specify "should take an :allow_blank option" do
+    @c.set_validations{validates_format(/.+_.+/, :value, :allow_blank=>true)}
+    @m.value = 'abc_'
+    @m.should_not be_valid
+    @m.value = '1_1'
+    @m.should be_valid
+    o = Object.new
+    @m.value = o
+    @m.should_not be_valid
+    def o.blank?
+      true
+    end
+    @m.should be_valid
+  end
+
+  specify "should take an :allow_missing option" do
+    @c.set_validations{validates_format(/.+_.+/, :value, :allow_missing=>true)}
+    @m.values.clear
+    @m.should be_valid
+    @m.value = nil
+    @m.should_not be_valid
+    @m.value = '1_1'
+    @m.should be_valid
+  end
+
+  specify "should take an :allow_nil option" do
+    @c.set_validations{validates_format(/.+_.+/, :value, :allow_nil=>true)}
+    @m.value = 'abc_'
+    @m.should_not be_valid
+    @m.value = '1_1'
+    @m.should be_valid
+    @m.value = nil
+    @m.should be_valid
+  end
+
+  specify "should take a :message option" do
+    @c.set_validations{validates_format(/.+_.+/, :value, :message=>"is so blah")}
+    @m.value = 'abc_'
+    @m.should_not be_valid
+    @m.errors.full_messages.should == ['value is so blah']
+    @m.value = '1_1'
+    @m.should be_valid
+  end
+
+  specify "should take multiple attributes in the same call" do
+    @c.columns :value, :value2
+    @c.set_validations{validates_presence([:value, :value2])}
+    @m.should_not be_valid
+    @m.value = 1
+    @m.should_not be_valid
+    @m.value2 = 1
+    @m.should be_valid
+  end
+
+  specify "should support validates_exact_length" do
+    @c.set_validations{validates_exact_length(3, :value)}
+    @m.should_not be_valid
+    @m.value = '123'
+    @m.should be_valid
+    @m.value = '12'
+    @m.should_not be_valid
+    @m.value = '1234'
+    @m.should_not be_valid
+  end
+  
+  specify "should support validate_format" do
+    @c.set_validations{validates_format(/.+_.+/, :value)}
+    @m.value = 'abc_'
+    @m.should_not be_valid
+    @m.value = 'abc_def'
+    @m.should be_valid
+  end
+  
+  specify "should support validates_includes with an array" do
+    @c.set_validations{validates_includes([1,2], :value)}
+    @m.should_not be_valid
+    @m.value = 1
+    @m.should be_valid
+    @m.value = 1.5
+    @m.should_not be_valid
+    @m.value = 2
+    @m.should be_valid    
+    @m.value = 3
+    @m.should_not be_valid 
+  end
+  
+  specify "should support validates_includes with a range" do
+    @c.set_validations{validates_includes(1..4, :value)}
+    @m.should_not be_valid
+    @m.value = 1
+    @m.should be_valid
+    @m.value = 1.5
+    @m.should be_valid
+    @m.value = 0
+    @m.should_not be_valid
+    @m.value = 5
+    @m.should_not be_valid    
+  end
+  
+  specify "should supports validates_integer" do
+    @c.set_validations{validates_integer(:value)}
+    @m.value = 'blah'
+    @m.should_not be_valid
+    @m.value = '123'
+    @m.should be_valid
+    @m.value = '123.1231'
+    @m.should_not be_valid
+  end
+  
+  specify "should support validates_length_range" do
+    @c.set_validations{validates_length_range(2..5, :value)}
+    @m.should_not be_valid
+    @m.value = '12345'
+    @m.should be_valid
+    @m.value = '1'
+    @m.should_not be_valid
+    @m.value = '123456'
+    @m.should_not be_valid
+  end
+
+  specify "should support validates_max_length" do
+    @c.set_validations{validates_max_length(5, :value)}
+    @m.should_not be_valid
+    @m.value = '12345'
+    @m.should be_valid
+    @m.value = '123456'
+    @m.should_not be_valid
+  end
+
+  specify "should support validates_min_length" do
+    @c.set_validations{validates_min_length(5, :value)}
+    @m.should_not be_valid
+    @m.value = '12345'
+    @m.should be_valid
+    @m.value = '1234'
+    @m.should_not be_valid
+  end
+
+  specify "should support validates_not_string" do
+    @c.set_validations{validates_not_string(:value)}
+    @m.value = 123
+    @m.should be_valid
+    @m.value = '123'
+    @m.should_not be_valid
+    @m.errors.full_messages.should == ['value is a string']
+    @m.meta_def(:db_schema){{:value=>{:type=>:integer}}}
+    @m.should_not be_valid
+    @m.errors.full_messages.should == ['value is not a valid integer']
+  end
+
+  specify "should support validates_numeric" do
+    @c.set_validations{validates_numeric(:value)}
+    @m.value = 'blah'
+    @m.should_not be_valid
+    @m.value = '123'
+    @m.should be_valid
+    @m.value = '123.1231'
+    @m.should be_valid
+    @m.value = '+1'
+    @m.should be_valid
+    @m.value = '-1'
+    @m.should be_valid
+    @m.value = '+1.123'
+    @m.should be_valid
+    @m.value = '-0.123'
+    @m.should be_valid
+    @m.value = '-0.123E10'
+    @m.should be_valid
+    @m.value = '32.123e10'
+    @m.should be_valid
+    @m.value = '+32.123E10'
+    @m.should be_valid
+    @m.should be_valid
+    @m.value = '.0123'
+  end
+
+  specify "should support validates_presence" do
+    @c.set_validations{validates_presence(:value)}
+    @m.should_not be_valid
+    @m.value = ''
+    @m.should_not be_valid
+    @m.value = 1234
+    @m.should be_valid
+    @m.value = nil
+    @m.should_not be_valid
+    @m.value = true
+    @m.should be_valid
+    @m.value = false
+    @m.should be_valid
+    @m.value = Time.now
+    @m.should be_valid
+  end
+  
+  it "should support validates_unique with a single attribute" do
+    @c.columns(:id, :username, :password)
+    @c.set_dataset MODEL_DB[:items]
+    @c.set_validations{validates_unique(:username)}
+    @c.dataset.extend(Module.new {
+      def fetch_rows(sql)
+        @db << sql
+        
+        case sql
+        when /COUNT.*username = '0records'/
+          yield({:v => 0})
+        when /COUNT.*username = '1record'/
+          yield({:v => 1})
+        end
+      end
+    })
+    
+    @user = @c.new(:username => "0records", :password => "anothertest")
+    @user.should be_valid
+    @user = @c.load(:id=>3, :username => "0records", :password => "anothertest")
+    @user.should be_valid
+
+    @user = @c.new(:username => "1record", :password => "anothertest")
+    @user.should_not be_valid
+    @user.errors.full_messages.should == ['username is already taken']
+    @user = @c.load(:id=>4, :username => "1record", :password => "anothertest")
+    @user.should_not be_valid
+    @user.errors.full_messages.should == ['username is already taken']
+
+    ds1 = @c.dataset.filter([[:username, '0records']])
+    ds2 = ds1.exclude(:id=>1)
+    @c.dataset.should_receive(:filter).with([[:username, '0records']]).twice.and_return(ds1)
+    ds1.should_receive(:exclude).with(:id=>1).once.and_return(ds2)
+
+    @user = @c.load(:id=>1, :username => "0records", :password => "anothertest")
+    @user.should be_valid
+    MODEL_DB.sqls.last.should == "SELECT COUNT(*) AS count FROM items WHERE ((username = '0records') AND (id != 1)) LIMIT 1"
+    @user = @c.new(:username => "0records", :password => "anothertest")
+    @user.should be_valid
+    MODEL_DB.sqls.last.should == "SELECT COUNT(*) AS count FROM items WHERE (username = '0records') LIMIT 1"
+  end
+  
+  it "should support validates_unique with multiple attributes" do
+    @c.columns(:id, :username, :password)
+    @c.set_dataset MODEL_DB[:items]
+    @c.set_validations{validates_unique([:username, :password])}
+    @c.dataset.extend(Module.new {
+      def fetch_rows(sql)
+        @db << sql
+        
+        case sql
+        when /COUNT.*username = '0records'/
+          yield({:v => 0})
+        when /COUNT.*username = '1record'/
+          yield({:v => 1})
+        end
+      end
+    })
+    
+    @user = @c.new(:username => "0records", :password => "anothertest")
+    @user.should be_valid
+    @user = @c.load(:id=>3, :username => "0records", :password => "anothertest")
+    @user.should be_valid
+
+    @user = @c.new(:username => "1record", :password => "anothertest")
+    @user.should_not be_valid
+    @user.errors.full_messages.should == ['username and password is already taken']
+    @user = @c.load(:id=>4, :username => "1record", :password => "anothertest")
+    @user.should_not be_valid
+    @user.errors.full_messages.should == ['username and password is already taken']
+
+    ds1 = @c.dataset.filter([[:username, '0records'], [:password, 'anothertest']])
+    ds2 = ds1.exclude(:id=>1)
+    @c.dataset.should_receive(:filter).with([[:username, '0records'], [:password, 'anothertest']]).twice.and_return(ds1)
+    ds1.should_receive(:exclude).with(:id=>1).once.and_return(ds2)
+
+    @user = @c.load(:id=>1, :username => "0records", :password => "anothertest")
+    @user.should be_valid
+    MODEL_DB.sqls.last.should == "SELECT COUNT(*) AS count FROM items WHERE (((username = '0records') AND (password = 'anothertest')) AND (id != 1)) LIMIT 1"
+    @user = @c.new(:username => "0records", :password => "anothertest")
+    @user.should be_valid
+    MODEL_DB.sqls.last.should == "SELECT COUNT(*) AS count FROM items WHERE ((username = '0records') AND (password = 'anothertest')) LIMIT 1"
+  end
+end 
diff -pruN 1.5.1-1/spec/hooks_spec.rb 3.0.0-1/spec/hooks_spec.rb
--- 1.5.1-1/spec/hooks_spec.rb	2008-04-22 20:16:00.000000000 +0100
+++ 3.0.0-1/spec/hooks_spec.rb	1970-01-01 01:00:00.000000000 +0100
@@ -1,272 +0,0 @@
-require File.join(File.dirname(__FILE__), "spec_helper")
-
-describe "Model hooks" do
-  before do
-    MODEL_DB.reset
-
-    @hooks = [
-      :after_initialize,
-      :before_create,
-      :after_create,
-      :before_update,
-      :after_update,
-      :before_save,
-      :after_save,
-      :before_destroy,
-      :after_destroy
-    ]
-    
-    # @hooks.each {|h| Sequel::Model.class_def(h) {}}
-  end
-  
-  specify "should be definable using def <hook name>" do
-    c = Class.new(Sequel::Model) do
-      def before_save
-        "hi there"
-      end
-    end
-    
-    c.new.before_save.should == 'hi there'
-  end
-  
-  specify "should be definable using a block" do
-    $adds = []
-    c = Class.new(Sequel::Model) do
-      before_save {$adds << 'hi'}
-    end
-    
-    c.new.before_save
-    $adds.should == ['hi']
-  end
-  
-  specify "should be definable using a method name" do
-    $adds = []
-    c = Class.new(Sequel::Model) do
-      def bye; $adds << 'bye'; end
-      before_save :bye
-    end
-    
-    c.new.before_save
-    $adds.should == ['bye']
-  end
-  
-  specify "should be additive" do
-    $adds = []
-    c = Class.new(Sequel::Model) do
-      before_save {$adds << 'hyiyie'}
-      before_save {$adds << 'byiyie'}
-    end
-    
-    c.new.before_save
-    $adds.should == ['hyiyie', 'byiyie']
-  end
-  
-  specify "should be inheritable" do
-    # pending
-    
-    $adds = []
-    a = Class.new(Sequel::Model) do
-      before_save {$adds << '123'}
-    end
-    
-    b = Class.new(a) do
-      before_save {$adds << '456'}
-      before_save {$adds << '789'}
-    end
-    
-    b.new.before_save
-    $adds.should == ['123', '456', '789']
-  end
-  
-  specify "should be overridable in descendant classes" do
-    $adds = []
-    a = Class.new(Sequel::Model) do
-      before_save {$adds << '123'}
-    end
-    
-    b = Class.new(a) do
-      def before_save; $adds << '456'; end
-    end
-    
-    a.new.before_save
-    $adds.should == ['123']
-    $adds = []
-    b.new.before_save
-    $adds.should == ['456']
-  end
-  
-  specify "should stop processing if a hook returns false" do
-    $flag = true
-    $adds = []
-    
-    a = Class.new(Sequel::Model) do
-      before_save {$adds << 'blah'; $flag}
-      before_save {$adds << 'cruel'}
-    end
-    
-    a.new.before_save
-    $adds.should == ['blah', 'cruel']
-
-    # chain should not break on nil
-    $adds = []
-    $flag = nil
-    a.new.before_save
-    $adds.should == ['blah', 'cruel']
-    
-    $adds = []
-    $flag = false
-    a.new.before_save
-    $adds.should == ['blah']
-    
-    b = Class.new(a) do
-      before_save {$adds << 'mau'}
-    end
-    
-    $adds = []
-    b.new.before_save
-    $adds.should == ['blah']
-  end
-end
-
-describe "Model#after_initialize" do
-  specify "should be called after initialization" do
-    $values1 = nil
-    $reached_after_initialized = false
-    
-    a = Class.new(Sequel::Model) do
-      columns :x, :y
-      after_initialize do
-        $values1 = @values.clone
-        $reached_after_initialized = true
-      end
-    end
-    
-    a.new(:x => 1, :y => 2)
-    $values1.should == {:x => 1, :y => 2}
-    $reached_after_initialized.should == true
-  end
-end
-
-describe "Model#before_create && Model#after_create" do
-  setup do
-    MODEL_DB.reset
-
-    @c = Class.new(Sequel::Model(:items)) do
-      columns :x
-      no_primary_key
-      
-      before_create {MODEL_DB << "BLAH before"}
-      after_create {MODEL_DB << "BLAH after"}
-    end
-  end
-  
-  specify "should be called around new record creation" do
-    @c.create(:x => 2)
-    MODEL_DB.sqls.should == [
-      'BLAH before',
-      'INSERT INTO items (x) VALUES (2)',
-      'BLAH after'
-    ]
-  end
-end
-
-describe "Model#before_update && Model#after_update" do
-  setup do
-    MODEL_DB.reset
-
-    @c = Class.new(Sequel::Model(:items)) do
-      before_update {MODEL_DB << "BLAH before"}
-      after_update {MODEL_DB << "BLAH after"}
-    end
-  end
-  
-  specify "should be called around record update" do
-    m = @c.load(:id => 2233)
-    m.save
-    MODEL_DB.sqls.should == [
-      'BLAH before',
-      'UPDATE items SET id = 2233 WHERE (id = 2233)',
-      'BLAH after'
-    ]
-  end
-end
-
-describe "Model#before_save && Model#after_save" do
-  setup do
-    MODEL_DB.reset
-
-    @c = Class.new(Sequel::Model(:items)) do
-      columns :x
-      before_save {MODEL_DB << "BLAH before"}
-      after_save {MODEL_DB << "BLAH after"}
-    end
-  end
-  
-  specify "should be called around record update" do
-    m = @c.load(:id => 2233)
-    m.save
-    MODEL_DB.sqls.should == [
-      'BLAH before',
-      'UPDATE items SET id = 2233 WHERE (id = 2233)',
-      'BLAH after'
-    ]
-  end
-  
-  specify "should be called around record creation" do
-    @c.no_primary_key
-    @c.create(:x => 2)
-    MODEL_DB.sqls.should == [
-      'BLAH before',
-      'INSERT INTO items (x) VALUES (2)',
-      'BLAH after'
-    ]
-  end
-end
-
-describe "Model#before_destroy && Model#after_destroy" do
-  setup do
-    MODEL_DB.reset
-
-    @c = Class.new(Sequel::Model(:items)) do
-      before_destroy {MODEL_DB << "BLAH before"}
-      after_destroy {MODEL_DB << "BLAH after"}
-      
-      def delete
-        MODEL_DB << "DELETE BLAH"
-      end
-    end
-  end
-  
-  specify "should be called around record update" do
-    m = @c.new(:id => 2233)
-    m.destroy
-    MODEL_DB.sqls.should == [
-      'BLAH before',
-      'DELETE BLAH',
-      'BLAH after'
-    ]
-  end
-end
-
-describe "Model#has_hooks?" do
-  setup do
-    @c = Class.new(Sequel::Model)
-  end
-  
-  specify "should return false if no hooks are defined" do
-    @c.has_hooks?(:before_save).should be_false
-  end
-  
-  specify "should return true if hooks are defined" do
-    @c.before_save {'blah'}
-    @c.has_hooks?(:before_save).should be_true
-  end
-  
-  specify "should return true if hooks are inherited" do
-    @d = Class.new(@c)
-    @d.has_hooks?(:before_save).should be_false
-    
-    @c.before_save :blah
-    @d.has_hooks?(:before_save).should be_true
-  end
-end
diff -pruN 1.5.1-1/spec/inflector_spec.rb 3.0.0-1/spec/inflector_spec.rb
--- 1.5.1-1/spec/inflector_spec.rb	2008-04-22 20:16:00.000000000 +0100
+++ 3.0.0-1/spec/inflector_spec.rb	1970-01-01 01:00:00.000000000 +0100
@@ -1,34 +0,0 @@
-require File.join(File.dirname(__FILE__), 'spec_helper')
-
-describe Inflector do
-  it "should transform words from singular to plural" do
-    "post".pluralize.should == "posts"
-    "octopus".pluralize.should =="octopi"
-    "the blue mailman".pluralize.should == "the blue mailmen"
-    "CamelOctopus".pluralize.should == "CamelOctopi"
-  end
-  
-  it "should transform words from plural to singular" do
-    "posts".singularize.should == "post"
-    "octopi".singularize.should == "octopus"
-    "the blue mailmen".singularize.should == "the blue mailman"
-    "CamelOctopi".singularize.should == "CamelOctopus"
-  end
-  
-  it "should transform class names to table names" do
-    "RawScaledScorer".tableize.should == "raw_scaled_scorers"
-    "egg_and_ham".tableize.should == "egg_and_hams"
-    "fancyCategory".tableize.should == "fancy_categories"
-  end
-  
-  it "should tranform table names to class names" do
-    "egg_and_hams".classify.should == "EggAndHam"
-    "post".classify.should == "Post"
-  end
-  
-  it "should create a foreign key name from a class name" do
-    "Message".foreign_key.should == "message_id"
-    "Message".foreign_key(false).should == "messageid"
-    "Admin::Post".foreign_key.should == "post_id"
-  end
-end
\ No newline at end of file
diff -pruN 1.5.1-1/spec/integration/database_test.rb 3.0.0-1/spec/integration/database_test.rb
--- 1.5.1-1/spec/integration/database_test.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/integration/database_test.rb	2009-04-13 02:03:44.000000000 +0100
@@ -0,0 +1,18 @@
+require File.join(File.dirname(__FILE__), 'spec_helper.rb')
+
+describe Sequel::Database do
+  specify "should provide disconnect functionality" do
+    INTEGRATION_DB.test_connection
+    INTEGRATION_DB.pool.size.should == 1
+    INTEGRATION_DB.disconnect
+    INTEGRATION_DB.pool.size.should == 0
+  end
+
+  specify "should raise Sequel::DatabaseError on invalid SQL" do
+    proc{INTEGRATION_DB << "SELECT"}.should raise_error(Sequel::DatabaseError)
+  end
+
+  specify "should not have the connection pool swallow non-StandardError based exceptions" do
+    proc{INTEGRATION_DB.pool.hold{raise Interrupt, "test"}}.should raise_error(Interrupt)
+  end
+end
diff -pruN 1.5.1-1/spec/integration/dataset_test.rb 3.0.0-1/spec/integration/dataset_test.rb
--- 1.5.1-1/spec/integration/dataset_test.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/integration/dataset_test.rb	2009-04-13 02:44:17.000000000 +0100
@@ -0,0 +1,276 @@
+require File.join(File.dirname(__FILE__), 'spec_helper.rb')
+
+describe "Simple Dataset operations" do
+  before do
+    INTEGRATION_DB.create_table!(:items) do
+      primary_key :id
+      Integer :number
+    end
+    @ds = INTEGRATION_DB[:items]
+    @ds.insert(:number=>10)
+    clear_sqls
+  end
+  after do
+    INTEGRATION_DB.drop_table(:items)
+  end
+
+  specify "should support sequential primary keys" do
+    @ds << {:number=>20}
+    @ds << {:number=>30}
+    @ds.order(:number).all.should == [
+      {:id => 1, :number=>10},
+      {:id => 2, :number=>20},
+      {:id => 3, :number=>30} ]   
+  end 
+
+  specify "should insert with a primary key specified" do
+    @ds.insert(:id=>100, :number=>20)
+    sqls_should_be(/INSERT INTO items \((number, id|id, number)\) VALUES \((100, 20|20, 100)\)/)
+    @ds.count.should == 2
+    @ds.order(:id).all.should == [{:id=>1, :number=>10}, {:id=>100, :number=>20}]
+  end
+
+  specify "should have insert return primary key value" do
+    @ds.insert(:number=>20).should == 2
+    sqls_should_be('INSERT INTO items (number) VALUES (20)')
+    @ds.filter(:id=>2).first[:number].should == 20
+  end
+
+  specify "should delete correctly" do
+    @ds.filter(1=>1).delete.should == 1
+    sqls_should_be('DELETE FROM items WHERE (1 = 1)')
+    @ds.count.should == 0
+  end
+
+  specify "should update correctly" do
+    @ds.update(:number=>:number+1).should == 1
+    sqls_should_be('UPDATE items SET number = (number + 1)')
+    @ds.all.should == [{:id=>1, :number=>11}]
+  end
+
+  specify "should fetch all results correctly" do
+    @ds.all.should == [{:id=>1, :number=>10}]
+    sqls_should_be('SELECT * FROM items')
+  end
+
+  specify "should fetch a single row correctly" do
+    @ds.first.should == {:id=>1, :number=>10}
+    sqls_should_be('SELECT * FROM items LIMIT 1')
+  end
+
+  specify "should alias columns correctly" do
+    @ds.select(:id___x, :number___n).first.should == {:x=>1, :n=>10}
+    sqls_should_be("SELECT id AS 'x', number AS 'n' FROM items LIMIT 1")
+  end
+end
+
+describe Sequel::Dataset do
+  before do
+    INTEGRATION_DB.create_table!(:test) do
+      String :name
+      Integer :value
+    end
+    @d = INTEGRATION_DB[:test]
+    clear_sqls
+  end
+  after do
+    INTEGRATION_DB.drop_table(:test)
+  end
+
+  specify "should return the correct record count" do
+    @d.count.should == 0
+    @d << {:name => 'abc', :value => 123}
+    @d << {:name => 'abc', :value => 456}
+    @d << {:name => 'def', :value => 789}
+    @d.count.should == 3
+  end
+
+  specify "should return the correct records" do
+    @d.to_a.should == []
+    @d << {:name => 'abc', :value => 123}
+    @d << {:name => 'abc', :value => 456}
+    @d << {:name => 'def', :value => 789}
+
+    @d.order(:value).to_a.should == [
+      {:name => 'abc', :value => 123},
+      {:name => 'abc', :value => 456},
+      {:name => 'def', :value => 789}
+    ]
+  end
+
+  specify "should update records correctly" do
+    @d << {:name => 'abc', :value => 123}
+    @d << {:name => 'abc', :value => 456}
+    @d << {:name => 'def', :value => 789}
+    @d.filter(:name => 'abc').update(:value => 530)
+    @d[:name => 'def'][:value].should == 789
+    @d.filter(:value => 530).count.should == 2
+  end
+
+  specify "should delete records correctly" do
+    @d << {:name => 'abc', :value => 123}
+    @d << {:name => 'abc', :value => 456}
+    @d << {:name => 'def', :value => 789}
+    @d.filter(:name => 'abc').delete
+    @d.count.should == 1
+    @d.first[:name].should == 'def'
+  end
+
+  specify "should be able to literalize booleans" do
+    proc {@d.literal(true)}.should_not raise_error
+    proc {@d.literal(false)}.should_not raise_error
+  end
+
+  specify "should correctly escape strings" do
+    INTEGRATION_DB.get("\\dingo".as(:a)) == "\\dingo"
+  end
+
+  specify "should correctly escape strings with quotes" do
+    INTEGRATION_DB.get("\\'dingo".as(:a)) == "\\'dingo"
+  end
+
+  specify "should properly escape binary data" do
+    INTEGRATION_DB.get("\1\2\3".to_sequel_blob.as(:a)) == "\1\2\3"
+  end
+end
+
+context "An SQLite dataset" do
+  before do
+    INTEGRATION_DB.create_table! :items do
+      primary_key :id 
+      Integer :value
+    end 
+    @d = INTEGRATION_DB[:items]
+    @d << {:value => 123}
+    @d << {:value => 456}
+    @d << {:value => 789}
+  end 
+  after do
+    INTEGRATION_DB.drop_table(:items)
+  end 
+  
+  specify "should correctly return avg" do
+    @d.avg(:value).to_i.should == 456
+  end 
+  
+  specify "should correctly return sum" do
+    @d.sum(:value).to_i.should == 1368
+  end 
+  
+  specify "should correctly return max" do
+    @d.max(:value).to_i.should == 789 
+  end 
+  
+  specify "should correctly return min" do
+    @d.min(:value).to_i.should == 123 
+  end 
+end
+
+describe "Simple Dataset operations" do
+  before do
+    INTEGRATION_DB.create_table!(:items) do
+      Integer :number
+      TrueClass :flag
+    end
+    @ds = INTEGRATION_DB[:items]
+  end
+  after do
+    INTEGRATION_DB.drop_table(:items)
+  end
+
+  specify "should deal with boolean conditions correctly" do
+    @ds.insert(:number=>1, :flag=>true)
+    @ds.insert(:number=>2, :flag=>false)
+    @ds.insert(:number=>3, :flag=>nil)
+    @ds.order!(:number)
+    @ds.filter(:flag=>true).map(:number).should == [1]
+    @ds.filter(:flag=>false).map(:number).should == [2]
+    @ds.filter(:flag=>nil).map(:number).should == [3]
+    @ds.exclude(:flag=>true).map(:number).should == [2, 3]
+    @ds.exclude(:flag=>false).map(:number).should == [1, 3]
+    @ds.exclude(:flag=>nil).map(:number).should == [1, 2]
+  end
+end
+
+describe "Simple Dataset operations in transactions" do
+  before do
+    INTEGRATION_DB.create_table!(:items_insert_in_transaction) do
+      primary_key :id
+      integer :number
+    end
+    @ds = INTEGRATION_DB[:items_insert_in_transaction]
+    clear_sqls
+  end
+  after do
+    INTEGRATION_DB.drop_table(:items_insert_in_transaction)
+  end
+
+  specify "should insert correctly with a primary key specified inside a transaction" do
+    INTEGRATION_DB.transaction do
+      @ds.insert(:id=>100, :number=>20)
+      sqls_should_be('Transaction.begin', /INSERT INTO items_insert_in_transaction \((number, id|id, number)\) VALUES \((100, 20|20, 100)\)/)
+      @ds.count.should == 1
+      @ds.order(:id).all.should == [{:id=>100, :number=>20}]
+    end
+  end
+  
+  specify "should have insert return primary key value inside a transaction" do
+    INTEGRATION_DB.transaction do
+      @ds.insert(:number=>20).should == 1
+      sqls_should_be('Transaction.begin', /INSERT INTO items_insert_in_transaction \(number\) VALUES \(20\)/)
+      @ds.count.should == 1
+      @ds.order(:id).all.should == [{:id=>1, :number=>20}]
+    end
+  end
+end
+
+describe "Dataset UNION, EXCEPT, and INTERSECT" do
+  before do
+    INTEGRATION_DB.create_table!(:i1){integer :number}
+    INTEGRATION_DB.create_table!(:i2){integer :number}
+    INTEGRATION_DB.create_table!(:i3){integer :number}
+    @ds1 = INTEGRATION_DB[:i1]
+    @ds1.insert(:number=>10)
+    @ds1.insert(:number=>20)
+    @ds2 = INTEGRATION_DB[:i2]
+    @ds2.insert(:number=>10)
+    @ds2.insert(:number=>30)
+    @ds3 = INTEGRATION_DB[:i3]
+    @ds3.insert(:number=>10)
+    @ds3.insert(:number=>40)
+    clear_sqls
+  end
+  
+  specify "should give the correct results for simple UNION, EXCEPT, and INTERSECT" do
+    @ds1.union(@ds2).order(:number).map{|x| x[:number].to_s}.should == %w'10 20 30'
+    unless defined?(Sequel::Dataset::UnsupportedIntersectExcept) and @ds1.class.ancestors.include?(Sequel::Dataset::UnsupportedIntersectExcept)
+      @ds1.except(@ds2).order(:number).map{|x| x[:number].to_s}.should == %w'20'
+      @ds1.intersect(@ds2).order(:number).map{|x| x[:number].to_s}.should == %w'10'
+    end
+  end
+  
+  specify "should give the correct results for compound UNION, EXCEPT, and INTERSECT" do
+    @ds1.union(@ds2).union(@ds3).order(:number).map{|x| x[:number].to_s}.should == %w'10 20 30 40'
+    @ds1.union(@ds2.union(@ds3)).order(:number).map{|x| x[:number].to_s}.should == %w'10 20 30 40'
+    unless defined?(Sequel::Dataset::UnsupportedIntersectExcept) and @ds1.class.ancestors.include?(Sequel::Dataset::UnsupportedIntersectExcept)
+      @ds1.union(@ds2).except(@ds3).order(:number).map{|x| x[:number].to_s}.should == %w'20 30'
+      @ds1.union(@ds2.except(@ds3)).order(:number).map{|x| x[:number].to_s}.should == %w'10 20 30'
+      @ds1.union(@ds2).intersect(@ds3).order(:number).map{|x| x[:number].to_s}.should == %w'10 '
+      @ds1.union(@ds2.intersect(@ds3)).order(:number).map{|x| x[:number].to_s}.should == %w'10 20'
+      
+      @ds1.except(@ds2).union(@ds3).order(:number).map{|x| x[:number].to_s}.should == %w'10 20 40'
+      @ds1.except(@ds2.union(@ds3)).order(:number).map{|x| x[:number].to_s}.should == %w'20'
+      @ds1.except(@ds2).except(@ds3).order(:number).map{|x| x[:number].to_s}.should == %w'20'
+      @ds1.except(@ds2.except(@ds3)).order(:number).map{|x| x[:number].to_s}.should == %w'10 20'
+      @ds1.except(@ds2).intersect(@ds3).order(:number).map{|x| x[:number].to_s}.should == %w''
+      @ds1.except(@ds2.intersect(@ds3)).order(:number).map{|x| x[:number].to_s}.should == %w'20'
+      
+      @ds1.intersect(@ds2).union(@ds3).order(:number).map{|x| x[:number].to_s}.should == %w'10 40'
+      @ds1.intersect(@ds2.union(@ds3)).order(:number).map{|x| x[:number].to_s}.should == %w'10'
+      @ds1.intersect(@ds2).except(@ds3).order(:number).map{|x| x[:number].to_s}.should == %w''
+      @ds1.intersect(@ds2.except(@ds3)).order(:number).map{|x| x[:number].to_s}.should == %w''
+      @ds1.intersect(@ds2).intersect(@ds3).order(:number).map{|x| x[:number].to_s}.should == %w'10'
+      @ds1.intersect(@ds2.intersect(@ds3)).order(:number).map{|x| x[:number].to_s}.should == %w'10'
+    end
+  end
+end
diff -pruN 1.5.1-1/spec/integration/eager_loader_test.rb 3.0.0-1/spec/integration/eager_loader_test.rb
--- 1.5.1-1/spec/integration/eager_loader_test.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/integration/eager_loader_test.rb	2009-04-22 17:14:15.000000000 +0100
@@ -0,0 +1,744 @@
+require File.join(File.dirname(__FILE__), 'spec_helper.rb')
+
+describe "Eagerly loading a tree structure" do
+  before do
+    INTEGRATION_DB.instance_variable_set(:@schemas, nil)
+    INTEGRATION_DB.create_table!(:nodes) do
+      primary_key :id
+      foreign_key :parent_id, :nodes
+    end
+    class ::Node < Sequel::Model
+      many_to_one :parent
+      one_to_many :children, :key=>:parent_id
+    
+      # Only useful when eager loading
+      many_to_one :ancestors, :eager_loader=>(proc do |key_hash, nodes, associations|
+        # Handle cases where the root node has the same parent_id as primary_key
+        # and also when it is NULL
+        non_root_nodes = nodes.reject do |n| 
+          if [nil, n.pk].include?(n.parent_id)
+            # Make sure root nodes have their parent association set to nil
+            n.associations[:parent] = nil
+            true
+          else
+            false
+          end
+        end
+        unless non_root_nodes.empty?
+          id_map = {}
+          # Create an map of parent_ids to nodes that have that parent id
+          non_root_nodes.each{|n| (id_map[n.parent_id] ||= []) << n}
+          # Doesn't cause an infinte loop, because when only the root node
+          # is left, this is not called.
+          Node.filter(Node.primary_key=>id_map.keys.sort).eager(:ancestors).all do |node|
+            # Populate the parent association for each node
+            id_map[node.pk].each{|n| n.associations[:parent] = node}
+          end
+        end
+      end)
+      many_to_one :descendants, :eager_loader=>(proc do |key_hash, nodes, associations|
+        id_map = {}
+        nodes.each do |n|
+          # Initialize an empty array of child associations for each parent node
+          n.associations[:children] = []
+          # Populate identity map of nodes
+          id_map[n.pk] = n
+        end
+        # Doesn't cause an infinite loop, because the :eager_loader is not called
+        # if no records are returned.  Exclude id = parent_id to avoid infinite loop
+        # if the root note is one of the returned records and it has parent_id = id
+        # instead of parent_id = NULL.
+        Node.filter(:parent_id=>id_map.keys.sort).exclude(:id=>:parent_id).eager(:descendants).all do |node|
+          # Get the parent from the identity map
+          parent = id_map[node.parent_id]
+          # Set the child's parent association to the parent 
+          node.associations[:parent] = parent
+          # Add the child association to the array of children in the parent
+          parent.associations[:children] << node
+        end
+      end)
+    end
+    
+    Node.insert(:parent_id=>1)
+    Node.insert(:parent_id=>1)
+    Node.insert(:parent_id=>1)
+    Node.insert(:parent_id=>2)
+    Node.insert(:parent_id=>4)
+    Node.insert(:parent_id=>5)
+    Node.insert(:parent_id=>6)
+    clear_sqls
+  end
+  after do
+    INTEGRATION_DB.drop_table :nodes
+    Object.send(:remove_const, :Node)
+  end
+
+  it "#descendants should get all descendants in one call" do
+    nodes = Node.filter(:id=>1).eager(:descendants).all
+    sqls_should_be('SELECT * FROM nodes WHERE (id = 1)',
+      'SELECT * FROM nodes WHERE ((parent_id IN (1)) AND (id != parent_id))',
+      'SELECT * FROM nodes WHERE ((parent_id IN (2, 3)) AND (id != parent_id))',
+      'SELECT * FROM nodes WHERE ((parent_id IN (4)) AND (id != parent_id))',
+      'SELECT * FROM nodes WHERE ((parent_id IN (5)) AND (id != parent_id))',
+      'SELECT * FROM nodes WHERE ((parent_id IN (6)) AND (id != parent_id))',
+      'SELECT * FROM nodes WHERE ((parent_id IN (7)) AND (id != parent_id))')
+    nodes.length.should == 1
+    node = nodes.first
+    node.pk.should == 1
+    node.children.length.should == 2
+    node.children.collect{|x| x.pk}.sort.should == [2, 3]
+    node.children.collect{|x| x.parent}.should == [node, node]
+    node = nodes.first.children.find{|x| x.pk == 2}
+    node.children.length.should == 1
+    node.children.first.pk.should == 4
+    node.children.first.parent.should == node
+    node = node.children.first
+    node.children.length.should == 1
+    node.children.first.pk.should == 5
+    node.children.first.parent.should == node
+    node = node.children.first
+    node.children.length.should == 1
+    node.children.first.pk.should == 6
+    node.children.first.parent.should == node
+    node = node.children.first
+    node.children.length.should == 1
+    node.children.first.pk.should == 7
+    node.children.first.parent.should == node
+    sqls_should_be
+  end
+
+  it "#ancestors should get all ancestors in one call" do
+    nodes = Node.filter(:id=>[7,3]).order(:id).eager(:ancestors).all
+    sqls_should_be('SELECT * FROM nodes WHERE (id IN (7, 3)) ORDER BY id',
+      'SELECT * FROM nodes WHERE (id IN (1, 6))',
+      'SELECT * FROM nodes WHERE (id IN (5))',
+      'SELECT * FROM nodes WHERE (id IN (4))',
+      'SELECT * FROM nodes WHERE (id IN (2))',
+      'SELECT * FROM nodes WHERE (id IN (1))')
+    nodes.length.should == 2
+    nodes.collect{|x| x.pk}.should == [3, 7]
+    nodes.first.parent.pk.should == 1
+    nodes.first.parent.parent.should == nil
+    node = nodes.last
+    node.parent.pk.should == 6
+    node = node.parent
+    node.parent.pk.should == 5
+    node = node.parent
+    node.parent.pk.should == 4
+    node = node.parent
+    node.parent.pk.should == 2
+    node = node.parent
+    node.parent.pk.should == 1
+    node.parent.parent.should == nil
+    sqls_should_be
+  end
+end
+
+describe "Association Extensions" do
+  before do
+    module ::FindOrCreate
+      def find_or_create(vals)
+        first(vals) || model.create(vals.merge(:author_id=>model_object.pk))
+      end 
+      def find_or_create_by_name(name)
+        first(:name=>name) || model.create(:name=>name, :author_id=>model_object.pk)
+      end
+    end
+    INTEGRATION_DB.instance_variable_set(:@schemas, nil)
+    INTEGRATION_DB.create_table!(:authors) do
+      primary_key :id
+    end
+    class ::Author < Sequel::Model
+      one_to_many :authorships, :extend=>FindOrCreate
+    end
+    INTEGRATION_DB.create_table!(:authorships) do
+      primary_key :id
+      foreign_key :author_id, :authors
+      String :name
+    end
+    class ::Authorship < Sequel::Model
+      many_to_one :author
+    end
+    @author = Author.create
+    clear_sqls
+  end
+  after do
+    INTEGRATION_DB.drop_table :authorships, :authors
+    Object.send(:remove_const, :Author)
+    Object.send(:remove_const, :Authorship)
+  end
+
+  it "should allow methods to be called on the dataset method" do
+    Authorship.count.should == 0
+    sqls_should_be('SELECT COUNT(*) AS \'count\' FROM authorships LIMIT 1')
+    authorship = @author.authorships_dataset.find_or_create_by_name('Bob')
+    sqls_should_be("SELECT * FROM authorships WHERE ((authorships.author_id = 1) AND (name = 'Bob')) LIMIT 1",
+      /INSERT INTO authorships \((author_id, name|name, author_id)\) VALUES \((1, 'Bob'|'Bob', 1)\)/,
+      "SELECT * FROM authorships WHERE (id = 1) LIMIT 1")
+    Authorship.count.should == 1
+    Authorship.first.should == authorship
+    sqls_should_be('SELECT COUNT(*) AS \'count\' FROM authorships LIMIT 1', "SELECT * FROM authorships LIMIT 1")
+    authorship.name.should == 'Bob'
+    authorship.author_id.should == @author.id
+    @author.authorships_dataset.find_or_create_by_name('Bob').should == authorship
+    sqls_should_be("SELECT * FROM authorships WHERE ((authorships.author_id = 1) AND (name = 'Bob')) LIMIT 1")
+    Authorship.count.should == 1
+    sqls_should_be('SELECT COUNT(*) AS \'count\' FROM authorships LIMIT 1')
+    authorship2 = @author.authorships_dataset.find_or_create(:name=>'Jim')
+    sqls_should_be("SELECT * FROM authorships WHERE ((authorships.author_id = 1) AND (name = 'Jim')) LIMIT 1",
+      /INSERT INTO authorships \((author_id, name|name, author_id)\) VALUES \((1, 'Jim'|'Jim', 1)\)/,
+      "SELECT * FROM authorships WHERE (id = 2) LIMIT 1")
+    Authorship.count.should == 2
+    sqls_should_be('SELECT COUNT(*) AS \'count\' FROM authorships LIMIT 1')
+    Authorship.order(:name).map(:name).should == ['Bob', 'Jim']
+    sqls_should_be('SELECT * FROM authorships ORDER BY name')
+    authorship2.name.should == 'Jim'
+    authorship2.author_id.should == @author.id
+    @author.authorships_dataset.find_or_create(:name=>'Jim').should == authorship2
+    sqls_should_be("SELECT * FROM authorships WHERE ((authorships.author_id = 1) AND (name = 'Jim')) LIMIT 1")
+  end
+end
+
+describe "has_many :through has_many and has_one :through belongs_to" do
+  before do
+    INTEGRATION_DB.instance_variable_set(:@schemas, nil)
+    INTEGRATION_DB.create_table!(:firms) do
+      primary_key :id
+    end
+    class ::Firm < Sequel::Model
+      one_to_many :clients
+      one_to_many :invoices, :read_only=>true, \
+        :dataset=>proc{Invoice.eager_graph(:client).filter(:client__firm_id=>pk)}, \
+        :after_load=>(proc do |firm, invs|
+          invs.each do |inv|
+            inv.client.associations[:firm] = inv.associations[:firm] = firm
+          end
+        end), \
+        :eager_loader=>(proc do |key_hash, firms, associations|
+          id_map = key_hash[Firm.primary_key]
+          firms.each{|firm| firm.associations[:invoices] = []}
+          Invoice.eager_graph(:client).filter(:client__firm_id=>id_map.keys).all do |inv|
+            id_map[inv.client.firm_id].each do |firm|
+              firm.associations[:invoices] << inv
+            end
+          end
+        end)
+    end
+
+    INTEGRATION_DB.create_table!(:clients) do
+      primary_key :id
+      foreign_key :firm_id, :firms
+    end
+    class ::Client < Sequel::Model
+      many_to_one :firm
+      one_to_many :invoices
+    end
+
+    INTEGRATION_DB.create_table!(:invoices) do
+      primary_key :id
+      foreign_key :client_id, :clients
+    end
+    class ::Invoice < Sequel::Model
+      many_to_one :client
+      many_to_one :firm, :key=>nil, :read_only=>true, \
+        :dataset=>proc{Firm.eager_graph(:clients).filter(:clients__id=>client_id)}, \
+        :after_load=>(proc do |inv, firm|
+          # Delete the cached associations from firm, because it only has the
+          # client with this invoice, instead of all clients of the firm
+          if c = firm.associations.delete(:clients)
+            firm.associations[:invoice_client] = c.first
+          end
+          inv.associations[:client] ||= firm.associations[:invoice_client]
+        end), \
+        :eager_loader=>(proc do |key_hash, invoices, associations|
+          id_map = {}
+          invoices.each do |inv|
+            inv.associations[:firm] = nil
+            (id_map[inv.client_id] ||= []) << inv
+          end
+          Firm.eager_graph(:clients).filter(:clients__id=>id_map.keys).all do |firm|
+            # Delete the cached associations from firm, because it only has the
+            # clients related the invoices being eagerly loaded, instead of all
+            # clients of the firm.
+            firm.associations[:clients].each do |client|
+              id_map[client.pk].each do |inv|
+                inv.associations[:firm] = firm
+                inv.associations[:client] = client
+              end
+            end
+          end
+        end)
+    end
+    @firm1 = Firm.create
+    @firm2 = Firm.create
+    @client1 = Client.create(:firm => @firm1)
+    @client2 = Client.create(:firm => @firm1)
+    @client3 = Client.create(:firm => @firm2)
+    @invoice1 = Invoice.create(:client => @client1)
+    @invoice2 = Invoice.create(:client => @client1)
+    @invoice3 = Invoice.create(:client => @client2)
+    @invoice4 = Invoice.create(:client => @client3)
+    @invoice5 = Invoice.create(:client => @client3)
+    clear_sqls
+  end
+  after do
+    INTEGRATION_DB.drop_table :invoices, :clients, :firms
+    Object.send(:remove_const, :Firm)
+    Object.send(:remove_const, :Client)
+    Object.send(:remove_const, :Invoice)
+  end
+
+  it "should return has_many :through has_many records for a single object" do
+    invs = @firm1.invoices.sort_by{|x| x.pk}
+    sqls_should_be("SELECT invoices.id, invoices.client_id, client.id AS 'client_id_0', client.firm_id FROM invoices LEFT OUTER JOIN clients AS 'client' ON (client.id = invoices.client_id) WHERE (client.firm_id = 1)")
+    invs.should == [@invoice1, @invoice2, @invoice3]
+    invs[0].client.should == @client1
+    invs[1].client.should == @client1
+    invs[2].client.should == @client2
+    invs.collect{|i| i.firm}.should == [@firm1, @firm1, @firm1]
+    invs.collect{|i| i.client.firm}.should == [@firm1, @firm1, @firm1]
+    sqls_should_be
+  end
+
+  it "should eagerly load has_many :through has_many records for multiple objects" do
+    firms = Firm.order(:id).eager(:invoices).all
+    sqls_should_be("SELECT * FROM firms ORDER BY id", "SELECT invoices.id, invoices.client_id, client.id AS 'client_id_0', client.firm_id FROM invoices LEFT OUTER JOIN clients AS 'client' ON (client.id = invoices.client_id) WHERE (client.firm_id IN (1, 2))")
+    firms.should == [@firm1, @firm2]
+    firm1, firm2 = firms
+    invs1 = firm1.invoices.sort_by{|x| x.pk}
+    invs2 = firm2.invoices.sort_by{|x| x.pk}
+    invs1.should == [@invoice1, @invoice2, @invoice3]
+    invs2.should == [@invoice4, @invoice5]
+    invs1[0].client.should == @client1
+    invs1[1].client.should == @client1
+    invs1[2].client.should == @client2
+    invs2[0].client.should == @client3
+    invs2[1].client.should == @client3
+    invs1.collect{|i| i.firm}.should == [@firm1, @firm1, @firm1]
+    invs2.collect{|i| i.firm}.should == [@firm2, @firm2]
+    invs1.collect{|i| i.client.firm}.should == [@firm1, @firm1, @firm1]
+    invs2.collect{|i| i.client.firm}.should == [@firm2, @firm2]
+    sqls_should_be
+  end
+
+  it "should return has_one :through belongs_to records for a single object" do
+    firm = @invoice1.firm
+    sqls_should_be("SELECT firms.id, clients.id AS 'clients_id', clients.firm_id FROM firms LEFT OUTER JOIN clients ON (clients.firm_id = firms.id) WHERE (clients.id = 1)")
+    firm.should == @firm1
+    @invoice1.client.should == @client1
+    @invoice1.client.firm.should == @firm1
+    sqls_should_be
+    firm.associations[:clients].should == nil
+  end
+
+  it "should eagerly load has_one :through belongs_to records for multiple objects" do
+    invs = Invoice.order(:id).eager(:firm).all
+    sqls_should_be("SELECT * FROM invoices ORDER BY id", "SELECT firms.id, clients.id AS 'clients_id', clients.firm_id FROM firms LEFT OUTER JOIN clients ON (clients.firm_id = firms.id) WHERE (clients.id IN (1, 2, 3))")
+    invs.should == [@invoice1, @invoice2, @invoice3, @invoice4, @invoice5]
+    invs[0].firm.should == @firm1
+    invs[0].client.should == @client1
+    invs[0].client.firm.should == @firm1
+    invs[0].firm.associations[:clients].should == nil
+    invs[1].firm.should == @firm1
+    invs[1].client.should == @client1
+    invs[1].client.firm.should == @firm1
+    invs[1].firm.associations[:clients].should == nil
+    invs[2].firm.should == @firm1
+    invs[2].client.should == @client2
+    invs[2].client.firm.should == @firm1
+    invs[2].firm.associations[:clients].should == nil
+    invs[3].firm.should == @firm2
+    invs[3].client.should == @client3
+    invs[3].client.firm.should == @firm2
+    invs[3].firm.associations[:clients].should == nil
+    invs[4].firm.should == @firm2
+    invs[4].client.should == @client3
+    invs[4].client.firm.should == @firm2
+    invs[4].firm.associations[:clients].should == nil
+    sqls_should_be
+  end
+end
+
+describe "Polymorphic Associations" do
+  before do
+    INTEGRATION_DB.instance_variable_set(:@schemas, nil)
+    INTEGRATION_DB.create_table!(:assets) do
+      primary_key :id
+      Integer :attachable_id
+      String :attachable_type
+    end
+    class ::Asset < Sequel::Model
+      m = method(:constantize)
+      many_to_one :attachable, :reciprocal=>:assets, \
+        :dataset=>(proc do
+          klass = m.call(attachable_type)
+          klass.filter(klass.primary_key=>attachable_id)
+        end), \
+        :eager_loader=>(proc do |key_hash, assets, associations|
+          id_map = {}
+          assets.each do |asset|
+            asset.associations[:attachable] = nil 
+            ((id_map[asset.attachable_type] ||= {})[asset.attachable_id] ||= []) << asset
+          end 
+          id_map.each do |klass_name, id_map|
+            klass = m.call(klass_name)
+            klass.filter(klass.primary_key=>id_map.keys).all do |attach|
+              id_map[attach.pk].each do |asset|
+                asset.associations[:attachable] = attach
+              end 
+            end 
+          end 
+        end)
+            
+      private
+
+      def _attachable=(attachable)
+        self[:attachable_id] = (attachable.pk if attachable)
+        self[:attachable_type] = (attachable.class.name if attachable)
+      end 
+    end 
+  
+    INTEGRATION_DB.create_table!(:posts) do
+      primary_key :id
+    end
+    class ::Post < Sequel::Model
+      one_to_many :assets, :key=>:attachable_id do |ds|
+        ds.filter(:attachable_type=>'Post')
+      end 
+      
+      private
+
+      def _add_asset(asset)
+        asset.attachable_id = pk
+        asset.attachable_type = 'Post'
+        asset.save
+      end
+      def _remove_asset(asset)
+        asset.attachable_id = nil
+        asset.attachable_type = nil
+        asset.save
+      end
+      def _remove_all_assets
+        Asset.filter(:attachable_id=>pk, :attachable_type=>'Post')\
+          .update(:attachable_id=>nil, :attachable_type=>nil)
+      end
+    end 
+  
+    INTEGRATION_DB.create_table!(:notes) do
+      primary_key :id
+    end
+    class ::Note < Sequel::Model
+      one_to_many :assets, :key=>:attachable_id do |ds|
+        ds.filter(:attachable_type=>'Note')
+      end 
+      
+      private
+
+      def _add_asset(asset)
+        asset.attachable_id = pk
+        asset.attachable_type = 'Note'
+        asset.save
+      end
+      def _remove_asset(asset)
+        asset.attachable_id = nil
+        asset.attachable_type = nil
+        asset.save
+      end
+      def _remove_all_assets
+        Asset.filter(:attachable_id=>pk, :attachable_type=>'Note')\
+          .update(:attachable_id=>nil, :attachable_type=>nil)
+      end
+    end
+    @post = Post.create
+    Note.create
+    @note = Note.create
+    @asset1 = Asset.create(:attachable=>@post)
+    @asset2 = Asset.create(:attachable=>@note)
+    @asset1.associations.clear
+    @asset2.associations.clear
+    clear_sqls
+  end
+  after do
+    INTEGRATION_DB.drop_table :assets, :posts, :notes
+    Object.send(:remove_const, :Asset)
+    Object.send(:remove_const, :Post)
+    Object.send(:remove_const, :Note)
+  end
+
+  it "should load the correct associated object for a single object" do
+    @asset1.attachable.should == @post
+    @asset2.attachable.should == @note
+    sqls_should_be("SELECT * FROM posts WHERE (id = 1) LIMIT 1", "SELECT * FROM notes WHERE (id = 2) LIMIT 1")
+  end
+
+  it "should eagerly load the correct associated object for a group of objects" do
+    assets = Asset.order(:id).eager(:attachable).all
+    sqls_should_be("SELECT * FROM assets ORDER BY id", "SELECT * FROM posts WHERE (id IN (1))", "SELECT * FROM notes WHERE (id IN (2))")
+    assets.should == [@asset1, @asset2]
+    assets[0].attachable.should == @post
+    assets[1].attachable.should == @note
+    sqls_should_be
+  end
+
+  it "should set items correctly" do
+    @asset1.attachable = @note
+    @asset2.attachable = @post
+    sqls_should_be("SELECT * FROM posts WHERE (id = 1) LIMIT 1", "SELECT * FROM notes WHERE (id = 2) LIMIT 1")
+    @asset1.attachable.should == @note
+    @asset1.attachable_id.should == @note.pk
+    @asset1.attachable_type.should == 'Note'
+    @asset2.attachable.should == @post
+    @asset2.attachable_id.should == @post.pk
+    @asset2.attachable_type.should == 'Post'
+    @asset1.attachable = nil
+    @asset1.attachable.should == nil
+    @asset1.attachable_id.should == nil
+    @asset1.attachable_type.should == nil
+    sqls_should_be
+  end
+
+  it "should add items correctly" do
+    @post.assets.should == [@asset1]
+    sqls_should_be("SELECT * FROM assets WHERE ((assets.attachable_id = 1) AND (attachable_type = 'Post'))")
+    @post.add_asset(@asset2)
+    sqls_should_be(/UPDATE assets SET ((attachable_id = 1|attachable_type = 'Post'|id = 2)(, )?){3} WHERE \(id = 2\)/)
+    @post.assets.should == [@asset1, @asset2]
+    @asset2.attachable.should == @post
+    @asset2.attachable_id.should == @post.pk
+    @asset2.attachable_type.should == 'Post'
+    sqls_should_be
+  end
+
+  it "should remove items correctly" do
+    @note.assets.should == [@asset2]
+    sqls_should_be("SELECT * FROM assets WHERE ((assets.attachable_id = 2) AND (attachable_type = 'Note'))")
+    @note.remove_asset(@asset2)
+    sqls_should_be(/UPDATE assets SET ((attachable_id = NULL|attachable_type = NULL|id = 2)(, )?){3} WHERE \(id = 2\)/)
+    @note.assets.should == []
+    @asset2.attachable.should == nil
+    @asset2.attachable_id.should == nil
+    @asset2.attachable_type.should == nil
+    sqls_should_be
+  end
+
+  it "should remove all items correctly" do
+    @post.remove_all_assets
+    @note.remove_all_assets
+    sqls_should_be(/UPDATE assets SET attachable_(id|type) = NULL, attachable_(type|id) = NULL WHERE \(\(attachable_(id|type) = (1|'Post')\) AND \(attachable_(type|id) = ('Post'|1)\)\)/,
+      /UPDATE assets SET attachable_(id|type) = NULL, attachable_(type|id) = NULL WHERE \(\(attachable_(id|type) = (2|'Note')\) AND \(attachable_(type|id) = ('Note'|2)\)\)/)
+    @asset1.reload.attachable.should == nil
+    @asset2.reload.attachable.should == nil
+  end
+end
+
+describe "many_to_one/one_to_many not referencing primary key" do
+  before do
+    INTEGRATION_DB.instance_variable_set(:@schemas, nil)
+    INTEGRATION_DB.create_table!(:clients) do
+      primary_key :id
+      String :name
+    end
+    class ::Client < Sequel::Model
+      one_to_many :invoices, :reciprocal=>:client, \
+        :dataset=>proc{Invoice.filter(:client_name=>name)}, \
+        :eager_loader=>(proc do |key_hash, clients, associations|
+          id_map = {}
+          clients.each do |client|
+            id_map[client.name] = client
+            client.associations[:invoices] = []
+          end 
+          Invoice.filter(:client_name=>id_map.keys.sort).all do |inv|
+            inv.associations[:client] = client = id_map[inv.client_name]
+            client.associations[:invoices] << inv 
+          end 
+        end)
+
+      private
+
+      def _add_invoice(invoice)
+        invoice.client_name = name
+        invoice.save
+      end
+      def _remove_invoice(invoice)
+        invoice.client_name = nil
+        invoice.save
+      end
+      def _remove_all_invoices
+        Invoice.filter(:client_name=>name).update(:client_name=>nil)
+      end
+    end 
+  
+    INTEGRATION_DB.create_table!(:invoices) do
+      primary_key :id
+      String :client_name
+    end
+    class ::Invoice < Sequel::Model
+      many_to_one :client, :key=>:client_name, \
+        :dataset=>proc{Client.filter(:name=>client_name)}, \
+        :eager_loader=>(proc do |key_hash, invoices, associations|
+          id_map = key_hash[:client_name]
+          invoices.each{|inv| inv.associations[:client] = nil}
+          Client.filter(:name=>id_map.keys).all do |client|
+            id_map[client.name].each{|inv| inv.associations[:client] = client}
+          end 
+        end)
+      
+      private
+
+      def _client=(client)
+        self.client_name = (client.name if client)
+      end
+    end
+
+    @client1 = Client.create(:name=>'X')
+    @client2 = Client.create(:name=>'Y')
+    @invoice1 = Invoice.create(:client_name=>'X')
+    @invoice2 = Invoice.create(:client_name=>'X')
+    clear_sqls
+  end
+  after do
+    INTEGRATION_DB.drop_table :invoices, :clients
+    Object.send(:remove_const, :Client)
+    Object.send(:remove_const, :Invoice)
+  end
+
+  it "should load all associated one_to_many objects for a single object" do
+    invs = @client1.invoices
+    sqls_should_be("SELECT * FROM invoices WHERE (client_name = 'X')")
+    invs.should == [@invoice1, @invoice2]
+    invs[0].client.should == @client1
+    invs[1].client.should == @client1
+    sqls_should_be
+  end
+
+  it "should load the associated many_to_one object for a single object" do
+    client = @invoice1.client
+    sqls_should_be("SELECT * FROM clients WHERE (name = 'X') LIMIT 1")
+    client.should == @client1
+  end
+
+  it "should eagerly load all associated one_to_many objects for a group of objects" do
+    clients = Client.order(:id).eager(:invoices).all
+    sqls_should_be("SELECT * FROM clients ORDER BY id", "SELECT * FROM invoices WHERE (client_name IN ('X', 'Y'))")
+    clients.should == [@client1, @client2]
+    clients[1].invoices.should == []
+    invs = clients[0].invoices.sort_by{|x| x.pk}
+    invs.should == [@invoice1, @invoice2]
+    invs[0].client.should == @client1
+    invs[1].client.should == @client1
+    sqls_should_be
+  end
+
+  it "should eagerly load the associated many_to_one object for a group of objects" do
+    invoices = Invoice.order(:id).eager(:client).all
+    sqls_should_be("SELECT * FROM invoices ORDER BY id", "SELECT * FROM clients WHERE (name IN ('X'))")
+    invoices.should == [@invoice1, @invoice2]
+    invoices[0].client.should == @client1
+    invoices[1].client.should == @client1
+    sqls_should_be
+  end
+
+  it "should set the associated object correctly" do
+    @invoice1.client = @client2
+    sqls_should_be("SELECT * FROM clients WHERE (name = 'X') LIMIT 1")
+    @invoice1.client.should == @client2
+    @invoice1.client_name.should == 'Y'
+    @invoice1.client = nil
+    @invoice1.client_name.should == nil
+    sqls_should_be
+  end
+
+  it "should add the associated object correctly" do
+    @client2.invoices.should == []
+    sqls_should_be("SELECT * FROM invoices WHERE (client_name = 'Y')")
+    @client2.add_invoice(@invoice1)
+    sqls_should_be(/UPDATE invoices SET (client_name = 'Y'|id = 1), (client_name = 'Y'|id = 1) WHERE \(id = 1\)/)
+    @client2.invoices.should == [@invoice1]
+    @invoice1.client_name.should == 'Y'
+    @invoice1.client = nil
+    @invoice1.client_name.should == nil
+    sqls_should_be
+  end
+
+  it "should remove the associated object correctly" do
+    invs = @client1.invoices.sort_by{|x| x.pk}
+    sqls_should_be("SELECT * FROM invoices WHERE (client_name = 'X')")
+    invs.should == [@invoice1, @invoice2]
+    @client1.remove_invoice(@invoice1)
+    sqls_should_be(/UPDATE invoices SET (client_name = NULL|id = 1), (client_name = NULL|id = 1) WHERE \(id = 1\)/)
+    @client1.invoices.should == [@invoice2]
+    @invoice1.client_name.should == nil
+    @invoice1.client.should == nil
+    sqls_should_be
+  end
+
+  it "should remove all associated objects correctly" do
+    invs = @client1.remove_all_invoices
+    sqls_should_be("UPDATE invoices SET client_name = NULL WHERE (client_name = 'X')")
+    @invoice1.refresh.client.should == nil
+    @invoice1.client_name.should == nil
+    @invoice2.refresh.client.should == nil
+    @invoice2.client_name.should == nil
+  end
+end
+
+describe "statistics associations" do
+  before do
+    INTEGRATION_DB.create_table!(:projects) do
+      primary_key :id
+      String :name
+    end
+    class ::Project < Sequel::Model
+      many_to_one :ticket_hours, :read_only=>true, :key=>:id,
+       :dataset=>proc{Ticket.filter(:project_id=>id).select{sum(hours).as(hours)}},
+       :eager_loader=>(proc do |kh, projects, a|
+        projects.each{|p| p.associations[:ticket_hours] = nil}
+        Ticket.filter(:project_id=>kh[:id].keys).
+         group(:project_id).
+         select{[project_id.as(project_id), sum(hours).as(hours)]}.
+         all do |t|
+          p = kh[:id][t.values.delete(:project_id)].first
+          p.associations[:ticket_hours] = t
+         end
+       end)  
+      def ticket_hours
+        if s = super
+          s[:hours]
+        end
+      end 
+    end 
+
+    INTEGRATION_DB.create_table!(:tickets) do
+      primary_key :id
+      foreign_key :project_id, :projects
+      Integer :hours
+    end
+    class ::Ticket < Sequel::Model
+      many_to_one :project
+    end
+
+    @project1 = Project.create(:name=>'X')
+    @project2 = Project.create(:name=>'Y')
+    @ticket1 = Ticket.create(:project=>@project1, :hours=>1)
+    @ticket2 = Ticket.create(:project=>@project1, :hours=>10)
+    @ticket3 = Ticket.create(:project=>@project2, :hours=>2)
+    @ticket4 = Ticket.create(:project=>@project2, :hours=>20)
+    clear_sqls
+  end
+  after do
+    INTEGRATION_DB.drop_table :tickets, :projects
+    Object.send(:remove_const, :Project)
+    Object.send(:remove_const, :Ticket)
+  end
+
+  it "should give the correct sum of ticket hours for each project" do
+    @project1.ticket_hours.to_i.should == 11
+    @project2.ticket_hours.to_i.should == 22
+  end
+
+  it "should give the correct sum of ticket hours for each project when eager loading" do
+    p1, p2 = Project.order(:name).eager(:ticket_hours).all
+    p1.ticket_hours.to_i.should == 11
+    p2.ticket_hours.to_i.should == 22
+  end
+end
diff -pruN 1.5.1-1/spec/integration/prepared_statement_test.rb 3.0.0-1/spec/integration/prepared_statement_test.rb
--- 1.5.1-1/spec/integration/prepared_statement_test.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/integration/prepared_statement_test.rb	2009-04-28 00:36:01.000000000 +0100
@@ -0,0 +1,130 @@
+require File.join(File.dirname(__FILE__), 'spec_helper.rb')
+
+describe "Prepared Statements and Bound Arguments" do
+  before do
+    INTEGRATION_DB.create_table!(:items) do
+      primary_key :id
+      integer :number
+    end
+    @c = Class.new(Sequel::Model(:items))
+    @ds = INTEGRATION_DB[:items]
+    @ds.insert(:number=>10)
+    @ds.meta_def(:ba) do |sym|
+      prepared_arg_placeholder == '$' ? :"#{sym}__int" : sym
+    end
+    clear_sqls
+  end
+  after do
+    INTEGRATION_DB.drop_table(:items)
+  end
+  
+  specify "should support bound variables with select, all, and first" do
+    @ds.filter(:number=>@ds.ba(:$n)).call(:select, :n=>10).should == [{:id=>1, :number=>10}]
+    @ds.filter(:number=>@ds.ba(:$n)).call(:all, :n=>10).should == [{:id=>1, :number=>10}]
+    @ds.filter(:number=>@ds.ba(:$n)).call(:first, :n=>10).should == {:id=>1, :number=>10}
+  end
+
+  specify "should support placeholder literal strings" do
+    @ds.filter("number = ?", @ds.ba(:$n)).call(:select, :n=>10).should == [{:id=>1, :number=>10}]
+  end
+
+  specify "should support datasets with static sql and placeholders" do
+    INTEGRATION_DB["SELECT * FROM items WHERE number = ?", @ds.ba(:$n)].call(:select, :n=>10).should == [{:id=>1, :number=>10}]
+  end
+
+  specify "should support subselects" do
+    @ds.filter(:id=>:$i).filter(:number=>@ds.select(:number).filter(:number=>@ds.ba(:$n))).filter(:id=>:$j).call(:select, :n=>10, :i=>1, :j=>1).should == [{:id=>1, :number=>10}]
+  end
+
+  specify "should support subselects with literal strings" do
+    @ds.filter(:id=>:$i, :number=>@ds.select(:number).filter("number = ?", @ds.ba(:$n))).call(:select, :n=>10, :i=>1).should == [{:id=>1, :number=>10}]
+  end
+
+  specify "should support subselects with static sql and placeholders" do
+    @ds.filter(:id=>:$i, :number=>INTEGRATION_DB["SELECT number FROM items WHERE number = ?", @ds.ba(:$n)]).call(:select, :n=>10, :i=>1).should == [{:id=>1, :number=>10}]
+  end
+
+  specify "should support subselects of subselects" do
+    @ds.filter(:id=>:$i).filter(:number=>@ds.select(:number).filter(:number=>@ds.select(:number).filter(:number=>@ds.ba(:$n)))).filter(:id=>:$j).call(:select, :n=>10, :i=>1, :j=>1).should == [{:id=>1, :number=>10}]
+  end
+
+  specify "should support bound variables with insert" do
+    @ds.call(:insert, {:n=>20, :i=>100}, :id=>@ds.ba(:$i), :number=>@ds.ba(:$n))
+    @ds.count.should == 2
+    @ds.order(:id).all.should == [{:id=>1, :number=>10}, {:id=>100, :number=>20}]
+  end
+
+  specify "should have insert return primary key value when using bound arguments" do
+    @ds.call(:insert, {:n=>20}, :number=>@ds.ba(:$n)).should == 2
+    @ds.filter(:id=>2).first[:number].should == 20
+  end
+
+  specify "should support bound variables with delete" do
+    @ds.filter(:number=>@ds.ba(:$n)).call(:delete, :n=>10).should == 1
+    @ds.count.should == 0
+  end
+
+  specify "should support bound variables with update" do
+    @ds.filter(:number=>@ds.ba(:$n)).call(:update, {:n=>10, :nn=>20}, :number=>:number+@ds.ba(:$nn)).should == 1
+    @ds.all.should == [{:id=>1, :number=>30}]
+  end
+  
+  specify "should support prepared statements with select, first, and all" do
+    @ds.filter(:number=>@ds.ba(:$n)).prepare(:select, :select_n)
+    INTEGRATION_DB.call(:select_n, :n=>10).should == [{:id=>1, :number=>10}]
+    @ds.filter(:number=>@ds.ba(:$n)).prepare(:all, :select_n)
+    INTEGRATION_DB.call(:select_n, :n=>10).should == [{:id=>1, :number=>10}]
+    @ds.filter(:number=>@ds.ba(:$n)).prepare(:first, :select_n)
+    INTEGRATION_DB.call(:select_n, :n=>10).should == {:id=>1, :number=>10}
+    if INTEGRATION_DB.class.adapter_scheme == :jdbc and INTEGRATION_DB.database_type == :sqlite
+      # Work around for open prepared statements on a table not allowing the
+      # dropping of a table when using SQLite over JDBC
+      INTEGRATION_DB.synchronize{|c| c.prepared_statements[:select_n][1].close}
+    end
+  end
+
+  specify "should support prepared statements with insert" do
+    @ds.prepare(:insert, :insert_n, :id=>@ds.ba(:$i), :number=>@ds.ba(:$n))
+    INTEGRATION_DB.call(:insert_n, :n=>20, :i=>100)
+    @ds.count.should == 2
+    @ds.order(:id).all.should == [{:id=>1, :number=>10}, {:id=>100, :number=>20}]
+  end
+
+  specify "should have insert return primary key value when using prepared statements" do
+    @ds.prepare(:insert, :insert_n, :number=>@ds.ba(:$n))
+    INTEGRATION_DB.call(:insert_n, :n=>20).should == 2
+    @ds.filter(:id=>2).first[:number].should == 20
+  end
+
+  specify "should support prepared statements with delete" do
+    @ds.filter(:number=>@ds.ba(:$n)).prepare(:delete, :delete_n)
+    INTEGRATION_DB.call(:delete_n, :n=>10).should == 1
+    @ds.count.should == 0
+  end
+
+  specify "should support prepared statements with update" do
+    @ds.filter(:number=>@ds.ba(:$n)).prepare(:update, :update_n, :number=>:number+@ds.ba(:$nn))
+    INTEGRATION_DB.call(:update_n, :n=>10, :nn=>20).should == 1
+    @ds.all.should == [{:id=>1, :number=>30}]
+  end
+  
+  specify "model datasets should return model instances when using select, all, and first with bound variables" do
+    @c.filter(:number=>@ds.ba(:$n)).call(:select, :n=>10).should == [@c.load(:id=>1, :number=>10)]
+    @c.filter(:number=>@ds.ba(:$n)).call(:all, :n=>10).should == [@c.load(:id=>1, :number=>10)]
+    @c.filter(:number=>@ds.ba(:$n)).call(:first, :n=>10).should == @c.load(:id=>1, :number=>10)
+  end
+  
+  specify "model datasets should return model instances when using select, all, and first with prepared statements" do
+    @c.filter(:number=>@ds.ba(:$n)).prepare(:select, :select_n)
+    INTEGRATION_DB.call(:select_n, :n=>10).should == [@c.load(:id=>1, :number=>10)]
+    @c.filter(:number=>@ds.ba(:$n)).prepare(:all, :select_n)
+    INTEGRATION_DB.call(:select_n, :n=>10).should == [@c.load(:id=>1, :number=>10)]
+    @c.filter(:number=>@ds.ba(:$n)).prepare(:first, :select_n)
+    INTEGRATION_DB.call(:select_n, :n=>10).should == @c.load(:id=>1, :number=>10)
+    if INTEGRATION_DB.class.adapter_scheme == :jdbc and INTEGRATION_DB.database_type == :sqlite
+      # Work around for open prepared statements on a table not allowing the
+      # dropping of a table when using SQLite over JDBC
+      INTEGRATION_DB.synchronize{|c| c.prepared_statements[:select_n][1].close}
+    end
+  end
+end
diff -pruN 1.5.1-1/spec/integration/schema_test.rb 3.0.0-1/spec/integration/schema_test.rb
--- 1.5.1-1/spec/integration/schema_test.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/integration/schema_test.rb	2009-04-28 00:35:18.000000000 +0100
@@ -0,0 +1,232 @@
+require File.join(File.dirname(__FILE__), 'spec_helper.rb')
+
+if INTEGRATION_DB.respond_to?(:schema_parse_table, true)
+describe "Database schema parser" do
+  before do
+    @iom = INTEGRATION_DB.identifier_output_method
+    @iim = INTEGRATION_DB.identifier_input_method
+    @defsch = INTEGRATION_DB.default_schema
+    clear_sqls
+  end
+  after do
+    INTEGRATION_DB.drop_table(:items) if INTEGRATION_DB.table_exists?(:items)
+    INTEGRATION_DB.identifier_output_method = @iom
+    INTEGRATION_DB.identifier_input_method = @iim
+    INTEGRATION_DB.default_schema = @defsch
+  end
+
+  specify "should handle a database with a identifier_output_method" do
+    INTEGRATION_DB.identifier_output_method = :reverse
+    INTEGRATION_DB.identifier_input_method = :reverse
+    INTEGRATION_DB.default_schema = nil if INTEGRATION_DB.default_schema
+    INTEGRATION_DB.create_table!(:items){Integer :number}
+    INTEGRATION_DB.schema(:items, :reload=>true).should be_a_kind_of(Array)
+    INTEGRATION_DB.schema(:items, :reload=>true).first.first.should == :number
+  end
+
+  specify "should not issue an sql query if the schema has been loaded unless :reload is true" do
+    INTEGRATION_DB.create_table!(:items){Integer :number}
+    INTEGRATION_DB.schema(:items, :reload=>true)
+    clear_sqls
+    INTEGRATION_DB.schema(:items)
+    sqls_should_be
+    clear_sqls
+    INTEGRATION_DB.schema(:items, :reload=>true)
+    sqls_should_be "PRAGMA table_info('items')"
+  end
+
+  specify "should raise an error when the table doesn't exist" do
+    proc{INTEGRATION_DB.schema(:no_table)}.should raise_error(Sequel::Error)
+  end
+
+  specify "should return the schema correctly" do
+    INTEGRATION_DB.create_table!(:items){Integer :number}
+    schema = INTEGRATION_DB.schema(:items, :reload=>true)
+    schema.should be_a_kind_of(Array)
+    schema.length.should == 1
+    col = schema.first
+    col.should be_a_kind_of(Array)
+    col.length.should == 2
+    col.first.should == :number
+    col_info = col.last
+    col_info.should be_a_kind_of(Hash)
+    col_info[:type].should == :integer
+    clear_sqls
+    INTEGRATION_DB.schema(:items)
+    sqls_should_be
+  end
+
+  specify "should parse primary keys from the schema properly" do
+    INTEGRATION_DB.create_table!(:items){Integer :number}
+    INTEGRATION_DB.schema(:items).collect{|k,v| k if v[:primary_key]}.compact.should == []
+    INTEGRATION_DB.create_table!(:items){primary_key :number}
+    INTEGRATION_DB.schema(:items).collect{|k,v| k if v[:primary_key]}.compact.should == [:number]
+    INTEGRATION_DB.create_table!(:items){Integer :number1; Integer :number2; primary_key [:number1, :number2]}
+    INTEGRATION_DB.schema(:items).collect{|k,v| k if v[:primary_key]}.compact.should == [:number1, :number2]
+  end
+
+  specify "should parse NULL/NOT NULL from the schema properly" do
+    INTEGRATION_DB.create_table!(:items){Integer :number, :null=>true}
+    INTEGRATION_DB.schema(:items).first.last[:allow_null].should == true
+    INTEGRATION_DB.create_table!(:items){Integer :number, :null=>false}
+    INTEGRATION_DB.schema(:items).first.last[:allow_null].should == false
+  end
+
+  specify "should parse defaults from the schema properly" do
+    INTEGRATION_DB.create_table!(:items){Integer :number}
+    INTEGRATION_DB.schema(:items).first.last[:default].should == nil
+    INTEGRATION_DB.create_table!(:items){Integer :number, :default=>0}
+    INTEGRATION_DB.schema(:items).first.last[:default].to_s.should == "0"
+    INTEGRATION_DB.create_table!(:items){String :a, :default=>"blah"}
+    INTEGRATION_DB.schema(:items).first.last[:default].should include('blah')
+  end
+
+  specify "should parse types from the schema properly" do
+    INTEGRATION_DB.create_table!(:items){Integer :number}
+    INTEGRATION_DB.schema(:items).first.last[:type].should == :integer
+    INTEGRATION_DB.create_table!(:items){Fixnum :number}
+    INTEGRATION_DB.schema(:items).first.last[:type].should == :integer
+    INTEGRATION_DB.create_table!(:items){Bignum :number}
+    INTEGRATION_DB.schema(:items).first.last[:type].should == :integer
+    INTEGRATION_DB.create_table!(:items){Float :number}
+    INTEGRATION_DB.schema(:items).first.last[:type].should == :float
+    INTEGRATION_DB.create_table!(:items){BigDecimal :number}
+    INTEGRATION_DB.schema(:items).first.last[:type].should == :decimal
+    INTEGRATION_DB.create_table!(:items){Numeric :number}
+    INTEGRATION_DB.schema(:items).first.last[:type].should == :decimal
+    INTEGRATION_DB.create_table!(:items){String :number}
+    INTEGRATION_DB.schema(:items).first.last[:type].should == :string
+    INTEGRATION_DB.create_table!(:items){Date :number}
+    INTEGRATION_DB.schema(:items).first.last[:type].should == :date
+    INTEGRATION_DB.create_table!(:items){Time :number}
+    INTEGRATION_DB.schema(:items).first.last[:type].should == :datetime
+    INTEGRATION_DB.create_table!(:items){DateTime :number}
+    INTEGRATION_DB.schema(:items).first.last[:type].should == :datetime
+    INTEGRATION_DB.create_table!(:items){File :number}
+    INTEGRATION_DB.schema(:items).first.last[:type].should == :blob
+    INTEGRATION_DB.create_table!(:items){TrueClass :number}
+    INTEGRATION_DB.schema(:items).first.last[:type].should == :boolean
+    INTEGRATION_DB.create_table!(:items){FalseClass :number}
+    INTEGRATION_DB.schema(:items).first.last[:type].should == :boolean
+  end
+end
+end
+
+if INTEGRATION_DB.respond_to?(:indexes)
+describe "Database index parsing" do
+  after do
+    INTEGRATION_DB.drop_table(:items)
+  end
+
+  specify "should parse indexes into a hash" do
+    INTEGRATION_DB.create_table!(:items){Integer :n; Integer :a}
+    INTEGRATION_DB.indexes(:items).should == {}
+    INTEGRATION_DB.add_index(:items, :n)
+    INTEGRATION_DB.indexes(:items).should == {:items_n_index=>{:columns=>[:n], :unique=>false}}
+    INTEGRATION_DB.drop_index(:items, :n)
+    INTEGRATION_DB.indexes(:items).should == {}
+    INTEGRATION_DB.add_index(:items, :n, :unique=>true, :name=>:blah_blah_index)
+    INTEGRATION_DB.indexes(:items).should == {:blah_blah_index=>{:columns=>[:n], :unique=>true}}
+    INTEGRATION_DB.add_index(:items, [:n, :a])
+    INTEGRATION_DB.indexes(:items).should == {:blah_blah_index=>{:columns=>[:n], :unique=>true}, :items_n_a_index=>{:columns=>[:n, :a], :unique=>false}}
+    INTEGRATION_DB.drop_index(:items, :n, :name=>:blah_blah_index)
+    INTEGRATION_DB.indexes(:items).should == {:items_n_a_index=>{:columns=>[:n, :a], :unique=>false}}
+    INTEGRATION_DB.drop_index(:items, [:n, :a])
+    INTEGRATION_DB.indexes(:items).should == {}
+  end
+  
+  specify "should not include a primary key index" do
+    INTEGRATION_DB.create_table!(:items){primary_key :n}
+    INTEGRATION_DB.indexes(:items).should == {}
+    INTEGRATION_DB.create_table!(:items){Integer :n; Integer :a; primary_key [:n, :a]}
+    INTEGRATION_DB.indexes(:items).should == {}
+  end
+end
+end
+
+describe "Database schema modifiers" do
+  before do
+    INTEGRATION_DB.create_table!(:items){Integer :number}
+    @ds = INTEGRATION_DB[:items]
+    @ds.insert([10])
+    clear_sqls
+  end
+  after do
+    INTEGRATION_DB.drop_table(:items) if INTEGRATION_DB.table_exists?(:items)
+  end
+
+  specify "should create tables correctly" do
+    INTEGRATION_DB.table_exists?(:items).should == true
+    INTEGRATION_DB.schema(:items, :reload=>true).map{|x| x.first}.should == [:number]
+    @ds.columns!.should == [:number]
+  end
+
+  specify "should add columns to tables correctly" do
+    INTEGRATION_DB.schema(:items, :reload=>true).map{|x| x.first}.should == [:number]
+    @ds.columns!.should == [:number]
+    INTEGRATION_DB.alter_table(:items){add_column :name, :text}
+    INTEGRATION_DB.schema(:items, :reload=>true).map{|x| x.first}.should == [:number, :name]
+    @ds.columns!.should == [:number, :name]
+    unless INTEGRATION_DB.database_type == :sqlite
+      INTEGRATION_DB.alter_table(:items){add_primary_key :id}
+      INTEGRATION_DB.schema(:items, :reload=>true).map{|x| x.first}.should == [:number, :name, :id]
+      @ds.columns!.should == [:number, :name, :id]
+      INTEGRATION_DB.alter_table(:items){add_foreign_key :item_id, :items}
+      INTEGRATION_DB.schema(:items, :reload=>true).map{|x| x.first}.should == [:number, :name, :id, :item_id]
+      @ds.columns!.should == [:number, :name, :id, :item_id]
+    end
+  end
+
+  specify "should remove columns from tables correctly" do
+    INTEGRATION_DB.create_table!(:items) do
+      primary_key :id
+      String :name
+      Integer :number
+      foreign_key :item_id, :items
+    end
+    @ds.insert(:number=>10)
+    INTEGRATION_DB.schema(:items, :reload=>true).map{|x| x.first}.should == [:id, :name, :number, :item_id]
+    @ds.columns!.should == [:id, :name, :number, :item_id]
+    INTEGRATION_DB.drop_column(:items, :number)
+    INTEGRATION_DB.schema(:items, :reload=>true).map{|x| x.first}.should == [:id, :name, :item_id]
+    @ds.columns!.should == [:id, :name, :item_id]
+    INTEGRATION_DB.drop_column(:items, :name)
+    INTEGRATION_DB.schema(:items, :reload=>true).map{|x| x.first}.should == [:id, :item_id]
+    @ds.columns!.should == [:id, :item_id]
+    INTEGRATION_DB.drop_column(:items, :item_id)
+    INTEGRATION_DB.schema(:items, :reload=>true).map{|x| x.first}.should == [:id]
+    @ds.columns!.should == [:id]
+  end
+end
+
+if INTEGRATION_DB.respond_to?(:tables)
+describe "Database#tables" do
+  before do
+    class ::String
+      @@xxxxx = 0
+      def xxxxx
+        "xxxxx#{@@xxxxx += 1}"
+      end
+    end
+    @iom = INTEGRATION_DB.identifier_output_method
+    @iim = INTEGRATION_DB.identifier_input_method
+    clear_sqls
+  end
+  after do
+    INTEGRATION_DB.identifier_output_method = @iom
+    INTEGRATION_DB.identifier_input_method = @iim
+  end
+
+  specify "should return an array of symbols" do
+    ts = INTEGRATION_DB.tables
+    ts.should be_a_kind_of(Array)
+    ts.each{|t| t.should be_a_kind_of(Symbol)}
+  end
+
+  specify "should respect the database's identifier_output_method" do
+    INTEGRATION_DB.identifier_output_method = :xxxxx
+    INTEGRATION_DB.identifier_input_method = :xxxxx
+    INTEGRATION_DB.tables.each{|t| t.to_s.should =~ /\Ax{5}\d+\z/}
+  end
+end
+end
diff -pruN 1.5.1-1/spec/integration/spec_helper.rb 3.0.0-1/spec/integration/spec_helper.rb
--- 1.5.1-1/spec/integration/spec_helper.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/integration/spec_helper.rb	2009-04-09 22:59:30.000000000 +0100
@@ -0,0 +1,61 @@
+require 'rubygems'
+unless Object.const_defined?('Sequel')
+  $:.unshift(File.join(File.dirname(__FILE__), "../../lib/"))
+  require 'sequel'
+end
+begin
+  require File.join(File.dirname(__FILE__), '../spec_config.rb')
+rescue LoadError
+end
+
+Sequel.virtual_row_instance_eval = true
+Sequel::Model.use_transactions = false
+
+$sqls = []
+def clear_sqls
+  $sqls.clear
+end 
+
+class Spec::Example::ExampleGroup
+  def start_logging
+    require 'logger'
+    INTEGRATION_DB.loggers << Logger.new(STDOUT)
+  end
+  def stop_logging
+     INTEGRATION_DB.loggers.clear
+  end
+end
+
+if defined?(INTEGRATION_DB) || defined?(INTEGRATION_URL) || ENV['SEQUEL_INTEGRATION_URL']
+  unless defined?(INTEGRATION_DB)
+    url = defined?(INTEGRATION_URL) ? INTEGRATION_URL : ENV['SEQUEL_INTEGRATION_URL']
+    INTEGRATION_DB = Sequel.connect(url)
+    #INTEGRATION_DB.instance_variable_set(:@server_version, 80100)
+  end
+  class Spec::Example::ExampleGroup
+    def sqls_should_be(*args)
+    end 
+  end
+else
+  sql_logger = Object.new
+  def sql_logger.info(str)
+    $sqls << str 
+  end
+  INTEGRATION_DB = Sequel.sqlite('', :loggers=>[sql_logger], :quote_identifiers=>false)
+  class Spec::Example::ExampleGroup
+    def sqls_should_be(*sqls)
+      sqls.zip($sqls).each do |should_be, is|
+        case should_be
+        when String
+          is.should == should_be
+        when Regexp
+          is.should =~ should_be
+        else
+          raise ArgumentError, "need String or RegExp"
+        end
+      end
+      $sqls.length.should == sqls.length
+      clear_sqls
+    end 
+  end
+end
diff -pruN 1.5.1-1/spec/integration/transaction_test.rb 3.0.0-1/spec/integration/transaction_test.rb
--- 1.5.1-1/spec/integration/transaction_test.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/integration/transaction_test.rb	2009-04-13 01:58:36.000000000 +0100
@@ -0,0 +1,95 @@
+require File.join(File.dirname(__FILE__), 'spec_helper.rb')
+
+describe "Database transactions" do
+  before do
+    INTEGRATION_DB.drop_table(:items) if INTEGRATION_DB.table_exists?(:items)
+    INTEGRATION_DB.create_table(:items){String :name; Integer :value}
+    @d = INTEGRATION_DB[:items]
+    clear_sqls
+  end
+  after do
+    INTEGRATION_DB.drop_table(:items) if INTEGRATION_DB.table_exists?(:items)
+  end
+
+  specify "should support transactions" do
+    INTEGRATION_DB.transaction do
+      @d << {:name => 'abc', :value => 1}
+    end
+
+    @d.count.should == 1
+  end
+
+  specify "should have #transaction yield the connection" do
+    INTEGRATION_DB.transaction do |conn|
+      conn.should_not == nil
+    end
+  end
+
+  specify "should correctly rollback transactions" do
+    proc do
+      INTEGRATION_DB.transaction do
+        @d << {:name => 'abc', :value => 1}
+        raise Interrupt, 'asdf'
+      end
+    end.should raise_error(Interrupt)
+
+    proc do
+      INTEGRATION_DB.transaction do
+        @d << {:name => 'abc', :value => 1}
+        raise Sequel::Rollback
+      end
+    end.should_not raise_error
+
+    @d.count.should == 0
+  end
+
+  specify "should support nested transactions" do
+    @db = INTEGRATION_DB
+    @db.transaction do
+      @db.transaction do
+        @d << {:name => 'abc', :value => 1}
+      end 
+    end 
+    @d.count.should == 1
+
+    @d.delete
+    proc {@db.transaction do
+      @d << {:name => 'abc', :value => 1}
+      @db.transaction do
+        raise Sequel::Rollback
+      end 
+    end}.should_not raise_error
+    @d.count.should == 0
+
+    proc {@db.transaction do
+      @d << {:name => 'abc', :value => 1}
+      @db.transaction do
+        raise Interrupt, 'asdf'
+      end 
+    end}.should raise_error(Interrupt)
+    @d.count.should == 0
+  end 
+
+  specify "should handle returning inside of the block by committing" do
+    def INTEGRATION_DB.ret_commit
+      transaction do
+        self[:items] << {:name => 'abc'}
+        return
+        self[:items] << {:name => 'd'}
+      end
+    end
+
+    @d.count.should == 0
+    INTEGRATION_DB.ret_commit
+    @d.count.should == 1
+    INTEGRATION_DB.ret_commit
+    @d.count.should == 2
+    proc do
+      INTEGRATION_DB.transaction do
+        raise Interrupt, 'asdf'
+      end
+    end.should raise_error(Interrupt)
+
+    @d.count.should == 2
+  end
+end
diff -pruN 1.5.1-1/spec/integration/type_test.rb 3.0.0-1/spec/integration/type_test.rb
--- 1.5.1-1/spec/integration/type_test.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/integration/type_test.rb	2009-04-13 01:14:03.000000000 +0100
@@ -0,0 +1,96 @@
+require File.join(File.dirname(__FILE__), 'spec_helper.rb')
+
+describe "Supported types" do
+  def create_items_table_with_column(name, type, opts={})
+    INTEGRATION_DB.create_table!(:items){column name, type, opts}
+    INTEGRATION_DB[:items]
+  end
+
+  specify "should support casting correctly" do
+    ds = create_items_table_with_column(:number, Integer)
+    ds.insert(:number => 1)
+    ds.select(:number.cast_string.as(:n)).map(:n).should == %w'1'
+    ds = create_items_table_with_column(:name, String)
+    ds.insert(:name=> '1')
+    ds.select(:name.cast_numeric.as(:n)).map(:n).should == [1]
+  end
+
+  specify "should support NULL correctly" do
+    ds = create_items_table_with_column(:number, Integer)
+    ds.insert(:number => nil)
+    ds.all.should == [{:number=>nil}]
+  end
+
+  specify "should support generic integer type" do
+    ds = create_items_table_with_column(:number, Integer)
+    ds.insert(:number => 2)
+    ds.all.should == [{:number=>2}]
+  end
+  
+  specify "should support generic fixnum type" do
+    ds = create_items_table_with_column(:number, Fixnum)
+    ds.insert(:number => 2)
+    ds.all.should == [{:number=>2}]
+  end
+  
+  specify "should support generic bignum type" do
+    ds = create_items_table_with_column(:number, Bignum)
+    ds.insert(:number => 2**34)
+    ds.all.should == [{:number=>2**34}]
+  end
+  
+  specify "should support generic float type" do
+    ds = create_items_table_with_column(:number, Float)
+    ds.insert(:number => 2.1)
+    ds.all.should == [{:number=>2.1}]
+  end
+  
+  specify "should support generic numeric type" do
+    ds = create_items_table_with_column(:number, Numeric, :size=>[15, 10])
+    ds.insert(:number => BigDecimal.new('2.123456789'))
+    ds.all.should == [{:number=>BigDecimal.new('2.123456789')}]
+    ds = create_items_table_with_column(:number, BigDecimal, :size=>[15, 10])
+    ds.insert(:number => BigDecimal.new('2.123456789'))
+    ds.all.should == [{:number=>BigDecimal.new('2.123456789')}]
+  end
+
+  specify "should support generic string type" do
+    ds = create_items_table_with_column(:name, String)
+    ds.insert(:name => 'Test User')
+    ds.all.should == [{:name=>'Test User'}]
+  end
+  
+  specify "should support generic date type" do
+    ds = create_items_table_with_column(:dat, Date)
+    d = Date.today
+    ds.insert(:dat => d)
+    ds.first[:dat].should == d
+  end
+  
+  specify "should support generic datetime type" do
+    ds = create_items_table_with_column(:tim, DateTime)
+    t = DateTime.now
+    ds.insert(:tim => t)
+    ds.first[:tim].strftime('%Y%m%d%H%M%S').should == t.strftime('%Y%m%d%H%M%S')
+    ds = create_items_table_with_column(:tim, Time)
+    t = Time.now
+    ds.insert(:tim => t)
+    ds.first[:tim].strftime('%Y%m%d%H%M%S').should == t.strftime('%Y%m%d%H%M%S')
+  end
+  
+  specify "should support generic file type" do
+    ds = create_items_table_with_column(:name, File)
+    ds.insert(:name => ("a\0"*300).to_sequel_blob)
+    ds.all.should == [{:name=>("a\0"*300).to_sequel_blob}]
+    ds.first[:name].should be_a_kind_of(::Sequel::SQL::Blob)
+  end
+  
+  specify "should support generic boolean type" do
+    ds = create_items_table_with_column(:number, TrueClass)
+    ds.insert(:number => true)
+    ds.all.should == [{:number=>true}]
+    ds = create_items_table_with_column(:number, FalseClass)
+    ds.insert(:number => true)
+    ds.all.should == [{:number=>true}]
+  end
+end
diff -pruN 1.5.1-1/spec/model/association_reflection_spec.rb 3.0.0-1/spec/model/association_reflection_spec.rb
--- 1.5.1-1/spec/model/association_reflection_spec.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/model/association_reflection_spec.rb	2009-03-16 20:53:19.000000000 +0000
@@ -0,0 +1,93 @@
+require File.join(File.dirname(__FILE__), "spec_helper")
+
+describe Sequel::Model::Associations::AssociationReflection, "#associated_class" do
+  before do
+    @c = Class.new(Sequel::Model)
+    class ::ParParent < Sequel::Model; end
+  end
+
+  it "should use the :class value if present" do
+    @c.many_to_one :c, :class=>ParParent
+    @c.association_reflection(:c).keys.should include(:class)
+    @c.association_reflection(:c).associated_class.should == ParParent
+  end
+  it "should figure out the class if the :class value is not present" do
+    @c.many_to_one :c, :class=>'ParParent'
+    @c.association_reflection(:c).keys.should_not include(:class)
+    @c.association_reflection(:c).associated_class.should == ParParent
+  end
+end
+
+describe Sequel::Model::Associations::AssociationReflection, "#primary_key" do
+  before do
+    @c = Class.new(Sequel::Model)
+    class ::ParParent < Sequel::Model; end
+  end
+
+  it "should use the :primary_key value if present" do
+    @c.many_to_one :c, :class=>ParParent, :primary_key=>:blah__blah
+    @c.association_reflection(:c).keys.should include(:primary_key)
+    @c.association_reflection(:c).primary_key.should == :blah__blah
+  end
+  it "should use the associated table's primary key if :primary_key is not present" do
+    @c.many_to_one :c, :class=>'ParParent'
+    @c.association_reflection(:c).keys.should_not include(:primary_key)
+    @c.association_reflection(:c).primary_key.should == :id
+  end
+end
+
+describe Sequel::Model::Associations::AssociationReflection, "#reciprocal" do
+  it "should use the :reciprocal value if present" do
+    @c = Class.new(Sequel::Model)
+    @d = Class.new(Sequel::Model)
+    @c.many_to_one :c, :class=>@d, :reciprocal=>:xx
+    @c.association_reflection(:c).keys.should include(:reciprocal)
+    @c.association_reflection(:c).reciprocal.should == :xx
+  end
+
+  it "should figure out the reciprocal if the :reciprocal value is not present" do
+    class ::ParParent < Sequel::Model; end
+    class ::ParParentTwo < Sequel::Model; end
+    class ::ParParentThree < Sequel::Model; end
+    ParParent.many_to_one :par_parent_two
+    ParParentTwo.one_to_many :par_parents
+    ParParent.many_to_many :par_parent_threes
+    ParParentThree.many_to_many :par_parents
+
+    ParParent.association_reflection(:par_parent_two).keys.should_not include(:reciprocal)
+    ParParent.association_reflection(:par_parent_two).reciprocal.should == :par_parents
+    ParParentTwo.association_reflection(:par_parents).keys.should_not include(:reciprocal)
+    ParParentTwo.association_reflection(:par_parents).reciprocal.should == :par_parent_two
+    ParParent.association_reflection(:par_parent_threes).keys.should_not include(:reciprocal)
+    ParParent.association_reflection(:par_parent_threes).reciprocal.should == :par_parents
+    ParParentThree.association_reflection(:par_parents).keys.should_not include(:reciprocal)
+    ParParentThree.association_reflection(:par_parents).reciprocal.should == :par_parent_threes
+  end
+end
+
+describe Sequel::Model::Associations::AssociationReflection, "#select" do
+  before do
+    @c = Class.new(Sequel::Model)
+    class ::ParParent < Sequel::Model; end
+  end
+
+  it "should use the :select value if present" do
+    @c.many_to_one :c, :class=>ParParent, :select=>[:par_parents__id]
+    @c.association_reflection(:c).keys.should include(:select)
+    @c.association_reflection(:c).select.should == [:par_parents__id]
+  end
+  it "should be the associated_table.* if :select is not present for a many_to_many associaiton" do
+    @c.many_to_many :cs, :class=>'ParParent'
+    @c.association_reflection(:cs).keys.should_not include(:select)
+    @c.association_reflection(:cs).select.should == :par_parents.*
+  end
+  it "should be if :select is not present for a many_to_one and one_to_many associaiton" do
+    @c.one_to_many :cs, :class=>'ParParent'
+    @c.association_reflection(:cs).keys.should_not include(:select)
+    @c.association_reflection(:cs).select.should == nil
+    @c.many_to_one :c, :class=>'ParParent'
+    @c.association_reflection(:c).keys.should_not include(:select)
+    @c.association_reflection(:c).select.should == nil
+  end
+end
+
diff -pruN 1.5.1-1/spec/model/associations_spec.rb 3.0.0-1/spec/model/associations_spec.rb
--- 1.5.1-1/spec/model/associations_spec.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/model/associations_spec.rb	2009-04-09 18:00:58.000000000 +0100
@@ -0,0 +1,1744 @@
+require File.join(File.dirname(__FILE__), "spec_helper")
+
+describe Sequel::Model, "associate" do
+  it "should use explicit class if given a class, symbol, or string" do
+    MODEL_DB.reset
+    klass = Class.new(Sequel::Model(:nodes))
+    class ::ParParent < Sequel::Model
+    end
+    
+    klass.associate :many_to_one, :par_parent0, :class=>ParParent
+    klass.associate :one_to_many, :par_parent1s, :class=>'ParParent'
+    klass.associate :many_to_many, :par_parent2s, :class=>:ParParent
+    
+    klass.association_reflection(:"par_parent0").associated_class.should == ParParent
+    klass.association_reflection(:"par_parent1s").associated_class.should == ParParent
+    klass.association_reflection(:"par_parent2s").associated_class.should == ParParent
+  end
+
+  it "should add a model_object and association_reflection accessors to the dataset, and return it with the current model object" do
+    MODEL_DB.reset
+    klass = Class.new(Sequel::Model(:nodes)) do
+      columns :id, :a_id
+    end
+    mod = Module.new do
+      def blah
+       filter{|o| o.__send__(association_reflection[:key]) > model_object.id*2}
+      end
+    end
+
+    klass.associate :many_to_one, :a, :class=>klass
+    klass.associate :one_to_many, :bs, :key=>:b_id, :class=>klass, :extend=>mod
+    klass.associate :many_to_many, :cs, :class=>klass
+    
+    node = klass.load(:id=>1)
+    node.a_dataset.model_object.should == node
+    node.bs_dataset.model_object.should == node
+    node.cs_dataset.model_object.should == node
+
+    node.a_dataset.association_reflection.should == klass.association_reflection(:a)
+    node.bs_dataset.association_reflection.should == klass.association_reflection(:bs)
+    node.cs_dataset.association_reflection.should == klass.association_reflection(:cs)
+
+    node.bs_dataset.blah.sql.should == 'SELECT * FROM nodes WHERE ((nodes.b_id = 1) AND (b_id > 2))'
+  end
+
+  it "should allow extending the dataset with :extend option" do
+    MODEL_DB.reset
+    klass = Class.new(Sequel::Model(:nodes)) do
+      columns :id, :a_id
+    end
+    mod = Module.new do
+      def blah
+       1
+      end
+    end
+    mod2 = Module.new do
+      def blar
+       2
+      end
+    end
+    
+    klass.associate :many_to_one, :a, :class=>klass, :extend=>mod
+    klass.associate :one_to_many, :bs, :class=>klass, :extend=>[mod]
+    klass.associate :many_to_many, :cs, :class=>klass, :extend=>[mod, mod2]
+    
+    node = klass.load(:id=>1)
+    node.a_dataset.blah.should == 1
+    node.bs_dataset.blah.should == 1
+    node.cs_dataset.blah.should == 1
+    node.cs_dataset.blar.should == 2
+  end
+
+  it "should clone an existing association with the :clone option" do
+    MODEL_DB.reset
+    klass = Class.new(Sequel::Model(:nodes))
+    
+    klass.many_to_one(:par_parent, :order=>:a){1}
+    klass.one_to_many(:par_parent1s, :class=>'ParParent', :limit=>12){4}
+    klass.many_to_many(:par_parent2s, :class=>:ParParent, :uniq=>true){2}
+
+    klass.many_to_one :par, :clone=>:par_parent, :select=>:b
+    klass.one_to_many :par1s, :clone=>:par_parent1s, :order=>:b, :limit=>10, :block=>nil
+    klass.many_to_many(:par2s, :clone=>:par_parent2s, :order=>:c){3}
+    
+    klass.association_reflection(:par).associated_class.should == ParParent
+    klass.association_reflection(:par1s).associated_class.should == ParParent
+    klass.association_reflection(:par2s).associated_class.should == ParParent
+    
+    klass.association_reflection(:par)[:order].should == :a
+    klass.association_reflection(:par).select.should == :b
+    klass.association_reflection(:par)[:block].call.should == 1
+    klass.association_reflection(:par1s)[:limit].should == 10
+    klass.association_reflection(:par1s)[:order].should == :b
+    klass.association_reflection(:par1s)[:block].should == nil
+    klass.association_reflection(:par2s)[:after_load].length.should == 1
+    klass.association_reflection(:par2s)[:order].should == :c
+    klass.association_reflection(:par2s)[:block].call.should == 3
+  end
+
+end
+
+describe Sequel::Model, "many_to_one" do
+  before do
+    MODEL_DB.reset
+    
+    @c2 = Class.new(Sequel::Model(:nodes)) do
+      unrestrict_primary_key
+      columns :id, :parent_id, :par_parent_id, :blah
+    end
+
+    @dataset = @c2.dataset
+  end
+
+  it "should use implicit key if omitted" do
+    @c2.many_to_one :parent, :class => @c2
+
+    d = @c2.new(:id => 1, :parent_id => 234)
+    p = d.parent
+    p.class.should == @c2
+    p.values.should == {:x => 1, :id => 1}
+
+    MODEL_DB.sqls.should == ["SELECT * FROM nodes WHERE (nodes.id = 234) LIMIT 1"]
+  end
+  
+  it "should use implicit class if omitted" do
+    class ::ParParent < Sequel::Model
+    end
+    
+    @c2.many_to_one :par_parent
+    
+    d = @c2.new(:id => 1, :par_parent_id => 234)
+    p = d.par_parent
+    p.class.should == ParParent
+    
+    MODEL_DB.sqls.should == ["SELECT * FROM par_parents WHERE (par_parents.id = 234) LIMIT 1"]
+  end
+
+  it "should use class inside module if given as a string" do
+    module ::Par 
+      class Parent < Sequel::Model
+      end
+    end
+    
+    @c2.many_to_one :par_parent, :class=>"Par::Parent"
+    
+    d = @c2.new(:id => 1, :par_parent_id => 234)
+    p = d.par_parent
+    p.class.should == Par::Parent
+    
+    MODEL_DB.sqls.should == ["SELECT * FROM parents WHERE (parents.id = 234) LIMIT 1"]
+  end
+
+  it "should use explicit key if given" do
+    @c2.many_to_one :parent, :class => @c2, :key => :blah
+
+    d = @c2.new(:id => 1, :blah => 567)
+    p = d.parent
+    p.class.should == @c2
+    p.values.should == {:x => 1, :id => 1}
+
+    MODEL_DB.sqls.should == ["SELECT * FROM nodes WHERE (nodes.id = 567) LIMIT 1"]
+  end
+
+  it "should use :primary_key option if given" do
+    @c2.many_to_one :parent, :class => @c2, :key => :blah, :primary_key => :pk
+    @c2.new(:id => 1, :blah => 567).parent
+    MODEL_DB.sqls.should == ["SELECT * FROM nodes WHERE (nodes.pk = 567) LIMIT 1"]
+  end
+
+  it "should use :select option if given" do
+    @c2.many_to_one :parent, :class => @c2, :key => :blah, :select=>[:id, :name]
+    @c2.new(:id => 1, :blah => 567).parent
+    MODEL_DB.sqls.should == ["SELECT id, name FROM nodes WHERE (nodes.id = 567) LIMIT 1"]
+  end
+
+  it "should use :conditions option if given" do
+    @c2.many_to_one :parent, :class => @c2, :key => :blah, :conditions=>{:a=>32}
+    @c2.new(:id => 1, :blah => 567).parent
+    MODEL_DB.sqls.should == ["SELECT * FROM nodes WHERE ((nodes.id = 567) AND (a = 32)) LIMIT 1"]
+
+    @c2.many_to_one :parent, :class => @c2, :key => :blah, :conditions=>:a
+    MODEL_DB.sqls.clear
+    @c2.new(:id => 1, :blah => 567).parent
+    MODEL_DB.sqls.should == ["SELECT * FROM nodes WHERE ((nodes.id = 567) AND a) LIMIT 1"]
+  end
+
+  it "should support :order, :limit (only for offset), and :dataset options, as well as a block" do
+    c2 = @c2
+    @c2.many_to_one :child_20, :class => @c2, :key=>:id, :dataset=>proc{c2.filter(:parent_id=>pk)}, :limit=>[10,20], :order=>:name do |ds|
+      ds.filter(:x.sql_number > 1)
+    end
+    @c2.load(:id => 100).child_20
+    MODEL_DB.sqls.should == ["SELECT * FROM nodes WHERE ((parent_id = 100) AND (x > 1)) ORDER BY name LIMIT 1 OFFSET 20"]
+  end
+
+  it "should return nil if key value is nil" do
+    @c2.many_to_one :parent, :class => @c2
+
+    d = @c2.new(:id => 1)
+    d.parent.should == nil
+  end
+
+  it "should cache negative lookup" do
+    @c2.many_to_one :parent, :class => @c2
+    ds = @c2.dataset
+    def ds.fetch_rows(sql, &block)
+      MODEL_DB.sqls << sql
+    end
+
+    d = @c2.new(:id => 1, :parent_id=>555)
+    MODEL_DB.sqls.should == []
+    d.parent.should == nil
+    MODEL_DB.sqls.should == ['SELECT * FROM nodes WHERE (nodes.id = 555) LIMIT 1']
+    d.parent.should == nil
+    MODEL_DB.sqls.should == ['SELECT * FROM nodes WHERE (nodes.id = 555) LIMIT 1']
+  end
+
+  it "should define a setter method" do
+    @c2.many_to_one :parent, :class => @c2
+
+    d = @c2.new(:id => 1)
+    d.parent = @c2.new(:id => 4321)
+    d.values.should == {:id => 1, :parent_id => 4321}
+
+    d.parent = nil
+    d.values.should == {:id => 1, :parent_id => nil}
+
+    e = @c2.new(:id => 6677)
+    d.parent = e
+    d.values.should == {:id => 1, :parent_id => 6677}
+  end
+  
+  it "should have the setter method respect the :primary_key option" do
+    @c2.many_to_one :parent, :class => @c2, :primary_key=>:blah
+
+    d = @c2.new(:id => 1)
+    d.parent = @c2.new(:id => 4321, :blah=>444)
+    d.values.should == {:id => 1, :parent_id => 444}
+
+    d.parent = nil
+    d.values.should == {:id => 1, :parent_id => nil}
+
+    e = @c2.new(:id => 6677, :blah=>8)
+    d.parent = e
+    d.values.should == {:id => 1, :parent_id => 8}
+  end
+  
+  it "should not persist changes until saved" do
+    @c2.many_to_one :parent, :class => @c2
+
+    d = @c2.load(:id => 1)
+    MODEL_DB.reset
+    d.parent = @c2.new(:id => 345)
+    MODEL_DB.sqls.should == []
+    d.save_changes
+    MODEL_DB.sqls.should == ['UPDATE nodes SET parent_id = 345 WHERE (id = 1)']
+  end
+
+  it "should set cached instance variable when accessed" do
+    @c2.many_to_one :parent, :class => @c2
+
+    d = @c2.load(:id => 1)
+    MODEL_DB.reset
+    d.parent_id = 234
+    d.associations[:parent].should == nil
+    ds = @c2.dataset
+    def ds.fetch_rows(sql, &block); MODEL_DB.sqls << sql; yield({:id=>234}) end
+    e = d.parent 
+    MODEL_DB.sqls.should == ["SELECT * FROM nodes WHERE (nodes.id = 234) LIMIT 1"]
+    d.associations[:parent].should == e
+  end
+
+  it "should set cached instance variable when assigned" do
+    @c2.many_to_one :parent, :class => @c2
+
+    d = @c2.create(:id => 1)
+    MODEL_DB.reset
+    d.associations[:parent].should == nil
+    d.parent = @c2.new(:id => 234)
+    e = d.parent 
+    d.associations[:parent].should == e
+    MODEL_DB.sqls.should == []
+  end
+
+  it "should use cached instance variable if available" do
+    @c2.many_to_one :parent, :class => @c2
+
+    d = @c2.create(:id => 1, :parent_id => 234)
+    MODEL_DB.reset
+    d.associations[:parent] = 42
+    d.parent.should == 42
+    MODEL_DB.sqls.should == []
+  end
+
+  it "should not use cached instance variable if asked to reload" do
+    @c2.many_to_one :parent, :class => @c2
+
+    d = @c2.create(:id => 1)
+    MODEL_DB.reset
+    d.parent_id = 234
+    d.associations[:parent] = 42
+    d.parent(true).should_not == 42 
+    MODEL_DB.sqls.should == ["SELECT * FROM nodes WHERE (nodes.id = 234) LIMIT 1"]
+  end
+  
+  it "should have the setter add to the reciprocal one_to_many cached association list if it exists" do
+    @c2.many_to_one :parent, :class => @c2
+    @c2.one_to_many :children, :class => @c2, :key=>:parent_id
+    ds = @c2.dataset
+    def ds.fetch_rows(sql, &block)
+      MODEL_DB.sqls << sql
+    end
+
+    d = @c2.new(:id => 1)
+    e = @c2.new(:id => 2)
+    MODEL_DB.sqls.should == []
+    d.parent = e
+    e.children.should_not(include(d))
+    MODEL_DB.sqls.should == ['SELECT * FROM nodes WHERE (nodes.parent_id = 2)']
+
+    MODEL_DB.reset
+    d = @c2.new(:id => 1)
+    e = @c2.new(:id => 2)
+    e.children.should_not(include(d))
+    MODEL_DB.sqls.should == ['SELECT * FROM nodes WHERE (nodes.parent_id = 2)']
+    d.parent = e
+    e.children.should(include(d))
+    MODEL_DB.sqls.should == ['SELECT * FROM nodes WHERE (nodes.parent_id = 2)']
+  end
+
+  it "should have the setter remove the object from the previous associated object's reciprocal one_to_many cached association list if it exists" do
+    @c2.many_to_one :parent, :class => @c2
+    @c2.one_to_many :children, :class => @c2, :key=>:parent_id
+    ds = @c2.dataset
+    def ds.fetch_rows(sql, &block)
+      MODEL_DB.sqls << sql
+    end
+
+    d = @c2.new(:id => 1)
+    e = @c2.new(:id => 2)
+    f = @c2.new(:id => 3)
+    e.children.should_not(include(d))
+    f.children.should_not(include(d))
+    MODEL_DB.reset
+    d.parent = e
+    e.children.should(include(d))
+    d.parent = f
+    f.children.should(include(d))
+    e.children.should_not(include(d))
+    d.parent = nil
+    f.children.should_not(include(d))
+    MODEL_DB.sqls.should == []
+  end
+
+  it "should get all matching records and only return the first if :key option is set to nil" do
+    c2 = @c2
+    @c2.one_to_many :children, :class => @c2, :key=>:parent_id
+    @c2.many_to_one :first_grand_parent, :class => @c2, :key=>nil, :eager_graph=>:children, :dataset=>proc{c2.filter(:children_id=>parent_id)}
+    ds = @c2.dataset
+    def ds.columns
+      [:id, :parent_id, :par_parent_id, :blah]
+    end
+    def ds.fetch_rows(sql, &block)
+      MODEL_DB.sqls << sql
+      yield({:id=>1, :parent_id=>0, :par_parent_id=>3, :blah=>4, :children_id=>2, :children_parent_id=>1, :children_par_parent_id=>5, :children_blah=>6})
+    end
+    p = @c2.new(:parent_id=>2)
+    fgp = p.first_grand_parent
+    MODEL_DB.sqls.should == ["SELECT nodes.id, nodes.parent_id, nodes.par_parent_id, nodes.blah, children.id AS children_id, children.parent_id AS children_parent_id, children.par_parent_id AS children_par_parent_id, children.blah AS children_blah FROM nodes LEFT OUTER JOIN nodes AS children ON (children.parent_id = nodes.id) WHERE (children_id = 2)"]
+    fgp.values.should == {:id=>1, :parent_id=>0, :par_parent_id=>3, :blah=>4}
+    fgp.children.first.values.should == {:id=>2, :parent_id=>1, :par_parent_id=>5, :blah=>6}
+  end
+
+  it "should not create the setter method if :read_only option is used" do
+    @c2.many_to_one :parent, :class => @c2, :read_only=>true
+    @c2.instance_methods.collect{|x| x.to_s}.should(include('parent'))
+    @c2.instance_methods.collect{|x| x.to_s}.should_not(include('parent='))
+  end
+
+  it "should not add associations methods directly to class" do
+    @c2.many_to_one :parent, :class => @c2
+    @c2.instance_methods.collect{|x| x.to_s}.should(include('parent'))
+    @c2.instance_methods.collect{|x| x.to_s}.should(include('parent='))
+    @c2.instance_methods(false).collect{|x| x.to_s}.should_not(include('parent'))
+    @c2.instance_methods(false).collect{|x| x.to_s}.should_not(include('parent='))
+  end
+
+  it "should raise an error if trying to set a model object that doesn't have a valid primary key" do
+    @c2.many_to_one :parent, :class => @c2
+    p = @c2.new
+    c = @c2.load(:id=>123)
+    proc{c.parent = p}.should raise_error(Sequel::Error)
+  end
+
+  it "should make the change to the foreign_key value inside a _association= method" do
+    @c2.many_to_one :parent, :class => @c2
+    @c2.private_instance_methods.collect{|x| x.to_s}.sort.should(include("_parent="))
+    p = @c2.new
+    c = @c2.load(:id=>123)
+    def p._parent=(x)
+      @x = x
+    end
+    p.should_not_receive(:parent_id=)
+    p.parent = c
+    p.instance_variable_get(:@x).should == c
+  end
+
+  it "should support (before|after)_(add|remove) callbacks" do
+    h = []
+    @c2.many_to_one :parent, :class => @c2, :before_add=>[proc{|x,y| h << x.pk; h << -y.pk}, :blah], :after_add=>proc{h << 3}, :before_remove=>:blah, :after_remove=>[:blahr]
+    @c2.class_eval do
+      @@blah = h
+      def []=(a, v)
+        a == :parent_id ? (@@blah << (v ? 4 : 5)) : super
+      end
+      def blah(x)
+        @@blah << x.pk
+      end
+      def blahr(x)
+        @@blah << 6
+      end
+    end
+    p = @c2.load(:id=>10)
+    c = @c2.load(:id=>123)
+    h.should == []
+    p.parent = c
+    h.should == [10, -123, 123, 4, 3]
+    p.parent = nil
+    h.should == [10, -123, 123, 4, 3, 123, 5, 6]
+  end
+
+  it "should support after_load association callback" do
+    h = []
+    @c2.many_to_one :parent, :class => @c2, :after_load=>[proc{|x,y| h << [x.pk, y.pk]}, :al]
+    @c2.class_eval do
+      @@blah = h
+      def al(v)
+        @@blah << v.pk
+      end
+      def @dataset.fetch_rows(sql)
+        yield({:id=>20})
+      end
+    end
+    p = @c2.load(:id=>10, :parent_id=>20)
+    parent = p.parent
+    h.should == [[10, 20], 20]
+    parent.pk.should == 20
+  end
+
+  it "should raise error and not call internal add or remove method if before callback returns false, even if raise_on_save_failure is false" do
+    # The reason for this is that assignment in ruby always returns the argument instead of the result
+    # of the method, so we can't return nil to signal that the association callback prevented the modification
+    p = @c2.new
+    c = @c2.load(:id=>123)
+    p.raise_on_save_failure = false
+    @c2.many_to_one :parent, :class => @c2, :before_add=>:ba, :before_remove=>:br
+    p.should_receive(:ba).once.with(c).and_return(false)
+    p.should_not_receive(:_parent=)
+    proc{p.parent = c}.should raise_error(Sequel::Error)
+    p.parent.should == nil
+    p.associations[:parent] = c
+    p.parent.should == c
+    p.should_receive(:br).once.with(c).and_return(false)
+    proc{p.parent = nil}.should raise_error(Sequel::Error)
+  end
+
+  it "should raise an error if a callback is not a proc or symbol" do
+    @c2.many_to_one :parent, :class => @c2, :before_add=>Object.new
+    proc{@c2.new.parent = @c2.load(:id=>1)}.should raise_error(Sequel::Error)
+  end
+
+  it "should call the remove callbacks for the previous object and the add callbacks for the new object" do
+    c = @c2.load(:id=>123)
+    d = @c2.load(:id=>321)
+    p = @c2.new
+    p.associations[:parent] = d
+    h = []
+    @c2.many_to_one :parent, :class => @c2, :before_add=>:ba, :before_remove=>:br, :after_add=>:aa, :after_remove=>:ar
+    @c2.class_eval do
+      @@blah = h
+      def []=(a, v)
+        a == :parent_id ? (@@blah << 5) : super
+      end
+      def ba(x)
+        @@blah << x.pk
+      end
+      def br(x)
+        @@blah << x.pk * -1
+      end
+      def aa(x)
+        @@blah << x.pk * 2
+      end
+      def ar(x)
+        @@blah << x.pk * -2
+      end
+    end
+    p.parent = c
+    h.should == [-321, 123, 5, 246, -642]
+  end
+end
+
+describe Sequel::Model, "one_to_many" do
+
+  before(:each) do
+    MODEL_DB.reset
+
+    @c1 = Class.new(Sequel::Model(:attributes)) do
+      unrestrict_primary_key
+      columns :id, :node_id
+    end
+
+    @c2 = Class.new(Sequel::Model(:nodes)) do
+      unrestrict_primary_key
+      attr_accessor :xxx
+      
+      def self.name; 'Node'; end
+      def self.to_s; 'Node'; end
+      columns :id
+    end
+    @dataset = @c2.dataset
+    
+    @c2.dataset.extend(Module.new {
+      def fetch_rows(sql)
+        @db << sql
+        yield Hash.new
+      end
+    })
+
+    @c1.dataset.extend(Module.new {
+      def fetch_rows(sql)
+        @db << sql
+        yield Hash.new
+      end
+    })
+  end
+
+  it "should use implicit key if omitted" do
+    @c2.one_to_many :attributes, :class => @c1 
+
+    n = @c2.new(:id => 1234)
+    a = n.attributes_dataset
+    a.should be_a_kind_of(Sequel::Dataset)
+    a.sql.should == 'SELECT * FROM attributes WHERE (attributes.node_id = 1234)'
+  end
+  
+  it "should use implicit class if omitted" do
+    class ::HistoricalValue < Sequel::Model
+    end
+    
+    @c2.one_to_many :historical_values
+    
+    n = @c2.new(:id => 1234)
+    v = n.historical_values_dataset
+    v.should be_a_kind_of(Sequel::Dataset)
+    v.sql.should == 'SELECT * FROM historical_values WHERE (historical_values.node_id = 1234)'
+    v.model.should == HistoricalValue
+  end
+  
+  it "should use class inside a module if given as a string" do
+    module ::Historical
+      class Value < Sequel::Model
+      end
+    end
+    
+    @c2.one_to_many :historical_values, :class=>'Historical::Value'
+    
+    n = @c2.new(:id => 1234)
+    v = n.historical_values_dataset
+    v.should be_a_kind_of(Sequel::Dataset)
+    v.sql.should == 'SELECT * FROM values WHERE (values.node_id = 1234)'
+    v.model.should == Historical::Value
+  end
+
+  it "should use explicit key if given" do
+    @c2.one_to_many :attributes, :class => @c1, :key => :nodeid
+    
+    n = @c2.new(:id => 1234)
+    a = n.attributes_dataset
+    a.should be_a_kind_of(Sequel::Dataset)
+    a.sql.should == 'SELECT * FROM attributes WHERE (attributes.nodeid = 1234)'
+  end
+
+  it "should define an add_ method" do
+    @c2.one_to_many :attributes, :class => @c1
+    
+    n = @c2.new(:id => 1234)
+    a = @c1.new(:id => 2345)
+    a.save
+    MODEL_DB.reset
+    a.should == n.add_attribute(a)
+    MODEL_DB.sqls.should == ['UPDATE attributes SET node_id = 1234 WHERE (id = 2345)']
+  end
+
+  it "should define a remove_ method" do
+    @c2.one_to_many :attributes, :class => @c1
+    
+    n = @c2.new(:id => 1234)
+    a = @c1.new(:id => 2345)
+    a.save
+    MODEL_DB.reset
+    a.should == n.remove_attribute(a)
+    MODEL_DB.sqls.should == ['UPDATE attributes SET node_id = NULL WHERE (id = 2345)']
+  end
+  
+  it "should have add_ method respect the :primary_key option" do
+    @c2.one_to_many :attributes, :class => @c1, :primary_key=>:xxx
+    
+    n = @c2.new(:id => 1234, :xxx=>5)
+    a = @c1.new(:id => 2345)
+    a.save
+    MODEL_DB.reset
+    a.should == n.add_attribute(a)
+    MODEL_DB.sqls.should == ['UPDATE attributes SET node_id = 5 WHERE (id = 2345)']
+  end
+
+  
+  it "should raise an error in add_ and remove_ if the passed object returns false to save (is not valid)" do
+    @c2.one_to_many :attributes, :class => @c1
+    n = @c2.new(:id => 1234)
+    a = @c1.new(:id => 2345)
+    def a.valid?; false; end
+    proc{n.add_attribute(a)}.should raise_error(Sequel::Error)
+    proc{n.remove_attribute(a)}.should raise_error(Sequel::Error)
+  end
+
+  it "should raise an error if the model object doesn't have a valid primary key" do
+    @c2.one_to_many :attributes, :class => @c1 
+    a = @c2.new
+    n = @c1.load(:id=>123)
+    proc{a.attributes_dataset}.should raise_error(Sequel::Error)
+    proc{a.attributes}.should raise_error(Sequel::Error)
+    proc{a.add_attribute(n)}.should raise_error(Sequel::Error)
+    proc{a.remove_attribute(n)}.should raise_error(Sequel::Error)
+    proc{a.remove_all_attributes}.should raise_error(Sequel::Error)
+  end
+  
+  it "should use :primary_key option if given" do
+    @c1.one_to_many :nodes, :class => @c2, :primary_key => :node_id, :key=>:id
+    n = @c1.load(:id => 1234, :node_id=>4321)
+    n.nodes_dataset.sql.should == "SELECT * FROM nodes WHERE (nodes.id = 4321)"
+  end
+
+  it "should support a select option" do
+    @c2.one_to_many :attributes, :class => @c1, :select => [:id, :name]
+
+    n = @c2.new(:id => 1234)
+    n.attributes_dataset.sql.should == "SELECT id, name FROM attributes WHERE (attributes.node_id = 1234)"
+  end
+  
+  it "should support a conditions option" do
+    @c2.one_to_many :attributes, :class => @c1, :conditions => {:a=>32}
+    n = @c2.new(:id => 1234)
+    n.attributes_dataset.sql.should == "SELECT * FROM attributes WHERE ((attributes.node_id = 1234) AND (a = 32))"
+    @c2.one_to_many :attributes, :class => @c1, :conditions => ~:a
+    n = @c2.new(:id => 1234)
+    n.attributes_dataset.sql.should == "SELECT * FROM attributes WHERE ((attributes.node_id = 1234) AND NOT a)"
+  end
+  
+  it "should support an order option" do
+    @c2.one_to_many :attributes, :class => @c1, :order => :kind
+
+    n = @c2.new(:id => 1234)
+    n.attributes_dataset.sql.should == "SELECT * FROM attributes WHERE (attributes.node_id = 1234) ORDER BY kind"
+  end
+  
+  it "should support an array for the order option" do
+    @c2.one_to_many :attributes, :class => @c1, :order => [:kind1, :kind2]
+
+    n = @c2.new(:id => 1234)
+    n.attributes_dataset.sql.should == "SELECT * FROM attributes WHERE (attributes.node_id = 1234) ORDER BY kind1, kind2"
+  end
+  
+  it "should return array with all members of the association" do
+    @c2.one_to_many :attributes, :class => @c1
+    
+    n = @c2.new(:id => 1234)
+    atts = n.attributes
+    atts.should be_a_kind_of(Array)
+    atts.size.should == 1
+    atts.first.should be_a_kind_of(@c1)
+    atts.first.values.should == {}
+    
+    MODEL_DB.sqls.should == ['SELECT * FROM attributes WHERE (attributes.node_id = 1234)']
+  end
+  
+  it "should accept a block" do
+    @c2.one_to_many :attributes, :class => @c1 do |ds|
+      ds.filter(:xxx => @xxx)
+    end
+    
+    n = @c2.new(:id => 1234)
+    atts = n.attributes
+    atts.should be_a_kind_of(Array)
+    atts.size.should == 1
+    atts.first.should be_a_kind_of(@c1)
+    atts.first.values.should == {}
+    
+    MODEL_DB.sqls.should == ['SELECT * FROM attributes WHERE ((attributes.node_id = 1234) AND (xxx IS NULL))']
+  end
+  
+  it "should support :order option with block" do
+    @c2.one_to_many :attributes, :class => @c1, :order => :kind do |ds|
+      ds.filter(:xxx => @xxx)
+    end
+    
+    n = @c2.new(:id => 1234)
+    atts = n.attributes
+    atts.should be_a_kind_of(Array)
+    atts.size.should == 1
+    atts.first.should be_a_kind_of(@c1)
+    atts.first.values.should == {}
+    
+    MODEL_DB.sqls.should == ['SELECT * FROM attributes WHERE ((attributes.node_id = 1234) AND (xxx IS NULL)) ORDER BY kind']
+  end
+  
+  it "should have the block argument affect the _dataset method" do
+    @c2.one_to_many :attributes, :class => @c1 do |ds|
+      ds.filter(:xxx => 456)
+    end
+    @c2.new(:id => 1234).attributes_dataset.sql.should == 'SELECT * FROM attributes WHERE ((attributes.node_id = 1234) AND (xxx = 456))'
+  end
+  
+  it "should support a :dataset option that is used instead of the default" do
+    c1 = @c1
+    @c2.one_to_many :all_other_attributes, :class => @c1, :dataset=>proc{c1.filter(:nodeid=>pk).invert}, :order=>:a, :limit=>10 do |ds|
+      ds.filter(:xxx => 5)
+    end
+    
+    @c2.new(:id => 1234).all_other_attributes_dataset.sql.should == 'SELECT * FROM attributes WHERE ((nodeid != 1234) AND (xxx = 5)) ORDER BY a LIMIT 10'
+    n = @c2.new(:id => 1234)
+    atts = n.all_other_attributes
+    atts.should be_a_kind_of(Array)
+    atts.size.should == 1
+    atts.first.should be_a_kind_of(@c1)
+    atts.first.values.should == {}
+    
+    MODEL_DB.sqls.should == ['SELECT * FROM attributes WHERE ((nodeid != 1234) AND (xxx = 5)) ORDER BY a LIMIT 10']
+  end
+  
+  it "should support a :limit option" do
+    @c2.one_to_many :attributes, :class => @c1 , :limit=>10
+    @c2.new(:id => 1234).attributes_dataset.sql.should == 'SELECT * FROM attributes WHERE (attributes.node_id = 1234) LIMIT 10'
+    @c2.one_to_many :attributes, :class => @c1 , :limit=>[10,10]
+    @c2.new(:id => 1234).attributes_dataset.sql.should == 'SELECT * FROM attributes WHERE (attributes.node_id = 1234) LIMIT 10 OFFSET 10'
+  end
+
+  it "should have the :eager option affect the _dataset method" do
+    @c2.one_to_many :attributes, :class => @c2 , :eager=>:attributes
+    @c2.new(:id => 1234).attributes_dataset.opts[:eager].should == {:attributes=>nil}
+  end
+  
+  it "should set cached instance variable when accessed" do
+    @c2.one_to_many :attributes, :class => @c1
+
+    n = @c2.new(:id => 1234)
+    MODEL_DB.reset
+    n.associations.include?(:attributes).should == false
+    atts = n.attributes
+    atts.should == n.associations[:attributes]
+    MODEL_DB.sqls.should == ['SELECT * FROM attributes WHERE (attributes.node_id = 1234)']
+  end
+
+  it "should use cached instance variable if available" do
+    @c2.one_to_many :attributes, :class => @c1
+
+    n = @c2.new(:id => 1234)
+    MODEL_DB.reset
+    n.associations[:attributes] = 42
+    n.attributes.should == 42
+    MODEL_DB.sqls.should == []
+  end
+
+  it "should not use cached instance variable if asked to reload" do
+    @c2.one_to_many :attributes, :class => @c1
+
+    n = @c2.new(:id => 1234)
+    MODEL_DB.reset
+    n.associations[:attributes] = 42
+    n.attributes(true).should_not == 42
+    MODEL_DB.sqls.should == ['SELECT * FROM attributes WHERE (attributes.node_id = 1234)']
+  end
+
+  it "should add item to cached instance variable if it exists when calling add_" do
+    @c2.one_to_many :attributes, :class => @c1
+
+    n = @c2.new(:id => 1234)
+    att = @c1.new(:id => 345)
+    MODEL_DB.reset
+    a = []
+    n.associations[:attributes] = a
+    n.add_attribute(att)
+    a.should == [att]
+  end
+
+  it "should set object to item's reciprocal cached association variable when calling add_" do
+    @c2.one_to_many :attributes, :class => @c1
+    @c1.many_to_one :node, :class => @c2
+
+    n = @c2.new(:id => 1234)
+    att = @c1.new(:id => 345)
+    n.add_attribute(att)
+    att.node.should == n
+  end
+
+  it "should remove item from cached instance variable if it exists when calling remove_" do
+    @c2.one_to_many :attributes, :class => @c1
+
+    n = @c2.load(:id => 1234)
+    att = @c1.load(:id => 345)
+    MODEL_DB.reset
+    a = [att]
+    n.associations[:attributes] = a
+    n.remove_attribute(att)
+    a.should == []
+  end
+
+  it "should remove item's reciprocal cached association variable when calling remove_" do
+    @c2.one_to_many :attributes, :class => @c1
+    @c1.many_to_one :node, :class => @c2
+
+    n = @c2.new(:id => 1234)
+    att = @c1.new(:id => 345)
+    att.associations[:node] = n
+    att.node.should == n
+    n.remove_attribute(att)
+    att.node.should == nil
+  end
+
+  it "should not create the add_, remove_, or remove_all_ methods if :read_only option is used" do
+    @c2.one_to_many :attributes, :class => @c1, :read_only=>true
+    im = @c2.instance_methods.collect{|x| x.to_s}
+    im.should(include('attributes'))
+    im.should(include('attributes_dataset'))
+    im.should_not(include('add_attribute'))
+    im.should_not(include('remove_attribute'))
+    im.should_not(include('remove_all_attributes'))
+  end
+
+  it "should not add associations methods directly to class" do
+    @c2.one_to_many :attributes, :class => @c1
+    im = @c2.instance_methods.collect{|x| x.to_s}
+    im.should(include('attributes'))
+    im.should(include('attributes_dataset'))
+    im.should(include('add_attribute'))
+    im.should(include('remove_attribute'))
+    im.should(include('remove_all_attributes'))
+    im2 = @c2.instance_methods(false).collect{|x| x.to_s}
+    im2.should_not(include('attributes'))
+    im2.should_not(include('attributes_dataset'))
+    im2.should_not(include('add_attribute'))
+    im2.should_not(include('remove_attribute'))
+    im2.should_not(include('remove_all_attributes'))
+  end
+
+  it "should populate the reciprocal many_to_one instance variable when loading the one_to_many association" do
+    @c2.one_to_many :attributes, :class => @c1, :key => :node_id
+    @c1.many_to_one :node, :class => @c2, :key => :node_id
+    
+    n = @c2.new(:id => 1234)
+    atts = n.attributes
+    MODEL_DB.sqls.should == ['SELECT * FROM attributes WHERE (attributes.node_id = 1234)']
+    atts.should be_a_kind_of(Array)
+    atts.size.should == 1
+    atts.first.should be_a_kind_of(@c1)
+    atts.first.values.should == {}
+    atts.first.node.should == n
+    
+    MODEL_DB.sqls.length.should == 1
+  end
+  
+  it "should use an explicit reciprocal instance variable if given" do
+    @c2.one_to_many :attributes, :class => @c1, :key => :node_id, :reciprocal=>:wxyz
+    
+    n = @c2.new(:id => 1234)
+    atts = n.attributes
+    MODEL_DB.sqls.should == ['SELECT * FROM attributes WHERE (attributes.node_id = 1234)']
+    atts.should be_a_kind_of(Array)
+    atts.size.should == 1
+    atts.first.should be_a_kind_of(@c1)
+    atts.first.values.should == {}
+    atts.first.associations[:wxyz].should == n
+    
+    MODEL_DB.sqls.length.should == 1
+  end
+  
+  it "should have an remove_all_ method that removes all associations" do
+    @c2.one_to_many :attributes, :class => @c1
+    @c2.new(:id => 1234).remove_all_attributes
+    MODEL_DB.sqls.first.should == 'UPDATE attributes SET node_id = NULL WHERE (node_id = 1234)'
+  end
+
+  it "should have the remove_all_ method respect the :primary_key option" do
+    @c2.one_to_many :attributes, :class => @c1, :primary_key=>:xxx
+    @c2.new(:id => 1234, :xxx=>5).remove_all_attributes
+    MODEL_DB.sqls.first.should == 'UPDATE attributes SET node_id = NULL WHERE (node_id = 5)'
+  end
+
+  it "remove_all should set the cached instance variable to []" do
+    @c2.one_to_many :attributes, :class => @c1
+    node = @c2.new(:id => 1234)
+    node.remove_all_attributes
+    node.associations[:attributes].should == []
+  end
+
+  it "remove_all should return the array of previously associated items if the cached instance variable exists" do
+    @c2.one_to_many :attributes, :class => @c1
+    attrib = @c1.new(:id=>3)
+    node = @c2.new(:id => 1234)
+    d = @c1.dataset
+    def d.fetch_rows(s); end
+    node.attributes.should == []
+    def attrib.save; self end
+    node.add_attribute(attrib)
+    node.associations[:attributes].should == [attrib]
+    node.remove_all_attributes.should == [attrib]
+  end
+
+  it "remove_all should return nil if the cached instance variable does not exist" do
+    @c2.one_to_many :attributes, :class => @c1
+    @c2.new(:id => 1234).remove_all_attributes.should == nil
+  end
+
+  it "remove_all should remove the current item from all reciprocal instance varaibles if it cached instance variable exists" do
+    @c2.one_to_many :attributes, :class => @c1
+    @c1.many_to_one :node, :class => @c2
+    d = @c1.dataset
+    def d.fetch_rows(s); end
+    d = @c2.dataset
+    def d.fetch_rows(s); end
+    attrib = @c1.new(:id=>3)
+    node = @c2.new(:id => 1234)
+    node.attributes.should == []
+    attrib.node.should == nil
+    def attrib.save; self end
+    node.add_attribute(attrib)
+    attrib.associations[:node].should == node 
+    node.remove_all_attributes
+    attrib.associations.fetch(:node, 2).should == nil
+  end
+
+  it "should add a getter method if the :one_to_one option is true" do
+    @c2.one_to_many :attributes, :class => @c1, :one_to_one=>true
+    att = @c2.new(:id => 1234).attribute
+    MODEL_DB.sqls.should == ['SELECT * FROM attributes WHERE (attributes.node_id = 1234)']
+    att.should be_a_kind_of(@c1)
+    att.values.should == {}
+  end
+
+  it "should not add a setter method if the :one_to_one option is true and :read_only option is true" do
+    @c2.one_to_many :attributes, :class => @c1, :one_to_one=>true, :read_only=>true
+    im = @c2.instance_methods.collect{|x| x.to_s}
+    im.should(include('attribute'))
+    im.should_not(include('attribute='))
+  end
+
+  it "should have the getter method raise an error if more than one record is found" do
+    @c2.one_to_many :attributes, :class => @c1, :one_to_one=>true
+    d = @c1.dataset
+    def d.fetch_rows(s); 2.times{yield Hash.new} end
+    proc{@c2.new(:id => 1234).attribute}.should raise_error(Sequel::Error)
+  end
+
+  it "should add a setter method if the :one_to_one option is true" do
+    @c2.one_to_many :attributes, :class => @c1, :one_to_one=>true
+    attrib = @c1.new(:id=>3)
+    d = @c1.dataset
+    def d.fetch_rows(s); yield({:id=>3}) end
+    @c2.new(:id => 1234).attribute = attrib
+    ['INSERT INTO attributes (node_id, id) VALUES (1234, 3)',
+      'INSERT INTO attributes (id, node_id) VALUES (3, 1234)'].should(include(MODEL_DB.sqls.first))
+    MODEL_DB.sqls.last.should == 'UPDATE attributes SET node_id = NULL WHERE ((node_id = 1234) AND (id != 3))'
+    MODEL_DB.sqls.length.should == 2
+    @c2.new(:id => 1234).attribute.should == attrib
+    MODEL_DB.sqls.clear
+    attrib = @c1.load(:id=>3)
+    @c2.new(:id => 1234).attribute = attrib
+    MODEL_DB.sqls.length.should == 2
+    MODEL_DB.sqls.first.should =~ /UPDATE attributes SET (node_id = 1234, id = 3|id = 3, node_id = 1234) WHERE \(id = 3\)/
+    MODEL_DB.sqls.last.should == 'UPDATE attributes SET node_id = NULL WHERE ((node_id = 1234) AND (id != 3))'
+  end
+
+  it "should use a transaction in the setter method if the :one_to_one option is true" do
+    @c2.one_to_many :attributes, :class => @c1, :one_to_one=>true
+    @c2.use_transactions = true
+    MODEL_DB.sqls.clear
+    attrib = @c1.load(:id=>3)
+    @c2.new(:id => 1234).attribute = attrib
+    MODEL_DB.sqls.length.should == 4
+    MODEL_DB.sqls.first.should == 'BEGIN'
+    MODEL_DB.sqls[1].should =~ /UPDATE attributes SET (node_id = 1234, id = 3|id = 3, node_id = 1234) WHERE \(id = 3\)/
+    MODEL_DB.sqls[2].should == 'UPDATE attributes SET node_id = NULL WHERE ((node_id = 1234) AND (id != 3))'
+    MODEL_DB.sqls.last.should == 'COMMIT'
+  end
+
+  it "should have the setter method for the :one_to_one option respect the :primary_key option" do
+    @c2.one_to_many :attributes, :class => @c1, :one_to_one=>true, :primary_key=>:xxx
+    attrib = @c1.new(:id=>3)
+    d = @c1.dataset
+    def d.fetch_rows(s); yield({:id=>3}) end
+    @c2.new(:id => 1234, :xxx=>5).attribute = attrib
+    ['INSERT INTO attributes (node_id, id) VALUES (5, 3)',
+      'INSERT INTO attributes (id, node_id) VALUES (3, 5)'].should(include(MODEL_DB.sqls.first))
+    MODEL_DB.sqls.last.should == 'UPDATE attributes SET node_id = NULL WHERE ((node_id = 5) AND (id != 3))'
+    MODEL_DB.sqls.length.should == 2
+    @c2.new(:id => 321, :xxx=>5).attribute.should == attrib
+    MODEL_DB.sqls.clear
+    attrib = @c1.load(:id=>3)
+    @c2.new(:id => 621, :xxx=>5).attribute = attrib
+    MODEL_DB.sqls.length.should == 2
+    MODEL_DB.sqls.first.should =~ /UPDATE attributes SET (node_id = 5, id = 3|id = 3, node_id = 5) WHERE \(id = 3\)/
+    MODEL_DB.sqls.last.should == 'UPDATE attributes SET node_id = NULL WHERE ((node_id = 5) AND (id != 3))'
+  end
+
+  it "should raise an error if the one_to_one getter would be the same as the association name" do
+    proc{@c2.one_to_many :song, :class => @c1, :one_to_one=>true}.should raise_error(Sequel::Error)
+  end
+
+  it "should not create remove_ and remove_all methods if :one_to_one option is used" do
+    @c2.one_to_many :attributes, :class => @c1, :one_to_one=>true
+    @c2.new.should_not(respond_to(:remove_attribute))
+    @c2.new.should_not(respond_to(:remove_all_attributes))
+  end
+
+  it "should make non getter and setter methods private if :one_to_one option is used" do 
+    @c2.one_to_many :attributes, :class => @c1, :one_to_one=>true do |ds| end
+    meths = @c2.private_instance_methods.collect{|x| x.to_s}
+    meths.should(include("attributes"))
+    meths.should(include("add_attribute"))
+    meths.should(include("attributes_dataset"))
+  end
+
+  it "should call an _add_ method internally to add attributes" do
+    @c2.one_to_many :attributes, :class => @c1
+    @c2.private_instance_methods.collect{|x| x.to_s}.sort.should(include("_add_attribute"))
+    p = @c2.load(:id=>10)
+    c = @c1.load(:id=>123)
+    def p._add_attribute(x)
+      @x = x
+    end
+    c.should_not_receive(:node_id=)
+    p.add_attribute(c)
+    p.instance_variable_get(:@x).should == c
+  end
+
+  it "should call a _remove_ method internally to remove attributes" do
+    @c2.one_to_many :attributes, :class => @c1
+    @c2.private_instance_methods.collect{|x| x.to_s}.sort.should(include("_remove_attribute"))
+    p = @c2.load(:id=>10)
+    c = @c1.load(:id=>123)
+    def p._remove_attribute(x)
+      @x = x
+    end
+    c.should_not_receive(:node_id=)
+    p.remove_attribute(c)
+    p.instance_variable_get(:@x).should == c
+  end
+
+  it "should support (before|after)_(add|remove) callbacks" do
+    h = []
+    @c2.one_to_many :attributes, :class => @c1, :before_add=>[proc{|x,y| h << x.pk; h << -y.pk}, :blah], :after_add=>proc{h << 3}, :before_remove=>:blah, :after_remove=>[:blahr]
+    @c2.class_eval do
+      @@blah = h
+      def _add_attribute(v)
+        @@blah << 4
+      end
+      def _remove_attribute(v)
+        @@blah << 5
+      end
+      def blah(x)
+        @@blah << x.pk
+      end
+      def blahr(x)
+        @@blah << 6
+      end
+    end
+    p = @c2.load(:id=>10)
+    c = @c1.load(:id=>123)
+    h.should == []
+    p.add_attribute(c)
+    h.should == [10, -123, 123, 4, 3]
+    p.remove_attribute(c)
+    h.should == [10, -123, 123, 4, 3, 123, 5, 6]
+  end
+
+  it "should support after_load association callback" do
+    h = []
+    @c2.one_to_many :attributes, :class => @c1, :after_load=>[proc{|x,y| h << [x.pk, y.collect{|z|z.pk}]}, :al]
+    @c2.class_eval do
+      @@blah = h
+      def al(v)
+        v.each{|x| @@blah << x.pk}
+      end
+    end
+    @c1.class_eval do
+      def @dataset.fetch_rows(sql)
+        yield({:id=>20})
+        yield({:id=>30})
+      end
+    end
+    p = @c2.load(:id=>10, :parent_id=>20)
+    attributes = p.attributes
+    h.should == [[10, [20, 30]], 20, 30]
+    attributes.collect{|a| a.pk}.should == [20, 30]
+  end
+
+  it "should raise error and not call internal add or remove method if before callback returns false if raise_on_save_failure is true" do
+    p = @c2.load(:id=>10)
+    c = @c1.load(:id=>123)
+    @c2.one_to_many :attributes, :class => @c1, :before_add=>:ba, :before_remove=>:br
+    p.should_receive(:ba).once.with(c).and_return(false)
+    p.should_not_receive(:_add_attribute)
+    p.should_not_receive(:_remove_attribute)
+    p.associations[:attributes] = []
+    proc{p.add_attribute(c)}.should raise_error(Sequel::Error)
+    p.attributes.should == []
+    p.associations[:attributes] = [c]
+    p.should_receive(:br).once.with(c).and_return(false)
+    proc{p.remove_attribute(c)}.should raise_error(Sequel::Error)
+    p.attributes.should == [c]
+  end
+
+  it "should return nil and not call internal add or remove method if before callback returns false if raise_on_save_failure is false" do
+    p = @c2.load(:id=>10)
+    c = @c1.load(:id=>123)
+    p.raise_on_save_failure = false
+    @c2.one_to_many :attributes, :class => @c1, :before_add=>:ba, :before_remove=>:br
+    p.should_receive(:ba).once.with(c).and_return(false)
+    p.should_not_receive(:_add_attribute)
+    p.should_not_receive(:_remove_attribute)
+    p.associations[:attributes] = []
+    p.add_attribute(c).should == nil
+    p.attributes.should == []
+    p.associations[:attributes] = [c]
+    p.should_receive(:br).once.with(c).and_return(false)
+    p.remove_attribute(c).should == nil
+    p.attributes.should == [c]
+  end
+end
+
+describe Sequel::Model, "many_to_many" do
+
+  before(:each) do
+    MODEL_DB.reset
+
+    @c1 = Class.new(Sequel::Model(:attributes)) do
+      unrestrict_primary_key
+      attr_accessor :yyy
+      def self.name; 'Attribute'; end
+      def self.to_s; 'Attribute'; end
+      columns :id
+    end
+
+    @c2 = Class.new(Sequel::Model(:nodes)) do
+      unrestrict_primary_key
+      attr_accessor :xxx
+      
+      def self.name; 'Node'; end
+      def self.to_s; 'Node'; end
+      columns :id
+    end
+    @dataset = @c2.dataset
+
+    [@c1, @c2].each do |c|
+      c.dataset.extend(Module.new {
+        def fetch_rows(sql)
+          @db << sql
+          yield Hash.new
+        end
+      })
+    end
+  end
+
+  it "should use implicit key values and join table if omitted" do
+    @c2.many_to_many :attributes, :class => @c1 
+
+    n = @c2.new(:id => 1234)
+    a = n.attributes_dataset
+    a.should be_a_kind_of(Sequel::Dataset)
+    a.sql.should == 'SELECT attributes.* FROM attributes INNER JOIN attributes_nodes ON ((attributes_nodes.attribute_id = attributes.id) AND (attributes_nodes.node_id = 1234))'
+  end
+  
+  it "should use implicit class if omitted" do
+    class ::Tag < Sequel::Model
+    end
+    
+    @c2.many_to_many :tags
+
+    n = @c2.new(:id => 1234)
+    a = n.tags_dataset
+    a.should be_a_kind_of(Sequel::Dataset)
+    a.sql.should == 'SELECT tags.* FROM tags INNER JOIN nodes_tags ON ((nodes_tags.tag_id = tags.id) AND (nodes_tags.node_id = 1234))'
+  end
+  
+  it "should use class inside module if given as a string" do
+    module ::Historical
+      class Tag < Sequel::Model
+      end
+    end
+    
+    @c2.many_to_many :tags, :class=>'::Historical::Tag'
+
+    n = @c2.new(:id => 1234)
+    a = n.tags_dataset
+    a.should be_a_kind_of(Sequel::Dataset)
+    a.sql.should == 'SELECT tags.* FROM tags INNER JOIN nodes_tags ON ((nodes_tags.tag_id = tags.id) AND (nodes_tags.node_id = 1234))'
+  end
+  
+  it "should use explicit key values and join table if given" do
+    @c2.many_to_many :attributes, :class => @c1, :left_key => :nodeid, :right_key => :attributeid, :join_table => :attribute2node
+    
+    n = @c2.new(:id => 1234)
+    a = n.attributes_dataset
+    a.should be_a_kind_of(Sequel::Dataset)
+    a.sql.should == 'SELECT attributes.* FROM attributes INNER JOIN attribute2node ON ((attribute2node.attributeid = attributes.id) AND (attribute2node.nodeid = 1234))'
+  end
+  
+  it "should support a conditions option" do
+    @c2.many_to_many :attributes, :class => @c1, :conditions => {:a=>32}
+    n = @c2.new(:id => 1234)
+    a = n.attributes_dataset
+    a.should be_a_kind_of(Sequel::Dataset)
+    a.sql.should == 'SELECT attributes.* FROM attributes INNER JOIN attributes_nodes ON ((attributes_nodes.attribute_id = attributes.id) AND (attributes_nodes.node_id = 1234)) WHERE (a = 32)'
+    @c2.many_to_many :attributes, :class => @c1, :conditions => ['a = ?', 32]
+    n = @c2.new(:id => 1234)
+    a = n.attributes_dataset
+    a.should be_a_kind_of(Sequel::Dataset)
+    a.sql.should == 'SELECT attributes.* FROM attributes INNER JOIN attributes_nodes ON ((attributes_nodes.attribute_id = attributes.id) AND (attributes_nodes.node_id = 1234)) WHERE (a = 32)'
+  end
+  
+  it "should support an order option" do
+    @c2.many_to_many :attributes, :class => @c1, :order => :blah
+
+    n = @c2.new(:id => 1234)
+    a = n.attributes_dataset
+    a.should be_a_kind_of(Sequel::Dataset)
+    a.sql.should == 'SELECT attributes.* FROM attributes INNER JOIN attributes_nodes ON ((attributes_nodes.attribute_id = attributes.id) AND (attributes_nodes.node_id = 1234)) ORDER BY blah'
+  end
+  
+  it "should support an array for the order option" do
+    @c2.many_to_many :attributes, :class => @c1, :order => [:blah1, :blah2]
+
+    n = @c2.new(:id => 1234)
+    a = n.attributes_dataset
+    a.should be_a_kind_of(Sequel::Dataset)
+    a.sql.should == 'SELECT attributes.* FROM attributes INNER JOIN attributes_nodes ON ((attributes_nodes.attribute_id = attributes.id) AND (attributes_nodes.node_id = 1234)) ORDER BY blah1, blah2'
+  end
+  
+  it "should support :left_primary_key and :right_primary_key options" do
+    @c2.many_to_many :attributes, :class => @c1, :left_primary_key=>:xxx, :right_primary_key=>:yyy
+    @c2.new(:id => 1234, :xxx=>5).attributes_dataset.sql.should == 'SELECT attributes.* FROM attributes INNER JOIN attributes_nodes ON ((attributes_nodes.attribute_id = attributes.yyy) AND (attributes_nodes.node_id = 5))'
+  end
+  
+  it "should support a select option" do
+    @c2.many_to_many :attributes, :class => @c1, :select => :blah
+
+    n = @c2.new(:id => 1234)
+    a = n.attributes_dataset
+    a.should be_a_kind_of(Sequel::Dataset)
+    a.sql.should == 'SELECT blah FROM attributes INNER JOIN attributes_nodes ON ((attributes_nodes.attribute_id = attributes.id) AND (attributes_nodes.node_id = 1234))'
+  end
+  
+  it "should support an array for the select option" do
+    @c2.many_to_many :attributes, :class => @c1, :select => [:attributes.*, :attribute_nodes__blah2]
+
+    n = @c2.new(:id => 1234)
+    a = n.attributes_dataset
+    a.should be_a_kind_of(Sequel::Dataset)
+    a.sql.should == 'SELECT attributes.*, attribute_nodes.blah2 FROM attributes INNER JOIN attributes_nodes ON ((attributes_nodes.attribute_id = attributes.id) AND (attributes_nodes.node_id = 1234))'
+  end
+  
+  it "should accept a block" do
+    @c2.many_to_many :attributes, :class => @c1 do |ds|
+      ds.filter(:xxx => @xxx)
+    end
+
+    n = @c2.new(:id => 1234)
+    n.xxx = 555
+    a = n.attributes
+    a.should be_a_kind_of(Array)
+    a.size.should == 1
+    a.first.should be_a_kind_of(@c1)
+    MODEL_DB.sqls.first.should == 'SELECT attributes.* FROM attributes INNER JOIN attributes_nodes ON ((attributes_nodes.attribute_id = attributes.id) AND (attributes_nodes.node_id = 1234)) WHERE (xxx = 555)'
+  end
+
+  it "should allow the :order option while accepting a block" do
+    @c2.many_to_many :attributes, :class => @c1, :order=>[:blah1, :blah2] do |ds|
+      ds.filter(:xxx => @xxx)
+    end
+
+    n = @c2.new(:id => 1234)
+    n.xxx = 555
+    a = n.attributes
+    a.should be_a_kind_of(Array)
+    a.size.should == 1
+    a.first.should be_a_kind_of(@c1)
+    MODEL_DB.sqls.first.should == 'SELECT attributes.* FROM attributes INNER JOIN attributes_nodes ON ((attributes_nodes.attribute_id = attributes.id) AND (attributes_nodes.node_id = 1234)) WHERE (xxx = 555) ORDER BY blah1, blah2'
+  end
+
+  it "should have the block argument affect the _dataset method" do
+    @c2.many_to_many :attributes, :class => @c1 do |ds|
+      ds.filter(:xxx => 456)
+    end
+    @c2.new(:id => 1234).attributes_dataset.sql.should == 'SELECT attributes.* FROM attributes INNER JOIN attributes_nodes ON ((attributes_nodes.attribute_id = attributes.id) AND (attributes_nodes.node_id = 1234)) WHERE (xxx = 456)'
+  end
+  
+  it "should support a :dataset option that is used instead of the default" do
+    c1 = @c1
+    @c2.many_to_many :attributes, :class => @c1, :dataset=>proc{c1.join_table(:natural, :an).filter(:an__nodeid=>pk)}, :order=> :a, :limit=>10, :select=>nil do |ds|
+      ds.filter(:xxx => @xxx)
+    end
+
+    n = @c2.new(:id => 1234)
+    n.xxx = 555
+    n.attributes_dataset.sql.should == 'SELECT * FROM attributes NATURAL JOIN an WHERE ((an.nodeid = 1234) AND (xxx = 555)) ORDER BY a LIMIT 10'
+    a = n.attributes
+    a.should be_a_kind_of(Array)
+    a.size.should == 1
+    a.first.should be_a_kind_of(@c1)
+    MODEL_DB.sqls.first.should == 'SELECT * FROM attributes NATURAL JOIN an WHERE ((an.nodeid = 1234) AND (xxx = 555)) ORDER BY a LIMIT 10'
+  end
+
+  it "should support a :limit option" do
+    @c2.many_to_many :attributes, :class => @c1 , :limit=>10
+    @c2.new(:id => 1234).attributes_dataset.sql.should == 'SELECT attributes.* FROM attributes INNER JOIN attributes_nodes ON ((attributes_nodes.attribute_id = attributes.id) AND (attributes_nodes.node_id = 1234)) LIMIT 10'
+    @c2.many_to_many :attributes, :class => @c1 , :limit=>[10, 10]
+    @c2.new(:id => 1234).attributes_dataset.sql.should == 'SELECT attributes.* FROM attributes INNER JOIN attributes_nodes ON ((attributes_nodes.attribute_id = attributes.id) AND (attributes_nodes.node_id = 1234)) LIMIT 10 OFFSET 10'
+  end
+
+  it "should have the :eager option affect the _dataset method" do
+    @c2.many_to_many :attributes, :class => @c2 , :eager=>:attributes
+    @c2.new(:id => 1234).attributes_dataset.opts[:eager].should == {:attributes=>nil}
+  end
+  
+  it "should define an add_ method" do
+    @c2.many_to_many :attributes, :class => @c1
+    
+    n = @c2.new(:id => 1234)
+    a = @c1.new(:id => 2345)
+    a.should == n.add_attribute(a)
+    ['INSERT INTO attributes_nodes (node_id, attribute_id) VALUES (1234, 2345)',
+     'INSERT INTO attributes_nodes (attribute_id, node_id) VALUES (2345, 1234)'
+    ].should(include(MODEL_DB.sqls.first))
+  end
+
+  it "should define a remove_ method" do
+    @c2.many_to_many :attributes, :class => @c1
+    
+    n = @c2.new(:id => 1234)
+    a = @c1.new(:id => 2345)
+    a.should == n.remove_attribute(a)
+    MODEL_DB.sqls.first.should == 'DELETE FROM attributes_nodes WHERE ((node_id = 1234) AND (attribute_id = 2345))'
+  end
+
+  it "should have the add_ method respect the :left_primary_key and :right_primary_key options" do
+    @c2.many_to_many :attributes, :class => @c1, :left_primary_key=>:xxx, :right_primary_key=>:yyy
+    
+    n = @c2.new(:id => 1234, :xxx=>5)
+    a = @c1.new(:id => 2345, :yyy=>8)
+    a.should == n.add_attribute(a)
+    ['INSERT INTO attributes_nodes (node_id, attribute_id) VALUES (5, 8)',
+     'INSERT INTO attributes_nodes (attribute_id, node_id) VALUES (8, 5)'
+    ].should(include(MODEL_DB.sqls.first))
+  end
+
+  it "should have the remove_ method respect the :left_primary_key and :right_primary_key options" do
+    @c2.many_to_many :attributes, :class => @c1, :left_primary_key=>:xxx, :right_primary_key=>:yyy
+    
+    n = @c2.new(:id => 1234, :xxx=>5)
+    a = @c1.new(:id => 2345, :yyy=>8)
+    a.should == n.remove_attribute(a)
+    MODEL_DB.sqls.first.should == 'DELETE FROM attributes_nodes WHERE ((node_id = 5) AND (attribute_id = 8))'
+  end
+
+  it "should raise an error if the model object doesn't have a valid primary key" do
+    @c2.many_to_many :attributes, :class => @c1 
+    a = @c2.new
+    n = @c1.load(:id=>123)
+    proc{a.attributes_dataset}.should raise_error(Sequel::Error)
+    proc{a.attributes}.should raise_error(Sequel::Error)
+    proc{a.add_attribute(n)}.should raise_error(Sequel::Error)
+    proc{a.remove_attribute(n)}.should raise_error(Sequel::Error)
+    proc{a.remove_all_attributes}.should raise_error(Sequel::Error)
+  end
+
+  it "should raise an error if trying to add/remove a model object that doesn't have a valid primary key" do
+    @c2.many_to_many :attributes, :class => @c1 
+    n = @c1.new
+    a = @c2.load(:id=>123)
+    proc{a.add_attribute(n)}.should raise_error(Sequel::Error)
+    proc{a.remove_attribute(n)}.should raise_error(Sequel::Error)
+  end
+
+  it "should provide an array with all members of the association" do
+    @c2.many_to_many :attributes, :class => @c1
+    
+    n = @c2.new(:id => 1234)
+    atts = n.attributes
+    atts.should be_a_kind_of(Array)
+    atts.size.should == 1
+    atts.first.should be_a_kind_of(@c1)
+
+    MODEL_DB.sqls.first.should == 'SELECT attributes.* FROM attributes INNER JOIN attributes_nodes ON ((attributes_nodes.attribute_id = attributes.id) AND (attributes_nodes.node_id = 1234))'
+  end
+
+  it "should set cached instance variable when accessed" do
+    @c2.many_to_many :attributes, :class => @c1
+
+    n = @c2.new(:id => 1234)
+    MODEL_DB.reset
+    n.associations.include?(:attributes).should == false
+    atts = n.attributes
+    atts.should == n.associations[:attributes]
+    MODEL_DB.sqls.length.should == 1
+  end
+
+  it "should use cached instance variable if available" do
+    @c2.many_to_many :attributes, :class => @c1
+
+    n = @c2.new(:id => 1234)
+    MODEL_DB.reset
+    n.associations[:attributes] = 42
+    n.attributes.should == 42
+    MODEL_DB.sqls.should == []
+  end
+
+  it "should not use cached instance variable if asked to reload" do
+    @c2.many_to_many :attributes, :class => @c1
+
+    n = @c2.new(:id => 1234)
+    MODEL_DB.reset
+    n.associations[:attributes] = 42
+    n.attributes(true).should_not == 42
+    MODEL_DB.sqls.length.should == 1
+  end
+
+  it "should add item to cached instance variable if it exists when calling add_" do
+    @c2.many_to_many :attributes, :class => @c1
+
+    n = @c2.new(:id => 1234)
+    att = @c1.new(:id => 345)
+    MODEL_DB.reset
+    a = []
+    n.associations[:attributes] = a
+    n.add_attribute(att)
+    a.should == [att]
+  end
+
+  it "should add item to reciprocal cached instance variable if it exists when calling add_" do
+    @c2.many_to_many :attributes, :class => @c1
+    @c1.many_to_many :nodes, :class => @c2
+
+    n = @c2.new(:id => 1234)
+    att = @c1.new(:id => 345)
+    att.associations[:nodes] = []
+    n.add_attribute(att)
+    att.nodes.should == [n]
+  end
+
+  it "should remove item from cached instance variable if it exists when calling remove_" do
+    @c2.many_to_many :attributes, :class => @c1
+
+    n = @c2.new(:id => 1234)
+    att = @c1.new(:id => 345)
+    MODEL_DB.reset
+    a = [att]
+    n.associations[:attributes] = a
+    n.remove_attribute(att)
+    a.should == []
+  end
+
+  it "should remove item from reciprocal cached instance variable if it exists when calling remove_" do
+    @c2.many_to_many :attributes, :class => @c1
+    @c1.many_to_many :nodes, :class => @c2
+
+    n = @c2.new(:id => 1234)
+    att = @c1.new(:id => 345)
+    att.associations[:nodes] = [n]
+    n.remove_attribute(att)
+    att.nodes.should == []
+  end
+
+  it "should not create the add_, remove_, or remove_all_ methods if :read_only option is used" do
+    @c2.many_to_many :attributes, :class => @c1, :read_only=>true
+    im = @c2.instance_methods.collect{|x| x.to_s}
+    im.should(include('attributes'))
+    im.should(include('attributes_dataset'))
+    im.should_not(include('add_attribute'))
+    im.should_not(include('remove_attribute'))
+    im.should_not(include('remove_all_attributes'))
+  end
+
+  it "should not add associations methods directly to class" do
+    @c2.many_to_many :attributes, :class => @c1
+    im = @c2.instance_methods.collect{|x| x.to_s}
+    im.should(include('attributes'))
+    im.should(include('attributes_dataset'))
+    im.should(include('add_attribute'))
+    im.should(include('remove_attribute'))
+    im.should(include('remove_all_attributes'))
+    im2 = @c2.instance_methods(false).collect{|x| x.to_s}
+    im2.should_not(include('attributes'))
+    im2.should_not(include('attributes_dataset'))
+    im2.should_not(include('add_attribute'))
+    im2.should_not(include('remove_attribute'))
+    im2.should_not(include('remove_all_attributes'))
+  end
+
+  it "should have an remove_all_ method that removes all associations" do
+    @c2.many_to_many :attributes, :class => @c1
+    @c2.new(:id => 1234).remove_all_attributes
+    MODEL_DB.sqls.first.should == 'DELETE FROM attributes_nodes WHERE (node_id = 1234)'
+  end
+
+  it "should have the remove_all_ method respect the :left_primary_key option" do
+    @c2.many_to_many :attributes, :class => @c1, :left_primary_key=>:xxx
+    @c2.new(:id => 1234, :xxx=>5).remove_all_attributes
+    MODEL_DB.sqls.first.should == 'DELETE FROM attributes_nodes WHERE (node_id = 5)'
+  end
+
+  it "remove_all should set the cached instance variable to []" do
+    @c2.many_to_many :attributes, :class => @c1
+    node = @c2.new(:id => 1234)
+    node.remove_all_attributes
+    node.associations[:attributes].should == []
+  end
+
+  it "remove_all should return the array of previously associated items if the cached instance variable exists" do
+    @c2.many_to_many :attributes, :class => @c1
+    attrib = @c1.new(:id=>3)
+    node = @c2.new(:id => 1234)
+    d = @c1.dataset
+    def d.fetch_rows(s); end
+    node.attributes.should == []
+    node.add_attribute(attrib)
+    node.associations[:attributes].should == [attrib]
+    node.remove_all_attributes.should == [attrib]
+  end
+
+  it "remove_all should return nil if the cached instance variable does not exist" do
+    @c2.many_to_many :attributes, :class => @c1
+    @c2.new(:id => 1234).remove_all_attributes.should == nil
+  end
+
+  it "remove_all should remove the current item from all reciprocal instance varaibles if it cached instance variable exists" do
+    @c2.many_to_many :attributes, :class => @c1
+    @c1.many_to_many :nodes, :class => @c2
+    d = @c1.dataset
+    def d.fetch_rows(s); end
+    d = @c2.dataset
+    def d.fetch_rows(s); end
+    attrib = @c1.new(:id=>3)
+    node = @c2.new(:id => 1234)
+    node.attributes.should == []
+    attrib.nodes.should == []
+    node.add_attribute(attrib)
+    attrib.associations[:nodes].should == [node]
+    node.remove_all_attributes
+    attrib.associations[:nodes].should == []
+  end
+
+  it "should call an _add_ method internally to add attributes" do
+    @c2.many_to_many :attributes, :class => @c1
+    @c2.private_instance_methods.collect{|x| x.to_s}.sort.should(include("_add_attribute"))
+    p = @c2.load(:id=>10)
+    c = @c1.load(:id=>123)
+    def p._add_attribute(x)
+      @x = x
+    end
+    p.add_attribute(c)
+    p.instance_variable_get(:@x).should == c
+    MODEL_DB.sqls.should == []
+  end
+
+  it "should call a _remove_ method internally to remove attributes" do
+    @c2.many_to_many :attributes, :class => @c1
+    @c2.private_instance_methods.collect{|x| x.to_s}.sort.should(include("_remove_attribute"))
+    p = @c2.load(:id=>10)
+    c = @c1.load(:id=>123)
+    def p._remove_attribute(x)
+      @x = x
+    end
+    p.remove_attribute(c)
+    p.instance_variable_get(:@x).should == c
+    MODEL_DB.sqls.should == []
+  end
+
+  it "should support (before|after)_(add|remove) callbacks" do
+    h = []
+    @c2.many_to_many :attributes, :class => @c1, :before_add=>[proc{|x,y| h << x.pk; h << -y.pk}, :blah], :after_add=>proc{h << 3}, :before_remove=>:blah, :after_remove=>[:blahr]
+    @c2.class_eval do
+      @@blah = h
+      def _add_attribute(v)
+        @@blah << 4
+      end
+      def _remove_attribute(v)
+        @@blah << 5
+      end
+      def blah(x)
+        @@blah << x.pk
+      end
+      def blahr(x)
+        @@blah << 6
+      end
+    end
+    p = @c2.load(:id=>10)
+    c = @c1.load(:id=>123)
+    h.should == []
+    p.add_attribute(c)
+    h.should == [10, -123, 123, 4, 3]
+    p.remove_attribute(c)
+    h.should == [10, -123, 123, 4, 3, 123, 5, 6]
+  end
+
+  it "should support after_load association callback" do
+    h = []
+    @c2.many_to_many :attributes, :class => @c1, :after_load=>[proc{|x,y| h << [x.pk, y.collect{|z|z.pk}]}, :al]
+    @c2.class_eval do
+      @@blah = h
+      def al(v)
+        v.each{|x| @@blah << x.pk}
+      end
+    end
+    @c1.class_eval do
+      def @dataset.fetch_rows(sql)
+        yield({:id=>20})
+        yield({:id=>30})
+      end
+    end
+    p = @c2.load(:id=>10, :parent_id=>20)
+    attributes = p.attributes
+    h.should == [[10, [20, 30]], 20, 30]
+    attributes.collect{|a| a.pk}.should == [20, 30]
+  end
+
+  it "should raise error and not call internal add or remove method if before callback returns false if raise_on_save_failure is true" do
+    p = @c2.load(:id=>10)
+    c = @c1.load(:id=>123)
+    @c2.many_to_many :attributes, :class => @c1, :before_add=>:ba, :before_remove=>:br
+    p.should_receive(:ba).once.with(c).and_return(false)
+    p.should_not_receive(:_add_attribute)
+    p.should_not_receive(:_remove_attribute)
+    p.associations[:attributes] = []
+    p.raise_on_save_failure = true
+    proc{p.add_attribute(c)}.should raise_error(Sequel::Error)
+    p.attributes.should == []
+    p.associations[:attributes] = [c]
+    p.should_receive(:br).once.with(c).and_return(false)
+    proc{p.remove_attribute(c)}.should raise_error(Sequel::Error)
+    p.attributes.should == [c]
+  end
+
+  it "should return nil and not call internal add or remove method if before callback returns false if raise_on_save_failure is false" do
+    p = @c2.load(:id=>10)
+    c = @c1.load(:id=>123)
+    p.raise_on_save_failure = false
+    @c2.many_to_many :attributes, :class => @c1, :before_add=>:ba, :before_remove=>:br
+    p.should_receive(:ba).once.with(c).and_return(false)
+    p.should_not_receive(:_add_attribute)
+    p.should_not_receive(:_remove_attribute)
+    p.associations[:attributes] = []
+    p.add_attribute(c).should == nil
+    p.attributes.should == []
+    p.associations[:attributes] = [c]
+    p.should_receive(:br).once.with(c).and_return(false)
+    p.remove_attribute(c).should == nil
+    p.attributes.should == [c]
+  end
+
+  it "should support a :uniq option that removes duplicates from the association" do
+    h = []
+    @c2.many_to_many :attributes, :class => @c1, :uniq=>true
+    @c1.class_eval do
+      def @dataset.fetch_rows(sql)
+        yield({:id=>20})
+        yield({:id=>30})
+        yield({:id=>20})
+        yield({:id=>30})
+      end
+    end
+    @c2.load(:id=>10, :parent_id=>20).attributes.should == [@c1.load(:id=>20), @c1.load(:id=>30)]
+  end
+end
+
+describe Sequel::Model, " association reflection methods" do
+  before do
+    MODEL_DB.reset
+    @c1 = Class.new(Sequel::Model(:nodes)) do
+      def self.name; 'Node'; end
+      def self.to_s; 'Node'; end
+    end
+  end
+  
+  it "#all_association_reflections should include all association reflection hashes" do
+    @c1.all_association_reflections.should == []
+
+    @c1.associate :many_to_one, :parent, :class => @c1
+    @c1.all_association_reflections.collect{|v| v[:name]}.should == [:parent]
+    @c1.all_association_reflections.collect{|v| v[:type]}.should == [:many_to_one]
+    @c1.all_association_reflections.collect{|v| v[:class]}.should == [@c1]
+
+    @c1.associate :one_to_many, :children, :class => @c1
+    @c1.all_association_reflections.sort_by{|x|x[:name].to_s}
+    @c1.all_association_reflections.sort_by{|x|x[:name].to_s}.collect{|v| v[:name]}.should == [:children, :parent]
+    @c1.all_association_reflections.sort_by{|x|x[:name].to_s}.collect{|v| v[:type]}.should == [:one_to_many, :many_to_one]
+    @c1.all_association_reflections.sort_by{|x|x[:name].to_s}.collect{|v| v[:class]}.should == [@c1, @c1]
+  end
+
+  it "#association_reflection should return nil for nonexistent association" do
+    @c1.association_reflection(:blah).should == nil
+  end
+
+  it "#association_reflection should return association reflection hash if association exists" do
+    @c1.associate :many_to_one, :parent, :class => @c1
+    @c1.association_reflection(:parent).should be_a_kind_of(Sequel::Model::Associations::AssociationReflection)
+    @c1.association_reflection(:parent)[:name].should == :parent
+    @c1.association_reflection(:parent)[:type].should == :many_to_one
+    @c1.association_reflection(:parent)[:class].should == @c1
+
+    @c1.associate :one_to_many, :children, :class => @c1
+    @c1.association_reflection(:children).should be_a_kind_of(Sequel::Model::Associations::AssociationReflection)
+    @c1.association_reflection(:children)[:name].should == :children
+    @c1.association_reflection(:children)[:type].should == :one_to_many
+    @c1.association_reflection(:children)[:class].should == @c1
+  end
+
+  it "#associations should include all association names" do
+    @c1.associations.should == []
+    @c1.associate :many_to_one, :parent, :class => @c1
+    @c1.associations.should == [:parent]
+    @c1.associate :one_to_many, :children, :class => @c1
+    @c1.associations.sort_by{|x|x.to_s}.should == [:children, :parent]
+  end
+
+  it "association reflections should be copied upon subclasing" do
+    @c1.associate :many_to_one, :parent, :class => @c1
+    c = Class.new(@c1)
+    @c1.associations.should == [:parent]
+    c.associations.should == [:parent]
+    c.associate :many_to_one, :parent2, :class => @c1
+    @c1.associations.should == [:parent]
+    c.associations.sort_by{|x| x.to_s}.should == [:parent, :parent2]
+    c.instance_methods.map{|x| x.to_s}.should include('parent')
+  end
+end
diff -pruN 1.5.1-1/spec/model/base_spec.rb 3.0.0-1/spec/model/base_spec.rb
--- 1.5.1-1/spec/model/base_spec.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/model/base_spec.rb	2009-04-29 00:52:21.000000000 +0100
@@ -0,0 +1,389 @@
+require File.join(File.dirname(__FILE__), "spec_helper")
+
+describe "Model attribute setters" do
+
+  before(:each) do
+    MODEL_DB.reset
+
+    @c = Class.new(Sequel::Model(:items)) do
+      columns :id, :x, :y, :"x y"
+    end
+    @o = @c.new
+  end
+
+  it "should mark the column value as changed" do
+    @o.changed_columns.should == []
+
+    @o.x = 2
+    @o.changed_columns.should == [:x]
+
+    @o.y = 3
+    @o.changed_columns.should == [:x, :y]
+
+    @o.changed_columns.clear
+
+    @o[:x] = 2
+    @o.changed_columns.should == [:x]
+
+    @o[:y] = 3
+    @o.changed_columns.should == [:x, :y]
+  end
+
+  it "should have columns that can't be called like normal ruby methods" do
+    @o.send(:"x y=", 3)
+    @o.changed_columns.should == [:"x y"]
+    @o.values.should == {:"x y"=>3}
+    @o.send(:"x y").should == 3
+  end
+end
+
+describe Sequel::Model, "dataset" do
+  before do
+    @a = Class.new(Sequel::Model(:items))
+    @b = Class.new(Sequel::Model)
+    
+    class Elephant < Sequel::Model(:ele1)
+    end
+    
+    class Maggot < Sequel::Model
+    end
+
+    class ShoeSize < Sequel::Model
+    end
+    
+    class BootSize < ShoeSize
+    end
+  end
+  
+  specify "should default to the plural of the class name" do
+    Maggot.dataset.sql.should == 'SELECT * FROM maggots'
+    ShoeSize.dataset.sql.should == 'SELECT * FROM shoe_sizes'
+  end
+  
+  specify "should return the dataset for the superclass if available" do
+    BootSize.dataset.sql.should == 'SELECT * FROM shoe_sizes'
+  end
+  
+  specify "should return the correct dataset if set explicitly" do
+    Elephant.dataset.sql.should == 'SELECT * FROM ele1'
+    @a.dataset.sql.should == 'SELECT * FROM items'
+  end
+  
+  specify "should raise if no dataset is explicitly set and the class is anonymous" do
+    proc {@b.dataset}.should raise_error(Sequel::Error)
+  end
+  
+  specify "should disregard namespaces for the table name" do
+    module BlahBlah
+      class MwaHaHa < Sequel::Model
+      end
+    end
+    
+    BlahBlah::MwaHaHa.dataset.sql.should == 'SELECT * FROM mwa_ha_has'
+  end
+end
+
+describe Sequel::Model, ".def_dataset_method" do
+  before do
+    @c = Class.new(Sequel::Model(:items)) do
+    end
+  end
+  
+  it "should add a method to the dataset and model if called with a block argument" do
+    @c.instance_eval do
+      def_dataset_method(:return_3){3}
+    end
+    @c.return_3.should == 3
+    @c.dataset.return_3.should == 3
+  end
+
+  it "should add all passed methods to the model if called without a block argument" do
+    @c.instance_eval do
+      def_dataset_method(:return_3, :return_4)
+    end
+    proc{@c.return_3}.should raise_error(NoMethodError)
+    proc{@c.return_4}.should raise_error(NoMethodError)
+    @c.dataset.instance_eval do
+      def return_3; 3; end
+      def return_4; 4; end
+    end
+    @c.return_3.should == 3
+    @c.return_4.should == 4
+  end
+
+  it "should cache calls and readd methods if set_dataset is used" do
+    @c.instance_eval do
+      def_dataset_method(:return_3){3}
+    end
+    @c.set_dataset :items
+    @c.return_3.should == 3
+    @c.dataset.return_3.should == 3
+  end
+
+  it "should readd methods to subclasses, if set_dataset is used in a subclass" do
+    @c.instance_eval do
+      def_dataset_method(:return_3){3}
+    end
+    c = Class.new(@c)
+    c.set_dataset :items
+    c.return_3.should == 3
+    c.dataset.return_3.should == 3
+  end
+end
+
+describe "A model class with implicit table name" do
+  before do
+    class Donkey < Sequel::Model
+    end
+  end
+  
+  specify "should have a dataset associated with the model class" do
+    Donkey.dataset.model.should == Donkey
+  end
+end
+
+describe "A model inheriting from a model" do
+  before do
+    class Feline < Sequel::Model
+    end
+    
+    class Leopard < Feline
+    end
+  end
+  
+  specify "should have a dataset associated with itself" do
+    Feline.dataset.model.should == Feline
+    Leopard.dataset.model.should == Leopard
+  end
+end
+
+describe "Model.db=" do
+  before do
+    $db1 = MockDatabase.new
+    $db2 = MockDatabase.new
+    
+    class BlueBlue < Sequel::Model(:items)
+      set_dataset $db1[:blue].filter(:x=>1)
+    end
+  end
+  
+  specify "should affect the underlying dataset" do
+    BlueBlue.db = $db2
+    
+    BlueBlue.dataset.db.should === $db2
+    BlueBlue.dataset.db.should_not === $db1
+  end
+
+  specify "should keep the same dataset options" do
+    BlueBlue.db = $db2
+    BlueBlue.dataset.sql.should == 'SELECT * FROM blue WHERE (x = 1)'
+  end
+
+  specify "should use the database for subclasses" do
+    BlueBlue.db = $db2
+    Class.new(BlueBlue).db.should === $db2
+  end
+end
+
+describe Sequel::Model, ".(allowed|restricted)_columns " do
+  before do
+    @c = Class.new(Sequel::Model(:blahblah)) do
+      columns :x, :y, :z
+      def _refresh(ds)
+        self
+      end
+    end
+    @c.strict_param_setting = false
+    @c.instance_variable_set(:@columns, [:x, :y, :z])
+  end
+  
+  it "should set the allowed columns correctly" do
+    @c.allowed_columns.should == nil
+    @c.set_allowed_columns :x
+    @c.allowed_columns.should == [:x]
+    @c.set_allowed_columns :x, :y
+    @c.allowed_columns.should == [:x, :y]
+  end
+
+  it "should set the restricted columns correctly" do
+    @c.restricted_columns.should == nil
+    @c.set_restricted_columns :x
+    @c.restricted_columns.should == [:x]
+    @c.set_restricted_columns :x, :y
+    @c.restricted_columns.should == [:x, :y]
+  end
+
+  it "should only set allowed columns by default" do
+    @c.set_allowed_columns :x, :y
+    i = @c.new(:x => 1, :y => 2, :z => 3)
+    i.values.should == {:x => 1, :y => 2}
+    i.set(:x => 4, :y => 5, :z => 6)
+    i.values.should == {:x => 4, :y => 5}
+    i.update(:x => 7, :y => 8, :z => 9)
+    i.values.delete(:id) # stupid specs
+    i.values.should == {:x => 7, :y => 8}
+  end
+
+  it "should not set restricted columns by default" do
+    @c.set_restricted_columns :z
+    i = @c.new(:x => 1, :y => 2, :z => 3)
+    i.values.should == {:x => 1, :y => 2}
+    i.set(:x => 4, :y => 5, :z => 6)
+    i.values.should == {:x => 4, :y => 5}
+    i.update(:x => 7, :y => 8, :z => 9)
+    i.values.delete(:id) # stupid specs
+    i.values.should == {:x => 7, :y => 8}
+  end
+
+  it "should have allowed take precedence over restricted" do
+    @c.set_allowed_columns :x, :y
+    @c.set_restricted_columns :y, :z
+    i = @c.new(:x => 1, :y => 2, :z => 3)
+    i.values.should == {:x => 1, :y => 2}
+    i.set(:x => 4, :y => 5, :z => 6)
+    i.values.should == {:x => 4, :y => 5}
+    i.update(:x => 7, :y => 8, :z => 9)
+    i.values.delete(:id) # stupid specs
+    i.values.should == {:x => 7, :y => 8}
+  end
+end
+
+describe Sequel::Model, ".(un)?restrict_primary_key\\??" do
+  before do
+    @c = Class.new(Sequel::Model(:blahblah)) do
+      set_primary_key :id
+      columns :x, :y, :z, :id
+      def refresh
+        self
+      end
+    end
+    @c.strict_param_setting = false
+    @c.instance_variable_set(:@columns, [:x, :y, :z])
+  end
+  
+  it "should restrict updates to primary key by default" do
+    i = @c.new(:x => 1, :y => 2, :id => 3)
+    i.values.should == {:x => 1, :y => 2}
+    i.set(:x => 4, :y => 5, :id => 6)
+    i.values.should == {:x => 4, :y => 5}
+  end
+
+  it "should allow updates to primary key if unrestrict_primary_key is used" do
+    @c.unrestrict_primary_key
+    i = @c.new(:x => 1, :y => 2, :id => 3)
+    i.values.should == {:x => 1, :y => 2, :id=>3}
+    i.set(:x => 4, :y => 5, :id => 6)
+    i.values.should == {:x => 4, :y => 5, :id=>6}
+  end
+
+  it "should have restrict_primary_key? return true or false depending" do
+    @c.restrict_primary_key?.should == true
+    @c.unrestrict_primary_key
+    @c.restrict_primary_key?.should == false
+    c1 = Class.new(@c)
+    c1.restrict_primary_key?.should == false
+    @c.restrict_primary_key
+    @c.restrict_primary_key?.should == true
+    c1.restrict_primary_key?.should == false
+    c2 = Class.new(@c)
+    c2.restrict_primary_key?.should == true
+  end
+end
+
+describe Sequel::Model, ".strict_param_setting" do
+  before do
+    @c = Class.new(Sequel::Model(:blahblah)) do
+      columns :x, :y, :z, :id
+      set_restricted_columns :z
+      def refresh
+        self
+      end
+    end
+    @c.instance_variable_set(:@columns, [:x, :y, :z])
+  end
+  
+  it "should be enabled by default" do
+    @c.strict_param_setting.should == true
+  end
+
+  it "should raise an error if a missing/restricted column/method is accessed" do
+    proc{@c.new(:z=>1)}.should raise_error(Sequel::Error)
+    proc{@c.create(:z=>1)}.should raise_error(Sequel::Error)
+    c = @c.new
+    proc{c.set(:z=>1)}.should raise_error(Sequel::Error)
+    proc{c.set_all(:id=>1)}.should raise_error(Sequel::Error)
+    proc{c.set_only({:x=>1}, :y)}.should raise_error(Sequel::Error)
+    proc{c.set_except({:x=>1}, :x)}.should raise_error(Sequel::Error)
+    proc{c.update(:z=>1)}.should raise_error(Sequel::Error)
+    proc{c.update_all(:id=>1)}.should raise_error(Sequel::Error)
+    proc{c.update_only({:x=>1}, :y)}.should raise_error(Sequel::Error)
+    proc{c.update_except({:x=>1}, :x)}.should raise_error(Sequel::Error)
+  end
+
+  it "should be disabled by strict_param_setting = false" do
+    @c.strict_param_setting = false
+    @c.strict_param_setting.should == false
+    proc{@c.new(:z=>1)}.should_not raise_error
+  end
+end
+
+describe Sequel::Model, ".[] optimization" do
+  before do
+    @c = Class.new(Sequel::Model(:a))
+    @c.instance_eval do
+      def simple_table
+        @simple_table
+      end
+    end
+  end
+
+  it "should set simple_pk to the literalized primary key column name if a single primary key" do
+    @c.simple_pk.should == 'id'
+    @c.set_primary_key :b
+    @c.simple_pk.should == 'b'
+    @c.set_primary_key :b__a.identifier
+    @c.simple_pk.should == 'b__a'
+  end
+
+  it "should have simple_pk be blank if compound or no primary key" do
+    @c.no_primary_key
+    @c.simple_pk.should == nil
+    @c.set_primary_key :b, :a
+    @c.simple_pk.should == nil
+  end
+
+  it "should have simple table set if passed a Symbol to set_dataset" do
+    @c.set_dataset :a
+    @c.simple_table.should == 'a'
+    @c.set_dataset :b
+    @c.simple_table.should == 'b'
+    @c.set_dataset :b__a
+    @c.simple_table.should == 'b.a'
+  end
+
+  it "should have simple_table = nil if passed a dataset to set_dataset" do
+    @c.set_dataset @c.db[:a]
+    @c.simple_table.should == nil
+  end
+
+  it "should have simple_table superclasses setting if inheriting" do
+    @c.set_dataset :a
+    Class.new(@c).simple_table.should == 'a'
+    @c.instance_variable_set(:@simple_table, nil)
+    Class.new(@c).simple_table.should == nil
+    @c.instance_variable_set(:@simple_table, "'b'")
+    Class.new(@c).simple_table.should == "'b'"
+  end
+
+  it "should use Dataset#with_sql if simple_table and simple_pk are true" do
+    @c.set_dataset :a
+    @c.dataset.should_receive(:with_sql).and_return(@c.dataset)
+    @c[1]
+  end
+
+  it "should not use Dataset#with_sql if either simple_table or simple_pk is nil" do
+    @c.set_dataset @c.dataset
+    @c.dataset.should_not_receive(:with_sql)
+    @c[1]
+  end
+end
diff -pruN 1.5.1-1/spec/model/dataset_methods_spec.rb 3.0.0-1/spec/model/dataset_methods_spec.rb
--- 1.5.1-1/spec/model/dataset_methods_spec.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/model/dataset_methods_spec.rb	2009-03-16 20:53:19.000000000 +0000
@@ -0,0 +1,78 @@
+require File.join(File.dirname(__FILE__), "spec_helper")
+
+describe Sequel::Model::DatasetMethods, "#destroy"  do
+  before do
+    @c = Class.new(Sequel::Model(:items)) do
+      @@destroyed = []
+      def destroy
+        @@destroyed << self
+      end
+      def self.destroyed
+        @@destroyed
+      end
+    end
+    @d = @c.dataset
+  end
+
+  it "should instantiate objects in the dataset and call destroy on each" do
+    def @d.fetch_rows(sql)
+      yield({:id=>1})
+      yield({:id=>2})
+    end
+    @d.destroy
+    @c.destroyed.collect{|x| x.values}.should == [{:id=>1}, {:id=>2}]
+  end
+
+  it "should return the number of records destroyed" do
+    def @d.fetch_rows(sql)
+      yield({:id=>1})
+      yield({:id=>2})
+    end
+    @d.destroy.should == 2
+    def @d.fetch_rows(sql)
+      yield({:id=>1})
+    end
+    @d.destroy.should == 1
+    def @d.fetch_rows(sql)
+    end
+    @d.destroy.should == 0
+  end
+end
+
+describe Sequel::Model::DatasetMethods, "#to_hash"  do
+  before do
+    @c = Class.new(Sequel::Model(:items)) do
+      set_primary_key :name
+    end
+    @d = @c.dataset
+  end
+
+  it "should result in a hash with primary key value keys and model object values" do
+    def @d.fetch_rows(sql)
+      yield({:name=>1})
+      yield({:name=>2})
+    end
+    h = @d.to_hash
+    h.should be_a_kind_of(Hash)
+    a = h.to_a
+    a.collect{|x| x[1].class}.should == [@c, @c]
+    a.sort_by{|x| x[0]}.collect{|x| [x[0], x[1].values]}.should == [[1, {:name=>1}], [2, {:name=>2}]]
+  end
+
+  it "should result in a hash with given value keys and model object values" do
+    def @d.fetch_rows(sql)
+      yield({:name=>1, :number=>3})
+      yield({:name=>2, :number=>4})
+    end
+    h = @d.to_hash(:number)
+    h.should be_a_kind_of(Hash)
+    a = h.to_a
+    a.collect{|x| x[1].class}.should == [@c, @c]
+    a.sort_by{|x| x[0]}.collect{|x| [x[0], x[1].values]}.should == [[3, {:name=>1, :number=>3}], [4, {:name=>2, :number=>4}]]
+  end
+
+  it "should raise an error if the class doesn't have a primary key" do
+    @c.no_primary_key
+    proc{@d.to_hash}.should raise_error(Sequel::Error)
+  end
+end
diff -pruN 1.5.1-1/spec/model/eager_loading_spec.rb 3.0.0-1/spec/model/eager_loading_spec.rb
--- 1.5.1-1/spec/model/eager_loading_spec.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/model/eager_loading_spec.rb	2009-03-16 20:53:19.000000000 +0000
@@ -0,0 +1,1165 @@
+require File.join(File.dirname(__FILE__), "spec_helper")
+
+describe Sequel::Model, "#eager" do
+  before(:each) do
+    MODEL_DB.reset
+    
+    class ::EagerAlbum < Sequel::Model(:albums)
+      columns :id, :band_id
+      many_to_one :band, :class=>'EagerBand', :key=>:band_id
+      one_to_many :tracks, :class=>'EagerTrack', :key=>:album_id
+      many_to_many :genres, :class=>'EagerGenre', :left_key=>:album_id, :right_key=>:genre_id, :join_table=>:ag
+      one_to_many :good_tracks, :class=>'EagerTrack', :key=>:album_id do |ds|
+        ds.filter(:name=>'Good')
+      end
+      many_to_one :band_name, :class=>'EagerBand', :key=>:band_id, :select=>[:id, :name]
+      one_to_many :track_names, :class=>'EagerTrack', :key=>:album_id, :select=>[:id, :name]
+      many_to_many :genre_names, :class=>'EagerGenre', :left_key=>:album_id, :right_key=>:genre_id, :join_table=>:ag, :select=>[:id]
+    end
+
+    class ::EagerBand < Sequel::Model(:bands)
+      columns :id, :p_k
+      one_to_many :albums, :class=>'EagerAlbum', :key=>:band_id, :eager=>:tracks
+      one_to_many :graph_albums, :class=>'EagerAlbum', :key=>:band_id, :eager_graph=>:tracks
+      many_to_many :members, :class=>'EagerBandMember', :left_key=>:band_id, :right_key=>:member_id, :join_table=>:bm
+      one_to_many :good_albums, :class=>'EagerAlbum', :key=>:band_id, :eager_block=>proc{|ds| ds.filter(:name=>'good')} do |ds|
+        ds.filter(:name=>'Good')
+      end
+      one_to_many :self_titled_albums, :class=>'EagerAlbum', :key=>:band_id, :allow_eager=>false do |ds|
+        ds.filter(:name=>name)
+      end
+      one_to_many :albums_by_name, :class=>'EagerAlbum', :key=>:band_id, :order=>:name, :allow_eager=>false
+      one_to_many :top_10_albums, :class=>'EagerAlbum', :key=>:band_id, :limit=>10
+    end
+    
+    class ::EagerTrack < Sequel::Model(:tracks)
+      columns :id, :album_id
+      many_to_one :album, :class=>'EagerAlbum', :key=>:album_id
+    end
+    
+    class ::EagerGenre < Sequel::Model(:genres)
+      columns :id, :xxx
+      many_to_many :albums, :class=>'EagerAlbum', :left_key=>:genre_id, :right_key=>:album_id, :join_table=>:ag
+    end
+    
+    class ::EagerBandMember < Sequel::Model(:members)
+      columns :id
+      many_to_many :bands, :class=>'EagerBand', :left_key=>:member_id, :right_key=>:band_id, :join_table=>:bm, :order =>:id
+    end
+    
+    EagerAlbum.dataset.extend(Module.new {
+      def columns
+        [:id, :band_id]
+      end
+
+      def fetch_rows(sql)
+        h = if sql =~ /101/
+          {:id => 101, :band_id=> 101}
+        else
+          {:id => 1, :band_id=> 2}
+        end
+        h.merge!(:x_foreign_key_x=>4) if sql =~ /ag\.genre_id/
+        @db << sql
+        yield h
+      end
+    })
+
+    EagerBand.dataset.extend(Module.new {
+      def fetch_rows(sql)
+        h = {:id => 2}
+        h.merge!(:x_foreign_key_x=>5) if sql =~ /bm\.member_id/
+        @db << sql
+        case sql
+        when /id IN (101)/
+        when /id > 100/
+          yield({:id => 101})
+          yield({:id => 102})
+        else
+          yield h
+        end
+      end
+    })
+    
+    EagerTrack.dataset.extend(Module.new {
+      def fetch_rows(sql)
+        @db << sql
+        yield({:id => 3, :album_id => 1})
+      end
+    })
+    
+    EagerGenre.dataset.extend(Module.new {
+      def fetch_rows(sql)
+        h = {:id => 4}
+        h.merge!(:x_foreign_key_x=>1) if sql =~ /ag\.album_id/
+        @db << sql
+        yield h
+      end
+    })
+    
+    EagerBandMember.dataset.extend(Module.new {
+      def fetch_rows(sql)
+        h = {:id => 5}
+        h.merge!(:x_foreign_key_x=>2) if sql =~ /bm\.band_id/
+        @db << sql
+        yield h
+      end
+    })
+  end
+  
+  it "should raise an error if called without a symbol or hash" do
+    proc{EagerAlbum.eager(Object.new)}.should raise_error(Sequel::Error)
+  end
+
+  it "should eagerly load a single many_to_one association" do
+    a = EagerAlbum.eager(:band).all
+    a.should be_a_kind_of(Array)
+    a.size.should == 1
+    a.first.should be_a_kind_of(EagerAlbum)
+    a.first.values.should == {:id => 1, :band_id => 2}
+    MODEL_DB.sqls.should == ['SELECT * FROM albums', 'SELECT * FROM bands WHERE (bands.id IN (2))']
+    a = a.first
+    a.band.should be_a_kind_of(EagerBand)
+    a.band.values.should == {:id => 2}
+    MODEL_DB.sqls.length.should == 2
+  end
+  
+  it "should eagerly load a single one_to_many association" do
+    a = EagerAlbum.eager(:tracks).all
+    a.should be_a_kind_of(Array)
+    a.size.should == 1
+    a.first.should be_a_kind_of(EagerAlbum)
+    a.first.values.should == {:id => 1, :band_id => 2}
+    MODEL_DB.sqls.should == ['SELECT * FROM albums', 'SELECT * FROM tracks WHERE (tracks.album_id IN (1))']
+    a = a.first
+    a.tracks.should be_a_kind_of(Array)
+    a.tracks.size.should == 1
+    a.tracks.first.should be_a_kind_of(EagerTrack)
+    a.tracks.first.values.should == {:id => 3, :album_id=>1}
+    MODEL_DB.sqls.length.should == 2
+  end
+  
+  it "should eagerly load a single many_to_many association" do
+    a = EagerAlbum.eager(:genres).all
+    a.should be_a_kind_of(Array)
+    a.size.should == 1
+    a.first.should be_a_kind_of(EagerAlbum)
+    a.first.values.should == {:id => 1, :band_id => 2}
+    MODEL_DB.sqls.should == ['SELECT * FROM albums', "SELECT genres.*, ag.album_id AS x_foreign_key_x FROM genres INNER JOIN ag ON ((ag.genre_id = genres.id) AND (ag.album_id IN (1)))"]
+    a = a.first
+    a.genres.should be_a_kind_of(Array)
+    a.genres.size.should == 1
+    a.genres.first.should be_a_kind_of(EagerGenre)
+    a.genres.first.values.should == {:id => 4}
+    MODEL_DB.sqls.length.should == 2
+  end
+  
+  it "should eagerly load multiple associations in a single call" do
+    a = EagerAlbum.eager(:genres, :tracks, :band).all
+    a.should be_a_kind_of(Array)
+    a.size.should == 1
+    a.first.should be_a_kind_of(EagerAlbum)
+    a.first.values.should == {:id => 1, :band_id => 2}
+    MODEL_DB.sqls.length.should == 4
+    MODEL_DB.sqls[0].should == 'SELECT * FROM albums'
+    MODEL_DB.sqls[1..-1].should(include('SELECT * FROM bands WHERE (bands.id IN (2))'))
+    MODEL_DB.sqls[1..-1].should(include('SELECT * FROM tracks WHERE (tracks.album_id IN (1))'))
+    MODEL_DB.sqls[1..-1].should(include('SELECT genres.*, ag.album_id AS x_foreign_key_x FROM genres INNER JOIN ag ON ((ag.genre_id = genres.id) AND (ag.album_id IN (1)))'))
+    a = a.first
+    a.band.should be_a_kind_of(EagerBand)
+    a.band.values.should == {:id => 2}
+    a.tracks.should be_a_kind_of(Array)
+    a.tracks.size.should == 1
+    a.tracks.first.should be_a_kind_of(EagerTrack)
+    a.tracks.first.values.should == {:id => 3, :album_id=>1}
+    a.genres.should be_a_kind_of(Array)
+    a.genres.size.should == 1
+    a.genres.first.should be_a_kind_of(EagerGenre)
+    a.genres.first.values.should == {:id => 4}
+    MODEL_DB.sqls.length.should == 4
+  end
+  
+  it "should eagerly load multiple associations in separate calls" do
+    a = EagerAlbum.eager(:genres).eager(:tracks).eager(:band).all
+    a.should be_a_kind_of(Array)
+    a.size.should == 1
+    a.first.should be_a_kind_of(EagerAlbum)
+    a.first.values.should == {:id => 1, :band_id => 2}
+    MODEL_DB.sqls.length.should == 4
+    MODEL_DB.sqls[0].should == 'SELECT * FROM albums'
+    MODEL_DB.sqls[1..-1].should(include('SELECT * FROM bands WHERE (bands.id IN (2))'))
+    MODEL_DB.sqls[1..-1].should(include('SELECT * FROM tracks WHERE (tracks.album_id IN (1))'))
+    MODEL_DB.sqls[1..-1].should(include('SELECT genres.*, ag.album_id AS x_foreign_key_x FROM genres INNER JOIN ag ON ((ag.genre_id = genres.id) AND (ag.album_id IN (1)))'))
+    a = a.first
+    a.band.should be_a_kind_of(EagerBand)
+    a.band.values.should == {:id => 2}
+    a.tracks.should be_a_kind_of(Array)
+    a.tracks.size.should == 1
+    a.tracks.first.should be_a_kind_of(EagerTrack)
+    a.tracks.first.values.should == {:id => 3, :album_id=>1}
+    a.genres.should be_a_kind_of(Array)
+    a.genres.size.should == 1
+    a.genres.first.should be_a_kind_of(EagerGenre)
+    a.genres.first.values.should == {:id => 4}
+    MODEL_DB.sqls.length.should == 4
+  end
+  
+  it "should allow cascading of eager loading for associations of associated models" do
+    a = EagerTrack.eager(:album=>{:band=>:members}).all
+    a.should be_a_kind_of(Array)
+    a.size.should == 1
+    a.first.should be_a_kind_of(EagerTrack)
+    a.first.values.should == {:id => 3, :album_id => 1}
+    MODEL_DB.sqls.length.should == 4
+    MODEL_DB.sqls.should == ['SELECT * FROM tracks', 
+      'SELECT * FROM albums WHERE (albums.id IN (1))',
+      'SELECT * FROM bands WHERE (bands.id IN (2))',
+      "SELECT members.*, bm.band_id AS x_foreign_key_x FROM members INNER JOIN bm ON ((bm.member_id = members.id) AND (bm.band_id IN (2)))"]
+    a = a.first
+    a.album.should be_a_kind_of(EagerAlbum)
+    a.album.values.should == {:id => 1, :band_id => 2}
+    a.album.band.should be_a_kind_of(EagerBand)
+    a.album.band.values.should == {:id => 2}
+    a.album.band.members.should be_a_kind_of(Array)
+    a.album.band.members.size.should == 1
+    a.album.band.members.first.should be_a_kind_of(EagerBandMember)
+    a.album.band.members.first.values.should == {:id => 5}
+    MODEL_DB.sqls.length.should == 4
+  end
+  
+  it "should cascade eagerly loading when the :eager association option is used" do
+    a = EagerBand.eager(:albums).all
+    a.should be_a_kind_of(Array)
+    a.size.should == 1
+    a.first.should be_a_kind_of(EagerBand)
+    a.first.values.should == {:id => 2}
+    MODEL_DB.sqls.should == ['SELECT * FROM bands', 
+                             'SELECT * FROM albums WHERE (albums.band_id IN (2))',
+                             'SELECT * FROM tracks WHERE (tracks.album_id IN (1))']
+    a = a.first
+    a.albums.should be_a_kind_of(Array)
+    a.albums.size.should == 1
+    a.albums.first.should be_a_kind_of(EagerAlbum)
+    a.albums.first.values.should == {:id => 1, :band_id => 2}
+    a = a.albums.first
+    a.tracks.should be_a_kind_of(Array)
+    a.tracks.size.should == 1
+    a.tracks.first.should be_a_kind_of(EagerTrack)
+    a.tracks.first.values.should == {:id => 3, :album_id => 1}
+    MODEL_DB.sqls.length.should == 3
+  end
+  
+  it "should respect :eager when lazily loading an association" do
+    a = EagerBand.all
+    a.should be_a_kind_of(Array)
+    a.size.should == 1
+    a.first.should be_a_kind_of(EagerBand)
+    a.first.values.should == {:id => 2}
+    MODEL_DB.sqls.should == ['SELECT * FROM bands']
+    a = a.first
+    a.albums
+    MODEL_DB.sqls.should == ['SELECT * FROM bands', 
+                             'SELECT * FROM albums WHERE (albums.band_id = 2)',
+                             'SELECT * FROM tracks WHERE (tracks.album_id IN (1))']
+    a.albums.should be_a_kind_of(Array)
+    a.albums.size.should == 1
+    a.albums.first.should be_a_kind_of(EagerAlbum)
+    a.albums.first.values.should == {:id => 1, :band_id => 2}
+    a = a.albums.first
+    a.tracks.should be_a_kind_of(Array)
+    a.tracks.size.should == 1
+    a.tracks.first.should be_a_kind_of(EagerTrack)
+    a.tracks.first.values.should == {:id => 3, :album_id => 1}
+    MODEL_DB.sqls.length.should == 3
+  end
+  
+  it "should cascade eagerly loading when the :eager_graph association option is used" do
+    EagerAlbum.dataset.extend(Module.new {
+      def fetch_rows(sql)
+        @db << sql
+        yield({:id=>1, :band_id=>2, :tracks_id=>3, :album_id=>1})
+      end
+    })
+    a = EagerBand.eager(:graph_albums).all
+    a.should be_a_kind_of(Array)
+    a.size.should == 1
+    a.first.should be_a_kind_of(EagerBand)
+    a.first.values.should == {:id => 2}
+    MODEL_DB.sqls.should == ['SELECT * FROM bands', 
+                             'SELECT albums.id, albums.band_id, tracks.id AS tracks_id, tracks.album_id FROM albums LEFT OUTER JOIN tracks ON (tracks.album_id = albums.id) WHERE (albums.band_id IN (2))']
+    a = a.first
+    a.graph_albums.should be_a_kind_of(Array)
+    a.graph_albums.size.should == 1
+    a.graph_albums.first.should be_a_kind_of(EagerAlbum)
+    a.graph_albums.first.values.should == {:id => 1, :band_id => 2}
+    a = a.graph_albums.first
+    a.tracks.should be_a_kind_of(Array)
+    a.tracks.size.should == 1
+    a.tracks.first.should be_a_kind_of(EagerTrack)
+    a.tracks.first.values.should == {:id => 3, :album_id => 1}
+    MODEL_DB.sqls.length.should == 2
+  end
+  
+  it "should respect :eager_graph when lazily loading an association" do
+    a = EagerBand.all
+    a.should be_a_kind_of(Array)
+    a.size.should == 1
+    a.first.should be_a_kind_of(EagerBand)
+    a.first.values.should == {:id => 2}
+    MODEL_DB.sqls.should == ['SELECT * FROM bands']
+    a = a.first
+    EagerAlbum.dataset.extend(Module.new {
+      def fetch_rows(sql)
+        @db << sql
+        yield({:id=>1, :band_id=>2, :tracks_id=>3, :album_id=>1})
+      end
+    })
+    a.graph_albums
+    MODEL_DB.sqls.should == ['SELECT * FROM bands', 
+                             'SELECT albums.id, albums.band_id, tracks.id AS tracks_id, tracks.album_id FROM albums LEFT OUTER JOIN tracks ON (tracks.album_id = albums.id) WHERE (albums.band_id = 2)']
+    a.graph_albums.should be_a_kind_of(Array)
+    a.graph_albums.size.should == 1
+    a.graph_albums.first.should be_a_kind_of(EagerAlbum)
+    a.graph_albums.first.values.should == {:id => 1, :band_id => 2}
+    a = a.graph_albums.first
+    a.tracks.should be_a_kind_of(Array)
+    a.tracks.size.should == 1
+    a.tracks.first.should be_a_kind_of(EagerTrack)
+    a.tracks.first.values.should == {:id => 3, :album_id => 1}
+    MODEL_DB.sqls.length.should == 2
+  end
+  
+  it "should respect :conditions when eagerly loading" do
+    EagerBandMember.many_to_many :good_bands, :clone=>:bands, :conditions=>{:a=>32}
+    a = EagerBandMember.eager(:good_bands).all
+    a.should be_a_kind_of(Array)
+    a.size.should == 1
+    a.first.should be_a_kind_of(EagerBandMember)
+    a.first.values.should == {:id => 5}
+    MODEL_DB.sqls.should == ['SELECT * FROM members', 'SELECT bands.*, bm.member_id AS x_foreign_key_x FROM bands INNER JOIN bm ON ((bm.band_id = bands.id) AND (bm.member_id IN (5))) WHERE (a = 32) ORDER BY id']
+    a = a.first
+    a.good_bands.should be_a_kind_of(Array)
+    a.good_bands.size.should == 1
+    a.good_bands.first.should be_a_kind_of(EagerBand)
+    a.good_bands.first.values.should == {:id => 2}
+    MODEL_DB.sqls.length.should == 2
+
+    MODEL_DB.sqls.clear
+    EagerBandMember.many_to_many :good_bands, :clone=>:bands, :conditions=>"x = 1"
+    a = EagerBandMember.eager(:good_bands).all
+    MODEL_DB.sqls.should == ['SELECT * FROM members', 'SELECT bands.*, bm.member_id AS x_foreign_key_x FROM bands INNER JOIN bm ON ((bm.band_id = bands.id) AND (bm.member_id IN (5))) WHERE (x = 1) ORDER BY id']
+  end
+  
+  it "should respect :order when eagerly loading" do
+    a = EagerBandMember.eager(:bands).all
+    a.should be_a_kind_of(Array)
+    a.size.should == 1
+    a.first.should be_a_kind_of(EagerBandMember)
+    a.first.values.should == {:id => 5}
+    MODEL_DB.sqls.should == ['SELECT * FROM members', 'SELECT bands.*, bm.member_id AS x_foreign_key_x FROM bands INNER JOIN bm ON ((bm.band_id = bands.id) AND (bm.member_id IN (5))) ORDER BY id']
+    a = a.first
+    a.bands.should be_a_kind_of(Array)
+    a.bands.size.should == 1
+    a.bands.first.should be_a_kind_of(EagerBand)
+    a.bands.first.values.should == {:id => 2}
+    MODEL_DB.sqls.length.should == 2
+  end
+  
+  it "should populate the reciprocal many_to_one association when eagerly loading the one_to_many association" do
+    a = EagerAlbum.eager(:tracks).all
+    a.should be_a_kind_of(Array)
+    a.size.should == 1
+    a.first.should be_a_kind_of(EagerAlbum)
+    a.first.values.should == {:id => 1, :band_id => 2}
+    MODEL_DB.sqls.should == ['SELECT * FROM albums', 'SELECT * FROM tracks WHERE (tracks.album_id IN (1))']
+    a = a.first
+    a.tracks.should be_a_kind_of(Array)
+    a.tracks.size.should == 1
+    a.tracks.first.should be_a_kind_of(EagerTrack)
+    a.tracks.first.values.should == {:id => 3, :album_id=>1}
+    a.tracks.first.album.should be_a_kind_of(EagerAlbum)
+    a.tracks.first.album.should == a
+    MODEL_DB.sqls.length.should == 2
+  end
+
+  it "should cache the negative lookup when eagerly loading a many_to_one association" do
+    a = EagerAlbum.eager(:band).filter(:id=>101).all
+    a.should be_a_kind_of(Array)
+    a.size.should == 1
+    a.first.should be_a_kind_of(EagerAlbum)
+    a.first.values.should == {:id => 101, :band_id => 101}
+    MODEL_DB.sqls.should == ['SELECT * FROM albums WHERE (id = 101)', 'SELECT * FROM bands WHERE (bands.id IN (101))']
+    a = a.first
+    a.associations.fetch(:band, 2).should == nil
+    a.band.should == nil
+    MODEL_DB.sqls.length.should == 2
+  end
+  
+  it "should cache the negative lookup when eagerly loading a *_to_many associations" do
+    a = EagerBand.eager(:albums).filter('id > 100').all
+    a.should be_a_kind_of(Array)
+    a.size.should == 2
+    a.first.should be_a_kind_of(EagerBand)
+    a.first.values.should == {:id => 101}
+    a.last.values.should == {:id => 102}
+    MODEL_DB.sqls.should == ['SELECT * FROM bands WHERE (id > 100)', 'SELECT * FROM albums WHERE (albums.band_id IN (101, 102))', "SELECT * FROM tracks WHERE (tracks.album_id IN (101))"]
+    a.first.associations[:albums].should be_a_kind_of(Array)
+    a.first.albums.length.should == 1
+    a.first.albums.first.should be_a_kind_of(EagerAlbum)
+    a.last.associations[:albums].should == []
+    a.last.albums.should == []
+    MODEL_DB.sqls.length.should == 3
+  end
+  
+  it "should use the association's block when eager loading by default" do
+    EagerAlbum.eager(:good_tracks).all
+    MODEL_DB.sqls.should == ['SELECT * FROM albums', "SELECT * FROM tracks WHERE ((tracks.album_id IN (1)) AND (name = 'Good'))"]
+  end
+
+  it "should use the eager_block option when eager loading if given" do
+    EagerBand.eager(:good_albums).all
+    MODEL_DB.sqls.should == ['SELECT * FROM bands', "SELECT * FROM albums WHERE ((albums.band_id IN (2)) AND (name = 'good'))"]
+    MODEL_DB.sqls.clear
+    EagerBand.eager(:good_albums=>:good_tracks).all
+    MODEL_DB.sqls.should == ['SELECT * FROM bands', "SELECT * FROM albums WHERE ((albums.band_id IN (2)) AND (name = 'good'))", "SELECT * FROM tracks WHERE ((tracks.album_id IN (1)) AND (name = 'Good'))"]
+  end
+
+  it "should raise an error when attempting to eagerly load an association with the :allow_eager option set to false" do
+    proc{EagerBand.eager(:self_titled_albums).all}.should raise_error(Sequel::Error)
+    proc{EagerBand.eager(:albums_by_name).all}.should raise_error(Sequel::Error)
+  end
+
+  it "should respect the association's :select option" do
+    EagerAlbum.eager(:band_name).all
+    MODEL_DB.sqls.should == ['SELECT * FROM albums', "SELECT id, name FROM bands WHERE (bands.id IN (2))"]
+    MODEL_DB.sqls.clear
+    EagerAlbum.eager(:track_names).all
+    MODEL_DB.sqls.should == ['SELECT * FROM albums', "SELECT id, name FROM tracks WHERE (tracks.album_id IN (1))"]
+    MODEL_DB.sqls.clear
+    EagerAlbum.eager(:genre_names).all
+    MODEL_DB.sqls.should == ['SELECT * FROM albums', "SELECT id, ag.album_id AS x_foreign_key_x FROM genres INNER JOIN ag ON ((ag.genre_id = genres.id) AND (ag.album_id IN (1)))"]
+  end
+
+  it "should respect the association's :primary_key option" do
+    EagerAlbum.many_to_one :special_band, :class=>:EagerBand, :primary_key=>:p_k, :key=>:band_id
+    EagerBand.dataset.extend(Module.new {
+      def fetch_rows(sql)
+        MODEL_DB.sqls << sql
+        yield({:p_k=>2, :id=>1})
+      end
+    })
+    as = EagerAlbum.eager(:special_band).all
+    MODEL_DB.sqls.should == ['SELECT * FROM albums', "SELECT * FROM bands WHERE (bands.p_k IN (2))"]
+    as.length.should == 1
+    as.first.special_band.should == EagerBand.load(:p_k=>2, :id=>1)
+    MODEL_DB.sqls.clear
+    EagerAlbum.one_to_many :special_tracks, :class=>:EagerTrack, :primary_key=>:band_id, :key=>:album_id
+    EagerTrack.dataset.extend(Module.new {
+      def fetch_rows(sql)
+        MODEL_DB.sqls << sql
+        yield({:album_id=>2, :id=>1})
+      end
+    })
+    as = EagerAlbum.eager(:special_tracks).all
+    MODEL_DB.sqls.should == ['SELECT * FROM albums', "SELECT * FROM tracks WHERE (tracks.album_id IN (2))"]
+    as.length.should == 1
+    as.first.special_tracks.should == [EagerTrack.load(:album_id=>2, :id=>1)]
+  end
+
+  it "should respect many_to_many association's :left_primary_key and :right_primary_key options" do
+    EagerAlbum.many_to_many :special_genres, :class=>:EagerGenre, :left_primary_key=>:band_id, :left_key=>:album_id, :right_primary_key=>:xxx, :right_key=>:genre_id, :join_table=>:ag
+    EagerGenre.dataset.extend(Module.new {
+      def fetch_rows(sql)
+        MODEL_DB.sqls << sql
+        yield({:x_foreign_key_x=>2, :id=>5})
+        yield({:x_foreign_key_x=>2, :id=>6})
+      end
+    })
+    as = EagerAlbum.eager(:special_genres).all
+    MODEL_DB.sqls.should == ['SELECT * FROM albums', "SELECT genres.*, ag.album_id AS x_foreign_key_x FROM genres INNER JOIN ag ON ((ag.genre_id = genres.xxx) AND (ag.album_id IN (2)))"]
+    as.length.should == 1
+    as.first.special_genres.should == [EagerGenre.load(:id=>5), EagerGenre.load(:id=>6)]
+  end
+
+  it "should use the :eager_loader association option when eager loading" do
+    EagerAlbum.many_to_one :special_band, :eager_loader=>(proc do |key_hash, records, assocs| 
+      item = EagerBand.filter(:album_id=>records.collect{|r| [r.pk, r.pk*2]}.flatten).order(:name).first
+      records.each{|r| r.associations[:special_band] = item}
+    end)
+    EagerAlbum.one_to_many :special_tracks, :eager_loader=>(proc do |key_hash, records, assocs| 
+      items = EagerTrack.filter(:album_id=>records.collect{|r| [r.pk, r.pk*2]}.flatten).all
+      records.each{|r| r.associations[:special_tracks] = items}
+    end)
+    EagerAlbum.many_to_many :special_genres, :class=>:EagerGenre, :eager_loader=>(proc do |key_hash, records, assocs| 
+      items = EagerGenre.inner_join(:ag, [:genre_id]).filter(:album_id=>records.collect{|r| r.pk}).all
+      records.each{|r| r.associations[:special_genres] = items}
+    end)
+    a = EagerAlbum.eager(:special_genres, :special_tracks, :special_band).all
+    a.should be_a_kind_of(Array)
+    a.size.should == 1
+    a.first.should be_a_kind_of(EagerAlbum)
+    a.first.values.should == {:id => 1, :band_id => 2}
+    MODEL_DB.sqls.length.should == 4
+    MODEL_DB.sqls[0].should == 'SELECT * FROM albums'
+    MODEL_DB.sqls[1..-1].should(include('SELECT * FROM bands WHERE (album_id IN (1, 2)) ORDER BY name LIMIT 1'))
+    MODEL_DB.sqls[1..-1].should(include('SELECT * FROM tracks WHERE (album_id IN (1, 2))'))
+    MODEL_DB.sqls[1..-1].should(include('SELECT * FROM genres INNER JOIN ag USING (genre_id) WHERE (album_id IN (1))'))
+    a = a.first
+    a.special_band.should be_a_kind_of(EagerBand)
+    a.special_band.values.should == {:id => 2}
+    a.special_tracks.should be_a_kind_of(Array)
+    a.special_tracks.size.should == 1
+    a.special_tracks.first.should be_a_kind_of(EagerTrack)
+    a.special_tracks.first.values.should == {:id => 3, :album_id=>1}
+    a.special_genres.should be_a_kind_of(Array)
+    a.special_genres.size.should == 1
+    a.special_genres.first.should be_a_kind_of(EagerGenre)
+    a.special_genres.first.values.should == {:id => 4}
+    MODEL_DB.sqls.length.should == 4
+  end
+  
+  it "should respect :after_load callbacks on associations when eager loading" do
+    EagerAlbum.many_to_one :al_band, :class=>'EagerBand', :key=>:band_id, :after_load=>proc{|o, a| a.id *=2}
+    EagerAlbum.one_to_many :al_tracks, :class=>'EagerTrack', :key=>:album_id, :after_load=>proc{|o, os| os.each{|a| a.id *=2}}
+    EagerAlbum.many_to_many :al_genres, :class=>'EagerGenre', :left_key=>:album_id, :right_key=>:genre_id, :join_table=>:ag, :after_load=>proc{|o, os| os.each{|a| a.id *=2}}
+    a = EagerAlbum.eager(:al_band, :al_tracks, :al_genres).all.first
+    a.should == EagerAlbum.load(:id => 1, :band_id => 2)
+    a.al_band.should == EagerBand.load(:id=>4)
+    a.al_tracks.should == [EagerTrack.load(:id=>6, :album_id=>1)]
+    a.al_genres.should == [EagerGenre.load(:id=>8)]
+  end
+end
+
+describe Sequel::Model, "#eager_graph" do
+  after(:all) do
+    class ::MockDataset
+      alias clone orig_clone
+    end
+  end
+
+  before(:all) do
+    class ::MockDataset
+      alias orig_clone clone
+      def clone(opts = {})
+        c = super()
+        c.opts = @opts.merge(opts)
+        c.instance_variable_set(:@columns, (@columns.dup if @columns))
+        c
+      end
+    end
+
+    class ::GraphAlbum < Sequel::Model(:albums)
+      dataset.opts[:from] = [:albums]
+      columns :id, :band_id
+      many_to_one :band, :class=>'GraphBand', :key=>:band_id
+      one_to_many :tracks, :class=>'GraphTrack', :key=>:album_id
+      many_to_many :genres, :class=>'GraphGenre', :left_key=>:album_id, :right_key=>:genre_id, :join_table=>:ag
+      many_to_one :previous_album, :class=>'GraphAlbum'
+    end
+
+    class ::GraphBand < Sequel::Model(:bands)
+      dataset.opts[:from] = [:bands]
+      columns :id, :vocalist_id
+      many_to_one :vocalist, :class=>'GraphBandMember', :key=>:vocalist_id
+      one_to_many :albums, :class=>'GraphAlbum', :key=>:band_id
+      many_to_many :members, :class=>'GraphBandMember', :left_key=>:band_id, :right_key=>:member_id, :join_table=>:bm
+      many_to_many :genres, :class=>'GraphGenre', :left_key=>:band_id, :right_key=>:genre_id, :join_table=>:bg
+    end
+    
+    class ::GraphTrack < Sequel::Model(:tracks)
+      dataset.opts[:from] = [:tracks]
+      columns :id, :album_id
+      many_to_one :album, :class=>'GraphAlbum', :key=>:album_id
+    end
+    
+    class ::GraphGenre < Sequel::Model(:genres)
+      dataset.opts[:from] = [:genres]
+      columns :id
+      many_to_many :albums, :class=>'GraphAlbum', :left_key=>:genre_id, :right_key=>:album_id, :join_table=>:ag
+    end
+    
+    class ::GraphBandMember < Sequel::Model(:members)
+      dataset.opts[:from] = [:members]
+      columns :id
+      many_to_many :bands, :class=>'GraphBand', :left_key=>:member_id, :right_key=>:band_id, :join_table=>:bm
+    end
+  end
+    
+  it "should raise an error if called without a symbol or hash" do
+    proc{GraphAlbum.eager_graph(Object.new)}.should raise_error(Sequel::Error)
+  end
+
+  it "should eagerly load a single many_to_one association" do
+    ds = GraphAlbum.eager_graph(:band)
+    ds.sql.should == 'SELECT albums.id, albums.band_id, band.id AS band_id_0, band.vocalist_id FROM albums LEFT OUTER JOIN bands AS band ON (band.id = albums.band_id)'
+    def ds.fetch_rows(sql, &block)
+      yield({:id=>1, :band_id=>2, :band_id_0=>2, :vocalist_id=>3})
+    end
+    a = ds.all
+    a.should be_a_kind_of(Array)
+    a.size.should == 1
+    a.first.should be_a_kind_of(GraphAlbum)
+    a.first.values.should == {:id => 1, :band_id => 2}
+    a = a.first
+    a.band.should be_a_kind_of(GraphBand)
+    a.band.values.should == {:id => 2, :vocalist_id=>3}
+  end
+
+  it "should eagerly load a single one_to_many association" do
+    ds = GraphAlbum.eager_graph(:tracks)
+    ds.sql.should == 'SELECT albums.id, albums.band_id, tracks.id AS tracks_id, tracks.album_id FROM albums LEFT OUTER JOIN tracks ON (tracks.album_id = albums.id)'
+    def ds.fetch_rows(sql, &block)
+      yield({:id=>1, :band_id=>2, :tracks_id=>3, :album_id=>1})
+    end
+    a = ds.all
+    a.should be_a_kind_of(Array)
+    a.size.should == 1
+    a.first.should be_a_kind_of(GraphAlbum)
+    a.first.values.should == {:id => 1, :band_id => 2}
+    a = a.first
+    a.tracks.should be_a_kind_of(Array)
+    a.tracks.size.should == 1
+    a.tracks.first.should be_a_kind_of(GraphTrack)
+    a.tracks.first.values.should == {:id => 3, :album_id=>1}
+  end
+
+  it "should eagerly load a single many_to_many association" do
+    ds = GraphAlbum.eager_graph(:genres)
+    ds.sql.should == 'SELECT albums.id, albums.band_id, genres.id AS genres_id FROM albums LEFT OUTER JOIN ag ON (ag.album_id = albums.id) LEFT OUTER JOIN genres ON (genres.id = ag.genre_id)'
+    def ds.fetch_rows(sql, &block)
+      yield({:id=>1, :band_id=>2, :genres_id=>4})
+    end
+    a = ds.all
+    a.should be_a_kind_of(Array)
+    a.size.should == 1
+    a.first.should be_a_kind_of(GraphAlbum)
+    a.first.values.should == {:id => 1, :band_id => 2}
+    a = a.first
+    a.genres.should be_a_kind_of(Array)
+    a.genres.size.should == 1
+    a.genres.first.should be_a_kind_of(GraphGenre)
+    a.genres.first.values.should == {:id => 4}
+  end
+
+  it "should eagerly load multiple associations in a single call" do 
+    ds = GraphAlbum.eager_graph(:genres, :tracks, :band)
+    ds.sql.should == 'SELECT albums.id, albums.band_id, genres.id AS genres_id, tracks.id AS tracks_id, tracks.album_id, band.id AS band_id_0, band.vocalist_id FROM albums LEFT OUTER JOIN ag ON (ag.album_id = albums.id) LEFT OUTER JOIN genres ON (genres.id = ag.genre_id) LEFT OUTER JOIN tracks ON (tracks.album_id = albums.id) LEFT OUTER JOIN bands AS band ON (band.id = albums.band_id)'
+    def ds.fetch_rows(sql, &block)
+      yield({:id=>1, :band_id=>2, :genres_id=>4, :tracks_id=>3, :album_id=>1, :band_id_0=>2, :vocalist_id=>6})
+    end
+    a = ds.all
+    a.should be_a_kind_of(Array)
+    a.size.should == 1
+    a.first.should be_a_kind_of(GraphAlbum)
+    a.first.values.should == {:id => 1, :band_id => 2}
+    a = a.first
+    a.band.should be_a_kind_of(GraphBand)
+    a.band.values.should == {:id => 2, :vocalist_id=>6}
+    a.tracks.should be_a_kind_of(Array)
+    a.tracks.size.should == 1
+    a.tracks.first.should be_a_kind_of(GraphTrack)
+    a.tracks.first.values.should == {:id => 3, :album_id=>1}
+    a.genres.should be_a_kind_of(Array)
+    a.genres.size.should == 1
+    a.genres.first.should be_a_kind_of(GraphGenre)
+    a.genres.first.values.should == {:id => 4}
+  end
+
+  it "should eagerly load multiple associations in separate calls" do 
+    ds = GraphAlbum.eager_graph(:genres).eager_graph(:tracks).eager_graph(:band)
+    ds.sql.should == 'SELECT albums.id, albums.band_id, genres.id AS genres_id, tracks.id AS tracks_id, tracks.album_id, band.id AS band_id_0, band.vocalist_id FROM albums LEFT OUTER JOIN ag ON (ag.album_id = albums.id) LEFT OUTER JOIN genres ON (genres.id = ag.genre_id) LEFT OUTER JOIN tracks ON (tracks.album_id = albums.id) LEFT OUTER JOIN bands AS band ON (band.id = albums.band_id)'
+    def ds.fetch_rows(sql, &block)
+      yield({:id=>1, :band_id=>2, :genres_id=>4, :tracks_id=>3, :album_id=>1, :band_id_0=>2, :vocalist_id=>6})
+    end
+    a = ds.all
+    a.should be_a_kind_of(Array)
+    a.size.should == 1
+    a.first.should be_a_kind_of(GraphAlbum)
+    a.first.values.should == {:id => 1, :band_id => 2}
+    a = a.first
+    a.band.should be_a_kind_of(GraphBand)
+    a.band.values.should == {:id => 2, :vocalist_id=>6}
+    a.tracks.should be_a_kind_of(Array)
+    a.tracks.size.should == 1
+    a.tracks.first.should be_a_kind_of(GraphTrack)
+    a.tracks.first.values.should == {:id => 3, :album_id=>1}
+    a.genres.should be_a_kind_of(Array)
+    a.genres.size.should == 1
+    a.genres.first.should be_a_kind_of(GraphGenre)
+    a.genres.first.values.should == {:id => 4}
+  end
+
+  it "should allow cascading of eager loading for associations of associated models" do
+    ds = GraphTrack.eager_graph(:album=>{:band=>:members})
+    ds.sql.should == 'SELECT tracks.id, tracks.album_id, album.id AS album_id_0, album.band_id, band.id AS band_id_0, band.vocalist_id, members.id AS members_id FROM tracks LEFT OUTER JOIN albums AS album ON (album.id = tracks.album_id) LEFT OUTER JOIN bands AS band ON (band.id = album.band_id) LEFT OUTER JOIN bm ON (bm.band_id = band.id) LEFT OUTER JOIN members ON (members.id = bm.member_id)'
+    def ds.fetch_rows(sql, &block)
+      yield({:id=>3, :album_id=>1, :album_id_0=>1, :band_id=>2, :members_id=>5, :band_id_0=>2, :vocalist_id=>6})
+    end
+    a = ds.all
+    a.should be_a_kind_of(Array)
+    a.size.should == 1
+    a.first.should be_a_kind_of(GraphTrack)
+    a.first.values.should == {:id => 3, :album_id => 1}
+    a = a.first
+    a.album.should be_a_kind_of(GraphAlbum)
+    a.album.values.should == {:id => 1, :band_id => 2}
+    a.album.band.should be_a_kind_of(GraphBand)
+    a.album.band.values.should == {:id => 2, :vocalist_id=>6}
+    a.album.band.members.should be_a_kind_of(Array)
+    a.album.band.members.size.should == 1
+    a.album.band.members.first.should be_a_kind_of(GraphBandMember)
+    a.album.band.members.first.values.should == {:id => 5}
+  end
+  
+  it "should allow cascading of eager loading for multiple *_to_many associations, eliminating duplicates caused by cartesian products" do
+    ds = GraphBand.eager_graph({:albums=>:tracks}, :members)
+    ds.sql.should == 'SELECT bands.id, bands.vocalist_id, albums.id AS albums_id, albums.band_id, tracks.id AS tracks_id, tracks.album_id, members.id AS members_id FROM bands LEFT OUTER JOIN albums ON (albums.band_id = bands.id) LEFT OUTER JOIN tracks ON (tracks.album_id = albums.id) LEFT OUTER JOIN bm ON (bm.band_id = bands.id) LEFT OUTER JOIN members ON (members.id = bm.member_id)'
+    def ds.fetch_rows(sql, &block)
+      yield({:id=>1, :vocalist_id=>2, :albums_id=>3, :band_id=>1, :tracks_id=>4, :album_id=>3, :members_id=>5})
+      yield({:id=>1, :vocalist_id=>2, :albums_id=>3, :band_id=>1, :tracks_id=>4, :album_id=>3, :members_id=>6})
+      yield({:id=>1, :vocalist_id=>2, :albums_id=>3, :band_id=>1, :tracks_id=>5, :album_id=>3, :members_id=>5})
+      yield({:id=>1, :vocalist_id=>2, :albums_id=>3, :band_id=>1, :tracks_id=>5, :album_id=>3, :members_id=>6})
+      yield({:id=>1, :vocalist_id=>2, :albums_id=>4, :band_id=>1, :tracks_id=>6, :album_id=>4, :members_id=>5})
+      yield({:id=>1, :vocalist_id=>2, :albums_id=>4, :band_id=>1, :tracks_id=>6, :album_id=>4, :members_id=>6})
+      yield({:id=>1, :vocalist_id=>2, :albums_id=>4, :band_id=>1, :tracks_id=>7, :album_id=>4, :members_id=>5})
+      yield({:id=>1, :vocalist_id=>2, :albums_id=>4, :band_id=>1, :tracks_id=>7, :album_id=>4, :members_id=>6})
+      yield({:id=>2, :vocalist_id=>2, :albums_id=>5, :band_id=>2, :tracks_id=>8, :album_id=>5, :members_id=>5})
+      yield({:id=>2, :vocalist_id=>2, :albums_id=>5, :band_id=>2, :tracks_id=>8, :album_id=>5, :members_id=>6})
+      yield({:id=>2, :vocalist_id=>2, :albums_id=>5, :band_id=>2, :tracks_id=>9, :album_id=>5, :members_id=>5})
+      yield({:id=>2, :vocalist_id=>2, :albums_id=>5, :band_id=>2, :tracks_id=>9, :album_id=>5, :members_id=>6})
+      yield({:id=>2, :vocalist_id=>2, :albums_id=>6, :band_id=>2, :tracks_id=>1, :album_id=>6, :members_id=>5})
+      yield({:id=>2, :vocalist_id=>2, :albums_id=>6, :band_id=>2, :tracks_id=>1, :album_id=>6, :members_id=>6})
+      yield({:id=>2, :vocalist_id=>2, :albums_id=>6, :band_id=>2, :tracks_id=>2, :album_id=>6, :members_id=>5})
+      yield({:id=>2, :vocalist_id=>2, :albums_id=>6, :band_id=>2, :tracks_id=>2, :album_id=>6, :members_id=>6})
+    end
+    a = ds.all
+    a.should == [GraphBand.load(:id=>1, :vocalist_id=>2), GraphBand.load(:id=>2, :vocalist_id=>2)]
+    members = a.map{|x| x.members}
+    members.should == [[GraphBandMember.load(:id=>5), GraphBandMember.load(:id=>6)], [GraphBandMember.load(:id=>5), GraphBandMember.load(:id=>6)]]
+    albums = a.map{|x| x.albums}
+    albums.should == [[GraphAlbum.load(:id=>3, :band_id=>1), GraphAlbum.load(:id=>4, :band_id=>1)], [GraphAlbum.load(:id=>5, :band_id=>2), GraphAlbum.load(:id=>6, :band_id=>2)]]
+    tracks = albums.map{|x| x.map{|y| y.tracks}}
+    tracks.should == [[[GraphTrack.load(:id=>4, :album_id=>3), GraphTrack.load(:id=>5, :album_id=>3)], [GraphTrack.load(:id=>6, :album_id=>4), GraphTrack.load(:id=>7, :album_id=>4)]], [[GraphTrack.load(:id=>8, :album_id=>5), GraphTrack.load(:id=>9, :album_id=>5)], [GraphTrack.load(:id=>1, :album_id=>6), GraphTrack.load(:id=>2, :album_id=>6)]]]
+  end
+  
+  it "should populate the reciprocal many_to_one association when eagerly loading the one_to_many association" do
+    MODEL_DB.reset
+    ds = GraphAlbum.eager_graph(:tracks)
+    ds.sql.should == 'SELECT albums.id, albums.band_id, tracks.id AS tracks_id, tracks.album_id FROM albums LEFT OUTER JOIN tracks ON (tracks.album_id = albums.id)'
+    def ds.fetch_rows(sql, &block)
+      @db << sql
+      yield({:id=>1, :band_id=>2, :tracks_id=>3, :album_id=>1})
+    end
+    a = ds.all
+    a.should be_a_kind_of(Array)
+    a.size.should == 1
+    a.first.should be_a_kind_of(GraphAlbum)
+    a.first.values.should == {:id => 1, :band_id => 2}
+    a = a.first
+    a.tracks.should be_a_kind_of(Array)
+    a.tracks.size.should == 1
+    a.tracks.first.should be_a_kind_of(GraphTrack)
+    a.tracks.first.values.should == {:id => 3, :album_id=>1}
+    a.tracks.first.album.should be_a_kind_of(GraphAlbum)
+    a.tracks.first.album.should == a
+    MODEL_DB.sqls.length.should == 1
+  end
+
+  it "should eager load multiple associations from the same table" do
+    ds = GraphBand.eager_graph(:vocalist, :members)
+    ds.sql.should == 'SELECT bands.id, bands.vocalist_id, vocalist.id AS vocalist_id_0, members.id AS members_id FROM bands LEFT OUTER JOIN members AS vocalist ON (vocalist.id = bands.vocalist_id) LEFT OUTER JOIN bm ON (bm.band_id = bands.id) LEFT OUTER JOIN members ON (members.id = bm.member_id)'
+    def ds.fetch_rows(sql, &block)
+      yield({:id=>2, :vocalist_id=>6, :vocalist_id_0=>6, :members_id=>5})
+    end
+    a = ds.all
+    a.should be_a_kind_of(Array)
+    a.size.should == 1
+    a.first.should be_a_kind_of(GraphBand)
+    a.first.values.should == {:id => 2, :vocalist_id => 6}
+    a = a.first
+    a.vocalist.should be_a_kind_of(GraphBandMember)
+    a.vocalist.values.should == {:id => 6}
+    a.members.should be_a_kind_of(Array)
+    a.members.size.should == 1
+    a.members.first.should be_a_kind_of(GraphBandMember)
+    a.members.first.values.should == {:id => 5}
+  end
+
+  it "should give you a graph of tables when called without .all" do 
+    ds = GraphAlbum.eager_graph(:band)
+    ds.sql.should == 'SELECT albums.id, albums.band_id, band.id AS band_id_0, band.vocalist_id FROM albums LEFT OUTER JOIN bands AS band ON (band.id = albums.band_id)'
+    def ds.fetch_rows(sql, &block)
+      yield({:id=>1, :band_id=>2, :band_id_0=>2, :vocalist_id=>3})
+    end
+    ds.first.should == {:albums=>GraphAlbum.load(:id => 1, :band_id => 2), :band=>GraphBand.load(:id => 2, :vocalist_id=>3)}
+  end
+
+  it "should not drop any associated objects if the graph could not be a cartesian product" do
+    ds = GraphBand.eager_graph(:members, :vocalist)
+    ds.sql.should == 'SELECT bands.id, bands.vocalist_id, members.id AS members_id, vocalist.id AS vocalist_id_0 FROM bands LEFT OUTER JOIN bm ON (bm.band_id = bands.id) LEFT OUTER JOIN members ON (members.id = bm.member_id) LEFT OUTER JOIN members AS vocalist ON (vocalist.id = bands.vocalist_id)'
+    def ds.fetch_rows(sql, &block)
+      yield({:id=>2, :vocalist_id=>6, :members_id=>5, :vocalist_id_0=>6})
+      yield({:id=>2, :vocalist_id=>6, :members_id=>5, :vocalist_id_0=>6})
+    end
+    a = ds.all
+    a.should be_a_kind_of(Array)
+    a.size.should == 1
+    a.first.should be_a_kind_of(GraphBand)
+    a.first.values.should == {:id => 2, :vocalist_id => 6}
+    a = a.first
+    a.vocalist.should be_a_kind_of(GraphBandMember)
+    a.vocalist.values.should == {:id => 6}
+    a.members.should be_a_kind_of(Array)
+    a.members.size.should == 2
+    a.members.first.should be_a_kind_of(GraphBandMember)
+    a.members.first.values.should == {:id => 5}
+    a.members.last.should be_a_kind_of(GraphBandMember)
+    a.members.last.values.should == {:id => 5}
+  end
+
+  it "should drop duplicate items that occur in sequence if the graph could be a cartesian product" do
+    ds = GraphBand.eager_graph(:members, :genres)
+    ds.sql.should == 'SELECT bands.id, bands.vocalist_id, members.id AS members_id, genres.id AS genres_id FROM bands LEFT OUTER JOIN bm ON (bm.band_id = bands.id) LEFT OUTER JOIN members ON (members.id = bm.member_id) LEFT OUTER JOIN bg ON (bg.band_id = bands.id) LEFT OUTER JOIN genres ON (genres.id = bg.genre_id)'
+    def ds.fetch_rows(sql, &block)
+      yield({:id=>2, :vocalist_id=>6, :members_id=>5, :genres_id=>7})
+      yield({:id=>2, :vocalist_id=>6, :members_id=>5, :genres_id=>8})
+      yield({:id=>2, :vocalist_id=>6, :members_id=>6, :genres_id=>7})
+      yield({:id=>2, :vocalist_id=>6, :members_id=>6, :genres_id=>8})
+    end
+    a = ds.all
+    a.should be_a_kind_of(Array)
+    a.size.should == 1
+    a.first.should be_a_kind_of(GraphBand)
+    a.first.values.should == {:id => 2, :vocalist_id => 6}
+    a = a.first
+    a.members.should be_a_kind_of(Array)
+    a.members.size.should == 2
+    a.members.first.should be_a_kind_of(GraphBandMember)
+    a.members.first.values.should == {:id => 5}
+    a.members.last.should be_a_kind_of(GraphBandMember)
+    a.members.last.values.should == {:id => 6}
+    a.genres.size.should == 2
+    a.genres.first.should be_a_kind_of(GraphGenre)
+    a.genres.first.values.should == {:id => 7}
+    a.genres.last.should be_a_kind_of(GraphGenre)
+    a.genres.last.values.should == {:id => 8}
+  end
+
+  it "should be able to be used in combination with #eager" do
+    MODEL_DB.reset
+    ds = GraphAlbum.eager_graph(:tracks).eager(:genres)
+    def ds.fetch_rows(sql, &block)
+      @db << sql
+      yield({:id=>1, :band_id=>2, :tracks_id=>3, :album_id=>1})
+    end
+    ds2 = GraphGenre.dataset
+    def ds2.fetch_rows(sql, &block)
+      @db << sql
+      yield({:id=>6, :x_foreign_key_x=>1})
+    end
+    a = ds.all
+    a.should be_a_kind_of(Array)
+    a.size.should == 1
+    a.first.should be_a_kind_of(GraphAlbum)
+    a.first.values.should == {:id => 1, :band_id => 2}
+    a = a.first
+    a.tracks.should be_a_kind_of(Array)
+    a.tracks.size.should == 1
+    a.tracks.first.should be_a_kind_of(GraphTrack)
+    a.tracks.first.values.should == {:id=>3, :album_id=>1}
+    a.genres.should be_a_kind_of(Array)
+    a.genres.size.should == 1
+    a.genres.first.should be_a_kind_of(GraphGenre)
+    a.genres.first.values.should == {:id=>6}
+    MODEL_DB.sqls.should == ['SELECT albums.id, albums.band_id, tracks.id AS tracks_id, tracks.album_id FROM albums LEFT OUTER JOIN tracks ON (tracks.album_id = albums.id)',
+    "SELECT genres.*, ag.album_id AS x_foreign_key_x FROM genres INNER JOIN ag ON ((ag.genre_id = genres.id) AND (ag.album_id IN (1)))"]
+  end
+
+  it "should handle no associated records for a single many_to_one association" do
+    ds = GraphAlbum.eager_graph(:band)
+    ds.sql.should == 'SELECT albums.id, albums.band_id, band.id AS band_id_0, band.vocalist_id FROM albums LEFT OUTER JOIN bands AS band ON (band.id = albums.band_id)'
+    def ds.fetch_rows(sql, &block)
+      yield({:id=>1, :band_id=>2, :band_id_0=>nil, :vocalist_id=>nil})
+    end
+    a = ds.all
+    a.should be_a_kind_of(Array)
+    a.size.should == 1
+    a.first.should be_a_kind_of(GraphAlbum)
+    a.first.values.should == {:id => 1, :band_id => 2}
+    a.first.associations.fetch(:band, 2).should == nil
+  end
+
+  it "should handle no associated records for a single one_to_many association" do
+    ds = GraphAlbum.eager_graph(:tracks)
+    ds.sql.should == 'SELECT albums.id, albums.band_id, tracks.id AS tracks_id, tracks.album_id FROM albums LEFT OUTER JOIN tracks ON (tracks.album_id = albums.id)'
+    def ds.fetch_rows(sql, &block)
+      yield({:id=>1, :band_id=>2, :tracks_id=>nil, :album_id=>nil})
+    end
+    a = ds.all
+    a.should be_a_kind_of(Array)
+    a.size.should == 1
+    a.first.should be_a_kind_of(GraphAlbum)
+    a.first.values.should == {:id => 1, :band_id => 2}
+    a.first.tracks.should == []
+  end
+
+  it "should handle no associated records for a single many_to_many association" do
+    ds = GraphAlbum.eager_graph(:genres)
+    ds.sql.should == 'SELECT albums.id, albums.band_id, genres.id AS genres_id FROM albums LEFT OUTER JOIN ag ON (ag.album_id = albums.id) LEFT OUTER JOIN genres ON (genres.id = ag.genre_id)'
+    def ds.fetch_rows(sql, &block)
+      yield({:id=>1, :band_id=>2, :genres_id=>nil})
+    end
+    a = ds.all
+    a.should be_a_kind_of(Array)
+    a.size.should == 1
+    a.first.should be_a_kind_of(GraphAlbum)
+    a.first.values.should == {:id => 1, :band_id => 2}
+    a.first.genres.should == []
+  end
+
+  it "should handle missing associated records when loading multiple associations" do 
+    ds = GraphAlbum.eager_graph(:genres, :tracks, :band)
+    ds.sql.should == 'SELECT albums.id, albums.band_id, genres.id AS genres_id, tracks.id AS tracks_id, tracks.album_id, band.id AS band_id_0, band.vocalist_id FROM albums LEFT OUTER JOIN ag ON (ag.album_id = albums.id) LEFT OUTER JOIN genres ON (genres.id = ag.genre_id) LEFT OUTER JOIN tracks ON (tracks.album_id = albums.id) LEFT OUTER JOIN bands AS band ON (band.id = albums.band_id)'
+    def ds.fetch_rows(sql, &block)
+      yield({:id=>1, :band_id=>2, :genres_id=>nil, :tracks_id=>3, :album_id=>1, :band_id_0=>nil, :vocalist_id=>nil})
+      yield({:id=>1, :band_id=>2, :genres_id=>nil, :tracks_id=>4, :album_id=>1, :band_id_0=>nil, :vocalist_id=>nil})
+      yield({:id=>1, :band_id=>2, :genres_id=>nil, :tracks_id=>5, :album_id=>1, :band_id_0=>nil, :vocalist_id=>nil})
+      yield({:id=>1, :band_id=>2, :genres_id=>nil, :tracks_id=>6, :album_id=>1, :band_id_0=>nil, :vocalist_id=>nil})
+    end
+    a = ds.all
+    a.should be_a_kind_of(Array)
+    a.size.should == 1
+    a.first.should be_a_kind_of(GraphAlbum)
+    a.first.values.should == {:id => 1, :band_id => 2}
+    a = a.first
+    a.tracks.should be_a_kind_of(Array)
+    a.tracks.size.should == 4
+    a.tracks.first.should be_a_kind_of(GraphTrack)
+    a.tracks.collect{|x|x[:id]}.should == [3,4,5,6]
+    a.associations.fetch(:band, 2).should == nil
+    a.genres.should == []
+  end
+
+  it "should handle missing associated records when cascading eager loading for associations of associated models" do
+    ds = GraphTrack.eager_graph(:album=>{:band=>:members})
+    ds.sql.should == 'SELECT tracks.id, tracks.album_id, album.id AS album_id_0, album.band_id, band.id AS band_id_0, band.vocalist_id, members.id AS members_id FROM tracks LEFT OUTER JOIN albums AS album ON (album.id = tracks.album_id) LEFT OUTER JOIN bands AS band ON (band.id = album.band_id) LEFT OUTER JOIN bm ON (bm.band_id = band.id) LEFT OUTER JOIN members ON (members.id = bm.member_id)'
+    def ds.fetch_rows(sql, &block)
+      yield({:id=>2, :album_id=>2, :album_id_0=>nil, :band_id=>nil, :members_id=>nil, :band_id_0=>nil, :vocalist_id=>nil})
+      yield({:id=>3, :album_id=>3, :album_id_0=>3, :band_id=>3, :members_id=>nil, :band_id_0=>nil, :vocalist_id=>nil})
+      yield({:id=>4, :album_id=>4, :album_id_0=>4, :band_id=>2, :members_id=>nil, :band_id_0=>2, :vocalist_id=>6})
+      yield({:id=>5, :album_id=>1, :album_id_0=>1, :band_id=>4, :members_id=>5, :band_id_0=>4, :vocalist_id=>8})
+      yield({:id=>5, :album_id=>1, :album_id_0=>1, :band_id=>4, :members_id=>6, :band_id_0=>4, :vocalist_id=>8})
+    end
+    a = ds.all
+    a.should be_a_kind_of(Array)
+    a.size.should == 4
+    a.first.should be_a_kind_of(GraphTrack)
+    a.collect{|x|x[:id]}.should == [2,3,4,5]
+    a[0].associations.fetch(:album, 2).should == nil
+    a[1].album.should be_a_kind_of(GraphAlbum)
+    a[1].album.values.should == {:id => 3, :band_id => 3}
+    a[1].album.associations.fetch(:band, 2).should == nil
+    a[2].album.should be_a_kind_of(GraphAlbum)
+    a[2].album.values.should == {:id => 4, :band_id => 2}
+    a[2].album.band.should be_a_kind_of(GraphBand)
+    a[2].album.band.values.should == {:id => 2, :vocalist_id=>6}
+    a[2].album.band.members.should == []
+    a[3].album.should be_a_kind_of(GraphAlbum)
+    a[3].album.values.should == {:id => 1, :band_id => 4}
+    a[3].album.band.should be_a_kind_of(GraphBand)
+    a[3].album.band.values.should == {:id => 4, :vocalist_id=>8}
+    a[3].album.band.members.size.should == 2
+    a[3].album.band.members.first.should be_a_kind_of(GraphBandMember)
+    a[3].album.band.members.first.values.should == {:id => 5}
+    a[3].album.band.members.last.should be_a_kind_of(GraphBandMember)
+    a[3].album.band.members.last.values.should == {:id => 6}
+  end
+
+  it "should respect the association's :primary_key option" do 
+    GraphAlbum.many_to_one :inner_band, :class=>'GraphBand', :key=>:band_id, :primary_key=>:vocalist_id
+    ds = GraphAlbum.eager_graph(:inner_band)
+    ds.sql.should == 'SELECT albums.id, albums.band_id, inner_band.id AS inner_band_id, inner_band.vocalist_id FROM albums LEFT OUTER JOIN bands AS inner_band ON (inner_band.vocalist_id = albums.band_id)'
+    def ds.fetch_rows(sql, &block)
+      yield({:id=>3, :band_id=>2, :inner_band_id=>5, :vocalist_id=>2})
+    end
+    as = ds.all
+    as.should == [GraphAlbum.load(:id=>3, :band_id=>2)]
+    as.first.inner_band.should == GraphBand.load(:id=>5, :vocalist_id=>2)
+
+    GraphAlbum.one_to_many :right_tracks, :class=>'GraphTrack', :key=>:album_id, :primary_key=>:band_id
+    ds = GraphAlbum.eager_graph(:right_tracks)
+    ds.sql.should == 'SELECT albums.id, albums.band_id, right_tracks.id AS right_tracks_id, right_tracks.album_id FROM albums LEFT OUTER JOIN tracks AS right_tracks ON (right_tracks.album_id = albums.band_id)'
+    def ds.fetch_rows(sql, &block)
+      yield({:id=>3, :band_id=>2, :right_tracks_id=>5, :album_id=>2})
+      yield({:id=>3, :band_id=>2, :right_tracks_id=>6, :album_id=>2})
+    end
+    as = ds.all
+    as.should == [GraphAlbum.load(:id=>3, :band_id=>2)]
+    as.first.right_tracks.should == [GraphTrack.load(:id=>5, :album_id=>2), GraphTrack.load(:id=>6, :album_id=>2)]
+  end
+
+  it "should respect many_to_many association's :left_primary_key and :right_primary_key options" do 
+    GraphAlbum.many_to_many :inner_genres, :class=>'GraphGenre', :left_key=>:album_id, :left_primary_key=>:band_id, :right_key=>:genre_id, :right_primary_key=>:xxx, :join_table=>:ag
+    ds = GraphAlbum.eager_graph(:inner_genres)
+    ds.sql.should == 'SELECT albums.id, albums.band_id, inner_genres.id AS inner_genres_id FROM albums LEFT OUTER JOIN ag ON (ag.album_id = albums.band_id) LEFT OUTER JOIN genres AS inner_genres ON (inner_genres.xxx = ag.genre_id)'
+    def ds.fetch_rows(sql, &block)
+      yield({:id=>3, :band_id=>2, :inner_genres_id=>5, :xxx=>12})
+      yield({:id=>3, :band_id=>2, :inner_genres_id=>6, :xxx=>22})
+    end
+    as = ds.all
+    as.should == [GraphAlbum.load(:id=>3, :band_id=>2)]
+    as.first.inner_genres.should == [GraphGenre.load(:id=>5), GraphGenre.load(:id=>6)]
+  end
+
+  it "should respect the association's :graph_select option" do 
+    GraphAlbum.many_to_one :inner_band, :class=>'GraphBand', :key=>:band_id, :graph_select=>:vocalist_id
+    GraphAlbum.eager_graph(:inner_band).sql.should == 'SELECT albums.id, albums.band_id, inner_band.vocalist_id FROM albums LEFT OUTER JOIN bands AS inner_band ON (inner_band.id = albums.band_id)'
+
+    GraphAlbum.one_to_many :right_tracks, :class=>'GraphTrack', :key=>:album_id, :graph_select=>[:album_id]
+    GraphAlbum.eager_graph(:right_tracks).sql.should == 'SELECT albums.id, albums.band_id, right_tracks.album_id FROM albums LEFT OUTER JOIN tracks AS right_tracks ON (right_tracks.album_id = albums.id)'
+
+    GraphAlbum.many_to_many :inner_genres, :class=>'GraphGenre', :left_key=>:album_id, :right_key=>:genre_id, :join_table=>:ag, :graph_select=>[]
+    GraphAlbum.eager_graph(:inner_genres).sql.should == 'SELECT albums.id, albums.band_id FROM albums LEFT OUTER JOIN ag ON (ag.album_id = albums.id) LEFT OUTER JOIN genres AS inner_genres ON (inner_genres.id = ag.genre_id)'
+  end
+
+  it "should respect the association's :graph_join_type option" do 
+    GraphAlbum.many_to_one :inner_band, :class=>'GraphBand', :key=>:band_id, :graph_join_type=>:inner
+    GraphAlbum.eager_graph(:inner_band).sql.should == 'SELECT albums.id, albums.band_id, inner_band.id AS inner_band_id, inner_band.vocalist_id FROM albums INNER JOIN bands AS inner_band ON (inner_band.id = albums.band_id)'
+
+    GraphAlbum.one_to_many :right_tracks, :class=>'GraphTrack', :key=>:album_id, :graph_join_type=>:right_outer
+    GraphAlbum.eager_graph(:right_tracks).sql.should == 'SELECT albums.id, albums.band_id, right_tracks.id AS right_tracks_id, right_tracks.album_id FROM albums RIGHT OUTER JOIN tracks AS right_tracks ON (right_tracks.album_id = albums.id)'
+
+    GraphAlbum.many_to_many :inner_genres, :class=>'GraphGenre', :left_key=>:album_id, :right_key=>:genre_id, :join_table=>:ag, :graph_join_type=>:inner
+    GraphAlbum.eager_graph(:inner_genres).sql.should == 'SELECT albums.id, albums.band_id, inner_genres.id AS inner_genres_id FROM albums INNER JOIN ag ON (ag.album_id = albums.id) INNER JOIN genres AS inner_genres ON (inner_genres.id = ag.genre_id)'
+  end
+
+  it "should respect the association's :graph_join_type option" do 
+    GraphAlbum.many_to_many :inner_genres, :class=>'GraphGenre', :left_key=>:album_id, :right_key=>:genre_id, :join_table=>:ag, :graph_join_table_join_type=>:inner
+    GraphAlbum.eager_graph(:inner_genres).sql.should == 'SELECT albums.id, albums.band_id, inner_genres.id AS inner_genres_id FROM albums INNER JOIN ag ON (ag.album_id = albums.id) LEFT OUTER JOIN genres AS inner_genres ON (inner_genres.id = ag.genre_id)'
+
+    GraphAlbum.many_to_many :inner_genres, :class=>'GraphGenre', :left_key=>:album_id, :right_key=>:genre_id, :join_table=>:ag, :graph_join_table_join_type=>:inner, :graph_join_type=>:right_outer
+    GraphAlbum.eager_graph(:inner_genres).sql.should == 'SELECT albums.id, albums.band_id, inner_genres.id AS inner_genres_id FROM albums INNER JOIN ag ON (ag.album_id = albums.id) RIGHT OUTER JOIN genres AS inner_genres ON (inner_genres.id = ag.genre_id)'
+  end
+
+  it "should respect the association's :conditions option" do 
+    GraphAlbum.many_to_one :active_band, :class=>'GraphBand', :key=>:band_id, :conditions=>{:active=>true}
+    GraphAlbum.eager_graph(:active_band).sql.should == "SELECT albums.id, albums.band_id, active_band.id AS active_band_id, active_band.vocalist_id FROM albums LEFT OUTER JOIN bands AS active_band ON ((active_band.id = albums.band_id) AND (active_band.active IS TRUE))"
+
+    GraphAlbum.one_to_many :right_tracks, :class=>'GraphTrack', :key=>:album_id, :conditions=>{:id=>(0..100)}
+    GraphAlbum.eager_graph(:right_tracks).sql.should == 'SELECT albums.id, albums.band_id, right_tracks.id AS right_tracks_id, right_tracks.album_id FROM albums LEFT OUTER JOIN tracks AS right_tracks ON ((right_tracks.album_id = albums.id) AND ((right_tracks.id >= 0) AND (right_tracks.id <= 100)))'
+
+    GraphAlbum.many_to_many :active_genres, :class=>'GraphGenre', :left_key=>:album_id, :right_key=>:genre_id, :join_table=>:ag, :conditions=>{true=>:active}
+    GraphAlbum.eager_graph(:active_genres).sql.should == "SELECT albums.id, albums.band_id, active_genres.id AS active_genres_id FROM albums LEFT OUTER JOIN ag ON (ag.album_id = albums.id) LEFT OUTER JOIN genres AS active_genres ON ((active_genres.id = ag.genre_id) AND ('t' = ag.active))"
+  end
+
+  it "should respect the association's :graph_conditions option" do 
+    GraphAlbum.many_to_one :active_band, :class=>'GraphBand', :key=>:band_id, :graph_conditions=>{:active=>true}
+    GraphAlbum.eager_graph(:active_band).sql.should == "SELECT albums.id, albums.band_id, active_band.id AS active_band_id, active_band.vocalist_id FROM albums LEFT OUTER JOIN bands AS active_band ON ((active_band.id = albums.band_id) AND (active_band.active IS TRUE))"
+
+    GraphAlbum.one_to_many :right_tracks, :class=>'GraphTrack', :key=>:album_id, :graph_conditions=>{:id=>(0..100)}
+    GraphAlbum.eager_graph(:right_tracks).sql.should == 'SELECT albums.id, albums.band_id, right_tracks.id AS right_tracks_id, right_tracks.album_id FROM albums LEFT OUTER JOIN tracks AS right_tracks ON ((right_tracks.album_id = albums.id) AND ((right_tracks.id >= 0) AND (right_tracks.id <= 100)))'
+
+    GraphAlbum.many_to_many :active_genres, :class=>'GraphGenre', :left_key=>:album_id, :right_key=>:genre_id, :join_table=>:ag, :graph_conditions=>{true=>:active}
+    GraphAlbum.eager_graph(:active_genres).sql.should == "SELECT albums.id, albums.band_id, active_genres.id AS active_genres_id FROM albums LEFT OUTER JOIN ag ON (ag.album_id = albums.id) LEFT OUTER JOIN genres AS active_genres ON ((active_genres.id = ag.genre_id) AND ('t' = ag.active))"
+  end
+
+  it "should respect the association's :graph_join_table_conditions option" do 
+    GraphAlbum.many_to_many :active_genres, :class=>'GraphGenre', :left_key=>:album_id, :right_key=>:genre_id, :join_table=>:ag, :graph_join_table_conditions=>{:active=>true}
+    GraphAlbum.eager_graph(:active_genres).sql.should == "SELECT albums.id, albums.band_id, active_genres.id AS active_genres_id FROM albums LEFT OUTER JOIN ag ON ((ag.album_id = albums.id) AND (ag.active IS TRUE)) LEFT OUTER JOIN genres AS active_genres ON (active_genres.id = ag.genre_id)"
+
+    GraphAlbum.many_to_many :active_genres, :class=>'GraphGenre', :left_key=>:album_id, :right_key=>:genre_id, :join_table=>:ag, :graph_conditions=>{true=>:active}, :graph_join_table_conditions=>{true=>:active}
+    GraphAlbum.eager_graph(:active_genres).sql.should == "SELECT albums.id, albums.band_id, active_genres.id AS active_genres_id FROM albums LEFT OUTER JOIN ag ON ((ag.album_id = albums.id) AND ('t' = albums.active)) LEFT OUTER JOIN genres AS active_genres ON ((active_genres.id = ag.genre_id) AND ('t' = ag.active))"
+  end
+
+  it "should respect the association's :graph_block option" do 
+    GraphAlbum.many_to_one :active_band, :class=>'GraphBand', :key=>:band_id, :graph_block=>proc{|ja,lja,js| {:active.qualify(ja)=>true}}
+    GraphAlbum.eager_graph(:active_band).sql.should == "SELECT albums.id, albums.band_id, active_band.id AS active_band_id, active_band.vocalist_id FROM albums LEFT OUTER JOIN bands AS active_band ON ((active_band.id = albums.band_id) AND (active_band.active IS TRUE))"
+
+    GraphAlbum.one_to_many :right_tracks, :class=>'GraphTrack', :key=>:album_id, :graph_block=>proc{|ja,lja,js| {:id.qualify(ja)=>(0..100)}}
+    GraphAlbum.eager_graph(:right_tracks).sql.should == 'SELECT albums.id, albums.band_id, right_tracks.id AS right_tracks_id, right_tracks.album_id FROM albums LEFT OUTER JOIN tracks AS right_tracks ON ((right_tracks.album_id = albums.id) AND ((right_tracks.id >= 0) AND (right_tracks.id <= 100)))'
+
+    GraphAlbum.many_to_many :active_genres, :class=>'GraphGenre', :left_key=>:album_id, :right_key=>:genre_id, :join_table=>:ag, :graph_block=>proc{|ja,lja,js| {true=>:active.qualify(lja)}}
+    GraphAlbum.eager_graph(:active_genres).sql.should == "SELECT albums.id, albums.band_id, active_genres.id AS active_genres_id FROM albums LEFT OUTER JOIN ag ON (ag.album_id = albums.id) LEFT OUTER JOIN genres AS active_genres ON ((active_genres.id = ag.genre_id) AND ('t' = ag.active))"
+  end
+
+  it "should respect the association's :graph_join_block option" do 
+    GraphAlbum.many_to_many :active_genres, :class=>'GraphGenre', :left_key=>:album_id, :right_key=>:genre_id, :join_table=>:ag, :graph_join_table_block=>proc{|ja,lja,js| {:active.qualify(ja)=>true}}
+    GraphAlbum.eager_graph(:active_genres).sql.should == "SELECT albums.id, albums.band_id, active_genres.id AS active_genres_id FROM albums LEFT OUTER JOIN ag ON ((ag.album_id = albums.id) AND (ag.active IS TRUE)) LEFT OUTER JOIN genres AS active_genres ON (active_genres.id = ag.genre_id)"
+
+    GraphAlbum.many_to_many :active_genres, :class=>'GraphGenre', :left_key=>:album_id, :right_key=>:genre_id, :join_table=>:ag, :graph_block=>proc{|ja,lja,js| {true=>:active.qualify(lja)}}, :graph_join_table_block=>proc{|ja,lja,js| {true=>:active.qualify(lja)}}
+    GraphAlbum.eager_graph(:active_genres).sql.should == "SELECT albums.id, albums.band_id, active_genres.id AS active_genres_id FROM albums LEFT OUTER JOIN ag ON ((ag.album_id = albums.id) AND ('t' = albums.active)) LEFT OUTER JOIN genres AS active_genres ON ((active_genres.id = ag.genre_id) AND ('t' = ag.active))"
+  end
+
+  it "should respect the association's :eager_grapher option" do 
+    GraphAlbum.many_to_one :active_band, :class=>'GraphBand', :key=>:band_id, :eager_grapher=>proc{|ds, aa, ta| ds.graph(GraphBand, {:active=>true}, :table_alias=>aa, :join_type=>:inner)}
+    GraphAlbum.eager_graph(:active_band).sql.should == "SELECT albums.id, albums.band_id, active_band.id AS active_band_id, active_band.vocalist_id FROM albums INNER JOIN bands AS active_band ON (active_band.active IS TRUE)"
+
+    GraphAlbum.one_to_many :right_tracks, :class=>'GraphTrack', :key=>:album_id, :eager_grapher=>proc{|ds, aa, ta| ds.graph(GraphTrack, nil, :join_type=>:natural, :table_alias=>aa)}
+    GraphAlbum.eager_graph(:right_tracks).sql.should == 'SELECT albums.id, albums.band_id, right_tracks.id AS right_tracks_id, right_tracks.album_id FROM albums NATURAL JOIN tracks AS right_tracks'
+
+    GraphAlbum.many_to_many :active_genres, :class=>'GraphGenre', :eager_grapher=>proc{|ds, aa, ta| ds.graph(:ag, {:album_id=>:id}, :table_alias=>:a123, :implicit_qualifier=>ta).graph(GraphGenre, [:album_id], :table_alias=>aa)}
+    GraphAlbum.eager_graph(:active_genres).sql.should == "SELECT albums.id, albums.band_id, active_genres.id AS active_genres_id FROM albums LEFT OUTER JOIN ag AS a123 ON (a123.album_id = albums.id) LEFT OUTER JOIN genres AS active_genres USING (album_id)"
+  end
+
+  it "should respect the association's :graph_only_conditions option" do 
+    GraphAlbum.many_to_one :active_band, :class=>'GraphBand', :key=>:band_id, :graph_only_conditions=>{:active=>true}
+    GraphAlbum.eager_graph(:active_band).sql.should == "SELECT albums.id, albums.band_id, active_band.id AS active_band_id, active_band.vocalist_id FROM albums LEFT OUTER JOIN bands AS active_band ON (active_band.active IS TRUE)"
+
+    GraphAlbum.one_to_many :right_tracks, :class=>'GraphTrack', :key=>:album_id, :graph_only_conditions=>nil, :graph_join_type=>:natural
+    GraphAlbum.eager_graph(:right_tracks).sql.should == 'SELECT albums.id, albums.band_id, right_tracks.id AS right_tracks_id, right_tracks.album_id FROM albums NATURAL JOIN tracks AS right_tracks'
+
+    GraphAlbum.many_to_many :active_genres, :class=>'GraphGenre', :left_key=>:album_id, :right_key=>:genre_id, :join_table=>:ag, :graph_only_conditions=>[:album_id]
+    GraphAlbum.eager_graph(:active_genres).sql.should == "SELECT albums.id, albums.band_id, active_genres.id AS active_genres_id FROM albums LEFT OUTER JOIN ag ON (ag.album_id = albums.id) LEFT OUTER JOIN genres AS active_genres USING (album_id)"
+  end
+
+  it "should respect the association's :graph_join_table_only_conditions option" do 
+    GraphAlbum.many_to_many :active_genres, :class=>'GraphGenre', :left_key=>:album_id, :right_key=>:genre_id, :join_table=>:ag, :graph_join_table_only_conditions=>{:active=>true}
+    GraphAlbum.eager_graph(:active_genres).sql.should == "SELECT albums.id, albums.band_id, active_genres.id AS active_genres_id FROM albums LEFT OUTER JOIN ag ON (ag.active IS TRUE) LEFT OUTER JOIN genres AS active_genres ON (active_genres.id = ag.genre_id)"
+
+    GraphAlbum.many_to_many :active_genres, :class=>'GraphGenre', :left_key=>:album_id, :right_key=>:genre_id, :join_table=>:ag, :graph_only_conditions=>(:price + 2 > 100), :graph_join_table_only_conditions=>"active"
+    GraphAlbum.eager_graph(:active_genres).sql.should == "SELECT albums.id, albums.band_id, active_genres.id AS active_genres_id FROM albums LEFT OUTER JOIN ag ON (active) LEFT OUTER JOIN genres AS active_genres ON ((price + 2) > 100)"
+  end
+
+  it "should create unique table aliases for all associations" do
+    GraphAlbum.eager_graph(:previous_album=>{:previous_album=>:previous_album}).sql.should == "SELECT albums.id, albums.band_id, previous_album.id AS previous_album_id, previous_album.band_id AS previous_album_band_id, previous_album_0.id AS previous_album_0_id, previous_album_0.band_id AS previous_album_0_band_id, previous_album_1.id AS previous_album_1_id, previous_album_1.band_id AS previous_album_1_band_id FROM albums LEFT OUTER JOIN albums AS previous_album ON (previous_album.id = albums.previous_album_id) LEFT OUTER JOIN albums AS previous_album_0 ON (previous_album_0.id = previous_album.previous_album_id) LEFT OUTER JOIN albums AS previous_album_1 ON (previous_album_1.id = previous_album_0.previous_album_id)"
+  end
+
+  it "should respect the association's :order" do
+    GraphAlbum.one_to_many :right_tracks, :class=>'GraphTrack', :key=>:album_id, :order=>[:id, :album_id]
+    GraphAlbum.eager_graph(:right_tracks).sql.should == 'SELECT albums.id, albums.band_id, right_tracks.id AS right_tracks_id, right_tracks.album_id FROM albums LEFT OUTER JOIN tracks AS right_tracks ON (right_tracks.album_id = albums.id) ORDER BY right_tracks.id, right_tracks.album_id'
+  end
+
+  it "should only qualify unqualified symbols, identifiers, or ordered versions in association's :order" do
+    GraphAlbum.one_to_many :right_tracks, :class=>'GraphTrack', :key=>:album_id, :order=>[:blah__id.identifier, :blah__id.identifier.desc, :blah__id.desc, :blah__id, :album_id, :album_id.desc, 1, 'RANDOM()'.lit, :a.qualify(:b)]
+    GraphAlbum.eager_graph(:right_tracks).sql.should == 'SELECT albums.id, albums.band_id, right_tracks.id AS right_tracks_id, right_tracks.album_id FROM albums LEFT OUTER JOIN tracks AS right_tracks ON (right_tracks.album_id = albums.id) ORDER BY right_tracks.blah__id, right_tracks.blah__id DESC, blah.id DESC, blah.id, right_tracks.album_id, right_tracks.album_id DESC, 1, RANDOM(), b.a'
+  end
+
+  it "should not respect the association's :order if :order_eager_graph is false" do
+    GraphAlbum.one_to_many :right_tracks, :class=>'GraphTrack', :key=>:album_id, :order=>[:id, :album_id], :order_eager_graph=>false
+    GraphAlbum.eager_graph(:right_tracks).sql.should == 'SELECT albums.id, albums.band_id, right_tracks.id AS right_tracks_id, right_tracks.album_id FROM albums LEFT OUTER JOIN tracks AS right_tracks ON (right_tracks.album_id = albums.id)'
+  end
+
+  it "should add the association's :order to the existing order" do
+    GraphAlbum.one_to_many :right_tracks, :class=>'GraphTrack', :key=>:album_id, :order=>[:id, :album_id]
+    GraphAlbum.order(:band_id).eager_graph(:right_tracks).sql.should == 'SELECT albums.id, albums.band_id, right_tracks.id AS right_tracks_id, right_tracks.album_id FROM albums LEFT OUTER JOIN tracks AS right_tracks ON (right_tracks.album_id = albums.id) ORDER BY band_id, right_tracks.id, right_tracks.album_id'
+  end
+
+  it "should add the association's :order for cascading associations" do
+    GraphBand.one_to_many :a_albums, :class=>'GraphAlbum', :key=>:band_id, :order=>:name
+    GraphAlbum.one_to_many :b_tracks, :class=>'GraphTrack', :key=>:album_id, :order=>[:id, :album_id]
+    GraphBand.eager_graph(:a_albums=>:b_tracks).sql.should == 'SELECT bands.id, bands.vocalist_id, a_albums.id AS a_albums_id, a_albums.band_id, b_tracks.id AS b_tracks_id, b_tracks.album_id FROM bands LEFT OUTER JOIN albums AS a_albums ON (a_albums.band_id = bands.id) LEFT OUTER JOIN tracks AS b_tracks ON (b_tracks.album_id = a_albums.id) ORDER BY a_albums.name, b_tracks.id, b_tracks.album_id'
+    GraphAlbum.one_to_many :albums, :class=>'GraphAlbum', :key=>:band_id, :order=>[:band_id, :id]
+    GraphAlbum.eager_graph(:albums=>{:albums=>:albums}).sql.should == 'SELECT albums.id, albums.band_id, albums_0.id AS albums_0_id, albums_0.band_id AS albums_0_band_id, albums_1.id AS albums_1_id, albums_1.band_id AS albums_1_band_id, albums_2.id AS albums_2_id, albums_2.band_id AS albums_2_band_id FROM albums LEFT OUTER JOIN albums AS albums_0 ON (albums_0.band_id = albums.id) LEFT OUTER JOIN albums AS albums_1 ON (albums_1.band_id = albums_0.id) LEFT OUTER JOIN albums AS albums_2 ON (albums_2.band_id = albums_1.id) ORDER BY albums_0.band_id, albums_0.id, albums_1.band_id, albums_1.id, albums_2.band_id, albums_2.id'
+  end
+
+  it "should add the associations :order for multiple associations" do
+    GraphAlbum.many_to_many :a_genres, :class=>'GraphGenre', :left_key=>:album_id, :right_key=>:genre_id, :join_table=>:ag, :order=>:id
+    GraphAlbum.one_to_many :b_tracks, :class=>'GraphTrack', :key=>:album_id, :order=>[:id, :album_id]
+    GraphAlbum.eager_graph(:a_genres, :b_tracks).sql.should == 'SELECT albums.id, albums.band_id, a_genres.id AS a_genres_id, b_tracks.id AS b_tracks_id, b_tracks.album_id FROM albums LEFT OUTER JOIN ag ON (ag.album_id = albums.id) LEFT OUTER JOIN genres AS a_genres ON (a_genres.id = ag.genre_id) LEFT OUTER JOIN tracks AS b_tracks ON (b_tracks.album_id = albums.id) ORDER BY a_genres.id, b_tracks.id, b_tracks.album_id'
+  end
+
+  it "should use the correct qualifier when graphing multiple tables with extra conditions" do
+    GraphAlbum.many_to_many :a_genres, :class=>'GraphGenre', :left_key=>:album_id, :right_key=>:genre_id, :join_table=>:ag
+    GraphAlbum.one_to_many :b_tracks, :class=>'GraphTrack', :key=>:album_id, :graph_conditions=>{:a=>:b}
+    GraphAlbum.eager_graph(:a_genres, :b_tracks).sql.should == 'SELECT albums.id, albums.band_id, a_genres.id AS a_genres_id, b_tracks.id AS b_tracks_id, b_tracks.album_id FROM albums LEFT OUTER JOIN ag ON (ag.album_id = albums.id) LEFT OUTER JOIN genres AS a_genres ON (a_genres.id = ag.genre_id) LEFT OUTER JOIN tracks AS b_tracks ON ((b_tracks.album_id = albums.id) AND (b_tracks.a = albums.b))'
+  end
+end
diff -pruN 1.5.1-1/spec/model/hooks_spec.rb 3.0.0-1/spec/model/hooks_spec.rb
--- 1.5.1-1/spec/model/hooks_spec.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/model/hooks_spec.rb	2009-04-09 19:57:40.000000000 +0100
@@ -0,0 +1,239 @@
+require File.join(File.dirname(__FILE__), "spec_helper")
+
+describe "Model#after_initialize" do
+  specify "should be called after initialization" do
+    $values1 = nil
+    $reached_after_initialized = false
+    
+    a = Class.new(Sequel::Model)
+    a.class_eval do
+      columns :x, :y
+      def after_initialize
+        $values1 = @values.clone
+        $reached_after_initialized = true
+      end
+    end
+    
+    a.new(:x => 1, :y => 2)
+    $values1.should == {:x => 1, :y => 2}
+    $reached_after_initialized.should == true
+  end
+end
+
+describe "Model#before_create && Model#after_create" do
+  before do
+    MODEL_DB.reset
+
+    @c = Class.new(Sequel::Model(:items))
+    @c.class_eval do
+      columns :x
+      no_primary_key
+      
+      def after_create
+        MODEL_DB << "BLAH after"
+      end
+    end
+  end
+  
+  specify "should be called around new record creation" do
+    @c.send(:define_method, :before_create){MODEL_DB << "BLAH before"}
+    @c.create(:x => 2)
+    MODEL_DB.sqls.should == [
+      'BLAH before',
+      'INSERT INTO items (x) VALUES (2)',
+      'BLAH after'
+    ]
+  end
+
+  specify ".create should cancel the save and raise an error if before_create returns false and raise_on_save_failure is true" do
+    @c.send(:define_method, :before_create){false}
+    proc{@c.load(:id => 2233).save}.should_not raise_error(Sequel::ValidationFailed)
+    proc{@c.create(:x => 2)}.should raise_error(Sequel::BeforeHookFailed)
+    MODEL_DB.sqls.should == []
+  end
+
+  specify ".create should cancel the save and return nil if before_create returns false and raise_on_save_failure is false" do
+    @c.send(:define_method, :before_create){false}
+    @c.raise_on_save_failure = false
+    @c.create(:x => 2).should == nil
+    MODEL_DB.sqls.should == []
+  end
+end
+
+describe "Model#before_update && Model#after_update" do
+  before do
+    MODEL_DB.reset
+
+    @c = Class.new(Sequel::Model(:items))
+    @c.class_eval do
+      def after_update; MODEL_DB << "BLAH after" end
+    end
+  end
+  
+  specify "should be called around record update" do
+    @c.send(:define_method, :before_update){MODEL_DB << "BLAH before"}
+    m = @c.load(:id => 2233)
+    m.save
+    MODEL_DB.sqls.should == [
+      'BLAH before',
+      'UPDATE items SET id = 2233 WHERE (id = 2233)',
+      'BLAH after'
+    ]
+  end
+
+  specify "#save should cancel the save and raise an error if before_update returns false and raise_on_save_failure is true" do
+    @c.send(:define_method, :before_update){false}
+    proc{@c.load(:id => 2233).save}.should_not raise_error(Sequel::ValidationFailed)
+    proc{@c.load(:id => 2233).save}.should raise_error(Sequel::BeforeHookFailed)
+    MODEL_DB.sqls.should == []
+  end
+
+  specify "#save should cancel the save and return nil if before_update returns false and raise_on_save_failure is false" do
+    @c.send(:define_method, :before_update){false}
+    @c.raise_on_save_failure = false
+    @c.load(:id => 2233).save.should == nil
+    MODEL_DB.sqls.should == []
+  end
+end
+
+describe "Model#before_save && Model#after_save" do
+  before do
+    MODEL_DB.reset
+
+    @c = Class.new(Sequel::Model(:items))
+    @c.class_eval do
+      columns :x
+      def after_save; MODEL_DB << "BLAH after" end
+    end
+  end
+  
+  specify "should be called around record update" do
+    @c.send(:define_method, :before_save){MODEL_DB << "BLAH before"}
+    m = @c.load(:id => 2233)
+    m.save
+    MODEL_DB.sqls.should == [
+      'BLAH before',
+      'UPDATE items SET id = 2233 WHERE (id = 2233)',
+      'BLAH after'
+    ]
+  end
+  
+  specify "should be called around record creation" do
+    @c.send(:define_method, :before_save){MODEL_DB << "BLAH before"}
+    @c.no_primary_key
+    @c.create(:x => 2)
+    MODEL_DB.sqls.should == [
+      'BLAH before',
+      'INSERT INTO items (x) VALUES (2)',
+      'BLAH after'
+    ]
+  end
+
+  specify "#save should cancel the save and raise an error if before_save returns false and raise_on_save_failure is true" do
+    @c.send(:define_method, :before_save){false}
+    proc{@c.load(:id => 2233).save}.should_not raise_error(Sequel::ValidationFailed)
+    proc{@c.load(:id => 2233).save}.should raise_error(Sequel::BeforeHookFailed)
+    MODEL_DB.sqls.should == []
+  end
+
+  specify "#save should cancel the save and return nil if before_save returns false and raise_on_save_failure is false" do
+    @c.send(:define_method, :before_save){false}
+    @c.raise_on_save_failure = false
+    @c.load(:id => 2233).save.should == nil
+    MODEL_DB.sqls.should == []
+  end
+end
+
+describe "Model#before_destroy && Model#after_destroy" do
+  before do
+    MODEL_DB.reset
+
+    @c = Class.new(Sequel::Model(:items))
+    @c.class_eval do
+      def after_destroy; MODEL_DB << "BLAH after"; end
+      
+      def delete
+        MODEL_DB << "DELETE BLAH"
+      end
+    end
+  end
+  
+  specify "should be called around record destruction" do
+    @c.send(:define_method, :before_destroy){MODEL_DB << "BLAH before"}
+    m = @c.load(:id => 2233)
+    m.destroy
+    MODEL_DB.sqls.should == [
+      'BLAH before',
+      'DELETE BLAH',
+      'BLAH after'
+    ]
+  end
+
+  specify "#destroy should cancel the destroy and raise an error if before_destroy returns false and raise_on_save_failure is true" do
+    @c.send(:define_method, :before_destroy){false}
+    proc{@c.load(:id => 2233).destroy}.should raise_error(Sequel::BeforeHookFailed)
+    MODEL_DB.sqls.should == []
+  end
+
+  specify "#destroy should cancel the destroy and return nil if before_destroy returns false and raise_on_save_failure is false" do
+    @c.send(:define_method, :before_destroy){false}
+    @c.raise_on_save_failure = false
+    @c.load(:id => 2233).destroy.should == nil
+    MODEL_DB.sqls.should == []
+  end
+end
+
+describe "Model#before_validation && Model#after_validation" do
+  before do
+    MODEL_DB.reset
+
+    @c = Class.new(Sequel::Model(:items))
+    @c.class_eval do
+      def after_validation; MODEL_DB << "BLAH after" end
+
+      def validate
+        errors.add(:id, 'not valid') unless id == 2233
+      end
+      columns :id
+    end
+  end
+  
+  specify "should be called around validation" do
+    @c.send(:define_method, :before_validation){MODEL_DB << "BLAH before"}
+    m = @c.load(:id => 2233)
+    m.should be_valid
+    MODEL_DB.sqls.should == ['BLAH before', 'BLAH after']
+
+    MODEL_DB.sqls.clear
+    m = @c.load(:id => 22)
+    m.should_not be_valid
+    MODEL_DB.sqls.should == ['BLAH before', 'BLAH after']
+  end
+
+  specify "should be called when calling save" do
+    @c.send(:define_method, :before_validation){MODEL_DB << "BLAH before"}
+    m = @c.load(:id => 2233)
+    m.save.should == m
+    MODEL_DB.sqls.should == ['BLAH before', 'BLAH after', 'UPDATE items SET id = 2233 WHERE (id = 2233)']
+
+    MODEL_DB.sqls.clear
+    m = @c.load(:id => 22)
+    m.raise_on_save_failure = false
+    m.save.should == nil
+    MODEL_DB.sqls.should == ['BLAH before', 'BLAH after']
+  end
+
+  specify "#save should cancel the save and raise an error if before_validation returns false and raise_on_save_failure is true" do
+    @c.send(:define_method, :before_validation){false}
+    proc{@c.load(:id => 2233).save}.should_not raise_error(Sequel::ValidationFailed)
+    proc{@c.load(:id => 2233).save}.should raise_error(Sequel::BeforeHookFailed)
+    MODEL_DB.sqls.should == []
+  end
+
+  specify "#save should cancel the save and return nil if before_validation returns false and raise_on_save_failure is false" do
+    @c.send(:define_method, :before_validation){false}
+    @c.raise_on_save_failure = false
+    @c.load(:id => 2233).save.should == nil
+    MODEL_DB.sqls.should == []
+  end
+end
diff -pruN 1.5.1-1/spec/model/inflector_spec.rb 3.0.0-1/spec/model/inflector_spec.rb
--- 1.5.1-1/spec/model/inflector_spec.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/model/inflector_spec.rb	2009-04-15 18:24:09.000000000 +0100
@@ -0,0 +1,26 @@
+require File.join(File.dirname(__FILE__), 'spec_helper')
+
+describe Sequel::Inflections do
+  before do
+    @plurals, @singulars, @uncountables = Sequel.inflections.plurals.dup, Sequel.inflections.singulars.dup, Sequel.inflections.uncountables.dup
+  end
+  after do
+    Sequel.inflections.plurals.replace(@plurals)
+    Sequel.inflections.singulars.replace(@singulars)
+    Sequel.inflections.uncountables.replace(@uncountables)
+  end
+
+  it "should be possible to clear the list of singulars, plurals, and uncountables" do
+    Sequel.inflections.clear(:plurals)
+    Sequel.inflections.plurals.should == []
+    Sequel.inflections.plural('blah', 'blahs')
+    Sequel.inflections.clear
+    Sequel.inflections.plurals.should == []
+    Sequel.inflections.singulars.should == []
+    Sequel.inflections.uncountables.should == []
+  end
+
+  it "should be yielded and returned by Sequel.inflections" do
+    Sequel.inflections{|i| i.should == Sequel::Inflections}.should == Sequel::Inflections
+  end
+end
diff -pruN 1.5.1-1/spec/model/model_spec.rb 3.0.0-1/spec/model/model_spec.rb
--- 1.5.1-1/spec/model/model_spec.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/model/model_spec.rb	2009-04-09 18:31:20.000000000 +0100
@@ -0,0 +1,473 @@
+require File.join(File.dirname(__FILE__), "spec_helper")
+
+describe Sequel::Model do
+  it "should have class method aliased as model" do
+    Sequel::Model.instance_methods.collect{|x| x.to_s}.should include("model")
+
+    model_a = Class.new(Sequel::Model(:items))
+    model_a.new.model.should be(model_a)
+  end
+
+  it "should be associated with a dataset" do
+    model_a = Class.new(Sequel::Model) { set_dataset MODEL_DB[:as] }
+
+    model_a.dataset.should be_a_kind_of(MockDataset)
+    model_a.dataset.opts[:from].should == [:as]
+
+    model_b = Class.new(Sequel::Model) { set_dataset MODEL_DB[:bs] }
+
+    model_b.dataset.should be_a_kind_of(MockDataset)
+    model_b.dataset.opts[:from].should == [:bs]
+
+    model_a.dataset.opts[:from].should == [:as]
+  end
+
+end
+
+describe Sequel::Model, "dataset & schema" do
+  before do
+    @model = Class.new(Sequel::Model(:items))
+  end
+
+  it "creates dynamic model subclass with set table name" do
+    @model.table_name.should == :items
+  end
+
+  it "defaults to primary key of id" do
+    @model.primary_key.should == :id
+  end
+
+  it "allow primary key change" do
+    @model.set_primary_key :ssn
+    @model.primary_key.should == :ssn
+  end
+
+  it "allows dataset change" do
+    @model.set_dataset(MODEL_DB[:foo])
+    @model.table_name.should == :foo
+  end
+
+  it "set_dataset should take a symbol" do
+    @model.db = MODEL_DB
+    @model.set_dataset(:foo)
+    @model.table_name.should == :foo
+  end
+
+  it "set_dataset should raise an error unless given a Symbol or Dataset" do
+    proc{@model.set_dataset(Object.new)}.should raise_error(Sequel::Error)
+  end
+
+  it "set_dataset should add the destroy method to the dataset" do
+    ds = MODEL_DB[:foo]
+    ds.should_not respond_to(:destroy)
+    @model.set_dataset(ds)
+    ds.should respond_to(:destroy)
+  end
+
+  it "doesn't raise an error on set_dataset if there is an error raised getting the schema" do
+    @model.meta_def(:get_db_schema){raise Sequel::Error}
+    proc{@model.set_dataset(MODEL_DB[:foo])}.should_not raise_error
+  end
+
+  it "doesn't raise an error on inherited if there is an error setting the dataset" do
+    @model.meta_def(:set_dataset){raise Sequel::Error}
+    proc{Class.new(@model)}.should_not raise_error
+  end
+end
+
+describe Sequel::Model, "constructor" do
+  
+  before(:each) do
+    @m = Class.new(Sequel::Model)
+    @m.columns :a, :b
+  end
+
+  it "should accept a hash" do
+    m = @m.new(:a => 1, :b => 2)
+    m.values.should == {:a => 1, :b => 2}
+    m.should be_new
+  end
+  
+  it "should accept a block and yield itself to the block" do
+    block_called = false
+    m = @m.new {|i| block_called = true; i.should be_a_kind_of(@m); i.values[:a] = 1}
+    
+    block_called.should be_true
+    m.values[:a].should == 1
+  end
+  
+end
+
+describe Sequel::Model, "new" do
+
+  before(:each) do
+    @m = Class.new(Sequel::Model) do
+      set_dataset MODEL_DB[:items]
+      columns :x, :id
+    end
+  end
+
+  it "should be marked as new?" do
+    o = @m.new
+    o.should be_new
+  end
+
+  it "should not be marked as new? once it is saved" do
+    o = @m.new(:x => 1)
+    o.should be_new
+    o.save
+    o.should_not be_new
+  end
+
+  it "should use the last inserted id as primary key if not in values" do
+    d = @m.dataset
+    def d.insert(*args)
+      super
+      1234
+    end
+
+    def d.first
+      {:x => 1, :id => 1234}
+    end
+
+    o = @m.new(:x => 1)
+    o.save
+    o.id.should == 1234
+
+    o = @m.load(:x => 1, :id => 333)
+    o.save
+    o.id.should == 333
+  end
+
+end
+
+describe Sequel::Model, ".subset" do
+  before do
+    MODEL_DB.reset
+
+    @c = Class.new(Sequel::Model(:items))
+  end
+
+  specify "should create a filter on the underlying dataset" do
+    proc {@c.new_only}.should raise_error(NoMethodError)
+    
+    @c.subset(:new_only) {:age.sql_number < 'new'}
+    
+    @c.new_only.sql.should == "SELECT * FROM items WHERE (age < 'new')"
+    @c.dataset.new_only.sql.should == "SELECT * FROM items WHERE (age < 'new')"
+    
+    @c.subset(:pricey) {:price.sql_number > 100}
+    
+    @c.pricey.sql.should == "SELECT * FROM items WHERE (price > 100)"
+    @c.dataset.pricey.sql.should == "SELECT * FROM items WHERE (price > 100)"
+    
+    @c.pricey.new_only.sql.should == "SELECT * FROM items WHERE ((price > 100) AND (age < 'new'))"
+    @c.new_only.pricey.sql.should == "SELECT * FROM items WHERE ((age < 'new') AND (price > 100))"
+  end
+
+  specify "should not override existing model methods" do
+    @c.meta_def(:active){true}
+    @c.subset(:active, :active)
+    @c.active.should == true
+  end
+end
+
+describe Sequel::Model, ".find" do
+
+  before(:each) do
+    MODEL_DB.reset
+    
+    @c = Class.new(Sequel::Model(:items))
+    
+    $cache_dataset_row = {:name => 'sharon', :id => 1}
+    @dataset = @c.dataset
+    $sqls = []
+    @dataset.extend(Module.new {
+      def fetch_rows(sql)
+        $sqls << sql
+        yield $cache_dataset_row
+      end
+    })
+  end
+  
+  it "should return the first record matching the given filter" do
+    @c.find(:name => 'sharon').should be_a_kind_of(@c)
+    $sqls.last.should == "SELECT * FROM items WHERE (name = 'sharon') LIMIT 1"
+
+    @c.find(:name.like('abc%')).should be_a_kind_of(@c)
+    $sqls.last.should == "SELECT * FROM items WHERE (name LIKE 'abc%') LIMIT 1"
+  end
+  
+  specify "should accept filter blocks" do
+    @c.find{:id.sql_number > 1}.should be_a_kind_of(@c)
+    $sqls.last.should == "SELECT * FROM items WHERE (id > 1) LIMIT 1"
+
+    @c.find {(:x.sql_number > 1) & (:y.sql_number < 2)}.should be_a_kind_of(@c)
+    $sqls.last.should == "SELECT * FROM items WHERE ((x > 1) AND (y < 2)) LIMIT 1"
+  end
+
+end
+
+describe Sequel::Model, ".fetch" do
+
+  before(:each) do
+    MODEL_DB.reset
+    @c = Class.new(Sequel::Model(:items))
+  end
+  
+  it "should return instances of Model" do
+    @c.fetch("SELECT * FROM items").first.should be_a_kind_of(@c)
+  end
+
+  it "should return true for .empty? and not raise an error on empty selection" do
+    rows = @c.fetch("SELECT * FROM items WHERE FALSE")
+    @c.send(:define_method, :fetch_rows){|sql| yield({:count => 0})}
+    proc {rows.empty?}.should_not raise_error
+  end
+end
+
+describe Sequel::Model, ".find_or_create" do
+
+  before(:each) do
+    MODEL_DB.reset
+    @c = Class.new(Sequel::Model(:items)) do
+      no_primary_key
+      columns :x
+    end
+  end
+
+  it "should find the record" do
+    @c.find_or_create(:x => 1)
+    MODEL_DB.sqls.should == ["SELECT * FROM items WHERE (x = 1) LIMIT 1"]
+    
+    MODEL_DB.reset
+  end
+  
+  it "should create the record if not found" do
+    @c.meta_def(:find) do |*args|
+      dataset.filter(*args).first
+      nil
+    end
+    
+    @c.find_or_create(:x => 1)
+    MODEL_DB.sqls.should == [
+      "SELECT * FROM items WHERE (x = 1) LIMIT 1",
+      "INSERT INTO items (x) VALUES (1)"
+    ]
+  end
+end
+
+describe Sequel::Model, ".all" do
+  
+  before(:each) do
+    MODEL_DB.reset
+    @c = Class.new(Sequel::Model(:items)) do
+      no_primary_key
+    end
+    
+    @c.dataset.meta_def(:all) {1234}
+  end
+  
+  it "should return all records in the dataset" do
+    @c.all.should == 1234
+  end
+  
+end
+
+class DummyModelBased < Sequel::Model(:blog)
+end
+
+describe Sequel::Model, "(:tablename)" do
+
+  it "should allow reopening of descendant classes" do
+    proc do
+      eval "class DummyModelBased < Sequel::Model(:blog); end"
+    end.should_not raise_error
+  end
+
+end
+
+describe Sequel::Model, "A model class without a primary key" do
+
+  before(:each) do
+    MODEL_DB.reset
+    @c = Class.new(Sequel::Model(:items)) do
+      columns :x
+      no_primary_key
+    end
+  end
+
+  it "should be able to insert records without selecting them back" do
+    i = nil
+    proc {i = @c.create(:x => 1)}.should_not raise_error
+    i.class.should be(@c)
+    i.values.to_hash.should == {:x => 1}
+
+    MODEL_DB.sqls.should == ['INSERT INTO items (x) VALUES (1)']
+  end
+
+  it "should raise when deleting" do
+    o = @c.new
+    proc {o.delete}.should raise_error
+  end
+
+  it "should insert a record when saving" do
+    o = @c.new(:x => 2)
+    o.should be_new
+    o.save
+    MODEL_DB.sqls.should == ['INSERT INTO items (x) VALUES (2)']
+  end
+
+end
+
+describe Sequel::Model, "attribute accessors" do
+  before do
+    MODEL_DB.reset
+    @dataset = Sequel::Dataset.new(MODEL_DB)
+    def @dataset.columns; [:x, :y]; end
+    @c = Class.new(Sequel::Model) do
+      def self.db_schema
+         set_columns(Array(@columns))
+        @db_schema = {:x=>{}, :y=>{}}
+      end
+      def self.set_dataset(ds, opts={}) 
+        @columns = ds.columns
+        db_schema
+      end
+    end
+  end
+
+  it "should be created on set_dataset" do
+    %w'x y x= y='.each do |x|
+      @c.instance_methods.collect{|y| y.to_s}.should_not include(x)
+    end
+    @c.set_dataset(@dataset)
+    %w'x y x= y='.each do |x|
+      @c.instance_methods.collect{|y| y.to_s}.should include(x)
+    end
+    o = @c.new
+    %w'x y x= y='.each do |x|
+      o.methods.collect{|y| y.to_s}.should include(x)
+    end
+
+    o.x.should be_nil
+    o.x = 34
+    o.x.should == 34
+  end
+
+  it "should be only accept one argument for the write accessor" do
+    @c.set_dataset(@dataset)
+    o = @c.new
+
+    o.x = 34
+    o.x.should == 34
+    proc{o.send(:x=)}.should raise_error
+    proc{o.send(:x=, 3, 4)}.should raise_error
+  end
+end
+
+describe Sequel::Model, ".[]" do
+
+  before(:each) do
+    MODEL_DB.reset
+
+    @c = Class.new(Sequel::Model(:items))
+
+    $cache_dataset_row = {:name => 'sharon', :id => 1}
+    @dataset = @c.dataset
+    $sqls = []
+    @dataset.extend(Module.new {
+      def fetch_rows(sql)
+        $sqls << sql
+        yield $cache_dataset_row
+      end
+    })
+  end
+
+  it "should return the first record for the given pk" do
+    @c[1].should be_a_kind_of(@c)
+    $sqls.last.should == "SELECT * FROM items WHERE (id = 1) LIMIT 1"
+    @c[9999].should be_a_kind_of(@c)
+    $sqls.last.should == "SELECT * FROM items WHERE (id = 9999) LIMIT 1"
+  end
+
+  it "should work correctly for custom primary key" do
+    @c.set_primary_key :name
+    @c['sharon'].should be_a_kind_of(@c)
+    $sqls.last.should == "SELECT * FROM items WHERE (name = 'sharon') LIMIT 1"
+  end
+
+  it "should work correctly for composite primary key" do
+    @c.set_primary_key [:node_id, :kind]
+    @c[3921, 201].should be_a_kind_of(@c)
+    $sqls.last.should =~ \
+    /^SELECT \* FROM items WHERE \((\(node_id = 3921\) AND \(kind = 201\))|(\(kind = 201\) AND \(node_id = 3921\))\) LIMIT 1$/
+  end
+end
+
+context "Model#inspect" do
+  before do
+    @o = Sequel::Model.load(:x => 333)
+  end
+  
+  specify "should include the class name and the values" do
+    @o.inspect.should == '#<Sequel::Model @values={:x=>333}>'
+  end
+end
+
+context "Model.db_schema" do
+  before do
+    @c = Class.new(Sequel::Model(:items)) do
+      def self.columns; orig_columns; end
+    end
+    @dataset = Sequel::Dataset.new(nil).from(:items)
+    @dataset.meta_def(:db){@db ||= Sequel::Database.new}
+    def @dataset.naked; self; end
+    def @dataset.columns; []; end
+    def @dataset.def_mutation_method(*names);  end
+  end
+  
+  specify "should use the database's schema_for_table and set the columns and dataset columns" do
+    d = @dataset.db
+    def d.schema(table, opts = {})
+      [[:x, {:type=>:integer}], [:y, {:type=>:string}]]
+    end
+    @c.dataset = @dataset
+    @c.db_schema.should == {:x=>{:type=>:integer}, :y=>{:type=>:string}}
+    @c.columns.should == [:x, :y]
+    @c.dataset.instance_variable_get(:@columns).should == [:x, :y]
+  end
+
+  specify "should restrict the schema and columns for datasets with a :select option" do
+    ds = @dataset.select(:x, :y___z)
+    d = ds.db
+    def d.schema(table, opts = {})
+      [[:x, {:type=>:integer}], [:y, {:type=>:string}]]
+    end
+    def @c.columns; [:x, :z]; end
+    @c.dataset = ds
+    @c.db_schema.should == {:x=>{:type=>:integer}, :z=>{}}
+  end
+
+  specify "should not use schema_for_table if the dataset uses multiple tables or custom sql" do
+    ds = @dataset.join(:x, :id)
+    d = ds.db
+    e = false
+    d.meta_def(:schema){|table| e = true}
+    def @c.columns; [:x]; end
+    @c.dataset = ds
+    @c.db_schema.should == {:x=>{}}
+    e.should == false
+  end
+
+  specify "should fallback to fetching records if schema_for_table raises an error" do
+    ds = @dataset.join(:x, :id)
+    d = ds.db
+    def d.schema(table)
+      raise StandardError
+    end
+    def @c.columns; [:x]; end
+    @c.dataset = ds
+    @c.db_schema.should == {:x=>{}}
+  end
+end
diff -pruN 1.5.1-1/spec/model/plugins_spec.rb 3.0.0-1/spec/model/plugins_spec.rb
--- 1.5.1-1/spec/model/plugins_spec.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/model/plugins_spec.rb	2009-04-09 18:31:41.000000000 +0100
@@ -0,0 +1,72 @@
+require File.join(File.dirname(__FILE__), "spec_helper")
+
+module Sequel::Plugins
+
+  module Timestamped
+    def self.apply(m, opts)
+      m.meta_def(:stamp_opts) {opts}
+      m.send(:define_method, :before_save){@values[:stamp] = Time.now}
+    end
+    
+    module InstanceMethods
+      def get_stamp(*args); @values[:stamp] end
+      def abc; timestamped_opts; end
+    end
+    
+    module ClassMethods
+      def deff; timestamped_opts; end
+    end
+
+    module DatasetMethods
+      def ghi; timestamped_opts; end
+    end
+  end
+
+end
+
+describe Sequel::Model, "using a plugin" do
+
+  it "should fail if the plugin is not found" do
+    proc do
+      c = Class.new(Sequel::Model)
+      c.class_eval do
+        plugin :something_or_other
+      end
+    end.should raise_error(LoadError)
+  end
+  
+  it "should apply the plugin to the class" do
+    c = nil
+    proc do
+      c = Class.new(Sequel::Model)
+      c.class_eval do
+        set_dataset MODEL_DB[:items]
+        plugin :timestamped, :a => 1, :b => 2
+      end
+    end.should_not raise_error(LoadError)
+    
+    c.should respond_to(:stamp_opts)
+    c.stamp_opts.should == {:a => 1, :b => 2}
+    
+    # instance methods
+    m = c.new
+    m.should respond_to(:get_stamp)
+    m.should respond_to(:abc)
+    m.abc.should == {:a => 1, :b => 2}
+    t = Time.now
+    m[:stamp] = t
+    m.get_stamp.should == t
+    
+    # class methods
+    c.should respond_to(:deff)
+    c.deff.should == {:a => 1, :b => 2}
+    
+    # dataset methods
+    c.dataset.should respond_to(:ghi)
+    c.dataset.ghi.should == {:a => 1, :b => 2}
+
+    # dataset methods called on the class
+    c.should respond_to(:ghi)
+    c.ghi.should == {:a => 1, :b => 2}
+  end
+end
diff -pruN 1.5.1-1/spec/model/record_spec.rb 3.0.0-1/spec/model/record_spec.rb
--- 1.5.1-1/spec/model/record_spec.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/model/record_spec.rb	2009-04-29 00:55:11.000000000 +0100
@@ -0,0 +1,1185 @@
+require File.join(File.dirname(__FILE__), "spec_helper")
+
+describe "Model#save server use" do
+  
+  before(:each) do
+    @c = Class.new(Sequel::Model(:items)) do
+      columns :id, :x, :y
+    end
+    @c.db = MockDatabase.new
+    db2 = @db2 = MockDatabase.new
+    @c.class_eval do
+      define_method(:after_save) do
+        model.db = db2
+        ds = model.dataset
+        def ds.fetch_rows(sql)
+          yield @db.execute(sql, @opts[:server])
+        end
+        @this = nil
+      end
+    end
+  end
+
+  it "should use the :default server if the model doesn't have one already specified" do
+    @c.db.should_receive(:execute).with("INSERT INTO items (x) VALUES (1)").and_return(10)
+    @db2.should_receive(:execute).with('SELECT * FROM items WHERE (id = 10) LIMIT 1', :default).and_return(:x=>1, :id=>10)
+    @c.new(:x=>1).save
+  end
+
+  it "should use the model's server if the model has one already specified" do
+    @c.dataset = @c.dataset.server(:blah)
+    @c.db.should_receive(:execute).with("INSERT INTO items (x) VALUES (1)").and_return(10)
+    @db2.should_receive(:execute).with('SELECT * FROM items WHERE (id = 10) LIMIT 1', :blah).and_return(:x=>1, :id=>10)
+    @c.new(:x=>1).save
+  end
+end
+
+describe "Model#save" do
+  
+  before(:each) do
+    @c = Class.new(Sequel::Model(:items)) do
+      columns :id, :x, :y
+    end
+    MODEL_DB.reset
+  end
+  
+  it "should insert a record for a new model instance" do
+    o = @c.new(:x => 1)
+    o.save
+    
+    MODEL_DB.sqls.first.should == "INSERT INTO items (x) VALUES (1)"
+  end
+
+  it "should use dataset's insert_select method if present" do
+    ds = @c.dataset = @c.dataset.clone
+    def ds.insert_select(hash)
+      execute("INSERT INTO items (y) VALUES (2)")
+      {:y=>2}
+    end
+    o = @c.new(:x => 1)
+    o.save
+    
+    o.values.should == {:y=>2}
+    MODEL_DB.sqls.first.should == "INSERT INTO items (y) VALUES (2)"
+  end
+
+  it "should update a record for an existing model instance" do
+    o = @c.load(:id => 3, :x => 1)
+    o.save
+    
+    MODEL_DB.sqls.first.should =~ 
+      /UPDATE items SET (id = 3, x = 1|x = 1, id = 3) WHERE \(id = 3\)/
+  end
+  
+  it "should update only the given columns if given" do
+    o = @c.load(:id => 3, :x => 1, :y => nil)
+    o.save(:y)
+    MODEL_DB.sqls.first.should == "UPDATE items SET y = NULL WHERE (id = 3)"
+  end
+  
+  it "should mark saved columns as not changed" do
+    o = @c.load(:id => 3, :x => 1, :y => nil)
+    o[:y] = 4
+    o.changed_columns.should == [:y]
+    o.save(:x)
+    o.changed_columns.should == [:y]
+    o.save(:y)
+    o.changed_columns.should == []
+  end
+
+  it "should store previous value of @new in @was_new and as well as the hash used for updating in @columns_updated until after hooks finish running" do
+    res = nil
+    @c.send(:define_method, :after_save){ res = [@columns_updated, @was_new]}
+    o = @c.new(:x => 1, :y => nil)
+    o[:x] = 2
+    o.save
+    res.should == [nil, true]
+    o.after_save
+    res.should == [nil, nil]
+
+    res = nil
+    o = @c.load(:id => 23,:x => 1, :y => nil)
+    o[:x] = 2
+    o.save
+    res.should == [{:id => 23,:x => 2, :y => nil}, nil]
+    o.after_save
+    res.should == [nil, nil]
+
+    res = nil
+    o = @c.load(:id => 23,:x => 2, :y => nil)
+    o[:x] = 2
+    o[:y] = 22
+    o.save(:x)
+    res.should == [{:x=>2},nil]
+    o.after_save
+    res.should == [nil, nil]
+  end
+  
+  it "should use Model's save_in_transaction setting by default" do
+    @c.use_transactions = true
+    @c.load(:id => 3, :x => 1, :y => nil).save(:y)
+    MODEL_DB.sqls.should == ["BEGIN", "UPDATE items SET y = NULL WHERE (id = 3)", "COMMIT"]
+    MODEL_DB.reset
+    @c.use_transactions = false
+    @c.load(:id => 3, :x => 1, :y => nil).save(:y)
+    MODEL_DB.sqls.should == ["UPDATE items SET y = NULL WHERE (id = 3)"]
+    MODEL_DB.reset
+  end
+
+  it "should inherit Model's save_in_transaction setting" do
+    @c.use_transactions = true
+    Class.new(@c).load(:id => 3, :x => 1, :y => nil).save(:y)
+    MODEL_DB.sqls.should == ["BEGIN", "UPDATE items SET y = NULL WHERE (id = 3)", "COMMIT"]
+    MODEL_DB.reset
+    @c.use_transactions = false
+    Class.new(@c).load(:id => 3, :x => 1, :y => nil).save(:y)
+    MODEL_DB.sqls.should == ["UPDATE items SET y = NULL WHERE (id = 3)"]
+    MODEL_DB.reset
+  end
+
+  it "should use object's save_in_transaction setting" do
+    o = @c.load(:id => 3, :x => 1, :y => nil)
+    o.use_transactions = false
+    @c.use_transactions = true
+    o.save(:y)
+    MODEL_DB.sqls.should == ["UPDATE items SET y = NULL WHERE (id = 3)"]
+    MODEL_DB.reset
+    o = @c.load(:id => 3, :x => 1, :y => nil)
+    o.use_transactions = true
+    @c.use_transactions = false 
+    o.save(:y)
+    MODEL_DB.sqls.should == ["BEGIN", "UPDATE items SET y = NULL WHERE (id = 3)", "COMMIT"]
+    MODEL_DB.reset
+  end
+
+  it "should use :transaction option if given" do
+    o = @c.load(:id => 3, :x => 1, :y => nil)
+    o.use_transactions = true
+    o.save(:y, :transaction=>false)
+    MODEL_DB.sqls.should == ["UPDATE items SET y = NULL WHERE (id = 3)"]
+    MODEL_DB.reset
+    o = @c.load(:id => 3, :x => 1, :y => nil)
+    o.use_transactions = false
+    o.save(:y, :transaction=>true)
+    MODEL_DB.sqls.should == ["BEGIN", "UPDATE items SET y = NULL WHERE (id = 3)", "COMMIT"]
+    MODEL_DB.reset
+  end
+end
+
+describe "Model#save_changes" do
+  
+  before(:each) do
+    MODEL_DB.reset
+
+    @c = Class.new(Sequel::Model(:items)) do
+      unrestrict_primary_key
+      columns :id, :x, :y
+    end
+  end
+  
+  it "should do nothing if no changed columns" do
+    o = @c.new(:id => 3, :x => 1, :y => nil)
+    o.save_changes
+    
+    MODEL_DB.sqls.should be_empty
+
+    o = @c.load(:id => 3, :x => 1, :y => nil)
+    o.save_changes
+    
+    MODEL_DB.sqls.should be_empty
+  end
+  
+  it "should update only changed columns" do
+    o = @c.load(:id => 3, :x => 1, :y => nil)
+    o.x = 2
+
+    o.save_changes
+    MODEL_DB.sqls.should == ["UPDATE items SET x = 2 WHERE (id = 3)"]
+    o.save_changes
+    o.save_changes
+    MODEL_DB.sqls.should == ["UPDATE items SET x = 2 WHERE (id = 3)"]
+    MODEL_DB.reset
+
+    o.y = 4
+    o.save_changes
+    MODEL_DB.sqls.should == ["UPDATE items SET y = 4 WHERE (id = 3)"]
+    o.save_changes
+    o.save_changes
+    MODEL_DB.sqls.should == ["UPDATE items SET y = 4 WHERE (id = 3)"]
+  end
+  
+  it "should not consider columns changed if the values did not change" do
+    o = @c.load(:id => 3, :x => 1, :y => nil)
+    o.x = 1
+
+    o.save_changes
+    MODEL_DB.sqls.should == []
+    o.x = 3
+    o.save_changes
+    MODEL_DB.sqls.should == ["UPDATE items SET x = 3 WHERE (id = 3)"]
+    MODEL_DB.reset
+
+    o[:y] = nil
+    o.save_changes
+    MODEL_DB.sqls.should == []
+    o[:y] = 4
+    o.save_changes
+    MODEL_DB.sqls.should == ["UPDATE items SET y = 4 WHERE (id = 3)"]
+  end
+
+  it "should update columns changed in a before_update hook" do
+    o = @c.load(:id => 3, :x => 1, :y => nil)
+    @c.send(:define_method, :before_update){self.x += 1}
+    o.save_changes
+    MODEL_DB.sqls.should == []
+    o.x = 2
+    o.save_changes
+    MODEL_DB.sqls.should == ["UPDATE items SET x = 3 WHERE (id = 3)"]
+    MODEL_DB.reset
+    o.save_changes
+    MODEL_DB.sqls.should == []
+    o.x = 4
+    o.save_changes
+    MODEL_DB.sqls.should == ["UPDATE items SET x = 5 WHERE (id = 3)"]
+    MODEL_DB.reset
+  end
+
+  it "should update columns changed in a before_save hook" do
+    o = @c.load(:id => 3, :x => 1, :y => nil)
+    @c.send(:define_method, :before_update){self.x += 1}
+    o.save_changes
+    MODEL_DB.sqls.should == []
+    o.x = 2
+    o.save_changes
+    MODEL_DB.sqls.should == ["UPDATE items SET x = 3 WHERE (id = 3)"]
+    MODEL_DB.reset
+    o.save_changes
+    MODEL_DB.sqls.should == []
+    o.x = 4
+    o.save_changes
+    MODEL_DB.sqls.should == ["UPDATE items SET x = 5 WHERE (id = 3)"]
+    MODEL_DB.reset
+  end
+end
+
+describe "Model#new?" do
+  
+  before(:each) do
+    MODEL_DB.reset
+
+    @c = Class.new(Sequel::Model(:items)) do
+      unrestrict_primary_key
+      columns :x
+    end
+  end
+  
+  it "should be true for a new instance" do
+    n = @c.new(:x => 1)
+    n.should be_new
+  end
+  
+  it "should be false after saving" do
+    n = @c.new(:x => 1)
+    n.save
+    n.should_not be_new
+  end
+end
+
+describe Sequel::Model, "w/ primary key" do
+  
+  it "should default to ':id'" do
+    model_a = Class.new Sequel::Model
+    model_a.primary_key.should be_equal(:id)
+  end
+
+  it "should be changed through 'set_primary_key'" do
+    model_a = Class.new(Sequel::Model) { set_primary_key :a }
+    model_a.primary_key.should be_equal(:a)
+  end
+
+  it "should support multi argument composite keys" do
+    model_a = Class.new(Sequel::Model) { set_primary_key :a, :b }
+    model_a.primary_key.should be_eql([:a, :b])
+  end
+
+  it "should accept single argument composite keys" do
+    model_a = Class.new(Sequel::Model) { set_primary_key [:a, :b] }
+    model_a.primary_key.should be_eql([:a, :b])
+  end
+  
+end
+
+describe Sequel::Model, "w/o primary key" do
+  it "should return nil for primary key" do
+    Class.new(Sequel::Model) { no_primary_key }.primary_key.should be_nil
+  end
+
+  it "should raise a Sequel::Error on 'this'" do
+    instance = Class.new(Sequel::Model) { no_primary_key }.new
+    proc { instance.this }.should raise_error(Sequel::Error)
+  end
+end
+
+describe Sequel::Model, "with this" do
+
+  before { @example = Class.new Sequel::Model(:examples); @example.columns :id, :a, :x, :y }
+
+  it "should return a dataset identifying the record" do
+    instance = @example.load :id => 3
+    instance.this.sql.should be_eql("SELECT * FROM examples WHERE (id = 3) LIMIT 1")
+  end
+
+  it "should support arbitary primary keys" do
+    @example.set_primary_key :a
+
+    instance = @example.load :a => 3
+    instance.this.sql.should be_eql("SELECT * FROM examples WHERE (a = 3) LIMIT 1")
+  end
+
+  it "should support composite primary keys" do
+    @example.set_primary_key :x, :y
+    instance = @example.load :x => 4, :y => 5
+
+    parts = [
+      'SELECT * FROM examples WHERE %s LIMIT 1',
+      '((x = 4) AND (y = 5))', 
+      '((y = 5) AND (x = 4))'
+    ].map { |expr| Regexp.escape expr }
+    regexp = Regexp.new parts.first % "(?:#{parts[1]}|#{parts[2]})"
+
+    instance.this.sql.should match(regexp)
+  end
+
+end
+
+describe "Model#pk" do
+  before(:each) do
+    @m = Class.new(Sequel::Model)
+    @m.columns :id, :x, :y
+  end
+  
+  it "should be default return the value of the :id column" do
+    m = @m.load(:id => 111, :x => 2, :y => 3)
+    m.pk.should == 111
+  end
+
+  it "should be return the primary key value for custom primary key" do
+    @m.set_primary_key :x
+    m = @m.load(:id => 111, :x => 2, :y => 3)
+    m.pk.should == 2
+  end
+
+  it "should be return the primary key value for composite primary key" do
+    @m.set_primary_key [:y, :x]
+    m = @m.load(:id => 111, :x => 2, :y => 3)
+    m.pk.should == [3, 2]
+  end
+
+  it "should raise if no primary key" do
+    @m.set_primary_key nil
+    m = @m.new(:id => 111, :x => 2, :y => 3)
+    proc {m.pk}.should raise_error(Sequel::Error)
+
+    @m.no_primary_key
+    m = @m.new(:id => 111, :x => 2, :y => 3)
+    proc {m.pk}.should raise_error(Sequel::Error)
+  end
+end
+
+describe "Model#pk_hash" do
+  before(:each) do
+    @m = Class.new(Sequel::Model)
+    @m.columns :id, :x, :y
+  end
+  
+  it "should be default return the value of the :id column" do
+    m = @m.load(:id => 111, :x => 2, :y => 3)
+    m.pk_hash.should == {:id => 111}
+  end
+
+  it "should be return the primary key value for custom primary key" do
+    @m.set_primary_key :x
+    m = @m.load(:id => 111, :x => 2, :y => 3)
+    m.pk_hash.should == {:x => 2}
+  end
+
+  it "should be return the primary key value for composite primary key" do
+    @m.set_primary_key [:y, :x]
+    m = @m.load(:id => 111, :x => 2, :y => 3)
+    m.pk_hash.should == {:y => 3, :x => 2}
+  end
+
+  it "should raise if no primary key" do
+    @m.set_primary_key nil
+    m = @m.new(:id => 111, :x => 2, :y => 3)
+    proc {m.pk_hash}.should raise_error(Sequel::Error)
+
+    @m.no_primary_key
+    m = @m.new(:id => 111, :x => 2, :y => 3)
+    proc {m.pk_hash}.should raise_error(Sequel::Error)
+  end
+end
+
+describe Sequel::Model, "#set" do
+  before do
+    MODEL_DB.reset
+    
+    @c = Class.new(Sequel::Model(:items)) do
+      set_primary_key :id
+      columns :x, :y, :id
+    end
+    @c.strict_param_setting = false
+    @c.instance_variable_set(:@columns, true)
+    @o1 = @c.new
+    @o2 = @c.load(:id => 5)
+  end
+
+  it "should filter the given params using the model columns" do
+    @o1.set(:x => 1, :z => 2)
+    @o1.values.should == {:x => 1}
+    MODEL_DB.sqls.should == []
+
+    @o2.set(:y => 1, :abc => 2)
+    @o2.values.should == {:y => 1, :id=> 5}
+    MODEL_DB.sqls.should == []
+  end
+  
+  it "should work with both strings and symbols" do
+    @o1.set('x'=> 1, 'z'=> 2)
+    @o1.values.should == {:x => 1}
+    MODEL_DB.sqls.should == []
+
+    @o2.set('y'=> 1, 'abc'=> 2)
+    @o2.values.should == {:y => 1, :id=> 5}
+    MODEL_DB.sqls.should == []
+  end
+  
+  it "should support virtual attributes" do
+    @c.send(:define_method, :blah=){|v| self.x = v}
+    @o1.set(:blah => 333)
+    @o1.values.should == {:x => 333}
+    MODEL_DB.sqls.should == []
+    @o1.set('blah'=> 334)
+    @o1.values.should == {:x => 334}
+    MODEL_DB.sqls.should == []
+  end
+  
+  it "should not modify the primary key" do
+    @o1.set(:x => 1, :id => 2)
+    @o1.values.should == {:x => 1}
+    MODEL_DB.sqls.should == []
+    @o2.set('y'=> 1, 'id'=> 2)
+    @o2.values.should == {:y => 1, :id=> 5}
+    MODEL_DB.sqls.should == []
+  end
+
+  it "should return self" do
+    returned_value = @o1.set(:x => 1, :z => 2)
+    returned_value.should == @o1
+    MODEL_DB.sqls.should == []
+  end
+end
+
+describe Sequel::Model, "#update" do
+  before do
+    MODEL_DB.reset
+    
+    @c = Class.new(Sequel::Model(:items)) do
+      set_primary_key :id
+      columns :x, :y, :id
+    end
+    @c.strict_param_setting = false
+    @c.instance_variable_set(:@columns, true)
+    @o1 = @c.new
+    @o2 = @c.load(:id => 5)
+  end
+  
+  it "should filter the given params using the model columns" do
+    @o1.update(:x => 1, :z => 2)
+    MODEL_DB.sqls.first.should == "INSERT INTO items (x) VALUES (1)"
+
+    MODEL_DB.reset
+    @o2.update(:y => 1, :abc => 2)
+    MODEL_DB.sqls.first.should == "UPDATE items SET y = 1 WHERE (id = 5)"
+  end
+  
+  it "should support virtual attributes" do
+    @c.send(:define_method, :blah=){|v| self.x = v}
+    @o1.update(:blah => 333)
+    MODEL_DB.sqls.first.should == "INSERT INTO items (x) VALUES (333)"
+  end
+  
+  it "should not modify the primary key" do
+    @o1.update(:x => 1, :id => 2)
+    MODEL_DB.sqls.first.should == "INSERT INTO items (x) VALUES (1)"
+    MODEL_DB.reset
+    @o2.update('y'=> 1, 'id'=> 2)
+    @o2.values.should == {:y => 1, :id=> 5}
+    MODEL_DB.sqls.first.should == "UPDATE items SET y = 1 WHERE (id = 5)"
+  end
+end
+
+describe Sequel::Model, "#(set|update)_(all|except|only)" do
+  before do
+    MODEL_DB.reset
+    
+    @c = Class.new(Sequel::Model(:items)) do
+      set_primary_key :id
+      columns :x, :y, :z, :id
+      set_allowed_columns :x
+      set_restricted_columns :y
+    end
+    @c.strict_param_setting = false
+    @c.instance_variable_set(:@columns, true)
+    @o1 = @c.new
+  end
+
+  it "#set_all should set all attributes" do
+    @o1.set_all(:x => 1, :y => 2, :z=>3, :id=>4)
+    @o1.values.should == {:x => 1, :y => 2, :z=>3}
+  end
+
+  it "#set_only should only set given attributes" do
+    @o1.set_only({:x => 1, :y => 2, :z=>3, :id=>4}, [:x, :y])
+    @o1.values.should == {:x => 1, :y => 2}
+    @o1.set_only({:x => 4, :y => 5, :z=>6, :id=>7}, :x, :y)
+    @o1.values.should == {:x => 4, :y => 5}
+  end
+
+  it "#set_except should not set given attributes" do
+    @o1.set_except({:x => 1, :y => 2, :z=>3, :id=>4}, [:y, :z])
+    @o1.values.should == {:x => 1}
+    @o1.set_except({:x => 4, :y => 2, :z=>3, :id=>4}, :y, :z)
+    @o1.values.should == {:x => 4}
+  end
+
+  it "#update_all should update all attributes" do
+    @c.new.update_all(:x => 1, :id=>4)
+    MODEL_DB.sqls.first.should == "INSERT INTO items (x) VALUES (1)"
+    MODEL_DB.reset
+    @c.new.update_all(:y => 1, :id=>4)
+    MODEL_DB.sqls.first.should == "INSERT INTO items (y) VALUES (1)"
+    MODEL_DB.reset
+    @c.new.update_all(:z => 1, :id=>4)
+    MODEL_DB.sqls.first.should == "INSERT INTO items (z) VALUES (1)"
+  end
+
+  it "#update_only should only update given attributes" do
+    @o1.update_only({:x => 1, :y => 2, :z=>3, :id=>4}, [:x])
+    MODEL_DB.sqls.first.should == "INSERT INTO items (x) VALUES (1)"
+    MODEL_DB.reset
+    @c.new.update_only({:x => 1, :y => 2, :z=>3, :id=>4}, :x)
+    MODEL_DB.sqls.first.should == "INSERT INTO items (x) VALUES (1)"
+  end
+
+  it "#update_except should not update given attributes" do
+    @o1.update_except({:x => 1, :y => 2, :z=>3, :id=>4}, [:y, :z])
+    MODEL_DB.sqls.first.should == "INSERT INTO items (x) VALUES (1)"
+    MODEL_DB.reset
+    @c.new.update_except({:x => 1, :y => 2, :z=>3, :id=>4}, :y, :z)
+    MODEL_DB.sqls.first.should == "INSERT INTO items (x) VALUES (1)"
+  end
+end
+
+describe Sequel::Model, "#destroy" do
+  before do
+    MODEL_DB.reset
+    @model = Class.new(Sequel::Model(:items))
+    @model.columns :id
+    @model.dataset.meta_def(:delete) {MODEL_DB.execute delete_sql}
+    
+    @instance = @model.load(:id => 1234)
+  end
+
+  it "should return self" do
+    @model.send(:define_method, :after_destroy){3}
+    @instance.destroy.should == @instance
+  end
+
+  it "should run within a transaction if use_transactions is true" do
+    @instance.use_transactions = true
+    @model.db.should_receive(:transaction)
+    @instance.destroy
+  end
+
+  it "should not run within a transaction if use_transactions is false" do
+    @instance.use_transactions = false
+    @model.db.should_not_receive(:transaction)
+    @instance.destroy
+  end
+
+  it "should run before_destroy and after_destroy hooks" do
+    @model.send(:define_method, :before_destroy){MODEL_DB.execute('before blah')}
+    @model.send(:define_method, :after_destroy){MODEL_DB.execute('after blah')}
+    @instance.destroy
+    
+    MODEL_DB.sqls.should == [
+      "before blah",
+      "DELETE FROM items WHERE (id = 1234)",
+      "after blah"
+    ]
+  end
+end
+
+describe Sequel::Model, "#exists?" do
+  before(:each) do
+    @model = Class.new(Sequel::Model(:items))
+    @m = @model.new
+  end
+
+  it "should returns true when #this.count > 0" do
+    @m.this.meta_def(:count) {1}
+    @m.exists?.should be_true
+  end
+
+  it "should return false when #this.count == 0" do
+    @m.this.meta_def(:count) {0}
+    @m.exists?.should be_false
+  end
+end
+
+describe Sequel::Model, "#each" do
+  before do
+    @model = Class.new(Sequel::Model(:items))
+    @model.columns :a, :b, :id
+    @m = @model.load(:a => 1, :b => 2, :id => 4444)
+  end
+  
+  specify "should iterate over the values" do
+    h = {}
+    @m.each {|k, v| h[k] = v}
+    h.should == {:a => 1, :b => 2, :id => 4444}
+  end
+end
+
+describe Sequel::Model, "#keys" do
+  before do
+    @model = Class.new(Sequel::Model(:items))
+    @model.columns :a, :b, :id
+    @m = @model.load(:a => 1, :b => 2, :id => 4444)
+  end
+  
+  specify "should return the value keys" do
+    @m.keys.size.should == 3
+    @m.keys.should include(:a, :b, :id)
+    
+    @m = @model.new()
+    @m.keys.should == []
+  end
+end
+
+describe Sequel::Model, "#==" do
+  specify "should compare instances by values" do
+    z = Class.new(Sequel::Model)
+    z.columns :id, :x
+    a = z.load(:id => 1, :x => 3)
+    b = z.load(:id => 1, :x => 4)
+    c = z.load(:id => 1, :x => 3)
+    
+    a.should_not == b
+    a.should == c
+    b.should_not == c
+  end
+
+  specify "should be aliased to #eql?" do
+    z = Class.new(Sequel::Model)
+    z.columns :id, :x
+    a = z.load(:id => 1, :x => 3)
+    b = z.load(:id => 1, :x => 4)
+    c = z.load(:id => 1, :x => 3)
+    
+    a.eql?(b).should == false
+    a.eql?(c).should == true
+    b.eql?(c).should == false
+  end
+end
+
+describe Sequel::Model, "#===" do
+  specify "should compare instances by class and pk if pk is not nil" do
+    z = Class.new(Sequel::Model)
+    z.columns :id, :x
+    y = Class.new(Sequel::Model)
+    y.columns :id, :x
+    a = z.load(:id => 1, :x => 3)
+    b = z.load(:id => 1, :x => 4)
+    c = z.load(:id => 2, :x => 3)
+    d = y.load(:id => 1, :x => 3)
+    
+    a.should === b
+    a.should_not === c
+    a.should_not === d
+  end
+
+  specify "should always be false if the primary key is nil" do
+    z = Class.new(Sequel::Model)
+    z.columns :id, :x
+    y = Class.new(Sequel::Model)
+    y.columns :id, :x
+    a = z.new(:x => 3)
+    b = z.new(:x => 4)
+    c = z.new(:x => 3)
+    d = y.new(:x => 3)
+    
+    a.should_not === b
+    a.should_not === c
+    a.should_not === d
+  end
+end
+
+describe Sequel::Model, "#hash" do
+  specify "should be the same only for objects with the same class and pk if the pk is not nil" do
+    z = Class.new(Sequel::Model)
+    z.columns :id, :x
+    y = Class.new(Sequel::Model)
+    y.columns :id, :x
+    a = z.load(:id => 1, :x => 3)
+    b = z.load(:id => 1, :x => 4)
+    c = z.load(:id => 2, :x => 3)
+    d = y.load(:id => 1, :x => 3)
+    
+    a.hash.should == b.hash
+    a.hash.should_not == c.hash
+    a.hash.should_not == d.hash
+  end
+
+  specify "should be the same only for objects with the same class and values if the pk is nil" do
+    z = Class.new(Sequel::Model)
+    z.columns :id, :x
+    y = Class.new(Sequel::Model)
+    y.columns :id, :x
+    a = z.new(:x => 3)
+    b = z.new(:x => 4)
+    c = z.new(:x => 3)
+    d = y.new(:x => 3)
+    
+    a.hash.should_not == b.hash
+    a.hash.should == c.hash
+    a.hash.should_not == d.hash
+  end
+end
+
+describe Sequel::Model, "#initialize" do
+  before do
+    @c = Class.new(Sequel::Model) do
+      columns :id, :x
+    end
+    @c.strict_param_setting = false
+  end
+  
+  specify "should accept values" do
+    m = @c.new(:x => 2)
+    m.values.should == {:x => 2}
+  end
+  
+  specify "should not modify the primary key" do
+    m = @c.new(:id => 1, :x => 2)
+    m.values.should == {:x => 2}
+  end
+  
+  specify "should accept no values" do
+    m = @c.new
+    m.values.should == {}
+  end
+  
+  specify "should accept a block to execute" do
+    m = @c.new {|o| o[:id] = 1234}
+    m.id.should == 1234
+  end
+  
+  specify "should accept virtual attributes" do
+    @c.send(:define_method, :blah=){|x| @blah = x}
+    @c.send(:define_method, :blah){@blah}
+    
+    m = @c.new(:x => 2, :blah => 3)
+    m.values.should == {:x => 2}
+    m.blah.should == 3
+  end
+  
+  specify "should convert string keys into symbol keys" do
+    m = @c.new('x' => 2)
+    m.values.should == {:x => 2}
+  end
+end
+
+describe Sequel::Model, ".create" do
+
+  before(:each) do
+    MODEL_DB.reset
+    @c = Class.new(Sequel::Model(:items)) do
+      unrestrict_primary_key
+      columns :x
+    end
+  end
+
+  it "should be able to create rows in the associated table" do
+    o = @c.create(:x => 1)
+    o.class.should == @c
+    MODEL_DB.sqls.should == ['INSERT INTO items (x) VALUES (1)',  "SELECT * FROM items WHERE (id IN ('INSERT INTO items (x) VALUES (1)')) LIMIT 1"]
+  end
+
+  it "should be able to create rows without any values specified" do
+    o = @c.create
+    o.class.should == @c
+    MODEL_DB.sqls.should == ["INSERT INTO items DEFAULT VALUES", "SELECT * FROM items WHERE (id IN ('INSERT INTO items DEFAULT VALUES')) LIMIT 1"]
+  end
+
+  it "should accept a block and run it" do
+    o1, o2, o3 =  nil, nil, nil
+    o = @c.create {|o4| o1 = o4; o3 = o4; o2 = :blah; o3.x = 333}
+    o.class.should == @c
+    o1.should === o
+    o3.should === o
+    o2.should == :blah
+    MODEL_DB.sqls.should == ["INSERT INTO items (x) VALUES (333)", "SELECT * FROM items WHERE (id IN ('INSERT INTO items (x) VALUES (333)')) LIMIT 1"]
+  end
+  
+  it "should create a row for a model with custom primary key" do
+    @c.set_primary_key :x
+    o = @c.create(:x => 30)
+    o.class.should == @c
+    MODEL_DB.sqls.should == ["INSERT INTO items (x) VALUES (30)", "SELECT * FROM items WHERE (x = 30) LIMIT 1"]
+  end
+end
+
+describe Sequel::Model, "#refresh" do
+  before do
+    MODEL_DB.reset
+    @c = Class.new(Sequel::Model(:items)) do
+      unrestrict_primary_key
+      columns :id, :x
+    end
+  end
+
+  specify "should reload the instance values from the database" do
+    @m = @c.new(:id => 555)
+    @m[:x] = 'blah'
+    @m.this.should_receive(:first).and_return({:x => 'kaboom', :id => 555})
+    @m.refresh
+    @m[:x].should == 'kaboom'
+  end
+  
+  specify "should raise if the instance is not found" do
+    @m = @c.new(:id => 555)
+    @m.this.should_receive(:first).and_return(nil)
+    proc {@m.refresh}.should raise_error(Sequel::Error)
+  end
+  
+  specify "should be aliased by #reload" do
+    @m = @c.new(:id => 555)
+    @m.this.should_receive(:first).and_return({:x => 'kaboom', :id => 555})
+    @m.reload
+    @m[:x].should == 'kaboom'
+  end
+
+  specify "should remove cached associations" do
+    @c.many_to_one :node, :class=>@c
+    @m = @c.new(:id => 555)
+    @m.associations[:node] = 15
+    @m.reload
+    @m.associations.should == {}
+  end
+end
+
+describe Sequel::Model, "typecasting" do
+  before do
+    MODEL_DB.reset
+    @c = Class.new(Sequel::Model(:items)) do
+      columns :x
+    end
+  end
+
+  after do
+    Sequel.datetime_class = Time
+  end
+
+  specify "should not convert if typecasting is turned off" do
+    @c.typecast_on_assignment = false
+    @c.instance_variable_set(:@db_schema, {:x=>{:type=>:integer}})
+    m = @c.new
+    m.x = '1'
+    m.x.should == '1'
+  end
+
+  specify "should convert to integer for an integer field" do
+    @c.instance_variable_set(:@db_schema, {:x=>{:type=>:integer}})
+    m = @c.new
+    m.x = '1'
+    m.x.should == 1
+    m.x = 1
+    m.x.should == 1
+    m.x = 1.3
+    m.x.should == 1
+  end
+
+  specify "should typecast '' to nil unless type is string or blob" do
+    [:integer, :float, :decimal, :boolean, :date, :time, :datetime].each do |x|
+      @c.instance_variable_set(:@db_schema, {:x=>{:type=>x}})
+      m = @c.new
+      m.x = ''
+      m.x.should == nil
+    end
+   [:string, :blob].each do |x|
+      @c.instance_variable_set(:@db_schema, {:x=>{:type=>x}})
+      m = @c.new
+      m.x = ''
+      m.x.should == ''
+    end
+  end
+
+  specify "should not typecast '' to nil if typecast_empty_string_to_nil is false" do
+    @c.instance_variable_set(:@db_schema, {:x=>{:type=>:integer}})
+    m = @c.new
+    m.typecast_empty_string_to_nil = false
+    proc{m.x = ''}.should raise_error
+    @c.typecast_empty_string_to_nil = false
+    proc{@c.new.x = ''}.should raise_error
+  end
+
+  specify "should not typecast nil if NULLs are allowed" do
+    @c.instance_variable_set(:@db_schema, {:x=>{:type=>:integer,:allow_null=>true}})
+    m = @c.new
+    m.x = nil
+    m.x.should == nil
+  end
+
+  specify "should raise an error if attempting to typecast nil and NULLs are not allowed" do
+    @c.instance_variable_set(:@db_schema, {:x=>{:type=>:integer,:allow_null=>false}})
+    proc{@c.new.x = nil}.should raise_error(Sequel::Error)
+    proc{@c.new.x = ''}.should raise_error(Sequel::Error)
+  end
+
+  specify "should not raise an error if NULLs are not allowed and typecasting is turned off" do
+    @c.typecast_on_assignment = false
+    @c.instance_variable_set(:@db_schema, {:x=>{:type=>:integer,:allow_null=>false}})
+    m = @c.new
+    m.x = nil
+    m.x.should == nil
+  end
+
+  specify "should not raise when typecasting nil to NOT NULL column but raise_on_typecast_failure is off" do
+    @c.raise_on_typecast_failure = false
+    @c.typecast_on_assignment = true
+    @c.instance_variable_set(:@db_schema, {:x=>{:type=>:integer,:allow_null=>false}})
+    m = @c.new
+    m.x = ''
+    m.x.should == nil
+    m.x = nil
+    m.x.should == nil
+  end
+
+  specify "should raise an error if invalid data is used in an integer field" do
+    @c.instance_variable_set(:@db_schema, {:x=>{:type=>:integer}})
+    proc{@c.new.x = 'a'}.should raise_error(Sequel::InvalidValue)
+  end
+
+  specify "should assign value if raise_on_typecast_failure is off and assigning invalid integer" do
+    @c.raise_on_typecast_failure = false
+    @c.instance_variable_set(:@db_schema, {:x=>{:type=>:integer}})
+    model = @c.new
+    model.x = '1d'
+    model.x.should == '1d'
+  end
+
+  specify "should convert to float for a float field" do
+    @c.instance_variable_set(:@db_schema, {:x=>{:type=>:float}})
+    m = @c.new
+    m.x = '1.3'
+    m.x.should == 1.3
+    m.x = 1
+    m.x.should == 1.0
+    m.x = 1.3
+    m.x.should == 1.3
+  end
+
+  specify "should raise an error if invalid data is used in an float field" do
+    @c.instance_variable_set(:@db_schema, {:x=>{:type=>:float}})
+    proc{@c.new.x = 'a'}.should raise_error(Sequel::InvalidValue)
+  end
+
+  specify "should assign value if raise_on_typecast_failure is off and assigning invalid float" do
+    @c.raise_on_typecast_failure = false
+    @c.instance_variable_set(:@db_schema, {:x=>{:type=>:float}})
+    model = @c.new
+    model.x = '1d'
+    model.x.should == '1d'
+  end
+
+  specify "should convert to BigDecimal for a decimal field" do
+    @c.instance_variable_set(:@db_schema, {:x=>{:type=>:decimal}})
+    m = @c.new
+    bd = BigDecimal.new('1.0')
+    m.x = '1.0'
+    m.x.should == bd
+    m.x = 1.0
+    m.x.should == bd
+    m.x = 1
+    m.x.should == bd
+    m.x = bd
+    m.x.should == bd
+  end
+
+  specify "should raise an error if invalid data is used in an decimal field" do
+    @c.instance_variable_set(:@db_schema, {:x=>{:type=>:decimal}})
+    proc{@c.new.x = Date.today}.should raise_error(Sequel::InvalidValue)
+  end
+
+  specify "should assign value if raise_on_typecast_failure is off and assigning invalid decimal" do
+    @c.raise_on_typecast_failure = false
+    @c.instance_variable_set(:@db_schema, {:x=>{:type=>:decimal}})
+    model = @c.new
+    time = Time.now
+    model.x = time
+    model.x.should == time
+  end
+
+  specify "should convert to string for a string field" do
+    @c.instance_variable_set(:@db_schema, {:x=>{:type=>:string}})
+    m = @c.new
+    m.x = '1.3'
+    m.x.should == '1.3'
+    m.x = 1
+    m.x.should == '1'
+    m.x = 1.3
+    m.x.should == '1.3'
+  end
+
+  specify "should convert to boolean for a boolean field" do
+    @c.instance_variable_set(:@db_schema, {:x=>{:type=>:boolean}})
+    m = @c.new
+    m.x = '1.3'
+    m.x.should == true
+    m.x = 1
+    m.x.should == true
+    m.x = 1.3
+    m.x.should == true
+    m.x = 't'
+    m.x.should == true
+    m.x = 'T'
+    m.x.should == true
+    m.x = true
+    m.x.should == true
+    m.x = nil
+    m.x.should == nil
+    m.x = ''
+    m.x.should == nil
+    m.x = []
+    m.x.should == nil
+    m.x = 'f'
+    m.x.should == false
+    m.x = 'F'
+    m.x.should == false
+    m.x = 'false'
+    m.x.should == false
+    m.x = 'FALSE'
+    m.x.should == false
+    m.x = '0'
+    m.x.should == false
+    m.x = 0
+    m.x.should == false
+    m.x = false
+    m.x.should == false
+  end
+
+  specify "should convert to date for a date field" do
+    @c.instance_variable_set(:@db_schema, {:x=>{:type=>:date}})
+    m = @c.new
+    y = Date.new(2007,10,21)
+    m.x = '2007-10-21'
+    m.x.should == y
+    m.x = Date.parse('2007-10-21')
+    m.x.should == y
+    m.x = Time.parse('2007-10-21')
+    m.x.should == y
+    m.x = DateTime.parse('2007-10-21')
+    m.x.should == y
+  end
+
+  specify "should raise an error if invalid data is used in a date field" do
+    @c.instance_variable_set(:@db_schema, {:x=>{:type=>:date}})
+    proc{@c.new.x = 'a'}.should raise_error(Sequel::InvalidValue)
+    proc{@c.new.x = 100}.should raise_error(Sequel::InvalidValue)
+  end
+
+  specify "should assign value if raise_on_typecast_failure is off and assigning invalid date" do
+    @c.raise_on_typecast_failure = false
+    @c.instance_variable_set(:@db_schema, {:x=>{:type=>:date}})
+    model = @c.new
+    model.x = 4
+    model.x.should == 4
+  end
+
+  specify "should convert to time for a time field" do
+    @c.instance_variable_set(:@db_schema, {:x=>{:type=>:time}})
+    m = @c.new
+    x = '10:20:30'
+    y = Time.parse(x)
+    m.x = x
+    m.x.should == y
+    m.x = y
+    m.x.should == y
+  end
+
+  specify "should raise an error if invalid data is used in a time field" do
+    @c.instance_variable_set(:@db_schema, {:x=>{:type=>:time}})
+    proc{@c.new.x = '0000'}.should raise_error
+    proc{@c.new.x = 'a'}.should_not raise_error # Valid Time
+    proc{@c.new.x = Date.parse('2008-10-21')}.should raise_error(Sequel::InvalidValue)
+    proc{@c.new.x = DateTime.parse('2008-10-21')}.should raise_error(Sequel::InvalidValue)
+  end
+
+  specify "should assign value if raise_on_typecast_failure is off and assigning invalid time" do
+    @c.raise_on_typecast_failure = false
+    @c.instance_variable_set(:@db_schema, {:x=>{:type=>:time}})
+    model = @c.new
+    model.x = '0000'
+    model.x.should == '0000'
+  end
+
+  specify "should convert to the Sequel.datetime_class for a datetime field" do
+    @c.instance_variable_set(:@db_schema, {:x=>{:type=>:datetime}})
+    m = @c.new
+    x = '2007-10-21T10:20:30-07:00'
+    y = Time.parse(x)
+    m.x = x
+    m.x.should == y
+    m.x = DateTime.parse(x)
+    m.x.should == y
+    m.x = Time.parse(x)
+    m.x.should == y
+    m.x = Date.parse('2007-10-21')
+    m.x.should == Time.parse('2007-10-21')
+    Sequel.datetime_class = DateTime
+    y = DateTime.parse(x)
+    m.x = x
+    m.x.should == y
+    m.x = DateTime.parse(x)
+    m.x.should == y
+    m.x = Time.parse(x)
+    m.x.should == y
+    m.x = Date.parse('2007-10-21')
+    m.x.should == DateTime.parse('2007-10-21')
+  end
+
+  specify "should raise an error if invalid data is used in a datetime field" do
+    @c.instance_variable_set(:@db_schema, {:x=>{:type=>:datetime}})
+    proc{@c.new.x = '0000'}.should raise_error(Sequel::InvalidValue)
+    proc{@c.new.x = 'a'}.should_not raise_error # Valid Time
+    Sequel.datetime_class = DateTime
+    proc{@c.new.x = '0000'}.should raise_error(Sequel::InvalidValue)
+    proc{@c.new.x = 'a'}.should raise_error(Sequel::InvalidValue)
+  end
+
+  specify "should assign value if raise_on_typecast_failure is off and assigning invalid datetime" do
+    @c.raise_on_typecast_failure = false
+    @c.instance_variable_set(:@db_schema, {:x=>{:type=>:datetime}})
+    model = @c.new
+    model.x = '0000'
+    model.x.should == '0000'
+    Sequel.datetime_class = DateTime
+    model = @c.new
+    model.x = '0000'
+    model.x.should == '0000'
+    model.x = 'a'
+    model.x.should == 'a'
+  end
+end
diff -pruN 1.5.1-1/spec/model/spec_helper.rb 3.0.0-1/spec/model/spec_helper.rb
--- 1.5.1-1/spec/model/spec_helper.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/model/spec_helper.rb	2009-04-29 00:35:29.000000000 +0100
@@ -0,0 +1,97 @@
+require 'rubygems'
+unless Object.const_defined?('Sequel')
+  $:.unshift(File.join(File.dirname(__FILE__), "../../lib/"))
+  require 'sequel/core'
+end
+unless Sequel.const_defined?('Model')
+  $:.unshift(File.join(File.dirname(__FILE__), "../../lib/"))
+  require 'sequel/model'
+end
+
+Sequel.virtual_row_instance_eval = true
+
+class MockDataset < Sequel::Dataset
+  def insert(*args)
+    @db.execute insert_sql(*args)
+  end
+  
+  def update(*args)
+    @db.execute update_sql(*args)
+  end
+  
+  def delete(*args)
+    @db.execute delete_sql(*args)
+  end
+  
+  def fetch_rows(sql)
+    return if sql =~ /information_schema/
+    @db.execute(sql)
+    yield({:id => 1, :x => 1})
+  end
+
+  def quoted_identifier(c)
+    "\"#{c}\""
+  end
+end
+
+class MockDatabase < Sequel::Database
+  @@quote_identifiers = false
+  self.identifier_input_method = nil
+  self.identifier_output_method = nil
+  attr_reader :sqls
+  
+  def execute(sql, opts={})
+    @sqls ||= []
+    @sqls << sql
+  end
+
+  def reset
+    @sqls = []
+  end
+
+  def schema(table_name, opts)
+    if table_name
+      [[:id, {:primary_key=>true}]]
+    else
+      {table_name=>[[:id, {:primary_key=>true}]]}
+    end
+  end
+
+   def transaction(opts={})
+    return yield if @transactions.include?(Thread.current)
+    execute('BEGIN')
+    begin
+      @transactions << Thread.current
+      yield
+    rescue Exception => e
+      execute('ROLLBACK')
+      transaction_error(e)
+    ensure
+      unless e
+        execute('COMMIT')
+      end
+      @transactions.delete(Thread.current)
+    end
+  end
+
+  def dataset(opts=nil); MockDataset.new(self, opts); end
+end
+
+class << Sequel::Model
+  alias orig_columns columns
+  def columns(*cols)
+    return if cols.empty?
+    define_method(:columns){cols}
+    @dataset.instance_variable_set(:@columns, cols) if @dataset
+    def_column_accessor(*cols)
+    @columns = cols
+    @db_schema = {}
+    cols.each{|c| @db_schema[c] = {}}
+  end
+  def simple_table
+    nil
+  end
+end
+
+Sequel::Model.use_transactions = false
+Sequel::Model.db = MODEL_DB = MockDatabase.new
diff -pruN 1.5.1-1/spec/model/validations_spec.rb 3.0.0-1/spec/model/validations_spec.rb
--- 1.5.1-1/spec/model/validations_spec.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/model/validations_spec.rb	2009-04-09 18:43:32.000000000 +0100
@@ -0,0 +1,149 @@
+require File.join(File.dirname(__FILE__), "spec_helper")
+
+describe Sequel::Model::Errors do
+  before do
+    @errors = Sequel::Model::Errors.new
+  end
+  
+  specify "should be clearable using #clear" do
+    @errors.add(:a, 'b')
+    @errors.should == {:a=>['b']}
+    @errors.clear
+    @errors.should == {}
+  end
+  
+  specify "should be empty if no errors are added" do
+    @errors.should be_empty
+    @errors[:blah] << "blah"
+    @errors.should_not be_empty
+  end
+  
+  specify "should return errors for a specific attribute using #on or #[]" do
+    @errors[:blah].should == []
+    @errors.on(:blah).should == []
+
+    @errors[:blah] << 'blah'
+    @errors[:blah].should == ['blah']
+    @errors.on(:blah).should == ['blah']
+
+    @errors[:bleu].should == []
+    @errors.on(:bleu).should == []
+  end
+  
+  specify "should accept errors using #[] << or #add" do
+    @errors[:blah] << 'blah'
+    @errors[:blah].should == ['blah']
+    
+    @errors.add :blah, 'zzzz'
+    @errors[:blah].should == ['blah', 'zzzz']
+  end
+  
+  specify "should return full messages using #full_messages" do
+    @errors.full_messages.should == []
+    
+    @errors[:blow] << 'blieuh'
+    @errors[:blow] << 'blich'
+    @errors[:blay] << 'bliu'
+    msgs = @errors.full_messages
+    msgs.size.should == 3
+    msgs.should include('blow blieuh', 'blow blich', 'blay bliu')
+  end
+
+  specify "should return the number of error messages using #count" do
+    @errors.count.should == 0
+    @errors.add(:a, 'b')
+    @errors.count.should == 1
+    @errors.add(:a, 'c')
+    @errors.count.should == 2
+    @errors.add(:b, 'c')
+    @errors.count.should == 3
+  end
+
+  specify "should return the array of error messages for a given attribute using #on" do
+    @errors.add(:a, 'b')
+    @errors.on(:a).should == ['b']
+    @errors.add(:a, 'c')
+    @errors.on(:a).should == ['b', 'c']
+    @errors.add(:b, 'c')
+    @errors.on(:a).should == ['b', 'c']
+  end
+
+  specify "should return nil if there are no error messages for a given attribute using #on" do
+    @errors.on(:a).should == nil
+    @errors.add(:b, 'b')
+    @errors.on(:a).should == nil
+  end
+end
+
+describe Sequel::Model do
+  before do
+    @c = Class.new(Sequel::Model) do
+      columns :score
+      def validate
+        errors[:score] << 'too low' if score < 87
+      end
+    end
+    
+    @o = @c.new
+  end
+  
+  specify "should supply a #valid? method that returns true if validations pass" do
+    @o.score = 50
+    @o.should_not be_valid
+    @o.score = 100
+    @o.should be_valid
+  end
+  
+  specify "should provide an errors object" do
+    @o.score = 100
+    @o.should be_valid
+    @o.errors.should be_empty
+    
+    @o.score = 86
+    @o.should_not be_valid
+    @o.errors[:score].should == ['too low']
+    @o.errors[:blah].should be_empty
+  end
+end
+
+describe "Model#save" do
+  before do
+    @c = Class.new(Sequel::Model(:people)) do
+      columns :id
+
+      def validate
+        errors[:id] << 'blah' unless id == 5
+      end
+    end
+    @m = @c.load(:id => 4)
+    MODEL_DB.reset
+  end
+
+  specify "should save only if validations pass" do
+    @m.raise_on_save_failure = false
+    @m.should_not be_valid
+    @m.save
+    MODEL_DB.sqls.should be_empty
+    
+    @m.id = 5
+    @m.should be_valid
+    @m.save.should_not be_false
+    MODEL_DB.sqls.should == ['UPDATE people SET id = 5 WHERE (id = 5)']
+  end
+  
+  specify "should skip validations if the :validate=>false option is used" do
+    @m.raise_on_save_failure = false
+    @m.should_not be_valid
+    @m.save(:validate=>false)
+    MODEL_DB.sqls.should == ['UPDATE people SET id = 4 WHERE (id = 4)']
+  end
+    
+  specify "should raise error if validations fail and raise_on_save_faiure is true" do
+    proc{@m.save}.should raise_error(Sequel::ValidationFailed)
+  end
+  
+  specify "should return nil if validations fail and raise_on_save_faiure is false" do
+    @m.raise_on_save_failure = false
+    @m.save.should == nil
+  end
+end
diff -pruN 1.5.1-1/spec/model_spec.rb 3.0.0-1/spec/model_spec.rb
--- 1.5.1-1/spec/model_spec.rb	2008-04-22 20:16:28.000000000 +0100
+++ 3.0.0-1/spec/model_spec.rb	1970-01-01 01:00:00.000000000 +0100
@@ -1,520 +0,0 @@
-require File.join(File.dirname(__FILE__), "spec_helper")
-
-describe Sequel::Model do
-  it "should have class method aliased as model" do
-    Sequel::Model.instance_methods.should include("model")
-
-    model_a = Class.new Sequel::Model
-    model_a.new.model.should be(model_a)
-  end
-
-  it "should be associated with a dataset" do
-    model_a = Class.new(Sequel::Model) { set_dataset MODEL_DB[:as] }
-
-    model_a.dataset.should be_a_kind_of(MockDataset)
-    model_a.dataset.opts[:from].should == [:as]
-
-    model_b = Class.new(Sequel::Model) { set_dataset MODEL_DB[:bs] }
-
-    model_b.dataset.should be_a_kind_of(MockDataset)
-    model_b.dataset.opts[:from].should == [:bs]
-
-    model_a.dataset.opts[:from].should == [:as]
-  end
-
-end
-
-describe Sequel::Model, "dataset & schema" do
-
-  before do
-    @model = Class.new(Sequel::Model(:items))
-  end
-
-  it "creates dynamic model subclass with set table name" do
-    @model.table_name.should == :items
-  end
-
-  it "defaults to primary key of id" do
-    @model.primary_key.should == :id
-  end
-
-  it "allow primary key change" do
-    @model.set_primary_key :ssn
-    @model.primary_key.should == :ssn
-  end
-
-  it "allows dataset change" do
-    @model.set_dataset(MODEL_DB[:foo])
-    @model.table_name.should == :foo
-  end
-
-  it "sets schema with implicit table name" do
-    @model.set_schema do
-      primary_key :ssn, :string
-    end
-    @model.primary_key.should == :ssn
-    @model.table_name.should == :items
-  end
-
-  it "sets schema with explicit table name" do
-    @model.set_schema :foo do
-      primary_key :id
-    end
-    @model.primary_key.should == :id
-    @model.table_name.should == :foo
-  end
-
-  it "puts the lotion in the basket or it gets the hose again" do
-    # just kidding!
-  end
-end
-
-describe Sequel::Model, "constructor" do
-  
-  before(:each) do
-    @m = Class.new(Sequel::Model)
-    @m.columns :a, :b
-  end
-
-  it "should accept a hash" do
-    m = @m.new(:a => 1, :b => 2)
-    m.values.should == {:a => 1, :b => 2}
-    m.should be_new
-  end
-  
-  it "should accept a block and yield itself to the block" do
-    block_called = false
-    m = @m.new {|i| block_called = true; i.should be_a_kind_of(@m); i.values[:a] = 1}
-    
-    block_called.should be_true
-    m.values[:a].should == 1
-  end
-  
-end
-
-describe Sequel::Model, "new" do
-
-  before(:each) do
-    @m = Class.new(Sequel::Model) do
-      set_dataset MODEL_DB[:items]
-    end
-  end
-
-  it "should be marked as new?" do
-    o = @m.new
-    o.should be_new
-  end
-
-  it "should not be marked as new? once it is saved" do
-    o = @m.new(:x => 1)
-    o.should be_new
-    o.save
-    o.should_not be_new
-  end
-
-  it "should use the last inserted id as primary key if not in values" do
-    d = @m.dataset
-    def d.insert(*args)
-      super
-      1234
-    end
-
-    def d.first
-      {:x => 1, :id => 1234}
-    end
-
-    o = @m.new(:x => 1)
-    o.save
-    o.id.should == 1234
-
-    o = @m.load(:x => 1, :id => 333)
-    o.save
-    o.id.should == 333
-  end
-
-end
-
-describe Sequel::Model, ".subset" do
-
-  before(:each) do
-    MODEL_DB.reset
-
-    @c = Class.new(Sequel::Model(:items))
-  end
-
-  it "should create a filter on the underlying dataset" do
-    proc {@c.new_only}.should raise_error(NoMethodError)
-    
-    @c.subset(:new_only) {:age == 'new'}
-    
-    @c.new_only.sql.should == "SELECT * FROM items WHERE (age = 'new')"
-    @c.dataset.new_only.sql.should == "SELECT * FROM items WHERE (age = 'new')"
-    
-    @c.subset(:pricey) {:price > 100}
-    
-    @c.pricey.sql.should == "SELECT * FROM items WHERE (price > 100)"
-    @c.dataset.pricey.sql.should == "SELECT * FROM items WHERE (price > 100)"
-    
-    # check if subsets are composable
-    @c.pricey.new_only.sql.should == "SELECT * FROM items WHERE (price > 100) AND (age = 'new')"
-    @c.new_only.pricey.sql.should == "SELECT * FROM items WHERE (age = 'new') AND (price > 100)"
-  end
-
-end
-
-describe Sequel::Model, ".find" do
-
-  before(:each) do
-    MODEL_DB.reset
-    
-    @c = Class.new(Sequel::Model(:items))
-    
-    $cache_dataset_row = {:name => 'sharon', :id => 1}
-    @dataset = @c.dataset
-    $sqls = []
-    @dataset.extend(Module.new {
-      def fetch_rows(sql)
-        $sqls << sql
-        yield $cache_dataset_row
-      end
-    })
-  end
-  
-  it "should return the first record matching the given filter" do
-    @c.find(:name => 'sharon').should be_a_kind_of(@c)
-    $sqls.last.should == "SELECT * FROM items WHERE (name = 'sharon') LIMIT 1"
-
-    @c.find {"name LIKE 'abc%'".lit}.should be_a_kind_of(@c)
-    $sqls.last.should == "SELECT * FROM items WHERE name LIKE 'abc%' LIMIT 1"
-  end
-  
-  it "should accept filter blocks" do
-    @c.find {:id == 1}.should be_a_kind_of(@c)
-    $sqls.last.should == "SELECT * FROM items WHERE (id = 1) LIMIT 1"
-
-    @c.find {:x > 1 && :y < 2}.should be_a_kind_of(@c)
-    $sqls.last.should == "SELECT * FROM items WHERE ((x > 1) AND (y < 2)) LIMIT 1"
-  end
-
-end
-
-describe Sequel::Model, ".fetch" do
-
-  before(:each) do
-    MODEL_DB.reset
-    @c = Class.new(Sequel::Model(:items))
-  end
-  
-  it "should return instances of Model" do
-    @c.fetch("SELECT * FROM items").first.should be_a_kind_of(@c)
-  end
-
-  it "should return true for .empty? and not raise an error on empty selection" do
-    rows = @c.fetch("SELECT * FROM items WHERE FALSE")
-    @c.class_def(:fetch_rows) {|sql| yield({:count => 0})}
-    proc {rows.empty?}.should_not raise_error
-  end
-  
-end
-
-### DEPRECATED
-describe Sequel::Model, "magic methods" do
-
-  before(:each) do
-    @c = Class.new(Sequel::Dataset) do
-      @@sqls = []
-    
-      def self.sqls; @@sqls; end
-    
-      def fetch_rows(sql)
-        @@sqls << sql
-        yield({:id => 123, :name => 'hey'})
-      end
-    end
-  
-    @m = Class.new(Sequel::Model(@c.new(nil).from(:items)))
-  end
-  
-  it "should support order_by_xxx" do
-    @m.order_by_name.should be_a_kind_of(@c)
-    @m.order_by_name.sql.should == "SELECT * FROM items ORDER BY name"
-  end
-
-  it "should support group_by_xxx" do
-    @m.group_by_name.should be_a_kind_of(@c)
-    @m.group_by_name.sql.should == "SELECT * FROM items GROUP BY name"
-  end
-
-  it "should support count_by_xxx" do
-    @m.count_by_name.should be_a_kind_of(@c)
-    @m.count_by_name.sql.should == "SELECT name, count(*) AS count FROM items GROUP BY name ORDER BY count"
-  end
-
-  it "should support filter_by_xxx" do
-    @m.filter_by_name('sharon').should be_a_kind_of(@c)
-    @m.filter_by_name('sharon').sql.should == "SELECT * FROM items WHERE (name = 'sharon')"
-  end
-  
-  it "should support all_by_xxx" do
-    all = @m.all_by_name('sharon')
-    all.class.should == Array
-    all.size.should == 1
-    all.first.should be_a_kind_of(@m)
-    all.first.values.should == {:id => 123, :name => 'hey'}
-    @c.sqls.should == ["SELECT * FROM items WHERE (name = 'sharon')"]
-  end
-  
-  it "should support find_by_xxx" do
-    @m.find_by_name('sharon').should be_a_kind_of(@m)
-    @m.find_by_name('sharon').values.should == {:id => 123, :name => 'hey'}
-    @c.sqls.should == ["SELECT * FROM items WHERE (name = 'sharon') LIMIT 1"] * 2
-  end
-
-  it "should support first_by_xxx" do
-    @m.first_by_name('sharon').should be_a_kind_of(@m)
-    @m.first_by_name('sharon').values.should == {:id => 123, :name => 'hey'}
-    @c.sqls.should == ["SELECT * FROM items ORDER BY name LIMIT 1"] * 2
-  end
-
-  it "should support last_by_xxx" do
-    @m.last_by_name('sharon').should be_a_kind_of(@m)
-    @m.last_by_name('sharon').values.should == {:id => 123, :name => 'hey'}
-    @c.sqls.should == ["SELECT * FROM items ORDER BY name DESC LIMIT 1"] * 2
-  end
-  
-end
-
-describe Sequel::Model, ".find_or_create" do
-
-  before(:each) do
-    MODEL_DB.reset
-    @c = Class.new(Sequel::Model(:items)) do
-      no_primary_key
-      columns :x
-    end
-  end
-
-  it "should find the record" do
-    @c.find_or_create(:x => 1)
-    MODEL_DB.sqls.should == ["SELECT * FROM items WHERE (x = 1) LIMIT 1"]
-    
-    MODEL_DB.reset
-  end
-  
-  it "should create the record if not found" do
-    @c.meta_def(:find) do |*args|
-      dataset.filter(*args).first
-      nil
-    end
-    
-    @c.find_or_create(:x => 1)
-    MODEL_DB.sqls.should == [
-      "SELECT * FROM items WHERE (x = 1) LIMIT 1",
-      "INSERT INTO items (x) VALUES (1)"
-    ]
-  end
-end
-
-describe Sequel::Model, ".delete_all" do
-
-  before(:each) do
-    MODEL_DB.reset
-    @c = Class.new(Sequel::Model(:items)) do
-      no_primary_key
-    end
-    
-    @c.dataset.meta_def(:delete) {MODEL_DB << delete_sql}
-  end
-
-  it "should delete all records in the dataset" do
-    @c.delete_all
-    MODEL_DB.sqls.should == ["DELETE FROM items"]
-  end
-
-end
-
-describe Sequel::Model, ".destroy_all" do
-
-  before(:each) do
-    MODEL_DB.reset
-    @c = Class.new(Sequel::Model(:items)) do
-      no_primary_key
-    end
-
-    @c.dataset.meta_def(:delete) {MODEL_DB << delete_sql}
-  end
-
-  it "should delete all records in the dataset" do
-    @c.dataset.meta_def(:destroy) {MODEL_DB << "DESTROY this stuff"}
-    @c.destroy_all
-    MODEL_DB.sqls.should == ["DESTROY this stuff"]
-  end
-  
-  it "should call dataset.destroy" do
-    @c.dataset.should_receive(:destroy).and_return(true)
-    @c.destroy_all
-  end
-end
-
-describe Sequel::Model, ".all" do
-  
-  before(:each) do
-    MODEL_DB.reset
-    @c = Class.new(Sequel::Model(:items)) do
-      no_primary_key
-    end
-    
-    @c.dataset.meta_def(:all) {1234}
-  end
-  
-  it "should return all records in the dataset" do
-    @c.all.should == 1234
-  end
-  
-end
-
-class DummyModelBased < Sequel::Model(:blog)
-end
-
-describe Sequel::Model, "(:tablename)" do
-
-  it "should allow reopening of descendant classes" do
-    proc do
-      eval "class DummyModelBased < Sequel::Model(:blog); end"
-    end.should_not raise_error
-  end
-
-end
-
-describe Sequel::Model, "A model class without a primary key" do
-
-  before(:each) do
-    MODEL_DB.reset
-    @c = Class.new(Sequel::Model(:items)) do
-      columns :x
-      no_primary_key
-    end
-  end
-
-  it "should be able to insert records without selecting them back" do
-    i = nil
-    proc {i = @c.create(:x => 1)}.should_not raise_error
-    i.class.should be(@c)
-    i.values.to_hash.should == {:x => 1}
-
-    MODEL_DB.sqls.should == ['INSERT INTO items (x) VALUES (1)']
-  end
-
-  it "should raise when deleting" do
-    o = @c.new
-    proc {o.delete}.should raise_error
-  end
-
-  it "should insert a record when saving" do
-    o = @c.new(:x => 2)
-    o.should be_new
-    o.save
-    MODEL_DB.sqls.should == ['INSERT INTO items (x) VALUES (2)']
-  end
-
-end
-
-describe Sequel::Model, "attribute accessors" do
-
-  before(:each) do
-    MODEL_DB.reset
-
-    @c = Class.new(Sequel::Model) do
-      def self.columns; orig_columns; end
-    end
-    @dataset = Object.new
-    def @dataset.db; end
-    def @dataset.set_model(blah); end
-    def @dataset.naked; self; end
-    def @dataset.columns; [:x, :y]; end
-    def @dataset.def_mutation_method(*names);  end
-  end
-
-  it "should be created on set_dataset" do
-    %w'x y x= y='.each do |x|
-      @c.instance_methods.include?(x).should == false
-    end
-    @c.set_dataset(@dataset)
-    %w'x y x= y='.each do |x|
-      @c.instance_methods.include?(x).should == true
-    end
-    o = @c.new
-    %w'x y x= y='.each do |x|
-      o.methods.include?(x).should == true
-    end
-
-    o.x.should be_nil
-    o.x = 34
-    o.x.should == 34
-  end
-
-  it "should be only accept one argument for the write accessor" do
-    @c.set_dataset(@dataset)
-    o = @c.new
-
-    o.x = 34
-    o.x.should == 34
-    proc{o.send(:x=)}.should raise_error
-    proc{o.send(:x=, 3, 4)}.should raise_error
-  end
-end
-
-describe Sequel::Model, ".[]" do
-
-  before(:each) do
-    MODEL_DB.reset
-
-    @c = Class.new(Sequel::Model(:items))
-
-    $cache_dataset_row = {:name => 'sharon', :id => 1}
-    @dataset = @c.dataset
-    $sqls = []
-    @dataset.extend(Module.new {
-      def fetch_rows(sql)
-        $sqls << sql
-        yield $cache_dataset_row
-      end
-    })
-  end
-
-  it "should return the first record for the given pk" do
-    @c[1].should be_a_kind_of(@c)
-    $sqls.last.should == "SELECT * FROM items WHERE (id = 1) LIMIT 1"
-    @c[9999].should be_a_kind_of(@c)
-    $sqls.last.should == "SELECT * FROM items WHERE (id = 9999) LIMIT 1"
-  end
-
-  it "should raise for boolean argument (mistaken comparison)" do
-    # This in order to prevent stuff like Model[:a == 'b']
-    proc {@c[:a == 1]}.should raise_error(Sequel::Error)
-    proc {@c[:a != 1]}.should raise_error(Sequel::Error)
-  end
-
-  it "should work correctly for custom primary key" do
-    @c.set_primary_key :name
-    @c['sharon'].should be_a_kind_of(@c)
-    $sqls.last.should == "SELECT * FROM items WHERE (name = 'sharon') LIMIT 1"
-  end
-
-  it "should work correctly for composite primary key" do
-    @c.set_primary_key [:node_id, :kind]
-    @c[3921, 201].should be_a_kind_of(@c)
-    $sqls.last.should =~ \
-    /^SELECT \* FROM items WHERE (\(node_id = 3921\) AND \(kind = 201\))|(\(kind = 201\) AND \(node_id = 3921\)) LIMIT 1$/
-  end
-end
-
-context "Model#inspect" do
-  setup do
-    @o = Sequel::Model.load(:x => 333)
-  end
-  
-  specify "should include the class name and the values" do
-    @o.inspect.should == '#<Sequel::Model @values={:x=>333}>'
-  end
-end
diff -pruN 1.5.1-1/spec/plugins_spec.rb 3.0.0-1/spec/plugins_spec.rb
--- 1.5.1-1/spec/plugins_spec.rb	2008-04-16 19:35:51.000000000 +0100
+++ 3.0.0-1/spec/plugins_spec.rb	1970-01-01 01:00:00.000000000 +0100
@@ -1,74 +0,0 @@
-require File.join(File.dirname(__FILE__), "spec_helper")
-
-module Sequel::Plugins
-
-  module Timestamped
-    def self.apply(m, opts)
-      m.class_def(:get_stamp) {@values[:stamp]}
-      m.meta_def(:stamp_opts) {opts}
-      m.before_save {@values[:stamp] = Time.now}
-    end
-    
-    module InstanceMethods
-      def abc; timestamped_opts; end
-    end
-    
-    module ClassMethods
-      def deff; timestamped_opts; end
-    end
-
-    module DatasetMethods
-      def ghi; timestamped_opts; end
-    end
-  end
-
-end
-
-describe Sequel::Model, "using a plugin" do
-
-  it "should fail if the plugin is not found" do
-    proc do
-      c = Class.new(Sequel::Model) do
-        is :something_or_other
-      end
-    end.should raise_error(LoadError)
-  end
-  
-  it "should apply the plugin to the class" do
-    c = nil
-    proc do
-      c = Class.new(Sequel::Model) do
-        set_dataset MODEL_DB[:items]
-        is :timestamped, :a => 1, :b => 2
-      end
-    end.should_not raise_error(LoadError)
-    
-    c.should respond_to(:stamp_opts)
-    c.stamp_opts.should == {:a => 1, :b => 2}
-    
-    # instance methods
-    m = c.new
-    m.should respond_to(:get_stamp)
-    m.should respond_to(:abc)
-    m.abc.should == {:a => 1, :b => 2}
-    t = Time.now
-    m[:stamp] = t
-    m.get_stamp.should == t
-    
-    # class methods
-    c.should respond_to(:deff)
-    c.deff.should == {:a => 1, :b => 2}
-    
-    # dataset methods
-    c.dataset.should respond_to(:ghi)
-    c.dataset.ghi.should == {:a => 1, :b => 2}
-  end
-  
-  it "should fail to apply if the plugin has DatasetMethod and the model has no datset" do
-    proc do
-      Class.new(Sequel::Model) do
-        is :timestamped, :a => 1, :b => 2
-      end
-    end.should raise_error(Sequel::Error)
-  end
-end
diff -pruN 1.5.1-1/spec/rcov.opts 3.0.0-1/spec/rcov.opts
--- 1.5.1-1/spec/rcov.opts	2008-04-16 19:35:51.000000000 +0100
+++ 3.0.0-1/spec/rcov.opts	2008-10-19 02:59:23.000000000 +0100
@@ -1,4 +1,6 @@
 --exclude
 gems
 --exclude
-spec
\ No newline at end of file
+spec
+--exclude
+00*
diff -pruN 1.5.1-1/spec/record_spec.rb 3.0.0-1/spec/record_spec.rb
--- 1.5.1-1/spec/record_spec.rb	2008-04-28 23:26:16.000000000 +0100
+++ 3.0.0-1/spec/record_spec.rb	1970-01-01 01:00:00.000000000 +0100
@@ -1,751 +0,0 @@
-require File.join(File.dirname(__FILE__), "spec_helper")
-
-describe "Model#save" do
-  
-  before(:each) do
-    MODEL_DB.reset
-
-    @c = Class.new(Sequel::Model(:items)) do
-      columns :id, :x, :y
-    end
-  end
-  
-  it "should insert a record for a new model instance" do
-    o = @c.new(:x => 1)
-    o.save
-    
-    MODEL_DB.sqls.first.should == "INSERT INTO items (x) VALUES (1)"
-  end
-
-  it "should update a record for an existing model instance" do
-    o = @c.load(:id => 3, :x => 1)
-    o.save
-    
-    MODEL_DB.sqls.first.should =~ 
-      /UPDATE items SET (id = 3, x = 1|x = 1, id = 3) WHERE \(id = 3\)/
-  end
-  
-  it "should update only the given columns if given" do
-    o = @c.load(:id => 3, :x => 1, :y => nil)
-    o.save(:y)
-    
-    MODEL_DB.sqls.first.should == "UPDATE items SET y = NULL WHERE (id = 3)"
-  end
-  
-  it "should mark saved columns as not changed" do
-    o = @c.new(:id => 3, :x => 1, :y => nil)
-    o[:y] = 4
-    o.changed_columns.should == [:y]
-    o.save(:x)
-    o.changed_columns.should == [:y]
-    o.save(:y)
-    o.changed_columns.should == []
-  end
-  
-end
-
-describe "Model#save_changes" do
-  
-  before(:each) do
-    MODEL_DB.reset
-
-    @c = Class.new(Sequel::Model(:items)) do
-      columns :id, :x, :y
-    end
-  end
-  
-  it "should do nothing if no changed columns" do
-    o = @c.new(:id => 3, :x => 1, :y => nil)
-    o.save_changes
-    
-    MODEL_DB.sqls.should be_empty
-
-    o = @c.load(:id => 3, :x => 1, :y => nil)
-    o.save_changes
-    
-    MODEL_DB.sqls.should be_empty
-  end
-  
-  it "should update only changed columns" do
-    o = @c.load(:id => 3, :x => 1, :y => nil)
-    o.x = 2
-
-    o.save_changes
-    MODEL_DB.sqls.should == ["UPDATE items SET x = 2 WHERE (id = 3)"]
-    o.save_changes
-    o.save_changes
-    MODEL_DB.sqls.should == ["UPDATE items SET x = 2 WHERE (id = 3)"]
-    MODEL_DB.reset
-
-    o.y = 4
-    o.save_changes
-    MODEL_DB.sqls.should == ["UPDATE items SET y = 4 WHERE (id = 3)"]
-    o.save_changes
-    o.save_changes
-    MODEL_DB.sqls.should == ["UPDATE items SET y = 4 WHERE (id = 3)"]
-  end
-  
-  it "should not consider columns changed if the values did not change" do
-    o = @c.load(:id => 3, :x => 1, :y => nil)
-    o.x = 1
-
-    o.save_changes
-    MODEL_DB.sqls.should == []
-    o.x = 3
-    o.save_changes
-    MODEL_DB.sqls.should == ["UPDATE items SET x = 3 WHERE (id = 3)"]
-    MODEL_DB.reset
-
-    o[:y] = nil
-    o.save_changes
-    MODEL_DB.sqls.should == []
-    o[:y] = 4
-    o.save_changes
-    MODEL_DB.sqls.should == ["UPDATE items SET y = 4 WHERE (id = 3)"]
-  end
-end
-
-describe "Model#update_values" do
-  before(:each) do
-    MODEL_DB.reset
-
-    @c = Class.new(Sequel::Model(:items)) do
-      columns :id, :x, :y
-    end
-  end
-  
-  it "should generate an update statement" do
-    o = @c.new(:id => 1)
-    o.update_values(:x => 1)
-    MODEL_DB.sqls.first.should == "UPDATE items SET x = 1 WHERE (id = 1)"
-  end
-  
-  it "should update attribute values" do
-    o = @c.new(:id => 1)
-    o.x.should be_nil
-    o.update_values(:x => 1)
-    o.x.should == 1
-  end
-  
-  it "should support string keys" do
-    o = @c.new(:id => 1)
-    o.x.should be_nil
-    o.update_values('x' => 1)
-    o.x.should == 1
-    MODEL_DB.sqls.first.should == "UPDATE items SET x = 1 WHERE (id = 1)"
-  end
-end
-
-describe "Model#set_values" do
-  before(:each) do
-    MODEL_DB.reset
-
-    @c = Class.new(Sequel::Model(:items)) do
-      columns :id, :x, :y
-    end
-  end
-  
-  it "should not touch the database" do
-    o = @c.new(:id => 1)
-    o.set_values(:x => 1)
-    MODEL_DB.sqls.should == []
-  end
-  
-  it "should update attribute values" do
-    o = @c.new(:id => 1)
-    o.x.should be_nil
-    o.set_values(:x => 1)
-    o.x.should == 1
-  end
-  
-  it "should support string keys" do
-    o = @c.new(:id => 1)
-    o.x.should be_nil
-    o.set_values('x' => 1)
-    o.x.should == 1
-  end
-end
-
-describe "Model#new?" do
-  
-  before(:each) do
-    MODEL_DB.reset
-
-    @c = Class.new(Sequel::Model(:items)) do
-    end
-  end
-  
-  it "should be true for a new instance" do
-    n = @c.new(:x => 1)
-    n.should be_new
-  end
-  
-  it "should be false after saving" do
-    n = @c.new(:x => 1)
-    n.save
-    n.should_not be_new
-  end
-  
-  ### DEPRECATED
-  it "should alias new_record? to new?" do
-    n = @c.new(:x => 1)
-    n.should respond_to(:new_record?)
-    n.should be_new_record
-    n.save
-    n.should_not be_new_record
-  end
-  
-end
-
-describe Sequel::Model, "w/ primary key" do
-  
-  it "should default to ':id'" do
-    model_a = Class.new Sequel::Model
-    model_a.primary_key.should be_equal(:id)
-  end
-
-  it "should be changed through 'set_primary_key'" do
-    model_a = Class.new(Sequel::Model) { set_primary_key :a }
-    model_a.primary_key.should be_equal(:a)
-  end
-
-  it "should support multi argument composite keys" do
-    model_a = Class.new(Sequel::Model) { set_primary_key :a, :b }
-    model_a.primary_key.should be_eql([:a, :b])
-  end
-
-  it "should accept single argument composite keys" do
-    model_a = Class.new(Sequel::Model) { set_primary_key [:a, :b] }
-    model_a.primary_key.should be_eql([:a, :b])
-  end
-  
-end
-
-describe Sequel::Model, "w/o primary key" do
-  it "should return nil for primary key" do
-    Class.new(Sequel::Model) { no_primary_key }.primary_key.should be_nil
-  end
-
-  it "should raise a Sequel::Error on 'this'" do
-    instance = Class.new(Sequel::Model) { no_primary_key }.new
-    proc { instance.this }.should raise_error(Sequel::Error)
-  end
-end
-
-describe Sequel::Model, "with this" do
-
-  before { @example = Class.new Sequel::Model(:examples); @example.columns :id, :a, :x, :y }
-
-  it "should return a dataset identifying the record" do
-    instance = @example.new :id => 3
-    instance.this.sql.should be_eql("SELECT * FROM examples WHERE (id = 3) LIMIT 1")
-  end
-
-  it "should support arbitary primary keys" do
-    @example.set_primary_key :a
-
-    instance = @example.new :a => 3
-    instance.this.sql.should be_eql("SELECT * FROM examples WHERE (a = 3) LIMIT 1")
-  end
-
-  it "should support composite primary keys" do
-    @example.set_primary_key :x, :y
-    instance = @example.new :x => 4, :y => 5
-
-    parts = [
-      'SELECT * FROM examples WHERE %s LIMIT 1',
-      '(x = 4) AND (y = 5)', 
-      '(y = 5) AND (x = 4)'
-    ].map { |expr| Regexp.escape expr }
-    regexp = Regexp.new parts.first % "(?:#{parts[1]}|#{parts[2]})"
-
-    instance.this.sql.should match(regexp)
-  end
-
-end
-
-describe "Model#pk" do
-  before(:each) do
-    @m = Class.new(Sequel::Model)
-    @m.columns :id, :x, :y
-  end
-  
-  it "should be default return the value of the :id column" do
-    m = @m.new(:id => 111, :x => 2, :y => 3)
-    m.pk.should == 111
-  end
-
-  it "should be return the primary key value for custom primary key" do
-    @m.set_primary_key :x
-    m = @m.new(:id => 111, :x => 2, :y => 3)
-    m.pk.should == 2
-  end
-
-  it "should be return the primary key value for composite primary key" do
-    @m.set_primary_key [:y, :x]
-    m = @m.new(:id => 111, :x => 2, :y => 3)
-    m.pk.should == [3, 2]
-  end
-
-  it "should raise if no primary key" do
-    @m.set_primary_key nil
-    m = @m.new(:id => 111, :x => 2, :y => 3)
-    proc {m.pk}.should raise_error(Sequel::Error)
-
-    @m.no_primary_key
-    m = @m.new(:id => 111, :x => 2, :y => 3)
-    proc {m.pk}.should raise_error(Sequel::Error)
-  end
-end
-
-describe "Model#pk_hash" do
-  before(:each) do
-    @m = Class.new(Sequel::Model)
-    @m.columns :id, :x, :y
-  end
-  
-  it "should be default return the value of the :id column" do
-    m = @m.new(:id => 111, :x => 2, :y => 3)
-    m.pk_hash.should == {:id => 111}
-  end
-
-  it "should be return the primary key value for custom primary key" do
-    @m.set_primary_key :x
-    m = @m.new(:id => 111, :x => 2, :y => 3)
-    m.pk_hash.should == {:x => 2}
-  end
-
-  it "should be return the primary key value for composite primary key" do
-    @m.set_primary_key [:y, :x]
-    m = @m.new(:id => 111, :x => 2, :y => 3)
-    m.pk_hash.should == {:y => 3, :x => 2}
-  end
-
-  it "should raise if no primary key" do
-    @m.set_primary_key nil
-    m = @m.new(:id => 111, :x => 2, :y => 3)
-    proc {m.pk_hash}.should raise_error(Sequel::Error)
-
-    @m.no_primary_key
-    m = @m.new(:id => 111, :x => 2, :y => 3)
-    proc {m.pk_hash}.should raise_error(Sequel::Error)
-  end
-end
-
-describe Sequel::Model, "set_with_params" do
-
-  before(:each) do
-    MODEL_DB.reset
-    
-    @c = Class.new(Sequel::Model(:items)) do
-      columns :x, :y, :id
-    end
-    @c.instance_variable_set(:@columns, true)
-    @o1 = @c.new
-    @o2 = @c.load(:id => 5)
-  end
-  
-  it "should filter the given params using the model columns" do
-    @o1.set_with_params(:x => 1, :z => 2)
-    @o1.values.should == {:x => 1}
-    MODEL_DB.sqls.should == []
-
-    @o2.set_with_params(:y => 1, :abc => 2)
-    @o2.values.should == {:y => 1, :id=> 5}
-    MODEL_DB.sqls.should == []
-  end
-  
-  it "should work with both strings and symbols" do
-    @o1.set_with_params('x'=> 1, 'z'=> 2)
-    @o1.values.should == {:x => 1}
-    MODEL_DB.sqls.should == []
-
-    @o2.set_with_params('y'=> 1, 'abc'=> 2)
-    @o2.values.should == {:y => 1, :id=> 5}
-    MODEL_DB.sqls.should == []
-  end
-  
-  it "should support virtual attributes" do
-    @c.class_def(:blah=) {|v| self.x = v}
-    @o1.set_with_params(:blah => 333)
-    @o1.values.should == {:x => 333}
-    MODEL_DB.sqls.should == []
-    @o1.set_with_params('blah'=> 334)
-    @o1.values.should == {:x => 334}
-    MODEL_DB.sqls.should == []
-  end
-end
-
-describe Sequel::Model, "update_with_params" do
-
-  before(:each) do
-    MODEL_DB.reset
-    
-    @c = Class.new(Sequel::Model(:items)) do
-      columns :x, :y, :id
-    end
-    @c.instance_variable_set(:@columns, true)
-    @o1 = @c.new
-    @o2 = @c.load(:id => 5)
-  end
-  
-  it "should filter the given params using the model columns" do
-    @o1.update_with_params(:x => 1, :z => 2)
-    MODEL_DB.sqls.first.should == "INSERT INTO items (x) VALUES (1)"
-
-    MODEL_DB.reset
-    @o2.update_with_params(:y => 1, :abc => 2)
-    MODEL_DB.sqls.first.should == "UPDATE items SET y = 1 WHERE (id = 5)"
-  end
-  
-  ### DEPRECATE
-  it "should be aliased by update_with" do
-    @o1.update_with(:x => 1, :z => 2)
-    MODEL_DB.sqls.first.should == "INSERT INTO items (x) VALUES (1)"
-
-    MODEL_DB.reset
-    @o2.update_with(:y => 1, :abc => 2)
-    MODEL_DB.sqls.first.should == "UPDATE items SET y = 1 WHERE (id = 5)"
-  end
-  
-  it "should support virtual attributes" do
-    @c.class_def(:blah=) {|v| self.x = v}
-    @o1.update_with_params(:blah => 333)
-    MODEL_DB.sqls.first.should == "INSERT INTO items (x) VALUES (333)"
-  end
-end
-
-### DEPRECATE
-describe Sequel::Model, "create_with_params" do
-
-  before(:each) do
-    MODEL_DB.reset
-    
-    @c = Class.new(Sequel::Model(:items)) do
-      columns :x, :y
-    end
-    @c.instance_variable_set(:@columns, true)
-  end
-  
-  it "should filter the given params using the model columns" do
-    @c.create_with_params(:x => 1, :z => 2)
-    MODEL_DB.sqls.first.should == "INSERT INTO items (x) VALUES (1)"
-
-    MODEL_DB.reset
-    @c.create_with_params(:y => 1, :abc => 2)
-    MODEL_DB.sqls.first.should == "INSERT INTO items (y) VALUES (1)"
-  end
-  
-  it "should be aliased by create_with" do
-    @c.create_with(:x => 1, :z => 2)
-    MODEL_DB.sqls.first.should == "INSERT INTO items (x) VALUES (1)"
-
-    MODEL_DB.reset
-    @c.create_with(:y => 1, :abc => 2)
-    MODEL_DB.sqls.first.should == "INSERT INTO items (y) VALUES (1)"
-  end
-  
-  it "should support virtual attributes" do
-    @c.class_def(:blah=) {|v| self.x = v}
-    o = @c.create_with(:blah => 333)
-    MODEL_DB.sqls.first.should == "INSERT INTO items (x) VALUES (333)"
-  end
-end
-
-describe Sequel::Model, "#destroy" do
-
-  before(:each) do
-    MODEL_DB.reset
-    @model = Class.new(Sequel::Model(:items))
-    @model.columns :id
-    @model.dataset.meta_def(:delete) {MODEL_DB.execute delete_sql}
-    
-    @instance = @model.new(:id => 1234)
-    #@model.stub!(:delete).and_return(:true)
-  end
-
-  it "should return self" do
-    @model.db.should_receive(:transaction)
-    @model.after_destroy{3}
-    @instance.destroy.should == @instance
-  end
-
-  it "should run within a transaction" do
-    @model.db.should_receive(:transaction)
-    @instance.destroy
-  end
-
-  it "should run before_destroy and after_destroy hooks" do
-    @model.before_destroy {MODEL_DB.execute('before blah')}
-    @model.after_destroy {MODEL_DB.execute('after blah')}
-    @instance.destroy
-    
-    MODEL_DB.sqls.should == [
-      "before blah",
-      "DELETE FROM items WHERE (id = 1234)",
-      "after blah"
-    ]
-  end
-end
-
-describe Sequel::Model, "#exists?" do
-  before(:each) do
-    @model = Class.new(Sequel::Model(:items))
-    @m = @model.new
-  end
-
-  it "should returns true when #this.count > 0" do
-    @m.this.meta_def(:count) {1}
-    @m.exists?.should be_true
-  end
-
-  it "should return false when #this.count == 0" do
-    @m.this.meta_def(:count) {0}
-    @m.exists?.should be_false
-  end
-end
-
-describe Sequel::Model, "#each" do
-  setup do
-    @model = Class.new(Sequel::Model(:items))
-    @model.columns :a, :b, :id
-    @m = @model.new(:a => 1, :b => 2, :id => 4444)
-  end
-  
-  specify "should iterate over the values" do
-    h = {}
-    @m.each {|k, v| h[k] = v}
-    h.should == {:a => 1, :b => 2, :id => 4444}
-  end
-end
-
-describe Sequel::Model, "#keys" do
-  setup do
-    @model = Class.new(Sequel::Model(:items))
-    @model.columns :a, :b, :id
-    @m = @model.new(:a => 1, :b => 2, :id => 4444)
-  end
-  
-  specify "should return the value keys" do
-    @m.keys.size.should == 3
-    @m.keys.should include(:a, :b, :id)
-    
-    @m = @model.new()
-    @m.keys.should == []
-  end
-end
-
-describe Sequel::Model, "#==" do
-  specify "should compare instances by values" do
-    z = Class.new(Sequel::Model)
-    z.columns :id, :x
-    a = z.new(:id => 1, :x => 3)
-    b = z.new(:id => 1, :x => 4)
-    c = z.new(:id => 1, :x => 3)
-    
-    a.should_not == b
-    a.should == c
-    b.should_not == c
-  end
-
-  specify "should be aliased to #eql?" do
-    z = Class.new(Sequel::Model)
-    z.columns :id, :x
-    a = z.new(:id => 1, :x => 3)
-    b = z.new(:id => 1, :x => 4)
-    c = z.new(:id => 1, :x => 3)
-    
-    a.eql?(b).should == false
-    a.eql?(c).should == true
-    b.eql?(c).should == false
-  end
-end
-
-describe Sequel::Model, "#===" do
-  specify "should compare instances by class and pk if pk is not nil" do
-    z = Class.new(Sequel::Model)
-    z.columns :id, :x
-    y = Class.new(Sequel::Model)
-    y.columns :id, :x
-    a = z.new(:id => 1, :x => 3)
-    b = z.new(:id => 1, :x => 4)
-    c = z.new(:id => 2, :x => 3)
-    d = y.new(:id => 1, :x => 3)
-    
-    a.should === b
-    a.should_not === c
-    a.should_not === d
-  end
-
-  specify "should always be false if the primary key is nil" do
-    z = Class.new(Sequel::Model)
-    z.columns :id, :x
-    y = Class.new(Sequel::Model)
-    y.columns :id, :x
-    a = z.new(:x => 3)
-    b = z.new(:x => 4)
-    c = z.new(:x => 3)
-    d = y.new(:x => 3)
-    
-    a.should_not === b
-    a.should_not === c
-    a.should_not === d
-  end
-end
-
-describe Sequel::Model, "#hash" do
-  specify "should be the same only for objects with the same class and pk if the pk is not nil" do
-    z = Class.new(Sequel::Model)
-    z.columns :id, :x
-    y = Class.new(Sequel::Model)
-    y.columns :id, :x
-    a = z.new(:id => 1, :x => 3)
-    b = z.new(:id => 1, :x => 4)
-    c = z.new(:id => 2, :x => 3)
-    d = y.new(:id => 1, :x => 3)
-    
-    a.hash.should == b.hash
-    a.hash.should_not == c.hash
-    a.hash.should_not == d.hash
-  end
-
-  specify "should be the same only for objects with the same class and values if the pk is nil" do
-    z = Class.new(Sequel::Model)
-    z.columns :id, :x
-    y = Class.new(Sequel::Model)
-    y.columns :id, :x
-    a = z.new(:x => 3)
-    b = z.new(:x => 4)
-    c = z.new(:x => 3)
-    d = y.new(:x => 3)
-    
-    a.hash.should_not == b.hash
-    a.hash.should == c.hash
-    a.hash.should_not == d.hash
-  end
-end
-
-describe Sequel::Model, "#initialize" do
-  setup do
-    @c = Class.new(Sequel::Model) do
-      columns :id, :x
-    end
-  end
-  
-  specify "should accept values" do
-    m = @c.new(:id => 1, :x => 2)
-    m.values.should == {:id => 1, :x => 2}
-  end
-  
-  specify "should accept no values" do
-    m = @c.new
-    m.values.should == {}
-  end
-  
-  specify "should accept nil values" do
-    m = @c.new(nil)
-    m.values.should == {}
-  end
-  
-  specify "should accept a block to execute" do
-    m = @c.new {|o| o[:id] = 1234}
-    m.id.should == 1234
-  end
-  
-  specify "should accept virtual attributes" do
-    @c.class_def(:blah=) {|x| @blah = x}
-    @c.class_def(:blah) {@blah}
-    
-    m = @c.new(:id => 1, :x => 2, :blah => 3)
-    m.values.should == {:id => 1, :x => 2}
-    m.blah.should == 3
-  end
-  
-  specify "should convert string keys into symbol keys" do
-    m = @c.new('id' => 1, 'x' => 2)
-    m.values.should == {:id => 1, :x => 2}
-  end
-end
-
-describe Sequel::Model, ".create" do
-
-  before(:each) do
-    MODEL_DB.reset
-    @c = Class.new(Sequel::Model(:items)) do
-      columns :x
-    end
-  end
-
-  it "should be able to create rows in the associated table" do
-    o = @c.create(:x => 1)
-    o.class.should == @c
-    MODEL_DB.sqls.should == ['INSERT INTO items (x) VALUES (1)',  "SELECT * FROM items WHERE (id IN ('INSERT INTO items (x) VALUES (1)')) LIMIT 1"]
-  end
-
-  it "should be able to create rows without any values specified" do
-    o = @c.create
-    o.class.should == @c
-    MODEL_DB.sqls.should == ["INSERT INTO items DEFAULT VALUES", "SELECT * FROM items WHERE (id IN ('INSERT INTO items DEFAULT VALUES')) LIMIT 1"]
-  end
-
-  it "should accept a block and run it" do
-    o1, o2, o3 =  nil, nil, nil
-    o = @c.create {|o3| o1 = o3; o2 = :blah; o3.x = 333}
-    o.class.should == @c
-    o1.should === o
-    o3.should === o
-    o2.should == :blah
-    MODEL_DB.sqls.should == ["INSERT INTO items (x) VALUES (333)", "SELECT * FROM items WHERE (id IN ('INSERT INTO items (x) VALUES (333)')) LIMIT 1"]
-  end
-  
-  it "should create a row for a model with custom primary key" do
-    @c.set_primary_key :x
-    o = @c.create(:x => 30)
-    o.class.should == @c
-    MODEL_DB.sqls.should == ["INSERT INTO items (x) VALUES (30)", "SELECT * FROM items WHERE (x = 30) LIMIT 1"]
-  end
-end
-
-describe Sequel::Model, "#refresh" do
-  setup do
-    MODEL_DB.reset
-    @c = Class.new(Sequel::Model(:items)) do
-      columns :x
-    end
-  end
-
-  specify "should reload the instance values from the database" do
-    @m = @c.new(:id => 555)
-    @m[:x] = 'blah'
-    @m.this.should_receive(:first).and_return({:x => 'kaboom', :id => 555})
-    @m.refresh
-    @m[:x].should == 'kaboom'
-  end
-  
-  specify "should raise if the instance is not found" do
-    @m = @c.new(:id => 555)
-    @m.this.should_receive(:first).and_return(nil)
-    proc {@m.refresh}.should raise_error(Sequel::Error)
-  end
-  
-  specify "should be aliased by #reload" do
-    @m = @c.new(:id => 555)
-    @m.this.should_receive(:first).and_return({:x => 'kaboom', :id => 555})
-    @m.reload
-    @m[:x].should == 'kaboom'
-  end
-
-  specify "should remove cached associations" do
-    @c.many_to_one :node, :class=>@c
-    @c.one_to_many :attributes, :class=>@c
-    @c.many_to_many :tags, :class=>@c
-    @m = @c.new(:id => 555)
-    @m.instance_variable_set(:@node, 15)
-    @m.instance_variable_set(:@attributes, [15])
-    @m.instance_variable_set(:@tags, [15])
-    @m.reload
-    @m.instance_variable_get(:@node).should == nil
-    @m.instance_variable_get(:@attributes).should == nil
-    @m.instance_variable_get(:@tags).should == nil
-  end
-end
diff -pruN 1.5.1-1/spec/schema_spec.rb 3.0.0-1/spec/schema_spec.rb
--- 1.5.1-1/spec/schema_spec.rb	2008-04-16 19:35:51.000000000 +0100
+++ 3.0.0-1/spec/schema_spec.rb	1970-01-01 01:00:00.000000000 +0100
@@ -1,69 +0,0 @@
-require File.join(File.dirname(__FILE__), "spec_helper")
-
-describe Sequel::Model, "table_exists?" do
-
-  before(:each) do
-    MODEL_DB.reset
-    @model = Class.new(Sequel::Model(:items))
-  end
-
-  it "should get the table name and question the model's db if table_exists?" do
-    @model.should_receive(:table_name).and_return(:items)
-    @model.db.should_receive(:table_exists?)
-    @model.table_exists?
-  end
-
-end
-
-describe Sequel::Model, "create_table" do
-
-  before(:each) do
-    MODEL_DB.reset
-    @model = Class.new(Sequel::Model) do
-      set_dataset MODEL_DB[:items]
-      set_schema do
-        text :name
-        float :price, :null => false
-      end
-    end
-  end
-
-  it "should get the create table SQL list from the db and execute it line by line" do
-    @model.create_table
-    MODEL_DB.sqls.should == ['CREATE TABLE items (name text, price float NOT NULL)']
-  end
-
-end
-
-describe Sequel::Model, "drop_table" do
-
-  before(:each) do
-    MODEL_DB.reset
-    @model = Class.new(Sequel::Model(:items))
-  end
-
-  it "should get the drop table SQL for the associated table and then execute the SQL." do
-    @model.should_receive(:table_name).and_return(:items)
-    @model.db.should_receive(:drop_table_sql).with(:items)
-    @model.db.should_receive(:execute).and_return(:true)
-    @model.drop_table
-  end
-
-end
-
-describe Sequel::Model, "create_table!" do
-
-  before(:each) do
-    MODEL_DB.reset
-    @model = Class.new(Sequel::Model(:items))
-  end
-  
-  it "should drop table if it exists and then create the table" do
-    @model.should_receive(:table_exists?).and_return(true)
-    @model.should_receive(:drop_table).and_return(true)
-    @model.should_receive(:create_table).and_return(true)
-
-    @model.create_table!
-  end
-
-end
diff -pruN 1.5.1-1/spec/spec_config.rb 3.0.0-1/spec/spec_config.rb
--- 1.5.1-1/spec/spec_config.rb	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/spec_config.rb	2008-10-19 18:05:27.000000000 +0100
@@ -0,0 +1,10 @@
+# database objects for running adapter specs
+# ADO_DB = Sequel.connect(:adapter => 'ado', :driver => "{Microsoft Access Driver (*.mdb)}; DBQ=c:\\Nwind.mdb")
+# INFORMIX_DB = Sequel.connect('informix://localhost/mydb')
+# INTEGRATION_URL = 'sqlite:/'
+# MYSQL_USER = 'root'
+# MYSQL_URL = "mysql://#{MYSQL_USER}@localhost/sandbox"
+# MYSQL_SOCKET_FILE = '/tmp/mysql.sock'
+# ORACLE_DB = Sequel.connect('oracle://hr:hr@localhost/XE')
+POSTGRES_URL = 'postgres:///sequel_test?user=_postgresql'
+# SQLITE_URL = 'sqlite:/'
diff -pruN 1.5.1-1/spec/spec_config.rb.example 3.0.0-1/spec/spec_config.rb.example
--- 1.5.1-1/spec/spec_config.rb.example	1970-01-01 01:00:00.000000000 +0100
+++ 3.0.0-1/spec/spec_config.rb.example	2008-10-19 02:59:23.000000000 +0100
@@ -0,0 +1,10 @@
+# database objects for running adapter specs
+# ADO_DB = Sequel.connect(:adapter => 'ado', :driver => "{Microsoft Access Driver (*.mdb)}; DBQ=c:\\Nwind.mdb")
+# INFORMIX_DB = Sequel.connect('informix://localhost/mydb')
+# INTEGRATION_URL = 'sqlite:/'
+# MYSQL_USER = 'root'
+# MYSQL_URL = "mysql://#{MYSQL_USER}@localhost/sandbox"
+# MYSQL_SOCKET_FILE = '/tmp/mysql.sock'
+# ORACLE_DB = Sequel.connect('oracle://hr:hr@localhost/XE')
+# POSTGRES_URL = 'postgres://postgres:postgres@localhost:5432/reality_spec'
+# SQLITE_URL = 'sqlite:/'
diff -pruN 1.5.1-1/spec/spec_helper.rb 3.0.0-1/spec/spec_helper.rb
--- 1.5.1-1/spec/spec_helper.rb	2008-04-28 22:18:21.000000000 +0100
+++ 3.0.0-1/spec/spec_helper.rb	1970-01-01 01:00:00.000000000 +0100
@@ -1,58 +0,0 @@
-require 'rubygems'
-unless Object.const_defined?('Sequel')
-  $:.unshift(File.join(File.dirname(__FILE__), "../../sequel_core/lib/"))
-  require 'sequel_core'
-end
-unless Sequel.const_defined?('Model')
-  $:.unshift(File.join(File.dirname(__FILE__), "../lib/"))
-  require 'sequel_model'
-end
-
-class MockDataset < Sequel::Dataset
-  def insert(*args)
-    @db.execute insert_sql(*args)
-  end
-  
-  def update(*args)
-    @db.execute update_sql(*args)
-  end
-  
-  def delete(*args)
-    @db.execute delete_sql(*args)
-  end
-  
-  def fetch_rows(sql)
-    @db.execute(sql)
-    yield({:id => 1, :x => 1})
-  end
-end
-
-class MockDatabase < Sequel::Database
-  attr_reader :sqls
-  
-  def execute(sql)
-    @sqls ||= []
-    @sqls << sql
-  end
-
-  def reset
-    @sqls = []
-  end
-
-  def transaction; yield; end
-  
-  def dataset; MockDataset.new(self); end
-end
-
-class << Sequel::Model
-  alias orig_columns columns
-  def columns(*cols)
-    return if cols.empty?
-    define_method(:columns){cols}
-    @dataset.instance_variable_set(:@columns, cols) if @dataset
-    define_method(:str_columns){cols.map{|x|x.to_s.freeze}}
-    def_column_accessor(*cols)
-  end
-end
-
-Sequel::Model.db = MODEL_DB = MockDatabase.new
diff -pruN 1.5.1-1/spec/spec.opts 3.0.0-1/spec/spec.opts
--- 1.5.1-1/spec/spec.opts	2008-04-16 19:35:51.000000000 +0100
+++ 3.0.0-1/spec/spec.opts	2008-10-19 02:59:23.000000000 +0100
@@ -1,5 +0,0 @@
---colour
---backtrace
---format
-specdoc
---diff
\ No newline at end of file
diff -pruN 1.5.1-1/spec/validations_spec.rb 3.0.0-1/spec/validations_spec.rb
--- 1.5.1-1/spec/validations_spec.rb	2008-04-28 23:26:16.000000000 +0100
+++ 3.0.0-1/spec/validations_spec.rb	1970-01-01 01:00:00.000000000 +0100
@@ -1,645 +0,0 @@
-require File.join(File.dirname(__FILE__), "spec_helper")
-
-describe "Validation::Errors" do
-  setup do
-    @errors = Validation::Errors.new
-    class Validation::Errors
-      attr_accessor :errors
-    end
-  end
-  
-  specify "should be clearable using #clear" do
-    @errors.errors = {1 => 2, 3 => 4}
-    @errors.clear
-    @errors.errors.should == {}
-  end
-  
-  specify "should be empty if no errors are added" do
-    @errors.should be_empty
-    @errors[:blah] << "blah"
-    @errors.should_not be_empty
-  end
-  
-  specify "should return errors for a specific attribute using #on or #[]" do
-    @errors[:blah].should == []
-    @errors.on(:blah).should == []
-
-    @errors[:blah] << 'blah'
-    @errors[:blah].should == ['blah']
-    @errors.on(:blah).should == ['blah']
-
-    @errors[:bleu].should == []
-    @errors.on(:bleu).should == []
-  end
-  
-  specify "should accept errors using #[] << or #add" do
-    @errors[:blah] << 'blah'
-    @errors[:blah].should == ['blah']
-    
-    @errors.add :blah, 'zzzz'
-    @errors[:blah].should == ['blah', 'zzzz']
-  end
-  
-  specify "should return full messages using #full_messages" do
-    @errors.full_messages.should == []
-    
-    @errors[:blow] << 'blieuh'
-    @errors[:blow] << 'blich'
-    @errors[:blay] << 'bliu'
-    msgs = @errors.full_messages
-    msgs.size.should == 3
-    msgs.should include('blow blieuh', 'blow blich', 'blay bliu')
-  end
-end
-
-describe Validation do
-  setup do
-    @c = Class.new do
-      include Validation
-      
-      def self.validates_coolness_of(attr)
-        validates_each(attr) {|o, a, v| o.errors[a] << 'is not cool' if v != :cool}
-      end
-    end
-    
-    @d = Class.new do
-      attr_accessor :errors
-      def initialize; @errors = Validation::Errors.new; end
-    end
-  end
-  
-  specify "should respond to validates, validations, has_validations?" do
-    @c.should respond_to(:validations)
-    @c.should respond_to(:has_validations?)
-  end
-  
-  specify "should acccept validation definitions using validates_each" do
-    @c.validates_each(:xx, :yy) {|o, a, v| o.errors[a] << 'too low' if v < 50}
-    
-    @c.validations[:xx].size.should == 1
-    @c.validations[:yy].size.should == 1
-    
-    o = @d.new
-    @c.validations[:xx].first.call(o, :aa, 40)
-    @c.validations[:yy].first.call(o, :bb, 60)
-    
-    o.errors.full_messages.should == ['aa too low']
-  end
-
-  specify "should return true/false for has_validations?" do
-    @c.has_validations?.should == false
-    @c.validates_each(:xx) {1}
-    @c.has_validations?.should == true
-  end
-  
-  specify "should provide a validates method that takes block with validation definitions" do
-    @c.validates do
-      coolness_of :blah
-    end
-    @c.validations[:blah].should_not be_empty
-
-    o = @d.new
-    @c.validations[:blah].first.call(o, :ttt, 40)
-    o.errors.full_messages.should == ['ttt is not cool']
-    o.errors.clear
-    @c.validations[:blah].first.call(o, :ttt, :cool)
-    o.errors.should be_empty
-  end
-end
-
-describe "A Validation instance" do
-  setup do
-    @c = Class.new do
-      attr_accessor :score
-      
-      include Validation
-      
-      validates_each :score do |o, a, v|
-        o.errors[a] << 'too low' if v < 87
-      end
-    end
-    
-    @o = @c.new
-  end
-  
-  specify "should supply a #valid? method that returns true if validations pass" do
-    @o.score = 50
-    @o.should_not be_valid
-    @o.score = 100
-    @o.should be_valid
-  end
-  
-  specify "should provide an errors object" do
-    @o.score = 100
-    @o.should be_valid
-    @o.errors.should be_empty
-    
-    @o.score = 86
-    @o.should_not be_valid
-    @o.errors[:score].should == ['too low']
-    @o.errors[:blah].should be_empty
-  end
-end
-
-describe Validation::Generator do
-  setup do
-    $testit = nil
-    
-    @c = Class.new do
-      include Validation
-      
-      def self.validates_blah
-        $testit = 1324
-      end
-    end
-  end
-  
-  specify "should instance_eval the block, sending everything to its receiver" do
-    Validation::Generator.new(@c) do
-      blah
-    end
-    $testit.should == 1324
-  end
-end
-
-describe "Validations" do
-  setup do
-    @c = Class.new do
-      attr_accessor :value
-      include Validation
-    end
-    @m = @c.new
-  end
-
-  specify "should validate acceptance_of" do
-    @c.validates_acceptance_of :value
-    @m.should be_valid
-    @m.value = '1'
-    @m.should be_valid
-  end
-  
-  specify "should validate acceptance_of with accept" do
-    @c.validates_acceptance_of :value, :accept => 'true'
-    @m.value = '1'
-    @m.should_not be_valid
-    @m.value = 'true'
-    @m.should be_valid
-  end
-  
-  specify "should validate acceptance_of with allow_nil => false" do
-    @c.validates_acceptance_of :value, :allow_nil => false
-    @m.should_not be_valid
-  end
-
-  specify "should validate confirmation_of" do
-    @c.send(:attr_accessor, :value_confirmation)
-    @c.validates_confirmation_of :value
-    
-    @m.value = 'blah'
-    @m.should_not be_valid
-    
-    @m.value_confirmation = 'blah'
-    @m.should be_valid
-  end
-
-  specify "should validate format_of" do
-    @c.validates_format_of :value, :with => /.+_.+/
-    @m.value = 'abc_'
-    @m.should_not be_valid
-    @m.value = 'abc_def'
-    @m.should be_valid
-  end
-  
-  specify "should raise for validate_format_of without regexp" do
-    proc {@c.validates_format_of :value}.should raise_error(ArgumentError)
-    proc {@c.validates_format_of :value, :with => :blah}.should raise_error(ArgumentError)
-  end
-  
-  specify "should validate length_of with maximum" do
-    @c.validates_length_of :value, :maximum => 5
-    @m.should_not be_valid
-    @m.value = '12345'
-    @m.should be_valid
-    @m.value = '123456'
-    @m.should_not be_valid
-  end
-
-  specify "should validate length_of with minimum" do
-    @c.validates_length_of :value, :minimum => 5
-    @m.should_not be_valid
-    @m.value = '12345'
-    @m.should be_valid
-    @m.value = '1234'
-    @m.should_not be_valid
-  end
-
-  specify "should validate length_of with within" do
-    @c.validates_length_of :value, :within => 2..5
-    @m.should_not be_valid
-    @m.value = '12345'
-    @m.should be_valid
-    @m.value = '1'
-    @m.should_not be_valid
-    @m.value = '123456'
-    @m.should_not be_valid
-  end
-
-  specify "should validate length_of with is" do
-    @c.validates_length_of :value, :is => 3
-    @m.should_not be_valid
-    @m.value = '123'
-    @m.should be_valid
-    @m.value = '12'
-    @m.should_not be_valid
-    @m.value = '1234'
-    @m.should_not be_valid
-  end
-  
-  specify "should validate length_of with allow_nil" do
-    @c.validates_length_of :value, :is => 3, :allow_nil => true
-    @m.should be_valid
-  end
-
-  specify "should validate numericality_of" do
-    @c.validates_numericality_of :value
-    @m.value = 'blah'
-    @m.should_not be_valid
-    @m.value = '123'
-    @m.should be_valid
-    @m.value = '123.1231'
-    @m.should be_valid
-  end
-
-  specify "should validate numericality_of with only_integer" do
-    @c.validates_numericality_of :value, :only_integer => true
-    @m.value = 'blah'
-    @m.should_not be_valid
-    @m.value = '123'
-    @m.should be_valid
-    @m.value = '123.1231'
-    @m.should_not be_valid
-  end
-  
-  specify "should validate presence_of" do
-    @c.validates_presence_of :value
-    @m.should_not be_valid
-    @m.value = ''
-    @m.should_not be_valid
-    @m.value = 1234
-    @m.should be_valid
-  end
-end
-
-context "Superclass validations" do
-  setup do
-    @c1 = Class.new do
-      include Validation
-      attr_accessor :value
-      validates_length_of :value, :minimum => 5
-    end
-    
-    @c2 = Class.new(@c1) do
-      validates_format_of :value, :with => /^[a-z]+$/
-    end
-  end
-  
-  specify "should be checked when validating" do
-    o = @c2.new
-    o.value = 'ab'
-    o.valid?.should == false
-    o.errors.full_messages.should == [
-      'value is too short'
-    ]
-
-    o.value = '12'
-    o.valid?.should == false
-    o.errors.full_messages.should == [
-      'value is too short',
-      'value is invalid'
-    ]
-
-    o.value = 'abcde'
-    o.valid?.should be_true
-  end
-  
-  specify "should be skipped if skip_superclass_validations is called" do
-    @c2.skip_superclass_validations
-
-    o = @c2.new
-    o.value = 'ab'
-    o.valid?.should be_true
-
-    o.value = '12'
-    o.valid?.should == false
-    o.errors.full_messages.should == [
-      'value is invalid'
-    ]
-
-    o.value = 'abcde'
-    o.valid?.should be_true
-  end
-end
-
-context ".validates with block" do
-  specify "should support calling .each" do
-    @c = Class.new do
-      attr_accessor :vvv
-      
-      include Validation
-      validates do
-        each :vvv do |o, a, v|
-          o.errors[a] << "is less than zero" if v.to_i < 0
-        end
-      end
-    end
-    
-    o = @c.new
-    o.vvv = 1
-    o.should be_valid
-    o.vvv = -1
-    o.should_not be_valid
-  end
-end
-
-describe Sequel::Model, "Validations" do
-
-  before(:all) do
-    class Person < Sequel::Model
-      columns :id,:name,:first_name,:last_name,:middle_name,:initials,:age, :terms
-    end
-
-    class Smurf < Person
-    end
-    
-    class Cow < Sequel::Model
-      columns :id, :name, :got_milk
-    end
-
-    class User < Sequel::Model
-      columns :id, :username, :password
-    end
-    
-    class Address < Sequel::Model
-      columns :id, :zip_code
-    end
-  end
-  
-  it "should validate the acceptance of a column" do
-    class Cow < Sequel::Model
-      validations.clear
-      validates_acceptance_of :got_milk, :accept => 'blah', :allow_nil => false
-    end
-    
-    @cow = Cow.new
-    @cow.should_not be_valid
-    @cow.errors.full_messages.should == ["got_milk is not accepted"]
-    
-    @cow.got_milk = "blah"
-    @cow.should be_valid
-  end
-  
-  it "should validate the confirmation of a column" do
-    class User < Sequel::Model
-      def password_confirmation
-        "test"
-      end
-      
-      validations.clear
-      validates_confirmation_of :password
-    end
-    
-    @user = User.new
-    @user.should_not be_valid
-    @user.errors.full_messages.should == ["password is not confirmed"]
-    
-    @user.password = "test"
-    @user.should be_valid
-  end
-  
-  it "should validate format of column" do
-    class Person < Sequel::Model
-      validates_format_of :first_name, :with => /^[a-zA-Z]+$/
-    end
-
-    @person = Person.new :first_name => "Lancelot99"
-    @person.valid?.should be_false
-    @person = Person.new :first_name => "Anita"
-    @person.valid?.should be_true
-  end
-  
-  # it "should allow for :with_exactly => /[a-zA-Z]/, which wraps the supplied regex with ^<regex>$" do
-  #   pending("TODO: Add this option to Validatable#validates_format_of")
-  # end
-
-  it "should validate length of column" do
-    class Person < Sequel::Model
-      validations.clear
-      validates_length_of :first_name, :maximum => 30
-      validates_length_of :last_name, :minimum => 30
-      validates_length_of :middle_name, :within => 1..5
-      validates_length_of :initials, :is => 2
-    end
-    
-    @person = Person.new(
-      :first_name => "Anamethatiswaytofreakinglongandwayoverthirtycharacters",
-      :last_name => "Alastnameunderthirtychars",
-      :initials => "LGC",
-      :middle_name => "danger"
-    )
-    
-    @person.should_not be_valid
-    @person.errors.full_messages.size.should == 4
-    @person.errors.full_messages.should include(
-      'first_name is too long',
-      'last_name is too short',
-      'middle_name is the wrong length',
-      'initials is the wrong length'
-    )
-    
-    @person.first_name  = "Lancelot"
-    @person.last_name   = "1234567890123456789012345678901"
-    @person.initials    = "LC"
-    @person.middle_name = "Will"
-    @person.should be_valid
-  end
-  
-  it "should validate numericality of column" do
-    class Person < Sequel::Model
-      validations.clear
-      validates_numericality_of :age
-    end
-    
-    @person = Person.new :age => "Twenty"
-    @person.should_not be_valid
-    @person.errors.full_messages.should == ['age is not a number']
-    
-    @person.age = 20
-    @person.should be_valid
-  end
-  
-  it "should validate the presence of a column" do
-    class Cow < Sequel::Model
-      validations.clear
-      validates_presence_of :name
-    end
-    
-    @cow = Cow.new
-    @cow.should_not be_valid
-    @cow.errors.full_messages.should == ['name is not present']
-    
-    @cow.name = "Betsy"
-    @cow.should be_valid
-  end
- 
-  it "should validate the uniqueness of a column" do
-    class User < Sequel::Model
-      validations.clear
-      validates do
-        uniqueness_of :username
-      end
-    end
-    User.dataset.extend(Module.new {
-      def fetch_rows(sql)
-        @db << sql
-        
-        case sql
-        when /COUNT.*username = '0records'/
-        when /COUNT.*username = '2records'/
-          yield({:v => 2})
-        when /COUNT.*username = '1record'/
-          yield({:v => 1})
-        when /username = '1record'/
-          yield({:id => 3, :username => "1record", :password => "test"})
-        end
-      end
-    })
-    
-    @user = User.new(:username => "2records", :password => "anothertest")
-    @user.should_not be_valid
-    @user.errors.full_messages.should == ['username is already taken']
-
-    @user = User.new(:username => "1record", :password => "anothertest")
-    @user.should_not be_valid
-    @user.errors.full_messages.should == ['username is already taken']
-
-    @user = User.new(:id=>4, :username => "1record", :password => "anothertest")
-    @user.instance_variable_set(:@new, false)
-    @user.should_not be_valid
-    @user.errors.full_messages.should == ['username is already taken']
-
-    @user = User.new(:id=>3, :username => "1record", :password => "anothertest")
-    @user.instance_variable_set(:@new, false)
-    @user.should be_valid
-    @user.errors.full_messages.should == []
-
-    @user = User.new(:username => "0records", :password => "anothertest")
-    @user.should be_valid
-    @user.errors.full_messages.should == []
-  end
-  
-  it "should have a validates block that contains multiple validations" do
-    class Person < Sequel::Model
-      validations.clear
-      validates do
-        format_of :first_name, :with => /^[a-zA-Z]+$/
-        length_of :first_name, :maximum => 30
-      end
-    end
-
-    Person.validations[:first_name].size.should == 2
-    
-    @person = Person.new :first_name => "Lancelot99"
-    @person.valid?.should be_false
-    
-    @person2 = Person.new :first_name => "Wayne"
-    @person2.valid?.should be_true
-  end
-
-  it "should allow 'longhand' validations direcly within the model." do
-    lambda {
-      class Person < Sequel::Model
-        validations.clear
-        validates_length_of :first_name, :maximum => 30
-      end
-    }.should_not raise_error
-    Person.validations.length.should eql(1)
-  end
-
-  it "should define a has_validations? method which returns true if the model has validations, false otherwise" do
-    class Person < Sequel::Model
-      validations.clear
-      validates do
-        format_of :first_name, :with => /\w+/
-        length_of :first_name, :maximum => 30
-      end
-    end
-
-    class Smurf < Person
-      validations.clear
-    end
-
-    Person.should have_validations
-    Smurf.should_not have_validations
-  end
-
-  it "should validate correctly instances initialized with string keys" do
-    class Can < Sequel::Model
-      columns :id, :name
-      
-      validates_length_of :name, :minimum => 4
-    end
-    
-    Can.new('name' => 'ab').should_not be_valid
-    Can.new('name' => 'abcd').should be_valid
-  end
-  
-end
-
-describe "Model#save!" do
-  setup do
-    @c = Class.new(Sequel::Model(:people)) do
-      def columns; [:id]; end
-      
-      validates_each :id do |o, a, v|
-        o.errors[a] << 'blah' unless v == 5
-      end
-    end
-    @m = @c.load(:id => 4)
-    MODEL_DB.reset
-  end
-  
-  specify "should save regardless of validations" do
-    @m.should_not be_valid
-    @m.save!
-    MODEL_DB.sqls.should == ['UPDATE people SET id = 4 WHERE (id = 4)']
-  end
-end
-
-describe "Model#save" do
-  setup do
-    @c = Class.new(Sequel::Model(:people)) do
-      columns :id
-
-      validates_each :id do |o, a, v|
-        o.errors[a] << 'blah' unless v == 5
-      end
-    end
-    @m = @c.load(:id => 4)
-    MODEL_DB.reset
-  end
-
-  specify "should save only if validations pass" do
-    @m.should_not be_valid
-    @m.save
-    MODEL_DB.sqls.should be_empty
-    
-    @m.id = 5
-    @m.should be_valid
-    @m.save.should_not be_false
-    MODEL_DB.sqls.should == ['UPDATE people SET id = 5 WHERE (id = 5)']
-  end
-  
-  specify "should return false if validations fail" do
-    @m.save.should == false
-  end
-end
